#!/bin/bash

# @author: <ziad.khoury.hanna@oceandatalab.com>

# Copyright (C) 2014-2024 OceanDataLab
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

set -o nounset   # stop if you attempt to use an undef variable
set -o errexit   # stop the script each time a command fails
set -o pipefail  # stop the script each time a command being piped into fails

# =============================================================================
# Compatibility Helpers
# =============================================================================
if readlink -f / 2>&1 >/dev/null;
then
  readlink_f() { readlink -f "${1}"; }
else
  # macOS's readlink  doesn't support "-f" but it has perl
  readlink_f() { LANG='' perl -MCwd -le 'print Cwd::abs_path shift' "${1}"; }
fi

# =============================================================================
# Globals
# =============================================================================
readonly PROGNAME="$(basename "${0}")";
readonly PROGDIR="$(readlink_f "$(dirname "${0}")")";

# =============================================================================
# Main
# =============================================================================
main()
{
  local root_dir;
  local docs_dir;

  root_dir="$(dirname "${PROGDIR}")";
  docs_dir="${root_dir}/doc";

  (
    cat <<EOF
<!DOCTYPE html>
<html lang="en" data-color-mode="auto" data-light-theme="light" data-dark-theme="dark_dimmed">
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width">
    <meta name="description" content="Documentation for Syntool's configuration.">
    <meta name="theme-color" content="#1e2327">
    <meta name="color-scheme" content="dark light">
    <title>Documentation for Syntool's configuration</title>
    <link rel="stylesheet" href="ghmd.css">
    <script async>
      function toggleMode() {
        var d = document.documentElement.dataset;
        d.colorMode = (d.colorMode == 'dark' ? 'light' : d.colorMode == 'light' ? 'auto' : 'dark');
      }
      document.documentElement.className = 'js';
    </script>
  </head>
  <body>
    <button title="Light mode" class="cmt light-mode" onclick="toggleMode()"><svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><path d="M12 7a5 5 0 0 1 5 5 5 5 0 0 1-5 5 5 5 0 0 1-5-5 5 5 0 0 1 5-5m0 2a3 3 0 0 0-3 3 3 3 0 0 0 3 3 3 3 0 0 0 3-3 3 3 0 0 0-3-3m0-7l2.39 3.42C13.65 5.15 12.84 5 12 5c-.84 0-1.65.15-2.39.42L12 2M3.34 7l4.16-.35A7.2 7.2 0 0 0 5.94 8.5c-.44.74-.69 1.5-.83 2.29L3.34 7m.02 10l1.76-3.77a7.131 7.131 0 0 0 2.38 4.14L3.36 17M20.65 7l-1.77 3.79a7.023 7.023 0 0 0-2.38-4.15l4.15.36m-.01 10l-4.14.36c.59-.51 1.12-1.14 1.54-1.86.42-.73.69-1.5.83-2.29L20.64 17M12 22l-2.41-3.44c.74.27 1.55.44 2.41.44.82 0 1.63-.17 2.37-.44L12 22z"></path></svg></button>
    <button title="Dark mode" class="cmt dark-mode" onclick="toggleMode()"><svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><path d="M17.75 4.09l-2.53 1.94.91 3.06-2.63-1.81-2.63 1.81.91-3.06-2.53-1.94L12.44 4l1.06-3 1.06 3 3.19.09m3.5 6.91l-1.64 1.25.59 1.98-1.7-1.17-1.7 1.17.59-1.98L15.75 11l2.06-.05L18.5 9l.69 1.95 2.06.05m-2.28 4.95c.83-.08 1.72 1.1 1.19 1.85-.32.45-.66.87-1.08 1.27C15.17 23 8.84 23 4.94 19.07c-3.91-3.9-3.91-10.24 0-14.14.4-.4.82-.76 1.27-1.08.75-.53 1.93.36 1.85 1.19-.27 2.86.69 5.83 2.89 8.02a9.96 9.96 0 0 0 8.02 2.89m-1.64 2.02a12.08 12.08 0 0 1-7.8-3.47c-2.17-2.19-3.33-5-3.49-7.82-2.81 3.14-2.7 7.96.31 10.98 3.02 3.01 7.84 3.12 10.98.31z"></path></svg></button>
    <button title="System preference" class="cmt system-mode" onclick="toggleMode()"><svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><path d="M7.5 2c-1.79 1.15-3 3.18-3 5.5s1.21 4.35 3.03 5.5C4.46 13 2 10.54 2 7.5A5.5 5.5 0 0 1 7.5 2m11.57 1.5l1.43 1.43L4.93 20.5 3.5 19.07 19.07 3.5m-6.18 2.43L11.41 5 9.97 6l.42-1.7L9 3.24l1.75-.12.58-1.65L12 3.1l1.73.03-1.35 1.13.51 1.67m-3.3 3.61l-1.16-.73-1.12.78.34-1.32-1.09-.83 1.36-.09.45-1.29.51 1.27 1.36.03-1.05.87.4 1.31M19 13.5a5.5 5.5 0 0 1-5.5 5.5c-1.22 0-2.35-.4-3.26-1.07l7.69-7.69c.67.91 1.07 2.04 1.07 3.26m-4.4 6.58l2.77-1.15-.24 3.35-2.53-2.2m4.33-2.7l1.15-2.77 2.2 2.54-3.35.23m1.15-4.96l-1.14-2.78 3.34.24-2.2 2.54M9.63 18.93l2.77 1.15-2.53 2.19-.24-3.34z"></path></svg></button>
    <article class="markdown-body">
EOF

    curl -X POST \
    -H 'Content-Type: text/plain' \
    -H 'Accept: application/vnd.github.v3+json' \
    --data-binary "@${docs_dir}/portalConfig.md" \
    'https://api.github.com/markdown/raw' \
    | sed 's/ id="user-content-/ id="/g' \
    | sed 's#<span aria-hidden="true" class="octicon octicon-link"></span>#<svg class="octicon octicon-link" viewBox="0 0 16 16" version="1.1" width="16" height="16" aria-hidden="true"><path fill-rule="evenodd" d="M7.775 3.275a.75.75 0 001.06 1.06l1.25-1.25a2 2 0 112.83 2.83l-2.5 2.5a2 2 0 01-2.83 0 .75.75 0 00-1.06 1.06 3.5 3.5 0 004.95 0l2.5-2.5a3.5 3.5 0 00-4.95-4.95l-1.25 1.25zm-4.69 9.64a2 2 0 010-2.83l2.5-2.5a2 2 0 012.83 0 .75.75 0 001.06-1.06 3.5 3.5 0 00-4.95 0l-2.5 2.5a3.5 3.5 0 004.95 4.95l1.25-1.25a.75.75 0 00-1.06-1.06l-1.25 1.25a2 2 0 01-2.83 0z"></path></svg>#g'

    cat <<EOF
</article></body></html>
EOF
  ) > "${docs_dir}/portalConfig.html";
}

main "$@";
exit 0;
