#!/bin/bash

# @author: <sylvain.herledan@oceandatalab.com>

# Copyright (C) 2014-2024 OceanDataLab
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

set -o nounset   # stop if you attempt to use an undef variable
set -o errexit   # stop the script each time a command fails
set -o pipefail  # stop the script each time a command being piped into fails

# =============================================================================
# Constants
# =============================================================================
readonly LIBS_BUNDLE_URL='https://ftp.odl.bzh/odl/syntool/syntool_libs.tar.gz';

# =============================================================================
# Compatibility Helpers
# =============================================================================
has()
{
  set +o errexit
  hash "$1" >/dev/null 2>/dev/null;
  if [ 0 -eq $? ];
  then
    set -o errexit
    return 0;
  else
    set -o errexit
    return 1;
  fi
}

if readlink -f /non-existant >/dev/null 2>/dev/null;
then
  readlink_f() { readlink -f "${1}"; }
else
  # macOS's readlink doesn't support "-f" but it has perl
  readlink_f() { LANG='' perl -MCwd -le 'print Cwd::abs_path shift' "${1}"; }
fi

if has md5sum;
then
  md5hash() { md5sum "${1}" | cut -d' ' -f1; }
elif has md5;
then
  # macOS doesn't have md5sum but has md5
  md5hash() { md5 -q "${1}"; }
fi

if sed --version >/dev/null 2>/dev/null;
then
  # GNU version
  sedi() { sed -i "$@"; }
else
  # macOS version
  sedi() { sed -i '' "$@"; }
fi

if realpath --canonicalize-missing --no-symlinks --relative-to=/tmp/b /tmp >/dev/null 2>&1;
then
  _relpath() {
    realpath --canonicalize-missing --no-symlinks --relative-to="${1}" "${2}";
  }
else
  # macOS doesn't have realpath but it has perl
  _relpath() {
    LANG='' perl -MPath::Class -le 'print file(shift)->relative(shift)' "${2}" "${1}";
  }
fi

# =============================================================================
# Globals
# =============================================================================
PROGDIR="$(readlink_f "$(dirname "${0}")")";
readonly PROGDIR;

# =============================================================================
# Helpers
# =============================================================================
refresh_app_version()
{
  local output_dir;
  local root_dir;
  local commit;
  local branch;
  local short_commit;
  local tag;

  output_dir="${1}"; shift;
  root_dir="$(dirname "$(dirname "${0}")")";

  if [ ! -d "${root_dir}/.git" ];
  then
    if [ -f "${root_dir}/VERSION" ];
    then
      cp "${root_dir}/VERSION" "${output_dir}/VERSION";
    else
      >&2 echo "Version file \"${root_dir}/VERSION\" missing.";
      >&2 echo 'Please use a proper release of the code.';
      exit 1;
    fi
  else
    cd "${root_dir}";
    commit="$(git rev-parse HEAD)";
    branch="$(git rev-parse --abbrev-ref HEAD)";
    short_commit="$(git rev-parse --short HEAD)";
    tag="$(git tag --contains "${commit}" | head -1)";
    cd - > /dev/null 2>/dev/null;

    if [ -n "${tag}" ];
    then
      echo "${tag} - ${short_commit}" \
        > "${output_dir}/VERSION";
    else
      echo "${branch} - ${short_commit}" \
        > "${output_dir}/VERSION";
    fi
  fi
}

trim()
{
  local string;

  string="${1}"; shift;
  # remove leading whitespace characters
  string="${string#"${string%%[![:space:]]*}"}";
  # remove trailing whitespace characters
  string="${string%"${string##*[![:space:]]}"}";
  printf '%s' "${string}";
}

find_in()
{
  local filename;

  filename="${1}"; shift;
  # >&2 echo 'Looking for "'"${filename}"'"...';
  for path in "$@";
  do
    # >&2 echo -n '  in "'"${path}"'"...';
    if [ -f "${path}/${filename}" ];
    then
      # >&2 echo ' found!';
      echo "${path}/${filename}";
      return 0;
    fi
    # >&2 echo ' nope';
  done

  # >&2 echo '  NOT FOUND!';
  return 1;
}

find_first_in()
{
  local delimiter;
  local files_list;

  delimiter="${1}"; shift;
  files_list="${1}"; shift;

  IFS="${delimiter}" read -ra filenames <<< "${files_list}";
  for filename in "${filenames[@]}";
  do
    if find_in "$(trim "${filename}")" "$@";
    then
      return 0;
    fi
  done

  return 1;
}

missing_file_error()
{
  local filename;

  filename="${1}"; shift;
  >&2 echo "Missing file: ${filename} in $*";
  exit 1
}

escape_regexp()
{
  local char;
  local string;

  if [ $# -gt 1 ];
  then
    char="${1}"; shift;
    string="${1}"; shift;
  elif [ $# -gt 0 ];
  then
    char='\/';
    string="${1}"; shift;
  else
    return
  fi
  echo "${string}" | sed -E 's/[-.*+?^${}()|[\\'"${char}"']|]/\\&/g';
}

relpath()
{
  local rel;

  rel="$(_relpath "${1}" "${2}")";
  case "${rel}" in
    .*) echo "${rel}";;
    *) echo "./${rel}";;
  esac
}

encode_email()
{
  echo "$1" | tr \
    'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz012356789@._+' \
    'AjYBeOW0yP9QSDunhV.zL@tKsIaHixF6+Zrgw7kRcGCEbfNmUXpq8o2J3dMv_T5l1';
}

# =============================================================================
# Resources generation
# =============================================================================
build_bundle_js_module()
{
  local root_dir;
  local output_dir;
  local top_imported_modules;
  local filename;
  local input_filepath;
  local output_filepath;
  local output_dirpath;
  local relative_to_input;
  local relative_to_output;

  root_dir="${1}"; shift;
  output_dir="${1}"; shift;
  top_imported_modules="${1}"; shift;
  filename="${1}"; shift;

  input_filepath="${root_dir}/${filename}";
  output_filepath="${output_dir}/${filename}";

  # Ignore files already bundled
  if [ -f "${output_filepath}" ]; then
    return;
  fi

  output_dirpath="$(dirname "${output_filepath}")";
  mkdir -p "${output_dirpath}";

  # Create esbuild's "--external" arguments from all top_imported_modules
  # except the one we are building now (ie. filename)
  sed -E "\!^$(escape_regexp '!' "${filename}")\$!d" "${top_imported_modules}" \
  | sed -E 's!^(.*)$!--external:"'"${root_dir}"'/\1"!' \
  | xargs npx esbuild@0.21.5 --bundle --format=esm --log-level=warning --log-limit=0 \
    --target=chrome61,firefox60,safari11,edge18 \
    --define:window.DEBUG=false \
    --outfile="${output_filepath}" \
    "${input_filepath}";

  # Fix the imported external modules URLs
  # esbuild uses relative URLs pointing to the source modules not the build ones
  # so we replace them using sed
  # See issue #484 (https://github.com/evanw/esbuild/issues/484)
  while read -r module_filename;
  do
    relative_to_input="$(relpath "${output_dirpath}" "${root_dir}/${module_filename}")";
    relative_to_output="$(relpath "${output_dirpath}" "${output_dir}/${module_filename}")";
    sedi -E "s!\"$(escape_regexp '!' "${relative_to_input}")\"!\"$(escape_regexp '!' "${relative_to_output}")\"!g" \
      "${output_filepath}";
  done < "${top_imported_modules}";
}

build_resources()
{
  local origin_dir;
  local root_dir;
  local output_dir;
  local html_path;
  local png_decode_worker_path;
  local png_decode_worker_md5;
  local version_path;
  local version;
  local build_date;
  local module_scripts;
  local sed_script;
  local importer_path;
  local importer_dirpath;
  local relative_to_importer;
  local imported_hash;

  root_dir="${1}"; shift;
  output_dir="${1}"; shift;
  html_path="${1}"; shift;
  png_decode_worker_path='js/png_decode_worker.js';

  refresh_app_version "${output_dir}";

  version_path="$(mktemp)";
  version="$(cat "${output_dir}/VERSION")";
  build_date="$(date '+%Y-%m-%d')";

  cat > "${version_path}" <<EOF
/**
 * Syntool webapp
 *
 * Copyright (c) 2014-2024 OceanDataLab
 *
 * Version: ${version}
 * Build date: ${build_date}
 */

EOF

  origin_dir="$(pwd)";
  cd "${root_dir}";

  # Combine CSS files
  sed -En 's/^(.*[[:space:]])?BUILD-PROD-BEGIN-COMBINE-CSS (.*)[[:space:]]+-->.*$/\2/p' "${html_path}" \
    | while read -r filename;
      do
        # Ignore files already combined
        if [ -f "${output_dir}/${filename}" ]; then
          continue;
        fi

        mkdir -p "$(dirname "${output_dir}/${filename}")";
        (
          # Remove all lines out side of the BEGIN and END
          echo '1,/BUILD-PROD-BEGIN-COMBINE-CSS '"$(escape_regexp "${filename}")"'[[:space:]]+-->/ d';
          echo '/BUILD-PROD-END-COMBINE-CSS/,$ d';

          # Replace href with data-prod-href
          echo '/[[:space:]]data-prod-href=/ {';
          echo '  s/([[:space:]])href="[^"]*"//';
          echo '  s/([[:space:]])data-prod-href=/\1href=/';
          echo '}';

          # Print the "href"s of all the <link> elements between BEGIN and END
          echo '/<link rel="stylesheet"/ s/^.*[[:space:]]href="([^"]*)".*$/\1/ p';
        ) | sed -Enf /dev/stdin "${html_path}" \
          | xargs cat "${version_path}" \
          > "${output_dir}/${filename}";

        if has zopfli; then
          zopfli "${output_dir}/${filename}";
        fi
      done;

  # Combine JS files
  sed -En 's/^(.*[[:space:]])?BUILD-PROD-BEGIN-COMBINE-JS (.*)[[:space:]]+-->.*$/\2/p' "${html_path}" \
    | while read -r filename;
      do
        # Ignore files already combined
        if [ -f "${output_dir}/${filename}" ]; then
          continue;
        fi

        mkdir -p "$(dirname "${output_dir}/${filename}")";
        (
          # Remove all lines out side of the BEGIN and END
          echo '1,/BUILD-PROD-BEGIN-COMBINE-JS '"$(escape_regexp "${filename}")"'[[:space:]]+-->/ d';
          echo '/BUILD-PROD-END-COMBINE-JS/,$ d';

          # Replace src with data-prod-src
          echo '/[[:space:]]data-prod-src=/ {';
          echo '  s/([[:space:]])src="[^"]*"//';
          echo '  s/([[:space:]])data-prod-src=/\1src=/';
          echo '}';

          # Print the "src"s of all the <script> elements between BEGIN and END
          echo '/<script/ s/^.*[[:space:]]src="([^"]*)".*$/\1/ p';
        ) | sed -Enf /dev/stdin "${html_path}" \
          | xargs cat "${version_path}" \
          > "${output_dir}/${filename}";

        if has zopfli; then
          zopfli "${output_dir}/${filename}";
        fi
      done;

  cd "${origin_dir}" > /dev/null 2>/dev/null;

  # Combine worker's imported scripts into it
  if [ ! -f "${output_dir}/${png_decode_worker_path}" ]; then
    mkdir -p "$(dirname "${output_dir}/${png_decode_worker_path}")";
    (
      cd "$(dirname "${root_dir}/${png_decode_worker_path}")";
      # Print all the imported scripts
      sed -n \
        -e $'/importScripts/ s/^.*importScripts([\'"]\\(.*\\)[\'"]).*$/\\1/ p' \
        "${root_dir}/${png_decode_worker_path}" \
        | xargs cat "${version_path}";

      # Print the contents of the worker without the imports
      sed -e '/importScripts/ d' "${root_dir}/${png_decode_worker_path}";
    ) > "${output_dir}/${png_decode_worker_path}";

    if has zopfli; then
      zopfli "${output_dir}/${png_decode_worker_path}";
    fi
  fi

  png_decode_worker_md5="$(md5hash "${output_dir}/${png_decode_worker_path}")";

  # Bundle JS modules
  module_scripts="$(mktemp)";
  sed_script="$(mktemp)";
  # shellcheck disable=SC2064
  trap "rm -f '${module_scripts}' '${sed_script}'" EXIT;

  # List all imported JS modules from the HTML
  sed -En -e '/<script[[:space:]](.*[[:space:]])?type="module"[[:space:]]?>/ {' \
    -e 's/^.*[[:space:]]src="([^"]*)".*$/\1/ p' \
    -e '}' "${html_path}" \
    > "${module_scripts}";
  # Ignore files with file name starting with '_'
  sedi -E '\!/_[^/]*! d' "${module_scripts}";

  # For each one of module_scripts
  while read -r filename;
  do
    if [ ! -f "${output_dir}/${filename}" ]; then
      # Build the module with it's dependencies into a single file
      build_bundle_js_module "${root_dir}" "${output_dir}" "${module_scripts}" \
        "${filename}";

      # Append a cache busting hash to the PNG decode worker's URL in that file
      sedi -E "s!\"($(escape_regexp '!' "${png_decode_worker_path}"))\"!\"\\1?build=${png_decode_worker_md5}\"!g" \
        "${output_dir}/${filename}";
    fi

    # Save the MD5 hash of that file for later
    md5hash "${output_dir}/${filename}" > "${output_dir}/${filename}.md5";
  done < "${module_scripts}";

  # Add cache busting hashes to imports in each one of module_scripts
  #
  # Note
  # ----
  # This logic has a major flaw in that it changes the contents of the module
  # file, which in turn will cause it to have a different MD5 hash when
  # appending it in the HTML (see "build_portal()"). Which in turn leads to
  # loading two version of the module, rendering the one in the HTML useless.
  # But this only happens if a module has both dependencies and dependents,
  # fortunately we don't have that yet!
  while read -r importer_filename;
  do
    importer_path="${output_dir}/${importer_filename}";
    importer_dirpath="$(dirname "${importer_path}")";
    while read -r imported_filename;
    do
      if [ "${imported_filename}" != "${importer_filename}" ];
      then
        relative_to_importer="$(relpath "${importer_dirpath}" "${output_dir}/${imported_filename}")";
        imported_hash="$(cat "${output_dir}/${imported_filename}.md5")";
        echo "s!\"($(escape_regexp '!' "${relative_to_importer}"))\"!\"\\1?build=${imported_hash}\"!g" \
          >> "${sed_script}";
      fi
    done < "${module_scripts}";
    sedi -Ef "${sed_script}" "${importer_path}";

    # Empty sed_script file
    # shellcheck disable=SC2188
    > "${sed_script}";
  done < "${module_scripts}";

  # Remove MD5 files
  xargs printf "${output_dir}/%s.md5\0" < "${module_scripts}" \
  | xargs -0 rm -f;

  # If modules_scripts is not empty (ie. we built some modules),
  # copy the CSS files linked in the JS modules, to the build dir.
  if [ -s "${module_scripts}" ];
  then
    # Equivalent pseudo-JS for the following commands:
    # ```js
    # module_scripts.map(filename => `${output_dir}/${filename}`)
    # .concatContentsOfFiles()
    # .replace(/"(css\/[^"]+\.css)"/, '\n$1\n')
    # .filterLinesMatching(/^css\/.*\.css$/)
    # .forEach(filename => {
    #   sed_script.push(`append "build=${md5hashOfFile(filename)}" to "${filename}"`);
    #   mkdirp(dirname(`${output_dir}/${filename}`));
    #   cp(`${root_dir}/${filename}`, `${output_dir}/${filename}`);
    #   if (zopfli) {
    #     zopfli(`${output_dir}/${filename}`);
    #   }
    # });
    # ```
    xargs printf "${output_dir}/%s\0" < "${module_scripts}" \
    | xargs -0 cat \
    | sed -E 's!"(css/[^"]+\.css)"!\
\1\
!g' \
    | sed -En '\!^css/.*\.css$! p' \
    | while read -r filename;
      do
        # Ignore files already copied
        if [ -f "${output_dir}/${filename}" ]; then
          continue;
        fi

        echo "s!\"($(escape_regexp '!' "${filename}"))\"!\"\\1?build=$(md5hash "${root_dir}/${filename}")\"!g" \
          >> "${sed_script}";
        mkdir -p "$(dirname "${output_dir}/${filename}")";
        cp "${root_dir}/${filename}" "${output_dir}/${filename}";
        if has zopfli; then
          zopfli "${output_dir}/${filename}";
        fi
      done;

    # Equivalent pseudo-JS for the following commands:
    # ```js
    # module_scripts
    # .forEach(filename => {
    #   applySedScriptInline(sed_script, `${output_dir}/${filename}`);
    #   if (zopfli) {
    #     zopfli(`${output_dir}/${filename}`);
    #   }
    # });
    # ```
    while read -r filename;
    do
      sedi -Ef "${sed_script}" "${output_dir}/${filename}";
      if has zopfli; then
        zopfli "${output_dir}/${filename}";
      fi
    done < "${module_scripts}";
  fi

  trap - EXIT;
  rm -f "${module_scripts}" "${sed_script}";

  # Remove temporary version file
  rm -f "${version_path}";
}

process_build_directives_for_variant()
{
  local root_dir;
  local custom_dir;
  local output_dir;
  local output_path;
  local variant;
  local sed_script;
  local sed_path;
  local inline_listing_path;
  local return_code;

  root_dir="${1}"; shift;
  custom_dir="${1}"; shift;
  output_dir="${1}"; shift;
  output_path="${1}"; shift;
  variant="${1:-}"; shift;

  if [ -n "${variant}" ];
  then
    variant="-${variant}";
  fi

  inline_listing_path="$(mktemp)";
  sed_path="$(mktemp)";

  # shellcheck disable=SC2064
  trap "rm -f '${inline_listing_path}' '${sed_path}'" EXIT;

  # Remove lines
  echo '/([[:space:]]|^)BUILD'"${variant}"'-BEGIN-REMOVE-LINES([[:space:]]|$)/,/([[:space:]]|^)BUILD'"${variant}"'-END-REMOVE-LINES([[:space:]]|$)/d' >> "${sed_path}";

  # Inline files
  sed -En 's/^(.*[[:space:]])?BUILD'"${variant}"'-INLINE (.*)[[:space:]]+-->.*$/\2/p' "${output_path}" > "${inline_listing_path}";
  while read -r filename;
  do
    local filepath;

    filepath="$(find_in "${filename}" "${custom_dir}" "${output_dir}" "${root_dir}")";
    echo '/BUILD'"${variant}"'-INLINE '"$(escape_regexp "${filename}")"'[[:space:]]+-->/ {' >> "${sed_path}";
    if [ -z "${filepath}" ];
    then
      missing_file_error "${filename}" "${custom_dir}" "${output_dir}" "${root_dir}";
    else
      echo "  r ${filepath}" >> "${sed_path}";
    fi
    echo '  d' >> "${sed_path}";
    echo '}' >> "${sed_path}";
  done < "${inline_listing_path}"

  sed -En 's/^(.*[[:space:]])?BUILD'"${variant}"'-OPTIONAL-INLINE (.*)[[:space:]]+-->.*$/\2/p' "${output_path}" > "${inline_listing_path}"
  while read -r files_list;
  do
    local filepath;

    filepath="$(find_first_in '|' "${files_list}" "${custom_dir}" "${output_dir}" "${root_dir}" || /usr/bin/true)";
    echo '/BUILD'"${variant}"'-OPTIONAL-INLINE '"$(escape_regexp "${files_list}")"'[[:space:]]+-->/ {' >> "${sed_path}";
    if [ -n "${filepath}" ];
    then
      echo "  r ${filepath}" >> "${sed_path}";
    fi
    echo '  d' >> "${sed_path}";
    echo '}' >> "${sed_path}";
  done < "${inline_listing_path}"

  sedi -Ef "${sed_path}" "${output_path}";

  # When inlining an HTML file, signal that we should repeat the
  # processing to handle the directives newly inlined
  grep 'r .*\.html$' < "${sed_path}" >/dev/null;
  return_code=$?;

  trap - EXIT;
  rm -f "${inline_listing_path}" "${sed_path}";

  return ${return_code};
}

process_build_directives()
{
  local root_dir;
  local custom_dir;
  local output_dir;
  local output_path;
  local html_variant;
  local debug;
  local stop;
  local variant;

  root_dir="${1}"; shift;
  custom_dir="${1}"; shift;
  output_dir="${1}"; shift;
  output_path="${1}"; shift;
  html_variant="${1}"; shift;
  debug="${1}"; shift;
  stop='false';

  if [ 'true' == "${debug}" ];
  then
    variant='DEV';
  else
    variant='PROD';
  fi

  while [ 'false' == "${stop}" ];
  do
    stop='true';

    process_build_directives_for_variant \
      "${root_dir}" \
      "${custom_dir}" \
      "${output_dir}" \
      "${output_path}" \
      "${variant}" \
      && stop='false';
    process_build_directives_for_variant \
      "${root_dir}" \
      "${custom_dir}" \
      "${output_dir}" \
      "${output_path}" \
      "${html_variant}" \
      && stop='false';
    process_build_directives_for_variant \
      "${root_dir}" \
      "${custom_dir}" \
      "${output_dir}" \
      "${output_path}" \
      && stop='false';
  done

  return 0;
}

process_build_preload()
{
  local root_dir;
  local custom_dir;
  local output_dir;
  local output_path;
  local debug;
  local preloads_path;

  root_dir="${1}"; shift;
  custom_dir="${1}"; shift;
  output_dir="${1}"; shift;
  output_path="${1}"; shift;
  debug="${1}"; shift;

  if [ 'true' == "${debug}" ];
  then
    # Only in production
    return;
  fi
  if ! grep -E '^(.*[[:space:]])?BUILD-PROD-INLINE-PRELOAD[[:space:]]+-->.*$' "${output_path}" >/dev/null;
  then
    # BUILD-PROD-INLINE-PRELOAD directive not found
    return;
  fi

  preloads_path="$(mktemp)";

  # shellcheck disable=SC2064
  trap "rm -f '${preloads_path}'" EXIT;

  (
    cat <<EOS
  # Preload all imported JS modules from the HTML
  /<script[[:space:]](.*[[:space:]])?type="module"[[:space:]]?>/ {
    s/^.*[[:space:]]src="([^"]*)".*$/<link rel="modulepreload" href="\1">/
    p
  }

  # Preload all included JS scripts in the HTML
  /<script[[:space:]](.*[[:space:]])?src="[^"]*"/ {
    s/^.*[[:space:]]src="([^"]*)".*$/<link rel="preload" href="\1" as="script">/
    p
  }
EOS
  ) | sed -Enf /dev/stdin "${output_path}" > "${preloads_path}";

  # Ignore files with file name starting with '_'
  sedi -E '\!/_[^/]*! d' "${preloads_path}";

  # Inline the preloads
  sedi -E -e '/BUILD-PROD-INLINE-PRELOAD/ {' -e "r ${preloads_path}" -e 'd' -e 'q' -e '}' "${output_path}";

  trap - EXIT;
  rm -f "${preloads_path}";
}

remove_build_directives()
{
  local output_path;

  output_path="${1}"; shift;
  (
    echo '/([[:space:]]|^)BUILD(-[^-]+)?-(BEGIN|END)-REMOVE-LINES([[:space:]]|$)/ d';
    echo '/([[:space:]]|^)BUILD(-[^-]+)?-(BEGIN|END)-COMBINE-(CSS|JS).*-->.*$/ d';
    echo '/BUILD(-[^-]+)?(-OPTIONAL)?-INLINE.*-->.*$/ d';
  ) | sedi -Ef /dev/stdin "${output_path}";
}

# =============================================================================
# Portal creation
# =============================================================================

build_portal()
{
  local root_dir;
  local html_path;
  local custom_dir;
  local output_path;
  local portal_title;
  local gmaps_api_key;
  local html_variant;
  local debug;
  local output_dir;
  local version;

  root_dir="${1}"; shift;
  html_path="${1}"; shift;
  custom_dir="${1}"; shift;
  output_path="${1}"; shift;
  portal_title="${1}"; shift;
  gmaps_api_key="${1}"; shift;
  html_variant="${1}"; shift;
  debug="${1}"; shift;
  output_dir="$(dirname "${output_path}")";

  mkdir -p "${output_dir}";
  cp "${html_path}" "${output_path}";

  refresh_app_version "${output_dir}";
  version="$(cat "${output_dir}/VERSION")";

  process_build_directives \
    "${root_dir}" \
    "${custom_dir}" \
    "${output_dir}" \
    "${output_path}" \
    "${html_variant}" \
    "${debug}";

  # Inline the version
  sedi -e "s!{{T__SYNTOOLWEB_VERSION}}!${version}!" "${output_path}";

  # Remove unused includes
  if [ "false" == "${gmaps_api_key}" ]; then
    sedi -e "/{{T__GMAPS_API_KEY}}/d" "${output_path}";
  else
    sedi -e "s/{{T__GMAPS_API_KEY}}/${gmaps_api_key}/g" "${output_path}";
  fi

  # Apply portal title
  sedi "s%{{T__PORTAL_TITLE}}%${portal_title}%g" "${output_path}";

  if [ 'false' == "${debug}" ];
  then
    # Append "?build={{T__BUILD_VERSION}}" to each module's src
    sedi -E -e '/<script[[:space:]](.*[[:space:]])?type="module"[[:space:]]?>/ {' \
      -e 's/([[:space:]]src)="([^"]*)"/\1="\2?build={{T__BUILD_VERSION}}"/' \
      -e '}' "${output_path}";
  fi

  (
    if [ 'false' == "${debug}" ];
    then
      # Replace href with data-prod-href
      echo '/[[:space:]]data-prod-href=/ {';
      echo '  s/([[:space:]])href="[^"]*"//';
      echo '  s/([[:space:]])data-prod-href=/\1href=/';
      echo '}';

      # Replace src with data-prod-src
      echo '/[[:space:]]data-prod-src=/ {';
      echo '  s/([[:space:]])src="[^"]*"//';
      echo '  s/([[:space:]])data-prod-src=/\1src=/';
      echo '}';
    fi

    # Add hashes
    sed -En -e '/\{\{T__BUILD_VERSION\}\}/ {' -e 's!.*="(.+)\?[^"]*\{\{T__BUILD_VERSION\}\}".*!\1!' -e p -e '}' "${output_path}" \
      | while read -r filename;
        do
          local filepath;
          local filemd5;

          filepath="$(
            find_in "${filename}" "${custom_dir}" "${output_dir}" "${root_dir}" ||
            find_in "${filename}.gz" "${custom_dir}" "${output_dir}" "${root_dir}" ||
            missing_file_error "${filename}" "${custom_dir}" "${output_dir}" "${root_dir}"
          )";
          if [ -n "${filepath}" ];
          then
            filemd5="$(md5hash "${filepath}")";
            echo 's!'"${filename}"'\?([^"]*)\{\{T__BUILD_VERSION\}\}!'"${filename}"'?\1'"${filemd5}"'!';
          fi
        done;

    # Encode email addresses
    sed -En -e '/[[:space:]]href="mailto:/ {' -e 's!.*[[:space:]]href="mailto:([^"?]*)(\?[^"]*)?".*!\1!' -e p -e '}' "${output_path}" \
      | while read -r email;
        do
          echo "s!([[:space:]]href=\"mailto:)${email}!\\1$(encode_email "${email}")!g";
        done;
  ) | sedi -Ef /dev/stdin "${output_path}";

  process_build_preload \
    "${root_dir}" \
    "${custom_dir}" \
    "${output_dir}" \
    "${output_path}" \
    "${debug}";

  remove_build_directives "${output_path}";
}

# =============================================================================
# Packaging
# =============================================================================
package()
{
  local src_dir;
  local build_dir;
  local output_dir;

  src_dir="${1}"; shift;
  build_dir="${1}"; shift;
  output_dir="${1}"; shift;

  # Copy static resources
  mkdir -p "${output_dir}";
  if [ "${build_dir}" != "${output_dir}" ];
  then
    mkdir -p "${output_dir}/css";
    mkdir -p "${output_dir}/js";
    cp -r "${build_dir}/css"/* "${output_dir}/css";
    cp -r "${build_dir}/js"/* "${output_dir}/js";
  fi

  if [ "${src_dir}" != "${output_dir}" ];
  then
    cp -r "${src_dir}/lib" "${output_dir}/lib";
    cp -r "${src_dir}/fonts" "${output_dir}/fonts";
    cp -r "${src_dir}/images" "${output_dir}/images";

    find "${output_dir}" -type f -exec chmod 644 {} \;;
    find "${output_dir}" -type d -exec chmod 755 {} \;;
  fi
}

# =============================================================================
# Main program
# =============================================================================
print_usage()
{
  local portal_dir;
  portal_dir="$(dirname "${0}")/portal";
  cat <<EOU
usage: ${0} <command(s)>
commands:
    ${0} (-h | --help)
    ${0} --prepare
    ${0} --package OUTPUT_DIR
    ${0} --portal PORTAL_CFG OUTPUT_PATH TITLE

Use '${0} (-h | --help) <command>' for available options.

Commands can be chained (except for '${0} (-h | --help)'), but their execution order will always be:
  1. --prepare
  2. --package
  3. --portal

examples:
  ${0} --help
      Will show this screen.

  ${0} --help --package
      Will show the usage for the --package command.

  ${0} --prepare --package ./dist
      Will concatinate the JS and CSS files and copy them along with other required rescources to "./dist".

  ${0} --prepare --package ./dist --portal '${portal_dir}' ./dist/index.html 'Syntool portal'
      Will generate a complete portal with title "Syntool portal" in "./dist" using resources found in "${portal_dir}".

  ${0} --package ./dist --portal '${portal_dir}' ./dist/index.html 'Syntool portal' --debug
      Will generate a debug version of a portal with title "Syntool portal" in "./dist" using resources found in "${portal_dir}".
EOU
}
print_prepare_usage()
{
  cat <<EOU
usage: ${0} --prepare [--debug]

Concatenate JS and CSS files into "<webapp_dir>/build".

options:
  --debug    Makes this command a no-op.
EOU
}
print_package_usage()
{
  cat <<EOU
usage: ${0} --package OUTPUT_DIR [--debug]

Copy all the required files to OUTPUT_DIR.

options:
  --debug    Copy source (non-concatenated) JS and CSS files instead of concatenated ones.
EOU
}
print_portal_usage()
{
  cat <<EOU
usage: ${0} --portal PORTAL_CFG OUTPUT_PATH TITLE [--google KEY] [--debug]

Generate portal web page entitled TITLE using resources found in PORTAL_CFG and save the result as OUTPUT_PATH.

options:
  --google KEY  Set the Google API Key (Google maps disabled otherwise).
  --debug       Generates a debug version of the portal.
EOU
}

main()
{
  local root_dir;
  local build_dir;
  local html_dir;
  local html_filename;
  local html_variant;
  local portal_path;
  local output_path;
  local package_path;
  local portal_title;
  local with_resources;
  local with_packaging;
  local with_portal;
  local gmap_key;
  local debug;

  root_dir="$(dirname "${PROGDIR}")";
  build_dir="${root_dir}/build";
  html_dir="${root_dir}/html";
  portal_path='/dev/null';
  output_path='/dev/null';
  package_path='/dev/null';
  portal_title='Syntool example';
  with_resources='false';
  with_packaging='false';
  with_portal='false';
  gmap_key='false';
  debug='false';

  if [ $# == 0 ];
  then
    print_usage;
    exit 0;
  fi

  while [[ $# -gt 0 ]];
  do
    case "${1}" in
    "-h"|"--help")
      shift;
      case "${1:-}" in
      "--prepare")
        print_prepare_usage;
        ;;
      "--portal")
        print_portal_usage;
        ;;
      "--package")
        print_package_usage;
        ;;
      *)
        print_usage;
        ;;
      esac
      exit 0;
      ;;
    "--google")
      shift;
      gmap_key="${1}"; shift;
      ;;
    "--debug")
      debug='true'; shift;
      ;;
    "--prepare")
      with_resources='true'; shift;
      ;;
    "--portal")
      with_portal='true'; shift;

      if [ $# -lt 3 ];
      then
        >&2 echo 'ERROR: not enough parameters';
        print_usage;
        exit 1;
      fi

      _portal_path="${1}"; shift;
      portal_path="$(readlink_f "${_portal_path}")";
      output_path="${1}"; shift
      portal_title="${1}"; shift;
      ;;
    "--package")
      with_packaging='true'; shift;

      if [ $# -lt 1 ];
      then
        >&2 echo 'ERROR: not enough parameters';
        print_usage;
        exit 1;
      fi

      local _package_path;

      _package_path="${1}"; shift;
      package_path="$(readlink_f "${_package_path}")";
      ;;
    *)
      >&2 echo "ERROR: Invalid option ${1}";
      print_usage;
      exit 1;
      ;;
    esac
  done

  # Retrieve libraries bundle if they are not available in the root directory
  if [ ! -d "${root_dir}/lib" ];
  then
    local tmp_path;
    tmp_path="$(mktemp)";
    wget -q "${LIBS_BUNDLE_URL}" -O "${tmp_path}";
    tar xzf "${tmp_path}" -C "${root_dir}";
    rm "${tmp_path}";
  fi

  if [ 'true' == "${with_resources}" ] && [ 'false' == "${debug}" ];
  then
    rm -rf "${build_dir}";
    mkdir -p "${build_dir}";

    for html_path in "${html_dir}/index.html" "${html_dir}"/index_*.html;
    do
      build_resources "${root_dir}" "${build_dir}" "${html_path}";
    done
  fi

  if [ 'true' == "${with_packaging}" ];
  then
    if [ 'true' == "${debug}" ];
    then
      package "${root_dir}" "${root_dir}" "${package_path}";
    else
      package "${root_dir}" "${build_dir}" "${package_path}";
    fi
  fi

  if [ 'true' == "${with_portal}" ];
  then
    shopt -s nullglob
    shopt -u nullglob
    find "${html_dir}" -name 'index.html' -o -name 'index_*.html' \
    | while read -r html_path;
      do
        html_filename="$(basename "${html_path}" .html)";
        html_variant="${html_filename#index}";
        html_variant="${html_variant#_}";
        build_portal \
          "${root_dir}" \
          "${html_path}" \
          "${portal_path}" \
          "${output_path%.html}${html_filename#index}.html" \
          "${portal_title}" \
          "${gmap_key}" \
          "${html_variant:-full}" \
          "${debug}";
      done
  fi
}

main "$@";
exit 0;
