/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/** */

// #region TS Lib overloads

declare function isNaN(number: undefined): true;
declare function isNaN(number: number | undefined): boolean;
declare function isFinite(number: undefined): false;
declare function isFinite(number: number | undefined): number is number;
declare function setTimeout(handler: TimerHandler, timeout?: number, ...arguments: any[]): TimeoutId;
interface Math {
  acos(x: number): Radians;
  asin(x: number): Radians;
  atan(x: number): Radians;
  atan2(y: number, x: number): Radians;

  cos(x: Radians): number;
  sin(x: Radians): number;
  tan(x: Radians): number;

  abs<T extends number | Date>(x: T): T;
  ceil<T extends number | Date>(x: T): T;
  floor<T extends number | Date>(x: T): T;
  round<T extends number | Date>(x: T): T;
  max<T extends number | Date>(...values: T[]): T;
  min<T extends number | Date>(...values: T[]): T;
}
interface ArrayConstructor {
  isArray(arg: Readonly<any>): arg is ReadonlyArray<any>;
}
interface ObjectConstructor {
  keys<T extends string>(o: Record<T, any>): T[];
  entries<K extends string, T>(o: Record<K, T>): [K, T][];
  getOwnPropertyDescriptors<T>(o: T): {[P in keyof T]: TypedPropertyDescriptor<T[P]>} & { [x: string | symbol]: PropertyDescriptor };
}
interface Date {
  valueOf(): Milliseconds;
  getTime(): Milliseconds;
  getMilliseconds(): Milliseconds;
  getUTCMilliseconds(): Milliseconds;
  setTime(time: Milliseconds): Milliseconds;
  setMilliseconds(ms: Milliseconds): Milliseconds;
  setUTCMilliseconds(ms: Milliseconds): Milliseconds;
}
interface DateConstructor {
  new(value: number | string | Date | undefined): Date;
  parse(s: string): Milliseconds;
  UTC(year: number, monthIndex: number, date?: number, hours?: number, minutes?: number, seconds?: number, ms?: number): Milliseconds;
  now(): Milliseconds;
}
interface PromiseConstructor {
  withResolvers<T>(): {
    promise: Promise<T>,
    resolve: (value: T | PromiseLike<T>) => void,
    reject: (reason?: any) => void,
  };
}

// #endregion
// #region Helper types

interface Stringifyable {
  toString(): string;
}

type Primitive = null | undefined | boolean | number | bigint | string | symbol;
type Class<R = any> = abstract new (...args: any) => R;
type Falsy = null | undefined | false | 0 | 0n | '';
type Func<T = any, A extends any[] = any[], R = any> = (this: T, ...args: A) => R;
type TypedArray = (
  | Int8Array
  | Uint8Array
  | Uint8ClampedArray
  | Int16Array
  | Uint16Array
  | Int32Array
  | Uint32Array
  | Float32Array
  | Float64Array
  | BigInt64Array
  | BigUint64Array
);
type NumberArray = (
  | Int8Array
  | Uint8Array
  | Uint8ClampedArray
  | Int16Array
  | Uint16Array
  | Int32Array
  | Uint32Array
  | Float32Array
  | Float64Array
  | number[]
);

type Keys<T> = (
  T extends Map<infer K, unknown> ? K
  : T extends WeakMap<infer K, unknown> ? K
  : T extends Record<any, unknown> ? keyof T
  : never
);
type Values<T> = (
  T extends Map<any, infer V> ? V
  : T extends WeakMap<any, infer V> ? V
  : T extends {[key: keyof any]: infer V} ? V
  : never
);

type ButNot<T, U> = T & {[K in Exclude<keyof U, keyof T>]?: never};
// type NotFunction<T> = Exclude<T, Function>;
type NotFunction<T> = ButNot<T, Function>;

type MethodName<T> = (
  {[P in keyof T]-?: T[P] extends Func ? P : never}[keyof T]
);
type InstanceMethodName<C extends Class> = MethodName<InstanceType<C>>;

type PartialMethods<T> = ThisType<T> & {[P in MethodName<T>]?: T[P]};
type PartialInstanceMethods<C extends Class> = PartialMethods<InstanceType<C>>;

/**
 * Helper to convert an interface (that may not have an indexer) to a type (that
 * has an implicit indexer) as workaround the issue described in TypeScript
 * issue #15300.
 * This workaround was proposed by @matgr1 and @lobsterkatie, see comment
 * https://github.com/microsoft/TypeScript/issues/15300#issuecomment-1320480061
 */
type IndexOptional<T> = Pick<T, keyof T>;

/// https://github.com/evanw/esbuild/releases/tag/v0.18.0
type SameShape<Out, In extends Out> = (
  In & {[Key in Exclude<keyof In, keyof Out>]: never}
);

type PromiseOrValue<T> = T | Promise<T>;
type OneOrArray<T> = T | T[];
type OneOrIterable<T> = T | Iterable<T>;
type OneOrMore<T> = [T, ...T[]];
type OmitFirstArrayElement<T extends Array<any>> = (
  T extends [any, ...(infer P)] ? P : never
);
interface NestedArray<T> extends Array<NestedArray<T> | T> {}
type OneOrNestedArray<T> = T | NestedArray<T>;

// Inspired by StackOverflow answer https://stackoverflow.com/a/52473108/2168921
type IfEquals<X, Y, A = X, B = never> =
  (<T>() => T extends X ? 1 : 2) extends
  (<T>() => T extends Y ? 1 : 2) ? A : B;

type WritableKey<T, P extends keyof T> = (
  IfEquals<{[Q in P]: T[P]}, {-readonly [Q in P]: T[P]}, P>
);
type WritableProperties<T> = {
  [P in keyof T]-?: T[P] extends Func ? never : WritableKey<T, P>
}[keyof T];
type PickWritableProperties<T> = Pick<T, WritableProperties<T>>;

type DeepPartial<T, Depth extends number = 1> = (
  Depth extends -1 ? T : {
    [P in keyof T]?: DeepPartial<T[P], [-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20][Depth]>;
  }
);

// #endregion
// #region Marker types

declare const _typeMarker: unique symbol;
declare const _extraMarkers: unique symbol;
interface IMarkedType<T, Marker = unknown> {
  readonly [_typeMarker]?: T;
  readonly [_extraMarkers]?: Marker;
}

type Marked<MaybeMarked, ExtraMarker> = (
  unknown extends ExtraMarker ? MaybeMarked :
  MaybeMarked extends IMarkedType<infer MarkedType, infer Marker>
  ? (
    Required<Marker> & Required<ExtraMarker> extends never
    ? never // Markers are incompatible
    : MarkedType & IMarkedType<MarkedType, Marker & ExtraMarker>
  )
  : MaybeMarked & IMarkedType<MaybeMarked, ExtraMarker>
);
type OmitMarker<MaybeMarked> = (
  MaybeMarked extends IMarkedType<infer MarkedType> ? MarkedType : MaybeMarked
);
type ExtractMarker<MaybeMarked> = (
  MaybeMarked extends IMarkedType<any, infer Marker> ? Marker : unknown
);
type MarkedLike<MaybeMarked, ToMark> = (
  Marked<ToMark, ExtractMarker<MaybeMarked>>
);

// #endregion
// #region JSON types

type JSONPrimitive = number | string | boolean | null;
type JSONArray = JSONValue[];
interface JSONObject {
  [key: string]: JSONValue;
}
type JSONValue = JSONPrimitive | JSONArray | JSONObject;

// #endregion
// #region Base types

type Named<T, M> = Marked<T, {name?: M}>;
type InUnit<T, U extends string> = Marked<T, {unit?: U}>;

/** The ID returned by `setTimeout()`. */
type TimeoutId = Named<number, 'TimeoutId'>;
/** Unsigned 32-bit integer. */
type Uint32 = Named<number, 'Uint32'>;
/** Milliseconds. */
type Milliseconds = InUnit<number, 'Milliseconds'>;
/** Hours. */
type Hours = InUnit<number, 'Hours'>;
/** Seconds since midnight, January 1, 1970 UTC. */
type Timestamp = InUnit<Uint32, 'Timestamp'>;
/** String containing HTML */
type HTML = InUnit<string, 'HTML'>;
/** Number in pixels */
type Pixels = InUnit<number, 'px'>;
/** Number in meters (m) */
type Meters = InUnit<number, 'm'>;
/** Number in meters squared (m²) */
type MetersSq = InUnit<number, 'm2'>;
/** Number in radians */
type Radians = InUnit<number, 'Radians'>;
/** Number in degrees */
type Degrees = InUnit<number, 'Degrees'>;
/** Number in degrees longitude */
type Lon = Named<Degrees, 'Longitude'>;
/** Number in degrees latitude */
type Lat = Named<Degrees, 'Latitude'>;

// #endregion
