/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {APIServer, URLShortenerCreator} from './api.js';
import {Dataset} from './dataset.js';
import {MapView} from './map_view.js';
import {NotificationCenter} from './notification_center.js';
import {Timeframe} from './timeline_store.js';

declare global {
  interface Config {
    /**
     * **MUST** be `'1.3.x'` where `x` can be any positive integer.
     */
    version: string;
  }
  interface ConfigPropsWithDefault {}
  interface ConfigOnlyProps {}
  interface ConfigRawOnlyProps {}
  type RawConfig = (
    & Omit<Config, keyof ConfigOnlyProps | keyof ConfigPropsWithDefault>
    & Partial<Pick<Config, keyof ConfigPropsWithDefault>>
    & ConfigRawOnlyProps
  );

  /* Services */
  interface Config {
    /**
     * A list of servers to be used to find and load the datasets.
     */
    servers: OneOrMore<ServerConfig>;
    /**
     * If provided, will be called to know how to shorten a given URL. Used, for
     * example, when user clicks on the "Share" button.
     */
    shortenerForURL?: URLShortenerCreator;
  }
}

type MapViewConfig = (
  Omit<MapView.Props, 'maxBounds' | 'restrictedBounds'>
  & {
    /**
     * The maximum extent of the map in the projection `projection`.
     */
    maxExtent: MapView.Props['maxBounds'];
    /**
     * The maximum navigable extent of the map in the projection `projection`.
     * If set to `null`, map navigation will be unrestricted.
     *
     * Default: `maxExtent`
     */
    restrictedExtent: MapView.Props['restrictedBounds'];
  }
);

declare global {
  interface ServerConfig {
    id: string;
    Class: typeof APIServer;
  }
}

declare global {
  /* Map */
  interface Config extends MapViewConfig {
    /**
     * If provided, will be used to set the zoom level of the map by default (if
     * the `extent` or `zoom` are not set in the URL of the page).
     */
    defaultZoom?: number;
  }

  /* Timeline */
  interface Config {
    /**
     * The default state of the coloc active toggle.
     *
     * Default: `true`
     */
    colocActive: boolean;
    /**
     * The default current datetime of the portal. If not given it will be set to
     * the client's local date and time and then a find nearest will be executed.
     * (example for `01/06/2006 12:00` use `new Date(Date.UTC(2006, 5, 1, 12))`).
     *
     * Default: `new Date()` then a find nearest will be executed.
     */
    currentDate: Date;
    /**
     * A list of predefined timespans that the users can select.
     *
     * Default:
     * ```
     * [
     *   ['6-Hour', '6h'],
     *   ['Daily', '1d', true],
     *   ['3-Day', '3d'],
     *   ['Weekly', '1w'],
     *   ['Bi-weekly', '2w'],
     * ]
     * ```
     */
    timeRanges: OneOrMore<TimeRange>;
    /**
     * Timeline years range from January 1st to January 1st (which means the
     * upper bound will not appear in the timeline).
     *
     * Default:
     * ```js
     * [currentYear - 15, currentYear + 1]
     * ```
     */
    yearsRange: [start: number, end: number];
  }
  type TimeRange = [label: string, length: Timeframe, selected?: boolean];

  /* Misc */
  interface Config {
    /**
     * Portals that have the same `storageId` and the same domain will share the
     * same local storage (local hotspots and user shapes).
     */
    storageId: string;
  }

  /* Predefined */
  interface Config {
    /**
     * An ordered list of available products.
     *
     * Default: `[]`
     */
    products: ProductConfig[];
  }

  interface Config extends NotificationCenter.Config {}
  interface ConfigPropsWithDefault {
    servers: true;
    currentDate: true;
    timeRanges: true;
    yearsRange: true;
    zoomOffset: true;
    restrictedExtent: true;
    additionalLayers: true;
    maxDatasetsWeight: true;
  }
  interface ConfigRawOnlyProps {
    /**
     * The URL to the server running a Syntool services server (or a compatible
     * one).
     *
     * @deprecated Use `servers` instead.
     */
    serviceHost?: string;
    /**
     * A list of URLs to use for requesting data (tiles, geoJSON, ...). All URLs
     * **MUST** be equivalent, a list is given in order to circumvent the
     * browser's limitation on the number of simultaneous requests to one
     * domain.
     *
     * @deprecated Use `servers` instead.
     */
    dataServers?: OneOrMore<string>;
    /**
     * If set to `true`, the products will be grouped in the products panel, if
     * possible.
     *
     * @deprecated Use `groupProductsBy` instead.
     */
    showGroups?: boolean;
    /**
     * An ordered list of groups.
     *
     * @deprecated Use product's `tags` and `groupProductsBy` instead.
     */
    groups?: Group[];
    /**
     * An ordered list of available products.
     *
     * Default: `[]`
     */
    products?: ProductRawConfig[];
  }

  /**
   * @deprecated Use product's `tags` and `groupProductsBy` instead.
   */
  interface Group {
    label?: HTML;
    products: string[];
  }

  /**
   * Specifies which granule info groups, resources and metadata fields should
   * be available to the user in the interface, and how they should be
   * presented.
   *
   * @note If a group has no corresponding configuration, it will be ignored and
   * will not be available to the user.
   */
  type ProductGranuleInfoConfig = GranuleInfoConfig | GranuleInfoConfigFn;

  /**
   * Specifies which granule info groups, resources and metadata fields should
   * be available to the user in the interface, and how they should be
   * presented.
   */
  interface GranuleInfoConfig {
    /**
     * The full ID, in the form `${groupId}/${resourceId}`, of the resource that
     * should be activated automatically upon the selection of its dataset.
     */
    autoResourceId?: `${GranuleInfoGroupId}/${GranuleInfoResourceId}`;
    /**
     * If given, specifies the configuration for known groups.
     */
    groups?: GranuleInfoGroupsConfig;
    /**
     * If given, will be call to specify the configuration for a group not in
     * `groups` field.
     */
    unknownGroup?: GranuleInfoUnknownGroupConfig;
  }

  /**
   * This function will be called in order to get the configuration for a given
   * dataset and its info.
   *
   * If this function returns `null` or `undefined`, the entire granule info
   * will be ignored and will not be available to the user.
   *
   * @param dataset The dataset to which the info apply.
   * @param info The info of the dataset that require the configuring.
   *
   * @note This function might be called multiple times, or just once and its
   * return value cached.
   */
  type GranuleInfoConfigFn = (
    (dataset: Dataset, info: GranuleInfo) => GranuleInfoConfig | null | undefined
  );

  interface GranuleInfoGroupsConfig {
    [groupId: GranuleInfoGroupId]: GranuleInfoGroupConfig;
  }

  /**
   * This function will be called in order to get the configuration for a given
   * info group associated to a dataset.
   *
   * If this function returns `null` or `undefined`, the entire info group will
   * be ignored and will not be available to the user.
   *
   * @param dataset The dataset with which the info group is associated.
   * @param group The info group that require the configuring.
   * @param groupId The ID of info group group referencing `groupId` of
   * GranuleInfo.
   *
   * @note This function might be called multiple times, or just once and its
   * return value cached.
   */
  type GranuleInfoUnknownGroupConfig = (
    (dataset: Dataset, group: GranuleInfoGroup, groupId: GranuleInfoGroupId)
    => GranuleInfoGroupConfig | null | undefined
  );

  /**
   * Specifies how the info group along with its resources and metadata fields
   * should be available to the user in the interface, and how they should be
   * presented.
   *
   * @note If no configuration is given for the `metadata` field, all the
   * metadata information included in this group will be ignored and will not be
   * available to the user.
   * @note If a resource has no corresponding configuration, it will be ignored
   * and will not be available to the user.
   */
  interface GranuleInfoGroupConfig {
    /**
     * The title of the group as presented to the user.
     */
    title: HTML;
    /**
     * Used to sort the groups before presenting them.
     */
    order?: number;
    /**
     * When set to true, the presented group will be collapsed by default.
     */
    defaultCollapsed?: boolean;
    /**
     * If given, specifies the configuration for the metadata fields.
     */
    metadata?: GranuleInfoMetadataConfig;
    /**
     * If given, specifies the configuration for known resources.
     */
    resources?: GranuleInfoResourcesConfig;
    /**
     * If given, will be call to specify the configuration for a resource not in
     * `resources` field.
     */
    unknownResource?: GranuleInfoUnknownResourceConfig;
    /**
     * If given, specifies defaults for missing metadata fields.
     */
    metadataDefaults?: GranuleInfoMetadata;
  }

  /**
   * Specifies which metadata fields should be available to the user in the
   * interface, and how they should be presented.
   */
  type GranuleInfoMetadataConfig = (
    GranuleInfoMetadataFieldsConfig | GranuleInfoMetadataConfigFn
  );

  /**
   * @note All metadata fields not specified, will be ignored and will not be
   * available to the user.
   */
  interface GranuleInfoMetadataFieldsConfig {
    [key: string]: HTML;
  }

  /**
   * This function will be called in order to get the configuration for a given
   * metadata field.
   *
   * If this function returns `null` or `undefined`, the metadata field will be
   * ignored and will not be available to the user.
   *
   * @param value The value of the metadata field.
   * @param key The ID of metadata field referencing key of GranuleInfoMetadata.
   * @param metadata All the metadata fields in the group.
   * @param dataset The dataset with which the metadata field is associated.
   *
   * @note This function might be called multiple times, or just once and its
   * return value cached.
   */
  type GranuleInfoMetadataConfigFn = (
    (value: string, key: string, metadata: GranuleInfoMetadata, dataset: Dataset)
    => GranuleInfoMetadataFieldConfig | null | undefined
  );

  /**
   * Specifies how a metadata field should be presented to the user in the
   * interface.
   */
  interface GranuleInfoMetadataFieldConfig {
    /**
     * The label of the metadata field.
     */
    label: HTML;
    /**
     * The description of the metadata field.
     */
    description?: string;
    /**
     * The value of the metadata field.
     */
    value: HTML;
    /**
     * Used to sort the metadata fields and resources of the group before
     * presenting them.
     *
     * Default: `0`
     */
    order?: number;
    /**
     * Whether or not a copy button can be displayed next to this field to allow
     * the user to copy it's `value`.
     *
     * Default: `true`
     */
    copyable?: boolean;
  }

  interface GranuleInfoResourcesConfig {
    [resourceId: GranuleInfoResourceId]: GranuleInfoResourceConfig;
  }

  /**
   * This function will be called in order to get the configuration for a given
   * resource associated to a dataset.
   *
   * If this function returns `null` or `undefined`, the resource will be
   * ignored and will not be available to the user.
   *
   * @param dataset The dataset with which the resource is associated.
   * @param resource The resource that require the configuring.
   * @param resourceId The ID of response `resource` referencing `resourceId` of
   * GranuleInfoResource.
   *
   * @note This function might be called multiple times, or just once and its
   * return value cached.
   */
  type GranuleInfoUnknownResourceConfig = (
    (dataset: Dataset, resource: GranuleInfoResource, resourceId: GranuleInfoResourceId)
    => GranuleInfoResourceConfig | null | undefined
  );

  type GranuleInfoResourceConfig = (
    GranuleInfoResourceBaseConfig
    & (GranuleInfoImageResourceConfig | GranuleInfoOtherResourceConfig)
  );

  interface GranuleInfoResourceBaseConfig {
    /**
     * The label of the resource.
     */
    label: HTML;
    /**
     * Used to sort the metadata fields and resources of the group before
     * presenting them.
     *
     * Default: `-1`
     */
    order?: number;
  }

  interface GranuleInfoImageResourceConfig {
    /**
     * The type of the resource.
     */
    type: 'image';
    /**
     * When `true`, the same image will be shown inline. When a `URL`, the image
     * located at its value will be shown inline. Otherwise the image is not
     * shown inline.
     *
     * Default: `false`
     */
    inline?: string | boolean | URL;
  }

  interface GranuleInfoOtherResourceConfig {
    /**
     * The type of the resource.
     */
    type: Exclude<GranuleInfoResourceType, 'image'>;
  }

  /**
   * The key is the group's label visible to the user in the interface.\
   * The value defines that group.
   *
   * @note Each filter defined in any of the groups inside this object must have
   * a unique pattern and a unique name.
   */
  interface DatasetFilterGroupsRaw {
    [label: string]: DatasetFilterGroupRaw,
  }
  /**
   * The key is the dataset filter's name and it's _not_ allowed to contain
   * percent-signs (`%`), ampersands (`&`), hash-signs (`#`), plus-signs (`+`),
   * commas (`,`), or exclamation-marks (`!`), also it's not allowed to start or
   * end with a whitespace (` `).
   *
   * @note The filters of a group are considered **_exhaustive_** and
   * **_non-overlapping_**. Meaning, when non of the group's filters are
   * selected, non of the product's datasets will be visible. Likewise, when all
   * of the group's filters are selected, the product's datasets will not be
   * filtered by that group.
   */
  interface DatasetFilterGroupRaw {
    /** The description of the group. */
    groupDescription?: string,

    [id: string]: Pattern | DatasetFilterRaw | undefined,
  }
  /**
   * Specifies the pattern to apply when it's active and optionally how it
   * should be presented to the user in the interface.
   */
  interface DatasetFilterRaw {
    /**
     * If given, will be used as the filter's label, instead of its name, when
     * presented to the user in the interface.
     *
     * @note The `label` field should only be used when the restrictions on the
     * name are too stringent as to hinder its clarity it the user. Like for a
     * cloud coverage filter of 10%, as the `%` is not allowed in the name the
     * `label` field can be used.
     */
    label?: string,
    /**
     * If given, will be used as a list of alternative/old names when loading a
     * saved Syntool state.
     *
     * @note The `aliases` field should be used when the name of a filter
     * changes or when a filter is replaced with other ones, to maintain
     * backwards compatibility with older saved states (like bookmarked URLs),
     * the old name can be added to the aliases list of the new filter(s).
     */
    aliases?: string[],
    /**
     * If given, will be used as description of the filter to be shown to the
     * user in the interface.
     */
    description?: string,
    /** The pattern to be applied when this filter is active. */
    pattern: Pattern,
  }

  interface Product {
    /**
     * If provided specifies which granule info groups, resources and metadata
     * fields should be available to the user in the interface, and how they
     * should be presented. Otherwise none will be available to the user in the
     * interface.
     */
    granuleInfoConfig?: ProductGranuleInfoConfig;
  }

  interface ProductConfigPropsWithDefault {}
  interface ProductConfigOnlyProps {}
  interface ProductConfigRawOnlyProps {}
  interface ProductOnlyProps {}
  type ProductRawConfig = (
    & Omit<ProductConfig, keyof ProductConfigOnlyProps | keyof ProductConfigPropsWithDefault>
    & Partial<Pick<ProductConfig, keyof ProductConfigPropsWithDefault>>
    & ProductConfigRawOnlyProps
  );
  interface ProductConfig extends Omit<Product, keyof ProductOnlyProps> {
    /**
     * If provided and `mustBeCurrent` is set to `false`, the product's datasets
     * are considered valid if their respective definition range intersects a
     * range of ± the value of this field in hours around the current date/time.
     */
    timelineValidity: Hours;
  }

  interface ProductOnlyProps {
    textLabel: true;
    searchTags: true;
    serverIds: true;
    timelineValidity: true;
    datasetFilters: true;
  }
  interface ProductConfigPropsWithDefault {
    description: true;
    tags: true;
    selected: true;
    unit: true;
    timelineValidity: true;
    permanent: true;
    excludedFromColoc: true;
    priority: true;
    datasetWeight: true;
    minPageSize: true;
    mapMinZoom: true;
    mapMaxZoom: true;
    opacity: true;
    outlines: true;
  }
  interface ProductConfigRawOnlyProps {
    /**
     * If `true` the outlines (in white around the data) of the product's
     * datasets will be hidden.
     *
     * @deprecated Use `outlines` instead.
     */
    noOutline?: boolean;
    datasetFilters?: DatasetFilterGroupsRaw;
  }
}
