/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {test, assert} from './utils/test_utils.js';
import {Pattern} from './pattern.js';

test('Pattern - parse', () => {
  test('empty', () => {
    assert.deepEqual(Pattern.parse(''), '');
    assert.deepEqual(Pattern.parse('!'), ['!', ['']]);
    assert.deepEqual(Pattern.parse('&'), ['&', ['', '']]);
    assert.deepEqual(Pattern.parse('|'), ['|', ['', '']]);
    assert.deepEqual(Pattern.parse('a&'), ['&', ['a', '']]);
    assert.deepEqual(Pattern.parse('a|'), ['|', ['a', '']]);
    assert.deepEqual(Pattern.parse('&a'), ['&', ['', 'a']]);
    assert.deepEqual(Pattern.parse('|a'), ['|', ['', 'a']]);
  });
  test('ignore whitespace around operators', () => {
    assert.deepEqual(Pattern.parse(' '), Pattern.parse(''));
    assert.deepEqual(Pattern.parse(' ! '), Pattern.parse('!'));
    assert.deepEqual(Pattern.parse(' & '), Pattern.parse('&'));
    assert.deepEqual(Pattern.parse(' | '), Pattern.parse('|'));
    assert.deepEqual(Pattern.parse(' a & '), Pattern.parse('a&'));
    assert.deepEqual(Pattern.parse(' a | '), Pattern.parse('a|'));
    assert.deepEqual(Pattern.parse(' & a '), Pattern.parse('&a'));
    assert.deepEqual(Pattern.parse(' | a '), Pattern.parse('|a'));
    assert.deepEqual(
      Pattern.parse('  ( a & b ) & c | ( ( ( d | e ) & ( f | ! ( g | ! ! h) & ! i & j '),
      Pattern.parse('(a&b)&c|(((d|e)&(f|!(g|!!h)&!i&j')
    );
  });
  test('keep whitespace inside atoms', () => {
    assert.deepEqual(Pattern.parse('a b'), Pattern.parse('a b'));
  });
  test('single', () => {
    assert.deepEqual(Pattern.parse('a'), 'a');
  });
  test('multi', () => {
    assert.deepEqual(Pattern.parse('a | c'), ['|', ['a', 'c']]);
  });
  test('unwrap | with single operand', () => {
    assert.deepEqual(Pattern.parse('a & b'), ['&', ['a', 'b']]);
  });
  test('flatten &, |', () => {
    assert.deepEqual(Pattern.parse('c & (d & e)'), ['&', ['c', 'd', 'e']]);
    assert.deepEqual(Pattern.parse('c | (d | e)'), ['|', ['c', 'd', 'e']]);
  });
  test('complex', () => {
    assert.deepEqual(
      Pattern.parse('(g | !h) & i'),
      ['&', [['|', ['g', ['!', ['h']]]], 'i']]
    );
    assert.deepEqual(
      Pattern.parse('(a & b) & c | (((d | e) & (f | !(g | !!h) & !i & j'),
      ['|', [
        ['&', ['a', 'b', 'c']],
        ['&', [
          ['|', ['d', 'e']],
          ['|', [
            'f',
            ['&', [
              ['!', [['|', ['g', 'h']]]],
              ['!', ['i']],
              'j',
            ]],
          ]],
        ]],
      ]]
    );
  });
  test('fail', () => {
    try {
      Pattern.parse(')', false);
      assert(false);
    } catch (error) {
      assert(error instanceof Pattern.ParseError);
      assert.equal(error.issue, 'Mismatched closing parenthesis');
      assert.equal(error.pattern, ')');
      assert.equal(error.position, 0);
    }
    try {
      Pattern.parse(',', false);
      assert(false);
    } catch (error) {
      assert(error instanceof Pattern.ParseError);
      assert.equal(error.issue, 'Commas "," are not allowed');
      assert.equal(error.pattern, ',');
      assert.equal(error.position, 0);
    }
    try {
      Pattern.parse('\\,', false);
      assert(false);
    } catch (error) {
      assert(error instanceof Pattern.ParseError);
      assert.equal(error.issue, 'Commas "," are not allowed');
      assert.equal(error.pattern, '\\,');
      assert.equal(error.position, 1);
    }
    try {
      Pattern.parse(',', true);
      assert(false);
    } catch (error) {
      assert(error instanceof Pattern.ParseError);
      assert.equal(error.issue, 'Commas "," are not allowed');
      assert.equal(error.pattern, ',');
      assert.equal(error.position, 0);
    }
  });
});
test('Pattern - exec', () => {
  /** @type {Pattern.ExecOperators<string>} */
  const operators = {
    '': (_, a) => a,
    [Pattern.Op.OR]: (_, ...a) => `(${a.join(' or ')})`,
    [Pattern.Op.AND]: (_, ...a) => `(${a.join(' and ')})`,
    [Pattern.Op.NOT]: (_, a) => `not ${a}`,
  };

  assert.equal(
    Pattern.AST.exec(
      Pattern.parse('(a_% & b\\%\\_) & c* | (((d? | e) & (f | !(g | !!h) & !i & j'),
      operators
    ),
    '((a_% and b%_ and c*) or ((d? or e) and (f or (not (g or h) and not i and j))))'
  );
  assert.equal(
    Pattern.AST.exec(
      Pattern.parse('(a?* & b\\*\\?) & c* | (((d? | e) & (f | !(g | !!h) & !i & j'),
      operators
    ),
    '((a?* and b\\*\\? and c*) or ((d? or e) and (f or (not (g or h) and not i and j))))'
  );
});
test('Pattern - stringify', () => {
  const p = 'a*?\\*\\?\\\\*b\\&b\\|c\\(\\)\\!\\\\|x&(d|!e)';
  assert.equal(Pattern.AST.stringify(Pattern.parse(p)), p);
});
test('Pattern - toRegExp', () => {
  assert.equal(
    Pattern.AST.toRegExp(Pattern.parse('a*b?\\*\\?\\\\*b')).toString(),
    '/^a.*b.\\*\\?\\\\.*b$/i'
  );
});
test('Pattern - simplify', () => {
  test('glob', () => {
    assert.equal(Pattern.simplify('**??**?*?*?'), '*');
    assert.equal(Pattern.simplify('???'), '???');
    assert.equal(Pattern.simplify('?a?*?b'), '?a*b');
    assert.equal(Pattern.simplify('?*a?*b?*c?*'), '*a*b*c*');
    assert.equal(Pattern.simplify('!*a*&!*b*'), '!*a*&!*b*');
  });
  test('basic algebra', () => {
    assert.equal(Pattern.simplify('!'), '*');
    assert.equal(Pattern.simplify('!*'), '');
    assert.equal(Pattern.simplify('!!a'), 'a');
    assert.equal(Pattern.simplify('a&'), '');
    assert.equal(Pattern.simplify('a&*'), 'a');
    assert.equal(Pattern.simplify('a|'), 'a');
    assert.equal(Pattern.simplify('a|*'), '*');
  });
});
