/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/// <reference path='../_types.d.ts' />
/** */

// #region Base types

/** Represents a position with x and y coordinates. */
type XY<T = number> = [x: T, y: T];
/** Represents a size with a certain width and height. */
type WH<T = number> = [width: T, height: T];
/** Represents a rectangular area in some coordinates. */
type LBRT<T = number> = [left: T, bottom: T, right: T, top: T];
/** Represents a rectangular area in some coordinates. */
type NSEW<T = number> = {north: T, east: T, south: T, west: T};

/** Represents a value in a projection. */
type Projected<T, P extends string = 'Map'> = Marked<T, {proj?: P}>;

/** Represents a geographical position with a certain latitude and longitude. */
type LonLat = Projected<[lon: Lon, lat: Lat], 'EPSG:4326'>;
/** Represents a rectangular geographical area on a map. */
type LonLatLBRT = Projected<[east: Lon, south: Lat, west: Lon, north: Lat], 'EPSG:4326'>;
/** Represents a rectangular geographical area on a map. */
type LonLatNSEW = Projected<{east: Lon, south: Lat, west: Lon, north: Lat}, 'EPSG:4326'>;

/** Absolute coordinates in the page. */
type Page<T> = Projected<T, 'Page'>;
/** Coordinates in the viewport of the client (i.e. browser) (like `Element.getClientBoundingRect()`). */
type Client<T> = Projected<T, 'Client'>;
/** Coordinates relative to the map's viewport (i.e. DOM element). */
type Viewport<T> = Projected<T, 'Viewport'>;

// #endregion
// #region Core Syntool types

type APIServerId = Named<string, 'APIServerId'>;
type ProductId = Named<string, 'ProductId'>;
type DatasetId = Named<string, 'DatasetId'>;

type GranuleInfoGroupId = Named<string, 'GranuleInfoGroupId'>;
type GranuleInfoResourceId = Named<string, 'GranuleInfoResourceId'>;
interface GranuleInfo {
  [groupId: GranuleInfoGroupId]: GranuleInfoGroup;
}
interface GranuleInfoGroup {
  // @ts-expect-error
  metadata: GranuleInfoMetadata;
  [resourceId: string]: GranuleInfoResource;
}
interface GranuleInfoMetadata {
  [key: string]: HTML;
}
interface GranuleInfoResource {
  /**
   * The URL of the linked resource (either a full URL or a path relative to
   * dataset)
   */
  url: string;
  /**
   * The description of this resource.
   */
  description?: string;

  // Indicate that there can be more key-value pairs
  [key: string]: string | undefined;
}
/**
 * @note Any other type from the three defined above will be ignored and the
 * resource will will not be available to the user.
 */
type GranuleInfoResourceType = 'link' | 'image' | 'spectrum' | string;

/**
 * Represents a complex pattern string used for case-insensitive pattern
 * matching against any string (usually the ID of datasets).
 *
 * #### Syntax
 * ```
 * <Pattern> ::= <operand>
 * <operand> ::= <operand> "|" <operand> # precedence: 0
 *             | <operand> "&" <operand> # precedence: 1
 *             | "!" <operand>           # precedence: 2
 *             | "(" <operand> ")"       # precedence: 3
 *             | <atom>                  # precedence: 4
 * <atom> ::= ""
 *         | <atom-char>
 *         | <atom-char> <atom-char>
 * <atom-char> ::= <regular-char> | <escaped-char> | "*" | "?"
 * <regular-char> ::= <char> except <special-char>
 * <escaped-char> ::= "\" <char>
 * <char> ::= any character except <illegal-char>
 * <illegal-char> ::= "," # comma
 * <special-char> ::= "(" # open group
 *                 | ")" # close group
 *                 | "|" # or operator
 *                 | "&" # and operator
 *                 | "!" # negation operator
 *                 | "\" # escape character
 *                 | "*" # matches any number of characters, even zero characters
 *                 | "?" # matches exactly one character
 * ```
 * Additionally:
 * * leading and trailing whitespace in atoms is ignored, and
 * * trailing closing parentheses (`)`) are optional.
 *
 * #### Examples
 * * `s1a-*` matches any string that starts with "s1a-"
 * * `s1?-*` matches any string that starts with "s1" followed by one character
 *   followed by "-"
 * * `*-s1-*|*-s2-*` matches any string that contains "-s1-" or "-s2-"
 */
type Pattern = Named<string, 'Pattern'>;
/**
 * Represents a legacy pattern string that doesn't support any operators and
 * checked with a case-insensitive contains test.
 *
 * @note Commas (`,`) are not allowed.
 */
type LegacyPattern = Named<string, 'LegacyPattern'>;
/**
 * Represents a string that when starting with an open parenthesis (`(`) will be
 * considered a complex pattern, otherwise will be considered a legacy pattern.
 */
type HybridPattern = Named<string, 'Pattern' | 'LegacyPattern'>;

type DatasetFiltersGroup = {
  /** The group's label when presented to the user in the interface. */
  label: string,
  /**
   * If given, will be used as description of the group to be shown to the user
   * in the interface.
   */
  description?: string,
  /**
   * @note The filters of a group are considered **_exhaustive_** and
   * **_non-overlapping_**. Meaning, when non of the group's filters are
   * selected, non of the product's datasets will be visible. Likewise, when all
   * of the group's filters are selected, the product's datasets will not be
   * filtered by that group.
   */
  filters: DatasetFilter[],
};

/**
 * Specifies the pattern to apply when it's active and optionally how it should
 * be presented to the user in the interface.
 */
type DatasetFilter = {
  /**
   * The dataset filter's name and it's _not_ allowed to contain percent-signs
   * (`%`), ampersands (`&`), hash-signs (`#`), plus-signs (`+`), commas (`,`),
   * or exclamation-marks (`!`), also it's not allowed to start or end with a
   * whitespace (` `).
   */
  id: string,
  /** The filter's label when presented to the user in the interface. */
  label: string,
  /**
   * If given, will be used as a list of alternative/old names when loading a
   * saved Syntool state.
   *
   * @note The `aliases` field should be used when the name of a filter changes
   * or when a filter is replaced with other ones, to maintain backwards
   * compatibility with older saved states (like bookmarked URLs), the old name
   * can be added to the aliases list of the new filter(s).
   */
  aliases?: string[],
  /**
   * If given, will be used as description of the filter to be shown to the user
   * in the interface.
   */
  description?: string,
  /** The pattern to be applied when this filter is active. */
  pattern: Pattern,
};

interface ProductOutlinesEnum {
  ALWAYS_SHOWN: 'ALWAYS_SHOWN';
  HIDDEN_IN_ZOOM_RANGE: 'HIDDEN_IN_ZOOM_RANGE';
  ALWAYS_HIDDEN: 'ALWAYS_HIDDEN';
}
type ProductOutlines = keyof ProductOutlinesEnum;

interface Product {
  /**
   * The ID used to reference the product in the portal.
   */
  id: ProductId;
  /**
   * The label of the product (used in the products side panel among other
   * places).
   */
  label: HTML;
  textLabel: string;
  /**
   * The type of the product's datasets and how they are rendered.
   */
  type: string;
  /**
   * The description of the product (shown as a tooltip in products side panel)
   */
  description: string | null;
  /**
   * The URL of a GitHub Flavored Markdown file containing information about
   * this product.
   */
  infoURL?: string | URL;
  searchTags: string[];
  /**
   * A list of tags associated to the product. Used for grouping and searching
   * in the products list. May be displayed to the user. Each tag may be of the
   * form `${key}:${value}` in which case it can be used for grouping and only
   * the value will be used for searching.
   */
  tags: string[];
  /** If `true` the product is selected by default. */
  selected: boolean;
  /**
   * Will be displayed in the "Display data" dialog.
   */
  unit: string | null;
  /** List of IDs of the API servers that handle the product. */
  serverIds: APIServerId[];
  /**
   * If `true` the product's datasets are considered valid (and thus added to
   * the map) only when the current date/time is inside their respective
   * definition range. Otherwise (if `false`) they are considered valid if their
   * respective definition range intersects the current timespan.
   */
  mustBeCurrent: boolean;
  /**
   * If provided and `mustBeCurrent` is set to `false`, the product's datasets
   * are considered valid if their respective definition range intersects a
   * range of ± the value of this field in hours around the current date/time.
   */
  timelineValidity: Milliseconds;
  /**
   * If `true` the product's datasets are excluded for the coverage and find
   * nearest requests.
   */
  permanent: boolean;
  /**
   * If `true` the product's datasets are excluded when computing colocations.
   */
  excludedFromColoc: boolean;
  /**
   * The priority of the product. Bigger values imply higher priority.
   */
  priority: number;
  /**
   * The average rendering weight of any one of the product's datasets.
   */
  datasetWeight: number;
  /**
   * The minimum number of the product's datasets that will be rendered (if
   * available) regardless of any pagination.
   */
  minPageSize: number;
  /**
   * If provided the product's datasets will be hidden for zoom level smaller
   * (farther away) than the value of this field.
   */
  mapMinZoom: number;
  /**
   * If provided the product's datasets will be hidden for zoom level bigger
   * (nearer) than the value of this field.
   */
  mapMaxZoom: number;
  /**
   * The default opacity of the datasets. `0.0` (fully transparent) to `1.0`
   * (fully opaque).
   */
  opacity: number;
  /**
   * Specifies the default z-order of the product's datasets. When datasets
   * overlap, z-order determines which one covers the other. A dataset with a
   * larger `stackLevel` generally covers an element with a lower one.
   */
  stackLevel: number;
  /**
   * When to show the outlines (in white around the data) of the product's
   * datasets. This takes precedence over the deprecated `noOutline` if both
   * specified.
   */
  outlines: ProductOutlines;
  /**
   * If provided specifies predefined dataset filtering options to be made
   * available to the user in the interface.
   *
   * @note Each filter defined in any of the groups inside this object must have
   * a unique pattern and a unique name.
   */
  datasetFilters: DatasetFiltersGroup[];
}

// #endregion
