/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/// <reference path='./api_.d.ts' />
import {Bounds} from '../gis.js';
import {GranuleInfo, validateValueType} from '../types.js';
import {
  $,
  isInvalidDate,
  getStartOfUTCDay,
  getEndOfUTCDay,
  groupArrayElementsBy,
  flattenArray,
  clamp,
  isAbortError,
  EventsHub,
} from '../utils.js';

/**
 * @template T
 * @template V
 * @param {Promise<T>} promise
 * @param {V} value
 * @returns {Promise<T | V>}
 */
function promiseWithAbortionValue(promise, value) {
  return promise.catch((error) => {
    if (isAbortError(error)) {
      return value;
    }
    throw error;
  });
}

/**
 * @extends {Map<APIServerId, APIServer>}
 */
class ServersMap extends Map {
  /**
   * @param {API} api
   * @param {() => void} onChange
   */
  constructor(api, onChange) {
    super();
    this._api = api;
    this._onChange = onChange;
  }

  /**
   * @param {APIServer} server
   * @returns {boolean}
   */
  add(server) {
    if (this.has(server.id)) {
      return false;
    }
    this.set(server.id, server);
    return true;
  }

  /**
   * @override
   * @param {APIServerId} id
   * @param {APIServer} server
   * @returns {this}
   */
  set(id, server) {
    if (server.id !== id) {
      throw new Error("API server's id must be identical (===) to the key");
    }
    super.set(id, server);
    server.api = this._api;
    this._onChange();
    return this;
  }

  /**
   * @override
   * @param {APIServerId} id
   * @returns {boolean}
   */
  delete(id) {
    const server = this.get(id);
    if (super.delete(id)) {
      if ($(server).api === this._api) {
        $(server).api = null;
      }
      this._onChange();
      return true;
    }
    return false;
  }

  /**
   * @override
   * @returns {void}
   */
  clear() {
    for (const server of this.values()) {
      if (server.api === this._api) {
        server.api = null;
      }
    }
    super.clear();
    this._onChange();
  }
}

export class API {
  // #region Private typedefs

  /** @typedef {import('../dataset.js').Dataset} Dataset */
  /** @typedef {import('../syntool.js').Syntool} Syntool */
  /** @typedef {import('./api_server.js').APIServer} APIServer */

  // #endregion
  // #region Public

  /**
   * @public
   * @param {Syntool} syntool
   */
  constructor(syntool) {
    this.syntool = syntool;
    this.hub = new EventsHub();
    this._servers = new ServersMap(
      this,
      () => this._updateServerIdsForProductWithId()
    );
    /** @type {Set<APIServerId>} */
    this.unreachableServers = new Set();

    this.syntool.hub.on('low-memory', (/** @type {number} */ level) => {
      if (level > 1) {
        this.clearCache();
      }
    });
  }

  /**
   * @public
   * @returns {ServersMap}
   */
  get servers() {
    return this._servers;
  }

  /**
   * @public
   * @returns {void}
   */
  init() {
    this.syntool.productsStore.hub.on(
      'change:products',
      this._updateServerIdsForProductWithId,
      this
    );
    if (this.syntool.productsStore.initialised) {
      this._updateServerIdsForProductWithId();
    } else {
      this.syntool.productsStore.hub.once(
        'init',
        this._updateServerIdsForProductWithId,
        this
      );
    }
  }

  /**
   * @public
   * @param {Product | undefined} product
   * @returns {APIServerId[]}
   */
  getServerIdsForProduct(product) {
    return product?.serverIds || [];
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {APIServerId[]}
   */
  getServerIdsForProductId(productId) {
    return this.getServerIdsForProduct(
      this.syntool.productsStore.getProduct(productId)
    );
  }
  /**
   * @public
   * @param {Dataset} dataset
   * @returns {APIServerId}
   */
  getServerIdForDataset(dataset) {
    return dataset.serverId;
  }
  /**
   * @public
   * @param {APIServerId} serverId
   * @returns {boolean}
   */
  isServerWithIdUnreachable(serverId) {
    return this.unreachableServers.has(serverId);
  }
  /**
   * @public
   * @param {API.ClearCacheOptions} [options]
   * @returns {void}
   */
  clearCache(options) {
    for (const server of this.servers.values()) {
      server.clearCache(options);
    }
  }

  /**
   * @public
   * @param {Dataset} dataset
   * @param {string} url
   * @returns {string}
   */
  getDataURLForDataset(dataset, url) {
    return $(this.servers.get(this.getServerIdForDataset(dataset))).getDataURL(url);
  }
  /**
   * @public
   * @param {APIServerId} serverId
   * @param {Promise<unknown>} promise
   * @returns {Promise<void>}
   */
  async updateServerReachabilityFromPromise(serverId, promise) {
    try {
      await promise;
      if (this.unreachableServers.has(serverId)) {
        this.unreachableServers.delete(serverId);
        this.hub.emit('change:unreachableServers');
      }
    } catch (/** @type {any} */ error) {
      if (!isAbortError(error) && !this.unreachableServers.has(serverId)) {
        this.unreachableServers.add(serverId);
        this.hub.emit('change:unreachableServers');
      }
    }
  }

  /**
   * @public
   * @template T
   * @template R
   * @param {Readonly<T[]>} elements
   * @param {(element: T) => APIServerId[]} getServerIdsForElement
   * @param {(server: APIServer, elements: T[]) => R} cb
   * @returns {R[]}
   */
  groupByServer(elements, getServerIdsForElement, cb) {
    const results = [];
    const elementsByServerId = (
      groupArrayElementsBy(elements, getServerIdsForElement)
    );
    for (const [serverId, serverElements] of elementsByServerId) {
      const server = this.servers.get(serverId);
      if (server) {
        results.push(cb(server, serverElements));
      }
    }
    return results;
  }

  /**
   * @public
   * @param {Date} minDate
   * @param {Date} maxDate
   * @param {ProductId[]} allProductIds
   * @param {AbortSignal} abortSignal
   * @returns {Promise<Array<API.DatasetInfo>>[]}
   */
  getDatasetInfos(minDate, maxDate, allProductIds, abortSignal) {
    const {absoluteMinDate, absoluteMaxDate} = this.syntool.timelineStore;
    minDate = clamp(minDate, absoluteMinDate, absoluteMaxDate);
    maxDate = clamp(maxDate, absoluteMinDate, absoluteMaxDate);

    const extent = this._extent();
    const promisesByServer = this.groupByServer(
      allProductIds,
      (productId) => this.getServerIdsForProductId(productId),
      (server, productIds) => (
        server.getDatasetInfos({
          productIds,
          minDate,
          maxDate,
          extent,
          patternsByProductId: this._patterns(productIds),
          filtersByProductId: this._filters(productIds),
        }, abortSignal)
        .map(async (promise) => {
          const infos = await promiseWithAbortionValue(promise, []);
          for (const info of /** @type {{serverId?: string}[]} */ (infos)) {
            if (!info.serverId) {
              info.serverId = server.id;
            }
          }
          return /** @type {API.DatasetInfo[]} */ (infos);
        })
      )
    );
    return flattenArray(promisesByServer);
  }
  /**
   * @public
   * @param {Dataset} dataset
   * @param {AbortSignal} abortSignal
   * @returns {Promise<GranuleInfo | null>}
   */
  async getDatasetMetadata(dataset, abortSignal) {
    const serverId = this.getServerIdForDataset(dataset);
    const server = this.servers.get(serverId);
    if (!server) {
      return null;
    }
    const granuleInfo = await server.getDatasetMetadata(dataset, abortSignal);
    const error = validateValueType(GranuleInfo, 'granuleInfo', granuleInfo);
    if (error) {
      console.error(
        `Dataset(${dataset.datasetId}) invalid granule info`,
        error,
        granuleInfo
      );
      return null;
    }
    return granuleInfo;
  }
  /**
   * @public
   * @param {ProductId[]} allProductIds
   * @param {AbortSignal} abortSignal
   * @returns {Promise<import('../coverage.js').CoverageMap | null>[]}
   */
  getCoverages(allProductIds, abortSignal) {
    const extent = this._extent();
    const minDate = new Date(this.syntool.timelineStore.minYear, 0, 1);
    const maxDate = new Date(this.syntool.timelineStore.maxYear, 0, 1);
    return this.groupByServer(
      allProductIds.filter((productId) => (
        !this.syntool.filters.product.permanent.withId(productId)
      )),
      (productId) => this.getServerIdsForProductId(productId),
      (server, productIds) => (
        promiseWithAbortionValue(
          server.getCoverage({
            productIds,
            minDate,
            maxDate,
            extent,
            patternsByProductId: this._patterns(productIds),
            filtersByProductId: this._filters(productIds),
          }, abortSignal),
          null
        )
      )
    );
  }

  /**
   * @public
   * @param {ProductId[]} allProductIds
   * @param {Date} date
   * @param {boolean} inPast
   * @param {boolean} inFuture
   * @param {boolean} inPresent
   * @param {AbortSignal} abortSignal
   * @returns {Promise<API.Nearest>}
   */
  async findNearest(allProductIds, date, inPast, inFuture, inPresent, abortSignal) {
    abortSignal.throwIfAborted();

    if (!(inPast || inFuture)) {
      return {nearest: date};
    }

    const coverage = (
      this.syntool.timelineStore.getCoverage()
      .getSubCoverageForProductIds(allProductIds)
    );
    if (coverage.isEmpty) {
      return {nearest: date};
    }

    const {colocActive} = this.syntool;
    const colocProductIds = (
      colocActive
      ? allProductIds.filter(this.syntool.filters.product.includedInColoc.withId)
      : []
    );
    const withColoc = (colocActive && coverage.hasColoc(colocProductIds));
    const extent = this._extent();
    const {absoluteMinDate, absoluteMaxDate} = this.syntool.timelineStore;
    let nearestPast;
    let nearestFuture;

    if (withColoc) {
      const colocDayDate = (
        inPresent && coverage.hasDataAtDate(date, colocActive, colocProductIds)
        ? date
        : coverage.findNearest(getStartOfUTCDay(date), {
          productIds: colocProductIds,

          coloc : true,
          year  : true,
          month : true,
          day   : true,
          past  : inPast,
          future: inFuture,
        })
      );

      if (colocDayDate === null) {
        return {nearest: date};
      }

      const minDate = getStartOfUTCDay(colocDayDate);
      const maxDate = getEndOfUTCDay(colocDayDate);
      const promisesByServer = this.groupByServer(
        colocProductIds,
        (productId) => this.getServerIdsForProductId(productId),
        (server, productIds) => server.getDatasetDates({
          productIds,
          minDate,
          maxDate,
          extent,
          patternsByProductId: this._patterns(productIds),
          filtersByProductId: this._filters(productIds),
        }, abortSignal)
      );

      let start = minDate.getTime();
      let end = maxDate.getTime();
      const infos = (
        flattenArray(await Promise.all(flattenArray(promisesByServer)))
      );
      for (const info of infos) {
        const min = Math.max(start, info.start.getTime());
        const max = Math.min(end, info.end.getTime());
        if (min < max) {
          start = min;
          end = max;
        } else {
          // use the requested day if no intersection
          start = minDate.getTime();
          end = maxDate.getTime();
          break;
        }
      }

      const colocDate = new Date((start + end) / 2);
      if (colocDate < date) {
        nearestPast = colocDate;
      } else {
        nearestFuture = colocDate;
      }
    } else {
      const promisesByServer = this.groupByServer(
        coverage.getProductIds(),
        (productId) => this.getServerIdsForProductId(productId),
        (server, productIds) => server.getNearest({
          productIds,
          date,
          past   : inPast,
          present: inPresent,
          future : inFuture,
          minDate: absoluteMinDate,
          maxDate: absoluteMaxDate,
          extent,
          patternsByProductId: this._patterns(productIds),
          filtersByProductId: this._filters(productIds),
        }, abortSignal)
      );

      const nearestsByServer = await Promise.all(promisesByServer);
      nearestPast = -Infinity;
      nearestFuture = Infinity;

      for (const {past, future} of nearestsByServer) {
        if (past) {
          for (const time of past) {
            if (time !== null && time > nearestPast) {
              nearestPast = time;
            }
          }
        }
        if (future) {
          for (const time of future) {
            if (time !== null && time < nearestFuture) {
              nearestFuture = time;
            }
          }
        }
      }
    }

    nearestPast = new Date(nearestPast);
    if (isInvalidDate(nearestPast) || nearestPast < absoluteMinDate) {
      nearestPast = null;
    }

    nearestFuture = new Date(nearestFuture);
    if (isInvalidDate(nearestFuture) || nearestFuture > absoluteMaxDate) {
      nearestFuture = null;
    }

    return {
      past: nearestPast,
      future: nearestFuture,
      nearest: (
        nearestPast && nearestFuture
        ? (
          // @ts-expect-error `Date`s can be used as numbers
          date - nearestPast < nearestFuture - date
          ? nearestPast
          : nearestFuture
        )
        : nearestPast || nearestFuture || date
      ),
    };
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {ProductId} [productId]
   * @returns {void}
   */
  _updateServerIdsForProductWithId(productId) {
    if (this.syntool.productsStore.initialised) {
      if (productId) {
        this._updateServerIdsForProduct(
          this.syntool.productsStore.getProduct(productId)
        );
      } else {
        this.syntool.productsStore.forEachProduct(
          this._updateServerIdsForProduct,
          this
        );
      }
    }
  }
  /**
   * @protected
   * @param {Product | undefined} product
   * @returns {void}
   */
  _updateServerIdsForProduct(product) {
    if (product) {
      const serverIds = [];
      for (const server of this.servers.values()) {
        if (server.handlesProduct(product)) {
          serverIds.push(server.id);
        }
      }
      product.serverIds = serverIds;
    }
  }
  /**
   * @protected
   * @returns {LBRT | null}
   */
  _extent() {
    return (
      this.syntool.geoFilterActive
      ? Bounds.toLBRT(this.syntool.mapView.getVisibleProjectedBounds())
      : null
    );
  }
  /**
   * @protected
   * @param {ProductId[]} productIds
   * @returns {Record<ProductId, HybridPattern>}
   */
  _patterns(productIds) {
    return this.syntool.productsStore.getComputedPatternsForProductIds(productIds);
  }
  /**
   * @protected
   * @param {ProductId[]} productIds
   * @returns {Record<ProductId, Record<string, boolean>>}
   */
  _filters(productIds) {
    return this.syntool.productsStore.getDatasetFiltersForProductIds(productIds);
  }

  // #endregion
}
