/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/// <reference path='./api_.d.ts' />
import {Pattern} from '../pattern.js';
import {Types} from '../types.js';

export class APIServer {
  // #region Private typedefs

  /** @typedef {import('../dataset.js').Dataset} Dataset */
  /** @typedef {import('../types.js').TypeChecker} TypeChecker */

  // #endregion
  // #region Public

  /**
   * @public
   * @type {TypeChecker}
   */
  static validateConfig = Types.any;

  /**
   * @public
   * @param {Readonly<any>} config
   * @param {Readonly<ProductRawConfig>} productConfig
   * @returns {boolean}
   */
  static instanceWithConfigHandlesProduct(config, productConfig) {
    return false;
  }
  /**
   * @public
   * @param {Readonly<any>} config
   * @returns {boolean}
   */
  static instanceWithConfigSupportsComplexPatterns(config) {
    return false;
  }

  /**
   * @public
   * @param {APIServerId | {id: APIServerId}} idOrConfig
   */
  constructor(idOrConfig) {
    this.id = typeof idOrConfig === 'string' ? idOrConfig : idOrConfig.id;
    /** @type {import('./api.js').API | null} */
    this.api = null;
  }

  /**
   * @public
   * @param {string} type
   * @returns {boolean}
   */
  handlesAllProductsOfType(type) {
    return false;
  }
  /**
   * @public
   * @param {ProductRawConfig | Product} product
   * @returns {boolean}
   */
  handlesProduct(product) {
    return this.handlesAllProductsOfType(product.type);
  }

  /**
   * @public
   * @returns {boolean}
   */
  get supportsComplexPatterns() {
    return false;
  }

  /**
   * @public
   * @param {import('./api.js').API.ClearCacheOptions} [options]
   * @returns {void}
   */
  clearCache(options) { /*empty*/ }

  /**
   * @public
   * @param {string} url
   * @returns {string}
   */
  getServiceURL(url) {
    return url;
  }
  /**
   * @public
   * @param {string} url
   * @returns {string}
   */
  getDataURL(url) {
    return url;
  }

  /**
   * @public
   * @param {APIServer.Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<APIServer.DatasetInfo[]>[]}
   */
  getDatasetInfos(request, signal) {
    return [];
  }
  /**
   * @public
   * @param {APIServer.Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<APIServer.DatasetDates[]>[]}
   */
  getDatasetDates(request, signal) {
    return [];
  }
  /**
   * @public
   * @param {Readonly<Dataset>} dataset
   * @param {AbortSignal} [signal]
   * @returns {Promise<GranuleInfo>}
   */
  getDatasetMetadata(dataset, signal) {
    return Promise.resolve({});
  }
  /**
   * @public
   * @param {APIServer.Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<import('../coverage.js').CoverageMap>}
   */
  getCoverage(request, signal) {
    return Promise.resolve({});
  }
  /**
   * @public
   * @param {APIServer.NearestRequest} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<APIServer.NearestResult>}
   */
  getNearest(request, signal) {
    return Promise.resolve({});
  }

  /**
   * @public
   * @template {(url: string, ...args: any[]) => Promise<any>} F
   * @param {F} fn
   * @param {boolean} [updatesReachability]
   * @returns {F}
   */
  getAjax(fn, updatesReachability = false) {
    const server = this;
    const api = server.api;
    const id = server.id;

    /**
     * @this {ThisParameterType<F>}
     */
    function serviceAjax() {
      arguments[0] = server.getServiceURL(arguments[0]);
      return /** @type {Function} */ (fn).apply(this, arguments);
    }

    if (api && updatesReachability) {
      return /** @type {F} */ (
        /**
         * @this {ThisParameterType<F>}
         */
        function() {
          const promise = (
            /** @type {Function} */ (serviceAjax).apply(this, arguments)
          );
          api.updateServerReachabilityFromPromise(id, promise);
          return promise;
        }
      );
    }

    return /** @type {F} */ (serviceAjax);
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {{name: string}} info
   * @param {HybridPattern | null | undefined} pattern
   * @returns {boolean}
   */
  _infoMatchesPattern(info, pattern) {
    return Pattern.test(pattern, info.name);
  }

  // #endregion
}
