/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {getCoverageFromInfos} from '../coverage.js';
import {Bounds, Geometry} from '../gis.js';
import {
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  $,
  timeToTimestamp,
  timeFromTimestamp,
  clamp,
  flattenArray,
} from '../utils.js';
import {APIServer} from './api_server.js';

/**
 * @param {APIServer.DatasetInfo[]} infos
 * @param {APIServer.NearestRequest} request
 * @returns {ProductNearestResult}
 */
function findNearestFromInfos(infos, request) {
  const {
    date,
    past: inPast,
    present: inPresent,
    future: inFuture,
    minDate,
    maxDate,
  } = request;

  const time = date.getTime();
  const min = minDate?.getTime() ?? -Infinity;
  const max = maxDate?.getTime() ?? Infinity;
  const timestamp = timeToTimestamp(time);
  let futureDelta = Infinity;
  let pastDelta = -Infinity;

  for (const info of infos) {
    const start = info.start.getTime();
    const end = info.end.getTime();

    // Skip datasets with an open-ended time range...
    if (
      // ...from both ends
      (start === DATE_MIN_TIME && end === DATE_MAX_TIME) ||
      // ...from one end and the current time is inside it
      (start === DATE_MIN_TIME && time < end) ||
      (end === DATE_MAX_TIME && time > start)
    ) {
      continue;
    }

    // Skip datasets that lie outside the date-range limit
    if (start >= max || end <= min) {
      continue;
    }

    const midTimestamp = (
      start === DATE_MIN_TIME
      ? timeToTimestamp(end)
      : (
        end === DATE_MAX_TIME
        ? timeToTimestamp(start)
        : timeToTimestamp(clamp((end + start) / 2, min, max))
      )
    );
    const delta = midTimestamp - timestamp;
    if (delta === 0) {
      if (inPresent) {
        pastDelta = 0;
        futureDelta = 0;
        break;
      }
    } else if (delta < 0) {
      if (inPast && delta > pastDelta) {
        pastDelta = delta;
      }
    } else if (inFuture && delta < futureDelta) {
      futureDelta = delta;
    }
  }

  return {
    past: (
      isFinite(pastDelta) ? timeFromTimestamp(timestamp + pastDelta) : null
    ),
    future: (
      isFinite(futureDelta) ? timeFromTimestamp(timestamp + futureDelta) : null
    ),
  };
}

// #region Exported typedefs

/**
 * @typedef LocalAPIServer.Request
 * @prop {Date} [minDate]
 * @prop {Date} [maxDate]
 * @prop {LBRT | null | undefined} [extent]
 * @prop {HybridPattern} pattern
 */

// #endregion

export class LocalAPIServer extends APIServer {
  // #region Typedefs

  /** @typedef {import('../coverage.js').CoverageMap} CoverageMap */
  /**
   * @typedef ProductNearestResult
   * @prop {Milliseconds | null} past
   * @prop {Milliseconds | null} future
   */

  // #endregion
  // #region Public

  /**
   * @override
   * @public
   * @param {Readonly<any>} config
   * @returns {boolean}
   */
  static instanceWithConfigSupportsComplexPatterns(config) {
    return true;
  }

  /**
   * @override
   * @public
   * @returns {boolean}
   */
  get supportsComplexPatterns() {
    return true;
  }

  /**
   * @override
   * @public
   * @param {APIServer.Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<APIServer.DatasetInfo[]>[]}
   */
  getDatasetInfos(request, signal) {
    const {patternsByProductId} = request;
    return [
      Promise.resolve(this._getInfos(request))
      .then((infos) => infos.filter(
        (info) => (
          this._infoMatchesPattern(info, patternsByProductId[info.productId])
        )
      ))
    ];
  }
  /**
   * @override
   * @public
   * @param {APIServer.Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<APIServer.DatasetDates[]>[]}
   */
  getDatasetDates(request, signal) {
    return [Promise.resolve(this._getInfos(request))];
  }
  /**
   * @override
   * @public
   * @param {APIServer.Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<CoverageMap>}
   */
  async getCoverage(request, signal) {
    if (request.productIds.length === 0) {
      return {};
    }

    const minDate = request.minDate;
    const maxDate = request.maxDate;
    /** @type {LocalAPIServer.Request} */
    const subrequest = {
      extent : request.extent,
      pattern: '',
    };

    const coverageEntryPromises = [];

    for (const productId of request.productIds) {
      subrequest.pattern = $(request.patternsByProductId[productId]);
      const coverageEntryPromise = (
        Promise.resolve(this._getInfosForProductWithId(productId, subrequest))
        .then(
          (infos) => [productId, getCoverageFromInfos(infos, minDate, maxDate)]
        )
      );
      coverageEntryPromises.push(coverageEntryPromise);
    }

    const coverageEntries = await Promise.all(coverageEntryPromises);
    return Object.fromEntries(coverageEntries);
  }
  /**
   * @override
   * @public
   * @param {APIServer.NearestRequest} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<APIServer.NearestResult>}
   */
  async getNearest(request, signal) {
    if (request.productIds.length === 0 || !(request.past || request.future)) {
      return {};
    }

    /** @type {LocalAPIServer.Request} */
    const subrequest = {
      extent : request.extent,
      pattern: '',
    };

    const nearestPromises = [];

    for (const productId of request.productIds) {
      subrequest.pattern = $(request.patternsByProductId[productId]);
      const nearestPromise = (
        Promise.resolve(this._getInfosForProductWithId(productId, subrequest))
        .then((infos) => findNearestFromInfos(infos, request))
      );
      nearestPromises.push(nearestPromise);
    }

    const productNearests = await Promise.all(nearestPromises);
    /** @type {APIServer.NearestResult} */
    const nearest = {};
    if (request.past) {
      nearest.past = productNearests.map(
        (productNearest) => productNearest.past
      );
    }
    if (request.future) {
      nearest.future = productNearests.map(
        (productNearest) => productNearest.future
      );
    }

    return nearest;
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {APIServer.Request} request
   * @returns {Promise<APIServer.DatasetInfo[]>}
   */
  async _getInfos(request) {
    if (request.productIds.length === 0) {
      return [];
    }

    /** @type {LocalAPIServer.Request} */
    const subrequest = {
      minDate: request.minDate,
      maxDate: request.maxDate,
      extent : request.extent,
      pattern: '',
    };

    const infosPromises = [];

    for (const productId of request.productIds) {
      subrequest.pattern = $(request.patternsByProductId[productId]);
      infosPromises.push(
        Promise.resolve(this._getInfosForProductWithId(productId, subrequest))
      );
    }

    return flattenArray(await Promise.all(infosPromises));
  }
  /**
   * @protected
   * @param {ProductId} productId
   * @param {LocalAPIServer.Request | null} [request]
   * @returns {Promise<APIServer.DatasetInfo[]>}
   */
  async _getInfosForProductWithId(productId, request) {
    const infos = await this._getAllInfosForProductWithId(productId);
    const filter = request && this._getInfoFilterForRequest(request);
    return filter ? infos.filter(filter, this) : infos;
  }
  /**
   * @param {ProductId} productId
   * @returns {PromiseOrValue<APIServer.DatasetInfo[]>}
   */
  _getAllInfosForProductWithId(productId) {
    return [];
  }
  /**
   * @param {LocalAPIServer.Request} request
   * @returns {(this: this, info: APIServer.DatasetInfo) => boolean}
   */
  _getInfoFilterForRequest(request) {
    const minTime = request.minDate?.getTime();
    const maxTime = request.maxDate?.getTime();
    const pattern = request.pattern;
    const extent = request.extent && Bounds.createFromLBRT(request.extent);
    const worldBounds = this.api?.syntool.mapView.getWorldProjectedBounds();

    return (info) => !!(
      // @ts-expect-error Date is coerced into its time when compared with a number
      (minTime == null || info.end >= minTime) &&
      // @ts-expect-error Date is coerced into its time when compared with a number
      (maxTime == null || info.start <= maxTime) &&
      (!pattern || this._infoMatchesPattern(info, pattern)) &&
      (!extent || !info.bounds || Bounds.intersects(extent, info.bounds, worldBounds)) &&
      (!extent || !info.geometry || Geometry.intersectsLatLonBounds(info.geometry, extent, worldBounds))
    );
  }

  // #endregion
}
