/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  getFirstDayWithDataInCoverage,
  getLastDayWithDataInCoverage,
  mergeCoverages,
} from '../coverage.js';
import {Pattern} from '../pattern.js';
import {Types} from '../types.js';
import {
  $,
  clamp,
  dateToTimestamp,
  asArray,
  appendToArray,
  groupArrayElementsBy,
  mapObject,
  binarySearch,
  urlEndingWithSlash,
  selectUrl,
} from '../utils.js';
import {APIServer} from './api_server.js';

/**
 * @param {Timestamp} start
 * @param {Timestamp} end
 * @param {DatasetServerInfo[]} infos
 * @returns {IntervalNode}
 */
function createIntervalNode(start, end, infos) {
  return {start, end, infos};
}
/**
 * @param {Timestamp} timestamp
 * @param {IntervalNode} node
 * @returns {-1 | 0 | 1}
 */
function compareTimestampWithOpenStartIntervalNode(timestamp, node) {
  if (timestamp <= node.start) {
    return -1;
  }
  if (timestamp > node.end) {
    return 1;
  }
  return 0;
}
/**
 * @param {Timestamp} timestamp
 * @param {IntervalNode} node
 * @returns {-1 | 0 | 1}
 */
function compareTimestampWithOpenEndIntervalNode(timestamp, node) {
  if (timestamp < node.start) {
    return -1;
  }
  if (timestamp >= node.end) {
    return 1;
  }
  return 0;
}
/**
 * @param {Timestamp} timestamp
 * @param {IntervalNode} node
 * @returns {-1 | 0 | 1}
 */
function compareTimestampWithClosedIntervalNode(timestamp, node) {
  if (timestamp < node.start) {
    return -1;
  }
  if (timestamp > node.end) {
    return 1;
  }
  return 0;
}
/**
 * @param {ReadonlyArray<IntervalNode>} nodes
 * @param {Timestamp} timestamp
 * @param {(timestamp: Timestamp, node: IntervalNode) => number} comparator
 * @returns {number}
 */
function indexOfIntervalContaining(nodes, timestamp, comparator) {
  return binarySearch(nodes, comparator.bind(null, timestamp));
}
/**
 * @param {IntervalNode[]} nodes
 * @param {Timestamp} start
 * @param {Timestamp} end
 * @param {ReadonlyArray<DatasetServerInfo>} infos
 * @returns {void}
 */
function addIntervalToCache(nodes, start, end, infos) {
  const node = createIntervalNode(start, end, []);
  let indexOfStart = indexOfIntervalContaining(
    nodes,
    start,
    compareTimestampWithClosedIntervalNode
  );
  let indexOfEnd = indexOfIntervalContaining(
    nodes,
    end,
    compareTimestampWithClosedIntervalNode
  );

  if (indexOfStart < 0) {
    indexOfStart = -(indexOfStart + 1);
  } else {
    node.start = $(nodes[indexOfStart]).start;
  }
  if (indexOfEnd < 0) {
    indexOfEnd = -(indexOfEnd + 1);
  } else {
    node.end = $(nodes[indexOfEnd]).end;
    indexOfEnd += 1;
  }

  /** @type {Map<DatasetId, DatasetServerInfo>} */
  const allInfos = new Map();

  for (let i = indexOfStart; i < indexOfEnd; i++) {
    for (const info of $(nodes[i]).infos) {
      allInfos.set(info.datasetId, info);
    }
  }
  for (const info of infos) {
    allInfos.set(info.datasetId, info);
  }

  node.infos = [...allInfos.values()];
  nodes.splice(indexOfStart, indexOfEnd - indexOfStart, node);
}
/**
 * @param {ReadonlyArray<IntervalNode>} nodes
 * @param {Timestamp} minTimestamp
 * @param {Timestamp} maxTimestamp
 * @returns {CacheResponse}
 */
function queryCacheBetween(nodes, minTimestamp, maxTimestamp) {
  let indexOfMin = indexOfIntervalContaining(
    nodes,
    minTimestamp,
    compareTimestampWithOpenEndIntervalNode
  );
  let indexOfMax = indexOfIntervalContaining(
    nodes,
    maxTimestamp,
    compareTimestampWithOpenStartIntervalNode
  );
  /** @type {Interval[]} */
  const missingIntervals = [];

  const foundMinInterval = (indexOfMin >= 0);
  const foundMaxInterval = (indexOfMax >= 0);

  if (!foundMinInterval) {
    indexOfMin = -(indexOfMin + 1);
  }
  if (!foundMaxInterval) {
    indexOfMax = -(indexOfMax + 1);
  }

  // type : IN
  // cache: ---[/////[----[////[---
  // query: ----[//[---------------
  // ===OR===
  // type : OUT
  // cache: ---[/////[-------[////[---
  // query: -----------[//[-----------
  // ===OR===
  // type  : IN-IN
  // cache : ---[/////[---[////[---[////[---[////[---
  // query : ------[//////////////////////////[------
  // naming: ------|A |        B        |  C  |------
  // ===OR===
  // type  : IN-OUT
  // cache : ---[/////[---[////[---[////[------------
  // query : ------[//////////////////////////[------
  // naming: ------|A |        B        |  C  |------
  // ===OR===
  // type  : OUT-IN
  // cache : -------------[////[---[////[---[////[---
  // query : ------[//////////////////////////[------
  // naming: ------|A     |      B      |  C  |------
  // ===OR===
  // type  : OUT-OUT
  // cache : -------------[////[---[////[------------
  // query : ------[//////////////////////////[------
  // naming: ------|A     |      B      |  C  |------

  /** @type {DatasetServerInfo[]} */
  let infos = [];

  if (indexOfMin === indexOfMax) {
    // handle type IN
    if (foundMinInterval && foundMaxInterval) {
      infos = $(nodes[indexOfMin]).infos.filter(
        (info) => (
          dateToTimestamp(info.start) < maxTimestamp &&
          dateToTimestamp(info.end) > minTimestamp
        )
      );
      return {infos, missingIntervals};
    }
    // handle type OUT
    if (!foundMinInterval && !foundMaxInterval) {
      missingIntervals.push([minTimestamp, maxTimestamp]);
      return {infos, missingIntervals};
    }
  }

  // handle (A)
  if (foundMinInterval) {
    // handle type IN-*
    infos = appendToArray(
      infos,
      $(nodes[indexOfMin]).infos.filter(
        (info) => (dateToTimestamp(info.end) > minTimestamp)
      )
    );
    indexOfMin++;
  } else {
    // handle type OUT-*
    missingIntervals.push([minTimestamp, $(nodes[indexOfMin]).start]);
  }

  // handle (B)
  for (let i = indexOfMin; i < indexOfMax; i++) {
    if (i > indexOfMin || foundMinInterval) {
      missingIntervals.push([$(nodes[i - 1]).end, $(nodes[i]).start]);
    }
    infos = appendToArray(infos, $(nodes[i]).infos);
  }

  // handle (C)
  if (foundMaxInterval) {
    // handle type *-IN
    infos = appendToArray(
      infos,
      $(nodes[indexOfMax]).infos.filter(
        (info) => (dateToTimestamp(info.start) < maxTimestamp)
      )
    );
    if (indexOfMax > 0) {
      missingIntervals.push([
        $(nodes[indexOfMax - 1]).end,
        $(nodes[indexOfMax]).start,
      ]);
    }
  } else {
    // handle type *-OUT
    missingIntervals.push([$(nodes[indexOfMax - 1]).end, maxTimestamp]);
  }

  return {infos, missingIntervals};
}
/**
 * @param {ReadonlyArray<IntervalNode>} nodes
 * @param {Interval[]} intervals
 * @returns {CacheResponse}
 */
function queryCacheIntervals(nodes, intervals) {
  /** @type {DatasetServerInfo[]} */
  let infos = [];
  /** @type {Interval[]} */
  let missingIntervals = [];
  for (const [min, max] of intervals) {
    const result = queryCacheBetween(nodes, min, max);
    infos = appendToArray(infos, result.infos);
    missingIntervals = appendToArray(missingIntervals, result.missingIntervals);
  }
  return {infos, missingIntervals};
}

/**
 * @param {RemoteAPIServer.Config} server
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @param {TypeCheckerGetProp} getProp
 * @returns {TypeCheckerResult}
 */
function validateServerProductIdMap(server, propName, propFullName, getProp) {
  const products = /** @type {{id: string}[]} */ (getProp('.products'));
  /** @type {Set<string>} */
  const productIds = new Set();
  for (const product of products) {
    productIds.add(product.id);
  }

  const productIdMap = server.productIdMap;
  for (const productId of Object.keys(productIdMap)) {
    if (!productIds.has(productId)) {
      console.warn([
        'Unknown product id `', productId, '` found in `', propFullName,
        '` (server with id "', server.id, '").'
      ].join(''));
    }
  }

  return null;
}
/**
 * @param {RemoteAPIServer.Config} server
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @param {TypeCheckerGetProp} getProp
 * @returns {TypeCheckerResult}
 */
function validateServerProductIdsByDatasetFilter(server, propName, propFullName, getProp) {
  const productFilterMap = (
    server[/** @type {'productFilterMap'} */ (propName)]
  );
  if (!productFilterMap) {
    return null;
  }

  const serverProductIds = groupArrayElementsBy(
    Object.keys(server.productIdMap),
    (productId) => $(server.productIdMap[productId])
  );
  /** @type {ProductRawConfig[]} */
  const products = getProp('.products');
  const productsById = new Map(products.map((product) => [product.id, product]));

  const errors = [];
  for (const [key, value] of Object.entries(productFilterMap)) {
    for (const serverProductId of asArray(value)) {
      const productIds = serverProductIds.get(serverProductId);
      if (!productIds) {
        errors.push([
          'Unknown server product id `', serverProductId, '` found in `',
          `${propFullName}.${key}`, '` (server with id "', server.id, '").'
        ].join(''));
        continue;
      }

      // eslint-disable-next-line no-labels
      nextProduct:
      for (const productId of productIds) {
        const product = $(productsById.get(productId));
        for (const group of Object.values(product.datasetFilters ?? {})) {
          if (group[key] !== undefined) {
            // eslint-disable-next-line no-labels
            continue nextProduct;
          }
        }
        errors.push([
          'Dataset filter "', key, '" missing from product with id `', productId,
          '`, found in `', propFullName, '` (server with id "', server.id, '").'
        ].join(''));
      }
    }
  }
  return errors;
}

const LIMITS_CACHE_KEY_PREFIX = 'limits:';
const DATASETS_CACHE_KEY_PREFIX = 'datasets:';
const COVERAGE_CACHE_KEY_PREFIX = 'coverage:';

// #region Exported typedefs

/** @typedef {Named<string, 'RemoteAPIServer.ServerProductId'>} RemoteAPIServer.ServerProductId */
/** @typedef {Omit<APIServer.DatasetInfo, 'productId'> & {productId: ServerProductId}} RemoteAPIServer.DatasetServerInfo */
/** @typedef {APIServer.DatasetInfo & {serverProductId: ServerProductId}} RemoteAPIServer.DatasetInfo */
/**
 * @typedef RemoteAPIServer.IntervalNode
 * @prop {Timestamp} start
 * @prop {Timestamp} end
 * @prop {RemoteAPIServer.DatasetServerInfo[]} infos
 */
/** @typedef {RemoteAPIServer.IntervalNode[]} RemoteAPIServer.DatasetInfosCache */
/** @typedef {[start: Timestamp, end: Timestamp]} RemoteAPIServer.Interval */
/**
 * @typedef RemoteAPIServer.CacheResponse
 * @prop {RemoteAPIServer.DatasetServerInfo[]} infos
 * @prop {RemoteAPIServer.Interval[]} missingIntervals
 */
/**
 * @typedef RemoteAPIServer.CacheLimits
 * @prop {Timestamp} minTimestamp
 * @prop {Timestamp} maxTimestamp
 */
/**
 * @typedef RemoteAPIServer.DatasetInfosCacheResponse
 * @prop {RemoteAPIServer.DatasetServerInfo[]} cachedInfos
 * @prop {RemoteAPIServer.DatasetInfosCacheMissingRequest[]} requests
 */
/**
 * @typedef RemoteAPIServer.DatasetInfosCacheMissingRequest
 * @prop {RemoteAPIServer.ServerProductId} productId
 * @prop {Timestamp} minDate
 * @prop {Timestamp} maxDate
 * @prop {LBRT | null | undefined} extent
 * @prop {HybridPattern[]} patterns
 * @prop {string} cacheKey
 */
/**
 * @typedef RemoteAPIServer.CoverageCacheResponse
 * @prop {Record<ProductId, ProductCoverage>} cachedCoverages
 * @prop {RemoteAPIServer.CoverageCacheMissingRequest[]} requests
 */
/**
 * @typedef RemoteAPIServer.CoverageCacheMissingRequest
 * @prop {ProductId[]} productIds
 * @prop {RemoteAPIServer.ServerProductId} serverProductId
 * @prop {HybridPattern} pattern
 * @prop {string} cacheKey
 */
/**
 * @typedef RemoteAPIServer.Config
 * @prop {string} id
 * @prop {string} serviceURL
 * @prop {string[]} [dataURLs]
 * @prop {Record<ProductId, ServerProductId | OneOrMore<ServerProductId>>} productIdMap
 * @prop {boolean} [supportsComplexPatterns = false]
 * @prop {Record<string, ServerProductId | OneOrMore<ServerProductId>>} [productFilterMap]
 */

// #endregion

export class RemoteAPIServer extends APIServer {
  // #region Private typedefs

  /** @typedef {import('../coverage.js').ProductCoverage} ProductCoverage */
  /** @typedef {import('../dataset.js').Dataset} Dataset */
  /** @typedef {import('../types.js').TypeChecker} TypeChecker */
  /** @typedef {import('../types.js').TypeCheckerPropName} TypeCheckerPropName */
  /** @typedef {import('../types.js').TypeCheckerGetProp} TypeCheckerGetProp */
  /** @typedef {import('../types.js').TypeCheckerResult} TypeCheckerResult */
  /** @typedef {RemoteAPIServer.ServerProductId} ServerProductId */
  /** @typedef {RemoteAPIServer.DatasetServerInfo} DatasetServerInfo */
  /** @typedef {RemoteAPIServer.IntervalNode} IntervalNode */
  /** @typedef {RemoteAPIServer.Interval} Interval */
  /** @typedef {RemoteAPIServer.CacheResponse} CacheResponse */
  /** @typedef {RemoteAPIServer.CacheLimits} CacheLimits */
  /** @typedef {RemoteAPIServer.DatasetInfosCacheResponse} DatasetInfosCacheResponse */
  /** @typedef {RemoteAPIServer.DatasetInfosCacheMissingRequest} DatasetInfosCacheMissingRequest */
  /** @typedef {RemoteAPIServer.CoverageCacheResponse} CoverageCacheResponse */
  /** @typedef {RemoteAPIServer.CoverageCacheMissingRequest} CoverageCacheMissingRequest */

  // #endregion
  // #region Public

  /**
   * @public
   * @override
   * @type {TypeChecker}
   */
  static validateConfig = Types.shape({
    serviceURL  : Types.URL.isRequired,
    dataURLs    : Types.arrayOf(Types.URL.isRequired),
    productIdMap: Types.allOfType([
      Types.objectOf(
        Types.oneOfType([
          Types.string,
          Types.arrayOf(Types.string.isRequired, [1]),
        ]).isRequired,
      ),
      validateServerProductIdMap,
    ]).isRequired,
    supportsComplexPatterns: Types.bool,
    productFilterMap: Types.allOfType([
      Types.objectOf(
        Types.oneOfType([
          Types.string,
          Types.arrayOf(Types.string.isRequired, [1]),
        ]).isRequired,
      ),
      validateServerProductIdsByDatasetFilter,
    ]),
  });

  /**
   * @public
   * @override
   * @param {Readonly<RemoteAPIServer.Config>} config
   * @param {Readonly<ProductRawConfig>} productConfig
   * @returns {boolean}
   */
  static instanceWithConfigHandlesProduct(config, productConfig) {
    return !!config.productIdMap[productConfig.id];
  }
  /**
   * @override
   * @public
   * @param {Readonly<RemoteAPIServer.Config>} config
   * @returns {boolean}
   */
  static instanceWithConfigSupportsComplexPatterns(config) {
    return !!config.supportsComplexPatterns;
  }

  /**
   * @public
   * @param {Readonly<RemoteAPIServer.Config>} config
   */
  constructor(config) {
    super(config);

    this.serviceURL = urlEndingWithSlash(config.serviceURL);
    this.productIdMap = (
      /** @type {Record<ProductId, OneOrMore<ServerProductId>>} */
      (mapObject(config.productIdMap, asArray))
    );
    if (config.supportsComplexPatterns) {
      Object.defineProperty(this, 'supportsComplexPatterns', {get: () => true});
    }

    if (config.dataURLs && config.dataURLs.length > 0) {
      this.dataURLs = config.dataURLs.map(urlEndingWithSlash);
    } else {
      this.dataURLs = [''];
    }

    /** @type {Map<string, unknown>} */
    this.cache = new Map();
    this.reverseProductIdsMap = groupArrayElementsBy(
      Object.keys(this.productIdMap),
      (productId) => $(this.productIdMap[productId])
    );

    const {productFilterMap = {}} = config;
    this.inclusionDatasetFiltersMap = groupArrayElementsBy(
      Object.keys(productFilterMap),
      (filter) => $(productFilterMap[filter])
    );
  }

  /**
   * @public
   * @override
   * @param {Readonly<ProductRawConfig | Product>} product
   * @returns {boolean}
   */
  handlesProduct(product) {
    return !!this.productIdMap[product.id];
  }

  /**
   * @public
   * @override
   * @param {Readonly<import('./api.js').API.ClearCacheOptions>} options
   * @returns {void}
   */
  clearCache(options) {
    const cache = this.cache;
    if (!options || (options.datasets && options.coverage)) {
      cache.clear();
    } else if (options.datasets) {
      for (const [key] of cache) {
        if (
          key.startsWith(DATASETS_CACHE_KEY_PREFIX) ||
          key.startsWith(LIMITS_CACHE_KEY_PREFIX)
        ) {
          cache.delete(key);
        }
      }
    } else if (options.coverage) {
      for (const [key] of cache) {
        if (key.startsWith(COVERAGE_CACHE_KEY_PREFIX)) {
          cache.delete(key);
        }
      }
    }
  }

  /**
   * @public
   * @override
   * @param {string} url
   * @returns {string}
   */
  getServiceURL(url) {
    return this.serviceURL + url;
  }
  /**
   * @public
   * @override
   * @param {string} url
   * @returns {string}
   */
  getDataURL(url) {
    return selectUrl(url, this.dataURLs) + url;
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {ProductId} productId
   * @returns {ServerProductId[]}
   */
  _serverProductIdsFromProductId(productId) {
    return this.productIdMap[productId] || [];
  }
  /**
   * @protected
   * @param {ServerProductId} serverProductId
   * @returns {ProductId[]}
   */
  _productIdsFromServerProductId(serverProductId) {
    return this.reverseProductIdsMap.get(serverProductId) || [];
  }
  /**
   * @protected
   * @param {Readonly<Dataset>} dataset
   * @returns {ServerProductId}
   */
  _serverProductIdFromDataset(dataset) {
    return (
      /** @type {RemoteAPIServer.DatasetInfo} */
      (/** @type {unknown} */ (dataset._info)).serverProductId
      || /** @type {ServerProductId} */ (dataset.productId)
    );
  }

  /**
   * @protected
   * @param {Readonly<DatasetServerInfo>} serverInfo
   * @param {ProductId} productId
   * @returns {APIServer.DatasetInfo}
   */
  _infoFromServerInfo(serverInfo, productId) {
    if (serverInfo.productId === productId) {
      return /** @type {APIServer.DatasetInfo} */ (serverInfo);
    }
    return Object.assign({}, serverInfo, {
      serverProductId: serverInfo.productId,
      productId,
      datasetId: serverInfo.datasetId.replace(serverInfo.productId, productId),
    });
  }
  /**
   * @protected
   * @param {Readonly<Record<ProductId, HybridPattern>>} patternsByProductId
   * @param {ReadonlyArray<DatasetServerInfo> | null | undefined} serverInfos
   * @returns {APIServer.DatasetInfo[]}
   */
  _infosFromServerInfos(patternsByProductId, serverInfos) {
    const infos = [];
    for (const serverInfo of serverInfos || []) {
      for (
        const productId
        of this._productIdsFromServerProductId(serverInfo.productId)
      ) {
        if (
          this._infoMatchesPattern(serverInfo, patternsByProductId[productId])
        ) {
          infos.push(this._infoFromServerInfo(serverInfo, productId));
        }
      }
    }
    return infos;
  }

  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @returns {boolean}
   */
  _hasCache(extent) {
    const extentStr = extent?.join(',');
    return this.cache.get('extent') === extentStr;
  }
  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @returns {Map<string, unknown>}
   */
  _getCache(extent) {
    const extentStr = extent?.join(',');
    if (this.cache.get('extent') !== extentStr) {
      this.cache.clear();
      this.cache.set('extent', extentStr);
    }
    return this.cache;
  }
  /**
   * @protected
   * @template T
   * @param {Readonly<LBRT> | null | undefined} extent
   * @param {string} key
   * @param {T} value
   * @returns {T}
   */
  _updateCache(extent, key, value) {
    if (this._hasCache(extent)) {
      this.cache.set(key, value);
    }
    return value;
  }
  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @param {string} key
   * @param {Timestamp} minTimestamp
   * @param {Timestamp} maxTimestamp
   * @returns {CacheLimits | undefined}
   */
  _updateLimitsCache(extent, key, minTimestamp, maxTimestamp) {
    key = LIMITS_CACHE_KEY_PREFIX + key.split(':', 2)[1];

    const limits = (
      this._hasCache(extent)
      ? /** @type {CacheLimits | undefined} */ (this.cache.get(key))
      : undefined
    );

    if (
      !isFinite(minTimestamp) ||
      !isFinite(maxTimestamp) ||
      minTimestamp > maxTimestamp ||
      (
        limits &&
        limits.minTimestamp <= minTimestamp &&
        limits.maxTimestamp >= maxTimestamp
      )
    ) {
      return limits;
    }

    const newLimits = {
      minTimestamp: (
        limits
        ? Math.min(limits.minTimestamp, minTimestamp)
        : minTimestamp
      ),
      maxTimestamp: (
        limits
        ? Math.max(limits.maxTimestamp, maxTimestamp)
        : maxTimestamp
      ),
    };
    return this._updateCache(extent, key, newLimits);
  }

  /**
   * @protected
   * @param {ServerProductId} serverProductId
   * @param {Record<string, boolean> | undefined} filters
   * @returns {boolean}
   */
  _skipServerProductForFilters(serverProductId, filters) {
    if (filters) {
      const inclusionFilters = (
        this.inclusionDatasetFiltersMap.get(serverProductId)
      );
      if (inclusionFilters) {
        for (const filter of inclusionFilters) {
          if (filters[filter]) {
            return false;
          }
        }
        return true;
      }
    }
    return false;
  }

  /**
   * @protected
   * @param {Readonly<APIServer.Request>} request
   * @returns {DatasetInfosCacheResponse}
   */
  _queryDatasetInfosCache(request) {
    /** @type {DatasetInfosCacheMissingRequest[]} */
    const requests = [];

    const minTimestamp = (
      request.minDate
      ? dateToTimestamp(request.minDate)
      : -Infinity
    );
    const maxTimestamp = (
      request.maxDate
      ? dateToTimestamp(request.maxDate)
      : Infinity
    );

    if (request.productIds.length === 0 || minTimestamp >= maxTimestamp) {
      return {cachedInfos: [], requests};
    }

    /** @type {Record<ProductId, HybridPattern>} */
    const patternsByProductId = {};
    /** @type {Set<ServerProductId>} */
    const serverProductIdsWithoutPattern = new Set();
    for (const productId of request.productIds) {
      patternsByProductId[productId] = (
        Pattern.simplify(request.patternsByProductId[productId], 'auto')
      );
      if (!patternsByProductId[productId]) {
        for (
          const serverProductId
          of this._serverProductIdsFromProductId(productId)
        ) {
          serverProductIdsWithoutPattern.add(serverProductId);
        }
      }
    }

    const extent = request.extent;
    const cache = this._getCache(extent);

    /** @type {Set<string>} */
    const queriedCacheKeys = new Set();
    /** @type {Map<DatasetId, DatasetServerInfo>} */
    const cachedInfos = new Map();
    for (const productId of request.productIds) {
      const pattern = $(patternsByProductId[productId]);
      if (pattern === '()') {
        // pattern matches nothing => response will be empty
        continue;
      }

      const filters = request.filtersByProductId[productId];
      for (
        const serverProductId of this._serverProductIdsFromProductId(productId)
      ) {
        if (pattern && serverProductIdsWithoutPattern.has(serverProductId)) {
          // Will be handled by a different iteration of this loop, the one
          // without a pattern
          continue;
        }
        if (this._skipServerProductForFilters(serverProductId, filters)) {
          continue;
        }

        const cacheKey = `${DATASETS_CACHE_KEY_PREFIX}${serverProductId}:`;
        const cacheKeyWithPattern = cacheKey + pattern;

        if (queriedCacheKeys.has(cacheKeyWithPattern)) {
          continue;
        }
        queriedCacheKeys.add(cacheKeyWithPattern);

        /** @type {Interval[]} */
        let missingIntervals = [[minTimestamp, maxTimestamp]];
        for (const key of [cacheKey, pattern && cacheKeyWithPattern]) {
          const infosCache = /** @type {IntervalNode[]} */ (cache.get(key));
          if (infosCache) {
            const result = queryCacheIntervals(infosCache, missingIntervals);
            for (const info of result.infos) {
              cachedInfos.set(info.datasetId, info);
            }
            missingIntervals = result.missingIntervals;
          }
        }

        const patterns = [pattern];
        for (const [minDate, maxDate] of missingIntervals) {
          requests.push({
            productId: serverProductId,
            minDate,
            maxDate,
            extent,
            patterns,
            cacheKey,
          });
        }
      }
    }

    return {cachedInfos: [...cachedInfos.values()], requests};
  }
  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @param {string} key
   * @param {Timestamp} minTimestamp
   * @param {Timestamp} maxTimestamp
   * @param {ReadonlyArray<DatasetServerInfo>} infos
   * @param {boolean} partialInfos
   * @returns {void}
   */
  _updateDatasetsCache(extent, key, minTimestamp, maxTimestamp, infos, partialInfos) {
    if (!this._hasCache(extent)) {
      return;
    }

    let minInfosTimestamp = Infinity;
    let maxInfosTimestamp = -Infinity;
    for (const info of infos) {
      const infoStartTimestamp = dateToTimestamp(info.start);
      if (infoStartTimestamp < minInfosTimestamp) {
        minInfosTimestamp = infoStartTimestamp;
      }
      if (infoStartTimestamp > maxInfosTimestamp) {
        maxInfosTimestamp = infoStartTimestamp;
      }
    }

    if (partialInfos) {
      maxTimestamp = maxInfosTimestamp;
    }

    let cache = /** @type {IntervalNode[]} */ (this.cache.get(key));
    if (!cache) {
      cache = [];
      this.cache.set(key, cache);
    }

    const limits = this._updateLimitsCache(
      extent,
      key,
      minInfosTimestamp,
      maxInfosTimestamp
    );
    if (limits) {
      minTimestamp = clamp(minTimestamp, limits.minTimestamp, limits.maxTimestamp);
      maxTimestamp = clamp(maxTimestamp, limits.minTimestamp, limits.maxTimestamp);
    }

    if (minTimestamp <= maxTimestamp) {
      addIntervalToCache(cache, minTimestamp, maxTimestamp, infos);
    }
  }
  /**
   * @protected
   * @param {Readonly<DatasetInfosCacheMissingRequest>} request
   * @param {ReadonlyArray<DatasetServerInfo>} infos
   * @param {boolean} partialInfos
   * @returns {void}
   */
  _updateDatasetsCaches(request, infos, partialInfos) {
    const {extent, productId, patterns, cacheKey, minDate, maxDate} = request;

    if (this._hasCache(extent)) {
      for (const pattern of patterns) {
        const patternMatches = Pattern.test(pattern);
        this._updateDatasetsCache(
          extent,
          cacheKey + pattern,
          minDate,
          maxDate,
          infos.filter(
            (info) => productId === info.productId && patternMatches(info.name)
          ),
          partialInfos
        );
      }
    }
  }

  /**
   * @protected
   * @param {Readonly<APIServer.Request>} request
   * @returns {CoverageCacheResponse}
   */
  _queryCoverageCache(request) {
    /** @type {Record<ProductId, ProductCoverage[]>} */
    const cachedCoverages = {};
    /** @type {Map<string, CoverageCacheMissingRequest>} */
    const requests = new Map();
    const extent = request.extent;
    const cache = this._getCache(extent);

    for (const productId of request.productIds) {
      const requestPattern = request.patternsByProductId[productId];
      const pattern = Pattern.simplify(requestPattern, 'auto');
      if (pattern === '()') {
        // pattern matches nothing => response will be empty
        cachedCoverages[productId] = [];
        continue;
      }

      const filters = request.filtersByProductId[productId];
      for (
        const serverProductId of this._serverProductIdsFromProductId(productId)
      ) {
        if (this._skipServerProductForFilters(serverProductId, filters)) {
          cachedCoverages[productId] ??= [];
          continue;
        }

        const key = `${COVERAGE_CACHE_KEY_PREFIX}${serverProductId}:${pattern}`;

        if (cache.has(key)) {
          cachedCoverages[productId] ??= [];
          $(cachedCoverages[productId]).push(
            /** @type {ProductCoverage} */ (cache.get(key))
          );
        } else {
          let r = requests.get(key);
          if (!r) {
            r = {productIds: [], serverProductId, pattern, cacheKey: key};
            requests.set(key, r);
          }
          r.productIds.push(productId);
        }
      }
    }

    return {
      cachedCoverages: mapObject(cachedCoverages, mergeCoverages),
      requests: [...requests.values()],
    };
  }
  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @param {string} key
   * @param {Readonly<ProductCoverage>} coverage
   * @returns {void}
   */
  _updateCoverageCache(extent, key, coverage) {
    // Get end of first day with coverage
    const firstDay = getFirstDayWithDataInCoverage(coverage);
    if (firstDay) {
      firstDay.setUTCDate(firstDay.getUTCDate() + 1);
      firstDay.setUTCSeconds(-1);

      // Get start of last day with coverage
      const lastDay = $(getLastDayWithDataInCoverage(coverage));

      this._updateLimitsCache(
        extent,
        key,
        dateToTimestamp(firstDay),
        dateToTimestamp(lastDay)
      );
    }
    this._updateCache(extent, key, coverage);
  }
  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @param {ReadonlyArray<{cacheKey: string, coverage?: ProductCoverage}>} requests
   * @returns {void}
   */
  _updateCoverageCaches(extent, requests) {
    for (const {cacheKey, coverage} of requests) {
      if (coverage) {
        this._updateCoverageCache(extent, cacheKey, coverage);
      }
    }
  }

  // #endregion
}
