/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {ajax, ajaxRetriable} from '../utils.js';

/**
 * @typedef URLShorteningRequest
 * @prop {string} url
 * @prop {import('../utils.js').AjaxOptions['method']} [method]
 * @prop {import('../utils.js').AjaxOptions['data']} [data]
 * @prop {import('../utils.js').AjaxOptions['responseType']} [responseType]
 * @prop {(response: unknown) => string | Error} [callback]
 * @prop {(error: Error) => string | Error} [errorback]
 */
/** @typedef {(url: string) => URLShorteningRequest} URLShortenerCreator */

/** @type {URLShortenerCreator | null} */
let requestForShorteningURL = null;

/**
 * @param {URLShortenerCreator | null} shortenerForURL
 * @returns {void}
 */
export function setShortenerForURL(shortenerForURL) {
  requestForShorteningURL = shortenerForURL;
}

/**
 * @param {((arg: any) => string | Error) | undefined} fn
 * @returns {((arg: unknown) => string) | undefined}
 */
function shortenWrap(fn) {
  return fn &&
    /**
     * @param {unknown} arg
     * @returns {string}
     */
    function(arg) {
      const result = fn(arg);

      if (result instanceof Error) {
        throw result;
      }
      return result;
    };
}
/**
 * @param {string} originalUrl
 * @returns {Promise<string>}
 */
export function shortenURL(originalUrl) {
  if (typeof requestForShorteningURL !== 'function') {
    return Promise.resolve(originalUrl);
  }

  const {url, method, data, responseType, callback, errorback} = (
    requestForShorteningURL(originalUrl)
  );

  return (
    ajaxRetriable(ajax)(url, {method, data, responseType})
    .then(shortenWrap(callback), shortenWrap(errorback))
  );
}
