/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {mergeCoverages, mergeCoverageMaps} from '../coverage.js';
import {Bounds} from '../gis.js';
import {Pattern} from '../pattern.js';
import {Types} from '../types.js';
import {
  $,
  emptyFunction,
  parseDecimal,
  timeFromTimestamp,
  dateToTimestamp,
  dateFromTimestamp,
  normalizeISODateString,
  flattenArray,
  groupArrayElementsBy,
  isBigEndian,
  ajax,
} from '../utils.js';
import {RemoteAPIServer} from './remote_api_server.js';

const COVERAGE_DELIMITER = 0xFFFFFFFF;

/**
 * @param {Uint32Array} serializedCoverage
 * @returns {ProductCoverage}
 */
function deserializeCoverage(serializedCoverage) {
  /** @type {ProductCoverage} */
  const coverage = Object.create(null);

  if (!serializedCoverage) {
    return coverage;
  }

  const yearsCount = serializedCoverage.length / 13;

  for (let i = 0; i < yearsCount; i++) {
    const year = $(serializedCoverage[i * 13]);

    coverage[year] = serializedCoverage.subarray(i * 13 + 1, (i + 1) * 13);
  }

  return coverage;
}

/**
 * @param {ServerGranuleInfoGroup | OldServerGranuleInfoGroup} group
 * @returns {group is ServerGranuleInfoGroup}
 */
function isServerGranuleInfoGroupInNewFormat(group) {
  return !!group.metadata?.syntool_id;
}
/**
 * @param {ServerGranuleInfoGroup} group
 * @returns {[id: string, GranuleInfoGroup]}
 */
function parseGranuleInfoGroup(group) {
  const groupId = group.metadata.syntool_id;
  const g = /** @type {GranuleInfoGroup} */ (group);
  delete g.metadata.syntool_id;
  return [groupId, g];
}
/**
 * @param {Readonly<Dataset>} dataset
 * @param {Readonly<OldServerGranuleInfoGroup>} oldGroup
 * @returns {[id: string, GranuleInfoGroup]}
 */
function parseOldGranuleInfoGroup(dataset, oldGroup) {
  const groupId = oldGroup.global?.feature_type?.toLowerCase() ?? 'default';
  /** @type {GranuleInfoGroup} */
  // @ts-expect-error
  const group = {metadata: {}};
  if (oldGroup.metadata) {
    for (const [key, value] of Object.entries(oldGroup.metadata)) {
      switch (key.toLowerCase()) {
      case 'spectrum':
        group[key] = {url: `${dataset.name}.${value}`};
        break;
      case 'profile':
        group[key] = {url: $(value)};
        break;
      case 'image':
        group[key] = {url: `features/${value}.png`};
        break;
      default:
        group.metadata[key] = $(value);
      }
    }
  }
  if (oldGroup.images) {
    for (const [key, value] of Object.entries(oldGroup.images)) {
      const autoSeparatorIndex = value.lastIndexOf('$');
      const urlSeparatorIndex = value.lastIndexOf('$', autoSeparatorIndex - 1);
      const description = value.substring(0, urlSeparatorIndex);
      const url = value.substring(urlSeparatorIndex + 1, autoSeparatorIndex);

      group[`images_${key}`] = {url: url, description: description};
    }
  }
  if (oldGroup.links) {
    for (const [key, value] of Object.entries(oldGroup.links)) {
      const separatorIndex = value.lastIndexOf('$');
      const description = value.substring(0, separatorIndex);
      const url = value.substring(separatorIndex + 1);

      group[`links_${key}`] = {url: url, description: description};
    }
  }

  return [groupId, group];
}

/**
 * @param {LegacyPattern} pattern
 * @returns {string}
 */
function escapeLegacyPattern(pattern) {
  return pattern.replace(/[\\_%]/g, '\\$&');
}

// #region Exported typedefs

/**
 * @typedef SyntoolAPIServer.DatasetInfosRequest
 * @prop {SyntoolAPIServer.DatasetsRequestData} data
 * @prop {OneOrMore<DatasetInfosCacheMissingRequest>} requestsGroup
 */
/**
 * @typedef SyntoolAPIServer.DatasetsRequestData
 * @prop {string} products
 * @prop {Timestamp} minDate
 * @prop {Timestamp} maxDate
 * @prop {string} [extent]
 * @prop {string} [patterns]
 * @prop {boolean} [dates_only]
 */
/** @typedef {Array<[start: Timestamp, end: Timestamp]> | null | undefined} SyntoolAPIServer.DatasetsDatesResponse */
/** @typedef {DatasetServerEvent[] | null | undefined} SyntoolAPIServer.DatasetsEventsResponse */
/**
 * @typedef SyntoolAPIServer.CoverageRequestData
 * @prop {string} products
 * @prop {string} [extent]
 * @prop {string} [patterns]
 * @prop {'big'} [endianess]
 */
/** @typedef {ArrayBuffer} SyntoolAPIServer.CoverageResponse */
/**
 * @typedef SyntoolAPIServer.NearestRequestData
 * @prop {Timestamp} date
 * @prop {string} products
 * @prop {string} [extent]
 * @prop {string} [patterns]
 * @prop {boolean} [past]
 * @prop {boolean} [present]
 * @prop {boolean} [future]
 */
/**
 * @typedef SyntoolAPIServer.NearestResponse
 * @prop {Array<Timestamp | null | undefined>} [past]
 * @prop {Array<Timestamp | null | undefined>} [future]
 */
/** @typedef {RemoteAPIServer.Config & ConfigExtras} SyntoolAPIServer.Config */

// #endregion

export class SyntoolAPIServer extends RemoteAPIServer {
  // #region Private typedefs

  /** @typedef {import('../dataset.js').Dataset} Dataset */
  /** @typedef {import('../coverage.js').ProductCoverage} ProductCoverage */
  /** @typedef {import('../coverage.js').CoverageMap} CoverageMap */
  /** @typedef {import('../types.js').TypeChecker} TypeChecker */
  /** @typedef {import('./api_server.js').APIServer.DatasetDates} DatasetDates */
  /** @typedef {import('./api_server.js').APIServer.Request} Request */
  /** @typedef {import('./api_server.js').APIServer.NearestRequest} NearestRequest */
  /** @typedef {import('./api_server.js').APIServer.NearestResult} NearestResult */
  /** @typedef {import('./api_server.js').APIServer.DatasetInfo} DatasetInfo */
  /** @typedef {RemoteAPIServer.DatasetServerInfo} DatasetServerInfo */
  /** @typedef {RemoteAPIServer.DatasetInfosCacheMissingRequest} DatasetInfosCacheMissingRequest */
  /** @typedef {RemoteAPIServer.CoverageCacheResponse} CoverageCacheResponse */
  /** @typedef {RemoteAPIServer.CoverageCacheMissingRequest} CoverageCacheMissingRequest */

  /**
   * @typedef ConfigExtras
   * @prop {number} [datasetsResponseLimit = 5000]
   */
  /**
   * @typedef DatasetServerEvent
   * @prop {string} productId
   * @prop {string} datasetId
   * @prop {string} title
   * @prop {string} start
   * @prop {string} end
   * @prop {string} mapMaxZoom
   * @prop {string} uri
   * @prop {string} polygon
   * @prop {unknown[]} resolutions
   * @prop {NSEW | null | undefined} mapBounds
   * @prop {unknown} [features]
   */
  /**
   * @typedef {{
   *   global?: {
   *     feature_type?: string,
   *     display_type?: string,
   *   },
   *   metadata?: {
   *     [key: string]: string | undefined,
   *     min     ?: string, // representing a number
   *     max     ?: string, // representing a number
   *     units   ?: string | 'k' | `log(${string})`,
   *     spectrum?: string, // the spectrum part of the URL: `${datasetDataURL}/${datasetId}.${spectrum}`
   *     profile ?: string, // profile URL relative to the dataset's data URL
   *     image   ?: string,
   *   },
   *   links?: {
   *     [label: string]: string, // `${desc}$${URL}`
   *   },
   *   images?: {
   *     [label: string]: string, // `${desc}$${URL}$${auto}` auto := 0 | 1 => shown automatically
   *   },
   * }} OldServerGranuleInfoGroup
   */
  /** @typedef {GranuleInfoGroup & {metadata: {syntool_id: string}}} ServerGranuleInfoGroup */
  /** @typedef {Array<ServerGranuleInfoGroup | OldServerGranuleInfoGroup>} ServerGranuleInfo */

  // #endregion
  // #region Public

  /**
   * @public
   * @override
   * @type {TypeChecker}
   */
  static validateConfig = Types.allOfType([
    RemoteAPIServer.validateConfig,
    Types.shape({
      datasetsResponseLimit: Types.integer.gt(0),
    }),
  ]);

  /**
   * @public
   * @param {SyntoolAPIServer.Config} config
   */
  constructor(config) {
    super(config);

    this.datasetsResponseLimit = config.datasetsResponseLimit || 5000;
  }

  /**
   * @public
   * @override
   * @param {Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<DatasetInfo[]>[]}
   */
  getDatasetInfos(request, signal) {
    const {cachedInfos, requests} = this._queryDatasetInfosCache(request);
    const {patternsByProductId} = request;
    const infoPromises = (
      this._groupDatasetInfosRequests(requests)
      .map(async ({data, requestsGroup}) => {
        const response = /** @type {SyntoolAPIServer.DatasetsEventsResponse} */ (
          await this.getAjax(ajax.getJSON, true)('datasets', data, signal)
        );

        const serverInfos = this._serverInfosFromEvents(requestsGroup, response);
        const infos = this._infosFromServerInfos(patternsByProductId, serverInfos);

        return infos;
      })
    );

    infoPromises.push(
      Promise.resolve(
        this._infosFromServerInfos(patternsByProductId, cachedInfos)
      )
    );
    return infoPromises;
  }
  /**
   * @public
   * @override
   * @param {Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<DatasetDates[]>[]}
   */
  getDatasetDates(request, signal) {
    const {cachedInfos, requests} = this._queryDatasetInfosCache(request);
    const infoPromises = (
      this._groupDatasetInfosRequests(requests)
      .map(async ({data}) => {
        data.dates_only = true;
        const response = /** @type {SyntoolAPIServer.DatasetsDatesResponse} */ (
          await this.getAjax(ajax.getJSON, true)('datasets', data, signal)
        );

        return this._infosFromDates(response);
      })
    );

    infoPromises.push(Promise.resolve(cachedInfos));
    return infoPromises;
  }
  /**
   * @public
   * @override
   * @param {Readonly<Dataset>} dataset
   * @param {AbortSignal} [signal]
   * @returns {Promise<GranuleInfo>}
   */
  async getDatasetMetadata(dataset, signal) {
    const serverGranuleInfo = /** @type {ServerGranuleInfo} */ (
      $(dataset._info).metadata
      ? $(dataset._info).metadata
      : await this.getAjax(ajax.getJSON, true)(
        'getGranuleInfo/',
        {
          product_id  : this._serverProductIdFromDataset(dataset),
          dataset_name: dataset.name,
        }
      )
    );
    return this._parseGranuleInfo(dataset, serverGranuleInfo);
  }
  /**
   * @public
   * @override
   * @param {Request} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<CoverageMap>}
   */
  async getCoverage(request, signal) {
    const {cachedCoverages, requests} = (
      /**
       * @type {CoverageCacheResponse & {
       *   requests: Array<
       *     CoverageCacheMissingRequest
       *     & {coverage?: ProductCoverage}
       *   >,
       * }}
       */
      (this._queryCoverageCache(request))
    );

    if (requests.length === 0) {
      return cachedCoverages;
    }

    const extents = this._wrapExtent(request.extent);
    const serverProductIds = requests.map((r) => r.serverProductId);
    const patterns = requests.map((r) => this._preparePattern(r.pattern));
    /** @type {SyntoolAPIServer.CoverageRequestData} */
    const data = {
      products: serverProductIds.join(','),
    };

    if (patterns.some(Boolean)) {
      data.patterns = patterns.join(',');
    }
    if (isBigEndian) {
      data.endianess = 'big';
    }

    const fetchedCoveragesByExtent = await Promise.all(
      extents.map(async (extent) => {
        data.extent = extent;
        const response = (
          await this.getAjax(ajax.getArrayBuffer, true)('coverage', data, signal)
        );
        return this._handleCoverageResponse(requests, response);
      })
    );
    this._updateCoverageCaches(request.extent, requests);
    return mergeCoverageMaps([cachedCoverages, ...fetchedCoveragesByExtent]);
  }

  /**
   * @public
   * @override
   * @param {NearestRequest} request
   * @param {AbortSignal} [signal]
   * @returns {Promise<NearestResult>}
   */
  async getNearest(request, signal) {
    if (request.productIds.length === 0) {
      return {};
    }

    const indexMap = new Map();
    const indices = [];
    const serverProductIds = [];
    const patterns = [];

    for (const productId of request.productIds) {
      const requestPattern = request.patternsByProductId[productId];
      const pattern = Pattern.simplify(requestPattern, 'auto');
      if (pattern === '()') {
        // pattern matches nothing => response will be empty
        continue;
      }

      const filters = request.filtersByProductId[productId];
      const productIndices = [];
      for (
        const serverProductId of this._serverProductIdsFromProductId(productId)
      ) {
        if (this._skipServerProductForFilters(serverProductId, filters)) {
          continue;
        }

        const key = `${serverProductId},${pattern}`;
        let index = indexMap.get(key);
        if (index === undefined) {
          index = serverProductIds.length;
          indexMap.set(key, index);
          serverProductIds.push(serverProductId);
          patterns.push(this._preparePattern(pattern));
        }
        productIndices.push(index);
      }
      indices.push(productIndices);
    }

    if (serverProductIds.length === 0) {
      return {};
    }

    const extents = this._wrapExtent(request.extent);
    /** @type {SyntoolAPIServer.NearestRequestData} */
    const data = {
      date    : dateToTimestamp(request.date),
      products: serverProductIds.join(','),
    };

    if (patterns.some(Boolean)) {
      data.patterns = patterns.join(',');
    }
    if (request.past) {
      data.past = true;
    }
    if (request.present) {
      data.present = true;
    }
    if (request.future) {
      data.future = true;
    }

    const serverNearestByExtent = await Promise.all(
      extents.map((extent) => {
        data.extent = extent;
        return /** @type {SyntoolAPIServer.NearestResponse} */ (
          this.getAjax(ajax.getJSON, true)('nearest', data, signal)
        );
      })
    );
    const {past: serverPast, future: serverFuture} = serverNearestByExtent[0] ?? {};
    for (const {past, future} of serverNearestByExtent) {
      for (let i = 0; i < serverProductIds.length; i++) {
        if (serverPast) {
          const serverPastTimestamp = serverPast[i];
          const pastTimestamp = past?.[i];
          if (
            pastTimestamp != null &&
            (
              serverPastTimestamp == null ||
              pastTimestamp > serverPastTimestamp
            )
          ) {
            serverPast[i] = pastTimestamp;
          }
        }
        if (serverFuture) {
          const serverFutureTimestamp = serverFuture[i];
          const futureTimestamp = future?.[i];
          if (
            futureTimestamp != null &&
            (
              serverFutureTimestamp == null ||
              futureTimestamp < serverFutureTimestamp
            )
          ) {
            serverFuture[i] = futureTimestamp;
          }
        }
      }
    }

    const past = [];
    const future = [];
    for (const productIndices of indices) {
      let pastTimestamp = null;
      let futureTimestamp = null;
      for (const index of productIndices) {
        const serverPastTimestamp = serverPast?.[index];
        const serverFutureTimestamp = serverFuture?.[index];
        if (
          serverPastTimestamp != null &&
          (pastTimestamp === null || serverPastTimestamp > pastTimestamp)
        ) {
          pastTimestamp = serverPastTimestamp;
        }
        if (
          serverFutureTimestamp != null &&
          (futureTimestamp === null || serverFutureTimestamp < futureTimestamp)
        ) {
          futureTimestamp = serverFutureTimestamp;
        }
      }
      if (serverPast) {
        past.push(
          pastTimestamp === null
          ? null
          : timeFromTimestamp(pastTimestamp)
        );
      }
      if (serverFuture) {
        future.push(
          futureTimestamp === null
          ? null
          : timeFromTimestamp(futureTimestamp)
        );
      }
    }

    const nearest = {};
    if (serverPast) {
      nearest.past = past;
    }
    if (serverFuture) {
      nearest.future = future;
    }
    return nearest;
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {Readonly<LBRT> | null | undefined} extent
   * @returns {(string | undefined)[]}
   */
  _wrapExtent(extent) {
    const worldBounds = this.api?.syntool.mapView.getWorldProjectedBounds();

    if (!extent || !worldBounds) {
      return [extent?.join(',')];
    }

    const worldWidth = Bounds.getWidth(worldBounds);
    const worldLeft = Bounds.getLeft(worldBounds);
    const worldRight = Bounds.getRight(worldBounds);

    const [left, bottom, right, top] = extent;
    const extentWidth = right - left;
    if (extentWidth >= worldWidth) {
      return [[worldLeft, bottom, worldRight, top].join(',')];
    }

    const extentLeft = /** @type {LBRT} */ ([left, bottom, worldRight, top]);
    const extentRight = /** @type {LBRT} */ ([worldLeft, bottom, right, top]);

    if (left < worldLeft) {
      extentLeft[0] = left + worldWidth;
    } else if (right > worldRight) {
      extentRight[2] = right - worldWidth;
    } else {
      return [extent.join(',')];
    }
    return [extentLeft.join(','), extentRight.join(',')];
  }
  /**
   * @param {HybridPattern | undefined} pattern
   * @returns {string}
   */
  _preparePattern(pattern) {
    if (!this.supportsComplexPatterns && pattern) {
      const _pattern = pattern;
      pattern = Pattern.toLegacy(pattern, 'auto');
      if (pattern == null) {
        console.warn(
          `SyntoolAPIServer(${this.id}): Ignoring unsupported pattern ` +
          `"${_pattern}".`
        );
      }
    }
    pattern ??= '';
    if (Pattern.isLegacy(pattern)) {
      return escapeLegacyPattern(pattern);
    }
    return pattern;
  }
  /**
   * @protected
   * @param {DatasetInfosCacheMissingRequest[]} requests
   * @returns {SyntoolAPIServer.DatasetInfosRequest[]}
   */
  _groupDatasetInfosRequests(requests) {
    const infoRequests = [];
    const groupedRequests = groupArrayElementsBy(
      requests,
      ({minDate, maxDate, extent, productId}) => (
        `${minDate},${maxDate},${extent?.join() ?? ''},${productId}`
      )
    );
    /** @type {Readonly<LBRT> | null | undefined} */
    let lastExtent;
    let lastWrappedExtents = this._wrapExtent(lastExtent);
    for (const requestsGroup of groupedRequests.values()) {
      const {minDate, maxDate, extent: _extent} = requestsGroup[0];
      const patternsByServerProductId = groupArrayElementsBy(
        requestsGroup,
        ({productId}) => productId,
        ({patterns}) => patterns.map(this._preparePattern, this),
        null
        // 👆 required, see https://github.com/microsoft/TypeScript/issues/55062
      );

      const productServerIds = [];
      const productPatterns = [];
      for (const [productId, patternGroups] of patternsByServerProductId) {
        const patterns = flattenArray(patternGroups);
        if (patterns.every(Boolean)) {
          // none of the patterns are falsy
          // => use all of them
          for (const pattern of patterns) {
            productServerIds.push(productId);
            productPatterns.push(pattern);
          }
        } else {
          // at least one of the patterns is falsy
          // => ignore all of them
          productServerIds.push(productId);
          productPatterns.push('');
        }
      }

      const products = productServerIds.join(',');
      const patterns = productPatterns.join(',');
      const wrappedExtents = (
        _extent === lastExtent
        ? lastWrappedExtents
        : this._wrapExtent(_extent)
      );

      for (const extent of wrappedExtents) {
        infoRequests.push({
          data: {products, minDate, maxDate, extent, patterns},
          requestsGroup,
        });
      }

      lastExtent = _extent;
      lastWrappedExtents = wrappedExtents;
    }
    return infoRequests;
  }
  /**
   * @protected
   * @param {DatasetInfosCacheMissingRequest[]} requests
   * @param {ReadonlyArray<DatasetServerEvent> | null | undefined} events
   * @returns {DatasetServerInfo[]}
   */
  _serverInfosFromEvents(requests, events) {
    const incompleteResponse = (events?.length === this.datasetsResponseLimit);
    const infos = (events || []).map((event) => /** @type {DatasetServerInfo} */ ({
      productId  : event.productId,
      datasetId  : event.datasetId,
      name       : event.title,
      start      : new Date(normalizeISODateString(event.start)),
      end        : new Date(normalizeISODateString(event.end)),
      mapMaxZoom : parseDecimal(event.mapMaxZoom),
      uri        : event.uri,
      polygon    : event.polygon,
      resolutions: event.resolutions,
      mapBounds  : event.mapBounds,
      metadata   : event.features,
    }));

    if (incompleteResponse) {
      this.api?.syntool.notificationCenter.showMessage({
        id: 'SyntoolAPIServer.incompleteResponseWarning',
        type: 'warning' /* NotificationMessageType.warning */,
        text: 'There may be more datasets to load',
        buttons: [
          {
            text   : 'load more',
            closes : true,
            onClick: () => (
              this.api?.syntool.loadDatasetInfos().catch(emptyFunction)
            ),
          },
        ],
        hideAfter: -1,
      });
    }

    for (const request of requests) {
      this._updateDatasetsCaches(request, infos, incompleteResponse);
    }
    return infos;
  }
  /**
   * @protected
   * @param {Readonly<SyntoolAPIServer.DatasetsDatesResponse>} dates
   * @returns {DatasetDates[]}
   */
  _infosFromDates(dates) {
    return (dates || []).map(([startTimestamp, endTimestamp]) => ({
      start: dateFromTimestamp(startTimestamp),
      end  : dateFromTimestamp(endTimestamp),
    }));
  }
  /**
   * @protected
   * @param {Array<CoverageCacheMissingRequest & {coverage?: ProductCoverage}>} requests
   * @param {SyntoolAPIServer.CoverageResponse} response
   * @returns {CoverageMap}
   */
  _handleCoverageResponse(requests, response) {
    const body = new Uint32Array(response);
    /** @type {CoverageMap} */
    const coverageMap = {};
    let offset;
    let index;
    let request;
    let coverage;

    for (let i = 0, l = body.length; i < l; i++) {
      offset = i++;
      index = $(body[offset]);
      while (i < l && body[i] !== COVERAGE_DELIMITER) {
        i += 13;
      }
      coverage = deserializeCoverage(body.subarray(offset + 1, i));

      request = requests[index];
      if (request) {
        request.coverage = mergeCoverages([request.coverage, coverage]);
        for (const productId of request.productIds) {
          coverageMap[productId] = (
            mergeCoverages([coverageMap[productId], coverage])
          );
        }
      }
    }

    return coverageMap;
  }

  /**
   * @protected
   * @param {Readonly<Dataset>} dataset
   * @param {ServerGranuleInfo} serverGranuleInfo
   * @returns {GranuleInfo}
   */
  _parseGranuleInfo(dataset, serverGranuleInfo) {
    /** @type {GranuleInfo} */
    const granuleInfo = {};
    for (const serverGroup of serverGranuleInfo) {
      const [groupId, group] = (
        isServerGranuleInfoGroupInNewFormat(serverGroup)
        ? parseGranuleInfoGroup(serverGroup)
        : parseOldGranuleInfoGroup(dataset, serverGroup)
      );

      // Update relative resource URLs
      for (const [resourceId, resource] of Object.entries(group)) {
        if (resourceId !== 'metadata') {
          resource.url = encodeURI(
            URL.canParse(resource.url)
            // `url` is a valid absolute URL
            ? resource.url
            // `url` is a relative URL
            : this.getDataURL($(dataset._info).uri + resource.url)
          );
        }
      }
      granuleInfo[groupId] = group;
    }
    return granuleInfo;
  }

  // #endregion
}
