/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {APIServer, SyntoolAPIServer} from './api.js';
import {Dataset} from './dataset.js';
import {Pattern} from './pattern.js';
import {productTag} from './regexp.js';
import {Types as is, assertValueType} from './types.js';
import {HOUR_IN_MILLISECONDS, $, textFromHTML, isObjectEmpty} from './utils.js';

const productTagRe = new RegExp(productTag);

/** @typedef {import('./types.js').TypeChecker} TypeChecker */
/** @typedef {import('./types.js').TypeCheckerPropName} TypeCheckerPropName */
/** @typedef {import('./types.js').TypeCheckerGetProp} TypeCheckerGetProp */
/** @typedef {import('./types.js').TypeCheckerResult} TypeCheckerResult */

/** @type {Record<string, TypeChecker>} */
const productFieldValidators = {};
/** @type {Array<(productConfig: ProductRawConfig) => ProductRawConfig>} */
const productNormalizers = [];
/** @type {TypeChecker[]} */
const configValidators = [];
/** @type {Record<string, TypeChecker>} */
const configFieldValidators = {};
/** @type {Array<(config: RawConfig) => RawConfig>} */
const configNormalizers = [];
/** @type {Array<(productConfig: ServerConfig) => ServerConfig>} */
const serverNormalizers = [];

/** @type {Map<APIServerId, APIServer>} */
let apiServers = new Map();

export {
  configFieldValidators,
  configValidators,
  productFieldValidators,

  configNormalizers,
  serverNormalizers,
  productNormalizers,

  assertValidConfig,
  assertValidServer,
  assertValidProduct,

  normalizeConfig,
  normalizeServer,
  normalizeProduct,
};

const GranuleInfoMetadataFieldConfig = is.shape({
  label: is.string.isRequired,
  description: is.string,
  value: is.string.isRequired,
  order: is.number,
  copyable: is.bool,
});
const GranuleInfoMetadataFieldsConfig = is.objectOf(is.string.isRequired);
const GranuleInfoMetadataConfigFn = is.func;
const GranuleInfoMetadataConfig = is.oneOfType([
  GranuleInfoMetadataFieldsConfig,
  GranuleInfoMetadataConfigFn,
]);
const GranuleInfoResourceType = is.oneOfType([
  is.oneOf(['link', 'image', 'spectrum']),
  is.string,
]);
const GranuleInfoResourceConfig = is.allOfType([
  is.shape({
    type: GranuleInfoResourceType.isRequired,
    label: is.string.isRequired,
    order: is.number,
  }),
  is.match([
    is.case(is.shape({type: is.value('image')}), is.shape({
      inline: is.oneOfType([is.bool, is.URL]),
    })),
    is.any,
  ]),
]);
const GranuleInfoResourcesConfig = (
  is.objectOf(GranuleInfoResourceConfig.isRequired)
);
const GranuleInfoUnknownResourceConfig = is.func;
const GranuleInfoGroupConfig = is.shape({
  title: is.string.isRequired,
  order: is.number,
  defaultCollapsed: is.bool,
  metadata: GranuleInfoMetadataConfig,
  resources: GranuleInfoResourcesConfig,
  unknownResource: GranuleInfoUnknownResourceConfig,
  metadataDefaults: is.GranuleInfoMetadata,
});
const GranuleInfoGroupsConfig = is.objectOf(GranuleInfoGroupConfig.isRequired);
const GranuleInfoUnknownGroupConfig = is.func;
const GranuleInfoConfig = is.shape({
  autoResourceId: is.string,
  groups: GranuleInfoGroupsConfig,
  unknownGroup: GranuleInfoUnknownGroupConfig,
});
const GranuleInfoConfigFn = is.func;
const ProductGranuleInfoConfig = is.oneOfType([
  GranuleInfoConfig,
  GranuleInfoConfigFn,
]);

const Server = is.allOfType([
  is.shape({
    id   : is.string.match(/^[^_]/i).isRequired,
    Class: is.subclassOf(APIServer),
  }),
  is.lazy((props, propName) => {
    const {Class = SyntoolAPIServer} = (
      /** @type {ServerConfig} */ (props[propName])
    );
    return Class.validateConfig;
  }),
]);

const Group = is.shape({
  label   : is.string,
  products: is.arrayOf(is.string.isRequired).isRequired,
});

const ProductType = is.lazy(
  () => is.oneOf(Object.keys(Dataset.CLASS_FOR_TYPE))
);
const ProductOutlines = is.oneOf([
  'ALWAYS_SHOWN',
  'HIDDEN_IN_ZOOM_RANGE',
  'ALWAYS_HIDDEN',
]);
const DatasetFilterId = is.string.match(/^[^%&#+,! ](?:[^%&#+,!]*[^%&#+,! ])?$/);
const DatasetFilter = is.oneOfType([
  is.Pattern,
  is.shape({
    label: is.string,
    aliases: is.arrayOf(DatasetFilterId.isRequired),
    description: is.string,
    pattern: is.Pattern.isRequired,
  }),
]);
const DatasetFiltersGroup = is.allOfType([
  is.shape({
    groupDescription: is.string,
  }).withIndexer(DatasetFilterId, DatasetFilter.isRequired),
  validateProductDatasetFiltersGroup,
]);
const DatasetFiltersGroups = is.allOfType([
  is.objectOf(DatasetFiltersGroup.isRequired),
  validateProductDatasetFiltersGroups,
]);
const BaseProduct = is.allOfType([
  is.shape({
    /* General */
    label            : is.string.isRequired,
    id               : is.string.match(/^[a-z0-9](?:[a-z0-9. _-]*[a-z0-9])?$/i).isRequired,
    type             : ProductType.isRequired,
    description      : is.string,
    infoURL          : is.URL,
    tags             : is.arrayOf(is.string.isRequired),
    selected         : is.bool,
    unit             : is.string,
    granuleInfoConfig: ProductGranuleInfoConfig,
    datasetFilters   : DatasetFiltersGroups,

    /* Timeline */
    mustBeCurrent    : is.bool.isRequired,
    timelineValidity : is.number.gt(0),
    permanent        : is.bool,
    excludedFromColoc: is.bool,

    /* Pagination */
    priority     : is.number,
    datasetWeight: is.number.gte(0),
    minPageSize  : is.number.gte(0),

    /* General Rendering */
    mapMinZoom: is.integer.gte(0),
    mapMaxZoom: is.integer.gte(0),
    opacity   : is.number.gte(0).lte(1),
    stackLevel: is.integer.gte(0).lte(1000).isRequired,
  }),
  is.oneOfType([
    /* General Rendering */

    // Replacement
    is.shape({outlines: ProductOutlines}),

    // Deprecated
    is.shape({noOutline: is.bool}),
  ]),
  is.lazy(() => is.shape(productFieldValidators)),
  /* Type Specific */
  is.lazy(() => {
    const datasetClasses = new Set(Object.values(Dataset.CLASS_FOR_TYPE));

    return is.match(Array.from(datasetClasses, (datasetClass) => {
      const productTypes = datasetClass.productTypes;
      return is.case(
        is.shape({
          type: (
            productTypes.length === 1
            ? is.value(productTypes[0])
            : is.oneOf(productTypes)
          ),
        }),
        datasetClass.validateConfig || is.any
      );
    }));
  }),
]);
const Product = is.allOfType([
  BaseProduct,
  validateProductHandledByOneServer,
  validateProductCanUseDatasetFilters,
]);

const AdditionalLayerInstance = is.allOfType([
  is.Layer,
  is.shape({syntoolIsOverlay: is.bool}),
]);
const TimeRange = is.tuple([is.string.isRequired, is.Timespan.isRequired, is.bool]);
const UniqueProduct = is.allOfType([
  BaseProduct,
  validateProductHandledByOneServerInConfig,
  validateProductCanUseDatasetFiltersInConfig,
  is.uniquePropInArray('id'),
]);
const UniqueServer = is.allOfType([
  Server,
  is.uniquePropInArray('id'),
]);

const ConfigVersion = is.allOfType([is.string, validateVersion('1.3.0')]);
const Config = is.allOfType([
  is.shape({
    version: ConfigVersion.isRequired,

    /* Services */
    shortenerForURL: is.func,

    /* Map */
    projection      : is.string.isRequired,
    maxExtent       : is.Bounds.isRequired,
    restrictedExtent: is.Bounds,
    zoomOffset      : is.integer.gte(0),
    defaultZoom     : is.integer.gte(0),
    baseLayers      : is.arrayOf(is.Layer.isRequired, [1]).isRequired,
    additionalLayers: is.arrayOf(AdditionalLayerInstance.isRequired),

    /* Timeline */
    colocActive   : is.bool,
    currentDate   : is.date,
    timeRanges    : is.arrayOf(TimeRange.isRequired, [1]),
    yearsRange    : is.arrayOf(is.integer.gt(1969).isRequired, 2),

    /* Misc */
    storageId        : is.string.isRequired,
    maxDatasetsWeight: is.number.gte(0),

    /* Predefined */
    products: is.arrayOf(UniqueProduct.isRequired),
  }),
  is.oneOfType([
    // Replacement
    is.shape({
      servers: is.arrayOf(UniqueServer.isRequired, [1]).isRequired,
    }),
    // Deprecated
    is.shape({
      serviceHost: is.URL.isRequired,
      dataServers: is.arrayOf(is.URL, [1]).isRequired,
    }),
  ]),
  // Deprecated
  is.shape({
    showGroups: is.bool,
    groups    : is.allOfType([
      is.arrayOf(Group.isRequired),
      validateGroupsProducts,
    ]),
  }),
  is.lazy(() => is.shape(configFieldValidators)),
  is.lazy(() => is.allOfType(configValidators)),
]);

const _types = {
  GranuleInfoMetadataFieldConfig,
  GranuleInfoMetadataFieldsConfig,
  GranuleInfoMetadataConfigFn,
  GranuleInfoMetadataConfig,
  GranuleInfoResourceType,
  GranuleInfoResourceConfig,
  GranuleInfoResourcesConfig,
  GranuleInfoUnknownResourceConfig,
  GranuleInfoGroupConfig,
  GranuleInfoGroupsConfig,
  GranuleInfoUnknownGroupConfig,
  GranuleInfoConfig,
  GranuleInfoConfigFn,
  ProductGranuleInfoConfig,

  Server,
  Group,
  ProductType,
  Product,

  Config,
};
/** @type {typeof _types | Record<string, TypeChecker>} */
const types = _types;
export {types as configTypes};

configNormalizers.push(normalizeConfigWithDefaults);
configNormalizers.push(normalizeConfigGroupsToTags);
configNormalizers.push(normalizeConfigProducts);
configNormalizers.push(normalizeConfigServiceHostToServers);
configNormalizers.push(normalizeConfigServers);

serverNormalizers.push(normalizeServerWithDefaults);

productNormalizers.push(normalizeProductWithDefaults);
productNormalizers.push(normalizeProductForDatasetClass);
productNormalizers.push(normalizeProductOutlines);
productNormalizers.push(normalizeProductDatasetFilters);

/**
 * @param {string} version
 * @returns {TypeChecker}
 */
function validateVersion(version) {
  const versionParts = version.split('.');

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    const propValue = props[propName];
    const propValueParts = propValue.split('.');

    if (
      propValueParts[0] === versionParts[0] &&
      propValueParts[1] === versionParts[1]
    ) {
      return null;
    }

    return [
      'Invalid `', propFullName, '` of value `', propValue, '`, ',
      'expected to be a compatible version to `', version, '`.'
    ].join('');
  }
  return validate;
}
/**
 * @param {any} groups
 * @param {TypeCheckerPropName} groupLabel
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateProductDatasetFiltersGroup(groups, groupLabel, propFullName) {
  /** @type {DatasetFilterGroupRaw} */
  const group = groups[groupLabel];
  const errors = [];
  let count = 0;
  for (const [key, value] of Object.entries(group)) {
    if (key === 'groupDescription') {
      continue;
    }
    count++;

    const pattern = typeof value === 'string' ? value : $(value).pattern;
    const normalizedPattern = Pattern.normalize(pattern);
    if (normalizedPattern === '' || normalizedPattern === '*') {
      errors.push([
        `Dataset filter \`${propFullName}.${key}\` expected to match something `,
        `but not everything, but found \`${pattern}\` that matches ${
          normalizedPattern === '' ? 'nothing' : 'everything'
        }.`
      ].join(''));
    }
  }
  if (count < 2) {
    errors.push([
      'Invalid `', propFullName, '` expected at least 2 filters, but found ',
      count === 0 ? 'none.' : `only ${count}.`
    ].join(''));
  }
  return errors;
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateProductDatasetFiltersGroups(props, propName, propFullName) {
  /** @type {DatasetFilterGroupsRaw} */
  const groups = props[propName];
  /** @type {Map<string, string>} */
  const groupLabelByFilterId = new Map();
  /** @type {Map<string, {key: string, pattern: Pattern}>} */
  const keyByNormalizedPattern = new Map();
  const errors = [];
  for (const [label, filters] of Object.entries(groups)) {
    for (const [key, value] of Object.entries(filters)) {
      if (key === 'groupDescription') {
        continue;
      }

      const sameIdGroup = groupLabelByFilterId.get(key);
      if (sameIdGroup) {
        errors.push([
          `Duplicate use of dataset filter id "${key}" in `,
          `\`${propFullName}.${label}.${key}\`, first defined in group \``,
          sameIdGroup, '`.'
        ].join(''));
      } else {
        groupLabelByFilterId.set(key, label);
      }

      const pattern = typeof value === 'string' ? value : $(value).pattern;
      const normalizedPattern = Pattern.normalize(pattern);
      const similar = keyByNormalizedPattern.get(normalizedPattern);
      if (similar) {
        const {key: similarKey, pattern: similarPattern} = similar;
        errors.push([
          `Dataset filter \`${propFullName}.${label}.${key}\` expected `,
          `to be unique, but found it to be equivalent to \`${similarKey}\` (`,
          `"${pattern}" is equivalent to "${similarPattern}").`
        ].join(''));
      } else {
        keyByNormalizedPattern.set(
          normalizedPattern,
          {key: `${label}.${key}`, pattern}
        );
      }
    }
  }
  return errors;
}
/**
 * @param {ServerConfig} config
 * @param {ProductRawConfig} product
 * @returns {boolean}
 */
function serverWithConfigHandlesProduct(config, product) {
  const ServerClass = config.Class || SyntoolAPIServer;
  return ServerClass.instanceWithConfigHandlesProduct(config, product);
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @param {TypeCheckerGetProp} getProp
 * @returns {TypeCheckerResult}
 */
function validateProductHandledByOneServerInConfig(props, propName, propFullName, getProp) {
  const configServers = getProp('.servers');
  if (!configServers) {
    // Old format, no validation needed
    return null;
  }

  /** @type {ProductRawConfig} */
  const product = props[propName];
  const handledByServers = [];
  let handledInternally = false;

  for (const server of apiServers.values()) {
    if (server.handlesProduct(product)) {
      if (server.id.startsWith('_')) {
        handledInternally = true;
      } else {
        handledByServers.push(server.id);
      }
      break;
    }
  }
  for (let i = 0, l = configServers.length; i < l; i++) {
    if (serverWithConfigHandlesProduct(configServers[i], product)) {
      handledByServers.push(configServers[i].id);
    }
  }

  if (handledInternally) {
    if (handledByServers.length === 0) {
      return null;
    }
    return [
      'Product `', propFullName, '` (with id "', product.id, '") is handled ',
      'internally. This product must not be handled by any server, remove ',
      'it from the following servers: ', handledByServers.join(', ')
    ].join('');
  }
  if (handledByServers.length === 0) {
    return [
      'Product `', propFullName, '` (with id "', product.id, '") is not ',
      'handled by any server.'
    ].join('');
  }
  return null;
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateProductHandledByOneServer(props, propName, propFullName) {
  /** @type {ProductRawConfig} */
  const product = props[propName];
  const handledByServers = [];
  let handledInternally = false;

  for (const server of apiServers.values()) {
    if (server.handlesProduct(product)) {
      if (server.id.startsWith('_')) {
        handledInternally = true;
      } else {
        handledByServers.push(server.id);
      }
      break;
    }
  }

  if (handledInternally) {
    if (handledByServers.length === 0) {
      return null;
    }
    return [
      'Product `', propFullName, '` (with id "', product.id, '") is handled ',
      'internally. This product must not be handled by any server, remove ',
      'it from the following servers: ', handledByServers.join(', ')
    ].join('');
  }
  if (handledByServers.length === 0) {
    return [
      'Product `', propFullName, '` (with id "', product.id, '") is not ',
      'handled by any server.'
    ].join('');
  }
  return null;
}
/**
 * @param {ServerConfig} config
 * @returns {boolean}
 */
function serverWithConfigSupportsComplexPatterns(config) {
  var ServerClass = config.Class || SyntoolAPIServer;
  return ServerClass.instanceWithConfigSupportsComplexPatterns(config);
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @param {TypeCheckerGetProp} getProp
 * @returns {TypeCheckerResult}
 */
function validateProductCanUseDatasetFiltersInConfig(props, propName, propFullName, getProp) {
  /** @type {ProductRawConfig} */
  const product = props[propName];
  if (isObjectEmpty(product.datasetFilters)) {
    // Dataset filters are not used
    return null;
  }

  const serversNotSupportingComplexPatterns = [];
  for (const server of apiServers.values()) {
    if (server.handlesProduct(product) && !server.supportsComplexPatterns) {
      serversNotSupportingComplexPatterns.push(server.id);
    }
  }

  const configServers = getProp('.servers');
  if (configServers) {
    for (const configServer of configServers) {
      if (
        serverWithConfigHandlesProduct(configServer, product) &&
        !serverWithConfigSupportsComplexPatterns(configServer)
      ) {
        serversNotSupportingComplexPatterns.push(configServer.id);
      }
    }
  } else {
    serversNotSupportingComplexPatterns.push('<default>');
  }

  if (serversNotSupportingComplexPatterns.length > 0) {
    return [
      'Product `', propFullName, '` (with id "', product.id, '") cannot use ',
      "`datasetFilters` field because it's handled by the following servers that ",
      "don't support complex patterns: ",
      serversNotSupportingComplexPatterns.join(', '),
    ].join('');
  }
  return null;
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateProductCanUseDatasetFilters(props, propName, propFullName) {
  /** @type {ProductRawConfig} */
  const product = props[propName];
  if (isObjectEmpty(product.datasetFilters)) {
    // Dataset filters are not used
    return null;
  }

  const serversNotSupportingComplexPatterns = [];
  for (const server of apiServers.values()) {
    if (server.handlesProduct(product) && !server.supportsComplexPatterns) {
      serversNotSupportingComplexPatterns.push(server.id);
    }
  }

  if (serversNotSupportingComplexPatterns.length > 0) {
    return [
      'Product `', propFullName, '` (with id "', product.id, '") cannot use ',
      "`datasetFilters` field because it's handled by the following servers that ",
      "don't support complex patterns: ",
      serversNotSupportingComplexPatterns.join(', '),
    ].join('');
  }
  return null;
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @param {TypeCheckerGetProp} getProp
 * @returns {TypeCheckerResult}
 */
function validateGroupsProducts(props, propName, propFullName, getProp) {
  propFullName = propFullName.substring(0, propFullName.lastIndexOf('.'));

  const groups = getProp('.groups');
  const products = getProp('.products');
  const productIds = Object.create(null);
  let group;
  let productId;
  let fullName;

  if (!groups) {
    return null;
  }

  for (let i = 0, l = products.length; i < l; i++) {
    productIds[products[i].id] = false;
  }

  const errors = [];
  for (let i = 0, l = groups.length; i < l; i++) {
    group = groups[i];

    for (let j = 0, k = group.products.length; j < k; j++) {
      productId = group.products[j];

      if (!(productId in productIds)) {
        fullName = propFullName + '.groups[' + i + '].products[' + j + ']';

        errors.push([
          'Invalid `', fullName, '` ',
          (
            group.label
            ? '(group with label "' + group.label + '") '
            : '(group without label) '
          ),
          'of value `', productId, '`, ',
          'expected to be the `id` of one of the products defined in ',
          '`', propFullName, '.products`.'
        ].join(''));
      }

      if (productIds[productId]) {
        fullName = propFullName + '.groups[' + i + '].products[' + j + ']';

        errors.push([
          'Invalid `', fullName, '` ',
          (
            group.label
            ? '(group with label "' + group.label + '") '
            : '(group without label) '
          ),
          'of value `', productId, '`, ',
          'expected to be unique in `', propFullName, '.groups', '`.'
        ].join(''));
      }

      productIds[productId] = true;
    }
  }

  for (let i = 0, l = products.length; i < l; i++) {
    if (!productIds[products[i].id]) {
      errors.push(
        'Product with id "' + products[i].id + '" not found in any group.'
      );
    }
  }

  return errors;
}

/**
 * @template {{}} T
 * @param {Array<(object: T) => T>} normalizers
 * @param {T} object
 * @returns {T}
 */
function applyNormalizers(normalizers, object) {
  return normalizers.reduce(
    (obj, normalizer) => normalizer(obj),
    Object.assign({}, object)
  );
}

/**
 * @param {ProductRawConfig} product
 * @returns {ProductRawConfig}
 */
function normalizeProductWithDefaults(product) {
  return Object.assign({
    /* General */
    description: null,
    tags       : [],
    selected   : false,
    unit       : null,

    /* Timeline */
    timelineValidity : 0,
    permanent        : false,
    excludedFromColoc: false,

    /* Pagination */
    priority     : 0,
    datasetWeight: 1,
    minPageSize  : 1,

    /* General Rendering */
    mapMinZoom: 0,
    mapMaxZoom: Infinity,
    opacity   : 1,
  }, product);
}
/**
 * @param {RawConfig} config
 * @returns {RawConfig}
 */
function normalizeConfigWithDefaults(config) {
  const currentYear = new Date().getUTCFullYear();
  return Object.assign({
    /* Map */
    restrictedExtent: config.maxExtent,
    zoomOffset      : 0,
    additionalLayers: [],

    /* Timeline */
    colocActive   : true,
    timeRanges    : [
      ['6-Hour', '6h;1d'],
      ['Daily', '1d;1w', true],
      ['3-Day', '3d;1w'],
      ['Weekly', '1w;1m'],
      ['Bi-weekly', '2w;1m'],
    ],
    yearsRange: [currentYear - 15, currentYear + 1],

    /* Misc */
    maxDatasetsWeight: 100,

    /* Predefine */
    products: [],
    groups  : [],
  }, config);
}
/**
 * @param {ServerConfig} server
 * @returns {ServerConfig}
 */
function normalizeServerWithDefaults(server) {
  return Object.assign({
    Class: SyntoolAPIServer,
  }, server);
}

/**
 * @param {RawConfig} config
 * @returns {RawConfig}
 */
function normalizeConfigServiceHostToServers(config) {
  if (!config.servers && config.serviceHost && config.dataServers) {
    config.servers = [
      /** @type {ServerConfig} */ ({
        id          : 'default',
        Class       : SyntoolAPIServer,
        serviceURL  : config.serviceHost,
        dataURLs    : config.dataServers,
        productIdMap: Object.fromEntries(
          config.products?.map((p) => [p.id, p.id]) ?? []
        ),
      })
    ];
  }
  return config;
}
/**
 * @param {RawConfig} config
 * @returns {RawConfig}
 */
function normalizeConfigServers(config) {
  config.servers = /** @type {Config['servers']} */ (
    config.servers?.map(normalizeServer, config)
  );
  return config;
}
/**
 * @param {ServerConfig} server
 * @returns {ServerConfig}
 */
function normalizeServer(server) {
  return applyNormalizers(serverNormalizers, server);
}
/**
 * @param {RawConfig} config
 * @returns {RawConfig}
 */
function normalizeConfigGroupsToTags(config) {
  const groups = config.groups;
  if (!groups) {
    return config;
  }

  const groupBy = 'Group';
  const products = config.products;

  for (const group of groups) {
    for (const productId of group.products) {
      const product = $(products.find((p) => p.id === productId));
      const hasGroupTag = product.tags?.some((tag) => {
        const parts = productTagRe.exec(tag);
        return parts?.[1] === groupBy;
      });

      if (!hasGroupTag && group.label) {
        const tag = (
          groupBy + ':' + textFromHTML(group.label.replace('<br>', '\n'))
        );
        if (product.tags) {
          product.tags.push(tag);
        } else {
          product.tags = [tag];
        }
      }
    }
  }
  return config;
}
/**
 * @param {RawConfig} config
 * @returns {RawConfig}
 */
function normalizeConfigProducts(config) {
  config.products = config.products.map(normalizeProduct, config);
  return config;
}
/**
 * @param {ProductRawConfig} product
 * @returns {ProductConfig}
 */
function normalizeProduct(product) {
  return /** @type {ProductConfig} */ (
    applyNormalizers(productNormalizers, product)
  );
}
/**
 * @param {ProductRawConfig} product
 * @returns {ProductRawConfig}
 */
function normalizeProductOutlines(product) {
  if (!('outlines' in product)) {
    product.outlines = (
      product.noOutline ? 'HIDDEN_IN_ZOOM_RANGE' : 'ALWAYS_SHOWN'
    );
  }
  return product;
}
/**
 * @param {ProductRawConfig} productConfig
 * @returns {ProductRawConfig}
 */
function normalizeProductDatasetFilters(productConfig) {
  const product = /** @type {Product} */ (/** @type {any} */ (productConfig));
  /** @type {DatasetFiltersGroup[]} */
  const groups = [];
  if (productConfig.datasetFilters) {
    for (
      const [groupLabel, group] of Object.entries(productConfig.datasetFilters)
    ) {
      const filters = [];
      for (const [id, value] of Object.entries(group)) {
        if (id === 'groupDescription') {
          continue;
        }
        let label;
        let aliases;
        let description;
        let pattern;
        if (typeof value === 'string') {
          pattern = value;
        } else {
          label = $(value).label;
          aliases = $(value).aliases;
          description = $(value).description;
          pattern = $(value).pattern;
        }
        label ??= id;
        filters.push({id, label, aliases, description, pattern});
      }
      groups.push({
        label: groupLabel,
        description: group.groupDescription,
        filters,
      });
    }
  }
  product.datasetFilters = groups;
  return productConfig;
}
/**
 * @param {ProductRawConfig} productConfig
 * @returns {ProductRawConfig}
 */
function normalizeProductForDatasetClass(productConfig) {
  const Class = Dataset.CLASS_FOR_TYPE[productConfig.type];
  return Class?.normalizeProductConfig(productConfig) ?? productConfig;
}

/**
 * @param {RawConfig} config
 * @returns {Config}
 */
function normalizeConfig(config) {
  return /** @type {Config} */ (applyNormalizers(configNormalizers, config));
}
/**
 * @param {RawConfig} config
 * @param {Map<APIServerId, APIServer>} servers
 * @returns {void}
 * @throws {TypeCheckerError}
 */
function assertValidConfig(config, servers) {
  apiServers = servers;
  assertValueType(types.Config, 'Config', config);
}
/**
 * @param {ServerConfig} serverConfig
 * @returns {void}
 * @throws {TypeCheckerError}
 */
function assertValidServer(serverConfig) {
  assertValueType(types.Server, 'Server', serverConfig);
}
/**
 * @param {ProductRawConfig} productConfig
 * @returns {void}
 * @throws {TypeCheckerError}
 */
function assertValidProduct(productConfig) {
  assertValueType(types.Product, 'Product', productConfig);
}
