/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  $,
  emptyFunctionReturningFalse,
  emptyFunctionReturningNull,
  emptyFunctionReturning,
  parseDecimal,
  numericComparator,
  reverseComparator,
  isObjectEmpty,
  filterObject,
  filterObjectToValuesArray,
} from './utils.js';

/**
 * Each bit represents a day's coverage. LSB is the 1st day of the month and MSB
 * is the 32nd day of the month.
 *
 * @typedef {Uint32} MonthCoverage
 */
/**
 * An array of 12 elements. Each representing the a month's coverage.
 * The 1st element (index `0`) corresponds the the 1st month of the year and
 * the 12th element (index `11`) corresponds the the 12th month of the year.
 *
 * @typedef {(
 *   | [
 *      MonthCoverage, MonthCoverage, MonthCoverage, MonthCoverage,
 *      MonthCoverage, MonthCoverage, MonthCoverage, MonthCoverage,
 *      MonthCoverage, MonthCoverage, MonthCoverage, MonthCoverage,
 *   ]
 *   | Uint32Array
 * )} YearCoverage
 */
/** @typedef {{[year: number]: YearCoverage}} ProductCoverage */
/** @typedef {{[productId: ProductId]: ProductCoverage}} CoverageMap */

/**
 * @param {Readonly<YearCoverage>} yearCoverage
 * @returns {boolean}
 */
function yearCoverageIsEmpty(yearCoverage) {
  return !yearCoverage.some(Boolean);
}
/**
 * @param {Readonly<ProductCoverage>} coverage
 * @returns {boolean}
 */
function coverageIsEmpty(coverage) {
  const yearCoverages = Object.values(coverage);
  return yearCoverages.every(yearCoverageIsEmpty);
}
/**
 * @param {Readonly<YearCoverage>} yearCoverage
 * @returns {YearCoverage}
 */
function cloneYearCoverage(yearCoverage) {
  return new Uint32Array(yearCoverage);
}
/**
 * @param {Readonly<ProductCoverage>} coverage
 * @returns {ProductCoverage}
 */
function cloneCoverage(coverage) {
  /** @type {ProductCoverage} */
  const clone = {};
  for (const year of Object.keys(coverage)) {
    const y = /** @type {number} */ (/** @type {unknown} */ (year));
    const yearCoverage = coverage[y];
    if (yearCoverage) {
      clone[y] = cloneYearCoverage(yearCoverage);
    }
  }
  return clone;
}
/**
 * @param {YearCoverage | undefined} union
 * @param {Readonly<YearCoverage>} yearCoverage
 * @returns {YearCoverage}
 */
function addYearCoverageToUnion(union, yearCoverage) {
  if (!union) {
    return cloneYearCoverage(yearCoverage);
  }
  for (let month = /** @type {_Month} */ (0); month < 12; month++) {
    //eslint-disable-next-line no-bitwise
    union[month] |= $(yearCoverage[month]);
  }
  return union;
}
/**
 * @param {Iterable<ProductCoverage | null | undefined>} coverages
 * @returns {ProductCoverage}
 */
export function mergeCoverages(coverages) {
  let shouldCloneCoverage = true;
  let union;
  for (const coverage of coverages) {
    if (coverage) {
      if (union) {
        if (shouldCloneCoverage) {
          union = cloneCoverage(union);
          shouldCloneCoverage = false;
        }
        for (const year of Object.keys(coverage)) {
          const y = /** @type {number} */ (/** @type {unknown} */ (year));
          const yearCoverage = coverage[y];
          if (yearCoverage) {
            union[y] = addYearCoverageToUnion(union[y], yearCoverage);
          }
        }
      } else {
        union = coverage;
      }
    }
  }
  return union || {};
}
/**
 * @param {YearCoverage | undefined} yearCoverage1
 * @param {Readonly<YearCoverage> | undefined} yearCoverage2
 * @returns {YearCoverage | null}
 */
function intersectYearCoverages(yearCoverage1, yearCoverage2) {
  if (!yearCoverage2) {
    return null;
  }
  if (!yearCoverage1) {
    return cloneYearCoverage(yearCoverage2);
  }
  let empty = true;
  for (let month = /** @type {_Month} */ (0); month < 12; month++) {
    //eslint-disable-next-line no-bitwise
    yearCoverage1[month] &= $(yearCoverage2[month]);
    empty &&= (yearCoverage1[month] === 0);
  }
  return empty ? null : yearCoverage1;
}
/**
 * @param {Iterable<ProductCoverage | null | undefined>} coverages
 * @returns {ProductCoverage}
 */
function getIntersectionOfCoverages(coverages) {
  let shouldCloneCoverage = true;
  let intersection;
  for (const coverage of coverages) {
    if (coverage) {
      if (intersection) {
        if (shouldCloneCoverage) {
          intersection = cloneCoverage(intersection);
          shouldCloneCoverage = false;
        }

        for (const year of Object.keys(intersection)) {
          const y = /** @type {number} */ (/** @type {unknown} */ (year));
          const yearIntersection = (
            intersectYearCoverages(intersection[y], coverage[y])
          );
          if (yearIntersection) {
            intersection[y] = yearIntersection;
          } else {
            delete intersection[y];
          }
        }
      } else {
        intersection = coverage;
      }
    }
  }
  return intersection || {};
}
/**
 * @param {YearCoverage} yearCoverage
 * @param {Date} currentDate
 * @param {FindNearestOptions} options
 * @returns {Date | null}
 */
function findNearestDayInYearCoverage(yearCoverage, currentDate, options) {
  if (!options.month && !options.day) {
    return null;
  }

  const inPast = !!options.past;
  const inFuture = !!options.future;
  const currentTime = currentDate.getTime();
  let currentMonth = currentDate.getUTCMonth();
  let currentDay = currentDate.getUTCDate() - 1;

  const nearestDate = new Date(currentTime);
  var nearestMonth;
  var nearestDay;

  let monthCoverage = $(yearCoverage[currentMonth]);
  if (monthCoverage !== 0 && options.day) {
    nearestDay = (
      findNearestDayInMonthCoverage(monthCoverage, currentDay, inPast, inFuture)
    );
    if (nearestDay !== null) {
      nearestDate.setUTCDate(nearestDay + 1);
      return nearestDate;
    }
  }

  if (!options.month) {
    return null;
  }

  for (;;) {
    nearestMonth = (
      findNearestMonthInYearCoverage(yearCoverage, currentMonth, inPast, inFuture)
    );
    if (nearestMonth === null) {
      return null;
    }

    if (nearestMonth < currentMonth) {
      // set `currentDay` to last day of `nearestMonth`
      nearestDate.setUTCDate(1);
      nearestDate.setUTCMonth(nearestMonth + 1);
      nearestDate.setUTCDate(0);
    } else {
      nearestDate.setUTCDate(1);
      nearestDate.setUTCMonth(nearestMonth);
    }
    currentDay = nearestDate.getUTCDate() - 1;
    currentMonth = nearestMonth;

    monthCoverage = $(yearCoverage[currentMonth]);
    //eslint-disable-next-line no-bitwise
    if ((monthCoverage & (1 << currentDay)) !== 0) {
      return nearestDate;
    }

    if (options.day) {
      // `nearestMonth` is in the correct time interval (past or future)
      // so any day in it will also be in the correct time interval
      // (past or future)
      nearestDay = (
        findNearestDayInMonthCoverage(monthCoverage, currentDay, true, true)
      );

      if (nearestDay !== null) {
        nearestDate.setUTCDate(nearestDay + 1);
        return nearestDate;
      }
    }
  }
}
/**
 * @param {ProductCoverage} coverage
 * @param {number} currentYear
 * @param {boolean} inPast
 * @param {boolean} inFuture
 * @returns {number | null}
 */
function findNearestYearInCoverage(coverage, currentYear, inPast, inFuture) {
  const years = Object.keys(coverage).map(parseDecimal);
  const minYear = Math.min(...years);
  const maxYear = Math.max(...years);
  const maxOffset = (
    Math.max(maxYear, currentYear) - Math.min(minYear, currentYear) + 1
  );

  for (let offset = 1; offset < maxOffset; offset++) {
    if (inFuture && coverage[currentYear + offset]?.some(Boolean)) {
      return currentYear + offset;
    }

    if (inPast && coverage[currentYear - offset]?.some(Boolean)) {
      return currentYear - offset;
    }
  }

  return null;
}
/**
 * @param {YearCoverage} yearCoverage
 * @param {number} currentMonth
 * @param {boolean} inPast
 * @param {boolean} inFuture
 * @returns {number | null}
 */
function findNearestMonthInYearCoverage(yearCoverage, currentMonth, inPast, inFuture) {
  for (let offset = 1; offset < 12; offset++) {
    if (
      inFuture &&
      currentMonth + offset < 12 &&
      yearCoverage[currentMonth + offset] !== 0
    ) {
      return currentMonth + offset;
    }

    if (
      inPast &&
      currentMonth - offset >= 0 &&
      yearCoverage[currentMonth - offset] !== 0
    ) {
      return currentMonth - offset;
    }
  }

  return null;
}
/**
 * @param {MonthCoverage} monthCoverage
 * @param {number} currentDay
 * @param {boolean} inPast
 * @param {boolean} inFuture
 * @returns {number | null}
 */
function findNearestDayInMonthCoverage(monthCoverage, currentDay, inPast, inFuture) {
  for (let offset = 1; offset < 31; offset++) {
    if (
      inFuture &&
      currentDay + offset < 31 &&
      //eslint-disable-next-line no-bitwise
      (monthCoverage & (1 << (currentDay + offset))) !== 0
    ) {
      return currentDay + offset;
    }

    if (
      inPast &&
      currentDay - offset >= 0 &&
      //eslint-disable-next-line no-bitwise
      (monthCoverage & (1 << (currentDay - offset))) !== 0
    ) {
      return currentDay - offset;
    }
  }

  return null;
}

export class Coverage {
  /**
   * @typedef FindNearestOptions
   * @prop {ProductId[]} [productIds]
   * @prop {boolean} [coloc]
   * @prop {boolean} [year]
   * @prop {boolean} [month]
   * @prop {boolean} [day]
   * @prop {boolean} [past]
   * @prop {boolean} [future]
   */
  /** @typedef {0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 10 | 11} _Month */

  /**
   * @param {CoverageMap} [coverageByProduct]
   */
  constructor(coverageByProduct) {
    this.coverageByProduct = coverageByProduct || {};
    this.isEmpty = isObjectEmpty(this.coverageByProduct);
    /** @type {ProductId[] | null} */
    this._productIds = null;
    /** @type {ProductCoverage[] | null} */
    this._array = null;
    /** @type {ProductCoverage | null} */
    this._union = null;
    /** @type {ProductCoverage | null} */
    this._intersection = null;
  }

  /**
   * @param {ProductId[]} productIds
   * @returns {Coverage}
   */
  getSubCoverageForProductIds(productIds) {
    if (!productIds) {
      return this;
    }

    const coverageByProduct = /** @type {CoverageMap} */ (filterObject(
      this.coverageByProduct,
      (coverage, productId) => productIds.indexOf(productId) !== -1
    ));
    if (Object.keys(coverageByProduct).length === this.getProductsCount()) {
      return this;
    }
    return new Coverage(coverageByProduct);
  }

  /**
   * @param {Coverage | null | undefined} coverage
   * @returns {boolean}
   */
  equals(coverage) {
    if (!coverage) {
      return this.isEmpty;
    }
    return (
      this === coverage ||
      this.coverageByProduct === coverage.coverageByProduct ||
      (this.isEmpty && coverage.isEmpty)
    );
  }
  /**
   * @returns {number}
   */
  getProductsCount() {
    return this.getProductIds().length;
  }
  /**
   * @returns {ProductId[]}
   */
  getProductIds() {
    if (!this._productIds) {
      this._productIds = Object.keys(this.coverageByProduct);
    }
    return this._productIds;
  }
  /**
   * @param {ProductId[]} [productIds]
   * @returns {ProductCoverage[]}
   */
  getArray(productIds) {
    if (this.isEmpty) {
      this._array = this._array || [];
      return this._array;
    }

    if (productIds) {
      if (productIds.length === 0) {
        return [];
      }
      if (productIds.length !== this.getProductsCount()) {
        return filterObjectToValuesArray(
          this.coverageByProduct,
          (coverage, productId) => productIds.indexOf(productId) !== -1
        );
      }
    }
    if (!this._array) {
      this._array = Object.values(this.coverageByProduct);
    }
    return this._array;
  }
  /**
   * @param {ProductId[]} [productIds]
   * @returns {ProductCoverage}
   */
  getUnion(productIds) {
    if (this.isEmpty) {
      this._union = this._union || {};
      return this._union;
    }

    const coverages = this.getArray(productIds);

    if (this._union && coverages.length === this.getProductsCount()) {
      return this._union;
    }

    const union = mergeCoverages(coverages);

    if (!this._union && coverages.length === this.getProductsCount()) {
      this._union = union;
    }
    return union;
  }
  /**
   * @param {ProductId[]} [productIds]
   * @returns {ProductCoverage}
   */
  getIntersection(productIds) {
    if (this.isEmpty) {
      this._intersection = this._intersection || {};
      return this._intersection;
    }

    const coverages = this.getArray(productIds);

    if (this._intersection && coverages.length === this.getProductsCount()) {
      return this._intersection;
    }

    const intersection = getIntersectionOfCoverages(coverages);

    if (!this._intersection && coverages.length === this.getProductsCount()) {
      this._intersection = intersection;
    }
    return intersection;
  }
  /**
   * @param {Date} currentDate
   * @param {boolean} coloc
   * @param {ProductId[]} productIds
   * @returns {boolean}
   */
  hasDataAtDate(currentDate, coloc, productIds) {
    if (this.isEmpty) {
      return false;
    }

    const currentYear = currentDate.getUTCFullYear();
    const currentMonth = currentDate.getUTCMonth();
    const currentDay = currentDate.getUTCDate() - 1;
    const coverage = (
      coloc
      ? this.getIntersection(productIds)
      : this.getUnion(productIds)
    );
    let yearCoverage;
    let monthCoverage;

    return !!(
      (yearCoverage = coverage[currentYear]) &&
      (monthCoverage = yearCoverage[currentMonth]) &&
      //eslint-disable-next-line no-bitwise
      (monthCoverage & (1 << currentDay))
    );
  }
  /**
   * @param {ProductId[]} productIds
   * @returns {boolean}
   */
  hasColoc(productIds) {
    const coverages = this.getArray(productIds);
    const length = coverages.length;

    if (length < 2) {
      return false;
    }

    return !coverageIsEmpty(getIntersectionOfCoverages(coverages));
  }
  /**
   * @param {Date} currentDate
   * @param {FindNearestOptions} options
   * @returns {Date | null}
   */
  findNearest(currentDate, options) {
    if (
      this.isEmpty ||
      (!options.year && !options.month && !options.day) ||
      (!options.past && !options.future)
    ) {
      return null;
    }

    const inPast = !!options.past;
    const inFuture = !!options.future;
    const currentTime = currentDate.getTime();
    let currentYear = currentDate.getUTCFullYear();
    let currentMonth = currentDate.getUTCMonth();
    let currentDay = currentDate.getUTCDate() - 1;
    let nearestDate;
    let nearestYear;
    let nearestDayDate;

    const coverage = (
      options.coloc
      ? this.getIntersection(options.productIds)
      : this.getUnion(options.productIds)
    );
    let yearCoverage = coverage[currentYear];

    if (yearCoverage && yearCoverage.some(Boolean)) {
      nearestDayDate = (
        findNearestDayInYearCoverage(yearCoverage, currentDate, options)
      );

      if (nearestDayDate !== null) {
        return nearestDayDate;
      }
    }

    if (!options.year) {
      return null;
    }

    // `nearestYear` will be in the correct time interval (past or future)
    // so any month/day in it will also be in the correct time interval
    // (past or future)
    options = Object.assign({}, options, {past: true, future: true});

    for (;;) {
      nearestYear = (
        findNearestYearInCoverage(coverage, currentYear, inPast, inFuture)
      );
      if (nearestYear === null) {
        return null;
      }

      if (nearestYear < currentYear) {
        // set `currentMonth` to last month of `nearestYear`
        // set `currentDay` to last day of `nearestYear`
        currentMonth = 11;
        currentDay = 30;
      } else {
        currentMonth = 0;
        currentDay = 0;
      }
      currentYear = nearestYear;
      nearestDate = new Date(currentTime);
      nearestDate.setUTCDate(1);
      nearestDate.setUTCMonth(currentMonth);
      nearestDate.setUTCDate(currentDay + 1);
      nearestDate.setUTCFullYear(currentYear);

      yearCoverage = $(coverage[currentYear]);
      //eslint-disable-next-line no-bitwise
      if (($(yearCoverage[currentMonth]) & (1 << currentDay)) !== 0) {
        return nearestDate;
      }

      nearestDayDate = findNearestDayInYearCoverage(
        yearCoverage, nearestDate, options
      );
      if (nearestDayDate !== null) {
        return nearestDayDate;
      }
    }
  }

  /**
   * @param {CoverageMap} coverageByProduct
   * @returns {Coverage}
   */
  static create(coverageByProduct) {
    return (
      (!coverageByProduct || isObjectEmpty(coverageByProduct))
      // eslint-disable-next-line no-use-before-define
      ? COVERAGE_EMPTY
      : new Coverage(coverageByProduct)
    );
  }

  /**
   * @param {Coverage} coverage1
   * @param {Coverage} coverage2
   * @returns {Coverage}
   */
  static merge(coverage1, coverage2) {
    if (coverage1.isEmpty) {
      return coverage2;
    }
    if (coverage2.isEmpty) {
      return coverage1;
    }
    return new Coverage(
      mergeCoverageMaps(
        [coverage1.coverageByProduct, coverage2.coverageByProduct]
      )
    );
  }
}

const emptyCoverageOverrides = {
  getProductsCount: emptyFunctionReturning(0),
  getProductIds   : emptyFunctionReturning([]),
  getArray        : emptyFunctionReturning([]),
  getUnion        : emptyFunctionReturning({}),
  getIntersection : emptyFunctionReturning({}),
  hasDataAtDate   : emptyFunctionReturningFalse,
  hasColoc        : emptyFunctionReturningFalse,
  findNearest     : emptyFunctionReturningNull,
};

/** @type {Coverage} */
export const COVERAGE_EMPTY = Object.assign(new Coverage(), emptyCoverageOverrides);
/** @type {Coverage} */
export const COVERAGE_NULL = Object.assign(new Coverage(), emptyCoverageOverrides);

/**
 * @param {{start: Date, end: Date}[]} infos
 * @param {Date} [minDate]
 * @param {Date} [maxDate]
 * @returns {ProductCoverage}
 */
export function getCoverageFromInfos(infos, minDate, maxDate) {
  /** @type {ProductCoverage} */
  const coverage = {};

  for (let {start, end} of infos) {
    if (start.getTime() === DATE_MIN_TIME && end.getTime() === DATE_MAX_TIME) {
      // permanent datasets are not included in coverage
      continue;
    }

    if (minDate) {
      start = new Date(Math.max(minDate, start));
    }
    if (maxDate) {
      end = new Date(Math.min(maxDate, end));
    }

    // Ensure `start` is at midnight
    start.setUTCHours(0, 0, 0, 0);

    // eslint-disable-next-line no-unmodified-loop-condition
    while (start < end) {
      const year = start.getUTCFullYear();
      const month = /** @type {_Month} */ (start.getUTCMonth());
      const day = start.getUTCDate() - 1;

      //eslint-disable-next-line no-bitwise
      (coverage[year] ??= new Uint32Array(12))[month] |= (1 << day);

      start.setUTCDate(start.getUTCDate() + 1);
    }
  }
  return coverage;
}

/**
 * @param {Iterable<CoverageMap>} coverageMaps
 * @returns {CoverageMap}
 */
export function mergeCoverageMaps(coverageMaps) {
  /** @type {CoverageMap} */
  const coverage = {};
  for (const coverageByProduct of coverageMaps) {
    for (const productId of Object.keys(coverageByProduct)) {
      coverage[productId] = mergeCoverages([
        coverage[productId],
        coverageByProduct[productId],
      ]);
    }
  }
  return coverage;
}

/**
 * @param {ProductCoverage} coverage
 * @returns {Date | undefined}
 */
export function getFirstDayWithDataInCoverage(coverage) {
  const years = Object.keys(coverage).map(parseDecimal).sort();
  for (const year of years) {
    const yearCoverage = $(coverage[year]);
    for (let month = 0; month < 12; month++) {
      const monthCoverage = $(yearCoverage[month]);
      if (monthCoverage !== 0) {
        for (let day = 0; day < 31; day++) {
          //eslint-disable-next-line no-bitwise
          if ((monthCoverage & (1 << day)) !== 0) {
            return new Date(Date.UTC(year, month, day + 1));
          }
        }
      }
    }
  }
  return undefined;
}
/**
 * @param {ProductCoverage} coverage
 * @returns {Date | undefined}
 */
export function getLastDayWithDataInCoverage(coverage) {
  const years = (
    Object.keys(coverage)
    .map(parseDecimal)
    .sort(reverseComparator(numericComparator))
  );
  for (const year of years) {
    const yearCoverage = $(coverage[year]);
    for (let month = 11; month > -1; month--) {
      const monthCoverage = $(yearCoverage[month]);
      if (monthCoverage !== 0) {
        for (let day = 30; day > -1; day--) {
          //eslint-disable-next-line no-bitwise
          if ((monthCoverage & (1 << day)) !== 0) {
            return new Date(Date.UTC(year, month, day + 1));
          }
        }
      }
    }
  }
  return undefined;
}
