/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Bounds, Geometry} from './gis.js';
import {Pattern} from './pattern.js';
import {$, createClass, extendClass, EventsHub} from './utils.js';

/**
 * @param {Readonly<DatasetInfo['mapBounds']>} mapBounds1
 * @param {Readonly<DatasetInfo['mapBounds']>} mapBounds2
 * @returns {boolean}
 */
function sameInfoMapBounds(mapBounds1, mapBounds2) {
  if (mapBounds1 === mapBounds2 || !mapBounds1 && !mapBounds2) {
    return true;
  }
  if (!mapBounds1 || !mapBounds2) {
    return false;
  }
  return (
    mapBounds1.north === mapBounds2.north &&
    mapBounds1.east  === mapBounds2.east  &&
    mapBounds1.south === mapBounds2.south &&
    mapBounds1.west  === mapBounds2.west
  );
}

/**
 * @param {Readonly<ArrayLike<unknown>>} array1
 * @param {Readonly<ArrayLike<unknown>>} array2
 * @return {boolean}
 */
function isEqualArrays(array1, array2) {
  if (array1.length !== array2.length) {
    return false;
  }
  for (let i = 0; i < array1.length; i++) {
    if (array1[i] !== array2[i]) {
      return false;
    }
  }
  return true;
}

/**
 * @param {Readonly<Projected<Geometry>>} geometry
 * @param {Readonly<Projected<Bounds>>} globalBounds
 * @returns {boolean}
 */
function isGeometryGlobalRect(geometry, globalBounds) {
  // Check if geometry is a polygon
  if (!Geometry.isPolygon(geometry)) {
    return false;
  }

  // Check if polygon is a rectangle
  const xys = Geometry.polygonToXYs(geometry);
  if (xys.length !== 5 || !isEqualArrays($(xys[0]), $(xys[4]))) {
    return false;
  }
  for (let i = 0; i < 4; i++) {
    if ($(xys[i])[0] !== $(xys[i + 1])[0] && $(xys[i])[1] !== $(xys[i + 1])[1]) {
      return false;
    }
  }

  // Check if rectangle has global bounds
  return isEqualArrays(
    Bounds.toLBRT(Geometry.getBounds(geometry)).map(Math.trunc),
    Bounds.toLBRT(globalBounds).map(Math.trunc)
  );
}

/**
 * @param {Readonly<Projected<Geometry>>} outline
 * @returns {Projected<Geometry>}
 */
function simplifyDatasetOutline(outline) {
  if (Geometry.isPolygon(outline)) {
    if (
      Geometry.polygonHolesCount(outline) === 0 &&
      Geometry.polygonCount(outline) <= 5
    ) {
      return outline;
    }
    const scale = Math.sqrt(Bounds.getArea(Geometry.getBounds(outline))) / 100;
    return Geometry.simplify(outline, Math.min(1000, scale));
  }
  if (Geometry.isCollection(outline)) {
    return Geometry.createCollection(
      Geometry.getCollectionComponents(outline).map(simplifyDatasetOutline)
    );
  }
  return Geometry.simplify(outline, 1000);
}

// #region Exported typedefs

/**
 * @typedef {ClassExtension<DatasetClass> & ThisType<any>} Dataset.Extension
 */
/**
 * @typedef Dataset.ClassProps
 * @prop {Array<Dataset.Extension>} [extensions]
 */

// #endregion

export class Dataset {
  // #region Private typedefs

  /**
   * @template {Class} C
   * @typedef {import('./utils.js').ClassExtension<C>} ClassExtension
   */
  /**
   * @template {Class} [C = Class]
   * @typedef {import('./utils.js').ClassDefinition<C>} ClassDefinition
   */
  /** @typedef {import('./api.js').API.DatasetInfo} DatasetInfo */
  /** @typedef {import('./syntool.js').Syntool} Syntool */
  /**
   * @typedef {typeof Dataset & {
   *   productTypes: string[],
   *   validateConfig: import('./types.js').TypeChecker,
   * }} DatasetClass
   */

  // #endregion
  // #region Public

  /** @type {Record<string, DatasetClass>} */
  static CLASS_FOR_TYPE = {};
  /** @type {Record<string, Dataset.Extension>} */
  static extensions = {};
  /**
   * @param {ClassDefinition<DatasetClass> & Dataset.ClassProps} props
   * @returns {DatasetClass}
   */
  static createClass(props) {
    const {extensions} = props;
    delete props.extensions;

    const Class = /** @type {DatasetClass} */ (
      createClass(Dataset, /** @type {ClassDefinition} */ (props))
    );

    if (extensions) {
      extendClass(
        Class,
        (
          /** @type {import('./utils.js').ClassExtensionWithInfer<DatasetClass>[]} */
          (extensions)
        )
      );
    }
    for (const type of Class.productTypes ?? []) {
      Dataset.CLASS_FOR_TYPE[type] = Class;
    }
    return Class;
  }
  /**
   * @public
   * @param {DatasetInfo} info
   * @param {Syntool} syntool
   * @returns {Dataset | null}
   */
  static createDataset(info, syntool) {
    const product = syntool.productsStore.getProduct(info.productId);

    if (!product) {
      console.error(
        `createDataset: Product with ID "${info.productId}" not found!`,
        info
      );
    } else if (info.start.getTime() >= info.end.getTime()) {
      console.error(
        'createDataset: Expected `info.start` < `info.end`. But found ' +
        `"${info.start.toISOString()}" and "${info.end.toISOString()}" ` +
        `for dataset with id "${info.datasetId}"`,
        info
      );
    } else {
      const DatasetClass = Dataset.CLASS_FOR_TYPE[product.type];
      if (DatasetClass) {
        return new DatasetClass(info, syntool);
      }
    }
    console.error(
      "createDataset: Can't create a dataset for product type " +
      `"${product?.type}" for dataset with id "${info.datasetId}"`,
      info
    );
    return null;
  }

  /**
   * @public
   * @param {ProductRawConfig} productConfig
   * @returns {ProductRawConfig}
   */
  static normalizeProductConfig(productConfig) {
    return productConfig;
  }

  /**
   * @public
   * @param {DatasetInfo} info
   * @param {Syntool} syntool
   */
  constructor(info, syntool) {
    this.syntool = syntool;
    this.hub = new EventsHub();
    this.serverId = info.serverId;      // to get server
    this.productId = info.productId;    // to get product
    this.datasetId = info.datasetId;    // id
    this.name = info.name;              // used as dataset name (shown to the user)
    this.start = info.start;            // timeline and datasets date filter
    this.end = info.end;                // timeline and datasets date filter
    /** @type {Projected<Bounds> | null} */
    this.bounds = null;
    /** @type {Projected<Geometry> | null} */
    this.geometry = null;
    this.hidden = true;
    /** @type {DatasetInfo | null} */
    this._info = null;
    /** @type {boolean | null} */
    this._global = null;
    this.initialize(info);
  }
  /**
   * @public
   * @returns {string}
   */
  toString() {
    return this.constructor.name + '(' + this.datasetId + ')';
  }

  /**
   * @public
   * @param {DatasetInfo} info
   * @returns {void}
   */
  initialize(info) {
    this.handleInfo(info);
  }
  /**
   * @public
   * @param {DatasetInfo} info
   * @returns {boolean}
   * @emits change
   */
  updateFromInfo(info) {
    if (this._info !== info && this.serverId === info.serverId) {
      this.handleInfo(info);
      this.hub.emit('change');
      return true;
    }
    return false;
  }

  /**
   * @protected
   * @param {DatasetInfo} info
   * @returns {void}
   */
  handleInfo(info) {
    this.start = info.start;  // timeline and datasets date filter
    this.end = info.end;      // timeline and datasets date filter
    if (!sameInfoMapBounds(this._info?.mapBounds, info.mapBounds)) {
      this.bounds = (
        info.mapBounds && info.mapBounds.north !== null
        ? Bounds.create(
          parseFloat(info.mapBounds.west),
          parseFloat(info.mapBounds.south),
          parseFloat(info.mapBounds.east),
          parseFloat(info.mapBounds.north)
        )
        : null
      );
    }
    if (this._info?.polygon !== info.polygon) {
      this.geometry = (
        info.polygon
        ? simplifyDatasetOutline(Geometry.fromWKT(info.polygon))
        : null
      );
    }

    if (this.bounds && !this.geometry) {
      this.geometry = Geometry.fromBounds(this.bounds);
    } else if (this.geometry && !this.bounds) {
      this.bounds = Geometry.getBounds(this.geometry);
    }

    this._info = info;
    this._global = null;
  }

  /**
   * @public
   * @returns {Product | undefined}
   */
  getProduct() {
    return this.syntool.productsStore.getProduct(this.productId);
  }
  /**
   * @public
   * @param {HybridPattern} pattern
   * @returns {boolean}
   */
  matchesPattern(pattern) {
    return Pattern.test(pattern, this.name);
  }
  /**
   * @public
   * @param {Projected<Bounds>} bounds
   * @param {Projected<Bounds> | null} [worldBounds]
   * @returns {boolean | 0}
   */
  intersectsBounds(bounds, worldBounds) {
    return (
      (!this.bounds || Bounds.intersects(this.bounds, bounds, worldBounds)) &&
      (
        !this.geometry ||
        // Global datasets have their geometry set to a point
        Geometry.isPoint(this.geometry) ||
        Geometry.intersectsLatLonBounds(this.geometry, bounds, worldBounds)
      )
    );
  }
  /**
   * @public
   * @param {Projected<XY>} projectedXY
   * @returns {boolean}
   */
  containsProjectedXY(projectedXY) {
    return false;
  }
  /**
   * @public
   * @returns {boolean}
   */
  hasResources() {
    return false;
  }

  /**
   * @public
   * @returns {boolean}
   */
  get canSetOpacity() {
    return !!(/** @type {any} */ (this)).setOpacity;
  }
  /**
   * @public
   * @returns {number}
   */
  getOpacity() {
    return 1;
  }

  /**
   * @public
   * @returns {boolean}
   */
  get canSetStackLevel() {
    return !!(/** @type {any} */ (this)).setStackLevel;
  }
  /**
   * @public
   * @returns {number}
   */
  getStackLevel() {
    return 0;
  }

  /**
   * @public
   * @returns {boolean}
   */
  isHidden() {
    return this.hidden;
  }
  /**
   * @public
   * @returns {void}
   */
  show() {
    if (this.hidden) {
      this.hidden = false;
      if (this.syntool.mapView.isDatasetInMap(this)) {
        this.handleVisibilityChange();
      }
    }
  }
  /**
   * @public
   * @returns {void}
   */
  hide() {
    if (!this.hidden) {
      this.hidden = true;
      if (this.syntool.mapView.isDatasetInMap(this)) {
        this.handleVisibilityChange();
      }
    }
  }

  /**
   * @public
   * @returns {boolean}
   */
  isVisible() {
    return this.syntool.mapView.isDatasetInMap(this) && !this.isHidden();
  }
  /**
   * @public
   * @returns {boolean}
   */
  isSelected() {
    return false;
  }

  /**
   * @public
   * @returns {boolean}
   */
  isGlobal() {
    if (this._global === null) {
      this._global = (
        this.geometry
        ? (
          Geometry.isPoint(this.geometry) ||
          isGeometryGlobalRect(
            this.geometry,
            this.syntool.mapView.getMaxProjectedBounds()
          )
        )
        : false
      );
    }
    return this._global;
  }

  /**
   * @protected
   * @returns {void}
   * @emits change:visibility
   */
  handleVisibilityChange() {
    this.hub.emit('change:visibility');
  }
  /**
   * @protected
   * @returns {void}
   */
  handleSelectedChange() { /*empty*/ }
  /**
   * @public
   * @returns {void}
   */
  willAddToMap() { /*empty*/ }
  /**
   * @public
   * @returns {void}
   */
  didAddToMap() {
    if (!this.isHidden()) {
      this.handleVisibilityChange();
    }
  }
  /**
   * @public
   * @returns {void}
   */
  willRemoveFromMap() { /*empty*/ }
  /**
   * @public
   * @returns {void}
   * @emits change:visibility
   */
  didRemoveFromMap() {
    if (!this.isHidden()) {
      this.handleVisibilityChange();
    }
    this.handleSelectedChange();
  }
  /**
   * @public
   * @returns {void}
   */
  destroy() { /*empty*/ }

  // #endregion
}
