/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {OpenLayers} from '../../vendors/openlayers.js';
import {$, inRange, modulo} from '../utils.js';

/** @typedef {Projected<import('.').OLLonLat, 'EPSG:4326'>} LatLon */
/** @typedef {Projected<import('.').OLBounds, 'EPSG:4326'>} LatLonBounds */
/** @typedef {import('.').OLSize} Size */
/** @typedef {import('.').OLLonLat} Point */
/** @typedef {import('.').OLBounds} Bounds */
/** @typedef {OpenLayers.PointLike} PointLike */

export const LatLon = {
  /**
   * @public
   * @param {Lat} lat
   * @param {Lon} lon
   * @returns {LatLon}
   */
  create(lat, lon) {
    return new OpenLayers.LonLat(lon, lat);
  },

  /**
   * @public
   * @param {Readonly<LonLat>} lonLat
   * @returns {LatLon}
   */
  createFromLonLat([lon, lat]) {
    return LatLon.create(lat, lon);
  },

  /**
   * @public
   * @param {Readonly<LatLon>} latLon
   * @param {LonLat} [lonLat]
   * @returns {LonLat}
   */
  toLonLat(latLon, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {LonLat} */ ([0, 0]);
    }
    lonLat[0] = latLon.lon;
    lonLat[1] = latLon.lat;
    return lonLat;
  },

  /**
   * @public
   * @param {Readonly<LatLon>} latLon
   * @returns {Lon}
   */
  getLon(latLon) {
    return latLon.lon;
  },

  /**
   * @public
   * @param {Readonly<LatLon>} latLon
   * @returns {Lat}
   */
  getLat(latLon) {
    return latLon.lat;
  },
};

export const LatLonBounds = {
  /**
   * @public
   * @param {Lat} lat1
   * @param {Lon} lon1
   * @param {Lat} lat2
   * @param {Lon} lon2
   * @returns {LatLonBounds}
   */
  create(lat1, lon1, lat2, lon2) {
    let left;
    let right;
    let top;
    let bottom;
    if (lat1 < lat2) {
      bottom = lat1;
      top = lat2;
    } else {
      bottom = lat2;
      top = lat1;
    }
    if (lon1 < lon2) {
      left = lon1;
      right = lon2;
    } else {
      left = lon2;
      right = lon1;
    }
    return new OpenLayers.Bounds(left, bottom, right, top);
  },

  /**
   * @public
   * @param {LatLonBounds | null | undefined} bounds
   * @param {Lat} lat1
   * @param {Lon} lon1
   * @param {Lat} lat2
   * @param {Lon} lon2
   * @returns {LatLonBounds}
   */
  createOrUpdate(bounds, lat1, lon1, lat2, lon2) {
    if (bounds) {
      if (lat1 < lat2) {
        bounds.bottom = lat1;
        bounds.top = lat2;
      } else {
        bounds.bottom = lat2;
        bounds.top = lat1;
      }
      if (lon1 < lon2) {
        bounds.left = lon1;
        bounds.right = lon2;
      } else {
        bounds.left = lon2;
        bounds.right = lon1;
      }
      bounds.centerLonLat = null;
    } else {
      bounds = LatLonBounds.create(lat1, lon1, lat2, lon2);
    }
    return bounds;
  },

  /**
   * @public
   * @param {Readonly<LonLatNSEW>} nsew
   * @returns {LatLonBounds}
   */
  createFromNESW(nsew) {
    return new OpenLayers.Bounds(nsew.west, nsew.south, nsew.east, nsew.north);
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {LatLonBounds}
   */
  createFromLatLonBounds(bounds) {
    return LatLonBounds.create(
      LatLonBounds.getBottom(bounds),
      LatLonBounds.getLeft(bounds),
      LatLonBounds.getTop(bounds),
      LatLonBounds.getRight(bounds)
    );
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lon}
   */
  getLeft(bounds) {
    return bounds.left;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lon}
   */
  getRight(bounds) {
    return bounds.right;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lat}
   */
  getTop(bounds) {
    return bounds.top;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lat}
   */
  getBottom(bounds) {
    return bounds.bottom;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lat}
   */
  getNorth(bounds) {
    return bounds.top;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lon}
   */
  getEast(bounds) {
    return bounds.right;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lat}
   */
  getSouth(bounds) {
    return bounds.bottom;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lon}
   */
  getWest(bounds) {
    return bounds.left;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Degrees}
   */
  getWidth(bounds) {
    return bounds.right - bounds.left;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Degrees}
   */
  getHeight(bounds) {
    return bounds.top - bounds.bottom;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {LatLon}
   */
  getCenterLatLon(bounds) {
    return bounds.getCenterLonLat();
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lat}
   */
  getCenterLat(bounds) {
    return (bounds.top + bounds.bottom) / 2;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @returns {Lon}
   */
  getCenterLon(bounds) {
    return (bounds.left + bounds.right) / 2;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @param {LonLat} [lonLat]
   * @returns {LonLat}
   */
  getCenter(bounds, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {LonLat} */ ([0, 0]);
    }
    lonLat[0] = LatLonBounds.getCenterLon(bounds);
    lonLat[1] = LatLonBounds.getCenterLat(bounds);
    return lonLat;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @param {LonLat} [lonLat]
   * @returns {LonLat}
   */
  getTopLeft(bounds, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {LonLat} */ ([0, 0]);
    }
    lonLat[0] = LatLonBounds.getLeft(bounds);
    lonLat[1] = LatLonBounds.getTop(bounds);
    return lonLat;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @param {LonLat} [lonLat]
   * @returns {LonLat}
   */
  getTopRight(bounds, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {LonLat} */ ([0, 0]);
    }
    lonLat[0] = LatLonBounds.getRight(bounds);
    lonLat[1] = LatLonBounds.getTop(bounds);
    return lonLat;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @param {LonLat} [lonLat]
   * @returns {LonLat}
   */
  getBottomLeft(bounds, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {LonLat} */ ([0, 0]);
    }
    lonLat[0] = LatLonBounds.getLeft(bounds);
    lonLat[1] = LatLonBounds.getBottom(bounds);
    return lonLat;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds
   * @param {LonLat} [lonLat]
   * @returns {LonLat}
   */
  getBottomRight(bounds, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {LonLat} */ ([0, 0]);
    }
    lonLat[0] = LatLonBounds.getRight(bounds);
    lonLat[1] = LatLonBounds.getBottom(bounds);
    return lonLat;
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds1
   * @param {Readonly<LatLonBounds>} bounds2
   * @param {Readonly<LatLonBounds> | null} [worldBounds]
   * @returns {boolean}
   */
  intersects(bounds1, bounds2, worldBounds) {
    return bounds1.intersectsBounds(
      bounds2,
      (worldBounds ? {worldBounds} : null)
    );
  },

  /**
   * @public
   * @param {Readonly<LatLonBounds>} bounds1
   * @param {Readonly<LatLonBounds>} bounds2
   * @returns {Readonly<LatLonBounds>}
   */
  intersection(bounds1, bounds2) {
    return LatLonBounds.create(
      Math.min(LatLonBounds.getTop(bounds1), LatLonBounds.getTop(bounds2)),
      Math.min(LatLonBounds.getRight(bounds1), LatLonBounds.getRight(bounds2)),
      Math.max(LatLonBounds.getBottom(bounds1), LatLonBounds.getBottom(bounds2)),
      Math.max(LatLonBounds.getLeft(bounds1), LatLonBounds.getLeft(bounds2))
    );
  },

  /**
   * @public
   * @param {LatLonBounds} bounds
   * @param {(x: number) => number} round
   * @returns {LatLonBounds}
   */
  round(bounds, round) {
    bounds.left   = round(bounds.left);
    bounds.bottom = round(bounds.bottom);
    bounds.right  = round(bounds.right);
    bounds.top    = round(bounds.top);
    return bounds;
  },
};

export const Size = {
  /**
   * @public
   * @param {number} width
   * @param {number} height
   * @returns {Size}
   */
  create(width, height) {
    return new OpenLayers.Size(width, height);
  },

  /**
   * @public
   * @param {Readonly<WH>} wh
   * @returns {Size}
   */
  createFromWH([width, height]) {
    return Size.create(width, height);
  },

  /**
   * @public
   * @param {Readonly<Size>} size
   * @param {WH} [wh]
   * @returns {WH}
   */
  toWH(size, wh) {
    if (!wh || wh.length === undefined) {
      wh = [0, 0];
    }
    wh[0] = Size.getWidth(size);
    wh[1] = Size.getHeight(size);
    return wh;
  },

  /**
   * @public
   * @param {Readonly<Size>} size
   * @returns {number}
   */
  getWidth(size) {
    return size.w;
  },

  /**
   * @public
   * @param {Readonly<Size>} size
   * @returns {number}
   */
  getHeight(size) {
    return size.h;
  },
};

export const Point = {
  /**
   * @public
   * @param {number} x
   * @param {number} y
   * @returns {Point}
   */
  create(x, y) {
    return new OpenLayers.LonLat(x, y);
  },

  /**
   * @public
   * @param {Readonly<XY>} xy
   * @returns {Point}
   */
  createFromXY([x, y]) {
    return Point.create(x, y);
  },

  /**
   * @public
   * @param {Readonly<Point>} point
   * @param {XY} [xy]
   * @returns {XY}
   */
  toXY(point, xy) {
    if (!xy || xy.length === undefined) {
      xy = [0, 0];
    }
    xy[0] = Point.getX(point);
    xy[1] = Point.getY(point);
    return xy;
  },

  /**
   * @public
   * @param {Readonly<Point>} point
   * @returns {number}
   */
  getX(point) {
    return point.lon;
  },

  /**
   * @public
   * @param {Readonly<Point>} point
   * @returns {number}
   */
  getY(point) {
    return point.lat;
  },
};

export const Bounds = {
  /**
   * @public
   * @param {number} left
   * @param {number} bottom
   * @param {number} right
   * @param {number} top
   * @returns {Bounds}
   */
  create(left, bottom, right, top) {
    return Bounds.createOrUpdate(null, left, bottom, right, top);
  },

  /**
   * @public
   * @param {Bounds | null | undefined} bounds
   * @param {number} left
   * @param {number} bottom
   * @param {number} right
   * @param {number} top
   * @returns {Bounds}
   */
  createOrUpdate(bounds, left, bottom, right, top) {
    if (!bounds) {
      bounds = new OpenLayers.Bounds();
    }
    bounds.left = left;
    bounds.bottom = bottom;
    bounds.right = right;
    bounds.top = top;
    bounds.centerLonLat = null;
    return bounds;
  },

  /**
   * @public
   * @returns {Bounds}
   */
  createZero() {
    return Bounds.create(0, 0, 0, 0);
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {Bounds}
   */
  createFromBounds(bounds) {
    return Bounds.create(
      Bounds.getLeft(bounds),
      Bounds.getBottom(bounds),
      Bounds.getRight(bounds),
      Bounds.getTop(bounds),
    );
  },

  /**
   * @public
   * @param {Readonly<LBRT>} lbrt
   * @returns {Bounds}
   */
  createFromLBRT([left, bottom, right, top]) {
    return Bounds.create(left, bottom, right, top);
  },

  /**
   * @public
   * @param {Readonly<NSEW>} nsew
   * @returns {Bounds}
   */
  createFromNESW(nsew) {
    return Bounds.create(nsew.west, nsew.south, nsew.east, nsew.north);
  },

  /**
   * @public
   * @param {Readonly<XY[]> | null | undefined} xys
   * @param {Readonly<Bounds> | null} [worldBounds]
   * @returns {Bounds}
   */
  createFromXYs(xys, worldBounds) {
    if (!xys || !Array.isArray(xys) || xys.length < 1) {
      return Bounds.createZero();
    }
    const worldWidth = worldBounds ? Bounds.getWidth(worldBounds) : 0;
    const firstXY = $(xys[0]);
    const center = firstXY[0] - worldWidth / 2;
    let left = firstXY[0];
    let bottom = firstXY[1];
    let right = left;
    let top = bottom;

    for (let [x, y] of xys) {
      if (worldWidth) {
        x = modulo(x - center, worldWidth) + center;
      }
      if (x < left) {
        left = x;
      } else if (x > right) {
        right = x;
      }
      if (y < bottom) {
        bottom = y;
      } else if (y > top) {
        top = y;
      }
    }
    return Bounds.create(left, bottom, right, top);
  },

  /**
   * @public
   * @param {Readonly<Bounds[]> | null | undefined} boundsList
   * @param {Readonly<Bounds> | null} [worldBounds]
   * @returns {Bounds}
   */
  createFromBoundsList(boundsList, worldBounds) {
    if (!boundsList || !Array.isArray(boundsList) || boundsList.length < 1) {
      return Bounds.createZero();
    }

    const worldWidth = worldBounds ? Bounds.getWidth(worldBounds) : 0;
    let bounds = $(boundsList[0]);
    let left = Bounds.getLeft(bounds);
    let bottom = Bounds.getBottom(bounds);
    let right = Bounds.getRight(bounds);
    let top = Bounds.getTop(bounds);
    const center = (left + right - worldWidth) / 2;

    let l;
    let b;
    let r;
    let t;
    for (bounds of boundsList) {
      l = Bounds.getLeft(bounds);
      b = Bounds.getBottom(bounds);
      r = Bounds.getRight(bounds);
      t = Bounds.getTop(bounds);
      if (worldWidth) {
        l = modulo(l - center, worldWidth) + center;
        r = modulo(r - center, worldWidth) + center;
      }
      if (l < left) {
        left = l;
      }
      if (r > right) {
        right = r;
      }
      if (b < bottom) {
        bottom = b;
      }
      if (t > top) {
        top = t;
      }
    }
    return Bounds.create(left, bottom, right, top);
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {LBRT} [lbrt]
   * @returns {LBRT}
   */
  toLBRT(bounds, lbrt) {
    if (!lbrt || lbrt.length === undefined) {
      lbrt = [0, 0, 0, 0];
    }
    lbrt[0] = Bounds.getLeft(bounds);
    lbrt[1] = Bounds.getBottom(bounds);
    lbrt[2] = Bounds.getRight(bounds);
    lbrt[3] = Bounds.getTop(bounds);
    return lbrt;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getLeft(bounds) {
    return bounds.left;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getRight(bounds) {
    return bounds.right;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getTop(bounds) {
    return bounds.top;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getBottom(bounds) {
    return bounds.bottom;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getWidth(bounds) {
    return bounds.right - bounds.left;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getHeight(bounds) {
    return bounds.top - bounds.bottom;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getCenterX(bounds) {
    return (bounds.left + bounds.right) / 2;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getCenterY(bounds) {
    return (bounds.top + bounds.bottom) / 2;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {XY} [xy]
   * @returns {XY}
   */
  getCenter(bounds, xy) {
    if (!xy || xy.length === undefined) {
      xy = [0, 0];
    }
    xy[0] = Bounds.getCenterX(bounds);
    xy[1] = Bounds.getCenterY(bounds);
    return xy;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {XY} [xy]
   * @returns {XY}
   */
  getTopLeft(bounds, xy) {
    if (!xy || xy.length === undefined) {
      xy = [0, 0];
    }
    xy[0] = Bounds.getLeft(bounds);
    xy[1] = Bounds.getTop(bounds);
    return xy;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {XY} [xy]
   * @returns {XY}
   */
  getTopRight(bounds, xy) {
    if (!xy || xy.length === undefined) {
      xy = [0, 0];
    }
    xy[0] = Bounds.getRight(bounds);
    xy[1] = Bounds.getTop(bounds);
    return xy;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {XY} [xy]
   * @returns {XY}
   */
  getBottomLeft(bounds, xy) {
    if (!xy || xy.length === undefined) {
      xy = [0, 0];
    }
    xy[0] = Bounds.getLeft(bounds);
    xy[1] = Bounds.getBottom(bounds);
    return xy;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {XY} [xy]
   * @returns {XY}
   */
  getBottomRight(bounds, xy) {
    if (!xy || xy.length === undefined) {
      xy = [0, 0];
    }
    xy[0] = Bounds.getRight(bounds);
    xy[1] = Bounds.getBottom(bounds);
    return xy;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {PointLike} point
   * @returns {Bounds}
   */
  extendWithPoint(bounds, {x, y}) {
    return bounds.add(x, y);
  },

  /**
   * @protected
   * @param {Readonly<Bounds>} bounds
   * @param {number} x
   * @param {number} y
   * @returns {boolean}
   */
  _contains(bounds, x, y) {
    return (
      inRange(x, bounds.left, bounds.right) &&
      inRange(y, bounds.bottom, bounds.top)
    );
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {XY} xy
   * @returns {boolean}
   */
  containsXY(bounds, [x, y]) {
    return Bounds._contains(bounds, x, y);
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @param {PointLike} point
   * @returns {boolean}
   */
  containsPoint(bounds, {x, y}) {
    return Bounds._contains(bounds, x, y);
  },

  /**
   * Determine whether bounds2 is contained within bounds1.
   *
   * @public
   * @param {Readonly<Bounds>} bounds1
   * @param {Readonly<Bounds>} bounds2
   * @param {boolean} [partial=false] - When true, if any of bounds2's corners
   * is within bounds1 consider bounds2 contained. Otherwise, the entire bound2
   * must be contained within bounds1.
   * @param {boolean} [inclusive=true] - Treat shared edges as contained.
   * @returns {boolean}
   */
  containsBounds(bounds1, bounds2, partial = false, inclusive = true) {
    return bounds1.containsBounds(bounds2, partial, inclusive);
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds1
   * @param {Readonly<Bounds>} bounds2
   * @param {Readonly<Bounds> | null} [worldBounds]
   * @returns {boolean}
   */
  intersects(bounds1, bounds2, worldBounds) {
    return bounds1.intersectsBounds(
      bounds2,
      (worldBounds ? {worldBounds} : null)
    );
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds1
   * @param {Readonly<Bounds>} bounds2
   * @returns {Bounds}
   */
  intersection(bounds1, bounds2) {
    return Bounds.create(
      Math.max(Bounds.getLeft(bounds1), Bounds.getLeft(bounds2)),
      Math.max(Bounds.getBottom(bounds1), Bounds.getBottom(bounds2)),
      Math.min(Bounds.getRight(bounds1), Bounds.getRight(bounds2)),
      Math.min(Bounds.getTop(bounds1), Bounds.getTop(bounds2))
    );
  },

  /**
   * @public
   * @param {Bounds} bounds
   * @param {(x: number) => number} round
   * @returns {Bounds}
   */
  round(bounds, round) {
    bounds.left   = round(bounds.left);
    bounds.bottom = round(bounds.bottom);
    bounds.right  = round(bounds.right);
    bounds.top    = round(bounds.top);
    return bounds;
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getPerimeter(bounds) {
    return 2 * (Bounds.getWidth(bounds) + Bounds.getHeight(bounds));
  },

  /**
   * @public
   * @param {Readonly<Bounds>} bounds
   * @returns {number}
   */
  getArea(bounds) {
    return (Bounds.getWidth(bounds) * Bounds.getHeight(bounds));
  },
};
