/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {DEG_IN_RAD, RAD_IN_DEG} from '../utils.js';

/** major semi-axes (in meters) (Equatorial Radius) */
const Va = /** @type {Meters} */ (6378137.0);
/** minor semi-axes (in meters) */
const Vb = /** @type {Meters} */ (6356752.314245);
/** flattening = (Va - Vb) / Va = 1 / 298.257223563 */
const Vf = 0.0033528106647474805;
/** the square of second eccentricity */
const Vε = (Va * Va - Vb * Vb) / (Vb * Vb);

/**
 * The difference between 1 and the smallest value greater than 1 that is
 * representable as a Number value
 */
const ε = Number.EPSILON;

export const Geography = {
  EARTH_RADIUS: Va,

  /**
   * Calculates the geodesic distance and bearings between two points.
   *
   * Based on Chris Veness's work
   * https://www.movable-type.co.uk/scripts/latlong-vincenty.html
   * (c) Chris Veness 2002-2019
   * MIT Licence
   *
   * @param {LonLat} lonLat1 - The point to calculate the distance from.
   * @param {LonLat} lonLat2 - The point to calculate the distance to.
   * @returns {{distance: Meters, initialBearing: Degrees, finalBearing: Degrees}}
   *         The geodesic distance between `lonLat1` and `lonLat2` in meters
   *         and the initial and final bearings in degrees.
   */
  geodesicDistanceAndBearings(lonLat1, lonLat2) {
    const L = (lonLat2[0] - lonLat1[0]) * DEG_IN_RAD;
    const U1 = Math.atan((1 - Vf) * Math.tan(lonLat1[1] * DEG_IN_RAD));
    const U2 = Math.atan((1 - Vf) * Math.tan(lonLat2[1] * DEG_IN_RAD));
    const sinU1 = Math.sin(U1);
    const cosU1 = Math.cos(U1);
    const sinU2 = Math.sin(U2);
    const cosU2 = Math.cos(U2);
    const antipodal = (
      Math.abs(lonLat2[0] - lonLat1[0]) > 90 ||
      Math.abs(lonLat2[1] - lonLat1[1]) > 90
    );
    let lambda = L;
    let lambdaP = 2 * Math.PI;
    let iterLimit = 20;
    let sinLambda;
    let cosLambda;
    let sinSigma;
    let cosSigma = antipodal ? -1 : 1;
    let sigma = antipodal ? Math.PI : 0;
    let alpha;
    let cosSqAlpha = 1;
    let cos2SigmaM = 1;
    let C;
    do {
      sinLambda = Math.sin(lambda);
      cosLambda = Math.cos(lambda);
      sinSigma = Math.sqrt(
        (cosU2 * sinLambda) * (cosU2 * sinLambda) +
        (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) * (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda)
      );
      if (sinSigma < ε && sinSigma > -ε) {
        // co-incident points
        break;
      }
      cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda;
      sigma = Math.atan2(sinSigma, cosSigma);
      alpha = Math.asin(cosU1 * cosU2 * sinLambda / sinSigma);
      cosSqAlpha = Math.cos(alpha) * Math.cos(alpha);
      cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha;
      C = Vf / 16 * cosSqAlpha * (4 + Vf * (4 - 3 * cosSqAlpha));
      lambdaP = lambda;
      lambda = (
        L + (1 - C) * Vf * Math.sin(alpha) *
        (
          sigma + C * sinSigma * (
            cos2SigmaM + C * cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM)
          )
        )
      );
    } while (Math.abs(lambda - lambdaP) > 1e-12 && --iterLimit > 0);
    if (iterLimit === 0) {
      // formula failed to converge
      return {distance: NaN, initialBearing: NaN, finalBearing: NaN};
    }

    const uSq = cosSqAlpha * Vε;
    const A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)));
    const B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)));
    const deltaSigma = (
      B * sinSigma *
      (
        cos2SigmaM + B / 4 *
        (
          cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM) -
          B / 6 * cos2SigmaM * (-3 + 4 * sinSigma * sinSigma) *
          (-3 + 4 * cos2SigmaM * cos2SigmaM)
        )
      )
    );
    const s = Vb * A * (sigma - deltaSigma);

    // geodesic distance, rounded to 1mm precision
    const distance = Number(s.toFixed(3));
    // the initial bearing, or forward azimuth
    const initialBearing = (
      Math.atan2(cosU2 * sinLambda,  cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) *
      RAD_IN_DEG
    );
    // the final bearing (in direction lonLat1→lonLat2)
    const finalBearing = (
      Math.atan2(cosU1 * sinLambda, -sinU1 * cosU2 + cosU1 * sinU2 * cosLambda) *
      RAD_IN_DEG
    );

    return {distance, initialBearing, finalBearing};
  },

  /**
   * Calculates the geodesic distance between two points.
   *
   * @param {LonLat} lonLat1 - The point to calculate the distance from.
   * @param {LonLat} lonLat2 - The point to calculate the distance to.
   * @returns {Meters} The geodesic distance between `lonLat1` and `lonLat2` in meters.
   */
  geodesicDistance(lonLat1, lonLat2) {
    return Geography.geodesicDistanceAndBearings(lonLat1, lonLat2).distance;
  },

  /**
   * Calculates the geodesic bearing between two points on the side of the first.
   *
   * @param {LonLat} lonLat1 - The point to calculate the bearing from.
   * @param {LonLat} lonLat2 - The point to calculate the bearing to.
   * @returns {Degrees} The geodesic bearing between `lonLat1` and `lonLat2` in degrees.
   */
  geodesicInitialBearing(lonLat1, lonLat2) {
    return (
      Geography.geodesicDistanceAndBearings(lonLat1, lonLat2).initialBearing
    );
  },

  /**
   * Calculates the geodesic bearing between two points on the side of the second.
   *
   * @param {LonLat} lonLat1 - The point to calculate the bearing from.
   * @param {LonLat} lonLat2 - The point to calculate the bearing to.
   * @returns {Degrees} The geodesic bearing between `lonLat1` and `lonLat2` in degrees.
   */
  geodesicFinalBearing(lonLat1, lonLat2) {
    return Geography.geodesicDistanceAndBearings(lonLat1, lonLat2).finalBearing;
  },

  /**
   * Calculates the geodesic bearings between two points.
   *
   * @param {LonLat} lonLat1 - The point to calculate the bearing from.
   * @param {LonLat} lonLat2 - The point to calculate the bearing to.
   * @returns {{initialBearing: Degrees, finalBearing: Degrees}}
   *         The geodesic bearings between `lonLat1` and `lonLat2` in degrees.
   */
  geodesicBearings(lonLat1, lonLat2) {
    const {initialBearing, finalBearing} = (
      Geography.geodesicDistanceAndBearings(lonLat1, lonLat2)
    );
    return {initialBearing, finalBearing};
  },

  /**
   * Finds the end point and the final bearing of a geodesic from a point, an
   * initial bearing and a distance.
   *
   * Based on Chris Veness's work
   * https://www.movable-type.co.uk/scripts/latlong-vincenty.html
   * (c) Chris Veness 2002-2019
   * MIT Licence
   *
   * @param {LonLat} lonLat1 - The starting point.
   * @param {Meters} distance - The geodesic distance between `lonLat1` and the end point in meters.
   * @param {Degrees} bearing - The initial bearing in degrees.
   * @returns {{lonLat: LonLat, finalBearing: Degrees}} The end point and final bearing.
   */
  geodesicDestinationAndBearing(lonLat1, distance, bearing) {
    const [lon1, lat1] = lonLat1;

    const s = distance;
    const alpha1 = bearing * DEG_IN_RAD;
    const sinAlpha1 = Math.sin(alpha1);
    const cosAlpha1 = Math.cos(alpha1);

    const tanU1 = (1 - Vf) * Math.tan(lat1 * DEG_IN_RAD);
    const cosU1 = 1 / Math.sqrt(1 + tanU1 * tanU1);
    const sinU1 = tanU1 * cosU1;
    const sigma1 = Math.atan2(tanU1, cosAlpha1);
    const sinAlpha = cosU1 * sinAlpha1;
    const cosSqAlpha = 1 - sinAlpha * sinAlpha;
    const uSq = cosSqAlpha * Vε;
    const A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)));
    const B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)));

    let sigma = s / (Vb * A);
    let sigmaP = 2 * Math.PI;
    let cos2SigmaM;
    let sinSigma;
    let cosSigma;
    let deltaSigma;
    do {
      cos2SigmaM = Math.cos(2 * sigma1  +  sigma);
      sinSigma = Math.sin(sigma);
      cosSigma = Math.cos(sigma);
      deltaSigma = (
        B * sinSigma *
        (
          cos2SigmaM + B / 4 *
          (
            cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM) -
            B / 6 * cos2SigmaM * (-3 + 4 * sinSigma * sinSigma) *
            (-3 + 4 * cos2SigmaM * cos2SigmaM)
          )
        )
      );
      sigmaP = sigma;
      sigma = s / (Vb * A) + deltaSigma;
    } while (Math.abs(sigma - sigmaP) > 1e-12);

    const tmp = sinU1 * sinSigma - cosU1 * cosSigma * cosAlpha1;
    const lat2 = Math.atan2(
      sinU1 * cosSigma + cosU1 * sinSigma * cosAlpha1,
      (1 - Vf) * Math.sqrt(sinAlpha * sinAlpha + tmp * tmp)
    );
    const lambda = Math.atan2(
      sinSigma * sinAlpha1,
      cosU1 * cosSigma - sinU1 * sinSigma * cosAlpha1
    );
    const C = Vf / 16 * cosSqAlpha * (4 + Vf * (4 - 3 * cosSqAlpha));
    const L = (
      lambda - (1 - C) * Vf * sinAlpha *
      (
        sigma + C * sinSigma *
        (cos2SigmaM + C * cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM))
      )
    );
    const finalBearing = Math.atan2(sinAlpha, -tmp) * RAD_IN_DEG;
    const lonLat = /** @type {LonLat} */ (
      [lon1 + L * RAD_IN_DEG, lat2 * RAD_IN_DEG]
    );

    return {lonLat, finalBearing};
  },

  /**
   * Finds the end point of a geodesic from a point, an initial bearing and a distance.
   *
   * @param {LonLat} lonLat1 - The starting point.
   * @param {Meters} distance - The geodesic distance between `lonLat1` and the end point in meters.
   * @param {Degrees} bearing - The initial bearing in degrees.
   * @returns {LonLat} The end point.
   */
  geodesicDestination(lonLat1, distance, bearing) {
    return (
      Geography.geodesicDestinationAndBearing(lonLat1, distance, bearing).lonLat
    );
  },
};
