/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/// <reference path='./geojson.d.ts' />
import {$, isXYEqual} from '../utils.js';
import {Feature} from './gis_feature.js';
import {Geometry} from './gis_geometry.js';
import {Projection} from './gis_projection.js';

const WGS84 = Projection.fromCode('WGS84');

/**
 * @param {Readonly<JSONValue> | null | undefined} value
 * @returns {value is XY}
 */
function isXY(value) {
  if (!Array.isArray(value)) {
    return false;
  }
  const [x, y] = /** @type {ReadonlyArray<number | null | undefined>} */ (value);
  return x !== null && y !== null && isFinite(x) && isFinite(y);
}
/**
 * @param {Readonly<JSONValue> | null | undefined} value
 * @returns {value is XY[]}
 */
function isPolygonRingCoords(value) {
  return (
    Array.isArray(value)
    && value.length >= 4
    && value.every(isXY)
    && isXYEqual(value[0], value[value.length - 1])
  );
}
/**
 * @param {Readonly<JSONValue> | null | undefined} value
 * @returns {value is XY[][]}
 */
function isPolygonCoords(value) {
  return (
    Array.isArray(value)
    && value.length >= 1
    && value.every(isPolygonRingCoords)
  );
}

/**
 * @template {JSONArray} T
 * @param {T} coordinates
 * @param {(position: XY) => XY} mapper
 * @returns {T}
 */
function mapCoordinates(coordinates, mapper) {
  if (Array.isArray(coordinates[0])) {
    return /** @type {T} */ (
      (/** @type {JSONArray[]} */ (coordinates))
      .map((coords) => mapCoordinates(coords, mapper))
    );
  }
  if (isXY(coordinates)) {
    return /** @type {any} */ (mapper(coordinates));
  }
  return coordinates;
}

/**
 * @template {JSONArray} T
 * @param {T} coordinates
 * @param {Readonly<Projection> | undefined | null} sourceProjection
 * @param {Readonly<Projection> | undefined | null} targetProjection
 * @returns {T}
 */
function changeCoordinatesProjection(coordinates, sourceProjection, targetProjection) {
  if (!sourceProjection && !targetProjection) {
    return coordinates;
  }
  return mapCoordinates(coordinates, (point) => {
    if (sourceProjection) {
      point = Projection.unprojectXY(
        sourceProjection,
        point,
        /** @type {LonLat} */ (point)
      );
    }
    if (targetProjection) {
      point = Projection.projectLonLat(
        targetProjection,
        /** @type {LonLat} */ (point),
        point
      );
    }
    return point;
  });
}
/**
 * @template {JSONArray} T
 * @param {T} coordinates
 * @param {Readonly<Projection> | undefined | null} sourceProjection
 * @param {Readonly<Projection> | undefined | null} targetProjection
 * @returns {T}
 */
function coordinatesWithProjection(coordinates, sourceProjection, targetProjection) {
  if (!sourceProjection && !targetProjection) {
    return coordinates;
  }
  return mapCoordinates(coordinates, (point) => {
    if (sourceProjection) {
      point = Projection.unprojectXY(
        sourceProjection,
        point,
      );
    }
    if (targetProjection) {
      point = Projection.projectLonLat(
        targetProjection,
        /** @type {LonLat} */ (point),
      );
    }
    return point;
  });
}

const GEOMETRY_TYPES = new Set([
  'Point', 'MultiPoint', 'LineString', 'MultiLineString',
  'Polygon', 'MultiPolygon', 'Box', 'GeometryCollection',
]);
let strictParsing = true;
const parseCoords = {
  /**
   * @param {Readonly<JSONValue>} coords
   * @returns {Geometry.Point}
   * @throws {TypeError} If `coords` is an invalid GeoJSON position.
   */
  point(coords) {
    if (!isXY(coords)) {
      throw new TypeError('Invalid GeoJSON position');
    }
    return Geometry.createPointFromXY(coords);
  },
  /**
   * @param {Readonly<JSONArray>} array
   * @returns {Geometry.Collection<Geometry.Point>}
   * @throws {TypeError} If `array` contains invalid GeoJSON positions.
   */
  multipoint(array) {
    return Geometry.createCollection(array.map(parseCoords.point));
  },
  /**
   * @param {Readonly<JSONValue>} coords
   * @returns {Geometry.Polyline}
   * @throws {TypeError} If `coords` is an invalid GeoJSON LineString coordinates.
   */
  linestring(coords) {
    if (
      !Array.isArray(coords)
      || (strictParsing && coords.length < 2)
      || !coords.every(isXY)
    ) {
      throw new TypeError('Invalid GeoJSON LineString coordinates');
    }
    if (coords.length < 2) {
      console.warn('Patched single point GeoJSON LineString');
      const xy = $(coords[0]);
      return Geometry.createPolylineFromXYs([xy, xy]);
    }
    return Geometry.createPolylineFromXYs(coords);
  },
  /**
   * @param {Readonly<JSONArray>} array
   * @returns {Geometry.Collection<Geometry.Polyline>}
   * @throws {TypeError} If `array` contains invalid GeoJSON LineString coordinates.
   */
  multilinestring(array) {
    return Geometry.createCollection(array.map(parseCoords.linestring));
  },
  /**
   * @param {Readonly<JSONValue>} coords
   * @returns {Geometry.Polygon}
   * @throws {TypeError} If `coords` is an invalid GeoJSON Polygon coordinates.
   */
  polygon(coords) {
    if (!isPolygonCoords(coords)) {
      throw new TypeError('Invalid GeoJSON Polygon coordinates');
    }
    return Geometry.createPolygonFromXYs(coords);
  },
  /**
   * @param {Readonly<JSONArray>} array
   * @returns {Geometry.Collection<Geometry.Polygon>}
   * @throws {TypeError} If `array` contains invalid GeoJSON Polygon coordinates.
   */
  multipolygon(array) {
    return Geometry.createCollection(array.map(parseCoords.polygon));
  },
  /**
   * @param {Readonly<JSONArray>} array
   * @returns {Geometry.Polygon}
   * @throws {TypeError} If `coords` is an invalid GeoJSON Box.
   */
  box(array) {
    if (array.length !== 2) {
      throw new TypeError('GeoJSON box coordinates must have 2 elements');
    }
    if (!isXY(array[0]) || !isXY(array[1])) {
      throw new TypeError('Invalid GeoJSON box coordinates');
    }
    return Geometry.createPolygonFromXYs([[
      [array[0][0], array[0][1]],
      [array[1][0], array[0][1]],
      [array[1][0], array[1][1]],
      [array[0][0], array[1][1]],
      [array[0][0], array[0][1]],
    ]]);
  },
};
const serializeCoords = {
  /**
   * @param {Readonly<Geometry.Point>} point
   * @returns {GeoJSON.Position}
   */
  Point(point) {
    return Geometry.pointToXY(point);
  },
  /**
   * @param {Readonly<Geometry.Collection<Geometry.Point>>} multiPoint
   * @returns {GeoJSON.Position[]}
   */
  MultiPoint(multiPoint) {
    return multiPoint.components.map(serializeCoords.Point);
  },
  /**
   * @param {Readonly<Geometry.Polyline>} lineString
   * @returns {GeoJSON.Position[]}
   */
  LineString(lineString) {
    return lineString.components.map(serializeCoords.Point);
  },
  /**
   * @param {Readonly<Geometry.Collection<Geometry.Polyline>>} multiLineString
   * @returns {GeoJSON.Position[][]}
   */
  MultiLineString(multiLineString) {
    return multiLineString.components.map(serializeCoords.LineString);
  },
  /**
   * @param {Readonly<Geometry.Polygon>} polygon
   * @returns {GeoJSON.Position[][]}
   */
  Polygon(polygon) {
    return polygon.components.map(serializeCoords.LineString);
  },
  /**
   * @param {Readonly<Geometry.Collection<Geometry.Polygon>>} multiPolygon
   * @returns {GeoJSON.Position[][][]}
   */
  MultiPolygon(multiPolygon) {
    return multiPolygon.components.map(serializeCoords.Polygon);
  },
};

/**
 * @param {Readonly<JSONValue> | undefined} json
 * @param {Readonly<Projection> | null} [sourceProjection]
 * @param {Readonly<Projection> | null} [targetProjection]
 * @returns {Geometry}
 * @throws {TypeError} If `json` is an invalid GeoJSON Geometry.
 */
function parseGeometry(json, sourceProjection, targetProjection) {
  if (
    !json ||
    typeof json !== 'object' ||
    Array.isArray(json) ||
    typeof json.type !== 'string'
  ) {
    throw new TypeError('Bad GeoJSON Geometry: ' + json);
  }

  if (json.type === 'GeometryCollection') {
    if (!Array.isArray(json.geometries)) {
      throw new TypeError('GeometryCollection must have geometries array');
    }
    return Geometry.createCollection(
      json.geometries.map(
        (geom) => parseGeometry(geom, sourceProjection, targetProjection)
      )
    );
  } else {
    if (!Array.isArray(json.coordinates)) {
      throw new TypeError('Geometry must have coordinates array');
    }

    const type = json.type.toLowerCase();
    const parse = (
      /** @type {{[type: string]: (coords: Readonly<JSONValue>) => Geometry}} */
      (parseCoords)[type]
    );
    if (!parse) {
      throw new TypeError('Unsupported geometry type: ' + json.type);
    }

    const coordinates = coordinatesWithProjection(
      json.coordinates,
      sourceProjection,
      targetProjection
    );
    return parse(coordinates);
  }
}

/**
 * @param {Readonly<JSONValue>} json
 * @param {Readonly<Projection> | null} [sourceProjection]
 * @param {Readonly<Projection> | null} [targetProjection]
 * @returns {Feature}
 * @throws {TypeError} If `json` is an invalid GeoJSON Feature.
 */
function parseFeature(json, sourceProjection, targetProjection) {
  if (
    !json ||
    typeof json !== 'object' ||
    Array.isArray(json) ||
    json.type !== 'Feature'
  ) {
    throw new TypeError('Bad GeoJSON Feature: ' + json);
  }

  return new Feature(
    parseGeometry(json.geometry, sourceProjection, targetProjection),
    typeof json.properties === 'object' ? json.properties : null
  );
}
/**
 * @param {Readonly<Geometry> | null} geometry
 * @param {Readonly<Projection> | null} [sourceProjection]
 * @param {Readonly<Projection> | null} [targetProjection]
 * @returns {GeoJSON.Geometry}
 */
function serializeGeometry(geometry, sourceProjection, targetProjection) {
  /** @typedef {Parameters<Values<typeof serializeCoords>>[0]} SingleGeometry */
  /** @typedef {ReturnType<Values<typeof serializeCoords>>} GeometryCoordinates */

  if (!geometry) {
    throw new TypeError('Bad GeoJSON geometry: ' + geometry);
  }

  const type = Geometry.getType(geometry);

  if (typeof type != 'string') {
    throw new TypeError('Bad GeoJSON geometry - no type: ' + geometry);
  }
  if (!GEOMETRY_TYPES.has(type)) {
    throw new TypeError('Unsupported GeoJSON geometry type: ' + type);
  }

  if (type === 'GeometryCollection') {
    const geometries = (
      /** @type {Geometry.Collection} */ (geometry).components.map(
        (geom) => serializeGeometry(geom, sourceProjection, targetProjection)
      )
    );
    return {type, geometries};
  } else {
    // TODO: [Antimeridian Cutting](https://www.rfc-editor.org/rfc/rfc7946#section-3.1.9)
    const serialize = (
      /** @type {(geometry: SingleGeometry) => GeometryCoordinates} */
      (serializeCoords[type])
    );
    const coordinates = changeCoordinatesProjection(
      serialize(/** @type {SingleGeometry} */ (geometry)),
      sourceProjection,
      targetProjection
    );
    // @ts-expect-error ZKH: Don't know why, maybe a bug in TS?
    return {type, coordinates};
  }
}

export const GeoJSON = {
  /** @typedef {import("./gis_projection.js").Projection.ProjectionLike} ProjectionLike */

  /**
   * In loose mode (`options.loose` set to `true`), LineString geometries with
   * a single point will be parsed as a LineString with two points with the same
   * coordinates.
   * A warning is also logged in the console.
   *
   * @param {Readonly<JSONValue>} json
   * @param {{
   *   sourceProjection?: ProjectionLike,
   *   targetProjection?: ProjectionLike,
   *   loose?: boolean,
   * }} [options]
   * @returns {Feature[]}
   * @throws {TypeError} If `json` is an invalid GeoJSON.
   */
  getFeatures(json, options) {
    if (!json || typeof json !== 'object' || Array.isArray(json)) {
      throw new TypeError('Bad GeoJSON: ' + json);
    }
    if (typeof json.type !== 'string') {
      throw new TypeError('Bad GeoJSON - no type: ' + json);
    }

    let sourceProjection = (
      options?.sourceProjection
      ? Projection.from(options.sourceProjection)
      : null
    );
    let targetProjection = (
      options?.targetProjection
      ? Projection.from(options.targetProjection)
      : null
    );
    if (sourceProjection === WGS84) {
      sourceProjection = null;
    }
    if (targetProjection === WGS84) {
      targetProjection = null;
    }
    if (sourceProjection === targetProjection) {
      sourceProjection = null;
      targetProjection = null;
    }

    strictParsing = !options?.loose;
    switch (json.type) {
    case 'FeatureCollection':
      return (
        Array.isArray(json.features)
        ? json.features.map(
          (feature) => parseFeature(feature, sourceProjection, targetProjection)
        )
        : []
      );
    case 'Feature':
      return [parseFeature(json, sourceProjection, targetProjection)];
    default:
      return [
        new Feature(parseGeometry(json, sourceProjection, targetProjection)),
      ];
    }
  },
  /**
   * In loose mode (`options.loose` set to `true`), LineString geometries with
   * a single point will be parsed as a LineString with two points with the same
   * coordinates.
   * A warning is also logged in the console.
   *
   * @param {Readonly<GeoJSON.Geometry> | null} geometry
   * @param {{
   *   sourceProjection?: ProjectionLike,
   *   targetProjection?: ProjectionLike,
   *   loose?: boolean,
   * }} [options]
   * @returns {Geometry | null}
   */
  parseGeometry(geometry, options) {
    let sourceProjection = (
      options?.sourceProjection
      ? Projection.from(options.sourceProjection)
      : null
    );
    let targetProjection = (
      options?.targetProjection
      ? Projection.from(options.targetProjection)
      : null
    );
    if (sourceProjection === WGS84) {
      sourceProjection = null;
    }
    if (targetProjection === WGS84) {
      targetProjection = null;
    }
    if (sourceProjection === targetProjection) {
      sourceProjection = null;
      targetProjection = null;
    }

    strictParsing = !options?.loose;
    return parseGeometry(
      /** @type {IndexOptional<typeof geometry>} */ (geometry),
      sourceProjection,
      targetProjection
    );
  },
  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {GeoJSON.Geometry}
   */
  serializeGeometry(geometry) {
    return serializeGeometry(
      Geometry.rewind(geometry, false /*counterclockwise*/)
    );
  },
};
