/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {OpenLayers} from '../../vendors/openlayers.js';
import {
  DEG_IN_RAD,
  $,
  parseDecimal,
  numericComparator,
  bezierPathToXYs,
} from '../utils.js';
import {Bounds} from './gis_base.js';
import {Geography} from './gis_geography.js';
import {Projection} from './gis_projection.js';

// #region Exported typedefs

/** @typedef {import('.').Point} Geometry.Point */
/** @typedef {import('.').LineString} Geometry.LineString */
/** @typedef {import('.').LinearRing} Geometry.LinearRing */
/** @typedef {import('.').Polygon} Geometry.Polygon */
/** @typedef {import('.').MultiLineString} Geometry.MultiLineString */
/** @typedef {import('.').MultiPolygon} Geometry.MultiPolygon */
/** @typedef {import('.').Polyline} Geometry.Polyline */
/**
 * @template {Geometry} [T = Geometry]
 * @typedef {import('.').GeometryCollection<T>} Geometry.Collection
 */
/** @typedef {import('.').Geometry} Geometry */

/**
 * @template T
 * @typedef Geometry.IterGeometry
 * @prop {MarkedLike<T, Point | Polyline | Polygon>} geometry
 * @prop {string} indexPath
 */
/**
 * @template T
 * @typedef Geometry.IterDeepGeometry
 * @prop {MarkedLike<T, Geometry>} geometry
 * @prop {string} indexPath
 */
/**
 * @template T
 * @typedef Geometry.IterEdge
 * @prop {MarkedLike<T, XY>} from
 * @prop {MarkedLike<T, XY>} to
 * @prop {string} indexPath
 */
/**
 * @template T
 * @typedef Geometry.IterVertex
 * @prop {MarkedLike<T, XY>} vertex
 * @prop {string} indexPath
 */

// #endregion

const {
  Point,
  LineString,
  LinearRing,
  Polygon,
  MultiLineString,
  MultiPolygon,
  Collection: GeometryCollection,
} = OpenLayers.Geometry;

/** @type {GeometryFunctions} */
export let Geometry;

/** @typedef {typeof _Geometry} _GeometryFunctions */
const _Geometry = {
  // #region Private typedefs

  /** @typedef {Geometry.Point} Point */
  /** @typedef {Geometry.LineString} LineString */
  /** @typedef {Geometry.LinearRing} LinearRing */
  /** @typedef {Geometry.Polygon} Polygon */
  /** @typedef {Geometry.MultiLineString} MultiLineString */
  /** @typedef {Geometry.MultiPolygon} MultiPolygon */
  /** @typedef {Geometry.Polyline} Polyline */
  /**
   * @template {Geometry} [T = Geometry]
   * @typedef {Geometry.Collection<T>} GeometryCollection
   */
  /** @typedef {import('../utils.js').BezierPath} BezierPath */

  // #endregion

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @returns {G}
   */
  clone(geometry) {
    let constructor;
    switch (geometry.CLASS_NAME) {
    case 'OpenLayers.Geometry.LinearRing':
      constructor = LinearRing;
      break;
    case 'OpenLayers.Geometry.Polygon':
      constructor = Polygon;
      break;
    case 'OpenLayers.Geometry.MultiPolygon':
      constructor = MultiPolygon;
      break;
    case 'OpenLayers.Geometry.Collection':
      constructor = GeometryCollection;
      break;
    default:
      return /** @type {G} */ (geometry.clone());
    }
    const clone = /** @type {G & GeometryCollection} */ (new constructor());
    const components = /** @type {GeometryCollection} */ (geometry).components;
    const length = components.length - (geometry instanceof LinearRing ? 1 : 0);
    for (let i = 0; i < length; i++) {
      clone.addComponent(Geometry.clone($(components[i])), i);
    }
    return clone;
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {'Point' | 'MultiPoint' | 'LineString' | 'MultiLineString'
   *           | 'Polygon' | 'MultiPolygon' | 'GeometryCollection' | null}
   */
  getType(geometry) {
    switch (geometry.CLASS_NAME) {
    case 'OpenLayers.Geometry.Point':
      return 'Point';
    case 'OpenLayers.Geometry.MultiPoint':
      return 'MultiPoint';
    case 'OpenLayers.Geometry.LineString':
    case 'OpenLayers.Geometry.LinearRing':
      return 'LineString';
    case 'OpenLayers.Geometry.MultiLineString':
      return 'MultiLineString';
    case 'OpenLayers.Geometry.Polygon':
      return 'Polygon';
    case 'OpenLayers.Geometry.MultiPolygon':
      return 'MultiPolygon';
    case 'OpenLayers.Geometry.Collection':
      return 'GeometryCollection';
    default:
      return null;
    }
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @returns {MarkedLike<G, Bounds>}
   */
  getBounds(geometry) {
    return geometry.getBounds();
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @param {Readonly<Projection>} projection
   * @returns {Meters}
   */
  getGeodesicLength(geometry, projection) {
    if (Geometry.isPoint(geometry)) {
      return 0;
    }
    if (Geometry.isPolyline(geometry)) {
      const points = geometry.components;
      if (points.length < 2) {
        return 0;
      }

      const lonLatProjection = Projection.fromCode('EPSG:4326');
      projection = Projection.from(projection) ?? lonLatProjection;

      let length = 0;
      /** @type {LonLat} */
      let ll;
      let ll1 = /** @type {LonLat} */ ([0, 0]);
      let ll2 = /** @type {LonLat} */ ([0, 0]);

      ll2 = Geometry.pointToXY($(points[0]), ll2);
      if (projection !== lonLatProjection) {
        Projection.unprojectXY(projection, /** @type {XY} */ (ll2), ll2);
      }

      for (let i = 1, len = points.length; i < len; i++) {
        ll = ll2;
        ll2 = Geometry.pointToXY($(points[i]), ll1);
        ll1 = ll;
        if (projection !== lonLatProjection) {
          Projection.unprojectXY(projection, /** @type {XY} */ (ll2), ll2);
        }
        length += Geography.geodesicDistance(ll1, ll2);
      }
      return length;
    }

    return /** @type {GeometryCollection} */ (geometry).components.reduce(
      (totalLength, geom) => (
        totalLength + Geometry.getGeodesicLength(geom, projection)
      ),
      0
    );
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @param {Readonly<Projection>} projection
   * @returns {MetersSq}
   */
  getGeodesicArea(geometry, projection) {
    if (Geometry.isPoint(geometry)) {
      return 0;
    }

    /** @type {number} */
    let area;
    /** @type {number} */
    let i;
    /** @type {number} */
    let len;

    if (Geometry.isPolygon(geometry)) {
      const rings = geometry.components;

      area = 0;
      if (rings.length > 0) {
        area += Geometry.getGeodesicArea($(rings[0]), projection);
        for (i = 1, len = rings.length; i < len; i++) {
          area -= Geometry.getGeodesicArea($(rings[i]), projection);
        }
      }
      return Math.abs(area);
    }
    if (geometry instanceof LinearRing) {
      const points = geometry.components;
      if (points.length < 3) {
        return 0;
      }

      const lonLatProjection = Projection.fromCode('EPSG:4326');
      projection = Projection.from(projection) ?? lonLatProjection;

      /** @type {LonLat} */
      let ll;
      let ll1 = /** @type {LonLat} */ ([0, 0]);
      let ll2 = /** @type {LonLat} */ ([0, 0]);

      ll2 = Geometry.pointToXY($(points[0]), ll2);
      if (projection !== lonLatProjection) {
        Projection.unprojectXY(projection, /** @type {XY} */ (ll2), ll2);
      }

      area = 0;
      for (i = 1, len = points.length; i < len; i++) {
        ll = ll2;
        ll2 = Geometry.pointToXY($(points[i]), ll1);
        ll1 = ll;
        if (projection !== lonLatProjection) {
          Projection.unprojectXY(projection, /** @type {XY} */ (ll2), ll2);
        }
        area += (
          (ll2[0] - ll1[0]) * DEG_IN_RAD *
          (2 + Math.sin(ll1[1] * DEG_IN_RAD) + Math.sin(ll2[1] * DEG_IN_RAD))
        );
      }

      return (
        Math.abs(area * Geography.EARTH_RADIUS * Geography.EARTH_RADIUS / 2)
      );
    }

    return /** @type {GeometryCollection} */ (geometry).components.reduce(
      (totalArea, geom) => (
        totalArea + Geometry.getGeodesicArea(geom, projection)
      ),
      0
    );
  },

  /**
   * @template {Readonly<Geometry>} G
   * @template {MarkedLike<G, XY>} X
   * @param {G} geometry
   * @param {X} [xy]
   * @returns {X | null}
   */
  getCentroidXY(geometry, xy) {
    const centroid = geometry.getCentroid();
    return centroid ? Geometry.pointToXY(centroid, xy) : null;
  },

  /**
   * @template {Geometry} G
   * @param {G} geometry
   * @param {number} dx
   * @param {number} dy
   * @returns {G}
   */
  translate(geometry, dx, dy) {
    geometry.move(dx, dy);
    return geometry;
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @param {number} dx
   * @param {number} dy
   * @returns {G}
   */
  translated(geometry, dx, dy) {
    return Geometry.translate(Geometry.clone(geometry), dx, dy);
  },

  /**
   * @template {Geometry} G
   * @param {G} geometry
   * @param {Degrees} angleDeg
   * @param {MarkedLike<G, Readonly<XY>>} centerXY
   * @returns {G}
   */
  rotate(geometry, angleDeg, centerXY) {
    const center = Geometry.createPointFromXY(centerXY);
    geometry.rotate(angleDeg, center);
    return geometry;
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @param {Degrees} angleDeg
   * @param {MarkedLike<G, Readonly<XY>>} centerXY
   * @returns {G}
   */
  rotated(geometry, angleDeg, centerXY) {
    return Geometry.rotate(Geometry.clone(geometry), angleDeg, centerXY);
  },

  /**
   * @template {Geometry} G
   * @param {G} geometry
   * @param {number} sx
   * @param {number} sy
   * @param {MarkedLike<G, Readonly<XY>>} centerXY
   * @returns {G}
   */
  scale(geometry, sx, sy, centerXY) {
    const center = Geometry.createPointFromXY(centerXY);
    geometry.resize(sy, center, sx / sy);
    return geometry;
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @param {number} sx
   * @param {number} sy
   * @param {MarkedLike<G, Readonly<XY>>} center
   * @returns {G}
   */
  scaled(geometry, sx, sy, center) {
    return Geometry.scale(Geometry.clone(geometry), sx, sy, center);
  },

  /**
   * @param {number} x
   * @param {number} y
   * @returns {Point}
   */
  createPoint(x, y) {
    return new Point(x, y);
  },

  /**
   * @template {Readonly<XY>} X
   * @param {X} xy
   * @returns {MarkedLike<X, Point>}
   */
  createPointFromXY([x, y]) {
    return Geometry.createPoint(x, y);
  },

  /**
   * @param {number} x
   * @param {number} y
   * @returns {Point}
   */
  createVirtualPoint(x, y) {
    const point = Geometry.createPoint(x, y);
    point.__virtual = true;
    return point;
  },

  /**
   * @template {Readonly<XY>} X
   * @param {X} xy
   * @returns {MarkedLike<X, Point>}
   */
  createVirtualPointFromXY([x, y]) {
    return Geometry.createVirtualPoint(x, y);
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {geometry is Point}
   */
  isPoint(geometry) {
    return geometry instanceof Point;
  },

  /**
   * @template {Point} G
   * @template {MarkedLike<G, XY>} X
   * @param {Readonly<G>} point
   * @param {X} [xy]
   * @returns {X}
   */
  pointToXY(point, xy) {
    if (!xy || xy.length === undefined) {
      xy = /** @type {X} */ ([0, 0]);
    }
    xy[0] = point.x;
    xy[1] = point.y;
    return xy;
  },

  /**
   * @template {Point} G
   * @param {G} point
   * @param {MarkedLike<G, Readonly<XY>>} xy
   * @returns {void}
   */
  pointMoveTo(point, [x, y]) {
    if (point.x !== x || point.y !== y) {
      point.x = x;
      point.y = y;
      point.bounds = Bounds.createOrUpdate(point.bounds, x, y, x, y);
      point.parent?.clearBounds();
    }
  },

  /**
   * @template {XY} T
   * @param {Readonly<T[]>} xys
   * @returns {MarkedLike<T, Polyline>}
   */
  createPolylineFromXYs(xys) {
    return new LineString(xys.map(Geometry.createPointFromXY));
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {geometry is Polyline}
   */
  isPolyline(geometry) {
    return geometry instanceof LineString;
  },

  /**
   * @template {Readonly<Polyline>} G
   * @param {G} polyline
   * @returns {MarkedLike<G, XY>[]}
   */
  polylineToXYs(polyline) {
    return polyline.components.map(pointToXY);
  },

  /**
   * @param {Readonly<Polyline>} polyline
   * @returns {number}
   */
  polylineCount(polyline) {
    return polyline.components.length;
  },

  /**
   * @template {Readonly<Polyline>} G
   * @param {G} polyline
   * @returns {Generator<MarkedLike<G, XY>, void>}
   */
  *polylinePoints(polyline) {
    for (const point of polyline.components) {
      yield Geometry.pointToXY(point);
    }
  },

  /**
   * @template {XY} T
   * @param {MarkedLike<T, Polyline>} polyline
   * @param {T} xy
   * @param {number} [index]
   * @returns {void}
   */
  polylineAddPoint(polyline, xy, index = polyline.components.length) {
    polyline.addComponent(Geometry.createPointFromXY(xy), index);
  },

  /**
   * @param {Polyline} polyline
   * @param {number} fromIndex
   * @param {number} toIndex
   * @returns {void}
   */
  polylineRemovePointsInRange(polyline, fromIndex = 0, toIndex = polyline.components.length) {
    polyline.components.splice(fromIndex, toIndex - fromIndex);
    polyline.clearBounds();
  },

  /**
   * @template {Polyline} G
   * @param {G} polyline
   * @param {number} tolerance
   * @param {object} [options]
   * @param {number} [options.fromIndex = 0]
   * @param {number} [options.toIndex = Geometry.polylineCount(polyline)]
   * @param {boolean} [options.inPlace = false]
   * @returns {G}
   */
  polylineSimplify(polyline, tolerance, options = {}) {
    const {
      fromIndex = 0,
      toIndex = polyline.components.length,
      inPlace = false,
    } = options;

    const intermediatePointsCount = toIndex - fromIndex - 2;
    if (intermediatePointsCount > 0) {
      const points = polyline.components;
      /** @type {number[]} */
      const pointIndicesToKeep = [];
      _douglasPeuckerReduction(
        points,
        fromIndex,
        toIndex - 1,
        tolerance * tolerance,
        pointIndicesToKeep
      );

      if (pointIndicesToKeep.length < intermediatePointsCount) {
        if (inPlace) {
          points.splice(
            fromIndex + 1,
            intermediatePointsCount,
            ...pointIndicesToKeep.sort(numericComparator).map(
              (i) => $(points[i])
            )
          );
          polyline.clearBounds();
        } else {
          polyline = /** @type {G} */ (
            new LineString(points.slice(0, fromIndex + 1))
          );
          polyline.addComponents(
            pointIndicesToKeep.sort(numericComparator).map((i) => $(points[i]))
          );
          polyline.addComponents(points.slice(toIndex - 1));
        }
      }
    }
    return polyline;
  },

  /**
   * @template {Polyline} G
   * @param {G} polyline
   * @param {object} options
   * @param {number} options.worldLeft
   * @param {number} options.worldRight
   * @param {number} options.deltaXWrapThreshold
   * @returns {G}
   */
  polylineFixXIDL(polyline, options) {
    const {worldLeft, worldRight, deltaXWrapThreshold} = options;
    const worldWidth = worldRight - worldLeft;
    const points = polyline.components;
    let lastX = $(points[0]).x;
    let offset = 0;
    for (const point of points) {
      const deltaX = lastX - (point.x + offset);
      if (Math.abs(deltaX) >= deltaXWrapThreshold) {
        offset += Math.sign(deltaX) * worldWidth;
      }
      if (offset !== 0) {
        point.x += offset;
        point.clearBounds();
      }
      lastX = point.x;
    }
    return polyline;
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {geometry is LinearRing}
   */
  isPolygonLinearRing(geometry) {
    return geometry instanceof LinearRing;
  },

  /**
   * @template {XY} T
   * @param {Readonly<T[]> | Readonly<T[][]>} xys
   * @param {object} [style]
   * @returns {MarkedLike<T, Polygon>}
   */
  createPolygonFromXYs(xys, style) {
    const rings = (
      Array.isArray($(xys[0])[0])
      ? (/** @type {T[][]} */ (xys))
      : [(/** @type {T[]} */ (xys))]
    );
    const polygon = /** @type {Polygon} */ (new Polygon(
      rings.map(
        (ringXYs) => new LinearRing(ringXYs.map(Geometry.createPointFromXY))
      )
    ));
    if (style && typeof style === 'object') {
      polygon.style = style;
    }
    return polygon;
  },

  /**
   * @template {Readonly<BezierPath>} T
   * @param {T} bezierPath
   * @param {number} segmentsCount
   * @param {object} [options]
   * @param {Polygon['style']} [options.style]
   * @param {boolean} [options.open = false]
   * @returns {MarkedLike<T, Polygon>}
   */
  createPolygonFromBezierPath(bezierPath, segmentsCount, options) {
    const polygon = Geometry.createPolygonFromXYs(
      bezierPathToXYs(bezierPath, segmentsCount),
      options?.style
    );
    const ring = polygon.components[0];

    ring.__open = !!options?.open;
    polygon.__bezier = bezierPath;

    return polygon;
  },

  /**
   * @template {Readonly<XY>} X
   * @param {X} originXY
   * @param {number} sidesCount
   * @returns {MarkedLike<X, Polygon>}
   */
  createRegularPolygon(originXY, sidesCount) {
    const origin = Geometry.createPointFromXY(originXY);
    return /** @type {Polygon} */ (
      Polygon.createRegularPolygon(origin, 1, sidesCount, 0)
    );
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {geometry is Polygon}
   */
  isPolygon(geometry) {
    return geometry instanceof Polygon;
  },

  /**
   * @param {Readonly<Polygon>} polygon
   * @returns {number}
   */
  polygonCount(polygon) {
    return Geometry.polylineCount(polygon.components[0]);
  },

  /**
   * @param {Readonly<Polygon>} polygon
   * @returns {number}
   */
  polygonHolesCount(polygon) {
    return polygon.components.length - 1;
  },

  /**
   * @template {Readonly<Polygon>} G
   * @param {G} polygon
   * @returns {MarkedLike<G, XY>[]}
   */
  polygonToXYs(polygon) {
    return Geometry.polylineToXYs(polygon.components[0]);
  },

  /**
   * @template {Readonly<Polygon>} G
   * @param {G} polygon
   * @returns {MarkedLike<G, XY>[][]}
   */
  polygonHolesToXYs(polygon) {
    const holes = polygon.components.slice(1);
    return holes.map(Geometry.polylineToXYs);
  },

  /**
   * @template {Polygon} G
   * @param {G} polygon
   * @param {MarkedLike<G, Readonly<XY>[]>} xys
   * @returns {G}
   */
  polygonAddHoleFromXYs(polygon, xys) {
    polygon.addComponent(new LinearRing(xys.map(Geometry.createPointFromXY)));
    return polygon;
  },

  /**
   * @template {Readonly<Polygon>} G
   * @param {G} polygon
   * @param {boolean} clockwise
   * @returns {MarkedLike<G, Polygon>}
   */
  polygonRewind(polygon, clockwise) {
    return /** @type {Polygon} */ (new Polygon(
      polygon.components.map((linearRing, i) => (
        rewindLinearRing((i === 0) !== !!clockwise, linearRing)
      ))
    ));
  },

  /**
   * @template {Readonly<Polygon>} G
   * @param {G} polygon
   * @returns {Generator<MarkedLike<G, Polyline>, void>}
   */
  *polygonRings(polygon) {
    yield* polygon.components;
  },

  /**
   * @template {Geometry} G
   * @param {G[]} components
   * @returns {MarkedLike<G, GeometryCollection<G>>}
   */
  createCollection(components) {
    return /** @type {GeometryCollection<G>} */ (
      new GeometryCollection(components)
    );
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {geometry is GeometryCollection}
   */
  isCollection(geometry) {
    const type = Geometry.getType(geometry);
    return (
      type === 'GeometryCollection' ||
      type === 'MultiPoint' ||
      type === 'MultiLineString' ||
      type === 'MultiPolygon'
    );
  },

  /**
   * @template C
   * @template {Readonly<GeometryCollection<C>>} G
   * @param {G} collection
   * @returns {MarkedLike<G, C>[]}
   */
  getCollectionComponents(collection) {
    return /** @type {MarkedLike<G, C>[]} */ (collection.components);
  },
  /**
   * @template C
   * @template {GeometryCollection<C>} G
   * @param {G} collection
   * @param {MarkedLike<G, C>[]} components
   * @returns {void}
   */
  setCollectionComponents(collection, components) {
    collection.components = /** @type {C[]} */ (components);
    collection.clearBounds();
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @param {number} tolerance
   * @returns {G}
   */
  simplify(geometry, tolerance) {
    return simplifyGeometry(tolerance, geometry);
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @param {boolean} clockwise
   * @returns {G}
   */
  rewind(geometry, clockwise) {
    return rewindGeometry(clockwise, geometry);
  },

  /**
   * @template {Readonly<Bounds>} B
   * @param {B} bounds
   * @returns {MarkedLike<B, Polygon>}
   */
  fromBounds(bounds) {
    return /** @type {Polygon} */ (bounds.toGeometry());
  },

  /**
   * @param {string} wkt
   * @returns {Geometry}
   */
  fromWKT(wkt) {
    return /** @type {Geometry} */ (OpenLayers.Geometry.fromWKT(wkt));
  },

  /**
   * @param {Readonly<Geometry>} geometry
   * @returns {string}
   */
  toWKT(geometry) {
    return geometry.toString();
  },

  /**
   * @template {Geometry} G
   * @param {G | null | undefined} geometry
   * @returns {Generator<Geometry.IterGeometry<G>, void>}
   */
  *geometries(geometry) {
    yield* iterGeometries(geometry);
  },

  /**
   * @template {Geometry} G
   * @param {G | null | undefined} geometry
   * @returns {Generator<Geometry.IterDeepGeometry<G>, void>}
   */
  *deepGeometries(geometry) {
    yield* iterDeepGeometries(geometry);
  },

  /**
   * @template {Geometry} G
   * @param {G | null | undefined} geometry
   * @returns {Generator<Geometry.IterEdge<G>, void>}
   */
  *edges(geometry) {
    yield* iterEdges(geometry);
  },

  /**
   * @template {Geometry} G
   * @param {G | null | undefined} geometry
   * @returns {Generator<Geometry.IterVertex<G>, void>}
   */
  *vertices(geometry) {
    yield* iterVertices(geometry);
  },

  /**
   * @template {Readonly<GeometryCollection>} G
   * @param {G | undefined} geometry
   * @param {string} indexPath
   * @returns {MarkedLike<G, Geometry> | undefined}
   */
  getComponentAtIndexPath(geometry, indexPath) {
    return indexPath.split('.').reduce(
      getComponentAtStringIndex,
      /** @type {Geometry | undefined} */ (geometry)
    );
  },

  /**
   * @template C
   * @template {Readonly<GeometryCollection<C>>} G
   * @param {G} geometry
   * @param {number} index
   * @returns {MarkedLike<G, C> | undefined}
   */
  getComponentAtIndex(geometry, index) {
    return /** @type {MarkedLike<G, C>[]} */ (geometry.components)[index];
  },

  /**
   * @template {Geometry} C
   * @template {Readonly<GeometryCollection<C>>} G
   * @param {G | null | undefined} geometry
   * @param {MarkedLike<G, C> | null | undefined} component
   * @param {number} index
   * @returns {void}
   */
  addComponentAtIndex(geometry, component, index) {
    if (geometry?.components && component) {
      geometry.addComponent(/** @type {C} */ (component), index);
    }
  },

  /**
   * @template {Readonly<GeometryCollection>} G
   * @param {G | null | undefined} geometry
   * @param {MarkedLike<G, Geometry> | null | undefined} component
   * @param {string | null | undefined} indexPath
   * @returns {void}
   */
  addComponentAtIndexPath(geometry, component, indexPath) {
    if (!geometry || !component || !indexPath || Geometry.isPoint(geometry)) {
      return;
    }

    const parentIndexPath = indexPath.split('.');
    const index = $(parentIndexPath.pop());
    const parent = (
      Geometry.getComponentAtIndexPath(geometry, parentIndexPath.join('.'))
    );
    if (parent && !Geometry.isPoint(parent)) {
      Geometry.addComponentAtIndex(parent, component, parseDecimal(index));
    }
  },

  /**
   * @param {GeometryCollection | null | undefined} geometry
   * @param {number} index
   * @param {boolean} [force]
   * @returns {void}
   */
  removeComponentAtIndex(geometry, index, force) {
    if (!geometry?.components) {
      return;
    }
    const canRemoveComponent = (
      geometry instanceof LinearRing ? geometry.components.length > 4
      : geometry instanceof LineString ? geometry.components.length > 2
      : true
    );
    if (force || canRemoveComponent) {
      geometry.components.splice(index, 1);
      geometry.clearBounds();
    }
  },

  /**
   * @param {GeometryCollection | null | undefined} geometry
   * @param {string | null | undefined} indexPath
   * @returns {void}
   */
  removeComponentAtIndexPath(geometry, indexPath) {
    if (!geometry || !indexPath || Geometry.isPoint(geometry)) {
      return;
    }

    const parentIndexPath = indexPath.split('.');
    const index = $(parentIndexPath.pop());
    const parent = (
      Geometry.getComponentAtIndexPath(geometry, parentIndexPath.join('.'))
    );
    if (parent && !Geometry.isPoint(parent)) {
      Geometry.removeComponentAtIndex(parent, parseDecimal(index));
    }
  },
};
Geometry = /** @type {GeometryFunctions} */ (_Geometry);

/** @typedef {typeof ProjectionExtension} _ProjectionExtension */
const ProjectionExtension = {
  /**
   * @template {Geometry} G
   * @param {Readonly<Projection.ProjectionLike>} projection
   * @param {Projected<G, 'EPSG:4326'>} geometry
   * @param {boolean} inPlace
   * @return {Projected<G>}
   */
  projectGeometry(projection, geometry, inPlace = false) {
    if (!inPlace) {
      geometry = Geometry.clone(geometry);
    }

    const tmp = /** @type {LonLat} */ ([0, 0]);
    for (const {geometry: geom} of Geometry.deepGeometries(geometry)) {
      let xy = /** @type {Projected<XY>} */ (tmp);

      if (Geometry.isPolygon(geom) && geom.__bezier) {
        const [xSegments, ySegments] = geom.__bezier;
        let i = xSegments.length;
        while (i--) {
          const xSegment = $(xSegments[i]);
          const ySegment = $(ySegments[i]);
          let j = xSegment.length;

          while (j--) {
            tmp[0] = $(xSegment[j]);
            tmp[1] = $(ySegment[j]);
            xy = Projection.projectLonLat(projection, tmp, xy);
            xSegment[j] = xy[0];
            ySegment[j] = xy[1];
          }
        }
      }
      if (Geometry.isPoint(geom)) {
        xy = Projection.projectLonLat(
          projection,
          Geometry.pointToXY(geom, tmp),
          xy
        );
        geom.x = xy[0];
        geom.y = xy[1];
      }
      geom.bounds = null;
    }
    return /** @type {Projected<G>} */ (geometry);
  },

  /**
   * @template {Geometry} G
   * @param {Readonly<Projection>} projection
   * @param {Projected<G>} geometry
   * @param {boolean} inPlace
   * @return {Projected<G, 'EPSG:4326'>}
   */
  unprojectGeometry(projection, geometry, inPlace = false) {
    if (!inPlace) {
      geometry = Geometry.clone(geometry);
    }

    const tmp = /** @type {Projected<XY>} */ ([0, 0]);
    for (const {geometry: geom} of Geometry.deepGeometries(geometry)) {
      let lonLat = /** @type {LonLat} */ (tmp);

      if (Geometry.isPolygon(geom) && geom.__bezier) {
        const [xSegments, ySegments] = geom.__bezier;
        let i = xSegments.length;
        while (i--) {
          const xSegment = $(xSegments[i]);
          const ySegment = $(ySegments[i]);
          let j = xSegment.length;

          while (j--) {
            tmp[0] = $(xSegment[j]);
            tmp[1] = $(ySegment[j]);
            lonLat = Projection.unprojectXY(projection, tmp, lonLat);
            xSegment[j] = lonLat[0];
            ySegment[j] = lonLat[1];
          }
        }
      }
      if (Geometry.isPoint(geom)) {
        lonLat = Projection.unprojectXY(
          projection,
          Geometry.pointToXY(geom, tmp),
          lonLat
        );
        geom.x = lonLat[0];
        geom.y = lonLat[1];
      }
      geom.bounds = null;
    }
    return /** @type {Projected<G, 'EPSG:4326'>} */ (geometry);
  },
};
Object.assign(Projection, ProjectionExtension);

/**
 * @template {Point} G
 * @param {Readonly<G>} point
 * @returns {MarkedLike<G, XY>}
 */
function pointToXY(point) {
  return Geometry.pointToXY(point);
}

/**
 * @template {Readonly<Geometry>} G
 * @param {G | undefined} geometry
 * @param {string} index
 * @returns {MarkedLike<G, Geometry> | undefined}
 */
function getComponentAtStringIndex(geometry, index) {
  return (
    !geometry || index === ''
    ? geometry
    : (
      Geometry.isPoint(geometry)
      ? undefined
      : Geometry.getComponentAtIndex(
        /** @type {GeometryCollection} */ (geometry),
        parseDecimal(index)
      )
    )
  );
}

/**
 * @param {Readonly<Point[]>} points
 * @param {number} startIndex
 * @param {number} lastIndex
 * @param {number} toleranceSquared
 * @param {number[]} indices
 * @returns {void}
 */
function _douglasPeuckerReduction(points, startIndex, lastIndex, toleranceSquared, indices) {
  const {x: x1, y: y1} = $(points[startIndex]);
  const {x: x2, y: y2} = $(points[lastIndex]);
  const dx = x2 - x1;
  const dy = y2 - y1;
  const bottomSquared = dx * dx + dy * dy;

  let maxDistanceSquared = 0;
  let indexFarthest = startIndex;
  for (let i = startIndex; i < lastIndex; i++) {
    const {x, y} = $(points[i]);

    let distanceSquared;
    if (bottomSquared === 0) {
      distanceSquared = (x - x1) * (x - x1) + (y - y1) * (y - y1);
    } else {
      const area = Math.abs(x1 * (y2 - y) - x2 * (y1 - y) - x * dy);
      distanceSquared = area * area / bottomSquared;
    }

    if (distanceSquared > maxDistanceSquared) {
      maxDistanceSquared = distanceSquared;
      indexFarthest = i;
    }
  }

  if (maxDistanceSquared > toleranceSquared) {
    indices.push(indexFarthest);
    _douglasPeuckerReduction(points, startIndex, indexFarthest, toleranceSquared, indices);
    _douglasPeuckerReduction(points, indexFarthest, lastIndex, toleranceSquared, indices);
  }
}

/**
 * @template {Readonly<LineString>} G
 * @param {number} tolerance
 * @param {G} lineString
 * @returns {G}
 */
function simplifyLineString(tolerance, lineString) {
  return Geometry.polylineSimplify(lineString, tolerance);
}

/**
 * @template {Readonly<MultiLineString>} G
 * @param {number} tolerance
 * @param {G} multiLineString
 * @returns {G}
 */
function simplifyMultiLineString(tolerance, multiLineString) {
  return /** @type {G} */ (new MultiLineString(
    multiLineString.components.map(simplifyLineString.bind(null, tolerance))
  ));
}

/**
 * @template {Readonly<LinearRing>} G
 * @param {number} tolerance
 * @param {G} linearRing
 * @returns {G}
 */
function simplifyLinearRing(tolerance, linearRing) {
  return /** @type {G} */ (new LinearRing(
    simplifyLineString(tolerance, linearRing).components
  ));
}

/**
 * @template {Readonly<Polygon>} G
 * @param {number} tolerance
 * @param {G} polygon
 * @returns {G}
 */
function simplifyPolygon(tolerance, polygon) {
  return /** @type {G} */ (new Polygon(
    polygon.components.map(simplifyLinearRing.bind(null, tolerance))
  ));
}

/**
 * @template {Readonly<MultiPolygon>} G
 * @param {number} tolerance
 * @param {G} multiPolygon
 * @returns {G}
 */
function simplifyMultiPolygon(tolerance, multiPolygon) {
  return /** @type {G} */ (new MultiPolygon(
    multiPolygon.components.map(simplifyPolygon.bind(null, tolerance))
  ));
}

/**
 * @template {Readonly<GeometryCollection>} G
 * @param {number} tolerance
 * @param {G} collection
 * @returns {G}
 */
function simplifyCollection(tolerance, collection) {
  return /** @type {G} */ (new GeometryCollection(
    collection.components.map(simplifyGeometry.bind(null, tolerance))
  ));
}

/**
 * @template {Readonly<Geometry>} G
 * @param {number} tolerance
 * @param {G} geometry
 * @returns {G}
 */
function simplifyGeometry(tolerance, geometry) {
  switch (Geometry.getType(geometry)) {
  case 'LineString':
    if (geometry instanceof LinearRing) {
      return /** @type {G} */ (simplifyLinearRing(tolerance, geometry));
    }
    return /** @type {G} */ (
      simplifyLineString(
        tolerance,
        /** @type {Readonly<LineString>} */ (geometry)
      )
    );
  case 'MultiLineString':
    return /** @type {G} */ (
      simplifyMultiLineString(
        tolerance,
        /** @type {Readonly<MultiLineString>} */ (geometry)
      )
    );
  case 'Polygon':
    return /** @type {G} */ (
      simplifyPolygon(tolerance, /** @type {Readonly<Polygon>} */ (geometry))
    );
  case 'MultiPolygon':
    return /** @type {G} */ (
      simplifyMultiPolygon(
        tolerance,
        /** @type {Readonly<MultiPolygon>} */ (geometry)
      )
    );
  case 'GeometryCollection':
    return /** @type {G} */ (
      simplifyCollection(
        tolerance,
        /** @type {Readonly<GeometryCollection>} */ (geometry)
      )
    );
  default:
    return Geometry.clone(geometry);
  }
}

/**
 * Taken from [Mapbox's geojson-rewind](https://github.com/mapbox/geojson-rewind)
 *
 * Just in case, here is the license for geojson-rewind:
 * Copyright (c) 2020, Mapbox
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * @param {LinearRing} linearRing
 * @returns {boolean}
 */
function isClockwiseLinearRing(linearRing) {
  const ring = linearRing.components;
  let area = 0;
  let err = 0;
  for (let i = 0, len = ring.length, j = len - 1; i < len; j = i++) {
    const p1 = $(ring[j]);
    const p2 = $(ring[i]);
    const k = (p2.x - p1.x) * (p1.y + p2.y);
    const m = area + k;
    err += Math.abs(area) >= Math.abs(k) ? area - m + k : k - m + area;
    area = m;
  }
  return area + err < 0;
}

/**
 * @template {Readonly<LinearRing>} G
 * @param {boolean} clockwise
 * @param {G} linearRing
 * @returns {G}
 */
function rewindLinearRing(clockwise, linearRing) {
  if (isClockwiseLinearRing(linearRing) !== !!clockwise) {
    return /** @type {G} */ (
      new LinearRing(linearRing.components.slice().reverse())
    );
  }
  return Geometry.clone(linearRing);
}

/**
 * @template {Readonly<Polygon>} G
 * @param {boolean} clockwise
 * @param {G} polygon
 * @returns {G}
 */
function rewindPolygon(clockwise, polygon) {
  return /** @type {G} */ (Geometry.polygonRewind(polygon, clockwise));
}

/**
 * @template {Readonly<MultiPolygon>} G
 * @param {boolean} clockwise
 * @param {G} multiPolygon
 * @returns {G}
 */
function rewindMultiPolygon(clockwise, multiPolygon) {
  return /** @type {G} */ (new MultiPolygon(
    multiPolygon.components.map(rewindPolygon.bind(null, clockwise))
  ));
}

/**
 * @template {Readonly<GeometryCollection>} G
 * @param {boolean} clockwise
 * @param {G} collection
 * @returns {G}
 */
function rewindCollection(clockwise, collection) {
  return /** @type {G} */ (new GeometryCollection(
    collection.components.map(rewindGeometry.bind(null, clockwise))
  ));
}

/**
 * @template {Readonly<Geometry>} G
 * @param {boolean} clockwise
 * @param {G} geometry
 * @returns {G}
 */
function rewindGeometry(clockwise, geometry) {
  switch (Geometry.getType(geometry)) {
  case 'Polygon':
    return /** @type {G} */ (
      rewindPolygon(
        clockwise,
        /** @type {Readonly<Polygon>} */ (geometry)
      )
    );
  case 'MultiPolygon':
    return /** @type {G} */ (
      rewindMultiPolygon(
        clockwise,
        /** @type {Readonly<MultiPolygon>} */ (geometry)
      )
    );
  case 'GeometryCollection':
    return /** @type {G} */ (
      rewindCollection(
        clockwise,
        /** @type {Readonly<GeometryCollection>} */ (geometry)
      )
    );
  case 'LineString':
    if (geometry instanceof LinearRing) {
      return /** @type {G} */ (rewindLinearRing(clockwise, geometry));
    }
    /*fallthrough*/
  default:
    return Geometry.clone(geometry);
  }
}

/**
 * @template {Readonly<Geometry>} G
 * @param {G | null | undefined} geometry
 * @param {string} [indexPath='']
 * @returns {Generator<Geometry.IterGeometry<G>, void>}
 */
function* iterGeometries(geometry, indexPath = '') {
  if (geometry) {
    if (Geometry.isCollection(geometry)) {
      let i = 0;
      for (const component of geometry.components) {
        yield* iterGeometries(component, indexPath + '.' + i++);
      }
    } else {
      yield {
        geometry: /** @type {Geometry.IterGeometry<G>['geometry']} */ (geometry),
        indexPath,
      };
    }
  }
}

/**
 * @template {Readonly<Geometry>} G
 * @param {G | null | undefined} geometry
 * @param {string} [indexPath='']
 * @returns {Generator<Geometry.IterDeepGeometry<G>, void>}
 */
function* iterDeepGeometries(geometry, indexPath = '') {
  if (!geometry) {
    return;
  }

  yield {geometry, indexPath};

  if (Geometry.isPoint(geometry)) {
    return;
  }

  const geometries = (
    /** @type {GeometryCollection<MarkedLike<G, Geometry>>} */
    (geometry).components
  );
  let numVert = geometries.length;

  if (numVert > 1 && geometries[numVert - 1] === geometries[0]) {
    numVert--;
  }

  for (let i = 0; i < numVert; ++i) {
    yield* iterDeepGeometries(geometries[i], indexPath + '.' + i);
  }
}

/**
 * @template {Readonly<Geometry>} G
 * @param {G | null | undefined} geometry
 * @param {string} [indexPath='']
 * @returns {Generator<Geometry.IterEdge<G>, void>}
 */
function* iterEdges(geometry, indexPath = '') {
  if (!geometry || Geometry.isPoint(geometry)) {
    return;
  }

  if (Geometry.isPolyline(geometry)) {
    for (let i = 1, len = geometry.components.length; i < len; ++i) {
      const prevVertex = $(geometry.components[i - 1]);
      const nextVertex = $(geometry.components[i]);

      if (Geometry.isPoint(prevVertex) && Geometry.isPoint(nextVertex)) {
        yield {
          from: Geometry.pointToXY(prevVertex),
          to: Geometry.pointToXY(nextVertex),
          indexPath: indexPath + '.' + i,
        };
      }
    }
    return;
  }

  const components = (
    /** @type {GeometryCollection<MarkedLike<G, Geometry>>} */
    (geometry).components
  );
  for (let i = 0; i < components.length; ++i) {
    yield* iterEdges(components[i], indexPath + '.' + i);
  }
}

/**
 * @template {Readonly<Geometry>} G
 * @param {G | null | undefined} geometry
 * @param {string} [indexPath='']
 * @returns {Generator<Geometry.IterVertex<G>, void>}
 */
function* iterVertices(geometry, indexPath = '') {
  for (
    const {geometry: geom, indexPath: ip}
    of iterDeepGeometries(geometry, indexPath)
  ) {
    if (Geometry.isPoint(geom)) {
      yield {vertex: Geometry.pointToXY(geom), indexPath: ip};
    }
  }
}
