/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$, modulo, idlRangeWrapCount} from '../utils.js';
import {Bounds} from './gis_base.js';
import {Geometry} from './gis_geometry.js';

/** @typedef {{x: number, y: number}} Point */
/** @typedef {{components: Point[]}} Polyline */
/** @typedef {Geometry.Polygon} Polygon */

/**
 * @param {boolean | 0} intersects0
 * @param {boolean | 0} intersects1
 * @returns {boolean | 0}
 */
function and(intersects0, intersects1) {
  return (
    intersects0 !== false &&
    intersects1 !== false &&
    (intersects0 === 0 || intersects1 === 0 ? 0 : true)
  );
}
/**
 * @param {boolean | 0} intersects0
 * @param {boolean | 0} intersects1
 * @returns {boolean | 0}
 */
function or(intersects0, intersects1) {
  return (
    intersects0 ||
    intersects1 ||
    (intersects0 === 0 || intersects1 === 0 ? 0 : false)
  );
}

/**
 * @param {number} x
 * @param {number} x0
 * @param {number} x1
 * @param {number} worldWidth
 * @returns {boolean | 0} Returns true if the point is on the line, 0 if it's
 * on an either extrema and false otherwise.
 */
function pointIntersectsLine1D(x, x0, x1, worldWidth) {
  if (worldWidth) {
    if (x1 - x0 < worldWidth) {
      x = modulo(x, worldWidth);
      x0 = modulo(x0, worldWidth);
      x1 = modulo(x1, worldWidth);
      if (x === x0 || x === x1) {
        return 0;
      }
      return (
        (x0 > x1 && (x > x0 || x < x1)) ||
        (x0 < x1 && x > x0 && x < x1)
      );
    }
    return true;
  }
  if (x === x0 || x === x1) {
    return 0;
  }
  return x0 < x && x < x1;
}

/**
 * @param {Point} point
 * @param {Point} p0
 * @param {Point} p1
 * @param {number} worldWidth
 * @returns {boolean | 0} Returns true if the point is on the line, 0 if it's
 * on an either extrema and false otherwise.
 */
function pointIntersectsLine({x, y}, p0, p1, worldWidth) {
  let x0 = p0.x;
  let y0 = p0.y;
  let x1 = p1.x;
  let y1 = p1.y;

  if ((y < y0 && y < y1) || (y > y0 && y > y1)) {
    return false;
  }
  if (y0 === y1) {
    if (y0 === y) {
      return (
        x0 > x1
        ? pointIntersectsLine1D(x, x1, x0, worldWidth)
        : pointIntersectsLine1D(x, x0, x1, worldWidth)
      );
    }
    return false;
  }

  let xi = (x1 - x0) / (y1 - y0) * (y - y0) + x0;
  if (worldWidth) {
    x = modulo(x, worldWidth);
    xi = modulo(xi, worldWidth);
  }
  return (
    x === xi
    ? (y === y0 || y === y1 ? 0 : true)
    : false
  );
}

/**
 * @param {Point} p11
 * @param {Point} p12
 * @param {Point} p21
 * @param {Point} p22
 * @param {number} worldWidth
 * @returns {boolean | 0} Returns true if the lines intersect, 0 if they touch
 * and false otherwise.
 */
function lineIntersectsLine(p11, p12, p21, p22, worldWidth) {
  let px = p11.x;
  let py = p11.y;
  let prx = p12.x;
  let pry = p12.y;
  let qx = p21.x;
  let qy = p21.y;
  let qsx = p22.x;
  let qsy = p22.y;

  if (
    (py < qy && py < qsy && pry < qy && pry < qsy) ||
    (py > qy && py > qsy && pry > qy && pry > qsy)
  ) {
    // vertical bounds don't intersect
    return false;
  }

  const rx = prx - px;
  const ry = pry - py;
  const sx = qsx - qx;
  const sy = qsy - qy;
  const rs = rx * sy - ry * sx;

  if (rs === 0) {
    // lines have the same slope, do they touch?
    return (
      (
        pointIntersectsLine(p11, p21, p22, worldWidth) !== false ||
        pointIntersectsLine(p12, p21, p22, worldWidth) !== false
      )
      ? 0
      : false
    );
  }
  if (!worldWidth) {
    /**
     * Uses the vector cross product approach described by Gareth Rees on
     * StackOverflow: http://stackoverflow.com/a/565282/786339
     */
    const qpx = qx - px;
    const qpy = qy - py;
    const qps = qpx * sy - qpy * sx;
    const qpr = qpx * ry - qpy * rx;
    if (
      (rs > 0 && qpr > 0 && qpr < rs && qps > 0 && qps < rs) ||
      (rs < 0 && qpr < 0 && qpr > rs && qps < 0 && qps > rs)
    ) {
      return true;
    }
    if (
      (rs > 0 && qpr >= 0 && qpr <= rs && qps >= 0 && qps <= rs) ||
      (rs < 0 && qpr <= 0 && qpr >= rs && qps <= 0 && qps >= rs)
    ) {
      return 0;
    }
    return false;
  }

  // [iy, jy] is the intersection of [py, pry] with [qy, qsy]
  const iy = Math.max(Math.min(py, pry), Math.min(qy, qsy));
  const jy = Math.min(Math.max(py, pry), Math.max(qy, qsy));

  // find the Xs corresponding to iy and jy for both lines
  // if a line is horizontal, its Xs remain unchanged
  const pix = ry ? (iy - py) * rx / ry + px : px;
  const pjx = ry ? (jy - py) * rx / ry + px : prx;
  const qix = sy ? (iy - qy) * sx / sy + qx : qx;
  const qjx = sy ? (jy - qy) * sx / sy + qx : qsx;

  // find the min and max Xs for both lines
  const p0x = Math.min(pix, pjx);
  const p1x = Math.max(pix, pjx);
  const q0x = Math.min(qix, qjx);
  const q1x = Math.max(qix, qjx);

  const sameSlopeSign = (pix < pjx && qix < qjx) || (pix > pjx && qix > qjx);
  if (sameSlopeSign) {
    // jy -------------p1--q1---------
    //                /, '
    //            , '/
    // iy -----q0---p0----------------
    return or(
      // both p0x and p1x in [q0x, q1x]
      and(
        pointIntersectsLine1D(p0x, q0x, q1x, worldWidth),
        pointIntersectsLine1D(p1x, q0x, q1x, worldWidth)
      ),
      // or both q0x and q1x in [p0x, p1x]
      and(
        pointIntersectsLine1D(q0x, p0x, p1x, worldWidth),
        pointIntersectsLine1D(q1x, p0x, p1x, worldWidth)
      )
    );
  } else {
    // jy ------q0--------p1----------
    //            `.   ,
    //            , `.
    // iy ----p0------q1--------------
    return or(
      // at least one of p0x and p1x in [q0x, q1x]
      or(
        pointIntersectsLine1D(p0x, q0x, q1x, worldWidth),
        pointIntersectsLine1D(p1x, q0x, q1x, worldWidth)
      ),
      // or at least one of q0x and q1x in [p0x, p1x]
      or(
        pointIntersectsLine1D(q0x, p0x, p1x, worldWidth),
        pointIntersectsLine1D(q1x, p0x, p1x, worldWidth)
      )
    );
  }
}

/**
 * Detect if a point lies inside a polygon.
 *
 * * Works on polygons with holes
 * * Works with degenerate/self-intersecting polygons
 *
 * Adapted from Rowan Winsemius's point-in-polygon-hao
 * See https: https://github.com/rowanwins/point-in-polygon-hao
 *
 * Just in case, here is the license for point-in-polygon-hao:
 * The MIT License (MIT)
 *
 * Copyright (c) 2021 Rowan Winsemius
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * @param {number} x
 * @param {number} y
 * @param {Geometry.Polygon} polygon
 * @returns {boolean | 0} Returns true if point is inside polygon, 0 if on the
 * edge and false otherwise.
 */
function _pointInPolygon(x, y, polygon) {
  let ii;
  let k = 0;
  let f;
  let x1;
  let y1;
  let x2;
  let y2;
  let currentP;
  let nextP;

  const numContours = polygon.components.length;
  for (let i = 0; i < numContours; i++) {
    const contour = $(polygon.components[i]).components;
    const contourLen = contour.length;

    currentP = $(contour[0]);

    x1 = currentP.x - x;
    y1 = currentP.y - y;

    for (ii = 1; ii < contourLen; ii++) {
      nextP = $(contour[ii]);

      x2 = nextP.x - x;
      y2 = nextP.y - y;

      if (y1 === 0 && y2 === 0) {
        if ((x2 <= 0 && x1 >= 0) || (x1 <= 0 && x2 >= 0)) {
          return 0;
        }
      } else if ((y2 >= 0 && y1 < 0) || (y2 < 0 && y1 >= 0)) {
        f = x1 * y2 - x2 * y1;
        if (f === 0) {
          return 0;
        }
        if ((f > 0 && y1 <= 0 && y2 > 0) || (f < 0 && y1 > 0 && y2 <= 0)) {
          k++;
        }
      }
      currentP = nextP;
      x1 = x2;
      y1 = y2;
    }
  }

  return k % 2 !== 0;
}

/**
 * @param {Point} point1
 * @param {Point} point2
 * @param {number} [worldWidth]
 * @returns {boolean}
 */
function pointIntersectsPoint(point1, point2, worldWidth) {
  if (point1.y !== point2.y) {
    return false;
  }

  let x1 = point1.x;
  let x2 = point2.x;
  if (worldWidth) {
    x1 = modulo(x1, worldWidth);
    x2 = modulo(x2, worldWidth);
  }
  return x1 === x2;
}

/**
 * @param {Point} point
 * @param {Polyline} polyline
 * @param {number} [worldWidth]
 * @returns {boolean}
 */
function pointIntersectsPolyline(point, polyline, worldWidth = 0) {
  for (let i = 1, len = polyline.components.length; i < len; i++) {
    if (
      pointIntersectsLine(
        point,
        $(polyline.components[i - 1]),
        $(polyline.components[i]),
        worldWidth
      ) !== false
    ) {
      return true;
    }
  }
  return false;
}

/**
 * @param {Point} point
 * @param {Polygon} polygon
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if point is inside polygon, 0 if on the
 * edge and false otherwise.
 */
function pointIntersectsPolygon(point, polygon, worldWidth) {
  let x = point.x;
  const y = point.y;

  if (!worldWidth) {
    if (polygon.bounds) {
      const {left, bottom, right, top} = polygon.bounds;
      if (x < left || x > right || y < bottom || y > top) {
        return false;
      }
    }
    return _pointInPolygon(x, y, polygon);
  }

  const bounds = Geometry.getBounds(polygon);
  const {left, bottom, right, top} = bounds;

  if (y < bottom || y > top) {
    return false;
  }

  const lwa = Math.floor(left / worldWidth);
  const rwa = Math.floor(right / worldWidth);
  const worldsAwayOffset = lwa * worldWidth;
  x = modulo(x - worldsAwayOffset, worldWidth) + worldsAwayOffset;

  let wc = idlRangeWrapCount(lwa, rwa);

  /** @type {boolean | 0} */
  let intersects = false;
  do {
    intersects = or(intersects, _pointInPolygon(x, y, polygon));
    if (intersects) {
      return true;
    }
    x += worldWidth;
  } while (wc--);
  return intersects;
}

/**
 * @param {Point} point
 * @param {Bounds} bounds
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if point is inside bounds, 0 if on the
 * edge and false otherwise.
 */
function pointIntersectsBounds(point, bounds, worldWidth) {
  const {left, bottom, right, top} = bounds;
  let x = point.x;
  const y = point.y;

  if (y < bottom || y > top) {
    return false;
  }
  if (worldWidth) {
    const worldLeft = (left + right - worldWidth) / 2;
    x = modulo(x - worldLeft, worldWidth) + worldLeft;
  }
  if (x < left || x > right) {
    return false;
  }
  if (x === left || x === right || y === bottom || y === top) {
    return 0;
  }
  return true;
}

/**
 * @param {Polyline} polyline1
 * @param {Polyline} polyline2
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if the polylines intersect, 0 if they
 * touch and false otherwise.
 */
function polylineIntersectsPolyline(polyline1, polyline2, worldWidth = 0) {
  let p11;
  let p12;
  let p21;
  let p22;
  let i1;
  let len1;
  let i2;
  let len2;
  /** @type {boolean | 0} */
  let intersects = false;
  for (i1 = 1, len1 = polyline1.components.length; i1 < len1; i1++) {
    p11 = $(polyline1.components[i1 - 1]);
    p12 = $(polyline1.components[i1]);
    for (i2 = 1, len2 = polyline2.components.length; i2 < len2; i2++) {
      p21 = $(polyline2.components[i2 - 1]);
      p22 = $(polyline2.components[i2]);
      intersects = (
        or(intersects, lineIntersectsLine(p11, p12, p21, p22, worldWidth))
      );
      if (intersects) {
        return intersects;
      }
    }
  }
  return intersects;
}

/**
 * @param {Polyline} polyline
 * @param {Polygon} polygon
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if polyline intersects polygon, 0 if it
 * touches it and false otherwise.
 */
function polylineIntersectsPolygon(polyline, polygon, worldWidth) {
  /** @type {boolean | 0} */
  let intersects = false;
  for (const point of polyline.components) {
    intersects = (
      or(intersects, pointIntersectsPolygon(point, polygon, worldWidth))
    );
    if (intersects) {
      return intersects;
    }
  }
  for (const ring of polygon.components) {
    intersects = (
      or(intersects, polylineIntersectsPolyline(polyline, ring, worldWidth))
    );
    if (intersects) {
      return intersects;
    }
  }
  return intersects;
}

/**
 * @param {Polyline} polyline
 * @param {Bounds} bounds
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if the polyline intersects bounds, 0 if
 * they touch and false otherwise.
 */
function polylineIntersectsBounds(polyline, bounds, worldWidth) {
  /** @type {boolean | 0} */
  let intersects = false;
  for (const point of polyline.components) {
    intersects = (
      or(intersects, pointIntersectsBounds(point, bounds, worldWidth))
    );
    if (intersects) {
      return intersects;
    }
  }

  const {left, bottom, right, top} = bounds;
  const boundsContour = {
    components: [
      {x: left, y: bottom},
      {x: right, y: bottom},
      {x: right, y: top},
      {x: left, y: top},
    ],
  };
  return or(
    intersects,
    polylineIntersectsPolyline(polyline, boundsContour, worldWidth)
  );
}

/**
 * @param {Polygon} polygon1
 * @param {Polygon} polygon2
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if the polygons intersect, 0 if they
 * touch and false otherwise.
 */
function polygonIntersectsPolygon(polygon1, polygon2, worldWidth) {
  /** @type {boolean | 0} */
  let intersects = false;
  const points = $(polygon1.components[0]).components;
  for (const point of points) {
    intersects = (
      or(intersects, pointIntersectsPolygon(point, polygon2, worldWidth))
    );
    if (intersects) {
      return intersects;
    }
  }
  for (const ring of polygon2.components) {
    intersects = (
      or(intersects, polylineIntersectsPolygon(ring, polygon1, worldWidth))
    );
    if (intersects) {
      return intersects;
    }
  }
  return intersects;
}

/**
 * @param {Polygon} polygon
 * @param {Bounds} bounds
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if the polygon intersects bounds, 0 if
 * they touch and false otherwise.
 */
function polygonIntersectsBounds(polygon, bounds, worldWidth) {
  /** @type {boolean | 0} */
  let intersects = false;
  const points = $(polygon.components[0]).components;
  for (const point of points) {
    intersects = (
      or(intersects, pointIntersectsBounds(point, bounds, worldWidth))
    );
    if (intersects) {
      return intersects;
    }
  }

  const {left, bottom, right, top} = bounds;
  const boundsContour = {
    components: [
      {x: left, y: bottom},
      {x: right, y: bottom},
      {x: right, y: top},
      {x: left, y: top},
    ],
  };
  return or(
    intersects,
    polylineIntersectsPolygon(boundsContour, polygon, worldWidth)
  );
}

const orderByType = {
  Point: 0,
  LineString: 1,
  Polygon: 2,
};
const intersectsByTypes = {
  Point: {
    Point: pointIntersectsPoint,
    LineString: pointIntersectsPolyline,
    Polygon: pointIntersectsPolygon,
    Bounds: pointIntersectsBounds,
  },
  LineString: {
    LineString: polylineIntersectsPolyline,
    Polygon: polylineIntersectsPolygon,
    Bounds: polylineIntersectsBounds,
  },
  Polygon: {
    Polygon: polygonIntersectsPolygon,
    Bounds: polygonIntersectsBounds,
  },
};

/**
 * @param {Geometry} geometry1
 * @param {Geometry} geometry2
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if the geometries intersect, 0 if they
 * touch and false otherwise.
 */
export function geometryIntersectsGeometry(geometry1, geometry2, worldWidth) {
  if (!geometry1 || !geometry2) {
    return false;
  }
  if (Geometry.isCollection(geometry1)) {
    /** @type {boolean | 0} */
    let intersects = false;
    for (const component of geometry1.components) {
      intersects = or(
        intersects,
        geometryIntersectsGeometry(component, geometry2, worldWidth)
      );
      if (intersects) {
        return true;
      }
    }
    return intersects;
  }
  if (Geometry.isCollection(geometry2)) {
    return geometryIntersectsGeometry(geometry2, geometry1, worldWidth);
  }

  const type1 = (
    /** @type {'Point' | 'LineString' | 'Polygon'} */
    (Geometry.getType(geometry1))
  );
  const type2 = (
    /** @type {'Point' | 'LineString' | 'Polygon'} */
    (Geometry.getType(geometry2))
  );
  if (orderByType[type1] <= orderByType[type2]) {
    // @ts-expect-error
    return intersectsByTypes[type1]?.[type2]?.(geometry1, geometry2, worldWidth);
  }
  // @ts-expect-error
  return intersectsByTypes[type2]?.[type1]?.(geometry2, geometry1, worldWidth);
}

/**
 * @param {Geometry} geometry
 * @param {Bounds} bounds
 * @param {number} [worldWidth]
 * @returns {boolean | 0} Returns true if the geometry intersects bounds, 0 if
 * they touch and false otherwise.
 */
export function geometryIntersectsBounds(geometry, bounds, worldWidth) {
  if (!geometry || !bounds) {
    return false;
  }
  if (Geometry.isCollection(geometry)) {
    /** @type {boolean | 0} */
    let intersects = false;
    for (const component of geometry.components) {
      intersects = (
        or(intersects, geometryIntersectsBounds(component, bounds, worldWidth))
      );
      if (intersects) {
        return true;
      }
    }
    return intersects;
  }

  const type = /** @type {'Point' | 'LineString' | 'Polygon'} */ (
    Geometry.getType(geometry)
  );
  // @ts-expect-error
  // eslint-disable-next-line new-cap
  return intersectsByTypes[type]?.Bounds(geometry, bounds, worldWidth);
}

/** @typedef {typeof GeometryExtension} _GeometryExtension */
const GeometryExtension = {
  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry1
   * @param {MarkedLike<G, Readonly<Geometry>>} geometry2
   * @param {MarkedLike<G, Readonly<Bounds>> | null} [worldBounds]
   * @returns {boolean | 0}
   */
  intersects(geometry1, geometry2, worldBounds) {
    return geometryIntersectsGeometry(
      geometry1,
      geometry2,
      worldBounds ? Bounds.getWidth(worldBounds) : 0
    );
  },

  /**
   * @template {Readonly<Geometry>} G
   * @param {G} geometry
   * @param {MarkedLike<G, Readonly<Bounds>>} bounds
   * @param {MarkedLike<G, Readonly<Bounds>> | null} [worldBounds]
   * @returns {boolean | 0}
   */
  intersectsLatLonBounds(geometry, bounds, worldBounds) {
    return geometryIntersectsBounds(
      geometry,
      bounds,
      worldBounds ? Bounds.getWidth(worldBounds) : 0
    );
  },
};
Object.assign(Geometry, GeometryExtension);
