/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {OpenLayers} from '../../vendors/openlayers.js';
import {
  DEG_IN_RAD,
  $,
  wrapAround,
  round,
  inRangeExclusive,
  inRangeInclusive,
  wrapClassMethods,
  extendClass,
  cloneBezierPath,
  Transform2D,
} from '../utils.js';

/**
 * The maximum safe value of a offsets (top, left) in CSS. This limit varies
 * between browsers. The limit is around 17 000 000, but Chromium converts big
 * numbers to the exponential notation which OpenLayers doesn't handle
 * correctly.
 *
 * @constant
 */
const MAX_OFFSET = 1_700_000;

// avoid pink tiles
OpenLayers.IMAGE_RELOAD_ATTEMPTS = 0;

// Make OpenLayers use the native isArray method
OpenLayers.Util.isArray = Array.isArray;

Object.assign(OpenLayers.String, {
  /**
   * @param {string} str
   * @param {string} sub
   * @returns {boolean}
   */
  startsWith(str, sub) {
    return str.startsWith(sub);
  },

  /**
   * @param {string} str
   * @param {string} sub
   * @returns {boolean}
   */
  contains(str, sub) {
    return str.includes(sub);
  },

  /**
   * @param {string} str
   * @returns {string}
   */
  trim(str) {
    return str.trim();
  },
});
wrapClassMethods(OpenLayers.LonLat, {
  /**
   * @this {Readonly<OpenLayers.LonLat>}
   * @returns {OpenLayers.LonLat}
   */
  clone() {
    const tmp = OpenLayers.Util.DEFAULT_PRECISION;
    OpenLayers.Util.DEFAULT_PRECISION = 0;
    const clone = new OpenLayers.LonLat(this.lon, this.lat);
    OpenLayers.Util.DEFAULT_PRECISION = tmp;
    return clone;
  },
});
wrapClassMethods(OpenLayers.Bounds, {
  /**
   * @this {Readonly<OpenLayers.Bounds>}
   * @returns {OpenLayers.Bounds}
   */
  clone() {
    const clone = new OpenLayers.Bounds();
    clone.left = this.left;
    clone.bottom = this.bottom;
    clone.right = this.right;
    clone.top = this.top;
    return clone;
  },
  /**
   * @this {Readonly<OpenLayers.Bounds>}
   * @param {Readonly<OpenLayers.Bounds>} bounds
   * @param {OpenLayers.BoundsOptions | boolean | null} [options]
   * @returns {boolean}
   */
  intersectsBounds(bounds, options) {
    if (typeof options === 'boolean') {
      options = {inclusive: options};
    }

    const {worldBounds, inclusive = true} = options || {};

    let {left: left1, bottom: bottom1, right: right1, top: top1} = this;
    let {left: left2, bottom: bottom2, right: right2, top: top2} = bounds;

    if (
      (top1 < bottom2 || top2 < bottom1) ||
      (!inclusive && (top1 === bottom2 || top2 === bottom1))
    ) {
      return false;
    }

    if (worldBounds) {
      const worldLeft = worldBounds.left;
      const worldRight = worldBounds.right;
      const worldWidth = worldBounds.getWidth();

      if (right1 - left1 < worldWidth) {
        const delta1 = Math.ceil((worldLeft - left1) / worldWidth) * worldWidth;
        left1 += delta1;
        right1 += delta1;
      } else {
        left1 = worldLeft;
        right1 = worldRight;
      }

      if (right2 - left2 < worldWidth) {
        const delta2 = Math.ceil((worldLeft - left2) / worldWidth) * worldWidth;
        left2 += delta2;
        right2 += delta2;
      } else {
        left2 = worldLeft;
        right2 = worldRight;
      }
    }

    const intersects = (
      (left1 < right2 && right1 > left2) ||
      (inclusive && (left1 === right2 || left2 === right1))
    );

    if (worldBounds && !intersects) {
      const worldRight = worldBounds.right;
      const worldWidth = worldBounds.getWidth();

      const inRange = inclusive ? inRangeExclusive : inRangeInclusive;
      const crosses1 = inRange(worldRight, left1, right1);
      const crosses2 = inRange(worldRight, left2, right2);
      if (crosses1 !== crosses2) {
        if (crosses1) {
          left2 += worldWidth;
          right2 += worldWidth;
        } else {
          left1 += worldWidth;
          right1 += worldWidth;
        }
        return (
          (left1 < right2 && right1 > left2) ||
          (inclusive && (left1 === right2 || left2 === right1))
        );
      }
    }
    return intersects;
  },
});
/** @typedef {typeof OLMapExtension} OpenLayersMapExtension */
export const OLMapExtension = {
  /**
   * @this {OpenLayers.Map}
   * @param {number} dLon
   * @returns {void}
   */
  _moveMaxExtentPixelBounds(dLon) {
    const dx = Math.round(dLon / $(this.getResolution()));
    this.minPx.x += dx;
    this.maxPx.x += dx;
  },

  /**
   * @this {OpenLayers.Map}
   * @param {number} dLon
   * @returns {void}
   */
  _moveLayerContainerOrigin(dLon) {
    this.layerContainerOrigin.lon -= dLon;
    this._moveMaxExtentPixelBounds(dLon);
  },

  /**
   * @this {OpenLayers.Map}
   * @param {OpenLayers.LonLat} center
   * @returns {void}
   */
  _updateRestrictedExtent(center) {
    if (
      !this.size ||
      !this.layerContainerOrigin ||
      !this.baseLayer?.wrapDateLine ||
      !this.restrictedExtent
    ) {
      return;
    }

    const {restrictedExtent} = this;
    const maxExtent = $(this.getMaxExtent());
    const worldWidth = maxExtent.getWidth();
    if (
      // restricted extent contains the IDL
      restrictedExtent.left < maxExtent.left ||
      restrictedExtent.right > maxExtent.right
    ) {
      // restricted extent containing left IDL (-180º)
      const leftRestrictedExtent = (
        restrictedExtent.left < maxExtent.left
        ? restrictedExtent
        : restrictedExtent.add(-worldWidth, 0)
      );
      // restricted extent containing right IDL (+180º)
      const rightRestrictedExtent = (
        restrictedExtent.left < maxExtent.left
        ? restrictedExtent.add(worldWidth, 0)
        : restrictedExtent
      );

      const deltaLeft = center.lon - leftRestrictedExtent.right;
      const deltaRight = rightRestrictedExtent.left - center.lon;

      // new center is outside one of the restricted extents
      if (deltaLeft > 0 || deltaRight > 0) {
        // switch to the restricted extent closest to the new center
        const newRestrictedExtent = (
          deltaLeft < deltaRight
          ? leftRestrictedExtent
          : rightRestrictedExtent
        );
        if (this.restrictedExtent !== newRestrictedExtent) {
          this._moveMaxExtentPixelBounds(
            newRestrictedExtent.left - this.restrictedExtent.left
          );
          this.restrictedExtent = newRestrictedExtent;
        }
      }
    }
  },

  /**
   * @this {OpenLayers.Map}
   * @param {number} dx
   * @param {number} dy
   * @returns {void}
   */
  moveByPx(dx, dy) {
    window.original();
    if (
      dx === 0 ||
      !this.baseLayer?.wrapDateLine ||
      !this.size ||
      !this.layerContainerOrigin
    ) {
      return;
    }

    const center = $(this.getCachedCenter());

    this._updateRestrictedExtent(center);

    this.baseLayer.wrapDateLine = false;
    const centerNoWrap = $(
      this.getLonLatFromViewPortPx({x: this.size.w / 2, y: this.size.h / 2})
    );
    this.baseLayer.wrapDateLine = true;

    // fix layerContainerOrigin and minPx/maxPx to stay in sync with
    // layerContainerDiv.style.left
    this._moveLayerContainerOrigin(centerNoWrap.lon - center.lon);
  },

  /**
   * @this {OpenLayers.Map}
   * @param {OpenLayers.LonLat | LonLat | null | undefined} lonlat
   * @param {number | string | null | undefined} [zoom]
   * @param {object} [options]
   * @param {boolean} [options.dragging = false]
   * @param {boolean} [options.forceZoomChange = false]
   * @returns {void}
   */
  moveTo(lonlat, zoom, options) {
    if (lonlat != null && !(lonlat instanceof OpenLayers.LonLat)) {
      lonlat = new OpenLayers.LonLat(lonlat);
    }
    if (zoom != null) {
      zoom = parseFloat(/** @type {string} */ (zoom));
      if (!this.fractionalZoom) {
        zoom = Math.round(zoom);
      }
    }
    if (this.baseLayer?.wrapDateLine) {
      const requestedZoom = zoom;
      zoom = this.adjustZoom(zoom);
      if (zoom !== requestedZoom) {
        // zoom was adjusted, so keep old lonlat to avoid panning
        lonlat = this.getCenter();
      }
    }
    if (!this.getCachedCenter() && !this.isValidLonLat(lonlat)) {
      lonlat = this.maxExtent.getCenterLonLat();
      this.center = lonlat.clone();
    }
    if (lonlat == null) {
      lonlat = this.center;
    }

    if (this.baseLayer?.wrapDateLine) {
      lonlat = lonlat.wrapDateLine(this.baseLayer.maxExtent);
    }

    const zoomChanged = (
      options?.forceZoomChange ||
      (this.isValidZoomLevel(zoom) && zoom !== this.getZoom())
    );
    const centerChanged = (
      this.isValidLonLat(lonlat) && !lonlat.equals(this.center)
    );

    this._updateRestrictedExtent(lonlat);
    if (
      this.baseLayer?.wrapDateLine &&
      !this.restrictedExtent &&
      !zoomChanged &&
      centerChanged &&
      this.center
    ) {
      const maxExtent = $(this.getMaxExtent());
      const worldWidth = maxExtent.getWidth();
      const lon = wrapAround(lonlat.lon, this.center.lon, worldWidth);
      this._moveLayerContainerOrigin(lon - lonlat.lon);
    }

    if (!zoomChanged && centerChanged && this.center) {
      // When zoom hasn't changed, OL will just slide the layerContainer using
      // CSS top and left.

      // We pre-compute their new expected values to check any of them gets
      // too big (> MAX_OFFSET).
      const originPx = this.getViewPortPxFromLonLat(this.layerContainerOrigin);
      const newPx = this.getViewPortPxFromLonLat(lonlat);

      if (
        originPx &&
        newPx &&
        (
          Math.abs(Math.round(originPx.x - newPx.x)) > MAX_OFFSET ||
          Math.abs(Math.round(originPx.y - newPx.y)) > MAX_OFFSET
        )
      ) {
        // top and/or left will get too big so to avoid the sliding mechanism,
        // we force a zoom change.
        zoom = this.getZoom();
        options ??= {};
        options.forceZoomChange = true;
      }
    }

    this.events.triggerEvent('prechangeresolution');
    window.original(lonlat, zoom, options);
  },
};
extendClass(OpenLayers.Map, OLMapExtension);
wrapClassMethods(OpenLayers.Layer, {
  /**
   * @this {OpenLayers.Layer & {__zIndex?: string, __zIndexArg?: string | number}}
   * @returns {number}
   */
  getZIndex() {
    if (!this.__zIndex) {
      this.__zIndex = /** @type {string} */ (window.original());
    }
    return /** @type {number} */ (/** @type {unknown} */ (this.__zIndex));
  },
  /**
   * @this {OpenLayers.Layer & {__zIndex?: string, __zIndexArg?: string | number}}
   * @param {string | number} zIndexArg
   * @returns {void}
   */
  setZIndex(zIndexArg) {
    if (
      (this.__zIndexArg && this.__zIndexArg === zIndexArg) ||
      (this.__zIndex && this.__zIndex === zIndexArg)
    ) {
      return;
    }
    // z-index must be an integer
    const zIndex = Math.floor(/** @type {number} */ (zIndexArg)).toString();
    if (this.__zIndex && this.__zIndex === zIndex) {
      return;
    }
    this.__zIndexArg = zIndexArg;
    this.__zIndex = zIndex;
    window.original(zIndex);
  },
});
wrapClassMethods(OpenLayers.Layer, {
  /**
   * @param {number} resolution
   * @param {boolean} [closest]
   * @returns {number}
   */
  getZoomForResolution(resolution, closest) {
    if (this.map.fractionalZoom || closest) {
      return window.original();
    }
    const resolutions = this.resolutions;
    const resolutionsCount = resolutions.length;
    let i = 0;
    while (i < resolutionsCount && round(resolutions[i] - resolution, -4) >= 0) {
      i++;
    }
    return Math.max(0, i - 1);
  },
});
wrapClassMethods(OpenLayers.Tile, {
  /**
   * @returns {boolean}
   */
  shouldDraw() {
    let shouldDraw = this.layer.displayOutsideMaxExtent;
    const maxExtent = this.layer.maxExtent;
    if (maxExtent) {
      const map = this.layer.map;
      const mapView = map.mapView;
      const mapMaxExtent = map.getMaxExtent();
      if (mapView) {
        shouldDraw ||= (
          (
            !mapMaxExtent ||
            mapView.shouldWrapDateLine() ||
            this.bounds.intersectsBounds(mapMaxExtent, {inclusive: false})
          ) &&
          this.bounds.intersectsBounds(maxExtent, {
            inclusive: false,
            worldBounds: mapView.getWorldProjectedBounds(),
          })
        );
      } else {
        shouldDraw ||= this.bounds.intersectsBounds(maxExtent, {
          inclusive: false,
          worldBounds: map.baseLayer?.wrapDateLine && map.getMaxExtent(),
        });
      }
    }
    return shouldDraw;
  },
});
extendClass(OpenLayers.Layer.Image, {
  /**
   * @public
   * @returns {Iterable<OpenLayers.Tile>}
   */
  get tiles() {
    return [this.tile];
  },
  /**
   * @public
   * @override
   * @this {OpenLayers.Layer.Image & {_tileEventsAdded?: true}}
   * @param {OpenLayers.Tile} tile
   * @returns {void}
   */
  addTileMonitoringHooks(tile) {
    if (!this._tileEventsAdded) {
      this.events.on({
        loadstart() {
          this.events.triggerEvent('tileloadstart', {tile});
        },
        loadend() {
          this.events.triggerEvent('tileloaded', {tile});
        },
        scope: this,
      });
      this._tileEventsAdded = true;
    }
  },
});
extendClass(OpenLayers.Layer.Grid, {
  /**
   * @public
   * @returns {Iterable<OpenLayers.Tile>}
   */
  get tiles() {
    const layer = this;
    return (function*() {
      for (const row of layer.grid) {
        yield* row;
      }
    })();
  },

  /**
   * @override
   * @this {OpenLayers.Layer.Grid}
   * @param {OpenLayers.Bounds} bounds
   * @returns {void}
   *
   * @note The body of this method is copied as is from OpenLayers, except for
   * the marked modification to fix the placement of tiles.
   * @note `@ts-ignore` comments are sprinkled all over because there's no way to
   * ignore errors for a block of code.
   * See [TypeScript#19573](https://github.com/Microsoft/TypeScript/issues/19573).
   */
  initGriddedTiles(bounds) {
    /* eslint-disable */
    // @ts-ignore
    this.clearTileQueue();

    // work out mininum number of rows and columns; this is the number of
    // tiles required to cover the viewport plus at least one for panning

    var viewSize = this.map.getSize();
    var minRows = Math.ceil(viewSize.h/this.tileSize.h) +
                  Math.max(1, 2 * this.buffer);
    var minCols = Math.ceil(viewSize.w/this.tileSize.w) +
                  Math.max(1, 2 * this.buffer);

    // @ts-ignore
    var origin = this.getTileOrigin();
    // @ts-ignore
    var resolution = this.getServerResolution();

    // @ts-ignore
    var tileLayout = this.calculateGridLayout(bounds, origin, resolution);

    //// <Origin in OpenLayers>
    // var tileoffsetx = Math.round(tileLayout.tileoffsetx); // heaven help us
    // var tileoffsety = Math.round(tileLayout.tileoffsety);
    //// </Origin in OpenLayers>
    //// <Modified>
    // @ts-ignore
    var tileoffsetx = tileLayout.tileoffsetx;
    // @ts-ignore
    var tileoffsety = tileLayout.tileoffsety;
    //// </Modified>

    // @ts-ignore
    var tileoffsetlon = tileLayout.tileoffsetlon;
    // @ts-ignore
    var tileoffsetlat = tileLayout.tileoffsetlat;

    var tilelon = tileLayout.tilelon;
    var tilelat = tileLayout.tilelat;

    var startX = tileoffsetx;
    var startLon = tileoffsetlon;

    var rowidx = 0;

    var layerContainerDivLeft = parseInt(this.map.layerContainerDiv.style.left);
    var layerContainerDivTop = parseInt(this.map.layerContainerDiv.style.top);

    var tileData = [], center = this.map.getCenter();
    do {
        var row = this.grid[rowidx++];
        if (!row) {
            row = [];
            this.grid.push(row);
        }

        tileoffsetlon = startLon;
        tileoffsetx = startX;
        var colidx = 0;

        do {
            var tileBounds =
                new OpenLayers.Bounds(tileoffsetlon,
                                      tileoffsetlat,
                                      tileoffsetlon + tilelon,
                                      tileoffsetlat + tilelat);

            var x = tileoffsetx;
            x -= layerContainerDivLeft;

            var y = tileoffsety;
            y -= layerContainerDivTop;

            var px = new OpenLayers.Pixel(x, y);
            var tile = row[colidx++];
            if (!tile) {
                // @ts-ignore
                tile = this.addTile(tileBounds, px);
                // @ts-ignore
                this.addTileMonitoringHooks(tile);
                row.push(tile);
            } else {
                tile.moveTo(tileBounds, px, false);
            }
            var tileCenter = tileBounds.getCenterLonLat();
            tileData.push({
                tile: tile,
                // @ts-ignore
                distance: Math.pow(tileCenter.lon - center.lon, 2) +
                    // @ts-ignore
                    Math.pow(tileCenter.lat - center.lat, 2)
            });

            tileoffsetlon += tilelon;
            tileoffsetx += this.tileSize.w;
        } while ((tileoffsetlon <= bounds.right + tilelon * this.buffer)
                  || colidx < minCols);

        tileoffsetlat -= tilelat;
        tileoffsety += this.tileSize.h;
    } while((tileoffsetlat >= bounds.bottom - tilelat * this.buffer)
            || rowidx < minRows);

    //shave off exceess rows and colums
    // @ts-ignore
    this.removeExcessTiles(rowidx, colidx);

    // store the resolution of the grid
    // @ts-ignore
    this.gridResolution = this.getServerResolution();

    //now actually draw the tiles
    tileData.sort(function(a, b) {
        return a.distance - b.distance;
    });
    for (var i=0, ii=tileData.length; i<ii; ++i) {
        // @ts-ignore
        tileData[i].tile.draw();
    }
    /* eslint-enable */
  },
});
wrapClassMethods(OpenLayers.Geometry.Polygon, {
  /** @typedef {import('../utils.js').BezierPath} BezierPath */
  /** @typedef {import('./gis_geometry.js').Geometry.Point} Point */
  /** @typedef {import('./gis_geometry.js').Geometry.Polygon} Polygon */

  /**
   * @this {Readonly<Polygon>}
   * @returns {Polygon}
   */
  clone() {
    /** @type {Polygon} */
    const geom = window.original();

    if (this.__bezier) {
      geom.__bezier = cloneBezierPath(this.__bezier);
    }
    return geom;
  },
  /**
   * @this {Polygon}
   * @param {number} dx
   * @param {number} dy
   * @returns {void}
   */
  move(dx, dy) {
    window.original();

    if (this.__bezier) {
      this.__bezier = Transform2D.transformPath(
        Transform2D.translation(dx, dy),
        /** @type {BezierPath} */ (this.__bezier)
      );
    }
  },
  /**
   * @this {Polygon}
   * @param {number} angle
   * @param {Point} origin
   * @returns {void}
   */
  rotate(angle, origin) {
    window.original();

    if (this.__bezier) {
      this.__bezier = Transform2D.transformPath(
        Transform2D.concat([
          Transform2D.translation(-origin.x, -origin.y),
          Transform2D.rotation(angle * DEG_IN_RAD),
          Transform2D.translation(origin.x, origin.y),
        ]),
        /** @type {BezierPath} */ (this.__bezier)
      );
    }
  },
  /**
   * @this {Polygon}
   * @param {number} scale
   * @param {Point} origin
   * @param {number} ratio
   * @returns {Polygon}
   */
  resize(scale, origin, ratio) {
    window.original();

    if (this.__bezier) {
      this.__bezier = Transform2D.transformPath(
        Transform2D.concat([
          Transform2D.translation(-origin.x, -origin.y),
          Transform2D.scale(scale * ratio, scale),
          Transform2D.translation(origin.x, origin.y),
        ]),
        /** @type {BezierPath} */ (this.__bezier)
      );
    }
    return this;
  },
});
wrapClassMethods(OpenLayers.Geometry.LinearRing, {
  /**
   * @this {Readonly<OpenLayers.Geometry.LinearRing>}
   * @returns {OpenLayers.Geometry.Point | false | null}
   */
  getCentroid() {
    const points = this.components;
    if (points.length === 0) {
      return null;
    }
    if (points.length === 1) {
      return $(points[0]).getCentroid();
    }
    const {left: minX, bottom: minY} = this.getBounds();
    let sumX = 0;
    let sumY = 0;
    for (let i = 0; i < points.length - 1; i++) {
      let {x: x0, y: y0} = $(points[i]);
      let {x: x1, y: y1} = $(points[i + 1]);
      x0 -= minX;
      x1 -= minX;
      y0 -= minY;
      y1 -= minY;
      sumX += (x0 + x1) * (x0 * y1 - x1 * y0);
      sumY += (y0 + y1) * (x0 * y1 - x1 * y0);
    }
    const area = this.getArea();
    const x = minX - sumX / (6 * area);
    const y = minY - sumY / (6 * area);
    return new OpenLayers.Geometry.Point(x, y);
  },
});
