/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {OpenLayers} from '../../vendors/openlayers.js';
import {LatLonBounds, Bounds} from './gis_base.js';

/**
 * @typedef Projection
 * @prop {import('../quantity.js').Unit} units
 * @prop {<T extends XY>(lonLat: Readonly<LonLat>, xy?: T) => T} projectLonLat
 * @prop {<T extends LonLat>(xy: Readonly<XY>, lonLat?: T) => T} unprojectXY
 */
/** @typedef {Projection | OpenLayers.Projection | string} Projection.ProjectionLike */

const MERCATOR_POLE = 20037508.34;
const PI = Math.PI;

/** @type {Projection} */
const mercatorProjection = {
  units: 'm',

  projectLonLat(lonLat, xy) {
    if (!xy || xy.length === undefined) {
      xy = /** @type {NonNullable<typeof xy>}> */ ([0, 0]);
    }
    xy[0] = lonLat[0] * MERCATOR_POLE / 180;
    xy[1] = (
      Math.log(Math.tan((90 + lonLat[1]) * PI / 360)) / PI * MERCATOR_POLE
    );
    return xy;
  },
  unprojectXY(xy, lonLat) {
    if (!lonLat || lonLat.length === undefined) {
      lonLat = /** @type {NonNullable<typeof lonLat>} */ ([0, 0]);
    }
    lonLat[0] = 180 * xy[0] / MERCATOR_POLE;
    lonLat[1] = (
      180 / PI * (2 * Math.atan(Math.exp((xy[1] / MERCATOR_POLE) * PI)) - PI / 2)
    );
    return lonLat;
  },
};
/** @type {Projection} */
const noopProjection = {
  units: 'degrees',

  projectLonLat(lonLat, xy) {
    if (!xy) {
      return /** @type {NonNullable<typeof xy>} */ (lonLat.slice());
    }
    xy[0] = lonLat[0];
    xy[1] = lonLat[1];
    return xy;
  },
  unprojectXY(xy, lonLat) {
    if (!lonLat) {
      return /** @type {NonNullable<typeof lonLat>} */ (xy.slice());
    }
    lonLat[0] = xy[0];
    lonLat[1] = xy[1];
    return lonLat;
  },
};

const projections = {
  WGS84        : noopProjection,
  'EPSG:4326'  : noopProjection,
  'EPSG:3857'  : mercatorProjection,
  'EPSG:900913': mercatorProjection,

  'urn:ogc:def:crs:OGC:1.3:CRS84': noopProjection,
  'urn:ogc:def:crs:EPSG::3857': mercatorProjection,
};

/** @type {ProjectionFunctions} */
export let Projection;

/** @typedef {typeof _Projection} _ProjectionFunctions */
const _Projection = {
  /** @type {typeof projections & Record<string, Projection>} */
  projections,

  /**
   * @param {string} projectionCode
   * @returns {Projection | undefined}
   */
  fromCode(projectionCode) {
    return Projection.projections[projectionCode];
  },
  /**
   * @param {Projection.ProjectionLike} projectionLike
   * @returns {Projection}
   * @throws {RangeError} If the requested projection is not supported.
   */
  from(projectionLike) {
    if (projectionLike instanceof OpenLayers.Projection) {
      projectionLike = projectionLike.getCode() || projectionLike.projCode;
    }
    const projection = (
      typeof projectionLike === 'string'
      ? Projection.fromCode(projectionLike)
      : projectionLike
    );
    if (!projection) {
      throw new RangeError(
        `The requested projection "${projectionLike}" is not one of the ` +
        `supported projections: ${supportedProjections()}`
      );
    }
    return projection;
  },
  /**
   * @param {Projection.ProjectionLike} projection
   * @returns {boolean}
   */
  isMercatorLike(projection) {
    return Projection.from(projection) === mercatorProjection;
  },
  /**
   * @template {XY} T
   * @param {Projection.ProjectionLike} projection
   * @param {LonLat} lonLat
   * @param {T} [xy]
   * @returns {T}
   */
  projectLonLat(projection, lonLat, xy) {
    return Projection.from(projection).projectLonLat(lonLat, xy);
  },
  /**
   * @template {LonLat} T
   * @param {Projection.ProjectionLike} projection
   * @param {XY} xy
   * @param {T} [lonLat]
   * @returns {T}
   */
  unprojectXY(projection, xy, lonLat) {
    return Projection.from(projection).unprojectXY(xy, lonLat);
  },

  /**
   * @template {Bounds} T
   * @param {Projection.ProjectionLike} projection
   * @param {LatLonBounds} latLonBounds
   * @param {T} [bounds]
   * @returns {T}
   */
  projectLatLonBounds(projection, latLonBounds, bounds) {
    const bl = LatLonBounds.getBottomLeft(latLonBounds);
    const br = LatLonBounds.getBottomRight(latLonBounds);
    const tl = LatLonBounds.getTopLeft(latLonBounds);
    const tr = LatLonBounds.getTopRight(latLonBounds);
    const ll = Projection.projectLonLat(projection, bl);
    const lr = Projection.projectLonLat(projection, br);
    const ul = Projection.projectLonLat(projection, tl);
    const ur = Projection.projectLonLat(projection, tr);
    const l  = Math.min(ll[0], ul[0]);
    const b  = Math.min(ll[1], lr[1]);
    const r  = Math.max(lr[0], ur[0]);
    const t  = Math.max(ul[1], ur[1]);
    return /** @type {T} */ (Bounds.createOrUpdate(bounds, l, b, r, t));
  },
  /**
   * @template {LatLonBounds} T
   * @param {Projection.ProjectionLike} projection
   * @param {Bounds} bounds
   * @param {T} [latLonBounds]
   * @returns {T}
   */
  unprojectBounds(projection, bounds, latLonBounds) {
    const bl = Bounds.getBottomLeft(bounds);
    const br = Bounds.getBottomRight(bounds);
    const tl = Bounds.getTopLeft(bounds);
    const tr = Bounds.getTopRight(bounds);
    const ll = Projection.unprojectXY(projection, bl);
    const lr = Projection.unprojectXY(projection, br);
    const ul = Projection.unprojectXY(projection, tl);
    const ur = Projection.unprojectXY(projection, tr);
    const l  = Math.min(ll[0], ul[0]);
    const b  = Math.min(ll[1], lr[1]);
    const r  = Math.max(lr[0], ur[0]);
    const t  = Math.max(ul[1], ur[1]);
    return /** @type {T} */ (
      LatLonBounds.createOrUpdate(latLonBounds, b, l, t, r)
    );
  },
};
Projection = /** @type {ProjectionFunctions} */ (_Projection);

function supportedProjections() {
  return Object.keys(Projection.projections).map(quote).join(', ');
}
/**
 * @param {string} str
 * @returns {string}
 */
function quote(str) {
  return `"${str}"`;
}
