/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  $,
  EventsDelegator,
  addDOMListener,
  removeDOMListener,
  DOMEventsDelegatorOptions,
} from './utils.js';

/** @type {Set<string>} */
const BROWSER_EVENTS = new Set();
/** @type {Map<string, string>} */
const MAP_EVENTS = new Map(Object.entries({
  'changing:extent': 'move',
  'change:extent'  : 'moveend',
  'change:zoom'    : 'zoomend',
}));

[
  'auxclick',
  'click',
  'dblclick',
  'mousedown',
  'mouseup',
  'mouseover',
  'mousemove',
  'mouseout',
  'mouseenter',
  'mouseleave',
  'mousewheel',
  'wheel',
  'contextmenu',
  'rightclick',
  'dblrightclick',
  'resize',
  'focusin',
  'focusout',
  'focus',
  'blur',
  'touchstart',
  'touchmove',
  'touchend',
  'touchcancel',
  'keydown',
].forEach(function(type) {
  BROWSER_EVENTS.add(type);
  BROWSER_EVENTS.add(type + '+passive');
  BROWSER_EVENTS.add(type + '+capture');
  BROWSER_EVENTS.add(type + '+passive+capture');
  BROWSER_EVENTS.add(type + '+capture+passive');
});

/**
 * @extends {EventsDelegator<number, MapHub, null>}
 */
export class MapHub extends EventsDelegator {
  // #region Private typedefs

  /** @type {import('./utils.js').EventsDelegatorOptions<number, MapHub, null>} */

  // #endregion
  // #region Public

  static delegatorOptions = {
    on(mapHub, type, callback) {
      if (BROWSER_EVENTS.has(type)) {
        addDOMListener(mapHub._node, type, callback);
      } else if (MAP_EVENTS.has(type)) {
        mapHub._map.events.registerPriority(
          $(MAP_EVENTS.get(type)),
          null,
          callback
        );
      } else {
        mapHub._listenerForType.set(type, callback);
      }
    },
    off(mapHub, type, callback) {
      if (BROWSER_EVENTS.has(type)) {
        removeDOMListener(mapHub._node, type, callback);
      } else if (MAP_EVENTS.has(type)) {
        mapHub._map.events.unregister(
          $(MAP_EVENTS.get(type)),
          null,
          callback
        );
      } else {
        mapHub._listenerForType.delete(type);
      }
    },
    getTarget(type, refs, args) {
      if (BROWSER_EVENTS.has(type)) {
        /** @type {Function} */ (DOMEventsDelegatorOptions.getTarget).apply(this, arguments);
      }
      return null;
    },
    matchesSelector(type, target, selector, mapHub, refs) {
      refs.distance = selector;
      return mapHub;
    },
    callListener(type, [callback, context], element, args, refs) {
      if (callback.apply(context || element, args) === false) {
        if (BROWSER_EVENTS.has(type)) {
          args[0].stopPropagation();
        } else {
          refs.propagationStopped = true;
          refs.returnValue = false;
        }
      }
    },
  };

  /**
   * @param {Element} node
   * @param {import('../vendors/openlayers.js').OpenLayers.Map} map
   */
  constructor(node, map) {
    // @ts-expect-error Passing `null` because we can't use `this` until after
    // `super()` is called
    super(null, MapHub.delegatorOptions);
    this.element = this;
    /** @type {Map<string, (...args: any[]) => unknown>} */
    this._listenerForType = new Map();
    this._node = node;
    this._map = map;
  }

  /**
   * @param {string} type
   * @param {any[]} args
   * @returns {unknown}
   */
  emit(type, ...args) {
    return this._listenerForType.get(type)?.(...args);
  }

  // #endregion
}
