/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {OpenLayers} from '../vendors/openlayers.js';
import {Z_INDEX} from './constants.js';
import {
  Size,
  Point,
  Bounds,
  LatLonBounds,
  Geography,
  Projection,
} from './gis.js';
import {MapHub} from './map_hub.js';
import {Quantity} from './quantity.js';
import {
  $,
  emptyFunction,
  raise,
  assignWithOwnProperties,
  memoGetters,
  debounce,
  clamp,
  PASSIVE,
  setElementCursor,
  disableTextSelection,
  getCenterOfTouches,
  getXYInFromClientXY,
  getXYInFromMouseEvent,
  getElementScaleInClient,
} from './utils.js';

const WORLD_LAT_LON_BOUNDS = LatLonBounds.create(-90, -180, 90, 180);

/**
 * @param {OpenLayers.Map} map
 * @returns {void}
 */
function fixOL2MapUpdateSizeOnResize(map) {
  /*
   * For some reason OpenLayers's auto updateSize mechanism doesn't always
   * update the size of the map.
   * But a custom one does.
   */
  // @ts-expect-error updateSizeDestroy is private
  if (map.updateSizeDestroy) {
    // @ts-expect-error updateSizeDestroy is private
    OpenLayers.Event.stopObserving(window, 'resize', map.updateSizeDestroy);
    // @ts-expect-error updateSizeDestroy is private
    map.updateSizeDestroy = null;
  } else {
    map.events.unregister('resize', map, map.updateSize);
  }
  const debouncedUpdateMapSize = debounce(100, map.updateSize.bind(map));
  window.addEventListener('resize', debouncedUpdateMapSize, PASSIVE);
  map.events.register('moveend', map, debouncedUpdateMapSize);
  map.events.register('zoomend', map, debouncedUpdateMapSize);
}

/**
 * @template T
 * @param {T} value
 * @returns {NonNullable<T>}
 * @throws {Error} If `value` is `null` or `undefined`
 */
function nonNullable(value) {
  return value ?? raise(new Error("Map's baseLayer is missing"));
}

export class MapView {
  // #region Private typedefs

  /** @typedef {import('./syntool.js').Syntool} Syntool */
  /** @typedef {import('./quantity.js').Unit} Unit */
  /** @typedef {MapView.MapUnits} MapUnits */
  /** @typedef {MapView.Props} Props */
  /** @typedef {MapView.State} State */
  /**
   * @typedef StateHook
   * @prop {(mapView: MapView) => Partial<State>} get
   * @prop {(state: Partial<State>, mapView: MapView) => void} set
   */

  // #endregion
  // #region Public

  static OBSERVE  = 1;
  static GESTURE  = 2;
  static DRAW     = 3;
  static EDIT     = 4;
  static SELECT   = 5;
  static NAVIGATE = 6;

  /**
   * @public
   * @param {MapView.Layer} layer
   * @returns {string}
   */
  static getLayerName(layer) {
    return layer.name;
  }

  /**
   * @public
   * @param {Syntool} syntool
   * @param {Props} props
   */
  constructor(syntool, props) {
    this.syntool = syntool;
    this.node = document.createElement('div');
    this.node.className = 'MapView';

    disableTextSelection(this.node);

    const {
      projection,
      maxBounds,
      restrictedBounds,
      zoomOffset,
      baseLayers,
      additionalLayers,
    } = props;
    const map = new OpenLayers.Map({
      // the theme css file is manually added to the html file
      theme: null,

      div             : this.node,
      size            : Size.create(0, 0),
      projection      : projection,
      units           : Projection.from(projection).units,
      maxExtent       : Bounds.createFromLBRT(maxBounds),
      restrictedExtent: (
        restrictedBounds && Bounds.createFromLBRT(restrictedBounds)
      ),
      center          : [0, 0],
      zoom            : 3 - zoomOffset,
      controls        : [],

      // the default value if the baseLayer doesn't specify it
      numZoomLevels: 13 - zoomOffset,

      // @ts-expect-error the following are not normal map options but will be
      // used to override map methods
      setLayerZIndex   : this._setLayerZIndex.bind(this),
      resetLayersZIndex: emptyFunction,

      /**
       * overwrite to use the restricted extent instead of the max extent
       *
       * @override
       * @this {OpenLayers.Map}
       * @param {number | null | undefined} zoom
       * @returns {number | null | undefined}
       */
      adjustZoom(zoom) {
        if (!this.baseLayer || zoom == null) {
          return zoom;
        }

        const resolutions = this.baseLayer.resolutions;
        const maxResolution = (
          $(this.getMaxExtent({restricted: true})).getWidth() / this.size.w
        );
        if ($(this.getResolutionForZoom(zoom)) > maxResolution) {
          for (let i = zoom|0, ii = resolutions.length; i < ii; ++i) {
            if (resolutions[i] <= maxResolution) {
              return i;
            }
          }
        }
        return zoom;
      },

      // Add reference to mapView
      mapView: this,
    });

    /** @type {State | null} */
    this._state = null;
    /** @type {StateHook[]} */
    this._stateHooks = [];
    this._projection = Projection.from(projection);
    this._worldProjectedBounds = (
      this.hasMercatorLikeProjection()
      ? this.getProjectedBoundsFromLonLatBounds(this.getWorldLatLonBounds())
      : null
    );
    /** @type {number} */
    this._zoomOffset = zoomOffset;
    /** @type {MapUnits | null} */
    this._resolution = null;
    this._map = map;
    this.hub = new MapHub(this.node, map);

    fixOL2MapUpdateSizeOnResize(map);

    this.addLayers(baseLayers);

    let over = 0;
    let under = 0;
    for (const layer of additionalLayers) {
      layer.isBaseLayer = false;
      layer.syntoolIsOverlay = !!layer.syntoolIsOverlay;
      map.addLayer(layer);
      if (layer.syntoolIsOverlay) {
        layer.setZIndex(Z_INDEX.OVERLAYS + over++);
      } else {
        layer.setZIndex(Z_INDEX.UNDERLAYS + under++);
      }
    }

    this.initialize(props);

    map.events.on({
      scope: this,
      /**
       * @param {object} event
       * @param {MapView.Layer} event.layer
       */
      preaddlayer({layer}) {
        this._resolution = null;
        this.hub.emit('layer:will-add-to-map', layer);
      },
      /**
       * @param {object} event
       * @param {MapView.Layer} event.layer
       */
      addlayer({layer}) {
        this.hub.emit('layer:did-add-to-map', layer);
      },
      /**
       * @param {object} event
       * @param {MapView.Layer} event.layer
       */
      preremovelayer({layer}) {
        this._resolution = null;
        this.hub.emit('layer:will-remove-from-map', layer);
      },
      /**
       * @param {object} event
       * @param {MapView.Layer} event.layer
       */
      removelayer({layer}) {
        this.hub.emit('layer:did-remove-from-map', layer);
      },
      /**
       * @param {object} event
       * @param {MapView.Layer} event.layer
       * @param {string} event.property
       */
      changelayer({layer, property}) {
        this.hub.emit('layer:change:' + property, layer);
      },
      /**
       * @param {object} event
       * @param {MapView.Layer} event.layer
       */
      changebaselayer({layer}) {
        this._resolution = null;
        this.hub.emit('change:background-layer', layer);
        this.hub.emit('change:extent');
        this._invalidateState();
      },
      moveend() {
        this._invalidateState();
      },
      prechangeresolution() {
        this._resolution = null;
      },
    });
  }

  /**
   * @public
   * @returns {void}
   */
  updateSize() {
    return this._map.updateSize();
  }
  /**
   * @public
   * @param {string | string[] | null} cursor
   * @returns {void}
   */
  setCursor(cursor) {
    setElementCursor(this.node, cursor);
  }

  /**
   * @public
   * @returns {State}
   */
  get state() {
    if (!this._state) {
      this._state = this._computeState();
    }
    return this._state;
  }
  /**
   * @public
   * @param {Partial<State>} state
   */
  set state(state) {
    this._applyState(state);
  }

  // #endregion
  // #region Protected - Extensions

  /**
   * @protected
   * @param {Props} props
   * @returns {void}
   */
  initialize(props) { /*empty*/ }

  // #endregion
  // #region Protected - State

  /**
   * @protected
   * @returns {State}
   */
  _computeState() {
    const backgroundLayerName = this.getBackgroundLayerName();
    const extent = Bounds.toLBRT(this.getViewportProjectedBounds());
    const center = this.getViewportCenterProjectedXY();
    const zoom = this.getZoom();
    return memoGetters(
      /** @type {State} */ (assignWithOwnProperties(
        {backgroundLayerName, extent, center, zoom},
        ...this._stateHooks.map(({get}) => get(this))
      ))
    );
  }
  /**
   * @protected
   * @param {Partial<State>} state
   * @returns {void}
   */
  _applyState(state) {
    const {backgroundLayerName, extent, center, zoom} = state;
    if (backgroundLayerName) {
      this.setBackgroundLayerByName(backgroundLayerName);
    }
    if (extent) {
      this.flyToProjectedBounds(Bounds.createFromLBRT(extent));
    } else if (isFinite(zoom)) {
      if (center) {
        this.setViewportCenterProjectedXY(center, zoom);
      } else {
        this.setZoom(zoom);
      }
    }
    this._stateHooks.forEach(({set}) => set(state, this));
  }
  /**
   * @protected
   * @returns {void}
   * @emits 'change:state'
   */
  _invalidateState() {
    this._state = null;
    this.hub.emit('change:state');
  }

  // #endregion
  // #region Public - Projection

  /**
   * @public
   * @returns {boolean}
   */
  hasMercatorLikeProjection() {
    return Projection.isMercatorLike(this.getProjection());
  }
  /**
   * @public
   * @returns {Projection}
   */
  getProjection() {
    return this._projection;
  }
  /**
   * @public
   * @returns {boolean}
   */
  shouldWrapDateLine() {
    return !!this._map.baseLayer?.wrapDateLine;
  }

  // #endregion
  // #region Public - World bounds

  /**
   * @public
   * @returns {Projected<Bounds> | null}
   */
  getWorldProjectedBounds() {
    return this._worldProjectedBounds;
  }
  /**
   * @public
   * @returns {LatLonBounds}
   */
  getWorldLatLonBounds() {
    return WORLD_LAT_LON_BOUNDS;
  }

  // #endregion
  // #region Public - Max bounds

  /**
   * @public
   * @returns {Projected<Bounds>}
   */
  getMaxProjectedBounds() {
    return nonNullable(this._map.getMaxExtent());
  }
  /**
   * @public
   * @returns {LatLonBounds}
   */
  getMaxLatLonBounds() {
    return this.getLonLatBoundsFromProjectedBounds(
      this.getMaxProjectedBounds()
    );
  }

  // #endregion
  // #region Public - Visible bounds

  /**
   * @public
   * @returns {Projected<Bounds>}
   */
  getVisibleProjectedBounds() {
    if (this.shouldWrapDateLine()) {
      return this.getViewportProjectedBounds();
    }
    return Bounds.intersection(
      this.getViewportProjectedBounds(),
      this.getMaxProjectedBounds()
    );
  }
  /**
   * @public
   * @returns {LatLonBounds}
   */
  getVisibleLatLonBounds() {
    return this.getLonLatBoundsFromProjectedBounds(
      this.getVisibleProjectedBounds()
    );
  }

  // #endregion
  // #region Public - Unit conversion

  /**
   * @public
   * @returns {Unit}
   */
  getMapUnits() {
    return /** @type {Unit} */ (nonNullable(this._map.getUnits()));
  }
  /**
   * @public
   * @template {Unit} U
   * @param {MapUnits} value
   * @param {U} toUnit
   * @returns {InUnit<number, U>}
   */
  convertToUnit(value, toUnit) {
    return Quantity.convert(value, this.getMapUnits(), toUnit);
  }
  /**
   * @public
   * @template {Unit} U
   * @param {InUnit<number, U>} value
   * @param {U} fromUnit
   * @returns {MapUnits}
   */
  convertFromUnit(value, fromUnit) {
    return /** @type {any} */ (
      Quantity.convert(value, fromUnit, this.getMapUnits())
    );
  }

  // #endregion
  // #region Public - Zoom

  /**
   * @public
   * @returns {number}
   */
  getZoom() {
    return this._map.getZoom() + this._zoomOffset;
  }
  /**
   * @public
   * @param {number} zoom
   * @returns {void}
   */
  setZoom(zoom) {
    this.setViewportCenterProjectedXY(
      this.getViewportCenterProjectedXY(),
      zoom
    );
  }

  // #endregion
  // #region Public - Resolution

  /**
   * @public
   * @param {Viewport<XY<Pixels>>} [xy = viewportCenter]
   * @returns {WH<Meters>}
   */
  getGeodesicPixelSize(xy) {
    const projectedXY = (
      xy
      ? this.getProjectedXYFromViewportXY(xy)
      : this.getViewportCenterProjectedXY()
    );
    const res = this.getResolution();
    /** @type {Projected<XY>} */
    const left = [projectedXY[0] - res / 2, projectedXY[1]];
    /** @type {Projected<XY>} */
    const right = [projectedXY[0] + res / 2, projectedXY[1]];
    /** @type {Projected<XY>} */
    const bottom = [projectedXY[0], projectedXY[1] - res / 2];
    /** @type {Projected<XY>} */
    const top = [projectedXY[0], projectedXY[1] + res / 2];

    return [
      Geography.geodesicDistance(
        this.getLonLatFromProjectedXY(left),
        this.getLonLatFromProjectedXY(right)
      ),
      Geography.geodesicDistance(
        this.getLonLatFromProjectedXY(bottom),
        this.getLonLatFromProjectedXY(top)
      ),
    ];
  }
  /**
   * @public
   * @returns {MapUnits}
   */
  getResolution() {
    if (!this._resolution) {
      this._resolution = nonNullable(this._map.getResolution());
    }
    return this._resolution;
  }
  /**
   * @public
   * @param {number} zoom
   * @returns {MapUnits}
   */
  getResolutionForZoom(zoom) {
    return nonNullable(this._map.getResolutionForZoom(zoom - this._zoomOffset));
  }
  /**
   * @public
   * @param {MapUnits} resolution
   * @param {boolean} [closest]
   * @returns {number}
   */
  getZoomForResolution(resolution, closest = false) {
    return (
      nonNullable(this._map.getZoomForResolution(resolution, !!closest))
      + this._zoomOffset
    );
  }

  // #endregion
  // #region Public - Viewport

  /**
   * @public
   * @returns {Viewport<WH<Pixels>>}
   */
  getViewportSize() {
    return Size.toWH(this._map.size);
  }
  /**
   * @public
   * @returns {Viewport<Pixels>}
   */
  getViewportWidth() {
    return Size.getWidth(this._map.size);
  }
  /**
   * @public
   * @returns {Viewport<Pixels>}
   */
  getViewportHeight() {
    return Size.getHeight(this._map.size);
  }
  /**
   * @public
   * @returns {Projected<Bounds>}
   */
  getViewportProjectedBounds() {
    return nonNullable(this._map.getExtent());
  }
  /**
   * @public
   * @returns {LatLonBounds}
   */
  getViewportLatLonBounds() {
    return this.getLonLatBoundsFromProjectedBounds(
      this.getViewportProjectedBounds()
    );
  }
  /**
   * @public
   * @returns {Projected<XY>}
   */
  getViewportCenterProjectedXY() {
    return Point.toXY(nonNullable(this._map.getCachedCenter()));
  }
  /**
   * @public
   * @returns {LonLat}
   */
  getViewportCenterLonLat() {
    return this.getLonLatFromProjectedXY(this.getViewportCenterProjectedXY());
  }
  /**
   * @public
   * @returns {Client<XY<Pixels>>}
   */
  getViewportCenterClientXY() {
    const mapClientRect = this._map.viewPortDiv.getBoundingClientRect();
    return [
      (mapClientRect.left + mapClientRect.right) / 2,
      (mapClientRect.top + mapClientRect.bottom) / 2,
    ];
  }
  /**
   * @public
   * @returns {Client<XY>}
   */
  getViewportScaleInClient() {
    return getElementScaleInClient(this._map.viewPortDiv);
  }

  // #endregion
  // #region Public - Navigation

  /**
   * @public
   * @param {Projected<XY>} center
   * @param {number} [zoom]
   * @returns {void}
   */
  setViewportCenterProjectedXY(center, zoom) {
    if (zoom == null) {
      zoom = this.getZoom();
    }
    zoom -= this._zoomOffset;
    zoom = clamp(zoom, 0, nonNullable(this._map.getNumZoomLevels()));
    this._map.setCenter(center, zoom);
  }
  /**
   * @public
   * @param {LonLat} center
   * @param {number} [zoom]
   * @returns {void}
   */
  setViewportCenterLonLat(center, zoom) {
    this.setViewportCenterProjectedXY(
      this.getProjectedXYFromLonLat(center),
      zoom
    );
  }
  /**
   * @public
   * @param {Viewport<Pixels>} dx
   * @param {Viewport<Pixels>} dy
   * @param {object} [options]
   * @param {boolean} [options.dragging = false]
   * @returns {void}
   */
  panBy(dx, dy, options) {
    this._map.pan(-dx, -dy, Object.assign({}, options, {animate: false}));
  }
  /**
   * @public
   * @param {Projected<Bounds>} bounds
   * @param {object} [options]
   * @param {number} [options.minZoom = 0]
   * @param {number} [options.maxZoom = Infinity]
   * @returns {void}
   */
  flyToProjectedBounds(bounds, options = {}) {
    let {minZoom = 0, maxZoom = Infinity} = options;
    if (minZoom > maxZoom) {
      minZoom = this.getZoom();
      maxZoom = minZoom;
    }

    const worldBounds = this.getWorldProjectedBounds();
    let center;
    if (worldBounds && this.shouldWrapDateLine()) {
      const worldLeft = Bounds.getLeft(worldBounds);
      const worldRight = Bounds.getRight(worldBounds);
      const worldWidth = Bounds.getWidth(worldBounds);

      // Unwrap bound if wrapped
      let [left, bottom, right, top] = Bounds.toLBRT(bounds);
      if (left > right) {
        right += Math.ceil((left - right) / worldWidth) * worldWidth;
        bounds = Bounds.createFromLBRT([left, bottom, right, top]);
      }

      // Get wrapped center of bounds
      center = Bounds.getCenter(bounds);
      if (center[0] < worldLeft) {
        center[0] += (
          Math.ceil((worldLeft - center[0]) / worldWidth) * worldWidth
        );
      } else if (center[0] > worldRight) {
        center[0] -= (
          Math.ceil((center[0] - worldRight) / worldWidth) * worldWidth
        );
      }
    } else {
      center = Bounds.getCenter(bounds);
    }

    const idealResolution = Math.max(
      Bounds.getWidth(bounds) / this.getViewportWidth(),
      Bounds.getHeight(bounds) / this.getViewportHeight()
    );
    let zoom = this.getZoomForResolution(idealResolution);
    zoom = clamp(zoom, minZoom, maxZoom);
    this.setViewportCenterProjectedXY(center, zoom);
  }
  /**
   * @public
   * @param {number} zoom
   * @param {LonLat} [anchorLonLat]
   * @returns {void}
   */
  zoomTo(zoom, anchorLonLat) {
    zoom = clamp(
      zoom,
      this._zoomOffset,
      nonNullable(this._map.getNumZoomLevels()) + this._zoomOffset
    );
    if (zoom === this.getZoom()) {
      return;
    }

    if (anchorLonLat) {
      const anchorProjectedXY = this.getProjectedXYFromLonLat(anchorLonLat);
      const anchorViewportXY = (
        this.getViewportXYFromProjectedXY(anchorProjectedXY)
      );
      const deltaX = this.getViewportWidth()  / 2 - anchorViewportXY[0];
      const deltaY = this.getViewportHeight() / 2 - anchorViewportXY[1];
      const newRes = this.getResolutionForZoom(zoom);
      /** @type {XY} */
      const newProjectedCenter = [
        anchorProjectedXY[0] + deltaX * newRes,
        anchorProjectedXY[1] - deltaY * newRes,
      ];
      this.setViewportCenterProjectedXY(newProjectedCenter, zoom);
    } else {
      this.setZoom(zoom);
    }
  }
  /**
   * @public
   * @param {LonLat} [anchorLonLat]
   * @returns {void}
   */
  zoomIn(anchorLonLat) {
    this.zoomTo(this.getZoom() + 1, anchorLonLat);
  }
  /**
   * @public
   * @param {LonLat} [anchorLonLat]
   * @returns {void}
   */
  zoomOut(anchorLonLat) {
    this.zoomTo(this.getZoom() - 1, anchorLonLat);
  }

  // #endregion
  // #region Public - Viewport <-> Client conversion

  /**
   * @public
   * @param {Viewport<XY<Pixels>>} viewportXY
   * @param {Client<XY<Pixels>>} [clientXY]
   * @returns {Client<XY<Pixels>>}
   */
  getClientXYFromViewportXY(viewportXY, clientXY) {
    const mapClientRect = this._map.viewPortDiv.getBoundingClientRect();
    if (!clientXY || clientXY.length === undefined) {
      clientXY = [0, 0];
    }
    clientXY[0] = viewportXY[0] + mapClientRect.left;
    clientXY[1] = viewportXY[1] + mapClientRect.top;
    return clientXY;
  }
  /**
   * @public
   * @param {Client<XY<Pixels>>} clientXY
   * @param {Viewport<XY<Pixels>>} [viewportXY]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportXYFromClientXY(clientXY, viewportXY) {
    return getXYInFromClientXY(this._map.viewPortDiv, clientXY, viewportXY);
  }
  /**
   * @public
   * @param {Viewport<Pixels>} viewportDistance
   * @returns {Client<Pixels>}
   */
  getClientDistanceFromViewportDistance(viewportDistance) {
    return viewportDistance / this.getViewportScaleInClient()[0];
  }
  /**
   * @public
   * @param {Client<Pixels>} clientDistance
   * @returns {Viewport<Pixels>}
   */
  getViewportDistanceFromClientDistance(clientDistance) {
    return clientDistance * this.getViewportScaleInClient()[0];
  }
  /**
   * @public
   * @param {Viewport<XY<Pixels>>} viewportVector
   * @param {Client<XY<Pixels>>} [clientVector]
   * @returns {Client<XY<Pixels>>}
   */
  getClientVectorFromViewportVector([x, y], clientVector) {
    const scale = this.getViewportScaleInClient();
    if (!clientVector || clientVector.length === undefined) {
      clientVector = [0, 0];
    }
    clientVector[0] = x / scale[0];
    clientVector[1] = y / scale[1];
    return clientVector;
  }
  /**
   * @public
   * @param {Client<XY<Pixels>>} clientVector
   * @param {Viewport<XY<Pixels>>} [viewportVector]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportVectorFromClientVector([x, y], viewportVector) {
    const scale = this.getViewportScaleInClient();
    if (!viewportVector || viewportVector.length === undefined) {
      viewportVector = [0, 0];
    }
    viewportVector[0] = x * scale[0];
    viewportVector[1] = y * scale[1];
    return viewportVector;
  }

  // #endregion
  // #region Public - <-> Projected conversion

  /**
   * @public
   * @param {Viewport<XY<Pixels>>} viewportXY
   * @param {Projected<XY<Pixels>>} [projectedXY]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedXYFromViewportXY([x, y], projectedXY) {
    // In order to get the exact projected XY (without any wrapping),
    // the base layer's `wrapDateLine` must be set to `false` before calling
    // `getLonLatFromViewPortPx()`

    const baseLayer = nonNullable(this._map.baseLayer);
    const wrapDateLine = baseLayer.wrapDateLine;
    baseLayer.wrapDateLine = false;
    const point = nonNullable(baseLayer.getLonLatFromViewPortPx({x, y}));
    baseLayer.wrapDateLine = wrapDateLine;
    return Point.toXY(point, projectedXY);
  }
  /**
   * @public
   * @param {Client<XY<Pixels>>} clientXY
   * @param {Projected<XY<Pixels>>} [projectedXY]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedXYFromClientXY(clientXY, projectedXY) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (projectedXY);
    return this.getProjectedXYFromViewportXY(
      this.getViewportXYFromClientXY(clientXY, viewportXY),
      projectedXY
    );
  }
  /**
   * @public
   * @param {Projected<XY<Pixels>>} projectedXY
   * @param {Viewport<XY<Pixels>>} [viewportXY]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportXYFromProjectedXY(projectedXY, viewportXY) {
    const {x, y} = nonNullable(
      this._map.getPixelFromLonLat(Point.createFromXY(projectedXY))
    );
    if (!viewportXY || viewportXY.length === undefined) {
      viewportXY = [0, 0];
    }
    viewportXY[0] = x;
    viewportXY[1] = y;
    return viewportXY;
  }
  /**
   * @public
   * @param {Projected<XY<Pixels>>} projectedXY
   * @param {Client<XY<Pixels>>} [clientXY]
   * @returns {Client<XY<Pixels>>}
   */
  getClientXYFromProjectedXY(projectedXY, clientXY) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (clientXY);
    return this.getClientXYFromViewportXY(
      this.getViewportXYFromProjectedXY(projectedXY, viewportXY),
      clientXY
    );
  }
  /**
   * @public
   * @param {Viewport<Pixels>} distance
   * @returns {MapUnits}
   */
  getProjectedDistanceFromViewportDistance(distance) {
    return distance * this.getResolution();
  }
  /**
   * @public
   * @param {Client<Pixels>} distance
   * @returns {MapUnits}
   */
  getProjectedDistanceFromClientDistance(distance) {
    return this.getProjectedDistanceFromViewportDistance(
      this.getViewportDistanceFromClientDistance(distance)
    );
  }
  /**
   * @public
   * @param {MapUnits} distance
   * @returns {Viewport<Pixels>}
   */
  getViewportDistanceFromProjectedDistance(distance) {
    return distance / this.getResolution();
  }
  /**
   * @public
   * @param {MapUnits} distance
   * @returns {Client<Pixels>}
   */
  getClientDistanceFromProjectedDistance(distance) {
    return this.getClientDistanceFromViewportDistance(
      this.getViewportDistanceFromProjectedDistance(distance)
    );
  }
  /**
   * @public
   * @param {Viewport<XY<Pixels>>} viewportVector
   * @param {Projected<XY<Pixels>>} [projectedVector]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedVectorFromViewportVector([x, y], projectedVector) {
    const resolution = this.getResolution();
    if (!projectedVector || projectedVector.length === undefined) {
      projectedVector = [0, 0];
    }
    projectedVector[0] = x * resolution;
    projectedVector[1] = -y * resolution;
    return projectedVector;
  }
  /**
   * @public
   * @param {Client<XY<Pixels>>} clientVector
   * @param {Projected<XY<Pixels>>} [projectedVector]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedVectorFromClientVector(clientVector, projectedVector) {
    const viewportVector = /** @type {Viewport<XY<Pixels>>} */ (projectedVector);
    return this.getProjectedVectorFromViewportVector(
      this.getViewportVectorFromClientVector(clientVector, viewportVector),
      projectedVector
    );
  }
  /**
   * @public
   * @param {Projected<XY<Pixels>>} projectedVector
   * @param {Viewport<XY<Pixels>>} [viewportVector]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportVectorFromProjectedVector([x, y], viewportVector) {
    const resolution = this.getResolution();
    if (!viewportVector || viewportVector.length === undefined) {
      viewportVector = [0, 0];
    }
    viewportVector[0] = x / resolution;
    viewportVector[1] = -y / resolution;
    return viewportVector;
  }
  /**
   * @public
   * @param {Projected<XY<Pixels>>} projectedVector
   * @param {Client<XY<Pixels>>} [clientVector]
   * @returns {Client<XY<Pixels>>}
   */
  getClientVectorFromProjectedVector(projectedVector, clientVector) {
    const viewportVector = /** @type {Viewport<XY<Pixels>>} */ (clientVector);
    return this.getClientVectorFromViewportVector(
      this.getViewportVectorFromProjectedVector(projectedVector, viewportVector),
      clientVector
    );
  }

  // #endregion
  // #region Public - ProjectedXY <-> LonLat conversion

  /**
   * @public
   * @template {LonLat} T
   * @param {Projected<XY>} projectedXY
   * @param {T} [lonLat]
   * @returns {T}
   */
  getLonLatFromProjectedXY(projectedXY, lonLat) {
    return Projection.unprojectXY(this.getProjection(), projectedXY, lonLat);
  }
  /**
   * @public
   * @template {LonLat} T
   * @param {Viewport<XY>} viewportXY
   * @param {T} [lonLat]
   * @returns {T}
   */
  getLonLatFromViewportXY(viewportXY, lonLat) {
    const projectedXY = /** @type {Projected<XY<Pixels>>} */ (lonLat);
    return this.getLonLatFromProjectedXY(
      this.getProjectedXYFromViewportXY(viewportXY, projectedXY),
      lonLat
    );
  }
  /**
   * @public
   * @template {LonLat} T
   * @param {Client<XY>} clientXY
   * @param {T} [lonLat]
   * @returns {T}
   */
  getLonLatFromClientXY(clientXY, lonLat) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (lonLat);
    return this.getLonLatFromViewportXY(
      this.getViewportXYFromClientXY(clientXY, viewportXY),
      lonLat
    );
  }
  /**
   * @public
   * @param {LonLat} lonLat
   * @param {Projected<XY>} [projectedXY]
   * @returns {Projected<XY>}
   */
  getProjectedXYFromLonLat(lonLat, projectedXY) {
    return Projection.projectLonLat(this.getProjection(), lonLat, projectedXY);
  }
  /**
   * @public
   * @param {LonLat} lonLat
   * @param {Viewport<XY>} [viewportXY]
   * @returns {Viewport<XY>}
   */
  getViewportXYFromLonLat(lonLat, viewportXY) {
    const projectedXY = /** @type {Projected<XY<Pixels>>} */ (viewportXY);
    return this.getViewportXYFromProjectedXY(
      this.getProjectedXYFromLonLat(lonLat, projectedXY),
      viewportXY
    );
  }
  /**
   * @public
   * @param {LonLat} lonLat
   * @param {Client<XY>} [clientXY]
   * @returns {Client<XY>}
   */
  getClientXYFromLonLat(lonLat, clientXY) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (clientXY);
    return this.getClientXYFromViewportXY(
      this.getViewportXYFromLonLat(lonLat, viewportXY),
      clientXY
    );
  }

  // #endregion
  // #region Public - ProjectedBounds <-> LonLatBounds conversion

  /**
   * @public
   * @template {LatLonBounds} T
   * @param {Projected<Bounds>} projectedBounds
   * @param {T} [lonLatBounds]
   * @returns {T}
   */
  getLonLatBoundsFromProjectedBounds(projectedBounds, lonLatBounds) {
    return Projection.unprojectBounds(
      this.getProjection(),
      projectedBounds,
      lonLatBounds
    );
  }
  /**
   * @public
   * @param {LatLonBounds} lonLatBounds
   * @param {Projected<Bounds>} [projectedBounds]
   * @returns {Projected<Bounds>}
   */
  getProjectedBoundsFromLonLatBounds(lonLatBounds, projectedBounds) {
    return Projection.projectLatLonBounds(
      this.getProjection(),
      lonLatBounds,
      projectedBounds
    );
  }

  // #endregion
  // #region Public - MouseEvent -> * conversion

  /**
   * @public
   * @param {MouseEvent} event
   * @param {Viewport<XY<Pixels>>} [viewportXY]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportXYFromMouseEvent(event, viewportXY) {
    return getXYInFromMouseEvent(this._map.viewPortDiv, event, viewportXY);
  }
  /**
   * @public
   * @param {MouseEvent} event
   * @param {Projected<XY<Pixels>>} [projectedXY]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedXYFromMouseEvent(event, projectedXY) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (projectedXY);
    return this.getProjectedXYFromViewportXY(
      this.getViewportXYFromMouseEvent(event, viewportXY),
      projectedXY
    );
  }
  /**
   * @public
   * @template {LonLat} T
   * @param {MouseEvent} event
   * @param {T} [lonLat]
   * @returns {T}
   */
  getLonLatFromMouseEvent(event, lonLat) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (lonLat);
    return this.getLonLatFromViewportXY(
      this.getViewportXYFromMouseEvent(event, viewportXY),
      lonLat
    );
  }

  // #endregion
  // #region Public - Touches -> * conversion

  /**
   * @public
   * @param {ArrayLike<Touch> & Iterable<Touch>} touches
   * @param {Client<XY<Pixels>>} [clientXY]
   * @returns {Client<XY<Pixels>>}
   */
  getClientXYFromTouches(touches, clientXY) {
    return getCenterOfTouches(touches, clientXY);
  }
  /**
   * @public
   * @param {ArrayLike<Touch> & Iterable<Touch>} touches
   * @param {Viewport<XY<Pixels>>} [viewportXY]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportXYFromTouches(touches, viewportXY) {
    const clientXY = /** @type {Client<XY<Pixels>>} */ (viewportXY);
    return this.getViewportXYFromClientXY(
      this.getClientXYFromTouches(touches, clientXY),
      viewportXY
    );
  }
  /**
   * @public
   * @param {ArrayLike<Touch> & Iterable<Touch>} touches
   * @param {Projected<XY<Pixels>>} [projectedXY]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedXYFromTouches(touches, projectedXY) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (projectedXY);
    return this.getProjectedXYFromViewportXY(
      this.getViewportXYFromTouches(touches, viewportXY),
      projectedXY
    );
  }
  /**
   * @public
   * @template {LonLat} T
   * @param {ArrayLike<Touch> & Iterable<Touch>} touches
   * @param {T} [lonLat]
   * @returns {T}
   */
  getLonLatFromTouches(touches, lonLat) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (lonLat);
    return this.getLonLatFromViewportXY(
      this.getViewportXYFromTouches(touches, viewportXY),
      lonLat
    );
  }

  // #endregion
  // #region Public - (Mouse|Touch|Pointer)Event -> * conversion

  /**
   * @public
   * @param {MouseEvent | TouchEvent | PointerEvent} event
   * @param {Viewport<XY<Pixels>>} [viewportXY]
   * @returns {Viewport<XY<Pixels>>}
   */
  getViewportXYFromPointerEvent(event, viewportXY) {
    return (
      /** @type {TouchEvent} */ (event).touches
      ? this.getViewportXYFromTouches(
        /** @type {TouchEvent} */ (event).touches,
        viewportXY
      )
      : this.getViewportXYFromMouseEvent(
        /** @type {MouseEvent | PointerEvent} */ (event),
        viewportXY
      )
    );
  }
  /**
   * @public
   * @param {MouseEvent | TouchEvent | PointerEvent} event
   * @param {Projected<XY<Pixels>>} [projectedXY]
   * @returns {Projected<XY<Pixels>>}
   */
  getProjectedXYFromPointerEvent(event, projectedXY) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (projectedXY);
    return this.getProjectedXYFromViewportXY(
      this.getViewportXYFromPointerEvent(event, viewportXY),
      projectedXY
    );
  }
  /**
   * @public
   * @template {LonLat} T
   * @param {MouseEvent | TouchEvent | PointerEvent} event
   * @param {T} [lonLat]
   * @returns {T}
   */
  getLonLatFromPointerEvent(event, lonLat) {
    const viewportXY = /** @type {Viewport<XY<Pixels>>} */ (lonLat);
    return this.getLonLatFromViewportXY(
      this.getViewportXYFromPointerEvent(event, viewportXY),
      lonLat
    );
  }

  // #endregion
  // #region Public - Background layer

  /**
   * @public
   * @returns {string}
   */
  getBackgroundLayerName() {
    return nonNullable(this._map.baseLayer).name;
  }
  /**
   * @public
   * @param {string} name
   * @returns {boolean}
   */
  setBackgroundLayerByName(name) {
    const layer = this._map.getLayersByName(name)[0];
    if (layer) {
      this._map.setBaseLayer(layer);
      return true;
    }
    return false;
  }

  // #endregion
  // #region Public - Add/remove layers

  /**
   * @public
   * @param {MapView.Layer} layer
   * @returns {void}
   */
  addLayer(layer) {
    if (layer && layer.map !== this._map) {
      this._map.addLayer(layer);
    }
  }
  /**
   * @public
   * @param {MapView.Layer[]} layers
   * @returns {void}
   */
  addLayers(layers) {
    layers.forEach(this.addLayer, this);
  }
  /**
   * @public
   * @param {MapView.Layer} layer
   * @returns {boolean}
   */
  hasLayer(layer) {
    return layer && layer.map === this._map;
  }
  /**
   * @public
   * @returns {MapView.Layer[]}
   */
  getLayers() {
    return this._map.layers;
  }
  /**
   * @public
   * @param {MapView.Layer} layer
   * @returns {void}
   */
  removeLayer(layer) {
    if (layer && layer.map === this._map) {
      this._map.removeLayer(layer);
    }
  }

  // #endregion
  // #region Protected - Layer Z-Index

  /**
   * @protected
   * @param {MapView.Layer} layer
   * @param {number} index
   * @returns {number}
   */
  getZIndexForLayer(layer, index) {
    if ('syntoolIsOverlay' in layer) {
      return layer.getZIndex();
    }
    if (layer.name.indexOf('Handler') === layer.name.length - 7) {
      return Z_INDEX.DRAWING_HANDLERS;
    }
    if (layer.isBaseLayer) {
      return Z_INDEX.MIN;
    }
    return Z_INDEX.MISC + index;
  }
  /**
   * @protected
   * @param {MapView.Layer} layer
   * @param {number} index
   * @returns {void}
   */
  _setLayerZIndex(layer, index) {
    layer.setZIndex(this.getZIndexForLayer(layer, index));
  }

  // #endregion
}
