/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/* eslint-disable max-lines */
import {
  extendClass,
  flattenArray,
  asyncIter,
  scheduleDelayedTask,
  scheduleTask,
  scheduleAnimationFrame,
  modulo,
  bisectLeft,
  getClientXYFromMouseEvent,
} from './utils.js';
import {Dataset} from './dataset.js';
import {Geometry} from './gis.js';
import {MapView} from './map_view.js';

/**
 * @param {Product} productA
 * @param {Product} productB
 * @returns {number}
 */
function compareProductsByPriority(productA, productB) {
  return productB.priority - productA.priority;
}
/**
 * @param {Product} productA
 * @param {Product} productB
 * @returns {number}
 */
function compareProductsByStackLevel(productA, productB) {
  return productA.stackLevel - productB.stackLevel;
}
/**
 * @param {Dataset} datasetA
 * @param {Dataset} datasetB
 * @returns {number}
 */
function compareDatasetsByStartAndID(datasetA, datasetB) {
  return (
    datasetA === datasetB
    ? 0
    : (
      (datasetA.start.getTime() - datasetB.start.getTime()) ||
      (datasetA.datasetId < datasetB.datasetId ? -1 : 1)
    )
  );
}
/**
 * @param {Dataset | null | undefined} datasetA
 * @param {Dataset | null | undefined} datasetB
 * @returns {number}
 */
function compareDatasetsByRenderOrder(datasetA, datasetB) {
  if (!datasetA) {
    return -1; // datasetA lower
  }
  if (!datasetB) {
    return 1; // datasetA higher
  }
  if (datasetA.isVisible() === datasetB.isVisible()) {
    return (
      (datasetA.getStackLevel() - datasetB.getStackLevel()) ||
      compareDatasetsByStartAndID(datasetA, datasetB)
    );
  }
  if (datasetA.isVisible()) {
    return 1; // datasetA higher
  }
  return -1; // datasetA lower
}
/**
 * @param {DatasetsRendering} datasetsRendering
 * @returns {void}
 */
function sortInMapDatasetsForProductWithId(datasetsRendering) {
  datasetsRendering.sortedInMapDatasets = (
    [...datasetsRendering.inMapDatasets]
    .sort(compareDatasetsByStartAndID)
  );
}

class InfosIterator {
  constructor() {
    /** @type {Set<DatasetInfo>} */
    this.infosSet = new Set();
    this.infosIter = this.infosSet.values();
    this.index = 0;
  }
  get length() {
    return this.infosSet.size;
  }
  reset() {
    this.infosSet.clear();
    this.infosIter = this.infosSet.values();
    this.index = 0;
  }
  /**
   * @param {Iterable<DatasetInfo> | null | undefined} iterable
   * @returns {void}
   */
  add(iterable) {
    if (iterable) {
      for (const value of iterable) {
        this.infosSet.add(value);
      }
    }
  }
  [Symbol.iterator]() {
    return this;
  }
  next() {
    const result = this.infosIter.next();
    if (!result.done) {
      this.index++;
    }
    return result;
  }
}

// #region Exported typedefs

/**
 * @typedef {(
 *   this: void,
 *   event: MouseEvent | TouchEvent,
 *   clientXY: Client<XY<Pixels>>,
 * ) => Dataset | null | undefined} DatasetFromEventGetter
 */

/** @typedef {typeof MapViewDatasetsExtension} MapViewDatasetsExtension */
/**
 * @typedef MapViewDatasetsExtension.Layer
 * @prop {Dataset} [dataset]
 * @prop {ProductId} [productId]
 * @prop {(event: Event) => import('./gis.js').Feature | null | undefined} [getFeatureFromEvent]
 */
/**
 * @typedef MapViewDatasetsExtension.Props
 * @prop {number} [maxDatasetsWeight = 100] The maximum cumulated weight of
 * datasets rendered at the same time.
 *
 * Default: `100`
 */
/**
 * @typedef MapViewDatasetsExtension.State
 * @prop {number} maxDatasetsWeight
 * @prop {Record<ProductId, number>} productsPage
 * @prop {Record<DatasetId, number>} datasetOpacities
 */

/**
 * @typedef MapViewDatasetsExtension.InstanceFields
 * @prop {DatasetFromEventGetter[]} _datasetFromEventGetters
 * @prop {number} _maxDatasetsWeight
 * @prop {Map<DatasetId, Dataset>} _datasetsById
 * @prop {Set<Dataset>} _datasetsInMap
 * @prop {Set<Dataset>} _datasetsShouldBeVisible
 * @prop {Set<Dataset>} _datasetsVisible
 * @prop {Map<ProductId, DatasetsRendering>} _datasetsRenderingByProductId
 * @prop {number} _updateDatasetsProgress
 * @prop {DatasetStatistics} _datasetsStatistics
 * @prop {InfosIterator} _infosIterator
 * @prop {(() => void) | null} cancelUpdateDatasets
 * @prop {(() => void) | null} _cancelUpdateDatasetsVisibility
 */

// #endregion

const MapViewDatasetsExtension = {
  // #region Private typedefs

  /** @typedef {import('./api.js').API.DatasetInfo} DatasetInfo */
  /** @typedef {import('./utils.js').AsyncIterOptions} AsyncIterOptions */
  /** @typedef {MapView.Layer} Layer */
  /** @typedef {MapViewDatasetsExtension.Props} Props */
  /** @typedef {MapViewDatasetsExtension.State} State */
  /**
   * @typedef DatasetStatistics
   * @prop {boolean} changing
   * @prop {number} inMapCount
   * @prop {number} shouldBeInMapCount
   * @prop {number} displayedCount
   * @prop {number} shouldBeDisplayedCount
   * @prop {number} displayableCount
   */
  /**
   * @typedef DatasetsRendering
   * @prop {Dataset[]} datasets
   * @prop {Set<Dataset>} inMapDatasets
   * @prop {Dataset[]} sortedInMapDatasets
   * @prop {number} visibleDatasetsCount
   * @prop {number} pageSize
   * @prop {number} currentPage
   */

  // #endregion
  // #region Protected - Extensions

  /**
   * @protected
   * @this {MapView}
   * @param {Props} props
   * @returns {void}
   */
  initialize(props) {
    window.original();

    this._datasetFromEventGetters = [];

    this._maxDatasetsWeight = props.maxDatasetsWeight || 100;

    this._datasetsById = new Map();
    this._datasetsInMap = new Set();
    this._datasetsShouldBeVisible = new Set();
    this._datasetsVisible = new Set();
    this._datasetsRenderingByProductId = new Map();

    this._updateDatasetsProgress = 1;
    this._datasetsStatistics = {
      changing              : false,
      inMapCount            : 0,
      shouldBeInMapCount    : 0,
      displayedCount        : 0,
      shouldBeDisplayedCount: 0,
      displayableCount      : 0,
    };

    this._infosIterator = new InfosIterator();
    this.cancelUpdateDatasets = null;
    this._cancelUpdateDatasetsVisibility = null;

    this._stateHooks.push(({
      /**
       * @this {MapView}
       * @returns {State}
       */
      get: () => {
        const productsPage = /** @type {Record<ProductId, number>} */ ({});
        for (const productId of this.syntool.productsStore.productIds()) {
          const datasetsRendering = (
            this._getDatasetsRenderingForProductWithId(productId)
          );
          if (datasetsRendering.currentPage > 0) {
            productsPage[productId] = datasetsRendering.currentPage;
          }
        }
        const datasetOpacities = /** @type {Record<DatasetId, number>} */ (
          Object.create(null)
        );
        for (const [datasetId, dataset] of this._datasetsById.entries()) {
          if (
            dataset.canSetOpacity &&
            dataset.getOpacity() !== dataset.getProduct()?.opacity
          ) {
            datasetOpacities[datasetId] = dataset.getOpacity();
          }
        }
        return {
          maxDatasetsWeight: this.maxDatasetsWeight,
          productsPage,
          datasetOpacities,
        };
      },
      /**
       * @this {MapView}
       * @param {Partial<State>} state
       * @returns {void}
       */
      set: ({
        // @ts-expect-error `this` has type `any`, see TypeScript issue #55048
        maxDatasetsWeight = this.maxDatasetsWeight,
        productsPage = {},
        datasetOpacities = {},
      }) => {
        let changed = false;

        if (this.maxDatasetsWeight !== maxDatasetsWeight) {
          this._maxDatasetsWeight = maxDatasetsWeight;
          this._distributePageSize();
          changed = true;
        }

        for (const productId of this.syntool.productsStore.productIds()) {
          if (this._setProductPage(productId, productsPage[productId])) {
            changed = true;
          }
        }

        for (const [datasetId, dataset] of this._datasetsById.entries()) {
          const opacity = (
            datasetOpacities[datasetId]
            ?? dataset.getProduct()?.opacity
            ?? dataset.getOpacity()
          );
          if (this._setDatasetOpacity(dataset, opacity)) {
            changed = true;
          }
        }

        if (changed) {
          // @ts-expect-error protected method
          this._invalidateState();
          this._updateDatasetsVisibility();
        }
      }
    }));

    this.syntool.productsStore.hub
      .on('change:opacity', this._handleProductOpacityChange, this)
      .on('change:stackLevel', this._updateDatasetsStackLevels, this);
    this.syntool.hub.on('low-memory', this._handleLowMemory, this);

    this.hub.on('change:extent', () => this.updateDatasets());
  },

  // #endregion
  // #region Protected - Layer Z-Index

  /**
   * @protected
   * @this {MapView}
   * @param {Layer} layer
   * @param {number} index
   * @returns {number}
   */
  getZIndexForLayer(layer, index) {
    if (layer.dataset || layer.productId) {
      return layer.getZIndex();
    }
    return window.original();
  },

  // #endregion
  // #region Public - Datasets

  /**
   * @public
   * @this {MapView}
   * @returns {number}
   */
  get updateDatasetsProgress() {
    return this._updateDatasetsProgress;
  },
  /**
   * @public
   * @this {MapView}
   * @returns {DatasetStatistics}
   */
  get datasetsStatistics() {
    return this._datasetsStatistics;
  },

  /**
   * @public
   * @this {MapView}
   * @param {DatasetId} datasetId
   * @returns {Dataset | undefined}
   */
  getDatasetWithId(datasetId) {
    return this._datasetsById.get(datasetId);
  },
  /**
   * @public
   * @this {MapView}
   * @param {ProductId} productId
   * @returns {Dataset[]}
   */
  getProductDatasets(productId) {
    return this._getDatasetsRenderingForProductWithId(productId).datasets;
  },

  /**
   * @public
   * @this {MapView}
   * @returns {void}
   */
  destroyOldDatasets() {
    for (const dataset of this._datasetsById.values()) {
      if (!this.isDatasetInMap(dataset) && dataset.hasResources()) {
        dataset.destroy();
      }
    }
  },
  /**
   * @public
   * @this {MapView}
   * @returns {void}
   */
  removeOldDatasets() {
    for (const dataset of this._datasetsById.values()) {
      if (!this.syntool.filters.dataset.inTimeRange(dataset)) {
        this.removeDataset(dataset);
      }
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @param {number} level
   * @returns {void}
   */
  _handleLowMemory(level) {
    if (level === 0) {
      this.destroyOldDatasets();
    } else {
      this.removeOldDatasets();
    }
  },

  /**
   * @public
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  removeDataset(dataset) {
    const productId = dataset.productId;
    const wasInMap = this.isDatasetInMap(dataset);

    if (wasInMap) {
      this._markDatasetAsNotInMap(dataset);
      this._removeDatasetFromMap(dataset);
      this._updateProductsInMapDatasets();
      this._updateDatasetsStackLevels();
      if (this._datasetsShouldBeVisible.delete(dataset)) {
        this._getDatasetsRenderingForProductWithId(productId)
          .visibleDatasetsCount--;
        this._updateDatasetsPagination();
      }
    }

    this._removeDataset(dataset);

    if (wasInMap) {
      this._updateDatasetsStatistics({
        inMapCount            : this.getDatasetsInMapCount(),
        shouldBeInMapCount    : this.getDatasetsInMapCount(),
        displayableCount      : this._datasetsShouldBeVisible.size,
        displayedCount        : this._datasetsVisible.size,
        shouldBeDisplayedCount: this._datasetsVisible.size,
      });
      this._emitInMapDatasetsChange();
    }
  },
  /**
   * @public
   * @this {MapView}
   * @param {DatasetId} datasetId
   * @returns {void}
   */
  removeDatasetWithId(datasetId) {
    const dataset = this.getDatasetWithId(datasetId);
    if (dataset) {
      this.removeDataset(dataset);
    }
  },
  /**
   * @public
   * @this {MapView}
   * @param {ProductId} productId
   * @returns {void}
   */
  removeProductWithId(productId) {
    const datasetsRendering = this._datasetsRenderingByProductId.get(productId);
    if (datasetsRendering) {
      this._datasetsRenderingByProductId.delete(productId);

      let wasInMap = false;
      for (const dataset of datasetsRendering.datasets) {
        if (this.isDatasetInMap(dataset)) {
          wasInMap = true;
          this._removeDatasetFromMap(dataset);
          this._datasetsShouldBeVisible.delete(dataset);
        }
        dataset.destroy();
        this._datasetsById.delete(dataset.datasetId);
        this.hub.emit('remove:dataset', dataset);
      }
      if (wasInMap) {
        this._updateDatasetsStackLevels();
        this._updateDatasetsPagination();
        this._updateDatasetsStatistics({
          inMapCount            : this.getDatasetsInMapCount(),
          shouldBeInMapCount    : this.getDatasetsInMapCount(),
          displayableCount      : this._datasetsShouldBeVisible.size,
          displayedCount        : this._datasetsVisible.size,
          shouldBeDisplayedCount: this._datasetsVisible.size,
        });
        this._emitInMapDatasetsChange();
      }
    }
  },
  /**
   * @public
   * @this {MapView}
   * @returns {void}
   */
  removeAllDatasets() {
    let wasInMap = false;
    for (const dataset of this._datasetsById.values()) {
      if (this.isDatasetInMap(dataset)) {
        wasInMap = true;
        this._removeDatasetFromMap(dataset);
      }
      dataset.destroy();
    }
    this._datasetsShouldBeVisible.clear();
    this._datasetsRenderingByProductId.clear();
    this._datasetsById.clear();
    this._updateProductsInMapDatasets();
    this.hub.emit('clear:datasets');
    if (wasInMap) {
      this._updateDatasetsStackLevels();
      this._updateDatasetsPagination();
      this._updateDatasetsStatistics({
        inMapCount            : this.getDatasetsInMapCount(),
        shouldBeInMapCount    : this.getDatasetsInMapCount(),
        displayableCount      : this._datasetsShouldBeVisible.size,
        displayedCount        : this._datasetsVisible.size,
        shouldBeDisplayedCount: this._datasetsVisible.size,
      });
      this._emitInMapDatasetsChange();
    }
  },

  /**
   * @public
   * @this {MapView}
   * @param {Iterable<DatasetInfo>} [infos]
   * @returns {void}
   */
  updateDatasets(infos) {
    this.cancelUpdateDatasets?.();
    const abortController = new AbortController();
    this.cancelUpdateDatasets = () => {
      abortController.abort();
      this.cancelUpdateDatasets = null;
    };
    this._updateDatasetsFromInfos(infos, abortController.signal);
  },

  // #endregion
  // #region Public - Datasets in map

  /**
   * @public
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {boolean}
   */
  isDatasetInMap(dataset) {
    return this._datasetsInMap.has(dataset);
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {number}
   *
   * @template TThis
   * @overload
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {number}
   *
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {number}
   */
  getDatasetsInMapCount(filter, thisArg) {
    if (filter) {
      let count = 0;
      for (const dataset of this._datasetsInMap) {
        if (filter.call(thisArg, dataset)) {
          count++;
        }
      }
      return count;
    }
    return this._datasetsInMap.size;
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {boolean}
   *
   * @template TThis
   * @overload
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {boolean}
   *
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {boolean}
   */
  hasDatasetsInMap(filter, thisArg) {
    if (filter) {
      for (const dataset of this._datasetsInMap) {
        if (filter.call(thisArg, dataset)) {
          return true;
        }
      }
      return false;
    }
    return this._datasetsInMap.size > 0;
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {Dataset[]}
   *
   * @template TThis
   * @overload
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {Dataset[]}
   *
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {Dataset[]}
   */
  getDatasetsInMap(filter, thisArg) {
    const datasets = [...this._datasetsInMap];
    return filter ? datasets.filter(filter, thisArg) : datasets;
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {Generator<Dataset, void>}
   *
   * @template TThis
   * @overload
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {Generator<Dataset, void>}
   *
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {Generator<Dataset, void>}
   */
  *datasetsInMap(filter, thisArg) {
    if (filter) {
      for (const dataset of this._datasetsInMap) {
        if (filter.call(thisArg, dataset)) {
          yield dataset;
        }
      }
    } else {
      yield* this._datasetsInMap.values();
    }
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {(this: void, dataset: Dataset) => boolean} fn
   * @returns {void}
   *
   * @template TThis
   * @overload
   * @param {(this: TThis, dataset: Dataset) => boolean} fn
   * @param {TThis} thisArg
   * @returns {void}
   *
   * @param {(dataset: Dataset) => boolean} fn
   * @param {unknown} [thisArg]
   * @returns {void}
   */
  forEachDatasetInMap(fn, thisArg) {
    this._datasetsInMap.forEach(fn, thisArg);
  },
  /**
   * @public
   * @this {MapView}
   * @param {ProductId} productId
   * @returns {Dataset[]}
   */
  getProductDatasetsInMap(productId) {
    return (
      this._getDatasetsRenderingForProductWithId(productId)
      .sortedInMapDatasets
    );
  },
  /**
   * @public
   * @this {MapView}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  hasDatasetsInMapForProductWithId(productId) {
    return this.getProductDatasetsInMap(productId).length > 0;
  },

  // #endregion
  // #region Public - Dataset related methods

  /**
   * @public
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {boolean}
   */
  isGlobalDataset(dataset) {
    return dataset.isGlobal();
  },
  /**
   * @public
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {boolean}
   */
  isRegionalDataset(dataset) {
    return !this.isGlobalDataset(dataset);
  },
  /**
   * @public
   * @this {MapView}
   * @param {Geometry} geometry
   * @param {Dataset} dataset
   * @param {Projected<import('./gis.js').Bounds> | null} [worldBounds]
   * @returns {boolean | 0}
   */
  isGeometryIntersectingDataset(
    geometry,
    dataset,
    // @ts-expect-error `this` has type `any`, see TypeScript issue #55048
    worldBounds = this.getWorldProjectedBounds()
  ) {
    return (
      dataset.geometry
      ? (
        this.isGlobalDataset(dataset) ||
        Geometry.intersects(dataset.geometry, geometry, worldBounds)
      )
      : false
    );
  },
  /**
   * @public
   * @this {MapView}
   * @param {Projected<XY>} projectedXY
   * @param {Dataset} dataset
   * @returns {boolean}
   */
  isProjectedXYInsideDataset(projectedXY, dataset) {
    return !!dataset.containsProjectedXY(projectedXY);
  },

  /**
   * @public
   * @this {MapView}
   * @param {Dataset} dataset
   * @param {number} opacity
   * @returns {boolean}
   */
  setDatasetOpacity(dataset, opacity) {
    if (this._setDatasetOpacity(dataset, opacity)) {
      // @ts-expect-error protected method
      this._invalidateState();
      return true;
    }
    return false;
  },
  /**
   * @public
   * @this {MapView}
   * @param {Iterable<Dataset>} datasets
   * @param {number} opacity
   * @returns {boolean}
   */
  setDatasetsOpacity(datasets, opacity) {
    let changed = false;
    for (const dataset of datasets) {
      if (this._setDatasetOpacity(dataset, opacity)) {
        changed = true;
      }
    }
    if (changed) {
      // @ts-expect-error protected method
      this._invalidateState();
      return true;
    }
    return false;
  },

  /**
   * @protected
   * @this {MapView}
   * @param {MouseEvent | TouchEvent} event
   * @param {Client<XY<Pixels>>} clientXY
   * @returns {Dataset | null}
   */
  _getDatasetFromEventAtClientXY(event, clientXY) {
    for (const datasetFromEventGetter of this._datasetFromEventGetters) {
      const dataset = datasetFromEventGetter(event, clientXY);
      if (dataset) {
        return dataset;
      }
    }
    for (const layer of this.getLayers()) {
      const feature = layer.getFeatureFromEvent?.(event);
      if (feature?.layer && feature.dataset) {
        return feature.dataset;
      }
    }
    return null;
  },
  /**
   * @public
   * @this {MapView}
   * @param {MouseEvent | TouchEvent} event
   * @param {Client<XY<Pixels>>} clientXY
   * @returns {Dataset | null}
   */
  getDatasetFromEventAtClientXY(event, clientXY) {
    return (
      this._getDatasetFromEventAtClientXY(event, clientXY) ||
      this.getDatasetUnderProjectedXY(
        this.getProjectedXYFromClientXY(clientXY),
        null,
        this.isRegionalDataset,
        this
      )
    );
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {MouseEvent | TouchEvent} event
   * @param {Client<XY<Pixels>>} clientXY
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {Dataset[]}
   *
   * @template TThis
   * @overload
   * @param {MouseEvent | TouchEvent} event
   * @param {Client<XY<Pixels>>} clientXY
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {Dataset[]}
   *
   * @param {MouseEvent | TouchEvent} event
   * @param {Client<XY<Pixels>>} clientXY
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {Dataset[]}
   */
  getDatasetsFromEventAtClientXY(event, clientXY, filter, thisArg) {
    const dataset = this._getDatasetFromEventAtClientXY(event, clientXY);
    return this.getDatasetsUnderProjectedXY(
      this.getProjectedXYFromClientXY(clientXY),
      (dataset ? [dataset] : []),
      filter,
      thisArg
    );
  },
  /**
   * @public
   * @this {MapView}
   * @param {MouseEvent} event
   * @returns {Dataset | null}
   */
  getDatasetFromMouseEvent(event) {
    return this.getDatasetFromEventAtClientXY(
      event,
      getClientXYFromMouseEvent(event)
    );
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {MouseEvent} event
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {Dataset[]}
   *
   * @template TThis
   * @overload
   * @param {MouseEvent} event
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {Dataset[]}
   *
   * @param {MouseEvent} event
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {Dataset[]}
   */
  getDatasetsFromMouseEvent(event, filter, thisArg) {
    return this.getDatasetsFromEventAtClientXY(
      event,
      getClientXYFromMouseEvent(event),
      filter,
      thisArg
    );
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {Projected<XY>} projectedXY
   * @param {Dataset[] | null | undefined} [datasets = []]
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {Dataset[]}
   *
   * @template TThis
   * @overload
   * @param {Projected<XY>} projectedXY
   * @param {Dataset[] | null | undefined} datasets
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {Dataset[]}
   *
   * @param {Projected<XY>} projectedXY
   * @param {Dataset[] | null | undefined} [datasets]
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {Dataset[]}
   */
  getDatasetsUnderProjectedXY(projectedXY, datasets, filter, thisArg) {
    const worldBounds = this.getWorldProjectedBounds();
    const point = Geometry.createPointFromXY(projectedXY);
    /** @type {Dataset} */
    let dataset; // used by comparator() below
    if (!datasets) {
      datasets = [];
    }
    for (const d of this.datasetsInMap(filter, thisArg)) {
      if (
        d.geometry
        ? this.isGeometryIntersectingDataset(point, d, worldBounds)
        : this.isProjectedXYInsideDataset(projectedXY, d)
      ) {
        dataset = d; // used by comparator() below
        const insertionIndex = bisectLeft(datasets, comparator);
        if (datasets[insertionIndex] !== dataset) {
          datasets.splice(insertionIndex, 0, dataset);
        }
      }
    }
    return datasets;

    /**
     * @param {Dataset} dataset2
     * @returns {number}
     */
    function comparator(dataset2) {
      return -compareDatasetsByRenderOrder(dataset, dataset2);
    }
  },
  /**
   * @public
   * @this {MapView}
   * @overload
   * @param {Projected<XY>} projectedXY
   * @param {Dataset | null | undefined} [topDataset]
   * @param {(this: void, dataset: Dataset) => boolean} [filter]
   * @returns {Dataset | null}
   *
   * @template TThis
   * @overload
   * @param {Projected<XY>} projectedXY
   * @param {Dataset | null} topDataset
   * @param {((this: TThis, dataset: Dataset) => boolean) | undefined} filter
   * @param {TThis} thisArg
   * @returns {Dataset | null}
   *
   * @param {Projected<XY>} projectedXY
   * @param {Dataset | null} [topDataset]
   * @param {(dataset: Dataset) => boolean} [filter]
   * @param {unknown} [thisArg]
   * @returns {Dataset | null}
   */
  getDatasetUnderProjectedXY(projectedXY, topDataset, filter, thisArg) {
    const worldBounds = this.getWorldProjectedBounds();
    const point = Geometry.createPointFromXY(projectedXY);
    for (const dataset of this.datasetsInMap(filter, thisArg)) {
      if (
        (
          dataset.geometry
          ? this.isGeometryIntersectingDataset(point, dataset, worldBounds)
          : this.isProjectedXYInsideDataset(projectedXY, dataset)
        ) &&
        compareDatasetsByRenderOrder(dataset, topDataset) > 0
      ) {
        topDataset = dataset;
      }
    }
    return topDataset || null;
  },

  // #endregion
  // #region Public - Pagination

  /**
   * @public
   * @this {MapView}
   * @returns {number}
   */
  get maxDatasetsWeight() {
    return this._maxDatasetsWeight;
  },
  /**
   * @public
   * @this {MapView}
   * @param {number} newWeight
   * @returns {void}
   */
  set maxDatasetsWeight(newWeight) {
    this._maxDatasetsWeight = newWeight;
    this._updateDatasetsPagination();
    // @ts-expect-error protected method
    this._invalidateState();
  },

  /**
   * @public
   * @this {MapView}
   * @param {number} delta
   * @returns {void}
   */
  incrementDataPageBy(delta) {
    let changed = false;

    for (const productId of this.syntool.productsStore.productIds()) {
      if (this._changeProductPageBy(productId, delta)) {
        changed = true;
      }
    }

    if (changed) {
      // @ts-expect-error protected method
      this._invalidateState();
      this._updateDatasetsVisibility();
    }
  },

  // #endregion
  // #region Protected

  /**
   * @protected
   * @this {MapView}
   * @param {ProductId} productId
   * @returns {DatasetsRendering}
   */
  _getDatasetsRenderingForProductWithId(productId) {
    let datasetsRendering = this._datasetsRenderingByProductId.get(productId);
    if (!datasetsRendering) {
      datasetsRendering = {
        datasets            : [],
        inMapDatasets       : new Set(),
        sortedInMapDatasets : [],
        visibleDatasetsCount: 0,
        pageSize            : 0,
        currentPage         : 0,
      };
      this._datasetsRenderingByProductId.set(productId, datasetsRendering);
    }
    return datasetsRendering;
  },

  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @param {number} opacity
   * @returns {boolean}
   */
  _setDatasetOpacity(dataset, opacity) {
    if (opacity != null && dataset.canSetOpacity) {
      const previous = dataset.getOpacity();
      // @ts-expect-error setOpacity() is not defined on Dataset
      dataset.setOpacity(opacity);
      return (previous !== dataset.getOpacity());
    }
    return false;
  },
  /**
   * @protected
   * @this {MapView}
   * @param {ProductId} productId
   * @returns {void}
   */
  _handleProductOpacityChange(productId) {
    const product = this.syntool.productsStore.getProduct(productId);
    if (!product) {
      return;
    }

    let changed = false;
    const {opacity} = product;
    for (const dataset of this.getProductDatasets(productId)) {
      if (this._setDatasetOpacity(dataset, opacity)) {
        changed = true;
      }
    }
    if (changed) {
      // @ts-expect-error protected method
      this._invalidateState();
    }
  },

  /**
   * @protected
   * @this {MapView}
   * @param {ProductId} productId
   * @param {number | null | undefined} page
   * @returns {boolean}
   */
  _setProductPage(productId, page) {
    const datasetsRendering = (
      this._getDatasetsRenderingForProductWithId(productId)
    );
    if (page == null || isNaN(page)) {
      page = 0;
    }
    if (datasetsRendering.currentPage !== page) {
      datasetsRendering.currentPage = page;
      return true;
    }
    return false;
  },
  /**
   * @protected
   * @this {MapView}
   * @param {ProductId} productId
   * @param {number} delta
   * @returns {boolean}
   */
  _changeProductPageBy(productId, delta) {
    const {pageSize, currentPage, visibleDatasetsCount} = (
      this._getDatasetsRenderingForProductWithId(productId)
    );
    if (pageSize > 0) {
      const maxPage = Math.ceil(visibleDatasetsCount / pageSize);
      const newPage = modulo(currentPage + delta, maxPage);
      return this._setProductPage(productId, newPage);
    }
    return false;
  },

  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _addDatasetToMap(dataset) {
    if (!this.isDatasetInMap(dataset)) {
      this.hub.emit('dataset:will-add-to-map', dataset);
      dataset.willAddToMap();
      this._datasetsInMap.add(dataset);
      dataset.didAddToMap();
      this.hub.emit('dataset:did-add-to-map', dataset);
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _removeDatasetFromMap(dataset) {
    if (this.isDatasetInMap(dataset)) {
      this.hub.emit('dataset:will-remove-from-map', dataset);
      dataset.willRemoveFromMap();
      this._datasetsInMap.delete(dataset);
      dataset.didRemoveFromMap();
      this._hideDataset(dataset);
      this.hub.emit('dataset:did-remove-from-map', dataset);
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @param {ProductId} productId
   * @param {Dataset} dataset
   * @returns {void}
   */
  _addProductDataset(productId, dataset) {
    this._getDatasetsRenderingForProductWithId(productId)
      .datasets.push(dataset);
  },
  /**
   * @protected
   * @this {MapView}
   * @param {ProductId} productId
   * @param {Dataset} dataset
   * @returns {void}
   */
  _removeProductDataset(productId, dataset) {
    const datasets = (
      this._getDatasetsRenderingForProductWithId(productId).datasets
    );
    const index = datasets.indexOf(dataset);

    if (index !== -1) {
      datasets.splice(index, 1);
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _showDataset(dataset) {
    dataset.show();
    this._datasetsVisible.add(dataset);
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _hideDataset(dataset) {
    dataset.hide();
    this._datasetsVisible.delete(dataset);
  },

  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _addDataset(dataset) {
    const datasetId = dataset.datasetId;

    this._datasetsById.set(datasetId, dataset);
    this._addProductDataset(dataset.productId, dataset);
    this.hub.emit('add:dataset', dataset);
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @param {DatasetInfo} info
   * @returns {void}
   */
  _updateDataset(dataset, info) {
    if (dataset.updateFromInfo(info)) {
      this.hub.emit('update:dataset', dataset);
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _removeDataset(dataset) {
    const datasetId = dataset.datasetId;
    const productId = dataset.productId;

    dataset.destroy();
    if (this._datasetsById.delete(datasetId)) {
      this._removeProductDataset(productId, dataset);
    }
    this.hub.emit('remove:dataset', dataset);
  },
  /**
   * @protected
   * @this {MapView}
   * @param {DatasetInfo} info
   * @returns {void}
   */
  _createOrUpdateDataset(info) {
    /** @type {Dataset | null | undefined} */
    let dataset = this.getDatasetWithId(info.datasetId);
    if (dataset) {
      this._updateDataset(dataset, info);
    } else {
      dataset = Dataset.createDataset(info, this.syntool);
      if (dataset) {
        this._addDataset(dataset);
      }
    }
  },

  /**
   * @protected
   * @this {MapView}
   * @param {number} phase
   * @param {number} progress
   * @param {number} totalProgress
   * @returns {void}
   */
  _setUpdateDatasetsProgress(phase, progress, totalProgress) {
    this._updateDatasetsProgress = (phase + progress / totalProgress) / 4;
    this.hub.emit('change:update-datasets-progress');
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {void}
   */
  _setUpdateDatasetsProgressComplete() {
    this._updateDatasetsProgress = 1;
    this.hub.emit('change:update-datasets-progress');
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Partial<DatasetStatistics>} partial
   * @returns {void}
   */
  _updateDatasetsStatistics(partial) {
    Object.assign(this._datasetsStatistics, partial);
    this.hub.emit('change:datasetsStatistics');
  },

  /**
   * @protected
   * @this {MapView}
   * @returns {void}
   */
  _emitInMapDatasetsChange() {
    this._setUpdateDatasetsProgressComplete();
    this.hub.emit('change:in-map-datasets');
  },

  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _markDatasetAsInMap(dataset) {
    this._getDatasetsRenderingForProductWithId(dataset.productId)
      .inMapDatasets.add(dataset);
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Dataset} dataset
   * @returns {void}
   */
  _markDatasetAsNotInMap(dataset) {
    this._getDatasetsRenderingForProductWithId(dataset.productId)
      .inMapDatasets.delete(dataset);
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {void}
   */
  _updateProductsInMapDatasets() {
    this._datasetsRenderingByProductId
      .forEach(sortInMapDatasetsForProductWithId);
  },
  /**
   * @protected
   * @this {MapView}
   * @param {Product} product
   * @returns {Dataset[]}
   */
  _getSortedInMapDatasetsForProduct(product) {
    return (
      this._getDatasetsRenderingForProductWithId(product.id)
      .sortedInMapDatasets
    );
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {void}
   */
  _updateDatasetsStackLevels() {
    const selectedProducts = this.syntool.productsStore.getSelectedProducts();
    const datasets = flattenArray(
      selectedProducts
      .sort(compareProductsByStackLevel)
      .map(this._getSortedInMapDatasetsForProduct, this)
    );

    let stackLevel = 0;
    for (const dataset of datasets) {
      if (dataset.canSetStackLevel) {
        // @ts-expect-error setStackLevel() is not defined on Dataset
        dataset.setStackLevel(stackLevel++);
      }
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {void}
   */
  _updateDatasetsPagination() {
    this._distributePageSize();
    this._updateDatasetsVisibility();
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {void}
   */
  _distributePageSize() {
    const products = this.syntool.productsStore.getSelectedProducts();
    const remainingCountByProduct = new Map();
    let remainingBudget = this.maxDatasetsWeight;

    products.sort(compareProductsByPriority);

    for (const product of products) {
      const datasetsRendering = (
        this._getDatasetsRenderingForProductWithId(product.id)
      );
      let pageSize = datasetsRendering.visibleDatasetsCount;
      let remainingCount = (product.datasetWeight > 0 ? pageSize : 0);

      if (remainingCount > 0) {
        pageSize = Math.min(pageSize, product.minPageSize);
        remainingBudget -= pageSize * product.datasetWeight;
        remainingCount -= pageSize;
      }

      datasetsRendering.pageSize = pageSize;
      if (remainingCount > 0) {
        remainingCountByProduct.set(product, remainingCount);
      }
    }

    let stale = false;
    while (!stale && remainingCountByProduct.size > 0 && remainingBudget > 0) {
      const budgetPerProduct = remainingBudget / remainingCountByProduct.size;

      stale = true;
      for (const [product, remainingCount] of remainingCountByProduct) {
        const count = Math.min(
          remainingCount,
          Math.floor(budgetPerProduct / product.datasetWeight)
        );

        if (count > 0) {
          const datasetsRendering = (
            this._getDatasetsRenderingForProductWithId(product.id)
          );
          datasetsRendering.pageSize += count;
          remainingBudget -= count * product.datasetWeight;

          if (remainingCount === count) {
            remainingCountByProduct.delete(product);
          } else {
            remainingCountByProduct.set(product, remainingCount - count);
          }
          stale = false;
          if (remainingBudget <= 0) {
            break;
          }
        }
      }
    }

    // Exhaust the remaining budget
    if (stale && remainingCountByProduct.size > 0 && remainingBudget > 0) {
      const product = remainingCountByProduct.keys().next().value;
      const datasetsRendering = (
        this._getDatasetsRenderingForProductWithId(product.id)
      );
      datasetsRendering.pageSize += 1;
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {Set<Dataset>}
   */
  _getDatasetsVisibleInCurrentPage() {
    const datasets = new Set();
    for (const product of this.syntool.productsStore.getSelectedProducts()) {
      const {
        pageSize,
        currentPage,
        visibleDatasetsCount,
        sortedInMapDatasets,
      } = (
        this._getDatasetsRenderingForProductWithId(product.id)
      );
      if (pageSize > 0) {
        const maxPage = Math.ceil(visibleDatasetsCount / pageSize);
        const wrappedCurrentPage = modulo(currentPage, maxPage);

        let index = 0;
        for (const dataset of sortedInMapDatasets) {
          if (
            this._datasetsShouldBeVisible.has(dataset) &&
            wrappedCurrentPage === Math.floor(index++ / pageSize)
          ) {
            datasets.add(dataset);
          }
        }
      }
    }
    return datasets;
  },

  // #endregion
  // #region Protected -

  /**
   * @protected
   * @this {MapView}
   * @param {Iterable<DatasetInfo> | null | undefined} infos
   * @param {AbortSignal} signal
   * @returns {Promise<void>}
   */
  async _updateDatasetsFromInfos(infos, signal) {
    this._infosIterator.add(infos);

    const options = /** @type {AsyncIterOptions} */ ({
      quantum : [100, 10],
      schedule: scheduleTask,
      signal,
    });

    this._setUpdateDatasetsProgress(0, 0, 1);
    for (const iter of asyncIter(this._infosIterator, options)) {
      // eslint-disable-next-line no-await-in-loop
      for (const info of await iter) {
        this._createOrUpdateDataset(info);
      }
      this._setUpdateDatasetsProgress(
        0,
        this._infosIterator.index,
        this._infosIterator.length
      );
    }
    if (!signal.aborted) {
      this._infosIterator.reset();
      this._updateDatasetsInMapAndVisibility(signal);
    }
  },
  /**
   * @protected
   * @this {MapView}
   * @param {AbortSignal} signal
   * @returns {Promise<void>}
   */
  async _updateDatasetsInMapAndVisibility(signal) {
    const {syntool} = this;
    const viewportBounds = this.getVisibleProjectedBounds();
    const worldBounds = this.getWorldProjectedBounds();
    const mapZoom = this.getZoom();
    const datasetsToAdd = [];
    const datasetsToRemove = [];
    let progress = 0;
    let totalProgress = this._datasetsById.size;

    const options = /** @type {AsyncIterOptions} */ ({
      quantum : [100, 10],
      schedule: scheduleTask,
      signal,
    });
    for (const iter of asyncIter(this._datasetsById.values(), options)) {
      // eslint-disable-next-line no-await-in-loop
      for (const dataset of await iter) {
        const shouldBeInMap = (
          syntool.filters.dataset.inMap(dataset, viewportBounds, worldBounds)
        );
        const shouldBeVisible = (
          shouldBeInMap && syntool.filters.dataset.zoom(dataset, mapZoom)
        );
        const datasetShouldBeVisibleChanged = (
          this._datasetsShouldBeVisible.has(dataset) !== shouldBeVisible
        );
        if (this.isDatasetInMap(dataset) !== shouldBeInMap) {
          if (shouldBeInMap) {
            datasetsToAdd.push(dataset);
          } else {
            datasetsToRemove.push(dataset);
          }
        }
        if (shouldBeInMap) {
          this._markDatasetAsInMap(dataset);
        } else {
          this._markDatasetAsNotInMap(dataset);
        }
        if (datasetShouldBeVisibleChanged) {
          if (shouldBeVisible) {
            this._datasetsShouldBeVisible.add(dataset);
            this._getDatasetsRenderingForProductWithId(dataset.productId)
              .visibleDatasetsCount++;
          } else {
            this._datasetsShouldBeVisible.delete(dataset);
            this._getDatasetsRenderingForProductWithId(dataset.productId)
              .visibleDatasetsCount--;
          }
        }
        progress++;
      }
      this._setUpdateDatasetsProgress(1, progress, totalProgress);
    }

    if (signal.aborted) {
      return;
    }

    this._cancelUpdateDatasetsVisibility?.();
    this._updateProductsInMapDatasets();
    this._updateDatasetsStackLevels();
    this._distributePageSize();

    const datasetsToShow = this._getDatasetsVisibleInCurrentPage();
    const inMapCount = this.getDatasetsInMapCount();
    const shouldBeInMapCount = (
      inMapCount + datasetsToAdd.length - datasetsToRemove.length
    );

    this._updateDatasetsStatistics({
      inMapCount            : inMapCount,
      shouldBeInMapCount    : shouldBeInMapCount,
      displayableCount      : this._datasetsShouldBeVisible.size,
      displayedCount        : this._datasetsVisible.size,
      shouldBeDisplayedCount: datasetsToShow.size,
    });

    options.quantum = 100;
    options.schedule = scheduleDelayedTask(10);

    progress = 0;
    totalProgress = datasetsToRemove.length + datasetsToAdd.length;
    if (totalProgress > 0) {
      this._setUpdateDatasetsProgress(2, progress, totalProgress);
      this._updateDatasetsStatistics({changing: true});
      this.hub.emit('changing:in-map-datasets');

      for (const iter of asyncIter(datasetsToRemove, options)) {
        // eslint-disable-next-line no-await-in-loop
        for (const dataset of await iter) {
          this._removeDatasetFromMap(dataset);
          progress++;
        }

        this._setUpdateDatasetsProgress(2, progress, totalProgress);
        this._updateDatasetsStatistics({
          inMapCount    : this.getDatasetsInMapCount(),
          displayedCount: this._datasetsVisible.size,
        });
        this.hub.emit('changing:in-map-datasets');
      }
    }

    const datasetsToHide = [];
    for (const dataset of this._datasetsVisible) {
      if (!datasetsToShow.has(dataset)) {
        datasetsToHide.push(dataset);
      }
    }
    if (datasetsToHide.length > 0) {
      totalProgress += datasetsToHide.length;
      this._setUpdateDatasetsProgress(2, progress, totalProgress);

      for (const iter of asyncIter(datasetsToHide, options)) {
        // eslint-disable-next-line no-await-in-loop
        for (const dataset of await iter) {
          this._hideDataset(dataset);
          progress++;
        }

        this._setUpdateDatasetsProgress(2, progress, totalProgress);
        this._updateDatasetsStatistics(
          {displayedCount: this._datasetsVisible.size}
        );
      }
    }

    options.quantum = [100, 10];
    options.schedule = scheduleAnimationFrame;
    for (const iter of asyncIter(datasetsToAdd, options)) {
      // eslint-disable-next-line no-await-in-loop
      for (const dataset of await iter) {
        this._addDatasetToMap(dataset);
        if (datasetsToShow.delete(dataset)) {
          this._showDataset(dataset);
        }
        progress++;
      }

      this._setUpdateDatasetsProgress(2, progress, totalProgress);
      this._updateDatasetsStatistics({
        inMapCount    : this.getDatasetsInMapCount(),
        displayedCount: this._datasetsVisible.size,
      });
      this.hub.emit('changing:in-map-datasets');
    }

    if (signal.aborted) {
      return;
    }

    this.cancelUpdateDatasets = null;
    if (datasetsToRemove.length + datasetsToAdd.length > 0) {
      this.hub.emit('change:in-map-datasets');
    }

    progress = 0;
    totalProgress = datasetsToShow.size;
    for (const iter of asyncIter(datasetsToShow, options)) {
      // eslint-disable-next-line no-await-in-loop
      for (const dataset of await iter) {
        this._showDataset(dataset);
        progress++;
      }

      this._setUpdateDatasetsProgress(3, progress, totalProgress);
      this._updateDatasetsStatistics(
        {displayedCount: this._datasetsVisible.size}
      );
    }
    this._setUpdateDatasetsProgressComplete();
    this._updateDatasetsStatistics({changing: false});
  },
  /**
   * @protected
   * @this {MapView}
   * @returns {Promise<void>}
   */
  async _updateDatasetsVisibility() {
    this._cancelUpdateDatasetsVisibility?.();

    const datasetsToShow = this._getDatasetsVisibleInCurrentPage();
    const datasetsToHide = [];
    for (const dataset of this._datasetsVisible) {
      if (!datasetsToShow.has(dataset)) {
        datasetsToHide.push(dataset);
      }
    }

    const changesCount = datasetsToHide.length + datasetsToShow.size;
    if (changesCount === 0) {
      this._cancelUpdateDatasetsVisibility = null;
      return;
    }

    const abortController = new AbortController();
    const {signal} = abortController;

    this._cancelUpdateDatasetsVisibility = () => {
      abortController.abort();
    };

    let progress = 0;
    this._setUpdateDatasetsProgress(3, progress, changesCount);
    this._updateDatasetsStatistics({
      changing              : true,
      displayedCount        : this._datasetsVisible.size,
      shouldBeDisplayedCount: datasetsToShow.size,
    });

    const options = /** @type {AsyncIterOptions} */ ({
      quantum : 100,
      schedule: scheduleDelayedTask(10),
      signal,
    });
    for (const iter of asyncIter(datasetsToHide, options)) {
      // eslint-disable-next-line no-await-in-loop
      for (const dataset of await iter) {
        this._hideDataset(dataset);
        progress++;
      }

      this._setUpdateDatasetsProgress(3, progress, changesCount);
      this._updateDatasetsStatistics(
        {displayedCount: this._datasetsVisible.size}
      );
    }

    options.quantum = [100, 10];
    options.schedule = scheduleAnimationFrame;
    for (const iter of asyncIter(datasetsToShow, options)) {
      // eslint-disable-next-line no-await-in-loop
      for (const dataset of await iter) {
        this._showDataset(dataset);
        progress++;
      }

      this._setUpdateDatasetsProgress(3, progress, changesCount);
      this._updateDatasetsStatistics(
        {displayedCount: this._datasetsVisible.size}
      );
    }
    this._setUpdateDatasetsProgressComplete();
    this._updateDatasetsStatistics({changing: false});
  },

  // #endregion
};
extendClass(MapView, MapViewDatasetsExtension);
