/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {configTypes, configFieldValidators} from './config.js';
import {NotificationMessageType, NotificationStore} from './notification_store.js';
import {Types as is} from './types.js';
import {SECOND_IN_MILLISECONDS, whenReady} from './utils.js';

// #region Exported typedefs

/** @typedef {NotificationStore.MessageButton} NotificationCenter.MessageButton */
/** @typedef {NotificationStore.Message} NotificationCenter.Message */
/** @typedef {NotificationStore.MessageLike} NotificationCenter.MessageLike */
/**
 * @typedef NotificationCenter.Config
 * @prop {NotificationStore.Message[]} [messages]
 */

// #endregion

whenReady(function() {
  const MessageType = is.oneOf(Object.keys(NotificationMessageType));
  const Message = is.shape({
    id      : is.string.isRequired,
    type    : MessageType.isRequired,
    text    : is.string.isRequired,
    showOnce: is.bool,
  });
  const UniqueMessage = is.allOfType([Message, is.uniquePropInArray('id')]);
  Object.assign(configTypes, {
    MessageType: MessageType,
    Message    : Message,
  });
  Object.assign(configFieldValidators, {
    messages: is.arrayOf(UniqueMessage.isRequired),
  });
});

export {NotificationMessageType, NotificationStore};
export class NotificationCenter {
  // #region Public

  /**
   * @param {NotificationStore} [store]
   */
  constructor(store) {
    if (!store) {
      store = new NotificationStore();
    }
    this.store = store;
    /** @type {Record<string, TimeoutId>} */
    this._closeTimerIdsByMessageId = {};

    store.hub.on('add-message', this.handleAddMessage, this);
    store.hub.on('update-message', this.handleUpdateMessage, this);
    store.hub.on('remove-message', this.handleRemoveMessage, this);
  }

  /**
   * @param {NotificationStore.Message} message
   * @returns {void}
   */
  handleAddMessage(message) {
    if (isFinite(message.hideAfter) && message.hideAfter >= 0) {
      this._closeTimerIdsByMessageId[message.id] = setTimeout(
        this.closeMessage.bind(this, message.id),
        message.hideAfter
      );
    }
  }
  /**
   * @param {NotificationStore.Message} message
   * @returns {void}
   */
  handleUpdateMessage(message) {
    this.handleRemoveMessage(message.id);
    this.handleAddMessage(message);
  }
  /**
   * @param {string} messageId
   * @returns {void}
   */
  handleRemoveMessage(messageId) {
    if (this._closeTimerIdsByMessageId[messageId]) {
      clearTimeout(this._closeTimerIdsByMessageId[messageId]);
      delete this._closeTimerIdsByMessageId[messageId];
    }
  }

  /**
   * @param {NotificationCenter.Config} config
   * @returns {Promise<void>}
   */
  async loadMessagesFromConfig({messages}) {
    if (messages) {
      const store = this.store;
      await store.loadShownOnceIds();
      for (const message of messages) {
        store.addMessage(message);
      }
    }
  }

  /**
   * @param {string} messageId
   * @returns {void}
   */
  closeMessage(messageId) {
    this.store.removeMessageWithId(messageId);
  }
  /**
   * @param {NotificationStore.MessageLike} message
   * @returns {string | null}
   */
  showMessage(message) {
    return this.store.addMessage(message);
  }
  /**
   * @param {string} infoText
   * @param {Milliseconds} [hideAfter = 10_000]
   * @returns {string | null}
   */
  showInfo(infoText, hideAfter) {
    if (!isFinite(hideAfter)) {
      hideAfter = 10 * SECOND_IN_MILLISECONDS;
    }
    return this.showMessage({
      text     : infoText,
      type     : NotificationMessageType.info,
      hideAfter: hideAfter,
    });
  }
  /**
   * @param {string} warningText
   * @param {Milliseconds} [hideAfter = 10_000]
   * @returns {string | null}
   */
  showWarning(warningText, hideAfter) {
    if (!isFinite(hideAfter)) {
      hideAfter = 10 * SECOND_IN_MILLISECONDS;
    }
    return this.showMessage({
      text     : warningText,
      type     : NotificationMessageType.warning,
      hideAfter: hideAfter,
    });
  }
  /**
   * @param {string} errorText
   * @param {Milliseconds} [hideAfter = 10_000]
   * @returns {string | null}
   */
  showError(errorText, hideAfter) {
    if (!isFinite(hideAfter)) {
      hideAfter = 10 * SECOND_IN_MILLISECONDS;
    }
    return this.showMessage({
      text     : errorText,
      type     : NotificationMessageType.error,
      hideAfter: hideAfter,
    });
  }

  // #endregion
}
