/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {UserStorage} from './user_storage.js';
import {$, EventsHub} from './utils.js';

// #region Exported typedefs

/**
 * @typedef NotificationStore.MessageButton
 * @prop {HTML} [text]
 * @prop {boolean} [closes]
 * @prop {ButtonProps} [props]
 * @prop {() => void} [onClick]
 */
/**
 * @typedef NotificationStore.Message
 * @prop {string} id
 * @prop {Values<NotificationMessageType>} type
 * @prop {HTML} text
 * @prop {boolean} [showOnce = false]
 * @prop {Milliseconds} addedAt
 *
 * @prop {HIDKey | null} [key]
 * @prop {HIDKey[] | null} [modifiers]
 * @prop {NotificationStore.MessageButton[]} [buttons = []]
 * @prop {Omit<NotificationStore.MessageButton, 'closes'>} [closeButton]
 * @prop {Milliseconds} [hideAfter = -1]
 */
/**
 * @typedef {Omit<Message, 'id' | 'addedAt' | 'hideAfter'>
 *           & {id?: string, hideAfter?: Milliseconds}} NotificationStore.MessageLike
 */

// #endregion

const AUTO_ID_PREFIX = '_NotificationCenter_';
const SHOWN_ONCE_STORAGE_KEY = 'NotificationStore.shownOnceIds';
let nextAutoId = 0;

export const NotificationMessageType = /** @type {const} */ ({
  info   : 'info',
  warning: 'warning',
  error  : 'error',
});

export class NotificationStore {
  // #region Private typedefs

  /** @typedef {import('./utils.js').HIDKey} HIDKey */
  /** @typedef {NotificationStore.Message} Message */
  /** @typedef {NotificationStore.MessageLike} MessageLike */
  /**
   * @typedef {Partial<PickWritableProperties<HTMLButtonElement>>
   *           & Record<string, string>} ButtonProps
   */

  // #endregion
  // #region Public

  constructor() {
    /** @type {Message[]} */
    this.messages = [];
    /** @type {Set<string>} */
    this.shownOnceIds = new Set();

    this.hub = new EventsHub();
  }

  /**
   * @returns {string[]}
   */
  toJSON() {
    return [...this.shownOnceIds];
  }
  /**
   * @param {string[]} json
   * @returns {void}
   */
  fromJSON(json) {
    for (const id of json) {
      if (!id.startsWith(AUTO_ID_PREFIX)) {
        this.shownOnceIds.add(id);
      }
    }
    this.saveShownOnceIds();
  }

  async loadShownOnceIds() {
    try {
      const shownOnceIds = await UserStorage.get(SHOWN_ONCE_STORAGE_KEY);
      if (shownOnceIds) {
        for (const id of shownOnceIds) {
          if (!id.startsWith(AUTO_ID_PREFIX)) {
            this.shownOnceIds.add(id);
          }
        }

        const l = this.messages.length;
        let i = l;
        while (i--) {
          const message = $(this.messages[i]);
          if (message.showOnce && this.shownOnceIds.has(message.id)) {
            this.messages.splice(i, 1);
            this.hub.emit('remove-message', message.id);
          }
        }
        if (this.messages.length < l) {
          this.hub.emit('change');
        }
      }
    } catch (error) {
      /* ignore errors */
    }
  }
  async saveShownOnceIds() {
    try {
      await UserStorage.set(SHOWN_ONCE_STORAGE_KEY, [...this.shownOnceIds]);
    } catch (error) {
      /* ignore errors */
    }
  }

  /**
   * @param {MessageLike} message
   * @returns {Message['id'] | null} `id` of the added message or `null` if it
   * was ignored (e.g. it was already shown).
   */
  addMessage(message) {
    const msg = /** @type {Message} */ (message);
    msg.addedAt = Date.now();
    if (msg.id) {
      if (msg.showOnce && this.shownOnceIds.has(msg.id)) {
        return null;
      }
      for (let i = 0, l = this.messages.length; i < l; i++) {
        if (this.messages[i]?.id === msg.id) {
          this.messages[i] = msg;
          this.hub.emit('update-message', msg);
          this.hub.emit('change');
          return msg.id;
        }
      }
    } else {
      msg.id = AUTO_ID_PREFIX + nextAutoId++;
    }
    this.messages.push(msg);
    this.hub.emit('add-message', msg);
    this.hub.emit('change');
    return msg.id;
  }
  /**
   * @param {string} id
   * @returns {Message | undefined}
   */
  getMessageWithId(id) {
    for (const message of this.messages) {
      if (message.id === id) {
        return message;
      }
    }
    return undefined;
  }
  /**
   * @param {string} id
   * @returns {void}
   */
  removeMessageWithId(id) {
    for (let i = 0, l = this.messages.length; i < l; i++) {
      const message = $(this.messages[i]);
      if (message.id === id) {
        if (!id.startsWith(AUTO_ID_PREFIX) && message.showOnce) {
          this.shownOnceIds.add(id);
          this.saveShownOnceIds();
        }
        this.messages.splice(i, 1);
        this.hub.emit('remove-message', id);
        this.hub.emit('change');
        return;
      }
    }
  }

  // #endregion
}
