/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  $,
  emptyFunctionReturningFalse,
  emptyFunctionReturningTrue,
  memo,
  RegExpHelpers,
} from './utils.js';

class ParseError extends Error {
  /**
   * @param {string} issue
   * @param {string} pattern
   * @param {number} position
   */
  constructor(issue, pattern, position) {
    super(`${issue} at index ${position} in "${pattern}."`);
    this.issue = issue;
    this.pattern = pattern;
    this.position = position;
  }
}

const OP_ESCAPE = /** @type {const} */ ('\\');
const OP_OPEN_PAREN = /** @type {const} */ ('(');
const OP_CLOSE_PAREN = /** @type {const} */ (')');
const OP_AND = /** @type {const} */ ('&');
const OP_OR = /** @type {const} */ ('|');
const OP_NOT = /** @type {const} */ ('!');

const orderedOperators = /** @type {const} */ ([OP_OR, OP_AND, OP_NOT]);
const LOW_OP = orderedOperators[0];
const precedence = /** @type {Record<Pattern.Operator, number>} */ (
  Object.fromEntries(orderedOperators.map((op, p) => [op, p]))
);

// #region Exported typedefs

/** @typedef {typeof orderedOperators extends ReadonlyArray<infer T> ? T : never} Pattern.Operator */
/** @typedef {Pattern} Pattern.Atom */
/** @typedef {Pattern.Atom | ASTNode} Pattern.AST */
/**
 * @template T
 * @typedef {{[K in Pattern.Operator]: ExecOperator<T>} & {'': ExecAtom<T>}} Pattern.ExecOperators
 */

// #endregion

export const Pattern = {
  // #region Private typedefs

  /** @typedef {[Pattern.Operator, OneOrMore<Pattern.AST>]} ASTNode */
  /**
   * @template T
   * @typedef {(parentOp: Pattern.Operator, atom: Pattern.Atom) => T} ExecAtom
   */
  /**
   * @template T
   * @typedef {(parentOp: Pattern.Operator, ...operands: T[]) => T} ExecOperator
   */

  // #endregion

  Op: {
    AND: OP_AND,
    OR: OP_OR,
    NOT: OP_NOT,
  },
  or,
  and,
  not,

  ParseError,
  isLegacy,
  canParse,
  parse,

  AST: {
    exec: execAST,
    split: splitAST,
    toRegExp: astToRegExp,
    stringify: stringifyAST,
    toLegacy: toLegacyAST,
    simplify: simplifyAST,
  },

  split,
  toRegExp,
  compile,
  test,
  escape,
  unescape,
  atomToLegacy,
  toLegacy,
  fromLegacy,
  toHybrid,
  simplify,
  normalize,
};

/**
 * @param {Pattern.Operator} op
 * @param {ReadonlyArray<Pattern.AST>} operands
 * @returns {Pattern.AST}
 */
function makeOperand(op, operands) {
  if (operands.length === 0) {
    return op === OP_NOT ? '*' : '';
  }
  if (operands.length === 1) {
    if (op !== OP_NOT) {
      // or(a) <=> a
      // and(aa) <=> aa
      return $(operands[0]);
    }
    if (typeof operands[0] !== 'string' && op === $(operands[0])[0]) {
      // not(not(x)) <=> x
      return $(operands[0])[1][0];
    }
    return [op, /** @type {OneOrMore<Pattern.AST>} */ (operands)];
  }
  if (
    op !== OP_NOT &&
    operands.every((o) => typeof o === 'string' || op === o[0])
  ) {
    // or(a, or(b, c)) <=> or(a, b, c)
    // and(a, and(b, c)) <=> and(a, b, c)
    operands = operands.flatMap((o) => (typeof o === 'string' ? o : o[1]));
  }
  return [op, /** @type {OneOrMore<Pattern.AST>} */ (operands)];
}

/**
 * @param {ReadonlyArray<Pattern.AST>} operands
 * @returns {Pattern.AST}
 */
function or(operands) {
  return makeOperand(OP_OR, operands);
}
/**
 * @param {ReadonlyArray<Pattern.AST>} operands
 * @returns {Pattern.AST}
 */
function and(operands) {
  return makeOperand(OP_AND, operands);
}
/**
 * @param {Pattern.AST} operand
 * @returns {Pattern.AST}
 */
function not(operand) {
  return makeOperand(OP_NOT, [operand]);
}

/**
 * @param {string | null | undefined} pattern
 * @returns {pattern is LegacyPattern}
 */
function isLegacy(pattern) {
  return !pattern || pattern[0] !== OP_OPEN_PAREN;
}
/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @returns {Pattern.AST}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @returns {Pattern.AST}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @returns {Pattern.AST}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @returns {Pattern.AST}
 */
function parse(pattern, legacy = false) {
  if (!pattern) {
    return pattern === '' && !legacy ? '' : '*';
  }
  if (legacy && (legacy !== 'auto' || isLegacy(pattern))) {
    const index = pattern.indexOf(',');
    if (index > -1) {
      throw new ParseError('Commas "," are not allowed', pattern, index);
    }
    return `*${escape(pattern)}*`;
  }

  /** @type {[Pattern.Operator, Pattern.AST[]][]} */
  const stack = [];
  /** @type {Pattern.AST[]} */
  let operands = [];
  /** @type {Pattern.Operator} */
  let op = LOW_OP;
  /** @type {Pattern.AST | null} */
  let operand = null;
  let escaping = false;
  let buffer = '';
  let i = 0;
  /** @type {string} */
  let c;

  function pushChar() {
    if (c !== ' ' || buffer.length > 0) {
      if (c === ',') {
        throw new ParseError('Commas "," are not allowed', $(pattern), i);
      }
      if (operand !== null) {
        throw new ParseError(`Unexpected character "${c}"`, $(pattern), i);
      }
      if ((c === '*' || c === '?' || c === OP_ESCAPE) && escaping) {
        buffer += OP_ESCAPE;
      }
      buffer += c;
    }
  }
  function buffer2operand() {
    operand = buffer.trim();
    buffer = '';
  }
  function pushOperand() {
    if (operand === null) {
      buffer2operand();
    }
    operands.push($(operand));
    operand = null;
  }
  function openParens() {
    if (buffer || operand !== null) {
      throw new ParseError(`Unexpected character "${c}"`, $(pattern), i);
    }
    stack.push([op, operands]);
    operands = [];
  }
  function closeParens() {
    operand = makeOperand(op, operands);
    if (stack.length === 0) {
      throw new ParseError('Mismatched closing parenthesis', $(pattern), i);
    }
    [op, operands] = $(stack.pop());
  }

  for (c of pattern) {
    if (escaping) {
      pushChar();
      escaping = false;
      i++;
      continue;
    }

    switch (c) {
    case OP_ESCAPE:
      escaping = true;
      break;
    case OP_OPEN_PAREN:
      openParens();
      op = LOW_OP;
      break;
    case OP_CLOSE_PAREN:
      pushOperand();
      while (op !== LOW_OP) {
        closeParens();
        pushOperand();
      }
      closeParens();
      break;
    case OP_OR:
    case OP_AND:
      if (operand === null) {
        buffer2operand();
      }
      // op has higher precedence
      while (op !== c && precedence[op] >= precedence[c]) {
        pushOperand();
        closeParens();
      }
      // op has lower precedence
      if (precedence[op] < precedence[c]) {
        /** @type {Pattern.AST | null} */
        const o = operand;
        operand = null;
        openParens();
        operand = o;
        op = c;
      }
      pushOperand();
      break;
    case OP_NOT:
      openParens();
      op = c;
      break;
    default:
      pushChar();
    }
    i++;
  }
  pushOperand();
  while (stack.length > 0) {
    closeParens();
    pushOperand();
  }

  return makeOperand(op, operands);
}
/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @returns {pattern is Pattern}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @returns {pattern is LegacyPattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @returns {pattern is HybridPattern}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @returns {boolean}
 */
function canParse(pattern, legacy = false) {
  try {
    parse(pattern, /** @type {any} */ (legacy));
    return true;
  } catch (error) {
    return false;
  }
}

/**
 * @template T
 * @param {Pattern.AST} ast
 * @param {Pattern.ExecOperators<T>} operators
 * @param {Pattern.Operator} [parentOp]
 * @returns {T}
 */
function execAST(ast, operators, parentOp = LOW_OP) {
  if (typeof ast === 'string') {
    return operators[''](parentOp, ast);
  }
  const [op, operands] = ast;
  const args = operands.map((operand) => execAST(operand, operators, op));
  return operators[op](parentOp, ...args);
}

/**
 * @param {Pattern.AST} ast
 * @param {Pattern.Operator} operator
 * @returns {Pattern.AST[]}
 */
function splitAST(ast, operator) {
  if (typeof ast === 'string' || ast[0] !== operator) {
    return [ast];
  }
  return ast[1];
}

/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {Pattern.Operator} operator
 * @param {false} [legacy]
 * @returns {Pattern[]}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {Pattern.Operator} operator
 * @param {true} legacy
 * @returns {Pattern[]}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {Pattern.Operator} operator
 * @param {'auto'} legacy
 * @returns {Pattern[]}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {Pattern.Operator} operator
 * @param {boolean | 'auto'} [legacy]
 * @returns {Pattern[]}
 */
function split(pattern, operator, legacy = false) {
  return (
    splitAST(parse(pattern, /** @type {any} */ (legacy)), operator)
    .map(stringifyASTNonLegacy)
  );
}

/**
 * @param {Pattern.AST} ast
 * @returns {Pattern}
 */
function stringifyASTNonLegacy(ast) {
  return stringifyAST(ast, false);
}

/**
 * @param {unknown} _
 * @param {boolean} esc
 * @param {string} c
 * @returns {string}
 */
function atom2re(_, esc, c) {
  if (!esc) {
    if (c === '*') {
      return '.*';
    }
    if (c === '?') {
      return '.';
    }
  }
  return `\\${c}`;
}
const {escape: reEscape} = RegExpHelpers;
const atom2reRe = (
  new RegExp(`(${reEscape(OP_ESCAPE)}?)([\\\\^$.+()[\\]{}|*?])`, 'g')
);
/** @type {Pattern.ExecOperators<string>} */
const toRegExpOperators = {
  '': (_, a) => `^${a.replace(atom2reRe, atom2re)}$`,
  [OP_OR]: (_, ...a) => a.join('|'),
  [OP_AND]: (_, ...a) => `(?=${a.join(')(?=')})^.*$`,
  [OP_NOT]: (_, a) => `(?!${a})^.*$`,
};
/**
 * @param {Pattern.AST} ast
 * @returns {RegExp}
 */
function astToRegExp(ast) {
  return new RegExp(execAST(simplifyAST(ast), toRegExpOperators), 'i');
}
/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @returns {RegExp}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @returns {RegExp}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @returns {RegExp}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @returns {RegExp}
 */
function toRegExp(pattern, legacy = false) {
  return astToRegExp(parse(pattern, /** @type {any} */ (legacy)));
}

/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @returns {(string: string) => boolean}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @returns {(string: string) => boolean}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @returns {(string: string) => boolean}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @returns {(string: string) => boolean}
 */
function compile(pattern, legacy = false) {
  const ast = simplifyAST(parse(pattern, /** @type {any} */ (legacy)));
  if (ast === '') {
    return emptyFunctionReturningFalse;
  }
  if (ast === '*') {
    return emptyFunctionReturningTrue;
  }

  const re = astToRegExp(ast);
  return re.test.bind(re);
}

const memoCompile = memo(compile, {key: ([p]) => p});

/**
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @returns {(string: string) => boolean}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {string} string
 * @returns {boolean}
 *
 * @param {HybridPattern | null | undefined} pattern
 * @param {string} [string]
 * @returns {boolean | ((string: string) => boolean)}
 */
function test(pattern, string) {
  if (string === undefined) {
    return memoCompile(pattern, 'auto');
  }
  return pattern ? memoCompile(pattern, 'auto')(string) : true;
}

const escapableChars = reEscape(
  '*?' + OP_ESCAPE + OP_OPEN_PAREN + OP_CLOSE_PAREN + OP_AND + OP_OR + OP_NOT
);
const escapeRe = new RegExp(`[${escapableChars}]`, 'g');

/**
 * @param {string} string
 * @returns {Pattern.Atom}
 */
function escape(string) {
  return string.replace(escapeRe, '\\$&');
}

const unescapeRe = new RegExp(`${reEscape(OP_ESCAPE)}(.)`, 'g');

/**
 * @param {Pattern.Atom} string
 * @returns {string}
 */
function unescape(string) {
  return string.replace(unescapeRe, '$1');
}

const legacyRe = new RegExp(
  `^\\*(?:(?:[^${escapableChars}]*(?:${reEscape(OP_ESCAPE)}.)*)*\\*)?$`
);

/**
 * @param {Pattern.Atom} atom
 * @returns {LegacyPattern | undefined}
 */
function atomToLegacy(atom) {
  return legacyRe.test(atom) ? unescape(atom.slice(1, -1)) : undefined;
}
/**
 * @param {Pattern.AST} ast
 * @returns {LegacyPattern | undefined}
 */
function toLegacyAST(ast) {
  ast = simplifyAST(ast);
  return typeof ast === 'string' ? atomToLegacy(ast) : undefined;
}
/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @returns {LegacyPattern | undefined}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @returns {LegacyPattern | undefined}
 */
function toLegacy(pattern, legacy = false) {
  return toLegacyAST(parse(pattern, /** @type {any} */ (legacy)));
}
/**
 * @param {LegacyPattern | null | undefined} pattern
 * @returns {Pattern}
 */
function fromLegacy(pattern) {
  return stringifyAST(parse(pattern, true));
}

/**
 * @overload
 * @param {Pattern} pattern
 * @param {false} [legacy]
 * @returns {HybridPattern}
 *
 * @overload
 * @param {LegacyPattern} pattern
 * @param {true} legacy
 * @returns {HybridPattern}
 *
 * @overload
 * @param {HybridPattern} pattern
 * @param {'auto'} legacy
 * @returns {HybridPattern}
 *
 * @param {Pattern | LegacyPattern | HybridPattern} pattern
 * @param {boolean | 'auto'} [legacy]
 * @returns {HybridPattern}
 */
function toHybrid(pattern, legacy = false) {
  if (legacy) {
    return pattern;
  }
  return OP_OPEN_PAREN + pattern;
}

/**
 * @param {Pattern.Operator} parentOp
 * @param {Pattern.Operator} op
 * @param {Pattern} pattern
 * @returns {Pattern}
 */
function group(parentOp, op, pattern) {
  return (
    precedence[parentOp] > precedence[op]
    ? OP_OPEN_PAREN + pattern + OP_CLOSE_PAREN
    : pattern
  );
}
const stringifyRe = new RegExp(
  '[' + OP_OPEN_PAREN + OP_CLOSE_PAREN + OP_AND + OP_OR + OP_NOT + ']',
  'g'
);
/** @type {Pattern.ExecOperators<Pattern>} */
const stringifyOperators = {
  '': (_, a) => a.replace(stringifyRe, '\\$&'),
  [OP_OR]: (_, ...a) => (a.length === 1 ? $(a[0]) : group(_, OP_OR, a.join(OP_OR))),
  [OP_AND]: (_, ...a) => (a.length === 1 ? $(a[0]) : group(_, OP_AND, a.join(OP_AND))),
  [OP_NOT]: (_, a) => group(_, OP_NOT, OP_NOT + a),
};

/**
 * @overload
 * @param {Pattern.AST} ast
 * @param {false} [legacy]
 * @returns {Pattern}
 *
 * @overload
 * @param {Pattern.AST} ast
 * @param {true} legacy
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {Pattern.AST} ast
 * @param {'auto'} legacy
 * @returns {HybridPattern}
 *
 * @param {Pattern.AST} ast
 * @param {boolean | 'auto'} [legacy]
 * @returns {Pattern | LegacyPattern | HybridPattern | undefined}
 */
function stringifyAST(ast, legacy = false) {
  if (!legacy) {
    return execAST(ast, stringifyOperators);
  }

  const legacyPattern = toLegacyAST(ast);
  if (legacy === 'auto') {
    return (
      legacyPattern === undefined
      ? OP_OPEN_PAREN + execAST(ast, stringifyOperators) + OP_CLOSE_PAREN
      : legacyPattern
    );
  }

  return legacyPattern;
}

/**
 * @param {unknown} a
 * @returns {boolean}
 */
function isAlways(a) {
  return a === '*';
}
/**
 * @param {unknown} a
 * @returns {boolean}
 */
function isNotAlways(a) {
  return a !== '*';
}
/**
 * @param {unknown} a
 * @returns {boolean}
 */
function isNever(a) {
  return a === '';
}
/**
 * @param {unknown} a
 * @returns {boolean}
 */
function isNotNever(a) {
  return a !== '';
}
const wildcardRe = /[?*]*\*[?*]*/g;
/** @type {Pattern.ExecOperators<Pattern.AST>} */
const simplifyOperators = {
  '': (_, a) => a.replace(wildcardRe, '*'),
  [OP_OR]: (_, ...a) => (a.some(isAlways) ? '*' : or(a.filter(isNotNever))),
  [OP_AND]: (_, ...a) => (a.some(isNever) ? '' : and(a.filter(isNotAlways))),
  [OP_NOT]: (_, a) => (isAlways(a) ? '' : (isNever(a) ? '*' : not(a))),
};
/**
 * @param {Pattern.AST} ast
 * @returns {Pattern.AST}
 */
function simplifyAST(ast) {
  return execAST(ast, simplifyOperators);
}
/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @param {false} [asLegacy]
 * @returns {Pattern}
 *
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} legacy
 * @param {true} asLegacy
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} legacy
 * @param {'auto'} asLegacy
 * @returns {HybridPattern}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @param {true} [asLegacy]
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @param {false} asLegacy
 * @returns {Pattern}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @param {'auto'} asLegacy
 * @returns {HybridPattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @param {'auto'} [asLegacy]
 * @returns {HybridPattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @param {false} asLegacy
 * @returns {Pattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @param {true} asLegacy
 * @returns {LegacyPattern | undefined}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @param {boolean | 'auto'} [asLegacy]
 * @returns {Pattern | LegacyPattern | HybridPattern | undefined}
 */
function simplify(pattern, legacy = false, asLegacy = legacy) {
  return stringifyAST(
    simplifyAST(parse(pattern, /** @type {any} */ (legacy))),
    /** @type {any} */ (asLegacy)
  );
}
/**
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} [legacy]
 * @param {false} [asLegacy]
 * @returns {Pattern}
 *
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} legacy
 * @param {true} asLegacy
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {Pattern | null | undefined} pattern
 * @param {false} legacy
 * @param {'auto'} asLegacy
 * @returns {HybridPattern}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @param {true} [asLegacy]
 * @returns {LegacyPattern | undefined}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @param {false} asLegacy
 * @returns {Pattern}
 *
 * @overload
 * @param {LegacyPattern | null | undefined} pattern
 * @param {true} legacy
 * @param {'auto'} asLegacy
 * @returns {HybridPattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @param {'auto'} [asLegacy]
 * @returns {HybridPattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @param {false} asLegacy
 * @returns {Pattern}
 *
 * @overload
 * @param {HybridPattern | null | undefined} pattern
 * @param {'auto'} legacy
 * @param {true} asLegacy
 * @returns {LegacyPattern | undefined}
 *
 * @param {Pattern | LegacyPattern | HybridPattern | null | undefined} pattern
 * @param {boolean | 'auto'} [legacy]
 * @param {boolean | 'auto'} [asLegacy]
 * @returns {Pattern | LegacyPattern | HybridPattern | undefined}
 */
function normalize(pattern, legacy = false, asLegacy = legacy) {
  return simplify(
    pattern?.toLowerCase(),
    /** @type {any} */ (legacy),
    /** @type {any} */ (asLegacy)
  );
}
