/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
// @ts-nocheck
/*eslint-disable no-extend-native */
import {flattenArray} from './utils/utils.js';

/*For all*/
polyfillQueueMicrotask();
polyfillObjectHasOwn();
polyfillObjectFromEntries();
polyfillArrayFlatMap();
polyfillPromiseFinally();
polyfillPromiseAllSettled();
polyfillPromiseWithResolvers();
polyfillAbortSignalAndController();
polyfillFetchAbortion();
polyfillURLCanParse();

function polyfillQueueMicrotask() {
  if (!window.queueMicrotask) {
    window.queueMicrotask = (fn) => {
      Promise.resolve(undefined).then(fn);
    };
  }
}

function polyfillObjectHasOwn() {
  if (!Object.hasOwn) {
    Object.defineProperty(Object, 'hasOwn', {
      value(object, property) {
        if (object == null) {
          throw new TypeError('Cannot convert undefined or null to object');
        }
        return Object.prototype.hasOwnProperty.call(Object(object), property);
      },
      configurable: true,
      enumerable: false,
      writable: true
    });
  }
}
function polyfillObjectFromEntries() {
  if (!Object.fromEntries) {
    Object.fromEntries = function(iterable) {
      const entries = (
        !Array.isArray(iterable) && typeof iterable.entries === 'function'
        ? iterable.entries()
        : iterable
      );
      const object = {};
      let entry;
      if (typeof entries.next === 'function') {
        while ((entry = entries.next()) && !entry.done) {
          object[entry.value[0]] = entry.value[1];
        }
      } else {
        for (let i = 0, l = entries.length; i < l; i++) {
          entry = entries[i];
          object[entry[0]] = entry[1];
        }
      }
      return object;
    };
  }
}
function polyfillArrayFlatMap() {
  Array.prototype.flatMap ??= function(callback, thisArg) {
    return flattenArray(this.map(callback, thisArg));
  };
}
function polyfillPromiseFinally() {
  if (typeof Promise.prototype.finally !== 'function') {
    Promise.prototype.finally = function(onFinally) {
      const isFunction = typeof onFinally == 'function';
      return this.then(
        isFunction ? async function(x) {
          await onFinally();
          return x;
        } : onFinally,
        isFunction ? async function(e) {
          await onFinally();
          throw e;
        } : onFinally
      );
    };
  }
}
function polyfillPromiseAllSettled() {
  function reflectOnFulfilled(value) {
    return {value, status: 'fulfilled'};
  }
  function reflectOnRejected(reason) {
    return {reason, status: 'rejected'};
  }

  if (typeof Promise.allSettled !== 'function') {
    Promise.allSettled = function(iterable) {
      return Promise.all(Array.from(iterable, function(item) {
        try {
          return (
            Promise.resolve(item).then(reflectOnFulfilled, reflectOnRejected)
          );
        } catch (error) {
          return Promise.reject(error);
        }
      }));
    };
  }
}
function polyfillPromiseWithResolvers() {
  if (typeof Promise.withResolvers !== 'function') {
    Promise.withResolvers = function() {
      let resolve;
      let reject;
      const promise = new this((rv, rj) => {
        resolve = rv;
        reject = rj;
      });
      return {promise, resolve, reject};
    };
  }
}
function polyfillAbortSignalAndController() {
  if (!window.AbortController) {
    window.AbortSignal = function AbortSignal() {
      var delegate = document.createDocumentFragment();

      this.addEventListener = delegate.addEventListener.bind(delegate);
      this.dispatchEvent = delegate.dispatchEvent.bind(delegate);
      this.removeEventListener = delegate.removeEventListener.bind(delegate);
      this.aborted = false;
      this.reason = undefined;
    };
    window.AbortController = function AbortController() {
      this.signal = new AbortSignal();
    };
    window.AbortController.prototype.abort = function abort(reason) {
      if (reason === undefined) {
        reason = new DOMException('The operation was aborted.', 'AbortError');
      }
      this.signal.reason = reason;
      this.signal.aborted = true;
      this.signal.dispatchEvent(new Event('abort'));
    };
  } else if (!Object.getOwnPropertyDescriptor(AbortSignal.prototype, 'reason')) {
    const AbortControllerAbort = AbortController.prototype.abort;
    const AbortSignalReason = Symbol('AbortSignalReason');
    AbortController.prototype.abort = function abort(reason) {
      if (reason === undefined) {
        reason = new DOMException('The operation was aborted.', 'AbortError');
      }
      this.signal[AbortSignalReason] = reason;
      return AbortControllerAbort.call(this);
    };
    Object.defineProperty(AbortSignal.prototype, 'reason', {
      configurable: true,
      enumerable: true,
      get() {
        return this[AbortSignalReason];
      },
    });
  }
  if (!AbortSignal.abort) {
    AbortSignal.abort = function(reason) {
      const abortController = new AbortController();
      abortController.abort(reason);
      return abortController.signal;
    };
  }
  if (!AbortSignal.timeout) {
    AbortSignal.timeout = function(time) {
      const abortController = new AbortController();
      setTimeout(function() {
        abortController.abort(
          new DOMException('The operation timed out.', 'TimeoutError')
        );
      }, time);
      return abortController.signal;
    };
  }
  if (!AbortSignal.prototype.throwIfAborted) {
    AbortSignal.prototype.throwIfAborted = function() {
      if (this.aborted) {
        throw this.reason;
      }
    };
  }
}
function polyfillFetchAbortion() {
  if (
    window.fetch &&
    window.Request &&
    !Object.hasOwn(window.Request.prototype, 'signal')
  ) {
    window.fetch = makeAbortableFetch(window.fetch);
    window.Request = makeAbortableRequest(window.Request);
  }

  function makeAbortableRequest(NativeRequest) {
    function Request(input, init) {
      var request = new NativeRequest(input, init);
      if (init && init.signal) {
        request.signal = init.signal;
      }
      return request;
    }
    Request.prototype = NativeRequest.prototype;
    return Request;
  }
  function makeAbortableFetch(fetch) {
    return function abortableFetch(input, init) {
      var signal = (
        input instanceof Request
        ? input.signal
        : (init ? init.signal : undefined)
      );

      if (signal) {
        // Return early if already aborted, thus avoiding making an HTTP request
        if (signal.aborted) {
          return Promise.reject(signal.reason);
        }

        // Turn an event into a promise that gets rejected once `abort` is
        // dispatched
        var abortPromise = new Promise(function(_, reject) {
          signal.addEventListener('abort', function handlerAbort() {
            signal.removeEventListener('abort', handlerAbort);
            reject(signal.reason);
          });
        });

        var responsePromise = fetch(input, init);
        // Return the fastest promise (don't need to wait for request to finish)
        var promise = new Promise(function(resolve, reject) {
          abortPromise.catch(reject);
          responsePromise.then(resolve, reject);
        });

        if (window.ReadableStream) {
          promise = promise.then(function cancelBodyOnAbort(response) {
            abortPromise.catch(function cancelBody(reason) {
              if (response.body && !response.body.locked) {
                response.body.cancel(reason);
              }
            });
            return (signal.aborted ? abortPromise : response);
          });
        }

        return promise;
      }

      return fetch(input, init);
    };
  }
}
function polyfillURLCanParse() {
  if (!URL.canParse) {
    URL.canParse = function(url, base) {
      try {
        if (base === undefined) {
          // In Safari 14 and earlier, calling the URL constructor with a base
          // URL whose value is `undefined` causes Safari to throw a TypeError;
          // see WebKit bug 216841.

          // eslint-disable-next-line no-new
          new URL(url);
        } else {
          // eslint-disable-next-line no-new
          new URL(url, base);
        }
        return true;
      } catch (error) {
        return false;
      }
    };
  }
}
