/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {assertValidProduct, normalizeProduct} from './config.js';
import {Dataset} from './dataset.js';
import {Pattern} from './pattern.js';
import {
  HOUR_IN_MILLISECONDS,
  $,
  clamp,
  floor,
  round,
  normalizeHTML,
  EventsHub,
} from './utils.js';

const NUM_STACK_SUB_LEVEL = 100;

/**
 * @param {string} tag
 * @returns {string}
 */
function stripTagKey(tag) {
  return tag.replace(/^[^:]+:/, '').trim();
}
/**
 * @param {Readonly<Product> | ProductId} productOrId
 * @returns {ProductId}
 */
function idFromProductOrId(productOrId) {
  return (typeof productOrId === 'string' ? productOrId : productOrId.id);
}
/**
 * @param {DatasetFiltersGroup} group
 * @returns {Pattern.AST}
 */
function emptyPatternGroupMarker(group) {
  return Pattern.or(
    ['', Pattern.and([Pattern.normalize(Pattern.escape(group.label)), ''])]
  );
}

// #region Exported typedefs

/**
 * @typedef ProductsStore.ProductState
 * @prop {number} opacity
 * @prop {number} stackLevel
 * @prop {LegacyPattern} pattern
 * @prop {Record<string, boolean>} filters
 */
/**
 * @typedef ProductsStore.State
 * @prop {Record<ProductId, ProductsStore.ProductState>} products
 * @prop {ProductId[]} selection
 */
/**
 * @typedef ProductsStore.PartialState
 * @prop {Record<ProductId, Partial<ProductsStore.ProductState>>} [products]
 * @prop {ProductId[]} [selection]
 */

// #endregion

export class ProductsStore {
  // #region Private typedefs

  /** @typedef {ProductsStore.ProductState} ProductState */
  /** @typedef {ProductsStore.State} State */
  /** @typedef {ProductsStore.PartialState} PartialState */
  /**
   * @typedef ProductFilter
   * @prop {Set<string>} activeFilters
   * @prop {boolean} customActive
   * @prop {LegacyPattern} customPattern
   * @prop {HybridPattern | undefined} computedPattern
   */

  // #endregion
  // #region Public

  /**
   * @public
   */
  constructor() {
    /** @type {Map<ProductId, Product>} */
    this._products = new Map();
    /** @type {Map<ProductId, Product>} */
    this._selection = new Map();
    /** @type {Map<ProductId, ProductFilter>} */
    this._filters = new Map();
    /** @type {Set<ProductId>} */
    this._deletableProductIds = new Set();
    /** @type {Map<number, number>} */
    this._stackLevelHistogram = new Map();

    this.hub = new EventsHub();
    this.initialised = false;
  }

  /**
   * @public
   * @param {ProductConfig[]} productConfigs
   * @returns {void}
   */
  init(productConfigs) {
    this._products.clear();
    this._selection.clear();
    this._filters.clear();
    this._deletableProductIds.clear();
    this._stackLevelHistogram.clear();

    for (const productConfig of productConfigs) {
      this._addProduct(productConfig);
    }

    this.hub.emit('change:products');
    this.hub.emit('change:selection');
    this.initialised = true;
    this.hub.emit('init');
  }

  /**
   * @public
   * @returns {JSONValue | undefined}
   */
  toJSON() {
    return undefined;
  }
  /**
   * @public
   * @param {JSONValue} json
   * @returns {void}
   */
  fromJSON(json) { /* empty */ }

  /**
   * @public
   * @returns {State}
   */
  get state() {
    /** @type {Record<ProductId, ProductState>} */
    const products = {};
    for (const product of this.products()) {
      products[product.id] = {
        opacity: product.opacity,
        stackLevel: product.stackLevel / NUM_STACK_SUB_LEVEL,
        pattern: (
          this.isProductPatternActive(product.id)
          ? this.getProductPattern(product.id)
          : ''
        ),
        filters: this.getProductDatasetFilters(product.id),
      };
    }
    return {products, selection: [...this.selectedProductIds()]};
  }
  /**
   * @public
   * @param {Readonly<PartialState>} state
   * @returns {void}
   */
  set state(state) {
    const {products = {}, selection = []} = state;
    this.selectOnlyProducts(selection);
    for (
      const [productId, {opacity, stackLevel, pattern, filters}]
      of Object.entries(products)
    ) {
      if (isFinite(opacity)) {
        this.setProductOpacity(productId, opacity);
      }
      if (isFinite(stackLevel)) {
        this.setProductStackLevel(
          productId,
          Math.floor(round(stackLevel * NUM_STACK_SUB_LEVEL, -10))
        );
      }
      this.setProductPattern(productId, pattern ?? '');
      if (!pattern) {
        this.deactivateProductPattern(productId);
      }
      this.setProductDatasetFilters(productId, filters);
    }
  }

  /**
   * @public
   * @param {Readonly<ProductRawConfig>} productConfig
   * @returns {void}
   * @throws {TypeCheckerError} If `productConfig` is invalid
   */
  addProduct(productConfig) {
    assertValidProduct(productConfig);
    const normalizedProductConfig = normalizeProduct(productConfig);
    const product = this._addProduct(normalizedProductConfig);
    this._deletableProductIds.add(product.id);
    this.hub.emit('change:products', product.id);
    if (this.isProductSelected(product.id)) {
      this.hub.emit('change:selection', product.id);
    }
    this.hub.emit('change:state');
  }
  /**
   * @public
   * @returns {ReadonlyArray<ProductId>}
   */
  getProductIds() {
    return [...this._products.keys()];
  }
  /**
   * @public
   * @returns {ReadonlyArray<Product>}
   */
  getProducts() {
    return [...this._products.values()];
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {Product | undefined}
   */
  getProduct(productId) {
    return this._products.get(productId);
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {boolean}
   */
  hasProduct(productId) {
    return this._products.has(productId);
  }
  /**
   * @public
   * @template T
   * @param {(this: T, product: Product, id: ProductId) => void} callback
   * @param {T} [thisArg]
   * @returns {void}
   */
  forEachProduct(callback, thisArg) {
    this._products.forEach(callback, thisArg);
  }
  /**
   * @public
   * @returns {Iterable<ProductId>}
   */
  productIds() {
    return this._products.keys();
  }
  /**
   * @public
   * @returns {Iterable<Product>}
   */
  products() {
    return this._products.values();
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {boolean}
   */
  canDeleteProduct(productId) {
    return this._deletableProductIds.has(productId);
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {void}
   */
  deleteProduct(productId) {
    if (this.canDeleteProduct(productId)) {
      this.deselectProduct(productId);
      this._products.delete(productId);
      this._filters.delete(productId);
      this._deletableProductIds.delete(productId);
      this.hub.emit('change:products', productId);
    }
  }
  /**
   * @public
   * @returns {void}
   */
  clear() {
    if (this._selection.size > 0) {
      this._selection.clear();
      this.hub.emit('change:selection');
    }
    if (this._products.size > 0) {
      this._products.clear();
      this._filters.clear();
      this._stackLevelHistogram.clear();
      this.hub.emit('change:products');
    }
    this.hub.emit('change:state');
  }

  /**
   * @public
   * @returns {number}
   */
  getSelectedProductsCount() {
    return this._selection.size;
  }
  /**
   * @public
   * @returns {ProductId[]}
   */
  getSelectedProductIds() {
    return [...this._selection.keys()];
  }
  /**
   * @public
   * @returns {Product[]}
   */
  getSelectedProducts() {
    return [...this._selection.values()];
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {boolean}
   */
  isProductSelected(productId) {
    return this._selection.has(productId);
  }
  /**
   * @public
   * @template T
   * @param {(this: T, product: Product, id: ProductId) => void} callback
   * @param {T} [thisArg]
   * @returns {void}
   */
  forEachSelectedProduct(callback, thisArg) {
    this._selection.forEach(callback, thisArg);
  }
  /**
   * @public
   * @returns {Iterable<ProductId>}
   */
  selectedProductIds() {
    return this._selection.keys();
  }
  /**
   * @public
   * @returns {Iterable<Product>}
   */
  selectedProducts() {
    return this._selection.values();
  }

  /**
   * @public
   * @param {Readonly<Product> | ProductId} productOrId
   * @returns {void}
   */
  selectProduct(productOrId) {
    const productId = idFromProductOrId(productOrId);
    if (this._selectProduct(productId)) {
      this.hub.emit('change:selection', productId);
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {ReadonlyArray<Product | ProductId>} productsOrIds
   * @returns {void}
   */
  selectProducts(productsOrIds) {
    var changed = false;
    for (const productOrId of productsOrIds) {
      if (this._selectProduct(idFromProductOrId(productOrId))) {
        changed = true;
      }
    }
    if (changed) {
      this.hub.emit('change:selection');
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {ReadonlyArray<Product | ProductId>} productsOrIds
   * @returns {void}
   */
  selectOnlyProducts(productsOrIds) {
    const productIdsToSelect = new Set();
    let changed = false;
    for (const productOrId of productsOrIds) {
      const productId = idFromProductOrId(productOrId);
      productIdsToSelect.add(productId);
      if (this._selectProduct(productId)) {
        changed = true;
      }
    }
    for (const productId of this.selectedProductIds()) {
      if (!productIdsToSelect.has(productId)) {
        this._selection.delete(productId);
        changed = true;
      }
    }
    if (changed) {
      this.hub.emit('change:selection');
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {Readonly<Product> | ProductId} productOrId
   * @returns {void}
   */
  deselectProduct(productOrId) {
    const productId = idFromProductOrId(productOrId);
    if (this.isProductSelected(productId)) {
      this._selection.delete(productId);
      this.hub.emit('change:selection', productId);
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {ReadonlyArray<Product | ProductId>} productsOrIds
   * @returns {void}
   */
  deselectProducts(productsOrIds) {
    let changed = false;
    for (const productOrId of productsOrIds) {
      const productId = idFromProductOrId(productOrId);
      if (this.isProductSelected(productId)) {
        this._selection.delete(productId);
        changed = true;
      }
    }
    if (changed) {
      this.hub.emit('change:selection');
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @returns {void}
   */
  deselectAllProducts() {
    if (this._selection.size === 0) {
      return;
    }
    this._selection.clear();
    this.hub.emit('change:selection');
    this.hub.emit('change:state');
  }

  /**
   * @protected
   * @param {ProductId} productId
   * @returns {ProductFilter | undefined}
   */
  _getProductFilter(productId) {
    let filter = this._filters.get(productId);
    if (!filter) {
      const product = this.getProduct(productId);
      if (product) {
        /** @type {Set<string>} */
        const activeFilters = new Set();
        for (const {filters} of product.datasetFilters) {
          for (const {id} of filters) {
            activeFilters.add(id);
          }
        }
        filter = {
          activeFilters,
          customActive: false,
          customPattern: '',
          computedPattern: undefined,
        };
        this._filters.set(productId, filter);
      }
    }
    return filter;
  }
  /**
   * @protected
   * @param {ProductId} productId
   * @returns {HybridPattern}
   */
  _computeProductPattern(productId) {
    const filter = this._getProductFilter(productId);
    if (!filter) {
      return '';
    }

    const {activeFilters, customActive, customPattern} = filter;
    /** @type {Pattern.AST[]} */
    const patterns = [];
    for (const group of $(this.getProduct(productId)).datasetFilters) {
      let activeGroupPatterns = [];
      for (const {id, pattern} of group.filters) {
        if (activeFilters.has(id)) {
          activeGroupPatterns.push(Pattern.parse(pattern));
        }
      }

      if (activeGroupPatterns.length === 0) {
        patterns.push(emptyPatternGroupMarker(group));
      } else if (activeGroupPatterns.length < group.filters.length) {
        patterns.push(Pattern.or(activeGroupPatterns));
      }
    }
    if (customActive && customPattern) {
      if (patterns.length === 0) {
        return customPattern;
      }

      patterns.push(Pattern.parse(customPattern, true));
    }
    if (patterns.length === 0) {
      return '';
    }
    return Pattern.AST.stringify(Pattern.and(patterns), 'auto');
  }

  /**
   * @public
   * @param {ProductId} productId
   * @returns {LegacyPattern}
   */
  getProductPattern(productId) {
    return this._filters.get(productId)?.customPattern ?? '';
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {boolean}
   */
  isProductPatternActive(productId) {
    return !!this._filters.get(productId)?.customActive;
  }
  /**
   * @public
   * @param {Iterable<ProductId>} productIds
   * @returns {Record<ProductId, LegacyPattern>}
   */
  getPatternsForProductIds(productIds) {
    /** @type {Record<ProductId, LegacyPattern>} */
    const patterns = {};
    for (const productId of productIds) {
      patterns[productId] = (
        this.isProductPatternActive(productId)
        ? this.getProductPattern(productId)
        : ''
      );
    }
    return patterns;
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {LegacyPattern} pattern
   * @returns {void}
   */
  setProductPattern(productId, pattern) {
    const filter = this._getProductFilter(productId);
    if (!filter) {
      return;
    }

    const before = this.getProductComputedPattern(productId);

    filter.customActive = true;
    filter.customPattern = pattern ?? '';
    filter.computedPattern = undefined;

    const after = this.getProductComputedPattern(productId);

    if (before !== after) {
      this.hub.emit('change:pattern');
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {void}
   */
  activateProductPattern(productId) {
    if (!this.isProductPatternActive(productId)) {
      const filter = this._getProductFilter(productId);
      if (filter) {
        filter.customActive = true;
        filter.computedPattern = undefined;
        this.hub.emit('change:pattern');
        this.hub.emit('change:state');
      }
    }
  }
  /**
   * @param {ProductId} productId
   * @returns {void}
   */
  deactivateProductPattern(productId) {
    const filter = this._filters.get(productId);
    if (filter?.customActive) {
      filter.customActive = false;
      filter.computedPattern = undefined;
      this.hub.emit('change:pattern');
      this.hub.emit('change:state');
    }
  }

  /**
   * @public
   * @param {ProductId} productId
   * @returns {Iterable<string>}
   */
  productActiveDatasetFilters(productId) {
    const filter = this._getProductFilter(productId);
    return filter ? filter.activeFilters.values() : [];
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {string} filter
   * @returns {boolean}
   */
  isProductDatasetFilterActive(productId, filter) {
    return !!this._filters.get(productId)?.activeFilters.has(filter);
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {ReadonlyArray<string>} filters
   * @returns {void}
   */
  setProductActiveDatasetFilters(productId, filters) {
    const filter = this._getProductFilter(productId);
    if (!filter) {
      return;
    }

    const before = this.getProductComputedPattern(productId);

    filter.activeFilters.clear();
    for (const id of filters) {
      filter.activeFilters.add(id);
    }
    filter.computedPattern = undefined;

    const after = this.getProductComputedPattern(productId);

    if (before !== after) {
      this.hub.emit('change:pattern');
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {string} filterId
   * @returns {void}
   */
  deactivateProductDatasetFilter(productId, filterId) {
    const filter = this._filters.get(productId);
    if (filter?.activeFilters.delete(filterId)) {
      filter.computedPattern = undefined;
      this.hub.emit('change:pattern');
      this.hub.emit('change:state');
    }
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {string} filterId
   * @returns {void}
   */
  activateProductDatasetFilter(productId, filterId) {
    const filter = this._getProductFilter(productId);
    if (!filter) {
      return;
    }

    if (!filter.activeFilters.has(filterId)) {
      filter.activeFilters.add(filterId);
      filter.computedPattern = undefined;
      this.hub.emit('change:pattern');
      this.hub.emit('change:state');
    }
  }

  /**
   * @public
   * @param {ProductId} productId
   * @returns {Record<string, boolean>}
   */
  getProductDatasetFilters(productId) {
    /** @type {Record<string, boolean>} */
    const patterns = {};

    const filter = this._getProductFilter(productId);
    if (!filter) {
      return patterns;
    }

    const {activeFilters} = filter;

    for (const group of $(this.getProduct(productId)).datasetFilters) {
      for (let {id} of group.filters) {
        patterns[id] = activeFilters.has(id);
      }
    }

    return patterns;
  }
  /**
   * @public
   * @param {Iterable<ProductId>} productIds
   * @returns {Record<ProductId, Record<string, boolean>>}
   */
  getDatasetFiltersForProductIds(productIds) {
    /** @type {Record<ProductId, Record<string, boolean>>} */
    const filters = {};
    for (const productId of productIds) {
      filters[productId] = this.getProductDatasetFilters(productId);
    }
    return filters;
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {Record<string, boolean> | undefined} filters
   * @returns {void}
   */
  setProductDatasetFilters(productId, filters) {
    const filter = this._getProductFilter(productId);
    if (!filter) {
      return;
    }

    const before = this.getProductComputedPattern(productId);
    const {activeFilters} = filter;

    activeFilters.clear();
    for (const group of $(this.getProduct(productId)).datasetFilters) {
      for (let {id, aliases} of group.filters) {
        let active = true;
        if (filters?.[id] === false) {
          active = false;
        } else if (filters && aliases) {
          for (const alias of aliases) {
            if (filters[alias] === false) {
              active = false;
            }
            if (filters[alias] != null) {
              break;
            }
          }
        }
        if (active) {
          activeFilters.add(id);
        }
      }
    }
    filter.computedPattern = undefined;

    const after = this.getProductComputedPattern(productId);

    if (before !== after) {
      this.hub.emit('change:pattern');
      this.hub.emit('change:state');
    }
  }

  /**
   * @public
   * @param {ProductId} productId
   * @returns {HybridPattern}
   */
  getProductComputedPattern(productId) {
    const filter = this._getProductFilter(productId);
    if (!filter) {
      return '';
    }
    if (filter.computedPattern == null) {
      filter.computedPattern = this._computeProductPattern(productId);
    }
    return $(filter.computedPattern);
  }
  /**
   * @public
   * @param {Iterable<ProductId>} productIds
   * @returns {Record<ProductId, HybridPattern>}
   */
  getComputedPatternsForProductIds(productIds) {
    /** @type {Record<ProductId, HybridPattern>} */
    const patterns = {};
    for (const productId of productIds) {
      patterns[productId] = this.getProductComputedPattern(productId);
    }
    return patterns;
  }

  /**
   * @public
   * @param {ProductId} productId
   * @returns {boolean}
   */
  canSetProductOpacity(productId) {
    const product = this.getProduct(productId);
    if (product) {
      return !!Dataset.CLASS_FOR_TYPE[product.type]?.prototype.canSetOpacity;
    }
    return false;
  }
  /**
   * @public
   * @param {ProductId} productId
   * @returns {boolean}
   */
  canSetProductStackLevel(productId) {
    const product = this.getProduct(productId);
    if (product) {
      return !!Dataset.CLASS_FOR_TYPE[product.type]?.prototype.canSetStackLevel;
    }
    return false;
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {number} opacity
   * @returns {void}
   */
  setProductOpacity(productId, opacity) {
    if (!this.canSetProductOpacity(productId)) {
      return;
    }

    const product = $(this.getProduct(productId));

    opacity = clamp(opacity, 0, 1);
    opacity = floor(opacity, -5);
    if (product.opacity === opacity) {
      return;
    }

    product.opacity = opacity;
    this.hub.emit('change:opacity', productId);
    this.hub.emit('change:state');
  }
  /**
   * @public
   * @param {ProductId} productId
   * @param {number} stackLevel
   * @returns {void}
   */
  setProductStackLevel(productId, stackLevel) {
    if (!this.canSetProductStackLevel(productId)) {
      return;
    }

    const product = $(this.getProduct(productId));

    if (product.stackLevel === stackLevel) {
      return;
    }

    product.stackLevel = stackLevel;
    this.hub.emit('change:stackLevel', productId);
    this.hub.emit('change:state');
  }
  /**
   * @public
   * @param {Record<ProductId, number>} stackLevelsByProduct
   * @returns {void}
   */
  setProductStackLevels(stackLevelsByProduct) {
    const allProductsSupportSettingStackLevel = (
      Object.keys(stackLevelsByProduct)
      .every(this.canSetProductStackLevel, this)
    );

    if (allProductsSupportSettingStackLevel) {
      for (const [productId, stackLevel] of Object.entries(stackLevelsByProduct)) {
        $(this.getProduct(productId)).stackLevel = stackLevel;
      }
      this.hub.emit('change:stackLevel');
      this.hub.emit('change:state');
    }
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {ProductConfig} productConfig
   * @returns {Product}
   */
  _addProduct(productConfig) {
    const label = normalizeHTML(productConfig.label);
    const {selected} = productConfig;
    const product = /** @type {Product} */ (Object.assign({}, productConfig, {
      /* General */
      label     : label.html,
      textLabel : label.text.trim(),
      searchTags: [...new Set(productConfig.tags.map(stripTagKey))],
      selected  : undefined,

      /* Timeline */
      timelineValidity: productConfig.timelineValidity * HOUR_IN_MILLISECONDS,
    }));

    this._products.set(product.id, product);

    if (this.canSetProductStackLevel(product.id)) {
      product.stackLevel = (
        Math.trunc(product.stackLevel) * NUM_STACK_SUB_LEVEL
      );
      this._stackLevelHistogram.set(
        product.stackLevel,
        (this._stackLevelHistogram.get(product.stackLevel) ?? 0) + 1
      );
      product.stackLevel += $(this._stackLevelHistogram.get(product.stackLevel));
    } else {
      product.stackLevel = 10000 * NUM_STACK_SUB_LEVEL;
    }

    if (selected) {
      this._selection.set(product.id, product);
    }
    return product;
  }

  /**
   * @protected
   * @param {ProductId} productId
   * @returns {boolean}
   */
  _selectProduct(productId) {
    if (this.hasProduct(productId) && !this.isProductSelected(productId)) {
      this._selection.set(productId, $(this.getProduct(productId)));
      return true;
    }
    return false;
  }

  // #endregion
}
