/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$, modulo, toFixed, h, wrap} from './utils.js';

/**
 * @typedef UnitDefinition
 * @prop {string} symbol
 * @prop {string} name
 * @prop {string} kind
 * @prop {number} precision
 * @prop {(value: number) => number} from
 * @prop {(value: number) => number} to
 * @prop {(value: number, toString: (value: number) => string) => string} format
 */

/**
 * @param {number} decimal
 * @returns {string}
 */
function decimalToSexagesimal(decimal) {
  if (decimal < 0) {
    decimal = -decimal;
  }

  const d = Math.trunc(decimal);
  const m = Math.trunc((decimal - d) * 60);
  const s = (decimal - d) * 3600 - m * 60;

  return `${d}°${m.toFixed().padStart(2)}'${s.toFixed().padStart(2)}"`;
}

/** @satisfies {Record<string, UnitPartialDefinition | string>} */
const PARTIAL_UNITS = {
  // #region Absolute Temperature

  K: {
    name: 'Kelvin',
    kind: 'temperature',
  },
  kelvin: 'K',
  '°C': {
    name: 'Celsius',
    kind: 'temperature',
    from: (celsius) => celsius + 273.15,
    to: (kelvin) => kelvin - 273.15,
  },
  celsius: '°C',
  '°F': {
    name: 'Fahrenheit',
    kind: 'temperature',
    from: (fahrenheit) => (fahrenheit - 32) * 5 / 9 + 273.15,
    to: (kelvin) => (kelvin - 273.15) * 9 / 5 + 32,
  },
  fahrenheit: '°F',

  // #endregion
  // #region Relative Temperature

  ΔK: {
    symbol: 'K',
    name: 'Kelvin',
    kind: 'delta-temperature',
  },
  'delta-kelvin': 'ΔK',
  'Δ°C': {
    symbol: '°C',
    name: 'Celsius',
    kind: 'delta-temperature',
  },
  'delta-celsius': 'Δ°C',
  'Δ°F': {
    symbol: '°F',
    name: 'Fahrenheit',
    kind: 'delta-temperature',
    scale: 5 / 9,
  },
  'delta-fahrenheit': 'Δ°F',

  // #endregion
  // #region Length

  m: {
    name: 'Meter',
    kind: 'length',
  },
  meter: 'm',
  mm: {
    name: 'Millimeter',
    kind: 'length',
    scale: 0.001,
  },
  millimeter: 'mm',
  cm: {
    name: 'Centimeter',
    kind: 'length',
    scale: 0.01,
  },
  centimeter: 'cm',
  dm: {
    name: 'Decimeter',
    kind: 'length',
    scale: 0.1,
  },
  decimeter: 'dm',
  km: {
    name: 'Kilometer',
    kind: 'length',
    scale: 1000,
  },
  kilometer: 'km',

  in: {
    name: 'Inch',
    kind: 'length',
    scale: 0.0254,
  },
  inch: 'in',
  ft: {
    name: 'Foot',
    kind: 'length',
    scale: 12 * 0.0254,
  },
  foot: 'ft',
  yd: {
    name: 'Yard',
    kind: 'length',
    scale: 36 * 0.0254,
  },
  yard: 'yd',
  mi: {
    name: 'Mile',
    kind: 'length',
    scale: 63360 * 0.0254,
  },
  mile: 'mi',
  nmi: {
    name: 'Nautical mile',
    kind: 'length',
    scale: 1852,
  },
  kmi: 'nmi',
  NM: 'nmi',

  // #endregion
  // #region Area

  'm²': {
    name: 'Square meter',
    kind: 'area',
  },
  m2: 'm²',
  'km²': {
    name: 'Square kilometer',
    kind: 'area',
    scale: 1e6,
  },
  km2: 'km²',
  'ft²': {
    name: 'Square foot',
    kind: 'area',
    scale: (39.3701 / 12) ** 2,
  },
  ft2: 'ft²',
  'NM²': {
    name: 'Square nautical mile',
    kind: 'area',
    scale: (39.3701 / 1852 * 39.3701) ** 2,
  },
  NM2: 'NM²',
  nmi2: 'NM²',
  'nmi²': 'NM²',

  // #endregion
  // #region Speed

  kn: {
    name: 'Knot',
    kind: 'speed',
    scale: 1852 / 3600,
  },
  kns: 'kn',
  kt: 'kn',
  kts: 'kn',
  knot: 'kn',
  knots: 'kn',
  'm/s': {
    name: 'Meter per second',
    kind: 'speed',
  },
  'm.s-1': 'm/s',
  'km/h': {
    name: 'Kilometer per hour',
    kind: 'speed',
    scale: 1000 / 3600,
  },
  mph: {
    name: 'Mile per hour',
    kind: 'speed',
    scale: 0.44704,
  },
  'mi/h': 'mph',

  // #endregion
  // #region Time

  s: {
    name: 'Second',
    kind: 'time',
  },
  mn: {
    name: 'Minute',
    kind: 'time',
    scale: 60,
  },
  hr: {
    name: 'Hour',
    kind: 'time',
    scale: 60 * 60,
  },
  d: {
    name: 'Day',
    kind: 'time',
    scale: 24 * 60 * 60,
  },
  'date-utc': {
    name: 'Date in UTC',
    kind: 'time',
    format: (value) => new Date(value).toISOString().slice(0, -5).replace('T', ' ') + ' UTC',
  },

  // #endregion
  // #region Angle

  '°': {
    name: 'Degree',
    kind: 'angle',
    from: (deg) => deg,
    to: (deg) => wrap(deg, 0, 360),
    format: (value, toString) => toString(value) + '°',
  },
  deg: '°',
  degree: '°',
  degrees: '°',

  lat: {
    extends: '°',
    name: 'Latitude in decimal degrees',
    precision: -2,
    format: (lat, toString) => toString(Math.abs(lat)) + '°' + (lat < 0 ? 'S' : 'N'),
  },
  'lat-dms': {
    extends: 'lat',
    name: 'Latitude in sexagesimal degrees (DMS notation)',
    format: (lat) => decimalToSexagesimal(lat) + (lat < 0 ? 'S' : 'N'),
  },
  lon: {
    extends: '°',
    name: 'Longitude in decimal degrees',
    precision: -2,
    format(lon, toString) {
      lon = wrap(lon, -180, 180);
      return toString(Math.abs(lon)) + '°' + (lon < 0 ? 'W' : 'E');
    },
  },
  'lon-dms': {
    extends: 'lon',
    name: 'Longitude in sexagesimal degrees (DMS notation)',
    format(lon) {
      lon = wrap(lon, -180, 180);
      return decimalToSexagesimal(lon) + (lon < 0 ? 'W' : 'E');
    },
  },

  // #endregion
  // #region Direction

  'dir-cartesian': {
    extends: '°',
    name: 'Degree counterclockwise from east',
    kind: 'direction',
    precision: 0,
  },
  'dir-meteorologic': {
    extends: '°',
    name: 'Degree clockwise from north (in meteorological convention)',
    kind: 'direction',
    precision: 0,
    from: (ne) => modulo(180 + 90 - ne, 360),
    to: (ccwe) => modulo(180 + 90 - ccwe, 360),
  },
  'dir-oceanographic': {
    extends: '°',
    name: 'Degree clockwise from north (in oceanographic convention)',
    kind: 'direction',
    precision: 0,
    from: (ne) => modulo(90 - ne, 360),
    to: (ccwe) => modulo(90 - ccwe, 360),
  },
  'dir-arrow': {
    extends: '°',
    name: 'An arrow',
    kind: 'direction',
    format: (ccwe) => h('i', {
      class: 'fa fa-long-arrow-right',
      'aria-hidden': 'true',
      style: `margin: 0 .2em;transform-origin: 50% 60%;transform: rotate(${-ccwe}deg)`,
    }),
  },

  // #endregion
};

/** @type {Record<string, UnitDefinition>} */
const UNITS = Object.create(null);

for (
  let [key, partial]
  of Object.entries(
    /** @type {Record<string, UnitPartialDefinition | string>} */
    (PARTIAL_UNITS)
  )
) {
  if (typeof partial !== 'string' && !partial.extends) {
    let {
      symbol = key,
      name = symbol,
      kind,
      precision = -3,
      from,
      to,
      scale = 1,
      prefix = '',
      suffix = '',
      format,
    } = partial;
    if (!from || !to) {
      from = (value) => value * scale;
      to = (value) => value / scale;
    }
    if (!format) {
      if (!prefix && !suffix) {
        suffix = ` ${symbol}`;
      }
      format = (value, toString) => `${prefix}${toString(value)}${suffix}`;
    }
    UNITS[key] = {
      symbol,
      name,
      kind: $(kind),
      precision,
      from,
      to,
      format,
    };
  }
}
for (
  let [key, partial]
  of Object.entries(
    /** @type {Record<string, UnitPartialDefinition | string>} */
    (PARTIAL_UNITS)
  )
) {
  const visitedUnits = new Set([key]);
  while (typeof partial === 'string') {
    if (visitedUnits.has(partial)) {
      throw new RangeError(
        `Circular definition of unit "${key}": ${[...visitedUnits, partial].join(' -> ')}`
      );
    }
    if (Object.hasOwn(UNITS, partial)) {
      UNITS[key] = $(UNITS[partial]);
      break;
    } else if (Object.hasOwn(PARTIAL_UNITS, partial)) {
      visitedUnits.add(partial);
      key = partial;
      partial = $(
        /** @type {Record<string, UnitPartialDefinition | string>} */
        (PARTIAL_UNITS)[partial]
      );
    } else {
      throw new RangeError(
        `Unit "${partial}" is not defined, but referenced by "${key}": ${[...visitedUnits, partial].join(' -> ')}`
      );
    }
  }
}
for (
  const [key, partial]
  of Object.entries(
    /** @type {Record<string, UnitPartialDefinition | string>} */
    (PARTIAL_UNITS)
  )
) {
  if (typeof partial !== 'string') {
    const {extends: parentName} = partial;
    if (parentName) {
      // @ts-expect-error -- extends is not optional
      delete partial.extends;

      const parent = UNITS[parentName];
      if (parent) {
        Object.setPrototypeOf(partial, parent);
        // @ts-expect-error -- TS doesn't know about the prototype of `partial`
        UNITS[key] = partial;
      } else {
        throw new RangeError(
          `Unit "${parentName}" is not defined, but referenced by "${key}"`
        );
      }
    }
  }
}

/** @typedef {'' | keyof PARTIAL_UNITS} Unit */
/**
 * @typedef FormatOptions
 * @prop {number} exp The precision of the return string as an exponent of 10.
 */

/**
 * Represents a possibly imprecise numeric value with an optional unit.
 * With support for unit conversion and formatting.
 *
 * @see https://en.wikipedia.org/wiki/Physical_quantity
 */
export class Quantity {
  // #region Private typedefs

  /**
   * @typedef UnitPartialBaseDefinition
   * @prop {UnitDefinition['symbol']} [symbol]
   * @prop {UnitDefinition['name']} [name]
   * @prop {UnitDefinition['precision']} [precision]
   */
  /**
   * @typedef UnitPartialExtendsDefinition
   * @prop {string} extends
   * @prop {UnitDefinition['kind']} [kind]
   */
  /**
   * @typedef UnitPartialNoExtendsDefinition
   * @prop {never} [extends]
   * @prop {UnitDefinition['kind']} kind
   */
  /**
   * @typedef UnitPartialScaleDefinition
   * @prop {number} [scale=1]
   * @prop {never} [from]
   * @prop {never} [to]
   */
  /**
   * @typedef UnitPartialNoScaleDefinition
   * @prop {never} [scale]
   * @prop {UnitDefinition['from']} from
   * @prop {UnitDefinition['to']} to
   */
  /**
   * @typedef UnitPartialFormatDefinition
   * @prop {UnitDefinition['format']} format
   * @prop {never} [prefix]
   * @prop {never} [suffix]
   */
  /**
   * @typedef UnitPartialNoFormatDefinition
   * @prop {never} [format]
   * @prop {string} [prefix]
   * @prop {string} [suffix]
   */
  /**
   * @typedef {(
   *   UnitPartialBaseDefinition
   *   & (UnitPartialExtendsDefinition | UnitPartialNoExtendsDefinition)
   *   & (UnitPartialScaleDefinition | UnitPartialNoScaleDefinition)
   *   & (UnitPartialFormatDefinition | UnitPartialNoFormatDefinition)
   * )} UnitPartialDefinition
   */

  // #endregion
  // #region Public

  /** @public */
  static units = UNITS;

  /**
   * @public
   * @template {Unit} F
   * @template {Unit} T
   * @param {InUnit<number, F>} value
   * @param {F | undefined} fromUnit
   * @param {T | undefined} toUnit
   * @returns {InUnit<number, T>}
   * @throws {TypeError} Cannot convert quantity
   */
  static convert(value, fromUnit, toUnit) {
    if (
      /** @type {Unit} */ (fromUnit) === /** @type {Unit} */ (toUnit)
      || fromUnit === undefined
      || toUnit === undefined
    ) {
      return /** @type {InUnit<number, T>} */ (value);
    }

    const fromDef = UNITS[fromUnit];
    const toDef = UNITS[toUnit];
    if (!fromDef || !toDef || fromDef.kind !== toDef.kind) {
      throw new TypeError(
        `Cannot convert value from "${fromUnit}" to "${toUnit}"`
      );
    }
    if (fromDef === toDef) {
      return /** @type {InUnit<number, T>} */ (value);
    }
    return toDef.to(fromDef.from(value));
  }

  /**
   * @public
   * @param {string | undefined} fromUnit
   * @param {string | undefined} toUnit
   * @returns {boolean}
   */
  static canConvert(fromUnit, toUnit) {
    if (fromUnit === toUnit || fromUnit === undefined || toUnit === undefined) {
      return true;
    }
    if (!fromUnit || !toUnit) {
      return false;
    }
    const fromKind = UNITS[fromUnit]?.kind;
    const toKind = UNITS[toUnit]?.kind;
    return !!fromKind && fromKind === toKind;
  }

  /**
   * @public
   * @template {Unit} U
   * @param {InUnit<number, U>} value
   * @param {U} [unit]
   * @param {FormatOptions} [options]
   * @returns {string}
   */
  static format(value, unit, options) {
    const u = unit && UNITS[unit];
    if (u) {
      return u.format(value, toFixed.bind(null, options?.exp ?? u.precision));
    } else {
      const v = (
        options
        ? toFixed(options.exp, value)
        : value
      );
      if (unit) {
        return v.toString() + ' ' + unit;
      } else {
        return v.toString();
      }
    }
  }

  /**
   * @public
   * @param {number | Quantity} value
   * @param {string} [unit]
   * @param {number} [min]
   * @param {number} [max]
   * @param {(value: number) => number} [transform]
   * @throws {TypeError} Cannot convert quantity
   */
  static from(value, unit, min, max, transform) {
    if (value instanceof Quantity) {
      const fromUnit = /** @type {Unit | undefined} */ (
        /** @type {unknown} */ (value.unit)
      );
      const toUnit = /** @type {Unit | undefined} */ (
        /** @type {unknown} */ (unit)
      );
      const sameUnit = fromUnit === unit;

      if (sameUnit && !transform && min === value.min && max === value.max) {
        return value;
      }

      if (min === undefined && value.min !== undefined) {
        min = (
          sameUnit
          ? value.min
          : Quantity.convert(value.min, fromUnit, toUnit)
        );
      }
      if (max === undefined && value.max !== undefined) {
        max = (
          sameUnit
          ? value.max
          : Quantity.convert(value.max, fromUnit, toUnit)
        );
      }
      value = (
        sameUnit
        ? value.value
        : Quantity.convert(value.value, fromUnit, toUnit)
      );
    }
    if (transform) {
      value = transform(value);
      min = min === undefined ? min : transform(min);
      max = max === undefined ? max : transform(max);
    }
    return new Quantity(value, unit, min, max);
  }

  /**
   * @public
   * @param {number} value The numerical value of the quantity
   * @param {string} [unit] The symbol of the unit of measurement of the quantity
   * @param {number} [min] For an imprecise value, the minimum possible value
   * @param {number} [max] For an imprecise value, the maximum possible value
   */
  constructor(value, unit, min, max) {
    this.value = value;
    this.unit = unit;
    this.min = min;
    this.max = max;
  }

  /**
   * @public
   * @returns {Quantity | undefined}
   */
  get minQuantity() {
    return (
      this.min === undefined
      ? undefined
      : new Quantity(this.min, this.unit)
    );
  }

  /**
   * @public
   * @returns {Quantity | undefined}
   */
  get maxQuantity() {
    return (
      this.max === undefined
      ? undefined
      : new Quantity(this.max, this.unit)
    );
  }

  /**
   * @public
   * @param {string} toUnit The symbol of the unit to check the convertibility to
   * @returns {boolean} `true` if the quantity can be converted to the requested
   * unit, `false` otherwise
   */
  canConvert(toUnit) {
    return Quantity.canConvert(this.unit, toUnit);
  }

  /**
   * @public
   * @param {string} toUnit The symbol of the unit to convert the quantity to
   * @returns {Quantity} A quantity with the requested unit representing the same value
   * @throws {TypeError} Cannot convert quantity
   */
  convert(toUnit) {
    return Quantity.from(this, toUnit);
  }

  /**
   * @public
   * @param {(value: number) => number} transform
   * @returns {Quantity}
   */
  transform(transform) {
    return new Quantity(
      transform(this.value),
      this.unit,
      (
        this.min === undefined
        ? undefined
        : transform(this.min)
      ),
      (
        this.max === undefined
        ? undefined
        : transform(this.max)
      )
    );
  }

  /**
   * @public
   * @param {'number' | 'string' | 'default' | string} hint
   * @returns {number | string}
   */
  [Symbol.toPrimitive](hint) {
    if (hint === 'string') {
      return this.toString();
    }
    return this.value;
  }

  /**
   * @public
   * @returns {number}
   */
  valueOf() {
    return this.value;
  }

  /**
   * @public
   * @param {string} [unit] The symbol of the unit to convert the quantity to
   * @returns {number} The value of the quantity in the requested unit
   * @throws {TypeError} Cannot convert value of the quantity to the requested unit
   */
  valueIn(unit) {
    const fromUnit = this.unit;
    if (!unit || !fromUnit || fromUnit === unit) {
      return this.value;
    }
    return Quantity.convert(
      this.value,
      /** @type {Unit} */ (this.unit),
      /** @type {Unit} */ (unit)
    );
  }

  /**
   * @public
   * @param {FormatOptions} [options]
   * @returns {string}
   */
  toString(options) {
    return Quantity.format(
      this.value,
      /** @type {Unit | undefined} */ (this.unit),
      options
    );
  }

  // #endregion
}
