/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$, asIterable, EventsHub, nameOfModifiedKey} from './utils.js';

/**
 * @typedef Shortcut
 * @prop {string} id
 * @prop {HIDKey[]} modifiers
 * @prop {OneOrIterable<HIDKey>} keys
 * @prop {boolean} [disabled = false]
 * @prop {Func} [handler]
 * @prop {unknown} [context]
 */

/**
 * @param {Shortcut} shortcut
 * @returns {Generator<string, void, unknown>}
 */
function* namesOfShortcut(shortcut) {
  const {modifiers, keys} = shortcut;
  for (const key of asIterable(keys)) {
    const name = nameOfModifiedKey(modifiers, key);
    if (name) {
      yield name;
    }
  }
}

export class ShortcutsStore {
  // #region Private typedefs

  /** @typedef {import('./utils.js').HIDKey} HIDKey */

  // #endregion
  // #region Public

  /**
   * @typedef Group
   * @prop {string} label
   * @prop {Shortcut[]} shortcuts
   */
  constructor() {
    /** @type {Map<string, string[]>} */
    this._shortcutIdsByKeyName = new Map();
    /** @type {Map<string, Group>} */
    this.groups = new Map();
    this.hub = new EventsHub();
  }

  toJSON() {
    /*empty*/
  }
  fromJSON() {
    /*empty*/
  }

  /**
   * @param {string} groupLabel
   * @param {Readonly<OneOrIterable<Shortcut | null | undefined>>} shortcuts
   * @returns {void}
   */
  add(groupLabel, shortcuts) {
    let group = this.groups.get(groupLabel);
    if (!group) {
      group = {label: groupLabel, shortcuts: []};
      this.groups.set(groupLabel, group);
    }

    let changed = false;
    for (const shortcut of asIterable(shortcuts)) {
      if (shortcut) {
        const existingShortcut = this.shortcutWithId(shortcut.id);
        if (existingShortcut) {
          console.warn(
            "ShortcutsStore: Can't add shortcut, a shortcut with the same " +
            `id (${existingShortcut.id}) already exists`,
            {existingShortcut: existingShortcut, newShortcut: shortcut}
          );
          continue;
        }
        group.shortcuts.push(shortcut);
        for (const name of namesOfShortcut(shortcut)) {
          const ids = this._shortcutIdsByKeyName.get(name);

          if (ids) {
            ids.push(shortcut.id);
          } else {
            this._shortcutIdsByKeyName.set(name, [shortcut.id]);
          }
        }
        changed = true;
      }
    }
    if (changed) {
      this.hub.emit('change');
    }
  }
  /**
   * @param {Iterable<Readonly<Group>>} groups
   * @returns {void}
   */
  addGroups(groups) {
    for (const group of groups) {
      this.add(group.label, group.shortcuts);
    }
  }

  /**
   * @param {string} name
   * @returns {string[] | undefined}
   */
  idsForKeyNamed(name) {
    return this._shortcutIdsByKeyName.get(name);
  }
  /**
   * @param {string} id
   * @returns {Shortcut | undefined}
   */
  shortcutWithId(id) {
    for (const {shortcuts} of this.groups.values()) {
      for (const shortcut of shortcuts) {
        if (shortcut.id === id) {
          return shortcut;
        }
      }
    }
    return undefined;
  }

  /**
   * @param {Iterable<string>} ids
   * @param {boolean} enabled
   * @returns {void}
   */
  setShortcutsWithIdsEnabled(ids, enabled) {
    const disabled = !enabled;
    let changed = false;
    const idsSet = new Set(ids);
    for (const {shortcuts} of this.groups.values()) {
      for (const shortcut of shortcuts) {
        if (idsSet.has(shortcut.id) && !shortcut.disabled === disabled) {
          shortcut.disabled = disabled;
          changed = true;
        }
      }
    }
    if (changed) {
      this.hub.emit('change');
    }
  }

  /**
   * @param {string} id
   * @returns {boolean}
   */
  removeShortcutWithId(id) {
    for (const {shortcuts} of this.groups.values()) {
      for (let i = 0; i < shortcuts.length; i++) {
        const shortcut = $(shortcuts[i]);
        if (shortcut.id === id) {
          for (const name of namesOfShortcut(shortcut)) {
            const ids = this._shortcutIdsByKeyName.get(name);
            const index = ids ? ids.indexOf(id) : -1;

            if (index !== -1) {
              $(ids).splice(index, 1);
            }
          }
          shortcuts.splice(i, 1);
          this.hub.emit('change');
          return true;
        }
      }
    }
    return false;
  }
  /**
   * @param {Iterable<string>} ids
   * @returns {void}
   */
  removeShortcutsWithIds(ids) {
    for (const id of ids) {
      this.removeShortcutWithId(id);
    }
  }

  // #endregion
}
