/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {API, setShortenerForURL} from './api.js';
import {assertValidConfig, normalizeConfig} from './config.js';
import {Coverage} from './coverage.js';
import {TimelineStore} from './timeline_store.js';
import {ProductsStore} from './products_store.js';
import {MapView} from './map_view.js';
import './map_view_datasets.js';
import {
  NotificationCenter,
  NotificationMessageType,
} from './notification_center.js';
import {
  $,
  emptyFunction,
  extendClass,
  assignWithOwnProperties,
  memoGetters,
  throttleOncePerEventLoop,
  isSettledPromiseRejected,
  getSettledPromiseReason,
  retriable,
  EventsHub,
  KeyboardManager,
  isTimeoutError,
  isOutOfMemoryError,
} from './utils.js';

const NO_DATA_AVAILABILITY_MESSAGE = (
  'Could not load data availability due to server or network errors!'
);
const PARTIAL_DATA_AVAILABILITY_MESSAGE = (
  'Data availability is incomplete due to server or network errors!'
);
const NO_DATASETS_MESSAGE = (
  'Could not find any datasets due to server or network errors!'
);
const PARTIAL_DATASETS_MESSAGE = (
  'Some datasets may be missing due to server or network errors!'
);

const NAME = Symbol('Syntool.NAME');
const CONSTRUCTOR = Symbol('Syntool.CONSTRUCTOR');
const INIT = Symbol('Syntool.INIT');
const STATE = Symbol('Syntool.STATE');
const FILTERS = Symbol('Syntool.FILTERS');

// #region Exported typedefs

/**
 * @typedef {{
 *   [NAME]?: string,
 *   [CONSTRUCTOR]?: (this: Syntool) => void,
 *   [INIT]?: (this: Syntool) => void,
 *   [STATE]?: Syntool.StateHook,
 *   [FILTERS]?: any, //FiltersMap,
 * }} Syntool.ExtensionExtras
 */
/** @typedef {ClassExtension<typeof Syntool> & Syntool.ExtensionExtras & ThisType<Syntool>} Syntool.Extension */
/**
 * @template TThis
 * @typedef {[(this: TThis) => void, TThis]} Syntool.UpdateFn
 */
/**
 * @typedef Syntool.StateHook
 * @prop {(syntool: Syntool) => Syntool.PartialState} get
 * @prop {(state: Syntool.PartialState, syntool: Syntool) => void} set
 */

// #endregion

export class Syntool {
  // #region Private typedefs

  /**
   * @template {Class} C
   * @typedef {import('./utils.js').ClassExtension<C>} ClassExtension
   */
  /**
   * @template {(...args: any[]) => any} F
   * @typedef {import('./utils.js').Retriable<F>} Retriable
   */
  /**
   * @template {(...args: any[]) => any} F
   * @typedef {import('./utils.js').ThrottledFunction<F>} ThrottledFunction
   */
  /** @typedef {import('./utils.js').RetriableExtraArgument} RetriableExtraArgument */
  /** @typedef {import('./api.js').API.DatasetInfo} DatasetInfo */
  /** @typedef {import('./coverage.js').CoverageMap} CoverageMap */

  // #endregion
  // #region Public

  /**
   * @public
   * @returns {typeof NAME}
   */
  static get NAME() {
    return NAME;
  }
  /**
   * @public
   * @returns {typeof CONSTRUCTOR}
   */
  static get CONSTRUCTOR() {
    return CONSTRUCTOR;
  }
  /**
   * @public
   * @returns {typeof INIT}
   */
  static get INIT() {
    return INIT;
  }
  /**
   * @public
   * @returns {typeof STATE}
   */
  static get STATE() {
    return STATE;
  }
  /**
   * @public
   * @returns {typeof FILTERS}
   */
  static get FILTERS() {
    return FILTERS;
  }

  /** @type {Syntool[]} */
  static instances = [];
  /**
   * @param {Syntool.Extension} extension
   */
  static extend(extension) {
    const {
      [NAME]: name,
      [CONSTRUCTOR]: constructor,
      [INIT]: init,
      [STATE]: stateHook,
      [FILTERS]: filters,
    } = extension;

    delete extension[NAME];
    delete extension[CONSTRUCTOR];
    delete extension[INIT];
    delete extension[STATE];
    delete extension[FILTERS];

    extension.meta = Object.assign(extension.meta || {}, {name});
    if (constructor || stateHook || filters) {
      /** @type {Syntool} */ (extension)._constructor = function() {
        window.original();
        constructor?.call(this);
        if (filters) {
          Object.assign(this.filters, filters);
        }
        if (stateHook) {
          this._stateHooks.push(stateHook);
        }
      };
    }
    if (init) {
      /** @type {Syntool} */ (extension)._init = function() {
        window.original();
        init.call(this);
      };
    }

    extendClass(
      this,
      (
        /** @type {import('./utils.js').ClassExtensionWithInfer<typeof this>} */
        (extension)
      )
    );
    for (const instance of this.instances) {
      constructor?.call(instance);
      if (filters) {
        Object.assign(instance.filters, filters);
      }
      if (stateHook) {
        instance._stateHooks.push(stateHook);
      }
      if (init && instance.initialised) {
        init.call(instance);
      }
    }
  }

  /**
   * @public
   * @param {Element} node
   */
  constructor(node) {
    node.classList.add('Syntool');

    const uiOverlayNode = document.createElement('div');
    uiOverlayNode.className = 'UIOverlay';
    node.insertBefore(uiOverlayNode, node.firstElementChild);

    const usableViewportNode = document.createElement('div');
    usableViewportNode.className = 'UsableViewport';
    uiOverlayNode.appendChild(usableViewportNode);

    this.node = node;
    this.uiOverlayNode = uiOverlayNode;
    this.usableViewportNode = usableViewportNode;
    this.hub = new EventsHub();
    this.initialised = false;
    this.api = new API(this);
    this.timelineStore = new TimelineStore();
    this.productsStore = new ProductsStore();
    this.keyboardManager = new KeyboardManager();
    this.notificationCenter = new NotificationCenter();

    const getFilter = (/** @type {string} */ name) => {
      const fn = /** @type {any} */ (this)[name];
      if (typeof fn !== 'function') {
        throw new TypeError(
          `Filter expected name of Syntool method but given "${name}"`
        );
      }

      const bound = fn.bind(this);
      const fnWithId = /** @type {any} */ (this)[`${name}WithId`];
      if (typeof fnWithId === 'function') {
        bound.withId = fnWithId.bind(this);
      }
      return bound;
    };

    /** @type {Syntool.Filters} */
    this.filters = new Proxy(/** @type {any} */ ({}), {
      set(target, prop, value, receiver) {
        const currentValue = Reflect.get(target, prop, receiver);
        if (currentValue && typeof currentValue !== typeof value) {
          throw new TypeError(
            'Cannot redefine filters property with different type. Expected ' +
            `type "${typeof currentValue}" but found type "${typeof value}".`
          );
        }
        if (typeof value === 'string') {
          value = getFilter(value);
        } else {
          value = Object.assign(currentValue ?? new Proxy({}, this), value);
        }
        return Reflect.set(target, prop, value, receiver);
      },
    });

    /** @type {MapView | undefined} */
    this._mapView = undefined;
    /** @type {Config | null} */
    this._config = null;
    /** @type {Syntool.State | null} */
    this._state = null;
    /** @type {Syntool.StateHook[]} */
    this._stateHooks = [];
    this._geoFilterActive = true;
    this._colocActive = false;
    /** @type {NotificationCenter.Message | null} */
    this._differentZoomWarning = null;

    /** @type {Syntool.UpdateFn<unknown>[]} */
    this._afterUpdateFns = [];
    /** @type {string} */
    this._updateState = '';
    this._update = throttleOncePerEventLoop(this._update);
    this.update = throttleOncePerEventLoop(this.update);
    this.updateUsableViewportSize = throttleOncePerEventLoop(
      this.updateUsableViewportSize.bind(this)
    );

    if (window.ResizeObserver) {
      const resizeObserver = (
        new ResizeObserver(this.updateUsableViewportSize)
      );
      resizeObserver.observe(this.usableViewportNode);
    } else {
      window.addEventListener('resize', this.updateUsableViewportSize);
    }

    this._constructor();
    Syntool.instances.push(this);
  }

  /**
   * @public
   * @returns {MapView}
   * @throws {Error} If not initialized yet.
   */
  get mapView() {
    if (!this._mapView) {
      throw new Error('Not initialized');
    }
    return this._mapView;
  }

  /**
   * Calls `fn` when the app is initialized.
   *
   * If the app is already initialized, the `fn` is called immediately
   * (synchronously), otherwise a `init` event listener is added to the app.
   *
   * @public
   * @overload
   * @param {(this: void) => void} fn The function to be called when the app is initialized.
   * @returns {void}
   *
   * @template TThis
   * @overload
   * @param {(this: TThis) => void} fn The function to be called when the app is initialized.
   * @param {TThis} thisArg The `this` used when calling `fn`.
   * @returns {void}
   *
   * @param {() => void} fn
   * @param {unknown} [thisArg]
   * @returns {void}
   */
  whenInitialised(fn, thisArg) {
    if (this.initialised) {
      fn.call(thisArg);
    } else {
      this.hub.once('init', fn.bind(thisArg));
    }
  }

  /**
   * @public
   * @param {PromiseOrValue<Config>} asyncConfig
   * @returns {Promise<void>}
   * @throws {TypeCheckerError} If config is invalid.
   * @emits beforeconfig
   * @emits config
   */
  async loadConfig(asyncConfig) {
    this.hub.emit('beforeconfig');

    const config = await asyncConfig;
    assertValidConfig(config, this.api.servers);

    console.log('Config is valid');
    this._config = normalizeConfig(config);

    this.hub.emit('config', this._config);
  }
  /**
   * @public
   * @returns {Config}
   * @throws {Error} If config not loaded yet.
   */
  get config() {
    if (this._config) {
      return this._config;
    }
    throw new Error('Config not loaded');
  }

  /**
   * @public
   * @returns {void}
   * @emits beforeinit
   * @emits init
   */
  init() {
    this._beforeInit();
    this.hub.emit('beforeinit');
    this._init();
    this.initialised = true;
    this.hub.emit('init');
    this._afterInit();
  }

  /**
   * @public
   * @returns {Syntool.State}
   */
  get state() {
    if (!this._state) {
      this._state = this._computeState();
    }
    return this._state;
  }
  /**
   * @public
   * @param {Syntool.PartialState} state
   * @returns {void}
   */
  set state(state) {
    this._applyState(state);
  }

  /**
   * @public
   * @returns {boolean}
   */
  get geoFilterActive() {
    return this._geoFilterActive;
  }
  /**
   * @public
   * @param {boolean | null | undefined} active
   * @returns {void}
   * @emits change:geo-filter-active
   */
  set geoFilterActive(active) {
    if (this._geoFilterActive === !active) {
      this._geoFilterActive = !!active;
      this.hub.emit('change:geo-filter-active');
      this._invalidateState();
    }
  }

  /**
   * @public
   * @returns {boolean}
   */
  get colocActive() {
    return this._colocActive;
  }
  /**
   * @public
   * @param {boolean | null | undefined} active
   * @returns {void}
   * @emits change:coloc-active
   */
  set colocActive(active) {
    if (this._colocActive === !active) {
      this._colocActive = !!active;
      this.hub.emit('change:coloc-active');
      this._invalidateState();
    }
  }

  /**
   * @public
   * @param {AbortSignal} [signal]
   * @returns {Promise<boolean>}
   */
  updateCoverage = retriable(
    /**
     * @this {Syntool}
     * @param {RetriableExtraArgument} extra
     * @returns {Promise<boolean>}
     */
    async function({signal, throwIfRetriable}) {
      const selectedProducts = this.productsStore.getSelectedProductIds();
      const promises = this.api.getCoverages(selectedProducts, signal);

      this._handleCoverageLoadingStart();
      await Promise.all(promises.map(this._handleCoverageLoadingPromise, this));

      if (signal.aborted) {
        return false;
      }

      if (this.timelineStore.isCoveragePartial()) {
        if (this.timelineStore.getCoverage().isEmpty) {
          throwIfRetriable();

          this.notificationCenter.showError(NO_DATA_AVAILABILITY_MESSAGE);
        } else {
          this.notificationCenter.showWarning(PARTIAL_DATA_AVAILABILITY_MESSAGE);
        }
      }

      // Loading is done we mark coverage loading as finished
      this._handleCoverageLoadingFinish();
      return true;
    },
    {abortsPreviousCall: true, withExtraArgument: true}
  );

  /**
   * @public
   * @param {ProductId[]} [productIds = this.productsStore.getSelectedProductIds()]
   * @param {Date} [minDate = this.timelineStore.minDate]
   * @param {Date} [maxDate = this.timelineStore.maxDate]
   * @param {AbortSignal} [signal]
   * @returns {Promise<boolean>}
   */
  loadDatasetInfos = retriable(
    /**
     * @this {Syntool}
     * @param {RetriableExtraArgument} extra
     * @param {ProductId[]} [productIds]
     * @param {Date} [minDate]
     * @param {Date} [maxDate]
     * @returns {Promise<boolean>}
     */
    async function(
      {signal, throwIfRetriable},
      productIds,
      minDate,
      maxDate
    ) {
      if (maxDate instanceof AbortSignal) {
        maxDate = undefined;
      } else if (minDate instanceof AbortSignal) {
        minDate = undefined;
      } else if (productIds instanceof AbortSignal) {
        productIds = undefined;
      }

      productIds ??= this.productsStore.getSelectedProductIds();
      minDate ??= this.timelineStore.minDate;
      maxDate ??= this.timelineStore.maxDate;

      const promises = (
        this.api.getDatasetInfos(minDate, maxDate, productIds, signal)
      );
      promises.forEach(this._processDatasetInfosPromise, this);

      const results = await Promise.allSettled(promises);

      if (signal.aborted) {
        return false;
      }

      const someRejected = results.some(isSettledPromiseRejected);
      if (someRejected) {
        const reasons = results.map(getSettledPromiseReason).filter(Boolean);
        if (!reasons.every(isTimeoutError)) {
          throwIfRetriable();
        }

        console.error(
          "Error loading datasets' info!",
          reasons.length === 1 ? reasons[0] : reasons
        );
        if (this.mapView.getDatasetsInMapCount() === 0) {
          this.notificationCenter.showError(NO_DATASETS_MESSAGE);
        } else {
          this.notificationCenter.showWarning(PARTIAL_DATASETS_MESSAGE);
        }
      }
      return true;
    },
    {abortsPreviousCall: true, withExtraArgument: true}
  );

  /**
   * @public
   * @returns {void}
   * @emits update
   */
  update() {
    this.loadDatasetInfos().catch(emptyFunction);
    this.updateCoverage().catch(emptyFunction);
    this.mapView.updateDatasets();
    this.hub.emit('update');
    this._afterUpdate();
  }

  /**
   * @public
   * @overload
   * @param {(this: void) => void} fn
   * @returns {void}
   *
   * @template TThis
   * @overload
   * @param {(this: TThis) => void} fn
   * @param {TThis} thisArg The `this` used when calling `fn`.
   * @returns {void}
   *
   * @param {() => void} fn
   * @param {unknown} [thisArg]
   * @returns {void}
   */
  afterUpdate(fn, thisArg) {
    if (
      !this.initialised
      || /** @type {ThrottledFunction<() => void>} */ (this._update).scheduled
      || /** @type {ThrottledFunction<() => void>} */ (this.update).scheduled
    ) {
      this._afterUpdateFns.push([fn, thisArg]);
    } else {
      fn.call(thisArg);
    }
  }

  /**
   * @public
   * @returns {void}
   * @emits change:usable-viewport
   */
  updateUsableViewportSize() {
    this.hub.emit('change:usable-viewport');
  }

  /**
   * @public
   * @template {(...args: any[]) => any} F
   * @param {F} func
   * @returns {Retriable<F>}
   * @emits low-memory
   */
  retryIfLowMemory(func) {
    if (typeof func !== 'function') {
      throw new TypeError('A function argument is required.');
    }

    return retriable(func, {
      maxRetries: 3,
      initialDelayMs: 100,
      shouldRetryAfterError: (error, retries) => {
        if (isOutOfMemoryError(error)) {
          this.hub.emit('low-memory', retries);
          return true;
        }
        return false;
      },
    });
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @returns {void}
   */
  _constructor() {
    /*empty*/
  }

  /**
   * @protected
   * @returns {void}
   */
  _beforeInit() {
    const config = this.config;

    this.colocActive = config.colocActive;
    this.api.init();
    config.servers.forEach((serverConfig) => {
      this.api.servers.add(new serverConfig.Class(serverConfig));
    });
    setShortenerForURL(config.shortenerForURL || null);
    this.notificationCenter.loadMessagesFromConfig(config);
    this.timelineStore.minYear = config.yearsRange[0];
    this.timelineStore.maxYear = config.yearsRange[1];
    this.timelineStore.currentDate = new Date();
    this.productsStore.init(config.products);

    this._mapView = new MapView(this, {
      projection       : config.projection,
      maxBounds        : config.maxExtent,
      restrictedBounds : config.restrictedExtent,
      zoomOffset       : config.zoomOffset,
      baseLayers       : config.baseLayers,
      additionalLayers : config.additionalLayers,
      maxDatasetsWeight: config.maxDatasetsWeight,
    });
    this.node.insertBefore(this.mapView.node, this.node.firstElementChild);
    this.mapView.updateSize();
  }
  /**
   * @protected
   * @returns {void}
   */
  _init() {
    /*empty*/
  }
  /**
   * @protected
   * @returns {void}
   */
  _afterInit() {
    this.timelineStore.hub
      .on('change:currentDate', this._invalidateState, this)
      .on('change:timeframeLength', this._invalidateState, this);
    this.productsStore.hub.on('change:state', this._invalidateState, this);
    this.mapView.hub.on('change:state', this._invalidateState, this);

    this.timelineStore.hub.on('change:range', this._update, this);
    this.productsStore.hub
      .on('change:selection', this._update, this)
      .on('change:pattern', this._update, this)
      .on('change:products', (productId) => {
        if (!this.productsStore.hasProduct(productId)) {
          // product deleted
          this.mapView.removeProductWithId(productId);
        }
      });
    this.mapView.hub.on('change:extent', this._update, this);
    this.hub.on('change:geo-filter-active', this._update, this);

    this.update();
  }
  /**
   * @protected
   * @returns {void}
   */
  _afterUpdate() {
    const callbacks = this._afterUpdateFns;
    this._afterUpdateFns = [];
    for (const [fn, thisArg] of callbacks) {
      fn.call(thisArg);
    }
  }

  /**
   * @protected
   * @returns {void}
   */
  _update() {
    const newState = JSON.stringify([
      this.timelineStore.minYear,
      this.timelineStore.maxYear,
      this.timelineStore.currentTime,
      this.timelineStore.timeframeLength,
      this.geoFilterActive && this.mapView.getVisibleProjectedBounds(),
      this.productsStore.getComputedPatternsForProductIds(
        this.productsStore.selectedProductIds()
      ),
    ]);
    if (this._updateState === newState) {
      this._afterUpdate();
    } else {
      this.update();
      this._updateState = newState;
    }
  }

  /**
   * @protected
   * @param {Promise<Iterable<DatasetInfo>>} promise
   * @returns {Promise<void>}
   */
  async _processDatasetInfosPromise(promise) {
    try {
      this.mapView.updateDatasets(await promise);
    } catch (_) {
      /*ignore*/
    }
  }

  /**
   * @protected
   * @param {number} zoom
   * @returns {void}
   */
  _showDifferentZoomWarning(zoom) {
    if (!this._differentZoomWarning) {
      this._differentZoomWarning = /** @type {NotificationCenter.Message} */ ({
        type: NotificationMessageType.warning,
        text: (
          'In order to fit the requested geographical region, the zoom ' +
          'level was changed!'
        ),
        buttons: [
          {
            text  : 'restore zoom',
            closes: true,
            props : {title: 'Restore the original zoom level'},
          },
        ],
        hideAfter: -1,
      });
    }
    const restoreZoomButton = $($(this._differentZoomWarning.buttons)[0]);
    restoreZoomButton.onClick = () => {
      this.mapView.setZoom(zoom);
    };
    this.notificationCenter.showMessage(this._differentZoomWarning);
    this.mapView.hub.off('change:zoom', this._hideDifferentZoomWarning, this);
    this.mapView.hub.on('change:zoom', this._hideDifferentZoomWarning, this);
  }
  /**
   * @protected
   * @returns {void}
   */
  _hideDifferentZoomWarning() {
    if (this._differentZoomWarning) {
      this.notificationCenter.closeMessage(this._differentZoomWarning.id);
    }
    this.mapView.hub.off('change:zoom', this._hideDifferentZoomWarning, this);
  }

  /**
   * @protected
   * @returns {void}
   */
  _handleCoverageLoadingStart() {
    this.timelineStore.setCoverage(null, {loading: true, hasErrors: false});
  }
  /**
   * @protected
   * @param {CoverageMap | null} partialCoverageByProduct
   * @returns {void}
   */
  _handleCoverageLoadingProgress(partialCoverageByProduct) {
    if (!partialCoverageByProduct) {
      return;
    }

    const partialCoverage = Coverage.create(partialCoverageByProduct);
    const currentCoverage = this.timelineStore.getCoverage();
    const newCoverage = Coverage.merge(currentCoverage, partialCoverage);
    this.timelineStore.setCoverage(newCoverage);
  }
  /**
   * @protected
   * @param {Error} error
   * @returns {void}
   */
  _handleCoverageLoadingError(error) {
    console.error('Error loading part of the coverage!', error);
    this.timelineStore.setCoverage(
      this.timelineStore.getCoverage(),
      {hasErrors: true}
    );
  }
  /**
   * @protected
   * @returns {void}
   */
  _handleCoverageLoadingFinish() {
    this.timelineStore.setCoverage(
      this.timelineStore.getCoverage(),
      {loading: false}
    );
  }
  /**
   * @protected
   * @param {Promise<CoverageMap | null>} promise
   * @returns {Promise<void>}
   */
  _handleCoverageLoadingPromise(promise) {
    return promise.then(
      this._handleCoverageLoadingProgress.bind(this),
      this._handleCoverageLoadingError.bind(this)
    );
  }

  // #endregion
  // #region Protected - State

  /**
   * @protected
   * @returns {Syntool.State}
   */
  _computeState() {
    const {products, selection} = this.productsStore.state;

    return /** @type {Syntool.State} */ (memoGetters(
      assignWithOwnProperties(
        {
          products,
          selection,
          mapView: this.mapView.state,
          timeline: this.timelineStore.state,
          geoFilterActive: this.geoFilterActive,
          colocActive: this.colocActive,
        },
        ...this._stateHooks.map(({get}) => get.call(this, this))
      )
    ));
  }
  /**
   * @protected
   * @param {Syntool.PartialState} state
   * @returns {void}
   */
  _applyState(state) {
    const {
      products,
      selection,
      mapView,
      timeline,
      geoFilterActive,
      colocActive,
    } = state;
    this.productsStore.state = {products, selection};
    if (mapView) {
      this.mapView.state = mapView;
    }
    if (timeline) {
      this.timelineStore.state = timeline;
    }
    this.geoFilterActive = geoFilterActive;
    this.colocActive = colocActive;

    this._stateHooks.forEach(({set}) => set.call(this, state, this));

    const zoom = mapView?.zoom;
    if (isFinite(zoom) && this.mapView.getZoom() !== zoom) {
      this._showDifferentZoomWarning(zoom);
    } else {
      this._hideDifferentZoomWarning();
    }
  }
  _invalidateState() {
    this._state = null;
    this.hub.emit('change:state');
  }

  // #endregion
}
