/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {EventsHub} from './utils.js';
import {Syntool} from './syntool.js';

/** @typedef {typeof SyntoolControlsExtension} SyntoolControlsExtension */
/** @typedef {{ui: {[key: string]: any}}} SyntoolControlsExtension.State */
/**
 * @typedef SyntoolControlsExtension.InstanceFields
 * @prop {{_: Control[]} & {[key: string]: Control}} _controls
 */

const SyntoolControlsExtension = {
  [Syntool.NAME]: 'Controls',
  /** @this {Syntool} */
  [Syntool.CONSTRUCTOR]() {
    this._controls = Object.create(null);
    this._controls._ = [];
  },
  [Syntool.STATE]: {
    /**
     * @this {Syntool}
     * @returns {Syntool.PartialState}
     */
    get() {
      /** @type {Syntool.PartialState['ui']} */
      const ui = {};
      for (const [key, control] of Object.entries(this._controls)) {
        if ('state' in control) {
          ui[key] = control.state;
        }
      }
      return {ui};
    },
    /**
     * @this {Syntool}
     * @param {Syntool.PartialState} state
     */
    set({ui = {}}) {
      for (const [key, control] of Object.entries(this._controls)) {
        if ('state' in control) {
          control.state = ui[key] ?? {};
        }
      }
    },
  },

  /**
   * @public
   * @this {Syntool}
   * @param {Record<string, Control>} controls
   * @returns {void}
   */
  addControls(controls) {
    for (const [key, control] of Object.entries(controls)) {
      this.addControl(key, control);
    }
  },
  /**
   * @public
   * @this {Syntool}
   * @param {string} key
   * @param {Control} control
   * @returns {void}
   */
  addControl(key, control) {
    if (typeof key !== 'string' && typeof key !== 'symbol' && !control) {
      control = key;
      key = '_';
    }
    if (!control) {
      throw new TypeError(
        `Expected control to be defined and not null. Found "${control}"`
      );
    }
    if (key !== '_' && this._controls[key] && this._controls[key] !== control) {
      throw new TypeError(`Control with key "${key}" already exists.`);
    }
    if (control.syntool !== this) {
      control.syntool?.removeControl(control);
      if (key === '_') {
        this._controls._.push(control);
      } else {
        this._controls[key] = control;
        control.hub?.on('change:state', this._handleStateChange, this);
      }
      control.syntool = this;
    }
  },

  /**
   * @public
   * @this {Syntool}
   * @overload
   * @param {'_'} key
   * @returns {Control[]}
   *
   * @overload
   * @param {string | Control} keyOrControl
   * @returns {Control | undefined}
   *
   * @param {string | Control} keyOrControl
   * @returns {Control | Control[] | undefined}
   */
  removeControl(keyOrControl) {
    if (keyOrControl === '_') {
      const controls = this._controls[keyOrControl];
      for (const control of controls) {
        control.syntool = null;
      }
      this._controls[keyOrControl] = [];
      return controls;
    }

    let key;
    let control;
    if (typeof keyOrControl !== 'string' && typeof keyOrControl !== 'symbol') {
      key = null;
      control = keyOrControl;
      if (control.syntool !== this) {
        return control;
      }
    } else {
      key = keyOrControl;
      control = this._controls[key];
    }
    if (control) {
      control.syntool = null;
      if (key === null) {
        for (const [k, c] of Object.entries(this._controls)) {
          if (control === c) {
            control.hub?.off('change:state', this._handleStateChange, this);
            delete this._controls[k];
            break;
          }
        }
        for (let i = 0; i < this._controls._.length; i++) {
          if (control === this._controls._[i]) {
            this._controls._.splice(i, 1);
            break;
          }
        }
      } else {
        delete this._controls[key];
      }
    }
    return control;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {string} key
   * @returns {Control | undefined}
   */
  getControl(key) {
    return this._controls[key];
  },

  /**
   * @public
   * @this {Syntool}
   * @returns {typeof this._controls}
   */
  get controls() {
    return this._controls;
  },

  /**
   * @protected
   * @this {Syntool}
   * @returns {void}
   */
  _handleStateChange() {
    this._invalidateState();
  },
};
Syntool.extend(SyntoolControlsExtension);

/**
 * @typedef Control.Options
 * @prop {boolean} [autoActivate = true]
 */
/**
 * @typedef Control.State
 * @prop {boolean} active
 */
/** @typedef {Shortcut & ShortcutExtras} Control.Shortcut */
/**
 * @template {Control.State} [State = Control.State]
 */
export class Control {
  /** @typedef {import('./shortcuts_store.js').Shortcut} Shortcut */
  /**
   * @typedef ShortcutExtras
   * @prop {string} groupLabel
   * @prop {boolean} [autoDisable = false]
   */

  /**
   * @public
   * @param {Control.Options} [options]
   */
  constructor({autoActivate = true} = {}) {
    /** @type {Syntool | null} */
    this._syntool = null;
    this._autoActivate = autoActivate;
    this._active = false;
    /** @type {Iterable<Control.Shortcut> | null} */
    this.shortcuts = null;
    this.hub = new EventsHub();
  }

  /**
   * @public
   * @returns {Syntool | null}
   */
  get syntool() {
    return this._syntool;
  }
  /**
   * @public
   * @param {Syntool | null} syntool
   * @returns {void}
   */
  set syntool(syntool) {
    if (this._syntool !== syntool) {
      if (this._syntool) {
        if (this._autoActivate) {
          this.active = false;
        }
        this.onRemove();
        this.removeShortcuts();
      }
      this._syntool = syntool;
      if (this._syntool) {
        this.onAdd();
        this.addShortcuts();
        if (this._autoActivate) {
          this.active = true;
        }
      }
    }
  }

  /**
   * @public
   * @returns {boolean}
   */
  get active() {
    return this._active;
  }
  /**
   * @public
   * @param {unknown} active
   * @returns {void}
   */
  set active(active) {
    if (this._active === !active) {
      this._active = !!active;
      if (active) {
        this.onActivate();
      } else {
        this.onDeactivate();
      }
      this.updateShortcutsState();
      this.hub.emit('change:active');
      this.hub.emit('change:state');
    }
  }

  /**
   * @public
   * @returns {State}
   */
  get state() {
    return /** @type {State} */ ({active: this.active});
  }
  /**
   * @public
   * @param {Partial<State>} state
   * @returns {void}
   */
  set state({active}) {
    this.active = active;
  }

  /**
   * @public
   * @returns {void}
   */
  onAdd() {
    /*Implemented in subclass*/
  }
  /**
   * @public
   * @returns {void}
   */
  onRemove() {
    /*Implemented in subclass*/
  }
  /**
   * @public
   * @returns {void}
   */
  onActivate() {
    /*Implemented in subclass*/
  }
  /**
   * @public
   * @returns {void}
   */
  onDeactivate() {
    /*Implemented in subclass*/
  }

  /**
   * @public
   * @returns {void}
   */
  addShortcuts() {
    if (this.syntool && this.shortcuts) {
      const groups = new Map();
      for (const shortcut of this.shortcuts) {
        let group = groups.get(shortcut.groupLabel);
        if (!group) {
          group = {label: shortcut.groupLabel, shortcuts: []};
          groups.set(shortcut.groupLabel, group);
        }
        if (shortcut.autoDisable) {
          shortcut.disabled = !this.active;
        }
        group.shortcuts.push(shortcut);
      }
      this.syntool.shortcutsStore.addGroups(groups.values());
    }
  }
  /**
   * @public
   * @returns {void}
   */
  removeShortcuts() {
    if (this.syntool && this.shortcuts) {
      const ids = new Set();
      for (const shortcut of this.shortcuts) {
        ids.add(shortcut.id);
      }
      this.syntool.shortcutsStore.removeShortcutsWithIds(ids);
    }
  }
  /**
   * @public
   * @returns {void}
   */
  updateShortcutsState() {
    if (this.syntool && this.shortcuts) {
      const ids = new Set();
      for (const shortcut of this.shortcuts) {
        if (shortcut.autoDisable) {
          ids.add(shortcut.id);
        }
      }
      this.syntool.shortcutsStore.setShortcutsWithIdsEnabled(ids, this.active);
    }
  }
}
