/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Syntool} from './syntool.js';
import {
  emptyFunctionReturningFalse,
  inRangeInclusive,
  asArray,
} from './utils.js';

/** @typedef {typeof SyntoolFiltersExtension} SyntoolFiltersExtension */

const IGNORE_MUST_BE_CURRENT = {ignoreMustBeCurrent: true};

const SyntoolFiltersExtension = {
  /** @typedef {import('./gis.js').Bounds} Bounds */
  /** @typedef {import('./dataset.js').Dataset} Dataset */

  [Syntool.NAME]: 'Filters',
  [Syntool.FILTERS]: /** @type {const} */ ({
    product: {
      selected       : 'filterProductSelected',
      permanent      : 'filterProductPermanent',
      includedInColoc: 'filterProductIncludedInColoc',
      zoom           : 'filterProductZoom',
      visible        : 'filterProductVisible',

      nonPermanentVisible  : 'filterProductNonPermanentVisible',
      withUnreachableServer: 'filterProductWithUnreachableServer',
      withReachableServer  : 'filterProductWithReachableServer',

      ofType: 'filterProductOfType',
    },
    dataset: {
      date   : 'filterDatasetDate',
      geo    : 'filterDatasetGeo',
      zoom   : 'filterDatasetZoom',
      pattern: 'filterDatasetPattern',

      inTimeline : 'filterDatasetInTimeline',
      inTimeRange: 'filterDatasetInTimeRange',
      inMap      : 'filterDatasetInMap',

      ofType: 'filterDatasetOfType',
    },
  }),

  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductSelected(product) {
    return this.filterProductSelectedWithId(product.id);
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductSelectedWithId(productId) {
    return this.productsStore.isProductSelected(productId);
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductPermanent(product) {
    return product.permanent;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductPermanentWithId(productId) {
    const product = this.productsStore.getProduct(productId);
    return !!product?.permanent;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductIncludedInColoc(product) {
    return !product.excludedFromColoc;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductIncludedInColocWithId(productId) {
    const product = this.productsStore.getProduct(productId);
    return product ? !product.excludedFromColoc : false;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @param {number} [zoom]
   * @returns {boolean}
   */
  filterProductZoom(product, zoom) {
    return inRangeInclusive(
      zoom == null ? this.mapView.getZoom() : zoom,
      product.mapMinZoom,
      product.mapMaxZoom
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @param {number} [zoom]
   * @returns {boolean}
   */
  filterProductZoomWithId(productId, zoom) {
    const product = this.productsStore.getProduct(productId);
    return product ? this.filterProductZoom(product, zoom) : false;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductVisible(product) {
    return (
      product.outlines !== 'ALWAYS_HIDDEN' ||
      this.filterProductZoom(product)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductVisibleWithId(productId) {
    const product = this.productsStore.getProduct(productId);
    return product ? this.filterProductVisible(product) : false;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductNonPermanentVisible(product) {
    return (
      !this.filterProductPermanent(product) &&
      this.filterProductVisible(product)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductNonPermanentVisibleWithId(productId) {
    return (
      !this.filterProductPermanentWithId(productId) &&
      this.filterProductVisibleWithId(productId)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductWithUnreachableServer(product) {
    for (const serverId of this.api.getServerIdsForProduct(product)) {
      if (this.api.isServerWithIdUnreachable(serverId)) {
        return true;
      }
    }
    return false;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductWithUnreachableServerWithId(productId) {
    const product = this.productsStore.getProduct(productId);
    return product ? this.filterProductWithUnreachableServer(product) : false;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Product} product
   * @returns {boolean}
   */
  filterProductWithReachableServer(product) {
    for (const serverId of this.api.getServerIdsForProduct(product)) {
      if (!this.api.isServerWithIdUnreachable(serverId)) {
        return true;
      }
    }
    return false;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {ProductId} productId
   * @returns {boolean}
   */
  filterProductWithReachableServerWithId(productId) {
    const product = this.productsStore.getProduct(productId);
    return product ? this.filterProductWithReachableServer(product) : false;
  },
  /**
   * @public
   * @this {Syntool}
   * @overload
   * @param {string | string[] | null | undefined} types
   * @returns {(p: Product | null | undefined) => boolean}
   *
   * @overload
   * @param {string | string[] | null | undefined} types
   * @param {Product | null | undefined} [product]
   * @returns {boolean}
   *
   * @param {string | string[] | null | undefined} types
   * @param {Product | null} [product]
   * @returns {boolean | ((p: Product | null | undefined) => boolean)}
   */
  filterProductOfType(types, product) {
    if (arguments.length > 1) {
      if (!product || !types) {
        return false;
      }
      if (!Array.isArray(types)) {
        return product.type === types;
      }
      return types.indexOf(product.type) !== -1;
    }

    if (!types) {
      return emptyFunctionReturningFalse;
    }

    const typesSet = new Set(asArray(types));
    return (p) => (p ? typesSet.has(p.type) : false);
  },
  /**
   * @public
   * @this {Syntool}
   * @overload
   * @param {string | string[] | null | undefined} types
   * @returns {(pId: ProductId) => boolean}
   *
   * @overload
   * @param {string | string[] | null | undefined} types
   * @param {ProductId | null | undefined} [productId]
   * @returns {boolean}
   *
   * @param {string | string[] | null | undefined} types
   * @param {ProductId | null} [productId]
   * @returns {boolean | ((pId: ProductId) => boolean)}
   */
  filterProductOfTypeWithId(types, productId) {
    if (arguments.length > 1) {
      const product = productId && this.productsStore.getProduct(productId);
      return product ? this.filterProductOfType(types, product) : false;
    }
    const productIsOfType = this.filterProductOfType(types);
    return (pId) => {
      const product = this.productsStore.getProduct(pId);
      return product ? productIsOfType(product) : false;
    };
  },

  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {object} [options]
   * @param {boolean} [options.ignoreMustBeCurrent = false]
   * @returns {boolean}
   */
  filterDatasetDate(dataset, options) {
    const product = dataset.getProduct();
    if (!product) {
      return false;
    }

    const mustBeCurrent = product.mustBeCurrent;
    const currentTimestamp = this.timelineStore.currentTime;
    const startTimestamp = dataset.start.getTime();
    const endTimestamp = dataset.end.getTime();

    // check if the dataset is in the current time range
    if (mustBeCurrent && !options?.ignoreMustBeCurrent) {
      return (
        startTimestamp <= currentTimestamp && currentTimestamp < endTimestamp
      );
    }

    const timelineValidity = product.timelineValidity;
    let minTimestamp = this.timelineStore.minTime;
    let maxTimestamp = this.timelineStore.maxTime;

    if (timelineValidity > 0) {
      minTimestamp = currentTimestamp - timelineValidity / 2;
      maxTimestamp = currentTimestamp + timelineValidity / 2;
    }
    return (minTimestamp < endTimestamp && startTimestamp < maxTimestamp);
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {Projected<Bounds>} [projectedBounds]
   * @param {Projected<Bounds> | null} [projectedWorldBounds]
   * @returns {boolean}
   */
  filterDatasetGeo(dataset, projectedBounds, projectedWorldBounds) {
    return !!dataset.intersectsBounds(
      projectedBounds ?? this.mapView.getVisibleProjectedBounds(),
      projectedWorldBounds ?? this.mapView.getWorldProjectedBounds()
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {Projected<Bounds>} [projectedBounds]
   * @param {Projected<Bounds> | null} [projectedWorldBounds]
   * @returns {boolean}
   */
  filterDatasetOptionalGeo(dataset, projectedBounds, projectedWorldBounds) {
    return (
      !this.geoFilterActive ||
      this.filterDatasetGeo(dataset, projectedBounds, projectedWorldBounds)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {number} [zoom]
   * @returns {boolean}
   */
  filterDatasetZoom(dataset, zoom) {
    return this.filterProductZoomWithId(dataset.productId, zoom);
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @returns {boolean}
   */
  filterDatasetPattern(dataset) {
    return dataset.matchesPattern(
      this.productsStore.getProductComputedPattern(dataset.productId)
    );
  },

  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {Projected<Bounds>} [projectedBounds]
   * @param {Projected<Bounds> | null} [projectedWorldBounds]
   * @returns {boolean}
   */
  filterDatasetInTimeline(dataset, projectedBounds, projectedWorldBounds) {
    return (
      this.filterProductSelectedWithId(dataset.productId) &&
      this.filterDatasetOptionalGeo(dataset, projectedBounds, projectedWorldBounds) &&
      this.filterDatasetPattern(dataset)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {Projected<Bounds>} [projectedBounds]
   * @param {Projected<Bounds> | null} [projectedWorldBounds]
   * @returns {boolean}
   */
  filterDatasetInTimeRange(dataset, projectedBounds, projectedWorldBounds) {
    return (
      this.filterProductSelectedWithId(dataset.productId) &&
      this.filterDatasetDate(dataset, IGNORE_MUST_BE_CURRENT) &&
      this.filterDatasetOptionalGeo(dataset, projectedBounds, projectedWorldBounds) &&
      this.filterDatasetPattern(dataset)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @param {Dataset} dataset
   * @param {Projected<Bounds>} [projectedBounds]
   * @param {Projected<Bounds> | null} [projectedWorldBounds]
   * @returns {boolean}
   */
  filterDatasetInMap(dataset, projectedBounds, projectedWorldBounds) {
    return (
      this.filterProductSelectedWithId(dataset.productId) &&
      this.filterDatasetDate(dataset) &&
      this.filterDatasetOptionalGeo(dataset, projectedBounds, projectedWorldBounds) &&
      this.filterDatasetPattern(dataset)
    );
  },
  /**
   * @public
   * @this {Syntool}
   * @overload
   * @param {string | string[] | null | undefined} types
   * @returns {(d: Dataset) => boolean}
   *
   * @overload
   * @param {string | string[] | null | undefined} types
   * @param {Dataset | null | undefined} [dataset]
   * @returns {boolean}
   *
   * @param {string | string[] | null | undefined} types
   * @param {Dataset | null} [dataset]
   * @returns {boolean | ((d: Dataset) => boolean)}
   */
  filterDatasetOfType(types, dataset) {
    if (arguments.length > 1) {
      if (types && dataset) {
        return this.filterProductOfType(types, dataset.getProduct());
      }
      return false;
    }

    const productIsOfType = this.filterProductOfType(types);
    return (d) => productIsOfType(d.getProduct());
  },
};
Syntool.extend(SyntoolFiltersExtension);
