/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {ShortcutsStore} from './shortcuts_store.js';
import {NotificationMessageType} from './notification_center.js';
import {
  SECOND_IN_MILLISECONDS,
  $,
  HIDKeys,
  eventMatchesKey,
  modifierKeysFromEvent,
  keyFromEvent,
  namesOfModifiedKeysFromEvent,
  preventDefaultAndStopEvent,
  renderModifiedKeys,
  asIterable,
} from './utils.js';
import {Syntool} from './syntool.js';

/** @typedef {typeof SyntoolShortcutsExtension} SyntoolShortcutsExtension */
/**
 * @typedef SyntoolShortcutsExtension.State
 * @prop {boolean} shortcutsActive
 */
/**
 * @typedef SyntoolShortcutsExtension.InstanceFields
 * @prop {ShortcutsStore} shortcutsStore
 * @prop {Set<string>} _inputElementTypesAllowingShortcuts
 * @prop {boolean} _shortcutsActive
 */

const SyntoolShortcutsExtension = {
  [Syntool.NAME]: 'Shortcuts',
  /** @this {Syntool} */
  [Syntool.CONSTRUCTOR]() {
    this._inputElementTypesAllowingShortcuts = new Set([
      'button',
      'checkbox',
      'file',
      'reset',
      'submit',
    ]);
    this.shortcutsStore = new ShortcutsStore();
    this._shortcutsActive = false;
  },
  /** @this {Syntool} */
  [Syntool.INIT]() {
    this.activateShortcuts();
  },
  [Syntool.STATE]: {
    /**
     * @this {Syntool}
     * @returns {Syntool.PartialState}
     */
    get() {
      return {shortcutsActive: this.shortcutsActive};
    },
    /**
     * @this {Syntool}
     * @param {Syntool.PartialState} state
     * @returns {void}
     */
    set({shortcutsActive}) {
      // @ts-expect-error `shortcutsActive`'s setter accepts type `unknown`
      this.shortcutsActive = shortcutsActive;
    },
  },

  // #region Public

  /**
   * @public
   * @this {Syntool}
   * @returns {void}
   */
  activateShortcuts() {
    if (!this._shortcutsActive) {
      this._shortcutsActive = true;
      this.keyboardManager.hub.on('change:down', this._handleKeyDown, this);
      this._invalidateState();
    }
  },
  /**
   * @public
   * @this {Syntool}
   * @returns {void}
   */
  deactivateShortcuts() {
    if (this._shortcutsActive) {
      this.keyboardManager.hub.off('change:down', this._handleKeyDown, this);
      this._shortcutsActive = false;
      this._invalidateState();
    }
  },
  /**
   * @public
   * @this {Syntool}
   * @returns {boolean}
   */
  get shortcutsActive() {
    return this._shortcutsActive;
  },
  /**
   * @public
   * @this {Syntool}
   * @param {unknown} active
   * @returns {void}
   */
  set shortcutsActive(active) {
    if (active) {
      this.activateShortcuts();
    } else {
      this.deactivateShortcuts();
    }
  },
  /**
   * @protected
   * @this {Syntool}
   * @param {string} id
   * @param {KeyboardEvent | null} event
   * @param {unknown[]} args
   * @param {boolean} notify
   * @returns {boolean}
   */
  _executeShortcutWithIdFromEventWithArgs(id, event, args, notify) {
    const shortcut = this.shortcutsStore.shortcutWithId(id);
    const handler = shortcut && !shortcut.disabled && shortcut.handler;
    if (typeof handler === 'function') {
      const activated = handler.apply($(shortcut).context, args);
      if (notify && typeof activated === 'string') {
        this._notifyActivationOfShortcutWithMessage(shortcut, event, activated);
      }
      return activated;
    }
    return false;
  },
  /**
   * @protected
   * @this {Syntool}
   * @param {string} id
   * @param {KeyboardEvent | null} event
   * @param {boolean} [notify]
   * @param {unknown[]} args
   * @returns {boolean}
   */
  _executeShortcutWithIdFromEvent(id, event, notify = true, ...args) {
    return this._executeShortcutWithIdFromEventWithArgs(id, event, args, notify);
  },
  /**
   * @public
   * @this {Syntool}
   * @param {string} id
   * @param {boolean} [notify]
   * @param {unknown[]} args
   * @returns {boolean}
   */
  executeShortcutWithId(id, notify = true, ...args) {
    return this._executeShortcutWithIdFromEventWithArgs(id, null, args, notify);
  },

  /**
   * @public
   * @this {Syntool}
   * @param {string} id
   * @returns {HTML | undefined}
   */
  renderShortcutWithId(id) {
    const shortcut = this.shortcutsStore.shortcutWithId(id);
    if (shortcut) {
      return renderModifiedKeys(
        shortcut.modifiers,
        [...asIterable(shortcut.keys)]
      ).join(' ');
    }
    return undefined;
  },

  // #endregion
  // #region Protected

  /**
   * @protected
   * @this {Syntool}
   * @param {EventTarget | null} eventTarget
   * @returns {boolean}
   */
  _shortcutsCanUseEventWithTarget(eventTarget) {
    if (eventTarget instanceof HTMLTextAreaElement) {
      return false;
    }
    if (eventTarget instanceof HTMLInputElement) {
      return this._inputElementTypesAllowingShortcuts.has(eventTarget.type);
    }
    return true;
  },
  /**
   * @protected
   * @this {Syntool}
   * @param {KeyboardEvent} event
   * @returns {boolean}
   */
  _shortcutsCanUseEvent(event) {
    return (
      // if we have explicit permission
      /** @type {{syntoolShortcutsCanUse?: boolean}} */
      (event).syntoolShortcutsCanUse ||

      // if it involves the <Escape> key
      eventMatchesKey(event, HIDKeys.escape) ||

      // if shortcuts are allowed in event's target:
      this._shortcutsCanUseEventWithTarget(event.target)
    );
  },
  /**
   * @protected
   * @this {Syntool}
   * @param {unknown} shortcut
   * @param {KeyboardEvent | null} event
   * @param {string} message
   * @returns {void}
   */
  _notifyActivationOfShortcutWithMessage(shortcut, event, message) {
    this.notificationCenter.showMessage({
      modifiers: event && modifierKeysFromEvent(event),
      key      : event && keyFromEvent(event),
      text     : message,
      type     : NotificationMessageType.info,
      hideAfter: 2 * SECOND_IN_MILLISECONDS,
    });
  },

  /**
   * @protected
   * @this {Syntool}
   * @param {KeyboardEvent} event
   * @returns {void}
   */
  _handleKeyDown(event) {
    if (this._shortcutsCanUseEvent(event)) {
      const names = namesOfModifiedKeysFromEvent(event);
      for (const name of names) {
        const ids = this.shortcutsStore.idsForKeyNamed(name);
        if (ids) {
          for (const id of ids.slice().reverse()) {
            if (id && this._handleShortcut(id, event)) {
              return;
            }
          }
        }
      }
    }
  },
  /**
   * @protected
   * @this {Syntool}
   * @param {string} id
   * @param {KeyboardEvent} event
   * @returns {boolean}
   */
  _handleShortcut(id, event) {
    const activated = (
      this._executeShortcutWithIdFromEvent(id, event, true /*notify*/)
    );
    if (activated) {
      preventDefaultAndStopEvent(event);
      return true;
    }
    return false;
  },

  // #endregion
};
Syntool.extend(SyntoolShortcutsExtension);
