/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {COVERAGE_NULL} from './coverage.js';
import {timespanUnit} from './regexp.js';
import {
  DAY_IN_MILLISECONDS,
  WEEK_IN_MILLISECONDS,
  UNIT_LENGTHS,
  LENGTH_FOR_UNIT,
  UNIT_WITH_LENGTH,
  $,
  clamp,
  floor,
  RegExpHelpers,
  EventsHub,
} from './utils.js';

const rangeRe = new RegExp(`(${RegExpHelpers.ufloat})?(${timespanUnit})`, 'g');

/**
 * @param {Milliseconds} length
 * @returns {Timeframe}
 */
function getTimelineTimeframeFromLength(length) {
  let timeframe = '';
  for (let unitIndex = UNIT_LENGTHS.length - 1; unitIndex >= 0; unitIndex--) {
    const unitLength = $(UNIT_LENGTHS[unitIndex]);

    if (unitLength === WEEK_IN_MILLISECONDS) {
      continue;
    }

    const value = Math.floor(length / unitLength);
    if (value > 0) {
      timeframe += (
        unitLength === DAY_IN_MILLISECONDS && value % 7 === 0
        ? `${value / 7}w`
        : `${value}${UNIT_WITH_LENGTH[unitLength]}`
      );
      length -= value * unitLength;
    }
  }
  return timeframe;
}
/**
 * @param {Timeframe} timeframe
 * @returns {Milliseconds | null}
 */
function getTimelineLengthFromTimeframe(timeframe) {
  if (!timeframe) {
    return null;
  }

  let length = 0;
  let match;

  timeframe = $(timeframe.split(';', 1)[0]);
  rangeRe.lastIndex = 0;
  while ((match = rangeRe.exec(timeframe)) !== null) {
    length += (
      (match[1] ? parseFloat(match[1]) : 1)
      * LENGTH_FOR_UNIT[/** @type {keyof LENGTH_FOR_UNIT} */ (match[2])]
    );
  }
  return length;
}

/**
 * @param {Timeframe} timeframe
 * @returns {Timeframe | null}
 */
export function normalizeTimeframe(timeframe) {
  const length = getTimelineLengthFromTimeframe(timeframe);
  return length === null ? null : getTimelineTimeframeFromLength(length);
}

// #region Exported typedefs

/** @typedef {string} Timeframe */
/** @typedef {[number, keyof LENGTH_FOR_UNIT]} Interval */
/**
 * @typedef TimelineStore.CoverageState
 * @prop {boolean} loading
 * @prop {boolean} hasErrors
 */
/**
 * @typedef TimelineStore.State
 * @prop {number} minYear
 * @prop {number} maxYear
 * @prop {number} currentTime
 * @prop {Timeframe} timeframe
 */

// #endregion

export class TimelineStore {
  // #region Private typedefs

  /** @typedef {import('./coverage.js').Coverage} Coverage */
  /** @typedef {TimelineStore.CoverageState} CoverageState */
  /** @typedef {TimelineStore.State} State */

  // #endregion
  // #region Public

  constructor() {
    const now = new Date();

    this._minYear = now.getUTCFullYear();
    this._maxYear = now.getUTCFullYear() + 1;
    this._currentDate = now;
    this._timeframeLength = DAY_IN_MILLISECONDS;
    this._coverage = COVERAGE_NULL;
    /** @type {CoverageState} */
    this._coverageState = {
      loading  : false,
      hasErrors: false,
    };

    this.hub = new EventsHub();
  }

  toJSON() {
    /* empty */
  }
  fromJSON() {
    /* empty */
  }

  /**
   * @public
   * @returns {State}
   */
  get state() {
    const {minYear, maxYear, currentTime, timeframe} = this;
    return {minYear, maxYear, currentTime, timeframe};
  }
  /**
   * @public
   * @param {Partial<State>} state
   * @returns {void}
   */
  set state(state) {
    const {
      minYear = this.minYear,
      maxYear = this.maxYear,
      currentTime = this.currentTime,
      timeframe = this.timeframe,
    } = state;
    this.minYear = minYear;
    this.maxYear = maxYear;
    this.currentTime = currentTime;
    this.timeframe = timeframe;
  }

  /**
   * @returns {number}
   */
  get minYear() {
    return this._minYear;
  }
  /**
   * @param {number} year
   * @returns {void}
   */
  set minYear(year) {
    if (!isFinite(year) || this._minYear === year) {
      return;
    }

    this._minYear = year;
    this.hub.emit('change:minYear');
    this.hub.emit('change:state');
  }

  /**
   * @returns {number}
   */
  get maxYear() {
    return this._maxYear;
  }
  /**
   * @param {number} year
   * @returns {void}
   */
  set maxYear(year) {
    if (!isFinite(year) || this._maxYear === year) {
      return;
    }

    this._maxYear = year;
    this.hub.emit('change:maxYear');
    this.hub.emit('change:state');
  }

  /**
   * @returns {Date}
   */
  get absoluteMinDate() {
    return new Date(Date.UTC(this._minYear, 0, 1));
  }
  /**
   * @returns {Date}
   */
  get absoluteMaxDate() {
    return new Date(Date.UTC(this._maxYear, 0, 1));
  }

  /**
   * @returns {Milliseconds}
   */
  get currentTime() {
    return this._currentDate.getTime();
  }
  /**
   * @param {Milliseconds} time
   * @returns {void}
   */
  set currentTime(time) {
    time = clamp(
      time,
      Date.UTC(this._minYear, 0, 1),
      Date.UTC(this._maxYear, 0, 1) - 1
    );
    // Drop milliseconds
    time = floor(time, 3);

    if (isNaN(time) || this.currentTime === time) {
      return;
    }

    this._currentDate = new Date(time);
    this.hub.emit('change:currentDate');
    this.hub.emit('change:range');
    this.hub.emit('change:state');
  }

  /**
   * @returns {Date}
   */
  get currentDate() {
    return this._currentDate;
  }
  /**
   * @param {Date} date
   * @returns {void}
   */
  set currentDate(date) {
    if (date) {
      this.currentTime = date.getTime();
    }
  }

  /**
   * @returns {Milliseconds}
   */
  get timeframeLength() {
    return this._timeframeLength;
  }
  /**
   * @param {Milliseconds} timeframeLength
   * @returns {void}
   */
  set timeframeLength(timeframeLength) {
    if (
      isNaN(timeframeLength) ||
      this._timeframeLength === timeframeLength
    ) {
      return;
    }

    this._timeframeLength = timeframeLength;
    this.hub.emit('change:timeframeLength');
    this.hub.emit('change:range');
    this.hub.emit('change:state');
  }

  /**
   * @returns {Timeframe}
   */
  get timeframe() {
    return getTimelineTimeframeFromLength(this.timeframeLength);
  }
  /**
   * @param {Timeframe} timeframe
   * @returns {void}
   */
  set timeframe(timeframe) {
    const length = getTimelineLengthFromTimeframe(timeframe);
    if (length !== null) {
      this.timeframeLength = length;
    }
  }

  /**
   * @returns {Milliseconds}
   */
  get minTime() {
    return this.currentTime - this.timeframeLength / 2;
  }
  /**
   * @returns {Milliseconds}
   */
  get maxTime() {
    return this.currentTime + this.timeframeLength / 2;
  }
  /**
   * @returns {Date}
   */
  get minDate() {
    return new Date(this.minTime);
  }
  /**
   * @returns {Date}
   */
  get maxDate() {
    return new Date(this.maxTime);
  }

  /**
   * @returns {Coverage}
   */
  getCoverage() {
    return this._coverage;
  }
  /**
   * @returns {boolean}
   */
  isCoverageLoading() {
    return this._coverageState.loading;
  }
  /**
   * @returns {boolean}
   */
  isCoveragePartial() {
    return this._coverageState.hasErrors;
  }
  /**
   * @returns {boolean}
   */
  isCoverageComplete() {
    return !this._coverageState.loading && !this._coverageState.hasErrors;
  }
  /**
   * @param {Coverage | null} coverage
   * @param {Partial<CoverageState>} [partialState]
   * @returns {void}
   */
  setCoverage(coverage, partialState) {
    const state = /** @type {CoverageState} */ (
      Object.assign({}, this._coverageState, partialState)
    );

    const stateChanged = (
      this._coverageState.loading !== state.loading ||
      this._coverageState.hasErrors !== state.hasErrors
    );
    const coverageChanged = !this._coverage.equals(coverage);

    if (coverageChanged) {
      this._coverage = coverage || COVERAGE_NULL;
    }
    if (stateChanged) {
      this._coverageState = state;
    }
    if (stateChanged || coverageChanged) {
      this.hub.emit('change:coverage');
    }
  }

  // #endregion
}
