/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Pattern as _Pattern} from './pattern.js';
import {LayerType} from './gis.js';
import {timespan} from './regexp.js';
import {RegExpHelpers as RE, Types as is} from './utils.js';

/** @typedef {import('./utils.js').TypeCheckerPropName} TypeCheckerPropName */
/** @typedef {import('./utils.js').TypeCheckerGetProp} TypeCheckerGetProp */
/** @typedef {import('./utils.js').TypeCheckerResult} TypeCheckerResult */
/** @typedef {import('./utils.js').TypeChecker} TypeChecker */
/** @typedef {import('./utils.js').RequireableTypeChecker} RequireableTypeChecker */
/** @typedef {import('./utils.js').StringTypeChecker} StringTypeChecker */
/** @typedef {import('./utils.js').NumberTypeChecker} NumberTypeChecker */
/** @typedef {import('./utils.js').ExactTypeChecker} ExactTypeChecker */
/** @typedef {import('./utils.js').ExactWithIndexerTypeChecker} ExactWithIndexerTypeChecker */

export const Bounds = is.arrayOf(is.number.neq(NaN).isRequired, 4);
export const Point = is.arrayOf(is.number.neq(NaN).isRequired, 2);
export const Size = is.arrayOf(is.number.gt(0).isRequired, 2);
export const SVGPaint = is.string;
export const URL = is.string.match(new RegExp('^' + RE.url + '$', 'i'));
export const AnyURL = is.allOfType([is.string, validateAnyURL]);
export const WKT = is.allOfType([is.string, validateWKT]);

export const uniquePropInArray = validateUniquePropInArray;

export const Timespan = is.string.match(new RegExp('^' + timespan + '$'));
export const Layer = is.lazy(function() {
  return LayerType;
});

export const GranuleInfoMetadata = is.objectOf(is.string.isRequired);
export const GranuleInfoResource = is.shape({
  url        : AnyURL.isRequired,
  description: is.string,
}).withIndexer(is.string.isRequired);
export const GranuleInfoGroup = is.shape({
  metadata: GranuleInfoMetadata.isRequired,
}).withIndexer(GranuleInfoResource.isRequired);
export const GranuleInfo = is.objectOf(GranuleInfoGroup.isRequired);

export const Pattern = is.allOfType([is.string, validatePattern]);

const extraTypes = {
  Bounds,
  Point,
  Size,
  SVGPaint,
  URL,
  AnyURL,
  WKT,

  uniquePropInArray,

  Timespan,
  Layer,
  Pattern,

  GranuleInfoMetadata,
  GranuleInfoResource,
  GranuleInfoGroup,
  GranuleInfo,
};

/** @type {typeof is & typeof extraTypes} */
export const Types = Object.assign(is, extraTypes);

export {validateValueType, assertValueType, TypeCheckerError} from './utils.js';

/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateAnyURL(props, propName, propFullName) {
  const propValue = props[propName];
  if (window.URL.canParse(propValue, 'a:/')) {
    return null;
  }
  return [
    'Invalid `', propFullName, '` of value `', propValue, '`, ',
    'expected to be a URL https://en.wikipedia.org/wiki/URL.'
  ].join('');
}
/**
 * @param {string} value
 * @returns {boolean}
 */
function isWKT(value) {
  const REGEX_WKT_GEOMETRY = new RegExp(RE.wktGeometry, 'ig');
  const REGEX_WKT_GEOMETRYCOLLECTION = new RegExp(
    'GEOMETRYCOLLECTION' + RE.optionalWhiteSpace +
    RE.funcVaArgs('_'),
    'ig'
  );

  let simplifiedValue = value.replace(REGEX_WKT_GEOMETRY, '_');
  simplifiedValue = simplifiedValue.replace(/\s+/g, '');

  do {
    value = simplifiedValue;
    simplifiedValue = value.replace(REGEX_WKT_GEOMETRYCOLLECTION, '_');
  } while (simplifiedValue !== value);

  return simplifiedValue === '_';
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateWKT(props, propName, propFullName) {
  const propValue = props[propName];
  if (isWKT(propValue)) {
    return null;
  }
  return [
    'Invalid `', propFullName, '` of value `', propValue, '`, ',
    'expected to be a WKT https://en.wikipedia.org/wiki/Well-known_text.'
  ].join('');
}
/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validatePattern(props, propName, propFullName) {
  var propValue = props[propName];
  try {
    _Pattern.parse(propValue);
    return null;
  } catch (error) {
    return [
      'Invalid `', propFullName, '` of value `', propValue, '`, ',
      'expected to be a valid pattern. ', /** @type {Error} */ (error).message
    ].join('');
  }
}
/**
 * @param {string} propName
 * @returns {TypeChecker}
 */
function validateUniquePropInArray(propName) {
  /**
   * @param {ReadonlyArray<Record<string, unknown>>} props
   * @param {number} index
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, index, propFullName) {
    const propValue = props[index]?.[propName];

    if (!propValue) {
      return null;
    }

    const errors = [];
    for (let i = index + 1, l = props.length; i < l; i++) {
      if (props[i]?.[propName] === propValue) {
        const parentFullName = propFullName.replace(/(\.[^.[\]]+|\[\d+\])$/, '');

        errors.push([
          'Invalid `', propFullName, '.', propName, '` of value ',
          '`', propValue, '`, expected to be unique in ',
          '`', parentFullName, '`.'
        ].join(''));
      }
    }
    return errors;
  }

  return /** @type {TypeChecker} */ (validate);
}
