/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {emptyFunctionReturning} from './utils.js';

/** @type {UserStorage[]} */
const instances = [];

const noop = emptyFunctionReturning(Promise.resolve());

export class UserStorage {
  /**
   * @typedef Impl
   * @prop {() => Promise<void>} initialize
   * @prop {<R, T>(iterator: (value: T, key: string, iterationNumber: number) => R) => Promise<R | void>} forEach
   * @prop {<V>(key: string) => Promise<V | null>} get
   * @prop {<V>(key: string, value: V, persist: boolean) => Promise<V>} set
   * @prop {(key: string) => Promise<void>} delete
   */

  static instances = instances;

  /** @type {Impl} */
  static impl = {
    initialize: noop,
    forEach   : noop,
    get       : emptyFunctionReturning(Promise.resolve(null)),
    set       : (key, value) => Promise.resolve(value),
    delete    : noop,
  };

  /**
   * @template R
   * @param {<T>(value: T, key: string, iterationNumber: number) => R} iterator
   * @param {string} [description]
   * @returns {Promise<R | void>}
   */
  static async forEach(iterator, description) {
    try {
      await UserStorage.impl.initialize();
      return await UserStorage.impl.forEach(iterator);
    } catch (error) {
      if (!description) {
        description = 'iterate';
      }
      console.error("Couldn't " + description, error);
      throw error;
    }
  }
  /**
   * @template V
   * @param {string} key
   * @param {string} [description]
   * @returns {Promise<V | null>}
   */
  static async get(key, description) {
    try {
      await UserStorage.impl.initialize();
      return await UserStorage.impl.get(key);
    } catch (error) {
      if (!description) {
        description = 'get persistent key';
      }
      console.error("Couldn't " + description, key, error);
      throw error;
    }
  }
  /**
   * @template V
   * @overload
   * @param {string} key
   * @param {V} value
   * @param {string} [description]
   * @returns {Promise<V>}
   *
   * @overload
   * @param {string} key
   * @param {V} value
   * @param {boolean} [persist]
   * @param {string} [description]
   * @returns {Promise<V>}
   *
   * @param {string} key
   * @param {V} value
   * @param {boolean | string} [descriptionOrPersist]
   * @param {string} [description]
   * @returns {Promise<V>}
   */
  static async set(key, value, descriptionOrPersist, description) {
    /** @type {boolean} */
    let persist;
    if (arguments.length === 3 && typeof descriptionOrPersist === 'string') {
      description = descriptionOrPersist;
      persist = false;
    } else {
      persist = /** @type {boolean} */ (descriptionOrPersist);
    }
    try {
      await UserStorage.impl.initialize();
      return await UserStorage.impl.set(key, value, persist);
    } catch (error) {
      if (!description) {
        description = 'set persistent key';
      }
      console.error("Couldn't " + description, key, value, error);
      throw error;
    }
  }
  /**
   * @param {string} key
   * @param {string} [description]
   * @returns {Promise<void>}
   */
  static async delete(key, description) {
    try {
      await UserStorage.impl.initialize();
      return await UserStorage.impl.delete(key);
    } catch (error) {
      if (!description) {
        description = 'remove persistent key';
      }
      console.error("Couldn't " + description, key, error);
      throw error;
    }
  }

  /**
   * @param {string} id
   * @param {string} label
   * @param {boolean} [persist]
   */
  constructor(id, label, persist) {
    this.id = id;
    this.label = label;
    this.persist = !!persist;
    this.prefix = this.id + '_';
    instances.push(this);
  }

  /**
   * @template R
   * @param {<T>(value: T, key: string, iterationNumber: number) => R} iterator
   * @param {string} [description]
   * @returns {Promise<R | void>}
   */
  forEach(iterator, description) {
    const cls = /** @type {typeof UserStorage} */ (this.constructor);
    return cls.forEach((value, key, iterationNumber) => {
      if (key.indexOf(this.prefix) === 0) {
        return iterator(
          value,
          key.substring(this.prefix.length),
          iterationNumber
        );
      }
      return undefined;
    }, description);
  }
  /**
   * @template V
   * @param {string} key
   * @param {string} [description]
   * @returns {Promise<V | null>}
   */
  get(key, description) {
    const cls = /** @type {typeof UserStorage} */ (this.constructor);
    return cls.get(this.prefix + key, description);
  }
  /**
   * @template V
   * @param {string} key
   * @param {V} value
   * @param {string} [description]
   * @returns {Promise<V>}
   */
  set(key, value, description) {
    const cls = /** @type {typeof UserStorage} */ (this.constructor);
    return cls.set(
      this.prefix + key,
      value,
      this.persist,
      description
    );
  }
  /**
   * @param {string} key
   * @param {string} [description]
   * @returns {Promise<void>}
   */
  delete(key, description) {
    const cls = /** @type {typeof UserStorage} */ (this.constructor);
    return cls.delete(this.prefix + key, description);
  }
}
