/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {MINUTE_IN_MILLISECONDS} from './date_utils.js';
import {EventsHub} from './events_hub.js';
import {retriable} from './promise_utils.js';
import {$} from './typescript.js';
import {serializeFormData} from './url_utils.js';
import {
  upsert,
  forEachObject,
  isAbortError,
  isTimeoutError,
  fetchWithTimeout,
} from './utils.js';

/**
 * @typedef PNGMetadata
 * @prop {number} metadata
 * @prop {number} height
 * @prop {1 | 2 | 3 | 4} channels
 * @prop {number[]} offsets
 * @prop {number[]} scales
 * @prop {number} nodataValue
 */
/**
 * @typedef PNG
 * @prop {PNGMetadata} metadata
 * @prop {Uint8Array} pixels
 */

const decodePNGAsync = (function() {
  /**
   * @typedef {[resolve: (png: PNG) => void, reject: (error: string) => void]} DecodeWorkerResolver
   */
  const decodingWorkers = {
    /** @type {Worker[]} */
    workers  : [],
    /** @type {{[id: number]: DecodeWorkerResolver | null}} */
    promises : {},
    currentId: 0,
  };
  let workerCount = window.navigator.hardwareConcurrency || 8;

  while (workerCount--) {
    const worker = new Worker('js/png_decode_worker.js');
    worker.onmessage = handlePNGDecodeWorkerMessage;
    decodingWorkers.workers.push(worker);
  }

  /**
   * @param {ArrayBuffer} rawPNG
   * @returns {Promise<PNG>}
   */
  return function(rawPNG) {
    return new Promise(function(resolve, reject) {
      const currentId = decodingWorkers.currentId++;
      const workerIndex = currentId % decodingWorkers.workers.length;
      const worker = $(decodingWorkers.workers[workerIndex]);

      decodingWorkers.promises[currentId] = [resolve, reject];
      worker.postMessage([currentId, rawPNG], [rawPNG]);
    });
  };

  /**
   * @param {MessageEvent<[number, 0, PNG] | [number, 1, string]>} event
   */
  function handlePNGDecodeWorkerMessage(event) {
    decodingWorkers.promises[event.data[0]]?.[event.data[1]](
      // @ts-expect-error
      event.data[2]
    );
    decodingWorkers.promises[event.data[0]] = null;
  }
})();

ajax.timeout = MINUTE_IN_MILLISECONDS;
getPNG.timeout = 5 * MINUTE_IN_MILLISECONDS;

/**
 * @typedef FetchResourceCacheEntry
 * @prop {number} usersCount
 * @prop {Promise<any>} promise
 * @prop {() => void} abort
 */
/** @type {Map<string, FetchResourceCacheEntry>} */
fetchResource.cache = new Map();

const MAX_CONCURRENT_REQUESTS_PER_ORIGIN = 100;
const concurrentRequestCountPerOrigin = new Map();
const blockedRequestQueuePerOrigin = new Map();

const hub = new EventsHub();
const statistics = {
  issuedRequestCount   : 0,
  liveRequestCount     : 0,
  completedRequestCount: 0,
  erroredRequestCount  : 0,
  canceledRequestCount : 0,
};

ajax.get = get;
ajax.getArrayBuffer = getArrayBuffer;
ajax.getPNG = getPNG;
ajax.getJSON = getJSON;
ajax.getJSONP = getJSONP;
ajax.post = post;

/** @typedef {Error & {response: Response}} FetchResponseError */

/**
 * @param {Response} response
 * @returns {Response}
 * @throws {FetchResponseError}
 */
function fetchResponseCheckStatus(response) {
  if (response.ok) {
    return response;
  } else {
    /** @type {Partial<FetchResponseError>} */
    const error = new Error(response.status + ' (' + response.statusText + ')');
    error.response = response;
    throw error;
  }
}
/**
 * @param {Response} response
 * @returns {Promise<ArrayBuffer>}
 */
function fetchResponseArrayBuffer(response) {
  return response.arrayBuffer();
}
/**
 * @param {Response} response
 * @returns {Promise<Blob>}
 */
function fetchResponseBlob(response) {
  return response.blob();
}
/**
 * @param {Response} response
 * @returns {Promise<JSONValue>}
 */
function fetchResponseJSON(response) {
  return response.json();
}
/**
 * @param {Response} response
 * @returns {Promise<string>}
 */
function fetchResponseText(response) {
  return response.text();
}
/**
 * @param {Response} response
 * @returns {Response}
 */
function fetchResponseRaw(response) {
  return response;
}

/**
 * @param {string} result
 * @returns {Promise<JSONValue>}
 */
function parseJSONPResponse(result) {
  const startIndex = result.indexOf('(');
  const endIndex = result.lastIndexOf(')');
  const json = result.substring(startIndex + 1, endIndex);

  return JSON.parse(json);
}
/**
 * @param {ArrayBuffer} arrayBuffer
 * @returns {Promise<PNG>}
 */
function parsePNGResponse(arrayBuffer) {
  return decodePNGAsync(arrayBuffer);
}

/**
 * @param {Error} error
 * @returns {boolean}
 */
function shouldRetryRequest(error) {
  if (error instanceof TypeError) {
    /* Network error */
    // return error instanceof TypeError;
  } else if (
    !isAbortError(error) &&
    !isTimeoutError(error) &&
    /** @type {FetchResponseError} */ (error).response
  ) {
    /* HTTP error */
    const status = /** @type {FetchResponseError} */ (error).response.status;

    return (
      status === 408 || // Request Timeout
      status === 503 || // Service Unavailable
      status === 504 || // Gateway Timeout
      status === 522 || // Connection timed out
      status === 524    // A timeout occurred
    );
  }

  return false;
}

/**
 * @typedef FetchOptions
 * @prop {string} method
 * @prop {Record<string, any>} [data]
 * @prop {'omit' | 'same-origin' | 'include'} [credentials]
 * @prop {'arraybuffer' | 'blob' | 'json' | 'text' | 'raw'} responseType
 * @prop {AbortSignal} [signal]
 * @prop {number} timeout
 */

/**
 * @param {string} url
 * @param {FetchOptions} options
 * @returns {Promise<any>}
 */
function _fetch(url, options) {
  const {
    method,
    data,
    credentials,
    responseType,
    signal,
    timeout,
  } = options;
  /** @type {FormData | undefined} */
  let body;

  if (data) {
    if (method === 'GET' || method === 'HEAD') {
      const searchParams = serializeFormData(data);

      if (url.indexOf('?') === -1) {
        url += '?' + searchParams;
      } else {
        url += '&' + searchParams;
      }
    } else {
      body = new FormData();
      forEachObject(data, (value, key) => {
        $(body).append(key, value);
      });
    }
  }

  let fetchResponseParser;

  switch (responseType) {
  case 'arraybuffer':
    fetchResponseParser = fetchResponseArrayBuffer;
    break;
  case 'blob':
    fetchResponseParser = fetchResponseBlob;
    break;
  case 'json':
    fetchResponseParser = fetchResponseJSON;
    break;
  case 'raw':
    fetchResponseParser = fetchResponseRaw;
    break;
  default:
    fetchResponseParser = fetchResponseText;
  }

  return (
    fetchWithTimeout(url, {
      method,
      body,
      credentials,
      signal,
      timeout,
    })
    .then(fetchResponseCheckStatus)
    // @ts-expect-error
    .then(fetchResponseParser)
  );
}

/**
 * @typedef AjaxOptions
 * @prop {string} [method = 'GET']
 * @prop {FetchOptions['data']} [data]
 * @prop {FetchOptions['credentials']} [credentials]
 * @prop {FetchOptions['responseType']} [responseType = 'text']
 * @prop {AbortSignal} [signal]
 * @prop {number} [timeout = ajax.timeout]
 */

/**
 * @param {string} url
 * @param {AjaxOptions} options
 * @returns {Promise<any>}
 */
async function ajax(url, options) {
  const method = options.method?.toUpperCase() ?? 'GET';
  let timeout = options.timeout;

  if (timeout === null || !isFinite(timeout)) {
    timeout = ajax.timeout;
  }

  options = {
    method,
    data        : options.data,
    credentials : options.credentials,
    responseType: options.responseType || 'text',
    signal      : options.signal,
    timeout,
  };

  statistics.issuedRequestCount++;
  statistics.liveRequestCount++;
  hub.emit('pre-send');

  const origin = new URL(url, window.location.origin).origin;
  let inAirRequestCount = concurrentRequestCountPerOrigin.get(origin) ?? 0;
  if (inAirRequestCount >= MAX_CONCURRENT_REQUESTS_PER_ORIGIN) {
    const queue = upsert(blockedRequestQueuePerOrigin, origin, []);
    await new Promise((resolve) => queue.push(resolve));
    inAirRequestCount = concurrentRequestCountPerOrigin.get(origin);
  }
  concurrentRequestCountPerOrigin.set(origin, inAirRequestCount + 1);

  try {
    const response = await _fetch(url, /** @type {FetchOptions} */ (options));
    statistics.liveRequestCount--;
    statistics.completedRequestCount++;
    hub.emit('success');
    return response;
  } catch (/** @type {any} */ error) {
    statistics.liveRequestCount--;
    if (isAbortError(error)) {
      statistics.canceledRequestCount++;
      hub.emit('cancel');
    } else {
      statistics.erroredRequestCount++;
      hub.emit('fail');
    }
    throw error;
  } finally {
    concurrentRequestCountPerOrigin.set(
      origin,
      concurrentRequestCountPerOrigin.get(origin) - 1
    );
    blockedRequestQueuePerOrigin.get(origin)?.shift()?.();
  }
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {AjaxOptions}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {AjaxOptions}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout]
 * @param {AbortSignal} [signal]
 * @returns {AjaxOptions}
 *
 * @param {string} url
 * @returns {AjaxOptions}
 */
function _getOptionsFromArgs(url) {
  const argsCount = arguments.length;
  if (argsCount === 0) {
    throw new TypeError('Requires at least 1 argument "url", none were given!');
  }
  if (argsCount === 1) {
    return {};
  }
  if (argsCount === 2) {
    if (arguments[1] instanceof AbortSignal) {
      return {signal: arguments[1]};
    }
    if (typeof arguments[1] === 'number') {
      return {timeout: arguments[1]};
    }
    return {data: arguments[1]};
  }
  if (argsCount === 3) {
    if (arguments[2] instanceof AbortSignal) {
      return {
        data  : arguments[1],
        signal: arguments[2],
      };
    }
    return {
      data   : arguments[1],
      timeout: arguments[2],
    };
  }
  return {
    data   : arguments[1],
    signal : arguments[3],
    timeout: arguments[2],
  };
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {Promise<string>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {Promise<string>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout = get.timeout]
 * @param {AbortSignal} [signal]
 * @returns {Promise<string>}
 *
 * @param {string} url
 * @returns {Promise<string>}
 */
function get(url) {
  return ajax(url, Object.assign({
    method      : 'GET',
    responseType: 'text',
    timeout     : /** @type {{timeout?: number}} */ (get).timeout,
  }, /** @type {Function} */ (_getOptionsFromArgs).apply(null, arguments)));
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {Promise<ArrayBuffer>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {Promise<ArrayBuffer>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout = getArrayBuffer.timeout]
 * @param {AbortSignal} [signal]
 * @returns {Promise<ArrayBuffer>}
 *
 * @param {string} url
 * @returns {Promise<ArrayBuffer>}
 */
function getArrayBuffer(url) {
  return ajax(url, Object.assign({
    method      : 'GET',
    responseType: 'arraybuffer',
    timeout     : /** @type {{timeout?: number}} */ (getArrayBuffer).timeout,
  }, /** @type {Function} */ (_getOptionsFromArgs).apply(null, arguments)));
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {Promise<PNG>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {Promise<PNG>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout = getPNG.timeout]
 * @param {AbortSignal} [signal]
 * @returns {Promise<PNG>}
 *
 * @param {string} url
 * @returns {Promise<PNG>}
 */
function getPNG(url) {
  return ajax(url, Object.assign({
    method      : 'GET',
    responseType: 'arraybuffer',
    timeout     : /** @type {{timeout?: number}} */ (getPNG).timeout,
  }, /** @type {Function} */ (_getOptionsFromArgs).apply(null, arguments))).then(parsePNGResponse);
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout = getJSON.timeout]
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @param {string} url
 * @returns {Promise<JSONValue>}
 */
function getJSON(url) {
  return ajax(url, Object.assign({
    method      : 'GET',
    responseType: 'json',
    timeout     : /** @type {{timeout?: number}} */ (getJSON).timeout,
  }, /** @type {Function} */ (_getOptionsFromArgs).apply(null, arguments)));
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout = getJSONP.timeout]
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @param {string} url
 * @returns {Promise<JSONValue>}
 */
function getJSONP(url) {
  const options = /** @type {Function} */ (_getOptionsFromArgs).apply(null, arguments);
  options.data = Object.assign({}, options.data, {callback: '_'});
  return ajax(url, Object.assign({
    method      : 'GET',
    responseType: 'text',
    timeout     : /** @type {{timeout?: number}} */ (getJSONP).timeout,
  }, options)).then(parseJSONPResponse);
}
/**
 * @overload
 * @param {string} url
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @overload
 * @param {string} url
 * @param {AjaxOptions['data']} [data]
 * @param {AjaxOptions['timeout']} [timeout = post.timeout]
 * @param {AbortSignal} [signal]
 * @returns {Promise<JSONValue>}
 *
 * @param {string} url
 * @returns {Promise<JSONValue>}
 */
function post(url) {
  return ajax(url, Object.assign({
    method      : 'POST',
    responseType: 'json',
    timeout     : /** @type {{timeout?: number}} */ (post).timeout,
  }, /** @type {Function} */ (_getOptionsFromArgs).apply(null, arguments)));
}

/**
 * @template {(...args: any[]) => any} F
 * @param {F} fn
 * @param {Parameters<typeof retriable<F>>[1]} [options]
 * @returns {ReturnType<typeof retriable<F>>}
 */
function ajaxRetriable(fn, options) {
  options = Object.assign({shouldRetryAfterError: shouldRetryRequest}, options);
  return retriable(fn, options);
}
/**
 * @template T
 * @param {(url: string, signal: AbortSignal) => Promise<T>} fetcher
 * @param {string} url
 * @returns {[promise: Promise<T>, abort: () => void]}
 */
function fetchResource(fetcher, url) {
  let resource = fetchResource.cache.get(url);
  if (!resource) {
    const abortController = new AbortController();

    resource = {
      usersCount: 0,
      promise   : fetcher(url, abortController.signal),
      abort() {
        resource = $(resource);
        if (resource.usersCount > 0) {
          resource.usersCount--;
          if (resource.usersCount === 0) {
            abortController.abort();
          }
        }
      },
    };
    fetchResource.cache.set(url, resource);
    resource.promise
      .catch(function(/** @type {Error} */ error) {
        if (isTimeoutError(error)) {
          abortController.abort();
        }
      })
      .finally(function() {
        fetchResource.cache.delete(url);
      });
  }

  resource.usersCount++;
  return [resource.promise, resource.abort];
}

export {
  hub as ajaxHub,
  statistics as ajaxStatistics,
  ajax,
  ajaxRetriable,
  fetchResource,
};
