/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {round, lerpN} from './math.js';
import {$} from './typescript.js';

/** @typedef {NumberArray} BezierSegment */
/** @typedef {[x: BezierSegment[], y: BezierSegment[]]} BezierPath */

/**
 * @template {Readonly<BezierPath>} T
 * @param {T} bezierPath
 * @param {number} segmentsCount
 * @returns {MarkedLike<T, XY>[]}
 */
export function bezierPathToXYs(bezierPath, segmentsCount) {
  const xs = bezierPath[0];
  const ys = bezierPath[1];
  const numPaths = xs.length;
  const step = numPaths / segmentsCount;
  /** @type {XY[]} */
  const points = [];
  let t = 0;

  for (let i = 0; i < numPaths; i++) {
    for (; t <= 1; t = round(t + step, -15)) {
      points.push([lerpN($(xs[i]), t), lerpN($(ys[i]), t)]);
    }
    t -= 1;
  }

  return points;
}

/**
 * @template {BezierPath} T
 * @param {Readonly<T>} bezierPath
 * @returns {T}
 */
export function cloneBezierPath(bezierPath) {
  return /** @type {T} */ ([
    bezierPath[0].map(cloneBezierSegment),
    bezierPath[1].map(cloneBezierSegment),
  ]);
}

/**
 * @param {Readonly<BezierSegment>} array
 * @returns {BezierSegment}
 */
function cloneBezierSegment(array) {
  return array.slice();
}
