/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {asIterable} from './iter.js';
import {wrapFn, namedFunction} from './fn_utils.js';
import {flattenArray, flattenArraySlice} from './utils.js';
import {$} from './typescript.js';

/**
 * @template {{}} T
 * @param {T} object
 * @param {MethodName<T>} methodName
 * @param {(this: T, ...args: any[]) => void} wrapper
 * @returns {void}
 * @note `object[methodName]` must be a function.
 */
export function wrapMethod(object, methodName, wrapper) {
  object[methodName] = /** @type {any} */ (
    wrapFn(/** @type {Func<T>} */ (object[methodName]), wrapper)
  );
}
/**
 * @template {{}} T
 * @param {T} object
 * @param {PartialMethods<T>} wrappers
 * @returns {void}
 * @note `object[key]` must be a function for all keys of `wrappers`.
 */
export function wrapMethods(object, wrappers) {
  for (const key in wrappers) {
    if (Object.hasOwn(wrappers, key)) {
      const methodName = /** @type {MethodName<T>} */ (key);
      const wrapper = wrappers[methodName];
      if (wrapper) {
        wrapMethod(object, methodName, /** @type {Func<T>} */ (wrapper));
      }
    }
  }
}
/**
 * @template {Class} C
 * @param {C} cls
 * @param {string & InstanceMethodName<C>} methodName
 * @param {(this: InstanceType<C>, ...args: any[]) => void} wrapper
 * @returns {void}
 * @note `methodName` must be the name of a method of `cls`.
 */
export function wrapClassMethod(cls, methodName, wrapper) {
  wrapMethod(cls.prototype, methodName, wrapper);
}
/**
 * @template {Class} C
 * @param {C} cls
 * @param {PartialInstanceMethods<C>} wrappers
 * @returns {void}
 * @note All keys of `wrappers` must be names of methods of `cls`.
 */
export function wrapClassMethods(cls, wrappers) {
  for (const key in wrappers) {
    if (Object.hasOwn(wrappers, key)) {
      const methodName = /** @type {string & InstanceMethodName<C>} */ (key);
      const wrapper = wrappers[methodName];
      if (wrapper) {
        wrapClassMethod(cls, methodName, wrapper);
      }
    }
  }
}

/** @typedef {Record<keyof any, unknown>} ClassInitializerFields */
/**
 * @template {Class} C
 * @typedef {(this: C, cls: C) => void} ClassInitializerFn
 */
/**
 * @template {Class} C
 * @typedef {ClassInitializerFields
 *           | ClassInitializerFn<C>
 *           | null | undefined | false} ClassInitializer
 */
/**
 * @template {Class} C
 * @typedef ClassExtensionMeta
 * @prop {string} [name]
 * @prop {ClassExtension<C>} [self]
 * @prop {ClassExtension<C>[]} [deps]
 */
/**
 * @template {Class} C
 * @typedef ClassExtensionStringMeta
 * @prop {ClassExtensionMeta<C>} [meta]
 */
/**
 * @template {Class} C
 * @typedef ClassObjectExtensionStatic
 * @prop {OneOrArray<ClassInitializer<C>>} [static]
 */
/**
 * @template {Class} C
 * @typedef {ClassObjectExtensionStatic<C>
 *           & ClassExtensionStringMeta<C>} ClassExtension
 */
/**
 * @template {Class} C
 * @typedef {ClassExtension<C>
 *           & Partial<InstanceType<C>>
 *           & ThisType<InstanceType<C>>
 *           & Record<keyof any, unknown>} ClassExtensionWithInfer
 */
/**
 * @template {Class} [C = Class]
 * @typedef ClassDefinition
 * @prop {OneOrArray<ClassInitializer<C>>} [static]
 * @prop {C} [constructor]
 */

/**
 * @template {Class} C
 * @this {((object: any, propertyDescriptors: PropertyDescriptorMap) => void) | void}
 * @param {ClassInitializer<C>} initializer
 * @returns {ClassInitializerFn<C> | null | undefined | false}
 * @throws {TypeError} If `initializer` is neither a function, an object,
 * `null`, `undefined`, nor `false`.
 */
function initializerFnFromInitializer(initializer) {
  if (!initializer || typeof initializer === 'function') {
    return initializer;
  }
  if (typeof initializer === 'object') {
    const classPropertyDescriptors = (
      Object.getOwnPropertyDescriptors(initializer)
    );
    delete classPropertyDescriptors.prototype;
    delete /** @type {any} */ (classPropertyDescriptors).constructor;
    if (
      classPropertyDescriptors.name && !classPropertyDescriptors.displayName
    ) {
      classPropertyDescriptors.displayName = classPropertyDescriptors.name;
    }
    const initializeWithPropertyDescriptors = this || Object.defineProperties;
    return function() {
      initializeWithPropertyDescriptors(this, classPropertyDescriptors);
    };
  }
  throw new TypeError(
    'Expected static initializer to be an optional function or object, ' +
    'but got `' + initializer + '`.'
  );
}
/**
 * @template {Class} C
 * @this {ClassInitializerFn<C>[]}
 * @param {OneOrArray<ClassInitializer<C>>} initializers
 * @returns {void}
 */
function addClassInitializerFnsFromInitializers(initializers) {
  if (initializers) {
    const initializerFns = this;
    for (const initializer of asIterable(initializers)) {
      const initializerFn = initializerFnFromInitializer(initializer);
      if (initializerFn) {
        initializerFns.push(initializerFn);
      }
    }
  }
}

/**
 * @template {Class} C
 * @this {C}
 * @param {ClassInitializerFn<C>} initialize
 * @returns {void}
 */
function callClassInitializer(initialize) {
  initialize.call(this, this);
}

/**
 * @param {Class | ClassDefinition | null | undefined | false} [superClass]
 * @param {ReadonlyArray<ClassDefinition | null | undefined | false>} args
 * @returns {Class}
 */
export function createClass(superClass, ...args) {
  if (arguments.length === 0) {
    return class {};
  }

  let propsIndex = 1;
  if (superClass === null || typeof superClass !== 'function') {
    superClass = undefined;
    propsIndex = 0;
  }

  /** @type {Class | undefined} */
  let constructor;
  /** @type {ClassInitializerFn<Class>[]} */
  const classInitializerFns = [];
  /** @type {(ClassDefinition | null | undefined | false)[]} */
  const props = flattenArraySlice(arguments, propsIndex);
  /** @type {PropertyDescriptorMap} */
  const propertyDescriptors = {};

  for (const properties of props) {
    if (properties) {
      if (Object.hasOwn(properties, 'static') && properties.static) {
        addClassInitializerFnsFromInitializers.call(
          classInitializerFns,
          properties.static
        );
      }
      if (
        Object.hasOwn(properties, 'constructor') &&
        properties.constructor
      ) {
        constructor = properties.constructor;
      }
      Object.assign(
        propertyDescriptors,
        Object.getOwnPropertyDescriptors(properties)
      );
    }
  }

  if (typeof constructor === 'function') {
    if (superClass === null) {
      constructor.prototype = Object.create(null);
      constructor.prototype.constructor = constructor;
    } else if (superClass) {
      constructor.prototype = Object.create(superClass.prototype);
      constructor.prototype.constructor = constructor;
      Object.setPrototypeOf(constructor, superClass);
    }
  } else {
    constructor = (
      superClass === undefined ? class {} : class extends superClass {}
    );
  }

  delete propertyDescriptors.static;
  delete /** @type {any} */ (propertyDescriptors).constructor;
  Object.defineProperties(constructor.prototype, propertyDescriptors);

  classInitializerFns.forEach(callClassInitializer, constructor);
  return constructor;
}

/**
 * @param {PropertyDescriptor} originalDescriptor
 * @param {PropertyDescriptor} extensionDescriptor
 * @returns {void}
 * @throws {TypeError} If trying to extend an accessor property with a data one
 * @throws {TypeError} If trying to extend a readable property with a non-readable one
 * @throws {TypeError} If trying to extend a writable property with a non-writable one
 * @throws {TypeError} If trying to extend a method with a value (and vice-versa)
 */
function assertCanExtendPropertyDescriptor(originalDescriptor, extensionDescriptor) {
  const originalValue = originalDescriptor.value;
  let originalGetter = originalDescriptor.get;
  let originalSetter = originalDescriptor.set;
  const isOriginalDataDescriptor = 'value' in originalDescriptor;
  const isExtensionDataDescriptor = 'value' in extensionDescriptor;
  const isOriginalReadable = isOriginalDataDescriptor || originalGetter;
  const isOriginalWritable = originalDescriptor.writable || originalSetter;
  const isExtensionReadable = isExtensionDataDescriptor || extensionDescriptor.get;
  const isExtensionWritable = extensionDescriptor.writable || extensionDescriptor.set;
  const isOriginalCallable = typeof originalValue === 'function';
  const isExtensionCallable = typeof extensionDescriptor.value === 'function';

  if (!isOriginalDataDescriptor && isExtensionDataDescriptor) {
    throw new TypeError("Can't extend an accessor property with a data one");
  }
  if (isOriginalReadable && !isExtensionReadable) {
    throw new TypeError("Can't extend a readable property with a non-readable one");
  }
  if (isOriginalWritable && !isExtensionWritable) {
    throw new TypeError("Can't extend a writable property with a non-writable one");
  }
  if (originalValue != null && isOriginalCallable !== isExtensionCallable) {
    throw new TypeError("Can't extend a method with a value (and vice-versa)");
  }
}
/**
 * @param {keyof any} key
 * @param {PropertyDescriptor} descriptor
 * @param {string} [prefix]
 * @returns {PropertyDescriptor}
 */
function synthesizeMissingPropertyAccessors(key, descriptor, prefix = '') {
  const originalValue = descriptor.value;
  const property = String(key);
  const propertyInstanceKey = '~' + property;
  delete descriptor.value;
  descriptor.get ??= namedFunction(
    prefix + '[fake] get ' + property,
    function() {
      if (Object.hasOwn(this, propertyInstanceKey)) {
        return this[propertyInstanceKey];
      }
      return originalValue;
    }
  );
  descriptor.set ??= namedFunction(
    prefix + '[fake] set ' + property,
    function(value) {
      this[propertyInstanceKey] = value;
    }
  );
  return descriptor;
}
/**
 * @param {any} object
 * @param {string | symbol} property
 * @returns {PropertyDescriptor | undefined}
 */
function getObjectPropertyDescriptor(object, property) {
  let descriptor;
  while (object && !descriptor) {
    descriptor = Object.getOwnPropertyDescriptor(object, property);
    object = Object.getPrototypeOf(object);
  }
  return descriptor;
}
/**
 * @param {any} object
 * @param {PropertyDescriptorMap} propertyDescriptors
 * @param {string | undefined} name
 * @returns {void}
 * @throws {TypeError} If trying to extend an accessor property with a data one
 * @throws {TypeError} If trying to extend a readable property with a non-readable one
 * @throws {TypeError} If trying to extend a writable property with a non-writable one
 * @throws {TypeError} If trying to extend a method with a value (and vice-versa)
 */
function extendObjectWithPropertyDescriptors(object, propertyDescriptors, name) {
  // eslint-disable-next-line guard-for-in
  for (const property in propertyDescriptors) {
    const originalDescriptor = getObjectPropertyDescriptor(object, property);
    let extensionDescriptor = $(propertyDescriptors[property]);

    if (originalDescriptor) {
      assertCanExtendPropertyDescriptor(originalDescriptor, extensionDescriptor);

      const originalValue = originalDescriptor.value;
      const isExtensionDataDescriptor = 'value' in extensionDescriptor;
      const isOriginalCallable = typeof originalValue === 'function';
      const prefix = name ? '[' + name + '] ' : undefined;

      if (isOriginalCallable) {
        extensionDescriptor.value = (
          wrapFn(originalValue, extensionDescriptor.value, prefix)
        );
      } else if (!isExtensionDataDescriptor) {
        synthesizeMissingPropertyAccessors(property, originalDescriptor, prefix);

        if (extensionDescriptor.get) {
          extensionDescriptor.get = (
            wrapFn($(originalDescriptor.get), extensionDescriptor.get, prefix)
          );
        }
        if (extensionDescriptor.set) {
          extensionDescriptor.set = (
            wrapFn($(originalDescriptor.set), extensionDescriptor.set, prefix)
          );
        }
      }
    }
    Object.defineProperty(object, property, extensionDescriptor);
  }
}

/**
 * @template {Class} C
 * @param {ClassExtension<C>} extension
 * @param {C} cls
 * @param {string | undefined} name
 * @returns {void}
 * @throws {TypeError} If trying to extend an accessor property with a data one
 * @throws {TypeError} If trying to extend a readable property with a non-readable one
 * @throws {TypeError} If trying to extend a writable property with a non-writable one
 * @throws {TypeError} If trying to extend a method with a value (and vice-versa)
 * @throws {TypeError} If a static initializer is neither a function, an object,
 * `null`, `undefined` nor `false`.
 */
function applyExtension(extension, cls, name) {
  const propertyDescriptors = Object.getOwnPropertyDescriptors(extension);

  delete propertyDescriptors.static;
  delete /** @type {any} */ (propertyDescriptors).constructor;
  delete propertyDescriptors.meta;

  extendObjectWithPropertyDescriptors(cls.prototype, propertyDescriptors, name);

  for (const initializer of asIterable(extension.static ?? [])) {
    const initializerFn = initializerFnFromInitializer.call(
      extendWithPropertyDescriptors,
      // @ts-expect-error ZKH: don't understand why, maybe bug in TS?
      initializer
    );
    if (initializerFn) {
      initializerFn.call(cls, cls);
    }
  }

  /**
   * @param {any} object
   * @param {PropertyDescriptorMap} descriptors
   * @returns {void}
   */
  function extendWithPropertyDescriptors(object, descriptors) {
    extendObjectWithPropertyDescriptors(object, descriptors, name);
  }
}

/** @type {WeakMap<Class, Set<ClassExtension<any>>>} */
const extensionsForClass = new WeakMap();

/**
 * Extends `this` with `extension`
 *
 * @template {Class} C
 * @this {C} The class to extend
 * @param {ClassExtension<C>} extension The extension to apply to `this`
 * @returns {void}
 * @throws {TypeError} If trying to extend an accessor property with a data one
 * @throws {TypeError} If trying to extend a readable property with a non-readable one
 * @throws {TypeError} If trying to extend a writable property with a non-writable one
 * @throws {TypeError} If trying to extend a method with a value (and vice-versa)
 * @throws {TypeError} If the class was already extended with `extension`
 * @throws {TypeError} If one or more dependencies of `extension` have not been
 * applied to the class
 * @throws {TypeError} If a static initializer is neither a function, an object,
 * `null`, `undefined` nor `false`.
 */
function extendClassWithExtension(extension) {
  const cls = this;
  const meta = getExtensionMeta(extension);
  const name = meta?.name;

  let appliedExtensions = extensionsForClass.get(cls);
  if (!appliedExtensions) {
    appliedExtensions = new Set();
    extensionsForClass.set(cls, appliedExtensions);
  } else if (appliedExtensions.has(extension)) {
    throw new TypeError(
      `Can't re-extend a class with the same extension${word(name)}`
    );
  }
  for (const dep of meta?.deps ?? []) {
    if (!appliedExtensions.has(dep)) {
      const depName = getExtensionMeta(dep)?.name;
      throw new TypeError(
        `Extension${word(name)} depends on a missing extension${word(depName)}`
      );
    }
  }

  applyExtension(extension, cls, name);
  appliedExtensions.add(meta?.self ?? extension);
}

/**
 * @template {Class} C
 * @param {ClassExtension<C>} extension
 * @returns {ClassExtensionMeta<C> | undefined}
 */
function getExtensionMeta(extension) {
  if (Object.hasOwn(extension, 'meta')) {
    return /** @type {ClassExtensionStringMeta<C>} */ (extension).meta;
  }
  return undefined;
}

/**
 * @param {string | null | undefined} string
 * @param {string} [prefix]
 * @returns {string}
 */
function word(string, prefix = ' ') {
  return string ? `${prefix}${string}` : '';
}

/**
 * Extends `cls` with all `extensions` in order.
 *
 * @template {Class} C
 * @param {C} cls The class to extend
 * @param {OneOrArray<ClassExtensionWithInfer<C>>[]} extensions The extensions to
 * apply to `cls`
 * @returns {void}
 * @throws {TypeError} If trying to extend an accessor property with a data one
 * @throws {TypeError} If trying to extend a readable property with a non-readable one
 * @throws {TypeError} If trying to extend a writable property with a non-writable one
 * @throws {TypeError} If trying to extend a method with a value (and vice-versa)
 * @throws {TypeError} If one or more extensions are classes and they dont
 * inherit directly from `cls`
 * @throws {TypeError} If applying an extension that was already applied to `cls`
 * @throws {TypeError} If one or more dependencies of an extension have not been
 * applied to `cls` before it
 * @throws {TypeError} If a static initializer is neither a function, an object,
 * `null`, `undefined` nor `false`.
 */
export function extendClass(cls, ...extensions) {
  /** @type {ClassExtension<C>[]} */ (flattenArray(extensions))
  .forEach(extendClassWithExtension, cls);
}

/**
 * @template T
 * @template {(...args: any[]) => any} F
 * @param {T} bindTo
 * @param {F} fn
 * @returns {F & ThisType<T>}
 */
function proxyBind(bindTo, fn) {
  return new Proxy(fn, {
    apply(target, thisArg, argumentsList) {
      return Reflect.apply(target, bindTo, argumentsList);
    },
  });
}

/**
 * @param {string | symbol} key
 * @returns {key is `_${string}`}
 */
function shouldProxyInstanceKey(key) {
  return typeof key === 'string' && key[0] !== '_';
}

/**
 * Wraps a class with a proxy that gives access to it's instance's properties.
 * The instance must be accessible via `classTarget.instance`.
 * Doesn't give access to any instance property prefixed with an "_"
 * (indicating that it's not public).
 *
 * @template {Class & {instance: InstanceType<Class>}} C
 * @param {C} classTarget The class to wrap
 * @returns {C & Omit<InstanceType<C>, `_${string}` | keyof C>} A proxy to the
 * class giving access to it's instance's properties.
 *
 * @example
 * const A = createInstanceProxy(class {
 *   static classProperty = 2;
 *   static classMethod() {
 *     return 1;
 *   }
 *   static instance = new this();
 *
 *   prototypeProperty = 5;
 *   _prototypeProperty = 10;
 *
 *   constructor() {
 *     this.instanceProperty = 'a';
 *     this._instanceProperty = 'b';
 *   }
 *
 *   instanceMethod() {
 *     return this._instanceProperty + this._prototypeProperty;
 *   }
 *   _instanceMethod() {
 *     return 11;
 *   }
 * });
 *
 * console.assert(A.classProperty === 2);
 * console.assert(A.classMethod() === 1);
 * console.assert(A.prototypeProperty === 5);
 * console.assert(A.instanceProperty === 'a');
 * console.assert(A.instanceMethod() === 'b10');
 * console.assert(A._prototypeProperty === undefined);
 * console.assert(A._instanceMethod === undefined);
 *
 * A.classProperty = 5;
 * console.assert(A.classProperty === 5);
 *
 * A.instanceProperty = 'c';
 * console.assert(A.instanceProperty === 'c');
 * console.assert(A.instance.instanceProperty === 'c');
 *
 * A.newClassProperty = 'z';
 * console.assert(A.newClassProperty === 'z');
 * console.assert(A.instance.newClassProperty === undefined);
 */
export function createInstanceProxy(classTarget) {
  const proxy = new Proxy(classTarget, {
    get(target, property, receiver) {
      if (
        !getObjectPropertyDescriptor(target, property) &&
        shouldProxyInstanceKey(property)
      ) {
        const value = target.instance[property];
        if (typeof value === 'function') {
          return proxyBind(target.instance, value);
        }
        return value;
      }
      return Reflect.get(target, property, receiver);
    },
    set(target, property, value, receiver) {
      if (
        !getObjectPropertyDescriptor(target, property) &&
        shouldProxyInstanceKey(property) &&
        getObjectPropertyDescriptor(target.instance, property)
      ) {
        receiver = target.instance;
        target = target.instance;
      }
      return Reflect.set(target, property, value, receiver);
    },

    // Needed for auto-complete in devtools
    ownKeys(target) {
      return [...new Set([
        ...Reflect.ownKeys(target),
        ...Reflect.ownKeys(target.prototype).filter(shouldProxyInstanceKey),
        ...Reflect.ownKeys(target.instance).filter(shouldProxyInstanceKey),
      ])];
    },
  });
  return /** @type {C & Omit<InstanceType<C>, `_${string}` | keyof C>} */ (proxy);
}
