/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/** */
export const SECOND_IN_MILLISECONDS = /** @type {Milliseconds} */ (1_000);
export const MINUTE_IN_MILLISECONDS = /** @type {Milliseconds} */ (60 * 1_000);
export const HOUR_IN_MILLISECONDS = /** @type {Milliseconds} */ (60 * 60 * 1_000);
export const DAY_IN_MILLISECONDS = /** @type {Milliseconds} */ (24 * 60 * 60 * 1_000);
export const WEEK_IN_MILLISECONDS = /** @type {Milliseconds} */ (7 * 24 * 60 * 60 * 1_000);
export const MONTH_IN_MILLISECONDS = /** @type {Milliseconds} */ (31 * 24 * 60 * 60 * 1_000);
export const YEAR_IN_MILLISECONDS = /** @type {Milliseconds} */ (365 * 24 * 60 * 60 * 1_000);
export const DECADE_IN_MILLISECONDS = /** @type {Milliseconds} */ (10 * 365 * 24 * 60 * 60 * 1_000);
export const CENTURY_IN_MILLISECONDS = /** @type {Milliseconds} */ (100 * 365 * 24 * 60 * 60 * 1_000);

export const UNIT_LENGTHS = /** @type {const} */ ([
  SECOND_IN_MILLISECONDS,
  MINUTE_IN_MILLISECONDS,
  HOUR_IN_MILLISECONDS,
  DAY_IN_MILLISECONDS,
  WEEK_IN_MILLISECONDS,
  MONTH_IN_MILLISECONDS,
  YEAR_IN_MILLISECONDS,
  DECADE_IN_MILLISECONDS,
  CENTURY_IN_MILLISECONDS,
]);
export const LENGTH_FOR_UNIT = /** @type {const} */ ({
  s: SECOND_IN_MILLISECONDS,
  M: MINUTE_IN_MILLISECONDS,
  h: HOUR_IN_MILLISECONDS,
  d: DAY_IN_MILLISECONDS,
  w: WEEK_IN_MILLISECONDS,
  m: MONTH_IN_MILLISECONDS,
  y: YEAR_IN_MILLISECONDS,
  D: DECADE_IN_MILLISECONDS,
  c: CENTURY_IN_MILLISECONDS,
});
export const UNIT_WITH_LENGTH = /** @type {const} */ ({
  [SECOND_IN_MILLISECONDS]: 's',
  [MINUTE_IN_MILLISECONDS]: 'M',
  [HOUR_IN_MILLISECONDS]: 'h',
  [DAY_IN_MILLISECONDS]: 'd',
  [WEEK_IN_MILLISECONDS]: 'w',
  [MONTH_IN_MILLISECONDS]: 'm',
  [YEAR_IN_MILLISECONDS]: 'y',
  [DECADE_IN_MILLISECONDS]: 'D',
  [CENTURY_IN_MILLISECONDS]: 'c',
});

export const DATE_MIN_TIME = /** @type {Milliseconds} */ (-8_640_000_000_000_000);
export const DATE_MAX_TIME = /** @type {Milliseconds} */ (8_640_000_000_000_000);

const ISO_DATE_REGEXP = /^(\d{4}-\d{2}-\d{2})[ T](\d{2}:\d{2}:\d{2})Z?$/;

/**
 * @param {Milliseconds} time - in milliseconds since 01/01/1970
 * @returns {Timestamp} `time` as Uint32 seconds since 01/01/1970
 */
export function timeToTimestamp(time) {
  // eslint-disable-next-line no-bitwise
  return (time / SECOND_IN_MILLISECONDS) >>> 0;
}

/**
 * @param {Timestamp} timestamp - as Uint32 seconds since 01/01/1970
 * @returns {Milliseconds} `timestamp` as milliseconds since 01/01/1970
 */
export function timeFromTimestamp(timestamp) {
  return timestamp * SECOND_IN_MILLISECONDS;
}

/**
 * @param {Date} date
 * @returns {Timestamp} `date` as Uint32 seconds since 01/01/1970
 */
export function dateToTimestamp(date) {
  return timeToTimestamp(date.getTime());
}

/**
 * @param {Timestamp} timestamp - as Uint32 seconds since 01/01/1970
 * @returns {Date} `timestamp` as a Date
 */
export function dateFromTimestamp(timestamp) {
  return new Date(timeFromTimestamp(timestamp));
}

/**
 * @param {Date?} date
 * @returns {boolean}
 */
export function isInvalidDate(date) {
  return !date || isNaN(date.getTime());
}

/**
 * @param {Date?} date
 * @returns {boolean}
 */
export function isValidDate(date) {
  return !isInvalidDate(date);
}

/**
 * @param {Date} date
 * @returns {Date}
 */
export function getStartOfUTCDay(date) {
  if (isInvalidDate(date)) {
    return date;
  }

  return new Date(Date.UTC(
    date.getUTCFullYear(),
    date.getUTCMonth(),
    date.getUTCDate()
  ));
}

/**
 * @param {Date} date
 * @returns {Date}
 */
export function getEndOfUTCDay(date) {
  if (isInvalidDate(date)) {
    return date;
  }

  return new Date(Date.UTC(
    date.getUTCFullYear(),
    date.getUTCMonth(),
    date.getUTCDate() + 1,
    0 /*hour*/,
    0 /*minute*/,
    0 /*second*/,
    -1 /*millisecond*/
  ));
}

/**
 * @param {string} stringDate
 * @returns {string}
 */
export function normalizeISODateString(stringDate) {
  if (typeof stringDate === 'string') {
    stringDate = stringDate.replace(ISO_DATE_REGEXP, '$1T$2Z');
  }
  return stringDate;
}

const isoDateRe = /^[+-]?0*(?<date>\d{4,}-\d{2}-\d{2})T(?<time>\d{2}:\d{2}:\d{2})\.(?<ms>\d{3})(?<tz>.*)$/;

/**
 * @param {Date} date
 * @param {string} [format]
 * @returns {string}
 */
export function formatUTCDate(date, format = '$<date> $<time> UTC') {
  return date.toISOString().replace(isoDateRe, format);
}
