/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {EventsDelegator} from './events_delegator.js';
import {$} from './typescript.js';

export const passiveEventsSupported = (function() {
  // See https://developer.mozilla.org/en-US/docs/Web/API/EventTarget/addEventListener
  /* MDN terms states: "Code samples added on or after August 20, 2010 are in the public domain." */
  let supported = false;

  try {
    const options = Object.defineProperty({}, 'passive', {
      //eslint-disable-next-line getter-return
      get() {
        supported = true;
      }
    });

    // @ts-expect-error "test" is not a known event type
    window.addEventListener('test', null, options);
  } catch (error) { /*ignored*/ }

  return supported;
})();
export const NONPASSIVE = (
  passiveEventsSupported ? {capture: false, passive: false} : false
);
export const PASSIVE = (
  passiveEventsSupported ? {capture: false, passive: true} : false
);
export const NONPASSIVE_CAPTURING = (
  passiveEventsSupported ? {capture: true, passive: false} : true
);
export const PASSIVE_CAPTURING = (
  passiveEventsSupported ? {capture: true, passive: true} : true
);

/**
 * @param {Event} event
 * @returns {void}
 */
export function preventEventDefault(event) {
  event.preventDefault();
}

/**
 * @param {Event} event
 * @returns {void}
 */
export function stopEventPropagation(event) {
  event.stopPropagation();
}

/**
 * @param {Event} event
 * @returns {void}
 */
export function preventDefaultAndStopEvent(event) {
  preventEventDefault(event);
  stopEventPropagation(event);
}

/**
 * @param {string} type
 * @returns {boolean | AddEventListenerOptions}
 */
function optionsForType(type) {
  const isPassive = type.indexOf('+passive') !== -1;
  const isCapture = type.indexOf('+capture') !== -1;
  return (
    isPassive
    ? (isCapture ? PASSIVE_CAPTURING : PASSIVE)
    : (isCapture ? NONPASSIVE_CAPTURING : NONPASSIVE)
  );
}

/**
 * @param {EventTarget} element
 * @param {string} type
 * @param {EventListenerOrEventListenerObject} callback
 * @param {boolean | AddEventListenerOptions} [options]
 * @returns {void}
 */
export function addDOMListener(element, type, callback, options) {
  if (!options) {
    options = optionsForType(type);
  }
  const plusIndex = type.indexOf('+');
  if (plusIndex !== -1) {
    type = type.substring(0, plusIndex);
  }
  if (type === 'blur' || type === 'focus') {
    // blur and focus do not bubble
    options = PASSIVE_CAPTURING;
  }
  element.addEventListener(type, callback, options);
}

/**
 * @param {EventTarget} element
 * @param {string} type
 * @param {EventListenerOrEventListenerObject} callback
 * @param {boolean | EventListenerOptions} [options]
 * @returns {void}
 */
export function removeDOMListener(element, type, callback, options) {
  if (!options) {
    options = optionsForType(type);
  }
  const plusIndex = type.indexOf('+');
  if (plusIndex !== -1) {
    type = type.substring(0, plusIndex);
  }
  if (type === 'blur' || type === 'focus') {
    // blur and focus do not bubble
    options = PASSIVE_CAPTURING;
  }
  element.removeEventListener(type, callback, options);
}

/**
 * @type {import('./events_delegator.js').EventsDelegatorOptions<string | Element, Element>}
 */
export const DOMEventsDelegatorOptions = {
  on : addDOMListener,
  off: removeDOMListener,

  getTarget(type, refs, [event]) {
    event.stopPropagation = function() {
      Event.prototype.stopPropagation.call(this);
      refs.propagationStopped = true;
    };
    event.stopImmediatePropagation = function() {
      Event.prototype.stopImmediatePropagation.call(this);
      refs.immediatePropagationStopped = true;
    };
    return event.target;
  },
  matchesSelector(type, element, selector, rootElement, refs) {
    /** @type {Element | null} */
    let el = element;
    while (el && el !== rootElement) {
      // if this is a match then we are done!
      if (
        el === selector ||
        (typeof selector === 'string' && el.matches(selector))
      ) {
        return el;
      }

      // if this element did not match but has a parent we should try
      // going up the tree to see if any of the parent elements match
      // for example if you are looking for a click on an <a> tag but there
      // is a <span> inside of the a tag that it is the target,
      // it should still work
      el = el.parentElement;
      refs.distance++;
    }
    return undefined;
  },
  callListener(type, listener, element, args, refs) {
    if (listener[0].apply(listener[1] || element, args) === false) {
      preventDefaultAndStopEvent(args[0]);
    }
  },
};

/**
 * @param {Element} element
 * @returns {EventsDelegator<string | Element, Element>}
 */
export function DOMEvents(element) {
  return EventsDelegator.for(element, DOMEventsDelegatorOptions);
}

/**
 * Calls `fn` when the document is loaded.
 *
 * If the document is already loaded (`document.readyState` is `'complete'`),
 * the `fn` is called immediately (synchronously),
 * otherwise a `load` event listener is added.
 *
 * @template T
 * @param {Func<T>} fn - The function to be called when the document is loaded.
 * @param {T} [thisArg] - The `this` used when calling `fn`.
 * @returns {void}
 */
export function whenReady(fn, thisArg) {
  if (document.readyState === 'complete') {
    fn.call($(thisArg));
  } else {
    window.addEventListener('load', function() {
      fn.call($(thisArg));
    });
  }
}

/**
 * @param {Element & ElementCSSInlineStyle} element
 * @param {string | string[] | null} cursor
 * @returns {void}
 */
export function setElementCursor(element, cursor) {
  if (cursor) {
    element.classList.add('has-global-cursor');

    if (Array.isArray(cursor)) {
      element.style.cursor = cursor.join(',');
    } else {
      element.style.cursor = cursor;
      if (element.style.cursor !== cursor) {
        element.style.cursor = '-webkit-' + cursor;
      }
    }
  } else {
    element.classList.remove('has-global-cursor');
    element.style.cursor = '';
  }
}
/**
 * @param {string | string[] | null} cursor
 * @returns {void}
 */
export function setGlobalCursor(cursor) {
  setElementCursor(document.body, cursor);
}

/**
 * @param {Element} node
 * @returns {void}
 */
export function removeAllChildrenOfNode(node) {
  // There are several ways to remove all children of a node:
  // 1. while (node.firstChild) node.removeChild(node.firstChild)
  // 2. while (node.firstChild) node.removeChild(node.lastChild)
  // 3. node.innerHTML = ''
  // 4. node.textContent = ''
  // 5. node.replaceChildren()
  //
  // But 3. seams to be fastest in general.
  node.innerHTML = '';
}

/**
 * @param {Window | Element} [element]
 * @returns {void}
 */
export function disableTextSelection(element) {
  (element || window).addEventListener(
    'selectstart',
    preventEventDefault,
    NONPASSIVE_CAPTURING
  );
}
/**
 * @param {Window | Element} [element]
 * @returns {void}
 */
export function enableTextSelection(element) {
  (element || window).removeEventListener(
    'selectstart',
    preventEventDefault,
    NONPASSIVE_CAPTURING
  );
}

/**
 * @param {MouseEvent | Touch} pointer
 * @param {Client<XY<Pixels>>} [clientXY]
 * @returns {Client<XY<Pixels>>}
 */
export function getClientXYOfPointer(pointer, clientXY) {
  if (!clientXY || clientXY.length === undefined) {
    clientXY = [0, 0];
  }
  clientXY[0] = pointer.clientX;
  clientXY[1] = pointer.clientY;
  return clientXY;
}
export const getClientXYFromMouseEvent = getClientXYOfPointer;
export const getClientXYOfTouch = getClientXYOfPointer;

/**
 * @param {ArrayLike<Touch> & Iterable<Touch>} touches
 * @param {Client<XY<Pixels>>} [xy]
 * @returns {Client<XY<Pixels>>}
 */
export function getCenterOfTouches(touches, xy) {
  let x = 0;
  let y = 0;
  for (const touch of touches) {
    x += touch.clientX;
    y += touch.clientY;
  }

  const l = touches.length || 1;
  if (!xy || xy.length === undefined) {
    xy = [0, 0];
  }
  xy[0] = x / l;
  xy[1] = y / l;
  return xy;
}

/**
 * @param {Element} element
 * @param {Client<XY<Pixels>>} clientXY
 * @param {XY<Pixels>} [xy]
 * @returns {XY<Pixels>}
 */
export function getXYInFromClientXY(element, clientXY, xy) {
  const clientRect = element.getBoundingClientRect();
  const width = element.clientWidth;
  const height = element.clientHeight;
  const inverseScaleX = width / clientRect.width;
  const inverseScaleY = height / clientRect.height;
  const clientAnchorX = (clientRect.left + clientRect.right) / 2;
  const clientAnchorY = (clientRect.top + clientRect.bottom) / 2;
  if (!xy || xy.length === undefined) {
    xy = [0, 0];
  }
  xy[0] = (clientXY[0] - clientAnchorX) * inverseScaleX + width / 2;
  xy[1] = (clientXY[1] - clientAnchorY) * inverseScaleY + height / 2;
  return xy;
}
/**
 * @param {Element} element
 * @param {MouseEvent} event
 * @param {XY<Pixels>} [xy]
 * @returns {XY<Pixels>}
 */
export function getXYInFromMouseEvent(element, event, xy) {
  if (!xy || xy.length === undefined) {
    xy = [0, 0];
  }
  if (
    event.eventPhase !== Event.NONE &&
    event.target === element &&
    !isNaN(event.offsetX) &&
    !isNaN(event.offsetY)
  ) {
    xy[0] = event.offsetX;
    xy[1] = event.offsetY;
    return xy;
  }
  return getXYInFromClientXY(element, getClientXYOfPointer(event, xy), xy);
}

/**
 * @param {Element} element
 * @returns {Client<XY>}
 */
export function getElementScaleInClient(element) {
  const clientRect = element.getBoundingClientRect();
  return [
    element.clientWidth / clientRect.width,
    element.clientHeight / clientRect.height,
  ];
}
