/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$} from './typescript.js';
import {asArray, upsert} from './utils.js';

const ROOT_SELECTOR = '**';

const _instances = new WeakMap();

// #region Exported typedefs

/**
 * @template TTarget
 * @template TThis
 * @template {any[]} [A = any[]]
 * @typedef {[
 *   callback: (this: NonNullable<TThis> | TTarget, ...args: A) => unknown,
 *   context: TThis,
 * ]} EventsDelegatorListener
 */
/**
 * @typedef EventsDelegatorContext
 * @prop {number} distance
 * @prop {boolean} propagationStopped
 * @prop {boolean} immediatePropagationStopped
 * @prop {unknown} returnValue
 */
/**
 * @template TSelector
 * @template TElement
 * @template [TTarget = TElement]
 * @typedef {EventsDelegatorBaseOptions<TSelector, TElement, TTarget> &
 *           ThisType<EventsDelegator<TSelector, TElement, TTarget>> &
 *           (EventsDelegatorOptionsWithHub<TElement>
 *            | EventsDelegatorOptionsWithoutHub<TElement>)} EventsDelegatorOptions
 */
/**
 * @template T
 * @typedef {(this: T, ...args: any[]) => void} EventsDelegatorCallback
 */

// #endregion

/**
 * @template [TSelector = string]
 * @template [TElement = Element]
 * @template [TTarget = TElement]
 */
export class EventsDelegator {
  // #region Private typedefs

  /**
   * @template TSelector
   * @template TElement
   * @template TTarget
   * @typedef EventsDelegatorBaseOptions
   * @prop {(
   *          type: string,
   *          context: EventsDelegatorContext,
   *          args: any[]
   *        ) => TTarget} getTarget
   * @prop {(
   *          type: string,
   *          target: TTarget,
   *          selector: TSelector,
   *          root: TElement,
   *          context: EventsDelegatorContext
   *        ) => TElement | null | undefined} matchesSelector
   * @prop {<T, A extends any[]>(
   *          type: string,
   *          listener: EventsDelegatorListener<TElement, T, A>,
   *          element: TElement,
   *          args: A,
   *          context: EventsDelegatorContext
   *        ) => void} callListener
   */
  /**
   * @template TElement
   * @typedef EventsDelegatorOptionsOnOff
   * @prop {(element: TElement, type: string, handler: (...args: any[]) => void) => void} on
   * @prop {(element: TElement, type: string, handler: (...args: any[]) => void) => void} off
   */
  /**
   * @template TElement
   * @typedef EventsDelegatorOptionsWithHub
   * @prop {EventsDelegatorOptionsOnOff<TElement>} hub
   */
  /**
   * @template TElement
   * @typedef {EventsDelegatorOptionsOnOff<TElement>} EventsDelegatorOptionsWithoutHub
   */

  // #endregion
  // #region Public

  /**
   * @template [TSelector = string]
   * @template {object} [TElement = Element]
   * @template [TTarget = TElement]
   * @param {TElement} element
   * @param {EventsDelegatorOptions<TSelector, TElement, TTarget>} options
   * @returns {EventsDelegator<TSelector, TElement, TTarget>}
   * @throws {TypeError} If `element` or `options` are not objects.
   */
  static for(element, options) {
    if (!element || typeof element !== 'object') {
      throw new TypeError('Expected an object "element" argument');
    }

    let instance = _instances.get(element);
    if (!instance) {
      instance = new EventsDelegator(element, options);
      _instances.set(element, instance);
    }
    return instance;
  }

  /**
   * @param {TElement} element
   * @param {EventsDelegatorOptions<TSelector, TElement, TTarget>} options
   * @throws {TypeError} If `options` is not an object.
   */
  constructor(element, options) {
    if (!options || typeof options !== 'object') {
      throw new TypeError('Expected an object "options" argument');
    }

    /** @type {Map<string, (...args: any[]) => void>} */
    this._handlersByType = new Map();
    /** @type {Map<string, Map<typeof ROOT_SELECTOR | TSelector, EventsDelegatorListener<TElement, any>[]>>} */
    this._listenersBySelectorByType = new Map();

    /** @type {TElement} */
    this.element = element;

    const hub = (
      /** @type {Partial<EventsDelegatorOptionsWithHub<TElement>>} */
      (options).hub
    );
    if (hub) {
      if (typeof hub.on !== 'function' || typeof hub.off !== 'function') {
        throw new Error();
      }
      /** @type {typeof hub.on} */
      this._on = hub.on.bind(hub);
      /** @type {typeof hub.off} */
      this._off = hub.off.bind(hub);
    } else {
      const {on, off} = (
        /** @type {Partial<EventsDelegatorOptionsWithoutHub<TElement>>} */
        (options)
      );
      if (typeof on !== 'function' || typeof off !== 'function') {
        throw new Error();
      }
      /** @type {typeof on} */
      this._on = on;
      /** @type {typeof off} */
      this._off = off;
    }
    if (
      typeof options.getTarget !== 'function' ||
      typeof options.matchesSelector !== 'function' ||
      typeof options.callListener !== 'function'
    ) {
      throw new Error();
    }
    /** @type {typeof options.getTarget} */
    this.getTarget = options.getTarget;
    /** @type {typeof options.matchesSelector} */
    this.matchesSelector = options.matchesSelector;
    /** @type {typeof options.callListener} */
    this.callListener = options.callListener;
  }

  /**
   * @template T
   * @overload
   * @param {string | string[]} types
   * @param {EventsDelegatorCallback<T>} callback
   * @param {T} [context]
   * @returns {this}
   *
   * @overload
   * @param {string | string[]} types
   * @param {NotFunction<TSelector> | null} selector
   * @param {EventsDelegatorCallback<T>} callback
   * @param {T} [context]
   * @returns {this}
   *
   * @param {string | string[]} types
   * @param {NotFunction<TSelector> | null | EventsDelegatorCallback<T>} selectorOrCallback
   * @param {EventsDelegatorCallback<T> | T} [callbackOrContext]
   * @param {T} [context]
   * @returns {this}
   */
  on(types, selectorOrCallback, callbackOrContext, context) {
    const selectorMissing = typeof selectorOrCallback === 'function';
    /** @type {NotFunction<TSelector> | null} */
    const selectorOrNull = /** @type {NotFunction<TSelector> | null} */ (
      selectorMissing ? null : selectorOrCallback
    );
    const selector = (selectorOrNull === null ? ROOT_SELECTOR : selectorOrNull);
    const callback = /** @type {EventsDelegatorCallback<T>} */ (
      selectorMissing ? selectorOrCallback : callbackOrContext
    );
    if (selectorMissing) {
      context = /** @type {T | undefined} */ (callbackOrContext);
    }

    for (const type of asArray(types)) {
      this._addListener(type, selector, callback, context);
    }
    return this;
  }

  /**
   * @template T
   * @overload
   * @param {string | string[]} types
   * @param {EventsDelegatorCallback<T>} callback
   * @param {T} [context]
   * @returns {this}
   *
   * @overload
   * @param {string | string[]} types
   * @param {NotFunction<TSelector> | null} selector
   * @param {EventsDelegatorCallback<T>} callback
   * @param {T} [context]
   * @returns {this}
   *
   * @param {string | string[]} types
   * @param {NotFunction<TSelector> | null | EventsDelegatorCallback<T>} selectorOrCallback
   * @param {EventsDelegatorCallback<T> | T} [callbackOrContext]
   * @param {T} [context]
   * @returns {this}
   */
  once(types, selectorOrCallback, callbackOrContext, context) {
    const selectorMissing = (
      callbackOrContext === undefined ||
      (context === undefined && typeof selectorOrCallback === 'function')
    );
    /** @type {NotFunction<TSelector> | null} */
    const selector = /** @type {NotFunction<TSelector> | null} */ (
      selectorMissing ? null : selectorOrCallback
    );
    const callback = /** @type {EventsDelegatorCallback<T>} */ (
      selectorMissing ? selectorOrCallback : callbackOrContext
    );
    if (selectorMissing) {
      context = /** @type {T | undefined} */ (callbackOrContext);
    }

    if (!callback) {
      return this;
    }

    const off = this.off.bind(
      this,
      types,
      selector,
      /** @type {EventsDelegatorCallback<unknown>} */ (listener),
      context
    );
    this.on(types, selector, listener, context);

    return this;

    /**
     * @this {ThisParameterType<typeof callback>}
     * @param {Parameters<typeof callback>} args
     */
    function listener() {
      off();
      /** @type {Function} */ (callback).apply(this, arguments);
    }
  }

  /**
   * @template T
   * @overload
   * @param {string | string[]} [types]
   * @param {EventsDelegatorCallback<T> | null} [callback]
   * @param {T | null} [context]
   * @returns {this}
   *
   * @overload
   * @param {string | string[]} [types]
   * @param {NotFunction<TSelector> | null} [selector]
   * @param {EventsDelegatorCallback<T> | null} [callback]
   * @param {T | null} [context]
   * @returns {this}
   *
   * @param {string | string[]} [types]
   * @param {NotFunction<TSelector> | null | EventsDelegatorCallback<T>} [selectorOrCallback]
   * @param {EventsDelegatorCallback<T> | null | T} [callbackOrContext]
   * @param {T | null} [context]
   * @returns {this}
   */
  off(types, selectorOrCallback, callbackOrContext, context) {
    if (!this._listenersBySelectorByType) {
      return this;
    }

    if (types == null) {
      // off()
      // off(null, ...)
      // off(undefined, ...)
      this._removeListener();
      return this;
    }

    const selectorMissing = (
      selectorOrCallback === undefined ||
      typeof selectorOrCallback === 'function'
    );
    /** @type {NotFunction<TSelector> | null} */
    const selectorOrNull = /** @type {NotFunction<TSelector> | null} */ (
      selectorMissing ? null : selectorOrCallback
    );
    const selector = (selectorOrNull === null ? ROOT_SELECTOR : selectorOrNull);
    const callback = /** @type {EventsDelegatorCallback<T> | null} */ (
      selectorMissing ? selectorOrCallback : callbackOrContext
    );
    if (selectorMissing) {
      context = /** @type {T | undefined} */ (callbackOrContext);
    }

    for (const type of asArray(types)) {
      this._removeListener(type, selector, callback, context);
    }
    return this;
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {string} type
   * @returns {(...args: any[]) => void}
   */
  _getHandler(type) {
    let handler = this._handlersByType.get(type);
    if (!handler) {
      handler = (
        /** @type {(...args: any[]) => void} */
        (this._handleEvent.bind(this, type))
      );
      this._handlersByType.set(type, handler);
    }
    return handler;
  }

  /**
   * @protected
   * @template T
   * @param {string} type
   * @param {typeof ROOT_SELECTOR | TSelector} selector
   * @param {EventsDelegatorCallback<T>} callback
   * @param {T} [context]
   */
  _addListener(type, selector, callback, context) {
    let listenersBySelector = this._listenersBySelectorByType.get(type);
    if (!listenersBySelector) {
      listenersBySelector = new Map();
      this._listenersBySelectorByType.set(type, listenersBySelector);
    }
    if (listenersBySelector.size === 0) {
      this._on(this.element, type, this._getHandler(type));
    }
    let listeners = listenersBySelector.get(selector);
    if (!listeners) {
      listeners = [];
      listenersBySelector.set(selector, listeners);
    }
    listeners.push([callback, context]);
  }
  /**
   * @protected
   * @template T
   * @param {string} [type]
   * @param {typeof ROOT_SELECTOR | TSelector} [selector]
   * @param {EventsDelegatorCallback<T> | null} [callback]
   * @param {T | null} [context]
   */
  _removeListener(type, selector, callback, context) {
    const element = this.element;
    const listenersBySelectorByType = this._listenersBySelectorByType;

    if (!type) {
      // example: #off()
      for (const t of listenersBySelectorByType.keys()) {
        this._off(element, t, this._getHandler(t));
      }
      listenersBySelectorByType.clear();
      return;
    }

    const listenersBySelector = listenersBySelectorByType.get(type);
    if (!listenersBySelector) {
      return;
    }

    if (!selector) {
      // example: #off('click')
      listenersBySelector.clear();
    } else if (callback) {
      // example: #off('click', '.sub-element', callback)
      // example: #off('click', '.sub-element', callback, context)
      const listeners = listenersBySelector.get(selector);
      if (listeners) {
        // remove the first occurrence of callback (with matching context)
        // from listeners
        for (let i = 0; i < listeners.length; i++) {
          const listener = $(listeners[i]);
          if (listener[0] === callback && listener[1] === context) {
            listeners.splice(i, 1);
            if (listeners.length === 0) {
              listenersBySelector.delete(selector);
            }
            break;
          }
        }
      }
    } else {
      // example: #off('click', '.sub-element')
      listenersBySelector.delete(selector);
    }
    if (listenersBySelector.size === 0) {
      this._off(element, type, this._getHandler(type));
    }
  }

  /**
   * @protected
   * @param {string} type
   * @param {any[]} args
   * @returns {any}
   */
  _handleEvent(type, ...args) {
    /**
     * @typedef Match
     * @prop {TElement} element
     * @prop {EventsDelegatorListener<TElement, any>[]} listeners
     */

    const root = this.element;
    const listenersBySelector = this._listenersBySelectorByType.get(type);

    if (!listenersBySelector) {
      return undefined;
    }

    const refs = {
      distance                   : 0,
      propagationStopped         : false,
      immediatePropagationStopped: false,
      returnValue                : undefined,
    };
    const target = this.getTarget(type, refs, args);
    /** @type {Map<number, Match[]>} */
    const matchesByDistance = new Map();

    for (const [selector, listeners] of listenersBySelector) {
      refs.distance = 0;
      const element = (
        selector === ROOT_SELECTOR
        ? root
        : this.matchesSelector(type, target, selector, root, refs)
      );
      if (element) {
        upsert(matchesByDistance, refs.distance, []).push({element, listeners});
      }
    }

    // eslint-disable-next-line no-labels
    propagation:
    for (const distance of [...matchesByDistance.keys()].sort()) {
      for (const {element, listeners} of $(matchesByDistance.get(distance))) {
        for (const listener of listeners) {
          this.callListener(type, listener, element, args, refs);
          if (refs.immediatePropagationStopped) {
            // eslint-disable-next-line no-labels
            break propagation;
          }
        }
      }
      if (refs.propagationStopped) {
        break;
      }
    }

    return refs.returnValue;
  }

  // #endregion
}
