/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$} from './typescript.js';

export class EventsHub {
  /**
   * @template T
   * @typedef {[callback: (this: T, ...args: any[]) => void, context: T]} Listener
   */

  constructor() {
    /** @type {Record<string, Listener<any>[]>} */
    this.events = {};
  }

  /**
   * @template T
   * @param {string} type
   * @param {(this: T, ...args: any[]) => void} func
   * @param {T} [ctx]
   * @returns {this}
   */
  on(type, func, ctx) {
    (this.events[type] = this.events[type] || []).push([func, ctx]);
    return this;
  }

  /**
   * @template T
   * @param {string} type
   * @param {(this: T, ...args: any[]) => void} func
   * @param {T} [ctx]
   * @returns {this}
   */
  once(type, func, ctx) {
    return this.on(type, function listener() {
      this.off(type, listener, this);
      /** @type {Function} */ (func).apply(ctx, arguments);
    }, this);
  }

  /**
   * @template T
   * @param {string} type
   * @param {(this: T, ...args: any[]) => void} func
   * @param {T} [ctx]
   * @returns {this}
   */
  off(type, func, ctx) {
    if (type) {
      const list = this.events[type];
      if (list) {
        if (func) {
          let i = list.length;
          while (i--) {
            const listener = $(list[i]);
            if (func === listener[0] && ctx === listener[1]) {
              list.splice(i, 1);
              break;
            }
          }
        } else {
          list.length = 0;
        }
      }
    } else {
      this.events = {};
    }
    return this;
  }

  /**
   * @param {string} type
   * @param {any[]} args
   * @returns {void}
   */
  emit(type, ...args) {
    const list = this.events[type]?.slice();
    if (list) {
      for (let i = 0, listener; (listener = list[i]); i++) {
        listener[0].apply(listener[1], args);
      }
    }
  }
}
