/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/// <reference path="./fn_utils_.d.ts" />
import {$} from './typescript.js';

const FIRST = Symbol('FIRST');
export const WRAPPED = Symbol('WRAPPED');
export const WRAPPER = Symbol('WRAPPER');
export const DECORATED = Symbol('DECORATED');
const NULL = /** @type {const} */ ([null, null, null]);

/** @type {Func | null} */
export let originalFn = null;
/** @type {any} */
let originalThis = null;
/** @type {any} */
let originalArgs = null;

/** @typedef {readonly [typeof originalFn, typeof originalThis, typeof originalArgs]} SavedOriginal */
/** @type {SavedOriginal[]} */
const stack = [];

Object.defineProperties(window, {
  originalFn: {
    configurable: false,
    enumerable  : false,
    get() {
      return originalFn;
    },
  },
  original: {
    configurable: false,
    enumerable  : false,
    get() {
      return originalFn && original;
    },
  },
});

/**
 * @template {Func} F
 * @typedef {F & {
 *   [FIRST]: Func;
 *   [WRAPPED]: WrappedFn<Func>;
 *   [WRAPPER]: WrappedFn<Func>;
 * }} WrappedFn
 */

function saveOriginal() {
  stack.push(originalFn ? [originalFn, originalThis, originalArgs] : NULL);
}
function restoreOriginal() {
  const last = $(stack.pop());
  originalFn = last[0];
  originalThis = last[1];
  originalArgs = last[2];
}

/**
 * @template {Func} F
 * @param {Func} func
 * @param {F} wrapper
 * @param {string} [prefix]
 * @returns {WrappedFn<F>}
 */
export function wrapFn(func, wrapper, prefix = 'wrapped ') {
  if (typeof func !== 'function') {
    throw new TypeError(
      'Expected 1st argument to be a function, but found "' + func + '"'
    );
  }
  if (typeof wrapper !== 'function') {
    throw new TypeError(
      'Expected 2nd argument to be a function, but found "' + wrapper + '"'
    );
  }
  const wrapping = /** @type {WrappedFn<F>} */ (
    functionCallsOriginal(wrapper) ? /** @type {F} */ (_wrap_) : wrapper
  );

  const wFunc = /** @type {WrappedFn<Func>} */ (func);
  const wWrapper = /** @type {WrappedFn<F>} */ (wrapper);
  const first = wFunc[FIRST] ?? func;
  wrapping[FIRST] = wWrapper[FIRST] = first;
  wrapping[WRAPPED] = wWrapper[WRAPPED] = wFunc[WRAPPER] ?? func;
  wrapping[WRAPPER] = wWrapper[WRAPPER] = wWrapper;
  namedFunction(prefix + first.name, wrapper);
  return wrapping;

  /**
   * @this {any}
   */
  function _wrap_() {
    saveOriginal();
    originalFn = func;
    originalThis = this;
    originalArgs = arguments;
    try {
      return /** @type {Function} */ (wrapper).apply(this, arguments);
    } finally {
      restoreOriginal();
    }
  }
}
export function original() {
  if (!originalFn) {
    throw new TypeError('no original saved');
  }

  const fn = originalFn;
  const thisArg = originalThis;
  const args = originalArgs;
  originalFn = null;
  originalThis = null;
  originalArgs = null;
  return fn.apply(thisArg, arguments.length ? arguments : args);
}

/**
 * @param {WrappedFn<Func>} func
 * @returns {Func[]}
 */
export function getWrapChain(func) {
  const chain = [];
  for (; func[WRAPPER]; func = func[WRAPPED]) {
    chain.push(func[WRAPPER]);
  }
  chain.push(func);
  return chain;
}

/**
 * @template {Func} F
 * @param {string} name
 * @param {F} fn
 * @returns {F}
 */
export function namedFunction(name, fn) {
  /** @type {{displayName?: string}} */ (fn).displayName = name;
  return Object.defineProperty(fn, 'name', {value: name});
}

/**
 * @template {Func} F
 * @param {Func} fn
 * @param {F} decorator
 * @returns {F}
 */
export function decoratedFn(fn, decorator) {
  const fname = fn.name;
  const dname = decorator.name;
  /** @type {{[DECORATED]?: Func}} */ (decorator)[DECORATED] = fn;
  if (fname) {
    return namedFunction(dname ? `${dname} ${fname}` : fname, decorator);
  }
  return decorator;
}

/**
 * @param {string} source
 * @returns {string}
 */
function removeCommentsAndWhitespaceFromCode(source) {
  return (
    source
    .replace(/\/\*.*\*\//g, '')
    .replace(/\/\/.*$/mg, '')
    .replace(/\s+/g, ' ')
    .trim()
  );
}

/**
 * @param {Function} func
 * @returns {boolean}
 */
function functionCallsOriginal(func) {
  const source = removeCommentsAndWhitespaceFromCode(func.toString());
  return /^class\b/.test(source) || /\boriginal(Fn)?\s*[(.]/m.test(source);
}
