/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {wrapFn} from './fn_utils.js';
import {first} from './iter.js';
import {$} from './typescript.js';

/**
 * @template {Func} F
 * @param {F} fn
 * @returns {(this: ThisParameterType<F>, ...args: Parameters<F>) => boolean}
 */
export function negate(fn) {
  return function wrapped() {
    return !fn.apply(this, /** @type {any} */ (arguments));
  };
}
/**
 * @template {Func} F
 * @param {F[]} fns
 * @returns {(this: ThisParameterType<F>, ...args: Parameters<F>) => boolean}
 */
export function intersect(...fns) {
  return function wrapped() {
    for (const fn of fns) {
      if (!fn.apply(this, /** @type {any} */ (arguments))) {
        return false;
      }
    }
    return true;
  };
}
/**
 * @template {Func} F
 * @param {F[]} fns
 * @returns {(this: ThisParameterType<F>, ...args: Parameters<F>) => boolean}
 */
export function union(...fns) {
  return function wrapped() {
    for (const fn of fns) {
      if (fn.apply(this, /** @type {any} */ (arguments))) {
        return true;
      }
    }
    return false;
  };
}
/**
 * @template {Func} G
 * @template {(this: ThisParameterType<F>, x: ReturnType<G>) => any} F
 * @param {F} f
 * @param {G} g
 * @returns {(this: ThisParameterType<F>, ...args: Parameters<G>) => ReturnType<F>}
 */
export function compose(f, g) {
  return function wrapped() {
    return f.call(this, g.apply(this, /** @type {any} */ (arguments)));
  };
}

/**
 * @template T
 * @template {any[]} A
 * @template R
 * @template [K = string]
 * @param {Func<T, A, R>} fn
 * @param {object} [options]
 * @param {(args: A) => K} [options.key]
 * @param {number} [options.capacity]
 * @returns {Func<T, A, R>}
 */
export function memo(fn, options) {
  const {
    key: getCacheKey = JSON.stringify,
    capacity = 100,
  } = options || {};
  /** @type {Map<K, R>} */
  const cache = new Map();

  return wrapFn(fn, function(/** @type {A} */ ...args) {
    const cacheKey = /** @type {K} */ (getCacheKey(args));

    // Get cached value
    const cached = cache.get(cacheKey);
    if (cached !== undefined) {
      // Readd it to the cache to extend it's lifetime
      cache.delete(cacheKey);
      cache.set(cacheKey, cached);

      // Return it
      return cached;
    }

    // No cached value, execute the function to get a new value or promise
    const valueOrPromise = /** @type {R} */ (fn.apply(this, args));

    if (cache.size === capacity) {
      // If cache is full, remove oldest cache entry
      cache.delete($(first(cache.keys())));
    }

    // Add value or promise to the cache
    cache.set(cacheKey, valueOrPromise);

    if (
      valueOrPromise
      && typeof /** @type {any} */ (valueOrPromise).then === 'function'
    ) {
      // If it's a thenable...
      Promise.resolve(valueOrPromise).then(
        // ... cache it's value when resolved, ...
        (data) => cache.set(cacheKey, data),
        // ... and remove it from the cache when rejected
        () => cache.delete(cacheKey)
      );
    }

    // Return the value or promise
    return valueOrPromise;
  });
}
