/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {h, isHTML, nodeFromHTML} from './html.js';
import {isApple, isWindows} from './is.js';
import {$} from './typescript.js';
import {asArray} from './utils.js';

/**
 * @typedef HIDKey
 * @prop {number | number[]} [code]
 * @prop {string} short
 * @prop {string} long
 * @prop {string} [html]
 * @prop {string} [_html]
 */

const Keys = {
  backspace: {
    code : 8,
    short: isApple ? '⌫' : 'Backspace',
    long : 'Backspace',
  },
  tab: {
    code : 9,
    short: isApple ? '⇥' : 'Tab',
    long : 'Tab',
  },
  enter: {
    code : 13,
    short: isApple ? '⏎' : 'Enter',
    long : isApple ? 'Return' : 'Enter',
  },
  shift: {
    code : 16,
    short: isApple ? '⇧' : 'Shift',
    long : 'Shift',
  },
  ctrl: {
    code : 17,
    short: isApple ? '⌃' : 'Ctrl',
    long : 'Control',
  },
  alt: {
    code : 18,
    short: isApple ? '⌥' : 'Alt',
    long : isApple ? 'Option' : 'Alt',
  },
  escape: {
    code : 27,
    short: isApple ? '⎋' : 'Esc',
    long : 'Escape',
  },
  space: {
    code : 32,
    short: 'Space',
    long : 'Space',
  },
  pageUp: {
    code : 33,
    short: isApple ? '⇞' : 'PageUp',
    long : 'Page Up',
  },
  pageDown: {
    code : 34,
    short: isApple ? '⇟' : 'PageDown',
    long : 'Page Down',
  },
  end: {
    code : 35,
    short: isApple ? '↘' : 'End',
    long : 'End',
  },
  home: {
    code : 36,
    short: isApple ? '↖' : 'Home',
    long : 'Home',
  },
  left: {
    code : 37,
    short: '◀',
    long : 'Left',
  },
  up: {
    code : 38,
    short: '▲',
    long : 'Up',
  },
  right: {
    code : 39,
    short: '▶',
    long : 'Right',
  },
  down: {
    code : 40,
    short: '▼',
    long : 'Down',
  },
  insert: {
    code : 45,
    short: 'Insert',
    long : 'Insert',
  },
  delete: {
    code : 46,
    short: isApple ? '⌦' : 'Del',
    long : 'Delete',
  },
  meta: {
    code : 91, // [91, 92],
    short: isApple ? '⌘' : (isWindows ? '⊞' : 'Meta'),
    long : isApple ? 'Command' : (isWindows ? 'Windows' : 'Meta'),
  },
  numpadTimes: {
    code : 106,
    short: 'Numpad *',
    long : 'Numpad *',
  },
  numpadPlus: {
    code : 107,
    short: 'Numpad +',
    long : 'Numpad +',
  },
  numpadMinus: {
    code : 109,
    short: 'Numpad -',
    long : 'Numpad -',
  },
  numpadPoint: {
    code : 110,
    short: 'Numpad .',
    long : 'Numpad .',
  },
  numpadDivide: {
    code : 111,
    short: 'Numpad /',
    long : 'Numpad /',
  },
  semicolon: {
    code : 186,
    short: ';',
    long : ';',
  },
  colon: {
    code : 186,
    short: ':',
    long : ':',
  },
  equal: {
    code : 187,
    short: '=',
    long : '=',
  },
  plus: {
    code : 187,
    short: '+',
    long : '+',
  },
  comma: {
    code : 188,
    short: ',',
    long : ',',
  },
  leftAngledBrackets: {
    code : 188,
    short: '<',
    long : '<',
  },
  minus: {
    code : 189,
    short: '-',
    long : '-',
  },
  underscore: {
    code : 189,
    short: '_',
    long : '_',
  },
  period: {
    code : 190,
    short: '.',
    long : '.',
  },
  rightAngledBrackets: {
    code : 190,
    short: '>',
    long : '>',
  },
  slash: {
    code : 191,
    short: '/',
    long : '/',
  },
  questionMark: {
    code : 191,
    short: '?',
    long : '?',
  },
  apostrophe: {
    code : 192,
    short: '`',
    long : '`',
  },
  tilde: {
    code : 192,
    short: 'Tilde',
    long : 'Tilde',
  },
  leftSquareBracket: {
    code : 219,
    short: '[',
    long : '[',
  },
  leftCurlyBracket: {
    code : 219,
    short: '{',
    long : '{',
  },
  backslash: {
    code : 220,
    short: '\\',
    long : '\\',
  },
  verticalLine: {
    code : 220,
    short: '|',
    long : '|',
  },
  rightSquareBracket: {
    code : 221,
    short: ']',
    long : ']',
  },
  rightCurlyBracket: {
    code : 219,
    short: '}',
    long : '}',
  },
  singleQuote: {
    code : 222,
    short: '\'',
    long : '\'',
  },
  doubleQuote: {
    code : 222,
    short: '"',
    long : '"',
  },
};

/** @type {typeof Keys & Record<string, HIDKey>} */
export const HIDKeys = Keys;

for (let i = 0; i < 26; i++) {
  const code = i + 65;
  const char = String.fromCharCode(code);
  HIDKeys[char] = {code, short: char, long: char};
}
for (let i = 0; i < 10; i++) {
  HIDKeys[i] = {code: i + 48, short: `${i}`, long: `${i}`};
  HIDKeys['numpad' + i] = {
    code : i + 96,
    short: `Numpad ${i}`,
    long : `Numpad ${i}`,
  };
}
for (let i = 0; i < 24; i++) {
  HIDKeys[`f${i + 1}`] = {
    code : i + 112,
    short: `F${i + 1}`,
    long : `F${i + 1}`,
  };
}

export const HIDMouse = {
  left: {
    short: 'LC',
    long : 'Left click',
    html : h('span', {class: 'mouse -left'}),
  },
  middle: {
    short: 'MC',
    long : 'Middle click',
    html : h('span', {class: 'mouse -middle'}),
  },
  right: {
    short: 'RC',
    long : 'Right click',
    html : h('span', {class: 'mouse -right'}),
  },
  doubleLeft: {
    short: 'LCx2',
    long : 'Left double-click',
    html : h('span', {class: 'mouse -left'}, 'x2'),
  },
  doubleMiddle: {
    short: 'RCx2',
    long : 'Middle double-click',
    html : h('span', {class: 'mouse -middle'}, 'x2'),
  },
  doubleRight: {
    short: 'RCx2',
    long : 'Right double-click',
    html : h('span', {class: 'mouse -right'}, 'x2'),
  },
  tripleLeft: {
    short: 'LCx3',
    long : 'Left triple-click',
    html : h('span', {class: 'mouse -left'}, 'x3'),
  },
  tripleMiddle: {
    short: 'RCx3',
    long : 'Middle triple-click',
    html : h('span', {class: 'mouse -middle'}, 'x3'),
  },
  tripleRight: {
    short: 'RCx3',
    long : 'Right triple-click',
    html : h('span', {class: 'mouse -right'}, 'x3'),
  },
  leftMove: {
    short: 'LD',
    long : 'Drag with left mouse button',
    html : h('span', {class: 'mouse -left'}, h('i', {class: 'fa fa-arrows'})),
  },
  middleMove: {
    short: 'RD',
    long : 'Drag with middle mouse button',
    html : h('span', {class: 'mouse -middle'}, h('i', {class: 'fa fa-arrows'})),
  },
  rightMove: {
    short: 'RD',
    long : 'Drag with right mouse button',
    html : h('span', {class: 'mouse -right'}, h('i', {class: 'fa fa-arrows'})),
  },
};

export const HIDModifiers = {
  shift     : HIDKeys.shift,
  alt       : HIDKeys.alt,
  ctrl      : HIDKeys.ctrl,
  meta      : HIDKeys.meta,
  ctrlOrMeta: isApple ? HIDKeys.meta : HIDKeys.ctrl,
};

const MODIFIERS = {
  none : 0,
  alt  : 1,
  ctrl : 2,
  meta : 4,
  shift: 8,
};

/**
 * @param {KeyboardEvent | MouseEvent | TouchEvent} event
 * @returns {number}
 */
function modifiersFromEvent(event) {
  return (
    (event.altKey   ? MODIFIERS.alt   : 0) +
    (event.ctrlKey  ? MODIFIERS.ctrl  : 0) +
    (event.metaKey  ? MODIFIERS.meta  : 0) +
    (event.shiftKey ? MODIFIERS.shift : 0)
  );
}
/**
 * @param {HIDKey | HIDKey[] | null} keys
 * @returns {number}
 */
function modifiersFromKeys(keys) {
  let modifiers = MODIFIERS.none;

  if (!keys) {
    return modifiers;
  }

  for (const key of asArray(keys)) {
    switch (key) {
    case HIDModifiers.shift:
      modifiers += MODIFIERS.shift;
      break;
    case HIDModifiers.alt:
      modifiers += MODIFIERS.alt;
      break;
    case HIDModifiers.ctrl:
      modifiers += MODIFIERS.ctrl;
      break;
    case HIDModifiers.meta:
      modifiers += MODIFIERS.meta;
      break;
    default:
    }
  }

  return modifiers;
}
/**
 * @param {number} modifiers
 * @param {number} code
 * @returns {string[]}
 */
function namesOfModifiedKeysWithCode(modifiers, code) {
  /*eslint-disable no-bitwise*/
  switch (code) {
  case HIDModifiers.alt.code:
    modifiers &= ~MODIFIERS.alt;
    break;
  case HIDModifiers.ctrl.code:
    modifiers &= ~MODIFIERS.ctrl;
    break;
  case HIDModifiers.meta.code:
    modifiers &= ~MODIFIERS.meta;
    break;
  case HIDModifiers.shift.code:
    modifiers &= ~MODIFIERS.shift;
    break;
  default:
  }
  /*eslint-enable no-bitwise*/

  const names = namesOfKeysWithCode(code);

  if (modifiers === MODIFIERS.none) {
    return names;
  }

  const stringifiedModifiers = stringifyModifiers(modifiers);
  const modifiedNames = names.map((name) => `${stringifiedModifiers}+${name}`);

  return modifiedNames;
}
/**
 * @param {number} modifiers
 * @returns {string}
 */
function stringifyModifiers(modifiers) {
  /*eslint-disable no-bitwise*/
  return (
    ((modifiers & MODIFIERS.alt)   ? 'a' : '') +
    ((modifiers & MODIFIERS.ctrl)  ? 'c' : '') +
    ((modifiers & MODIFIERS.meta)  ? 'm' : '') +
    ((modifiers & MODIFIERS.shift) ? 's' : '')
  );
  /*eslint-enable no-bitwise*/
}

/**
 * @param {number} code
 * @returns {HIDKey | undefined}
 */
function keyWithCode(code) {
  // @ts-expect-error HIDKeys[undefined] === undefined
  return HIDKeys[namesOfKeysWithCode(code)[0]];
}
/**
 * @param {number} code
 * @returns {string[]}
 */
function namesOfKeysWithCode(code) {
  const names = [];

  for (const k in HIDKeys) {
    if (Object.hasOwn(HIDKeys, k)) {
      const keyCode = $(HIDKeys[k]).code;
      if (
        code === keyCode
        || (Array.isArray(keyCode) && keyCode.indexOf(code))
      ) {
        names.push(k);
      }
    }
  }
  return names;
}

/**
 * @param {HIDKey[]} modifierKeys
 * @param {HIDKey} key
 * @returns {string | null}
 */
export function nameOfModifiedKey(modifierKeys, key) {
  let modifiers = modifiersFromKeys(modifierKeys);

  /*eslint-disable no-bitwise*/
  switch (key) {
  case HIDModifiers.alt:
    modifiers &= ~MODIFIERS.alt;
    break;
  case HIDModifiers.ctrl:
    modifiers &= ~MODIFIERS.ctrl;
    break;
  case HIDModifiers.meta:
    modifiers &= ~MODIFIERS.meta;
    break;
  case HIDModifiers.shift:
    modifiers &= ~MODIFIERS.shift;
    break;
  default:
  }
  /*eslint-enable no-bitwise*/

  const name = nameOfKey(key);

  if (modifiers === MODIFIERS.none) {
    return name;
  }

  if (name) {
    return stringifyModifiers(modifiers) + '+' + name;
  }
  return null;
}
/**
 * @param {HIDKey} key
 * @returns {string | null}
 */
export function nameOfKey(key) {
  for (const k in HIDKeys) {
    if (key === HIDKeys[k]) {
      return k;
    }
  }
  return null;
}

/**
 * @param {HIDKey[]} modifierKeys
 * @param {HIDKey | undefined} key
 * @returns {string}
 */
export function renderModifiedKey(modifierKeys, key) {
  const renderedKeys = renderKeys(modifierKeys);
  renderedKeys.push(renderKey(key));

  return renderedKeys.join(isApple ? '' : ' + ');
}
/**
 * @param {HIDKey[]} modifierKeys
 * @param {HIDKey[]} keys
 * @returns {string[]}
 */
export function renderModifiedKeys(modifierKeys, keys) {
  if (isApple) {
    return (
      Array.isArray(keys)
      ? [keys.map((key) => renderModifiedKey(modifierKeys, key)).join(' / ')]
      : [renderModifiedKey(modifierKeys, keys)]
    );
  }

  const renderedKeys = renderKeys(modifierKeys);
  renderedKeys.push(renderKeys(keys).join(' / '));

  return [renderedKeys.join(' + ')];
}
/**
 * @param {HIDKey[]} keys
 * @returns {string[]}
 */
export function renderKeys(keys) {
  if (!keys) {
    return [];
  }
  if (Array.isArray(keys)) {
    return keys.map(renderKey);
  }
  return [renderKey(keys)];
}
/**
 * @param {string | HIDKey | undefined} key
 * @returns {string}
 */
export function renderKey(key) {
  if (key === undefined) {
    return h('kbd', {class: 'unknown-key', title: 'Unknown key'}, '??');
  }
  if (typeof key === 'string') {
    return isHTML(key) ? key : h('kbd', key);
  }
  if (key._html) {
    return key._html;
  }

  let html = key.html || h('kbd', key.short);

  if (key.long && (key.html || key.long !== key.short)) {
    const node = /** @type {HTMLElement} */ (nodeFromHTML(html));
    node.title = key.long;
    html = node.outerHTML;
  }
  key._html = html;
  return html;
}

/**
 * @param {KeyboardEvent | MouseEvent | TouchEvent} event
 * @returns {boolean}
 */
export function isModifiedEvent(event) {
  return modifiersFromEvent(event) !== MODIFIERS.none;
}
/**
 * @param {KeyboardEvent} event
 * @param {HIDKey} key
 * @returns {boolean}
 */
export function eventMatchesKey(event, key) {
  return event.keyCode === key.code;
}
/**
 * @param {KeyboardEvent} event
 * @param {HIDKey[]} modifierKeys
 * @param {HIDKey} key
 * @returns {boolean}
 */
export function eventMatchesModifiedKey(event, modifierKeys, key) {
  return (
    eventMatchesKey(event, key) &&
    modifiersFromEvent(event) === modifiersFromKeys(modifierKeys)
  );
}
/**
 * @param {KeyboardEvent} event
 * @returns {HIDKey | undefined}
 */
export function keyFromEvent(event) {
  return keyWithCode(event.keyCode);
}
/**
 * @param {KeyboardEvent | MouseEvent | TouchEvent} event
 * @returns {HIDKey[]}
 */
export function modifierKeysFromEvent(event) {
  var modifiers = [];
  if (event.altKey) {
    modifiers.push(HIDModifiers.alt);
  }
  if (event.ctrlKey) {
    modifiers.push(HIDModifiers.ctrl);
  }
  if (event.metaKey) {
    modifiers.push(HIDModifiers.meta);
  }
  if (event.shiftKey) {
    modifiers.push(HIDModifiers.shift);
  }
  return modifiers;
}
/**
 * @param {KeyboardEvent} event
 * @returns {string[]}
 */
export function namesOfModifiedKeysFromEvent(event) {
  return namesOfModifiedKeysWithCode(modifiersFromEvent(event), event.keyCode);
}
/**
 * @param {KeyboardEvent} event
 * @returns {string}
 */
export function renderKeyFromEvent(event) {
  return renderModifiedKey(modifierKeysFromEvent(event), keyFromEvent(event));
}
