/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {namedFunction} from './fn_utils.js';
import {isIterable} from './iter.js';
import {$} from './typescript.js';
import {asArray} from './utils.js';

const XMLNS = 'http://www.w3.org/2000/xmlns/';
const SVG_NS = 'http://www.w3.org/2000/svg';
const selfClosingTags = new Set(
  'area base br col embed hr img input keygen link meta param source track wbr'
  .split(' ')
);
const NS_PREFIX_RE = /^([^:]*):/;

/** @typedef {Record<string | number | symbol, any>} HTMLProps */
/** @typedef {OneOrArray<Exclude<Primitive | Stringifyable, symbol>>} HTMLChildren */

/**
 * @this {unknown}
 * @param {string | null} tag
 * @param {HTMLProps | null} props
 * @param {HTMLChildren | null} children
 * @returns {string | null}
 */
function createHTML(tag, props, children) {
  const childrenHtml = (
    children == null
    ? ''
    : (
      Array.isArray(children)
      ? children.filter(Boolean).join('')
      : children.toString()
    )
  );

  if (!tag) {
    return childrenHtml;
  }

  let attributes = '';

  for (const propName in props) {
    if (Object.hasOwn(props, propName)) {
      let propValue = props[propName];

      if (propValue === true) {
        attributes += ' ' + propName + '="' + propName + '"';
      } else if (propValue != null && propValue !== false) {
        attributes += (
          ' ' + propName + '="' +
          propValue.toString().replace(/"/g, '&quot;') + '"'
        );
      }
    }
  }

  return '<' + tag + attributes + (
    childrenHtml || !selfClosingTags.has(tag)
    ? ('>' + childrenHtml + '</' + tag + '>')
    : ' />'
  );
}

/**
 * @param {string} name
 * @returns {string}
 */
function getNSPrefix(name) {
  return NS_PREFIX_RE.exec(name)?.[1] ?? '';
}
/**
 * @param {Node} ns
 * @param {Element} node
 * @param {string} name
 * @param {string | false | null} value
 * @returns {void}
 */
function setAttribute(ns, node, name, value) {
  const prefix = getNSPrefix(name);
  if (prefix) {
    const nsuri = (prefix === 'xmlns' ? XMLNS : ns.lookupNamespaceURI(prefix));
    if (value == null || value === false) {
      node.removeAttributeNS(nsuri, name);
    } else {
      node.setAttributeNS(nsuri, name, value);
    }
  } else if (value == null || value === false) {
    node.removeAttribute(name);
  } else {
    node.setAttribute(name, value);
  }
}

/** @typedef {HTMLProps & {$parent?: Node}} ElementProps */
/** @typedef {OneOrIterable<Primitive | Node | Stringifyable>} ElementChildren */
/**
 * @this {Document | Node | string | undefined | void}
 * @param {string | Element | DocumentFragment | null} tag
 * @param {ElementProps | null} props
 * @param {ElementChildren | null} children
 * @returns {Element | DocumentFragment}
 */
function createElement(tag, props, children) {
  let doc = (
    this instanceof Document ? this
    : this instanceof Node ? (this.ownerDocument || document)
    : document
  );
  /** @type {Node | undefined} */
  let ns;
  let node;
  if (tag) {
    if (tag instanceof Element || tag instanceof DocumentFragment) {
      node = tag;
    } else {
      ns = props?.$parent || doc;
      const prefix = getNSPrefix(tag);
      const nsuri = (
        this instanceof Element && !prefix ? this.namespaceURI
        : prefix ? ns.lookupNamespaceURI(prefix)
        : typeof this === 'string' ? this
        : undefined
      );
      node = nsuri ? doc.createElementNS(nsuri, tag) : doc.createElement(tag);
    }

    if (props) {
      if (node instanceof DocumentFragment) {
        throw new TypeError(
          'Unexpected props argument when given a DocumentFragment as tag'
        );
      }
      if (!ns) {
        ns = props.$parent || node;
      }
      for (const propName in props) {
        if (propName[0] !== '$' && Object.hasOwn(props, propName)) {
          let propValue = props[propName];

          if (propValue === true) {
            propValue = propName;
          }
          setAttribute(ns, node, propName, propValue);
        }
      }
    }
  } else {
    node = doc.createDocumentFragment();
  }

  if (children instanceof Node) {
    node.appendChild(children);
  } else if (typeof children !== 'string' && isIterable(children)) {
    for (const child of children) {
      if (child instanceof Node) {
        node.appendChild(child);
      } else if (child) {
        if (node instanceof DocumentFragment) {
          throw new TypeError(
            'Unexpected non-Node child when given a DocumentFragment as tag'
          );
        }
        node.insertAdjacentHTML('beforeend', child.toString());
      }
    }
  } else if (children != null) {
    if (node instanceof DocumentFragment) {
      throw new TypeError(
        'Unexpected non-Node child when given a DocumentFragment as tag'
      );
    }
    node.innerHTML = children.toString();
  }

  return node;
}

/**
 * @template {(this: any, tag: any, props: any, children: any) => any} F
 * @param {string} name
 * @param {F} fn
 * @param {ThisParameterType<F>} [thisArg]
 */
function overload(name, fn, thisArg) {
  return namedFunction(name, function(tagOrChildren, propsOrChildren, children) {
    if (thisArg === undefined) {
      thisArg = this;
    }
    if (arguments.length < 1) {
      return fn.call(thisArg, '', null, null);
    }
    if (arguments.length === 1) {
      if (Array.isArray(tagOrChildren)) {
        return fn.call(thisArg, '', null, tagOrChildren);
      }
      return fn.call(thisArg, tagOrChildren, null, null);
    }
    const tag = tagOrChildren;
    if (arguments.length === 2) {
      if (
        Array.isArray(propsOrChildren) ||
        propsOrChildren instanceof Node ||
        typeof propsOrChildren !== 'object'
      ) {
        return fn.call(thisArg, tag, null, propsOrChildren);
      }
      return fn.call(thisArg, tag, propsOrChildren, null);
    }
    return fn.call(thisArg, tag, propsOrChildren, children);
  });
}

/**
 * @type {{
 *   (this: unknown, tag?: null): null;
 *   (this: unknown, children: HTMLChildren & any[]): string;
 *   (this: unknown, tag: string, children?: HTMLChildren): string;
 *   (this: unknown, tag: string, props: HTMLProps | null, children?: HTMLChildren): string;
 * }}
 */
export const h = overload('h', createHTML);

/**
 * @type {{
 *   (this: Document | Node | string | undefined | void, tag?: null): DocumentFragment;
 *   (this: Document | Node | string | undefined | void, children: OneOrIterable<Node>): DocumentFragment;
 *   (
 *     this: Document | Node | string | undefined | void,
 *     tag: string | Element,
 *     children?: ElementChildren
 *   ): HTMLElement;
 *   (
 *     this: Document | Node | string | undefined | void,
 *     tag: DocumentFragment,
 *     children?: OneOrIterable<Node>
 *   ): DocumentFragment;
 *   (
 *     this: Document | Node | string | undefined | void,
 *     tag: string | Element,
 *     props: ElementProps | null,
 *     children?: ElementChildren
 *   ): HTMLElement;
 * }}
 */
export const H = overload('H', createElement);

/**
 * @type {{
 *   (this: void, tag?: null): DocumentFragment;
 *   (this: void, children: OneOrIterable<Node>): DocumentFragment;
 *   (
 *     this: void,
 *     tag: string | Element,
 *     children?: ElementChildren
 *   ): SVGElement;
*   (
*     this: Document | Node | string | undefined | void,
*     tag: DocumentFragment,
*     children?: OneOrIterable<Node>
*   ): DocumentFragment;
 *   (
 *     this: void,
 *     tag: string | Element,
 *     props: ElementProps | null,
 *     children?: ElementChildren
 *   ): SVGElement;
 * }}
 */
export const S = overload('S', createElement, SVG_NS);

/**
 * @param {string[]} strings
 * @param {any[]} substitutions
 * @returns {HTMLStyleElement}
 */
export function css(strings, ...substitutions) {
  return /** @type {HTMLStyleElement} */ (
    H('style', strings && String.raw({raw: asArray(strings)}, ...substitutions))
  );
}

const parser = new DOMParser();
const CACHE_HTML_MAX_LENGTH = 100;
const parsingCache = new Map();

/**
 * @param {HTML} html
 * @returns {{html: HTML, text: string}}
 */
export function normalizeHTML(html) {
  const cache = parsingCache.get(html);
  if (cache?.html) {
    return cache;
  }

  const originalHtml = html;
  const doc = parser.parseFromString(html || '', 'text/html');
  html = doc.body.innerHTML;
  const text = doc.body.textContent || '';
  const parsed = Object.freeze({html, text});

  if (html.length < CACHE_HTML_MAX_LENGTH) {
    parsingCache.set(originalHtml, parsed);
  }
  return parsed;
}
/**
 * @param {HTML} html
 * @returns {string}
 */
export function textFromHTML(html) {
  const cache = parsingCache.get(html);
  if (cache) {
    return cache.text;
  }

  const doc = parser.parseFromString(html || '', 'text/html');
  const text = doc.body.textContent || '';
  const parsed = Object.freeze({text});

  if (html.length < CACHE_HTML_MAX_LENGTH) {
    parsingCache.set(html, parsed);
  }
  return parsed.text;
}
/**
 * @param {HTML} html
 * @returns {Element | null}
 */
export function nodeFromHTML(html) {
  const doc = parser.parseFromString(html || '', 'text/html');
  return doc.body.firstElementChild;
}
/**
 * @param {HTML} html
 * @returns {Element[]}
 */
export function nodesFromHTML(html) {
  const doc = parser.parseFromString(html || '', 'text/html');
  return [...doc.body.children];
}
/**
 * @param {string} string
 * @returns {boolean}
 */
export function isHTML(string) {
  return textFromHTML(string) !== string;
}

const _node = document.createElement('div');

/**
 * @param {string} str
 * @returns {string}
 */
export function camelCaseToDataAttribute(str) {
  _node.dataset[str] = '';
  const dataAttribute = _node.getAttributeNames()[0];
  delete _node.dataset[str];
  return $(dataAttribute);
}
/**
 * @param {Record<string | number, string | number>} object
 * @returns {Record<string, string | number>}
 */
export function objectToDataAttributes(object) {
  /** @type {Record<string, string | number>} */
  const dataAttributes = {};
  for (const property in object) {
    if (Object.hasOwn(object, property)) {
      dataAttributes[camelCaseToDataAttribute(property)] = $(object[property]);
    }
  }
  return dataAttributes;
}
