/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  emptyFunctionReturningFalse,
} from './empty_function.js';
import {getNowMs, scheduleTask} from './scheduling.js';

const ITERATOR_DONE = /** @type {IteratorReturnResult<any>} */ (
  Object.freeze({done: true})
);

/** @type {IterableIterator<never>} */
export const EMPTY_ITERATOR = Object.freeze({
  next: () => ITERATOR_DONE,
  [Symbol.iterator]: () => EMPTY_ITERATOR,
});

/**
 * @param {unknown} object
 * @returns {object is Iterable<unknown>}
 */
export function isIterable(object) {
  return typeof Object(object)[Symbol.iterator] === 'function';
}

/**
 * @template T
 * @param {Iterable<T> | T} iterableOrElement
 * @returns {Iterable<T>}
 */
export function asIterable(iterableOrElement) {
  return (
    isIterable(iterableOrElement)
    ? iterableOrElement
    : [iterableOrElement]
  );
}

/**
 * @template T
 * @param {Iterable<T>} iterable
 * @returns {T | undefined}
 */
export function first(iterable) {
  return iterable[Symbol.iterator]().next().value;
}

/**
 * @overload
 * @param {number} stop
 * @param {number} [step]
 * @returns {Generator<number, void>}
 *
 * @overload
 * @param {number} start
 * @param {number} stop
 * @param {number} [step]
 * @returns {Generator<number, void>}
 *
 * @param {number} start
 * @param {number | undefined} stop
 * @param {number} [step]
 * @returns {Generator<number, void>}
 */
export function* range(start, stop, step) {
  if (stop === undefined) {
    stop = start;
    start = 0;
  }
  if (step === undefined) {
    step = Math.sign(stop - start);
  }
  if (start <= stop) {
    for (; start < stop; start += step) {
      yield start;
    }
  } else {
    for (; start > stop; start += step) {
      yield start;
    }
  }
}

/** @type {() => boolean} */
const isInputPending = (
  (
    // @ts-expect-error
    // eslint-disable-next-line compat/compat
    navigator.scheduling?.isInputPending?.bind(navigator.scheduling, {includeContinuous: true})
  ) ?? emptyFunctionReturningFalse
);

/**
 * @template T
 * @typedef {Iterable<T> & Pick<AsyncIterOptions, 'quantum' | 'schedule'>} AsyncIterIterable
 */
/**
 * @typedef AsyncIterOptions
 * @prop {number | Readonly<OneOrMore<number>>} [quantum = 5]
 * @prop {import('./scheduling.js').Scheduler<() => void>} [schedule = scheduleTask]
 * @prop {AbortSignal} [signal]
 */
/**
 * @template T
 * @param {Iterable<T>} iterable
 * @param {Readonly<AsyncIterOptions>} [options={}]
 * @return {IterableIterator<Promise<AsyncIterIterable<T>>>}
 */
export function asyncIter(
  iterable,
  {quantum: quanta = 5, schedule = scheduleTask, signal} = {}
) {
  if (signal?.aborted) {
    return EMPTY_ITERATOR;
  }

  let waitTime = 0;

  const iterator = iterable[Symbol.iterator]();
  let iteratorDone = false;

  /** @type {AsyncIterIterable<T>} */
  const quantumIterator = {
    // An iterator object is faster than a generator
    // see https://jsperf.app/tozuse/1/preview
    [Symbol.iterator]() {
      const quantum = Math.max(
        waitTime,
        (
          Array.isArray(quanta)
          ? (quanta.length > 1 ? quanta.shift() : quanta[0])
          : quanta
        )
      );

      const deadline = getNowMs() + quantum;
      return {
        next() {
          if (signal?.aborted || isInputPending() || getNowMs() > deadline) {
            return ITERATOR_DONE;
          }

          const result = iterator.next();
          if (result.done) {
            iteratorDone = true;
          }
          return result;
        },
        [Symbol.iterator]() {
          return this;
        },
      };
    },
    get quantum() {
      return quanta;
    },
    set quantum(q) {
      quanta = q;
    },
    get schedule() {
      return schedule;
    },
    set schedule(s) {
      schedule = s;
    },
  };

  // Iteration of this iterator is slow (each iteration will wait for the next
  // quantum) so code simplicity is more important than its speed.
  return (function*() {
    yield Promise.resolve(quantumIterator);

    let waitStart = 0;

    // eslint-disable-next-line no-unmodified-loop-condition
    while (!signal?.aborted && !iteratorDone) {
      // still work to be done and quantum used up
      // => schedule a new quantum for later
      waitStart = getNowMs();
      yield (
        schedule(emptyFunction, signal)
        .catch(emptyFunction)
        .then(continueIteration)
      );
    }

    function continueIteration() {
      waitTime = getNowMs() - waitStart;
      return quantumIterator;
    }
  })();
}
