/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {PASSIVE_CAPTURING, NONPASSIVE} from './dom_utils.js';
import {EventsHub} from './events_hub.js';
import {HIDKeys as Keys, keyFromEvent} from './hid.js';

export class KeyboardManager {
  // #region Private typedefs

  /**
   * @typedef {import('./hid.js').HIDKey} HIDKey
   */

  // #endregion
  // #region Public

  /**
   * @public
   */
  constructor() {
    /** @type {Set<HIDKey>} */
    this.keysPressed = new Set();

    this.hub = new EventsHub();

    this.handleBlur = this.handleBlur.bind(this);
    this.handleEventWithModifiers = this.handleEventWithModifiers.bind(this);
    this.handleKeydown = this.handleKeydown.bind(this);
    this.handleKeyup = this.handleKeyup.bind(this);

    window.addEventListener('blur', this.handleBlur, PASSIVE_CAPTURING);
    window.addEventListener('mousemove', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('mouseup', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('mousedown', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('wheel', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('touchstart', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('touchend', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('touchmove', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('touchcancel', this.handleEventWithModifiers, PASSIVE_CAPTURING);
    window.addEventListener('keydown', this.handleKeydown, NONPASSIVE);
    window.addEventListener('keyup', this.handleKeyup, NONPASSIVE);
  }

  /**
   * @public
   * @returns {boolean}
   */
  isShiftDown() {
    return this.isKeyDown(Keys.shift);
  }
  /**
   * @public
   * @param {HIDKey} key
   * @returns {boolean}
   */
  isKeyDown(key) {
    return this.keysPressed.has(key);
  }

  // #endregion
  // #region Event handlers

  /**
   * @protected
   * @param {FocusEvent} event
   * @returns {void}
   */
  handleBlur(event) {
    if (event.target === window) {
      this.keysPressed.clear();
      this.hub.emit('change:up');
      this.hub.emit('change');
    }
  }
  /**
   * @protected
   * @param {FocusEvent | KeyboardEvent | MouseEvent | TouchEvent} event
   * @returns {void}
   */
  handleEventWithModifiers(event) {
    this.updateModifiersFromEvent(event);
  }
  /**
   * @protected
   * @param {KeyboardEvent} event
   * @returns {void}
   */
  handleKeydown(event) {
    let changed = (
      this.updateModifiersFromEvent(event, true /*withoutChangeEvent*/)
    );

    changed = (
      this._updateKey(keyFromEvent(event), true, event, true /*force*/)
      || changed
    );
    if (changed) {
      this.hub.emit('change', event);
    }
  }
  /**
   * @protected
   * @param {KeyboardEvent} event
   * @returns {void}
   */
  handleKeyup(event) {
    let changed = (
      this.updateModifiersFromEvent(event, true /*withoutChangeEvent*/)
    );

    changed = this._updateKey(keyFromEvent(event), false, event) || changed;
    if (changed) {
      this.hub.emit('change', event);
    }
  }

  // #endregion
  // #region Protected

  /**
   * @protected
   * @param {HIDKey | undefined} key
   * @param {boolean} state
   * @param {Event} event
   * @param {boolean} [force]
   * @returns {boolean}
   */
  _updateKey(key, state, event, force = false) {
    if (!key) {
      return false;
    }
    if (force || state !== this.keysPressed.has(key)) {
      if (state) {
        this.keysPressed.add(key);
      } else {
        this.keysPressed.delete(key);
      }
      this.hub.emit('change:' + (state ? 'down' : 'up'), event, key);
      return true;
    }
    return false;
  }
  /**
   * @protected
   * @param {FocusEvent | KeyboardEvent | MouseEvent | TouchEvent} event
   * @param {boolean} [withoutChangeEvent]
   * @returns {boolean}
   */
  updateModifiersFromEvent(event, withoutChangeEvent = false) {
    let changed = false;

    if ('shiftKey' in event) {
      changed = this._updateKey(Keys.shift, event.shiftKey, event) || changed;
    }
    if ('ctrlKey' in event) {
      changed = this._updateKey(Keys.ctrl, event.ctrlKey, event) || changed;
    }
    if ('altKey' in event) {
      changed = this._updateKey(Keys.alt, event.altKey, event) || changed;
    }
    if ('metaKey' in event) {
      changed = this._updateKey(Keys.meta, event.metaKey, event) || changed;
    }

    if (changed && !withoutChangeEvent) {
      this.hub.emit('change', event);
    }

    return changed;
  }

  // #endregion
}
