/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$} from './typescript.js';

const ε = Number.EPSILON;
export const MAX_IDL_WRAPS = 100;
export const DEG_IN_RAD = Math.PI / 180;
export const RAD_IN_DEG = 180 / Math.PI;

/**
 * Limits a number to min and max values.
 *
 * @template {number | Date} V
 * @param {V} value
 * @param {V} min
 * @param {V} max
 * @returns {V} Returns `value` if it's in the range [`min`, `max`], otherwise `min` or `max`.
 */
export function clamp(value, min, max) {
  if (value < min) {
    return min;
  }
  if (value > max) {
    return max;
  }
  return value;
}
/**
 * @param {number} number
 * @param {number} min
 * @param {number} max
 * @returns {number}
 */
export function wrap(number, min, max) {
  return modulo(number - min, max - min) + min;
}
/**
 * @param {number} number
 * @param {number} center
 * @param {number} width
 * @returns {number}
 */
export function wrapAround(number, center, width) {
  const offset = center - width / 2;
  return modulo(number - offset, width) + offset;
}
/**
 * @param {number} left
 * @param {number} right
 * @param {number} worldCenter
 * @param {number} worldWidth
 * @returns {[leftWorldsAway: number, rightWorldsAway: number]}
 */
export function idlRangeWorldsAway(left, right, worldCenter, worldWidth) {
  return [
    Math.round((left - worldCenter) / worldWidth + ε) | 0,
    Math.round((right - worldCenter) / worldWidth - ε) | 0,
  ];
}
/**
 * @param {number} leftWorldsAway
 * @param {number} rightWorldsAway
 * @param {number} [max]
 * @returns {number}
 */
export function idlRangeWrapCount(leftWorldsAway, rightWorldsAway, max = MAX_IDL_WRAPS) {
  const wrapCount = rightWorldsAway - leftWorldsAway;
  if (wrapCount > max) {
    console.warn('Limiting the wrap count to avoid blocking for a long time');
    return max;
  }
  return wrapCount;
}

/**
 * @param {number} value
 * @param {number} exp
 * @returns {number}
 */
export function round(value, exp) {
  var scale = parseFloat('1e' + (-exp));
  return Math.round(value * scale) / scale;
}
/**
 * @param {number} value
 * @param {number} exp
 * @returns {number}
 */
export function floor(value, exp) {
  var scale = parseFloat('1e' + (-exp));
  return Math.floor(value * scale) / scale;
}
/**
 * @param {number} value
 * @param {number} exp
 * @returns {number}
 */
export function ceil(value, exp) {
  var scale = parseFloat('1e' + (-exp));
  return Math.ceil(value * scale) / scale;
}

/**
 * Decimal adjustment of a number.
 * From https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/round
 * MDN terms states: "Code samples added on or after August 20, 2010 are in the public domain."
 *
 * @param {'round' | 'floor' | 'ceil'} type  The type of adjustment.
 * @param {number} value The number.
 * @param {number} exp   The exponent (the 10 logarithm of the adjustment base).
 * @returns {number} The adjusted value.
 */
function decimalAdjust(type, value, exp) {
  /*eslint-disable no-implicit-coercion*/
  // If the exp is undefined or zero...
  if (typeof exp === 'undefined' || +exp === 0) {
    return Math[type](value);
  }
  value = +value;
  exp = +exp;
  // If the value is not a number or the exp is not an integer...
  if (isNaN(value) || !(typeof exp === 'number' && exp % 1 === 0)) {
    return NaN;
  }
  // If the value is negative...
  if (value < 0) {
    return -decimalAdjust(type, -value, exp);
  }
  // Shift
  let split = /** @type {[string, string?]} */ (value.toString().split('e'));
  value = Math[type](+(split[0] + 'e' + (split[1] ? (+split[1] - exp) : -exp)));
  // Shift back
  split = /** @type {[string, string?]} */ (value.toString().split('e'));
  return +(split[0] + 'e' + (split[1] ? (+split[1] + exp) : exp));
  /*eslint-enable no-implicit-coercion*/
}

/**
 * Decimal round
 * @param {number} value
 * @param {number} exp
 * @returns {number}
 */
export function round10(value, exp) {
  return decimalAdjust('round', value, exp);
}
/**
 * Decimal floor
 * @param {number} value
 * @param {number} exp
 * @returns {number}
 */
export function floor10(value, exp) {
  return decimalAdjust('floor', value, exp);
}
/**
 * Decimal ceil
 * @param {number} value
 * @param {number} exp
 * @returns {number}
 */
export function ceil10(value, exp) {
  return decimalAdjust('ceil', value, exp);
}

/**
 * @param {number} exp
 * @param {number} value
 * @returns {string}
 */
export function toFixed(exp, value) {
  if (exp > 100) {
    return '0';
  }
  if (exp < -100) {
    return value.toString();
  }
  if (exp > 0) {
    return round(value, exp).toFixed(0);
  }
  return value.toFixed(-exp);
}

/**
 * Calculates the modulo of the given number.
 *
 * @param {number} number
 * @param {number} mod
 * @returns {number} Returns the mathematical modulo of `number`, always in [0, `module`[.
 */
export function modulo(number, mod) {
  // This version is slightly faster but appears to be more prone to floating
  // point errors.
  // return ((number % mod) + mod) % mod;

  // This version is slightly slower but appears to be less prone to floating
  // point errors.
  return number - Math.floor(number / mod) * mod;
}

/**
 * @param {number | null} a
 * @param {number | null} b
 * @param {number | null} t
 * @returns {number}
 */
export function lerp(a, b, t) {
  if (a === null || b === null || t === null) {
    return NaN;
  }
  if (t === 0) {
    return a;
  }
  if (t === 1) {
    return b;
  }
  return (b - a) * t + a;
}
/**
 * @param {number} a
 * @param {number} c0
 * @param {number} b
 * @param {number} t
 * @returns {number}
 */
export function lerp2(a, c0, b, t) {
  if (t === 0) {
    return a;
  }
  if (t === 1) {
    return b;
  }
  return lerp(lerp(a, c0, t), lerp(c0, b, t), t);
}
/**
 * @param {number} a
 * @param {number} c0
 * @param {number} c1
 * @param {number} b
 * @param {number} t
 * @returns {number}
 */
export function lerp3(a, c0, c1, b, t) {
  if (t === 0) {
    return a;
  }
  if (t === 1) {
    return b;
  }
  return lerp2(lerp(a, c0, t), lerp(c0, c1, t), lerp(c1, b, t), t);
}
/**
 * @param {NumberArray} points
 * @param {number} t
 * @returns {number}
 */
export function lerpN(points, t) {
  var n = points.length;

  if (n === 0) {
    return 0;
  }
  if (t === 0) {
    return $(points[0]);
  }
  if (t === 1) {
    return $(points[n - 1]);
  }

  switch (n) {
  case 1:
    return $(points[0]);
  case 2:
    return lerp($(points[0]), $(points[1]), t);
  case 3:
    return lerp2($(points[0]), $(points[1]), $(points[2]), t);
  case 4:
    return lerp3($(points[0]), $(points[1]), $(points[2]), $(points[3]), t);
  default:
    points = points.slice();
    while (--n) {
      for (let i = 0; i < n; i++) {
        points[i] = $(points[i]) + ($(points[i + 1]) - $(points[i])) * t;
      }
    }
    return $(points[0]);
  }
}

/**
 * @param {number | null} a
 * @param {number | null} b
 * @param {number | null} x
 * @returns {number}
 */
export function inverseLerp(a, b, x) {
  if (a === null || b === null || x === null) {
    return NaN;
  }
  return (x - a) / (b - a);
}

/**
 * @param {XY} p1
 * @param {XY} p2
 * @returns {XY}
 */
export function getXYMovement(p1, p2) {
  return [p2[0] - p1[0], p2[1] - p1[1]];
}

/**
 * @param {XY} p1
 * @param {XY} p2
 * @returns {number}
 */
export function getXYDistance(p1, p2) {
  return Math.hypot(p2[0] - p1[0], p2[1] - p1[1]);
}

/**
 * @param {XY} xy
 * @returns {number}
 */
export function getXYLength(xy) {
  return Math.hypot(xy[0], xy[1]);
}

/**
 * @param {XY | null | undefined} p1
 * @param {XY | null | undefined} p2
 * @returns {boolean}
 */
export function isXYEqual(p1, p2) {
  return p1 === p2 || (!!p1 && !!p2 && p1[0] === p2[0] && p1[1] === p2[1]);
}

/**
 * @param {XY[] | null | undefined} xys1
 * @param {XY[] | null | undefined} xys2
 * @returns {boolean}
 */
export function areXYsEqual(xys1, xys2) {
  if (xys1 === xys2) {
    return true;
  }
  if (!xys1 || !xys2) {
    return false;
  }

  const l = xys1.length;
  if (l !== xys2.length) {
    return false;
  }

  for (let i = 0; i < l; i++) {
    if (!isXYEqual(xys1[i], xys2[i])) {
      return false;
    }
  }
  return true;
}
