/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  emptyFunctionReturningFalse,
  emptyFunctionReturningTrue,
} from './empty_function.js';
import {scheduleDelayedTask} from './scheduling.js';
import {isAbortError, anyAbortSignal} from './utils.js';

/**
 * @param {PromiseSettledResult<unknown>} result
 * @returns {result is PromiseFulfilledResult<unknown>}
 */
export function isSettledPromiseFulfilled(result) {
  return result.status === 'fulfilled';
}
/**
 * @param {PromiseSettledResult<unknown>} result
 * @returns {result is PromiseRejectedResult}
 */
export function isSettledPromiseRejected(result) {
  return result.status === 'rejected';
}
/**
 * @template T
 * @param {PromiseSettledResult<T>} result
 * @returns {T | undefined}
 */
export function getSettledPromiseValue(result) {
  return /** @type {{value?: T}} */ (result).value;
}
/**
 * @param {PromiseSettledResult<unknown>} result
 * @returns {any}
 */
export function getSettledPromiseReason(result) {
  return /** @type {PromiseRejectedResult} */ (result).reason;
}

/**
 * @param {Milliseconds} ms
 * @param {AbortSignal} [signal]
 * @returns {Promise<void>}
 */
export function delay(ms, signal) {
  return scheduleDelayedTask(ms, emptyFunction, signal);
}

const RETRY_ERROR = Symbol('RETRY_ERROR');

/**
 * @template T
 * @typedef RetriableOptions
 * @prop {boolean} [abortsPreviousCall = false]
 * @prop {number} [maxRetries = 3]
 * @prop {Milliseconds} [initialDelayMs = 1000]
 * @prop {number} [backoffMultiplier = 2]
 * @prop {boolean} [withExtraArgument = false]
 * @prop {(value: T, retries: number) => boolean} [shouldRetryAfterValue = () => false]
 * @prop {(error: unknown, retries: number) => boolean} [shouldRetryAfterError = () => true]
 */
/**
 * @typedef RetriableExtraArgument
 * @prop {AbortSignal} signal
 * @prop {() => boolean} canRetry
 * @prop {() => void} throwIfRetriable
 * @prop {() => void} abort
 */
/**
 * @template {unknown[]} P
 * @typedef {P[0] extends RetriableExtraArgument ? OmitFirstArrayElement<P> : P} _OmitExtraArgument
 */
/**
 * @template {unknown[]} P
 * @typedef {P extends [...unknown[], AbortSignal] ? P : [...args: P, signal?: AbortSignal]} _AddAbortSignal
 */
/**
 * @template {(...args: any[]) => any} F
 * @typedef {_AddAbortSignal<_OmitExtraArgument<Parameters<F>>>} _Parameters
 */
/**
 * @template {(...args: any[]) => any} F
 * @typedef {(this: ThisParameterType<F>, ...args: _Parameters<F>) => Promise<Awaited<ReturnType<F>>>} Retriable
 */

/**
 * @template {(...args: any[]) => any} F
 * @param {F} fn
 * @param {RetriableOptions<Awaited<ReturnType<F>>>} [options]
 * @returns {Retriable<F>}
 */
export function retriable(fn, {
  abortsPreviousCall = false,
  maxRetries = 3,
  initialDelayMs = 1000,
  backoffMultiplier = 2,
  withExtraArgument = false,
  shouldRetryAfterValue = emptyFunctionReturningFalse,
  shouldRetryAfterError = emptyFunctionReturningTrue,
} = {}) {
  /** @type {AbortController | null} */
  let lastAbortController = null;

  return async function(...args) {
    let retries = 0;
    let delayMs = initialDelayMs;
    let signal;

    if (args[args.length - 1] instanceof AbortSignal) {
      signal = args[args.length - 1];
    }

    if (withExtraArgument || abortsPreviousCall) {
      const abortController = new AbortController();
      if (signal) {
        signal = anyAbortSignal([signal, abortController.signal]);
        args[args.length - 1] = signal;
      } else {
        signal = abortController.signal;
      }

      if (withExtraArgument) {
        args.unshift({
          signal,
          canRetry() {
            return retries < maxRetries;
          },
          throwIfRetriable() {
            if (retries < maxRetries) {
              throw RETRY_ERROR;
            }
          },
          abort() {
            abortController.abort();
          },
        });
      }
      if (abortsPreviousCall) {
        lastAbortController?.abort();
        lastAbortController = abortController;
      }
    }

    do {
      try {
        // eslint-disable-next-line no-await-in-loop
        const value = await fn.apply(this, args);
        const shouldRetry = (
          retries < maxRetries && shouldRetryAfterValue(value, retries)
        );
        if (!shouldRetry) {
          return value;
        }
      } catch (/** @type {any} */ error) {
        const shouldRetry = (
          retries < maxRetries &&
          !isAbortError(error) &&
          (error === RETRY_ERROR || shouldRetryAfterError(error, retries))
        );
        if (!shouldRetry) {
          throw error;
        }
      }

      // Retry
      retries++;
      delayMs *= backoffMultiplier;
      // eslint-disable-next-line no-await-in-loop
      await delay(delayMs, signal);
    } while (true);
  };
}
