/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/// <reference path='./scheduling_.d.ts' />
import {decoratedFn} from './fn_utils.js';
import {isAbortError} from './utils.js';

/** @typedef {Named<number, 'ImmediateId'>} ImmediateId */

let setImmediateId = 1;
const {
  port1: setImmediateReceivePort,
  port2: setImmediateSendPort,
} = new MessageChannel();
const setImmediateCallbacks = new Map();

/** @param {MessageEvent<number>} event */
setImmediateReceivePort.onmessage = function(event) {
  const id = event.data;
  const callback = setImmediateCallbacks.get(id);
  if (callback) {
    setImmediateCallbacks.delete(id);
    callback[0](...callback[1]);
  }
};

/** @type {() => Milliseconds} */
export const getNowMs = (
  window.performance
  ? window.performance.now.bind(window.performance)
  : Date.now.bind(Date)
);

/**
 * @template {(...args: any[]) => any} F
 * @param {Milliseconds} ms
 * @param {F} fn
 * @param {Parameters<F>} params
 * @returns {TimeoutId}
 */
export function setTimeoutWithMs(ms, fn, ...params) {
  return setTimeout(fn, ms, ...params);
}

/**
 * @template {(...args: any[]) => any} F
 * @param {F} fn
 * @param {Parameters<F>} params
 * @returns {ImmediateId}
 */
export function setImmediate(fn, ...params) {
  setImmediateCallbacks.set(setImmediateId, [fn, params]);
  setImmediateSendPort.postMessage(setImmediateId);
  return setImmediateId++;
}

/**
 * @param {ImmediateId} id
 * @returns {void}
 */
export function clearImmediate(id) {
  setImmediateCallbacks.delete(id);
}

/**
 * @template {(...args: any[]) => any} F
 * @param {(value: ReturnType<F>) => void} resolve
 * @param {(error: any) => void} reject
 * @param {F} fn
 * @param {Parameters<F>} args
 * @returns {void}
 */
function tryCall(resolve, reject, fn, args) {
  try {
    resolve(fn.apply(undefined, args));
  } catch (error) {
    reject(error);
  }
}

/**
 * @template {(...args: any[]) => any} F
 * @typedef {(fn: F, signal?: ?AbortSignal) => Promise<ReturnType<F>>} Scheduler
 */

/**
 * @template {(...args: any[]) => any} F
 * @template {(cb: (...args: Parameters<F>) => void, ...requestArgs: any) => void} R
 * @param {R} request
 * @param {OmitFirstArrayElement<Parameters<R>>} requestArgs
 * @returns {Scheduler<F>}
 */
export function createCallbackScheduler(request, ...requestArgs) {
  return (fn, signal) => new Promise((resolve, reject) => {
    if (signal?.aborted) {
      reject(signal.reason);
    } else {
      request((...args) => {
        if (signal?.aborted) {
          reject(signal.reason);
        } else {
          tryCall(resolve, reject, fn, args);
        }
      }, ...requestArgs);
    }
  });
}

/**
 * @template {(...args: any[]) => any} F
 * @template {(cb: (...args: Parameters<F>) => void, ...requestArgs: any) => any} R
 * @param {R} request
 * @param {(id: ReturnType<R>) => void} cancel
 * @param {OmitFirstArrayElement<Parameters<R>>} requestArgs
 * @returns {Scheduler<F>}
 */
export function createRequestScheduler(request, cancel, ...requestArgs) {
  return (fn, signal) => new Promise((resolve, reject) => {
    if (signal?.aborted) {
      reject(signal.reason);
      return;
    }

    signal?.addEventListener('abort', abortHandler);

    const id = request((...args) => {
      signal?.removeEventListener('abort', abortHandler);
      tryCall(resolve, reject, fn, args);
    }, ...requestArgs);

    function abortHandler() {
      // @ts-expect-error -- signal is non-null since the event was triggered
      signal.removeEventListener('abort', abortHandler);
      cancel(id);
      // @ts-expect-error -- signal is non-null since the event was triggered
      reject(signal.reason);
    }
  });
}

export const scheduleMicrotask = createCallbackScheduler(queueMicrotask);
export const scheduleTask = createRequestScheduler(setImmediate, clearImmediate);

/**
 * @template {(...args: any[]) => any} F
 *
 * @overload
 * @param {Milliseconds} ms
 * @returns {Scheduler<F>}
 *
 * @overload
 * @param {Milliseconds} ms
 * @param {Parameters<Scheduler<F>>[0]} fn
 * @param {Parameters<Scheduler<F>>[1]} signal
 * @returns {ReturnType<Scheduler<F>>}
 *
 * @param {Milliseconds} ms
 * @param {Parameters<Scheduler<F>>[0]} [fn]
 * @param {Parameters<Scheduler<F>>[1]} [signal]
 * @returns {Scheduler<F> | ReturnType<Scheduler<F>>}
 */
export function scheduleDelayedTask(ms, fn, signal) {
  const scheduler = createRequestScheduler(setTimeout, clearTimeout, ms);
  return fn ? scheduler(fn, signal) : scheduler;
}

export const scheduleAnimationFrame = (
  createRequestScheduler(requestAnimationFrame, cancelAnimationFrame)
);

/**
 * @param {FrameRequestCallback} fn
 * @param {AbortSignal} [signal]
 * @returns {Promise<ReturnType<void>>}
 */
export function scheduleNextAnimationFrame(fn, signal) {
  return scheduleAnimationFrame(
    () => scheduleAnimationFrame(fn, signal),
    signal
  );
}

/// requestIdleCallback and cancelIdleCallback are not available in Safari which
/// causes an error to be thrown.
// /**
//  * @template {(...args: any[]) => any} F
//  *
//  * @overload
//  * @param {IdleRequestOptions} [options]
//  * @returns {Scheduler<F>}
//  *
//  * @overload
//  * @param {IdleRequestOptions} [options]
//  * @param {Parameters<Scheduler<F>>[0]} fn
//  * @param {Parameters<Scheduler<F>>[1]} signal
//  * @returns {ReturnType<Scheduler<F>>}
//  *
//  * @param {IdleRequestOptions} [options]
//  * @param {Parameters<Scheduler<F>>[0]} [fn]
//  * @param {Parameters<Scheduler<F>>[1]} [signal]
//  * @returns {Scheduler<F> | ReturnType<Scheduler<F>>}
//  */
// export function scheduleIdleWithOptions(options, fn, signal) {
//   const scheduler = (
//     createRequestScheduler(requestIdleCallback, cancelIdleCallback, options)
//   );
//   return fn ? scheduler(fn, signal) : scheduler;
// }
// export const scheduleIdle = scheduleIdleWithOptions();

/**
 * @template T
 * @param {any} prevArgs
 * @param {T} args
 * @returns {T}
 */
function keepLastCallArgs(prevArgs, args) {
  return args;
}

/**
 * @template {(...args: any[]) => any} F
 * @callback ArgsReducer
 * @param {?Parameters<F> | undefined} prevArgs
 * @param {Parameters<F>} newArgs
 * @returns {?Parameters<F> | undefined}
 */
/**
 * @template {(...args: any[]) => any} F
 * @typedef DebounceOptions
 * @prop {ArgsReducer<F>} [argsReducer = keepLastCallArgs]
 * @prop {?Parameters<F>} [initialArgs]
 */

/**
 * @template {(...args: any[]) => any} F
 * @param {Milliseconds} ms
 * @param {F} fn
 * @param {DebounceOptions<F>} [options]
 * @returns {DebouncedFunction<F>}
 */
export function debounce(
  ms,
  fn,
  {argsReducer = keepLastCallArgs, initialArgs} = {}
) {
  /** @type {TimeoutId | undefined} */
  let id;
  let deadline = 0;
  /** @type {ThisParameterType<F> | undefined} */
  let callThis;
  let callArgs = initialArgs;

  function call() {
    const remaining = deadline - getNowMs();
    if (remaining > 0) {
      id = setTimeout(call, remaining);
      return;
    }

    const thisArg = callThis;
    const args = callArgs || [];
    id = undefined;
    callThis = undefined;
    callArgs = initialArgs;
    fn.apply(thisArg, args);
  }
  /**
   * @this {ThisParameterType<F>}
   * @param {Parameters<F>} args
   * @returns {void}
   */
  function debounced(...args) {
    callThis = this;
    callArgs = argsReducer(callArgs, args);
    deadline = getNowMs() + ms;
    call();
  }
  Object.defineProperty(debounced, 'scheduled', {
    get() {
      return id !== undefined;
    },
  });
  debounced.cancel = () => {
    clearTimeout(id);
    id = undefined;
    callThis = undefined;
    callArgs = initialArgs;
  };
  debounced.invoke = () => {
    clearTimeout(id);
    deadline = -1;
    call();
  };

  return decoratedFn(fn, debounced);
}

/**
 * @template {(...args: any[]) => any} F
 * @typedef {DebounceOptions<F>} ThrottleOptions
 */
/**
 * @template {(...args: any[]) => any} F
 * @param {Scheduler<() => void>} schedule
 * @param {F} fn
 * @param {ThrottleOptions<F>} [options]
 * @returns {ThrottledFunction<F>}
 */
export function throttleWith(
  schedule,
  fn,
  {argsReducer = keepLastCallArgs, initialArgs} = {}
) {
  /** @type {AbortController | null} */
  let abortController;
  /** @type {ThisParameterType<F> | undefined} */
  let callThis;
  let callArgs = initialArgs;

  function call() {
    const thisArg = callThis;
    const args = callArgs || [];
    abortController = null;
    callThis = undefined;
    callArgs = initialArgs;
    fn.apply(thisArg, args);
  }
  /**
   * @param {Error} error
   * @returns {void}
   */
  function logError(error) {
    if (!isAbortError(error)) {
      console.error('Uncaught error in throttled function:', fn, error);
    }
  }
  /**
   * @this {ThisParameterType<F>}
   * @param {Parameters<F>} args
   * @returns {void}
   */
  function throttled(...args) {
    callThis = this;
    callArgs = argsReducer(callArgs, args);
    if (!abortController) {
      abortController = new AbortController();
      schedule(call, abortController.signal).catch(logError);
    }
  }
  Object.defineProperty(throttled, 'scheduled', {
    get() {
      return !!abortController;
    },
  });
  throttled.cancel = () => {
    if (abortController) {
      abortController.abort();
      abortController = null;
      callThis = undefined;
      callArgs = initialArgs;
    }
  };
  throttled.invoke = () => {
    abortController?.abort();
    call();
  };

  return decoratedFn(fn, throttled);
}

/**
 * @template {Func} F
 * @param {Milliseconds} ms
 * @param {F} fn
 * @param {ThrottleOptions<F>} [options = {}]
 * @returns {ThrottledFunction<F>}
 */
export function throttleOncePerMs(ms, fn, options) {
  const schedule = scheduleDelayedTask(ms);
  return throttleWith(schedule, fn, options);
}

/**
 * @template {Func} F
 * @param {F} fn
 * @param {ThrottleOptions<F>} [options = {}]
 * @returns {ThrottledFunction<F>}
 */
export function throttleOncePerEventLoop(fn, options) {
  return throttleWith(scheduleTask, fn, options);
}

/**
 * @template {Func} F
 * @param {F} fn
 * @param {ThrottleOptions<F>} [options = {}]
 * @returns {ThrottledFunction<F>}
 */
export function throttleOncePerAnimationFrame(fn, options) {
  return throttleWith(scheduleAnimationFrame, fn, options);
}
