/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {comparisonComparator} from './comparator.js';
import {$} from './typescript.js';

/**
 * @template T
 * @param {-1 | 0 | 1} side
 * @param {readonly T[]} array
 * @param {T | ((b: T) => number)} valueOrComparator
 * @returns {number}
 */
function _bisect(side, array, valueOrComparator) {
  let left = 0;
  let right = array.length - 1;
  const comparator = /** @type {(b: T) => number} */ (
    typeof valueOrComparator === 'function'
    ? valueOrComparator
    : comparisonComparator.bind(null, valueOrComparator)
  );
  let mid;
  let comparatorResult;

  while (left <= right) {
    mid = Math.floor((left + right) / 2);
    comparatorResult = comparator($(array[mid]));

    if (comparatorResult < 0 || (comparatorResult === 0 && side === -1)) {
      right = mid - 1;
    } else if (comparatorResult > 0 || (comparatorResult === 0 && side === 1)) {
      left = mid + 1;
    } else {
      return mid;
    }
  }
  return side ? left : -(left + 1);
}
/**
 * @template T
 * @param {readonly T[]} array
 * @param {T | ((b: T) => number)} valueOrComparator
 * @returns {number}
 */
export function bisectLeft(array, valueOrComparator) {
  return _bisect(-1, array, valueOrComparator);
}
/**
 * @template T
 * @param {readonly T[]} array
 * @param {T | ((b: T) => number)} valueOrComparator
 * @returns {number}
 */
export function bisectRight(array, valueOrComparator) {
  return _bisect(1, array, valueOrComparator);
}
/**
 * @template T
 * @param {readonly T[]} array
 * @param {T | ((b: T) => number)} valueOrComparator
 * @returns {number}
 */
export function binarySearch(array, valueOrComparator) {
  return _bisect(0, array, valueOrComparator);
}
/**
 * @template T
 * @param {readonly T[]} array
 * @param {T | ((b: T) => number)} valueOrComparator
 * @param {T} [defaultValue]
 * @returns {T | undefined}
 */
export function findInSorted(array, valueOrComparator, defaultValue) {
  const index = binarySearch(array, valueOrComparator);
  if (index < 0) {
    return defaultValue;
  }
  return array[index];
}

/**
 * @template T
 * @param {T[]} array
 * @param {T} value
 * @param {(b: T) => number} [comparator]
 * @returns {void}
 */
export function addToSorted(array, value, comparator) {
  const index = bisectLeft(array, comparator || value);
  array.splice(index, 0, value);
}
/**
 * @template T
 * @param {T[]} array
 * @param {T} value
 * @param {(b: T) => number} [comparator]
 * @returns {void}
 */
export function addToUniqueSorted(array, value, comparator) {
  const index = bisectLeft(array, comparator || value);
  if (array[index] !== value) {
    array.splice(index, 0, value);
  }
}
