/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$} from './typescript.js';

/** @type {{name: string, failed: boolean}} */
let currentTest;

/**
 * @param {unknown} a
 * @param {unknown} b
 * @returns {boolean}
 */
export function deepEquals(a, b) {
  if (Object.is(a, b)) {
    return true;
  }
  if (Array.isArray(a)) {
    if (!Array.isArray(b) || a.length !== b.length) {
      return false;
    }
    for (let i = 0; i < a.length; i++) {
      if (!deepEquals(a[i], b[i])) {
        return false;
      }
    }
    return true;
  }
  const ka = Object.keys(/** @type {any} */ (a));
  const kb = Object.keys(/** @type {any} */ (b));
  if (ka.length !== kb.length) {
    return false;
  }
  for (let i = 0; i < ka.length; i++) {
    if (
      !deepEquals(
        /** @type {any} */ (a)[$(ka[i])],
        /** @type {any} */ (b)[$(ka[i])]
      )
    ) {
      return false;
    }
  }
  return true;
}

/**
 * @param {boolean} condition
 * @param {string} [message]
 * @returns {asserts condition}
 */
export function assert(condition, message) {
  if (!condition) {
    console.error(
      new Error('Assertion failed' + (message ? `: ${message}` : ''))
    );
    currentTest.failed = true;
  }
}

/**
 * @template T
 * @param {unknown} actual
 * @param {T} expected
 * @param {string} [message]
 * @returns {asserts actual is T}
 */
assert.equal = function(actual, expected, message = `Actual !== Expected\nActual: ${actual}\nExpected: ${expected}`) {
  assert(Object.is(actual, expected), message);
};

/**
 * @template T
 * @param {unknown} actual
 * @param {T} expected
 * @param {string} [message]
 * @returns {asserts actual is T}
 */
assert.deepEqual = function(actual, expected, message = `Actual !== Expected (deep)\nActual: ${actual}\nExpected: ${expected}`) {
  assert(deepEquals(actual, expected), message);
};

/**
 * @param {string} name
 * @param {() => void} exec
 * @returns {void}
 */
export function test(name, exec) {
  console.group(name);
  const prevTest = currentTest;
  currentTest = {name, failed: false};
  try {
    exec();
  } catch (error) {
    console.error(error);
    currentTest.failed = true;
  }
  if (!currentTest.failed) {
    console.log('OK');
  }
  currentTest = prevTest;
  console.groupEnd();
}
