/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$} from './typescript.js';

/** @typedef {Float64Array & [number, number, number, number, number, number]} Transform2D */

/**
 * @returns {Transform2D}
 */
export function identity() {
  const out = new Float64Array(6);
  out[0] = 1;
  out[3] = 1;
  return /** @type {Transform2D} */ (out);
}
/**
 * @param {number} dx
 * @param {number} dy
 * @returns {Transform2D}
 */
export function translation(dx, dy) {
  const out = new Float64Array(6);
  out[0] = 1;
  out[3] = 1;
  out[4] = dx;
  out[5] = dy;
  return /** @type {Transform2D} */ (out);
}
/**
 * @param {number} alpha
 * @returns {Transform2D}
 */
export function rotation(alpha) {
  const s = Math.sin(alpha);
  const c = Math.cos(alpha);
  const out = new Float64Array(6);
  out[0] = c;
  out[1] = s;
  out[2] = -s;
  out[3] = c;
  return /** @type {Transform2D} */ (out);
}
/**
 * @param {number} sx
 * @param {number} sy
 * @returns {Transform2D}
 */
export function scale(sx, sy) {
  const out = new Float64Array(6);
  out[0] = sx;
  out[3] = sy;
  return /** @type {Transform2D} */ (out);
}
/**
 * @param {Transform2D} a
 * @param {Readonly<Transform2D>} b
 * @returns {Transform2D}
 */
function multiply(a, b) {
  const [a0, a1, a2, a3, a4, a5] = a;
  const [b0, b1, b2, b3, b4, b5] = b;
  a[0] = a0 * b0 + a2 * b1;
  a[1] = a1 * b0 + a3 * b1;
  a[2] = a0 * b2 + a2 * b3;
  a[3] = a1 * b2 + a3 * b3;
  a[4] = a0 * b4 + a2 * b5 + a4;
  a[5] = a1 * b4 + a3 * b5 + a5;
  return a;
}
/**
 * @param {readonly Transform2D[]} transforms
 * @returns {Transform2D}
 */
export function concat(transforms) {
  return transforms.reduceRight(multiply, identity());
}
/**
 * @param {Readonly<Transform2D>} t
 * @returns {Transform2D | null}
 */
export function inverse(t) {
  const [aa, ab, ac, ad, atx, aty] = t;

  let det = aa * ad - ab * ac;
  if (!det) {
    return null;
  }
  det = 1.0 / det;

  const out = new Float64Array(6);
  out[0] = ad * det;
  out[1] = -ab * det;
  out[2] = -ac * det;
  out[3] = aa * det;
  out[4] = (ac * aty - ad * atx) * det;
  out[5] = (ab * atx - aa * aty) * det;
  return /** @type {Transform2D} */ (out);
}

/**
 * @param {Readonly<Transform2D>} transform
 * @param {XY} point
 * @returns {XY}
 */
export function transformPoint(transform, point) {
  const [x, y] = point;
  point[0] = transform[0] * x + transform[2] * y + transform[4];
  point[1] = transform[1] * x + transform[3] * y + transform[5];
  return point;
}
/**
 * @param {Readonly<Transform2D>} transform
 * @param {XY[]} points
 * @returns {XY[]}
 */
export function transformPoints(transform, points) {
  points.forEach(transformPoint.bind(null, transform));
  return points;
}
/**
 * @param {Readonly<Transform2D>} transform
 * @param {BezierPath} path
 * @returns {BezierPath}
 */
export function transformPath(transform, path) {
  /** @typedef {import('../utils.js').BezierPath} BezierPath */

  const [t0, t1, t2, t3, t4, t5] = transform;
  const [px, py] = path;
  const {length} = px;

  for (let i = 0; i < length; i++) {
    const pxi = $(px[i]);
    const pyi = $(py[i]);
    const l = pxi.length;

    for (let j = 0; j < l; j++) {
      const x = $(pxi[j]);
      const y = $(pyi[j]);
      pxi[j] = x * t0 + y * t2 + t4;
      pyi[j] = x * t1 + y * t3 + t5;
    }
  }
  return path;
}
/**
 * @param {Readonly<Transform2D>} transform
 * @param {Readonly<XY>} point
 * @returns {XY}
 */
export function transformedPoint(transform, [x, y]) {
  return [
    transform[0] * x + transform[2] * y + transform[4],
    transform[1] * x + transform[3] * y + transform[5],
  ];
}
/**
 * @param {Readonly<Transform2D>} transform
 * @param {Readonly<XY[]>} points
 * @returns {XY[]}
 */
export function transformedPoints(transform, points) {
  return points.map(transformedPoint.bind(null, transform));
}
/**
 * @param {Readonly<Transform2D>} transform
 * @param {Readonly<BezierPath>} path
 * @returns {BezierPath}
 */
export function transformedPath(transform, path) {
  /** @typedef {import('../utils.js').BezierPath} BezierPath */

  const [t0, t1, t2, t3, t4, t5] = transform;
  const [px, py] = path;
  const {length} = px;
  const xs = [];
  const ys = [];

  for (let i = 0; i < length; i++) {
    const pxi = $(px[i]);
    const pyi = $(py[i]);
    const l = pxi.length;
    const xi = new Float64Array(l);
    const yi = new Float64Array(l);

    for (let j = 0; j < l; j++) {
      const x = $(pxi[j]);
      const y = $(pyi[j]);
      xi[j] = x * t0 + y * t2 + t4;
      yi[j] = x * t1 + y * t3 + t5;
    }

    xs.push(xi);
    ys.push(yi);
  }
  return [xs, ys];
}
