/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/**
 * Heavily inspired by [Facebook's
 * prop-types](https://github.com/facebook/prop-types)
 *
 * Just in case, here is the MIT license for prop-types:
 * MIT License
 *
 * Copyright (c) 2013-present, Facebook, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
import {DATE_MIN_TIME, DATE_MAX_TIME} from './date_utils.js';
import {$} from './typescript.js';
import {asArray, appendToIndexPath, getPropAtIndexPath} from './utils.js';

const isPrototypeOf = Object.prototype.isPrototypeOf;
const ANONYMOUS = /** @type {const} */ ('<<anonymous>>');
const IGNORE = /** @type {const} */ ('__IGNORE__');
const AND = /** @type {const} */ ('AND');
const OR = /** @type {const} */ ('OR');

createMatchTypeChecker.IGNORE = IGNORE;

export const Types = {
  any    : createRequireableTypeChecker(() => null),
  none   : validateNone,
  bool   : createPrimitiveTypeChecker('boolean'),
  func   : createPrimitiveTypeChecker('function'),
  string : createStringTypeChecker(),
  number : createNumberTypeChecker(),
  integer: createIntegerTypeChecker(),

  date    : createDateTypeChecker(false),
  dateOrMs: createDateTypeChecker(true),

  required  : createRequiredTypeChecker,
  value     : createValueTypeChecker,
  instanceOf: createInstanceTypeChecker,
  subclassOf: createSubclassTypeChecker,
  oneOf     : createEnumTypeChecker,
  oneOfType : createUnionTypeChecker,
  allOfType : createIntersectTypeChecker,
  arrayOf   : createArrayOfTypeChecker,
  objectOf  : createObjectOfTypeChecker,
  tuple     : createTupleTypeChecker,
  shape     : createShapeTypeChecker,
  match     : createMatchTypeChecker,
  case      : createCaseTypeChecker,
  lazy      : createLazyTypeChecker,
};

const TypedArray = Object.getPrototypeOf(Uint8Array);

/**
 * @param {unknown} obj
 * @returns {obj is ArrayLike<unknown>}
 */
function isArrayLike(obj) {
  return Array.isArray(obj) || obj instanceof TypedArray;
}

/** @typedef {Exclude<keyof any, symbol>} TypeCheckerPropName */
/** @typedef {(indexPath: string) => any} TypeCheckerGetProp */
/** @typedef {string | string[] | null} TypeCheckerResult */
/**
 * @callback TypeChecker
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @param {TypeCheckerGetProp} getProp
 * @returns {TypeCheckerResult}
 */

/**
 * @param {TypeChecker} typeChecker
 * @param {string} name
 * @param {unknown} value
 * @returns {string | null}
 */
export function validateValueType(typeChecker, name, value) {
  if (typeof typeChecker !== 'function') {
    throw new TypeError(
      'Invalid argument supplied to validateValueType, expected a function.'
    );
  }

  const object = {[name]: value};
  const fullName = appendToIndexPath('', name);

  /**
   * @param {string} indexPath
   * @returns {any}
   */
  function getProp(indexPath) {
    if (indexPath[0] === '.' || indexPath[0] === '[') {
      indexPath = fullName + indexPath;
    }
    return getPropAtIndexPath(object, indexPath, 'parent');
  }

  const errors = typeChecker(object, name, fullName, getProp);
  return makeList(AND, errors);
}

export class TypeCheckerError extends TypeError {
  /**
   * @public
   * @param {string} message
   * @param {string} name
   * @param {unknown} value
   */
  constructor(message, name, value) {
    super(message);
    this.name = name;
    this.value = value;
  }
}

/**
 * @param {TypeChecker} typeChecker
 * @param {string} name
 * @param {unknown} value
 * @returns {void}
 * @throws {TypeCheckerError}
 */
export function assertValueType(typeChecker, name, value) {
  const error = validateValueType(typeChecker, name, value);
  if (error) {
    throw new TypeCheckerError(error, name, value);
  }
}

/** @typedef {TypeChecker & {isRequired: TypeChecker}} RequireableTypeChecker */
/**
 * @param {TypeChecker} validate
 * @param {boolean} [consumeNullish]
 * @returns {RequireableTypeChecker}
 */
function createRequireableTypeChecker(validate, consumeNullish = true) {
  /**
   * @param {boolean} isRequired
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function checkType(isRequired, props, propName, propFullName, getProp) {
    propFullName = propFullName || `${propName}`;

    if (!Object.hasOwn(props, propName)) {
      if (isRequired) {
        return 'Missing required `' + propFullName + '`.';
      }
      if (consumeNullish) {
        return null;
      }
    }
    if (props[propName] == null) {
      if (isRequired) {
        return [
          '`', propFullName, '` is required, but its value is `',
          String(props[propName]), '`.'
        ].join('');
      }
      if (consumeNullish) {
        return null;
      }
    }
    return validate(props, propName, propFullName, getProp);
  }

  const chainedCheckType = /** @type {RequireableTypeChecker} */ (
    checkType.bind(null, false)
  );
  chainedCheckType.isRequired = checkType.bind(null, true);

  return chainedCheckType;
}

/**
 * @param {TypeChecker} typeChecker
 * @returns {typeChecker is RequireableTypeChecker}
 */
function isRequireableTypeChecker(typeChecker) {
  return (
    typeof /** @type {RequireableTypeChecker} */ (typeChecker).isRequired
    === 'function'
  );
}
/**
 * @param {TypeChecker} typeChecker
 * @param {boolean} [consumeNullish]
 * @returns {RequireableTypeChecker}
 */
function asRequireableTypeChecker(typeChecker, consumeNullish = true) {
  return (
    isRequireableTypeChecker(typeChecker)
    ? typeChecker
    : createRequireableTypeChecker(typeChecker, consumeNullish)
  );
}

/**
 * @typedef {RequireableTypeChecker & {
 *   len: (length: number | [min?: number | undefined, max?: number | undefined]) => StringTypeChecker,
 *   match: (regexp: RegExp) => StringTypeChecker,
 * }} StringTypeChecker
 */
/**
 * @param {TypeChecker} validate
 * @returns {StringTypeChecker}
 */
function createChainableStringTypeChecker(validate) {
  const chainedCheckType = /** @type {StringTypeChecker} */ (
    asRequireableTypeChecker(validate)
  );

  chainedCheckType.len = function(length) {
    /** @type {number | null} */
    let expectedLength = null;
    let min = 0;
    let max = Infinity;
    if (Array.isArray(length)) {
      if (!isNaN(length[0])) {
        min = Math.floor(Math.max(0, $(length[0])));
      }
      if (!isNaN(length[1])) {
        max = Math.ceil(Math.max(0, $(length[1])));
      }
    } else if (isFinite(length) && length >= 0) {
      expectedLength = Math.floor(length);
    } else if (arguments.length > 1) {
      throw new TypeError(
        'Invalid argument supplied to len, expected a positive ' +
        'finite number or a couple of the form [min, max].'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {string} */ (props[propName]);
      if (expectedLength !== null && propValue.length !== expectedLength) {
        return [
          'Invalid `', propFullName, '` string of length ',
          '`', propValue.length, '`, expected a string of length ',
          '`', expectedLength, '`.'
        ].join('');
      } else if (propValue.length < min || propValue.length >= max) {
        return [
          'Invalid `', propFullName, '` string of length ',
          '`', propValue.length, '`, expected a string of length in the ',
          'interval [`', min, '`, `', max, '`[.'
        ].join('');
      }
      return null;
    }
    return createChainableStringTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };
  chainedCheckType.match = function(regexp) {
    if (!(regexp instanceof RegExp)) {
      throw new TypeError(
        'Invalid argument supplied to match, expected a RegExp.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {string} */ (props[propName]);
      if (regexp.test(propValue)) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected to match ', regexp, '.'
      ].join('');
    }
    return createChainableStringTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };

  return chainedCheckType;
}

/**
 * @typedef {RequireableTypeChecker & {
 *   eq: (expectation: number | string) => NumberTypeChecker,
 *   neq: (expectation: number | string) => NumberTypeChecker,
 *   gt: (limit: number | string) => NumberTypeChecker,
 *   gte: (limit: number | string) => NumberTypeChecker,
 *   lt: (limit: number | string) => NumberTypeChecker,
 *   lte: (limit: number | string) => NumberTypeChecker,
 * }} NumberTypeChecker
 */
/**
 * @param {TypeChecker} validate
 * @returns {NumberTypeChecker}
 */
function createChainableNumberTypeChecker(validate) {
  const chainedCheckType = /** @type {NumberTypeChecker} */ (
    asRequireableTypeChecker(validate)
  );

  chainedCheckType.eq = function(expectation) {
    if (typeof expectation !== 'number' && typeof expectation !== 'string') {
      throw new TypeError(
        'Invalid argument supplied to eq, expected a number or string.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {number} */ (props[propName]);
      let expectationValue;

      if (typeof expectation === 'string') {
        expectationValue = /** @type {unknown} */ (props[expectation]);
      } else {
        expectationValue = expectation;
      }

      if (
        propValue === expectationValue ||
        (isNaN(propValue) && isNaN(/** @type {number} */ (expectationValue)))
      ) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected equal to `', expectationValue, '`.'
      ].join('');
    }
    return createChainableNumberTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };
  chainedCheckType.neq = function(expectation) {
    if (typeof expectation !== 'number' && typeof expectation !== 'string') {
      throw new TypeError(
        'Invalid argument supplied to neq, expected a number or string.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {number} */ (props[propName]);
      let expectationValue;

      if (typeof expectation === 'string') {
        expectationValue = /** @type {unknown} */ (props[expectation]);
      } else {
        expectationValue = expectation;
      }

      if (
        propValue !== expectationValue &&
        isNaN(propValue) !== isNaN(/** @type {number} */ (expectationValue))
      ) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected different from `', expectationValue, '`.'
      ].join('');
    }
    return createChainableNumberTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };
  chainedCheckType.gt = function(limit) {
    if (typeof limit !== 'number' && typeof limit !== 'string') {
      throw new TypeError(
        'Invalid argument supplied to gt, expected a number or string.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {number} */ (props[propName]);
      let limitValue;

      if (typeof limit === 'string') {
        limitValue = /** @type {unknown} */ (props[limit]);
      } else {
        limitValue = limit;
      }

      if (propValue > /** @type {number} */ (limitValue)) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected greater than `', limitValue, '`.'
      ].join('');
    }
    return createChainableNumberTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };
  chainedCheckType.gte = function(limit) {
    if (typeof limit !== 'number' && typeof limit !== 'string') {
      throw new TypeError(
        'Invalid argument supplied to gte, expected a number or string.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {number} */ (props[propName]);
      let limitValue;

      if (typeof limit === 'string') {
        limitValue = /** @type {unknown} */ (props[limit]);
      } else {
        limitValue = limit;
      }

      if (propValue >= /** @type {number} */ (limitValue)) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected greater than or equal to `', limitValue, '`.'
      ].join('');
    }
    return createChainableNumberTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };
  chainedCheckType.lt = function(limit) {
    if (typeof limit !== 'number' && typeof limit !== 'string') {
      throw new TypeError(
        'Invalid argument supplied to lt, expected a number or string.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {number} */ (props[propName]);
      let limitValue;

      if (typeof limit === 'string') {
        limitValue = /** @type {unknown} */ (props[limit]);
      } else {
        limitValue = limit;
      }

      if (propValue < /** @type {number} */ (limitValue)) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected less than `', limitValue, '`.'
      ].join('');
    }
    return createChainableNumberTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };
  chainedCheckType.lte = function(limit) {
    if (typeof limit !== 'number' && typeof limit !== 'string') {
      throw new TypeError(
        'Invalid argument supplied to lte, expected a number or string.'
      );
    }
    /**
     * @param {any} props
     * @param {TypeCheckerPropName} propName
     * @param {string} [propFullName]
     * @returns {TypeCheckerResult}
     */
    function checkValid(props, propName, propFullName) {
      const propValue = /** @type {number} */ (props[propName]);
      let limitValue;

      if (typeof limit === 'string') {
        limitValue = /** @type {unknown} */ (props[limit]);
      } else {
        limitValue = limit;
      }

      if (propValue <= /** @type {number} */ (limitValue)) {
        return null;
      }
      return [
        'Invalid `', propFullName, '` of value `', propValue, '`, ',
        'expected less than or equal to `', limitValue, '`.'
      ].join('');
    }
    return createChainableNumberTypeChecker(
      createIntersectTypeChecker([
        chainedCheckType,
        checkValid,
      ])
    );
  };

  return chainedCheckType;
}

/**
 * @param {any} props
 * @param {TypeCheckerPropName} propName
 * @param {string} propFullName
 * @returns {TypeCheckerResult}
 */
function validateNone(props, propName, propFullName) {
  if (props[propName] === undefined) {
    return null;
  }
  return 'Unexpected `' + propFullName + '`.';
}
/**
 * @param {ReturnType<getType>} expectedType
 * @returns {RequireableTypeChecker}
 */
function createPrimitiveTypeChecker(expectedType) {
  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    const propValue = props[propName];
    const propType = getType(propValue);
    if (propType !== expectedType) {
      // `propValue` being instance of, say, date/regexp, pass the 'object'
      // check, but we can offer a more precise error message here rather than
      // 'of type `object`'.
      const preciseType = getPreciseType(propValue);

      return [
        'Invalid `', propFullName, '` of type `', preciseType, '`, ',
        'expected `', expectedType, '`.'
      ].join('');
    }
    return null;
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @returns {StringTypeChecker}
 */
function createStringTypeChecker() {
  return createChainableStringTypeChecker(createPrimitiveTypeChecker('string'));
}
/**
 * @returns {NumberTypeChecker}
 */
function createNumberTypeChecker() {
  return createChainableNumberTypeChecker(createPrimitiveTypeChecker('number'));
}
/**
 * @returns {NumberTypeChecker}
 */
function createIntegerTypeChecker() {
  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    const propValue = /** @type {number} */ (props[propName]);
    /** @type {string} */
    let propType = getType(propValue);
    if (propType === 'number' && Math.floor(propValue) === propValue) {
      propType = 'integer';
    }
    if (propType !== 'integer') {
      // `propValue` being instance of, say, date/regexp, pass the 'object'
      // check, but we can offer a more precise error message here rather than
      // 'of type `object`'.
      const preciseType = getPreciseType(propValue);

      return [
        'Invalid `', propFullName, '` of type `', preciseType, '`, ',
        'expected `integer`.'
      ].join('');
    }
    return null;
  }
  return createChainableNumberTypeChecker(validate);
}
/**
 * @param {boolean} canBeNumeric
 * @returns {NumberTypeChecker}
 */
function createDateTypeChecker(canBeNumeric) {
  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    const propValue = props[propName];
    const isNumeric = (
      canBeNumeric &&
      getType(propValue) === 'number' &&
      /** @type {number} */ (propValue) >= DATE_MIN_TIME &&
      /** @type {number} */ (propValue) <= DATE_MAX_TIME
    );
    const isDateInstance = propValue instanceof Date;
    const isValid = (isNumeric || isDateInstance) && !isNaN(propValue);
    if (isValid) {
      return null;
    }
    const expectation = (
      canBeNumeric
      ? 'expected `number` or valid `Date` instance'
      : 'expected valid `Date` instance'
    );
    if (!isNumeric && !isDateInstance) {
      // `propValue` being instance of, say, date/regexp, pass the 'object'
      // check, but we can offer a more precise error message here rather than
      // 'of type `object`'.
      const preciseType = getPreciseType(propValue);

      return [
        'Invalid `', propFullName, '` of type `', preciseType, '`, ',
        expectation, '.'
      ].join('');
    }
    return [
      'Invalid `', propFullName, '` of value `', propValue, '`, ',
      expectation, '.'
    ].join('');
  }
  return createChainableNumberTypeChecker(validate);
}
/**
 * @param {TypeChecker} [typeChecker]
 * @returns {TypeChecker}
 */
function createRequiredTypeChecker(typeChecker = Types.any) {
  if (typeof typeChecker !== 'function') {
    throw new TypeError(
      'Invalid 1st argument supplied to required, expected a function.'
    );
  }
  return asRequireableTypeChecker(typeChecker, false).isRequired;
}
/**
 * @param {unknown} expectedValue
 * @returns {RequireableTypeChecker}
 */
function createValueTypeChecker(expectedValue) {
  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    const propValue = props[propName];
    if (Object.is(propValue, expectedValue)) {
      return null;
    }

    return [
      'Invalid `', propFullName, '` of value `', propValue, '`, ',
      'expected `', expectedValue, '`.'
    ].join('');
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {Class} expectedClass
 * @returns {RequireableTypeChecker}
 */
function createInstanceTypeChecker(expectedClass) {
  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    if (!(props[propName] instanceof expectedClass)) {
      const expectedClassName = expectedClass.name || ANONYMOUS;
      const actualClassName = (
        getClassName(/** @type {Class} */ (props[propName]))
      );
      return [
        'Invalid `', propFullName, '` of type `', actualClassName, '`, ',
        'expected instance of `', expectedClassName, '`.'
      ].join('');
    }
    return null;
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {Class} superclass
 * @returns {RequireableTypeChecker}
 */
function createSubclassTypeChecker(superclass) {
  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    if (
      !isPrototypeOf.call(superclass, /** @type {Object} */ (props[propName]))
    ) {
      const superclassName = superclass.name || ANONYMOUS;
      const actualClassName = (
        getClassName(/** @type {Class} */ (props[propName]))
      );
      return [
        'Invalid `', propFullName, '` of value `', actualClassName, '`, ',
        'expected subclass of `', superclassName, '`.'
      ].join('');
    }
    return null;
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {unknown[]} expectedValues
 * @returns {RequireableTypeChecker}
 */
function createEnumTypeChecker(expectedValues) {
  if (!Array.isArray(expectedValues)) {
    throw new TypeError(
      'Invalid argument supplied to oneOf, expected an instance of array.'
    );
  }
  if (expectedValues.length === 1) {
    return createValueTypeChecker(expectedValues[0]);
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName) {
    const propValue = props[propName];
    for (const expectedValue of expectedValues) {
      if (Object.is(propValue, expectedValue)) {
        return null;
      }
    }

    const valuesString = JSON.stringify(expectedValues);
    return [
      'Invalid `', propFullName, '` of value `', propValue, '`, ',
      'expected one of ', valuesString, '.'
    ].join('');
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {TypeChecker[]} arrayOfTypeCheckers
 * @returns {RequireableTypeChecker}
 */
function createUnionTypeChecker(arrayOfTypeCheckers) {
  if (!Array.isArray(arrayOfTypeCheckers)) {
    throw new TypeError(
      'Invalid argument supplied to oneOfType, expected an instance of array.'
    );
  }
  if (arrayOfTypeCheckers.length === 1) {
    return asRequireableTypeChecker($(arrayOfTypeCheckers[0]), false);
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    const errors = [];
    for (const checker of arrayOfTypeCheckers) {
      let checkerErrors = checker(props, propName, propFullName, getProp);
      checkerErrors = makeList(AND, checkerErrors);
      if (checkerErrors) {
        errors.push(checkerErrors);
      } else {
        return null;
      }
    }
    return makeList('Invalid `' + propFullName + '`', OR, errors);
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {TypeChecker[]} arrayOfTypeCheckers
 * @returns {RequireableTypeChecker}
 */
function createIntersectTypeChecker(arrayOfTypeCheckers) {
  if (!Array.isArray(arrayOfTypeCheckers)) {
    throw new TypeError(
      'Invalid argument supplied to allOfType, expected an instance of array.'
    );
  }
  if (arrayOfTypeCheckers.length === 1) {
    return asRequireableTypeChecker($(arrayOfTypeCheckers[0]), false);
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    const errors = [];
    for (const checker of arrayOfTypeCheckers) {
      const checkerErrors = checker(props, propName, propFullName, getProp);
      if (Array.isArray(checkerErrors)) {
        errors.push.apply(errors, checkerErrors);
      } else if (checkerErrors) {
        errors.push(checkerErrors);
      }
    }
    return errors;
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @typedef {RequireableTypeChecker
 *           & {isExact: RequireableTypeChecker}} ExactTypeChecker
 */
/**
 * @param {TypeChecker[]} arrayOfTypeCheckers
 * @returns {ExactTypeChecker}
 */
function createTupleTypeChecker(arrayOfTypeCheckers) {
  if (!Array.isArray(arrayOfTypeCheckers)) {
    throw new TypeError(
      'Invalid argument supplied to tuple, expected an instance of array.'
    );
  }

  /**
   * @param {boolean} isExact
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(isExact, props, propName, propFullName, getProp) {
    const propValue = props[propName];
    if (!isArrayLike(propValue)) {
      const propType = getType(propValue);
      return [
        'Invalid `', propFullName, '` of type `', propType, '`, ',
        'expected a tuple of length `', arrayOfTypeCheckers.length, '`',
        isExact ? '' : ' or more', '.'
      ].join('');
    }

    const errors = [];
    if (isExact) {
      if (propValue.length !== arrayOfTypeCheckers.length) {
        errors.push([
          'Invalid `', propFullName, '` tuple of length ',
          '`', propValue.length, '`, expected a tuple of length ',
          '`', arrayOfTypeCheckers.length, '`.'
        ].join(''));
      }
    } else if (propValue.length > arrayOfTypeCheckers.length) {
      errors.push([
        'Invalid `', propFullName, '` tuple of length ',
        '`', propValue.length, '`, expected at most ',
        '`', arrayOfTypeCheckers.length, '`.'
      ].join(''));
    }
    for (let i = 0; i < arrayOfTypeCheckers.length; i++) {
      const checker = /** @type {TypeChecker} */ (arrayOfTypeCheckers[i]);
      const checkerErrors = (
        checker(propValue, i, appendToIndexPath(propFullName, i), getProp)
      );
      if (Array.isArray(checkerErrors)) {
        errors.push.apply(errors, checkerErrors);
      } else if (checkerErrors) {
        errors.push(checkerErrors);
      }
    }
    return errors;
  }

  const exactValidate = /** @type {ExactTypeChecker} */ (
    createRequireableTypeChecker(validate.bind(null, false))
  );
  exactValidate.isExact = (
    createRequireableTypeChecker(validate.bind(null, true))
  );

  return exactValidate;
}
/**
 * @param {TypeChecker} typeChecker
 * @param {number | [min?: number | undefined, max?: number | undefined]} [length]
 * @returns {RequireableTypeChecker}
 */
function createArrayOfTypeChecker(typeChecker, length) {
  if (typeof typeChecker !== 'function') {
    throw new TypeError(
      'Invalid 1st argument supplied to arrayOf, expected a function.'
    );
  }

  /** @type {number | null} */
  let expectedLength = null;
  let min = 0;
  let max = Infinity;
  if (Array.isArray(length)) {
    if (!isNaN(length[0])) {
      min = Math.floor(Math.max(0, $(length[0])));
    }
    if (!isNaN(length[1])) {
      max = Math.ceil(Math.max(0, $(length[1])));
    }
  } else if (isFinite(length) && length >= 0) {
    expectedLength = Math.floor(length);
    min = expectedLength;
    max = expectedLength + 1;
  } else if (arguments.length > 1) {
    throw new TypeError(
      'Invalid 2nd argument supplied to arrayOf, expected a positive ' +
      'finite number or a couple of the form [min, max].'
    );
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    const propValue = props[propName];
    if (!isArrayLike(propValue)) {
      const propType = getType(propValue);
      return [
        'Invalid `', propFullName, '` of type `', propType, '`, ',
        'expected an array.'
      ].join('');
    }

    const errors = [];
    if (expectedLength !== null && propValue.length !== expectedLength) {
      errors.push([
        'Invalid `', propFullName, '` array of length ',
        '`', propValue.length, '`, expected an array of length ',
        '`', expectedLength, '`.'
      ].join(''));
    } else if (propValue.length < min || propValue.length >= max) {
      errors.push([
        'Invalid `', propFullName, '` array of length ',
        '`', propValue.length, '`, expected an array of length in the ',
        'interval [`', min, '`, `', max, '`[.'
      ].join(''));
    }
    for (let i = 0; i < propValue.length && i < max; i++) {
      const checkerErrors = (
        typeChecker(propValue, i, appendToIndexPath(propFullName, i), getProp)
      );
      if (Array.isArray(checkerErrors)) {
        errors.push.apply(errors, checkerErrors);
      } else if (checkerErrors) {
        errors.push(checkerErrors);
      }
    }
    return errors;
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {TypeChecker} typeChecker
 * @returns {RequireableTypeChecker}
 */
function createObjectOfTypeChecker(typeChecker) {
  if (typeof typeChecker !== 'function') {
    throw new TypeError(
      'Invalid argument supplied to objectOf, expected a function.'
    );
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    const propValue = /** @type {object} */ (props[propName]);
    const propType = getType(propValue);
    if (propType !== 'object') {
      return [
        'Invalid `', propFullName, '` of type `', propType, '`, ',
        'expected an object.'
      ].join('');
    }
    const errors = [];
    for (const key in propValue) {
      if (Object.hasOwn(propValue, key)) {
        const checkerErrors = typeChecker(
          propValue,
          key,
          appendToIndexPath(propFullName, key),
          getProp
        );
        if (Array.isArray(checkerErrors)) {
          errors.push.apply(errors, checkerErrors);
        } else if (checkerErrors) {
          errors.push(checkerErrors);
        }
      }
    }
    return errors;
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @typedef {ExactTypeChecker & {
 *   withIndexer: {
 *     (valueChecker: TypeChecker): RequireableTypeChecker,
 *     (keyChecker: TypeChecker, valueChecker: TypeChecker): RequireableTypeChecker,
 *   },
 * }} ExactWithIndexerTypeChecker
 */
/**
 * @param {Record<string, TypeChecker>} shapeTypes
 * @returns {ExactWithIndexerTypeChecker}
 */
function createShapeTypeChecker(shapeTypes) {
  /**
   * @param {TypeChecker | null} keyChecker
   * @param {TypeChecker | null} valueChecker
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(keyChecker, valueChecker, props, propName, propFullName, getProp) {
    const propValue = /** @type {object} */ (props[propName]);
    const propType = getType(propValue);
    if (propType !== 'object') {
      return [
        'Invalid `', propFullName, '` of type `', propType, '`, ',
        'expected `object`.'
      ].join('');
    }
    const errors = [];
    let checkerErrors;
    for (const key in shapeTypes) {
      if (Object.hasOwn(shapeTypes, key)) {
        const checker = /** @type {TypeChecker} */ (shapeTypes[key]);
        checkerErrors = (
          checker(propValue, key, appendToIndexPath(propFullName, key), getProp)
        );
        if (Array.isArray(checkerErrors)) {
          errors.push.apply(errors, checkerErrors);
        } else if (checkerErrors) {
          errors.push(checkerErrors);
        }
      }
    }
    if (keyChecker || valueChecker) {
      for (const key in propValue) {
        if (!Object.hasOwn(shapeTypes, key) && Object.hasOwn(propValue, key)) {
          checkerErrors = keyChecker?.(
            {key},
            'key',
            appendToIndexPath(propFullName, key),
            getProp
          );
          if (checkerErrors) {
            for (const error of asArray(checkerErrors)) {
              errors.push('[Key] ' + error);
            }
          }
          checkerErrors = valueChecker?.(
            propValue,
            key,
            appendToIndexPath(propFullName, key),
            getProp
          );
          if (Array.isArray(checkerErrors)) {
            errors.push.apply(errors, checkerErrors);
          } else if (checkerErrors) {
            errors.push(checkerErrors);
          }
        }
      }
    }
    return errors;
  }

  const shapeValidate = /** @type {ExactWithIndexerTypeChecker} */ (
    createRequireableTypeChecker(validate.bind(null, null, null))
  );
  shapeValidate.isExact = (
    createRequireableTypeChecker(validate.bind(null, validateNone, null))
  );
  /**
   * @overload
   * @param {TypeChecker} valueChecker
   * @returns {RequireableTypeChecker}
   *
   * @overload
   * @param {TypeChecker} keyChecker
   * @param {TypeChecker} valueChecker
   * @returns {RequireableTypeChecker}
   *
   * @param {TypeChecker} valueOrKeyChecker
   * @param {TypeChecker} [valueChecker]
   * @returns {RequireableTypeChecker}
   */
  shapeValidate.withIndexer = function(valueOrKeyChecker, valueChecker) {
    if (typeof valueOrKeyChecker !== 'function') {
      throw new TypeError(
        'Invalid argument supplied to shape().withIndexer, expected a ' +
        'function as 1st argument.'
      );
    }
    let keyChecker = null;
    if (valueChecker) {
      if (typeof valueChecker !== 'function') {
        throw new TypeError(
          'Invalid argument supplied to shape().withIndexer, expected an ' +
          'optional function as 2nd argument.'
        );
      }
      keyChecker = valueOrKeyChecker;
    } else {
      valueChecker = valueOrKeyChecker;
    }
    return createRequireableTypeChecker(validate.bind(null, keyChecker, valueChecker));
  };

  return shapeValidate;
}
/**
 * @param {TypeChecker[]} arrayOfTypeCheckers
 * @returns {RequireableTypeChecker}
 */
function createMatchTypeChecker(arrayOfTypeCheckers) {
  if (!Array.isArray(arrayOfTypeCheckers)) {
    throw new TypeError(
      'Invalid argument supplied to match, expected an instance of array.'
    );
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    for (const checker of arrayOfTypeCheckers) {
      const checkerErrors = checker(props, propName, propFullName, getProp);
      if (checkerErrors !== IGNORE) {
        return makeList(AND, checkerErrors);
      }
    }
    return [
      'Invalid `', propFullName, '` of value `', props[propName], '`, ',
      'expected to match one of the cases.'
    ].join('');
  }
  return createRequireableTypeChecker(validate);
}
/**
 * @param {TypeChecker} conditionChecker
 * @param {TypeChecker} typeChecker
 * @returns {TypeChecker}
 */
function createCaseTypeChecker(conditionChecker, typeChecker) {
  if (
    typeof conditionChecker !== 'function' ||
    typeof typeChecker !== 'function'
  ) {
    throw new TypeError(
      'Invalid argument supplied to case, expected a function.'
    );
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    const checkerErrors = (
      conditionChecker(props, propName, propFullName, getProp)
    );
    if (checkerErrors && checkerErrors.length > 0) {
      // should be ignored by the match checker
      return IGNORE;
    }
    return typeChecker(props, propName, propFullName, getProp);
  }
  return validate;
}
/**
 * @param {(...args: Parameters<TypeChecker>) => TypeChecker} getTypeChecker
 * @returns {RequireableTypeChecker}
 */
function createLazyTypeChecker(getTypeChecker) {
  if (typeof getTypeChecker !== 'function') {
    throw new TypeError(
      'Invalid argument supplied to lazy, expected a function.'
    );
  }

  /**
   * @param {any} props
   * @param {TypeCheckerPropName} propName
   * @param {string} propFullName
   * @param {TypeCheckerGetProp} getProp
   * @returns {TypeCheckerResult}
   */
  function validate(props, propName, propFullName, getProp) {
    const typeChecker = getTypeChecker(props, propName, propFullName, getProp);
    const errors = typeChecker(props, propName, propFullName, getProp);
    return errors;
  }
  return createRequireableTypeChecker(validate);
}

/**
 * Equivalent of `typeof` but with special handling for array and regexp.
 *
 * @param {unknown} value
 * @returns {"string" | "number" | "bigint" | "boolean" | "symbol"
 *           | "undefined" | "object" | "function" | "array"}
 */
function getType(value) {
  const type = typeof value;
  if (Array.isArray(value)) {
    return 'array';
  }
  if (value instanceof RegExp) {
    // Old webkits (at least until Android 4.0) return 'function' rather than
    // 'object' for typeof a RegExp. We'll normalize this here so that /bla/
    // passes PropTypes.object.
    return 'object';
  }
  return type;
}

/**
 * This handles more types than `getType`. Only used for error messages.
 * See `createPrimitiveTypeChecker`.
 *
 * @param {unknown} value
 * @returns {ReturnType<getType> | 'date' | 'regexp' | 'typed-array'}
 */
function getPreciseType(value) {
  var type = getType(value);
  if (type === 'object') {
    if (value instanceof Date) {
      return 'date';
    } else if (value instanceof RegExp) {
      return 'regexp';
    } else if (value instanceof TypedArray) {
      return 'typed-array';
    }
  }
  return type;
}

/**
 * Returns class name of the object, if any.
 *
 * @param {Class} value
 * @returns {string}
 */
function getClassName(value) {
  if (!value.constructor || !value.constructor.name) {
    return ANONYMOUS;
  }
  return value.constructor.name;
}

/**
 * @param {string} line
 * @param {number} index
 * @returns {string}
 */
function indentOtherLine(line, index) {
  if (index === 0) {
    // skip first line
    return line;
  }
  return '  ' + line;
}
/**
 * @param {string} lines
 * @returns {string}
 */
function indentOtherLines(lines) {
  return lines.split('\n').map(indentOtherLine).join('\n');
}
/**
 * @param {string} lines
 * @returns {string}
 */
function addBulletPoint(lines) {
  return '* ' + indentOtherLines(lines);
}
/**
 * @overload
 * @param {string} title
 * @param {typeof AND | typeof OR} grouping
 * @param {TypeCheckerResult} items
 * @returns {string | null}
 *
 * @overload
 * @param {typeof AND | typeof OR} grouping
 * @param {TypeCheckerResult} items
 * @returns {string | null}
 *
 * @param {string | typeof AND | typeof OR} titleOrGrouping
 * @param {typeof AND | typeof OR | TypeCheckerResult} groupingOrItems
 * @param {TypeCheckerResult} [items]
 * @returns {string | null}
 */
function makeList(titleOrGrouping, groupingOrItems, items) {
  let title;
  let grouping;
  if (items === undefined) {
    title = '';
    grouping = titleOrGrouping;
    items = groupingOrItems;
  } else {
    title = titleOrGrouping;
    grouping = groupingOrItems;
  }

  if (!items) {
    return null;
  }
  items = asArray(items);
  if (items.length === 0) {
    return null;
  }
  if (items.length === 1) {
    if (items[0]) {
      return title ? `${title}, ${items[0]}` : items[0];
    }
    return null;
  }
  title += title ? ', a' : 'A';
  if (grouping === AND) {
    title += 'll of the following should be fixed:';
  } else if (grouping === OR) {
    title += 't least one of the following should be fixed:';
  }
  return title + '\n' + items.map(addBulletPoint).join('\n');
}
