/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {$} from './typescript.js';
import {asArray} from './utils.js';

/**
 * Parses a query string and returns it them as a key-value map.
 *
 * @param {string} qs - The query string to parse
 * @returns {Record<string, string>} A key-value map of the query string
 * parameters.
 *
 * @note Like the `URLSearchParams` constructor, this function interprets plus
 * signs (`+`) as spaces.
 */
export function parseQueryString(qs) {
  return Object.fromEntries(new URLSearchParams(qs).entries());
}
/**
 * @param {string | URL} [url]
 * @returns {Record<string, string>}
 */
export function parseURLQueryString(url = window.location.href) {
  return Object.fromEntries(
    new URL(url, window.location.origin).searchParams.entries()
  );
}

/**
 * @param {Record<string, string>} data
 * @returns {string}
 */
export function stringifyQueryString(data) {
  return '?' + serializeFormData(data);
}

/**
 * @param {Record<string, OneOrArray<string>>} data
 * @returns {string}
 *
 * @note Like with `URLSearchParams`, this function serializes spaces as plus
 * signs (`+`).
 */
export function serializeFormData(data) {
  const searchParams = new URLSearchParams();
  for (const k in data) {
    if (Object.hasOwn(data, k) && data[k] !== undefined) {
      for (const v of asArray($(data[k]))) {
        searchParams.append(k, v);
      }
    }
  }
  return searchParams.toString();
}

/**
 * @param {string | URL} url
 * @param {Record<string, string>} newQs
 * @returns {URL}
 */
export function updateUrlQueryString(url, newQs) {
  url = new URL(url, window.location.origin);
  for (let k in newQs) {
    if (Object.hasOwn(newQs, k)) {
      const v = newQs[k];
      if (v === undefined) {
        url.searchParams.delete(k);
      } else {
        url.searchParams.set(k, v);
      }
    }
  }
  return url;
}

/**
 * @param {string} url
 * @returns {string}
 */
export function urlEndingWithSlash(url) {
  if (url[url.length - 1] !== '/') {
    return url + '/';
  }
  return url;
}
