/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {wrapFn} from './fn_utils.js';
import {or} from './regexp.js';
import {$} from './typescript.js';

const URL_HASH_FACTOR = (Math.sqrt(5) - 1) / 2;

/**
 * @param {string} string
 * @returns {number}
 */
export function parseDecimal(string) {
  return parseInt(string, 10);
}

/**
 * @param {string} string
 * @returns {number}
 */
export function parseHex(string) {
  return parseInt(string, 16);
}

/**
 * @param {number} value
 * @param {number} min
 * @param {number} max
 * @returns {boolean} Whether or not `value` is in [`min`, `max`[
 */
export function inRange(value, min, max) {
  return min <= value && value < max;
}

/**
 * @param {number} value
 * @param {number} min
 * @param {number} max
 * @returns {boolean} Whether or not `value` is in [`min`, max]
 */
export function inRangeInclusive(value, min, max) {
  return min <= value && value <= max;
}

/**
 * @param {number} value
 * @param {number} min
 * @param {number} max
 * @returns {boolean} Whether or not `value` is in ]`min`, `max`[
 */
export function inRangeExclusive(value, min, max) {
  return min < value && value < max;
}

/**
 * @param {number} length
 * @returns {number[]} An array filled with `length` zeros
 */
export function zerosArray(length) {
  const array = [];
  for (; length > 9; length -= 10) {
    array.push(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  }
  for (; length > 0; length--) {
    array.push(0);
  }
  return array;
}

/**
 * @template T
 * @param {number} times
 * @param {T} element
 * @returns {T[]} An array filled with `times` `element`s
 */
export function repeat(times, element) {
  return Array.from({length: times}, () => element);
}

/**
 * @template T
 * @param {Array<T> | T} arrayOrElement
 * @returns {Array<T>}
 */
export function asArray(arrayOrElement) {
  return Array.isArray(arrayOrElement) ? arrayOrElement : [arrayOrElement];
}

/**
 * @template T
 * @param {T[]} destination
 * @param {readonly T[]} elements
 * @returns {T[]} An array resulting from the concatenation of both arguments.
 * @note The returned array is usually `destination` itself, but might be a new
 *   one.
 */
export function appendToArray(destination, elements) {
  // To avoid exceeding the maximum call stack size
  // we append the elements in chunks of size 100000
  if (elements.length < 100_000) {
    destination.push(...elements);
    return destination;
  }
  return destination.concat(elements);
}

/**
 * @overload
 * @param {null | undefined} array
 * @returns {null}
 *
 * @template T
 * @overload
 * @param {readonly T[]} array
 * @returns {FlatArray<readonly T[], 1>[]}
 *
 * @param {readonly T[] | null | undefined} array
 * @returns {FlatArray<readonly T[], 1>[] | null}
 */
export function flattenArray(array) {
  if (!array) {
    return null;
  }
  if (array.length === 0) {
    return [];
  }
  if (array.length === 1) {
    return (Array.isArray(array[0]) ? array[0] : array).slice();
  }
  if (array.flat) {
    return array.flat();
  }
  return flattenArraySlice(array);
}

/**
 * @overload
 * @param {null | undefined} array
 * @param {number} [begin = 0]
 * @param {number} [end = array.length]
 * @returns {null}
 *
 * @template T
 * @overload
 * @param {Readonly<ArrayLike<T>>} array
 * @param {number} [begin = 0]
 * @param {number} [end = array.length]
 * @returns {FlatArray<readonly T[], 1>[]}
 *
 * @param {Readonly<ArrayLike<T>> | null | undefined} array
 * @param {number} [begin = 0]
 * @param {number} [end = array.length]
 * @returns {FlatArray<readonly T[], 1>[] | null}
 */
export function flattenArraySlice(array, begin, end) {
  if (!array) {
    return null;
  }
  // @ts-expect-error -- Older browsers don't support Array#flat
  if (Array.prototype.flat) {
    return Array.prototype.slice.call(array, begin, end).flat();
  }

  const length = array.length;
  let start = begin || 0;
  start = (start < 0) ? Math.max(0, length + start) : start;
  let upTo = (typeof end === 'number') ? Math.min(end, length) : length;
  // @ts-expect-error -- undefined < 0 is false so its ok
  if (end < 0) {
    upTo = length + /** @type {number} */ (end);
  }

  /** @type {FlatArray<readonly T[], 1>[]} */
  let flat = [];
  /** @type {T} */
  let element;
  for (; start < upTo; start++) {
    element = $(array[start]);
    if (Array.isArray(element)) {
      if (element.length < 1_000) {
        flat.push(...element);
      } else {
        flat = flat.concat(element);
      }
    } else {
      flat.push(/** @type {FlatArray<readonly T[], 1>} */ (element));
    }
  }
  return flat;
}

/**
 * @template E
 * @template K
 * @template U
 * @template T
 *
 * @overload
 * @param {readonly E[]} array
 * @param {(this: T, element: E, index: number, array: readonly E[]) => K | K[]} getKeys
 * @param {NotFunction<T>} [context]
 * @returns {Map<K, OneOrMore<E>>}
 *
 * @overload
 * @param {readonly E[]} array
 * @param {(this: T, element: E, index: number, array: readonly E[]) => K | K[]} getKeys
 * @param {(this: T, element: E, index: number, array: readonly E[]) => U} mapFn
 * @param {T} [context]
 * @returns {Map<K, OneOrMore<U>>}
 *
 * @param {readonly E[]} array
 * @param {(this: T, element: E, index: number, array: readonly E[]) => K | K[]} getKeys
 * @param {((this: T, element: E, index: number, array: readonly E[]) => U) | NotFunction<T>} [mapFnOrContext]
 * @param {T} [context]
 * @returns {Map<K, OneOrMore<E>> | Map<K, OneOrMore<U>>}
 */
export function groupArrayElementsBy(array, getKeys, mapFnOrContext, context) {
  let mapFn;
  if (!context && typeof mapFnOrContext !== 'function') {
    context = mapFnOrContext;
  } else {
    mapFn = (
      /** @type {(this: T, element: E, index: number, array: readonly E[]) => U} */
      (mapFnOrContext)
    );
  }

  const groups = new Map();
  let i = 0;
  for (const element of array) {
    const keys = asArray(getKeys.call($(context), element, i, array));
    for (const key of keys) {
      const group = upsert(groups, key, []);
      group.push(mapFn ? mapFn.call($(context), element, i, array) : element);
    }
    i++;
  }
  return groups;
}

/**
 * @template {Map<any, any> | WeakMap<any, any> | Record<any, any>} M
 * @param {M} map
 * @param {Keys<M>} key
 * @param {Values<M>} defaultValue
 * @returns {Values<M>}
 */
export function upsert(map, key, defaultValue) {
  if (map instanceof Map || map instanceof WeakMap) {
    if (!map.has(key)) {
      map.set(key, defaultValue);
    }
    return map.get(key);
  }

  if (!(key in map)) {
    map[key] = defaultValue;
  }
  return map[key];
}

/**
 * @param {object | null | undefined} obj
 * @returns {boolean}
 */
export function isObjectEmpty(obj) {
  if (!obj) {
    return true;
  }
  //eslint-disable-next-line guard-for-in
  for (const _ in obj) {
    return false;
  }
  return true;
}

/**
 * @template TThis
 * @template {{}} T
 * @template U
 * @param {T | null | undefined} object
 * @param {(this: TThis, value: T[string & keyof T], key: keyof T, object: T) => U} callback
 * @param {TThis} [context]
 * @returns {T extends {} ? Record<keyof T, U> : null}
 */
export function mapObject(object, callback, context) {
  if (!object) {
    // @ts-expect-error
    return null;
  }
  /** @type {any} */
  const result = {};
  for (const name in object) {
    if (Object.hasOwn(object, name)) {
      result[name] = callback.call($(context), object[name], name, object);
    }
  }
  return result;
}

/**
 * @template TThis
 * @template {{}} T
 * @template U
 * @param {T | null | undefined} object
 * @param {(this: TThis, value: T[string & keyof T], key: keyof T, object: T) => U | U[]} callback
 * @param {TThis} [context]
 * @returns {T extends {} ? U[] : null}
 */
export function mapObjectToFlatArray(object, callback, context) {
  if (!object) {
    // @ts-expect-error
    return null;
  }
  /** @type {any} */
  const result = [];
  for (const name in object) {
    if (Object.hasOwn(object, name)) {
      const item = callback.call($(context), object[name], name, object);
      if (Array.isArray(item)) {
        result.push(...item);
      } else {
        result.push(item);
      }
    }
  }
  return result;
}

/**
 * @template TThis
 * @template {{}} T
 * @param {T | null | undefined} object
 * @param {(this: TThis, value: T[string & keyof T], key: keyof T, object: T) => boolean} callback
 * @param {TThis} [context]
 * @returns {T extends {} ? Partial<T> : null}
 */
export function filterObject(object, callback, context) {
  if (!object) {
    // @ts-expect-error
    return null;
  }
  /** @type {any} */
  const result = {};
  for (const name in object) {
    if (Object.hasOwn(object, name)) {
      const value = object[name];
      if (callback.call($(context), value, name, object)) {
        result[name] = value;
      }
    }
  }
  return result;
}

/**
 * @template TThis
 * @template {{}} T
 * @param {T | null | undefined} object
 * @param {(this: TThis, value: T[string & keyof T], key: keyof T, object: T) => boolean} callback
 * @param {TThis} [context]
 * @returns {T extends {} ? Values<T>[] : null}
 */
export function filterObjectToValuesArray(object, callback, context) {
  if (!object) {
    // @ts-expect-error
    return null;
  }
  /** @type {any} */
  const result = [];
  for (const name in object) {
    if (Object.hasOwn(object, name)) {
      const value = object[name];
      if (callback.call($(context), value, name, object)) {
        result.push(value);
      }
    }
  }
  return result;
}

/**
 * @template TThis
 * @template {{}} T
 * @param {T} object
 * @param {(this: TThis, value: T[string & keyof T], key: keyof T, object: T) => void} callback
 * @param {TThis} [context]
 * @returns {void}
 */
export function forEachObject(object, callback, context) {
  for (const name in object) {
    if (Object.hasOwn(object, name)) {
      callback.call(/** @type {TThis} */ (context), object[name], name, object);
    }
  }
}

/**
 * @template {{}} T
 * @param {T} object
 * @returns {{[P in T[keyof T]]: (keyof T)[]}}
 */
export function reverseObject(object) {
  /** @type {any} */
  const reverse = {};
  forEachObject(object, function(value, key) {
    if (reverse[value]) {
      reverse[value].push(key);
    } else {
      reverse[value] = [key];
    }
  });
  return reverse;
}

/**
 * @template T
 * @param {string} id
 * @param {T[]} urls
 * @returns {T}
 */
export function selectUrl(id, urls) {
  let hash = 1;
  for (let i = 0, len = id.length; i < len; i++) {
    hash *= id.charCodeAt(i) * URL_HASH_FACTOR;
    hash -= Math.floor(hash);
  }
  return $(urls[Math.floor(hash * urls.length)]);
}

/**
 * Returns a random number [`min`, `max`[.
 *
 * @param {number} min
 * @param {number} max
 * @returns {number} Returns a random number [`min`, `max`[.
 */
export function randomBetween(min, max) {
  return min + (Math.random() * (max - min));
}

/**
 * Returns a random integer [`min`, `max`[.
 *
 * @param {number} min
 * @param {number} max
 * @returns {number} Returns a random integer [`min`, `max`[.
 */
export function randomIntBetween(min, max) {
  return Math.floor(randomBetween(min, max));
}

/**
 * Generate a random `length` characters long string.
 *
 * @param {number} length - The length of the string to generate.
 * @param {string?} charset - A string containing the characters to use.
 * @returns {string} A random string of length `length`.
 */
export function randomString(length, charset) {
  charset = charset || randomString.charset;
  const charsetLength = charset.length;
  let text = '';

  while (length--) {
    text += charset.charAt(randomIntBetween(0, charsetLength));
  }

  return text;
}
randomString.charset = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';

/**
 * @param {string} string
 * @param {Record<string, Stringifyable>} context
 * @returns {string}
 */
export function interpolateString(string, context) {
  return string.replace(/([^\\])\$\{([^}]*)\}/g, function(_, firstChar, name) {
    return firstChar + (context[name] == null ? '' : context[name]);
  });
}

/**
 * @template {Error & {name?: unknown}} T
 * @param {T} error
 * @returns {T is Error & {name: 'AbortError'}}
 */
export function isAbortError(error) {
  return error.name === 'AbortError';
}

/**
 * @template {Error & {name?: unknown}} T
 * @param {T} error
 * @returns {T is Error & {name: 'TimeoutError'}}
 */
export function isTimeoutError(error) {
  return error.name === 'TimeoutError';
}

/**
 * @param {unknown} error
 * @returns {boolean}
 */
export function isOutOfMemoryError(error) {
  return (
    // Firefox
    error === 'out of memory' ||
    (
      error instanceof RangeError &&
      (
        // Chromium
        error.message === 'Array buffer allocation failed' ||
        // Safari
        error.message === 'Out of memory'
      )
    )
  );
}

/**
 * Inspired by the `AbortSignal.any()` proposal:
 * https://github.com/shaseley/abort-signal-any
 *
 * @param {AbortSignal[]} signals
 * @returns {AbortSignal}
 */
export function anyAbortSignal(signals) {
  const abortController = new AbortController();
  for (const signal of signals) {
    if (signal.aborted) {
      abortController.abort(signal.reason);
      break;
    }

    // eslint-disable-next-line func-style
    const handleAbort = () => abortController.abort(signal.reason);

    signal.addEventListener('abort', handleAbort);
    abortController.signal.addEventListener('abort', function() {
      signal.removeEventListener('abort', handleAbort);
    });
  }
  return abortController.signal;
}

/**
 * @template T
 * @param {(T | Falsy)[]} array
 * @returns {T[]}
 */
function truthy(array) {
  return /** @type {T[]} */ (array.filter(Boolean));
}

/**
 * @param {RequestInfo | URL} input
 * @param {RequestInit & {timeout?: number}} [init]
 * @returns {Promise<Response>}
 */
export function fetchWithTimeout(input, init = {}) {
  const {timeout, ...initWithoutTimeout} = init;

  if (isFinite(timeout) && timeout > 0) {
    initWithoutTimeout.signal = anyAbortSignal(
      truthy([init.signal, AbortSignal.timeout(timeout)])
    );
  }
  return fetch(input, initWithoutTimeout);
}

/**
 * @param {Error} error
 * @returns {never}
 * @throws `error`
 */
export function raise(error) {
  throw error;
}

/**
 * @overload
 * @param {undefined} value
 * @returns {'undefined'}
 *
 * @overload
 * @param {null} value
 * @returns {'null'}
 *
 * @overload
 * @param {boolean} value
 * @returns {'boolean'}
 *
 * @overload
 * @param {number} value
 * @returns {'number'}
 *
 * @overload
 * @param {bigint} value
 * @returns {'bigint'}
 *
 * @overload
 * @param {string} value
 * @returns {'string'}
 *
 * @overload
 * @param {symbol} value
 * @returns {'symbol'}
 *
 * @overload
 * @param {Func | Class} value
 * @returns {'function'}
 *
 * @overload
 * @param {unknown[]} value
 * @returns {'array'}
 *
 * @overload
 * @param {unknown} value
 * @returns {string}
 *
 * @param {unknown} value
 * @returns {string}
 */
export function typeOf(value) {
  const type = typeof value;
  return (
    value === null
    ? 'null'
    : (
      type === 'object'
      ? (
        Array.isArray(value)
        ? 'array'
        : (
          /** @type {{}} */ (value).constructor
          ? /** @type {{}} */ (value).constructor.name.toLowerCase()
          : type
        )
      )
      : type
    )
  );
}

/**
 * @template T
 * @param {T} value
 * @returns {T} A clone of the given `value` if it's an object, an array or a
 * Date, otherwise the given `value` as is.
 */
export function clone(value) {
  switch (typeOf(value)) {
  case 'object':
    return Object.assign({}, value);
  case 'array':
    return /** @type {T} */ ((/** @type {unknown[]} */ (value)).slice());
  case 'date':
    return /** @type {T} */ (new Date(/** @type {Date} */ (value)));
  default:
    return value;
  }
}

/**
 * @param {(target: Record<any, any>, value: any, key: string, source: Record<any, any>) => void} fn
 * @returns {(target: Record<any, any>, ...sources: any[]) => Record<any, any>}
 */
export function createAssigner(fn) {
  return (target, ...sources) => {
    for (const source of sources) {
      if (source) {
        for (let key in source) {
          if (Object.hasOwn(source, key)) {
            fn(target, source[key], key, source);
          }
        }
      }
    }
    return target;
  };
}
export const deepAssign = createAssigner((target, value, key) => {
  const type = typeOf(value);
  if (type === 'object' && type === typeOf(target[key])) {
    deepAssign(target[key], value);
  } else {
    target[key] = (type === 'object' ? Object.assign({}, value) : value);
  }
});

/**
 * @param {object} target
 * @param {any[]} sources
 * @returns {object}
 */
export function assignWithOwnProperties(target, ...sources) {
  for (const source of sources) {
    Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
  }
  return target;
}

/**
 * @template {{}} T
 * @param {T} object
 * @returns {T}
 */
export function memoGetters(object) {
  const propertyDescriptors = Object.getOwnPropertyDescriptors(object);
  // eslint-disable-next-line guard-for-in
  for (const key in propertyDescriptors) {
    const propertyDescriptor = $(propertyDescriptors[key]);
    if (propertyDescriptor.get && !propertyDescriptor.set) {
      propertyDescriptor.get = wrapFn(propertyDescriptor.get, function() {
        const value = window.original();
        propertyDescriptor.value = value;
        delete propertyDescriptor.set;
        delete propertyDescriptor.get;
        Object.defineProperty(this, key, propertyDescriptor);
        return value;
      });
      Object.defineProperty(object, key, propertyDescriptor);
    }
  }
  return object;
}

/**
 * @param {unknown} a
 * @param {unknown} b
 * @param {number} [depth]
 * @param {string[] | null} [keys]
 * @returns {boolean}
 */
export function equals(a, b, depth = 1, keys = null) {
  if (
    Object.is(a, b) ||
    (
      a instanceof Date &&
      b instanceof Date &&
      a.getTime() === b.getTime()
    )
  ) {
    return true;
  }

  if (depth < 1) {
    return false;
  }

  if (keys?.length === 1) {
    // @ts-expect-error keys[0] is expected to be a string
    return equals(a?.[keys[0]], b?.[keys[0]], depth - 1);
  }

  if (typeof a !== 'object' || !a || typeof b !== 'object' || !b) {
    return false;
  }

  if (!keys?.length) {
    keys = Object.keys(a);
    if (keys.length !== Object.keys(b).length) {
      return false;
    }
  }

  for (const key of keys) {
    if (
      !equals(
        /** @type {{[x: string]: unknown}} */ (a)[key],
        /** @type {{[x: string]: unknown}} */ (b)[key],
        depth - 1
      )
    ) {
      return false;
    }
  }
  return true;
}

/**
 * @param {any} a
 * @param {any} b
 * @param {string[]?} [keys]
 * @returns {boolean}
 */
export function shallowEqual(a, b, keys) {
  return equals(a, b, 1, keys);
}

const reEscapeChar = /\\([^u])/ig;
const reEscapeUnicode = /\\u([0-1a-f]{4})/ig;
const reJSProp = /^[_$a-zA-Z\xA0-\uFFFF][_$a-zA-Z0-9\xA0-\uFFFF]*$/;
const reSplitIndexPath = new RegExp(
  or(
    // bare key
    '([^.[]+)',
    // key in brackets
    '\\[' + or(
      '((?:[^"\'`][^\\[]*)?)',
      // in optionally in quotes (with support for escapes)
      '(["\'`])((?:(?!\\3)[^\\\\]|\\\\(?:[^u]|u[0-1a-f]{4}))*?)\\3'
    ) + '\\]',
    // empty key (like "..", ".[" or ".$")
    '\\.(?=\\.|\\[|$)'
  ),
  'ig'
);

/**
 * @param {string} _
 * @param {string} hexCode
 * @returns {string}
 */
function reEscapeUnicodeReplacer(_, hexCode) {
  return String.fromCharCode(parseInt(hexCode, 16));
}

/**
 * @param {string} indexPath
 * @param {'keep' | 'skip' | 'parent'} [onEmptyKey]
 * @returns {string[]}
 */
export function splitIndexPath(indexPath, onEmptyKey = 'keep') {
  const path = [];
  let match;
  reSplitIndexPath.lastIndex = 0;
  while ((match = reSplitIndexPath.exec(indexPath))) {
    let [, bare, key = bare?.trim(), quote, stringWithEscapes] = match;
    if (quote) {
      key = (
        $(stringWithEscapes)
          .replace(reEscapeChar, '$1')
          .replace(reEscapeUnicode, reEscapeUnicodeReplacer)
      );
    }
    if (key) {
      path.push(key);
    } else {
      switch (onEmptyKey) {
      case 'keep':
        path.push('');
        break;
      case 'parent':
        path.pop();
        break;
      default:
      }
    }
  }
  return path;
}

/**
 * Appends a given index to a given index path.
 *
 * @param {string} indexPath The index path to append to
 * @param {number | string} index The index to append
 * @returns {string} The complete index path
 */
export function appendToIndexPath(indexPath, index) {
  return indexPath + (
    typeof index === 'string' && reJSProp.test(index)
      ? (indexPath ? '.' : '') + index
      : `[${JSON.stringify(index)}]`
  );
}
/**
 * @param {any} object
 * @param {string} indexPath
 * @param {'keep' | 'skip' | 'parent'} [onEmptyKey]
 * @returns {unknown}
 */
export function getPropAtIndexPath(object, indexPath, onEmptyKey = 'keep') {
  if (!indexPath) {
    return object;
  }
  if (!object) {
    return undefined;
  }
  const path = splitIndexPath(indexPath, onEmptyKey);
  let i = 0;
  let l = path.length;
  while (object != null && i < l) {
    object = object[/** @type {string} */ (path[i++])];
  }
  return (i && i === l) ? object : undefined;
}
