/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  wrapFn,
  inverseLerp,
  asArray,
  mapObject,
  Dataset,
  Types as is,
} from '../../syntool.js';
import {colorFromCSSColor, colorToCSSColor} from '../../utils/color.js';
import {colorPaletteGetColor} from '../../utils/color_palette.js';

const DataField = is.shape({
  channel: is.oneOfType([is.string, is.number]),
  min    : is.number.isRequired,
  max    : is.number.gte('min').isRequired,
});
const DataFieldRef = is.lazy(function(props, propName, propFullName, getProp) {
  const dataFields = getProp(propFullName + '...dataFields');
  const fieldNames = Object.keys(dataFields);
  return is.oneOf(fieldNames);
});

class RenderValue {
  constructor(value, min = NaN, max = NaN) {
    this.value = value;
    this.min = min;
    this.max = max;
  }
  get normalizedValue() {
    return (
      isNaN(this.min) || isNaN(this.max)
      ? undefined
      : inverseLerp(this.min, this.max, this.value)
    );
  }
  valueOf() {
    return this.value;
  }
  color(colorPalette) {
    if (!colorPalette) {
      return undefined;
    }
    const {normalizedValue} = this;
    return (
      normalizedValue === undefined
      ? colorFromCSSColor(this.value)
      : colorPaletteGetColor(colorPalette, normalizedValue)
    );
  }
  cssColor(colorPalette) {
    if (!colorPalette) {
      return undefined;
    }
    const {normalizedValue} = this;
    return (
      normalizedValue === undefined
      ? this.value
      : colorToCSSColor(this.color(colorPalette))
    );
  }
}

class ProductRenderField {
  constructor(dataFields, {dataFieldName, value}) {
    this.dataFields = dataFields;
    this.dataFieldName = dataFieldName;
    this.value = value;
  }

  get isUniform() {
    return this.dataFieldName == null;
  }
  get dataField() {
    return (this.isUniform ? undefined : this.dataFields[this.dataFieldName]);
  }

  getDataValue(dataValues, defaultValue) {
    if (this.isUniform) {
      return defaultValue;
    }

    const {channel} = this.dataField;
    return (
      dataValues && Object.hasOwn(dataValues, channel)
      ? dataValues[channel]
      : defaultValue
    );
  }
  getRenderValue(dataValues) {
    if (this.isUniform) {
      return new RenderValue(this.value);
    }

    const {channel, min, max} = this.dataField;
    if (dataValues && Object.hasOwn(dataValues, channel)) {
      return new RenderValue(dataValues[channel], min, max);
    }

    return undefined;
  }
  getNormalizedValue(dataValues, defaultValue) {
    if (this.isUniform) {
      return this.value;
    }

    const {channel, min, max} = this.dataField;
    return (
      dataValues && Object.hasOwn(dataValues, channel)
      ? inverseLerp(min, max, dataValues[channel])
      : defaultValue
    );
  }
  getValueAsColor(colorPalette, dataValues, defaultColor) {
    if (this.isUniform) {
      return colorFromCSSColor(this.value);
    }
    if (!colorPalette) {
      return defaultColor;
    }

    const value = this.getNormalizedValue(dataValues);
    return (
      value === undefined
      ? defaultColor
      : colorPaletteGetColor(colorPalette, value)
    );
  }
  getValueAsCSSColor(colorPalette, dataValues, defaultColor) {
    if (this.isUniform) {
      return this.value;
    }
    return colorToCSSColor(
      this.getValueAsColor(colorPalette, dataValues, defaultColor)
    );
  }
}

function normalizeProductConfig(productConfig) {
  productConfig.dataFields = Object.fromEntries(
    Object.entries(productConfig.dataFields).map(([fieldName, field]) => [
      fieldName,
      (
        Object.hasOwn(field, 'channel')
        ? field
        : Object.assign({channel: fieldName}, field)
      ),
    ])
  );
  productConfig.renderFields = Object.fromEntries(
    Object.entries(productConfig.renderFields).map(([fieldName, field]) => [
      fieldName,
      new ProductRenderField(
        productConfig.dataFields,
        (
          (
            typeof field === 'string' &&
            Object.hasOwn(productConfig.dataFields, field)
          )
          ? {dataFieldName: field}
          : {value: field}
        )
      ),
    ])
  );
  return productConfig;
}

export const dataFieldsExtension = {
  meta: {name: 'DataFieldsExtension'},
  static() {
    const validateConfig = is.shape({
      dataFields: is.objectOf(DataField.isRequired).isRequired,
      renderFields: is.lazy(() => (
        this.renderFields
        ? is.shape(
          mapObject(
            this.renderFields,
            (renderField) => is.oneOfType(
              asArray(renderField)
              .map((type) => (type === 'ref' ? DataFieldRef : type))
            )
          )
        ).isRequired
        : is.none
      )),
    });

    this.validateConfig = (
      this.validateConfig
      ? is.allOfType([this.validateConfig, validateConfig])
      : validateConfig
    );
    this.normalizeProductConfig = (
      this.normalizeProductConfig
      ? wrapFn(
        this.normalizeProductConfig,
        (productConfig) => window.original(normalizeProductConfig(productConfig))
      )
      : normalizeProductConfig
    );
  },

  _renderFields: null,
  get renderFields() {
    if (!this._renderFields) {
      const cls = this.constructor;
      const product = this.getProduct();
      const renderFields = {};
      for (const fieldName in cls.renderFields) {
        if (cls.renderFields && Object.hasOwn(cls.renderFields, fieldName)) {
          const field = product.renderFields[fieldName];
          if (field) {
            renderFields[fieldName] = field;
          }
        }
      }
      this._renderFields = renderFields;
    }
    return this._renderFields;
  },

  getUniformFieldValue(fieldName, defaultValue = false) {
    const field = this.renderFields[fieldName];
    return (field?.isUniform ? field.value : defaultValue);
  },
  getReferencedFieldName(fieldName, defaultName = '') {
    const field = this.renderFields[fieldName];
    return field?.dataFieldName ?? defaultName;
  },
  getReferencedField(fieldName, defaultValue = false) {
    const field = this.renderFields[fieldName];
    return field?.dataField ?? defaultValue;
  },

  getRenderValues(dataValues) {
    return mapObject(
      this.renderFields,
      (field) => field.getRenderValue(dataValues)
    );
  },
};
Object.assign(Dataset.extensions, {dataFields: dataFieldsExtension});
