/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  SECOND_IN_MILLISECONDS,
  emptyFunction,
  isAbortError,
  ajaxRetriable,
  fetchResource,
  Dataset,
  NotificationMessageType,
} from '../../syntool.js';
import {auxiliaryDataExtension} from './auxiliary_data_extension.js';

export const dataLoadingExtension = {
  meta: {
    name: 'AuxiliaryDataLoadingExtension',
    deps: [auxiliaryDataExtension],
  },

  initialize(info) {
    window.original();
    this.requestDataWhenInMap = false;
    this.dataRequested = false;
    this.cancelLoadData = null;
    this.dataLoadErrorMessage = {
      type: NotificationMessageType.error,
      text: (
        'Could not load dataset "' + this.datasetId + '" due to server or ' +
        'network errors!'
      ),
      buttons: [
        {
          text   : 'retry',
          closes : true,
          onClick: this.reloadData.bind(this),
        },
      ],
      hideAfter: 10 * SECOND_IN_MILLISECONDS,
    };
    this.dataLoadParseErrorMessage = {
      type: NotificationMessageType.error,
      text: 'Failed to load dataset "' + this.datasetId + '"!',
      hideAfter: 10 * SECOND_IN_MILLISECONDS,
    };
  },
  destroy() {
    this.dataRequested = false;
    window.original();
  },
  didAddToMap() {
    window.original();
    if (this.requestDataWhenInMap) {
      this.loadData();
    }
  },
  willRemoveFromMap() {
    window.original();
    if (this.requestDataWhenInMap && this.cancelLoadData) {
      this.cancelLoadData();
    }
  },
  handleVisibilityChange() {
    window.original();
    if (!this.requestDataWhenInMap) {
      if (this.isVisible()) {
        this.loadData();
      } else if (this.cancelLoadData) {
        this.cancelLoadData();
      }
    }
  },

  dataLoader: null,
  getDataURL() {
    return this.getAuxiliaryDataURL();
  },
  async loadData() {
    if (!this.dataLoader || this.dataRequested) {
      return;
    }

    var [dataPromise, abortFetch] = fetchResource(
      ajaxRetriable(this.dataLoader),
      this.getDataURL()
    );

    this.dataRequested = true;
    this.cancelLoadData = () => {
      abortFetch();
      this.dataRequested = false;
      this.cancelLoadData = null;
    };

    this.hub.emit('data-load:start');
    try {
      const data = await dataPromise;
      this.cancelLoadData = null;
      try {
        await this.handleDataLoad(data);
      } catch (error) {
        console.error('Dataset(' + this.datasetId + ') failed to load data', error);
        this.syntool.notificationCenter.showMessage(this.dataLoadParseErrorMessage);
      }
    } catch (error) {
      this.dataRequested = false;
      this.cancelLoadData = null;
      if (!isAbortError(error)) {
        console.error(
          'Dataset(' + this.datasetId + ') could not load data',
          error
        );
        this.syntool.notificationCenter.showMessage(this.dataLoadErrorMessage);
        await this.handleDataLoadError(error);
      }
    } finally {
      this.hub.emit('data-load:end');
    }
  },
  reloadData() {
    // Retry loading the data
    if (this.cancelLoadData) {
      this.cancelLoadData();
    }
    if (this.syntool.mapView.isDatasetInMap(this)) {
      this.loadData();
    }
  },
  handleDataLoadError: emptyFunction,
  handleDataLoad     : emptyFunction,
};

Object.assign(Dataset.extensions, {dataLoading: dataLoadingExtension});
