/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {ajax, LatLonBounds, Bounds, Dataset} from '../../syntool.js';
import {dataLoadingExtension} from './data_loading_extension.js';
import {dataFieldsExtension} from './data_fields_extension.js';
import {VectorDataMatrix} from './vector_data_matrix.js';

export const dataMatrixExtension = {
  meta: {
    name: 'AuxiliaryDataMatrixExtension',
    deps: [dataLoadingExtension, dataFieldsExtension],
  },

  dataLoader: ajax.getPNG,

  initialize(info) {
    window.original();
    this.dataMatrix = null;
  },
  hasResources() {
    return window.original() || !!this.dataMatrix;
  },
  destroy() {
    this.dataMatrix = null;
    window.original();
  },

  getValuesAtProjectedXY(xy, worldBounds, fieldNames) {
    if (this.dataMatrix) {
      const pixel = this.dataMatrix.indexFromProjectedXY(
        xy,
        worldBounds ? Bounds.getLeft(worldBounds) : NaN,
        worldBounds ? Bounds.getRight(worldBounds) : NaN
      );
      if (!pixel) {
        return null;
      }

      const uniformFields = [];
      const dataFieldNames = [];

      for (const fieldName of fieldNames) {
        uniformFields.push(this.getUniformFieldValue(fieldName));
        dataFieldNames.push(this.getReferencedFieldName(fieldName));
      }

      const values = this.dataMatrix.getRealValuesAt(
        pixel[0],
        pixel[1],
        dataFieldNames
      );
      if (!values) {
        return null;
      }

      for (let i = 0, l = uniformFields.length; i < l; i++) {
        if (uniformFields[i] !== false) {
          values[i] = uniformFields[i];
        }
      }
      return values;
    }
    return null;
  },
  handleDataLoad(response) {
    window.original();
    const wraps = (
      this.syntool.mapView.hasMercatorLikeProjection() &&
      LatLonBounds.getWidth(
        this.syntool.mapView.getLonLatBoundsFromProjectedBounds(this.bounds)
      ) === 360
    );
    response.metadata.north = Bounds.getTop(this.bounds);
    response.metadata.east = Bounds.getRight(this.bounds);
    response.metadata.south = Bounds.getBottom(this.bounds);
    response.metadata.west = Bounds.getLeft(this.bounds);
    this.dataMatrix = new VectorDataMatrix(
      this.getProduct().dataFields,
      response.metadata,
      response.pixels,
      null,
      wraps
    );
  },
};

Object.assign(Dataset.extensions, {dataMatrix: dataMatrixExtension});
