/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {emptyFunction, ajax, GeoJSON, Dataset} from '../../syntool.js';
import {COLOR_WHITE} from '../../utils/color.js';
import {auxiliaryDataExtension} from './auxiliary_data_extension.js';
import {colorPaletteExtension} from './color_palette_extension.js';
import {dataLoadingExtension} from './data_loading_extension.js';
import {dataFieldsExtension} from './data_fields_extension.js';

export const geoJsonExtension = {
  meta: {
    name: 'GeoJSONAuxiliaryDataExtension',
    deps: [
      auxiliaryDataExtension,
      colorPaletteExtension,
      dataLoadingExtension,
      dataFieldsExtension,
    ],
  },

  getDataURL() {
    return this.getAuxiliaryDataURL('geojson.json');
  },

  dataLoader: ajax.getJSON,
  handleDataLoad(json) {
    window.original();

    const features = GeoJSON.getFeatures(json, {
      sourceProjection: json?.crs?.properties?.name,
      targetProjection: this.syntool.mapView.getProjection(),

      // Enable loose mode to allow for LineString geometries with a single
      // point in the GeoJSON.
      // This invalid geometry was observed in datasets of the products:
      // * "Drifters surface (SPLASH-CARTHE)" (3857_CARTHE_Splash_drifters)
      //   notable dataset with ID "0325_2017-04-25"
      // * "ARGO profilers 2002/2016" (900913_ARGO_profilers) notable dataset
      //   with ID "6901245_22"
      loose: true,
    });

    for (const feature of features) {
      feature.dataset = this;
    }

    return this.handleFeaturesLoad(features);
  },

  handleFeaturesLoad: emptyFunction,

  updateFeaturesStyle(features, styledCallback) {
    const colorField = this.renderFields.color;
    let updatedStyles = false;

    if (colorField && !colorField.isUniform) {
      this.withColorPalette(() => {
        for (const feature of features) {
          if (this.updateFeatureStyle(feature)) {
            feature.layer?.drawFeature(feature);
          }
        }
        updatedStyles = true;
        styledCallback?.();
      });
    }
    if (!updatedStyles) {
      for (const feature of features) {
        if (this.updateFeatureStyle(feature)) {
          feature.layer?.drawFeature(feature);
        }
      }
    }
    if (!colorField || colorField.isUniform) {
      styledCallback?.();
    }
  },
  updateFeatureStyle(feature) {
    const colorField = this.renderFields.color;
    if (!colorField) {
      return false;
    }

    let values = feature.properties.fields;
    if (Array.isArray(values)) {
      values = values[0];
    }

    const color = colorField.getValueAsCSSColor(
      this.colorPalette,
      values,
      COLOR_WHITE
    );
    const style = feature.style || {};
    if (style.strokeColor !== color || style.fillColor !== color) {
      style.strokeColor = color;
      style.fillColor = color;
      feature.style = style;
      return true;
    }
    return false;
  },
};

Object.assign(Dataset.extensions, {geoJson: geoJsonExtension});
