/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Z_INDEX, round, Dataset} from '../../syntool.js';
import {OpenLayers} from '../../vendors/openlayers.js';
import {LayerManager} from '../layer_managers/layer_manager.js';

export const layerExtension = {
  meta: {name: 'LayerExtension'},
  static: {
    BASE_Z_INDEX: Z_INDEX.DATASETS,
  },

  initialize(info) {
    window.original();
    this.layerManager = this.createLayerManager();
    this.layerManager.setDataset(this);
    this.layerManager.setVisibility(this, this.isVisible());
    this.setOpacity(this.getProduct().opacity);
  },
  createLayerManager() {
    return new LayerManager(() => this.createLayer(), this.syntool.mapView);
  },
  createLayer() {
    return new OpenLayers.Layer(this.name);
  },
  getOpacity() {
    return this.layerManager.getOpacity(this);
  },
  setOpacity(opacity) {
    opacity = round(opacity, -3);
    if (this.getOpacity() !== opacity) {
      const wasVisible = this.isVisible();
      this.layerManager.setOpacity(this, opacity);
      this.hub.emit('change:opacity');
      if (this.isVisible() !== wasVisible) {
        this.handleVisibilityChange();
      }
    }
  },
  getStackLevel() {
    var zIndex = this.layerManager.getZIndex(this);
    if (zIndex < 0) {
      return window.original();
    }
    return zIndex - this.constructor.BASE_Z_INDEX;
  },
  setStackLevel(stackLevel) {
    if (stackLevel !== this.getStackLevel()) {
      this.layerManager.setZIndex(
        this,
        stackLevel + this.constructor.BASE_Z_INDEX
      );
      this.hub.emit('change:stackLevel');
    }
  },
  isVisible() {
    return window.original() && this.getOpacity() > 0;
  },
  handleVisibilityChange() {
    window.original();
    this.layerManager.setVisibility(this, this.isVisible());
  },
  didAddToMap() {
    window.original();
    this.layerManager.setInMap(this, true);
  },
  willRemoveFromMap() {
    window.original();
    this.layerManager.setInMap(this, false);
  },
  destroy() {
    window.original();
    this.layerManager.destroy(this);
  },
};

Object.assign(Dataset.extensions, {layer: layerExtension});
