/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  SECOND_IN_MILLISECONDS,
  validateValueType,
  Config,
  Dataset,
  NotificationMessageType,
} from '../../syntool.js';

class DatasetGranuleInfoGroup {
  constructor(dataset, id, group, config) {
    this._dataset = dataset;
    this._group = group;
    this._metadata = undefined;
    this.id = id;
    this.config = config;
    this._configuredResources = new Map();
    this._configuredMetadataFields = new Map();
  }

  get rawGroup() {
    return this._group;
  }
  get metadata() {
    if (this._metadata === undefined) {
      this._metadata = (
        Object.assign({}, this.config.metadataDefaults, this._group.metadata)
      );
    }
    return this._metadata;
  }

  getConfiguredMetadataField(key, value = this.metadata[key]) {
    if (value === undefined) {
      return undefined;
    }

    let configuredMetadataField = this._configuredMetadataFields.get(key);
    if (configuredMetadataField) {
      return configuredMetadataField;
    }

    const metadata = this.metadata;
    const config = this.config.metadata;
    if (config) {
      if (typeof config === 'function') {
        const fieldConfig = config(value, key, metadata, this._dataset);
        const error = validateValueType(
          Config.types.GranuleInfoMetadataFieldConfig,
          'metadata()',
          fieldConfig
        );

        if (error) {
          console.error(
            `Ignoring granule info metadata field "${key}" for`,
            `Dataset(${this._dataset.datasetId}).`,
            error,
            {this: this, fn: config, returned: fieldConfig}
          );
        } else if (fieldConfig !== undefined) {
          configuredMetadataField = {value, config: fieldConfig};
        }
      } else {
        const fieldConfig = config[key];

        if (fieldConfig !== undefined) {
          configuredMetadataField = {value, config: {label: fieldConfig}};
        }
      }
    }

    this._configuredMetadataFields.set(key, configuredMetadataField);
    return configuredMetadataField;
  }
  *configuredMetadataFields() {
    const metadata = this.metadata;
    const config = this.config.metadata;
    if (config) {
      for (const [key, value] of Object.entries(metadata)) {
        const configuredMetadataField = this.getConfiguredMetadataField(key, value);
        if (configuredMetadataField) {
          yield [key, configuredMetadataField];
        }
      }
    }
  }

  getConfiguredResource(resourceId, resource = this._group[resourceId]) {
    if (resourceId === 'metadata') {
      return undefined;
    }

    let configuredResource = this._configuredResources.get(resourceId);
    if (configuredResource) {
      return configuredResource;
    }

    if (resource) {
      const config = this.config;
      let resourceConfig = config.resources?.[resourceId];
      if (!resourceConfig && typeof config.unknownResource === 'function') {
        resourceConfig = config.unknownResource(this._dataset, resource, resourceId);

        const error = validateValueType(
          Config.types.GranuleInfoResourceConfig,
          'unknownResource()',
          resourceConfig
        );

        if (error) {
          console.error(
            `Ignoring granule info resource "${resourceId}" for`,
            `Dataset(${this._dataset.datasetId}).`,
            error,
            {this: this, fn: config.unknownResource, returned: resourceConfig}
          );
          resourceConfig = undefined;
        }
      }
      if (resourceConfig) {
        configuredResource = {resource, config: resourceConfig};
      }
    }

    this._configuredResources.set(resourceId, configuredResource);
    return configuredResource;
  }
  *configuredResources() {
    for (const [resourceId, resource] of Object.entries(this._group)) {
      const configuredResource = this.getConfiguredResource(resourceId, resource);
      if (configuredResource) {
        yield [resourceId, configuredResource];
      }
    }
  }
}

class DatasetGranuleInfo {
  constructor(dataset, granuleInfo, config) {
    this._dataset = dataset;
    this._granuleInfo = granuleInfo;
    this._config = config;
    this._warned = false;
    this._configuredGroups = new Map();
  }

  get rawGroups() {
    return this._granuleInfo;
  }
  get config() {
    const config = this._config;
    if (typeof config === 'function') {
      this._config = config(this._dataset, this._granuleInfo);

      const error = validateValueType(
        Config.types.GranuleInfoConfig,
        'granuleInfoConfig()',
        this._config
      );
      if (error) {
        console.error(
          `Ignoring granule info for Dataset(${this._dataset.datasetId}).`,
          error,
          {this: this, fn: config, returned: this._config},
        );
        this._config = undefined;
        this._warned = true;
      }
    }
    if (!this._config && !this._warned) {
      console.warn(
        `Ignoring granule info for Dataset(${this._dataset.datasetId}),`,
        'granule info config missing or returned falsy',
        {granuleInfo: this._granuleInfo, config}
      );
      this._warned = true;
    }
    return this._config;
  }
  get autoResourceId() {
    return this.config?.autoResourceId;
  }

  getConfiguredGroup(groupId, group = this._granuleInfo[groupId]) {
    let configuredGroup = this._configuredGroups.get(groupId);
    if (configuredGroup) {
      return configuredGroup;
    }

    const config = this.config;
    if (group && config) {
      let groupConfig = config.groups?.[groupId];
      if (!groupConfig && typeof config.unknownGroup === 'function') {
        groupConfig = config.unknownGroup(this._dataset, group, groupId);

        const error = validateValueType(
          Config.types.GranuleInfoGroupConfig,
          'unknownGroup()',
          groupConfig
        );
        if (error) {
          console.error(
            `Ignoring granule info group "${groupId}" for`,
            `Dataset(${this._dataset.datasetId}).`,
            error,
            {this: this, fn: config.unknownGroup, returned: groupConfig}
          );
          groupConfig = undefined;
        }
      }
      if (groupConfig) {
        configuredGroup = new DatasetGranuleInfoGroup(
          this._dataset,
          groupId,
          group,
          groupConfig
        );
      }
    }

    this._configuredGroups.set(groupId, configuredGroup);
    return configuredGroup;
  }
  *configuredGroups() {
    if (this.config) {
      for (const [groupId, group] of Object.entries(this._granuleInfo)) {
        const configuredGroup = this.getConfiguredGroup(groupId, group);
        if (configuredGroup) {
          yield [groupId, configuredGroup];
        }
      }
    }
  }

  getConfiguredResourceWithId(absoluteId) {
    const [groupId, resourceId] = absoluteId.split('/');
    const group = this.getConfiguredGroup(groupId);
    return group?.getConfiguredResource(resourceId);
  }
}

export const metadataExtension = {
  meta: {name: 'GetMetadataExtension'},
  initialize(info) {
    window.original();
    this._metadataPromise = null;
    this.metadataErrorMessage = {
      type     : NotificationMessageType.error,
      text     : (
        'Could not load metadata for dataset "' + this.datasetId +
        '" due to server or network errors!'
      ),
      hideAfter: 10 * SECOND_IN_MILLISECONDS,
    };
  },
  getMetadata() {
    if (!this._metadataPromise) {
      this._metadataPromise = (
        this.syntool.api.getDatasetMetadata(this)
        .then((granuleInfo) => (
          granuleInfo && Object.keys(granuleInfo).length > 0
          ? new DatasetGranuleInfo(
            this,
            granuleInfo,
            this.getProduct().granuleInfoConfig
          )
          : null
        ))
      );
      this._metadataPromise.catch((error) => {
        this._metadataPromise = null;
        console.error(
          'Dataset(' + this.datasetId + ') could not load metadata',
          error
        );
        this.syntool.notificationCenter.showMessage(this.metadataErrorMessage);
      });
    }
    return this._metadataPromise;
  },
};

Object.assign(Dataset.extensions, {metadata: metadataExtension});
