/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {inRange, createClass, modulo, wrap, Bounds} from '../../syntool.js';

var NO_DATA_VALUE = 0xFF;

/**
 * @typedef {Object} VectorDataMatrix~Field
 * @property {number} channel
 * @property {number} min
 * @property {number} max
 */

/**
 * @typedef {Object} VectorDataMatrix~VectorData
 * @property {Object<number>} raw
 * @property {?Object<number>} real
 */

/**
 * @typedef {Object} VectorDataMatrix~Metadata
 * @property {number} width
 * @property {number} height
 * @property {number} north
 * @property {number} east
 * @property {number} south
 * @property {number} west
 * @property {number} channels
 * @property {?number} nodataValue
 * @property {number[]} offsets
 * @property {number[]} scales
 */

/**
 * @constructor
 * @param {Object<VectorDataMatrix~Field>} fields
 * @param {VectorDataMatrix~Metadata} metadata
 * @param {Uint8Array} data
 * @param {?Bounds} bounds
 */
export const VectorDataMatrix = createClass([
  /*Public*/ {
    constructor: function VectorDataMatrix(fields, metadata, data, bounds, wraps) {
      var leftOffset = 0;
      var topOffset = 0;
      var width = metadata.width;
      var height = metadata.height;

      if (bounds) {
        var boundsTop = Bounds.getTop(bounds);
        var boundsBottom = Bounds.getBottom(bounds);
        var boundsLeft = Bounds.getLeft(bounds);
        var boundsRight = Bounds.getRight(bounds);
        var scaleh = (metadata.east  - metadata.west)  / metadata.width;
        var scalev = (metadata.north - metadata.south) / metadata.height;

        leftOffset = Math.floor((boundsLeft - metadata.west)  / scaleh);
        topOffset  = Math.floor(-(boundsTop - metadata.north) / scalev);
        width  = Math.ceil((boundsRight   - metadata.west)  / scaleh) - leftOffset;
        height = Math.ceil(-(boundsBottom - metadata.north) / scalev) - topOffset;
        bounds = {
          left  : boundsLeft,
          right : boundsRight,
          top   : boundsTop,
          bottom: boundsBottom,
          width : boundsRight - boundsLeft,
          height: boundsTop - boundsBottom,
        };
      } else {
        bounds = {
          left  : metadata.west,
          right : metadata.east,
          top   : metadata.north,
          bottom: metadata.south,
          width : metadata.east - metadata.west,
          height: metadata.north - metadata.south,
        };
      }

      // The only acceptable nodataValue is NO_DATA_VALUE
      if (
        !isNaN(metadata.nodataValue) &&
        metadata.nodataValue !== NO_DATA_VALUE
      ) {
        throw new Error(
          'VectorDataMatrix: Only supported value for nodataValue is ' +
          NO_DATA_VALUE + ', but found `' + metadata.nodataValue + '`.'
        );
      }

      this.fields = fields;
      this.metadata = metadata;
      this.hasMaskedData = !isNaN(metadata.nodataValue);
      this.data = data;
      this.bounds = bounds;
      this.wraps = !!wraps;
      this._topOffset = topOffset;
      this._leftOffset = leftOffset;
      this._width = width;
      this._height = height;
      this._fieldNames = Object.keys(this.fields);
      this._tempRawValuesArray = new Uint8Array(metadata.channels);
    },

    /**
     * @return {number} Returns the width in pixels.
     */
    getWidth() {
      return this._width;
    },

    /**
     * @return {number} Returns the height in pixels.
     */
    getHeight() {
      return this._height;
    },

    /**
     * @param {number} i - Zero based column index.
     * @param {number} j - Zero based row index.
     * @param {?number[]} values - The array to store the raw values to return.
     * @return {?number[]} Returns the raw values stored in the matrix at the given index or `null`.
     */
    getRawValuesAt(i, j, values) {
      if (this.wraps) {
        i = modulo(i, this._width);
      }
      if (
        !this.data ||
        !inRange(i, 0, this._width) ||
        !inRange(j, 0, this._height)
      ) {
        return null;
      }

      i += this._leftOffset;
      j += this._topOffset;

      var dataIndex = (j * this.metadata.width + i) * this.metadata.channels;
      values = values || new Uint8Array(this.metadata.channels);
      for (var channelId = 0; channelId < this.metadata.channels; ++channelId) {
        values[channelId] = this.data[dataIndex + channelId];
      }
      return values;
    },

    /**
     * @param {number} i - Zero based column index.
     * @param {number} j - Zero based row index.
     * @param {string} fieldName - The name of the field for which the value is returned.
     * @return {number} Returns the value for the given field at the given index or `NaN`.
     */
    getRealValueAt(i, j, fieldName) {
      var rawValues = this.getRawValuesAt(i, j, this._tempRawValuesArray);

      if (!rawValues) {
        return NaN;
      }

      // Check if all rawValues are NO_DATA_VALUE
      // In which case we have no data at this location and should return `NaN`
      if (this.hasMaskedData) {
        i = rawValues.length;
        while (i-- && rawValues[i] === NO_DATA_VALUE) {
          /*empty*/
        }
        if (i === -1) {
          // all rawValues are NO_DATA_VALUE
          return NaN;
        }
      }

      var field = this.fields[fieldName];
      var fieldChannel = field.channel;
      var rawValue = rawValues[fieldChannel];

      // If we have data at this location and the rawValue is NO_DATA_VALUE
      // then it's a rounding error in the data and we should subtract 1
      if (rawValue === NO_DATA_VALUE) {
        rawValue--;
      }
      return (
        this.metadata.offsets[fieldChannel] +
        this.metadata.scales[fieldChannel] * rawValue
      );
    },

    /**
     * @param {number} i - Zero based column index.
     * @param {number} j - Zero based row index.
     * @param {?string[]} fieldNames - The list of names of the fields for which the values are returned.
     * @param {?number[]} result - The array to store the result in.
     * @return {?number[]} Returns the list of values for the given fields at the given index or `null`.
     */
    getRealValuesAt(i, j, fieldNames, result) {
      var rawValues = this.getRawValuesAt(i, j, this._tempRawValuesArray);
      if (!rawValues) {
        return null;
      }

      // Check if all rawValues are NO_DATA_VALUE
      // In which case we have no data at this location and should return `null`
      if (this.hasMaskedData) {
        i = rawValues.length;
        while (i-- && rawValues[i] === NO_DATA_VALUE) {
          /*empty*/
        }
        if (i === -1) {
          // all rawValues are NO_DATA_VALUE
          return null;
        }
      }

      if (!fieldNames) {
        fieldNames = this._fieldNames;
      }
      result = result || [];
      for (i = 0; i < fieldNames.length; i++) {
        var fieldName = fieldNames[i];
        var field = this.fields[fieldName];
        var fieldChannel = field ? field.channel : null;
        var rawValue = fieldChannel === null ? NaN : rawValues[fieldChannel];

        // Since we have data at this location, if the rawValue is NO_DATA_VALUE
        // then it's a rounding error in the data and we should subtract 1
        if (rawValue === NO_DATA_VALUE) {
          rawValue--;
        }
        result[i] = (
          this.metadata.offsets[fieldChannel] +
          this.metadata.scales[fieldChannel] * rawValue
        );
      }

      return result;
    },

    /**
     * @param {[number, number]} xy - The projected coordinates to convert.
     * @param {number} [worldWest] - The west-most horizontal value in the projection, used for wrapping on the IDL.
     * @param {number} [worldEast] - The east-most horizontal value in the projection, used for wrapping on the IDL.
     * @return {?[number, number]} Returns the index corresponding to `xy` or `null`.
     */
    indexFromProjectedXY(xy, worldWest, worldEast) {
      let [x, y] = xy;
      const {top, bottom, left, right, width, height} = this.bounds;

      if (!inRange(y, bottom, top)) {
        return null;
      }

      if (!isNaN(worldWest) && !isNaN(worldEast)) {
        if (x < worldWest || x >= worldEast) {
          x = wrap(x, worldWest, worldEast);
        }

        if (right > worldEast && x < left) {
          x = x - worldWest + worldEast;
        } else if (left < worldWest && x > right) {
          x = x - worldEast + worldWest;
        }
      }

      if (!this.wraps && !inRange(x, left, right)) {
        return null;
      }

      const i =  (x - left) * this._width  / width;
      const j = -(y - top)  * this._height / height;

      return [i | 0, j | 0];
    },

    /**
     * @param {number} i - Zero based column index.
     * @param {number} j - Zero based row index.
     * @return {?[number, number]} Returns the corresponding projected coordinates or `null`.
     */
    projectedXYFromIndex(i, j) {
      if (
        (!this.wraps && !inRange(i, 0, this._width)) ||
        !inRange(j, 0, this._height)
      ) {
        return null;
      }

      var scaleh = this.bounds.width / this._width;
      var scalev = this.bounds.height / this._height;

      return [
        this.bounds.left + i * scaleh,
        this.bounds.top  - j * scalev,
      ];
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {VectorDataMatrix});
