/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunctionThrowingNotImplemented,
  parseDecimal,
  idlRangeWorldsAway,
  Bounds,
  Dataset,
} from '../syntool.js';
import {getLayerOptions} from '../utils/ol_utils.js';
import {OpenLayers} from '../vendors/openlayers.js';
import {layerExtension} from './extensions/layer_extension.js';
import {metadataExtension} from './extensions/metadata_extension.js';
import {auxiliaryDataExtension} from './extensions/auxiliary_data_extension.js';

export class OLImageLayer extends OpenLayers.Layer.Image {
  /**
   * Parameters:
   * name - {String} A name for the layer.
   * url - {String} Relative or absolute path to the image
   * extent - {<OpenLayers.Bounds>} The extent represented by the image
   * size - {[Number, Number]} The size (in pixels) of the image
   * options - {Object} Hashtable of extra options to tag onto the layer
   */
  constructor(name, url, extent, size, options) {
    super(name, url, extent, new OpenLayers.Size(size[0], size[1]), options);
    this.tileWrapped = null;
    this.clone = emptyFunctionThrowingNotImplemented;
  }

  get tiles() {
    return [this.tile, this.tileWrapped];
  }

  /**
   * Method: destroy
   * Destroy this layer
   */
  destroy() {
    if (this.tileWrapped) {
      this.removeTileMonitoringHooks(this.tileWrapped);
      this.tileWrapped.destroy();
      this.tileWrapped = null;
    }
    super.destroy();
  }

  /**
   * Method: moveTo
   * Create the tile for the image or resize it for the new resolution
   *
   * Parameters:
   * bounds - {<OpenLayers.Bounds>}
   * zoomChanged - {Boolean}
   * dragging - {Boolean}
   */
  moveTo(bounds, zoomChanged, dragging) {
    OpenLayers.Layer.prototype.moveTo.apply(this, arguments);

    this._setClipPath(bounds);

    //determine new tile size
    const tileWidth = this.extent.getWidth() / this.map.getResolution();
    const tileHeight = this.extent.getHeight() / this.map.getResolution();
    this.tileSize = new OpenLayers.Size(tileWidth, tileHeight);

    let wrappedExtent = this.extent;
    let wrappingExtent;

    const worldBounds = this.map.mapView.getWorldProjectedBounds();
    if (worldBounds) {
      const worldWidth = worldBounds.getWidth();
      const viewCenterX = Bounds.getCenterX(
        this.map.mapView.shouldWrapDateLine() ? bounds : worldBounds
      );

      const left = this.extent.left;
      const right = this.extent.right;
      const [leftWorldsAway, rightWorldsAway] = (
        idlRangeWorldsAway(left, right, viewCenterX, worldWidth)
      );
      const wrappedLeft = left - leftWorldsAway * worldWidth;
      const wrappedRight = right - rightWorldsAway * worldWidth;
      const wrapping = (leftWorldsAway < rightWorldsAway);

      wrappedExtent = Bounds.create(
        wrappedLeft,
        this.extent.bottom,
        wrappedLeft + this.extent.getWidth(),
        this.extent.top
      );
      if (wrapping) {
        wrappingExtent = Bounds.create(
          wrappedRight - this.extent.getWidth(),
          this.extent.bottom,
          wrappedRight,
          this.extent.top
        );
      }
    }

    this.tile = (
      this._createOrUpdateTileWithPosition(bounds, wrappedExtent, this.tile)
    );
    this.tileWrapped = (
      this._createOrUpdateTileWithPosition(bounds, wrappingExtent, this.tileWrapped)
    );
  }
  addTileMonitoringHooks(tile) {
    tile.onLoadStart = function() {
      this.events.triggerEvent('tileloadstart', {tile});
    };
    tile.onLoadEnd = function() {
      this.events.triggerEvent('tileloaded', {tile});
    };
    tile.events.on({
      loadstart: tile.onLoadStart,
      loadend: tile.onLoadEnd,
      unload: tile.onLoadEnd,
      scope: this,
    });
  }
  setUrl(url) {
    super.setUrl(url);
    this.tileWrapped?.draw();
  }

  _setClipPath(visibleExtent) {
    if (this.map.mapView.shouldWrapDateLine()) {
      this.div.style.clipPath = '';
      return;
    }

    const visibleWidth = Bounds.getWidth(visibleExtent);
    const visibleHeight = Bounds.getHeight(visibleExtent);

    if (visibleWidth === 0 || visibleHeight === 0) {
      this.div.style.clipPath = 'inset(100% 100% 0 0)';
      return;
    }

    const {w, h} = this.map.getSize();
    const maxExtent = this.map.getMaxExtent();
    const extent = Bounds.intersection(visibleExtent, maxExtent);

    const topRatio = (
      -(Bounds.getTop(extent) - Bounds.getTop(visibleExtent)) / visibleHeight
    );
    const rightRatio = (
      -(Bounds.getRight(extent) - Bounds.getRight(visibleExtent)) / visibleWidth
    );
    const bottomRatio = (
      (Bounds.getBottom(extent) - Bounds.getBottom(visibleExtent)) / visibleHeight
    );
    const leftRatio = (
      (Bounds.getLeft(extent) - Bounds.getLeft(visibleExtent)) / visibleWidth
    );

    const div = this.isFixed ? this.div : this.map.layerContainerDiv;
    const offsetLeft = parseDecimal(div.style.left || '0');
    const offsetTop = parseDecimal(div.style.top || '0');

    const top = h * topRatio - offsetTop;
    const left = w * leftRatio - offsetLeft;
    const bottom = h * bottomRatio + offsetTop + 100 - h;
    const right = w * rightRatio + offsetLeft + 100 - w;

    this.div.style.clipPath = `inset(${top}% ${right}% ${bottom}% ${left}%)`;
  }
  _createOrUpdateTileWithPosition(bounds, extent, tile) {
    if (extent && Bounds.intersects(bounds, extent)) {
      //determine new position (upper left corner of new bounds)
      const position = this.map.getLayerPxFromLonLat({
        lon: extent.left,
        lat: extent.top,
      });

      if (tile) {
        //just resize the tile and set it's new position
        tile.size = this.tileSize.clone();
        tile.position = position;
      } else {
        //create the new tile
        tile = new OpenLayers.Tile.Image(
          this,
          position,
          this.extent,
          null,
          this.tileSize
        );
        this.addTileMonitoringHooks(tile);
      }
      tile.draw();
    } else {
      tile?.clear();
    }
    return tile;
  }
}

var imageExtension = {
  initialize(info) {
    window.original();
    this._resolutions = info.resolutions;
  },
  createLayer() {
    var size = this._resolutions.split(',');
    var layerOptions = Object.assign(getLayerOptions(), {
      //////////////
      // Options  //
      //////////////

      /* OpenLayers.Layer */
      maxExtent: this.bounds,

      /* Custom */
      productId: this.productId,
      datasetId: this.datasetId,
    });

    return new OLImageLayer(
      this.name,
      this.getAuxiliaryDataURL(),
      this.bounds,
      size,
      layerOptions
    );
  },
};

export const ImageDataset = Dataset.createClass({
  static: {
    name        : 'ImageDataset',
    productTypes: ['IMAGE'],
  },
  extensions: [
    layerExtension,
    metadataExtension,
    auxiliaryDataExtension,
    imageExtension,
  ],
});
window.ODL = Object.assign(window.ODL || {}, {ImageDataset});
