/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  whenReady,
  Dataset,
  Syntool,
  Types,
} from '../syntool.js';
import {layerExtension} from './extensions/layer_extension.js';
import {sharedLayerExtension} from './extensions/shared_layer_extension.js';
import {LayerAPIServer} from './layer_dataset/layer_api_server.js';

whenReady(function() {
  Syntool.api.servers.add(new LayerAPIServer('_layers'));
});

var productLayerExtension = {
  createLayer() {
    return this.getProduct().layer;
  },
  handleVisibilityChange() {
    window.original();
    if (this.isVisible() && this.getProduct().updateLayer) {
      this.syntool.timelineStore.hub.on(
        'change:currentDate',
        this.handleCurrentDateChange,
        this
      );
      this.handleCurrentDateChange();
    } else {
      this.syntool.timelineStore.hub.off(
        'change:currentDate',
        this.handleCurrentDateChange,
        this
      );
    }
  },

  handleCurrentDateChange() {
    this.getProduct().updateLayer(
      this.layerManager.layer,
      this.syntool.timelineStore.currentDate
    );
  },
};

function validateLayerProduct(props, propName, propFullName) {
  var propValue = props[propName];
  if (
    propValue.validFrom && propValue.validTill &&
    propValue.validFrom >= propValue.validTill
  ) {
    return [
      'Invalid `', propFullName, '` of value `', propValue, '`, ',
      'expected `validFrom` date less than `validTill` date.'
    ].join('');
  }
  return null;
}

export const LayerDataset = Dataset.createClass({
  static: {
    name          : 'LayerDataset',
    productTypes  : ['LAYER'],
    validateConfig: Types.allOfType([
      Types.shape({
        layer      : Types.Layer.isRequired,
        validFrom  : Types.date,
        validTill  : Types.date,
        updateLayer: Types.func,
      }),
      validateLayerProduct,
    ]),
    normalizeProductConfig(productConfig) {
      if (!('validFrom' in productConfig)) {
        productConfig.validFrom = new Date(DATE_MIN_TIME); // min date
      }
      if (!('validTill' in productConfig)) {
        productConfig.validTill = new Date(DATE_MAX_TIME); // max date
      }
      return productConfig;
    },
    APIServer: LayerAPIServer,
  },
  extensions: [
    layerExtension,
    sharedLayerExtension,
    productLayerExtension,
  ],
});
window.ODL = Object.assign(window.ODL || {}, {LayerDataset});
