/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {createClass, clamp} from '../../syntool.js';
import {OpenLayers} from '../../vendors/openlayers.js';

export const LayerManager = createClass([
  /*Public*/ {
    constructor: function LayerManager(layerOrLayerCreator, mapView) {
      var createLayer;
      var layer;

      if (typeof layerOrLayerCreator === 'function') {
        createLayer = layerOrLayerCreator;
        layer = null;
      } else if (layerOrLayerCreator instanceof OpenLayers.Layer) {
        createLayer = null;
        layer = layerOrLayerCreator;
      } else {
        throw new TypeError(
          "Expected first argument of the LayerManager's constructor to be " +
          'either a function or an OpenLayers.Layer instance, but found "' +
          layerOrLayerCreator + '"'
        );
      }

      this.mapView = mapView;
      this._dataset = null;
      this._createLayer = createLayer;
      this.layer = layer;
      this.layerProps = {
        inMap     : mapView.hasLayer(layer),
        visibility: 0,
        opacity   : layer ? layer.opacity : 1,
        zIndex    : layer ? layer.getZIndex() : 0,
      };
      this.propsByDatasetId = new Map();
      layer?.setVisibility(false);
    },

    setDataset(dataset) {
      this._dataset = dataset;
      if (this.layer) {
        this.layer.dataset = dataset;
        this.layer.productId = dataset.productId;
      }
    },
    setInMap(dataset, inMap) {
      var props = this._getPropsForDataset(dataset);

      if (props.inMap === inMap) {
        return;
      }
      props.inMap = inMap;

      this._update(Object.assign({
        inMap: this.layerProps.inMap + (inMap ? 1 : -1),
      }, this._getZIndexProps()));
    },

    getVisibility(dataset) {
      var props = this._getPropsForDataset(dataset);

      return props.visibility;
    },
    setVisibility(dataset, visibility) {
      var props = this._getPropsForDataset(dataset);

      if (props.visibility === visibility) {
        return;
      }
      props.visibility = visibility;
      this._update({
        visibility: this.layerProps.visibility + (visibility ? 1 : -1),
      });
    },

    getOpacity(dataset) {
      var props = this._getPropsForDataset(dataset);

      return props.opacity;
    },
    setOpacity(dataset, opacity) {
      var props = this._getPropsForDataset(dataset);

      opacity = clamp(opacity, 0, 1);
      if (props.opacity === opacity) {
        return;
      }
      props.opacity = opacity;
      if (this.layerProps.opacity !== opacity) {
        this._update({opacity: opacity});
      }
    },

    getZIndex(dataset) {
      return this._getPropsForDataset(dataset).zIndex;
    },
    setZIndex(dataset, zIndex) {
      var props = this._getPropsForDataset(dataset);

      if (props.zIndex === zIndex) {
        return;
      }
      props.zIndex = zIndex;
      this._update(this._getZIndexProps());
    },

    destroy(dataset) {
      this.propsByDatasetId.delete(dataset.datasetId);
      if (this.layer && this._createLayer && this.propsByDatasetId.size === 0) {
        this.layer.destroy();
        this.layer = null;
      }
    },
  },
  /*Protected*/ {
    _getPropsForDataset(dataset) {
      let props = this.propsByDatasetId.get(dataset.datasetId);
      if (!props) {
        props = {
          inMap     : false,
          visibility: false,
          opacity   : 1,
          zIndex    : 0,
        };
        this.propsByDatasetId.set(dataset.datasetId, props);
      }
      return props;
    },
    _getZIndexProps() {
      for (const [, datasetProps] of this.propsByDatasetId) {
        if (datasetProps.inMap) {
          return {zIndex: datasetProps.zIndex};
        }
      }
      return null;
    },
    _update(partialProps) {
      var props = this.layerProps;
      var nextProps = (
        partialProps
        ? Object.assign({}, props, partialProps)
        : props
      );

      if (!this.layer) {
        if (nextProps.inMap) {
          this.layer = this._createLayer();
          this.layer.dataset = this._dataset;
          this.layer.productId = this._dataset.productId;
          this.layer.setVisibility(nextProps.visibility > 0);
          this.layer.setOpacity(nextProps.opacity);
          this.layer.setZIndex(nextProps.zIndex);
          this.mapView.addLayer(this.layer);
        }
        this.layerProps = nextProps;

        return;
      }

      if (!nextProps.inMap) {
        this.mapView.removeLayer(this.layer);
      }

      if (!props.visibility !== !nextProps.visibility) {
        this.layer.setVisibility(nextProps.visibility > 0);
      }

      if (props.opacity !== nextProps.opacity) {
        this.layer.setOpacity(nextProps.opacity);
      }

      if (props.zIndex !== nextProps.zIndex) {
        this.layer.setZIndex(nextProps.zIndex);
      }

      if (nextProps.inMap) {
        this.mapView.addLayer(this.layer);
      }
      this.layerProps = nextProps;
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {LayerManager});
