/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {createClass, appendToArray, Dataset} from '../../syntool.js';

function notInArrayFilter(item) {
  return this.indexOf(item) === -1;
}

export const VectorLayerManager = createClass([
  /*Public*/ {
    constructor: function VectorLayerManager(createLayer, mapView) {
      this.mapView = mapView;
      this._createLayer = createLayer;
      this._productId = null;
      this.layer = null;
      this.layerProps = {
        hidden: false,
        inMap : 0,
        zIndex: 0,
      };
      this.propsByDatasetId = new Map();
    },

    setHidden(hidden) {
      this._update({hidden});
    },

    setDataset(dataset) {
      this._productId = dataset.productId;
      if (this.layer) {
        this.layer.productId = dataset.productId;
      }
    },
    setInMap(dataset, inMap) {
      const props = this._getPropsForDataset(dataset);

      if (props.inMap !== inMap) {
        props.inMap = inMap;
        this._update(
          Object.assign(
            {inMap: this.layerProps.inMap + (inMap ? 1 : -1)},
            this._getZIndexProps()
          ),
          inMap ? props.features : null,
          inMap ? null : dataset
        );
      }
    },

    getVisibility(dataset) {
      return this._getPropsForDataset(dataset).visibility;
    },
    setVisibility(dataset, visibility) {
      this._getPropsForDataset(dataset).visibility = visibility;
    },

    getOpacity(dataset) {
      return this._getPropsForDataset(dataset).opacity;
    },
    setOpacity(dataset, opacity) {
      this._getPropsForDataset(dataset).opacity = opacity;
    },

    getZIndex(dataset) {
      return this._getPropsForDataset(dataset).zIndex;
    },
    setZIndex(dataset, zIndex) {
      const props = this._getPropsForDataset(dataset);
      if (props.zIndex !== zIndex) {
        props.zIndex = zIndex;
        this._update(this._getZIndexProps());
      }
    },

    addFeatures(dataset, features) {
      const props = this._getPropsForDataset(dataset);

      props.features = appendToArray(props.features, features);
      if (props.inMap) {
        this._update(null, features);
      }
    },
    getFeatures(dataset) {
      return this._getPropsForDataset(dataset).features;
    },
    removeAllFeatures(dataset) {
      const props = this._getPropsForDataset(dataset);

      if (props.inMap) {
        this._update(null, null, dataset);
      }
      props.features.length = 0;
    },
    removeFeatures(dataset, features) {
      const props = this._getPropsForDataset(dataset);
      const length = props.features.length;

      if (length > 0 && features.length > 0) {
        props.features = props.features.filter(notInArrayFilter, features);
        if (props.inMap && props.features.length < length) {
          this._update(
            null,
            null,
            props.features.length === 0 ? dataset : features
          );
        }
      }
    },
    drawFeature(dataset, feature) {
      this.layer?.drawFeature(feature);
    },
    redraw(dataset, force) {
      const props = this._getPropsForDataset(dataset);
      if (props?.features?.length > 0) {
        this.layer?.drawFeatures(props.features, force);
      }
    },

    destroy(dataset) {
      if (this.propsByDatasetId.delete(dataset.datasetId)) {
        this.removeAllFeatures(dataset);
      }

      if (this.layer && this.propsByDatasetId.size === 0) {
        this.layer.destroy();
        this.layer = null;
      }
    },
  },
  /*Protected*/ {
    _getPropsForDataset(dataset) {
      if (!this.propsByDatasetId.has(dataset.datasetId)) {
        this.propsByDatasetId.set(dataset.datasetId, {
          inMap   : false,
          opacity : 1,
          zIndex  : 0,
          features: [],
        });
      }
      return this.propsByDatasetId.get(dataset.datasetId);
    },
    _getZIndexProps() {
      for (const datasetProps of this.propsByDatasetId.values()) {
        if (datasetProps.inMap) {
          return {zIndex: datasetProps.zIndex};
        }
      }
      return null;
    },
    _update(partialProps, addedFeatures, removedFeatures) {
      const props = this.layerProps;
      const nextProps = (
        partialProps
        ? Object.assign({}, props, partialProps)
        : props
      );

      if (!this.layer) {
        this.layer = this._createLayer();
        this.layer.productId = this._productId;
        this.layer.appliesDatasetProperties = true;
        this.layer.setZIndex(nextProps.zIndex);
        if (addedFeatures) {
          this.layer.addFeatures(addedFeatures);
        }

        if (!nextProps.hidden && nextProps.inMap) {
          this.mapView.addLayer(this.layer);
        }
        this.layerProps = nextProps;

        return;
      }

      if (nextProps.hidden || !nextProps.inMap) {
        this.mapView.removeLayer(this.layer);
      }

      if (props.zIndex !== nextProps.zIndex) {
        this.layer.setZIndex(nextProps.zIndex);
      }

      if (removedFeatures) {
        if (removedFeatures instanceof Dataset) {
          this.layer.removeAllFeaturesOfDataset(removedFeatures);
        } else {
          this.layer.removeFeatures(removedFeatures);
        }
      }
      if (addedFeatures) {
        this.layer.addFeatures(addedFeatures);
      }

      if (!nextProps.hidden && nextProps.inMap) {
        this.mapView.addLayer(this.layer);
      }
      this.layerProps = nextProps;
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {VectorLayerManager});
