/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import '../map_view_extensions/map_view_shapes.js';
import {
  emptyFunctionReturningFalse,
  emptyFunctionReturningNull,
  clamp,
  round,
  Pattern,
  Geometry,
  Feature,
  Types,
  Dataset,
} from '../syntool.js';
import {setFeatureVisibility} from '../utils/ol_utils.js';
import {Shape} from './shape_dataset/shape.js';
import {
  getDatasetIdForShapeId,
  getInfoForShape,
  ShapesAPIServer,
} from './shape_dataset/shapes_api_server.js';
import './shape_dataset/shapes_core.js';
import {ShapesStore} from './shape_dataset/shapes_store.js';

Object.assign(ShapesAPIServer, {
  getDatasetIdForShapeId,
  getInfoForShape,
});

var shapeExtension = {
  initialize(info) {
    this.shape = null;
    this.shapeFeature = null;
    this.readOnly = false;
    window.original();
    this._opacity = this.getProduct().opacity;
  },
  handleInfo(info) {
    var bounds = this.bounds;
    var geometry = this.geometry;
    info.mapBounds = null;
    info.polygon = null;
    window.original();
    this.readOnly = info.readOnly;
    if (this.shape === info.shape) {
      this.bounds = bounds;
      this.geometry = geometry;
    } else {
      this.shape = info.shape;
      this.handleShapeChange(this.shape);
    }
  },
  matchesPattern(pattern) {
    return Pattern.test(pattern, this.shape.text);
  },
  intersectsBounds(bounds, worldBounds) {
    return (
      window.original() &&
      Geometry.intersectsLatLonBounds(this.geometry, bounds, worldBounds)
    );
  },
  hasResources() {
    return window.original() || !!this.shapeFeature;
  },
  getOpacity() {
    return this._opacity;
  },
  setOpacity(opacity) {
    opacity = round(opacity, -3);
    opacity = clamp(opacity, 0, 1);
    if (this.getOpacity() !== opacity) {
      const wasVisible = this.isVisible();
      this._opacity = opacity;
      this.hub.emit('change:opacity');
      if (this.isVisible() !== wasVisible) {
        this.handleVisibilityChange();
      }
      this.shapeFeature?.layer?.drawFeature(this.shapeFeature);
    }
  },
  isGlobal: emptyFunctionReturningFalse,
  getOutlineGeometry: emptyFunctionReturningNull,
  handleVisibilityChange() {
    window.original();
    if (this.shapeFeature) {
      setFeatureVisibility(this.shapeFeature, this.isVisible());
    }
  },
  handleSelectedChange() {
    // check dataset's product still exists
    if (this.getProduct()) {
      var feature = this.shapeFeature;
      if (feature) {
        feature.renderIntent = this.isSelected() ? 'select' : 'default';
        if (feature.layer) {
          feature.layer.drawFeature(feature);
          feature.layer.setZIndex(
            this.syntool.mapView.getZIndexForLayer(feature.layer)
          );
        }
      }
    }
  },
  didAddToMap() {
    window.original();
    if (!this.shapeFeature) {
      this.shapeFeature = new Feature(this.geometry);
      if (
        Geometry.isCollection(this.geometry) &&
        Geometry.isPoint(Geometry.getComponentAtIndex(this.geometry, 0))
      ) {
        Geometry.getComponentAtIndex(this.geometry, 0).__virtual = true;
      }
      this.shapeFeature.dataset = this;
    }
    setFeatureVisibility(this.shapeFeature, this.isVisible());
    this.syntool.mapView.shapesLayer.addFeature(this.shapeFeature);
  },
  willRemoveFromMap() {
    window.original();
    this.syntool.mapView.shapesLayer.removeAllFeaturesOfDataset(this);
  },
  destroy() {
    this.shapeFeature?.destroy();
    this.shapeFeature = null;
    window.original();
  },

  getLength() {
    return Shape.getLength(this.shownShape);
  },
  getArea() {
    return Shape.getArea(this.shownShape);
  },
  updateShape(shape) {
    if (this.shownShape === shape) {
      return;
    }
    this.handleShapeChange(shape);
    this.hub.emit('changing');
  },

  handleShapeChange(shape) {
    this.shownShape = shape;
    this.geometry = Shape.createGeometry(shape);
    this.bounds = Geometry.getBounds(this.geometry);

    if (this.shapeFeature) {
      this.shapeFeature.geometry = this.geometry;
      this.shapeFeature.layer?.drawFeature(this.shapeFeature, true /*force*/);
    }
  },
};

export const ShapeDataset = Dataset.createClass({
  static: {
    name          : 'ShapeDataset',
    productTypes  : ['USER_SHAPES', 'SERVER_SHAPES'],
    validateConfig: Types.match([
      Types.case(
        Types.shape({type: Types.value('USER_SHAPES')}),
        Types.allOfType([
          Types.shape({
            style: Shape.types.ShapeStyleMap,
          }),
          Types.uniquePropInArray('type'),
        ])
      ),
      Types.case(
        Types.shape({type: Types.value('SERVER_SHAPES')}),
        Types.shape({
          style : Shape.types.ShapeStyleMap,
          shapes: Types.arrayOf(
            Types.allOfType([
              Shape.types.Shape.isRequired,
              Types.uniquePropInArray('id'),
            ])
          ).isRequired,
        })
      ),
    ]),
    APIServer: ShapesAPIServer,
    Shape,
    ShapesStore,
  },
  extensions: [shapeExtension],
});
window.ODL = Object.assign(window.ODL || {}, {ShapeDataset});
