/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  SECOND_IN_MILLISECONDS,
  DAY_IN_MILLISECONDS,
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  RAD_IN_DEG,
  round,
  inRange,
  Bounds,
  Geometry,
  Projection,
  Types as is,
  validateValueType,
  Config,
  Syntool,
} from '../../syntool.js';
import {Color} from '../../utils/color.js';

const ARROW_HEAD_COLD_COLOR = '#0000ff';
const ARROW_HEAD_HOT_COLOR = '#ff0000';
const ARROW_HEAD_NONE_STYLE = {
  fillOpacity: 0,
};
const ARROW_HEAD_COLD_STYLE = {
  strokeWidth: 1,
  strokeColor: ARROW_HEAD_COLD_COLOR,
  fillColor  : ARROW_HEAD_COLD_COLOR,
  fillOpacity: 1,
};
const ARROW_HEAD_HOT_STYLE = {
  strokeWidth: 1,
  strokeColor: ARROW_HEAD_HOT_COLOR,
  fillColor  : ARROW_HEAD_HOT_COLOR,
  fillOpacity: 1,
};

const ShapeTypes = {};
const ShapeHandlers = {};

const ShapeArrow = is.shape({
  directionFlipped: is.bool,
  sideFlipped     : is.oneOfType([is.bool, is.value('none')]),
  ratio           : is.number.gte(0).lte(1),
  density         : is.number.gte(0).lte(1),
  offset          : is.number.gte(0).lte(1),
  size            : is.number.gt(0),
}).isExact;
const OldShape = is.shape({
  author  : is.string.isRequired,
  text    : is.string.isRequired,
  wkt     : is.WKT.isRequired,
  date    : is.dateOrMs.isRequired,
  duration: is.number.gte(0),
}).isExact;
const BaseShape = is.shape({
  id   : is.string,
  type : is.string.isRequired,
  text : is.string,
  start: is.dateOrMs,
  end  : is.dateOrMs,
  color: Color,
});
function caseForShapeType(shapeType) {
  return is.case(
    is.shape({type: is.value(shapeType)}),
    ShapeHandlers[shapeType].typeChecker
  );
}
const ShapeType = is.allOfType([
  BaseShape,
  validateShapeStartEnd,
  is.lazy(() => is.match(Object.keys(ShapeTypes).map(caseForShapeType))),
]);
const ShapeStyle = is.shape({
  pointRadius    : is.number.gt(0),
  fillColor      : is.SVGPaint,
  fillOpacity    : is.number.gte(0).lte(1),
  strokeColor    : is.SVGPaint,
  strokeOpacity  : is.number.gte(0).lte(1),
  strokeWidth    : is.oneOfType([is.string, is.number.gte(0)]),
  strokeDashstyle: is.string,
  fontColor      : is.SVGPaint,
  fontSize       : is.number.gte(0),
});
const ShapeStyleMap = is.shape({
  ALL     : ShapeStyle,
  TEXT    : ShapeStyle,
  ARROW   : ShapeStyle,
  LINE    : ShapeStyle,
  ELLIPSE : ShapeStyle,
  MUSHROOM: ShapeStyle,
  WKT     : ShapeStyle,
});

Config.configValidators.push(validateContainsUserShapesProduct);

export const Shape = {
  types: {
    Shape: ShapeType,
    ShapeArrow,
    ShapeStyle,
    ShapeStyleMap,
  },
  ShapeHandlers,
  addShapeType,

  create        : createShape,
  same          : sameShape,
  createGeometry,
  toInternalJSON: shapeToInternalJSON,
  toExternalJSON: shapeToExternalJSON,
  fromJSON      : shapeFromJSON,

  getLength: callOnShapeHandler.bind(null, 'getLength', 0),
  getArea  : callOnShapeHandler.bind(null, 'getArea', 0),

  getControlPoints         : callOnShapeHandler.bind(null, 'getControlPoints', null),
  controlPointDidStartDrag : callOnShapeHandler.bind(null, 'controlPointDidStartDrag', null),
  changesOnControlPointMove: callOnShapeHandler.bind(null, 'changesOnControlPointMove', null),
  controlPointDidEndDrag   : callOnShapeHandler.bind(null, 'controlPointDidEndDrag', null),
  updatesFromChanges       : callOnShapeHandler.bind(null, 'updatesFromChanges', null),

  getArrowSizeForResolution,
  getInitialShapeArrowAtResolution,

  utils: {
    styles: {
      move     : {fillColor: '#0F0', pointRadius: 5, cursor: 'move'},
      rotate   : {fillColor: '#00F', pointRadius: 5, cursor: 'se-resize'},
      secondary: {fillColor: '#FF0', pointRadius: 5, cursor: 'move'},
      virtual  : {fillOpacity: 0.3, strokeOpacity: 0.3, cursor: 'copy'},
    },

    rotateOffset: 30,     // in viewport px
    minScaleDistance: 10, // in viewport px

    projectedDistance,
    projectedPoint,
    projectedPoints,
    projectedArrow,
    projectedWKT,

    unprojectedDistance,
    unprojectedPoint,
    unprojectedPoints,
    unprojectedArrow,
    unprojectedWKT,

    correctedDistance: roundNumber,
    correctedPoint,
    correctedPoints,
    correctedArrow,
    correctedWKT,

    sameDate,
    samePoint,
    sameArray,
    sameArrow,

    createDoubleArrowHeadGeometry,
    createArrowHeadGeometries,
  },
};

function addShapeType(type, handler) {
  if (!Shape[type]) {
    Shape[type] = type;
    ShapeTypes[type] = type;
    ShapeHandlers[type] = handler;
  }
}

function validateShapeStartEnd(props, propName, propFullName) {
  const propValue = props[propName];
  if (
    propValue.start &&
    propValue.end &&
    propValue.start - propValue.end >= 0
  ) {
    return 'Invalid shape, expected `start` date less than `end` date.';
  }
  return null;
}
function validateContainsUserShapesProduct(props, propName, propFullName) {
  const config = props[propName];
  if (!config || !config.products) {
    return null;
  }

  const products = config.products;
  for (const product of products) {
    if (product.type === 'USER_SHAPES') {
      return null;
    }
  }

  return [
    'Invalid `', propFullName, '.products`, expected to include one product ',
    'of type `USER_SHAPES`.'
  ].join('');
}

function projectedDistance(distInMeters) {
  return Syntool.mapView.convertFromUnit(distInMeters, 'm');
}
function projectedPoint(point) {
  return Projection.projectLonLat(Syntool.mapView.getProjection(), point);
}
function projectedPoints(points) {
  return points.map(projectedPoint);
}
function projectedArrow(arrow) {
  return arrow && Object.assign({}, arrow, {size: projectedDistance(arrow.size)});
}
function projectedWKT(wkt) {
  return Geometry.toWKT(
    Projection.projectGeometry(
      Syntool.mapView.getProjection(),
      Geometry.fromWKT(wkt),
      true
    )
  );
}
function roundExportedNumber(number) {
  return round(number, -4);
}
function unprojectedDistance(distInMapUnits) {
  return roundExportedNumber(
    Syntool.mapView.convertToUnit(distInMapUnits, 'm')
  );
}
function unprojectedPoint(point) {
  return (
    Projection.unprojectXY(Syntool.mapView.getProjection(), point)
    .map(roundExportedNumber)
  );
}
function unprojectedPoints(points) {
  return points.map(unprojectedPoint);
}
function unprojectedArrow(arrow) {
  return arrow && Object.assign({}, arrow, {size: unprojectedDistance(arrow.size)});
}
function unprojectedWKT(wkt) {
  return (
    Geometry.toWKT(
      Projection.unprojectGeometry(
        Syntool.mapView.getProjection(),
        Geometry.fromWKT(wkt),
        true
      )
    )
    .replace(/\d+\.\d+/g, roundExportedNumber)
  );
}

function wrapGeometry(geometry) {
  const worldBounds = Syntool.mapView.getWorldProjectedBounds();

  if (worldBounds) {
    const worldLeft = Bounds.getLeft(worldBounds);
    const worldRight = Bounds.getRight(worldBounds);
    const bounds = Geometry.getBounds(geometry);

    if (
      !inRange(Bounds.getLeft(bounds), worldLeft, worldRight) ||
      !inRange(Bounds.getRight(bounds), worldLeft, worldRight)
    ) {
      const worldWidth = Bounds.getWidth(worldBounds);
      const centerX = Bounds.getCenterX(bounds);
      const worldsAway = Math.round(centerX / worldWidth);

      return Geometry.translated(geometry, -worldsAway * worldWidth, 0);
    }
  }

  return geometry;
}

function roundNumber(number) {
  return round(number, -6);
}
function correctedPoint(point) {
  const x = roundNumber(point[0]);
  const y = roundNumber(point[1]);
  const worldBounds = Syntool.mapView.getWorldProjectedBounds();
  if (worldBounds) {
    const worldWidth = Bounds.getWidth(worldBounds);
    const worldsAway = Math.round(x / worldWidth);

    return [x - worldsAway * worldWidth, y];
  }
  return [x, y];
}
function correctedPoints(points) {
  if (points.length === 0) {
    return points;
  }

  let i = points.length - 1;
  let x = roundNumber(points[i][0]);
  let minX = x;
  let maxX = x;

  while (i--) {
    x = roundNumber(points[i][0]);

    if (x < minX) {
      minX = x;
    } else if (x > maxX) {
      maxX = x;
    }
  }

  const worldBounds = Syntool.mapView.getWorldProjectedBounds();
  const worldWidth = worldBounds ? Bounds.getWidth(worldBounds) : 0;
  let worldsAway = 0;

  if (
    worldBounds &&
    (maxX < Bounds.getLeft(worldBounds) || minX > Bounds.getRight(worldBounds))
  ) {
    worldsAway = Math.round(((minX + maxX) / 2) / worldWidth);
  }

  return points.map(correctedPoints_point, worldsAway * worldWidth);
}
//eslint-disable-next-line camelcase
function correctedPoints_point(point) {
  return [roundNumber(point[0]) - this, roundNumber(point[1])];
}
function correctedArrow(arrow) {
  return arrow && Object.assign({}, arrow, {size: roundNumber(arrow.size)});
}
function correctedWKT(wkt) {
  return Geometry.toWKT(
    Geometry.rewind(
      wrapGeometry(Geometry.fromWKT(wkt.replace(/\d+\.\d+/g, roundNumber))),
      false /*counterclockwise*/
    )
  );
}

function sameDate(a, b) {
  return (a === b || (a && b && a.getTime() === b.getTime()));
}
function samePoint(a, b) {
  return (a === b || (a && b && a[0] === b[0] && a[1] === b[1]));
}
function sameArray(a, b, sameElement) {
  if (a === b) {
    return true;
  }

  if (!a || !b) {
    return false;
  }

  if (a.length !== b.length) {
    return false;
  }

  let l = a.length;

  while (l--) {
    if (!sameElement(a[l], b[l])) {
      return false;
    }
  }

  return true;
}
function sameArrow(a, b) {
  if (a === b) {
    return true;
  }

  if (!a || !b) {
    return false;
  }

  if (
    a.directionFlipped !== b.directionFlipped ||
    a.sideFlipped !== b.sideFlipped           ||
    a.ratio !== b.ratio                       ||
    a.density !== b.density                   ||
    a.offset !== b.offset                     ||
    a.size !== b.size
  ) {
    return false;
  }

  return true;
}

function migrateProps(props) {
  if (props.type) {
    return props;
  }

  const validationError = validateValueType(OldShape, 'shape', props);

  if (validationError) {
    console.warn('createShape', validationError, props);
    return null;
  }

  let start;
  let end;

  if (props.date instanceof Date) {
    start = props.date.getTime();
  } else if (isFinite(props.date)) {
    start = props.date;
  } else if (typeof props.date === 'string') {
    start = Date.parse(props.date);
    if (isNaN(start)) {
      start = undefined;
    }
  }
  if (start !== undefined) {
    if (isFinite(props.duration)) {
      end = start + Math.max(
        SECOND_IN_MILLISECONDS, props.duration * DAY_IN_MILLISECONDS
      );
    } else if (props.date) {
      end = start + SECOND_IN_MILLISECONDS;
    }
  }

  const newProps = Object.assign({}, props);

  newProps.type = ShapeTypes.WKT;
  delete newProps.date;
  delete newProps.duration;
  if (start) {
    newProps.start = start;
  }
  if (end) {
    newProps.end = end;
  }

  return newProps;
}
function projectShape(shape) {
  if (ShapeHandlers[shape.type]) {
    ShapeHandlers[shape.type].project.apply(null, arguments);
  } else {
    console.warn(
      'projectShape',
      'Trying to projected a shape with unknown type',
      shape
    );
  }
  return shape;
}
function unprojectShape(shape) {
  if (ShapeHandlers[shape.type]) {
    ShapeHandlers[shape.type].unproject.apply(null, arguments);
  } else {
    console.warn(
      'unprojectShape',
      'Trying to unprojected a shape with unknown type',
      shape
    );
  }
  return shape;
}
function createShape(props) {
  props = migrateProps(props);
  if (!props) {
    return null;
  }

  const validationError = validateValueType(ShapeType, 'shape', props);

  if (validationError) {
    console.warn('createShape', validationError, props);
    return null;
  }

  let start = DATE_MIN_TIME;
  let end = DATE_MAX_TIME;

  if (props.start instanceof Date) {
    start = props.start.getTime();
  } else if (isFinite(props.start)) {
    start = props.start;
  }
  if (props.end instanceof Date) {
    end = props.end.getTime();
  } else if (isFinite(props.end)) {
    end = props.end;
  }

  const shape = {
    type : props.type,
    text : (props.text || '').trim(),
    start: new Date(start),
    end  : new Date(end),
    color: props.color,
  };

  shape.toJSON = shapeToExternalJSON.bind(null, shape);

  if (props.id) {
    shape.id = props.id;
  }

  if (ShapeHandlers[shape.type]) {
    ShapeHandlers[shape.type].create(shape, props);
  } else {
    console.warn(
      'createShape', 'Trying to create a shape with unknown type', props
    );
    return null;
  }
  return shape;
}
function sameShape(a, b) {
  if (a === b) {
    return true;
  }

  if (!a || !b) {
    return false;
  }

  if (
    a.type !== b.type           ||
    a.text !== b.text           ||
    !sameDate(a.start, b.start) ||
    !sameDate(a.end, b.end)     ||
    a.color !== b.color
  ) {
    return false;
  }

  if (ShapeHandlers[a.type] && !ShapeHandlers[a.type].same(a, b)) {
    return false;
  }
  return true;
}
function createGeometry(shape) {
  if (ShapeHandlers[shape.type]) {
    return ShapeHandlers[shape.type].createGeometry.apply(null, arguments);
  }
  return null;
}
function shapeToInternalJSON(shape) {
  const json = Object.assign({}, shape);

  delete json.id;
  delete json.toJSON;

  json.start = shape.start.getTime();
  if (json.start === DATE_MIN_TIME) {
    delete json.start;
  }
  json.end = shape.end.getTime();
  if (json.end === DATE_MAX_TIME) {
    delete json.end;
  }
  return json;
}
function shapeToExternalJSON(shape) {
  return unprojectShape(shapeToInternalJSON(shape));
}
function shapeFromJSON(json) {
  return createShape(projectShape(json));
}
function callOnShapeHandler(name, defaultReturn, shape, ...args) {
  const handler = ShapeHandlers[shape.type];
  return (
    (handler && typeof handler[name] === 'function')
    ? handler[name](shape, ...args)
    : defaultReturn
  );
}

function getArrowSizeForResolution(resolution) {
  return 25/*px*/ * resolution/*m/px*/;
}
function getInitialShapeArrowAtResolution(shape, resolution) {
  const handler = ShapeHandlers[shape.type];

  if (handler && typeof handler.getInitialArrowAtResolution === 'function') {
    return handler.getInitialArrowAtResolution.apply(null, arguments);
  }
  if (handler && handler.hasArrow) {
    return {
      size: getArrowSizeForResolution(resolution),
    };
  }
  return null;
}
function createDoubleArrowHeadGeometry(arrow, arrowXY, arrowSize) {
  const arrowOffset = (arrow.directionFlipped ? 1 : -1) * arrowSize / 2;
  const arrowHeadLowStyle = (
    arrow.sideFlipped === 'none' ? ARROW_HEAD_NONE_STYLE
    : arrow.sideFlipped ? ARROW_HEAD_HOT_STYLE
    : ARROW_HEAD_COLD_STYLE
  );
  const arrowHeadHighStyle = (
    arrow.sideFlipped === 'none' ? ARROW_HEAD_NONE_STYLE
    : arrow.sideFlipped ? ARROW_HEAD_COLD_STYLE
    : ARROW_HEAD_HOT_STYLE
  );
  const arrowHeadLow = Geometry.createPolygonFromXYs([
    [arrowXY[0] - arrowSize * arrow.ratio, arrowXY[1] + arrowOffset],
    [arrowXY[0], arrowXY[1] - arrowOffset],
    [arrowXY[0], arrowXY[1] + arrowOffset],
  ], arrowHeadLowStyle);
  const arrowHeadHigh = Geometry.createPolygonFromXYs([
    [arrowXY[0] + arrowSize * (1 - arrow.ratio), arrowXY[1] + arrowOffset],
    [arrowXY[0], arrowXY[1] - arrowOffset],
    [arrowXY[0], arrowXY[1] + arrowOffset],
  ], arrowHeadHighStyle);

  return Geometry.createCollection([arrowHeadLow, arrowHeadHigh]);
}
function createArrowHeadGeometries(arrow, points) {
  const segmentsCount = points.length - 1;
  const arrowHeadsCount = Math.floor(arrow.density * (segmentsCount - 1)) + 1;
  const arrowHeads = [];
  const segmentOffset = Math.round(arrow.offset * (segmentsCount - 1));
  const size = arrow.size;

  for (let i = 0; i < arrowHeadsCount; i++) {
    const segmentIndex = (
      (segmentOffset + Math.floor(i / arrowHeadsCount * segmentsCount)) %
      segmentsCount
    );
    const prevPoint = points[segmentIndex];
    const lastPoint = points[segmentIndex + 1];
    const angle = (
      Math.atan2(lastPoint[1] - prevPoint[1], lastPoint[0] - prevPoint[0]) *
      RAD_IN_DEG - 90
    );
    const center = [
      (prevPoint[0] + lastPoint[0]) / 2,
      (prevPoint[1] + lastPoint[1]) / 2,
    ];

    arrowHeads.push(
      Geometry.rotate(
        createDoubleArrowHeadGeometry(arrow, center, size),
        angle,
        center
      )
    );
  }

  if (arrowHeadsCount === 1) {
    return arrowHeads[0];
  }
  return Geometry.createCollection(arrowHeads);
}
