/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Geometry, LocalAPIServer, Pattern} from '../../syntool.js';
import {Shape} from './shape.js';

var infosMap = new WeakMap();

export function getDatasetIdForShapeId(productId, shapeId) {
  return productId + '-' + shapeId;
}
export function getInfoForShape(productId, shape) {
  var info = infosMap.get(shape);
  if (!info) {
    info = {
      productId: productId,
      datasetId: getDatasetIdForShapeId(productId, shape.id),
      name     : shape.id,
      start    : shape.start,
      end      : shape.end,
      bounds   : Geometry.getBounds(Shape.createGeometry(shape)),
      shape    : shape,
      readOnly : false,
    };
    infosMap.set(shape, info);
  }
  return info;
}
function getInfoForReadOnlyShape(productId, shape) {
  var info = infosMap.get(shape);
  if (!info) {
    info = getInfoForShape(productId, shape);
    info.readOnly = true;
  }
  return info;
}

export class ShapesAPIServer extends LocalAPIServer {
  /*Public*/
  handlesAllProductsOfType(type) {
    return type === 'USER_SHAPES' || type === 'SERVER_SHAPES';
  }

  /*Protected*/
  _infoMatchesPattern(info, pattern) {
    return (
      info.shape
      ? Pattern.test(pattern, info.shape.text)
      : super._infoMatchesPattern(info, pattern)
    );
  }
  _getAllInfosForProductWithId(productId) {
    var syntool = this.api.syntool;
    var shapes;

    if (!syntool.shapesStore) {
      return [];
    }

    if (
      syntool.userShapesProduct &&
      productId === syntool.userShapesProduct.id
    ) {
      shapes = syntool.shapesStore.localShapes;
      return (
        shapes
        ? Object.values(shapes).map(getInfoForShape.bind(null, productId))
        : []
      );
    }

    shapes = syntool.shapesStore.getShapesForProduct(productId);
    return (
      shapes
      ? Object.values(shapes).map(getInfoForReadOnlyShape.bind(null, productId))
      : []
    );
  }
}
