/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Syntool} from '../../syntool.js';
import {
  getDatasetIdForShapeId,
  getInfoForShape,
  ShapesAPIServer,
} from './shapes_api_server.js';
import {ShapesStore} from './shapes_store.js';
import './syntool_filters.js';

const SHAPES_API_SERVER_ID = '_shapes';

function getDatasetShapeId(dataset) {
  return dataset.shape.id;
}

Syntool.extend({
  [Syntool.NAME]: 'ShapesCore',
  [Syntool.CONSTRUCTOR]() {
    this.shapesStore = new ShapesStore();
    this.api.servers.add(new ShapesAPIServer(SHAPES_API_SERVER_ID));
  },
  [Syntool.INIT]() {
    this.productsStore.hub.on(
      'change:products',
      this.handleUserShapesProductsChange,
      this
    );
    if (this.productsStore.initialised) {
      this.handleUserShapesProductsChange();
    }
  },

  duplicateSelectedShapes() {
    if (!this.mapView.getSelectedDatasets) {
      return false;
    }
    var datasets = (
      this.mapView.getSelectedDatasets(this.filters.dataset.isShape)
    );
    if (datasets.length === 0) {
      return false;
    }

    var duplicateShapes = datasets.map(function(dataset) {
      return this.shapesStore.duplicate(dataset.shape);
    }, this);

    this.shapesStore.add(duplicateShapes);
    return true;
  },
  removeSelectedShapes() {
    if (!this.mapView.getSelectedDatasets) {
      return false;
    }
    var datasets = (
      this.mapView.getSelectedDatasets(this.filters.dataset.isEditableShape)
    );
    if (datasets.length === 0) {
      return false;
    }
    this.shapesStore.remove(datasets.map(getDatasetShapeId));
    return true;
  },

  _getDatasetForUserShapeWithId(userShapeId) {
    return this.mapView.getDatasetWithId(
      getDatasetIdForShapeId(this.userShapesProduct.id, userShapeId)
    );
  },
  _getDatasetForUserShape(userShape) {
    return this._getDatasetForUserShapeWithId(userShape.id);
  },
  handleUserShapesProductsChange() {
    if (!this.userShapesProduct) {
      this.productsStore.forEachProduct((product) => {
        if (product.type === 'USER_SHAPES') {
          this.userShapesProduct = product;
          this.shapesStore.hub
            .on('add', this.handleAddUserShape, this)
            .on('update', this.handleUpdateUserShape, this)
            .on('remove', this.handleRemoveUserShape, this);
          this.shapesStore.loadLocalShapes()
            .then(() => {
              if (this.filters.product.selected(this.userShapesProduct)) {
                this.update();
              }
            });
          this.hub.emit('change:userShapesProduct', product);
        }
      });
    }

    const products = this.productsStore.getProducts();
    const serverShapesProducts = products.filter(
      this.filters.product.serverShapesProduct
    );
    this.shapesStore.clearProductsShapes();
    if (serverShapesProducts.length > 0) {
      this.shapesStore.loadShapesFromProducts(serverShapesProducts);
    }
    this.update();
  },
  handleAddUserShape(userShapes) {
    this.mapView.hub.once('change:in-map-datasets', () => {
      // select newly added shapes
      if (this.mapView.selectDatasets) {
        this.mapView.selectDatasets(
          userShapes.map(this._getDatasetForUserShape, this)
        );
      }
    });
    this.update();
  },
  handleUpdateUserShape(userShapes) {
    this.mapView.updateDatasets(
      userShapes.map(
        (shape) => Object.assign(
          getInfoForShape(this.userShapesProduct.id, shape),
          {serverId: SHAPES_API_SERVER_ID}
        )
      )
    );
    this.update();
  },
  handleRemoveUserShape(userShapeIds) {
    userShapeIds.forEach((userShapeId) => {
      var dataset = this._getDatasetForUserShapeWithId(userShapeId);
      if (dataset) {
        this.mapView.removeDataset(dataset);
      }
    });
    this.update();
  },
});
