/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  emptyFunction,
  scheduleTask,
  negate,
  randomString,
  EventsHub,
  UserStorage,
} from '../../syntool.js';
import {HashMap} from '../../utils/hash_map.js';
import {Shape} from './shape.js';

var MAX_HISTORY_SIZE = 1000;

function getShapeId(shape) {
  return shape.id;
}

export class ShapesStore {
  constructor() {
    this.hub = new EventsHub();
    this.userStorage = new UserStorage('user_shapes', 'shapes', true);
    this.shapesByProductId = new Map();
    this.serverShapes = new Map();
    this.localShapes = {};
    this.idsByHash = new HashMap(this.sameShapeFromId.bind(this));
    this.history = {
      undo: [],
      redo: [],
    };
  }

  get canUndo() {
    return this.history.undo.length > 0;
  }
  get canRedo() {
    return this.history.redo.length > 0;
  }

  toJSON() {
    return Object.values(this.localShapes).map(Shape.toExternalJSON);
  }
  fromJSON(json) {
    var shapes = json.map(this.createFromJSON, this).filter(Boolean);
    var newShapes = shapes.filter(negate(this.isShapeDuplicate), this);
    this.add(newShapes);
  }

  sameShapeFromId(shape, id) {
    return JSON.stringify(shape) === JSON.stringify(this.get(id));
  }
  isShapeDuplicate(shape) {
    return shape && this.idsByHash.has(shape);
  }

  loadShapesFromProducts(products) {
    for (const product of products) {
      this.loadShapesFromProduct(product);
    }
  }
  loadShapesFromProduct(product) {
    const shapes = this.serverShapes;
    const shapesByProductId = this.shapesByProductId;

    const productShapes = [];
    if (shapesByProductId.has(product.id)) {
      for (const shape of shapesByProductId.get(product.id)) {
        shapes.delete(shape.id);
        this.idsByHash.delete(shape);
      }
    }
    shapesByProductId.set(product.id, productShapes);

    for (const shapeJSON of product.shapes) {
      const shape = this.createFromJSON(Object.assign({}, shapeJSON));

      if (shape) {
        shape.id = shapeJSON.id ? shapeJSON.id : productShapes.length;
        shapes.set(shape.id, shape);
        productShapes.push(shape);
        this.idsByHash.add(shape);
      }
    }
  }
  clearProductsShapes() {
    // console.log('ShapesStore#clearProductsShapes');

    this.serverShapes.clear();
    this.shapesByProductId.clear();
  }
  loadLocalShapes() {
    // console.log('ShapesStore#loadLocalShapes');

    var self = this;

    return (
      this.userStorage.forEach(function(value, key) {
        value.id = key;

        var shape = Shape.create(value);

        if (shape) {
          self.localShapes[key] = shape;
          self.idsByHash.add(shape);
        }
      }, 'load local shapes').catch(emptyFunction)
    );
  }

  _addIdToShape(shape) {
    if (shape) {
      var shapeId;
      var prefix = (
        (
          shape.start.getTime() === DATE_MIN_TIME
          ? ''
          : shape.start.toISOString().replace(/\.\d+Z$|\D/g, '') + '_'
        ) +
        (
          shape.end.getTime() === DATE_MAX_TIME
          ? ''
          : shape.end.toISOString().replace(/\.\d+Z$|\D/g, '') + '_'
        )
      );

      do {
        shapeId = prefix + randomString(5);
      } while (shapeId in this.localShapes);
      shape.id = shapeId;
    }
    return shape;
  }
  createFromJSON(json) {
    // console.log('ShapesStore#createFromJSON', json);

    return this._addIdToShape(Shape.fromJSON(json));
  }
  create(props) {
    // console.log('ShapesStore#create', props);

    return this._addIdToShape(Shape.create(props));
  }
  duplicate(shape) {
    // console.log('ShapesStore#duplicate', shape);

    return this.create(Shape.toInternalJSON(shape));
  }
  add(shapes) {
    // console.log('ShapesStore#add', shapes);

    if (Array.isArray(shapes)) {
      shapes = shapes.filter(Boolean);
    } else {
      shapes = [shapes];
    }
    if (shapes.length > 0) {
      this.applyAction({type: 'add', payload: shapes});
    }
  }
  get(id) {
    return this.serverShapes.get(id) || this.localShapes[id] || null;
  }
  getShapesForProduct(productId) {
    return this.shapesByProductId.get(productId) || [];
  }
  update(shapeIds, updates) {
    // console.log('ShapesStore#update', shapeIds, updates);

    if (!Array.isArray(shapeIds)) {
      shapeIds = [shapeIds];
    }

    var payload = shapeIds.map(function(shapeId) {
      var oldShape;
      var newShapeProps;
      var newShape;

      oldShape = this.localShapes[shapeId];
      if (!oldShape) {
        console.warn(
          'ShapesStore#update',
          'Cannot update a non-existing shape (' + shapeId + ').\n' +
          'You should create it first using ShapesStore#create'
        );
        return null;
      }

      newShapeProps = Object.assign({}, oldShape, updates);
      if (oldShape.arrow && updates.arrow) {
        newShapeProps.arrow = Object.assign({}, oldShape.arrow, updates.arrow);
      }
      newShape = Shape.create(newShapeProps);
      if (!newShape || Shape.same(oldShape, newShape)) {
        return null;
      }

      return newShape;
    }, this).filter(Boolean);

    this.applyAction({type: 'update', payload: payload});
  }
  remove(shapeIds) {
    // console.log('ShapesStore#remove', shapeIds);

    if (!Array.isArray(shapeIds)) {
      shapeIds = [shapeIds];
    }
    this.applyAction({type: 'remove', payload: shapeIds});
  }

  _pushUndo(action, oldShapes, newShapes) {
    this.history.undo.push(this.reverseAction(action, oldShapes, newShapes));
    if (this.history.undo.length > MAX_HISTORY_SIZE) {
      this.history.undo.shift();
    }
    this.history.redo = [];
  }
  undo() {
    // console.log('ShapesStore#undo');

    var self = this;

    if (this.history.undo.length > 0) {
      this.applyAction(
        this.history.undo[this.history.undo.length - 1],
        function(action, oldShapes, newShapes) {
          self.history.undo.pop();
          self.history.redo.push(self.reverseAction(action, oldShapes, newShapes));
        }
      );
      return true;
    }
    return false;
  }
  redo() {
    // console.log('ShapesStore#redo');

    var self = this;

    if (this.history.redo.length > 0) {
      this.applyAction(
        this.history.redo[this.history.redo.length - 1],
        function(action, oldShapes, newShapes) {
          self.history.undo.push(self.reverseAction(action, oldShapes, newShapes));
          self.history.redo.pop();
        }
      );
      return true;
    }
    return false;
  }

  reverseAction(action, oldShapes, newShapes) {
    switch (action.type) {
    case 'add':
      return {
        type   : 'remove',
        payload: newShapes.map(getShapeId),
      };
    case 'update':
      return {
        type   : 'update',
        payload: oldShapes,
      };
    case 'remove':
      return {
        type   : 'add',
        payload: oldShapes,
      };
    default:
      return null;
    }
  }
  applyAction(action, done) {
    var oldShapes = null;
    var newShapes = null;

    if (!action || !action.payload || action.payload.length === 0) {
      return;
    }

    switch (action.type) {
    case 'add':
      action.payload.forEach((shape) => {
        this.localShapes[shape.id] = shape;
        this.idsByHash.add(shape);

        this.userStorage
          .set(shape.id, Shape.toInternalJSON(shape))
          .catch(emptyFunction);
      });
      newShapes = action.payload;
      // console.log('Added shapes', action.payload);
      break;
    case 'update':
      oldShapes = action.payload.map((shape) => {
        var oldShape = this.localShapes[shape.id];

        this.localShapes[shape.id] = shape;
        this.idsByHash.delete(oldShape);
        this.idsByHash.add(shape);

        this.userStorage
          .set(shape.id, Shape.toInternalJSON(shape))
          .catch(emptyFunction);

        return oldShape;
      });
      newShapes = action.payload;
      // console.log('Updated shapes', action.payload);
      break;
    case 'remove':
      oldShapes = action.payload.map((shapeId) => {
        var oldShape = this.localShapes[shapeId];

        this.idsByHash.delete(oldShape);
        delete this.localShapes[shapeId];

        this.userStorage.delete(shapeId).catch(emptyFunction);
        return oldShape;
      });
      // console.log('Removed shapes with ids', action.payload);
      break;
    default:
      return;
    }

    if (done) {
      done(action, oldShapes, newShapes);
    } else {
      this._pushUndo(action, oldShapes, newShapes);
    }
    this.emitEventForAction(action);
  }
  emitEventForAction(action) {
    scheduleTask(() => this.hub.emit(action.type, action.payload));
  }
}
