/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Dataset, Types} from '../syntool.js';
import {colorFromCSSColor} from '../utils/color.js';
import {OLLayer} from '../utils/ol_layer.js';
import {layerExtension} from './extensions/layer_extension.js';
import {metadataExtension} from './extensions/metadata_extension.js';
import {dataFieldsExtension} from './extensions/data_fields_extension.js';
import {auxiliaryDataExtension} from './extensions/auxiliary_data_extension.js';
import {dataLoadingExtension} from './extensions/data_loading_extension.js';
import {dataMatrixExtension} from './extensions/data_matrix_extension.js';
import {StreamlineParticles} from './streamlines_dataset/streamline_particles.js';
import {StreamlineRendererWebGL} from './streamlines_dataset/streamline_renderer_webgl.js';
import {StreamlineRendererCanvas} from './streamlines_dataset/streamline_renderer_canvas.js';

var streamlinesExtension = {
  initialize(info) {
    window.original();
    this._renderer = null;
    this._updateAnimatingState = (
      this.syntool.retryIfLowMemory(this._updateAnimatingState)
    );
  },
  createLayer() {
    return new OLLayer(this.name);
  },
  hasResources() {
    return window.original() || !!this._renderer;
  },
  destroy() {
    if (this._renderer) {
      this._renderer.destroy();
      this._renderer = null;
    }
    window.original();
  },
  handleVisibilityChange() {
    window.original();
    this._updateAnimatingState();
  },
  handleDataLoad(response) {
    window.original();
    this._updateAnimatingState();
  },

  _updateAnimatingState() {
    const shouldBeAnimating = this.isVisible();
    if (shouldBeAnimating && !this._renderer && this.dataMatrix) {
      var product = this.getProduct();
      var color = colorFromCSSColor(this.getUniformFieldValue('color'));
      var RendererClass;

      if (StreamlineRendererWebGL.isSupported) {
        RendererClass = StreamlineRendererWebGL;
      } else if (StreamlineRendererCanvas.isSupported) {
        RendererClass = StreamlineRendererCanvas;
      } else {
        this.syntool.notificationCenter.showError(
          'Syntool uses the WebGL technology to display streamlines but ' +
          'your browser does not seem to support it. Syntool will still ' +
          'work but streamlines will not be displayed. You can learn more ' +
          'about WebGL at https://get.webgl.org'
        );
        return;
      }

      this._renderer = new RendererClass({
        mapView       : this.syntool.mapView,
        container     : this.layerManager.layer.div,
        color         : color,
        particlesCount: Math.floor(product.density * 7500),
        particlesSpeed: product.speed,
        dataMatrix    : this.dataMatrix,

        getReferencedFieldName: this.getReferencedFieldName.bind(this),
        getUniformFieldValue  : this.getUniformFieldValue.bind(this),
      });
      this.layerManager.layer.setRenderer(this._renderer);
    }
    this._renderer?.setAnimating(shouldBeAnimating);
  },
};

export const StreamlinesDataset = Dataset.createClass({
  static: {
    name        : 'StreamlinesDataset',
    productTypes: ['STREAMLINES'],
    renderFields: {
      modulus: ['ref', Types.number],
      angle  : ['ref', Types.number],
      color  : [Types.Color],
    },
    validateConfig: Types.shape({
      density: Types.number.gt(0).lte(2),
      speed: Types.number.gt(0),
    }),
    normalizeProductConfig(productConfig) {
      if (!('density' in productConfig)) {
        productConfig.density = 1;
      }
      if (!('speed' in productConfig)) {
        productConfig.speed = 1;
      }
      return productConfig;
    },
    Particles: StreamlineParticles,
    WebGLRenderer: StreamlineRendererWebGL,
    CanvasRenderer: StreamlineRendererCanvas,
  },
  extensions: [
    layerExtension,
    metadataExtension,
    dataFieldsExtension,
    auxiliaryDataExtension,
    dataLoadingExtension,
    dataMatrixExtension,
    streamlinesExtension,
  ],
});
window.ODL = Object.assign(window.ODL || {}, {StreamlinesDataset});
