/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {DEG_IN_RAD, randomBetween, clamp, modulo} from '../../syntool.js';

export class StreamlineParticles {
  // #region Public

  constructor(
    numParticles,
    speed,
    getReferencedFieldName,
    getUniformFieldValue,
    dataMatrix
  ) {
    numParticles = Math.ceil(numParticles);

    this.speed = speed;
    this.speedScale = speed;
    this.totalParticlesCount = numParticles;
    this.visibleParticlesCount = numParticles;
    this.bounds = dataMatrix.bounds;
    this.wraps = dataMatrix.wraps;
    this.dataWidth  = dataMatrix.getWidth();
    this.dataHeight = dataMatrix.getHeight();

    this.particles = new Float32Array(numParticles * 6);
    this.particlesTTL = new Uint8Array(numParticles);
    this._data = new Float32Array(this.dataWidth * this.dataHeight * 2);

    const uniformFields = [
      getUniformFieldValue('modulus'),
      getUniformFieldValue('angle'),
    ];
    const fieldNames = [
      getReferencedFieldName('modulus'),
      getReferencedFieldName('angle'),
    ];
    const valuesArray = [0, 0];

    for (let j = 0, k = 0; j < this.dataHeight; j++) {
      for (let i = 0; i < this.dataWidth; i++) {
        const values = (
          dataMatrix.getRealValuesAt(i, j, fieldNames, valuesArray)
        );

        if (values) {
          if (uniformFields[0] !== false) {
            values[0] = uniformFields[0];
          }
          if (uniformFields[1] !== false) {
            values[1] = uniformFields[1];
          }
          this._data[k++] = values[0] * Math.cos(values[1] * DEG_IN_RAD);
          this._data[k++] = values[0] * Math.sin(values[1] * DEG_IN_RAD);
        } else {
          this._data[k++] = NaN;
          this._data[k++] = NaN;
        }
      }
    }

    this._randomParticleX = null;
    this._randomParticleY = null;
  }

  updateViewportBounds(width, height, visibleBounds, zoom) {
    const {
      top: intersectionTop,
      bottom: intersectionBottom,
      left: intersectionMainLeft,
      right: intersectionMainRight,
      wrapLeft: intersectionWrapLeft,
      wrapRight: intersectionWrapRight,
    } = visibleBounds;

    const intersectionMainWidth = intersectionMainRight - intersectionMainLeft;
    const intersectionWrapWidth = intersectionWrapRight - intersectionWrapLeft;
    const intersectionHeight    = intersectionTop - intersectionBottom;

    const intersectsMainBounds = (
      intersectionMainWidth > 0 && intersectionHeight > 0
    );
    const intersectsWrapBounds = (
      intersectionWrapWidth > 0 && intersectionHeight > 0
    );
    let intersectionWidth;

    if (intersectsMainBounds && intersectsWrapBounds) {
      intersectionWidth = intersectionMainWidth + intersectionWrapWidth;
      this._randomParticleX = function() {
        const r = randomBetween(0, intersectionWidth);

        if (r < intersectionMainWidth) {
          return r + intersectionMainLeft;
        }

        return r - intersectionMainWidth + intersectionWrapLeft;
      };
    } else if (intersectsMainBounds) {
      intersectionWidth = intersectionMainWidth;
      this._randomParticleX = randomBetween.bind(null, intersectionMainLeft, intersectionMainRight);
    } else if (intersectsWrapBounds) {
      intersectionWidth = intersectionWrapWidth;
      this._randomParticleX = randomBetween.bind(null, intersectionWrapLeft, intersectionWrapRight);
    } else {
      intersectionWidth = 0;
      this._randomParticleX = null;
    }

    this._randomParticleY = randomBetween.bind(null, intersectionBottom, intersectionTop);

    const dataResolution = (
      (this.bounds.width * this.bounds.height) /
      (this.dataWidth * this.dataHeight)
    );
    const renderedResolution = (
      (intersectionWidth * intersectionHeight) / (width * height)
    );
    const resolution = Math.sqrt(renderedResolution / dataResolution);
    const densityFactor = Math.pow(2, Math.ceil(Math.log2(resolution * 10)));

    this.visibleParticlesCount = this.totalParticlesCount * densityFactor;
    this.speedScale = this.speed * 2 * 100_000 / Math.pow(2, zoom);
  }
  updateParticlePositions() {
    const {
      totalParticlesCount,
      visibleParticlesCount,
      particles,
      particlesTTL,
      _data: data,
      dataWidth,
      dataHeight,
      bounds: {
        left: boundsLeft,
        right: boundsRight,
        top: boundsTop,
        bottom: boundsBottom,
        width: boundsWidth,
        height: boundsHeight,
      },
      wraps,
      speedScale,
    } = this;
    let x;
    let y;
    let px;
    let py;
    let x1;
    let y1;
    let i11;
    let i12;
    let i21;
    let i22;
    let d11;
    let d12;
    let d21;
    let d22;
    let dx;
    let dy;
    let a;
    let dvx;
    let dvy;
    for (
      let index = 0,
          CURR_X = 0,
          CURR_Y = 1,
          CURR_Z = 2,
          NEXT_X = 3,
          NEXT_Y = 4,
          NEXT_Z = 5;
      index < totalParticlesCount;
      (
        index++,
        CURR_X += 6,
        CURR_Y += 6,
        CURR_Z += 6,
        NEXT_X += 6,
        NEXT_Y += 6,
        NEXT_Z += 6
      )
    ) {
      // copy old positions
      particles[CURR_X] = particles[NEXT_X];
      particles[CURR_Y] = particles[NEXT_Y];
      particles[CURR_Z] = particles[NEXT_Z];

      if (particlesTTL[index] === 0 && index < visibleParticlesCount) {
        // The particle is dead (✝ RIP)

        // Hide it
        // z = -1 => not visible
        particles[CURR_Z] = -1;
        particles[NEXT_Z] = -1;

        if (this._randomParticleX && this._randomParticleY) {
          // Schedule its reincarnation
          particlesTTL[index] = Math.floor(randomBetween(100, 200));
          particles[NEXT_X] = this._randomParticleX();
          particles[NEXT_Y] = this._randomParticleY();
        }
        continue;
      }
      if (particlesTTL[index] > 0) {
        particlesTTL[index]--;
      }

      x = particles[CURR_X];
      y = particles[CURR_Y];
      px = (x - boundsLeft) * dataWidth / boundsWidth;
      py = (boundsTop - y) * dataHeight / boundsHeight;

      if (wraps) {
        px = modulo(px, dataWidth);
      }

      // <Bilinear interpolation>

      /**
       *           x1     px x2
       *      |    :    | ¦  :    |
       *    ——+—————————+—¦———————+——
       *      |    :    | ¦  :    |
       *  y1⋅⋅| ⋅⋅ * ⋅⋅ | ¦⋅ * ⋅⋅ |⋅⋅y1
       *   |  |    :    | ¦  :    |
       * dy|——+—————————+—¦———————+——
       *  py--|----:----|-⨯--:----|--py
       *  y2⋅⋅| ⋅⋅ * ⋅⋅ | ¦⋅ * ⋅⋅ |⋅⋅y2
       *      |    :    | ¦  :    |
       *    ——+—————————+—¦———————+——
       *      |    :    | ¦  :    |
       *           x1<———>px x2
       *               dx
       */

      // we translate by half a pixel so that we interpolate from the center
      // of the pixel instead of its top left corner
      x1 = (px - 0.5)|0;
      y1 = (py - 0.5)|0;

      i11 = (y1 * dataWidth + x1) * 2;
      i12 = i11 + dataWidth * 2;
      i21 = i11 + 2;
      i22 = i12 + 2;
      d11 = data[i11];
      d12 = data[i12];
      d21 = data[i21];
      d22 = data[i22];

      dx = px - x1;
      dy = py - y1;
      dvx = 0;
      dvy = 0;

      if (!Number.isNaN(d11)) {
        a = (1 - dx) * (1 - dy);
        dvx += d11 * a;
        dvy += data[i11 + 1] * a;
      }
      if (!Number.isNaN(d12)) {
        a = (1 - dx) * dy;
        dvx += d12 * a;
        dvy += data[i12 + 1] * a;
      }
      if (!Number.isNaN(d21)) {
        a = dx * (1 - dy);
        dvx += d21 * a;
        dvy += data[i21 + 1] * a;
      }
      if (!Number.isNaN(d22)) {
        a = dx * dy;
        dvx += d22 * a;
        dvy += data[i22 + 1] * a;
      }

      // </Bilinear interpolation>

      if (dvx === 0 && dvy === 0) {
        // Particle stopped, hide it
        // z = -1 => not visible
        particles[CURR_Z] = -1;
        particles[NEXT_Z] = -1;
        continue;
      }

      // z = 0 => visible
      particles[NEXT_Z] = 0;

      // horizontal
      x += dvx * speedScale;
      if (wraps) {
        const currX = particles[CURR_X];
        if (currX < boundsLeft && x < boundsLeft) {
          particles[CURR_X] += boundsWidth;
          x += boundsWidth;
        } else if (currX > boundsRight && x > boundsRight) {
          particles[CURR_X] -= boundsWidth;
          x -= boundsWidth;
        }
      } else if (x < boundsLeft) {
        x = boundsLeft;
        // Kill the particle
        particlesTTL[index] = 0;
      } else if (x > boundsRight) {
        x = boundsRight;
        // Kill the particle
        particlesTTL[index] = 0;
      }
      particles[NEXT_X] = x;

      // vertical
      y += dvy * speedScale;
      if (y < boundsBottom) {
        y = boundsBottom;
        // Kill the particle
        particlesTTL[index] = 0;
      } else if (y > boundsTop) {
        y = boundsTop;
        // Kill the particle
        particlesTTL[index] = 0;
      }
      particles[NEXT_Y] = y;

      // check mask
      px = (x - boundsLeft) * dataWidth / boundsWidth;
      py = (boundsTop - y) * dataHeight / boundsHeight;
      if (wraps) {
        px = modulo(px, dataWidth);
      }
      // Find the coordinates of the data pixel that will contain the particle
      // Note: x|0 === Math.round(x - 0.5)
      px = clamp(px|0, 0, dataWidth  - 1);
      py = clamp(py|0, 0, dataHeight - 1);
      if (Number.isNaN(this._data[(py * dataWidth + px) * 2])) {
        // Kill the particle
        particlesTTL[index] = 0;
        // and hide it
        // z = -1 => not visible
        particles[CURR_Z] = -1;
        particles[NEXT_Z] = -1;
      }
    }
  }

  // #endregion
}
