/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {emptyFunction, idlRangeWorldsAway, createClass, Bounds} from '../../syntool.js';
import {StreamlineParticles} from './streamline_particles.js';

export const StreamlineRenderer = createClass([
  /*Public*/ {
    constructor: function StreamlineRenderer(options) {
      this.mapView = options.mapView;
      this.container = options.container;
      this.canvas = null;
      this._color = options.color;
      this._isAnimated = false;
      this._rafId = null;

      this.handleAnimationFrame = this.handleAnimationFrame.bind(this);

      this._particles = new StreamlineParticles(
        options.particlesCount,
        options.particlesSpeed,
        options.getReferencedFieldName,
        options.getUniformFieldValue,
        options.dataMatrix
      );
      this.intersectsMainBounds = false;
      this.intersectsWrapBounds = false;
      this._top = 0;
      this._left = 0;
      this._dx = 0;
      this._dy = 0;
      this.newCanvas();
      this.init();
      this.setExtent(this.mapView.getViewportProjectedBounds());
    },

    setAnimating(isAnimated) {
      if (this._isAnimated !== isAnimated) {
        this._isAnimated = isAnimated;
        if (isAnimated) {
          this.startAnimation();
        } else {
          this.stopAnimation();
        }
      }
    },
  },
  /*Map layer interface*/ {
    destroy() {
      this.container = null;
    },
    setSize(width, height) {
      let top = 0;
      let left = 0;

      if (!this.mapView.shouldWrapDateLine()) {
        const viewportExtent = this.mapView.getViewportProjectedBounds();
        const viewportWidth = Bounds.getWidth(viewportExtent);
        const viewportHeight = Bounds.getHeight(viewportExtent);

        if (viewportWidth > 0 && viewportHeight > 0) {
          const visibleExtent = this.mapView.getVisibleProjectedBounds();
          const topRatio = (
            -(Bounds.getTop(visibleExtent) - Bounds.getTop(viewportExtent))
            / viewportHeight
          );
          const leftRatio = (
            (Bounds.getLeft(visibleExtent) - Bounds.getLeft(viewportExtent))
            / viewportWidth
          );

          top = height * topRatio;
          left = width * leftRatio;
          width *= Bounds.getWidth(visibleExtent) / viewportWidth;
          height *= Bounds.getHeight(visibleExtent) / viewportHeight;
        } else {
          width = 0;
          height = 0;
        }
      }

      this._top = top;
      this._left = left;

      var canvas = this.canvas;
      canvas.style.top = (this._top - this._dy) + 'px';
      canvas.style.left = (this._left - this._dx) + 'px';
      canvas.style.width = width + 'px';
      canvas.style.height = height + 'px';
      canvas.width = width;
      canvas.height = height;
      this.resized();
    },
    setExtent(viewportBounds) {
      if (this.canvas) {
        const {width, height} = this.canvas;
        const {bounds, wraps} = this._particles;

        let wrapDelta = NaN;
        const worldBounds = this.mapView.getWorldProjectedBounds();
        if (worldBounds) {
          const worldWidth = Bounds.getWidth(worldBounds);
          const viewCenterX = Bounds.getCenterX(
            this.mapView.shouldWrapDateLine() ? viewportBounds : worldBounds
          );
          const [leftWorldsAway, rightWorldsAway] = idlRangeWorldsAway(
            bounds.left,
            bounds.right,
            viewCenterX,
            worldWidth
          );
          wrapDelta = -(leftWorldsAway || rightWorldsAway || NaN) * worldWidth;
        }

        const visibleMapBounds = (
          this.mapView.shouldWrapDateLine()
          ? viewportBounds
          : Bounds.intersection(
            viewportBounds,
            this.mapView.getMaxProjectedBounds()
          )
        );
        const visibleBounds = {
          top   : Bounds.getTop(visibleMapBounds),
          bottom: Bounds.getBottom(visibleMapBounds),
          left  : Bounds.getLeft(visibleMapBounds),
          right : Bounds.getRight(visibleMapBounds),
        };

        const top = Math.min(bounds.top, visibleBounds.top);
        const bottom = Math.max(bounds.bottom, visibleBounds.bottom);
        const left = Math.max(bounds.left, visibleBounds.left);
        const right = Math.min(bounds.right, visibleBounds.right);
        const wrapLeft = Math.max(bounds.left, visibleBounds.left - wrapDelta);
        const wrapRight = Math.min(bounds.right, visibleBounds.right - wrapDelta);

        this._particles.updateViewportBounds(
          width,
          height,
          {
            top,
            bottom,
            left: wraps ? visibleBounds.left : left,
            right: wraps ? visibleBounds.right : right,
            wrapLeft: wraps ? NaN : wrapLeft,
            wrapRight: wraps ? NaN : wrapRight,
          },
          this.mapView.getZoom()
        );

        this.intersectsMainBounds = (left < right && bottom < top);
        this.intersectsWrapBounds = (wrapLeft < wrapRight && bottom < top);
        this.updateViewportBounds(width, height, visibleBounds, wrapDelta);
      }
    },
    invalidate() {
      // Hide canvas when invalid
      this.canvas.style.visibility = 'hidden';
    },
    setVisibility: emptyFunction,

    resetOffset() {
      this.setOffset(-this._dx, -this._dy);
    },
    setOffset(dx, dy) {
      this._dx += dx;
      this._dy += dy;
      if (this.canvas) {
        this.canvas.style.left = (this._left - this._dx) + 'px';
        this.canvas.style.top = (this._top - this._dy) + 'px';
      }
    },
  },
  /*Protected*/ {
    init   : emptyFunction,
    resized: emptyFunction,
    destroy: emptyFunction,
    render : emptyFunction,

    updateViewportBounds: emptyFunction,

    newCanvas() {
      var oldCanvas = this.canvas;
      var canvas = document.createElement('canvas');
      canvas.style.position = 'absolute';
      canvas.style.left = (this._left - this._dx) + 'px';
      canvas.style.top = (this._top - this._dy) + 'px';

      this.canvas = canvas;
      this.invalidate();
      this.container.appendChild(this.canvas);

      if (oldCanvas) {
        this.container.removeChild(oldCanvas);
        canvas.setAttribute('style', oldCanvas.getAttribute('style'));
        canvas.width = oldCanvas.width;
        canvas.height = oldCanvas.height;
      }
    },
    startAnimation() {
      if (this._rafId === null) {
        this.setExtent(this.mapView.getViewportProjectedBounds());
        this._rafId = requestAnimationFrame(this.handleAnimationFrame);
      }
    },
    stopAnimation() {
      if (this._rafId !== null) {
        cancelAnimationFrame(this._rafId);
        this._rafId = null;
      }
    },

    handleAnimationFrame() {
      if (this._isAnimated) {
        this._rafId = requestAnimationFrame(this.handleAnimationFrame);
        if (this._particles) {
          this._particles.updateParticlePositions();
          this.render();
          // Show canvas after render
          this.canvas.style.visibility = 'visible';
        }
      }
    },
  },
]);
