/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {createClass} from '../../syntool.js';
import {colorToCSSColor} from '../../utils/color.js';
import {StreamlineRenderer} from './streamline_renderer.js';

export const StreamlineRendererCanvas = createClass(StreamlineRenderer, [
  /*Public*/ {
    static: {
      get isSupported() {
        let value = false;
        try {
          const canvas = document.createElement('canvas');
          const context = canvas.getContext('2d');
          value = !!context;
        } catch (e) {
          /* ignore */
        }
        Object.defineProperty(this, 'isSupported', {value});
        return value;
      },
    },

    constructor: function StreamlineRendererCanvas(options) {
      this._offsetX = 0;
      this._offsetY = 0;
      this._scaleX  = 1;
      this._scaleY  = 1;
      this._wrapDelta = 0;
      this._shouldReset = true;

      this.handleContextLost = this.handleContextLost.bind(this);
      this.handleContextRestored = this.handleContextRestored.bind(this);

      StreamlineRenderer.call(this, options);
    },
  },
  /*Protected*/ {
    handleContextLost() {
      this.stopAnimation();
    },
    handleContextRestored() {
      if (this._isAnimated) {
        this.startAnimation();
      }
    },

    newCanvas() {
      this.canvas?.removeEventListener('contextlost', this.handleContextLost);
      this.canvas?.removeEventListener('contextrestored', this.handleContextRestored);

      StreamlineRenderer.prototype.newCanvas.call(this);

      this.context = this.canvas?.getContext('2d', {storage: 'discardable'});
      if (this.context) {
        this.canvas.addEventListener('contextlost', this.handleContextLost);
        this.canvas.addEventListener('contextrestored', this.handleContextRestored);
        this.resized();
      }
    },
    getContext() {
      return this.context;
    },
    updateViewportBounds(width, height, viewportBounds, wrapDelta) {
      this._shouldReset = true;
      this._offsetX = -viewportBounds.left;
      this._offsetY = -viewportBounds.top;
      this._scaleX  =  width  / (viewportBounds.right - viewportBounds.left);
      this._scaleY  = -height / (viewportBounds.top - viewportBounds.bottom);
      this._wrapDelta = this._scaleX * wrapDelta;
    },

    resized() {
      this.context.lineWidth = 1.2;
      this.context.lineCap = 'round';
      this.context.strokeStyle = colorToCSSColor(this._color);
    },
    render() {
      var ctx = this.getContext();
      if (!ctx) {
        return;
      }

      var canvas = ctx.canvas;
      if (this._shouldReset) {
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        this._shouldReset = false;
      } else {
        ctx.globalCompositeOperation = 'destination-in';
        ctx.fillStyle = 'rgba(0, 0, 0, .96)';
        ctx.fillRect(0, 0, canvas.width, canvas.height);

        // ctx.globalCompositeOperation = 'source-atop';
        // ctx.fillStyle = 'rgba(0, 0, 255, 0.01)';
        // ctx.fillRect(0, 0, canvas.width, canvas.height);

        ctx.globalCompositeOperation = 'source-over';
      }

      this.renderParticles(ctx);
    },
    renderParticles(ctx) {
      var particles = this._particles.particles;
      var canvasWidth = ctx.canvas.width;

      ctx.beginPath();
      for (var i = 0, l = particles.length; i < l; i += 6) {
        this.renderParticle(ctx, canvasWidth, particles, i);
      }
      ctx.stroke();
    },
    renderParticle(ctx, canvasWidth, particles, index) {
      var CURR_X = index + 0;
      var CURR_Y = index + 1;
      var CURR_Z = index + 2;
      var NEXT_X = index + 3;
      var NEXT_Y = index + 4;

      if (particles[CURR_Z] < 0) {
        return;
      }

      var x1 = (particles[CURR_X] + this._offsetX) * this._scaleX;
      var y1 = (particles[CURR_Y] + this._offsetY) * this._scaleY;
      var x2 = (particles[NEXT_X] + this._offsetX) * this._scaleX;
      var y2 = (particles[NEXT_Y] + this._offsetY) * this._scaleY;

      var x1Inside = (x1 >= 0 && x1 < canvasWidth);
      var x2Inside = (x2 >= 0 && x2 < canvasWidth);

      if (!x1Inside || !x2Inside) {
        // At least one is outside
        if ((!x1Inside && x2Inside) || (x1Inside && !x2Inside)) {
          // Only one outside => cross-IDL
          ctx.moveTo(x1, y1);
          ctx.lineTo(x2, y2);
        }

        x1 += this._wrapDelta;
        x2 += this._wrapDelta;
      }

      ctx.moveTo(x1, y1);
      ctx.lineTo(x2, y2);
    },
  },
]);
