/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {NONPASSIVE, PASSIVE, createClass} from '../../syntool.js';
import {colorToNormalizedRGBArray} from '../../utils/color.js';
import {StreamlineRenderer} from './streamline_renderer.js';

var vshColorMask = `
  attribute vec3 aVertexPosition;

  uniform mat4 uPMatrix;

  void main(void) {
    gl_Position = uPMatrix * vec4(aVertexPosition, 1.0);
  }
`;
var fshColorMask = `
  precision highp float;

  void main(void) {
    gl_FragColor = vec4(0.0, 0.0, 1.0, 0.005);
  }
`;
var vshFBOTexture = `
  attribute vec3 aVertexPosition;
  attribute vec2 aTextureCoord;

  uniform mat4 uPMatrix;

  varying vec2 vTextureCoord;

  void main(void) {
    vTextureCoord = aTextureCoord;
    gl_Position = uPMatrix * vec4(aVertexPosition, 1.0);
  }
`;
var fshFBOTexture = `
  precision highp float;

  varying vec2 vTextureCoord;

  uniform sampler2D uSampler;

  void main(void) {
    vec4 texelColor = texture2D(uSampler, vTextureCoord);
      if (texelColor.a < 0.55)
          texelColor.a = 0.0;
      else
          texelColor = texelColor * texelColor.a;
    gl_FragColor = texelColor;
  }
`;
var vshParticles = `
  attribute vec3 aVertexPosition;

  uniform mat4 uPMatrix;

  varying vec4 vPosition;

  void main(void) {
    gl_Position = uPMatrix * vec4(aVertexPosition, 1.0);
    vPosition = gl_Position;
  }
`;
var fshParticles = `
  precision highp float;

  uniform vec3 uColor;

  varying vec4 vPosition;

  void main(void) {
    vec4 color;
    color.rgb = uColor;

    if (vPosition.z > 0.0) {
      color.a = 0.0;
    } else {
      color.a = 1.0;
    }

    gl_FragColor = color;
  }
`;

function glGetShader(gl, type, src) {
  if (gl.isContextLost()) {
    return null;
  }

  var shader = gl.createShader(type);

  if (gl.isContextLost()) {
    return null;
  }

  gl.shaderSource(shader, src);
  gl.compileShader(shader);

  if (
    !gl.getShaderParameter(shader, gl.COMPILE_STATUS) &&
    !gl.isContextLost()
  ) {
    console.log(gl.getShaderInfoLog(shader));
    return null;
  }

  return shader;
}
function glGetVBO(gl, dataArray, usage) {
  if (gl.isContextLost()) {
    return null;
  }

  var vbo = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, vbo);
  gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(dataArray), usage);
  return vbo;
}
function glCreateAndSetupTexture(gl) {
  if (gl.isContextLost()) {
    return null;
  }

  var texture = gl.createTexture();
  gl.bindTexture(gl.TEXTURE_2D, texture);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
  return texture;
}
function glSetFramebuffer(gl, fbo, width, height, clearColor) {
  if (gl.isContextLost()) {
    return;
  }

  // make this the framebuffer we are rendering to.
  gl.bindFramebuffer(gl.FRAMEBUFFER, fbo);

  // Tell webgl the viewport setting needed for framebuffer.
  gl.viewport(0, 0, width, height);
  if (clearColor) {
    //eslint-disable-next-line no-bitwise
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
  }
}
function createShaderProgram(gl, vertexShader, fragmentShader) {
  if (gl.isContextLost()) {
    return null;
  }

  vertexShader = glGetShader(gl, gl.VERTEX_SHADER, vertexShader);
  fragmentShader = glGetShader(gl, gl.FRAGMENT_SHADER, fragmentShader);

  var shaderProgram = gl.createProgram();
  gl.attachShader(shaderProgram, vertexShader);
  gl.attachShader(shaderProgram, fragmentShader);
  gl.linkProgram(shaderProgram);

  if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) {
    console.log('Could not initialise shaders');
  }

  gl.useProgram(shaderProgram);

  shaderProgram.vertexPositionAttribute = gl.getAttribLocation(shaderProgram, 'aVertexPosition');
  shaderProgram.textureCoordAttribute = gl.getAttribLocation(shaderProgram, 'aTextureCoord');

  shaderProgram.uSampler = gl.getUniformLocation(shaderProgram, 'uSampler');
  shaderProgram.colorUniform = gl.getUniformLocation(shaderProgram, 'uColor');
  shaderProgram.pMatrixUniform = gl.getUniformLocation(shaderProgram, 'uPMatrix');

  return shaderProgram;
}
function loseContext(gl) {
  if (gl && !gl.isContextLost()) {
    // This is the recommended mechanism for applications to
    // programmatically halt their use of the WebGL API.
    // @see https://www.khronos.org/registry/webgl/extensions/WEBGL_lose_context/
    var loseContextExtension = gl.getExtension('WEBGL_lose_context');
    if (loseContextExtension) {
      loseContextExtension.loseContext();
    }
  }
}

/**
 * Generates an orthogonal projection matrix with the given bounds
 *
 * @param {number} left - Left bound of the frustum
 * @param {number} right - Right bound of the frustum
 * @param {number} bottom - Bottom bound of the frustum
 * @param {number} top - Top bound of the frustum
 * @param {number} near - Near bound of the frustum
 * @param {number} far - Far bound of the frustum
 * @returns {Float32Array} 4x4 frustum matrix
 */
function orthogonalProjectionMatrix(left, right, bottom, top, near, far) {
  var lr = 1 / (left - right);
  var bt = 1 / (bottom - top);
  var nf = 1 / (near - far);
  var out = new Float32Array(16);
  out[0] = -2 * lr;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = -2 * bt;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = 2 * nf;
  out[11] = 0;
  out[12] = (left + right) * lr;
  out[13] = (top + bottom) * bt;
  out[14] = (far + near) * nf;
  out[15] = 1;
  return out;
}

export const StreamlineRendererWebGL = createClass(StreamlineRenderer, [
  /*Public*/ {
    static: {
      get isSupported() {
        let value = false;
        try {
          const canvas = document.createElement('canvas');
          const context = (
            canvas.getContext('webgl') ||
            canvas.getContext('experimental-webgl')
          );
          value = !!context;
        } catch (e) {
          /* ignore */
        }
        Object.defineProperty(this, 'isSupported', {value});
        return value;
      },
    },

    constructor: function StreamlineRendererWebGL(options) {
      this.gl = null;
      this._shaderPrograms = [];
      this._fbo = null;
      this._fboTexture = null;
      this._fboVerticesPositionBuffer = null;
      this._fboVerticesTextureCoordBuffer = null;
      this._particlesVerticesPositionBuffer = null;
      this._mainMatrix = null;
      this._wrapMatrix = null;
      this._pMatrix = orthogonalProjectionMatrix(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
      this._shouldReset = true;

      this.handleContextLost = this.handleContextLost.bind(this);
      this.handleContextRestored = this.handleContextRestored.bind(this);

      StreamlineRenderer.call(this, options);

      this._color = new Float32Array(colorToNormalizedRGBArray(this._color));
    },

    destroy() {
      this.newCanvas();
    },
  },
  /*Map layer interface*/ {
    setVisibility(visible) {
      if (visible) {
        // We should be visible, so we need a valid WebGL context
        if (this.gl && this.gl.isContextLost()) {
          // The context that we had is lost, to get a new one we need
          // a new canvas
          this.newCanvas();
        }
        if (!this.gl) {
          // We don't have a context, so we create one and initialize it
          this.init();
        }
      } else if (this.gl) {
        // We are not visible and we have a WebGL context, so we discard it
        // by discarding the canvas and replacing it with a new one
        this.newCanvas();
      }
    },
  },
  /*Protected*/ {
    handleContextLost(event) {
      event.preventDefault();
      this.deinit();
    },
    handleContextRestored(event) {
      this.init();
    },

    newCanvas() {
      if (this.canvas) {
        this.canvas.removeEventListener(
          'webglcontextlost',
          this.handleContextLost,
          NONPASSIVE
        );
        this.canvas.removeEventListener(
          'webglcontextrestored',
          this.handleContextRestored,
          PASSIVE
        );
        this.deinit();
        loseContext(this.gl);
        this.gl = null;
      }

      StreamlineRenderer.prototype.newCanvas.call(this);

      if (this.canvas) {
        this.canvas.addEventListener(
          'webglcontextlost',
          this.handleContextLost,
          NONPASSIVE
        );
        this.canvas.addEventListener(
          'webglcontextrestored',
          this.handleContextRestored,
          PASSIVE
        );
      }
    },
    getContext() {
      if (!this.gl) {
        var webGLOptions = {premultipliedAlpha: true, antialias: false};
        this.gl = (
          this.canvas.getContext('webgl', webGLOptions) ||
          this.canvas.getContext('experimental-webgl', webGLOptions)
        );
      }
      return this.gl;
    },
    updateViewportBounds(width, height, viewportBounds, wrapDelta) {
      var gl = this.getContext();
      if (!gl) {
        return;
      }

      this._shouldReset = true;

      if (this.intersectsMainBounds) {
        this._mainMatrix = orthogonalProjectionMatrix(
          viewportBounds.left, viewportBounds.right,
          viewportBounds.bottom, viewportBounds.top,
          -1.0, 1.0
        );
      } else {
        this._mainMatrix = null;
      }

      if (this.intersectsWrapBounds) {
        this._wrapMatrix = orthogonalProjectionMatrix(
          viewportBounds.left - wrapDelta, viewportBounds.right - wrapDelta,
          viewportBounds.bottom, viewportBounds.top,
          -1.0, 1.0
        );
      } else {
        this._wrapMatrix = null;
      }

      if (gl.viewportWidth !== width || gl.viewportHeight !== height) {
        gl.viewportWidth = width;
        gl.viewportHeight = height;
        this.resetTextures(gl);
      }
    },

    init() {
      var gl = this.getContext();
      if (gl && !gl.isContextLost()) {
        gl.viewportWidth = gl.canvas.width;
        gl.viewportHeight = gl.canvas.height;
        gl.clearColor(0.0, 0.0, 0.0, 0.0);
        gl.clearDepth(1.0);
        gl.lineWidth(2.0);
        gl.disable(gl.DEPTH_TEST);
        gl.depthFunc(gl.LEQUAL);
        gl.enable(gl.BLEND);
        gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

        this._shouldReset = true;
        this.initShaders(gl);
        this.initBuffers(gl);
        this.initTextures(gl);

        if (this._isAnimated) {
          this.startAnimation();
        }
      }
    },
    deinit() {
      this.stopAnimation();

      var gl = this.gl;
      if (gl && !gl.isContextLost()) {
        var numTextureUnits = gl.getParameter(gl.MAX_TEXTURE_IMAGE_UNITS);
        for (var unit = 0; unit < numTextureUnits; ++unit) {
          gl.activeTexture(gl.TEXTURE0 + unit);
          gl.bindTexture(gl.TEXTURE_2D, null);
          gl.bindTexture(gl.TEXTURE_CUBE_MAP, null);
        }
        gl.bindBuffer(gl.ARRAY_BUFFER, null);
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, null);
        gl.bindRenderbuffer(gl.RENDERBUFFER, null);
        gl.bindFramebuffer(gl.FRAMEBUFFER, null);

        this._shaderPrograms.forEach(function(program) {
          var shaders = gl.getAttachedShaders(program);
          shaders.forEach(gl.deleteShader, gl);
          gl.deleteProgram(program);
        });
        if (this._fbo) {
          gl.deleteFramebuffer(this._fbo);
        }
        if (this._fboTexture) {
          gl.deleteTexture(this._fboTexture);
        }
        if (this._fboVerticesPositionBuffer) {
          gl.deleteBuffer(this._fboVerticesPositionBuffer);
        }
        if (this._fboVerticesTextureCoordBuffer) {
          gl.deleteBuffer(this._fboVerticesTextureCoordBuffer);
        }
        if (this._particlesVerticesPositionBuffer) {
          gl.deleteBuffer(this._particlesVerticesPositionBuffer);
        }
      }

      this._shaderPrograms.length = 0;
      this._fbo = null;
      this._fboTexture = null;
      this._fboVerticesPositionBuffer = null;
      this._fboVerticesTextureCoordBuffer = null;
      this._particlesVerticesPositionBuffer = null;
    },

    initShaders(gl) {
      if (gl.isContextLost()) {
        return;
      }

      console.log('Init shaders');
      this._shaderPrograms = [
        createShaderProgram(gl, vshColorMask, fshColorMask),
        createShaderProgram(gl, vshFBOTexture, fshFBOTexture),
        createShaderProgram(gl, vshParticles, fshParticles),
      ].filter(Boolean);
    },
    initBuffers(gl) {
      if (gl.isContextLost()) {
        return;
      }

      console.log('Init buffers');

      var vertices = [
        1.0,  1.0,  0.0,
        -1.0,  1.0,  0.0,
        1.0, -1.0,  0.0,
        -1.0, -1.0,  0.0
      ];

      this._fboVerticesPositionBuffer = glGetVBO(gl, vertices, gl.STATIC_DRAW);

      var textureCoordinates = [
        1.0, 1.0,
        0.0, 1.0,
        1.0, 0.0,
        0.0, 0.0
      ];

      this._fboVerticesTextureCoordBuffer = glGetVBO(gl, textureCoordinates, gl.STATIC_DRAW);
      this._particlesVerticesPositionBuffer = gl.createBuffer();
    },
    initTextures(gl) {
      if (gl.isContextLost()) {
        return;
      }

      console.log('Init textures');

      gl.activeTexture(gl.TEXTURE0);
      this._fboTexture = glCreateAndSetupTexture(gl);
      this.resetTextures(gl);

      this._fbo = gl.createFramebuffer();
      gl.bindFramebuffer(gl.FRAMEBUFFER, this._fbo);
      gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, this._fboTexture, 0);

      gl.useProgram(this._shaderPrograms[1]);
      gl.uniform1i(this._shaderPrograms[1].uSampler, 0);
    },

    resetTextures(gl) {
      if (gl.isContextLost()) {
        return;
      }

      if (this._fboTexture) {
        gl.texImage2D(
          gl.TEXTURE_2D,
          0,
          gl.RGBA,
          gl.viewportWidth,
          gl.viewportHeight,
          0,
          gl.RGBA,
          gl.UNSIGNED_BYTE,
          null
        );
      }
    },

    render() {
      var gl = this.getContext();
      if (!gl || gl.isContextLost()) {
        return;
      }

      // When the context is lost, in some edge cases, we get back the context
      // without getting 'webglcontextrestored' event, and then render() gets
      // called without being initialized.
      if (this._shaderPrograms.length === 0) {
        this.init();
      }

      // draw in the FBO
      glSetFramebuffer(gl, this._fbo, gl.viewportWidth, gl.viewportHeight, this._shouldReset);

      if (this._shouldReset) {
        this._shouldReset = false;
      } else {
        this.renderColoredAlpha(gl);
      }

      this.renderParticles(gl);

      // draw on screen
      glSetFramebuffer(gl, null, gl.viewportWidth, gl.viewportHeight, true);

      this.renderFBOTexture(gl);
    },
    renderColoredAlpha(gl) {
      if (gl.isContextLost()) {
        return;
      }

      var shaderProgram = this._shaderPrograms[0];
      gl.useProgram(shaderProgram);

      gl.enableVertexAttribArray(shaderProgram.vertexPositionAttribute);

      gl.bindBuffer(gl.ARRAY_BUFFER, this._fboVerticesPositionBuffer);
      gl.vertexAttribPointer(shaderProgram.vertexPositionAttribute, 3, gl.FLOAT, false, 0, 0);

      gl.uniformMatrix4fv(shaderProgram.pMatrixUniform, false, this._pMatrix);

      gl.drawArrays(gl.TRIANGLE_STRIP, 0, 4);

      gl.disableVertexAttribArray(shaderProgram.vertexPositionAttribute);
    },
    renderParticles(gl) {
      if (gl.isContextLost()) {
        return;
      }

      var shaderProgram = this._shaderPrograms[2];
      gl.useProgram(shaderProgram);

      gl.enableVertexAttribArray(shaderProgram.vertexPositionAttribute);

      gl.bindBuffer(gl.ARRAY_BUFFER, this._particlesVerticesPositionBuffer);
      gl.bufferData(gl.ARRAY_BUFFER, this._particles.particles, gl.DYNAMIC_DRAW);
      gl.vertexAttribPointer(shaderProgram.vertexPositionAttribute, 3, gl.FLOAT, false, 0, 0);

      gl.uniform3fv(shaderProgram.colorUniform, this._color);

      if (this._mainMatrix) {
        gl.uniformMatrix4fv(shaderProgram.pMatrixUniform, false, this._mainMatrix);
        gl.drawArrays(gl.LINES, 0, this._particles.particles.length / 3);
      }
      if (this._wrapMatrix) {
        gl.uniformMatrix4fv(shaderProgram.pMatrixUniform, false, this._wrapMatrix);
        gl.drawArrays(gl.LINES, 0, this._particles.particles.length / 3);
      }

      gl.disableVertexAttribArray(shaderProgram.vertexPositionAttribute);
    },
    renderFBOTexture(gl) {
      if (gl.isContextLost()) {
        return;
      }

      var shaderProgram = this._shaderPrograms[1];
      gl.useProgram(shaderProgram);

      gl.enableVertexAttribArray(shaderProgram.vertexPositionAttribute);
      gl.enableVertexAttribArray(shaderProgram.textureCoordAttribute);

      gl.bindBuffer(gl.ARRAY_BUFFER, this._fboVerticesPositionBuffer);
      gl.vertexAttribPointer(shaderProgram.vertexPositionAttribute, 3, gl.FLOAT, false, 0, 0);

      gl.bindBuffer(gl.ARRAY_BUFFER, this._fboVerticesTextureCoordBuffer);
      gl.vertexAttribPointer(shaderProgram.textureCoordAttribute, 2, gl.FLOAT, false, 0, 0);

      gl.uniformMatrix4fv(shaderProgram.pMatrixUniform, false, this._pMatrix);

      gl.drawArrays(gl.TRIANGLE_STRIP, 0, 4);

      gl.disableVertexAttribArray(shaderProgram.vertexPositionAttribute);
      gl.disableVertexAttribArray(shaderProgram.textureCoordAttribute);
    },
  },
]);
