/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  clamp,
  inverseLerp,
  Bounds,
  Feature,
  Types,
  Dataset,
} from '../syntool.js';
import {colorToCSSColor} from '../utils/color.js';
import {colorPaletteGetColor} from '../utils/color_palette.js';
import {OLVectorLayer} from '../utils/ol_vector_layer.js';
import {layerExtension} from './extensions/layer_extension.js';
import {metadataExtension} from './extensions/metadata_extension.js';
import {dataFieldsExtension} from './extensions/data_fields_extension.js';
import {auxiliaryDataExtension} from './extensions/auxiliary_data_extension.js';
import {dataLoadingExtension} from './extensions/data_loading_extension.js';
import {dataMatrixExtension} from './extensions/data_matrix_extension.js';
import {colorPaletteExtension} from './extensions/color_palette_extension.js';
import {BarbsRenderer, ArrowRenderer} from './vector_field_dataset/vector_fabric.js';

/* 6*6 = 36 vectors displayed at zoom 0 */
var NB_VECTORS_AT_ZOOM_0 = 6;

var vectorFieldExtension = {
  initialize(info) {
    window.original();
    this.render = throttleOncePerAnimationFrame(this.render);
    this.renderWithColorPalette = this.renderWithColorPalette.bind(this);
  },
  createLayer() {
    return new OLVectorLayer(this.name);
  },
  handleVisibilityChange() {
    window.original();
    if (this.isVisible()) {
      this.syntool.mapView.hub.on('change:extent', this.render, this);
      this.render();
    } else {
      this.syntool.mapView.hub.off('change:extent', this.render, this);
    }
  },
  didAddToMap() {
    window.original();
    this.render();
  },
  handleDataLoad(response) {
    window.original();
    this._finishedFirstRender = Promise.withResolvers();
    this.render();
    return this._finishedFirstRender.promise;
  },

  async render() {
    await this.withColorPalette(this.renderWithColorPalette);
    setTimeout(() => this._finishedFirstRender?.resolve(), 100);
  },
  renderWithColorPalette() {
    if (!this.isVisible() || !this.dataMatrix) {
      return;
    }

    const product = this.getProduct();
    const renderer = this.constructor.Renderers[product.type];
    if (!renderer) {
      console.warn(
        `Could not find ${this.constructor.name} renderer for type`,
        product.type
      );
      return;
    }

    const viewportBounds = this.syntool.mapView.getVisibleProjectedBounds();
    const worldBounds = this.syntool.mapView.getWorldProjectedBounds();
    if (
      !Bounds.intersects(
        Bounds.createFromLBRT([
          this.dataMatrix.bounds.left,
          this.dataMatrix.bounds.bottom,
          this.dataMatrix.bounds.right,
          this.dataMatrix.bounds.top,
        ]),
        viewportBounds,
        worldBounds
      )
    ) {
      return;
    }

    const {density} = product;
    const minStep = this.dataMatrix.bounds.width / this.dataMatrix.getWidth();
    const zoomLevel = this.syntool.mapView.getZoom();
    const numberOfVectors = NB_VECTORS_AT_ZOOM_0 * Math.pow(2, zoomLevel) * density;
    const maxBounds = this.syntool.mapView.getMaxProjectedBounds();
    const step = Math.max(minStep, Bounds.getWidth(maxBounds) / numberOfVectors);
    const featuresPool = this.layerManager.layer.getFeatures();
    const newFeatures = [];

    if (step > 0) {
      const uniformFields = [
        this.getUniformFieldValue('modulus'),
        this.getUniformFieldValue('angle'),
        this.getUniformFieldValue('color'),
      ];
      const dataFields = [
        this.getReferencedFieldName('modulus'),
        this.getReferencedFieldName('angle'),
        this.getReferencedFieldName('color'),
      ];
      const modulusField = this.getReferencedField('modulus');
      const colorField = this.getReferencedField('color');
      const valuesArray = [0, 0, 0];
      const modulusMin = modulusField ? modulusField.min : NaN;
      const modulusMax = modulusField ? modulusField.max : NaN;

      const worldLeft = worldBounds ? Bounds.getLeft(worldBounds) : NaN;
      const worldRight = worldBounds ? Bounds.getRight(worldBounds) : NaN;
      const minX = Bounds.getLeft(viewportBounds);
      const minY = Bounds.getBottom(viewportBounds);
      const maxX = Bounds.getRight(viewportBounds);
      const maxY = Bounds.getTop(viewportBounds);

      let index;
      let x;
      let y;
      let xy = [0, 0];
      let values;
      let feature;
      for (y = minY; y <= maxY; y += step) {
        for (x = minX; x <= maxX; x += step) {
          xy[0] = x;
          xy[1] = y;
          index = this.dataMatrix.indexFromProjectedXY(xy, worldLeft, worldRight);

          if (!index) {
            continue;
          }

          values = this.dataMatrix.getRealValuesAt(
            index[0],
            index[1],
            dataFields,
            valuesArray
          );

          // Check for no data
          if (!values) {
            continue;
          }

          if (uniformFields[0] !== false) {
            values[0] = uniformFields[0];
          }
          if (uniformFields[1] !== false) {
            values[1] = uniformFields[1];
          }
          if (uniformFields[2] !== false) {
            values[2] = uniformFields[2];
          }

          // Check for no data
          if (values.every(isNaN)) {
            continue;
          }

          if (colorField) {
            values[2] = (
              this.colorPalette
              ? colorToCSSColor(colorPaletteGetColor(
                this.colorPalette,
                clamp(
                  inverseLerp(colorField.min, colorField.max, values[2]),
                  0,
                  1
                )
              ))
              : '#FFF'
            );
          }

          // Use Vector objects from pool in priority
          feature = featuresPool.pop();
          if (!feature) {
            feature = new Feature();
            newFeatures.push(feature);
          }

          renderer({
            scale: step,
            feature,
            x,
            y,
            modulus: values[0],
            angle  : values[1],
            color  : values[2],
            modulusMin,
            modulusMax,
          });
        }
      }
    }

    // Hide unused features
    let i = featuresPool.length;
    while (i--) {
      featuresPool[i].style.display = 'none';
    }

    if (0 < newFeatures.length) {
      //console.log(newFeatures.length + ' new features');
      this.layerManager.layer.addFeatures(newFeatures);
    }
    this.layerManager.layer.redrawAllFeatures(true /*force*/);
  },
};

export const VectorFieldDataset = Dataset.createClass({
  static: {
    name        : 'VectorFieldDataset',
    productTypes: ['ARROWS', 'BARBS'],
    renderFields: {
      modulus: ['ref', Types.number],
      angle  : ['ref', Types.number],
      color  : ['ref', Types.Color],
    },
    validateConfig: Types.shape({
      density: Types.number.gt(0).lte(2),
    }),
    normalizeProductConfig(productConfig) {
      if (!('density' in productConfig)) {
        productConfig.density = 1;
      }
      return productConfig;
    },
    Renderers: {
      ARROWS: ArrowRenderer,
      BARBS: BarbsRenderer,
    },
  },
  extensions: [
    layerExtension,
    metadataExtension,
    dataFieldsExtension,
    auxiliaryDataExtension,
    dataLoadingExtension,
    dataMatrixExtension,
    colorPaletteExtension,
    vectorFieldExtension,
  ],
});
window.ODL = Object.assign(window.ODL || {}, {VectorFieldDataset});
