/**
 * @author: <sylvain.herledan@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {clamp, inverseLerp, Geometry, Quantity} from '../../syntool.js';

const ZERO_POINT = [0, 0];
const BARBS_MAX_KNOTS = 160;
const BARBS_SIZE = 15;
const ARROW_XYS = [
  [0,   -0.05],
  [0.7, -0.05],
  [0.7, -0.2],
  [1,    0],
  [0.7,  0.2],
  [0.7,  0.05],
  [0,    0.05],
  [0,   -0.05],
];

export {
  updateBarbsFeature as BarbsRenderer,
  updateArrowFeature as ArrowRenderer,
};

/**
 * Create barbs for the tail of the arrow in wind barbs representation.
 *
 * The tail is a composition of:
 * - triangles (1 per 50 knots)
 * - long line (1 per 10 knots)
 * - short line (1 per 5 knots)
 *
 * @see https://www.weather.gov/hfo/windbarbinfo
 *
 * @param value Speed in knots
 *
 * @return A list of geometries that represent the speed as wind barbs
 */
function getBarbsFeathers(value) {
  var feathers = [];
  var x = -10;

  value = Math.round(value / 5);
  while (value > 0) {
    if (value >= 10) {
      feathers.push(Geometry.createPolygonFromXYs([
        [x, 0],
        [x + 1.5, 0],
        [x, 3],
        [x, 0],
      ]));
      x += 2.5;
      value -= 10;
    } else if (value >= 2) {
      feathers.push(Geometry.createPolylineFromXYs([
        [x, 0],
        [x - 1.5, 3],
      ]));
      x += 1;
      value -= 2;
    } else {
      feathers.push(Geometry.createPolylineFromXYs([
        [Math.max(-9, x), 0],
        [Math.max(-9, x) - 1, 2],
      ]));
      value -= 1;
    }
  }

  return feathers;
}

/**
 * Updates a vector with barbs according to measured speed.
 *
 * @param scale Scale to apply to the vector (so that it appears at the
 * same size for all zoom levels).
 * @param feature The feature to update.
 * @param x Horizontal coordinate of the vector's center.
 * @param y Vertical coordinate of the vector's center.
 * @param modulus Speed in m/s.
 * @param angle Counter-clockwise from east direction of the vector.
 * @param color Color of the vector as CSS color (eg. #FFCC33).
 */
function updateBarbsFeature({scale, feature, x, y, modulus, angle, color}) {
  // Convert speed from m/s to knots
  const speedInKnots = (
    clamp(Quantity.convert(modulus, 'm/s', 'knots'), 0, BARBS_MAX_KNOTS)
  );

  let geometry;
  let fillColor = color;
  if (speedInKnots < 2.5) {
    // Calm winds
    geometry = Geometry.createPoint(0, 0);
    fillColor = 'none';
  } else {
    /* Geometry representations for wind barbs. */
    const feathers = getBarbsFeathers(speedInKnots);
    const base = Geometry.createPolylineFromXYs([[-10, 0], [0, 0]]);

    feathers.push(base);
    geometry = Geometry.createCollection(feathers);
  }

  scale /= BARBS_SIZE;
  feature.geometry = Geometry.translate(
    Geometry.scale(
      Geometry.rotate(geometry, angle, ZERO_POINT),
      scale,
      scale,
      ZERO_POINT
    ),
    x,
    y
  );
  feature.style = {
    fillColor,
    fillOpacity  : 0.8,
    strokeColor  : color,
    strokeOpacity: 0.7,
    strokeWidth  : 1.5,
    pointRadius  : 4,
  };
}

/**
 * Updates a vector with an arrow tip.
 *
 * @param modulusMin Data value represented with 0 length arrow
 * @param modulusMax Data value represented with full scale arrow
 * @param scale Scale to apply to the vector (so that it appears at the
 * same size for all zoom levels).
 * @param feature The feature to update.
 * @param x Horizontal coordinate of the vector's center.
 * @param y Vertical coordinate of the vector's center.
 * @param modulus Speed in m/s.
 * @param angle Counter-clockwise from east direction of the vector.
 * @param color Color of the vector as CSS color (eg. #FFCC33).
 */
function updateArrowFeature({modulusMin, modulusMax, scale, feature, x, y, modulus, angle, color}) {
  // mapped modulus from [modulusMin, modulusMax] to [0, 1]
  var normalizedSpeed = (
    isNaN(modulusMin) || isNaN(modulusMax)
    ? 0.5
    : clamp(inverseLerp(modulusMin, modulusMax, modulus), 0, 1)
  );

  // Scale to apply to the vector (proportional to modulus)
  var weightedScale = scale * normalizedSpeed;

  feature.geometry = Geometry.translate(
    Geometry.scale(
      Geometry.rotate(
        Geometry.createPolygonFromXYs(ARROW_XYS),
        angle,
        ZERO_POINT
      ),
      weightedScale,
      weightedScale,
      ZERO_POINT
    ),
    x,
    y
  );
  feature.style = {
    fillColor    : color,
    fillOpacity  : 0.9,
    strokeColor  : '#FFFFFF',
    strokeOpacity: 0.2 + 0.8 * normalizedSpeed,
    strokeWidth  : 1 /* + 3 * normalizedSpeed*/,
  };
}
