/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  parseDecimal,
  interpolateString,
  range,
  Size,
  Point,
  Bounds,
  Types,
  Dataset,
} from '../syntool.js';
import {getGridLayerOptions} from '../utils/ol_utils.js';
import {OpenLayers} from '../vendors/openlayers.js';
import {layerExtension} from './extensions/layer_extension.js';
import {metadataExtension} from './extensions/metadata_extension.js';
import {auxiliaryDataExtension} from './extensions/auxiliary_data_extension.js';
import {OLOptionalImageTile} from './zxy_dataset/ol_optional_image_tile.js';
import {TilesMask} from './zxy_dataset/tiles_mask.js';

function getXYZ(bounds) {
  var origin = this.getTileOrigin();
  var res = this.getServerResolution();
  var x = Math.round(
    (bounds.left - origin.lon) / (res * this.tileSize.w)
  );
  var y = Math.round(
    (origin.lat - bounds.top) / (res * this.tileSize.h)
  );
  var z = (
    this.serverResolutions == null
    ? this.map.mapView.getZoom()
    : this.serverResolutions.indexOf(res)
  );

  if (this.wrapDateLine) {
    var limit = Math.pow(2, z);
    x = ((x % limit) + limit) % limit;
  }

  return {x, y, z};
}
function getURL(bounds) {
  return this.dataset.syntool.api.getDataURLForDataset(
    this.dataset,
    interpolateString(this.url, this.getXYZ(bounds))
  );
}
function getTileBounds(viewPortPx) {
  var origin = this.getTileOrigin();
  var resolution = this.map.getResolution();
  var tileMapWidth = resolution * this.tileSize.w;
  var tileMapHeight = resolution * this.tileSize.h;
  var mapLocation = this.getLonLatFromViewPortPx(viewPortPx);
  var tileLeft = origin.lon + tileMapWidth * Math.floor(
    (mapLocation.lon - origin.lon) / tileMapWidth
  );
  var tileBottom = origin.lat + tileMapHeight * Math.floor(
    (mapLocation.lat - origin.lat) / tileMapHeight
  );

  return Bounds.create(
    tileLeft,
    tileBottom,
    tileLeft + tileMapWidth,
    tileBottom + tileMapHeight
  );
}

var zxyExtension = {
  initialize(info) {
    window.original();
    this._mapMaxZoom = info.mapMaxZoom;
    this._resolutions = info.resolutions;
  },
  createLayer() {
    var product = this.getProduct();
    var layerOptions = Object.assign(getGridLayerOptions(), {
      //////////////
      // Options  //
      //////////////

      /* OpenLayers.Layer */
      maxExtent: this.bounds,

      /* OpenLayers.Layer.Grid */
      tileSize  : Size.createFromWH(product.tileSize),
      tileOrigin: Point.createFromXY(product.tileOrigin),
      tileClass : OLOptionalImageTile,

      serverResolutions: (
        this._mapMaxZoom === 0
        ? undefined
        : Array.from(
          range(this._mapMaxZoom + 1),
          this.syntool.mapView.getResolutionForZoom,
          this.syntool.mapView
        )
      ),

      /* OpenLayers.Layer.XYZ */
      sphericalMercator: true,

      /* Custom */
      mask   : this.createMask(),
      dataset: this,

      ////////////////
      // Overrides  //
      ////////////////

      /* OpenLayers.Layer.Grid */
      getTileBounds: getTileBounds,

      /* OpenLayers.Layer.XYZ */
      getXYZ: getXYZ,
      getURL: getURL,
    });

    return new OpenLayers.Layer.XYZ(
      this.name,
      this.uri + '${z}/${x}/${y}.png',
      layerOptions
    );
  },

  createMask() {
    var resolutions = this._resolutions;
    var l = resolutions.length;

    if (l > 1 && resolutions[l - 2].indexOf('*') !== -1) {
      var mask = new TilesMask();
      var maskRes = resolutions[l - 2].split('*');
      var minZoom = parseDecimal(maskRes[0]);
      var maxZoom = parseDecimal(maskRes[1]);
      var encodedData = resolutions[l - 1];

      mask.init(minZoom, maxZoom, encodedData);

      return mask;
    }
    return null;
  },
};

export const ZXYDataset = Dataset.createClass({
  static: {
    name          : 'ZXYDataset',
    productTypes  : ['ZXY'],
    validateConfig: Types.shape({
      tileOrigin: Types.Point.isRequired,
      tileSize  : Types.Size.isRequired,
    }),
    OLOptionalImageTile,
    TilesMask,
  },
  extensions: [
    layerExtension,
    metadataExtension,
    auxiliaryDataExtension,
    zxyExtension,
  ],
});
window.ODL = Object.assign(window.ODL || {}, {ZXYDataset});
