/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/**
 * The deprecated.js module includes all methods, properties, and constructors
 * that are no-longer supported. If you use any of these, you have to explicitly
 * import this module in your application.
 *
 * You are strongly encouraged to avoid using deprecated API. The documentation
 * here should point you to the supported alternatives.
 */
import {
  asyncIter,
  extendClass,
  assignWithOwnProperties,
  TimelineStore,
  MapView,
  Control,
  Syntool,
  Quantity,
} from './syntool.js';

const deprecationsSuppressed = new Set();
function deprecated(what, alternative = '') {
  if (typeof what == 'function') {
    what = `${what.name}()`;
  }
  if (!deprecationsSuppressed.has(what)) {
    deprecationsSuppressed.add(what);
    if (typeof alternative == 'function') {
      alternative = `${alternative.name}()`;
    }
    console.warn(
      (
        alternative
        ? `${what} is deprecated, please use ${alternative} instead.`
        : what
      ) +
      ' (This warning will only be given once)'
    );
    console.trace();
  }
}

assignWithOwnProperties(ODL, {
  preemptibleLoop(options) {
    deprecated('preemptibleLoop()', asyncIter);
    const abortController = new AbortController();
    const signal = abortController.signal;
    const iterator = {
      done: false,
      next() {
        return {done: this.done};
      },
      [Symbol.iterator]() {
        return this;
      },
    };
    const context = Object.assign({signal}, options);

    (async function() {
      for (let iter of asyncIter(iterator, context)) {
        // eslint-disable-next-line no-await-in-loop
        iter = await iter;
        iter.quantum = context.quantum;
        iter.schedule = context.schedule;
        // eslint-disable-next-line no-unused-vars
        for (const _ of iter) {
          iterator.done = !context.iterCallback?.();
        }
        context.laterCallback?.();
      }
      if (!signal.aborted) {
        // no more work to be done
        context.doneCallback?.();
      }
      context.finallyCallback?.();
    })();

    return function() {
      abortController.abort();
    };
  },

  /**
   * @template T
   * @returns {{
   *   promise: Promise<T>,
   *   resolve: (value: T | PromiseLike<T>) => void,
   *   reject: (reason?: unknown) => void
   * }}
   */
  createDeferredPromise() {
    deprecated('createDeferredPromise()', 'Promise.withResolvers()');
    return Promise.withResolvers();
  },

  get Unit() {
    deprecated('Unit', 'Quantity');
    return Quantity;
  },
});
Object.assign(ODL.Config, {
  validate(config, servers) {
    deprecated('Config.validate()', ODL.Config.assertValid);
    try {
      ODL.Config.assertValid(config, servers);
      return null;
    } catch (error) {
      return error.message;
    }
  },
});
Object.assign(ODL.Geography, {
  /**
   * Calculates the geodesic distance between two points.
   *
   * @deprecated Use Geography.geodesicDistance() instead
   * @param  {[lon: number, lat: number]} lonLat1 - The point to calculate the distance from.
   * @param  {[lon: number, lat: number]} lonLat2 - The point to calculate the distance to.
   * @return {number} The geodesic distance between `lonLat1` and `lonLat2` in meters.
   */
  greatCircleDistance(lonLat1, lonLat2) {
    deprecated('greatCircleDistance()', ODL.Geography.geodesicDistance);
    return ODL.Geography.geodesicDistance(lonLat1, lonLat2);
  },

  /**
   * Calculates the geodesic bearing between two points on the side of the first.
   *
   * @deprecated Use Geography.geodesicInitialBearing() instead
   * @param  {[lon: number, lat: number]} lonLat1 - The point to calculate the bearing from.
   * @param  {[lon: number, lat: number]} lonLat2 - The point to calculate the bearing to.
   * @return {number} The geodesic bearing between `lonLat1` and `lonLat2` in degrees.
   */
  greatCircleBearing(lonLat1, lonLat2) {
    deprecated('greatCircleBearing()', ODL.Geography.geodesicInitialBearing);
    return ODL.Geography.geodesicInitialBearing(lonLat1, lonLat2);
  },

  /**
   * Finds the end point of a geodesic from a point, an initial bearing and a distance.
   *
   * @deprecated Use Geography.geodesicDestination() instead
   * @param  {[lon: number, lat: number]} lonLat1 - The starting point.
   * @param  {number} distance - The geodesic distance between `lonLat1` and the end point in meters.
   * @param  {number} bearing - The initial bearing in degrees.
   * @return {[lon: number, lat: number]} The end point.
   */
  greatCircleWaypoint(lonLat1, distance, bearing) {
    deprecated('greatCircleWaypoint()', ODL.Geography.geodesicDestination);
    return ODL.Geography.geodesicDestination(lonLat1, distance, bearing);
  },
});
Object.assign(ODL.Geometry, {
  getCentroid(geometry) {
    deprecated('getCentroid()', ODL.Geometry.getCentroidXY);
    return geometry.getCentroid();
  },
  forEachGeometry(geometry, callback, thisArg) {
    deprecated('forEachGeometry()', ODL.Geometry.deepGeometries);
    for (const {geometry: g, indexPath: ip} of ODL.Geometry.deepGeometries(geometry)) {
      callback.call(thisArg, g, ip);
    }
  },
  forEachEdge(geometry, callback, thisArg) {
    deprecated('forEachEdge()', ODL.Geometry.edges);
    for (const {from, to, indexPath: ip} of ODL.Geometry.edges(geometry)) {
      callback.call(thisArg, [from, to], ip);
    }
  },
  forEachVertex(geometry, callback, thisArg) {
    deprecated('forEachVertex()', ODL.Geometry.vertices);
    for (const {vertex, indexPath: ip} of ODL.Geometry.vertices(geometry)) {
      callback.call(thisArg, vertex, ip);
    }
  },
});

Object.assign(Syntool, {
  addInstanceConstructor(constructor) {
    deprecated('addInstanceConstructor()', this.extend);
    this.extend({[Syntool.CONSTRUCTOR]: constructor});
  },
  addInstanceInitialiser(init) {
    deprecated('addInstanceInitialiser()', this.extend);
    this.extend({[Syntool.INIT]: init});
  },
});
Syntool.extend({
  isGeoFilterActive() {
    deprecated('isGeoFilterActive()', 'geoFilterActive');
    return this.geoFilterActive;
  },
  toggleGeoFilter() {
    deprecated('toggleGeoFilter()', 'geoFilterActive');
    this.geoFilterActive = !this.geoFilterActive;
  },
  isColocActive() {
    deprecated('isColocActive()', 'colocActive');
    return this.colocActive;
  },
  toggleColoc() {
    deprecated('toggleColoc()', 'colocActive');
    this.colocActive = !this.colocActive;
  },

  isScreenshotModeActive() {
    deprecated('isScreenshotModeActive()', 'screenshotModeActive');
    return this.screenshotModeActive;
  },
  toggleScreenshotMode() {
    deprecated('toggleScreenshotMode()', 'screenshotModeActive');
    this.screenshotModeActive = !this.screenshotModeActive;
  },
});
extendClass(Control, {
  activate() {
    deprecated('activate()', 'active');
    if (this.active) {
      return false;
    }
    this.active = true;
    return this.active;
  },
  deactivate() {
    deprecated('deactivate()', 'active');
    if (!this.active) {
      return false;
    }
    this.active = false;
    return !this.active;
  },
});
extendClass(MapView, {
  getProjectedWorldBounds() {
    deprecated('getProjectedWorldBounds()', this.getWorldProjectedBounds);
    return this.getWorldProjectedBounds();
  },

  getDatasetFromEvent(event) {
    deprecated(
      'getDatasetFromEvent()',
      'getDatasetFromMouseEvent() or getDatasetFromEventAtClientXY()'
    );
    return this.getDatasetFromMouseEvent(event);
  },
  getDatasetsFromEvent(event, filter, thisArg) {
    deprecated(
      'getDatasetsFromEvent()',
      'getDatasetsFromMouseEvent() or getDatasetsFromEventAtClientXY()'
    );
    return this.getDatasetsFromMouseEvent(event, filter, thisArg);
  },

  getMeasuringRulerStartLonLat() {
    deprecated('getMeasuringRulerStartLonLat()', 'measuringRulerStartLonLat');
    return this.measuringRulerStartLonLat;
  },
  getMeasuringRulerEndLonLat() {
    deprecated('getMeasuringRulerEndLonLat()', 'measuringRulerEndLonLat');
    return this.measuringRulerEndLonLat;
  },
  getMeasuringRulerDistance() {
    deprecated('getMeasuringRulerDistance()', 'measuringRulerDistance');
    return this.measuringRulerDistance;
  },
  getMeasuringRulerBearing() {
    deprecated('getMeasuringRulerBearing()', 'measuringRulerBearing');
    return this.measuringRulerBearing;
  },

  toggleOutlinesVisibility() {
    deprecated('toggleOutlinesVisibility()', 'outlinesHidden');
    this.outlinesHidden = !this.outlinesHidden;
  },

  isTilesPixelInterpolationActive() {
    deprecated(
      'isTilesPixelInterpolationActive()',
      'tilesPixelInterpolationActive'
    );
    return this.tilesPixelInterpolationActive;
  },
  setTilesPixelInterpolationActive(on) {
    deprecated(
      'setTilesPixelInterpolationActive()',
      'tilesPixelInterpolationActive'
    );
    on = !!on;
    if (
      !this.tilesPixelInterpolationSupported ||
      this.tilesPixelInterpolationActive === on
    ) {
      return false;
    }
    this.tilesPixelInterpolationActive = on;
    return true;
  },
});
extendClass(TimelineStore, {
  getMinYear() {
    deprecated('getMinYear()', 'minYear');
    return this.minYear;
  },
  getMaxYear() {
    deprecated('getMaxYear()', 'maxYear');
    return this.maxYear;
  },
  getCurrentDate() {
    deprecated('getCurrentDate()', 'currentDate');
    return this.currentDate;
  },
  getCurrentTime() {
    deprecated('getCurrentTime()', 'currentTime');
    return this.currentTime;
  },
  getAbsoluteMinDate() {
    deprecated('getAbsoluteMinDate()', 'absoluteMinDate');
    return this.absoluteMinDate;
  },
  getAbsoluteMaxDate() {
    deprecated('getAbsoluteMaxDate()', 'absoluteMaxDate');
    return this.absoluteMaxDate;
  },
  getMinDate() {
    deprecated('getMinDate()', 'minDate');
    return this.minDate;
  },
  getMinTime() {
    deprecated('getMinTime()', 'minTime');
    return this.minTime;
  },
  getMaxDate() {
    deprecated('getMaxDate()', 'maxDate');
    return this.maxDate;
  },
  getMaxTime() {
    deprecated('getMaxTime()', 'maxTime');
    return this.maxTime;
  },
  getTimeframeLength() {
    deprecated('getTimeframeLength()', 'timeframeLength');
    return this.timeframeLength;
  },
  getTimeframe() {
    deprecated('getTimeframe()', 'timeframe');
    return this.timeframe;
  },

  setMinYear(year) {
    deprecated('setMinYear()', 'minYear');
    this.minYear = year;
  },
  setMaxYear(year) {
    deprecated('setMaxYear()', 'maxYear');
    this.maxYear = year;
  },
  setCurrentTime(time) {
    deprecated('setCurrentTime()', 'currentTime');
    this.currentTime = time;
  },
  setCurrentDate(date) {
    deprecated('setCurrentDate()', 'currentDate');
    this.currentDate = date;
  },
  setTimeframeLength(timeframeLength) {
    deprecated('setTimeframeLength()', 'timeframeLength');
    this.timeframeLength = timeframeLength;
  },
  setTimeframe(timeframe) {
    deprecated('setTimeframe()', 'timeframe');
    this.timeframe = timeframe;
  },
});

// eslint-disable-next-line no-unused-expressions
ODL.Dialog && extendClass(ODL.Dialog, {
  show() {
    if (!this.syntool) {
      deprecated(
        'Showing a dialog not added to a Syntool instance is deprecated, ' +
        'please add it to a Syntool instance before calling show(). ' +
        'Adding this dialog to the global Syntool instance for backwards ' +
        'compatibility with `Syntool.addControl(this)`'
      );
      Syntool.addControl(this);
    }
    return window.original();
  },
  isShown() {
    deprecated('isShown()', 'shown');
    return this.shown;
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.CoverageTimelineView && extendClass(ODL.CoverageTimelineView, {
  toggleColocActive(colocActive) {
    deprecated('toggleColocActive()', 'colocActive');
    this.colocActive = (
      colocActive === true || colocActive === false
      ? colocActive
      : !this.colocActive
    );
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.ProductsPanel && extendClass(ODL.ProductsPanel, {
  // `.open()`` replaced with `.open`!!!
  // open() {
  //   deprecated('open()', 'open');
  //   this.open = true;
  // },
  close() {
    deprecated('close()', 'open');
    this.open = false;
  },
  toggle() {
    deprecated('toggle()', 'open');
    this.open = !this.open;
  },

  getGroupByTag() {
    deprecated('getGroupByTag()', 'groupByTag');
    return this.groupByTag;
  },
  getSearch() {
    deprecated('getSearch()', 'search');
    return this.search;
  },
  setSearch(search) {
    deprecated('setSearch()', 'search');
    this.search = search;
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.TimelinePanel && extendClass(ODL.TimelinePanel, {
  toggleDetailsView() {
    deprecated('toggleDetailsView()', 'withDetails');
    this.withDetails = !this.withDetails;
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.VectorDataMatrix && extendClass(ODL.VectorDataMatrix, {
  /**
   * @deprecated Use VectorDataMatrix#projectedXYFromIndex() instead
   * @param {number} i - Zero based column index.
   * @param {number} j - Zero based row index.
   * @return {?{lon: number, lat: number}} Returns the corresponding geographical coordinates or `null`.
   */
  lonLatFromIndex(i, j) {
    deprecated('lonLatFromIndex()', 'projectedXYFromIndex()');
    const xy = this.projectedXYFromIndex(i, j);
    return xy && {lon: xy[0], lat: xy[1]};
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.TrajectoriesDataset && Object.assign(ODL.TrajectoriesDataset.TrajectoryMarker, {
  getTimestampFromRelativePostion(sortedTimestamps, relativePosition) {
    const replacement = (
      ODL.TrajectoriesDataset.TrajectoryMarker.getTimestampFromRelativePosition
    );
    deprecated('getTimestampFromRelativePostion()', replacement);
    return replacement(sortedTimestamps, relativePosition);
  },
  getPositionIndexFromRelativePostion(sortedTimestamps, relativePosition) {
    const replacement = (
      ODL.TrajectoriesDataset.TrajectoryMarker.getPositionIndexFromRelativePosition
    );
    deprecated('getPositionIndexFromRelativePostion()', replacement);
    return replacement(sortedTimestamps, relativePosition);
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.Feature && extendClass(ODL.Feature, {
  get attributes() {
    deprecated('attributes', 'properties');
    return this.properties;
  },
  set attributes(attributes) {
    deprecated('attributes', 'properties');
    this.properties = attributes;
  },
});
// eslint-disable-next-line no-unused-expressions
ODL.ProcessingToolsPanel && extendClass(ODL.ProcessingToolsPanel, {
  hide() {
    deprecated('hide()', 'hidden = true');
    this.hidden = true;
  },
  unhide() {
    deprecated('unhide()', 'hidden = false');
    this.hidden = false;
  },
});
