/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  ajax,
  whenReady,
  Types as is,
  H,
  NotificationMessageType,
  Config,
  Syntool,
} from '../syntool.js';
import './import.js';

whenReady(function() {
  Object.assign(Config.configFieldValidators, {
    importPolicyForURL: is.func,
  });
});

Syntool.extend({
  [Syntool.NAME]: 'AutoImport',
  [Syntool.INIT]() {
    this._autoImportImportedProductURLs = new Map();

    const {importPolicyForURL} = this.config;
    const imports = new Set(
      new URL(window.location, window.location.origin)
      .searchParams.getAll('import')
    );
    const ignoredUrls = new Set();
    let index = 1;
    for (const untrustedHref of imports) {
      try {
        const untrustedUrl = new URL('', untrustedHref);

        // "file:" and "javascript:" URLs have their `host` set to ""
        if (!untrustedUrl.host) {
          ignoredUrls.add(untrustedHref);
          continue;
        }

        const key = index++;

        // Prevent XSS from untrustedUrl
        const anchor = H('a', {href: untrustedUrl.href});
        anchor.textContent = untrustedUrl.href;

        const policy = importPolicyForURL?.(untrustedUrl);
        switch (policy) {
        case 'allow':
          this._autoImportImportFile(untrustedUrl, key);
          break;
        case 'deny':
          ignoredUrls.add(untrustedHref);
          console.warn(
            `Auto import from URL "${untrustedUrl}" denied`,
            untrustedUrl,
            policy
          );
          break;
        default:
        case 'ask':
          this.notificationCenter.showMessage({
            type: NotificationMessageType.info,
            text: `Detected a linked external file at ${anchor.outerHTML}`,
            closeButton: {
              text : 'ignore',
              props: {title: 'Ignore the linked file'},
              onClick: () => this._autoImportRemoveImportURLs([untrustedHref]),
            },
            buttons: [
              {
                text   : 'import',
                closes : true,
                props  : {title: 'Try to import the linked file'},
                onClick: () => this._autoImportImportFile(untrustedUrl, key),
              },
            ],
            hideAfter: -1,
          });
          break;
        }
      } catch (error) {
        console.error(error);
        ignoredUrls.add(untrustedHref);
      }
    }
    this._autoImportRemoveImportURLs(ignoredUrls);

    this.productsStore.hub.on('change:products', (productId) => {
      if (!this.productsStore.hasProduct(productId)) {
        // product deleted
        this._autoImportRemoveImportURLs(
          [this._autoImportImportedProductURLs.get(productId)]
        );
        this._autoImportImportedProductURLs.delete(productId);
      }
    });
  },

  _autoImportRemoveImportURLs(urlsToRemove) {
    const currentURL = new URL(window.location, window.location.origin);
    const {searchParams} = currentURL;
    const imports = new Set(searchParams.getAll('import'));
    for (const url of urlsToRemove) {
      imports.delete(url);
    }

    // Remove all imports from search params of location
    searchParams.delete('import');

    // Add all other imports to search params of location
    for (const url of imports) {
      searchParams.append('import', url);
    }

    // Update location with new search params
    window.history.replaceState(null, null, currentURL);
  },

  async _autoImportImportFile(url, key) {
    const response = await ajax(url, {
      credentials : 'include',
      responseType: 'raw',
      timeout     : ajax.get.timeout,
    });
    const mimeType = response.headers.get('content-type').split(';')[0];
    const string = await response.text();

    const id = `URL_IMPORT_${key}`;
    this.importFromString(string, mimeType === 'text/plain' ? null : mimeType, {
      shapes: {
        product: {
          id,
          label: `<i>Linked shapes ${key}</i>`,
          description: `External linked shapes from:\n${url}`,
          importedFrom: url.href,
        },
        checkUserShapesDuplicates: false,
        onAddShapes: () => {
          this._autoImportImportedProductURLs.set(id, url.href);
        },
      },
    });
  }
});
