/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {negate, forEachObject, Syntool} from '../syntool.js';
import './import.js';

const backupBase = {
  '@type'   : 'backup',
  '@version': 1,
};

function isValidBackupJSON(backup) {
  return (
    backup['@type'] === backupBase['@type'] &&
    backup['@version'] === backupBase['@version']
  );
}

Syntool.extend({
  [Syntool.NAME]: 'Backup',
  [Syntool.CONSTRUCTOR]() {
    this.backupStoresMap = {
      controls: {
        toJSON: () => {
          const json = {};
          let empty = true;
          for (const [key, control] of Object.entries(this.controls)) {
            if (
              typeof control.toJSON === 'function' &&
              typeof control.fromJSON === 'function'
            ) {
              json[key] = control.toJSON();
              empty = false;
            }
          }
          return empty ? undefined : json;
        },
        fromJSON: (json) => {
          if (!json) {
            return;
          }
          for (const [key, control] of Object.entries(this.controls)) {
            if (
              Object.hasOwn(json, key) &&
              typeof control.toJSON === 'function' &&
              typeof control.fromJSON === 'function'
            ) {
              control.fromJSON(json[key]);
            }
          }
        },
      },
    };
    this.importersFromJSON.push(
      (json, options) => this.importBackupFromJSON(json, options)
    );
  },

  importBackupFromJSON(json, options) {
    const backupJson = json.filter(isValidBackupJSON);
    const remainingJson = json.filter(negate(isValidBackupJSON));

    for (const backup of backupJson) {
      forEachObject(this.backupStoresMap, function(store, key) {
        if (key in backup) {
          store.fromJSON(backup[key]);
        }
      });
    }

    const importedCount = json.length - remainingJson.length;
    if (importedCount > 0) {
      console.log(
        'Import::Imported ' + importedCount + ' backup' +
        (importedCount === 1 ? '' : 's')
      );
      this.notificationCenter.showInfo(
        'Imported ' + importedCount + ' backup' +
        (importedCount === 1 ? '' : 's')
      );
    }

    return remainingJson;
  },
  exportBackupToJSON() {
    return Object.assign({}, backupBase, this.backupStoresMap);
  },
});
