/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {HIDKeys, HIDMouse, HIDModifiers, Bounds, Syntool} from '../syntool.js';

function toggleOpacity(opacity) {
  return opacity > 0 ? 0 : 1;
}

function descendingNumberComparator(a, b) {
  return b - a;
}

function descendingProductsStackLevelComparator(a, b) {
  return descendingNumberComparator(a.stackLevel, b.stackLevel);
}

function getBoundsFromDataset(dataset) {
  return dataset.bounds;
}

Syntool.extend({
  [Syntool.NAME]: 'BasicShortcuts',
  [Syntool.INIT]() {
    this.shortcutsStore.addGroups([
      {
        label    : 'Global',
        shortcuts: [
          {
            id         : 'move-current-time-forward',
            keys       : HIDKeys.right,
            description: 'Go forward a third of the selected timespan in the time',
            context    : this,
            handler    : function() {
              this.timelineStore.currentTime += (
                this.timelineStore.timeframeLength / 3
              );
              return true;
            },
          },
          {
            id         : 'move-current-time-backward',
            keys       : HIDKeys.left,
            description: 'Go back a third of the selected timespan in the time',
            context    : this,
            handler    : function() {
              this.timelineStore.currentTime -= (
                this.timelineStore.timeframeLength / 3
              );
              return true;
            },
          },
          {
            id         : 'unselect-all',
            keys       : HIDKeys.escape,
            description: 'Unselect all datasets',
            context    : this,
            handler    : function() {
              return this.mapView.unselectAllDatasets();
            },
          },
          {
            id         : 'toggle-tile-interpolation',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.A,
            description: "Toggle interpolation of the datasets' tiles",
            context    : this,
            handler    : function() {
              if (this.mapView.tilesPixelInterpolationSupported) {
                this.mapView.tilesPixelInterpolationActive = (
                  !this.mapView.tilesPixelInterpolationActive
                );
                return (
                  'Interpolation is <em>' +
                  (this.mapView.tilesPixelInterpolationActive ? 'on' : 'off') +
                  '</em>'
                );
              }
              return 'Interpolation is not supported by your browser';
            },
          },
          {
            id         : 'toggle-geo-filter',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.G,
            description: 'Toggle geographic filtering of the datasets',
            context    : this,
            handler    : function() {
              this.geoFilterActive = !this.geoFilterActive;
              return (
                'Geographic filter is <em>' +
                (this.geoFilterActive ? 'on' : 'off') +
                '</em>'
              );
            },
          },
          {
            id         : 'toggle-dataset-outline-visibility',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.S,
            description: 'Toggle the visibility of the outline of datasets',
            context    : this,
            handler    : function() {
              this.mapView.outlinesHidden = !this.mapView.outlinesHidden;
              return 'The dataset outlines are <em>' + (
                this.mapView.outlinesHidden ? 'hidden' : 'visible'
              ) + '</em>';
            },
          },
        ],
      },
      {
        label    : 'Map',
        shortcuts: [
          {
            id         : 'map-zoom-in',
            keys       : HIDMouse.doubleLeft,
            description: 'Zoom in',
          },
          {
            id         : 'map-zoom-out',
            modifiers  : [HIDModifiers.ctrlOrMeta],
            keys       : HIDMouse.doubleLeft,
            description: 'Zoom out',
          },
          {
            id         : 'map-measure',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDMouse.leftMove,
            description: 'Measure a distance',
          },
        ],
      },
      {
        label: 'Change the current datasets page',
        shortcuts: [
          {
            id         : 'pagination-previous',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.B,
            description: 'Show previous datasets page',
            context    : this,
            handler    : function() {
              this.mapView.incrementDataPageBy(-1);
              return true;
            },
          },
          {
            id         : 'pagination-next',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.N,
            description: 'Show next datasets page',
            context    : this,
            handler    : function() {
              this.mapView.incrementDataPageBy(1);
              return true;
            },
          },
        ],
      },
      {
        label    : 'Datasets selection (in map and extended timeline)',
        shortcuts: [
          {
            id         : 'map-dataset-select',
            keys       : HIDMouse.left,
            description: 'Select/deselect clicked dataset',
          },
          {
            id         : 'map-dataset-select-add',
            modifiers  : [HIDModifiers.ctrlOrMeta],
            keys       : HIDMouse.left,
            description: 'Add/remove clicked dataset to/from current selection',
          },
        ],
      },
      {
        label    : 'Selected dataset (only one selected)',
        shortcuts: [
          {
            id         : 'timeline-center-on-dataset',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.C,
            description: "Center the timeline on the selected dataset's time",
            context    : this,
            handler    : function() {
              switch (this.mapView.getSelectedDatasetsCount()) {
              case 0:
                return 'No datasets selected, select a dataset first';
              case 1:
                if (this._centerTimelineOnSelectedDataset()) {
                  return 'Centered timeline on the selected dataset';
                }
                return false;
              default:
                return "Can't center timeline on multiple datasets";
              }
            },
          },
          {
            id         : 'selected-dataset-bring-to-front',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.pageUp,
            description: 'Bring to front among the same product (temporarily)',
            context    : this,
            handler    : function() {
              switch (this.mapView.getSelectedDatasetsCount()) {
              case 0:
                return 'No datasets selected, select a dataset first';
              case 1:
                if (this._moveSelectedDataset('up')) {
                  return true;
                }
                return "Can't bring dataset to front temporarily";
              default:
                return "Can't bring multiple datasets to front";
              }
            },
          },
          {
            id         : 'selected-dataset-send-to-back',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.pageDown,
            description: 'Send to back among the same product (temporarily)',
            context    : this,
            handler    : function() {
              switch (this.mapView.getSelectedDatasetsCount()) {
              case 0:
                return 'No datasets selected, select a dataset first';
              case 1:
                if (this._moveSelectedDataset('down')) {
                  return true;
                }
                return "Can't send dataset to back temporarily";
              default:
                return "Can't send multiple datasets to back";
              }
            },
          },
        ],
      },
      {
        label    : 'Selected dataset(s) (within the same product)',
        shortcuts: [
          {
            id         : 'selected-datasets-product-bring-forward',
            keys       : [HIDKeys.pageUp],
            description: 'Bring the product forward',
            context    : this,
            handler    : function() {
              switch (this.mapView.getSelectedDatasetProductsCount()) {
              case 0:
                return (
                  'No datasets selected, select one or more datasets of the ' +
                  'same product first'
                );
              case 1:
                if (this._moveSelectedDatasetProduct('up')) {
                  return true;
                }
                return "Can't bring the product forward";
              default:
                return "Can't bring multiple products forward";
              }
            },
          },
          {
            id         : 'selected-datasets-product-send-back',
            keys       : [HIDKeys.pageDown],
            description: 'Send the product back',
            context    : this,
            handler    : function() {
              switch (this.mapView.getSelectedDatasetProductsCount()) {
              case 0:
                return (
                  'No datasets selected, select one or more datasets of the ' +
                  'same product first'
                );
              case 1:
                if (this._moveSelectedDatasetProduct('down')) {
                  return true;
                }
                return "Can't send the product back";
              default:
                return "Can't send multiple products back";
              }
            },
          },
        ],
      },
      {
        label    : 'Selected dataset(s)',
        shortcuts: [
          {
            id         : 'selected-datasets-products-toggle-visiblity',
            keys       : HIDKeys.backspace,
            description: "Toggle their product's visibility",
            context    : this,
            handler    : function() {
              return this._toggleSelectedDatasetProductsOpacity();
            },
          },
          {
            id         : 'selected-datasets-products-decrement-opacity',
            keys       : [HIDKeys.M],
            description: "Reduce their product's opacity by 10%",
            context    : this,
            handler    : function() {
              return this._incrementSelectedDatasetProductsOpacity(-0.1);
            },
          },
          {
            id         : 'selected-datasets-products-increment-opacity',
            keys       : [HIDKeys.P],
            description: "Increase their product's opacity by 10%",
            context    : this,
            handler    : function() {
              return this._incrementSelectedDatasetProductsOpacity(0.1);
            },
          },
          {
            id         : 'map-zoom-on-datasets',
            modifiers  : [HIDModifiers.shift],
            keys       : HIDKeys.Z,
            description: 'Zoom on selected datasets',
            context    : this,
            handler    : function() {
              var datasets = this.mapView.getSelectedDatasets();
              if (datasets.length === 0) {
                return 'No datasets selected, select one or more datasets first';
              }

              var boundedDatasets = (
                datasets.filter(this.mapView.isRegionalDataset, this.mapView)
              );
              if (boundedDatasets.length === 0) {
                return "Can't zoom on global datasets";
              }

              var boundsList = (
                boundedDatasets.map(getBoundsFromDataset).filter(Boolean)
              );
              if (boundsList.length === 0) {
                return "Can't zoom on boundless datasets";
              }

              const worldBounds = this.mapView.getWorldProjectedBounds();
              var bounds = Bounds.createFromBoundsList(boundsList, worldBounds);
              var prevZoom = this.mapView.getZoom();

              // Only zoom if bounds is not empty (for a single point dataset)
              if (Bounds.getWidth(bounds) > 0 && Bounds.getHeight(bounds) > 0) {
                var minZoom = 0;
                var maxZoom = Infinity;
                for (const dataset of datasets) {
                  var product = dataset.getProduct();
                  if (product.mapMinZoom > minZoom) {
                    minZoom = product.mapMinZoom;
                  }
                  if (product.mapMaxZoom < maxZoom) {
                    maxZoom = product.mapMaxZoom;
                  }
                }
                this.mapView.flyToProjectedBounds(bounds, {minZoom, maxZoom});
              } else {
                this.mapView.setViewportCenterProjectedXY(Bounds.getCenter(bounds));
              }

              return (
                this.mapView.getZoom() === prevZoom
                ? 'Centered map around selected datasets'
                : 'Zoomed on selected datasets'
              );
            },
          },
        ],
      },
    ]);

    const updateBasicNavigationShortcuts = () => {
      this.shortcutsStore.setShortcutsWithIdsEnabled([
        'map-zoom-in',
        'map-zoom-out',
      ], this.mapView.navigationActive);
    };
    const updateBasicMeasurementShortcuts = () => {
      this.shortcutsStore.setShortcutsWithIdsEnabled([
        'map-measure',
      ], this.mapView.measuringRulerActive);
    };
    const updateBasicSelectionShortcuts = () => {
      this.shortcutsStore.setShortcutsWithIdsEnabled([
        'unselect-all',
        'map-dataset-select',
        'map-dataset-select-add',
        'timeline-center-on-dataset',
        'selected-dataset-bring-to-front',
        'selected-dataset-send-to-back',
        'selected-datasets-product-bring-forward',
        'selected-datasets-product-send-back',
        'selected-datasets-products-toggle-visiblity',
        'selected-datasets-products-decrement-opacity',
        'selected-datasets-products-increment-opacity',
        'map-zoom-on-datasets',
      ], (
        this.mapView.datasetInteractionsActive &&
        this.mapView.datasetSelectionActive
      ));
    };
    this.mapView.hub
      .on('change:navigationActive', updateBasicNavigationShortcuts)
      .on('change:measuringRulerActive', updateBasicMeasurementShortcuts)
      .on('change:datasetSelectionActive', updateBasicSelectionShortcuts)
      .on('change:datasetInteractionsActive', updateBasicSelectionShortcuts);
    updateBasicNavigationShortcuts();
    updateBasicMeasurementShortcuts();
    updateBasicSelectionShortcuts();
  },

  _incrementSelectedDatasetProductsOpacity(deltaOpacity) {
    let didSomething = false;
    for (const product of this.mapView.getSelectedDatasetProducts()) {
      if (this.productsStore.canSetProductOpacity(product.id)) {
        this.productsStore.setProductOpacity(
          product.id,
          product.opacity + deltaOpacity
        );
        didSomething = true;
      }
    }
    return didSomething;
  },
  _toggleSelectedDatasetProductsOpacity() {
    let didSomething = false;
    for (const product of this.mapView.getSelectedDatasetProducts()) {
      if (this.productsStore.canSetProductOpacity(product.id)) {
        this.productsStore.setProductOpacity(
          product.id,
          toggleOpacity(product.opacity)
        );
        didSomething = true;
      }
    }
    return didSomething;
  },
  _moveSelectedDatasetProduct(direction) {
    var product = this.mapView.getSelectedDatasetProduct();
    if (!product || !this.productsStore.canSetProductStackLevel(product.id)) {
      return false;
    }

    var products = this.catalogStore.getSelectedProducts();

    // Sort products by stack level
    products.sort(descendingProductsStackLevelComparator);

    var index = products.indexOf(product);
    var dir = direction === 'up' ? -1 : 1;

    if (index + dir < 0 || index + dir === products.length) {
      return false;
    }

    var otherProduct = products[index + dir];

    if (!this.productsStore.canSetProductStackLevel(otherProduct.id)) {
      return false;
    }

    var stackLevelsByProduct = {};
    stackLevelsByProduct[otherProduct.id] = product.stackLevel;
    stackLevelsByProduct[product.id] = otherProduct.stackLevel;

    this.productsStore.setProductStackLevels(stackLevelsByProduct);
    return true;
  },
  _moveSelectedDataset(direction) {
    var selectedDataset = this.mapView.getSelectedDataset();
    if (!selectedDataset || !selectedDataset.setStackLevel) {
      return false;
    }

    var product = selectedDataset.getProduct();
    var inMapDatasets = this.mapView.getProductDatasetsInMap(product.id);
    var initialStackLevel = selectedDataset.getStackLevel();
    var currentStackLevel = initialStackLevel;
    var dir = direction === 'up' ? 1 : -1;

    for (var i = 0, len = inMapDatasets.length; i < len; ++i) {
      var dataset = inMapDatasets[i];
      var stackLevel = dataset.getStackLevel();
      var stackLevelDelta = stackLevel - initialStackLevel;

      // check if the dataset is in way of the requested movement
      if (stackLevelDelta * dir > 0) {
        dataset.setStackLevel(stackLevel - dir);
        currentStackLevel += dir;
      }
    }
    selectedDataset.setStackLevel(currentStackLevel);
    return true;
  },
  _centerTimelineOnSelectedDataset() {
    var dataset = this.mapView.getSelectedDataset();
    if (dataset) {
      this.timelineStore.currentTime = (
        (dataset.start.getTime() + dataset.end.getTime()) / 2
      );
      return true;
    }
    return false;
  },
});
