/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  emptyFunctionReturningTrue,
  emptyFunctionReturningFalse,
  ajax,
  ajaxRetriable,
  isValidDate,
  whenReady,
  Config,
  Syntool,
  Types as is,
  UserStorage,
} from '../syntool.js';

var STORAGE_KEY = 'Changelog.lastViewedServerTime';

function lastModifiedTimeFromResponse(response) {
  var lastModified = new Date(response.headers.get('last-modified'));
  return isValidDate(lastModified) ? lastModified.getTime() : null;
}
function serverTimeFromResponse(response) {
  var now = new Date(response.headers.get('date'));
  return isValidDate(now) ? now.getTime() : null;
}
function serverTimeFromError(error) {
  return serverTimeFromResponse(error.response);
}

export class Changelog {
  constructor(url) {
    this.url = url;
    this._checkPromise = null;
  }

  get lastViewedTime() {
    return UserStorage.get(STORAGE_KEY).catch(emptyFunction);
  }
  get lastModifiedTime() {
    return this._check().then(lastModifiedTimeFromResponse);
  }
  get accessible() {
    return this._check().then(
      emptyFunctionReturningTrue,
      emptyFunctionReturningFalse
    );
  }
  get updatedSinceLastViewed() {
    return (
      Promise.all([this.lastModifiedTime, this.lastViewedTime])
      .then(function([lastModifiedTime, lastViewedTime]) {
        return (lastModifiedTime && lastModifiedTime > lastViewedTime);
      }, emptyFunctionReturningFalse)
    );
  }

  updateLastViewedTime() {
    return (
      this._check()
      .then(serverTimeFromResponse, serverTimeFromError)
      .then(function(now) {
        if (now) {
          UserStorage.set(STORAGE_KEY, now).catch(emptyFunction);
        }
      })
    );
  }

  _check() {
    if (!this._checkPromise) {
      this._checkPromise = (
        ajaxRetriable(ajax)(this.url, {method: 'HEAD', responseType: 'raw'})
      );
      this._checkPromise.catch(function(error) {
        console.warn('Error checking CHANGELOG!', error);
      });
    }
    return this._checkPromise;
  }
}
window.ODL = Object.assign(window.ODL || {}, {Changelog});

whenReady(function() {
  Object.assign(Config.configFieldValidators, {
    changelogURL: is.URL,
  });
});

Syntool.extend({
  [Syntool.NAME]: 'Changelog',

  changelog: null,

  [Syntool.INIT]() {
    var config = this.config;

    if (config.changelogURL) {
      this.changelog = new Changelog(config.changelogURL);
    }
  },
});
