/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {wrapMethods, isOutOfMemoryError, Syntool} from '../syntool.js';

wrapMethods(window, {
  onerror(message, url, lineno, colno, error) {
    if (!Syntool.initialised) {
      return window.original();
    }
    try {
      let msg;
      if (error instanceof Error) {
        msg = 'Error: ' + error.message;
        console.error(error);
      } else {
        msg = (
          'Error: ' + message +
          (url ? '\nurl: ' + url : '') +
          (lineno ? '\nline: ' + lineno : '') +
          (colno ? '\ncolumn: ' + colno : '') +
          (error ? '\nerror: ' + error : '')
        );
        console.error({message, url, lineno, colno, error});
      }
      if (isOutOfMemoryError(error)) {
        Syntool.hub.emit('low-memory', 1);
      }
      Syntool.notificationCenter.showError(
        msg + '\n\nSee the JavaScript console for technical details.',
        -1
      );
      return true;
    } catch (err) {
      console.error(err);
      return window.original();
    }
  },
  onunhandledrejection(event) {
    event.preventDefault();
    if (!Syntool.initialised) {
      window.original();
      return;
    }
    try {
      var error = event.reason;
      console.warn('Unhandled promise rejection', error);
      if (isOutOfMemoryError(error)) {
        Syntool.hub.emit('low-memory', 1);
      }
      Syntool.notificationCenter.showError(
        'Unhandled promise rejection: ' +
        (error instanceof Error ? error.message : error) +
        '\n\nSee the JavaScript console for technical details.',
        -1
      );
    } catch (err) {
      console.error(err);
      window.original();
    }
  },
});
