/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  isAbortError,
  isTimeoutError,
  retriable,
  HIDKeys,
  HIDModifiers,
  Syntool,
} from '../syntool.js';

/**
 * Rethrows any passed error, except abort errors.
 *
 * @param {Error} error
 * @returns {void}
 */
function ignoreAbortError(error) {
  if (!isAbortError(error)) {
    throw error;
  }
}

Syntool.extend({
  [Syntool.NAME]: 'GoToNearest',
  [Syntool.INIT]() {
    this.shortcutsStore.add('Global', [
      {
        id         : 'jump-to-nearest-dataset',
        description: 'Jump to the date of the nearest dataset',
        context    : this,
        handler    : function(...args) {
          this.gotoNearest(
            true /*inPast*/,
            true /*inFuture*/,
            true /*inPresent*/,
            ...args
          ).catch(ignoreAbortError);
          return true;
        },
      },
      {
        id         : 'jump-to-previous-dataset',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.left,
        description: 'Jump to the date of the previous dataset',
        context    : this,
        handler    : function() {
          this.gotoNearest(
            true /*inPast*/,
            false /*inFuture*/,
            false /*inPresent*/
          ).catch(ignoreAbortError);
          return true;
        },
      },
      {
        id         : 'jump-to-next-dataset',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.right,
        description: 'Jump to the date of the next dataset',
        context    : this,
        handler    : function() {
          this.gotoNearest(
            false /*inPast*/,
            true /*inFuture*/,
            false /*inPresent*/
          ).catch(ignoreAbortError);
          return true;
        },
      },
    ]);
  },

  gotoNearest: retriable(async function gotoNearest(
    {signal, canRetry, abort},
    inPast,
    inFuture,
    inPresent,
    ifNeeded
  ) {
    if (ifNeeded instanceof AbortSignal) {
      ifNeeded = false;
    }

    const selectedProductIds = (
      this.productsStore.getSelectedProductIds()
      .filter(this.filters.product.nonPermanentVisible.withId)
    );
    const {currentDate} = this.timelineStore;

    try {
      await new Promise((resolve) => this.afterUpdate(resolve));
      this.hub.on('update', abort);

      const updated = await (
        this.timelineStore.isCoverageLoading()
        ? new Promise((resolve) => {
          const handler = () => {
            if (!this.timelineStore.isCoverageLoading()) {
              this.timelineStore.hub.off('change:coverage', handler);
              resolve(true);
            }
          };
          this.timelineStore.hub.on('change:coverage', handler);
        })
        : this.updateCoverage(signal)
      );
      if (!updated || signal.aborted) {
        return false;
      }

      const nearest = await this.api.findNearest(
        selectedProductIds,
        currentDate,
        inPast,
        inFuture,
        inPresent,
        signal
      );

      const {currentTime} = this.timelineStore;
      const nearestTime = nearest.nearest.getTime();

      if (currentTime === nearestTime || signal.aborted) {
        return false;
      }
      if (ifNeeded) {
        let needed;
        if (this.colocActive) {
          needed = false;
          for (const productId of selectedProductIds) {
            if (
              this.filters.product.includedInColoc.withId(productId) &&
              !this.mapView.hasDatasetsInMapForProductWithId(productId)
            ) {
              needed = true;
              break;
            }
          }
        } else {
          needed = true;
          for (const productId of selectedProductIds) {
            if (this.mapView.hasDatasetsInMapForProductWithId(productId)) {
              needed = false;
              break;
            }
          }
        }
        if (!needed) {
          return false;
        }
      }
      this.timelineStore.currentTime = nearestTime;
      return true;
    } catch (error) {
      if (isAbortError(error) || isTimeoutError(error)) {
        return false;
      }
      if (canRetry()) {
        throw error;
      }

      console.error('Error finding nearest data!', error);
      this.notificationCenter.showError(
        'Could not find nearest data due to server or network errors!'
      );
      return false;
    } finally {
      this.hub.off('update', abort);
    }
  }, {abortsPreviousCall: true, withExtraArgument: true}),
});
