/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {whenReady, Config, Syntool, Types as is} from '../syntool.js';
import {updateURLParams} from '../utils/url_utils.js';
import {hotspotFromPermalink, HotspotsStore} from './hotspots/hotspots_store.js';

var openedHotspot = hotspotFromPermalink(window.location.href);

/* remove hotspot URL params */
updateURLParams({
  name       : undefined,
  description: undefined,
});

export {openedHotspot, hotspotFromPermalink};

function getDefaultTimespan(timeRanges) {
  for (var i = 0; i < timeRanges.length; i++) {
    var range = timeRanges[i][1];
    var isDefault = !!timeRanges[i][2];

    if (isDefault) {
      return range;
    }
  }
  return timeRanges.length > 0 ? timeRanges[0][1] : '1d';
}

whenReady(function() {
  var HotspotPermalink = is.string.match(getHotspotPermalinkRegExp());
  var Hotspot = is.oneOfType([
    HotspotPermalink,
    is.shape({
      name       : is.string.isRequired,
      description: is.AnyURL,
      date       : is.oneOfType([is.date, is.integer.gte(0)]),
      timespan   : is.Timespan,
      zoom       : is.integer.gte(0),
      extent     : is.Bounds,
      products   : is.arrayOf(is.string.isRequired),
      opacity    : is.arrayOf(is.number.gte(0).lte(100).isRequired),
      stackLevel : is.arrayOf(is.number.gte(0).isRequired),
      pattern    : is.arrayOf(is.string.isRequired),
    }),
  ]);
  Object.assign(Config.types, {
    Hotspot: Hotspot,
  });
  Object.assign(Config.configFieldValidators, {
    hotspots: is.arrayOf(Hotspot.isRequired),
  });

  function getHotspotPermalinkRegExp() {
    var URL_QUERY_PARAM = '(?:[^=#]+=[^=#]+)?';
    var HOTSPOT_QUERY_PARAMS = ['(?:', URL_QUERY_PARAM, '&)*name=[^=#](?:&', URL_QUERY_PARAM, ')*'].join('');
    var HOTSPOT_PERMALINK = '^[^?]*\\?' + HOTSPOT_QUERY_PARAMS;

    return new RegExp(HOTSPOT_PERMALINK);
  }
});

Syntool.extend({
  [Syntool.NAME]: 'Hotspots',
  [Syntool.CONSTRUCTOR]() {
    this.hotspotsStore = new HotspotsStore();
  },
  [Syntool.INIT]() {
    var syntool = this;
    var config = syntool.config;

    function handleHotspotsStoreInit() {
      syntool.hotspotsStore.setDefaultTimespan(
        getDefaultTimespan(config.timeRanges)
      );
    }

    this.hotspotsStore.init(config.hotspots);
    if (this.hotspotsStore.initialised) {
      handleHotspotsStoreInit();
    } else {
      this.hotspotsStore.hub.once('init', handleHotspotsStoreInit);
    }
  },
});
