/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  createClass,
  randomString,
  parseURLQueryString,
  stringifyQueryString,
  EventsHub,
  UserStorage,
} from '../../syntool.js';
import {
  appStateToJSON,
  appStateFromJSON,
  appStateJSONToURLParameters,
  appStateJSONFromURLParameters,
} from '../../utils/app_state.js';

export function hotspotFromPermalink(url) {
  const urlParams = parseURLQueryString(url);
  const json = appStateJSONFromURLParameters(urlParams);

  if (urlParams.name) {
    json.name = urlParams.name;
  }
  if (urlParams.description) {
    json.description = urlParams.description;
  }

  return hotspotFromJSON(json);
}
function permalinkFromHotspot(hotspot) {
  const urlParams = appStateJSONToURLParameters(hotspotToJSON(hotspot));

  if (hotspot.name) {
    urlParams.name = hotspot.name;
  }
  if (hotspot.description) {
    urlParams.description = hotspot.description;
  }

  return stringifyQueryString(urlParams);
}
function hotspotFromJSON(json) {
  if (json.date instanceof Date) {
    json.date = json.date.getTime();
  }
  if (json.extent && !Array.isArray(json.extent)) {
    json.extent = json.extent.split(',').map(parseFloat);
  }
  if (json.products && !Array.isArray(json.products)) {
    json.products = json.products.split(',');
  }
  if (json.opacity && !Array.isArray(json.opacity)) {
    json.opacity = json.opacity.split(',').map(parseFloat);
  }
  if (json.stackLevel && !Array.isArray(json.stackLevel)) {
    json.stackLevel = json.stackLevel.split(',').map(parseFloat);
  }

  var hotspot = appStateFromJSON(json);

  if (!hotspot.date && !hotspot.products && !hotspot.extent) {
    return null;
  }

  if (json.name) {
    hotspot.name = json.name;
  }
  if (json.description) {
    hotspot.description = json.description;
  }
  hotspot.permalink = permalinkFromHotspot(hotspot);

  return hotspot;
}
function hotspotToJSON(hotspot) {
  var json = appStateToJSON(hotspot);

  if (hotspot.name) {
    json.name = hotspot.name;
  }
  if (hotspot.description) {
    json.description = hotspot.description;
  }

  return json;
}

export const HotspotsStore = createClass([
  /*Public*/ {
    static: {hotspotFromPermalink},

    constructor: function HotspotsStore() {
      this.hub = new EventsHub();
      this.userStorage = new UserStorage('hotspots', 'hotspots', true);
      this.serverHotspots = {};
      this.localHotspots = {};
      this.defaultTimespan = null;
      this.initialised = false;
    },

    init(predefinedHotspots) {
      var self = this;

      this.loadServerHotspots(predefinedHotspots || []);
      return (
        this.loadLocalHotspots().then(function() {
          self.initialised = true;
          self.hub.emit('init');
        })
      );
    },

    loadServerHotspots(predefinedHotspots) {
      var serverHotspots;

      serverHotspots = {};
      predefinedHotspots.forEach(function(permalinkOrJson, index) {
        var hotspot = (
          typeof permalinkOrJson === 'string'
          ? hotspotFromPermalink(permalinkOrJson)
          : hotspotFromJSON(permalinkOrJson)
        );

        if (!hotspot) {
          console.warn(
            "HotspotsStore: Couldn't parse hotspot at index",
            index,
            permalinkOrJson
          );
          return;
        }

        hotspot.id = index;
        serverHotspots[hotspot.id] = hotspot;
      });
      this.serverHotspots = serverHotspots;
      this.hub.emit('change');
    },

    loadLocalHotspots() {
      var self = this;
      var localHotspots = {};

      return (
        this.userStorage.forEach(function(json, key) {
          var hotspot = hotspotFromJSON(json);
          if (hotspot) {
            hotspot.id = key;
            localHotspots[key] = hotspot;
          }
        }, 'load local hotspots')
        .catch(emptyFunction)
        .then(function() {
          self.localHotspots = localHotspots;
          self.hub.emit('change');
        })
      );
    },

    setDefaultTimespan(defaultTimespan) {
      this.defaultTimespan = defaultTimespan;

      if (this.initialised) {
        this._updateLocalHotspotsWithDefaultTimespan();
      }
    },
  },
  /*Public*/ {
    toJSON() {
      return Object.values(this.localHotspots).map(hotspotToJSON, this);
    },
    fromJSON(json) {
      json.forEach(function(hotspotJson) {
        var hotspot = hotspotFromJSON(hotspotJson);
        if (hotspot) {
          var hotspotId = this._newId();
          hotspot.id = hotspotId;
          this.localHotspots[hotspotId] = hotspot;
          this.userStorage.set(hotspotId, hotspotJson)
            .catch(emptyFunction);
        }
      }, this);
      this.hub.emit('change');
    },

    addHotspot(hotspot) {
      var hotspotId = hotspot.id || this._newId();
      var hotspotJson = hotspotToJSON(hotspot);

      hotspot = hotspotFromJSON(hotspotJson);
      hotspot.id = hotspotId;

      this.localHotspots[hotspotId] = hotspot;
      this.userStorage.set(hotspotId, hotspotJson)
        .catch(emptyFunction);
      this.hub.emit('change');
    },

    add(name, description, appState) {
      var hotspot = Object.assign({
        name       : name,
        description: description,
      }, appState);

      this.addHotspot(hotspot);
    },

    addFromPermalink(permalink, name, description) {
      var hotspot;

      hotspot = hotspotFromPermalink(permalink);
      if (!hotspot) {
        console.warn("HotspotsStore: Couldn't parse permalink", permalink);
        return;
      }

      if (name || !hotspot.name) {
        hotspot.name = name || 'Untitled hotspot';
      }
      if (description) {
        hotspot.description = description;
      }

      this.addHotspot(hotspot);
    },

    get(id) {
      return this.serverHotspots[id] || this.localHotspots[id] || null;
    },

    hasHotspotSameHotspot(hotspot) {
      var permalink;
      var ids;
      var i;

      permalink = permalinkFromHotspot(hotspot);

      ids = Object.keys(this.serverHotspots);
      i = ids.length;
      while (i--) {
        if (this.serverHotspots[ids[i]].permalink === permalink) {
          return true;
        }
      }

      ids = Object.keys(this.localHotspots);
      i = ids.length;
      while (i--) {
        if (this.localHotspots[ids[i]].permalink === permalink) {
          return true;
        }
      }

      return false;
    },

    remove(hotspotId) {
      delete this.localHotspots[hotspotId];
      this.userStorage.delete(hotspotId).catch(emptyFunction);
      this.hub.emit('change');
    },
  },
  /*Protected*/ {
    _newId() {
      var hotspotId;
      do {
        hotspotId = randomString(5);
      } while (hotspotId in this.localHotspots);
      return hotspotId;
    },
    _updateLocalHotspotsWithDefaultTimespan() {
      var localHotspots;
      var hotspot;
      var i;

      if (!this.defaultTimespan) {
        return;
      }

      localHotspots = Object.values(this.localHotspots);
      i = localHotspots.length;
      while (i--) {
        hotspot = localHotspots[i];
        if (!hotspot.timespan) {
          hotspot.timespan = this.defaultTimespan;
          UserStorage.set(
            hotspot.id,
            hotspotToJSON(hotspot),
            true /*persist*/,
            'set default timespan on hotspot with id'
          ).catch(emptyFunction);
        }
      }
      this.hub.emit('change');
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {HotspotsStore});
