/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Syntool} from '../syntool.js';

const JSON_MIME_TYPE = 'application/json';

function isFileExtension(type) {
  return type[0] === '.';
}
function isMimeType(type) {
  return type[0] !== '.' && type.includes('/');
}
function getFileExtension(file) {
  var name = file.name;
  var lastDotIndex = name.lastIndexOf('.');
  if (lastDotIndex === -1) {
    return '';
  }
  return name.substring(lastDotIndex).toLowerCase();
}
function toLowerCase(string) {
  return string.toLowerCase();
}

Syntool.extend({
  [Syntool.NAME]: 'Import',
  [Syntool.CONSTRUCTOR]() {
    this.acceptedImportFormats = ['Syntool JSON'];
    this.acceptedImportFileTypes = ['.json', JSON_MIME_TYPE];
    this.importersFromFile = [
      (file, options) => this._importFromFileAsString(file, options),
    ];
    this.importersFromString = [
      (string, mimeType, options) => (
        this._importFromStringAsJSON(string, mimeType, options)
      ),
    ];
    this.importersFromJSON = [];
  },

  get acceptedImportFileExtensions() {
    return this.acceptedImportFileTypes.filter(isFileExtension).map(toLowerCase);
  },
  get acceptedImportMimeTypes() {
    return this.acceptedImportFileTypes.filter(isMimeType);
  },

  async importFromJSON(json, options) {
    if (json.length === 0) {
      console.log('Import::input is empty');
      this.notificationCenter.showInfo('Nothing to import, input is empty');
      return true;
    }

    let remainingJson = json;
    for (const importer of this.importersFromJSON) {
      // eslint-disable-next-line no-await-in-loop
      remainingJson = await importer(remainingJson, options);
    }
    if (remainingJson.length > 0) {
      console.warn('Import::Cannot deserialize a part of the input');
      this.notificationCenter.showWarning(
        'Could not import ' + remainingJson.length + ' malformed input item' +
        (remainingJson.length === 1 ? '' : 's')
      );
    }
    return true;
  },
  async _importFromString(string, mimeType, options) {
    for (const importer of this.importersFromString) {
      // eslint-disable-next-line no-await-in-loop
      if (await importer(string, mimeType, options)) {
        return true;
      }
    }
    return false;
  },
  async importFromString(string, mimeType, options) {
    if (await this._importFromString(string, mimeType, options)) {
      return true;
    }
    console.error('Import::Cannot deserialize input string', string, mimeType);
    this.notificationCenter.showError(
      'Cannot import: input is malformed or not supported!'
    );
    return false;
  },
  async importFromFile(file, options) {
    for (const importer of this.importersFromFile) {
      // eslint-disable-next-line no-await-in-loop
      if (await importer(file, options)) {
        return true;
      }
    }
    console.error('Import::Cannot deserialize input file', file);
    this.notificationCenter.showError(
      'Cannot import: file "' + file.name + '" is malformed or not supported!'
    );
    return false;
  },
  importFromFiles(files, options) {
    return Promise.all(
      [...files].map((file) => this.importFromFile(file, options))
    );
  },

  canImportFromMimeType(mimeType) {
    if (mimeType) {
      for (const acceptedMimeType of this.acceptedImportMimeTypes) {
        if (
          mimeType === acceptedMimeType ||
          (
            acceptedMimeType.endsWith('/*') &&
            mimeType.startsWith(
              acceptedMimeType.substring(0, acceptedMimeType.length - 1)
            )
          )
        ) {
          return true;
        }
      }
    }
    return false;
  },
  canImportFromFile(file) {
    return (
      !file ||
      this.canImportFromMimeType(file.type) ||
      this.acceptedImportFileExtensions.includes(getFileExtension(file))
    );
  },
  _importFromFileAsString(file, options) {
    const fileReader = new FileReader();
    const {promise, resolve} = Promise.withResolvers();

    fileReader.onerror = (event) => {
      console.error('Import::Cannot read file as text', fileReader.error);
      resolve(false);
    };
    fileReader.onload = (event) => {
      resolve(this._importFromString(fileReader.result, file.type, options));
    };
    fileReader.readAsText(file);
    return promise;
  },
  _importFromStringAsJSON(string, mimeType, options) {
    if (mimeType && mimeType !== JSON_MIME_TYPE) {
      return false;
    }

    try {
      let json = JSON.parse(string);
      if (Array.isArray(json)) {
        json = json.filter(Boolean);
      } else if (json) {
        json = [json];
      }
      if (json && json.length > 0) {
        return this.importFromJSON(json, options);
      }
      return true;
    } catch (error) {
      return false;
    }
  },
});
