/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  flattenArray,
  isSafari,
  UserStorage,
  Syntool,
} from '../syntool.js';
import {localforage} from '../vendors/localforage.js';

const NOT_SUPPORTED_ERROR = new Error('Persistent storage not supported!');
const CANT_PERSIST_MESSAGE = (
  'In-browser storage may be cleared by the browser under storage pressure.' +
  '\nAnything you create $like MAY BE LOST when the browser runs out of ' +
  'storage space!'
);
const CANT_USE_STORAGE = (
  'In-browser storage is disabled.\nAnything you create $like ' +
  'WILL BE LOST after you refresh or close this page!'
);

// eslint-disable-next-line compat/compat -- Checking support
const supportsPersistence = !!navigator.storage?.persist;
let hasPersistence = false;
let noPersistenceMessageShown = false;

let initPromise = null;

UserStorage.NOT_SUPPORTED_ERROR = NOT_SUPPORTED_ERROR;
UserStorage.impl = {
  initialize() {
    if (!initPromise) {
      initPromise = initialize();
    }
    return initPromise;
  },

  forEach(iterator) {
    return localforage.iterate(iterator);
  },
  get(key) {
    return localforage.getItem(key);
  },
  async set(key, value, persist) {
    const result = await localforage.setItem(key, value);
    if (persist) {
      await tryToPersist();
    }
    return result;
  },
  delete(key) {
    return localforage.removeItem(key);
  },
};

async function initialize() {
  const {storageId} = Syntool.config;
  let oldStores;

  if (
    // Migrate old WebSQL store.
    // See https://github.com/localForage/localForage/blob/master/CHANGELOG.md#15
    /* The Apache v2.0 licence for localForage is already included in the project */
    isSafari &&
    localforage.supports(localforage.INDEXEDDB) &&
    (
      localforage.supports(localforage.WEBSQL) ||
      localforage.supports(localforage.LOCALSTORAGE)
    )
  ) {
    oldStores = [
      localforage.createInstance({
        driver: [localforage.WEBSQL, localforage.LOCALSTORAGE],
      }),
      localforage.createInstance({
        driver: [localforage.WEBSQL, localforage.LOCALSTORAGE],
        name  : 'Syntool - ' + storageId,
      }),
    ];
  } else {
    oldStores = [localforage.createInstance()];
  }

  localforage.config({
    name: 'Syntool - ' + storageId
  });

  const enabled = await localStorageIsEnabled();
  if (enabled) {
    const migrated = await migrateFromLocalForageInstances(oldStores);
    if (migrated) {
      await tryToPersist();
    }
  } else {
    console.warn("Can't use local storage!");
    showWarningMessage(CANT_USE_STORAGE);

    // eslint-disable-next-line func-style
    const nope = () => Promise.reject(NOT_SUPPORTED_ERROR);
    Object.assign(UserStorage, {
      forEach: () => Promise.resolve(),
      get    : nope,
      set    : nope,
      delete : nope,
    });
  }
}

function showWarningMessage(message) {
  const labels = UserStorage.instances.map(
    (instance) => instance.persist && instance.label
  ).filter(Boolean);
  const like = (
    labels.length === 0
    ? ''
    : ' ' + (
      labels.length === 1
      ? labels[0]
      : labels.slice(0, -1).join(', ') + ' and ' + labels[labels.length - 1]
    )
  );
  Syntool.notificationCenter.showWarning(message.replace(' $like', like), -1);
}

async function localStorageIsEnabled() {
  try {
    const KEY = '______TEST_____';
    await localforage.setItem(KEY, KEY);
    const value = await localforage.getItem(KEY);
    localforage.removeItem(KEY);
    return (value === KEY);
  } catch (error) {
    return false;
  }
}
async function migrateFromLocalForageInstance(instance) {
  const setPromises = [];
  try {
    await instance.iterate((value, key) => {
      setPromises.push(localforage.setItem(key, value));
    });
    await instance.clear();
  } catch (error) {
    console.error(error);
  }
  return Promise.all(setPromises);
}
async function migrateFromLocalForageInstances(instances) {
  if (instances && instances.length > 0) {
    const sets = flattenArray(
      await Promise.all(instances.map(migrateFromLocalForageInstance))
    );
    if (sets.length > 0) {
      console.log('migrated %d items in local storage', sets.length);
      return true;
    }
    return false;
  }
  return false;
}
async function tryToPersist() {
  if (supportsPersistence && !hasPersistence) {
    let persistent = await navigator.storage.persisted();
    if (!persistent) {
      persistent = await navigator.storage.persist();
    }
    // eslint-disable-next-line require-atomic-updates
    hasPersistence = persistent;
    if (!persistent && !noPersistenceMessageShown) {
      console.warn("Can't persist local storage!");
      showWarningMessage(CANT_PERSIST_MESSAGE);
      noPersistenceMessageShown = true;
    }
  }
}
