/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {OpenLayers} from '../vendors/openlayers.js';
import {
  assignWithOwnProperties,
  wrapMethods,
  extendClass,
  flattenArray,
} from '../syntool.js';

export function patchOLGoogleLayer(google) {
  function getRealElementsOfChild(child) {
    const isJustAContainer = (
      child instanceof HTMLDivElement &&
      (child.getAttribute('class') || '') === '' &&
      (child.getAttribute('style') || '') === ''
    );
    return isJustAContainer ? getRealElementsOf(child) : child;
  }
  function getRealElementsOf(node) {
    return flattenArray(Array.from(node.children, getRealElementsOfChild));
  }

  // Update OpenLayers.Layer.Google.v3 for Google Maps version 3.52
  Object.assign(OpenLayers.Layer.Google.v3, {
    /**
     * Method: loadMapObject
     * Load the GMap and register appropriate event listeners. If we can't
     *     load GMap2, then display a warning message.
     */
    loadMapObject() {
      if (!this.type) {
        this.type = google.maps.MapTypeId.ROADMAP;
      }
      let mapObject;
      let cache = OpenLayers.Layer.Google.cache[this.map.id];
      if (cache) {
        // there are already Google layers added to this map
        mapObject = cache.mapObject;
        // increment the layer count
        ++cache.count;
        this.mapObject = mapObject;
      } else {
        // this is the first Google layer for this map

        const container = this.map.viewPortDiv;
        const div = document.createElement('div');
        div.id = this.map.id + '_GMapContainer';
        div.classList.add('OLGoogleMapsContainer');
        div.style.position = 'absolute';
        div.style.width = '100%';
        div.style.height = '100%';
        // Hide "Sorry, we have no imagery here." text
        div.style.color = 'transparent';
        container.appendChild(div);

        // create GMap and shuffle elements
        mapObject = new google.maps.Map(div, {
          center                : {lat: 90, lng: 0},
          zoom                  : this.map.getZoom() || 0,
          mapTypeId             : this.type,
          backgroundColor       : '#1b2854',
          clickableIcons        : false,
          disableDefaultUI      : true,
          disableDoubleClickZoom: true,
          gestureHandling       : 'none',
          keyboardShortcuts     : false,
          streetViewControl     : false
        });

        // cache elements for use by any other google layers added to
        // this same map
        cache = {mapObject, ready: false, count: 1};
        OpenLayers.Layer.Google.cache[this.map.id] = cache;
        this.mapObject = mapObject;

        const center = (
          this.getMapObjectLonLatFromOLLonLat(this.map.getCenter())
        );
        if (center) {
          mapObject.setCenter(center);
          this.repositionMapElements();
        } else {
          this.repositionListener = google.maps.event.addListenerOnce(
            mapObject,
            'center_changed',
            () => this.repositionMapElements()
          );
        }
      }
      this.setGMapVisibility();
    },

    /**
     * Method: repositionMapElements
     *
     * Waits until powered by and terms of use elements are available and then
     * moves them so they are clickable.
     */
    repositionMapElements() {
      // This is the first time any Google layer in this mapObject has been
      // made visible.  The mapObject needs to know the container size.
      google.maps.event.trigger(this.mapObject, 'resize');

      if (!this._repositionMapElements()) {
        new MutationObserver((_, observer) => {
          if (this._repositionMapElements()) {
            observer.disconnect();
          }
        }).observe(this.mapObject.getDiv(), {childList: true, subtree: true});
      }
    },

    /**
     * Moves "powered by" and "terms of use" elements so they are clickable.
     */
    _repositionMapElements() {
      const div = this.mapObject.getDiv().querySelector('.gm-style');
      if (!div || !div.querySelector('.gmnoprint')) {
        return false;
      }

      const cache = OpenLayers.Layer.Google.cache[this.map.id];
      const syntool = this.map.mapView.syntool;
      const container = syntool.usableViewportNode;

      // move the ToS and branding stuff up to the container div
      const attribContainer = div.querySelector('.gmnoprint').parentElement;
      attribContainer.style.zIndex = '';
      attribContainer.classList.add('olLayerGoogleV3-nn');
      container.insertBefore(attribContainer, container.firstElementChild);
      cache.termsOfUse = attribContainer;

      // depends on value of zIndex, which is not robust
      for (const child of attribContainer.children) {
        if (child.style.zIndex === '1000001') {
          child.style.zIndex = '';
          if (child.classList.contains('gm-style-cc')) {
            child.classList.add('olLayerGoogleTOU');
          } else {
            child.classList.add('olLayerGoogleAttribution');
          }
        }
      }

      const children = flattenArray(
        Array.from(
          div.children,
          (child, i) => (i === 0 ? [] : getRealElementsOfChild(child))
        )
      );
      // depends on value of zIndex, which is not robust
      for (const child of children) {
        if (child.style.zIndex === '1000000') {
          child.style.display = '';
          child.style.zIndex = '';
          child.classList.add('olLayerGooglePoweredBy', 'olLayerGoogleV3-nn');
          container.insertBefore(child, container.firstElementChild);
          cache.poweredBy = child;
          break;
        }
      }
      cache.ready = true;
      this.setGMapVisibility();

      new MutationObserver((mutations) => {
        const uiContainer = syntool.uiOverlayNode;
        const nextSibling = syntool.usableViewportNode.nextSibling;
        for (const {target, addedNodes} of mutations) {
          if (addedNodes && target === div.parentElement) {
            for (const node of addedNodes) {
              node.classList.add('olLayerGoogleV3-nn', 'olLayerGoogleOverlay');
              uiContainer.insertBefore(node, nextSibling);
            }
          }
        }
      }).observe(this.mapObject.getDiv(), {childList: true, subtree: true});

      return true;
    },

    setMapObjectCenter(center, zoom) {
      const cache = OpenLayers.Layer.Google.cache[this.map.id];
      if (cache.ready && cache.wasHidden) {
        this.mapObject.moveCamera({center, zoom});
        return;
      }

      this.mapObject.panTo(center);
      this.mapObject.setZoom(zoom);
      cache.wasHidden = false;
    },

    /**
     * Method: setGMapVisibility
     * Display the GMap container and associated elements.
     *
     * Parameters:
     * visible - {Boolean} Display the GMap elements.
     */
    setGMapVisibility() {
      const cache = OpenLayers.Layer.Google.cache[this.map.id];
      if (cache) {
        const layers = this.map.layers;
        let type = this.type;
        let visible = false;
        let layer;
        for (let i = layers.length - 1; i >= 0; --i) {
          layer = layers[i];
          if (
            layer instanceof OpenLayers.Layer.Google &&
            layer.visibility === true &&
            layer.inRange === true
          ) {
            type = layer.type;
            visible = true;
            break;
          }
        }
        if (visible) {
          this.mapObject.setMapTypeId(type);
          cache.layer = layer;
        } else {
          cache.wasHidden = true;
          cache.layer = null;
        }
        this.map.mapView.syntool.node.classList.toggle('-with-google-maps', visible);
      }
    },
  });

  /////////////////////////////
  // Track Google maps tiles //
  /////////////////////////////
  function addLoadingImage(image) {
    image.addEventListener('load', removeLoadingImage);
    image.addEventListener('error', removeLoadingImage);
    googleMapsTileLoadStart(image);
  }
  function removeLoadingImage() {
    this.removeEventListener('load', removeLoadingImage);
    this.removeEventListener('error', removeLoadingImage);
    googleMapsTileLoaded(this);
  }

  // Google maps does the following steps in order:
  // 1. set onload and onerror properties of the Image instance
  // 2. set src property of the Image instance to the tile URL
  // 3. call __gm_captureTile, if set, passing the tile URL
  // 4. When the image is done loading, onload and onerror are reset to null

  if (Object.getOwnPropertyDescriptor(Image.prototype, 'src').configurable) {
    // The Image's src property is generally configurable. So we extend it to
    // detect the start of a Google maps tile loading.
    extendClass(Image, {
      get src() {
        return window.original();
      },
      set src(src) {
        if (new URL(src, window.location).origin.endsWith('.googleapis.com')) {
          addLoadingImage(this);
        }
        window.original();
      },
    });
  } else {
    // Fallback for when Image's src property is non-configurable, as is the
    // case in Firefox Focus for iOS, see mozilla-mobile/focus-ios#3719
    // https://github.com/mozilla-mobile/focus-ios/issues/3719.
    // We save a reference to the tile's Image instance to be loaded by
    // observing the onload property that is set in step 1.
    // Then when __gm_captureTile is called in step 3, we check that tile's URL
    // matches the src of the image.

    let image;
    extendClass(Image, {
      get onload() {
        return window.original();
      },
      set onload(onload) {
        // Verify that we are in step 1 and not step 4.
        if (onload) {
          image = this;
        }
        window.original();
      },
    });
    window.__gm_captureTile = (url) => {
      if (image.src === url) {
        addLoadingImage(image);
        image = null;
      }
    };
  }

  ///////////////////////////////////////////////
  // Add tiles management to Google maps layer //
  ///////////////////////////////////////////////
  function googleMapsTileLoadStart(image) {
    for (const {layer} of Object.values(OpenLayers.Layer.Google.cache)) {
      if (layer && !layer.tilesLoading.has(image)) {
        const tile = {layer, isLoading: true, getImage: () => image};
        layer.tilesLoading.set(image, tile);
        layer.events.triggerEvent('tileloadstart', {tile});
      }
    }
  }
  function googleMapsTileLoaded(image) {
    for (const {layer} of Object.values(OpenLayers.Layer.Google.cache)) {
      const tile = layer?.tilesLoading.get(image);
      if (tile) {
        layer.tilesLoading.delete(image);
        layer.events.triggerEvent('tileloaded', {tile});
      }
    }
  }

  extendClass(OpenLayers.Layer.Google, {
    get tiles() {
      return this.tilesLoading.values();
    },
  });
  wrapMethods(OpenLayers.Layer.Google.v3, {
    loadMapObject() {
      const existingCache = OpenLayers.Layer.Google.cache[this.map.id];

      window.original();

      this.tilesLoading = new Map();
      if (!existingCache) {
        // Add property accessors to new cache
        assignWithOwnProperties(OpenLayers.Layer.Google.cache[this.map.id], {
          _ready: false,
          get ready() {
            return this._ready;
          },
          set ready(ready) {
            if (!this._ready && ready) {
              this._ready = true;
              const tile = this.layer?.tilesLoading.has(null);
              if (tile) {
                this.layer.events.triggerEvent('tileloaded', {tile});
                this.layer.tilesLoading.delete(null);
              }
            }
          },

          _layer: null,
          get layer() {
            return this._layer;
          },
          set layer(layer) {
            const prevLayer = this._layer;
            if (prevLayer === layer) {
              return;
            }

            const tilesLoading = new Map(prevLayer?.tilesLoading.entries());
            if (prevLayer) {
              for (const tile of prevLayer.tilesLoading.values()) {
                prevLayer.events.triggerEvent('tileloaded', {tile});
              }
              prevLayer.tilesLoading.clear();
            }
            this._layer = layer;
            if (layer) {
              for (const [image, prevLayerTile] of tilesLoading) {
                const tile = Object.assign({}, prevLayerTile, {layer});
                layer.tilesLoading.set(image, tile);
                layer.events.triggerEvent('tileloadstart', {tile});
              }
            }
          },
        });
      }
    },
  });
}
