/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {getXYDistance, Geometry, Feature} from '../../syntool.js';
import {PointHandlerWithDragging} from './point_handler.js';

export class OLArrowShapeHandler extends PointHandlerWithDragging {
  /*Public*/
  constructor(options) {
    super(options);
    this.type = Shape.ARROW;
    this.shapeProps = options?.shapeProps || null;
    this.firstProjectedXY = null;
    this.lastProjectedXY = null;
  }

  activate(mapView) {
    if (!super.activate(mapView)) {
      return false;
    }
    this.listenToEvent(this.mapView.hub, 'change:zoom', this['change:zoom'], this);
    return true;
  }
  deactivate() {
    this.firstProjectedXY = null;
    super.deactivate();
  }

  /*Event handlers*/
  'change:zoom'() {
    if (this.lastProjectedXY) {
      this.modifyFeature(this.lastProjectedXY);
    }
  }
  handleMouseLeave() {
    if (this.firstProjectedXY) {
      this.modifyFeature(this.firstProjectedXY);
    } else {
      super.handleMouseLeave();
    }
  }
  handleDragStart(event) {
    this.placeAt(event);
  }
  handleDragMove(recognizer) {
    this.modifyFeature(this.getXYFromEvent(recognizer.event, recognizer));
  }
  handleDragEnd(recognizer) {
    this.placeAt(recognizer.event, recognizer);
  }
  handleDragCancel() {
    this.firstProjectedXY = null;
    this.lastProjectedXY = null;
    this.destroyFeature(true /*force*/);
  }

  /*Protected*/
  placeAt(event, recognizer) {
    const projectedXY = this.getXYFromEvent(event, recognizer);
    if (!this.firstProjectedXY) {
      this.firstProjectedXY = projectedXY;
      this.modifyFeature(projectedXY);
      return;
    }
    this.modifyFeature(projectedXY);
    this.finalize();
  }

  getShape(projectedXY = this.lastProjectedXY) {
    return Object.assign({}, this.shapeProps, {
      type: Shape.ARROW,
      points: [this.firstProjectedXY, projectedXY],
      arrowSize: Shape.getArrowSizeForResolution(
        this.mapView.convertToUnit(this.mapView.getResolution(), 'm')
      ),
    });
  }

  createGeometry(projectedXY) {
    this.lastProjectedXY = projectedXY;
    const distanceToFirstPoint = (
      this.firstProjectedXY
      ? getXYDistance(projectedXY, this.firstProjectedXY)
      : NaN
    );
    if (distanceToFirstPoint > 0) {
      return Shape.createGeometry(this.getShape(projectedXY));
    } else {
      return Geometry.createPointFromXY(projectedXY);
    }
  }

  modifyFeature(projectedXY) {
    const shape = this.getShape(projectedXY);
    const geometry = this.createGeometry(projectedXY);
    if (this.point) {
      this.point.geometry = geometry;
      this.point.properties.shape = shape;
      this.layer.drawFeature(this.point, true /*force*/);
    } else {
      this.point = new Feature(geometry, {shape});
      this.layer.addFeature(this.point);
    }
  }
}
