/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {RAD_IN_DEG, getXYLength, getXYMovement, Feature} from '../../syntool.js';
import {PointHandlerWithDragging} from './point_handler.js';

export class OLEllipseShapeHandler extends PointHandlerWithDragging {
  /*Public*/
  constructor(options) {
    super(options);
    this.type = Shape.ELLIPSE;
    this.shapeProps = options?.shapeProps || null;
    this.lastCenterXY = null;
  }

  activate(mapView) {
    if (!super.activate(mapView)) {
      return false;
    }
    this.listenToEvent(this.mapView.hub, 'change:zoom', this['change:zoom'], this);
    return true;
  }
  deactivate() {
    super.deactivate();
  }

  /*Event handlers*/
  'change:zoom'() {
    if (this.lastCenterXY) {
      this.modifyFeature(this.lastCenterXY);
    }
  }
  handleDragStart(event) {
    this.modifyFeature(this.getXYFromEvent(event));
  }
  handleDragMove(recognizer) {
    this.modifyFeature(this.getXYFromEvent(event, recognizer));
  }
  handleDragEnd(recognizer) {
    this.modifyFeature(this.getXYFromEvent(event, recognizer));
    this.finalize();
  }
  handleDragCancel() {
    this.lastCenterXY = null;
    this.destroyFeature(true /*force*/);
  }

  /*Protected*/
  getShape(projectedXY) {
    if (projectedXY) {
      this.lastCenterXY = projectedXY;
    } else {
      projectedXY = this.lastCenterXY;
    }

    const resolution = (
      this.mapView.convertToUnit(this.mapView.getResolution(), 'm')
    );

    if (this.isDragging) {
      const {shape} = this.point.properties;
      const {center, arrow} = shape;
      const delta = getXYMovement(center, projectedXY);
      const radius = getXYLength(delta);
      if (radius < resolution) {
        return shape;
      }

      return Object.assign({}, shape, {
        rx: radius,
        ry: radius,
        angle: (Math.atan2(delta[1], delta[0]) * RAD_IN_DEG) % 360,
        arrow: arrow && Object.assign({}, arrow,
          Shape.getInitialShapeArrowAtResolution(shape, resolution)
        ),
      });
    }

    const {
      rx = 100,
      ry = 100,
      arrow = null,
    } = this.shapeProps;
    const newProps = {
      type: Shape.ELLIPSE,
      center: projectedXY,
      rx: rx /*px*/ * resolution /*m/px*/,
      ry: ry /*px*/ * resolution /*m/px*/,
    };

    if (arrow) {
      if (!('text' in this.shapeProps)) {
        newProps.text = 'C';
      }

      newProps.arrow = Object.assign(
        {},
        arrow,
        Shape.getInitialShapeArrowAtResolution(newProps, resolution)
      );
    }

    return Object.assign({}, this.shapeProps, newProps);
  }

  createGeometry(shape) {
    return Shape.createGeometry(shape);
  }

  modifyFeature(projectedXY) {
    const shape = this.getShape(projectedXY);
    const geometry = this.createGeometry(shape);

    if (this.point) {
      this.point.geometry = geometry;
      this.point.properties.shape = shape;
      this.layer.drawFeature(this.point, true /*force*/);
    } else {
      this.point = new Feature(geometry, {shape});
      this.layer.addFeature(this.point);
    }
  }
}
