/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {
  throttleOncePerEventLoop,
  getXYDistance,
  Geometry,
  Feature,
} from '../../syntool.js';
import {PointHandlerWithDragging} from './point_handler.js';

export class PathHandler extends PointHandlerWithDragging {
  /*Public*/
  constructor(options) {
    super(options);

    this.line = null;
    this.maxVertices = options?.maxVertices;
    this.canExtendExisting = options?.canExtendExisting;
    this.extendingShapeDataset = null;
    this.extendingFromControlPoint = null;

    this._history = [];
    this._historyIndex = -1;
    this._continuousStartIndex = 0;

    this.geometryChanged = (
      throttleOncePerEventLoop(this.geometryChanged.bind(this))
    );
  }

  get canUndo() {
    return this._historyIndex >= 0;
  }
  get canRedo() {
    return this._historyIndex + 1 < this._history.length;
  }
  get canFinish() {
    return this.getCurrentPointIndex() > 1;
  }

  undo() {
    if (this.canUndo) {
      if (this._historyIndex === this._history.length) {
        this._updateHistoryState();
      }
      this._historyIndex--;
      this._applyHistoryState();
      return true;
    }
    return false;
  }
  redo() {
    if (this.canRedo) {
      this._historyIndex++;
      this._applyHistoryState();
      return true;
    }
    return false;
  }

  getCurrentPointIndex() {
    if (!this.line) {
      return 0;
    }
    const count = Geometry.polylineCount(this.line.geometry);
    return (
      Geometry.isPolygonLinearRing(this.line.geometry)
      ? count - 2
      : count - 1
    );
  }

  /*Event handlers*/
  handleMouseLeave() {
    const currentPointIndex = this.getCurrentPointIndex();
    if (currentPointIndex === 0) {
      this.layer.destroyFeatures();
      this.line = null;
    } else {
      const lastPoint = (
        Geometry.getComponentAtIndex(this.line.geometry, currentPointIndex - 1)
      );
      this.modifyFeature(Geometry.pointToXY(lastPoint));
    }
  }
  handleDragStart(event) {
    this._continuousStartIndex = this.getCurrentPointIndex();
    this.placeAt(event);
    this._pushHistoryState();
  }
  handleDragMove(recognizer) {
    this.placeAt(recognizer.event, recognizer);
  }
  handleDragEnd(recognizer) {
    this.placeAt(recognizer.event, recognizer);
    this.simplify(this._continuousStartIndex);
  }
  handleDragCancel() {
    if (this.line) {
      Geometry.polylineRemovePointsInRange(
        this.line.geometry,
        this._continuousStartIndex
      );
      this.geometryChanged();
    }
  }

  /*Protected*/
  _updateHistoryState() {
    if (this._historyIndex === -1) {
      return;
    }
    const state = {
      geometry: Geometry.clone(this.line.geometry),
      extendingShapeDataset: this.extendingShapeDataset,
      extendingFromControlPoint: this.extendingFromControlPoint,
    };
    this._history[this._historyIndex] = state;
  }
  _applyHistoryState() {
    if (this._historyIndex === -1) {
      this.layer.destroyFeatures();
      this.line = null;
      this.extendingShapeDataset = null;
      this.extendingFromControlPoint = null;
      if (this.mapView.lastMouseMoveEvent) {
        this.handleMouseMove(this.mapView.lastMouseMoveEvent);
      }
      this.geometryChanged();
      return;
    }

    const state = this._history[this._historyIndex];
    const geometry = Geometry.clone(state.geometry);

    const count = Geometry.polylineCount(geometry);
    const currentPointIndex = (
      Geometry.isPolygonLinearRing(geometry)
      ? count - 2
      : count - 1
    );
    const lastPointIndex = currentPointIndex - 1;
    const lastPoint = Geometry.getComponentAtIndex(geometry, lastPointIndex);
    const currentPoint = Geometry.getComponentAtIndex(geometry, currentPointIndex);

    this.point.geometry = currentPoint;
    if (this.mapView.lastMouseMoveEvent) {
      this.handleMouseMove(this.mapView.lastMouseMoveEvent);
    } else {
      this.modifyFeature(Geometry.pointToXY(lastPoint));
    }

    Geometry.setCollectionComponents(
      this.line.geometry,
      Geometry.getCollectionComponents(geometry)
    );
    this.extendingShapeDataset = state.extendingShapeDataset;
    this.extendingFromControlPoint = state.extendingFromControlPoint;
    this.geometryChanged();
  }
  _pushHistoryState() {
    this._updateHistoryState();
    this._historyIndex++;
    this._history.length = this._historyIndex;
  }

  tryExtendingExisting(event) {
    if (
      this.canExtendExisting &&
      !this.extendingShapeDataset &&
      this.getCurrentPointIndex() === 0
    ) {
      const controlPoint = (
        this.mapView.shapesEditingControl?.getControlPointFromEvent(event)
      );
      if (controlPoint) {
        this.extendingShapeDataset = (
          this.mapView.shapesEditingControl.feature.dataset
        );
        this.extendingFromControlPoint = controlPoint;
        return true;
      }
    }
    return false;
  }
  getXYFromEvent(event, recognizer) {
    return super.getXYFromEvent(event, recognizer, this.isDragging);
  }

  geometryChanged() {
    this.layer?.redrawAllFeatures(true /*force*/);
    this.hub.emit('change');
  }

  placeAt(event, recognizer) {
    const projectedXY = (
      this.tryExtendingExisting(event)
      ? [this.extendingFromControlPoint.x, this.extendingFromControlPoint.y]
      : this.getXYFromEvent(event, recognizer)
    );
    const currentPointIndex = this.getCurrentPointIndex();
    const lastPointIndex = currentPointIndex - 1;
    const lastPoint = (
      this.line &&
      Geometry.getComponentAtIndex(this.line.geometry, lastPointIndex)
    );
    const distanceFromLastPoint = (
      lastPoint
      ? getXYDistance(projectedXY, Geometry.pointToXY(lastPoint))
      : NaN
    );
    if (distanceFromLastPoint < this.mapView.getResolution() + 1) {
      return;
    }

    if (!this.isDragging) {
      this._pushHistoryState();
    }

    this.modifyFeature(projectedXY);

    const lineGeometry = this.line.geometry;
    Geometry.polylineAddPoint(lineGeometry, projectedXY);
    this.point.geometry = (
      Geometry.getComponentAtIndex(lineGeometry, currentPointIndex + 1)
    );
    this.geometryChanged();
    if (Geometry.polylineCount(lineGeometry) === this.maxVertices + 1) {
      this.finishGeometry();
    }
  }

  simplify(fromIndex = 0) {
    Geometry.polylineSimplify(
      this.line.geometry,
      this.mapView.getResolution(),
      {
        fromIndex,
        toIndex: this.getCurrentPointIndex(),
        inPlace: true,
      }
    );
    this.geometryChanged();
  }

  getShape() {
    switch (this.type) {
    case Shape.LINE:
      return {
        type: this.type,
        points: Geometry.polylineToXYs(this.getGeometry()),
      };
    default:
      return super.getShape();
    }
  }
  getGeometry() {
    return this.line?.geometry;
  }
  finishGeometry() {
    const cancel = !this.canFinish;
    Geometry.removeComponentAtIndex(
      this.line.geometry,
      this.getCurrentPointIndex(),
      true /*force*/
    );
    if (this.point) {
      this.layer.removeFeature(this.point);
    }
    this.finalize(cancel);
  }

  modifyFeature(projectedXY) {
    if (!this.line) {
      const lineGeometry = Geometry.createPolylineFromXYs([projectedXY]);
      const pointGeometry = Geometry.getComponentAtIndex(lineGeometry, 0);
      this.point = new Feature(pointGeometry);
      this.line = new Feature(lineGeometry);
      this.layer.addFeatures([this.line, this.point]);
      return;
    }
    super.modifyFeature(projectedXY);
  }
  destroyFeature(force) {
    super.destroyFeature(force);
    this.line = null;
    this.extendingShapeDataset = null;
    this.extendingFromControlPoint = null;
    this._history.length = 0;
    this._historyIndex = -1;
  }
}
