/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {
  extendClass,
  EventsHub,
  Geometry,
  Feature,
  MapView,
} from '../../syntool.js';
import {eventsListeningExtension} from '../../utils/events_listening_extension.js';

let nextHandlerId = 0;

export class PointHandler {
  /*Public*/
  constructor(options) {
    this.id = options?.id ?? (
      `${options?.name || this.constructor.name}_${nextHandlerId++}`
    );
    this.name = options?.name;
    this.type = options?.type ?? Shape.WKT;
    this.cssClass = options?.cssClass;
    this.modifiers = options?.modifiers;
    this.keys = options?.keys;
    this.snapToControlPoints = options?.snapToControlPoints ?? true;
    this.dragMovementThreshold = 20;
    this.active = false;
    this.mapView = null;
    this.hub = new EventsHub();
    this.point = null;
    this.layer = null;
  }

  activate(mapView) {
    if (this.active || !mapView) {
      return false;
    }
    this.active = true;
    this.mapView = mapView;
    this.layer = mapView.shapesDrawingLayer;

    this.listenToEvent(this.mapView.hub, 'mousemove', MapView.DRAW, this.handleMouseMove, this);
    this.listenToEvent(this.mapView.hub, 'mouseleave+passive+capture', this.handleMouseLeave, this);
    this.listenToEvent(this.mapView.hub, 'click', MapView.DRAW, this.handleClick, this);
    this.listenToEvent(this.mapView.hub, 'tap-1', MapView.DRAW, this.handleTap, this);

    this.hub.emit('activate');

    this.mapView.setDragMovementThreshold?.(this.dragMovementThreshold);
    if (this.mapView.lastMouseMoveEvent) {
      this.handleMouseMove(this.mapView.lastMouseMoveEvent);
    }
    return true;
  }
  deactivate() {
    if (!this.active) {
      return;
    }

    this.stopAllListening();
    this.active = false;

    this.cancel();
    this.layer.destroyFeatures();
    this.mapView.restoreDragMovementThreshold?.();

    this.hub.emit('deactivate');
    this.mapView = null;
    this.layer = null;
  }

  get canFinish() {
    return false;
  }

  /*Event handlers*/
  handleMouseMove(event) {
    this.modifyFeature(this.getXYFromEvent(event));
  }
  handleMouseLeave() {
    if (this.point) {
      this.layer.destroyFeatures();
      this.point = null;
    }
  }
  handleClick(event) {
    if (this.mapView.isValidClick(event)) {
      this.placeAt(event);
    }
  }
  handleTap(event, recognizer) {
    this.placeAt(event, recognizer);
  }

  /*Protected*/
  getXYFromEvent(event, recognizer, ignoreSnap) {
    if (this.snapToControlPoints && !ignoreSnap) {
      const controlPoint = (
        this.mapView.shapesEditingControl?.getControlPointFromEvent(event)
      );
      if (controlPoint) {
        return [controlPoint.x, controlPoint.y];
      }
      const feature = this.mapView.shapesLayer.getFeatureFromEvent(event);
      const geometry = (
        this.mapView.shapesEditingControl.feature === feature &&
        feature?.geometry
      );
      if (Geometry.isPoint(geometry)) {
        return Geometry.pointToXY(geometry);
      }
    }
    return (
      recognizer
      ? this.mapView.getProjectedXYFromClientXY(recognizer.location)
      : this.mapView.getProjectedXYFromPointerEvent(event)
    );
  }

  geometryChanged() {
    this.point.properties.shape = this.getShape();
    this.layer.redrawAllFeatures(true /*force*/);
  }

  placeAt(event, recognizer) {
    this.modifyFeature(this.getXYFromEvent(event, recognizer));
    this.finalize();
  }

  getShape() {
    switch (this.type) {
    case Shape.TEXT:
      return {
        type    : this.type,
        location: Geometry.pointToXY(this.getGeometry()),
        text    : 'Text goes here',
      };
    case Shape.WKT:
      return {
        type: this.type,
        wkt : Geometry.toWKT(this.getGeometry()),
      };
    default:
      return null;
    }
  }
  getGeometry() {
    return this.point?.geometry;
  }

  modifyFeature(projectedXY) {
    if (!this.point) {
      this.point = new Feature(Geometry.createPointFromXY(projectedXY));
      this.point.properties.shape = this.getShape();
      this.layer.addFeature(this.point);
      return;
    }

    Geometry.pointMoveTo(this.point.geometry, projectedXY);
    this.geometryChanged();
  }
  destroyFeature(force) {
    if (this.layer && force) {
      this.layer.destroyFeatures();
    }
    this.point = null;
  }

  finalize(cancel) {
    if (!cancel) {
      this.hub.emit('done', this.getShape());
    }
    this.destroyFeature(cancel);
  }
  cancel() {
    this.finalize(true);
  }
}
extendClass(PointHandler, eventsListeningExtension);

export class PointHandlerWithDragging extends PointHandler {
  /*Public*/
  constructor(options) {
    super(options);

    this.dragMovementThreshold = 2;
    this._dragging = false;
  }

  activate(mapView) {
    if (!super.activate(mapView)) {
      return false;
    }
    this.mapView.showPanningHelp?.();
    this.listenToEvent(this.mapView.hub, 'drag', MapView.DRAW, this.handleDrag, this);
    return true;
  }
  deactivate() {
    if (!this.active) {
      return;
    }

    this.mapView.restoreDragMovementThreshold?.();
    this.mapView.hidePanningHelp?.();
    if (this._dragging) {
      this.mapView.cancelDrag();
    }
    this._dragging = false;
    super.deactivate();
  }

  get isDragging() {
    return this._dragging;
  }

  /*Event handlers*/
  handleMouseLeave() {
    if (!this.isDragging) {
      super.handleMouseLeave();
    }
  }
  handleDrag(recognizer, captureDrag) {
    if (this.mapView.isDragForPanning(recognizer)) {
      return true;
    }

    const capturedDrag = captureDrag({
      context: this,
      move   : this._handleDragMove,
      end    : this._handleDragEnd,
      cancel : this._handleDragCancel,
    });

    if (capturedDrag) {
      this.handleDragStart(recognizer.startEvent, recognizer);

      this._dragging = true;
      this.handleDragMove(recognizer);
    }
    return !capturedDrag;
  }
  _handleDragMove(recognizer) {
    this.handleDragMove(recognizer);
  }
  _handleDragEnd(recognizer) {
    this.handleDragEnd(recognizer);
    this._dragging = false;
  }
  _handleDragCancel() {
    this._dragging = false;
    this.handleDragCancel();
  }

  handleDragStart(event, recognizer) { /*empty*/ }
  handleDragMove(recognizer) { /*empty*/ }
  handleDragEnd(recognizer) { /*empty*/ }
  handleDragCancel() { /*empty*/ }

  /*Protected*/
  placeAt(event, recognizer) {
    this.modifyFeature(this.getXYFromEvent(event, recognizer));
    if (!this.isDragging) {
      this.finalize();
    }
  }
}
