/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Shape} from '../datasets/shape_dataset/shape.js';
import {
  HIDKeys,
  HIDModifiers,
  Geometry,
  Types as is,
  Config,
  Syntool,
} from '../syntool.js';
import {OLArrowShapeHandler} from './shapes/ol_arrow_shape_handler.js';
import {OLEllipseShapeHandler} from './shapes/ol_ellipse_shape_handler.js';
import {OLGeometryHandler} from './shapes/ol_geometry_handler.js';
import {OLMushroomShapeHandler} from './shapes/ol_mushroom_shape_handler.js';
import {PathHandler} from './shapes/path_handler.js';
import {PointHandler} from './shapes/point_handler.js';
import {PolygonHandler} from './shapes/polygon_handler.js';
import '../map_view_extensions/map_view_shapes_drawing.js';

var DrawableShape = is.shape({
  title        : is.string.isRequired,
  cssClassNames: is.string,
  wkt          : is.WKT.isRequired,
});

Object.assign(Config.configFieldValidators, {
  drawableShapes: is.arrayOf(DrawableShape.isRequired),
});

Syntool.extend({
  [Syntool.NAME]: 'ShapesDrawing',
  [Syntool.CONSTRUCTOR]() {
    this._shapesDrawingHandlers = new Map();
    this._shapesDrawingActive = false;
    this._shapesDrawingActiveHandlerId = null;
  },
  [Syntool.INIT]() {
    const defaultHandlers = [
      new PointHandler({
        id: 'point',
        name: 'Point',
        keys: HIDKeys.D,
      }),
      new OLArrowShapeHandler({
        id: 'arrow',
        name: 'Arrow',
      }),
      new PathHandler({
        id: 'polyline',
        name: 'Polyline',
        type: Shape.LINE,
        keys: HIDKeys.L,
        canExtendExisting: true,
      }),
      new PolygonHandler({
        id: 'polygon',
        name: 'Polygon',
        keys: HIDKeys.P,
      }),
      new OLEllipseShapeHandler({
        id: 'circle',
        name: 'Eddy/Ellipse',
        shapeProps: {
          rx   : 100,
          ry   : 100,
          angle: 0,
          arrow: {},
        },
      }),
      new OLMushroomShapeHandler({
        id: 'mushroom',
        name: 'Hammerhead/Mushroom',
      }),
      new PointHandler({
        id: 'text',
        name: 'Text',
        type: Shape.TEXT,
      }),
    ];

    this.shortcutsStore.add(
      'User shapes drawing',
      [
        {
          id         : 'user-shapes-undo-drawing',
          modifiers  : [HIDModifiers.ctrlOrMeta],
          keys       : HIDKeys.Z,
          description: 'Undo',
          handler    : () => {
            const activeDrawingHandler = this.shapesDrawingActiveHandler;
            if (activeDrawingHandler) {
              activeDrawingHandler.undo?.();
              return true;
            }
            return false;
          },
        },
        {
          id         : 'user-shapes-redo-drawing',
          modifiers  : [HIDModifiers.ctrlOrMeta, HIDModifiers.shift],
          keys       : HIDKeys.Z,
          description: 'Redo',
          handler    : () => {
            const activeDrawingHandler = this.shapesDrawingActiveHandler;
            if (activeDrawingHandler) {
              activeDrawingHandler.redo?.();
              return true;
            }
            return false;
          },
        },
        {
          id         : 'user-shapes-finish-drawing',
          keys       : HIDKeys.enter,
          description: 'Finalize the shape being drawn',
          context    : this,
          handler    : this.commitShapeDrawing,
        },
        {
          id         : 'user-shapes-cancel-drawing',
          keys       : HIDKeys.escape,
          description: 'Cancel the shape being drawn',
          context    : this,
          handler    : this.cancelShapeDrawing,
        },
      ]
    );
    for (const handler of defaultHandlers) {
      this.addShapesDrawingHandler(handler);
    }
    if (this.config.drawableShapes) {
      for (const drawableShape of this.config.drawableShapes) {
        this.addShapesDrawingHandler(new OLGeometryHandler({
          name: drawableShape.title,
          cssClass: drawableShape.cssClassNames,
          geometry: Geometry.fromWKT(drawableShape.wkt),
        }));
      }
    }
    this.shapesDrawingActive = true;
    this._updateShapesDrawingShortcuts();
  },
  [Syntool.STATE]: {
    get() {
      return {
        shapesDrawingActive: this.shapesDrawingActive,
        shapesDrawingActiveHandler: this.shapesDrawingActiveHandlerId,
      };
    },
    set({shapesDrawingActive, shapesDrawingActiveHandler}) {
      this.shapesDrawingActive = shapesDrawingActive;
      this.shapesDrawingActiveHandlerId = shapesDrawingActiveHandler;
    },
  },

  get shapesDrawingActive() {
    return this._shapesDrawingActive;
  },
  set shapesDrawingActive(active) {
    if (this._shapesDrawingActive === !active) {
      this._shapesDrawingActive = !!active;
      if (this._shapesDrawingActive) {
        this.mapView.hub
          .on('change:selection', this.handleShapesDrawingSelectionChange, this)
          .on('picking:start', this.cancelShapeDrawing, this);
      } else {
        this.shapesDrawingActiveHandler = null;
        this.cancelShapeDrawing();
        this.mapView.hub
          .off('change:selection', this.handleShapesDrawingSelectionChange, this)
          .off('picking:start', this.cancelShapeDrawing, this);
      }
      const shortcutIds = Array.from(
        this._shapesDrawingHandlers.keys(),
        (id) => this.getShortcutIdForHandlerWithId(id)
      );
      this.shortcutsStore.setShortcutsWithIdsEnabled(
        shortcutIds,
        this._shapesDrawingActive
      );
      this.hub.emit('change:shapesDrawingActive');
      this._invalidateState();
    }
  },

  get shapesDrawingActiveHandlerId() {
    return this._shapesDrawingActiveHandlerId;
  },
  set shapesDrawingActiveHandlerId(handlerId) {
    if (
      this.shapesDrawingActive &&
      this._shapesDrawingActiveHandlerId !== handlerId
    ) {
      const oldHandler = (
        this.getShapesDrawingHandlerWithId(this._shapesDrawingActiveHandlerId)
      );
      if (oldHandler) {
        this._shapesDrawingActiveHandlerId = null;
        oldHandler.deactivate();
        this.mapView.activateDatasetInteractions?.();
      }
      const handler = this.getShapesDrawingHandlerWithId(handlerId);
      if (handler?.active) {
        this._shapesDrawingActiveHandlerId = handlerId;
        this.mapView.deactivateDatasetInteractions?.();
        this.productsStore.selectProduct(this.userShapesProduct);
      } else {
        handler?.activate(this.mapView);
      }
      this._updateShapesDrawingShortcuts();
      this.hub.emit('change:shapesDrawingActiveHandler');
      this.hub.emit('change:shapesDrawingActiveHandler:state');
      this._invalidateState();
    }
  },

  get shapesDrawingActiveHandler() {
    return this.getShapesDrawingHandlerWithId(this.shapesDrawingActiveHandlerId);
  },
  set shapesDrawingActiveHandler(handler) {
    this.shapesDrawingActiveHandlerId = handler?.id;
  },

  _updateShapesDrawingShortcuts() {
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-undo-drawing'],
      this.shapesDrawingActiveHandler?.canUndo
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-redo-drawing'],
      this.shapesDrawingActiveHandler?.canRedo
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-finish-drawing'],
      this.shapesDrawingActiveHandler?.canFinish
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-cancel-drawing'],
      this.shapesDrawingActiveHandlerId !== null
    );
  },

  getShortcutIdForHandlerWithId(handlerId) {
    return 'user-shapes-toggle-drawing-' + handlerId;
  },
  addShapesDrawingHandler(handler) {
    if (handler.active) {
      handler.deactivate();
    }
    this._shapesDrawingHandlers.set(handler.id, handler);
    this.shortcutsStore.add('User shapes drawing', {
      id         : this.getShortcutIdForHandlerWithId(handler.id),
      modifiers  : handler.modifiers,
      keys       : handler.keys,
      description: 'Draw ' + handler.name,
      handler    : () => this.toggleDrawingWith(handler.id),
    });

    handler.hub
      .on('activate', () => {
        this.shapesDrawingActiveHandler = handler;
      })
      .on('deactivate', () => {
        if (this.shapesDrawingActiveHandler === handler) {
          this.shapesDrawingActiveHandler = null;
        }
      })
      .on('change', () => {
        if (this.shapesDrawingActiveHandler === handler) {
          this._updateShapesDrawingShortcuts();
          this.hub.emit('change:shapesDrawingActiveHandler:state');
        }
      })
      .on('done', (shapeOrGeometry) => {
        this.handleShapesDrawingDone(handler, shapeOrGeometry);
      });
    this.hub.emit('change:shapesDrawingHandlers');
  },
  getShapesDrawingHandlerWithId(id) {
    return this._shapesDrawingHandlers.get(id);
  },

  get shapesDrawingHandlers() {
    return this._shapesDrawingHandlers.values();
  },

  toggleDrawingWith(id) {
    if (this.shapesDrawingActiveHandlerId === id) {
      this.shapesDrawingActiveHandlerId = null;
    } else {
      this.shapesDrawingActiveHandlerId = id;
    }
    return true;
  },
  commitShapeDrawing() {
    const handler = this.shapesDrawingActiveHandler;
    if (handler?.canFinish) {
      handler.finishGeometry();
      return true;
    }
    return false;
  },
  cancelShapeDrawing() {
    if (this.shapesDrawingActiveHandler) {
      this.shapesDrawingActiveHandler = null;
      return true;
    }
    return false;
  },

  /*Protected - Event handlers*/
  handleShapesDrawingSelectionChange() {
    if (this.mapView.hasSelectedDatasets()) {
      this.cancelShapeDrawing();
    }
  },
  handleShapesDrawingDone(handler, shape) {
    const {extendingShapeDataset, extendingFromControlPoint} = handler;
    this.shapesDrawingActiveHandler = null;

    if (shape) {
      if (extendingShapeDataset) {
        const shapeToExtend = extendingShapeDataset.shape;
        const updates = Shape.updatesFromChanges(
          shapeToExtend,
          {extend: {from: extendingFromControlPoint, shape}},
          extendingShapeDataset.shapeFeature.geometry
        );

        if (updates) {
          const extendedShape = Object.assign({}, shapeToExtend, updates);

          if (!Shape.same(shapeToExtend, extendedShape)) {
            this.shapesStore.update(extendedShape.id, extendedShape);
            return;
          }
        }
      }

      shape.start = this.timelineStore.minDate;
      shape.end = this.timelineStore.maxDate;

      this.shapesStore.add(this.shapesStore.create(shape));
    }
  },
});
