/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {isApple, HIDKeys, HIDModifiers, Syntool} from '../syntool.js';
import '../map_view_extensions/map_view_shapes_editing.js';

Syntool.extend({
  [Syntool.NAME]: 'ShapesEditing',
  [Syntool.CONSTRUCTOR]() {
    this._shapesEditingActive = false;
  },
  [Syntool.INIT]() {
    this.mapView.shapesEditingControl.onChange = (dataset) => {
      this.shapesStore.update(dataset.shape.id, dataset.shownShape);
    };
    this.shortcutsStore.add('User shapes editing', [
      {
        id         : 'user-shapes-undo',
        modifiers  : [HIDModifiers.ctrlOrMeta],
        keys       : HIDKeys.Z,
        description: 'Undo',
        handler    : () => {
          this.shapesStore.undo();
          return true;
        },
      },
      {
        id         : 'user-shapes-redo',
        modifiers  : [HIDModifiers.ctrlOrMeta, HIDModifiers.shift],
        keys       : HIDKeys.Z,
        description: 'Redo',
        handler    : () => {
          this.shapesStore.redo();
          return true;
        },
      },
      this.mapView.getSelectedDatasets && {
        id         : 'user-shapes-duplicate-selected-shapes',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.D,
        description: 'Duplicate the selected shape(s)',
        handler    : () => {
          if (this.duplicateSelectedShapes()) {
            return 'Duplicated shape(s)';
          }
          return false;
        },
      },
      this.mapView.getSelectedDatasets && {
        id         : 'user-shapes-remove-selected-shapes',
        keys       : isApple ? HIDKeys.backspace : HIDKeys.delete,
        description: 'Remove the selected shape(s) or the hovered vertex',
        handler    : () => {
          if (this.removeUserShapeHoveredVertex()) {
            return 'Removing vertex';
          }
          if (this.removeSelectedShapes()) {
            return 'Removing shape(s)';
          }
          return false;
        },
      },
    ]);
    this.shapesEditingActive = true;
  },
  [Syntool.STATE]: {
    get() {
      return {shapesEditingActive: this.shapesEditingActive};
    },
    set({shapesEditingActive}) {
      this.shapesEditingActive = shapesEditingActive;
    },
  },

  get shapesEditingActive() {
    return this._shapesEditingActive;
  },
  set shapesEditingActive(active) {
    if (this._shapesEditingActive === !active) {
      this._shapesEditingActive = !!active;
      if (this._shapesEditingActive) {
        this.mapView.shapesEditingControl.activate();
        this.shapesStore.hub
          .on('add', this._updateShapesEditingShortcuts, this)
          .on('update', this._updateShapesEditingShortcuts, this)
          .on('remove', this._updateShapesEditingShortcuts, this);
        this.mapView.hub
          .on('change:selection', this._updateShapesEditingShortcuts, this);
      } else {
        this.mapView.shapesEditingControl.deactivate();
        this.shapesStore.hub
          .off('add', this._updateShapesEditingShortcuts, this)
          .off('update', this._updateShapesEditingShortcuts, this)
          .off('remove', this._updateShapesEditingShortcuts, this);
        this.mapView.hub
          .off('change:selection', this._updateShapesEditingShortcuts, this);
      }
      this._updateShapesEditingShortcuts();
      this.hub.emit('change:shapesEditingActive');
      this._invalidateState();
    }
  },

  removeUserShapeHoveredVertex() {
    return this.mapView.shapesEditingControl.removeHoveredVertex();
  },

  _updateShapesEditingShortcuts() {
    const selectedShapeDatasets = (
      this.mapView.getSelectedDatasets(this.filters.dataset.isShape)
    );
    const selectedEditableShapeDatasets = (
      selectedShapeDatasets.filter(this.filters.dataset.isEditableShape)
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-undo'],
      this._shapesEditingActive && this.shapesStore.canUndo
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-redo'],
      this._shapesEditingActive && this.shapesStore.canRedo
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-remove-selected-shapes'],
      this._shapesEditingActive && selectedEditableShapeDatasets.length > 0
    );
    this.shortcutsStore.setShortcutsWithIdsEnabled(
      ['user-shapes-duplicate-selected-shapes'],
      this._shapesEditingActive && selectedShapeDatasets.length > 0
    );
  },
});
