/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {NotificationMessageType, Syntool} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';
import './import.js';

Syntool.extend({
  [Syntool.NAME]: 'ShapesImportExport',
  [Syntool.CONSTRUCTOR]() {
    this.importersFromJSON.push(
      (json, options) => this.importShapesFromJSON(json, options)
    );

    this.shapesImportOptions = {};
    this.shapesExporters = new Map();
    this.shapesExporters.set('json', {
      label: 'ShapesJSON',
      canExportShapeOfType(type) {
        return true;
      },
      export: (shapes, options) => {
        options = Object.assign({
          label   : 'shape' + (shapes.length === 1 ? '' : 's'),
          fileName: 'syntool-shapes-' + (new Date().toISOString()) + '.json',
        }, options);
        if (options.toClipboard) {
          options.fileName = null;
        }
        this.exportJSON(this.exportShapesToJSON(shapes), options);
      },
    });
  },
  _prepareShapesImportFromJSON(json, options) {
    const {
      checkUserShapesDuplicates = true,
      checkImportDuplicates = true,
    } = options?.shapes ?? {};

    const newShapesSet = new Set();
    const shapes = [];
    const duplicateShapes = [];
    const remainingJson = [];
    for (const element of json) {
      const shape = this.shapesStore.createFromJSON(element);
      if (shape) {
        const shapeJSON = shape.toJSON();

        let duplicate = false;
        if (checkImportDuplicates) {
          const hash = JSON.stringify(shapeJSON);
          duplicate = newShapesSet.has(hash);
          if (!duplicate) {
            newShapesSet.add(hash);
          }
        }

        if (
          !duplicate &&
          checkUserShapesDuplicates &&
          this.shapesStore.isShapeDuplicate(shape)
        ) {
          duplicate = true;
        }

        (duplicate ? duplicateShapes : shapes).push(shape);
      } else {
        remainingJson.push(element);
      }
    }
    return {shapes, duplicateShapes, remainingJson};
  },
  importShapesFromJSON(json, options) {
    const {
      product = this.userShapesProduct,
      onBeforeAddShapes,
      onAddShapes,
    } = options?.shapes ?? {};

    const addShapes = (shapes, duplicateShapes) => {
      onBeforeAddShapes?.(shapes, duplicateShapes);

      if (product === this.userShapesProduct) {
        this.shapesStore.add(shapes);
      } else {
        shapes = shapes.map((shape) => shape.toJSON());
        if (product.shapes) {
          product.shapes.push(...shapes);
        } else {
          product.shapes = shapes;
        }

        if (this.productsStore.hasProduct(product.id)) {
          this.shapesStore.loadShapesFromProduct(product);
        } else {
          this.productsStore.addProduct(
            Object.assign(product, {
              tags: this.userShapesProduct.tags,
              type: 'SERVER_SHAPES',
              selected: true,
              mustBeCurrent: false,
              opacity: 1,
              stackLevel: 150,
              style: this.userShapesProduct.style,
            }, product)
          );
        }
      }

      onAddShapes?.(shapes, duplicateShapes);
    };

    const {
      shapes,
      duplicateShapes,
      remainingJson,
    } = this._prepareShapesImportFromJSON(json, options);

    addShapes(shapes, duplicateShapes);

    const addedCount = shapes.length;
    const duplicateCount = duplicateShapes.length;
    if (addedCount + duplicateCount === 0) {
      return remainingJson;
    }

    const intoProduct = this.productsStore.getProduct(product.id);
    const addedPlural = (addedCount === 1 ? '' : 's');
    const duplicatePlural = (duplicateCount === 1 ? '' : 's');
    const importedMessage = `Imported ${addedCount} shape${addedPlural}`;
    const intoMessage = (
      intoProduct === this.userShapesProduct
      ? ''
      : ` into "${intoProduct.label}"`
    );
    const duplicatesMessage = (
      duplicateCount > 0
      ? ` and skipped ${duplicateCount} duplicate${duplicatePlural}`
      : ''
    );
    const message = importedMessage + intoMessage + duplicatesMessage;
    console.log('Import::' + message);
    if (duplicateCount === 0) {
      this.notificationCenter.showInfo(message);
    } else {
      this.notificationCenter.showMessage({
        type: NotificationMessageType.warning,
        text: message,
        buttons: [
          {
            text   : `import ${duplicateCount} duplicate${duplicatePlural}`,
            closes : true,
            props  : {title: `Import the duplicate${duplicatePlural} anyway`},
            onClick: () => {
              addShapes(duplicateShapes);
              const newMessage = (
                `Imported ${duplicateCount} duplicate shape${duplicatePlural}` +
                intoMessage
              );
              console.log(`Import::${newMessage}`);
              this.notificationCenter.showInfo(newMessage);
            },
          },
        ],
        hideAfter: -1,
      });
    }

    return remainingJson;
  },
  exportShapesToJSON(shapes) {
    return shapes.map(Shape.toExternalJSON);
  },

  exportShapes(shapes, options = {}) {
    const {format = 'json'} = options;
    this.shapesExporters.get(format)?.export(shapes, options);
  },
});
