/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  normalizeISODateString,
  Projection,
  Geometry,
  GeoJSON,
  Syntool,
} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';
import './shapes_import_export.js';

const JSON_MIME_TYPE = 'application/json';
const GEOJSON_MIME_TYPE = 'application/geo+json';

function shapeFromFeature(feature, proj) {
  const geometry = feature.geometry;
  const shape = {};

  switch (geometry?.type) {
  case 'LineString':
    if (feature.properties?.author == null) {
      shape.type = Shape.LINE;
      shape.points = geometry.coordinates;
      if (proj) {
        shape.points = shape.points.map((xy) => proj.unprojectXY(xy));
      }
      break;
    }
    /*fallthrough*/
  case 'Point':
  case 'MultiPoint':
  case 'MultiLineString':
  case 'Polygon':
  case 'MultiPolygon':
  case 'GeometryCollection':
    shape.type = Shape.WKT;
    shape.wkt = Geometry.toWKT(GeoJSON.parseGeometry(geometry, {
      sourceProjection: proj,
    }));
    shape.author = feature.properties?.author;
    break;
  default:
    return null;
  }

  const date = feature.properties?.date;
  const start = feature.properties?.timespan?.start;
  const end = feature.properties?.timespan?.end;

  shape.id = feature.id;
  if (start == null && end == null && date != null) {
    shape.start = new Date(normalizeISODateString(date)).getTime();
    shape.end = shape.start + 1000;
  } else {
    if (start != null) {
      shape.start = new Date(normalizeISODateString(start)).getTime();
    }
    if (end != null) {
      shape.end = new Date(normalizeISODateString(end)).getTime();
    }
  }
  shape.text = feature.properties?.text ?? feature.properties?.name;
  shape.color = feature.properties?.color;

  return shape;
}
function shapeToFeature(shape) {
  const json = Shape.toExternalJSON(shape);
  const feature = {};
  switch (shape.type) {
  case Shape.TEXT:
    feature.geometry = {
      type: 'Point',
      coordinates: json.location,
    };
    break;
  case Shape.LINE:
  case Shape.ARROW:
    feature.geometry = {
      type: 'LineString',
      coordinates: json.points,
    };
    break;
  case Shape.WKT:
    feature.geometry = GeoJSON.serializeGeometry(Geometry.fromWKT(json.wkt));
    if (json.author) {
      feature.properties = {author: json.author};
    }
    break;
  default:
    return null;
  }

  feature.id = shape.id;
  feature.type = 'Feature';

  const hasStart = json.start != null;
  const hasEnd = json.end != null;
  const hasTimespan = hasStart || hasEnd;

  if (hasTimespan || json.text || json.color) {
    if (!feature.properties) {
      feature.properties = {};
    }
    if (hasTimespan) {
      feature.properties.timespan = {};
      if (hasStart) {
        feature.properties.timespan.start = shape.start.toISOString();
      }
      if (hasEnd) {
        feature.properties.timespan.end = shape.end.toISOString();
      }
    }
    if (shape.text) {
      feature.properties.text = shape.text;
    }
    if (shape.color) {
      feature.properties.color = shape.color;
    }
  }
  return feature;
}

Syntool.extend({
  [Syntool.NAME]: 'ShapesImportExportGeoJSON',
  [Syntool.CONSTRUCTOR]() {
    this.acceptedImportFormats.push('GeoJSON');
    this.acceptedImportFileTypes.push('.geojson', GEOJSON_MIME_TYPE);
    this.importersFromString.push((string, mimeType, options) => {
      if (mimeType === GEOJSON_MIME_TYPE) {
        return this._importFromString(string, JSON_MIME_TYPE, options);
      }
      return false;
    });
    this.importersFromJSON.unshift(
      (json, options) => this._importShapesFromGeoJSON(json, options)
    );

    this.shapesExporters.set('geojson', {
      label: 'GeoJSON',
      canExportShapeOfType(type) {
        switch (type) {
        case Shape.TEXT:
        case Shape.LINE:
        case Shape.ARROW:
        case Shape.WKT:
          return true;
        default:
          return false;
        }
      },
      export: (shapes, options) => {
        options = Object.assign({
          label   : 'shape' + (shapes.length === 1 ? '' : 's') + ' as GeoJSON',
          fileName: 'syntool-shapes-' + (new Date().toISOString()) + '.geojson',
          fileType: GEOJSON_MIME_TYPE,
        }, options);
        if (options.toClipboard) {
          options.fileName = null;
        }
        this.exportJSON(this.exportShapesToGeoJSON(shapes), options);
      },
    });
  },

  exportShapesToGeoJSON(shapes) {
    const features = shapes.map(shapeToFeature).filter(Boolean);
    if (features.length === 1) {
      return features[0];
    }
    return {type: 'FeatureCollection', features};
  },

  _importShapesFromGeoJSON(mixedJson, options, proj = null) {
    const json = [];
    for (const element of mixedJson) {
      switch (element.type) {
      case 'Feature':
        json.push(shapeFromFeature(element, proj) || element);
        break;
      case 'FeatureCollection':
        try {
          const projName = element.crs?.properties?.name;
          proj = projName ? Projection.from(projName) : null;
          json.push(
            ...this._importShapesFromGeoJSON(element.features, options, proj)
          );
        } catch (error) {
          console.warn(
            'GeoJSON::Ignoring feature collection with unsupported "crs"',
            element,
            error
          );
          json.push(element);
        }
        break;
      default:
        json.push(element);
      }
    }
    return json;
  },
});
