/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  H,
  parseDecimal,
  modulo,
  Geometry,
  Syntool,
} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';
import {
  colorFromCSSColor,
  colorGetRed,
  colorGetGreen,
  colorGetBlue,
  colorGetAlpha,
} from '../utils/color.js';
import './shapes_import_export.js';

const KML_NS = 'http://www.opengis.net/kml/2.2';
const GX_NS = 'http://www.google.com/kml/ext/2.2';
const ATOM_NS = 'http://www.w3.org/2005/Atom';
const KML_MIME_TYPE = 'application/vnd.google-earth.kml+xml';

function KML(tag) {
  if (tag === 'kml') {
    KML.doc = document.implementation.createDocument(KML_NS, 'kml', null);
    arguments[0] = KML.doc.documentElement;
  }
  return H.apply(KML.doc, arguments);
}

function sign(number) {
  return number < 0 ? -1 : 1;
}
function ensureShortestArc(lonLats) {
  // Ensure that points are always linked along the shortest arc (|Δlon| <= 180)

  let lastLon = NaN;
  let lastShiftedLon = NaN;
  let lonShift = 0;
  return lonLats.map(([lon, lat]) => {
    const lon360 = modulo(lon, 360);
    let shiftedLon;

    if (isNaN(lastShiftedLon)) {
      shiftedLon = lon360;
    } else {
      const lastLon360 = modulo(lastLon, 360);
      const deltaLon = Number((lon360 - lastLon360).toFixed(6));

      if (Math.abs(deltaLon) === 180) {
        // We have to pick one of the two possible arcs:
        //   * (+) positive: increasing longitudes from last point to current
        //   * (-) negative: decreasing longitudes from last point to current
        //
        // if reversed ==  1: we pick the arc corresponding to the sign of Δlon
        // if reversed == -1: we pick the opposite arc
        const reversed = sign(lon) * sign(lastLon);

        shiftedLon = lastShiftedLon + reversed * deltaLon;
        lonShift = shiftedLon - lon360;
      } else {
        const worldsAway = Math.round(deltaLon / 360);
        lonShift -= worldsAway * 360;
        shiftedLon = lon360 + lonShift;
      }
    }
    lastLon = lon;
    lastShiftedLon = shiftedLon;

    return [shiftedLon, lat];
  });
}

function kmlGetNormalStyleFromNode(styleNode) {
  switch (styleNode?.tagName) {
  case 'styleUrl':
    if (styleNode.textContent[0] === '#') {
      if (kmlGetNormalStyleFromNode.visited.has(styleNode.textContent)) {
        console.warn('KML: recursion linked style detected, bailing out');
      } else {
        kmlGetNormalStyleFromNode.visited.add(styleNode.textContent);
        const styleSelector = styleNode.ownerDocument.getElementById(
          styleNode.textContent.substr(1)
        );
        return kmlGetNormalStyleFromNode(styleSelector);
      }
    } else {
      console.warn('KML: only same file "#..." <styleUrl> is supported');
    }
    styleNode = null;
    break;
  case 'StyleMap':
    for (const pair of styleNode.children) {
      if (pair.getElementsByTagName('key')[0]?.textContent === 'normal') {
        const style = pair.getElementsByTagName('Style')[0];
        if (style) {
          styleNode = style;
          break;
        } else {
          return kmlGetNormalStyleFromNode(
            pair.getElementsByTagName('styleUrl')[0]
          );
        }
      }
    }
    styleNode = null;
    break;
  case 'Style':
    break;
  default:
    styleNode = null;
  }
  kmlGetNormalStyleFromNode.visited.clear();
  return styleNode;
}
kmlGetNormalStyleFromNode.visited = new Set();

function kmlColorToCSSColor(color, colorMode) {
  /*eslint-disable no-bitwise*/
  color = parseInt(color, 16) & 0xFFFFFFFF;
  const alpha = color >>> 24;
  color &= 0xFFFFFF;
  if (colorMode === 'random') {
    color = Math.floor(color * Math.random());
  }
  const red = color & 0xFF;
  const green = (color >>> 8) & 0xFF;
  const blue = (color >>> 16) & 0xFF;
  /*eslint-enable no-bitwise*/

  return [
    '#',
    red.toString(16).padStart(2, '0'),
    green.toString(16).padStart(2, '0'),
    blue.toString(16).padStart(2, '0'),
    alpha.toString(16).padStart(2, '0'),
  ].join('');
}
function kmlGetPlacemarkCSSColor(placemark, type) {
  const styleNodes = placemark.querySelectorAll(
    'Placemark > Style, Placemark > StyleMap, Placemark > styleUrl'
  );
  for (const styleNode of styleNodes) {
    const style = kmlGetNormalStyleFromNode(styleNode);
    if (style) {
      const colorStyle = style.querySelector(type);
      if (colorStyle) {
        const color = colorStyle.querySelector('color')?.textContent;
        const colorMode = colorStyle.querySelector('colorMode')?.textContent;
        if (color) {
          return kmlColorToCSSColor(color, colorMode);
        }
      }
    }
  }
  return undefined;
}
function kmlCoordinatesToLonLat(coordinates) {
  return coordinates.trim().split(',').map(parseFloat).slice(0, 2);
}
function kmlLineCoordinatesToLonLats(coordinates) {
  const pointCoordinates = coordinates.trim().split(/\s+/g);
  const lonLats = pointCoordinates.map(kmlCoordinatesToLonLat);
  return ensureShortestArc(lonLats);
}
function kmlRingCoordinatesToLonLats(coordinates) {
  const pointCoordinates = coordinates.trim().split(/\s+/g);
  return pointCoordinates.map(kmlCoordinatesToLonLat);
}
function kmlCoordinatesToPolygon(outerCoordinates, innerCoordinates) {
  if (!outerCoordinates) {
    return null;
  }
  const polygon = Geometry.createPolygonFromXYs(
    kmlRingCoordinatesToLonLats(outerCoordinates)
  );
  if (innerCoordinates) {
    for (const coordinates of innerCoordinates) {
      Geometry.polygonAddHoleFromXYs(
        polygon,
        kmlRingCoordinatesToLonLats(coordinates)
      );
    }
  }
  return polygon;
}
function geometryFromKML(node) {
  switch (node?.tagName) {
  case 'Point': {
    const coordinates = node.querySelector('coordinates')?.textContent;
    if (coordinates) {
      return Geometry.createPointFromXY(kmlCoordinatesToLonLat(coordinates));
    }
    break;
  }
  case 'LineString': {
    const coordinates = node.querySelector('coordinates')?.textContent;
    if (coordinates) {
      return Geometry.createPolylineFromXYs(
        kmlLineCoordinatesToLonLats(coordinates)
      );
    }
    break;
  }
  case 'Polygon': {
    const outerCoordinates = (
      node.querySelector('outerBoundaryIs LinearRing coordinates')
      ?.textContent
    );
    const innerCoordinates = Array.from(
      node.querySelectorAll('innerBoundaryIs LinearRing coordinates'),
      (n) => n.textContent
    );
    return kmlCoordinatesToPolygon(outerCoordinates, innerCoordinates);
  }
  case 'MultiGeometry':
    return Geometry.createCollection(
      Array.from(node.children, geometryFromKML).filter(Boolean)
    );
  default:
  }
  return null;
}
function isTextPlacemark(placemark) {
  if (
    !placemark.querySelector('name')?.textContent ||
    !placemark.querySelector('Placemark > Point') ||
    placemark.getElementsByTagNameNS(ATOM_NS, 'author')[0]
  ) {
    return false;
  }

  const color = kmlGetPlacemarkCSSColor(placemark, 'IconStyle');
  return (color && parseInt(color.substr(-2), 16) < 26 /* 1% opacity */);
}

function kmlColorFromCSSColor(cssColor) {
  const color = colorFromCSSColor(cssColor);
  return [
    (colorGetAlpha(color) * 255).toString(16).padStart(2, '0'),
    colorGetBlue(color).toString(16).padStart(2, '0'),
    colorGetGreen(color).toString(16).padStart(2, '0'),
    colorGetRed(color).toString(16).padStart(2, '0'),
  ].join('');
}
function kmlInlineStyleFromColor(color) {
  return KML('Style', [
    KML('IconStyle', KML('color', color)),
    KML('LineStyle', KML('color', color)),
    KML('PolyStyle', KML('color', '66' + color.substr(2))),
  ]);
}
function kmlCoordinatesFromLonLat([lon, lat]) {
  return `${Number(lon.toFixed(6))},${Number(lat.toFixed(6))}`;
}
function kmlLineCoordinatesFromLonLats(lonLats, wrapDateLine) {
  let lastLonLat = lonLats[0];
  let lastKmlLonLat = lonLats[0];
  let coords = kmlCoordinatesFromLonLat(lastLonLat);
  for (let i = 1; i < lonLats.length; i++) {
    let lonLat = lonLats[i];
    let deltaLon = lonLat[0] - lastLonLat[0];

    // KML points will always be linked along the shortest arc (|Δlon| <= 180)
    // So we should split long arcs into shorter ones
    // But only if the projection requires date line wrapping
    if (wrapDateLine && Math.abs(deltaLon) > 180) {
      const deltaLat = lonLat[1] - lastLonLat[1];
      lonLat = [
        lastLonLat[0] + Math.sign(deltaLon) * 180,
        lastLonLat[1] + deltaLat / Math.abs(deltaLon) * 180,
      ];
      deltaLon = Math.sign(deltaLon) * 180;
      i--;
    }

    let kmlLonLat = lonLat;
    if (Math.abs(deltaLon) === 180) {
      // Pick the correct 180° arc from two possibilities:
      //   * (+) positive: increasing longitudes from last point to current
      //   * (-) negative: decreasing longitudes from last point to current
      //
      // if reversed ==  1: we pick the arc corresponding to the sign of Δlon
      // if reversed == -1: we pick the opposite arc
      const reversed = sign(lonLat[0]) * sign(lastKmlLonLat[0]);
      const lon360 = modulo(lonLat[0], 360);
      const lastLon360 = modulo(lastKmlLonLat[0], 360);
      const kmlSign = reversed * Math.sign(lon360 - lastLon360);
      if (kmlSign !== Math.sign(deltaLon)) {
        // flip the sign of current longitude (lonLat[0])
        kmlLonLat = [
          lonLat[0] - 360 * (2 * Math.trunc(lonLat[0] / 360) + sign(lonLat[0])),
          lonLat[1],
        ];
      }
    }

    coords += ` ${kmlCoordinatesFromLonLat(kmlLonLat)}`;
    lastLonLat = lonLat;
    lastKmlLonLat = kmlLonLat;
  }
  return [
    KML('coordinates', coords),
    KML('tessellate', 1),
  ];
}
function kmlRingCoordinatesFromLonLats(lonLats, wrapDateLine) {
  if (!wrapDateLine) {
    lonLats = ensureShortestArc(lonLats);
  }
  return [
    KML('coordinates', lonLats.map(kmlCoordinatesFromLonLat).join(' ')),
    KML('tessellate', 1),
  ];
}
function kmlInnerBoundaryFromLonLats(lonLats) {
  const mapView = this;
  const wrapDateLine = mapView.shouldWrapDateLine();

  return KML('innerBoundaryIs',
    KML('LinearRing', kmlRingCoordinatesFromLonLats(lonLats, wrapDateLine))
  );
}
function kmlGeometryFromGeometry(geometry) {
  const mapView = this;
  const wrapDateLine = mapView.shouldWrapDateLine();

  if (Geometry.isPoint(geometry)) {
    return KML('Point',
      KML('coordinates', kmlCoordinatesFromLonLat(Geometry.pointToXY(geometry)))
    );
  }
  if (Geometry.isPolyline(geometry)) {
    return KML('LineString',
      kmlLineCoordinatesFromLonLats(
        Geometry.polylineToXYs(geometry),
        wrapDateLine
      )
    );
  }
  if (Geometry.isPolygon(geometry)) {
    geometry = Geometry.rewind(geometry, false /*counterclockwise*/);
    const outerLonLats = Geometry.polygonToXYs(geometry);
    const innerLonLats = Geometry.polygonHolesToXYs(geometry);
    return KML('Polygon', [
      KML('outerBoundaryIs',
        KML('LinearRing',
          kmlRingCoordinatesFromLonLats(outerLonLats, wrapDateLine)
        )
      ),
      ...innerLonLats.map(kmlInnerBoundaryFromLonLats, mapView),
    ]);
  }
  if (Geometry.isCollection(geometry)) {
    return KML('MultiGeometry',
      Geometry.getCollectionComponents(geometry)
      .map(kmlGeometryFromGeometry, mapView)
    );
  }
  return null;
}
function kmlGeometryFromShapeJSON(shape) {
  const mapView = this;
  const wrapDateLine = mapView.shouldWrapDateLine();

  switch (shape.type) {
  case Shape.LINE:
  case Shape.ARROW:
    return KML('LineString',
      kmlLineCoordinatesFromLonLats(shape.points, wrapDateLine)
    );
  case Shape.TEXT:
    return KML([
      KML('Point', KML('coordinates', kmlCoordinatesFromLonLat(shape.location))),
      KML('styleUrl', '#syntoolTextStyle'),
    ]);
  case Shape.WKT:
    return KML([
      shape.author ? KML('atom:author', KML('atom:name', shape.author)) : null,
      kmlGeometryFromGeometry.call(mapView, Geometry.fromWKT(shape.wkt)),
    ]);
  default:
    return null;
  }
}
function kmlFromShape(shape) {
  const json = Shape.toExternalJSON(shape);
  const geometryNode = kmlGeometryFromShapeJSON.call(this, json);

  if (!geometryNode) {
    return null;
  }

  const timeSpanChildren = [];
  if (json.start != null) {
    timeSpanChildren.push(KML('begin', shape.start.toISOString()));
  }
  if (json.end != null) {
    timeSpanChildren.push(KML('end', shape.end.toISOString()));
  }

  return KML('Placemark', {id: shape.id}, [
    KML('name', shape.text ? shape.text : shape.id),
    timeSpanChildren.length > 0 ? KML('TimeSpan', timeSpanChildren) : null,
    KML('styleUrl', shape.text ? '#syntoolStyle' : '#syntoolNoLabelStyle'),
    shape.color ? kmlInlineStyleFromColor(kmlColorFromCSSColor(shape.color)) : null,
    geometryNode,
  ]);
}

Syntool.extend({
  [Syntool.NAME]: 'ShapesImportExportKML',
  [Syntool.CONSTRUCTOR]() {
    this.acceptedImportFormats.push('KML (partial support)');
    this.acceptedImportFileTypes.push('.kml', KML_MIME_TYPE);
    this.importersFromString.push((string, mimeType, options) => (
      this._importShapesFromStringAsKML(string, mimeType, options)
    ));

    this.shapesExporters.set('kml', {
      label: (
        'KML <small style="font-weight:normal;font-style:italic;">(partial ' +
        'support)</small>'
      ),
      canExportShapeOfType(type) {
        switch (type) {
        case Shape.TEXT:
        case Shape.LINE:
        case Shape.ARROW:
        case Shape.WKT:
          return true;
        default:
          return false;
        }
      },
      export: (shapes, options) => {
        options = Object.assign({
          label   : 'shape' + (shapes.length === 1 ? '' : 's') + ' as KML',
          fileName: 'syntool-shapes-' + (new Date().toISOString()) + '.kml',
          fileType: KML_MIME_TYPE,
        }, options);
        if (options.toClipboard) {
          options.fileName = null;
        }
        this.exportText(this.exportShapesToKML(shapes), options);
      },
    });
  },
  exportShapesToKML(shapes) {
    const kml = KML(KML('kml', {
      'xmlns:gx': GX_NS,
      'xmlns:atom': ATOM_NS,
    }), KML('Document', [
      KML('name', 'Syntool shapes ' + new Date().toISOString()),
      KML('Style', {id: 'syntoolStyle'}, [
        KML('LineStyle', KML('gx:labelVisibility', 1)),
        KML('PolyStyle', KML('color', '66dadada')),
      ]),
      KML('Style', {id: 'syntoolNoLabelStyle'}, [
        KML('LabelStyle', KML('color', '00000000')),
        KML('PolyStyle', KML('color', '66dadada')),
      ]),
      KML('Style', {id: 'syntoolTextStyle'},
        KML('IconStyle', KML('color', '01FFFFFF'))
      ),
      ...shapes.map(kmlFromShape, this.mapView),
    ]));
    return `<?xml version="1.0" encoding="UTF-8"?>\n${kml.outerHTML}`;
  },

  _importShapesFromStringAsKML(string, mimeType, options) {
    if (mimeType && mimeType !== KML_MIME_TYPE) {
      return false;
    }

    const parser = new DOMParser();
    const kml = parser.parseFromString(string, 'application/xml');

    if (kml.documentElement.tagName === 'parsererror') {
      console.warn(kml.documentElement.textContent);
      return false;
    }

    const json = [];
    for (const placemark of kml.querySelectorAll('Placemark')) {
      const shape = {};

      const geometryNode = placemark.querySelector([
        'Placemark > Point',
        'Placemark > LineString',
        'Placemark > Polygon',
        'Placemark > MultiGeometry',
      ].join(','));

      if (isTextPlacemark(placemark)) {
        shape.type = Shape.TEXT;
        shape.location = Geometry.pointToXY(geometryFromKML(geometryNode));
      } else if (geometryNode) {
        switch (geometryNode.tagName) {
        case 'LineString': {
          const coordinates = (
            geometryNode.querySelector('coordinates')?.textContent
          );
          if (coordinates) {
            shape.type = Shape.LINE;
            shape.points = kmlLineCoordinatesToLonLats(coordinates);
          }
          break;
        }
        case 'Point':
        case 'Polygon':
        case 'MultiGeometry': {
          const geometry = geometryFromKML(geometryNode);
          if (geometry) {
            shape.type = Shape.WKT;
            shape.wkt = Geometry.toWKT(geometry);

            const author = (
              placemark
                .getElementsByTagNameNS(ATOM_NS, 'author')[0]
                ?.getElementsByTagNameNS(ATOM_NS, 'name')?.[0]
                ?.textContent
            );
            if (author) {
              shape.author = author;
            }
          }
          break;
        }
        default:
          break;
        }
      }
      if (!shape.type) {
        continue;
      }

      if (placemark.id) {
        shape.id = placemark.id;
      }

      const name = placemark.querySelector('name')?.textContent;
      if (name && name !== placemark.id) {
        shape.text = name;
      }

      const begin = placemark.querySelector('TimeSpan begin')?.textContent;
      const end = placemark.querySelector('TimeSpan end')?.textContent;

      if (begin || end) {
        if (begin) {
          shape.start = new Date(begin).getTime();
        }
        if (end) {
          shape.end = new Date(end).getTime();
        }
      } else {
        const when = placemark.querySelector('TimeStamp when')?.textContent;
        if (when) {
          if (when.length < 7) {
            // gYear (YYYY)
            shape.start = Date.UTC(parseDecimal(when));
            shape.end = Date.UTC(parseDecimal(when) + 1) - 1000;
          } else if (when.length < 10) {
            // gYearMonth (YYYY-MM)
            const [year, month] = when.split('-').map(parseDecimal);
            shape.start = Date.UTC(year, month - 1);
            shape.end = Date.UTC(year, month) - 1000;
          } else if (when.length < 15) {
            // date (YYYY-MM-DD)
            const [year, month, day] = when.split('-').map(parseDecimal);
            shape.start = Date.UTC(year, month - 1, day);
            shape.end = Date.UTC(year, month - 1, day + 1) - 1000;
          } else {
            // dateTime (YYYY-MM-DDThh:mm:ssZ)
            // dateTime (YYYY-MM-DDThh:mm:sszzzzzz)
            shape.start = (new Date(when)).getTime();
            shape.end = shape.start + 1000;
          }
        }
      }
      if (shape.type !== Shape.TEXT) {
        shape.color = kmlGetPlacemarkCSSColor(placemark, 'LineStyle, IconStyle');
      }
      json.push(shape);
    }
    return this.importFromJSON(json, options);
  },
});
