/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {HOUR_IN_MILLISECONDS, Config, Syntool, Types} from '../syntool.js';

Object.assign(Config.configFieldValidators, {
  animationSpeed: Types.number.gt(0),
});
Config.configNormalizers.push((config) => Object.assign({
  animationSpeed: 2.5 * HOUR_IN_MILLISECONDS /* 2.5 hours / s */,
}, config));
Syntool.extend({
  [Syntool.NAME]: 'TimelineAnimation',

  timelineAnimationTimeFactor: 1,

  [Syntool.INIT]() {
    this.timelineAnimationBaseStepLength = this.config.animationSpeed / 5;
    this._timelineAnimationInterval = null;
  },
  [Syntool.STATE]: {
    get() {
      const {timelineAnimationBaseStepLength, timelineAnimating} = this;
      return {timelineAnimationBaseStepLength, timelineAnimating};
    },
    set({
      timelineAnimationBaseStepLength = 4 * HOUR_IN_MILLISECONDS,
      timelineAnimating,
    }) {
      this.timelineAnimationBaseStepLength = timelineAnimationBaseStepLength;
      this.timelineAnimating = timelineAnimating;
    },
  },

  get timelineAnimationBaseStepLength() {
    if (!this._timelineAnimationBaseStepLength) {
      this._timelineAnimationBaseStepLength = 4 * HOUR_IN_MILLISECONDS;
    }
    return this._timelineAnimationBaseStepLength;
  },
  set timelineAnimationBaseStepLength(newValue) {
    if (isNaN(newValue) || this._timelineAnimationBaseStepLength === newValue) {
      return;
    }

    this._timelineAnimationBaseStepLength = newValue;
    this.hub.emit('change:timelineAnimationBaseStepLength');
    this._invalidateState();
  },

  get timelineAnimating() {
    return (this._timelineAnimationInterval !== null);
  },
  set timelineAnimating(animating) {
    if (animating) {
      this.startAnimatingTimeline();
    } else {
      this.stopAnimatingTimeline();
    }
  },
  isAnimatingTimeline() {
    return this.timelineAnimating;
  },
  startAnimatingTimeline() {
    if (!this.timelineAnimating) {
      this._timelineAnimationInterval = (
        window.setInterval(this._animateTimeline.bind(this), 200)
      );
      this.hub.emit('change:timelineAnimating');
      this._invalidateState();
    }
  },
  stopAnimatingTimeline() {
    if (this.timelineAnimating) {
      window.clearInterval(this._timelineAnimationInterval);
      this._timelineAnimationInterval = null;
      this.hub.emit('change:timelineAnimating');
      this._invalidateState();
    }
  },
  toggleAnimatingTimeline() {
    this.timelineAnimating = !this.timelineAnimating;
  },

  _advanceCurrentDateBy(ms) {
    if (ms === 0) {
      return;
    }
    this.timelineStore.currentTime += ms;
  },
  _animateTimeline() {
    this._advanceCurrentDateBy(
      this.timelineAnimationBaseStepLength *
      this.timelineAnimationTimeFactor
    );
  },
});
