/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
// MapView Extensions
import './map_view_extensions/map_view_dataset_selection.js';
import './map_view_extensions/map_view_clickable_features.js';
import './map_view_extensions/map_view_dataset_outlines.js';
import './map_view_extensions/map_view_dataset_picking.js';
import './map_view_extensions/map_view_measuring_ruler.js';
import './map_view_extensions/map_view_navigation.js';
import './map_view_extensions/map_view_pixel_interpolation.js';

// Datasets
import './datasets/geojson_dataset.js';
import './datasets/image_dataset.js';
import './datasets/layer_dataset.js';
import './datasets/time_layer_dataset.js';
import './datasets/moored_dataset.js';
import './datasets/ncwms_dataset.js';
import './datasets/shape_dataset.js';
import './datasets/streamlines_dataset.js';
import './datasets/trajectories_dataset.js';
import './datasets/vector_field_dataset.js';
import './datasets/zxy_dataset.js';

// Optional features
import './features/abort_loading_tiles.js';
import './features/backup.js';
import './features/basic_shortcuts.js';
import './features/catalog_store.js';
import './features/changelog.js';
import './features/global_error_handler.js';
import './features/goto_nearest.js';
import './features/auto_goto_nearest.js';
import './features/import.js';
import './features/export.js';
import './features/localforage_user_storage.js';
import './features/timeline_animation.js';
import {getStateFromURL, getStateURL} from './features/url_state.js';
import {FlagsStore} from './features/flags_store.js';
import './features/hotspots.js';
import './features/monitor_loading_tiles.js';
import './features/auto_import.js';
import './features/shapes_editing.js';
import './features/shapes_drawing.js';
import './features/shapes_import_export.js';
import './features/shapes_import_export_kml.js';
import './features/shapes_import_export_geojson.js';
import './features/shapes_shortcuts.js';
import {patchOLGoogleLayer} from './features/ol_google_patch.js';

// Shape types
import './shape_types/arrow_shape.js';
import './shape_types/ellipse_shape.js';
import './shape_types/line_shape.js';
import './shape_types/mushroom_shape.js';
import './shape_types/text_shape.js';
import './shape_types/wkt_shape.js';

// Processing tools
import {ProcessingToolsPanel} from './processing_tools/processing_tools.js';
import './processing_tools/basic_processing_tool.js';

import {
  NONPASSIVE,
  getNowMs,
  appendToIndexPath,
  scheduleTask,
  deepAssign,
  H,
  HIDMouse,
  HIDModifiers,
  whenReady,
  Syntool,
  throttleOncePerEventLoop,
  TypeCheckerError,
} from './syntool.js';
import {AttributionControl} from './ui/components/attribution_control.js';
import {createBackgroundLayerPicker} from './ui/components/background_layer_picker.js';
import {ChangelogDialog} from './ui/components/changelog_dialog.js';
import {SelectedDatasetsOpacityControl} from './ui/components/selected_datasets_opacity_control.js';
import {DateTimeControl} from './ui/components/date_time_control.js';
import {HotspotsDialog} from './ui/components/hotspots_dialog.js';
import {LatLonGridControl} from './ui/components/lat_lon_grid_control.js';
import {NotificationsControl} from './ui/components/notifications_control.js';
import {ProductsDisplaySettingsDialog} from './ui/components/products_display_settings_dialog.js';
import {ProductsPanel} from './ui/components/products_panel.js';
import {ProgressBar} from './ui/components/progress_bar.js';
import {ScaleLineControl} from './ui/components/scale_line_control.js';
import {SelectionInfoPanel} from './ui/components/selection_info_panel.js';
import {SelectionMetadataUI} from './ui/components/selection_metadata_ui.js';
import {SelectionShapeUI} from './ui/components/selection_shape_ui.js';
import {ShapesToolbarControl} from './ui/components/shapes_toolbar_control.js';
import {ShortcutsDialog} from './ui/components/shortcuts_dialog.js';
import {TimelinePanel} from './ui/components/timeline_panel.js';
import {TitleBar} from './ui/components/title_bar.js';
import {TourControl} from './ui/components/tour_control.js';
import {ValueUnderCursorControl} from './ui/components/value_under_cursor_control.js';
import {ZoomControl} from './ui/components/zoom_control.js';
import {AboutDialog} from './ui/about.js';
import * as UIUtils from './ui/utils.js';
import {showPermalink} from './ui/utils.js';

import './ui/import_export.js';
import './ui/products_settings_list_view_shapes_patch.js';
import './ui/screenshot_mode.js';
import './deprecated.js';
import {decodeEmail} from './utils/email.js';

window.ODL = Object.assign(window.ODL || {}, UIUtils);

if (window.google) {
  patchOLGoogleLayer(window.google);
}

whenReady(function() {
  // Specify a canonical URL for search engines like Google
  // See: https://developers.google.com/search/docs/crawling-indexing/consolidate-duplicate-urls
  document.head.appendChild(
    H('link', {rel: 'canonical', href: new URL('.', location).href})
  );

  // Prevent pinch-to-zoom on iOS 10+ Safari (which stopped respecting user-scalable=no)
  document.addEventListener('touchmove', function(event) {
    if (!isNaN(event.scale) && event.scale !== 1) {
      event.preventDefault();
    }
  }, NONPASSIVE);

  // Prevent double-tap-to-zoom on iOS 10+ Safari (which stopped respecting user-scalable=no)
  var lastTouchEnd = 0;
  document.addEventListener('touchend', function(event) {
    var now = getNowMs();
    if (now - lastTouchEnd <= 300) {
      event.preventDefault();
    }
    lastTouchEnd = now;
  }, NONPASSIVE);
});
whenReady(function() {
  Syntool.backupStoresMap.timeline = Syntool.timelineStore;
  Syntool.backupStoresMap.products = Syntool.productsStore;
  Syntool.backupStoresMap.shortcuts = Syntool.shortcutsStore;
  Syntool.backupStoresMap.notifications = Syntool.notificationCenter.store;
  Syntool.backupStoresMap.hotspots = Syntool.hotspotsStore;
  Syntool.backupStoresMap.shapes = Syntool.shapesStore;
});
whenReady(function() {
  Syntool.addControls({titleBar: new TitleBar()});

  const {titleBar} = Syntool.controls;
  titleBar.addGroup('map');
  titleBar.addGroup('state');
  titleBar.addGroup('tools');
  titleBar.addGroup('misc');

  // Ensure all DOMContentLoaded listeners have been called before loading
  // and validating the config and initializing the app.
  scheduleTask(async function() {
    try {
      await Syntool.loadConfig(window.config());
    } catch (error) {
      if (error instanceof TypeCheckerError) {
        // eslint-disable-next-line require-atomic-updates
        window[error.name] = error.value;
        let access = appendToIndexPath('', error.name);
        if (access[0] === '[') {
          access = `window${access}`;
        }
        console.log(
          'The config has been added on the global scope.\n' +
          'You can use `%s` to access it.',
          access
        );
      }
      console.error(error);
      // eslint-disable-next-line no-alert
      alert('Error loading config.\n\nSee the JavaScript console for more details.');
      return;
    }

    await FlagsStore.migrateCookies();
    await FlagsStore.load();
    Syntool.init();
  });
});

Syntool.whenInitialised(function() {
  if (Syntool.mapView.hasMercatorLikeProjection()) {
    Syntool.addControls({latLonGrid: new LatLonGridControl()});
  }

  const aboutDialogNode = document.getElementById('AboutDialogContent');
  if (aboutDialogNode) {
    aboutDialogNode.remove();
    aboutDialogNode.hidden = false;
  }
  const tourContentNode = document.getElementById('TourContent');
  tourContentNode?.remove();

  const selectionInfo = new SelectionInfoPanel();
  Syntool.addControls({
    progressBar: new ProgressBar(),
    products: new ProductsPanel(),
    attribution: new AttributionControl(),
    dateTime: new DateTimeControl(),
    displaySettings: new ProductsDisplaySettingsDialog(),
    hotspotsDialog: new HotspotsDialog(),
    notifications: new NotificationsControl(),
    processingTools: new ProcessingToolsPanel(),
    scaleLine: new ScaleLineControl(),
    selectedDatasetsOpacity: new SelectedDatasetsOpacityControl(),
    selectionInfo,
    selectionMetadataUI: new SelectionMetadataUI({selectionInfo}),
    selectionShapeUI: new SelectionShapeUI({selectionInfo}),
    shapesToolbar: new ShapesToolbarControl(),
    shortcutsDialog: new ShortcutsDialog(),
    timeline: new TimelinePanel(),
    valueUnderCursor: new ValueUnderCursorControl(),
    zoom: new ZoomControl(),
    aboutDialog: new AboutDialog('about-dialog', {
      storageShowKey: 'showAbout',
      title: 'About',
      content: aboutDialogNode,
    }),
    tour: new TourControl(tourContentNode.querySelectorAll('syntool-tour-step')),
  });

  // Forget old unused "showTour" flag
  FlagsStore.delete('showTour');

  const {
    node: backgroundLayerSelect,
    setLayerName,
  } = createBackgroundLayerPicker(
    Syntool.config.baseLayers,
    Syntool.mapView.getBackgroundLayerName(),
    (layerName) => Syntool.mapView.setBackgroundLayerByName(layerName)
  );
  Syntool.mapView.hub.on(
    'change:background-layer',
    () => setLayerName(Syntool.mapView.getBackgroundLayerName())
  );

  const {titleBar} = Syntool.controls;
  const displaySettingsTooltip = (
    'Colorbar, transparency, stacking order and filtering for each product'
  );
  let displaySettingsButton;

  titleBar.addButton('map', H('div', [
    H('i', {class: '-icon fa fa-globe'}),
    backgroundLayerSelect,
  ]));
  titleBar.addButton('map', displaySettingsButton = {
    id: 'settingsButton',
    icon: 'svg svg-layers',
    class: '-warning-indicator',
    label: Syntool.controls.displaySettings.props.title,
    description: displaySettingsTooltip,
    onClick() {
      Syntool.controls.displaySettings.toggle();
    },
  });

  const handleFilteredProductsChange = throttleOncePerEventLoop(() => {
    const patternByProductId = (
      Syntool.productsStore.getComputedPatternsForProductIds(
        Syntool.productsStore.selectedProductIds()
      )
    );
    const hasFilteredProducts = Object.values(patternByProductId).some(Boolean);

    displaySettingsButton.class = (
      '-warning-indicator' + (hasFilteredProducts ? ' -with-indicator' : '')
    );
    displaySettingsButton.description = displaySettingsTooltip + (
      hasFilteredProducts
      ? '\n⚠️ Some data may be hidden by the applied filters'
      : ''
    );

    const button = document.getElementById(displaySettingsButton.id);
    if (button) {
      button.classList.toggle('-with-indicator', hasFilteredProducts);
      button.title = displaySettingsButton.description;
    }
  });
  Syntool.productsStore.hub
    .on('change:selection', handleFilteredProductsChange)
    .on('change:pattern', handleFilteredProductsChange);
  handleFilteredProductsChange();

  titleBar.addButton('state', {
    id: 'hotspotsButton',
    icon: 'fa fa-bookmark',
    label: 'Hotspots',
    description: 'Saved case studies',
    onClick() {
      Syntool.controls.hotspotsDialog.toggle();
    },
  });
  titleBar.addButton('state', {
    id: 'permalinkButton',
    icon: 'fa fa-link',
    label: 'Share',
    description: 'Create a short URL from the current view',
    onClick() {
      showPermalink(getStateURL(), Syntool);
    },
  });

  var feedbackDialogNode = document.getElementById('FeedbackDialogContent');
  if (feedbackDialogNode) {
    feedbackDialogNode.remove();

    const items = [];
    for (const link of feedbackDialogNode.querySelectorAll('a[href]')) {
      const url = new URL(link.href);
      if (url.protocol === 'mailto:') {
        const email = decodeEmail(url.pathname);
        /**
         * A mailto URL has an [opaque path][1], so its pathname can't be changed.
         * Thus we need to reconstruct its href but with the decoded pathname.
         * [1]: https://url.spec.whatwg.org/#url-opaque-path
         */
        link.href = url.protocol + email + url.search;
      }
      items.push({
        label: link.textContent,
        description: link.title || undefined,
        onClick() {
          const w = window.open(
            link.href,
            url.protocol === 'mailto:' ? '_self' : '_blank'
          );
          if (w) {
            w.opener = null;
          }
        },
      });
    }

    titleBar.addButton('misc', {
      id: 'buttonFeedback',
      icon: 'fa fa-envelope',
      label: 'Feedback',
      description: 'Send feedback',
      items,
    });
  }

  titleBar.addButton('misc', {
    id: 'buttonHelp',
    icon: 'fa fa-question',
    label: 'Help',
    items: [
      {
        label: '👀 Quick Tour',
        description: 'Start a quick guided tour of the interface',
        onClick() {
          Syntool.controls.tour.start();
        },
      },
      {
        label: '⌨️ Keyboard Shortcuts',
        description: 'See all available keyboard shortcuts',
        onClick() {
          Syntool.controls.shortcutsDialog.show();
        },
      },
      {
        label: '🏠 About',
        onClick() {
          Syntool.controls.aboutDialog.show();
        },
      },
    ],
  });

  if (Syntool.changelog) {
    Promise.all(
      [Syntool.changelog.accessible, Syntool.changelog.updatedSinceLastViewed]
    )
    .then(function([accessible, updatedSinceLastViewed]) {
      if (accessible) {
        Syntool.addControls({
          changelogDialog: new ChangelogDialog(Syntool.changelog),
        });
        Syntool.controls.titleBar.addButton('misc', {
          id: 'showChangelogButton',
          class: updatedSinceLastViewed ? ' -active' : '',
          icon: 'fa fa-bell',
          description: "What's new!",
          onClick: function() {
            this.classList.remove('-active');
            Syntool.controls.changelogDialog.toggle();
          },
        });
      }
    });
  }

  [
    Syntool.controls.titleBar,
    Syntool.controls.hotspotsDialog,
    Syntool.controls.processingTools,
    Syntool.controls.selectedDatasetsOpacity,
    Syntool.controls.shapesToolbar,
    Syntool.controls.shortcutsDialog,
    Syntool.controls.zoom,
    Syntool.controls.aboutDialog,
    Syntool.controls.feedbackDialog,
    ...Syntool.controls.tour.steps,
  ].forEach(
    (control) => control?.node.classList.add('h-hidden-in-screenshot-mode')
  );

  Syntool.controls.displaySettings.hub
    .on('hide', function() {
      FlagsStore.set('displaySettingsDialogOpen', false);
    })
    .on('show', function() {
      FlagsStore.set('displaySettingsDialogOpen', true);
    })
    .on('click:add', function() {
      Syntool.controls.products.focusOnSearch();
    });

  var displaySettingsDialogOpen = (
    FlagsStore.has('displaySettingsDialogOpen')
    ? FlagsStore.get('displaySettingsDialogOpen')
    : true
  );
  if (displaySettingsDialogOpen && window.innerWidth >= 500) {
    Syntool.controls.displaySettings.show();
  }

  Syntool.controls.selectionInfo.view.hub.on('render', function() {
    Syntool.controls.processingTools.hidden = (
      !Syntool.controls.selectionInfo.view.isEmpty
    );
  });

  function addProcessingToolsButton() {
    Syntool.controls.titleBar.addButton('tools', {
      id: 'showProcessingToolsButton',
      icon: 'fa fa-gears',
      description: 'Show/hide the processing tools panel',
      onClick() {
        Syntool.executeShortcutWithId('toggle-processing-tools');
      },
    });
    Syntool.controls.processingTools.hub.on('change:active', function() {
      Syntool.mapView.unselectAllDatasets();

      var button = document.getElementById('showProcessingToolsButton');
      button.classList.toggle('-active', Syntool.controls.processingTools.active);
    });
  }
  if (Syntool.controls.processingTools.canActivate()) {
    addProcessingToolsButton();
  } else {
    Syntool.controls.processingTools.hub.once(
      'change:selection',
      addProcessingToolsButton
    );
  }

  Syntool.shortcutsStore.add(
    'Products selection (in products list and display data)',
    [
      {
        id         : 'products-select',
        keys       : HIDMouse.left,
        description: 'Select/deselect the clicked product',
      },
      {
        id         : 'products-select-one',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDMouse.left,
        description: 'Only select the clicked product (and deselect all others)',
      },
    ]
  );

  Syntool.controls.aboutDialog.showIfNeeded();

  // Remove splashscreen
  var splashscreen = document.getElementById('splashscreen');
  if (splashscreen) {
    splashscreen.addEventListener('animationend', function(event) {
      if (event.animationName === 'splashscreen-hide') {
        this.remove();
      }
    });
    splashscreen.classList.add('hidden');
  }
});

Syntool.hub.once('beforeinit', prepareAppState);

function prepareAppState() {
  function getDefaultTimeframe(timeRanges) {
    for (var i = 0; i < timeRanges.length; i++) {
      var range = timeRanges[i][1];
      var isDefault = !!timeRanges[i][2];

      if (isDefault) {
        return range;
      }
    }
    return timeRanges.length > 0 ? timeRanges[0][1] : '1d';
  }

  var config = Syntool.config;
  const appState = deepAssign({}, Syntool.state, {
    timeline: {
      currentTime: config.currentDate?.getTime(),
      timeframe: getDefaultTimeframe(config.timeRanges),
    },
    mapView: {extent: undefined, center: undefined, zoom: config.defaultZoom},
    catalogSelection: undefined,
  }, getStateFromURL());

  if (!appState.mapView.extent && !isFinite(appState.mapView.zoom)) {
    appState.mapView.extent = config.maxExtent;
  }
  if (isNaN(appState.timeline.currentTime)) {
    Syntool.hub.once('init', function() {
      Syntool.executeShortcutWithId('jump-to-nearest-dataset');
    });
  }
  Syntool.state = appState;
}
