/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
// MapView Extensions
import './map_view_extensions/map_view_clickable_features.js';
import './map_view_extensions/map_view_dataset_outlines.js';
import './map_view_extensions/map_view_pixel_interpolation.js';

// Datasets
import './datasets/geojson_dataset.js';
import './datasets/image_dataset.js';
import './datasets/layer_dataset.js';
import './datasets/time_layer_dataset.js';
import './datasets/moored_dataset.js';
import './datasets/ncwms_dataset.js';
import {ShapeDataset} from './datasets/shape_dataset.js';
import './datasets/streamlines_dataset.js';
import './datasets/trajectories_dataset.js';
import './datasets/vector_field_dataset.js';
import './datasets/zxy_dataset.js';

// Optional features
import './features/abort_loading_tiles.js';
import './features/monitor_loading_tiles.js';
// import './features/goto_nearest.js';
import {patchOLGoogleLayer} from './features/ol_google_patch.js';

// Shape types
import './shape_types/arrow_shape.js';
import './shape_types/ellipse_shape.js';
import './shape_types/line_shape.js';
import './shape_types/mushroom_shape.js';
import './shape_types/text_shape.js';
import './shape_types/wkt_shape.js';

import {
  scheduleTask,
  deepAssign,
  ajaxHub,
  ajaxStatistics,
  whenReady,
  Syntool,
} from './syntool.js';
import {DateTimeControl} from './ui/components/date_time_control.js';
import {LatLonGridControl} from './ui/components/lat_lon_grid_control.js';
import {ScaleLineControl} from './ui/components/scale_line_control.js';
import {appStateFromURL} from './utils/app_state.js';

Syntool.extend({
  [Syntool.NAME]: 'NoCoverage',
  [Syntool.CONSTRUCTOR]() {
    this.updateCoverage = () => Promise.resolve(false);
  },
});

if (window.google) {
  patchOLGoogleLayer(window.google);
}

let resolveReadyPromise;
let readyPromiseResolved;
let readyCounter = 0;
notReadyYet();

function notReadyYet() {
  if (window.__SEASHOT_READY__ && !readyPromiseResolved) {
    return;
  }

  let resolveReady;
  window.__SEASHOT_READY__ = new Promise((resolve) => (resolveReady = resolve));
  readyPromiseResolved = false;
  resolveReadyPromise = () => {
    resolveReady(readyCounter++);
    readyPromiseResolved = true;
  };
}
function signalEndOfActivities() {
  console.log('End of activities');
  resolveReadyPromise();
}

whenReady(function() {
  // Ensure all DOMContentLoaded listeners have been called before loading
  // and validating the config and initializing the app.
  scheduleTask(function() {
    Syntool.loadConfig(window.config())
      .then(() => {
        Syntool.config.maxDatasetsWeight = Infinity;
      })
      .then(() => Syntool.init());
  });
});

Syntool.whenInitialised(function() {
  if (Syntool.mapView.hasMercatorLikeProjection()) {
    Syntool.addControls({latLonGrid: new LatLonGridControl()});
  }
  Syntool.addControls({
    dateTime: new DateTimeControl(),
    scaleLine: new ScaleLineControl(),
  });

  const datasetsWithActivity = new Set();
  let requestId;

  ajaxHub
    .on('success', updateActivity)
    .on('fail', updateActivity)
    .on('cancel', updateActivity)
    .on('pre-send', updateActivity);
  Syntool.mapView.hub
    .on('change:update-datasets-progress', updateActivity)
    .on('dataset:did-add-to-map', didAddDatasetToMap)
    .on('dataset:did-remove-from-map', didRemoveDatasetFromMap);

  function didAddDatasetToMap(dataset) {
    dataset.hub
      .on('data-load:start', addDatasetToActivity, dataset)
      .on('data-load:end', removeDatasetFromActivity, dataset);
  }
  function didRemoveDatasetFromMap(dataset) {
    dataset.hub
      .off('data-load:start', addDatasetToActivity, dataset)
      .off('data-load:end', removeDatasetFromActivity, dataset);
    removeDatasetFromActivity(dataset);
  }

  function addDatasetToActivity(dataset = this) {
    datasetsWithActivity.add(dataset);
    updateActivity();
  }
  function removeDatasetFromActivity(dataset = this) {
    datasetsWithActivity.delete(dataset);
    updateActivity();
  }

  function updateActivity() {
    clearTimeout(requestId);
    notReadyYet();
    if (
      ajaxStatistics.liveRequestCount === 0 &&
      Syntool.mapView.updateDatasetsProgress === 1 &&
      datasetsWithActivity.size === 0
    ) {
      requestId = setTimeout(signalEndOfActivities, 100);
    }
  }
});

Syntool.hub.once('beforeinit', prepareAppState);

async function prepareAppState() {
  const seashotConfig = await getSeashotConfig();
  if (seashotConfig.userShapes && seashotConfig.userShapes.length > 0) {
    Syntool.whenInitialised(function createUserShapes() {
      function* userShapeInfos() {
        let id = 0;
        for (const userShape of seashotConfig.userShapes) {
          const shape = ShapeDataset.Shape.fromJSON(userShape);
          if (shape) {
            shape.id = (id++).toString();
            yield ShapeDataset.APIServer.getInfoForShape(
              Syntool.userShapesProduct.id,
              shape
            );
          }
        }
      }

      if (Syntool.userShapesProduct) {
        if (
          seashotConfig.selection.indexOf(Syntool.userShapesProduct.id) === -1
        ) {
          seashotConfig.selection.push(Syntool.userShapesProduct.id);
        }
        Syntool.mapView.updateDatasets(userShapeInfos());
      }
    });
  }
  Syntool.state = seashotConfig;
}

function getSeashotConfig() {
  return Promise.resolve(window.__SEASHOT_CONFIG__ || getUrlAppState());
}

function getUrlAppState() {
  function getDefaultTimeframe(timeRanges) {
    for (var i = 0; i < timeRanges.length; i++) {
      var range = timeRanges[i][1];
      var isDefault = !!timeRanges[i][2];
      if (isDefault) {
        return range;
      }
    }
    return timeRanges.length > 0 ? timeRanges[0][1] : '1d';
  }

  const config = Syntool.config;
  const appState = deepAssign({
    timeline: {
      currentTime: config.currentDate?.getTime(),
      timeframe: getDefaultTimeframe(config.timeRanges),
    },
    mapView: {zoom: config.defaultZoom},
    geoFilterActive: true,
  }, appStateFromURL());

  if (!appState.mapView.extent && !isFinite(appState.mapView.zoom)) {
    appState.mapView.extent = config.maxExtent;
  }
  return appState;
}
