/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  createClass,
  inRangeInclusive,
  EventsHub,
  getClientXYFromMouseEvent,
  getClientXYOfTouch,
  getCenterOfTouches,
  getXYDistance,
  getXYLength,
} from '../syntool.js';
import {isMouseEvent} from '../utils/dom_utils.js';

var POSSIBLE = 'POSSIBLE';
var BEGAN = 'BEGAN';
var CHANGED = 'CHANGED';
var ENDED = 'ENDED';
var CANCELLED = 'CANCELLED';
var FAILED = 'FAILED';
var RECOGNIZED = ENDED;

const States = {
  possible  : POSSIBLE,
  began     : BEGAN,
  changed   : CHANGED,
  ended     : ENDED,
  cancelled : CANCELLED,
  failed    : FAILED,
  recognized: RECOGNIZED,
};

const MIN_PAN_MOVEMENT = 4;
const MAX_TAP_DOWN = 250;
const MAX_TAP_MOVEMENT = MIN_PAN_MOVEMENT;
const MAX_MULTI_TAP_INTERVAL = 300;
const MAX_MULTI_TAP_MOVEMENT = 10;
const MAX_CLICK_DOWN = MAX_TAP_DOWN;
const MAX_CLICK_MOVEMENT = MIN_PAN_MOVEMENT;
const MAX_MULTI_CLICK_INTERVAL = MAX_MULTI_TAP_INTERVAL;
const MAX_MULTI_CLICK_MOVEMENT = MIN_PAN_MOVEMENT;

function cloneTouch(touch) {
  return {
    identifier: touch.identifier,
    target    : touch.target,
    clientX   : touch.clientX,
    clientY   : touch.clientY,
  };
}
function getTouchId(touch) {
  return touch.identifier;
}
function getTouchWithId(touches, id) {
  for (var i = 0; i < touches.length; i++) {
    if (touches[i].identifier === id) {
      return touches[i];
    }
  }
  return null;
}
function _getTouchWithId(id) {
  // This function is supposed to be called with the touches bound to `this`
  return getTouchWithId(this, id);
}
function getTouchesWithIds(touches, ids) {
  return ids.map(_getTouchWithId, touches).filter(Boolean);
}
function getTouches(event, trackedTouches) {
  var eventTouches = event.touches;
  var changedTouches = Array.from(event.changedTouches, cloneTouch);
  var allTouches = [];
  var newTouches = [];
  var movedTouches = [];
  var removedTouches = [];
  var cancelledTouches = [];
  var disappearedTouches = [];

  if (changedTouches.length > 0) {
    if (event.type === 'touchstart') {
      newTouches = changedTouches;
    } else if (event.type === 'touchmove') {
      movedTouches = changedTouches;
    } else if (event.type === 'touchend') {
      removedTouches = changedTouches;
    } else {
      cancelledTouches = changedTouches;
    }
  }

  trackedTouches.forEach(function(touch) {
    var id = touch.identifier;
    var j;

    for (j = 0; j < eventTouches.length; j++) {
      if (eventTouches[j].identifier === id) {
        allTouches.push(cloneTouch(eventTouches[j]));
        return;
      }
    }
    for (j = 0; j < removedTouches.length; j++) {
      if (removedTouches[j].identifier === id) {
        return;
      }
    }
    for (j = 0; j < cancelledTouches.length; j++) {
      if (cancelledTouches[j].identifier === id) {
        return;
      }
    }
    disappearedTouches.push(touch);
  });
  if (newTouches.length > 0) {
    allTouches = allTouches.concat(newTouches);
  }

  // console.log(
  //   '[' + event.type + ']',
  //   'all: ' + allTouches.map(getTouchId).join(','),
  //   'new: ' + newTouches.map(getTouchId).join(','),
  //   'moved: ' + movedTouches.map(getTouchId).join(','),
  //   'removed: ' + removedTouches.map(getTouchId).join(','),
  //   'cancelled: ' + cancelledTouches.map(getTouchId).join(','),
  //   'disappeared: ' + disappearedTouches.map(getTouchId).join(',')
  // );
  return {
    allTouches        : allTouches,
    newTouches        : newTouches,
    movedTouches      : movedTouches,
    removedTouches    : removedTouches,
    cancelledTouches  : cancelledTouches,
    disappearedTouches: disappearedTouches,
  };
}
function getDistanceBetweenTouches(touches) {
  return getXYDistance(
    getClientXYOfTouch(touches[0]),
    getClientXYOfTouch(touches[1])
  );
}

const TapRecognizer = createClass([
  /*Public*/ {
    constructor: function TapRecognizer(name) {
      this.name = name;
      this.allowedMouseButtons = [];
      this.numberOfTouchesRequired = 1;
      this.maxTouchMovement = MAX_TAP_MOVEMENT;
      this.maxMouseMovement = MAX_CLICK_MOVEMENT;
      this.maxTouchDuration = MAX_TAP_DOWN;
      this.maxMouseDuration = MAX_CLICK_DOWN;
      this.onRecognized = null;

      this._state = POSSIBLE;
      this._timerId = null;
      this._allTouchesDown = false;
      this.location = null;

      this._timerFired = this._timerFired.bind(this);
    },

    cancel() {
      if (this._state === POSSIBLE) {
        this._clearTimer();
        this._state = CANCELLED;
        this.location = null;
      }
    },
  },
  /*Event handlers*/ {
    handleMouseDown(event) {
      if (this.allowedMouseButtons.indexOf(event.button) === -1) {
        this._state = FAILED;
        this.location = null;
        return;
      }

      if (this._state !== POSSIBLE) {
        this._state = POSSIBLE;
      }
      this.location = getClientXYFromMouseEvent(event);
      this._rewindTimer(this.maxMouseDuration);
    },
    handleMouseMove(event) {
      if (
        this._state === POSSIBLE &&
        this.location &&
        getXYDistance(
          getClientXYFromMouseEvent(event),
          this.location
        ) > this.maxMouseMovement
      ) {
        this._clearTimer();
        this._state = FAILED;
        this.location = null;
      }
    },
    handleMouseUp(event) {
      if (
        this._state === POSSIBLE &&
        this.allowedMouseButtons.indexOf(event.button) !== -1
      ) {
        this._clearTimer();
        this._state = RECOGNIZED;
        this.onRecognized?.(event, this);
      }
    },

    handleTouchStart(event, touches) {
      var allTouches = touches.allTouches;
      var fingerCount = allTouches.length;

      if (fingerCount > this.numberOfTouchesRequired) {
        this._state = FAILED;
        this.location = null;
        return;
      }

      if (this._state !== POSSIBLE) {
        this._state = POSSIBLE;
      }
      this._allTouchesDown = (fingerCount === this.numberOfTouchesRequired);
      this.location = getCenterOfTouches(allTouches);
      this._rewindTimer(this.maxTouchDuration);
    },
    handleTouchMove(event, touches) {
      if (this._state === POSSIBLE) {
        var location = getCenterOfTouches(touches.allTouches);
        if (getXYDistance(location, this.location) > this.maxTouchMovement) {
          this._clearTimer();
          this._state = FAILED;
          this.location = null;
        }
      }
    },
    handleTouchEnd(event, touches) {
      if (
        this._state === POSSIBLE &&
        this._allTouchesDown &&
        touches.allTouches.length === 0
      ) {
        this._clearTimer();
        this._state = RECOGNIZED;
        this.onRecognized?.(event, this);
      }
    },
    handleTouchCancel(event, touches) {
      this.cancel();
    },
  },
  /*Protected*/ {
    _clearTimer() {
      window.clearTimeout(this._timerId);
      this._timerId = null;
    },
    _rewindTimer(timeout) {
      this._clearTimer();
      this._timerId = window.setTimeout(this._timerFired, timeout);
    },
    _timerFired() {
      this._state = FAILED;
      this.location = null;
    },
  },
]);

const MultiTapRecognizer = createClass([
  /*Public*/ {
    constructor: function MultiTapRecognizer(name) {
      this.name = name;
      this.allowedMouseButtons = [];
      this.numberOfTouchesRequired = 1;
      this.numberOfTapsRequired = 1;
      this.maxTouchMovement = MAX_MULTI_TAP_MOVEMENT;
      this.maxMouseMovement = MAX_MULTI_CLICK_MOVEMENT;
      this.maxInterTapsInterval = MAX_MULTI_TAP_INTERVAL;
      this.maxInterClicksInterval = MAX_MULTI_CLICK_INTERVAL;
      this.onRecognized = null;

      this._state = POSSIBLE;
      this._timerId = null;
      this._taps = 0;
      this._event = null;
      this.location = null;

      this._tapRecognizer = new TapRecognizer(name + '._tapRecognizer');
      this._tapRecognizer.onRecognized = this.handleTap.bind(this);

      this._timerFired = this._timerFired.bind(this);
    },

    setMaxTouchMovement(maxTouchMovement) {
      this._tapRecognizer.maxTouchMovement = maxTouchMovement;
    },
    setMaxMouseMovement(maxMouseMovement) {
      this._tapRecognizer.maxMouseMovement = maxMouseMovement;
    },

    cancel() {
      this._tapRecognizer.cancel();
      if (this._state === POSSIBLE) {
        this._state = CANCELLED;
        this._clearTimer();
        this._event = null;
        this.location = null;
      }
    },
  },
  /*Event handlers*/ {
    handleMouseDown(event) {
      this._tapRecognizer.allowedMouseButtons = this.allowedMouseButtons;
      this._tapRecognizer.handleMouseDown(event);
      this.checkRecognized();
    },
    handleMouseMove(event) {
      this._tapRecognizer.handleMouseMove(event);
      this.checkRecognized();

      if (
        this._state === POSSIBLE &&
        this.location &&
        getXYDistance(
          getClientXYFromMouseEvent(event),
          this.location
        ) > this.maxMouseMovement
      ) {
        this._clearTimer();
        this._timerFired();
      }
    },
    handleMouseUp(event) {
      this._tapRecognizer.handleMouseUp(event);
    },
    handleTouchStart(event, touches) {
      this._tapRecognizer.numberOfTouchesRequired = this.numberOfTouchesRequired;
      this._tapRecognizer.handleTouchStart(event, touches);
      this.checkRecognized();
    },
    handleTouchMove(event, touches) {
      this._tapRecognizer.handleTouchMove(event, touches);
      this.checkRecognized();
    },
    handleTouchEnd(event, touches) {
      this._tapRecognizer.handleTouchEnd(event, touches);
    },
    handleTouchCancel(event, touches) {
      this._tapRecognizer.handleTouchCancel(event, touches);
    },
    handleTap(event, recognizer) {
      if (this._state !== POSSIBLE) {
        this._state = POSSIBLE;
        this._taps = 0;
      }
      if (this._taps === 0) {
        this.location = recognizer.location;
      } else {
        var movement = getXYDistance(recognizer.location, this.location);
        var maxMovement = (
          isMouseEvent(event)
          ? this.maxMouseMovement
          : this.maxTouchMovement
        );
        if (movement > maxMovement) {
          if (this._taps === this.numberOfTapsRequired) {
            this._state = RECOGNIZED;
            this.onRecognized?.(this._event, this);
          }
          this._state = POSSIBLE;
          this._taps = 0;
          this.location = recognizer.location;
        }
      }

      this._taps++;
      this._event = event;
      this._rewindTimer(
        isMouseEvent(event)
        ? this.maxInterClicksInterval
        : this.maxInterTapsInterval
      );
    },
  },
  /*Protected*/ {
    checkRecognized() {
      if (
        this._state === POSSIBLE &&
        this._taps === this.numberOfTapsRequired &&
        this._tapRecognizer._state === FAILED
      ) {
        this._state = RECOGNIZED;
        this.onRecognized?.(this._event, this);
        this._clearTimer();
      }
    },

    _clearTimer() {
      window.clearTimeout(this._timerId);
      this._timerId = null;
    },
    _rewindTimer(timeout) {
      this._clearTimer();
      this._timerId = window.setTimeout(this._timerFired, timeout);
    },
    _timerFired() {
      if (this._taps === this.numberOfTapsRequired) {
        this._state = RECOGNIZED;
        this.onRecognized?.(this._event, this);
      } else {
        this._state = FAILED;
        this._event = null;
        this.location = null;
      }
    },
  },
]);

const PinchRecognizer = createClass([
  /*Public*/ {
    constructor: function PinchRecognizer(name) {
      this.name = name;
      this.hub = new EventsHub();

      this._touchIds = [];
      this._startDistance = null;
      this.location = null;
      this.scale = 1;
    },

    cancel() {
      if (this._touchIds.length > 0) {
        this.hub.emit('cancel', null, this);
        this._touchIds.length = 0;
        this._startDistance = null;
        this.location = null;
        this.scale = 1;
      }
    },
  },
  /*Event handlers*/ {
    handleTouchStart(event, touches) {
      var allTouches = touches.allTouches;
      var fingerCount = allTouches.length;

      if (fingerCount === 2) {
        this._touchIds = allTouches.map(getTouchId);
        this._startDistance = getDistanceBetweenTouches(allTouches);
        this.location = getCenterOfTouches(allTouches);
        this.hub.emit('start', event, this);
      }
    },
    handleTouchMove(event, touches) {
      if (this._touchIds.length > 0) {
        var allTouches = touches.allTouches;
        var pinchTouches = getTouchesWithIds(allTouches, this._touchIds);
        var distance = getDistanceBetweenTouches(pinchTouches);
        this.location = getCenterOfTouches(pinchTouches);
        this.scale = distance / this._startDistance;
        this.hub.emit('change', event, this);
      }
    },
    handleTouchEnd(event, touches) {
      if (this._touchIds.length > 0) {
        var allTouches = touches.allTouches;
        var pinchTouches = getTouchesWithIds(allTouches, this._touchIds);
        if (pinchTouches.length < 2) {
          this.hub.emit('end', event, this);
          this._touchIds.length = 0;
          this._startDistance = null;
          this.location = null;
          this.scale = 1;
        } else {
          var distance = getDistanceBetweenTouches(pinchTouches);
          this.location = getCenterOfTouches(pinchTouches);
          this.scale = distance / this._startDistance;
          this.hub.emit('change', event, this);
        }
      }
    },
    handleTouchCancel(event, touches) {
      if (this._touchIds.length > 0) {
        var allTouches = touches.allTouches;
        var pinchTouches = getTouchesWithIds(allTouches, this._touchIds);
        if (pinchTouches.length < 2) {
          this.hub.emit('cancel', event, this);
          this._touchIds.length = 0;
          this._startDistance = null;
          this.location = null;
          this.scale = 1;
        } else {
          var distance = getDistanceBetweenTouches(pinchTouches);
          this.location = getCenterOfTouches(pinchTouches);
          this.scale = distance / this._startDistance;
          this.hub.emit('change', event, this);
        }
      }
    },
  },
]);

function isMouseButtonPressed(event, button) {
  switch (button) {
  case 1:
    button = 2;
    break;
  case 2:
    button = 1;
    break;
  default:
  }

  //eslint-disable-next-line no-bitwise
  return (event.buttons & (1 << button)) > 0;
}

const PanRecognizer = createClass([
  /*Public*/ {
    constructor: function PanRecognizer(name) {
      this.name = name;
      this.hub = new EventsHub();
      this.allowedMouseButtons = [0];
      this.minimumNumberOfTouches = 1;
      this.maximumNumberOfTouches = Infinity;
      this.movementThreshold = MIN_PAN_MOVEMENT;

      this._state = POSSIBLE;
      this._button = null;
      this._previousLocation = null;
      this._numberOfTouches = 0;
      this.startEvent = null;
      this.startLocation = null;
      this.event = null;
      this.translation = [0, 0];
    },

    get location() {
      var xy = this.startLocation.slice();
      xy[0] += this.translation[0];
      xy[1] += this.translation[1];
      return xy;
    },
    get mouseButton() {
      return this._button;
    },
    get numberOfTouches() {
      return this._numberOfTouches;
    },

    getTranslationIn(element) {
      const clientRect = element.getBoundingClientRect();
      return [
        this.translation[0] * element.clientWidth / clientRect.width,
        this.translation[1] * element.clientHeight / clientRect.height,
      ];
    },

    cancel() {
      this.handleDragCancel();
    },
  },
  /*Event handlers*/ {
    handleMouseDown(event) {
      if (this.allowedMouseButtons.indexOf(event.button) !== -1) {
        this._button = event.button;
        this.handleStart(event, getClientXYFromMouseEvent(event));
      }
    },
    handleMouseMove(event) {
      if (this._button !== null && isMouseButtonPressed(event, this._button)) {
        this.handleMove(event, getClientXYFromMouseEvent(event));
      }
    },
    handleMouseUp(event) {
      if (this._button === event.button) {
        this._button = null;
        this.handleEnd(event, getClientXYFromMouseEvent(event));
      }
    },

    handleTouchStart(event, touches) {
      this._numberOfTouches = touches.allTouches.length;
      if (
        inRangeInclusive(
          this._numberOfTouches,
          this.minimumNumberOfTouches,
          this.maximumNumberOfTouches
        )
      ) {
        this.handleStart(event, getCenterOfTouches(touches.allTouches));
      }
    },
    handleTouchMove(event, touches) {
      this._numberOfTouches = touches.allTouches.length;
      this.handleMove(event, getCenterOfTouches(touches.allTouches));
    },
    handleTouchEnd(event, touches) {
      this._numberOfTouches = touches.allTouches.length;
      if (this._numberOfTouches < this.minimumNumberOfTouches) {
        this.handleEnd(event, getCenterOfTouches(touches.removedTouches));
      } else {
        var xy = getCenterOfTouches(touches.allTouches);
        this._previousLocation = xy;
        this.handleMove(event, xy);
      }
    },
    handleTouchCancel(event, touches) {
      this._numberOfTouches = touches.allTouches.length;
      if (this._numberOfTouches < this.minimumNumberOfTouches) {
        this.handleDragCancel(event);
      } else {
        var xy = getCenterOfTouches(touches.allTouches);
        this._previousLocation = xy;
        this.handleMove(event, xy);
      }
    },

    handleStart(event, xy) {
      this._previousLocation = xy;

      if (this._state === BEGAN || this._state === CHANGED) {
        return;
      }

      this._state = POSSIBLE;
      this.startEvent = event;
      this.startLocation = xy;
      this.translation = [0, 0];
    },
    handleMove(event, xy) {
      this.event = event;
      this.translation[0] += xy[0] - this._previousLocation[0];
      this.translation[1] += xy[1] - this._previousLocation[1];
      this._previousLocation = xy;

      if (this._state === POSSIBLE) {
        if (getXYLength(this.translation) < this.movementThreshold) {
          return;
        }
        this._state = BEGAN;
      } else if (this._state === BEGAN) {
        this._state = CHANGED;
      } else if (this._state !== CHANGED) {
        return;
      }

      this.hub.emit(this._state === BEGAN ? 'start' : 'change', event, this);
      this.event = null;
    },
    handleEnd(event, xy) {
      if (this._state === BEGAN || this._state === CHANGED) {
        this._state = ENDED;
        this.event = event;
        this.translation[0] += xy[0] - this._previousLocation[0];
        this.translation[1] += xy[1] - this._previousLocation[1];
        this.hub.emit('end', event, this);
        this.startEvent = null;
        this.event = null;
        this._numberOfTouches = 0;
      }
    },
    handleDragCancel(event) {
      if (this._state === BEGAN || this._state === CHANGED) {
        this._state = CANCELLED;
        this.event = event;
        this.hub.emit('cancel', event, this);
        this.startEvent = null;
        this.event = null;
        this._numberOfTouches = 0;
      }
    },
  },
]);

export {
  MIN_PAN_MOVEMENT,
  MAX_TAP_DOWN,
  MAX_TAP_MOVEMENT,
  MAX_MULTI_TAP_INTERVAL,
  MAX_MULTI_TAP_MOVEMENT,
  MAX_CLICK_DOWN,
  MAX_CLICK_MOVEMENT,
  MAX_MULTI_CLICK_INTERVAL,
  MAX_MULTI_CLICK_MOVEMENT,

  States,

  getTouches,

  TapRecognizer,
  MultiTapRecognizer,
  PinchRecognizer,
  PanRecognizer,
};
window.ODL = Object.assign(window.ODL || {}, {
  GestureRecognizer: {
    MIN_PAN_MOVEMENT,
    MAX_TAP_DOWN,
    MAX_TAP_MOVEMENT,
    MAX_MULTI_TAP_INTERVAL,
    MAX_MULTI_TAP_MOVEMENT,
    MAX_CLICK_DOWN,
    MAX_CLICK_MOVEMENT,
    MAX_MULTI_CLICK_INTERVAL,
    MAX_MULTI_CLICK_MOVEMENT,

    States,

    getTouches,

    TapRecognizer,
    MultiTapRecognizer,
    PinchRecognizer,
    PanRecognizer,
  },
});
