/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {SECOND_IN_MILLISECONDS, getNowMs} from '../syntool.js';
import {CircularBuffer} from './circular_buffer.js';

const DECELERATION_RATE = 0.995;

export class Inertia {
  constructor() {
    this.velocitySamples = new CircularBuffer(10, 100);
    this.lastPoint = null;
    this.decelerating = false;
    this._rafId = null;
    this._callback = null;
  }

  begin() {
    this.stopDecelerating();
    this.velocitySamples.clear();
    this.lastPoint = null;
  }
  stopDecelerating() {
    if (this.decelerating) {
      window.cancelAnimationFrame(this._rafId);
      this._callback?.(0, 0, true);
      this.decelerating = false;
      this._callback = null;
    }
  }
  update(x, y) {
    const now = getNowMs();
    if (this.lastPoint) {
      const dx = x - this.lastPoint[0];
      const dy = y - this.lastPoint[1];
      const dt = now - this.lastPoint[2];
      if (Math.abs(dx * dy) > 0 && dt > 8) {
        this.velocitySamples.push([dx, dy, dt]);
      }
    }
    this.lastPoint = [x, y, now];
  }
  apply(callback) {
    const velocitySample = this.velocitySamples.getTail();
    this.velocitySamples.clear();

    if (!velocitySample) {
      return false;
    }

    const [dx, dy, dt] = velocitySample;
    let vx = (dx / dt) * SECOND_IN_MILLISECONDS;
    let vy = (dy / dt) * SECOND_IN_MILLISECONDS;

    const speed = Math.hypot(vx, vy);
    if (speed < 250) {
      return false;
    }

    let lastTime = getNowMs();
    this.decelerating = true;
    this._callback = callback;

    const handleAnimationFrame = () => {
      const now = getNowMs();
      const deltaT = now - lastTime;
      const decelerationFactor = Math.exp(Math.log(DECELERATION_RATE) * deltaT);

      const n = DECELERATION_RATE * (1 - decelerationFactor) * 200;
      const deltaX = vx / SECOND_IN_MILLISECONDS * n;
      const deltaY = vy / SECOND_IN_MILLISECONDS * n;

      vx *= decelerationFactor;
      vy *= decelerationFactor;

      const end = (Math.abs(vx) <= 10 && Math.abs(vy) <= 10);

      callback(deltaX, deltaY, end);

      if (end) {
        this.decelerating = false;
      } else {
        lastTime = now;
        this._rafId = window.requestAnimationFrame(handleAnimationFrame);
      }
    };

    this._rafId = window.requestAnimationFrame(handleAnimationFrame);
    return true;
  }
}
window.ODL = Object.assign(window.ODL || {}, {Inertia});
