/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  Z_INDEX,
  wrapFn,
  extendClass,
  Config,
  MapView,
  Types as is,
} from '../syntool.js';
import {FeaturesRenderer} from '../utils/features_renderer.js';
import {OLVectorLayer} from '../utils/ol_vector_layer.js';
import './map_view_dataset_resources.js';

const MarkerPolygonSymbol = is.arrayOf(is.number.isRequired);
const MarkerNodeSymbol = is.shape({
  node: is.instanceOf(SVGElement).isRequired,
  width: is.number.gt(0).isRequired,
  height: is.number.gt(0).isRequired,
});
const MarkerSymbol = is.oneOfType([MarkerPolygonSymbol, MarkerNodeSymbol]);
const ImageMarkerStyle = is.shape({
  rotation       : is.number,
  externalGraphic: is.URL.isRequired,
  graphicWidth   : is.number.gt(0).isRequired,
  graphicHeight  : is.number.gt(0).isRequired,
  graphicOpacity : is.number.gte(0).lte(1),
  graphicXOffset : is.number,
  graphicYOffset : is.number,
  graphicHFlip   : is.bool,
  graphicVFlip   : is.bool,
  graphicRotationOffset: is.number.gte(-360).lte(360),
  graphicRotation: is.oneOfType([is.bool, is.number]),
  noWrap         : is.bool,
});
const PolygonMarkerGraphicName = is.lazy(function(props, propName, propFullName, getProp) {
  let graphicNames = [...FeaturesRenderer.symbols.keys()];
  const markerSymbols = getProp('.markerSymbols');
  if (markerSymbols) {
    graphicNames = graphicNames.concat(Object.keys(markerSymbols));
  }
  return is.oneOf(graphicNames);
});
const PolygonMarkerStyle = is.shape({
  rotation       : is.number,
  graphicName    : PolygonMarkerGraphicName.isRequired,
  pointRadius    : is.number.gt(0).isRequired,
  fillColor      : is.SVGPaint,
  fillOpacity    : is.number.gte(0).lte(1),
  strokeColor    : is.SVGPaint,
  strokeOpacity  : is.number.gte(0).lte(1),
  strokeWidth    : is.oneOfType([is.string, is.number]),
  strokeDashstyle: is.string,
  graphicOpacity : is.number.gte(0).lte(1),
  graphicHFlip   : is.bool,
  graphicVFlip   : is.bool,
  graphicRotationOffset: is.number.gte(-360).lte(360),
  graphicRotation: is.oneOfType([is.bool, is.number]),
  noWrap         : is.bool,
});
const MarkerStateStyle = is.oneOfType([ImageMarkerStyle, PolygonMarkerStyle]);
const MarkerStyle = is.oneOfType([
  MarkerStateStyle,
  is.shape({
    default : MarkerStateStyle.isRequired,
    selected: MarkerStateStyle,
  }),
]);
const MarkerFn = is.func;
const Marker = is.oneOfType([MarkerStyle, MarkerFn]);

const MarkerSymbols = is.objectOf(MarkerSymbol.isRequired);

Object.assign(Config, {
  applyMarkerStyleDefaults,
  applyMarkerDefaults,
});
Object.assign(Config.types, {
  MarkerSymbol,
  MarkerStyle,
  MarkerFn,
  Marker,
});
Object.assign(Config.configFieldValidators, {
  markerSymbols: MarkerSymbols,
});
Config.configNormalizers.push(function normalizeConfigWithDefaults(config) {
  return Object.assign({markerSymbols: {}}, config);
});

function applyMarkerStateStyleDefaults(marker) {
  return Object.assign({
    strokeWidth: 1,
    strokeOpacity: 1,
    fillOpacity: 1,
    graphicOpacity: 1,
    graphicXOffset: -marker.graphicWidth / 2,
    graphicYOffset: -marker.graphicHeight / 2,
    graphicRotationOffset: marker.rotation == null ? 0 : -marker.rotation,
    graphicRotation: marker.rotation == null ? true : !isNaN(marker.rotation),
  }, marker);
}
function applyMarkerStyleDefaults(marker) {
  return {
    default: applyMarkerStateStyleDefaults(marker.default || marker),
    selected: marker.selected && applyMarkerStateStyleDefaults(marker.selected),
  };
}
function applyMarkerDefaults(marker, defaultMarker) {
  if (typeof marker === 'function') {
    return wrapFn(marker, function() {
      return applyMarkerStyleDefaults(window.original() || defaultMarker);
    });
  }
  return applyMarkerStyleDefaults(marker || defaultMarker);
}

export {
  MarkerStyle,
  Marker,

  applyMarkerStyleDefaults,
  applyMarkerDefaults,
};

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      for (const [name, symbol] of Object.entries(this.syntool.config.markerSymbols)) {
        FeaturesRenderer.symbols.set(name, symbol);
      }

      var clickableFeaturesLayer = new OLVectorLayer('ClickableFeatures', {
        style: {
          base: (feature) => {
            const {markerStyle, rotation} = feature.properties;
            const selected = feature.dataset.isSelected();
            const opacity = feature.dataset.getOpacity();

            const state = (selected ? 'selected' : 'default');
            const markerStateStyle = (markerStyle[state] || markerStyle.default);
            const style = Object.assign({}, markerStateStyle);

            const {graphicRotationOffset, graphicRotation} = markerStateStyle;
            style.graphicRotation = (
              graphicRotation === true
              ? (
                isNaN(rotation)
                ? graphicRotationOffset
                : graphicRotationOffset + rotation
              )
              : graphicRotationOffset + (graphicRotation || 0)
            );

            const {graphicOpacity} = markerStateStyle;
            style.graphicOpacity = (
              graphicOpacity === true
              ? opacity
              : graphicOpacity ?? 1
            );

            return style;
          },
        },
      });
      this._clickableFeaturesLayer = clickableFeaturesLayer;
      this._hiddenClickableFeatures = new WeakSet();
      this.addLayer(clickableFeaturesLayer);

      this.hub
        .on(
          'dataset:did-add-to-map',
          this.handleClickableFeaturesDatasetDidAddToMap,
          this
        )
        .on(
          'dataset:will-remove-from-map',
          this.handleClickableFeaturesDatasetWillRemoveFromMap,
          this
        )
        .on(
          'dataset:change:selected',
          this.handleClickableFeaturesDatasetSelectedChange,
          this
        );
    },
  },
  /*Protected - Layer Z-Index*/ {
    getZIndexForLayer(layer, index) {
      if (layer === this._clickableFeaturesLayer) {
        return Z_INDEX.CLICKABLE_FEATURES;
      }
      return window.original();
    },
  },
  /*Protected - Dataset handlers*/ {
    handleClickableFeaturesDatasetDidAddToMap(dataset) {
      this.updateClickableFeaturesVisibility(dataset);
      const clickableFeatures = (
        this.getResourceForDataset(dataset, 'clickableFeatures')
      );
      if (clickableFeatures) {
        this._clickableFeaturesLayer.addFeatures(clickableFeatures);
      }
    },
    handleClickableFeaturesDatasetWillRemoveFromMap(dataset) {
      if (this.syntool.productsStore.isProductSelected(dataset.productId)) {
        this._clickableFeaturesLayer.removeAllFeaturesOfDataset(dataset);
      } else {
        this._clickableFeaturesLayer.removeAllFeaturesOfProductWithId(
          dataset.productId
        );
      }
    },
    handleClickableFeaturesDatasetSelectedChange(dataset) {
      if (this.hasClickableFeatures(dataset)) {
        this._clickableFeaturesLayer.drawFeatures(
          this.getResourceForDataset(dataset, 'clickableFeatures')
        );
      }
    },
  },
  /*Public - Dataset resources*/ {
    destroyResourcesForDataset(dataset) {
      if (this.hasResourceForDataset(dataset, 'clickableFeatures')) {
        const clickableFeatures = (
          this.getResourceForDataset(dataset, 'clickableFeatures')
        );
        for (const feature of clickableFeatures || []) {
          this._hiddenClickableFeatures.delete(feature);
          feature.destroy();
        }
      }
      window.original();
    },
  },
  /*Public - Dataset clickable features*/ {
    addClickableFeature(dataset, feature, hidden) {
      let clickableFeatures = (
        this.getResourceForDataset(dataset, 'clickableFeatures')
      );
      if (!clickableFeatures) {
        clickableFeatures = [];
        this.setResourceForDataset(
          dataset,
          'clickableFeatures',
          clickableFeatures
        );
      }

      feature.dataset = dataset;
      clickableFeatures.push(feature);
      if (hidden) {
        this._hiddenClickableFeatures.add(feature);
      }
      this.updateClickableFeatureVisibility(dataset, feature);
      if (this.isDatasetInMap(dataset)) {
        this._clickableFeaturesLayer.addFeature(feature);
      }
    },
    removeClickableFeature(dataset, feature) {
      const clickableFeatures = (
        this.getResourceForDataset(dataset, 'clickableFeatures')
      );
      feature.dataset = null;
      if (clickableFeatures) {
        const index = clickableFeatures.indexOf(feature);
        if (index !== -1) {
          clickableFeatures.splice(index, 1);
        }
      }
      this._clickableFeaturesLayer.removeFeature(feature);
      this._hiddenClickableFeatures.delete(feature);
    },
    hasClickableFeatures(dataset) {
      const clickableFeatures = (
        this.getResourceForDataset(dataset, 'clickableFeatures')
      );
      return clickableFeatures?.length > 0;
    },
    showClickableFeature(dataset, feature, redraw) {
      if (this._hiddenClickableFeatures.has(feature) || redraw) {
        this._hiddenClickableFeatures.delete(feature);
        this.updateClickableFeatureVisibility(dataset, feature, redraw);
      }
    },
    hideClickableFeature(dataset, feature) {
      if (!this._hiddenClickableFeatures.has(feature)) {
        this._hiddenClickableFeatures.add(feature);
        this.updateClickableFeatureVisibility(dataset, feature);
      }
    },
    updateClickableFeatureVisibility(dataset, feature, redraw) {
      const visible = (
        this.isDatasetInMap(dataset) &&
        !this._hiddenClickableFeatures.has(feature)
      );
      const changed = (feature.style?.display !== 'none') !== visible;
      if (changed) {
        if (visible) {
          feature.style.display = '';
        } else if (feature.style) {
          feature.style.display = 'none';
        } else {
          feature.style = {display: 'none'};
        }
      }
      if (redraw || changed) {
        feature.layer?.drawFeature(feature, redraw /*force*/);
      }
    },
    updateClickableFeaturesVisibility(dataset) {
      if (this.hasClickableFeatures(dataset)) {
        const clickableFeatures = (
          this.getResourceForDataset(dataset, 'clickableFeatures')
        );
        for (const clickableFeature of clickableFeatures) {
          this.updateClickableFeatureVisibility(dataset, clickableFeature);
        }
      }
    },
    redrawClickableFeatures(dataset) {
      if (this.hasClickableFeatures(dataset)) {
        this._clickableFeaturesLayer.drawFeatures(
          this.getResourceForDataset(dataset, 'clickableFeatures'),
          true /*force*/
        );
      }
    },
  },
  /*Protected - Dataset user selectability*/ {
    isDatasetUserSelectable(dataset) {
      if (!dataset) {
        return false;
      }
      if (window.original?.()) {
        return true;
      }
      const clickableFeatures = (
        this.getResourceForDataset(dataset, 'clickableFeatures')
      );
      if (clickableFeatures) {
        for (const feature of clickableFeatures) {
          if (feature.layer?.isFeatureVisible(feature)) {
            return true;
          }
        }
      }
      return false;
    },
  },
]);
