/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {extendClass, MapView} from '../syntool.js';
import './map_view_gesture_recognizers.js';

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      this._datasetInteractionsActive = false;
      this._hoveredDataset = null;

      this._stateHooks.push({
        get: ({datasetInteractionsActive}) => ({datasetInteractionsActive}),
        set: ({datasetInteractionsActive}) => {
          this.datasetInteractionsActive = datasetInteractionsActive;
        },
      });

      this.activateDatasetInteractions();
    },
  },
  /*Public - Dataset selection*/ {
    activateDatasetInteractions() {
      if (this._datasetInteractionsActive) {
        return false;
      }
      this._datasetInteractionsActive = true;
      this.hub
        .on('mousemove', MapView.SELECT, this.handleInteractionsMouseMove, this)
        .on('tap-1', MapView.SELECT, this.handleInteractionsTap, this)
        .on('click', MapView.SELECT, this.handleInteractionsMouseClick, this);
      this.hub.emit('change:datasetInteractionsActive');
      this._invalidateState();
      return true;
    },
    deactivateDatasetInteractions() {
      if (!this._datasetInteractionsActive) {
        return false;
      }
      this.hub
        .off('mousemove', MapView.SELECT, this.handleInteractionsMouseMove, this)
        .off('tap-1', MapView.SELECT, this.handleInteractionsTap, this)
        .off('click', MapView.SELECT, this.handleInteractionsMouseClick, this);
      this._hoveredDataset = null;
      this._datasetInteractionsActive = false;
      this.hub.emit('change:datasetInteractionsActive');
      this._invalidateState();
      return true;
    },
    get datasetInteractionsActive() {
      return this._datasetInteractionsActive;
    },
    set datasetInteractionsActive(on) {
      if (on) {
        this.activateDatasetInteractions();
      } else {
        this.deactivateDatasetInteractions();
      }
    },
  },
  /*Protected - Dataset interactions*/ {
    _setHoveredDataset(dataset) {
      if (dataset !== this._hoveredDataset) {
        if (this._hoveredDataset) {
          this.hub.emit('dataset:unhover', this._hoveredDataset);
        }
        this._hoveredDataset = dataset;
        if (dataset) {
          this.hub.emit('dataset:hover', dataset);
        }
      }
    },
    handleInteractionsMouseMove(event) {
      var dataset = this.getDatasetFromMouseEvent(event);
      if (dataset !== this._hoveredDataset) {
        this._setHoveredDataset(dataset);
      }
    },
    handleInteractionsMouseClick(event) {
      if (this.isValidClick(event)) {
        this.handleInteractionsClick(event);
      }
    },
    handleInteractionsTap(event, recognizer) {
      this.handleInteractionsClick(event, recognizer);
    },
    handleInteractionsClick(event, recognizer) {
      if (event.target instanceof SVGGElement) {
        // Ignore clicks that have a <g> element as target
        // Because if <g> is the target then none of its children were clicked
        // and we are only interested in the children
        return;
      }

      var dataset = (
        recognizer
        ? this.getDatasetFromEventAtClientXY(event, recognizer.location)
        : this.getDatasetFromMouseEvent(event)
      );
      this._setHoveredDataset(dataset);
      this.hub.emit('dataset:click', dataset);
    },
  },
]);
