/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  Z_INDEX,
  extendClass,
  Geometry,
  Feature,
  Dataset,
  MapView,
} from '../syntool.js';
import {OLVectorLayer} from '../utils/ol_vector_layer.js';
import './map_view_dataset_resources.js';

const SHOWN_STYLE = null;
const HIDDEN_STYLE = {display: 'none'};

extendClass(Dataset, [
  /*Protected - Outlines*/ {
    getOutlineGeometry() {
      return (
        this.geometry && !Geometry.isPoint(this.geometry)
        ? this.geometry
        : null
      );
    },
  },
]);

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      var outlinesLayer = new OLVectorLayer('Outlines', {
        style: {
          base: {
            strokeColor: '#eee',
            strokeOpacity: 0.5,
            strokeWidth: 2,
            fillOpacity: 0,
            pointerEvents: 'none',
          },
          default: (feature) => {
            const outlinesHidden = this.outlinesHidden;

            if (outlinesHidden) {
              return HIDDEN_STYLE;
            }

            const dataset = feature.dataset;
            const product = dataset.getProduct();
            const hideMode = product.outlines;

            switch (hideMode) {
            case 'ALWAYS_SHOWN':
            default:
              return SHOWN_STYLE;
            case 'HIDDEN_IN_ZOOM_RANGE':
              return (
                this.syntool.filters.product.zoom(product)
                ? HIDDEN_STYLE
                : SHOWN_STYLE
              );
            case 'ALWAYS_HIDDEN':
              return HIDDEN_STYLE;
            }
          },
          select: {
            strokeColor  : '#dc0',
            strokeOpacity: 1,
          },
        },
      });
      this._outlinesHidden = false;
      this._outlinesLayer = outlinesLayer;

      this.addLayer(outlinesLayer);

      this._stateHooks.push({
        get: ({outlinesHidden}) => ({outlinesHidden}),
        set: ({outlinesHidden}) => {
          this.outlinesHidden = outlinesHidden;
        },
      });

      this.hub
        .on(
          'change:zoom',
          this.handleOutlinesZoomChange,
          this
        )
        .on(
          'dataset:did-add-to-map',
          this.handleOutlinesDatasetDidAddToMap,
          this
        )
        .on(
          'dataset:will-remove-from-map',
          this.handleOutlinesDatasetWillRemoveFromMap,
          this
        )
        .on(
          'dataset:change:selected',
          this.handleOutlinesDatasetSelectedChange,
          this
        );
    },
  },
  /*Protected - Layer Z-Index*/ {
    getZIndexForLayer(layer, index) {
      if (layer === this._outlinesLayer) {
        return Z_INDEX.OUTLINES;
      }
      return window.original();
    },
  },
  /*Public - Dataset resources*/ {
    destroyResourcesForDataset(dataset) {
      if (this.hasResourceForDataset(dataset, 'outlineFeature')) {
        this.getDatasetOutline(dataset).destroy();
      }
      window.original();
    },
  },
  /*Public - Dataset outline*/ {
    hasDatasetOutline(dataset) {
      return this.hasResourceForDataset(dataset, 'outlineFeature');
    },
    getDatasetOutline(dataset) {
      return this.getResourceForDataset(dataset, 'outlineFeature');
    },
    isDatasetOutline(feature) {
      return feature.layer === this._outlinesLayer;
    },

    get outlinesHidden() {
      return this._outlinesHidden;
    },
    set outlinesHidden(hidden) {
      if (this.outlinesHidden === !hidden) {
        this._outlinesHidden = !!hidden;
        this._outlinesLayer.redrawAllFeatures();
        this._invalidateState();
      }
    },
  },
  /*Protected - Dataset handlers*/ {
    handleOutlinesZoomChange() {
      if (this.outlinesHidden) {
        return;
      }

      for (const feature of this._outlinesLayer.features) {
        if (feature.dataset.getProduct().outlines === 'HIDDEN_IN_ZOOM_RANGE') {
          this._outlinesLayer.drawFeature(feature);
        }
      }
    },
    handleOutlinesDatasetDidAddToMap(dataset) {
      var outlineFeature = this.getResourceForDataset(dataset, 'outlineFeature');
      if (!outlineFeature && this.isRegionalDataset(dataset)) {
        var geometry = dataset.getOutlineGeometry();
        if (geometry) {
          outlineFeature = new Feature(geometry);
          outlineFeature.dataset = dataset;
          this.setResourceForDataset(dataset, 'outlineFeature', outlineFeature);
        }
      }
      this._outlinesLayer.addFeature(outlineFeature);
    },
    handleOutlinesDatasetWillRemoveFromMap(dataset) {
      if (this.syntool.productsStore.isProductSelected(dataset.productId)) {
        this._outlinesLayer.removeAllFeaturesOfDataset(dataset);
      } else {
        this._outlinesLayer.removeAllFeaturesOfProductWithId(dataset.productId);
      }
    },
    handleOutlinesDatasetSelectedChange(dataset) {
      var feature = this.getDatasetOutline(dataset);
      if (feature) {
        feature.renderIntent = dataset.isSelected() ? 'select' : 'default';
        feature.layer?.drawFeature(feature);
      }
    },
  },
  /*Protected - Dataset user selectability*/ {
    isDatasetUserSelectable(dataset) {
      if (!dataset) {
        return false;
      }
      if (window.original?.()) {
        return true;
      }
      var outlineFeature = this.getResourceForDataset(dataset, 'outlineFeature');
      return (outlineFeature?.layer === this._outlinesLayer);
    },
  },
]);
