/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {extendClass, HIDModifiers, Dataset, MapView} from '../syntool.js';
import './map_view_dataset_interactions.js';

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      this.multipleSelectKey = HIDModifiers.ctrlOrMeta;

      this._datasetSelectionActive = false;
      this._selectedDatasets = new Set();

      this._stateHooks.push({
        get: (mapView) => ({
          datasetSelectionActive: this.datasetSelectionActive,
          get selectedDatasetIds() {
            return Array.from(mapView._selectedDatasets, (d) => d.datasetId);
          },
        }),
        set: ({datasetSelectionActive, selectedDatasetIds = []}) => {
          this.datasetSelectionActive = datasetSelectionActive;
          if (datasetSelectionActive) {
            this.selectDatasets(
              selectedDatasetIds.map(this.getDatasetWithId, this)
            );
          }
        },
      });

      this.activateDatasetSelection();
    },
  },
  /*Public - Dataset selection*/ {
    activateDatasetSelection() {
      if (this._datasetSelectionActive) {
        return false;
      }
      this._datasetSelectionActive = true;
      this.hub
        .on('dataset:will-remove-from-map', this.unselectDataset, this)
        .on('dataset:click', this.handleDatasetClick, this);
      this.hub.emit('change:datasetSelectionActive');
      this._invalidateState();
      return true;
    },
    deactivateDatasetSelection() {
      if (!this._datasetSelectionActive) {
        return false;
      }
      this.hub
        .off('dataset:will-remove-from-map', this.unselectDataset, this)
        .off('dataset:click', this.handleDatasetClick, this);
      this._datasetSelectionActive = false;
      this.hub.emit('change:datasetSelectionActive');
      if (!this.unselectAllDatasets()) {
        this._invalidateState();
      }
      return true;
    },
    get datasetSelectionActive() {
      return this._datasetSelectionActive;
    },
    set datasetSelectionActive(on) {
      if (on) {
        this.activateDatasetSelection();
      } else {
        this.deactivateDatasetSelection();
      }
    },

    isDatasetUserSelectable(dataset) {
      return !!window.original && window.original();
    },

    selectClickDataset(dataset) {
      this.handleDatasetClick(
        dataset && this.isDatasetInMap(dataset)
        ? dataset
        : null
      );
    },
    selectDataset(dataset) {
      if (!dataset || !this.isDatasetInMap(dataset)) {
        return false;
      }

      var changed = this._deselectAllDatasets(dataset);
      var selected = this._selectDataset(dataset);

      changed = changed || selected;
      if (changed) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return changed;
    },
    selectDatasets(datasets) {
      var datasetsSet = new Set(datasets);
      var changed = (
        this._deselectAllDatasetsExcept(datasetsSet.has, datasetsSet)
      );
      for (var l = 0; l < datasets.length; ++l) {
        if (this.isDatasetInMap(datasets[l])) {
          var selected = this._selectDataset(datasets[l]);
          changed = changed || selected;
        }
      }
      if (changed) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return changed;
    },
    addDatasetToSelection(dataset) {
      if (!this.isDatasetInMap(dataset)) {
        return false;
      }
      var changed = this._selectDataset(dataset);
      if (changed) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return changed;
    },
    isDatasetSelected(dataset) {
      return this._selectedDatasets.has(dataset);
    },
    hasSelectedDatasets(filter, thisArg) {
      if (!filter) {
        return this._selectedDatasets.size > 0;
      }
      for (const dataset of this._selectedDatasets) {
        if (filter.call(thisArg, dataset)) {
          return true;
        }
      }
      return false;
    },
    getSelectedDatasetsCount(filter, thisArg) {
      if (!filter) {
        return this._selectedDatasets.size;
      }
      var count = 0;
      for (const dataset of this._selectedDatasets) {
        if (filter.call(thisArg, dataset)) {
          count++;
        }
      }
      return count;
    },
    getSelectedDataset(filter, thisArg) {
      if (!filter) {
        if (this._selectedDatasets.size !== 1) {
          return null;
        }
        for (const dataset of this._selectedDatasets) {
          return dataset;
        }
      }
      var selectedDataset = null;
      for (const dataset of this._selectedDatasets) {
        if (filter.call(thisArg, dataset)) {
          if (selectedDataset) {
            // More than one
            return null;
          } else {
            selectedDataset = dataset;
          }
        }
      }
      return selectedDataset;
    },
    getSelectedDatasets(filter, thisArg) {
      if (!filter) {
        return [...this._selectedDatasets];
      }
      var selectedDatasets = [];
      for (const dataset of this._selectedDatasets) {
        if (filter.call(thisArg, dataset)) {
          selectedDatasets.push(dataset);
        }
      }
      return selectedDatasets;
    },
    getSelectedDatasetProduct(filter, thisArg) {
      var selectedProduct = null;
      for (const dataset of this._selectedDatasets) {
        var product = dataset.getProduct();
        if (!filter || filter.call(thisArg, product)) {
          if (selectedProduct) {
            // More than one
            return null;
          } else {
            selectedProduct = product;
          }
        }
      }
      return selectedProduct;
    },
    getSelectedDatasetProducts(filter, thisArg) {
      var selectedProducts = new Set();
      for (const dataset of this._selectedDatasets) {
        var product = dataset.getProduct();
        if (!filter || filter.call(thisArg, product)) {
          selectedProducts.add(product);
        }
      }
      return selectedProducts;
    },
    getSelectedDatasetProductsCount(filter, thisArg) {
      return this.getSelectedDatasetProducts(filter, thisArg).size;
    },
    getSelectedDatasetsByProduct(filter, thisArg) {
      const selectedDatasetsByProduct = new Map();
      for (const dataset of this._selectedDatasets) {
        const product = dataset.getProduct();
        if (!filter || filter.call(thisArg, dataset)) {
          let datasets = selectedDatasetsByProduct.get(product);
          if (!datasets) {
            datasets = new Set();
            selectedDatasetsByProduct.set(product, datasets);
          }
          datasets.add(dataset);
        }
      }
      return selectedDatasetsByProduct;
    },
    forEachSelectedDataset(callback, thisArg) {
      this._selectedDatasets.forEach(callback, thisArg);
    },
    unselectDataset(dataset) {
      var deselected = this._deselectDataset(dataset);
      if (deselected) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return deselected;
    },
    unselectDatasets(datasets) {
      var changed = false;
      for (var l = 0; l < datasets.length; ++l) {
        var deselected = this._deselectDataset(datasets[l]);
        changed = changed || deselected;
      }
      if (changed) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return changed;
    },
    unselectAllDatasets() {
      var deselected = this._deselectAllDatasets();
      if (deselected) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return deselected;
    },
    unselectAllDatasetsExcept(dataset) {
      var deselected = this._deselectAllDatasetsExceptOne(dataset);
      if (deselected) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
      return deselected;
    },
  },
  /*Public - Dataset selection handler*/ {
    onDatasetSelectedChange(dataset) {
      dataset.handleSelectedChange();
      this.hub.emit('dataset:change:selected', dataset);
    },
  },
  /*Protected - Dataset selection*/ {
    _selectDataset(dataset) {
      if (
        dataset &&
        !this._selectedDatasets.has(dataset) &&
        this.datasetSelectionActive
      ) {
        this._selectedDatasets.add(dataset);
        this.onDatasetSelectedChange(dataset);
        return true;
      }
      return false;
    },
    _deselectDataset(dataset) {
      if (dataset && this._selectedDatasets.has(dataset)) {
        this._selectedDatasets.delete(dataset);
        this.onDatasetSelectedChange(dataset);
        return true;
      }
      return false;
    },
    _deselectAllDatasets() {
      var deselected = this._selectedDatasets.size > 0;
      this._selectedDatasets.forEach(this._deselectDataset, this);
      return deselected;
    },
    _deselectAllDatasetsExcept(exceptPredicate, thisArg) {
      var countBeforeDeselect = this._selectedDatasets.size;
      for (const dataset of this._selectedDatasets) {
        if (!exceptPredicate.call(thisArg, dataset)) {
          this._deselectDataset(dataset);
        }
      }
      return countBeforeDeselect > this._selectedDatasets.size;
    },
    _deselectAllDatasetsExceptOne(exception) {
      if (!exception) {
        return this._deselectAllDatasets();
      }
      return this._deselectAllDatasetsExcept(function(dataset) {
        return dataset === exception;
      });
    },
  },
  /*Protected - Dataset selection interactions*/ {
    handleDatasetClick(dataset) {
      var changed = false;

      if (this.syntool.keyboardManager.isKeyDown(this.multipleSelectKey)) {
        if (dataset) {
          changed = (
            dataset.isSelected()
            ? this._deselectDataset(dataset)
            : this._selectDataset(dataset)
          );
        }
      } else if (dataset) {
        if (dataset.isSelected()) {
          changed = (
            this.getSelectedDatasetsCount() === 1
            ? this._deselectDataset(dataset)
            : this._deselectAllDatasetsExceptOne(dataset)
          );
        } else {
          changed = this._deselectAllDatasets();
          changed = this._selectDataset(dataset) || changed;
        }
      } else {
        changed = this._deselectAllDatasets();
      }
      if (changed) {
        this.hub.emit('change:selection');
        this._invalidateState();
      }
    },
  },
]);
extendClass(Dataset, {
  isSelected() {
    return this.syntool.mapView.isDatasetSelected(this);
  },
});
