/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  extendClass,
  setGlobalCursor,
  disableTextSelection,
  enableTextSelection,
  MapView,
} from '../syntool.js';
import {disableImageDrag, enableImageDrag} from '../utils/dom_utils.js';
import {PanRecognizer, MIN_PAN_MOVEMENT} from './gesture_recognizers.js';
import './map_view_gesture_recognizers.js';

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      this._dragHolder = null;

      this._panRecognizer = new PanRecognizer('drag');
      this._panRecognizer.allowedMouseButtons = [0, 1];
      this._panRecognizer.hub
        .on('start', this.handleDragStart, this)
        .on('change', this.handleDragMove, this)
        .on('end', this.handleDragEnd, this)
        .on('cancel', this.handleDragCancel, this);
      this.addGestureRecognizer(this._panRecognizer);
    },
  },
  /*Public - Dragging*/ {
    setDragMovementThreshold(threshold) {
      this._panRecognizer.movementThreshold = threshold;
      this._oneFingerSingleTapRecognizer.setMaxTouchMovement(threshold);
      this._oneFingerSingleTapRecognizer.setMaxMouseMovement(threshold);
      this._oneFingerDoubleTapRecognizer.setMaxTouchMovement(threshold);
      this._oneFingerDoubleTapRecognizer.setMaxMouseMovement(threshold);
      this._twoFingersSingleTapRecognizer.setMaxTouchMovement(threshold);
      this._twoFingersSingleTapRecognizer.setMaxMouseMovement(threshold);
    },
    restoreDragMovementThreshold() {
      this.setDragMovementThreshold(MIN_PAN_MOVEMENT);
    },
    cancelDrag() {
      this._panRecognizer.cancel();
    },
  },
  /*Protected - Drag event handlers*/ {
    handleDragStart(event, recognizer) {
      disableTextSelection();
      disableImageDrag();

      var mapView = this;
      mapView._dragHolder = null;
      this.hub.emit('drag', recognizer, function captureDrag(dragHandler) {
        if (!mapView._dragHolder) {
          mapView._dragHolder = dragHandler;
          if (dragHandler.cursor) {
            setGlobalCursor(dragHandler.cursor);
          }
          return true;
        }
        return false;
      });
      this.hub.emit('drag:start', recognizer);
    },
    handleDragMove(event, recognizer) {
      if (this._dragHolder) {
        this._dragHolder.move.call(this._dragHolder.context, recognizer);
      }
      this.hub.emit('dragging', recognizer);
    },
    handleDragEnd(event, recognizer) {
      enableTextSelection();
      enableImageDrag();

      if (this._dragHolder) {
        this._dragHolder.end.call(this._dragHolder.context, recognizer);
        if (this._dragHolder.cursor) {
          setGlobalCursor(null);
        }
        this._dragHolder = null;
      }
      this.hub.emit('drag:end', recognizer);
    },
    handleDragCancel(event) {
      enableTextSelection();
      enableImageDrag();

      if (this._dragHolder) {
        this._dragHolder.cancel.call(this._dragHolder.context);
        if (this._dragHolder.cursor) {
          setGlobalCursor(null);
        }
        this._dragHolder = null;
      }
      this.hub.emit('drag:cancel');
    },
  },
]);
