/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  PASSIVE,
  PASSIVE_CAPTURING,
  extendClass,
  getXYDistance,
  preventEventDefault,
  getClientXYFromMouseEvent,
  MapView,
} from '../syntool.js';
import {isMouseEvent} from '../utils/dom_utils.js';
import {
  MAX_CLICK_MOVEMENT,
  getTouches,
  MultiTapRecognizer,
  PinchRecognizer,
} from './gesture_recognizers.js';

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      this._windowHasTrackingMouseListeners = false;
      this._elementsWithTrackingTouchListeners = new WeakSet();
      this._touchesState = {
        trackedTouches: [],
      };
      this._recognizers = [];
      this._lastMouseDownByButton = [];

      this.handleTouchStart = this.handleTouchStart.bind(this);
      this.handleTouchMove = this.handleTouchMove.bind(this);
      this.handleTouchEnd = this.handleTouchEnd.bind(this);
      this.handleTouchCancel = this.handleTouchCancel.bind(this);

      this.handleMouseDown = this.handleMouseDown.bind(this);
      this.handleMouseMove = this.handleMouseMove.bind(this);
      this.handleMouseUp = this.handleMouseUp.bind(this);

      // Prevent the generation of mouse events from touch events
      this.hub.on('touchstart', preventEventDefault);

      this.hub.on('touchstart', MapView.GESTURE, this.handleTouchStart);
      this.hub.on('touchmove', MapView.GESTURE, this.handleTouchMove);
      this.hub.on('touchend', MapView.GESTURE, this.handleTouchEnd);
      this.hub.on('touchcancel', MapView.GESTURE, this.handleTouchCancel);
      this.hub.on('mousedown', MapView.GESTURE, this.handleMouseDown);
      this.hub.on('mousemove', MapView.GESTURE, this.handleMouseMove);

      this._pinchRecognizer = new PinchRecognizer('pinch');
      this._oneFingerSingleTapRecognizer = new MultiTapRecognizer('tap-1');
      this._oneFingerDoubleTapRecognizer = new MultiTapRecognizer('tap-1x2');
      this._twoFingersSingleTapRecognizer = new MultiTapRecognizer('tap-2');

      this._recognizers.push(
        this._pinchRecognizer,
        this._oneFingerSingleTapRecognizer,
        this._oneFingerDoubleTapRecognizer,
        this._twoFingersSingleTapRecognizer
      );

      this._pinchRecognizer.hub.on('start', this.hub.emit.bind(this.hub, 'pinch:start'));
      this._pinchRecognizer.hub.on('change', this.hub.emit.bind(this.hub, 'pinch:change'));
      this._pinchRecognizer.hub.on('end', this.hub.emit.bind(this.hub, 'pinch:end'));
      this._pinchRecognizer.hub.on('cancel', this.hub.emit.bind(this.hub, 'pinch:cancel'));

      this._oneFingerDoubleTapRecognizer.numberOfTapsRequired = 2;
      this._twoFingersSingleTapRecognizer.numberOfTouchesRequired = 2;
      this._oneFingerSingleTapRecognizer.onRecognized = this.hub.emit.bind(this.hub, 'tap-1');
      this._oneFingerDoubleTapRecognizer.onRecognized = this.hub.emit.bind(this.hub, 'tap-1x2');
      this._twoFingersSingleTapRecognizer.onRecognized = this.hub.emit.bind(this.hub, 'tap-2');
    },
  },
  /*Public - Gesture recognizers*/ {
    addGestureRecognizer(recognizer) {
      if (this._recognizers.indexOf(recognizer) === -1) {
        this._recognizers.push(recognizer);
        return true;
      }
      return false;
    },
    removeGestureRecognizer(recognizer) {
      recognizer.cancel?.();
      var index = this._recognizers.indexOf(recognizer);
      if (index !== -1) {
        this._recognizers.splice(index, 1);
        return true;
      }
      return false;
    },
  },
  /*Protected - Gesture base event handlers*/ {
    addTrackingListeners(event, move, end, cancel) {
      if (isMouseEvent(event)) {
        if (!this._windowHasTrackingMouseListeners) {
          window.addEventListener('mousemove', move, PASSIVE_CAPTURING);
          window.addEventListener('mouseup', end, PASSIVE_CAPTURING);
          this._windowHasTrackingMouseListeners = true;
        }
        return;
      }

      /**
       * If the target element is removed from the document events won't
       * necessarily bubble up.
       * If there is any risk of an element being removed while it is being
       * touched, the best practice is to attach the touch listeners directly
       * to the target.
       */

      if (
        event.target === event.currentTarget ||
        this._elementsWithTrackingTouchListeners.has(event.target)
      ) {
        // We are already listening on the target
        return;
      }

      event.target.addEventListener('touchmove', move, PASSIVE);
      event.target.addEventListener('touchend', end, PASSIVE);
      event.target.addEventListener('touchcancel', cancel, PASSIVE);
      this._elementsWithTrackingTouchListeners.add(event.target);
    },
    removeTrackingListeners(event, move, end, cancel) {
      if (isMouseEvent(event)) {
        if (this._windowHasTrackingMouseListeners) {
          window.removeEventListener('mousemove', move, PASSIVE_CAPTURING);
          window.removeEventListener('mouseup', end, PASSIVE_CAPTURING);
          this._windowHasTrackingMouseListeners = false;
        }
        return;
      }

      if (
        event.target === event.currentTarget ||
        !this._elementsWithTrackingTouchListeners.has(event.target)
      ) {
        // No need to remove the listener on the target
        return;
      }

      event.target.removeEventListener('touchmove', move, PASSIVE);
      event.target.removeEventListener('touchend', end, PASSIVE);
      event.target.removeEventListener('touchcancel', cancel, PASSIVE);
      this._elementsWithTrackingTouchListeners.delete(event.target);
    },

    handleMouseDown(event) {
      this.addTrackingListeners(
        event,
        this.handleMouseMove,
        this.handleMouseUp
      );
      var clientXY = getClientXYFromMouseEvent(event);
      this._lastMouseDownByButton[event.button] = clientXY;
      for (const recognizer of this._recognizers) {
        recognizer.handleMouseDown?.(event);
      }
    },
    handleMouseMove(event) {
      for (const recognizer of this._recognizers) {
        recognizer.handleMouseMove?.(event);
      }
    },
    handleMouseUp(event) {
      if (event.buttons === 0) {
        this.removeTrackingListeners(
          event,
          this.handleMouseMove,
          this.handleMouseUp
        );
      }
      for (const recognizer of this._recognizers) {
        recognizer.handleMouseUp?.(event);
      }
    },

    handleTouchStart(event) {
      var touches = getTouches(
        event,
        this._touchesState.trackedTouches
      );
      this._touchesState.trackedTouches = touches.allTouches;

      this.addTrackingListeners(
        event,
        this.handleTouchMove,
        this.handleTouchEnd,
        this.handleTouchCancel
      );
      for (const recognizer of this._recognizers) {
        recognizer.handleTouchStart?.(event, touches);
      }
    },
    handleTouchMove(event) {
      var touches = getTouches(
        event,
        this._touchesState.trackedTouches
      );
      this._touchesState.trackedTouches = touches.allTouches;
      for (const recognizer of this._recognizers) {
        recognizer.handleTouchMove?.(event, touches);
      }
    },
    handleTouchEnd(event) {
      var touches = getTouches(
        event,
        this._touchesState.trackedTouches
      );
      this._touchesState.trackedTouches = touches.allTouches;
      if (event.targetTouches.length === 0) {
        this.removeTrackingListeners(
          event,
          this.handleTouchMove,
          this.handleTouchEnd,
          this.handleTouchCancel
        );
      }
      for (const recognizer of this._recognizers) {
        recognizer.handleTouchEnd?.(event, touches);
      }
    },
    handleTouchCancel(event) {
      var touches = getTouches(
        event,
        this._touchesState.trackedTouches
      );
      this._touchesState.trackedTouches = touches.allTouches;
      if (event.targetTouches.length === 0) {
        this.removeTrackingListeners(
          event,
          this.handleTouchMove,
          this.handleTouchEnd,
          this.handleTouchCancel
        );
      }
      for (const recognizer of this._recognizers) {
        recognizer.handleTouchCancel?.(event, touches);
      }
    },

    /**
     * The browser's mouse click event will fire even if the mouse has traveled
     * a long distance while the it's button was held down.
     * So we check to see it's a valid (a.k.a real) click.
     * This is only applicable to a MouseEvent, TouchEvents don't have this
     * problem, so this method ignores them and returns `true`.
     *
     * @param  {*} event The click event to check
     * @return {Boolean} Whether or not the event represents a real click.
     */
    isValidClick(event) {
      if (
        // If the event is not a MouseEvent, then it's a TouchEvent and we don't
        // need to worry about it
        event instanceof MouseEvent &&

        // If the event is a MouseEvent converted from a TouchEvent, then we
        // will not have had a mousedown event before
        this._lastMouseDownByButton[event.button]
      ) {
        const movement = getXYDistance(
          this._lastMouseDownByButton[event.button],
          getClientXYFromMouseEvent(event)
        );
        return movement < MAX_CLICK_MOVEMENT;
      }
      return true;
    },
  },
]);
