/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  Z_INDEX,
  extendClass,
  repeat,
  getXYLength,
  isXYEqual,
  Geometry,
  Feature,
  Geography,
  Quantity,
  MapView,
} from '../syntool.js';
import {OLVectorLayer} from '../utils/ol_vector_layer.js';
import {setFeatureVisibility} from '../utils/ol_utils.js';
import {MIN_PAN_MOVEMENT} from './gesture_recognizers.js';
import './map_view_gesture_recognizers.js';
import './map_view_dragging.js';

const MEASURING_RULER_NUM_SEGMENTS = 20;
const MEASURING_RULER_POINT_THRESHOLD = MIN_PAN_MOVEMENT;

function renderMeasuringRulerDistance(m) {
  const metricUnit = m < 1000 ? 'm' : 'km';
  const imperialUnit = m < 1000 ? 'ft' : 'NM';
  const distance = new Quantity(m, 'm');

  return (
    distance.convert(metricUnit).toString({exp: -2}) + '\n' +
    distance.convert(imperialUnit).toString({exp: -2})
  );
}
function renderMeasuringRulerBearing(bearing) {
  return bearing.toFixed(2) + '° N';
}

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      this._measuringRulerActive = false;
      this._measuringRulerLayer = null;
      this._measuringRulerLineFeature = null;
      this._measuringRulerDragStartViewportXY = null;
      this._measuringRulerStartLonLat = null;
      this._measuringRulerEndLonLat = null;
      this._measuringRulerDistance = 0;
      this._measuringRulerBearing = 0;

      this._stateHooks.push({
        get: ({
          measuringRulerActive,
          measuringRulerStartLonLat,
          measuringRulerEndLonLat,
        }) => ({
          measuringRulerActive,
          measuringRulerStartLonLat,
          measuringRulerEndLonLat,
        }),
        set: ({
          measuringRulerActive,
          measuringRulerStartLonLat,
          measuringRulerEndLonLat,
        }) => {
          this.measuringRulerActive = measuringRulerActive;
          this.setMeasuringRulerBetween(
            measuringRulerStartLonLat,
            measuringRulerEndLonLat
          );
        },
      });

      this.activateMeasuringRuler();
    },
  },
  /*Protected - Layer Z-Index*/ {
    getZIndexForLayer(layer, index) {
      if (layer === this._measuringRulerLayer) {
        return Z_INDEX.DRAWING_HANDLERS;
      }
      return window.original();
    },
  },
  /*Public - Measure ruler*/ {
    activateMeasuringRuler() {
      if (this._measuringRulerActive) {
        return false;
      }
      this._measuringRulerActive = true;
      this.syntool.keyboardManager.hub.on(
        'change',
        this.handleMeasuringRulerKeyChange,
        this
      );
      this.hub
        .on('drag', MapView.DRAW, this.handleMeasuringRulerDrag, this)
        .on('click', MapView.DRAW, this.handleMeasuringRulerClear, this)
        .on('tap-1', MapView.DRAW, this.handleMeasuringRulerClear, this);
      this.handleMeasuringRulerKeyChange();
      if (this._measuringRulerLayer) {
        this.addLayer(this._measuringRulerLayer);
      }
      this._updateMeasuringRulerGeometry();
      this.hub.emit('change:measuringRulerActive');
      this._invalidateState();
      return true;
    },
    deactivateMeasuringRuler() {
      if (!this._measuringRulerActive) {
        return false;
      }
      this.syntool.keyboardManager.hub.off(
        'change',
        this.handleMeasuringRulerKeyChange,
        this
      );
      this.hub
        .off('drag', MapView.DRAW, this.handleMeasuringRulerDrag, this)
        .off('click', MapView.DRAW, this.handleMeasuringRulerClear, this)
        .off('tap-1', MapView.DRAW, this.handleMeasuringRulerClear, this);
      if (this._isMeasuringRulerTriggered()) {
        this.setCursor(null);
      }
      if (this._measuringRulerLayer) {
        this.removeLayer(this._measuringRulerLayer);
      }
      if (this._measuringRulerDragStartViewportXY) {
        this.cancelDrag();
      }
      this._measuringRulerActive = false;
      this.hub.emit('change:measuringRulerActive');
      this._invalidateState();
      return true;
    },
    get measuringRulerActive() {
      return this._measuringRulerActive;
    },
    set measuringRulerActive(on) {
      if (on) {
        this.activateMeasuringRuler();
      } else {
        this.deactivateMeasuringRuler();
      }
    },
    get measuringRulerStartLonLat() {
      return this._measuringRulerStartLonLat;
    },
    set measuringRulerStartLonLat(lonlat) {
      if (this._measuringRulerStartLonLat !== lonlat) {
        this.setMeasuringRulerBetween(lonlat, this.measuringRulerEndLonLat);
      }
    },
    get measuringRulerEndLonLat() {
      return this._measuringRulerEndLonLat;
    },
    set measuringRulerEndLonLat(lonlat) {
      if (this._measuringRulerEndLonLat !== lonlat) {
        this.setMeasuringRulerBetween(this.measuringRulerStartLonLat, lonlat);
      }
    },
    get measuringRulerDistance() {
      return this._measuringRulerDistance;
    },
    get measuringRulerBearing() {
      return this._measuringRulerBearing;
    },

    setMeasuringRulerBetween(startLonLat, endLonLat) {
      startLonLat ||= endLonLat;
      endLonLat ||= startLonLat;
      if (
        !isXYEqual(this._measuringRulerStartLonLat, startLonLat) ||
        !isXYEqual(this._measuringRulerEndLonLat, endLonLat)
      ) {
        this._measuringRulerStartLonLat = startLonLat;
        this._measuringRulerEndLonLat = endLonLat;
        this._updateMeasuringRuler();
        this.hub.emit('change:measure');
        this._invalidateState();
      }
    },
  },
  /*Protected - Measure ruler*/ {
    _isMeasuringRulerTriggered() {
      return (
        this.syntool.keyboardManager.keysPressed.size === 1 &&
        this.syntool.keyboardManager.isShiftDown()
      );
    },
    handleMeasuringRulerKeyChange() {
      this.setCursor(this._isMeasuringRulerTriggered() ? 'crosshair' : null);
    },
    handleMeasuringRulerDrag(recognizer, captureDrag) {
      if (recognizer.event.button || !this._isMeasuringRulerTriggered()) {
        return true;
      }

      const capturedDrag = captureDrag({
        cursor : 'crosshair',
        context: this,
        move   : this.handleMeasuringRulerDragMove,
        end    : this.handleMeasuringRulerDragEnd,
        cancel : this.handleMeasuringRulerDragCancel,
      });

      if (capturedDrag) {
        this._measuringRulerDragStartViewportXY = (
          this.getViewportXYFromClientXY(recognizer.startLocation)
        );
        this._measuringRulerStartLonLat = (
          this.getLonLatFromViewportXY(this._measuringRulerDragStartViewportXY)
        );
        this._updateMeasuringRulerEndLocation(recognizer);
      }
      return !capturedDrag;
    },
    handleMeasuringRulerDragMove(recognizer) {
      this._updateMeasuringRulerEndLocation(recognizer);
    },
    handleMeasuringRulerDragEnd(recognizer) {
      this._updateMeasuringRulerEndLocation(recognizer);
      this._measuringRulerDragStartViewportXY = null;
      this.hub.emit('change:measure');
      this._invalidateState();
    },
    handleMeasuringRulerDragCancel() {
      this._measuringRulerDragStartViewportXY = null;
      this._measuringRulerStartLonLat = null;
      this._measuringRulerEndLonLat = null;
      this._updateMeasuringRuler();
      this.hub.emit('change:measure');
      this._invalidateState();
    },
    handleMeasuringRulerClear(event, recognizer) {
      if (!this._isMeasuringRulerTriggered()) {
        return true;
      }

      if (recognizer || this.isValidClick(event)) {
        this.setMeasuringRulerBetween(
          recognizer
          ? recognizer.startLocation
          : this.getLonLatFromMouseEvent(event)
        );
      }
      return false;
    },

    _updateMeasuringRulerEndLocation(recognizer) {
      if (
        getXYLength(recognizer.translation) > MEASURING_RULER_POINT_THRESHOLD
      ) {
        const clientXY = this.getClientXYFromViewportXY(
          this._measuringRulerDragStartViewportXY
        );
        clientXY[0] += recognizer.translation[0];
        clientXY[1] += recognizer.translation[1];
        this._measuringRulerEndLonLat = this.getLonLatFromClientXY(clientXY);
      } else {
        this._measuringRulerEndLonLat = this._measuringRulerStartLonLat;
      }
      this._updateMeasuringRuler();
      this.hub.emit('changing:measure');
    },
    _updateMeasuringRuler() {
      if (
        isXYEqual(
          this._measuringRulerStartLonLat,
          this._measuringRulerEndLonLat
        )
      ) {
        this._measuringRulerDistance = 0;
        this._measuringRulerBearing = 0;
      } else {
        const {distance, initialBearing} = (
          Geography.geodesicDistanceAndBearings(
            this._measuringRulerStartLonLat,
            this._measuringRulerEndLonLat
          )
        );
        this._measuringRulerDistance = distance;
        this._measuringRulerBearing = initialBearing;
      }
      if (this.measuringRulerActive) {
        this._updateMeasuringRulerGeometry();
      }
    },
    _updateMeasuringRulerGeometry() {
      if (!this._measuringRulerLayer) {
        this._measuringRulerLayer = new OLVectorLayer('MeasuringRuler', {
          style: {
            base: {
              strokeColor      : '#6CC',
              fillColor        : '#AAA',
              fillOpacity      : 0.4,
              pointRadius      : 0,
              strokeOpacity    : 1,
              strokeWidth      : 2,
              strokeDashstyle  : 'dashdot',
              fontColor        : '#fff',
              fontSize         : '12px',
              fontFamily       : 'Helvetica,Arial,sans-serif',
              fontWeight       : 'bold',
              labelAlign       : 'cb',
              labelXOffset     : 0,
              labelYOffset     : 20,
              labelOutlineColor: '#000',
              labelOutlineWidth: 2,
              pointerEvents    : 'none',
            },
          },
        });
        this._measuringRulerLineFeature = new Feature(
          Geometry.createPolylineFromXYs(
            repeat(MEASURING_RULER_NUM_SEGMENTS + 1, [0, 0])
          )
        );
        this._measuringRulerLayer.addFeature(this._measuringRulerLineFeature);
        this.addLayer(this._measuringRulerLayer);
      }

      const isVisible = this._measuringRulerDistance > 0;

      if (isVisible) {
        const lineGeometry = this._measuringRulerLineFeature.geometry;
        const startLonLat = this._measuringRulerStartLonLat;
        const distance = this._measuringRulerDistance;
        let bearing = this._measuringRulerBearing;
        const segmentsCount = MEASURING_RULER_NUM_SEGMENTS;
        const step = distance / segmentsCount;

        const lonLats = [startLonLat];
        Geometry.pointMoveTo(
          Geometry.getComponentAtIndex(lineGeometry, 0),
          this.getProjectedXYFromLonLat(startLonLat)
        );
        for (let i = 0; i < segmentsCount; i++) {
          const {lonLat, finalBearing} = (
            Geography.geodesicDestinationAndBearing(lonLats[i], step, bearing)
          );
          if (isNaN(finalBearing)) {
            lonLats.push(this._measuringRulerEndLonLat);
          } else {
            lonLats.push(lonLat);
            bearing = finalBearing;
          }

          Geometry.pointMoveTo(
            Geometry.getComponentAtIndex(lineGeometry, i + 1),
            this.getProjectedXYFromLonLat(lonLats[i + 1])
          );
        }

        this._measuringRulerLineFeature.style.label = (
          renderMeasuringRulerDistance(distance) + '\n' +
          renderMeasuringRulerBearing(this._measuringRulerBearing)
        );
      }

      if (
        !setFeatureVisibility(this._measuringRulerLineFeature, isVisible) &&
        isVisible
      ) {
        this._measuringRulerLayer.drawFeature(
          this._measuringRulerLineFeature,
          true /*force*/
        );
      }
    },
  },
]);
