/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  SECOND_IN_MILLISECONDS,
  emptyFunction,
  extendClass,
  HIDModifiers,
  renderKey,
  isMacOS,
  preventDefaultAndStopEvent,
  NotificationMessageType,
  MapView,
  HIDMouse,
} from '../syntool.js';
import {Gestures} from '../ui/utils.js';
import {Inertia} from './inertia.js';
import './map_view_gesture_recognizers.js';
import './map_view_dragging.js';

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      this._navigationActive = false;
      this._inertia = new Inertia();
      this._navigationLastTranslation = null;

      this.zoomWheelSensitivity = 1;
      this._zoomFraction = 0;
      this._zoomWheelTimeoutId = null;

      this.navigationPanKey = HIDModifiers.ctrlOrMeta;
      this._navigationPanMessage = {
        id: 'MapViewNavigation.panHelp',
        type: NotificationMessageType.info,
        hideAfter: -1,
        showOnce: true,
      };

      this._stateHooks.push({
        get: ({navigationActive}) => ({navigationActive}),
        set: ({navigationActive}) => {
          this.navigationActive = navigationActive;
        },
      });

      this.handleNavigationMouseWheel = Gestures.getWheelHandler(
        {direction: 'y'},
        this.handleNavigationMouseWheel,
        this
      );

      this.activateNavigation();
    },
  },
  /*Public - Navigation interactions*/ {
    activateNavigation() {
      if (this._navigationActive) {
        return false;
      }
      this._navigationActive = true;
      this.hub
        .on('mousedown', MapView.NAVIGATE, this.handleNavigationInteractionStart, this)
        .on('touchstart', MapView.NAVIGATE, this.handleNavigationInteractionStart, this)
        .on('wheel', MapView.NAVIGATE, this.handleNavigationInteractionStart, this)
        .on('tap-2', MapView.NAVIGATE, this.handleNavigationTap, this)
        .on('tap-1x2', MapView.NAVIGATE, this.handleNavigationDoubleTap, this)
        .on('dblclick', MapView.NAVIGATE, this.handleNavigationDoubleClick, this)
        .on('wheel', MapView.NAVIGATE, this.handleNavigationMouseWheel, this)
        .on('drag', MapView.NAVIGATE, this.handleNavigationDrag, this)
        .on('pinch:change', MapView.NAVIGATE, this.handleNavigationPinchChange, this)
        .on('pinch:end', MapView.NAVIGATE, this.handleNavigationPinchEnd, this)
        .on('pinch:cancel', MapView.NAVIGATE, this.handleNavigationPinchCancel, this);
      this.hub.emit('change:navigationActive');
      this._invalidateState();
      return true;
    },
    deactivateNavigation() {
      if (!this._navigationActive) {
        return false;
      }
      this.hub
        .off('mousedown', MapView.NAVIGATE, this.handleNavigationInteractionStart, this)
        .off('touchstart', MapView.NAVIGATE, this.handleNavigationInteractionStart, this)
        .off('wheel', MapView.NAVIGATE, this.handleNavigationInteractionStart, this)
        .off('tap-2', MapView.NAVIGATE, this.handleNavigationTap, this)
        .off('tap-1x2', MapView.NAVIGATE, this.handleNavigationDoubleTap, this)
        .off('dblclick', MapView.NAVIGATE, this.handleNavigationDoubleClick, this)
        .off('wheel', MapView.NAVIGATE, this.handleNavigationMouseWheel, this)
        .off('drag', MapView.NAVIGATE, this.handleNavigationDrag, this)
        .off('pinch:change', MapView.NAVIGATE, this.handleNavigationPinchChange, this)
        .off('pinch:end', MapView.NAVIGATE, this.handleNavigationPinchEnd, this)
        .off('pinch:cancel', MapView.NAVIGATE, this.handleNavigationPinchCancel, this);
      this._navigationActive = false;
      this.hub.emit('change:navigationActive');
      this._invalidateState();
      return true;
    },
    get navigationActive() {
      return this._navigationActive;
    },
    set navigationActive(on) {
      if (on) {
        this.activateNavigation();
      } else {
        this.deactivateNavigation();
      }
    },

    showPanningHelp() {
      const hasFinePointer = window.matchMedia('(any-pointer: fine)').matches;
      const hasCoarsePointer = window.matchMedia('(any-pointer: coarse)').matches;
      const waysToPanTheMap = [
        `Hold down ${renderKey(this.navigationPanKey)}`,
        hasFinePointer && renderKey(HIDMouse.middleMove),
        hasCoarsePointer && 'use two-finger gesture',
      ].filter(Boolean);
      const lastIndex = waysToPanTheMap.length - 1;
      this._navigationPanMessage.text = (
        waysToPanTheMap.slice(0, -1).join(', ') +
        (lastIndex > 0 ? ' or ' + waysToPanTheMap[lastIndex] : '') +
        ' to pan the map'
      );
      this.syntool.notificationCenter.showMessage(this._navigationPanMessage);
    },
    hidePanningHelp() {
      this._navigationPanMessage.showOnce = false;
      this.syntool.notificationCenter.closeMessage(this._navigationPanMessage.id);
      this._navigationPanMessage.showOnce = true;
    },
    isDragForPanning(recognizer) {
      return (
        recognizer.mouseButton === 1 ||
        recognizer.numberOfTouches > 1 ||
        this.syntool.keyboardManager.isKeyDown(this.navigationPanKey)
      );
    },
  },
  /*Protected - Navigation interactions*/ {
    _zoomByFraction(delta, anchorLonLat) {
      const zoom = this.getZoom() + this._zoomFraction + delta;
      const whole = Math.round(zoom);
      this.zoomTo(whole, anchorLonLat);
      this._zoomFraction = (whole === this.getZoom() ? (zoom - whole) : 0);
    },

    handleNavigationInteractionStart(event) {
      this._inertia.stopDecelerating();
    },
    handleNavigationTap(event, recognizer) {
      this.zoomOut();
    },
    handleNavigationDoubleTap(event, recognizer) {
      this.zoomIn(this.getLonLatFromClientXY(recognizer.location));
    },

    handleNavigationDoubleClick(event) {
      preventDefaultAndStopEvent(event);
      if (!this.isValidClick(event)) {
        return;
      }
      var anchorLonLat = this.getLonLatFromMouseEvent(event);
      if (this.syntool.keyboardManager.isKeyDown(HIDModifiers.ctrlOrMeta)) {
        this.zoomOut(anchorLonLat);
      } else {
        this.zoomIn(anchorLonLat);
      }
    },
    handleNavigationMouseWheel(event, delta) {
      clearTimeout(this._zoomWheelTimeoutId);

      if (isMacOS && event.deltaMode !== event.DOM_DELTA_PIXEL) {
        this._zoomFraction = 0;
        this._zoomWheelTimeoutId = null;
      } else {
        this._zoomWheelTimeoutId = (
          setTimeout(() => (this._zoomFraction = 0), SECOND_IN_MILLISECONDS)
        );
      }
      this._zoomByFraction(
        delta * this.zoomWheelSensitivity,
        this.getLonLatFromMouseEvent(event)
      );
    },
    handleNavigationDrag(recognizer, captureDrag) {
      var capturedDrag = captureDrag({
        cursor : 'grabbing',
        context: this,
        move   : this.handleNavigationDragMove,
        end    : this.handleNavigationDragEnd,
        cancel : emptyFunction,
      });

      if (capturedDrag) {
        this._inertia.begin();
        this._navigationLastTranslation = [0, 0];
        this.handleNavigationDragMove(recognizer);
      }
      return !capturedDrag;
    },
    handleNavigationDragMove(recognizer) {
      const translation = recognizer.getTranslationIn(this.node);
      this._inertia.update(translation[0], translation[1]);
      this.panBy(
        translation[0] - this._navigationLastTranslation[0],
        translation[1] - this._navigationLastTranslation[1],
        {dragging: true}
      );
      this._navigationLastTranslation = translation;
    },
    handleNavigationDragEnd(recognizer) {
      const translation = recognizer.getTranslationIn(this.node);
      this._inertia.update(translation[0], translation[1]);
      const dragging = this._inertia.apply((dx, dy, end) => {
        this.panBy(dx, dy, {dragging: !end});
      });
      this.panBy(
        translation[0] - this._navigationLastTranslation[0],
        translation[1] - this._navigationLastTranslation[1],
        {dragging}
      );
    },

    handleNavigationPinchChange(event, recognizer) {
      var containerTop = parseInt(this._map.layerContainerDiv.style.top, 10);
      var containerLeft = parseInt(this._map.layerContainerDiv.style.left, 10);
      var anchorClientXY = recognizer.location;
      var anchorViewportXY = this.getViewportXYFromClientXY(anchorClientXY);
      var anchorContainerX = anchorViewportXY[0] - containerLeft;
      var anchorContainerY = anchorViewportXY[1] - containerTop;

      this._map.layerContainerDiv.style.transformOrigin = (
        anchorContainerX + 'px ' + anchorContainerY + 'px'
      );
      this._map.layerContainerDiv.style.transform = (
        'scale(' + recognizer.scale + ')'
      );
    },
    handleNavigationPinchEnd(event, recognizer) {
      this._map.layerContainerDiv.style.transformOrigin = '';
      this._map.layerContainerDiv.style.transform = '';

      var anchorClientXY = recognizer.location;
      var anchorViewportXY = this.getViewportXYFromClientXY(anchorClientXY);
      var anchorProjectedXY = this.getProjectedXYFromViewportXY(anchorViewportXY);
      var deltaX = this.getViewportWidth() / 2 - anchorViewportXY[0];
      var deltaY = this.getViewportHeight() / 2 - anchorViewportXY[1];
      var requestedRes = this.getResolution() / recognizer.scale;
      var zoom = this.getZoomForResolution(requestedRes, true /*closest*/);
      var newRes = this.getResolutionForZoom(zoom);
      var newProjectedCenter = [
        anchorProjectedXY[0] + deltaX * newRes,
        anchorProjectedXY[1] - deltaY * newRes,
      ];
      this.setViewportCenterProjectedXY(newProjectedCenter, zoom);
    },
    handleNavigationPinchCancel(event) {
      this._map.layerContainerDiv.style.transform = '';
    },
  },
]);
