/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Z_INDEX, extendClass, MapView} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';
import {OLVectorLayer} from '../utils/ol_vector_layer.js';
import '../datasets/shape_dataset/syntool_filters.js';

Z_INDEX.SHAPES = Z_INDEX.OUTLINES + 1000;

const STYLE_MAP = {
  base: {
    strokeColor      : '#AAAAEE',
    strokeOpacity    : 1.0,
    strokeWidth      : 2,
    strokeDashstyle  : 'solid',
    fillColor        : '#AAAAAA',
    fillOpacity      : 0.4,
    pointRadius      : 10,
    fontFamily       : 'Helvetica,Arial,sans-serif',
    fontWeight       : 'bold',
    fontColor        : '#FFFFFF',
    fontSize         : 12,
    labelAlign       : 'lm',
    labelXOffset     : 15,
    labelYOffset     : 0,
    labelOutlineColor: '#000000',
    labelOutlineWidth: 2,
    labelSelect      : true,
    cursor           : 'pointer',
    lineHitWidth     : 10,
  },
  select: {
    strokeColor  : '#CA1D03',
    strokeOpacity: 1.0,
    fillColor    : '#CA1D03',
    fontColor    : '#CA1D03',
    opacity      : 1,
  },
  edit: {
    strokeColor   : '#CA1D03',
    strokeOpacity : 1.0,
    fillColor     : '#CA1D03',
    fillOpacity   : 0.4,
    fontColor     : '#CA1D03',
    opacity       : 1,
    __show_virtual: true,
  },
  temporary: {
    strokeColor: '#66CCCC',
    fillColor  : '#66CCCC',
    fillOpacity: 0.4,
    opacity    : 1,
    pointerEvents: 'none',
    labelSelect: false,
    lineHitWidth: 0,
  },
};

function centeredLabelInWKT(wkt) {
  const type = wkt.substring(0, wkt.indexOf('(')).trim().toUpperCase();
  return (type !== 'POINT' && type !== 'LINESTRING');
}
export function shapeStyleForIntent(product, shape, intent) {
  if (intent === 'base' && shape) {
    const style = Object.assign({}, STYLE_MAP.base, {
      label: shape.text || '',
    });
    const centeredLabel = (
      shape.type === Shape.ELLIPSE ||
      (shape.type === Shape.WKT && centeredLabelInWKT(shape.wkt))
    );

    if (centeredLabel) {
      Object.assign(style, {labelAlign: 'cm', labelXOffset: 0});
    }
    var productStyle = product.style;
    if (productStyle) {
      Object.assign(style, productStyle.ALL, productStyle[shape.type]);
    }
    if (shape.color) {
      Object.assign(style, {strokeColor: shape.color, fillColor: shape.color});
    }
    return style;
  }
  return STYLE_MAP[intent];
}

extendClass(MapView, [
  /*Protected - Extensions*/ {
    initialize() {
      window.original();

      var shapesLayer = new OLVectorLayer('Shapes', {
        style(feature, intent) {
          const {dataset} = feature;
          const style = shapeStyleForIntent(
            dataset.getProduct(),
            dataset.shownShape,
            intent
          );

          if (intent === 'base') {
            style.opacity = dataset.getOpacity();
          }
          return style;
        },
      });

      this.shapesLayer = shapesLayer;
      this.addLayer(shapesLayer);
    },
  },
  /*Protected - Layer Z-Index*/ {
    getZIndexForLayer(layer, index) {
      if (layer === this.shapesLayer) {
        return (
          (
            this.hasSelectedDatasets &&
            this.hasSelectedDatasets(this.syntool.filters.dataset.isShape)
          )
          ? Z_INDEX.EDITING_HANDLERS - 1
          : Z_INDEX.SHAPES
        );
      }
      return window.original();
    },
  },
  /*Protected - Dataset user selectability*/ {
    isDatasetUserSelectable(dataset) {
      if (!dataset) {
        return false;
      }
      if (window.original && window.original()) {
        return true;
      }
      return dataset.isVisible();
    },
  },
]);
