/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  shallowEqual,
  createClass,
  DOMEvents,
  Bounds,
  Geometry,
  Feature,
  MapView,
} from '../../syntool.js';
import {eventsListeningExtension} from '../../utils/events_listening_extension.js';

const CONTROL_POINT_BASE_STYLE = {
  strokeColor: '#CA1D03',
  fillColor  : '#CA1D03',
  fontColor  : '#CA1D03',
  fontOpacity: 1,
  opacity    : 1,
  label: undefined,
  labelAlign: 'cm',
  labelXOffset: 0,
};

export const Modify = createClass([
  eventsListeningExtension,
  /*Public*/ {
    constructor: function Modify(mapView, layer, options) {
      this.mapView = mapView;
      this.active = false;
      this.layer = layer;
      this.vertexRenderIntent = options.vertexRenderIntent;
      this.feature = null;
      this.controlPoints = {};
      this.controlPointFeatures = {};
      this.event = null;

      var center = null;
      var draggedFeature = null;
      var draggedControlPointKey = null;

      this.dragHandler = {
        selector: MapView.DRAW,
        cursor  : 'grabbing',
        dragging: false,
        context : this,

        start(recognizer, captureDrag) {
          if (this.mapView.datasetInteractionsActive === false) {
            return true;
          }

          var feature = (
            this.event &&
            (
              this.layer.getFeatureFromEvent(this.event) ||
              this.feature?.layer?.getFeatureFromEvent(this.event)
            )
          );
          var canDrag = (
            !recognizer.event.button &&
            recognizer.numberOfTouches < 2 &&
            this.mapView.datasetInteractionsActive &&
            feature &&
            (feature === this.feature || feature._controlPointKey)
          );

          if (canDrag && captureDrag(this.dragHandler)) {
            draggedFeature = feature;
            draggedControlPointKey = (
              draggedFeature === this.feature
              ? null
              : draggedFeature._controlPointKey
            );

            var newControlPointKey = this.onDragControlPointStart(
              draggedControlPointKey
            );
            if (newControlPointKey) {
              draggedFeature = this.controlPointFeatures[newControlPointKey];
              draggedControlPointKey = newControlPointKey;
            }
            if (draggedFeature) {
              center = Bounds.getCenter(
                Geometry.getBounds(draggedFeature.geometry)
              );
            }
            this.dragHandler.dragging = true;
            this.dragHandler.move.call(this, recognizer);
            return false;
          }

          draggedFeature = null;
          draggedControlPointKey = null;
          return true;
        },
        move(recognizer) {
          if (!draggedFeature) {
            return;
          }
          if (
            draggedFeature !== this.feature &&
            draggedControlPointKey !== draggedFeature._controlPointKey
          ) {
            this.onDragControlPointComplete(draggedControlPointKey);
            draggedFeature = null;
            return;
          }

          var translation = this.mapView.getProjectedVectorFromClientVector(
            recognizer.translation
          );
          var newControlPointKey = this.onDragControlPointMove(
            draggedControlPointKey,
            center[0] + translation[0],
            center[1] + translation[1]
          );

          if (newControlPointKey) {
            draggedFeature = this.controlPointFeatures[newControlPointKey];
            draggedControlPointKey = newControlPointKey;
            if (draggedFeature) {
              center = Bounds.getCenter(
                Geometry.getBounds(draggedFeature.geometry)
              );
            }
          }
        },
        end(recognizer) {
          this.dragHandler.dragging = false;

          if (draggedFeature) {
            this.onDragControlPointComplete(draggedControlPointKey);
          }
        },
        cancel() {
          this.dragHandler.dragging = false;

          if (draggedFeature) {
            this.onDragControlPointCancel(draggedControlPointKey);
          }
        },
      };
    },

    activate() {
      if (this.active) {
        return false;
      }
      this.active = true;
      this.listenToEvent(this.mapView.hub, 'touchstart+passive+capture', this.setEvent, this);
      this.listenToEvent(this.mapView.hub, 'mousedown+passive+capture', this.setEvent, this);
      this.listenToEvent(DOMEvents(document), 'mousemove+passive+capture', this.setEvent, this);
      this.listenToEvent(this.mapView.hub, 'click', this.dragHandler.selector, this.handleClick, this);
      this.listenToEvent(this.mapView.hub, 'tap-1', this.dragHandler.selector, this.handleClick, this);
      this.listenToEvent(this.mapView.hub, 'drag', this.dragHandler.selector, this.dragHandler.start, this);
      return true;
    },
    deactivate() {
      if (this.active) {
        this.active = false;
        this.stopAllListening();
        this.setControlPoints({});
        return true;
      }
      return false;
    },
    destroy() {
      this.mapView = null;
      this.layer = null;
    },

    setFeature(feature) {
      if (this.feature !== feature) {
        this.feature = feature;
        return true;
      }
      return false;
    },
    setControlPoints(controlPoints) {
      if (this.controlPoints !== controlPoints) {
        var prevControlPoints = this.controlPoints;
        this.controlPoints = controlPoints || {};
        this.resetControlPointFeatures(prevControlPoints);
      }
    },

    getControlPointKeyFromEvent(event = this.event) {
      const feature = (event && this.layer.getFeatureFromEvent(event));
      return feature?._controlPointKey || null;
    },
    getControlPointFromEvent(event) {
      const key = this.getControlPointKeyFromEvent(event);
      return key && this.controlPoints[key];
    },
    getHoveredControlPointKey() {
      return this.getControlPointKeyFromEvent(this.event);
    },
    getHoveredControlPoint() {
      return this.getControlPointFromEvent(this.event);
    },
    resetControlPointFeatures(prevControlPoints) {
      var featuresToAdd = [];
      var featuresToUpdate = [];
      var featuresToAddOnTop = [];
      var featuresToUpdateOnTop = [];
      var prevControlPoint;
      var controlPoint;
      var controlPointFeature;
      var reusableControlPointFeatures = [];
      var updatedControlPoints = {};
      var dataset = this.feature && this.feature.dataset;
      var hoveredControlPointKey = this.getHoveredControlPointKey();
      var key;
      var controlPointStyle;

      //eslint-disable-next-line guard-for-in
      for (key in this.controlPointFeatures) {
        prevControlPoint = prevControlPoints?.[key];
        controlPoint = this.controlPoints[key];
        controlPointFeature = this.controlPointFeatures[key];

        if (controlPoint) {
          controlPointStyle = (
            (
              prevControlPoint &&
              shallowEqual(prevControlPoint.style, controlPoint.style)
            )
            ? controlPointFeature.style
            : Object.assign({}, CONTROL_POINT_BASE_STYLE, controlPoint.style)
          );

          // control point still here => update
          controlPointFeature.dataset = dataset;
          if (
            this.updateControlPointFeature(
              controlPointFeature,
              controlPoint,
              controlPointStyle
            ) ||
            controlPoint.onTop
          ) {
            if (controlPoint.onTop) {
              featuresToUpdateOnTop.push(controlPointFeature);
            } else {
              featuresToUpdate.push(controlPointFeature);
            }
          }
          updatedControlPoints[key] = true;
        } else {
          // control point removed => mark feature for reuse
          controlPointFeature.dataset = null;
          reusableControlPointFeatures.push(controlPointFeature);
          delete this.controlPointFeatures[key];
        }
      }

      for (key in this.controlPoints) {
        if (updatedControlPoints[key]) {
          // control point already updated in the loop above
          continue;
        }

        controlPoint = this.controlPoints[key];
        controlPointStyle = (
          Object.assign({}, CONTROL_POINT_BASE_STYLE, controlPoint.style)
        );

        // new control point
        controlPointFeature = reusableControlPointFeatures.pop();
        if (controlPointFeature) {
          // recycle old feature
          this.updateControlPointFeature(
            controlPointFeature,
            controlPoint,
            controlPointStyle
          );
          if (controlPoint.onTop) {
            featuresToUpdateOnTop.push(controlPointFeature);
          } else {
            featuresToUpdate.push(controlPointFeature);
          }
        } else {
          // create new feature
          controlPointFeature = (
            this.createControlPointFeature(controlPoint, controlPointStyle)
          );
          if (controlPoint.onTop) {
            featuresToAddOnTop.push(controlPointFeature);
          } else {
            featuresToAdd.push(controlPointFeature);
          }
        }
        controlPointFeature.dataset = dataset;
        controlPointFeature._controlPointKey = key;
        this.controlPointFeatures[key] = controlPointFeature;
      }

      this.layer.removeFeatures(reusableControlPointFeatures);
      this.layer.addFeatures(featuresToAdd);
      this.layer.drawFeatures(featuresToUpdate, true /*force*/);
      this.layer.addFeatures(featuresToAddOnTop);
      this.layer.drawFeatures(featuresToUpdateOnTop, true /*force*/);

      // Since all the features have been reset we need to verify that the saved
      // event, used to get the hovered control point (via its target), still
      // corresponds to the same hovered control point, otherwise we reset it
      if (hoveredControlPointKey !== this.getHoveredControlPointKey()) {
        this.setEvent(null);
      }
    },
  },
  /*Protected*/ {
    onDragControlPointStart   : emptyFunction,
    onDragControlPointMove    : emptyFunction,
    onDragControlPointComplete: emptyFunction,
    onDragControlPointCancel  : emptyFunction,

    setEvent(event) {
      if (!event || event.type !== 'mousemove' || event.buttons === 0) {
        this.event = event;
      }
    },
    handleClick(event) {
      if (this.mapView.datasetInteractionsActive !== false && this.feature) {
        var feature = event && this.layer.getFeatureFromEvent(event);
        return !(feature && feature._controlPointKey);
      }
      return true;
    },

    createControlPointFeature(controlPoint, style) {
      const feature = new Feature(
        Geometry.createPoint(controlPoint.x, controlPoint.y),
        null,
        style
      );
      feature.renderIntent = this.vertexRenderIntent;
      return feature;
    },
    updateControlPointFeature(feature, controlPoint, style) {
      var changed = false;

      if (
        feature.geometry.x !== controlPoint.x ||
        feature.geometry.y !== controlPoint.y
      ) {
        feature.geometry.x = controlPoint.x;
        feature.geometry.y = controlPoint.y;
        feature.geometry.clearBounds();
        changed = true;
      }

      if (!shallowEqual(feature.style, style)) {
        feature.style = style;
        changed = true;
      }
      return changed;
    },
  },
]);
