/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Geometry} from '../../syntool.js';
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {TapRecognizer} from '../gesture_recognizers.js';
import {Modify} from './ol_modify_control.js';
import '../../datasets/shape_dataset/syntool_filters.js';
import '../map_view_gesture_recognizers.js';

export class OLModifyShape extends Modify {
  /*Public*/
  state = {};

  constructor(mapView, layer, options) {
    super(mapView, layer, options);

    this._tapToRemoveRecognizer = new TapRecognizer('OLModifyShape.tapToRemove');
    this._tapToRemoveRecognizer.allowedMouseButtons = [0];
    this._tapToRemoveRecognizer.onRecognized = this.handleTapToRemove.bind(this);
  }

  onChange() {
    /* empty */
  }

  activate() {
    if (!super.activate()) {
      return false;
    }
    const dataset = this.mapView.getSelectedDataset();
    this.setFeature(
      this.mapView.syntool.filters.dataset.isEditableShape(dataset)
      ? dataset.shapeFeature
      : null
    );
    this.mapView.addGestureRecognizer(this._tapToRemoveRecognizer);
    this.listenToEvent(
      this.mapView.hub,
      'change:selection',
      this.handleSelectionChange,
      this
    );
    this.listenToEvent(
      this.mapView.hub,
      'change:zoom',
      this.updateControlPoints,
      this
    );
    return true;
  }
  deactivate() {
    this.setFeature(null);
    this.mapView.removeGestureRecognizer(this._tapToRemoveRecognizer);
    return super.deactivate();
  }

  removeHoveredVertex() {
    if (!this.feature) {
      return false;
    }

    if (!this.dragHandler.dragging) {
      return this._removeVertexOfControlPoint(this.getHoveredControlPoint());
    }

    return false;
  }

  /*Protected*/

  handleSelectionChange() {
    const dataset = this.mapView.getSelectedDataset();
    this.setFeature(
      this.mapView.syntool.filters.dataset.isEditableShape(dataset)
      ? dataset.shapeFeature
      : null
    );
  }
  handleVisibilityChange() {
    this.layer.setVisibility(this.feature.dataset.isVisible());
  }
  handleTapToRemove(event, recognizer) {
    if (this.mapView.datasetInteractionsActive !== false && this.feature) {
      const feature = event && this.layer.getFeatureFromEvent(event);
      const controlPointKey = feature?._controlPointKey;
      if (controlPointKey) {
        this._removeVertexOfControlPoint(this.controlPoints[controlPointKey]);
      }
    }
  }

  setFeature(feature) {
    const dataset = feature?.dataset;
    const prevFeature = this.feature;
    if (super.setFeature(feature)) {
      if (prevFeature) {
        prevFeature.dataset.hub
          .off('change', this.updateControlPoints, this)
          .off('change:visibility', this.handleVisibilityChange, this);
      }
      this.state = {};
      if (dataset) {
        this.handleVisibilityChange();
        dataset.hub
          .on('change', this.updateControlPoints, this)
          .on('change:visibility', this.handleVisibilityChange, this);
        if (Geometry.isPoint(this.feature.geometry)) {
          this.feature.renderIntent = this.vertexRenderIntent;
          this.feature.layer.drawFeature(this.feature);
        }
      }
      this.updateControlPoints();
      return true;
    }
    return false;
  }

  updateControlPoints() {
    this.setControlPoints(
      this.feature && !Geometry.isPoint(this.feature.geometry)
      ? Shape.getControlPoints(
        this.feature.dataset.shownShape,
        this.state,
        this.feature.geometry,
        this.mapView.getResolution()
      )
      : null
    );
  }
  changesOnControlPointMove(key, x, y) {
    const shape = this.feature.dataset.shownShape;

    return Shape.changesOnControlPointMove(
      shape,
      key,
      x,
      y,
      this.controlPoints,
      this.state,
      this.feature.geometry,
      this.mapView.getResolution()
    );
  }

  onDragControlPointStart(key) {
    if (!this.feature) {
      return null;
    }

    const shape = this.feature.dataset.shownShape;
    Shape.controlPointDidStartDrag(
      shape,
      key,
      this.state,
      this.controlPoints,
      this.feature
    );

    if (key && this.controlPoints[key].virtual) {
      if (this.applyChangesToDataset({addPoint: this.controlPoints[key]})) {
        this.updateControlPoints();
        return key.substring(1);
      }
    }
    return null;
  }
  onDragControlPointMove(key, x, y) {
    if (!this.feature) {
      return;
    }

    const changes = this.changesOnControlPointMove(key, x, y);

    if (changes && this.applyChangesToDataset(changes)) {
      this.updateControlPoints();
    }
  }
  onDragControlPointComplete(key) {
    if (!this.feature) {
      return;
    }

    const shape = this.feature.dataset.shownShape;
    Shape.controlPointDidEndDrag(
      shape,
      key,
      this.state,
      this.controlPoints,
      this.feature
    );
    this.state = {};

    this.onChange(this.feature.dataset);
  }
  onDragControlPointCancel(key) {
    if (!this.feature) {
      return;
    }

    this.feature.dataset.updateShape(this.feature.dataset.shape);
    this.state = {};
  }

  _removeVertexOfControlPoint(controlPoint) {
    if (
      controlPoint &&
      !controlPoint.virtual &&
      this.applyChangesToDataset({removePoint: controlPoint})
    ) {
      this.onChange(this.feature.dataset);
      this.updateControlPoints();

      // Since the hovered control point has been removed, we reset the
      // saved event, used to get the hovered control point (via its target)
      // So that it's no longer considered hovered, which prevents chain
      // removal of vertices.
      this.setEvent(null);
      return true;
    }
    return false;
  }

  applyChangesToDataset(changes) {
    const dataset = this.feature.dataset;
    const shape = dataset.shownShape;
    const geometry = dataset.shapeFeature.geometry;
    const updates = Shape.updatesFromChanges(shape, changes, geometry);

    if (updates) {
      const updatedShape = Object.assign({}, shape, updates);

      if (!Shape.same(shape, updatedShape)) {
        dataset.updateShape(updatedShape);
        return true;
      }
    }
    return false;
  }
}
