/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  formatUTCDate,
  intersect,
  clamp,
  wrap,
  h,
  preventDefaultAndStopEvent,
  DOMEvents,
  Bounds,
  Geometry,
  Feature,
  Projection,
  Syntool,
} from '../syntool.js';
import {linkStyleSheet} from '../utils/dom_utils.js';
import {eventsListeningExtension} from '../utils/events_listening_extension.js';
import {OLVectorLayer} from '../utils/ol_vector_layer.js';
import {Dialog} from '../ui/components/dialog.js';

const dateTimePattern = [
  [
    '[12][0-9]{3}',
    '(?:0[0-9]|1[012])',
    '(?:[012][0-9]|3[01])',
  ].join('-'),
  [
    '(?:[01][0-9]|2[0-3])',
    '[0-5][0-9]',
    '[0-5][0-9]',
  ].join(':'),
].join('[ T]');
let spatialFilterLayer = null;

const toolDefaults = {
  linkedWithMap: true,
  geometry     : null,
  minDate      : null,
  maxDate      : null,
  withTime     : false,
  renderControls,
  unmountControls,
  validateInput,
  canGetFromMap,
};

function getSpatialFilterLayer() {
  if (!spatialFilterLayer) {
    spatialFilterLayer = new OLVectorLayer('ProcessingSpatialFilterHandler', {
      style: {
        base: {
          strokeColor  : '#FFFF00',
          strokeOpacity: 1.0,
          strokeWidth  : 2,
          fillColor    : '#FFFF00',
          fillOpacity  : 0.4,
          pointRadius  : 10,
          pointerEvents: 'none',
        },
      },
    });
    Syntool.mapView.addLayer(spatialFilterLayer);
  }
  return spatialFilterLayer;
}

function createBasicProcessingTool(tool) {
  tool = Object.assign({}, toolDefaults, tool, eventsListeningExtension);

  tool.renderControls = tool.renderControls.bind(tool);
  tool.unmountControls = tool.unmountControls.bind(tool);
  tool.validateInput = tool.validateInput.bind(tool);
  tool.canGetFromMap = tool.canGetFromMap.bind(tool);

  var execute = tool.execute;
  tool.execute = function() {
    if (!tool.validateInput()) {
      if (tool.withTime) {
        execute.call(tool, tool.geometry, tool.minDate, tool.maxDate);
      } else {
        execute.call(tool, tool.geometry);
      }
    }
  };

  return tool;
}
function renderControls(node) {
  var tool = this;
  var hub = DOMEvents(node);

  if (tool.linkedWithMap) {
    tool.geometry = getGeometryFromMapAs(tool.type);
  }
  tool.handleMeasureChange = handleMeasureChangeForTool.bind(null, tool, node);
  tool.updateGeometryFromMap = updateGeometryFromMapForTool.bind(null, tool, node);
  tool.updateSpatialControls = updateSpatialControlsForTool.bind(null, tool, node);
  tool.updateMapFeatures = updateMapFeaturesForTool.bind(null, tool, node);
  tool.updateExecuteButton = updateExecuteButtonForTool.bind(null, tool, node);
  tool.updateGeometry = updateGeometryForTool.bind(null, tool, node);
  tool.updateTime = updateTimeForTool.bind(null, tool, node);
  tool.validateDateField = validateDateFieldForTool.bind(null, tool, node);

  tool.listenToEvent(Syntool.mapView.hub, 'change:measuringRulerActive', tool.handleMeasureChange);
  tool.listenToEvent(Syntool.mapView.hub, 'changing:measure', tool.handleMeasureChange);
  tool.listenToEvent(Syntool.mapView.hub, 'change:measure', tool.handleMeasureChange);
  tool.listenToEvent(Syntool.mapView.hub, 'change:in-map-datasets', tool.updateExecuteButton);

  if (tool.withTime) {
    tool.updateTemporalControls = function() {
      tool.minDate = Syntool.timelineStore.minDate;
      tool.maxDate = Syntool.timelineStore.maxDate;
      updateTemporalControlsForTool(tool, node);
    };
    tool.listenToEvent(Syntool.timelineStore.hub, 'change:currentDate', tool.updateTemporalControls);
    tool.listenToEvent(Syntool.timelineStore.hub, 'change:timeframeLength', tool.updateTemporalControls);
  } else {
    tool.updateTemporalControls = emptyFunction;
  }

  node.innerHTML = renderControlsForTool(tool);
  updateControlsForTool(tool, node);

  switch (tool.type) {
  case 'POINT':
    tool.listenToEvent(hub, 'change', '.js-lon', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-lat', tool.updateGeometry);
    break;
  case 'LINE':
    tool.listenToEvent(hub, 'change', '.js-lon-start', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-lat-start', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-lon-end', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-lat-end', tool.updateGeometry);
    break;
  case 'BOX':
    tool.listenToEvent(hub, 'change', '.js-north', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-east', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-south', tool.updateGeometry);
    tool.listenToEvent(hub, 'change', '.js-west', tool.updateGeometry);
    break;
  default:
  }

  if (tool.withTime) {
    tool.listenToEvent(hub, 'input', '.js-start-date', tool.validateDateField);
    tool.listenToEvent(hub, 'input', '.js-end-date', tool.validateDateField);

    tool.listenToEvent(hub, 'change', '.js-start-date', tool.updateTime);
    tool.listenToEvent(hub, 'change', '.js-end-date', tool.updateTime);
  }

  // tool.listenToEvent(hub, 'click', '.js-getFromMap', function(event) {
  //   tool.linkedWithMap = !tool.linkedWithMap;
  //   updateGetFromMapButtonForTool(tool, node);
  //   const tip = node.querySelector('.js-spatial-tip');
  //   if (tip) {
  //     tip.hidden = !tool.linkedWithMap;
  //   }
  //   if (tool.linkedWithMap) {
  //     tool.updateGeometryFromMap();
  //   }
  // });
  tool.listenToEvent(hub, 'submit', '.js-form', function(event) {
    preventDefaultAndStopEvent(event);
    tool.execute();
  });
}
function unmountControls(node) {
  var layer = getSpatialFilterLayer();
  layer.destroyFeatures();

  this.geometry = null;
  this.minDate = null;
  this.maxDate = null;

  this.stopAllListening();

  this.handleMeasureChange = emptyFunction;
  this.updateGeometryFromMap = emptyFunction;
  this.updateSpatialControls = emptyFunction;
  this.updateMapFeatures = emptyFunction;
  this.updateTemporalControls = emptyFunction;
  this.updateExecuteButton = emptyFunction;
  this.updateGeometry = emptyFunction;
  this.updateTime = emptyFunction;
  this.validateDateField = emptyFunction;
}

// function renderGetFromMapButton() {
//   return h('button', {
//     class: 'UIButton ProcessingTools-Button --GetFromMap js-getFromMap',
//     type : 'button',
//     title: '[Un-]Link with measure tool',
//   }, h('i', {class: '-icon fa fa-map-marker'}));
// }
function renderNumberField(name, label, placeholder, before, after, scale) {
  return h('label', {class: 'ProcessingTools-Field -' + name}, [
    before && h('span', {class: 'ProcessingTools-FieldBeforeInput'}, before),
    h('input', {
      class      : 'ProcessingTools-FieldInput -number js-' + name,
      type       : 'number',
      title      : label,
      placeholder: placeholder,
      step       : 0.01,
      min        : isFinite(scale) && -scale,
      max        : isFinite(scale) && scale,
      required   : true,
    }),
    after && h('span', {class: 'ProcessingTools-FieldAfterInput'}, after),
  ]);
}
function renderPointControls() {
  return h('div', {class: 'ProcessingTools-PointControls'}, [
    // renderGetFromMapButton(),
    h('div', {class: 'ProcessingTools-Point'}),
    renderNumberField('lon', 'Longitude', 'Lon', '⬌', '°'),
    renderNumberField('lat', 'Latitude', 'Lat', '⬍', '°', 90),
  ]);
}
function renderLineControls() {
  return h('div', {class: 'ProcessingTools-LineControls'}, [
    // renderGetFromMapButton(),
    h('div', {class: 'ProcessingTools-LinePoint'}, [
      h('div', {class: 'ProcessingTools-Point -withRightLink'}),
      renderNumberField('lon-start', 'Longitude', 'Lon', '⬌', '°'),
      renderNumberField('lat-start', 'Latitude', 'Lat', '⬍', '°', 90),
    ]),
    h('div', {class: 'ProcessingTools-LinePoint'}, [
      h('div', {class: 'ProcessingTools-Point -withLeftLink'}),
      renderNumberField('lon-end', 'Longitude', 'Lon', '⬌', '°'),
      renderNumberField('lat-end', 'Latitude', 'Lat', '⬍', '°', 90),
    ]),
  ]);
}
function renderBoxControls() {
  return h('div', {class: 'ProcessingTools-BoxControls'}, [
    // renderGetFromMapButton(),
    renderNumberField('north', 'North', 'North', '', '°', 90),
    renderNumberField('west', 'West', 'West', '', '°'),
    renderNumberField('east', 'East', 'East', '', '°'),
    renderNumberField('south', 'South', 'South', '', '°', 90),
  ]);
}
function renderSpatialControlsTip(what) {
  return h('span', {class: 'ProcessingTools-Tip js-spatial-tip'}, [
    h('strong', 'Tip: '),
    `You can ${what} on the map while holding the shift key.`,
  ]);
}
function renderTimeControls() {
  return h('div', {class: 'ProcessingTools-TimeControls'}, [
    h('div', {class: 'ProcessingTools-TimeFields'}, [
      h('label', {class: 'ProcessingTools-Field'}, [
        h('input', {
          class      : 'ProcessingTools-FieldInput -time js-start-date',
          type       : 'text',
          title      : 'Start date: yyyy-mm-dd HH:MM:SS',
          placeholder: 'yyyy-mm-dd HH:MM:SS',
          size       : 19,
          pattern    : dateTimePattern,
          required   : true,
        }),
      ]),
      h('label', {class: 'ProcessingTools-Field'}, [
        h('input', {
          class      : 'ProcessingTools-FieldInput -time js-end-date',
          type       : 'text',
          title      : 'End date: yyyy-mm-dd HH:MM:SS',
          placeholder: 'yyyy-mm-dd HH:MM:SS',
          size       : 19,
          pattern    : dateTimePattern,
          required   : true,
        }),
      ]),
    ]),
    h('svg', {
      class  : 'ProcessingTools-TimeArrow',
      xmlns  : 'http://www.w3.org/2000/svg',
      viewBox: '0 0 90 130',
    }, [
      h('path', {
        d             : 'M65 5v80h20l-40 40l-40 -40h20v-80z',
        fill          : 'white',
        stroke        : 'black',
        'stroke-width': 3,
      }),
      h('g', {
        'font-size'  : 25,
        'text-anchor': 'middle',
        transform    : 'translate(45,32)',
      }, [
        h('text', {y: 0}, 'T'),
        h('text', {y: 25}, 'I'),
        h('text', {y: 50}, 'M'),
        h('text', {y: 75}, 'E'),
      ]),
    ]),
  ]);
}
function renderValidationMessage() {
  return h('span', {
    class: 'ProcessingTools-ValidationMessage js-validation-message',
  });
}
function renderExecuteButton() {
  return h('button', {
    class: 'UIButton ProcessingTools-Button --ExecuteButton js-execute',
  }, [h('i', {class: '-icon fa fa-gear'}), h('span', 'Execute')]);
}
function renderControlsForTool(tool) {
  var controls = [];

  switch (tool.type) {
  case 'POINT':
    controls.push(renderPointControls());
    controls.push(renderSpatialControlsTip('click'));
    break;
  case 'LINE':
    controls.push(renderLineControls());
    controls.push(renderSpatialControlsTip('drag'));
    break;
  case 'BOX':
    controls.push(renderBoxControls());
    controls.push(renderSpatialControlsTip('drag'));
    break;
  default:
  }

  if (tool.withTime) {
    controls.push(renderTimeControls());
  }
  controls.push(renderExecuteButton());
  controls.push(renderValidationMessage());

  return h('form', {class: 'js-form'}, controls);
}
function handleMeasureChangeForTool(tool, node) {
  // updateGetFromMapButtonForTool(tool, node);
  if (tool.linkedWithMap) {
    tool.updateGeometryFromMap();
  }
}
// function updateGetFromMapButtonForTool(tool, node) {
//   var button = node.querySelector('.js-getFromMap');
//   button.disabled = !tool.canGetFromMap();
//   button.classList.toggle('-active', tool.linkedWithMap);
// }
function updateGeometryFromMapForTool(tool, node) {
  tool.geometry = getGeometryFromMapAs(tool.type) ?? tool.geometry;
  tool.updateSpatialControls();
  tool.updateMapFeatures();
  tool.updateExecuteButton();
}
function updateSpatialControlsForTool(tool, node) {
  var geometry = tool.geometry;

  switch (tool.type) {
  case 'POINT':
    var formattedXY = (
      geometry && Geometry.isPoint(geometry)
      ? formatNumbers(Geometry.pointToXY(geometry))
      : ['', '']
    );
    node.querySelector('.js-lon').value = formattedXY[0];
    node.querySelector('.js-lat').value = formattedXY[1];
    break;
  case 'LINE':
    var formattedXYs = (
      geometry &&
      Geometry.isPolyline(geometry) &&
      Geometry.polylineCount(geometry) === 2
      ? Geometry.polylineToXYs(geometry).map(formatNumbers)
      : ['', '', '', '']
    );
    node.querySelector('.js-lon-start').value = formattedXYs[0][0];
    node.querySelector('.js-lat-start').value = formattedXYs[0][1];
    node.querySelector('.js-lon-end').value = formattedXYs[1][0];
    node.querySelector('.js-lat-end').value = formattedXYs[1][1];
    break;
  case 'BOX':
    var formattedBounds;
    if (geometry && Geometry.isPolygon(geometry)) {
      var bounds = Bounds.toLBRT(Geometry.getBounds(geometry));
      if (bounds[2] > 360) {
        // cross IDL
        bounds[2] -= 360;
      }
      formattedBounds = formatNumbers(bounds);
    } else {
      formattedBounds = ['', '', '', ''];
    }
    node.querySelector('.js-north').value = formattedBounds[3];
    node.querySelector('.js-south').value = formattedBounds[1];
    node.querySelector('.js-east').value = formattedBounds[2];
    node.querySelector('.js-west').value = formattedBounds[0];
    break;
  default:
  }
}
function updateMapFeaturesForTool(tool, node) {
  var geometry = tool.geometry;
  var layer = getSpatialFilterLayer();

  layer.destroyFeatures();

  if (!geometry) {
    return;
  }

  switch (tool.type) {
  case 'POINT':
    if (!Geometry.isPoint(geometry)) {
      return;
    }
    break;
  case 'LINE':
    if (
      !Geometry.isPolyline(geometry) ||
      Geometry.polylineCount(geometry) !== 2
    ) {
      return;
    }
    break;
  case 'BOX':
    if (!Geometry.isPolygon(geometry)) {
      return;
    }
    break;
  default:
    return;
  }

  layer.addFeature(new Feature(geometryToMap(geometry)));
}
function updateTemporalControlsForTool(tool, node) {
  node.querySelector('.js-start-date').value = formatDate(tool.minDate);
  node.querySelector('.js-end-date').value = formatDate(tool.maxDate);
}
function updateExecuteButtonForTool(tool, node) {
  var validity = tool.validateInput();
  var canExecute = !validity;

  node.querySelector('.js-validation-message').textContent = validity;
  node.querySelector('.js-execute').disabled = !canExecute;
}
function updateControlsForTool(tool, node) {
  // updateGetFromMapButtonForTool(tool, node);
  tool.updateSpatialControls();
  tool.updateMapFeatures();
  tool.updateTemporalControls();
  tool.updateExecuteButton();
}
function updateGeometryForTool(tool, node) {
  var lon0;
  var lat0;
  var lon1;
  var lat1;
  var latMin;
  var latMax;

  tool.geometry = null;
  switch (tool.type) {
  case 'POINT':
    lon0 = wrap(parseFloat(node.querySelector('.js-lon').value), -180, 180);
    lat0 = clamp(parseFloat(node.querySelector('.js-lat').value), -90, 90);
    if (!isNaN(lon0) && !isNaN(lat0)) {
      tool.geometry = Geometry.createPoint(lon0, lat0);
    }
    break;
  case 'LINE':
    lon0 = parseFloat(node.querySelector('.js-lon-start').value);
    lat0 = clamp(parseFloat(node.querySelector('.js-lat-start').value), -90, 90);
    lon1 = parseFloat(node.querySelector('.js-lon-end').value);
    lat1 = clamp(parseFloat(node.querySelector('.js-lat-end').value), -90, 90);
    if (!isNaN(lon0) && !isNaN(lat0) && !isNaN(lon1) && !isNaN(lat1)) {
      tool.geometry = Geometry.createPolylineFromXYs([
        [lon0, lat0],
        [lon1, lat1],
      ]);
    }
    break;
  case 'BOX':
    lon0 = parseFloat(node.querySelector('.js-west').value);
    lat0 = clamp(parseFloat(node.querySelector('.js-north').value), -90, 90);
    lon1 = parseFloat(node.querySelector('.js-east').value);
    lat1 = clamp(parseFloat(node.querySelector('.js-south').value), -90, 90);
    if (!isNaN(lon0) && !isNaN(lat0) && !isNaN(lon1) && !isNaN(lat1)) {
      latMin = Math.min(lat0, lat1);
      latMax = Math.max(lat0, lat1);
      if (lon0 > lon1) {
        // cross IDL
        lon1 += 360;
      }
      tool.geometry = Geometry.createPolygonFromXYs([
        [lon0, latMin],
        [lon1, latMin],
        [lon1, latMax],
        [lon0, latMax],
      ]);
    }
    break;
  default:
  }

  if (tool.geometry) {
    tool.updateSpatialControls();
  }
  tool.updateMapFeatures();
  tool.updateExecuteButton();
}
function updateTimeForTool(tool, node) {
  tool.minDate = dateFromField(node.querySelector('.js-start-date'));
  tool.maxDate = dateFromField(node.querySelector('.js-end-date'));
  tool.updateExecuteButton();
}
function validateDateFieldForTool(tool, node, event) {
  var field = event.target;

  field.setCustomValidity('');
  if (field.validity.valid && !validateDateString(field.value.trim())) {
    field.setCustomValidity('Invalid date!');
  }
}
function canGetFromMap() {
  return (
    Syntool.mapView.measuringRulerActive &&
    Syntool.mapView.measuringRulerStartLonLat &&
    (this.type === 'POINT' || Syntool.mapView.measuringRulerDistance > 0)
  );
}
function validateInput() {
  var timeValid = (
    !this.withTime ||
    (
      this.minDate &&
      this.maxDate &&
      !isNaN(this.minDate.getTime()) &&
      !isNaN(this.maxDate.getTime()) &&
      this.minDate.getTime() <= this.maxDate.getTime()
    )
  );
  var geometryValid;

  switch (this.type) {
  case 'POINT':
    geometryValid = (
      this.geometry &&
      Geometry.isPoint(this.geometry)
    );
    break;
  case 'LINE':
    geometryValid = (
      this.geometry &&
      Geometry.isPolyline(this.geometry) &&
      Geometry.polylineCount(this.geometry) === 2 &&
      Bounds.getPerimeter(Geometry.getBounds(this.geometry)) > 0
    );
    break;
  case 'BOX':
    geometryValid = (
      this.geometry &&
      Geometry.isPolygon(this.geometry) &&
      Bounds.getArea(Geometry.getBounds(this.geometry)) > 0
    );
    break;
  default:
    geometryValid = false;
  }

  if (!timeValid || !geometryValid) {
    return 'Invalid input!';
  }
  return '';
}

function formatNumber(number) {
  return number.toFixed(2);
}
function formatNumbers(numbers) {
  return numbers.map(formatNumber);
}
function formatDate(date) {
  return formatUTCDate(date, '$<date> $<time>');
}
function validateDateString(string) {
  var date = parseDate(string);

  return (date && formatDate(date) === string);
}
function parseDate(dateString) {
  var date = new Date(dateString + 'Z');

  if (isNaN(date.getTime())) {
    return null;
  }
  return date;
}
function dateFromField(field) {
  if (field.validity.valid) {
    return parseDate(field.value);
  }
  return null;
}
function getGeometryFromMapAs(type) {
  if (!Syntool.mapView.measuringRulerActive) {
    return null;
  }

  var start = Syntool.mapView.measuringRulerStartLonLat;
  var end = Syntool.mapView.measuringRulerEndLonLat;

  switch (type) {
  case 'POINT':
    return start ? Geometry.createPointFromXY(start) : null;
  case 'LINE':
    return (
      Syntool.mapView.measuringRulerDistance > 0
      ? Geometry.createPolylineFromXYs([start, end])
      : null
    );
  case 'BOX':
    return (
      Syntool.mapView.measuringRulerDistance > 0
      ? Geometry.fromBounds(Bounds.createFromXYs([start, end]))
      : null
    );
  default:
    return null;
  }
}

function geometryToMap(geometry) {
  return Projection.projectGeometry(Syntool.mapView.getProjection(), geometry);
}
function getMapGeometryFrom(geometry) {
  return geometryToMap(geometry);
}
function isDatasetIntersectingGeometry(geometry) {
  const geom = getMapGeometryFrom(geometry);
  return (dataset) => (
    Syntool.mapView.isGeometryIntersectingDataset(geom, dataset)
  );
}
function hasIntersectingDatasets(types, geometry) {
  return Syntool.mapView.hasDatasetsInMap(
    intersect(
      Syntool.filters.dataset.ofType(types),
      isDatasetIntersectingGeometry(geometry)
    )
  );
}
function getIntersectingDatasets(types, geometry) {
  return Syntool.mapView.getDatasetsInMap(
    intersect(
      Syntool.filters.dataset.ofType(types),
      isDatasetIntersectingGeometry(geometry)
    )
  );
}

linkStyleSheet('css/basic_processing_tool.css');
export const BasicProcessingTools = {
  dialog      : new Dialog('processing-dialog'),
  toolDefaults,
  createTool  : createBasicProcessingTool,

  getMapGeometryFrom,
  isDatasetIntersectingGeometry,
  hasIntersectingDatasets,
  getIntersectingDatasets,
};
Syntool.addControl(BasicProcessingTools.dialog);
window.ODL = Object.assign(window.ODL || {}, {BasicProcessingTools});
