/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  HIDModifiers,
  HIDKeys,
  H,
  removeAllChildrenOfNode,
  DOMEvents,
  Control,
} from '../syntool.js';
import {linkStyleSheet} from '../utils/dom_utils.js';

linkStyleSheet('css/processing_tools.css');

// type ProcessingTool = {
//   id          : string,
//   label       : string,
//   icon       ?: ?string,
//   description?: ?string,

//   // all functions are called on the tool (this === tool)
//   renderControls : (node: HTMLElement) => void,
//   unmountControls: (node: HTMLElement) => void,
// };

export class ProcessingToolsPanel extends Control {
  constructor(options) {
    super(options);

    this.tools = [];
    this.toolsById = new Map();
    this._selectedToolId = null;

    this.node = H('div', {class: 'UISidebar ProcessingTools'}, [
      this.toolbarNode = H('div', {class: 'UISegmentedControl ProcessingTools-Toolbar'}),
      this.descriptionNode = H('div', {class: 'ProcessingTools-Description'}),
      this.controlsNode = H('div', {class: 'ProcessingTools-Controls'}),
    ]);

    this.shortcuts = [
      {
        groupLabel : 'Global',
        id         : 'toggle-processing-tools',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.T,
        description: 'Toggle processing tools panel',
        handler    : () => {
          if (this.hidden) {
            this.active = false;
            this.active = true;
          } else {
            this.toggle();
          }
          return true;
        },
      },
    ];

    const self = this;
    DOMEvents(this.toolbarNode).on('change', 'input', function() {
      self.handleSelectToolWithId(this.value);
    });
  }

  get state() {
    const {selectedToolId} = this;
    return Object.assign(super.state, {selectedToolId});
  }
  set state(state) {
    super.state = state;
    this.selectedToolId = state.selectedToolId;
  }

  get active() {
    return super.active;
  }
  set active(active) {
    super.active = this.canActivate() && active;
  }

  get hidden() {
    return this.node.hidden;
  }
  set hidden(hidden) {
    if (this.canActivate()) {
      this.node.hidden = hidden;
    }
  }

  get selectedToolId() {
    return this._selectedToolId;
  }
  set selectedToolId(id) {
    this.selectToolWithId(id);
  }

  addTool(tool) {
    if (tool.id && !this.toolsById.has(tool.id)) {
      this.tools.push(tool);
      this.toolsById.set(tool.id, tool);
      this.toolbarNode.appendChild(
        H('label', {
          class: 'UISegmentedControl-Segment ProcessingTools-Tool',
          title: tool.icon ? tool.label : null,
        }, [
          H('input', {
            type : 'radio',
            name : 'ProcessingTools-Tool',
            value: tool.id,
          }),
          H('span', tool.icon ? H('i', {class: tool.icon}) : tool.label),
        ])
      );
      if (!this._selectedToolId) {
        this.selectToolWithId(tool.id);
      }
    }
  }
  addTools(tools) {
    tools.forEach(this.addTool, this);
  }
  canActivate() {
    return this._selectedToolId !== null;
  }
  toggle(active = !this.active) {
    this.active = active;
    return (this.active === !!active);
  }

  selectToolWithId(id) {
    if (!this.toolsById.has(id)) {
      return;
    }

    var selectedToolInput = this.toolbarNode.querySelector(':checked');
    if (selectedToolInput) {
      selectedToolInput.checked = false;
    }
    selectedToolInput = this.toolbarNode.querySelector('[value="' + id + '"]');
    if (selectedToolInput) {
      selectedToolInput.checked = true;
    }

    this.handleSelectToolWithId(id);
  }

  onAdd() {
    this.syntool.usableViewportNode.appendChild(this.node);
  }
  onRemove() {
    this.syntool.usableViewportNode.removeChild(this.node);
  }
  onActivate() {
    var tool = this.toolsById.get(this._selectedToolId);

    if (tool.description) {
      this.descriptionNode.innerHTML = tool.description;
      this.descriptionNode.style.display = '';
    } else {
      this.descriptionNode.style.display = 'none';
    }

    removeAllChildrenOfNode(this.controlsNode);
    tool.renderControls(this.controlsNode);
    this.controlsNode.style.display = (
      this.controlsNode.children.length === 0 ? 'none' : ''
    );

    this.node.classList.add('-active');
  }
  onDeactivate() {
    var tool = this.toolsById.get(this._selectedToolId);

    tool.unmountControls(this.controlsNode);
    this.node.classList.remove('-active');
  }

  handleSelectToolWithId(id) {
    if (this._selectedToolId === id) {
      return;
    }

    var prevTool = this.toolsById.get(this._selectedToolId);
    var tool = this.toolsById.get(id);

    this._selectedToolId = id;

    if (prevTool) {
      prevTool.unmountControls(this.controlsNode);
      removeAllChildrenOfNode(this.controlsNode);
    }

    this.node.setAttribute('data-tool-id', id);

    this.hub.emit('change:selection');
    this.hub.emit('change:state');
    if (!this._active) {
      return;
    }

    if (tool.description) {
      this.descriptionNode.innerHTML = tool.description;
      this.descriptionNode.style.display = '';
    } else {
      removeAllChildrenOfNode(this.descriptionNode);
      this.descriptionNode.style.display = 'none';
    }

    tool.renderControls(this.controlsNode);
    if (this.controlsNode.children.length === 0) {
      this.controlsNode.style.display = 'none';
    } else {
      this.controlsNode.style.display = '';
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProcessingToolsPanel});
