/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {RAD_IN_DEG, Geometry, Quantity, Types, Syntool} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';

var DEFAULT_ARROW_SIZE = 15000; /*15km*/
var NO_FILL_STYLE = {fillOpacity: 0};
var utils = Shape.utils;

Shape.addShapeType('ARROW', {
  typeChecker: Types.shape({
    points   : Types.arrayOf(Types.Point.isRequired, 2).isRequired,
    arrowSize: Types.number.gt(0),
  }),
  project(shape) {
    shape.points = utils.projectedPoints(shape.points);
    shape.arrowSize = utils.projectedDistance(shape.arrowSize);
  },
  unproject(shape) {
    shape.points = utils.unprojectedPoints(shape.points);
    shape.arrowSize = utils.unprojectedDistance(shape.arrowSize);
  },
  create(shape, props) {
    shape.points = utils.correctedPoints(props.points);
    shape.arrowSize = utils.correctedDistance(
      props.arrowSize || DEFAULT_ARROW_SIZE
    );
  },
  same(a, b) {
    return (
      a.arrowSize === b.arrowSize &&
      utils.sameArray(a.points, b.points, utils.samePoint)
    );
  },
  createGeometry(shape) {
    var angle = (
      Math.atan2(
        shape.points[1][1] - shape.points[0][1],
        shape.points[1][0] - shape.points[0][0]
      ) *
      RAD_IN_DEG + 90
    );
    var geometry = Geometry.createPolylineFromXYs(shape.points);
    var arrowHead = createArrowHead(shape.points[1], shape.arrowSize, angle);

    return Geometry.createCollection([geometry, arrowHead]);
  },

  getLength(shape) {
    var geometry = Geometry.createPolylineFromXYs(shape.points);
    var lengthInMeters = Geometry.getGeodesicLength(
      geometry,
      Syntool.mapView.getProjection()
    );
    return Quantity.convert(lengthInMeters, 'm', 'km');
  },
  getControlPoints(shape) {
    return {
      start: {x: shape.points[0][0], y: shape.points[0][1]},
      end  : {x: shape.points[1][0], y: shape.points[1][1]},
      move : {
        x: (shape.points[0][0] + shape.points[1][0]) / 2,
        y: (shape.points[0][1] + shape.points[1][1]) / 2,

        style: utils.styles.move,
      },
    };
  },
  changesOnControlPointMove(shape, key, x, y, controlPoints) {
    if (Syntool.keyboardManager.isShiftDown()) {
      var dx0 = shape.points[1][0] - shape.points[0][0];
      var dy0 = shape.points[1][1] - shape.points[0][1];
      var dx1 = x - shape.points[0][0];
      var dy1 = y - shape.points[0][1];
      var l0 = dx0 * dx0 + dy0 * dy0;
      var l1 = dx1 * dx1 + dy1 * dy1;
      var a0 = Math.atan2(dy0, dx0);
      var a1 = Math.atan2(dy1, dx1);
      var angle = Math.abs(a0 - a1);
      var scale = Math.cos(angle) * Math.sqrt(l1 / l0);

      x = shape.points[0][0] + dx0 * scale;
      y = shape.points[0][1] + dy0 * scale;
    }

    switch (key) {
    case null:
    case 'move':
      var center = [
        (shape.points[0][0] + shape.points[1][0]) / 2,
        (shape.points[0][1] + shape.points[1][1]) / 2,
      ];
      return {
        points: [
          [
            shape.points[0][0] + x - center[0],
            shape.points[0][1] + y - center[1],
          ],
          [
            shape.points[1][0] + x - center[0],
            shape.points[1][1] + y - center[1],
          ],
        ],
      };
    case 'start':
      return {points: [[x, y], shape.points[1]]};
    case 'end':
      return {points: [shape.points[0], [x, y]]};
    default:
      return null;
    }
  },
  updatesFromChanges(shape, changes) {
    return changes;
  },
});

function createArrowHead(xy, size, angle) {
  return Geometry.rotate(
    Geometry.createPolygonFromXYs([
      [xy[0] - size / 2, xy[1] + size],
      xy,
      [xy[0] + size / 2, xy[1] + size],
    ], NO_FILL_STYLE),
    angle,
    xy
  );
}
