/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DEG_IN_RAD,
  RAD_IN_DEG,
  Transform2D,
  Geometry,
  Projection,
  Quantity,
  Types,
  Syntool,
} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';
import {
  createBezierEllipsePath as createEllipse,
} from '../utils/bezier_ellipse.js';

var NUM_SEGMENTS = 20;
var DEFAULT_ARROW_SIZE = 15000; /*15km*/
var DEFAULT_ELLIPSE_ARROW = {
  directionFlipped: false,
  sideFlipped     : false,
  ratio           : 0.5,
  density         : 0.2,
  offset          : 0.125,
  size            : DEFAULT_ARROW_SIZE,
};
var utils = Shape.utils;

Shape.addShapeType('ELLIPSE', {
  hasArrow: true,

  typeChecker: Types.shape({
    center: Types.Point.isRequired,
    rx    : Types.number.gt(0).isRequired,
    ry    : Types.number.gt(0),
    angle : Types.number.neq(NaN),
    arrow : Shape.types.ShapeArrow,
  }),
  project(shape) {
    shape.center = utils.projectedPoint(shape.center);
    shape.rx = utils.projectedDistance(shape.rx);
    shape.ry = utils.projectedDistance(shape.ry);
    shape.arrow = utils.projectedArrow(shape.arrow);
  },
  unproject(shape) {
    shape.center = utils.unprojectedPoint(shape.center);
    shape.rx = utils.unprojectedDistance(shape.rx);
    shape.ry = utils.unprojectedDistance(shape.ry);
    shape.arrow = utils.unprojectedArrow(shape.arrow);
  },
  create(shape, props) {
    shape.center = utils.correctedPoint(props.center);
    shape.rx = utils.correctedDistance(props.rx);
    shape.ry = utils.correctedDistance(isFinite(props.ry) ? props.ry : props.rx);
    shape.arrow = utils.correctedArrow(props.arrow);
    shape.angle = isFinite(props.angle) ? props.angle : 0;
    if (props.arrow) {
      shape.arrow = Object.assign({}, DEFAULT_ELLIPSE_ARROW, props.arrow);
    } else {
      shape.arrow = null;
    }
  },
  same(a, b) {
    return (
      a.rx === b.rx &&
      a.ry === b.ry &&
      a.angle === b.angle &&
      utils.samePoint(a.center, b.center) &&
      utils.sameArrow(a.arrow, b.arrow)
    );
  },
  createGeometry(shape) {
    var geometry = Geometry.createPolygonFromBezierPath(
      createEllipse(shape.center[0], shape.center[1], shape.rx, shape.ry),
      NUM_SEGMENTS
    );

    if (shape.arrow) {
      var arrow = Object.assign({}, DEFAULT_ELLIPSE_ARROW, shape.arrow);
      var segmentsCount = NUM_SEGMENTS;
      var arrowHeadsCount = Math.floor(arrow.density * (segmentsCount - 1)) + 1;
      var arrowHeads = Array(arrowHeadsCount);
      var segmentOffset = Math.round(arrow.offset * (segmentsCount - 1));

      for (var i = 0; i < arrowHeadsCount; i++) {
        var segmentIndex = (
          (segmentOffset + Math.floor(i / arrowHeadsCount * segmentsCount)) %
          segmentsCount
        );
        var theta = 2 * Math.PI / segmentsCount;
        var alpha = segmentIndex * theta;
        var point = [
          shape.center[0] + shape.rx * Math.cos(alpha),
          shape.center[1] + shape.ry * Math.sin(alpha),
        ];
        var angle = RAD_IN_DEG * Math.atan2(
          -shape.rx * Math.sin(alpha) * Math.sin(theta / 2),
          -shape.ry * Math.cos(alpha) * Math.sin(theta / 2)
        );

        arrowHeads[i] = Geometry.rotate(
          utils.createDoubleArrowHeadGeometry(arrow, point, arrow.size),
          angle,
          point
        );
      }

      geometry = Geometry.createCollection([
        geometry,
        (
          arrowHeadsCount === 1
          ? arrowHeads[0]
          : Geometry.createCollection(arrowHeads)
        ),
      ]);
    }

    return Geometry.rotate(geometry, shape.angle, shape.center);
  },

  getLength(shape) {
    var geometry = Shape.createGeometry(Object.assign({}, shape, {arrow: null}));
    var lengthInMeters = Geometry.getGeodesicLength(
      geometry,
      Syntool.mapView.getProjection()
    );

    return Quantity.convert(lengthInMeters, 'm', 'km');
  },
  getArea(shape) {
    var geometry = Shape.createGeometry(Object.assign({}, shape, {arrow: null}));
    var areaInMeters2 = Geometry.getGeodesicArea(
      geometry,
      Syntool.mapView.getProjection()
    );

    return Quantity.convert(areaInMeters2, 'm2', 'km2');
  },
  getControlPoints(shape) {
    var x = shape.center[0];
    var y = shape.center[1];
    var rx = shape.rx;
    var ry = shape.ry;
    var angle = shape.angle;

    var transform = Transform2D.concat([
      Transform2D.translation(-x, -y),
      Transform2D.rotation(angle * DEG_IN_RAD),
      Transform2D.translation(x, y),
    ]);
    var right = Transform2D.transformPoint(transform, [x + rx, y]);
    var bottom = Transform2D.transformPoint(transform, [x, y - ry]);

    return {
      scaleX: {x: right[0], y: right[1]},
      scaleY: {x: bottom[0], y: bottom[1]},
    };
  },
  changesOnControlPointMove(shape, key, x, y, controlPoints) {
    var dx;
    var dy;
    var changes;

    switch (key) {
    case null:
      return {center: [x, y]};
    case 'scaleX':
      dx = x - shape.center[0];
      dy = y - shape.center[1];

      changes = {rx: Math.hypot(dx, dy)};
      if (!Syntool.keyboardManager.isShiftDown()) {
        changes.angle = (Math.atan2(dy, dx) * RAD_IN_DEG) % 360;
      }
      return changes;
    case 'scaleY':
      dx = x - shape.center[0];
      dy = y - shape.center[1];

      changes = {ry: Math.hypot(dx, dy)};
      if (!Syntool.keyboardManager.isShiftDown()) {
        changes.angle = (Math.atan2(dy, dx) * RAD_IN_DEG + 90) % 360;
      }
      return changes;
    default:
      return null;
    }
  },
  updatesFromChanges(shape, changes) {
    return changes;
  },

  getInitialArrowAtResolution(shape, resolution) {
    var centerInNorthernHemisphere = (
      Projection.unprojectXY(
        Syntool.mapView.getProjection(),
        shape.center
      )[1] > 0
    );

    return {
      directionFlipped: centerInNorthernHemisphere,
      sideFlipped     : true,
      size            : Shape.getArrowSizeForResolution(resolution),
    };
  },
});
