/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  getXYDistance,
  Transform2D,
  Bounds,
  Geometry,
  Quantity,
  Types,
  Syntool,
} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';

const MIN_VIRTUAL_SEGMENT_LENGTH = 40; /*px*/
const DEFAULT_ARROW_SIZE = 15000; /*15km*/
const DEFAULT_LINE_ARROW = {
  directionFlipped: false,
  sideFlipped     : false,
  ratio           : 0.5,
  density         : 0.5,
  offset          : 0.5,
  size            : DEFAULT_ARROW_SIZE,
};
const HIDDEN_STYLE = {display: 'none'};
const {utils} = Shape;

Shape.addShapeType('LINE', {
  hasArrow: true,

  typeChecker: Types.shape({
    points: Types.arrayOf(Types.Point.isRequired).isRequired,
    arrow : Shape.types.ShapeArrow,
  }),
  project(shape) {
    shape.points = utils.projectedPoints(shape.points);
    shape.arrow = utils.projectedArrow(shape.arrow);
  },
  unproject(shape) {
    shape.points = utils.unprojectedPoints(shape.points);
    shape.arrow = utils.unprojectedArrow(shape.arrow);
  },
  create(shape, props) {
    shape.points = props.points ? utils.correctedPoints(props.points) : [];
    shape.arrow = (
      props.arrow
      ? Object.assign({}, DEFAULT_LINE_ARROW, utils.correctedArrow(props.arrow))
      : null
    );
  },
  same(a, b) {
    return (
      utils.sameArrow(a.arrow, b.arrow) &&
      utils.sameArray(a.points, b.points, utils.samePoint)
    );
  },
  createGeometry(shape) {
    let geometry = Geometry.createPolylineFromXYs(shape.points);

    if (shape.arrow) {
      const arrowHeads = utils.createArrowHeadGeometries(
        Object.assign({}, DEFAULT_LINE_ARROW, shape.arrow),
        shape.points
      );

      geometry = Geometry.createCollection([geometry, arrowHeads]);
    }

    return geometry;
  },

  getLength(shape) {
    const geometry = (
      Shape.createGeometry(Object.assign({}, shape, {arrow: null}))
    );
    const lengthInMeters = Geometry.getGeodesicLength(
      geometry,
      Syntool.mapView.getProjection()
    );

    return Quantity.convert(lengthInMeters, 'm', 'km');
  },
  getControlPoints(shape, state, geometry, resolution) {
    if (state.lastShape === shape) {
      if (state.movePos || state.rotatePos) {
        state.lastControlPoints = Object.assign({}, state.lastControlPoints, {
          move: state.movePos,
          rotate: state.rotatePos,
        });
      } else if (state.lastResolution !== resolution) {
        const {rotate} = state.lastControlPoints;

        state.lastControlPoints = Object.assign({}, state.lastControlPoints, {
          rotate: Object.assign(rotate, {
            x: rotate._x + rotate.xOffset * resolution,
            y: rotate._y + rotate.yOffset * resolution,
          }),
        });
        updateVirtualControlPointsVisibility(state.lastControlPoints, resolution);
        state.lastResolution = resolution;
      }
      return state.lastControlPoints;
    }

    const points = shape.points;
    const len = points.length;

    let [minX, minY] = points[0];
    let [maxX, maxY] = points[0];

    for (const [x, y] of points) {
      if (x < minX) {
        minX = x;
      } else if (x > maxX) {
        maxX = x;
      }
      if (y < minY) {
        minY = y;
      } else if (y > maxY) {
        maxY = y;
      }
    }

    const midX = (minX + maxX) / 2;
    const midY = (minY + maxY) / 2;

    const controlPoints = {};
    controlPoints.move = (
      state.movePos || {
        x: midX,
        y: midY,

        style: utils.styles.move,
        onTop: true,
      }
    );

    const dx = maxX - controlPoints.move.x;
    const dy = minY - controlPoints.move.y;
    const d = utils.rotateOffset / Math.hypot(dx, dy);
    const xOffset = dx * d;
    const yOffset = dy * d;

    controlPoints.rotate = (
      state.rotatePos || {
        x: maxX + xOffset * resolution,
        y: minY + yOffset * resolution,

        _x: maxX,
        _y: minY,
        xOffset,
        yOffset,

        style: utils.styles.rotate,
        onTop: true,
      }
    );

    for (let i = 0; i < len; ++i) {
      controlPoints[i] = {
        x    : points[i][0],
        y    : points[i][1],
        index: i,
      };
    }

    // add virtual vertices in the middle of each edge
    for (let i = 0; i < len - 1; ++i) {
      const index = i + 1;
      if (
        points[i][0] === points[index][0] &&
        points[i][1] === points[index][1]
      ) {
        continue;
      }

      const x = (points[i][0] + points[index][0]) / 2;
      const y = (points[i][1] + points[index][1]) / 2;
      const segLen = getXYDistance(points[i], points[index]);
      const style = utils.styles.virtual;

      controlPoints['_' + index] = {x, y, segLen, virtual: true, style, index};
    }

    updateVirtualControlPointsVisibility(controlPoints, resolution);

    state.lastShape = shape;
    state.lastControlPoints = controlPoints;
    state.lastResolution = resolution;
    return controlPoints;
  },
  controlPointDidStartDrag(shape, key, state, controlPoints) {
    if (key === 'rotate') {
      state.movePos = Object.assign({}, controlPoints.move);
      state.rotatePos = Object.assign({}, controlPoints.rotate);
    }
  },
  controlPointDidEndDrag(shape, key, state) {
    if (key === 'rotate') {
      state.movePos = null;
      state.rotatePos = null;
    }
  },
  changesOnControlPointMove(shape, key, x, y, controlPoints, state, geometry, resolution) {
    switch (key) {
    case null: {
      const bounds = Geometry.getBounds(geometry);
      const centerX = Bounds.getCenterX(bounds);
      const centerY = Bounds.getCenterY(bounds);

      return {translate: [x - centerX, y - centerY]};
    }
    case 'move':
      return {translate: [x - controlPoints.move.x, y - controlPoints.move.y]};
    case 'rotate': {
      const dx0 = controlPoints.rotate.x - controlPoints.move.x;
      const dy0 = controlPoints.rotate.y - controlPoints.move.y;
      const dx1 = x - controlPoints.move.x;
      const dy1 = y - controlPoints.move.y;

      const angle = (
        Syntool.keyboardManager.isShiftDown()
        ? 0
        : Math.atan2(dy1, dx1) - Math.atan2(dy0, dx0)
      );

      const xOffset = controlPoints.rotate.x - controlPoints.rotate._x;
      const yOffset = controlPoints.rotate.y - controlPoints.rotate._y;
      const lOffset = Math.hypot(xOffset, yOffset);
      const lMin = utils.minScaleDistance * resolution;
      const l0 = Math.hypot(dx0, dy0) - lOffset;
      const l1 = Math.hypot(dx1, dy1) - lOffset;
      const scale = Math.max(lMin, l1) / l0;

      if (Syntool.keyboardManager.isShiftDown()) {
        x = controlPoints.move.x + (dx0 - xOffset) * scale + xOffset;
        y = controlPoints.move.y + (dy0 - yOffset) * scale + yOffset;
      } else if (l1 < lMin) {
        const lScale = (lMin + lOffset) / (l1 + lOffset);
        x = controlPoints.move.x + dx1 * lScale;
        y = controlPoints.move.y + dy1 * lScale;
      }

      state.rotatePos._x = x - xOffset;
      state.rotatePos._y = y - yOffset;
      state.rotatePos.x = x;
      state.rotatePos.y = y;
      return {
        origin: [controlPoints.move.x, controlPoints.move.y],
        scale : scale,
        rotate: angle,
      };
    }
    default:
      if (!controlPoints[key].virtual) {
        return {
          movePoint: {
            point: controlPoints[key],
            to   : [x, y],
          },
        };
      }
      return null;
    }
  },
  updatesFromChanges({points}, changes) {
    const {
      extend,
      movePoint,
      addPoint,
      removePoint,
      origin,
      rotate,
      scale,
      translate,
    } = changes;

    if (movePoint) {
      points = points.slice();
      points[movePoint.point.index] = movePoint.to;
    }
    if (extend) {
      if (
        extend.shape?.type === Shape.LINE &&
        extend.from &&
        !extend.from.virtual &&
        (extend.from.index === 0 || extend.from.index === points.length - 1) &&
        extend.shape.points.length > 0
      ) {
        const extensionPoints = extend.shape.points;

        // Remove first point which already exist in this shape
        extensionPoints.shift();

        if (extend.from.index === 0) {
          points = [...extensionPoints.reverse(), ...points];
        } else {
          points = [...points, ...extensionPoints];
        }
      }
    } else if (addPoint) {
      if (!movePoint) {
        points = points.slice();
      }
      points.splice(addPoint.index, 0, [addPoint.x, addPoint.y]);
    } else if (removePoint) {
      if (points.length > 2) {
        if (!movePoint) {
          points = points.slice();
        }
        points.splice(removePoint.index, 1);
      }
    }

    const transforms = [];
    if (origin) {
      transforms.push(Transform2D.translation(-origin[0], -origin[1]));
    }
    if (rotate != null && !Syntool.keyboardManager.isShiftDown()) {
      transforms.push(Transform2D.rotation(rotate));
    }
    if (scale != null) {
      transforms.push(Transform2D.scale(scale, scale));
    }
    if (origin) {
      transforms.push(Transform2D.translation(origin[0], origin[1]));
    }
    if (translate) {
      transforms.push(Transform2D.translation(translate[0], translate[1]));
    }

    if (transforms.length > 0) {
      points = (
        Transform2D.transformedPoints(Transform2D.concat(transforms), points)
      );
    }

    return {points};
  },

  getInitialArrowAtResolution(shape, resolution) {
    return {
      sideFlipped: 'none',
      size: Shape.getArrowSizeForResolution(resolution),
    };
  },
});

function updateVirtualControlPointsVisibility(controlPoints, resolution) {
  for (const [key, controlPoint] of Object.entries(controlPoints)) {
    if (controlPoint.virtual) {
      const style = (
        controlPoint.segLen / resolution < MIN_VIRTUAL_SEGMENT_LENGTH
        ? HIDDEN_STYLE
        : utils.styles.virtual
      );
      controlPoints[key] = Object.assign({}, controlPoint, {style});
    }
  }
}
