/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  RAD_IN_DEG,
  DEG_IN_RAD,
  clamp,
  modulo,
  getXYLength,
  Transform2D,
  Bounds,
  Geometry,
  Types,
  Syntool,
} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';
import {BEZIER_ELLIPSE_CONSTANT} from '../utils/bezier_ellipse.js';

const PI_2 = Math.PI / 2;

var NUM_SEGMENTS = 20;
var utils = Shape.utils;
var styles = {
  move      : utils.styles.move,
  rotate    : utils.styles.rotate,
  top       : {label: '2'},
  baseAnchor: {label: '5'},
  baseWidth : {label: '7'},
  baseHeight: {label: '6'},

  leftLobeCenter : {label: '1'},
  leftLobeRx     : {label: '3'},
  leftLobeRy     : {label: '4'},
  rightLobeCenter: {label: '1'},
  rightLobeRx    : {label: '3'},
  rightLobeRy    : {label: '4'},
};

Shape.addShapeType('MUSHROOM', {
  typeChecker: Types.shape({
    baseAnchor: Types.arrayOf(Types.number.gte(-1).lte(1).isRequired, 2),
    baseWidth : Types.number.gte(0).isRequired,
    baseHeight: Types.number.gte(0).isRequired,
    baseAngle : Types.number.gte(-45).lte(45),

    leftLobeCenter : Types.Point.isRequired,
    leftLobeRx     : Types.number.gte(0).isRequired,
    leftLobeRy     : Types.number.gte(0).isRequired,
    rightLobeCenter: Types.Point.isRequired,
    rightLobeRx    : Types.number.gte(0).isRequired,
    rightLobeRy    : Types.number.gte(0).isRequired,
  }),
  project(shape) {
    shape.baseWidth = utils.projectedDistance(shape.baseWidth);
    shape.baseHeight = utils.projectedDistance(shape.baseHeight);
    shape.leftLobeCenter = utils.projectedPoint(shape.leftLobeCenter);
    shape.leftLobeRx = utils.projectedDistance(shape.leftLobeRx);
    shape.leftLobeRy = utils.projectedDistance(shape.leftLobeRy);
    shape.rightLobeCenter = utils.projectedPoint(shape.rightLobeCenter);
    shape.rightLobeRx = utils.projectedDistance(shape.rightLobeRx);
    shape.rightLobeRy = utils.projectedDistance(shape.rightLobeRy);
  },
  unproject(shape) {
    shape.baseWidth = utils.unprojectedDistance(shape.baseWidth);
    shape.baseHeight = utils.unprojectedDistance(shape.baseHeight);
    shape.leftLobeCenter = utils.unprojectedPoint(shape.leftLobeCenter);
    shape.leftLobeRx = utils.unprojectedDistance(shape.leftLobeRx);
    shape.leftLobeRy = utils.unprojectedDistance(shape.leftLobeRy);
    shape.rightLobeCenter = utils.unprojectedPoint(shape.rightLobeCenter);
    shape.rightLobeRx = utils.unprojectedDistance(shape.rightLobeRx);
    shape.rightLobeRy = utils.unprojectedDistance(shape.rightLobeRy);
  },
  create(shape, props) {
    var centers = utils.correctedPoints([
      props.leftLobeCenter,
      props.rightLobeCenter,
    ]);
    shape.baseAnchor = (props.baseAnchor ? props.baseAnchor.slice() : [0, 1]);
    shape.baseWidth = utils.correctedDistance(props.baseWidth);
    shape.baseHeight = utils.correctedDistance(props.baseHeight);
    shape.baseAngle = props.baseAngle || 0;
    shape.leftLobeCenter = centers[0];
    shape.leftLobeRx = utils.correctedDistance(props.leftLobeRx);
    shape.leftLobeRy = utils.correctedDistance(props.leftLobeRy);
    shape.rightLobeCenter = centers[1];
    shape.rightLobeRx = utils.correctedDistance(props.rightLobeRx);
    shape.rightLobeRy = utils.correctedDistance(props.rightLobeRy);
  },
  same(a, b) {
    return (
      a.baseWidth   === b.baseWidth   &&
      a.baseHeight  === b.baseHeight  &&
      a.baseAngle   === b.baseAngle   &&
      a.leftLobeRx  === b.leftLobeRx  &&
      a.leftLobeRy  === b.leftLobeRy  &&
      a.rightLobeRx === b.rightLobeRx &&
      a.rightLobeRy === b.rightLobeRy &&
      utils.samePoint(a.baseAnchor, b.baseAnchor) &&
      utils.samePoint(a.leftLobeCenter,  b.leftLobeCenter) &&
      utils.samePoint(a.rightLobeCenter, b.rightLobeCenter)
    );
  },
  createGeometry(shape) {
    var baseWidth = shape.baseWidth;
    var baseHeight = shape.baseHeight;
    var baseAngle = shape.baseAngle;
    var headLeftCenterX = shape.leftLobeCenter[0];
    var headLeftCenterY = shape.leftLobeCenter[1];
    var headLeftRadiusX = shape.leftLobeRx;
    var headLeftRadiusY = shape.leftLobeRy;
    var headRightCenterX = shape.rightLobeCenter[0];
    var headRightCenterY = shape.rightLobeCenter[1];
    var headRightRadiusX = shape.rightLobeRx;
    var headRightRadiusY = shape.rightLobeRy;

    var headCenterX = (headLeftCenterX + headRightCenterX) / 2;
    var headCenterY = (headLeftCenterY + headRightCenterY) / 2;
    var headDistX = headRightCenterX - headLeftCenterX;
    var headDistY = headRightCenterY - headLeftCenterY;
    var headDist = Math.hypot(headDistX, headDistY);
    var headDist2 = headDist / 2;
    var headAngle = Math.atan2(headDistY, headDistX);

    var normalize = Transform2D.concat([
      Transform2D.translation(-headCenterX, -headCenterY),
      Transform2D.rotation(-headAngle),
    ]);
    var denormalize = Transform2D.inverse(normalize);
    var nLeftCenter = Transform2D.transformedPoint(normalize, [headLeftCenterX, headLeftCenterY]);
    var nRightCenter = Transform2D.transformedPoint(normalize, [headRightCenterX, headRightCenterY]);
    var nHeadPath = createMushroomHeadPath(
      nLeftCenter[0], nLeftCenter[1], headLeftRadiusX, headLeftRadiusY,
      nRightCenter[0], nRightCenter[1], headRightRadiusX, headRightRadiusY
    );

    var normalizeBase = Transform2D.concat([
      Transform2D.translation(-shape.baseAnchor[0] * headDist2, shape.baseAnchor[1] * headDist2),
      Transform2D.rotation(-baseAngle * DEG_IN_RAD),
    ]);
    var denormalizeBase = Transform2D.concat([
      Transform2D.inverse(normalizeBase),
      denormalize,
    ]);
    var nbLeftLegPoints = [
      [-baseWidth * 0.50, 0],
      [-baseWidth * 0.50, -baseHeight * 0.8],
      [-baseWidth * 0.75, -baseHeight],
    ];
    var nbRightLegPoints = [
      [baseWidth * 0.50, 0],
      [baseWidth * 0.50, -baseHeight * 0.8],
      [baseWidth * 0.75, -baseHeight],
    ];

    var headPath = Transform2D.transformedPath(denormalize, nHeadPath);
    var leftLegPoints = Transform2D.transformedPoints(denormalizeBase, nbLeftLegPoints);
    var rightLegPoints = Transform2D.transformedPoints(denormalizeBase, nbRightLegPoints);

    var head = Geometry.createPolygonFromBezierPath(headPath, NUM_SEGMENTS, {open: true});
    var leftLeg = Geometry.createPolylineFromXYs(leftLegPoints);
    var rightLeg = Geometry.createPolylineFromXYs(rightLegPoints);

    return Geometry.createCollection([leftLeg, rightLeg, head]);
  },

  getControlPoints(shape) {
    var leftCenterX = shape.leftLobeCenter[0];
    var leftCenterY = shape.leftLobeCenter[1];
    var rightCenterX = shape.rightLobeCenter[0];
    var rightCenterY = shape.rightLobeCenter[1];
    var headCenterX = (leftCenterX + rightCenterX) / 2;
    var headCenterY = (leftCenterY + rightCenterY) / 2;
    var headDistX = rightCenterX - leftCenterX;
    var headDistY = rightCenterY - leftCenterY;
    var headDist = Math.hypot(headDistX, headDistY);
    var headDist2 = headDist / 2;
    var headAngle = Math.atan2(headDistY, headDistX);
    var leftRx  = shape.leftLobeRx;
    var leftRy  = shape.leftLobeRy;
    var rightRx = shape.rightLobeRx;
    var rightRy = shape.rightLobeRy;

    var fromHead = Transform2D.concat([
      Transform2D.rotation(headAngle),
      Transform2D.translation(headCenterX, headCenterY),
    ]);
    var fromBase = Transform2D.concat([
      Transform2D.rotation(shape.baseAngle * DEG_IN_RAD),
      Transform2D.translation(shape.baseAnchor[0] * headDist2, -shape.baseAnchor[1] * headDist2),
      fromHead,
    ]);

    var move;
    var rotate;
    var top;
    var baseAnchor;
    var baseWidth;
    var baseHeight;
    var leftLobeRx;
    var leftLobeRy;
    var rightLobeRx;
    var rightLobeRy;

    var A0 = (leftRx * leftRy) || 1;
    var A1 = (rightRx * rightRy) || 1;
    var A = A0 + A1;
    var f = (A1 - A0) / A;

    move = [headCenterX, headCenterY];
    rotate = Transform2D.transformedPoint(fromHead, [headDist * 3 / 4, -shape.baseHeight]);
    top = Transform2D.transformedPoint(fromHead, [f * headDist2, Math.sqrt(A / 2)]);
    baseAnchor = Transform2D.transformedPoint(fromBase, [0, 0]);
    baseWidth = Transform2D.transformedPoint(fromBase, [shape.baseWidth / 2, -0.3 * shape.baseHeight]);
    baseHeight = Transform2D.transformedPoint(fromBase, [0, -shape.baseHeight]);
    leftLobeRx = Transform2D.transformedPoint(fromHead, [-leftRx - headDist2, 0]);
    leftLobeRy = Transform2D.transformedPoint(fromHead, [-headDist2, -leftRy]);
    rightLobeRx = Transform2D.transformedPoint(fromHead, [rightRx + headDist2, 0]);
    rightLobeRy = Transform2D.transformedPoint(fromHead, [headDist2, -rightRy]);

    return {
      move      : {x: move[0], y: move[1], style: styles.move, onTop: true},
      rotate    : {x: rotate[0], y: rotate[1], style: styles.rotate, onTop: true},
      top       : {x: top[0], y: top[1], style: styles.top},
      baseAnchor: {x: baseAnchor[0], y: baseAnchor[1], style: styles.baseAnchor},
      baseWidth : {x: baseWidth[0], y: baseWidth[1], style: styles.baseWidth},
      baseHeight: {x: baseHeight[0], y: baseHeight[1], style: styles.baseHeight},

      leftLobeCenter : {x: leftCenterX, y: leftCenterY, style: styles.leftLobeCenter},
      leftLobeRx     : {x: leftLobeRx[0], y: leftLobeRx[1], style: styles.leftLobeRx},
      leftLobeRy     : {x: leftLobeRy[0], y: leftLobeRy[1], style: styles.leftLobeRy},
      rightLobeCenter: {x: rightCenterX, y: rightCenterY, style: styles.rightLobeCenter},
      rightLobeRx    : {x: rightLobeRx[0], y: rightLobeRx[1], style: styles.rightLobeRx},
      rightLobeRy    : {x: rightLobeRy[0], y: rightLobeRy[1], style: styles.rightLobeRy},
    };
  },
  changesOnControlPointMove(shape, key, x, y, controlPoints, state, geometry) {
    var leftCenterX = shape.leftLobeCenter[0];
    var leftCenterY = shape.leftLobeCenter[1];
    var rightCenterX = shape.rightLobeCenter[0];
    var rightCenterY = shape.rightLobeCenter[1];
    var headCenterX = (leftCenterX + rightCenterX) / 2;
    var headCenterY = (leftCenterY + rightCenterY) / 2;
    var headDistX = rightCenterX - leftCenterX;
    var headDistY = rightCenterY - leftCenterY;
    var headDist = Math.hypot(headDistX, headDistY);
    var headDist2 = headDist / 2;
    var headAngle = Math.atan2(headDistY, headDistX);

    var toHead = Transform2D.concat([
      Transform2D.translation(-headCenterX, -headCenterY),
      Transform2D.rotation(-headAngle),
    ]);
    var toBase = Transform2D.concat([
      toHead,
      Transform2D.translation(-shape.baseAnchor[0] * headDist2, shape.baseAnchor[1] * headDist2),
      Transform2D.rotation(-shape.baseAngle * DEG_IN_RAD),
    ]);
    var fromHead = Transform2D.inverse(toHead);
    var localXY;

    var baseHeight;
    var baseAngle;
    var leftLobeRx;
    var leftLobeRy;
    var rightLobeRx;
    var rightLobeRy;
    var dx0;
    var dy0;
    var dx1;
    var dy1;
    var dx;
    var dy;
    var a0;
    var a1;
    var angle;
    var l0;
    var l1;
    var l;
    var scale;
    var transform;
    var f;
    var A0;
    var A1;
    var s;
    var baseAnchor;

    switch (key) {
    case null:
      var bounds = Geometry.getBounds(geometry);
      var centerX = Bounds.getCenterX(bounds);
      var centerY = Bounds.getCenterY(bounds);
      return {
        leftLobeCenter: [
          x - centerX + leftCenterX,
          y - centerY + leftCenterY,
        ],
        rightLobeCenter: [
          x - centerX + rightCenterX,
          y - centerY + rightCenterY,
        ],
      };
    case 'move':
      return {
        leftLobeCenter: [
          x - controlPoints.move.x + leftCenterX,
          y - controlPoints.move.y + leftCenterY,
        ],
        rightLobeCenter: [
          x - controlPoints.move.x + rightCenterX,
          y - controlPoints.move.y + rightCenterY,
        ],
      };
    case 'rotate':
      dx0 = controlPoints[key].x - controlPoints.move.x;
      dy0 = controlPoints[key].y - controlPoints.move.y;
      dx1 = x - controlPoints.move.x;
      dy1 = y - controlPoints.move.y;

      if (Syntool.keyboardManager.isShiftDown()) {
        angle = 0;
      } else {
        a0 = Math.atan2(dy0, dx0);
        a1 = Math.atan2(dy1, dx1);
        angle = a1 - a0;
      }

      l0 = (dx0 * dx0) + (dy0 * dy0);
      l1 = (dx1 * dx1) + (dy1 * dy1);
      scale = Math.sqrt(l1 / l0);

      transform = Transform2D.concat([
        Transform2D.translation(-headCenterX, -headCenterY),
        Transform2D.rotation(angle),
        Transform2D.scale(scale, scale),
        Transform2D.translation(headCenterX, headCenterY),
      ]);

      return {
        baseWidth      : shape.baseWidth * scale,
        baseHeight     : shape.baseHeight * scale,
        leftLobeCenter : Transform2D.transformedPoint(transform, shape.leftLobeCenter),
        leftLobeRx     : shape.leftLobeRx * scale,
        leftLobeRy     : shape.leftLobeRy * scale,
        rightLobeCenter: Transform2D.transformedPoint(transform, shape.rightLobeCenter),
        rightLobeRx    : shape.rightLobeRx * scale,
        rightLobeRy    : shape.rightLobeRy * scale,
      };
    case 'top':
      localXY = Transform2D.transformedPoint(toHead, [x, y]);

      leftLobeRx  = shape.leftLobeRx;
      leftLobeRy  = shape.leftLobeRy;
      rightLobeRx = shape.rightLobeRx;
      rightLobeRy = shape.rightLobeRy;
      f = headDist ? clamp(localXY[0] / (headDist / 2), -1, 1) : 0;
      A0 = (leftLobeRx * leftLobeRy) || 1;
      A1 = (rightLobeRx * rightLobeRy) || 1;
      s = Math.max(localXY[1], 0);

      if (Syntool.keyboardManager.isShiftDown()) {
        s = Math.sqrt((A0 + A1) / 2);
      }

      a0 = s * Math.sqrt((1 - f) / A0);
      a1 = s * Math.sqrt((1 + f) / A1);

      if (leftLobeRx === 0 && leftLobeRy === 0) {
        leftLobeRx = a0;
        leftLobeRy = a0;
      } else if (leftLobeRx === 0) {
        leftLobeRx = a0;
        leftLobeRy *= a0;
      } else {
        leftLobeRx *= a0;
        leftLobeRy *= a0;
      }
      if (rightLobeRx === 0 && rightLobeRy === 0) {
        rightLobeRx = a1;
        rightLobeRy = a1;
      } else if (rightLobeRx === 0) {
        rightLobeRx = a1;
        rightLobeRy *= a1;
      } else {
        rightLobeRx *= a1;
        rightLobeRy *= a1;
      }

      return {
        leftLobeRx : leftLobeRx,
        leftLobeRy : leftLobeRy,
        rightLobeRx: rightLobeRx,
        rightLobeRy: rightLobeRy,
      };
    case 'leftLobeCenter':
      if (Syntool.keyboardManager.isShiftDown()) {
        localXY = Transform2D.transformedPoint(toHead, [x, y]);
        return {
          leftLobeCenter: Transform2D.transformedPoint(fromHead, [localXY[0], 0]),
        };
      } else {
        return {
          leftLobeCenter: [x, y],
        };
      }
    case 'leftLobeRx':
      localXY = Transform2D.transformedPoint(toHead, [x, y]);
      if (Syntool.keyboardManager.isShiftDown()) {
        return {
          leftLobeRx: Math.max(0, -localXY[0] - headDist / 2),
        };
      } else {
        dx = localXY[0] - headDist / 2;
        dy = localXY[1];
        l = Math.hypot(dx, dy);

        return {
          leftLobeCenter: Transform2D.transformedPoint(
            fromHead,
            [headDist * (dx / l + 0.5), headDist * dy / l]
          ),
          leftLobeRx: Math.max(0, l - headDist),
          leftLobeRy: Math.max(0, shape.leftLobeRy + headDist * dy / l),
        };
      }
    case 'leftLobeRy':
      localXY = Transform2D.transformedPoint(toHead, [x, y]);
      if (Syntool.keyboardManager.isShiftDown()) {
        return {
          leftLobeRy: Math.max(0, -localXY[1]),
        };
      } else {
        return {
          leftLobeCenter: Transform2D.transformedPoint(fromHead, [localXY[0], 0]),
          leftLobeRx    : Math.max(0, shape.leftLobeRx + localXY[0] + headDist / 2),
          leftLobeRy    : Math.max(0, -localXY[1]),
        };
      }
    case 'rightLobeCenter':
      if (Syntool.keyboardManager.isShiftDown()) {
        localXY = Transform2D.transformedPoint(toHead, [x, y]);
        return {
          rightLobeCenter: Transform2D.transformedPoint(fromHead, [localXY[0], 0]),
        };
      } else {
        return {
          rightLobeCenter: [x, y],
        };
      }
    case 'rightLobeRx':
      localXY = Transform2D.transformedPoint(toHead, [x, y]);
      if (Syntool.keyboardManager.isShiftDown()) {
        return {
          rightLobeRx: Math.max(0, localXY[0] - headDist / 2),
        };
      } else {
        dx = localXY[0] + headDist / 2;
        dy = localXY[1];
        l = Math.hypot(dx, dy);

        return {
          rightLobeCenter: Transform2D.transformedPoint(
            fromHead,
            [headDist * (dx / l - 0.5), headDist * dy / l]
          ),
          rightLobeRx: Math.max(0, l - headDist),
          rightLobeRy: Math.max(0, shape.rightLobeRy + headDist * dy / l),
        };
      }
    case 'rightLobeRy':
      localXY = Transform2D.transformedPoint(toHead, [x, y]);
      if (Syntool.keyboardManager.isShiftDown()) {
        return {
          rightLobeRy: Math.max(0, -localXY[1]),
        };
      } else {
        return {
          rightLobeCenter: Transform2D.transformedPoint(fromHead, [localXY[0], 0]),
          rightLobeRx    : Math.max(0, shape.rightLobeRx - localXY[0] + headDist / 2),
          rightLobeRy    : Math.max(0, -localXY[1]),
        };
      }
    case 'baseAnchor':
      localXY = Transform2D.transformedPoint(toHead, [x, y]);
      baseAnchor = [
        clamp(localXY[0] / headDist2, -1, 1),
        clamp(-localXY[1] / headDist2, -1, 1),
      ];

      localXY = Transform2D.transformedPoint(
        Transform2D.concat([
          Transform2D.rotation(shape.baseAngle * DEG_IN_RAD),
          Transform2D.translation(
            (shape.baseAnchor[0] - baseAnchor[0]) * headDist2,
            -(shape.baseAnchor[1] - baseAnchor[1]) * headDist2
          ),
          Transform2D.rotation(-shape.baseAngle * DEG_IN_RAD),
        ]),
        [0, -shape.baseHeight]
      );
      baseHeight = (localXY[1] < 0 ? getXYLength(localXY) : 0);
      baseAngle = PI_2 + Math.atan2(localXY[1], localXY[0]);

      return {
        baseAnchor,
        baseHeight: baseHeight,
        baseAngle : shape.baseAngle + baseAngle * RAD_IN_DEG,
      };
    case 'baseWidth':
      localXY = Transform2D.transformedPoint(toBase, [x, y]);
      return {
        baseWidth: clamp(localXY[0] * 2, 0, headDist),
      };
    case 'baseHeight':
      localXY = Transform2D.transformedPoint(toBase, [x, y]);
      baseHeight = (localXY[1] < 0 ? getXYLength(localXY) : 0);
      baseAngle = PI_2 + Math.atan2(localXY[1], localXY[0]);

      if (Syntool.keyboardManager.isShiftDown()) {
        return {
          baseHeight: baseHeight * Math.cos(baseAngle),
        };
      } else {
        return {
          baseHeight: baseHeight,
          baseAngle : shape.baseAngle + baseAngle * RAD_IN_DEG,
        };
      }
    default:
      return null;
    }
  },
  updatesFromChanges(shape, changes) {
    var updates = {};

    if (changes.leftLobeCenter) {
      updates.leftLobeCenter = changes.leftLobeCenter;
    }
    if ('leftLobeRx' in changes) {
      updates.leftLobeRx = changes.leftLobeRx;
    }
    if ('leftLobeRy' in changes) {
      updates.leftLobeRy = changes.leftLobeRy;
    }
    if (changes.rightLobeCenter) {
      updates.rightLobeCenter = changes.rightLobeCenter;
    }
    if ('rightLobeRx' in changes) {
      updates.rightLobeRx = changes.rightLobeRx;
    }
    if ('rightLobeRy' in changes) {
      updates.rightLobeRy = changes.rightLobeRy;
    }
    if (changes.baseAnchor) {
      updates.baseAnchor = [
        clamp(changes.baseAnchor[0], -1, 1),
        clamp(changes.baseAnchor[1], -1, 1),
      ];
    }
    if ('baseWidth' in changes) {
      updates.baseWidth = changes.baseWidth;
    }
    if ('baseAngle' in changes) {
      var angle = changes.baseAngle;

      angle = modulo(angle, 360);
      angle = (angle > 180 ? angle - 360 : angle);
      updates.baseAngle = clamp(angle, -45, 45);

      if (changes.baseAngle !== updates.baseAngle && changes.baseHeight) {
        changes.baseHeight *= Math.cos(
          (updates.baseAngle - changes.baseAngle) * DEG_IN_RAD
        );
      }
    }
    if ('baseHeight' in changes) {
      updates.baseHeight = Math.max(0, changes.baseHeight);
    }
    return updates;
  },
});

function createMushroomHeadPath(o0x, o0y, r0x, r0y, o1x, o1y, r1x, r1y) {
  var _0x = o0x - r0x;
  var e0x = o0x - r0x * BEZIER_ELLIPSE_CONSTANT;
  var e1x = o1x + r1x * BEZIER_ELLIPSE_CONSTANT;
  var _1x = o1x + r1x;
  var _0y = o0y - r0y;
  var e0y = o0y - r0y * BEZIER_ELLIPSE_CONSTANT;
  var E0y = o0y + r0y * BEZIER_ELLIPSE_CONSTANT * 3;
  var E1y = o1y + r1y * BEZIER_ELLIPSE_CONSTANT * 3;
  var e1y = o1y - r1y * BEZIER_ELLIPSE_CONSTANT;
  var _1y = o1y - r1y;

  return [
    [
      [o1x, e1x, _1x, _1x],
      [_1x, _1x, _0x, _0x],
      [_0x, _0x, e0x, o0x],
    ],
    [
      [_1y, _1y, e1y, o1y],
      [o1y, E1y, E0y, o0y],
      [o0y, e0y, _0y, _0y],
    ],
  ];
}
