/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  RAD_IN_DEG,
  getXYDistance,
  Bounds,
  Geometry,
  Quantity,
  Types,
  Syntool,
} from '../syntool.js';
import {Shape} from '../datasets/shape_dataset/shape.js';

const MIN_VIRTUAL_SEGMENT_LENGTH = 40; /*px*/
const HIDDEN_STYLE = {display: 'none'};
const {utils} = Shape;

Shape.addShapeType('WKT', {
  typeChecker: Types.shape({
    author: Types.string,
    wkt   : Types.WKT.isRequired,
  }),
  project(shape) {
    shape.wkt = utils.projectedWKT(shape.wkt);
  },
  unproject(shape) {
    shape.wkt = utils.unprojectedWKT(shape.wkt);
  },
  create(shape, props) {
    shape.author = (props.author || '').trim();
    shape.wkt = utils.correctedWKT(props.wkt);
  },
  same(a, b) {
    return (
      a.author === b.author &&
      a.wkt    === b.wkt
    );
  },
  createGeometry(shape) {
    return Geometry.fromWKT(shape.wkt);
  },

  getLength(shape) {
    const geometry = Shape.createGeometry(shape);
    const lengthInMeters = Geometry.getGeodesicLength(
      geometry,
      Syntool.mapView.getProjection()
    );

    return Quantity.convert(lengthInMeters, 'm', 'km');
  },
  getArea(shape) {
    const geometry = Shape.createGeometry(shape);
    const areaInMeters2 = Geometry.getGeodesicArea(
      geometry,
      Syntool.mapView.getProjection()
    );

    return Quantity.convert(areaInMeters2, 'm2', 'km2');
  },
  getControlPoints(shape, state, geometry, resolution) {
    if (state.lastGeometry === geometry) {
      if (state.movePos || state.rotatePos) {
        state.lastControlPoints = Object.assign({}, state.lastControlPoints, {
          move: state.movePos,
          rotate: state.rotatePos,
        });
      } else if (state.lastResolution !== resolution) {
        const {rotate} = state.lastControlPoints;

        state.lastControlPoints = Object.assign({}, state.lastControlPoints, {
          rotate: Object.assign(rotate, {
            x: rotate._x + rotate.xOffset * resolution,
            y: rotate._y + rotate.yOffset * resolution,
          }),
        });
        updateVirtualControlPointsVisibility(state.lastControlPoints, resolution);
        state.lastResolution = resolution;
      }
      return state.lastControlPoints;
    }

    const controlPoints = (
      collectGeometryVertexControlPoints(geometry, utils.styles.virtual)
    );
    const bounds = Geometry.getBounds(geometry);

    updateVirtualControlPointsVisibility(controlPoints, resolution);

    controlPoints.move = (
      state.movePos || {
        x: Bounds.getCenterX(bounds),
        y: Bounds.getCenterY(bounds),

        style: utils.styles.move,
        onTop: true,
      }
    );

    const maxX = Bounds.getRight(bounds);
    const minY = Bounds.getBottom(bounds);
    const dx = maxX - controlPoints.move.x;
    const dy = minY - controlPoints.move.y;
    const d = utils.rotateOffset / Math.hypot(dx, dy);
    const xOffset = dx * d;
    const yOffset = dy * d;

    controlPoints.rotate = (
      state.rotatePos || {
        x: maxX + xOffset * resolution,
        y: minY + yOffset * resolution,

        _x: maxX,
        _y: minY,
        xOffset,
        yOffset,

        style: utils.styles.rotate,
        onTop: true,
      }
    );

    state.lastGeometry = geometry;
    state.lastControlPoints = controlPoints;
    return controlPoints;
  },
  controlPointDidStartDrag(shape, key, state, controlPoints) {
    if (key === 'rotate') {
      state.movePos = Object.assign({}, controlPoints.move);
      state.rotatePos = Object.assign({}, controlPoints.rotate);
    }
  },
  controlPointDidEndDrag(shape, key, state) {
    if (key === 'rotate') {
      state.movePos = null;
      state.rotatePos = null;
    }
  },
  changesOnControlPointMove(shape, key, x, y, controlPoints, state, geometry, resolution) {
    switch (key) {
    case null:
    case 'move': {
      const bounds = Geometry.getBounds(geometry);
      const centerX = Bounds.getCenterX(bounds);
      const centerY = Bounds.getCenterY(bounds);

      return {translate: [x - centerX, y - centerY]};
    }
    case 'rotate': {
      const dx0 = controlPoints[key].x - controlPoints.move.x;
      const dy0 = controlPoints[key].y - controlPoints.move.y;
      const dx1 = x - controlPoints.move.x;
      const dy1 = y - controlPoints.move.y;

      const angle = (
        Syntool.keyboardManager.isShiftDown()
        ? 0
        : Math.atan2(dy1, dx1) - Math.atan2(dy0, dx0)
      );

      const xOffset = controlPoints.rotate.x - controlPoints.rotate._x;
      const yOffset = controlPoints.rotate.y - controlPoints.rotate._y;
      const lOffset = Math.hypot(xOffset, yOffset);
      const lMin = utils.minScaleDistance * resolution;
      const l0 = Math.hypot(dx0, dy0) - lOffset;
      const l1 = Math.hypot(dx1, dy1) - lOffset;
      const scale = Math.max(lMin, l1) / l0;

      if (Syntool.keyboardManager.isShiftDown()) {
        x = controlPoints.move.x + (dx0 - xOffset) * scale + xOffset;
        y = controlPoints.move.y + (dy0 - yOffset) * scale + yOffset;
      } else if (l1 < lMin) {
        const lScale = (lMin + lOffset) / (l1 + lOffset);
        x = controlPoints.move.x + dx1 * lScale;
        y = controlPoints.move.y + dy1 * lScale;
      }

      state.rotatePos._x = x - xOffset;
      state.rotatePos._y = y - yOffset;
      state.rotatePos.x = x;
      state.rotatePos.y = y;
      return {
        origin: [controlPoints.move.x, controlPoints.move.y],
        scale : scale,
        rotate: angle,
      };
    }
    default:
      if (!controlPoints[key].virtual) {
        return {
          movePoint: {
            point: controlPoints[key],
            to   : [x, y],
          },
        };
      }
      return null;
    }
  },
  updatesFromChanges(shape, changes, geometry) {
    const {
      movePoint,
      addPoint,
      removePoint,
      origin,
      rotate,
      scale,
      translate,
    } = changes;

    if (movePoint) {
      Geometry.translate(
        Geometry.getComponentAtIndexPath(geometry, movePoint.point.indexPath),
        movePoint.to[0] - movePoint.point.x,
        movePoint.to[1] - movePoint.point.y
      );
    }
    if (addPoint) {
      Geometry.addComponentAtIndexPath(
        geometry,
        Geometry.createPoint(addPoint.x, addPoint.y),
        addPoint.indexPath
      );
    } else if (removePoint) {
      Geometry.removeComponentAtIndexPath(geometry, removePoint.indexPath);
    }

    if (rotate != null && !Syntool.keyboardManager.isShiftDown()) {
      Geometry.rotate(geometry, rotate * RAD_IN_DEG, origin);
    }
    if (scale != null) {
      Geometry.scale(geometry, scale, scale, origin);
    }
    if (translate) {
      Geometry.translate(geometry, translate[0], translate[1]);
    }
    return {wkt: Geometry.toWKT(geometry)};
  },
});

function collectGeometryVertexControlPoints(geometry, style) {
  const controlPoints = {};

  for (const {vertex: [x, y], indexPath} of Geometry.vertices(geometry)) {
    controlPoints[indexPath] = {x, y, indexPath};
  }
  for (const {from, to, indexPath} of Geometry.edges(geometry)) {
    if (from[0] === to[0] && from[1] === to[1]) {
      continue;
    }

    const x = (from[0] + to[0]) / 2;
    const y = (from[1] + to[1]) / 2;
    const segLen = getXYDistance(from, to);

    controlPoints['_' + indexPath] = (
      {x, y, segLen, virtual: true, style, indexPath}
    );
  }

  return controlPoints;
}

function updateVirtualControlPointsVisibility(controlPoints, resolution) {
  for (const [key, controlPoint] of Object.entries(controlPoints)) {
    if (controlPoint.virtual) {
      const style = (
        controlPoint.segLen / resolution < MIN_VIRTUAL_SEGMENT_LENGTH
        ? HIDDEN_STYLE
        : utils.styles.virtual
      );
      controlPoints[key] = Object.assign({}, controlPoint, {style});
    }
  }
}
