/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  PASSIVE,
  wrap,
  createClass,
  EventsHub,
  Projection,
  Quantity,
} from '../../syntool.js';
import {eventsListeningExtension} from '../../utils/events_listening_extension.js';
import '../../map_view_extensions/map_view_mouse_position.js';

let lonLat = [0, 0];

function renderDecimals(lon, lat) {
  return (
    Quantity.format(lat, 'lat').padStart(8) + ' ' +
    Quantity.format(lon, 'lon').padStart(9)
  );
}
function renderSexagesimals(lon, lat) {
  return (
    Quantity.format(lat, 'lat-dms').padStart(10) + ' ' +
    Quantity.format(lon, 'lon-dms').padStart(11)
  );
}

const DECIMAL = 'decimal';
const SEXAGESIMALS = 'sexagesimals';
const renderers = {
  [DECIMAL]: renderDecimals,
  [SEXAGESIMALS]: renderSexagesimals,
};

export const CoordinatesView = createClass([
  eventsListeningExtension,
  /*Public*/ {
    static: {
      DECIMAL,
      SEXAGESIMALS,
    },

    constructor: function CoordinatesView(mapView) {
      this.mapView = mapView;
      this.hub = new EventsHub();
      this._active = false;
      this._format = DECIMAL;
      this._followingMouse = false;
      this.lon = 0;
      this.lat = 0;

      this.node = document.createElement('span');
      this.node.className = 'CoordinatesView';
      this.node.title = (
        "Coordinates of the mouse cursor (or the map's center when the " +
        'crosshairs icon is visible on the left)'
      );

      this.node.addEventListener('click', () => {
        this.format = (this.format === DECIMAL ? SEXAGESIMALS : DECIMAL);
      }, PASSIVE);

      this.activate();
    },

    activate() {
      if (this._active || !this.mapView) {
        return;
      }
      this._active = true;
      this.handleMousePositionChange(null);
      this.listenToEvent(
        this.mapView.hub,
        'change:mouse-position',
        this.handleMousePositionChange,
        this
      );
      this.listenToEvent(
        this.mapView.hub,
        'changing:extent',
        this.handleMapViewExtentChange,
        this
      );
      this.hub.emit('change:active');
    },
    deactivate() {
      if (!this._active) {
        return;
      }
      this.stopAllListening();
      this._active = false;
      this.hub.emit('change:active');
    },
    get active() {
      return this._active;
    },
    set active(on) {
      if (on) {
        this.activate();
      } else {
        this.deactivate();
      }
    },

    get format() {
      return this._format;
    },
    set format(format) {
      if (this._format !== format && renderers[this.format]) {
        this._format = format;
        this.render();
        this.hub.emit('change:format');
      }
    },

    updateFromProjectedXY(xy, projection) {
      if (xy) {
        lonLat = Projection.unprojectXY(projection, xy, lonLat);
        this.update(lonLat[0], lonLat[1]);
      } else {
        this.node.textContent = '';
      }
    },
    update(lon, lat) {
      this.lon = wrap(lon, -180, 180);
      this.lat = lat;
      this.render();
    },
    render() {
      this.node.textContent = renderers[this.format](this.lon, this.lat);
    },
  },
  /*Event handlers*/ {
    handleMousePositionChange(position) {
      this._followingMouse = position !== null;
      this.node.classList.toggle('-center', !this._followingMouse);
      if (position) {
        this.updateFromProjectedXY(position, this.mapView.getProjection());
      } else {
        this.handleMapViewExtentChange();
      }
    },
    handleMapViewExtentChange() {
      if (!this._followingMouse) {
        this.updateFromProjectedXY(
          this.mapView.getViewportCenterProjectedXY(),
          this.mapView.getProjection()
        );
      }
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {CoordinatesView});
