/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  PASSIVE,
  parseDecimal,
  debounce,
  H,
  DOMEvents,
  removeAllChildrenOfNode,
  setGlobalCursor,
  RegExpHelpers,
  Control,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {drag} from '../utils/gestures.js';

linkStyleSheet('css/dialog_arrow.css');
linkStyleSheet('css/dialog.css');

const pinToggleNode = H('label', {class: 'Dialog-Toggle --Pin js-pin'}, [
  H('input', {
    type    : 'checkbox',
    class   : 'Dialog-ToggleCheckbox --Pin js-pin-toggle',
    checked : false,
    tabindex: 0,
  }),
  H('i', {class: 'fa fa-thumb-tack'}),
]);
const closeButtonNode = H('button', {
  class: 'Dialog-Button --Close js-close-button',
  title: 'Hide this dialog',
}, H('i', {class: 'fa fa-times'}));

let nextId = 0;
let topDialog = null;

function setTopDialog(dialog) {
  if (topDialog !== dialog) {
    if (topDialog) {
      topDialog.node.classList.remove('-on-top');
    }
    topDialog = dialog;
    if (topDialog) {
      topDialog.node.classList.add('-on-top');
    }
  }
}

const CSS_PX_RE = new RegExp(`^${RegExpHelpers.cssPixels}$`);
function parseCSSPx(string) {
  return CSS_PX_RE.test(string) ? parseFloat(string) : NaN;
}

export class Dialog extends Control {
  /*Public*/
  constructor(id, props, options) {
    super(options);

    if (!props && typeof id !== 'string') {
      props = id;
      id = null;
    }

    id ||= ('Dialog-' + nextId++);
    this.id = id;
    this.node = H('div', {id, class: 'Dialog', style: 'display: none'}, [
      H('div', {class: 'Dialog-TitleBar js-title-bar'}, [
        H('i', {class: 'fa fa-arrows'}),
        this.titleNode = H('span', {class: 'Dialog-Title js-title'}),
      ]),
      this.contentNode = H('div', {class: 'Dialog-Content js-content'}),
      H('div', {class: 'Dialog-Resizer js-resizer'}),
    ]);

    this.props = Object.assign({
      title    : 'Untitled',
      content  : null,
      arrow    : false,
      closable : true,
      pinnable : false,
      resizable: false,

      autoLocationReset: true,
      pinTooltip: 'Pin dialog',

      context       : null,
      onBeforeShow  : null,
      onShow        : null,
      onHide        : null,
      onPinnedChange: null,
      onResize      : null,
    }, props);

    this.handleWindowResize = debounce(100, this.handleWindowResize.bind(this));

    this._shown = false;
    this._usableViewportRect = null;
    this._dragStartPosition = null;
    this._dragStartStyle = null;
    this._resizeStartMaxHeight = null;
    this._resizeStartHeight = null;
    this._defaultMaxHeight = 0;
    this._removeResizeGesture = null;

    DOMEvents(document)
    .on('DOMContentLoaded', this._invalidateUsableViewportRect, this);

    DOMEvents(this.node)
    .on([
      'mousedown+capture+passive',
      'touchstart+capture+passive',
      'focus+capture+passive',
    ], () => {
      setTopDialog(this);
    })
    .on('change', '.js-pin-toggle', (event) => {
      this.pinned = event.target.checked;
    })
    .on('click', '.js-close-button', () => {
      this.hide();
    });

    const titleBarNode = this.node.querySelector('.js-title-bar');
    drag(
      titleBarNode,
      {moveTolerance: 5},
      this.handleTitleBarDragChange,
      this
    );

    this._propsDidChange();
  }

  get state() {
    return Object.assign(super.state, (
      this.props.pinnable
      ? {shown: this.shown, pinned: this.pinned}
      : {shown: this.shown}
    ));
  }
  set state(state) {
    super.state = state;
    this.shown = state.shown;
    if (this.props.pinnable) {
      this.pinned = state.pinned;
    }
  }

  get shown() {
    return this._shown;
  }
  set shown(shown) {
    if (this._shown === !shown) {
      this.toggle();
    }
  }

  get pinned() {
    var pinToggle = this.node.querySelector('.js-pin-toggle');
    return pinToggle ? pinToggle.checked : false;
  }
  set pinned(pinned) {
    var pinToggle = this.node.querySelector('.js-pin-toggle');
    if (pinToggle) {
      pinToggle.checked = pinned;
      if (typeof this.props.onPinnedChange === 'function') {
        this.props.onPinnedChange.call(this.props.context, this, pinned);
      }
      this.hub.emit('change:pinned');
      this.hub.emit('change:state');
    }
  }

  setProps(props) {
    if (props) {
      var prevProps = this.props;
      this.props = Object.assign({}, this.props, props);
      this._propsDidChange(prevProps);
    }
  }

  hide() {
    if (this.shown) {
      this._hide();
      this._shown = false;
      if (this.active) {
        this._onHide();
      }
      this.hub.emit('change:shown');
      this.hub.emit('change:state');
    }
  }
  show(props) {
    this.setProps(props);

    const wasHidden = !this.shown;

    if (this.active) {
      setTopDialog(this);
      this._onBeforeShow();
      this._show();
      this._shown = true;
      if (this._shouldResetLocation()) {
        this.resetLocation();
      }
      this._onShow();
    }

    if (wasHidden) {
      this._shown = true;
      this.hub.emit('change:shown');
      this.hub.emit('change:state');
    }
  }
  toggle() {
    if (this.active && this.shown && this._shouldResetLocation()) {
      this.show();
    } else if (this.shown) {
      this.hide();
    } else {
      this.show();
    }
  }
  clear() {
    if (this.shown) {
      this.titleNode.title = '';
      removeAllChildrenOfNode(this.titleNode);
      removeAllChildrenOfNode(this.contentNode);
    }
  }
  resetLocation() {
    this.node.style.left = '';
    this.node.style.top = '';
  }

  /*Protected - Control overrides*/
  onActivate() {
    this.syntool.uiOverlayNode.appendChild(this.node);
    this.syntool.hub
    .on('change:usable-viewport', this._invalidateUsableViewportRect, this);

    this._propsDidChange(
      Object.assign({}, this.props, {resizable: false, onResize: null})
    );
    if (this.shown) {
      this.show();
    }
  }
  onDeactivate() {
    if (this.shown) {
      this._onHide();
    }
    this._propsDidChange(
      Object.assign({}, this.props, {resizable: false, onResize: null})
    );

    this.node.remove();
    this.syntool.hub
    .off('change:usable-viewport', this._invalidateUsableViewportRect, this);
  }

  /*Protected - Gesture handlers*/
  handleWindowResize() {
    if (typeof this.props.onResize === 'function' && this.shown) {
      this.props.onResize.call(this.props.context, this);
    }
  }
  handleTitleBarDragChange(props) {
    const node = this.node;

    switch (props.state) {
    case 'start':
      setGlobalCursor('grabbing');
      break;
    case 'cancel':
    case 'end':
      setGlobalCursor(null);
      break;
    default:
    }
    switch (props.state) {
    case 'possible':
      node.classList.add('user-interacting');
      break;
    case 'end':
    case 'cancel':
      node.classList.remove('user-interacting');
      break;
    default:
    }

    switch (props.state) {
    case 'possible': {
      let {left, top} = node.style;
      this._dragStartStyle = [left, top];
      left = parseCSSPx(left);
      top = parseCSSPx(top);
      if (isNaN(left) || isNaN(top)) {
        const computedStyle = window.getComputedStyle(node);
        if (isNaN(left)) {
          left = parseCSSPx(computedStyle.left);
        }
        if (isNaN(top)) {
          top = parseCSSPx(computedStyle.top);
        }
      }
      this._dragStartPosition = [left, top];
      break;
    }
    case 'cancel':
      node.style.left = this._dragStartStyle[0];
      node.style.top = this._dragStartStyle[1];
      break;
    default:
      node.style.left = (this._dragStartPosition[0] + props.movement[0]) + 'px';
      node.style.top = (this._dragStartPosition[1] + props.movement[1]) + 'px';
    }

    if (
      props.state === 'possible'
      && props.event.target.closest('.js-close-button')
    ) {
      props.stop();
    }
  }
  handleResizerDragChange(props) {
    const node = this.node;
    const wrapper = this.contentNode;
    let clientRect;
    let height;
    let maxHeight;

    switch (props.state) {
    case 'possible':
      node.classList.add('user-interacting');
      break;
    case 'end':
    case 'cancel':
      node.classList.remove('user-interacting');
      break;
    default:
    }

    switch (props.state) {
    case 'possible':
      this._resizeStartMaxHeight = wrapper.style.maxHeight;

      if (this._resizeStartMaxHeight) {
        wrapper.style.maxHeight = '';
      }
      this._defaultMaxHeight = parseDecimal(
        window.getComputedStyle(wrapper).maxHeight
      );
      if (this._resizeStartMaxHeight) {
        wrapper.style.maxHeight = this._resizeStartMaxHeight;
      }

      clientRect = wrapper.getBoundingClientRect();
      height = Math.floor(clientRect.height);
      this._resizeStartHeight = height;
      maxHeight = (
        this._resizeStartMaxHeight
        ? parseDecimal(this._resizeStartMaxHeight)
        : this._defaultMaxHeight
      );
      break;
    case 'cancel':
      wrapper.style.maxHeight = this._resizeStartMaxHeight;
      break;
    default:
      maxHeight = Math.floor(Math.max(0, this._resizeStartHeight + props.movement));
      wrapper.style.maxHeight = maxHeight + 'px';

      clientRect = wrapper.getBoundingClientRect();
      height = Math.floor(clientRect.height);
      if (height < Math.min(maxHeight, this._defaultMaxHeight)) {
        // when resized to the max and that max is smaller than the default
        // max-height, revert to it
        wrapper.style.maxHeight = '';
      }
      if (typeof this.props.onResize === 'function' && this.shown) {
        this.props.onResize.call(this.props.context, this);
      }
    }

    switch (props.state) {
    case 'end':
    case 'cancel':
      setGlobalCursor(null);
      break;
    default:
      setGlobalCursor(
        (maxHeight > 0 ? 'n' : '') +
        (maxHeight > height ? '' : 's') +
        '-resize'
      );
    }
  }

  /*Protected*/
  _hide() {
    this.node.style.display = 'none';
  }
  _show() {
    this.node.style.display = '';
  }

  _onHide() {
    if (typeof this.props.onHide === 'function') {
      this.props.onHide.call(this.props.context, this);
    }
    this.hub.emit('hide');
  }
  _onBeforeShow() {
    if (typeof this.props.onBeforeShow === 'function') {
      this.props.onBeforeShow.call(this.props.context, this);
    }
    this.hub.emit('before-show');
  }
  _onShow() {
    if (typeof this.props.onShow === 'function') {
      this.props.onShow.call(this.props.context, this);
    }
    this.hub.emit('show');
  }

  /*Protected*/
  _propsDidChange(prevProps) {
    const props = this.props;

    if (!prevProps || prevProps.pinnable !== props.pinnable) {
      let pinNode = this.node.querySelector('.js-pin');
      if (props.pinnable === false) {
        if (pinNode) {
          this.node.removeChild(pinNode);
        }
      } else if (!pinNode) {
        pinNode = pinToggleNode.cloneNode(true);
        this.titleNode.parentNode.insertBefore(
          pinNode,
          this.titleNode.nextElementSibling
        );
      }
    }
    if (!prevProps || prevProps.closable !== props.closable) {
      let closeNode = this.node.querySelector('.js-close-button');
      if (props.closable === false) {
        if (closeNode) {
          this.node.removeChild(closeNode);
        }
      } else if (!closeNode) {
        closeNode = closeButtonNode.cloneNode(true);
        this.titleNode.parentNode.appendChild(closeNode);
      }
    }
    if (!prevProps || prevProps.arrow !== props.arrow) {
      let arrowNode = this.node.querySelector('.js-arrow');
      if (props.arrow === false) {
        if (arrowNode) {
          this.node.removeChild(arrowNode);
        }
      } else {
        if (!arrowNode) {
          arrowNode = document.createElement('div');
          this.node.insertBefore(arrowNode, this.node.firstElementChild);
        }
        arrowNode.className = 'Dialog-Arrow -' + props.arrow + ' js-arrow';
      }
    }
    if (!prevProps || prevProps.pinTooltip !== props.pinTooltip) {
      const pinToggle = this.node.querySelector('.js-pin');
      if (pinToggle) {
        pinToggle.title = props.pinTooltip;
      }
    }
    if (!prevProps || prevProps.title !== props.title) {
      this.titleNode.innerHTML = props.title || '';
      this.titleNode.title = props.title || '';
    }
    if (!prevProps || prevProps.content !== props.content) {
      this.contentNode.style.cssText = '';
      removeAllChildrenOfNode(this.contentNode);
      if (typeof props.content === 'string') {
        this.contentNode.innerHTML = props.content;
      } else if (typeof props.content === 'function') {
        props.content(this.contentNode);
      } else if (props.content instanceof Node) {
        this.contentNode.appendChild(props.content);
      }
    }
    if (!prevProps || prevProps.resizable !== props.resizable) {
      this.node.classList.toggle('-no-resize', !props.resizable);
      if (this.active && props.resizable) {
        this._defaultMaxHeight = parseFloat(
          window.getComputedStyle(this.contentNode).maxHeight
        );
        this._removeResizeGesture = drag(
          this.node.querySelector('.js-resizer'),
          {direction: 'y'},
          this.handleResizerDragChange,
          this
        );
      } else {
        this._defaultMaxHeight = 0;
        if (this._removeResizeGesture) {
          this._removeResizeGesture();
        }
      }
    }
    if (!prevProps || prevProps.onResize !== props.onResize) {
      if (this.active && typeof props.onResize === 'function') {
        window.addEventListener('resize', this.handleWindowResize, PASSIVE);
      } else {
        window.removeEventListener('resize', this.handleWindowResize, PASSIVE);
      }
    }
  }
  _invalidateUsableViewportRect() {
    this._usableViewportRect = null;
  }
  _getUsableViewport() {
    if (!this._usableViewportRect) {
      this._usableViewportRect = (
        this.syntool.usableViewportNode.getBoundingClientRect()
      );
    }
    return this._usableViewportRect;
  }
  _shouldResetLocation() {
    return (
      this.props.autoLocationReset
      && !this._isAtDefaultLocation()
      && !this._isInViewport()
    );
  }
  _isAtDefaultLocation() {
    return (!this.node.style.left && !this.node.style.top);
  }
  _isInViewport() {
    if (!this.shown) {
      // Can't getBoundingClientRect() for a `display: none` element
      return false;
    }

    const titleBarNode = this.node.querySelector('.js-title-bar');
    const toolBarRect = titleBarNode.getBoundingClientRect();
    const rect = this._getUsableViewport();
    return (
      toolBarRect.left < rect.right &&
      toolBarRect.right > rect.left &&
      toolBarRect.top < rect.bottom &&
      toolBarRect.bottom > rect.top
    );
  }
}
window.ODL = Object.assign(window.ODL || {}, {Dialog});
