/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  NONPASSIVE,
  clamp,
  lerp,
  inverseLerp,
  interpolateString,
  HIDKeys,
  keyFromEvent,
  H,
  S,
  preventDefaultAndStopEvent,
  Syntool,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {Gestures} from '../utils.js';

linkStyleSheet('css/gauge_component.css');

const attributesOverrideForProps = {
  disabled: undefined,
  slicesCount: undefined,
  title: undefined,
  min: undefined,
  max: undefined,
  value: undefined,
  withSlider: undefined,
  onChange: undefined,
  onClick: undefined,
};

function renderGauge(props) {
  const slicesCount = props.slicesCount;
  const width = 100;
  const height = 100;
  const cx = width / 2;
  const cy = height / 2;
  const radius = width / 2;

  const gradeDelta = Math.PI / slicesCount;
  const gradePoints = [];
  for (let i = 0; i <= slicesCount; i++) {
    gradePoints.push(
      [
        cx + radius * Math.cos(i * gradeDelta),
        cy - radius * Math.sin(i * gradeDelta),
      ]
    );
  }

  return S('svg', Object.assign({}, props, attributesOverrideForProps, {
    version            : '1.1',
    xmlns              : 'http://www.w3.org/2000/svg',
    preserveAspectRatio: 'xMidYMid',
    viewBox            : `0 0 ${width} ${height}`,
    class              : 'GaugeComponent-Gauge js-gauge',
  }), S('g', [
    S('title'),
    S('path', {
      class: 'background-area',
      d    : [
        'M' + (cx - radius),
        cy + ' A' + radius,
        radius + ' 0 0,1 ' + (cx + radius),
        cy + ' Z',
      ].join(','),
      stroke: 'none',
      fill  : 'rgba(85, 85, 85, 0.2)',
    }),
    S('path', {
      class : 'filled-area',
      stroke: 'none',
      fill  : 'hsl(210, 70%, 30%)',

      'pointer-events': 'none',
    }),
    S('line', {
      class      : 'filled-area-edge',
      x1         : cx,
      y1         : cy,
      x2         : cx - radius,
      y2         : cy,
      stroke     : 'black',
      strokeWidth: 5,

      'pointer-events': 'none',
    }),
    S('text', {
      class        : 'value-text',
      x            : cx,
      y            : height,
      dy           : -10,
      fill         : 'currentColor',
      'text-anchor': 'middle',
      'font-size'  : '36px',
    }),
    ...gradePoints.map(([x2, y2]) => S('line', {
      class: 'slice-edge',
      x1: cx,
      y1: cy,
      x2,
      y2,
      stroke: 'black',
      strokeWidth: 1,

      'pointer-events': 'none',
    })),
  ]));
}

export class GaugeComponent {
  constructor(props) {
    this.props = {
      slicesCount: 4,
      title      : '${value}%',
      min        : 0,
      max        : 100,
      value      : 50,
      disabled   : false,
      withSlider : false,
    };
    this.node = H('div', {class: `GaugeComponent ${props.class || ''}`}, [
      H('input', {
        type : 'range',
        min  : 0,
        max  : 1,
        step : 'any',
        class: 'GaugeComponent-Slider js-slider',

        tabindex: -1,
      }),
      renderGauge(this.props),
    ]);

    this.handleClick = this.handleClick.bind(this);
    this.handleSliderInput = this.handleSliderInput.bind(this);
    this.handleKeyDown = this.handleKeyDown.bind(this);
    this.handleWheel = Gestures.getWheelHandler(this.handleWheel, this);

    this.addEventHandlers();
    this.setProps(props);
  }

  setProps(partialProps) {
    const prevProps = this.props;
    this.props = Object.assign({}, this.props, partialProps);
    if (this.node && prevProps.disabled !== this.props.disabled) {
      if (this.props.disabled) {
        this.removeEventHandlers();
      } else {
        this.addEventHandlers();
      }
      if (this.props.disabled) {
        this.node.setAttribute('disabled', 'disabled');
      } else {
        this.node.removeAttribute('disabled');
      }
      this.node.querySelector('.js-slider').disabled = this.props.disabled;
    }
    if (this.node && prevProps.withSlider !== this.props.withSlider) {
      this.node.classList.toggle('-has-slider', this.props.withSlider);
    }
    this.updateValue();
  }

  addEventHandlers() {
    const node = this.node;
    const gaugeElement = node.querySelector('.js-gauge');
    const sliderElement = node.querySelector('.js-slider');
    const areaNode = node.querySelector('.background-area');
    let dragged = false;

    const handleClick = (event) => {
      if (!(areaNode.contains(event.target) && dragged)) {
        this.handleClick(event);
      }
    };

    function handleSliderKeydown(event) {
      event.preventDefault();
      event.syntoolShortcutsCanUse = true;
    }

    gaugeElement.addEventListener('click', handleClick, NONPASSIVE);
    sliderElement.addEventListener('input', this.handleSliderInput, NONPASSIVE);
    sliderElement.addEventListener('keydown', handleSliderKeydown, NONPASSIVE);
    node.addEventListener('keydown', this.handleKeyDown, NONPASSIVE);
    node.addEventListener('wheel', this.handleWheel, NONPASSIVE);

    let dragStartValue = NaN;
    const removeDragGesture = Gestures.drag(
      areaNode,
      {moveTolerance: 2},
      (props) => {
        switch (props.state) {
        case 'possible':
          dragged = false;
          break;
        case 'start':
          dragged = true;
          break;
        default:
        }
        switch (props.state) {
        case 'possible':
          dragStartValue = this.props.value;
          if (node.focus) {
            node.focus();
          }
          break;
        case 'cancel':
          this.emitChange(dragStartValue);
          break;
        default:
          this.emitChangeFromClientLocation(props.currentLocation);
        }
      }
    );

    this.removeEventHandlers = function() {
      gaugeElement.removeEventListener('click', handleClick, NONPASSIVE);
      sliderElement.removeEventListener('input', this.handleSliderInput, NONPASSIVE);
      sliderElement.removeEventListener('keydown', handleSliderKeydown, NONPASSIVE);
      node.removeEventListener('keydown', this.handleKeyDown, NONPASSIVE);
      node.removeEventListener('wheel', this.handleWheel, NONPASSIVE);
      removeDragGesture();
    };
  }
  removeEventHandlers() {
    /*empty*/
  }

  emitChange(value) {
    if (value !== this.props.value && this.props.onChange) {
      this.props.onChange(value);
    }
  }
  emitChangeFromClientLocation(clientLocation) {
    const rect = this.node.getBoundingClientRect();
    const cx = rect.x + rect.width / 2;
    const cy = rect.y + rect.height / 2;
    const dx = clientLocation[0] - cx;
    const dy = clientLocation[1] - cy;
    const alpha = Math.atan2(-dy, dx) / Math.PI;
    const value = 1.5 - clamp((alpha + 2.5) % 2, 0.5, 1.5);

    ////////////////////////////////////////////////////////////////////
    // alpha           // alpha + 2.5       // (alpha + 2.5) % 2      //
    // =============== // ================= // ====================== //
    //        0.5      //        3          //           1            //
    //        __       //        __         //           __           //
    //      /    \     //      /    \       //         /    \         //
    //   1 |------| 0  // 3.5 |------| 2.5  //    1.5 |------| 0.5    //
    //  -1  \ __ /     // 1.5  \ __ /       //         \ __ /         //
    //       -0.5      //        2          //        1.9  0          //
    ////////////////////////////////////////////////////////////////////
    // clamp((alpha + 2.5) % 2, 0.5, 1.5) // value = 1.5 - clamp(...) //
    // ================================== // ======================== //
    //                 1                  //            0.5           //
    //                 __                 //            __            //
    //               /    \               //          /    \          //
    //          1.5 |------| 0.5          //       0 |------| 1       //
    //               \ __ /               //          \ __ /          //
    //              1.5  0.5              //           0  1           //
    ////////////////////////////////////////////////////////////////////

    this.emitChange(lerp(this.props.min, this.props.max, value));
  }
  emitIncrement(value, delta) {
    if (Syntool.keyboardManager.isKeyDown(HIDKeys.shift)) {
      delta *= 2;
    } else if (Syntool.keyboardManager.isKeyDown(HIDKeys.alt)) {
      delta /= 10;
    }
    this.emitChange(value + delta * Math.abs(this.props.max - this.props.min));
  }

  handleClick(event) {
    if (this.props.onChange) {
      event.preventDefault();
      this.props.onClick(event);
    }
  }
  handleSliderInput(event) {
    const value = parseFloat(event.target.value);
    this.emitChange(lerp(this.props.min, this.props.max, value));
    this.updateValue();
  }
  handleKeyDown(event) {
    switch (keyFromEvent(event)) {
    case HIDKeys.up:
      this.emitIncrement(this.props.value, 0.1);
      break;
    case HIDKeys.down:
      this.emitIncrement(this.props.value, -0.1);
      break;
    default:
      return;
    }

    preventDefaultAndStopEvent(event);
  }
  handleWheel(event, delta) {
    this.emitIncrement(this.props.value, delta * 10 / 100);
  }

  updateValue() {
    const node = this.node;
    const value = inverseLerp(this.props.min, this.props.max, this.props.value);
    const valueString = value.toString();

    if (node.dataset.value !== valueString) {
      const gaugeElement = node.querySelector('.js-gauge');
      const titleElement = node.querySelector('title');
      const filledAreaElement = node.querySelector('.filled-area');
      const filledAreaEdgeElement = node.querySelector('.filled-area-edge');
      const valueTextElement = node.querySelector('.value-text');
      const width = gaugeElement.viewBox.baseVal.width;
      const height = gaugeElement.viewBox.baseVal.height;

      const alpha = (1 - value) * Math.PI;
      const cx = width / 2;
      const cy = height / 2;
      const radius = width / 2;
      const x = cx + radius * Math.cos(alpha);
      const y = cy - radius * Math.sin(alpha);

      node.dataset.value = valueString;
      titleElement.textContent = interpolateString(
        this.props.title,
        this.props
      );
      filledAreaElement.setAttribute('d', [
        'M' + (cx - radius),
        cy + ' A' + radius,
        radius + ' 0 0,1 ' + x,
        y + ' L' + cx,
        cy + ' Z',
      ].join(','));
      filledAreaEdgeElement.setAttribute('x2', x);
      filledAreaEdgeElement.setAttribute('y2', y);
      valueTextElement.textContent = Math.round(value * 100) + '%';
    }

    const sliderElement = node.querySelector('.js-slider');
    if (sliderElement.value !== valueString) {
      sliderElement.value = value;
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {GaugeComponent});
