/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  deepAssign,
  h,
  preventDefaultAndStopEvent,
  DOMEvents,
} from '../../syntool.js';
import {openedHotspot} from '../../features/hotspots.js';
import {removeNodes, linkStyleSheet} from '../../utils/dom_utils.js';
import {showConfirm, showPermalink} from '../utils.js';
import {Dialog} from './dialog.js';

linkStyleSheet('css/hotspots-dialog.css');

function hotspotNameComparator(hotspotA, hotspotB) {
  var nameA = hotspotA.name;
  var nameB = hotspotB.name;

  if (nameA < nameB) {
    return -1;
  }
  if (nameA === nameB) {
    return 0;
  }
  return 1;
}

function renderFormHTML() {
  return h('form', {class: 'HotspotCreationForm', action: '#'}, [
    h('label', 'Create local hotspot:'),
    h('label', {
      for: 'hotspot-name',
      class: 'sr-only',
    }, 'Hotspot name:'),
    h('input', {
      type: 'text',
      id: 'hotspot-name',
      name: 'name',
      required: true,
      placeholder: 'name',
      value: '',
    }),
    h('label', {
      for: 'hotspot-description',
      class: 'sr-only',
    }, 'Hotspot description link:'),
    h('input', {
      type: 'url',
      id: 'hotspot-description',
      name: 'description',
      placeholder: 'description link',
      value: '',
    }),
    h('button', 'Create'),
  ]);
}
function renderListHTML() {
  return h('ul', {class: 'HotspotsList'}, [
    h('li', {class: 'HotspotsList-SectionHeader --Server'}, 'Predefined hotspots'),
    h('li', {class: 'HotspotsList-SectionHeader --Local'}, 'Local hotspots'),
  ]);
}
function renderMessageHTML() {
  return [
    h('em', [
      'All local hotspots are based on cookies and available only from ',
      'this browser.',
    ]),
    h('br'),
    h('br'),
    h('em',
      'Want to participate to the growth of the predefined set of hotspots?'
    ),
    h('br'),
    h('em',
      h('strong', [
        'Post the hotspot link on ',
        h('a', {href: 'https://forum.oceandatalab.com/syntool'},
          'Syntool forum'
        ),
        ' in the "Feature requests" section',
      ])
    ),
  ].join('');
}
function renderHTML() {
  return [
    renderFormHTML(),
    h('div', {class: 'HotspotsList-Wrapper'}, [
      renderListHTML(),
      renderMessageHTML(),
    ]),
  ].join('');
}
function renderHotspot(removable, hotspot) {
  var removeBtn = removable && h('button',
    {type : 'button', class: 'HotspotsList-Button --Remove', title: 'Remove'},
    h('i', {class: 'fa fa-trash'})
  );
  var descBtn = hotspot.description && h('a', {
    href  : hotspot.description,
    target: '_blank',
    class : 'HotspotsList-Button --Description',
    title : 'Description',
  }, h('i', {class: 'fa fa-info'}));
  var linkBtn = h('button',
    {type: 'button', class: 'HotspotsList-Button --Share', title: 'Share'},
    h('i', {class: 'fa fa-link'})
  );
  var mainBtn = h('a',
    {href: hotspot.permalink, class: 'HotspotsList-Button --GoTo'},
    hotspot.name || 'Untitled hotspot'
  );

  return h('li', {class: 'HotspotsList-Hotspot', 'data-id': hotspot.id}, [
    mainBtn,
    linkBtn,
    descBtn,
    removeBtn,
  ]);
}

function getAbsoluteURL(relativeURL) {
  var a = document.createElement('A');
  a.href = relativeURL;
  return a.href;
}

export class HotspotsDialog extends Dialog {
  constructor(options) {
    super('hotspots-dialog', {
      title  : 'Hotspots',
      content: renderHTML(),
    }, options);
    this.store = null;

    var self = this;
    DOMEvents(this.node)
      .on('click', '.HotspotsList-Button.--GoTo', function(event) {
        event.preventDefault();
        var hotspotNode = this.parentElement;
        const hotspot = self.store.get(hotspotNode.dataset.id);
        self.syntool.state = deepAssign(
          {},
          self.syntool.state,
          {
            mapView: {center: undefined, zoom: undefined},
            catalogSelection: undefined,
          },
          hotspot
        );
        self.hide();
      })
      .on('click', '.HotspotsList-Button.--Share', function(event) {
        event.preventDefault();
        var hotspotNode = this.parentElement;
        var hotspot = self.store.get(hotspotNode.dataset.id);
        showPermalink(getAbsoluteURL(hotspot.permalink), self.syntool);
      })
      .on('click', '.HotspotsList-Button.--Remove', function(event) {
        event.preventDefault();
        var hotspotNode = this.parentElement;
        self.store.remove(hotspotNode.dataset.id);
      })
      .on('submit', '.HotspotCreationForm', function(event) {
        preventDefaultAndStopEvent(event);

        var hotspotName = this.querySelector('[name="name"]').value;
        var description = this.querySelector('[name="description"]').value;

        if (hotspotName) {
          self.store.add(hotspotName, description, self.syntool.state);
          this.reset();
        }
      });
  }

  onAdd() {
    super.onAdd();

    this.store = this.syntool.hotspotsStore;
    if (this.store.initialised) {
      this.handleHotspotsInit();
    } else {
      this.store.hub.once('init', this.handleHotspotsInit, this);
    }
    this.store.hub.on('change', this.handleHotspotsChange, this);
    this.handleHotspotsChange();
  }
  onRemove() {
    super.onRemove();

    this.store.hub
      .off('init', this.handleHotspotsInit, this)
      .off('change', this.handleHotspotsChange, this);
    this.store = null;
  }

  handleHotspotsInit() {
    if (
      !openedHotspot ||
      !openedHotspot.name ||
      this.store.hasHotspotSameHotspot(openedHotspot)
    ) {
      return;
    }

    var store = this.store;

    showConfirm({
      syntool: this.syntool,
      title  : 'Save hotspot?',
      message: 'Do you want to save this hotspot locally?',
      details: 'Hotspot name: ' + openedHotspot.name,
      no     : 'No, thanks',
      yes    : 'Save',
    }).then(function(answer) {
      if (answer) {
        store.addHotspot(openedHotspot);
      }
    });
  }
  handleHotspotsChange() {
    var hotspotsList = this.contentNode.querySelector('.HotspotsList');
    var serverListDivider = hotspotsList.querySelector(
      '.HotspotsList-SectionHeader.--Server'
    );
    var localListDivider = hotspotsList.querySelector(
      '.HotspotsList-SectionHeader.--Local'
    );
    var hotspots;

    removeNodes(hotspotsList.querySelectorAll('.HotspotsList-Hotspot'));

    hotspots = Object.values(this.store.serverHotspots);
    hotspots.sort(hotspotNameComparator);
    serverListDivider.insertAdjacentHTML(
      'afterend',
      hotspots.map(renderHotspot.bind(null, false)).join('')
    );

    hotspots = Object.values(this.store.localHotspots);
    hotspots.sort(hotspotNameComparator);
    localListDivider.insertAdjacentHTML(
      'afterend',
      hotspots.map(renderHotspot.bind(null, true)).join('')
    );
  }
}
window.ODL = Object.assign(window.ODL || {}, {HotspotsDialog});
