/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  round10,
  wrap,
  bisectRight,
  addToUniqueSorted,
  HIDModifiers,
  HIDKeys,
  LatLonBounds,
  Bounds,
  Projection,
  Control,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {magnitude} from '../../utils/math.js';

linkStyleSheet('css/lat_lon_grid_control.css');

function round(number) {
  return Math.round(number * 1e6) / 1e6;
}

function formatDegrees(deg, suffixes) {
  deg = wrap(deg, -180, 180);
  const rounded = round10(Math.abs(deg), -6);
  return (
    rounded + '°' + (
      rounded === 0
      ? ''
      : ' ' + (deg < 0 ? suffixes[1] : suffixes[0])
    )
  );
}

const tmp = [0, 0];

export class LatLonGridControl extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this.node = null;
    this.lonMinDist = 150;
    this.latMinDist = 150;
    this._withGridLines = false;
    this._ticks = [];
    this._lines = [];
    this._positionsCache = {};
    this._clientRect = null;
    this._projection = null;
    this._extent = null;
    this._resolution = null;
    this._tickIndex = null;
    this._lineIndex = null;

    this.node = document.createElement('div');
    this.node.className = 'LatLonGridControl';

    this.shortcuts = [
      {
        groupLabel : 'Global',
        id         : 'toggle-grid-visibility',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.L,
        description: 'Toggle the visibility of the grid',
        autoDisable: true,
        handler    : () => {
          this.withGridLines = !this.withGridLines;
          return true;
        },
      },
    ];

    this.render = throttleOncePerAnimationFrame(this.render);
  }

  get state() {
    const {withGridLines} = this;
    return Object.assign(super.state, {withGridLines});
  }
  set state(state) {
    super.state = state;
    this.withGridLines = state.withGridLines;
  }

  get withGridLines() {
    return this._withGridLines;
  }
  set withGridLines(withGridLines) {
    if (this._withGridLines !== withGridLines) {
      this._withGridLines = withGridLines;
      this.render();
      this.hub.emit('change:withGridLines');
      this.hub.emit('change:state');
    }
  }

  onActivate() {
    this.syntool.usableViewportNode.appendChild(this.node);
    this.syntool.hub.on('change:usable-viewport', this.render, this);
    this.syntool.mapView.hub.on('changing:extent', this.render, this);
    this.render();
  }
  onDeactivate() {
    this.syntool.hub.off('change:usable-viewport', this.render, this);
    this.syntool.mapView.hub.off('changing:extent', this.render, this);
    this.syntool.usableViewportNode.removeChild(this.node);
  }

  render() {
    if (!this.active) {
      return;
    }

    const {mapView} = this.syntool;
    const mapCenterClientXY = mapView.getViewportCenterClientXY();
    const scale = mapView.getViewportScaleInClient();

    this._clientRect = this.node.getBoundingClientRect();
    this._projection = mapView.getProjection();
    this._resolution = mapView.getResolution();
    this._centerXY = [
      (mapCenterClientXY[0] - this._clientRect.left) * scale[0],
      (mapCenterClientXY[1] - this._clientRect.top) * scale[1],
    ];
    this._projectedCenterXY = mapView.getViewportCenterProjectedXY();

    let bounds = Bounds.createFromXYs([
      this.projectRelativeXY([0, 0]),
      this.projectRelativeXY([this.node.clientWidth, this.node.clientHeight]),
    ]);

    const maxProjectedBounds = mapView.getMaxProjectedBounds();
    if (mapView.shouldWrapDateLine()) {
      bounds = Bounds.create(
        Bounds.getLeft(bounds),
        Math.max(Bounds.getBottom(maxProjectedBounds), Bounds.getBottom(bounds)),
        Bounds.getRight(bounds),
        Math.min(Bounds.getTop(maxProjectedBounds), Bounds.getTop(bounds))
      );
    } else {
      bounds = Bounds.create(
        Math.max(Bounds.getLeft(maxProjectedBounds), Bounds.getLeft(bounds)),
        Math.max(Bounds.getBottom(maxProjectedBounds), Bounds.getBottom(bounds)),
        Math.min(Bounds.getRight(maxProjectedBounds), Bounds.getRight(bounds)),
        Math.min(Bounds.getTop(maxProjectedBounds), Bounds.getTop(bounds))
      );
    }

    var latLonBounds = LatLonBounds.round(
      Projection.unprojectBounds(this._projection, bounds),
      round
    );

    this._extent = {
      left: LatLonBounds.getLeft(latLonBounds),
      right: LatLonBounds.getRight(latLonBounds),
      top: LatLonBounds.getTop(latLonBounds),
      bottom: LatLonBounds.getBottom(latLonBounds),
    };
    this._tickIndex = 0;
    this._lineIndex = 0;

    this.renderVertical('top');
    this.renderVertical('bottom');
    this.renderHorizontal('left');
    this.renderHorizontal('right');

    this.reuseRemainingTicks();
    this.reuseRemainingLines();
  }

  /*Protected*/
  projectRelativeXY(xy) {
    xy[0] = (
      this._projectedCenterXY[0] +
      (xy[0] - this._centerXY[0]) * this._resolution
    );
    xy[1] = (
      this._projectedCenterXY[1] -
      (xy[1] - this._centerXY[1]) * this._resolution
    );
    return xy;
  }

  renderTick(position, projectedXY, deg) {
    var x = Math.round(
      this._centerXY[0] +
      (projectedXY[0] - this._projectedCenterXY[0]) / this._resolution
    );
    var y = Math.round(
      this._centerXY[1] -
      (projectedXY[1] - this._projectedCenterXY[1]) / this._resolution
    );
    var index = this._tickIndex++;
    var tick;
    var label;

    if (this._ticks.length > index) {
      tick = this._ticks[index];
    } else {
      tick = document.createElement('div');
      label = document.createElement('span');
      label.className = 'LatLonGridControlLabel';
      tick.appendChild(label);
      tick.label = label;
      this._ticks.push(tick);
    }
    if (tick.parentElement !== this.node) {
      this.node.appendChild(tick);
    }

    tick.className = 'LatLonGridControlTick -' + position;
    switch (position) {
    case 'top':
      tick.style.top    = '0';
      tick.style.bottom = '';
      tick.style.left   = x + 'px';
      tick.style.right  = '';
      if (this.withGridLines) {
        this.renderLine('vertical', x, y);
      }
      break;
    case 'bottom':
      tick.style.top    = '';
      tick.style.bottom = '0';
      tick.style.left   = x + 'px';
      tick.style.right  = '';
      break;
    case 'left':
      tick.style.top    = y + 'px';
      tick.style.bottom = '';
      tick.style.left   = '0';
      tick.style.right  = '';
      if (this.withGridLines) {
        this.renderLine('horizontal', x, y);
      }
      break;
    case 'right':
      tick.style.top    = y + 'px';
      tick.style.bottom = '';
      tick.style.left   = '';
      tick.style.right  = '0';
      break;
    default:
    }
    tick.label.textContent = formatDegrees(
      deg,
      position === 'top' || position === 'bottom' ? 'EW' : 'NS'
    );
  }
  reuseRemainingTicks() {
    var index = this._tickIndex;
    while (index < this._ticks.length) {
      this._ticks[index++].remove();
    }
  }

  renderLine(position, x, y) {
    var index = this._lineIndex++;
    var line;

    if (this._lines.length > index) {
      line = this._lines[index];
    } else {
      line = document.createElement('div');
      this._lines.push(line);
    }
    if (line.parentElement !== this.node) {
      this.node.appendChild(line);
    }

    line.className = 'LatLonGridControlLine -' + position;
    switch (position) {
    case 'vertical':
      line.style.top    = '0';
      line.style.bottom = '';
      line.style.left   = x + 'px';
      line.style.right  = '';
      break;
    case 'horizontal':
      line.style.top    = y + 'px';
      line.style.bottom = '';
      line.style.left   = '0';
      line.style.right  = '';
      break;
    default:
    }
  }
  reuseRemainingLines() {
    var index = this._lineIndex;
    while (index < this._lines.length) {
      this._lines[index++].remove();
    }
  }

  getClosestScale(number) {
    var sign;
    var pow10;
    var firstDigit;
    var scale;

    if (number === 0) {
      return 0;
    }

    if (number < 0) {
      number = -number;
      sign = -1;
    } else {
      sign = 1;
    }

    pow10 = magnitude(number);
    firstDigit = number / pow10 | 0;
    if (firstDigit < 1) {
      scale = 1;
    } else if (firstDigit < 2) {
      scale = 2;
    } else if (firstDigit < 5) {
      scale = pow10 < 10 ? 5 : 4.5;
    } else {
      scale = pow10 < 10 ? 10 : 9;
    }

    return sign * scale * pow10;
  }

  addToCache(position, location) {
    var cache = this._positionsCache[this._resolution];
    if (!cache) {
      cache = {};
      this._positionsCache[this._resolution] = cache;
    }
    if (!(position in cache)) {
      cache[position] = [0];
    }
    addToUniqueSorted(cache[position], location);
  }
  getFromCache(position, location) {
    var cache = this._positionsCache[this._resolution];
    if (cache && position in cache) {
      cache = cache[position];
      var index = bisectRight(cache, location);
      if (location < 0) {
        if (index === cache.length) {
          return 0;
        }
        return cache[index];
      }
      if (index === 0) {
        return 0;
      }
      return cache[index - 1];
    }
    return 0;
  }

  renderVertical(position) {
    var minDist = this.lonMinDist * this._resolution;
    var last;
    var lonlat;
    var projectedXY;
    var counter;

    switch (this.syntool.mapView.getZoom()) {
    case 0:
      minDist /= 2;
      /*fallthrough*/
    case 1:
    case 2:
      minDist /= 2;
      /*fallthrough*/
    default:
    }

    last = this.getFromCache(position, Math.max(0, this._extent.left));
    counter = 0;
    while (round(last) <= this._extent.right) {
      tmp[0] = last;
      tmp[1] = this._extent[position];
      projectedXY = Projection.projectLonLat(this._projection, tmp, tmp);

      if (round(last) >= this._extent.left) {
        this.renderTick(position, projectedXY, last);
      }

      projectedXY[0] += minDist;
      lonlat = Projection.unprojectXY(this._projection, projectedXY, tmp);
      last += this.getClosestScale(lonlat[0] - last);

      if (++counter % 2000 === 0) {
        counter = 0;
        this.addToCache(position, round(last));
      }
    }

    last = this.getFromCache(position, Math.min(0, this._extent.right));
    counter = 0;
    while (round(last) >= this._extent.left) {
      tmp[0] = last;
      tmp[1] = this._extent[position];
      projectedXY = Projection.projectLonLat(this._projection, tmp, tmp);

      if (last !== 0 && round(last) <= this._extent.right) {
        this.renderTick(position, projectedXY, last);
      }

      projectedXY[0] -= minDist;
      lonlat = Projection.unprojectXY(this._projection, projectedXY, tmp);
      last += this.getClosestScale(lonlat[0] - last);

      if (++counter % 2000 === 0) {
        counter = 0;
        this.addToCache(position, round(last));
      }
    }
  }
  renderHorizontal(position) {
    var minDist = this.latMinDist * this._resolution;
    var last;
    var lonlat;
    var projectedXY;
    var counter;

    switch (this.syntool.mapView.getZoom()) {
    case 0:
      minDist /= 2;
      /*fallthrough*/
    case 1:
      minDist /= 2;
      /*fallthrough*/
    case 2:
      minDist /= 2;
      /*fallthrough*/
    default:
    }

    last = this.getFromCache(position, Math.max(0, this._extent.bottom));
    counter = 0;
    while (round(last) <= this._extent.top) {
      tmp[0] = this._extent[position];
      tmp[1] = last;
      projectedXY = Projection.projectLonLat(this._projection, tmp, tmp);

      if (round(last) >= this._extent.bottom) {
        this.renderTick(position, projectedXY, last);
      }

      projectedXY[1] += minDist;
      lonlat = Projection.unprojectXY(this._projection, projectedXY, tmp);
      last += this.getClosestScale(lonlat[1] - last);

      if (++counter % 2000 === 0) {
        counter = 0;
        this.addToCache(position, round(last));
      }
    }

    last = this.getFromCache(position, Math.min(0, this._extent.top));
    counter = 0;
    while (round(last) >= this._extent.bottom) {
      tmp[0] = this._extent[position];
      tmp[1] = last;
      projectedXY = Projection.projectLonLat(this._projection, tmp, tmp);

      if (last !== 0 && round(last) <= this._extent.top) {
        this.renderTick(position, projectedXY, last);
      }

      projectedXY[1] -= minDist;
      lonlat = Projection.unprojectXY(this._projection, projectedXY, tmp);
      last += this.getClosestScale(lonlat[1] - last);

      if (++counter % 2000 === 0) {
        counter = 0;
        this.addToCache(position, round(last));
      }
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {LatLonGridControl});
