/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  parseDecimal,
  H,
  EventsHub,
  DOMEvents,
  renderModifiedKey,
} from '../../syntool.js';
import {removeNodes, linkStyleSheet} from '../../utils/dom_utils.js';

linkStyleSheet('css/notifications_view.css');

export class NotificationsView {
  constructor() {
    this.node = document.createElement('div');
    this.node.className = 'NotificationsView';
    this._store = null;
    this._updateTimerId = null;
    this._timersByMessageId = {};

    this.hub = new EventsHub();

    // Event handlers
    // ==============
    this.handleUpdateTimer = this.handleUpdateTimer.bind(this);

    const view = this;
    DOMEvents(this.node)
    .on('click', '.js-custom-btn', function(event) {
      event.preventDefault();

      const buttonIndex = parseDecimal(this.getAttribute('data-btn-index'));
      const message = view.getMessageForNode(this);

      message.buttons[buttonIndex].onClick();
    })
    .on('click', '.js-close', function(event) {
      event.preventDefault();
      const messageNode = view.getMessageNodeForNode(this);

      if (!this.getAttribute('data-btn-index')) {
        const message = view.getMessageForNode(this);
        message.closeButton?.onClick?.();
      }
      if (messageNode) {
        view.closeMessageWithNode(messageNode);
      }
    });

    this.hub.on('close:message', (id) => this.store?.removeMessageWithId(id));
  }

  get store() {
    return this._store;
  }
  set store(store) {
    if (this.store !== store) {
      if (this.store) {
        this.store.hub.off('change', this.render, this);
      }
      this._store = store;
      if (this.store) {
        this.store.hub.on('change', this.render, this);
        this.render();
      }
    }
  }

  getMessageNodeForNode(node) {
    let parent = node && node.parentNode;

    while (parent && parent !== this.node) {
      node = parent;
      parent = parent.parentNode;
    }

    return parent && node;
  }
  getMessageForNode(node) {
    const messageNode = this.getMessageNodeForNode(node);
    const messageId = messageNode?.getAttribute('data-id');
    return this.store?.getMessageWithId(messageId);
  }
  closeMessageWithNode(messageNode) {
    this.hub.emit('close:message', messageNode.getAttribute('data-id'));
  }

  handleUpdateTimer() {
    if (!this.store) {
      return;
    }
    let i = this.store.messages.length;
    while (i--) {
      const message = this.store.messages[i];
      if (message.hideAfter > 0) {
        const node = this.node.querySelector('[data-id="' + message.id + '"]');
        if (node) {
          this.updateMessageRemainingTime(node, message);
        }
      }
    }
  }

  updateMessage(node, message) {
    const body = node.querySelector('.NotificationMessage-Body');
    const close = node.querySelector('.NotificationMessage-Close');

    node.className = 'NotificationMessage -' + message.type;
    node.setAttribute('data-id', message.id);
    if (message.key) {
      body.innerHTML = (
        renderModifiedKey(message.modifiers, message.key) + ' : ' +
        message.text
      );
    } else {
      body.innerHTML = message.text;
    }

    const customButtons = [
      ...node.querySelectorAll('.NotificationMessage-CustomBtn'),
    ];
    if (message.buttons) {
      for (let i = 0, l = message.buttons.length; i < l; i++) {
        const button = message.buttons[i];
        const customButton = (
          customButtons[i] || document.createElement('button')
        );
        customButton.textContent = button.text;

        const props = Object.assign({type: 'button', className: ''}, button.props);
        props.className = 'NotificationMessage-CustomBtn ' + props.className;
        for (let [key, value] of Object.entries(props)) {
          if (key in customButton) {
            customButton[key] = value;
          } else {
            customButton.setAttribute(key, value);
          }
        }
        if (button.onClick) {
          customButton.classList.add('js-custom-btn');
          customButton.dataset.btnIndex = i;
        }
        if (button.closes) {
          customButton.classList.add('js-close');
        }

        node.insertBefore(customButton, close);
      }
    }
    if (message.buttons && customButtons.length > message.buttons.length) {
      removeNodes(customButtons.slice(message.buttons.length));
    }
    this.updateMessageRemainingTime(node, message);
  }
  updateMessageRemainingTime(node, message) {
    const remainingSeconds = (
      message.hideAfter >= 0 && message.addedAt
      ? Math.ceil((message.addedAt + message.hideAfter - Date.now()) / 1000)
      : -1
    );
    const close = node.querySelector('.NotificationMessage-Close');
    const label = message.closeButton?.text ?? 'close';
    const props = Object.assign({
      className: '',
      title: 'Close this message',
    }, message.closeButton?.props);
    props.className = `NotificationMessage-Close ${props.className} js-close`;
    for (let [key, value] of Object.entries(props)) {
      if (key in close) {
        close[key] = value;
      } else {
        close.setAttribute(key, value);
      }
    }
    const buttonText = (
      remainingSeconds < 0 ? label : `${label} (${remainingSeconds})`
    );
    if (close.textContent !== buttonText) {
      close.textContent = buttonText;
    }
  }

  render() {
    const nodes = this.node.children;
    const nodesByMessageId = {};
    let needsTimer = false;
    let i;

    i = nodes.length;
    while (i--) {
      nodesByMessageId[nodes[i].getAttribute('data-id')] = nodes[i];
    }

    i = this.store?.messages.length ?? 0;
    while (i--) {
      const message = this.store.messages[i];
      let node = nodesByMessageId[message.id];

      if (node) {
        delete nodesByMessageId[message.id];
      } else {
        node = H('div', {class: 'NotificationMessage'}, [
          H('span', {class: 'NotificationMessage-Body'}),
          H('button', {
            class: 'NotificationMessage-Close js-close',
            title: 'Close this message',
          }, 'close'),
        ]);
      }

      if (message.hideAfter > 0) {
        needsTimer = true;
      }

      this.updateMessage(node, message);
      this.node.insertBefore(node, this.node.firstChild);
    }

    const remainingNodes = Object.values(nodesByMessageId);

    i = remainingNodes.length;
    while (i--) {
      this.node.removeChild(remainingNodes[i]);
    }

    if (needsTimer) {
      if (this._updateTimerId === null) {
        this._updateTimerId = setInterval(this.handleUpdateTimer, 100);
      }
    } else if (this._updateTimerId !== null) {
      clearInterval(this._updateTimerId);
      this._updateTimerId = null;
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {NotificationsView});
