/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {throttleOncePerEventLoop, H, DOMEvents} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {
  ProductsSettingsListView,
} from '../products_settings/products_settings_list_view.js';
import {Dialog} from './dialog.js';

linkStyleSheet('css/products_display_settings_dialog.css');

export class ProductsDisplaySettingsDialog extends Dialog {
  constructor(options) {
    super('ProductsDisplaySettingsDialog', {
      title    : 'Display data',
      resizable: true,
      pinnable : true,

      pinTooltip: "Toggle dialog's full opacity",

      onPinnedChange(dialog, pinned) {
        dialog.node.style.opacity = pinned ? 1 : '';
      },
    }, options);

    /** @type {import('../../utils/catalog_store.js').CatalogStore | null} */
    this.store = null;
    this.listView = new ProductsSettingsListView(this.contentNode);

    this._addButtonNode = H('button', {
      class: 'Dialog-Button --Add js-add-button',
      title: 'Add more products',
    }, [H('i', {class: 'fa fa-plus'}), 'more']);
    this.titleNode.parentNode.insertBefore(
      this._addButtonNode,
      this.titleNode.nextElementSibling
    );

    DOMEvents(this._addButtonNode).on('click', this.handleAddClick, this);

    this._renderListView = throttleOncePerEventLoop(() => {
      if (this.syntool) {
        this.listView.render(this._getProductIds());
      }
    });
    this._handleDatasetSelectionChange = throttleOncePerEventLoop(() => {
      if (this.syntool) {
        this.listView.selectedDatasetsByProduct = (
          this.syntool.mapView.getSelectedDatasetsByProduct()
        );
        this.listView.render(this._getProductIds());
      }
    });

    this.listView.hub
      .on('remove-product', this.handleRemoveProduct, this)
      .on('select-product', this.handleSelectProduct, this)
      .on('select-one-product', this.handleSelectOneProduct, this)
      .on('deselect-product', this.handleDeselectProduct, this)
      .on('change:opacity', this.handleOpacityChange, this)
      .on('change:stackLevels', this.handleStackLevelsChange, this)
      .on('select:dataset-filter', this.handleSelectDatasetFilter, this)
      .on('select-one:dataset-filter', this.handleSelectOneDatasetFilter, this)
      .on('deselect:dataset-filter', this.handleDeselectDatasetFilter, this)
      .on('change:filterPattern', this.handleFilterPatternChange, this)
      .on('change:filter:activate', this.handleFilterActivateChange, this)
      .on('change:filter:deactivate', this.handleFilterDeactivateChange, this);
  }

  /*Protected - Control overrides*/
  onAdd() {
    super.onAdd();

    this.store = this.syntool.catalogStore;
    this.listView.store = this.store.productsStore;
    this.listView.render(this._getProductIds());
  }
  onRemove() {
    super.onRemove();

    this.store = null;
    this.listView.store = null;
  }
  onActivate() {
    super.onActivate();

    this.store.hub.on('change:selection', this._renderListView);
    this.store.productsStore.hub
      .on('change:selection', this._renderListView)
      .on('change:opacity', this._renderListView)
      .on('change:stackLevel', this._renderListView)
      .on('change:pattern', this._renderListView);

    this.syntool.hub.on('change:screenshot-mode', this._renderListView);
    this.syntool.mapView.hub
      .on('change:selection', this._handleDatasetSelectionChange);
  }
  onDeactivate() {
    super.onDeactivate();

    this.store.hub.off('change:selection', this._renderListView);
    this.store.productsStore.hub
      .off('change:selection', this._renderListView)
      .off('change:opacity', this._renderListView)
      .off('change:stackLevel', this._renderListView)
      .off('change:pattern', this._renderListView);

    this.syntool.hub.off('change:screenshot-mode', this._renderListView);
    this.syntool.mapView.hub
      .off('change:selection', this._handleDatasetSelectionChange);
  }

  /*Protected*/
  _getProductIds() {
    let productIds = this.store.getSelectedProductIds();
    if (this.syntool.screenshotModeActive) {
      productIds = (
        productIds.filter(this._productWithIdHasVisibleDatasets, this)
      );
    }
    return productIds;
  }
  _productWithIdHasVisibleDatasets(productId) {
    return (
      this.syntool.mapView.hasDatasetsInMapForProductWithId(productId) &&
      this.syntool.filters.product.zoom.withId(productId)
    );
  }

  /*Protected - Event handlers*/
  /**
   * @protected
   * @param {ProductId} ProductId
   * @returns {void}
   */
  handleRemoveProduct(productId) {
    this.store.deselectProduct(productId);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @returns {void}
   */
  handleSelectProduct(productId) {
    this.store.productsStore.selectProduct(productId);
    this.syntool.executeShortcutWithId(
      'auto-jump-to-nearest-dataset',
      true /*notify*/,
      productId
    );
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @returns {void}
   */
  handleSelectOneProduct(productId) {
    this.store.productsStore.selectOnlyProducts([productId]);
    this.syntool.executeShortcutWithId(
      'auto-jump-to-nearest-dataset',
      true /*notify*/,
      productId
    );
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @returns {void}
   */
  handleDeselectProduct(productId) {
    this.store.productsStore.deselectProduct(productId);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @param {number} opacity
   * @returns {void}
   */
  handleOpacityChange(productId, opacity) {
    this.store.productsStore.setProductOpacity(productId, opacity);
  }
  /**
   * @protected
   * @param {Record<ProductId, number>} stackLevelsByProduct
   * @returns {void}
   */
  handleStackLevelsChange(stackLevelsByProduct) {
    this.store.productsStore.setProductStackLevels(stackLevelsByProduct);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @param {string} filter
   * @returns {void}
   */
  handleSelectDatasetFilter(productId, filter) {
    this.store.productsStore.activateProductDatasetFilter(productId, filter);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @param {string} group
   * @param {string} filter
   * @returns {void}
   */
  handleSelectOneDatasetFilter(productId, group, filter) {
    const store = this.store.productsStore;
    const product = store.getProduct(productId);
    if (!product?.datasetFilters) {
      return;
    }

    const activeFilters = new Set(store.productActiveDatasetFilters(productId));
    for (const {label, filters} of product.datasetFilters) {
      if (label === group) {
        for (const {id} of filters) {
          if (filter === id) {
            activeFilters.add(id);
          } else {
            activeFilters.delete(id);
          }
        }
        break;
      }
    }
    store.setProductActiveDatasetFilters(productId, activeFilters);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @param {string} filter
   * @returns {void}
   */
  handleDeselectDatasetFilter(productId, filter) {
    this.store.productsStore.deactivateProductDatasetFilter(productId, filter);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @param {LegacyPattern} pattern
   * @returns {void}
   */
  handleFilterPatternChange(productId, pattern) {
    this.store.productsStore.setProductPattern(productId, pattern);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @returns {void}
   */
  handleFilterActivateChange(productId) {
    this.store.productsStore.activateProductPattern(productId);
  }
  /**
   * @protected
   * @param {ProductId} ProductId
   * @returns {void}
   */
  handleFilterDeactivateChange(productId) {
    this.store.productsStore.deactivateProductPattern(productId);
  }
  /**
   * @protected
   * @returns {void}
   */
  handleAddClick() {
    this.hub.emit('click:add');
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProductsDisplaySettingsDialog});
