/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  H,
  DOMEvents,
  whenReady,
  Config,
  RegExpHelpers,
  Pattern,
  Control,
  Types as is,
} from '../../syntool.js';
import {FlagsStore} from '../../features/flags_store.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {ProductsListHeaderView} from '../products_panel/products_list_header_view.js';
import {ProductsListView} from '../products_panel/products_list_view.js';
import {MarkdownDialog} from './markdown_dialog.js';

linkStyleSheet('css/products_panel.css');

const productTagRe = new RegExp(RegExpHelpers.productTag);

function getProductId(product) {
  return product.id;
}

function computeGroupTagsForProducts(products) {
  const groupTags = new Map();

  products.forEach(function(product) {
    for (const tag of product.tags) {
      const parts = productTagRe.exec(tag);
      const key = parts[1];
      const groupId = key && key.toLowerCase();
      if (key && !groupTags.has(groupId)) {
        groupTags.set(groupId, key);
      }
    }
  });

  return [...groupTags.values()].sort();
}

whenReady(function() {
  Object.assign(Config.configFieldValidators, {
    groupProductsBy  : is.oneOfType([is.value(false), is.string]),
    defaultGroupLabel: is.string,
    sortGroups       : is.func,
    collapsibleGroups: is.bool,
  });
  Config.configNormalizers.push(function normalizeConfigWithDefaults(config) {
    return Object.assign({
      groupProductsBy  : config.showGroups ? 'Group' : false,
      defaultGroupLabel: null,
      collapsibleGroups: false,
    }, config);
  });
});

export class ProductsPanel extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this.node = H('div', {class: 'UISidebar ProductsListPanel -closed'}, [
      H('div', {class: 'lid horizontal-lid h-hidden-in-screenshot-mode'}, [
        'Products',
        this.productsWarningIndicatorNode = H('div', {
          id: 'products-warning-indicator',
          class: 'warning-icon',
          title: 'Some products are hidden',
        }),
        this.productsErrorIndicatorNode = H('div', {
          id: 'products-error-indicator',
          class: 'error-icon',
          title: (
            'Some products are unavailable due to server or network errors, ' +
            'they appear with a red checkbox in the products list. You can ' +
            'filter them out using the toggle button located in the list ' +
            'header.'
          ),
        }),
      ]),
    ]);

    this._search = '';
    this.productInfoDialog = new MarkdownDialog('product-info-dialog');
    this.productsListHeaderView = new ProductsListHeaderView();
    this.productsListView = new ProductsListView();

    this.node.appendChild(this.productsListHeaderView.node);
    this.node.appendChild(this.productsListView.node);

    DOMEvents(this.node).on('click', '.lid', () => (this.open = !this.open));
    this.productsListHeaderView.hub
      .on('show-unreachable-toggle', this.handleShowUnreachableToggle, this)
      .on('group-by-change', this.handleGroupByChange, this)
      .on('search-change', this.handleSearchChange, this);
    this.productsListView.hub
      .on('change:collapsedGroups', () => this.hub.emit('change:state'))
      .on('select-product', this.handleSelectProduct, this)
      .on('select-one-product', this.handleSelectOneProduct, this)
      .on('deselect-product', this.handleDeselectProduct, this)
      .on('click:product-info', this.handleProductInfoClick, this)
      .on('click:product-delete', this.handleProductDeleteClick, this)
      .on('click:product-zoom-in', this.handleProductZoomInClick, this)
      .on('click:product-zoom-out', this.handleProductZoomOutClick, this);
  }

  get open() {
    return !this.node.classList.contains('-closed');
  }
  set open(open) {
    this.node.classList.toggle('-closed', !open);
    this.hub.emit('change:state');
  }
  get groupByTag() {
    return this.productsListView.groupBy;
  }
  set groupByTag(groupTag) {
    this.productsListHeaderView.selectedGroupTag = groupTag;
    this.productsListView.groupBy = groupTag;
    this.productsListHeaderView.render();
    this.productsListView.render();
    this.hub.emit('change:state');
  }
  get search() {
    return this._search;
  }
  set search(search) {
    this.productsListHeaderView.search = search;
  }
  get sortGroups() {
    return this.productsListView.sortGroups;
  }
  get defaultGroupLabel() {
    return this.productsListView.defaultGroupLabel;
  }
  set defaultGroupLabel(label) {
    if (this.productsListView.defaultGroupLabel !== label) {
      this.productsListView.defaultGroupLabel = label;
      this.productsListView.render();
      this.hub.emit('change:state');
    }
  }
  get groupsCollapsible() {
    return this.productsListView.groupsCollapsible;
  }
  set groupsCollapsible(collapsible) {
    this.productsListView.groupsCollapsible = collapsible;
    this.hub.emit('change:state');
  }
  get collapsedGroups() {
    return this.productsListView.collapsedGroups;
  }
  set collapsedGroups(collapsedGroups) {
    this.productsListView.collapsedGroups = collapsedGroups;
  }
  get hasUnreachableProducts() {
    return this.productsListHeaderView.hasUnreachableProducts;
  }
  get hideUnreachableProducts() {
    return this.productsListHeaderView.hideUnreachableProducts;
  }
  set hideUnreachableProducts(hide) {
    this.productsListHeaderView.hideUnreachableProducts = hide;
    this.hub.emit('change:state');
  }

  get state() {
    return Object.assign(super.state, {
      open: this.open,
      search: this.search,
      groupByTag: this.groupByTag,
      defaultGroupLabel: this.defaultGroupLabel,
      groupsCollapsible: this.groupsCollapsible,
      collapsedGroups: [...this.collapsedGroups],
      hideUnreachableProducts: this.hideUnreachableProducts,
    });
  }
  set state(state) {
    const {
      open,
      search = '',
      groupByTag = this.groupByTag,
      defaultGroupLabel = this.defaultGroupLabel,
      groupsCollapsible = this.groupsCollapsible,
      collapsedGroups = [],
      hideUnreachableProducts = this.hideUnreachableProducts,
    } = state;
    super.state = state;
    this.open = open;
    this.search = search;
    this.groupByTag = groupByTag;
    this.defaultGroupLabel = defaultGroupLabel;
    this.groupsCollapsible = groupsCollapsible;
    this.collapsedGroups = new Set(collapsedGroups);
    this.hideUnreachableProducts = hideUnreachableProducts;
  }

  setGroupByTag(groupTag, options) {
    if (options && options.save) {
      FlagsStore.set('groupProductsBy', groupTag);
    }
    this.groupByTag = groupTag;
  }
  focusOnSearch() {
    this.open = true;
    this.productsListHeaderView.focusOnSearch();
  }

  showInfoForProductWithId(productId) {
    if (!this.syntool) {
      return;
    }

    const product = this.syntool.catalogStore.getProduct(productId);

    if (product.infoURL) {
      this.productInfoDialog.show({
        title     : product.textLabel.trim(),
        contentURL: product.infoURL,
        onError   : {
          console: (
            'Error loading info for product with id "' + productId + '" ' +
            'from URL "' + product.infoURL + '"!'
          ),
          header: 'Error loading info for product!',
        },
        onShow(dialog) {
          dialog.contentNode.scrollTop = 0;
        },
      });
    }
  }

  /*Protected - Control overrides*/
  onAdd() {
    this.productsListView.syntool = this.syntool;
    this.productsListView.store = this.syntool.catalogStore;
    this.productsListView.productIds = this._getProductIds();
    this.syntool.addControl(this.productInfoDialog);
    this.productsListView.render();
  }
  onRemove() {
    this.syntool.removeControl(this.productInfoDialog);
    this.productsListView.syntool = null;
    this.productsListView.store = null;
    this.productsListView.productIds = [];
  }
  onActivate() {
    const {syntool} = this;

    syntool.api.hub
      .on('change:unreachableServers', this.handleUnreachableChange, this)
      .on('change:unreachableServers', this.updateErrorIndicator, this);

    syntool.mapView.hub
      .on('change:zoom', this.updateWarningIndicator, this)
      .on('change:in-map-datasets', this.updateWarningIndicator, this)
      .on('change:zoom', this.productsListView.render, this.productsListView);

    syntool.productsStore.hub
      .on('change:selection', this.updateWarningIndicator, this)
      .on('change:pattern', this.updateWarningIndicator, this)
      .on('change:opacity', this.updateWarningIndicator, this)
      .on('change:selection', this.productsListView.render, this.productsListView)
      .on('change:pattern', this.productsListView.render, this.productsListView)
      .on('change:opacity', this.productsListView.render, this.productsListView);

    syntool.catalogStore.hub
      .on('change:products', this.handleProductsChange, this)
      .on('change:selection', this.updateWarningIndicator, this)
      .on('change:selection', this.productsListView.render, this.productsListView);

    syntool.whenInitialised(() => {
      if (this.syntool !== syntool) {
        return;
      }

      const {config} = syntool;
      const groupTag = (
        FlagsStore.get('groupProductsBy') ?? config.groupProductsBy
      );

      this.productsListHeaderView.selectedGroupTag = groupTag;
      this.productsListView.groupBy = groupTag;
      this.productsListView.defaultGroupLabel = config.defaultGroupLabel;
      this.productsListView.sortGroups = config.sortGroups;
      this.productsListView.collapsibleGroups = config.collapsibleGroups;
      this.handleProductsChange();
      this.updateWarningIndicator();
    });

    this.productInfoDialog.active = true;
    this.productsListHeaderView.render();
    this.productsListView.render();
    syntool.usableViewportNode.appendChild(this.node);
  }
  onDeactivate() {
    this.productInfoDialog.active = false;

    this.syntool.api.hub
      .off('change:unreachableServers', this.handleUnreachableChange, this)
      .off('change:unreachableServers', this.updateErrorIndicator, this);

    this.syntool.mapView.hub
      .off('change:zoom', this.updateWarningIndicator, this)
      .off('change:in-map-datasets', this.updateWarningIndicator, this)
      .off('change:zoom', this.productsListView.render, this.productsListView);

    this.syntool.productsStore.hub
      .off('change:selection', this.updateWarningIndicator, this)
      .off('change:pattern', this.updateWarningIndicator, this)
      .off('change:opacity', this.updateWarningIndicator, this)
      .off('change:selection', this.productsListView.render, this.productsListView)
      .off('change:pattern', this.productsListView.render, this.productsListView)
      .off('change:opacity', this.productsListView.render, this.productsListView);

    this.syntool.catalogStore.hub
      .off('change:products', this.handleProductsChange, this)
      .off('change:selection', this.updateWarningIndicator, this)
      .off('change:selection', this.productsListView.render, this.productsListView);

    this.syntool.usableViewportNode.removeChild(this.node);
  }

  /*Protected - Event handlers*/
  handleUnreachableChange() {
    this.productsListHeaderView.hasUnreachableProducts = (
      this.syntool.catalogStore.getProducts()
      .some(this.productShownUnreachable, this)
    );
    this.productsListView.render();
  }

  handleShowUnreachableToggle() {
    this.hideUnreachableProducts = !this.hideUnreachableProducts;
  }

  handleGroupByChange(groupTag) {
    this.setGroupByTag(groupTag, {save: true});
  }
  handleSearchChange(search) {
    this._search = search.trim();
    this.productsListView.productIds = this._getProductIds();
    this.productsListView.render();
    this.hub.emit('change:state');
  }

  handleSelectProduct(productId) {
    this.syntool.catalogStore.selectProduct(productId);
    this.syntool.executeShortcutWithId(
      'auto-jump-to-nearest-dataset',
      true /*notify*/,
      productId
    );
  }
  handleSelectOneProduct(productId) {
    this.syntool.catalogStore.selectOnlyProducts([productId]);
    this.syntool.executeShortcutWithId(
      'auto-jump-to-nearest-dataset',
      true /*notify*/,
      productId
    );
  }
  handleDeselectProduct(productId) {
    this.syntool.catalogStore.deselectProduct(productId);
  }
  handleProductInfoClick(productId) {
    this.showInfoForProductWithId(productId);
  }
  handleProductDeleteClick(productId) {
    this.syntool.catalogStore.deleteProduct(productId);
  }
  handleProductZoomInClick(productId) {
    const product = this.syntool.catalogStore.getProduct(productId);
    this.syntool.mapView.setZoom(product.mapMinZoom);
  }
  handleProductZoomOutClick(productId) {
    const product = this.syntool.catalogStore.getProduct(productId);
    this.syntool.mapView.setZoom(product.mapMaxZoom);
  }

  handleProductsChange() {
    this.productsListHeaderView.groupTags = (
      computeGroupTagsForProducts(this.syntool.catalogStore.getProducts())
    );
    this.productsListView.productIds = this._getProductIds();
    this.productsListHeaderView.render();
    this.productsListView.render();
    this.updateErrorIndicator();
  }

  /*Protected*/
  _getProductIds() {
    let products = this.syntool.catalogStore.searchProducts(this.search);
    if (this.hideUnreachableProducts) {
      products = products.filter(this.isProductSelectedOrReachable, this);
    }
    return products.map(getProductId);
  }

  isProductSelectedOrReachable(product) {
    return (
      this.syntool.catalogStore.isProductSelected(product.id) ||
      this.syntool.filters.product.withReachableServer(product)
    );
  }
  productShownUnreachable(product) {
    return (
      !this.syntool.catalogStore.isProductSelected(product.id) &&
      !this.syntool.filters.product.withReachableServer(product)
    );
  }

  updateWarningIndicator() {
    this.productsWarningIndicatorNode.style.display = (
      this.syntool.catalogStore.getSelectedProducts()
        .some(this.productHasWarnings, this)
      ? 'block'
      : ''
    );
  }
  updateErrorIndicator() {
    this.productsErrorIndicatorNode.style.display = (
      this.syntool.catalogStore.getSelectedProducts()
        .some(this.syntool.filters.product.withUnreachableServer)
      ? 'block'
      : ''
    );
  }

  productHasWarnings(product) {
    return (
      product.opacity === 0 ||
      !this.syntool.filters.product.selected(product) ||
      '' === Pattern.simplify(
        this.syntool.productsStore.getProductComputedPattern(product.id),
        'auto',
        false
      ) ||
      (
        this.syntool.mapView.hasDatasetsInMapForProductWithId(product.id) &&
        !this.syntool.filters.product.zoom(product)
      )
    );
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProductsPanel});
