/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  clamp,
  throttleOncePerAnimationFrame,
  ajaxHub,
  ajaxStatistics,
  Control,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';

linkStyleSheet('css/progress_bar.css');

export class ProgressBar extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this.node = document.createElement('div');
    this._ajaxProgress = 1;
    this._mapViewProgress = 1;
    this._progressNode = null;

    this.node.className = 'ProgressBar';
    this.render = throttleOncePerAnimationFrame(this.render);

    ajaxHub.on('success', this._updateAjax, this);
    ajaxHub.on('fail', this._updateAjax, this);
    ajaxHub.on('cancel', this._updateAjax, this);
    ajaxHub.on('pre-send', this._updateAjax, this);
  }

  get progress() {
    return 0.5 * (
      clamp(this._ajaxProgress, 0, 1) +
      clamp(this._mapViewProgress, 0, 1)
    );
  }

  onActivate() {
    this.syntool.mapView.hub.on(
      'change:update-datasets-progress',
      this._updateMapView,
      this
    );
    this.syntool.uiOverlayNode.insertBefore(
      this.node,
      this.syntool.uiOverlayNode.firstElementChild
    );
  }
  onDeactivate() {
    this.syntool.uiOverlayNode.removeChild(this.node);
    this.syntool.mapView.hub.off(
      'change:update-datasets-progress',
      this._updateMapView,
      this
    );
  }

  render() {
    var node = this.node;
    var progressNode = this._progressNode;
    var progress = this.progress;

    if (!progressNode) {
      progressNode = document.createElement('div');
      progressNode.className = 'progress';
      node.appendChild(progressNode);
      this._progressNode = progressNode;
    }

    if (node.classList.contains('is-loading')) {
      progressNode.style.width = (progress * 100) + '%';
      if (progress === 1) {
        node.classList.remove('is-loading');
      }
    } else if (progressNode.style.width !== '0') {
      progressNode.style.width = '0';
      if (progress < 1) {
        window.requestAnimationFrame(() => {
          if (this.progress < 1) {
            node.classList.add('is-loading');
            progressNode.style.width = (this.progress * 100) + '%';
          }
        });
      }
    } else if (progress < 1) {
      node.classList.add('is-loading');
      progressNode.style.width = (progress * 100) + '%';
    }
  }

  /*Protected*/
  _updateAjax() {
    if (ajaxStatistics.liveRequestCount <= 0) {
      this._ajaxProgress = 1;
      if (ajaxStatistics.liveRequestCount < 0) {
        console.log('liveRequestCount < 0', ajaxStatistics.liveRequestCount);
      }
    } else {
      this._ajaxProgress = 0.15 + 0.8 / (ajaxStatistics.liveRequestCount + 1);
    }
    this.render();
  }
  _updateMapView() {
    this._mapViewProgress = this.syntool.mapView.updateDatasetsProgress;
    this.render();
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProgressBar});
