/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {throttleOncePerAnimationFrame, Quantity, Control} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {magnitude} from '../../utils/math.js';

linkStyleSheet('css/scale_line_control.css');

function getBarLen(maxLen) {
  var pow10 = magnitude(maxLen);
  var firstChar = maxLen / pow10 | 0;

  var barLen;
  if (firstChar > 5) {
    barLen = 5;
  } else if (firstChar > 2) {
    barLen = 2;
  } else {
    barLen = 1;
  }

  return barLen * pow10;
}

export class ScaleLineControl extends Control {
  constructor(options) {
    super(options);

    this.maxWidth = 100;
    this.topOutUnits = 'km';
    this.topInUnits = 'm';
    this.bottomOutUnits = 'NM';
    this.bottomInUnits = 'ft';

    this.node = document.createElement('div');
    this.node.className = 'ScaleLineControl';

    this._top = document.createElement('div');
    this._top.className = 'ScaleLineControlTop';
    this.node.appendChild(this._top);

    this._bottom = document.createElement('div');
    this._bottom.className = 'ScaleLineControlBottom';
    this.node.appendChild(this._bottom);

    this.render = throttleOncePerAnimationFrame(this.render);
  }

  onActivate() {
    this.syntool.uiOverlayNode.appendChild(this.node);
    this.syntool.mapView.hub.on('changing:extent', this.render, this);
    this.render();
  }
  onDeactivate() {
    this.syntool.mapView.hub.off('changing:extent', this.render, this);
    this.syntool.uiOverlayNode.removeChild(this.node);
  }

  render() {
    const {mapView} = this.syntool;

    var res = mapView.getResolution();
    if (!res) {
      return;
    }

    // convert maxWidth to map units
    var resInches = mapView.convertToUnit(res, 'in');
    if (mapView.hasMercatorLikeProjection()) {
      var pixelSize = mapView.getGeodesicPixelSize();
      var resInMeters = Math.sqrt(pixelSize[0] * pixelSize[1]);
      resInches = Quantity.convert(resInMeters, 'm', 'in');
    }
    var maxSizeInches = resInches * this.maxWidth;

    // decide whether to use large or small scale units
    var topUnits;
    var bottomUnits;
    if (maxSizeInches > 100000) {
      topUnits = this.topOutUnits;
      bottomUnits = this.bottomOutUnits;
    } else {
      topUnits = this.topInUnits;
      bottomUnits = this.bottomInUnits;
    }

    // and to map units units
    var topMax = Quantity.convert(maxSizeInches, 'in', topUnits);
    var bottomMax = Quantity.convert(maxSizeInches, 'in', bottomUnits);

    // now trim this down to useful block length
    var topRounded = getBarLen(topMax);
    var bottomRounded = getBarLen(bottomMax);

    // and back to display units
    topMax = Quantity.convert(topRounded, topUnits, 'in');
    bottomMax = Quantity.convert(bottomRounded, bottomUnits, 'in');

    // and to pixel units
    var topPx = topMax / resInches;
    var bottomPx = bottomMax / resInches;

    // now set the pixel widths
    // and the values inside them

    if (this.topOutUnits === '' || this.topInUnits === '') {
      this._top.style.visibility = 'hidden';
    } else {
      this._top.style.visibility = 'visible';
      this._top.style.width = Math.round(topPx) + 'px';
      this._top.innerHTML = topRounded + ' ' + topUnits;
    }

    if (this.bottomOutUnits === '' || this.bottomInUnits === '') {
      this._bottom.style.visibility = 'hidden';
    } else {
      this._bottom.style.visibility = 'visible';
      this._bottom.style.width = Math.round(bottomPx) + 'px';
      this._bottom.innerHTML = bottomRounded + ' ' + bottomUnits;
    }

    if (resInches > 100000) {
      topUnits = this.topOutUnits;
      bottomUnits = this.bottomOutUnits;
    } else {
      topUnits = this.topInUnits;
      bottomUnits = this.bottomInUnits;
    }

    var resTop = Quantity.convert(resInches, 'in', topUnits);
    var resButtom = Quantity.convert(resInches, 'in', bottomUnits);
    this.node.title = [
      resTop.toFixed(2) + ' ' + topUnits + '/px',
      resButtom.toFixed(2) + ' ' + bottomUnits + '/px',
    ].join('\n');
  }
}
window.ODL = Object.assign(window.ODL || {}, {ScaleLineControl});
