/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {clamp, Control} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {GaugeComponent} from './gauge_component.js';

linkStyleSheet('css/selected_datasets_opacity_control.css');

export class SelectedDatasetsOpacityControl extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this.node = document.createElement('div');
    this.node.id = 'selected-datasets-opacity';

    this.gauge = new GaugeComponent({
      class   : 'SelectedDatasetsOpacityGauge',
      tabindex: '0',
      title   : 'Selected dataset(s) opacity: ${value}%',
      value   : 0,
      onChange: this.handleChange.bind(this),
      onClick : this.handleClick.bind(this),
    });

    this.node.hidden = true;
    this.node.appendChild(this.gauge.node);
  }

  onActivate() {
    this.syntool.uiOverlayNode.appendChild(this.node);
    this.syntool.mapView.hub
      .on('change:selection', this.updateDatasetsOpacity, this);
    this.syntool.productsStore.hub
      .on('change:opacity', this.updateDatasetsOpacity, this);
  }
  onDeactivate() {
    this.syntool.uiOverlayNode.removeChild(this.node);
    this.syntool.mapView.hub
      .off('change:selection', this.updateDatasetsOpacity, this);
    this.syntool.productsStore.hub
      .off('change:opacity', this.updateDatasetsOpacity, this);
  }

  updateDatasetsOpacity() {
    this._setOpacity(this.getSelectedDatasetsOpacity());
  }
  getSelectedDatasetsOpacity(selectedDatasets) {
    if (!selectedDatasets) {
      selectedDatasets = this.syntool.mapView.getSelectedDatasets();
    }

    var maxOpacity = NaN;
    for (var i = 0, l = selectedDatasets.length; i < l; i++) {
      var dataset = selectedDatasets[i];
      if (dataset.setOpacity) {
        var opacity = dataset.getOpacity();
        if (isNaN(maxOpacity) || opacity > maxOpacity) {
          maxOpacity = opacity;
        }
      }
    }
    return maxOpacity;
  }
  setSelectedDatasetsOpacity(opacity) {
    const selectedDatasets = this.syntool.mapView.getSelectedDatasets();
    this.syntool.mapView.setDatasetsOpacity(selectedDatasets, opacity);
    this._setOpacity(this.getSelectedDatasetsOpacity(selectedDatasets));
  }
  incrementSelectedDatasetsOpacity(delta) {
    this.setSelectedDatasetsOpacity(
      this.getSelectedDatasetsOpacity() + delta
    );
  }
  toggleSelectedDatasetsOpacity() {
    this.setSelectedDatasetsOpacity(
      this.getSelectedDatasetsOpacity() > 0 ? 0 : 1
    );
  }

  /*Event handlers*/
  handleChange(value) {
    this.setSelectedDatasetsOpacity(clamp(value / 100, 0, 1));
  }
  handleClick() {
    this.gauge.setProps({withSlider: !this.gauge.props.withSlider});
  }

  /*Protected*/
  _setOpacity(opacity) {
    if (isNaN(opacity)) {
      this.node.hidden = true;
    } else {
      this.gauge.setProps({value: Math.round(opacity * 100)});
      this.node.hidden = false;
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {SelectedDatasetsOpacityControl});
