/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {scheduleAnimationFrame, Control, UserStorage} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {SpectrumBuilder} from '../../utils/spectrum_builder.js';
import {DatasetMetadataView} from '../selection_panel/dataset_metadata_view.js';
import {Dialog} from './dialog.js';

import '../../features/syntool_clipboard.js';

linkStyleSheet('css/selection_metadata_ui.css');

const STORAGE_KEY = 'SelectionMetadataUI.collapsedBlocks';

export class SelectionMetadataUI extends Control {
  constructor(options) {
    super(options);

    this._selectionInfo = options.selectionInfo;

    this.spectrumBuilder = new SpectrumBuilder();
    this.spectrumBuilder.initPalette();

    this.view = new DatasetMetadataView();
    this.view.hub
      .on('change:collapsedBlocks', this.handleCollapsedBlocksChange, this)
      .on('click:resource', this.handleResourceClick, this)
      .on('click:copy', this.handleCopyClick, this)
      .on('click:filter', this.handleFilterClick, this);

    (async () => {
      try {
        this.fromJSON({collapsedBlocks: await UserStorage.get(STORAGE_KEY)});
      } catch (error) {
        /*ignore*/
      }
    })();

    this.dialog = new Dialog('dataset-dialog');
    this._dialogContents = null;
  }

  toJSON() {
    return {
      collapsedBlocks: Object.fromEntries(this.view.collapsedBlocks.entries()),
    };
  }
  fromJSON({collapsedBlocks} = {}) {
    if (collapsedBlocks) {
      this.view.collapsedBlocks = new Map(
        Array.isArray(collapsedBlocks)
        ? collapsedBlocks.map((id) => [id, true])
        : Object.entries(collapsedBlocks)
      );
    }
  }

  onAdd() {
    this.syntool.addControl(this.dialog);
  }
  onRemove() {
    this.syntool.removeControl(this.dialog);
    if (this.syntool.backupStoresMap) {
      delete this.syntool.backupStoresMap.collapsedSelectionMetadataBlocks;
    }
  }
  onActivate() {
    this.view.syntool = this.syntool;
    this._selectionInfo.view.appendChild(this.view);
    this._selectionInfo.view.hub
      .on('will-change:datasets', this.handleWillChangeDatasets, this);
    this.dialog.active = true;
  }
  onDeactivate() {
    this.dialog.active = false;
    this._selectionInfo.view.hub
      .off('will-change:datasets', this.handleWillChangeDatasets, this);
    this._selectionInfo.view.removeChild(this.view);
    this.view.syntool = null;
  }

  /**
   * Shows a popup containing a spectrum generated from the data located at `url`.
   *
   * @param  {string} title - The title of the popup shown.
   * @param  {string} url - The URL of the spectrum data.
   * @return {void}
   */
  showSpectrum(title, url) {
    const dataset = this.view.dataset;
    const datasetId = dataset.datasetId;
    if (this._dialogContents === url) {
      this.dialog.show();
    } else {
      this._dialogContents = url;
      this.dialog.show({
        title  : title,
        content: (node) => {
          this.spectrumBuilder.clear();
          node.appendChild(this.spectrumBuilder.element);
          this.spectrumBuilder.loadJson(url, datasetId);
        },
      });
    }
  }

  /**
   * Shows a popup containing an image, located at `url`.
   *
   * @param  {string} title - The title of the popup shown.
   * @param  {string} label - What does this image represent for the dataset.
   * @param  {string} url - The URL of the image.
   * @return {void}
   */
  showImage(title, label, url) {
    const selectionInfo = this._selectionInfo;
    const dataset = this.view.dataset;
    const datasetId = dataset.datasetId;
    if (this._dialogContents === url) {
      this.dialog.show();
    } else {
      this._dialogContents = url;
      this.dialog.show({
        title  : title,
        content(node) {
          const imageNode = new Image();
          imageNode.className = 'Dialog-Image -loading';
          imageNode.onload = function(event) {
            imageNode.classList.remove('-loading');
            imageNode.classList.add('-loaded');
          };
          imageNode.onerror = function(event) {
            console.error(
              'Error loading ' + label + ' image for dataset with id "' +
              datasetId + '" from URL "' + url + '"!',
              event
            );
            selectionInfo.syntool?.notificationCenter.showError(
              'Error: ' + label + ' image not found!'
            );
            imageNode.classList.remove('-loading');
            imageNode.classList.add('-errored');
          };
          imageNode.src = url;
          imageNode.style.display = 'none';
          node.appendChild(imageNode);
          scheduleAnimationFrame(() => {
            imageNode.style.display = '';
          });
        },
      });
    }
  }

  handleWillChangeDatasets(datasets, nextDatasets) {
    this.dialog.hide();
  }

  async handleCollapsedBlocksChange() {
    try {
      await UserStorage.set(STORAGE_KEY, this.toJSON().collapsedBlocks);
    } catch (error) {
      /* ignore errors */
    }
  }
  handleResourceClick(configuredResource) {
    if (!configuredResource) {
      return;
    }

    const dataset = this.view.dataset;

    const {resource, config} = configuredResource;
    const url = resource.url;
    const label = config.label;

    switch (config.type) {
    case 'spectrum':
      this.showSpectrum(label + ': ' + dataset.datasetId, url);
      break;
    case 'image':
      this.showImage(label + ': ' + dataset.datasetId, label, url);
      break;
    default:
    }
  }
  handleCopyClick(label, value) {
    this._selectionInfo.syntool.copyToClipboard(value, {label});
  }
  handleFilterClick(filter, value) {
    if (filter === 'datasetId') {
      const {productId} = this.view.dataset;
      const {productsStore} = this._selectionInfo.syntool;

      if (productsStore.getProductPattern(productId) === value) {
        productsStore.setProductPattern(productId, '');
      } else {
        productsStore.setProductPattern(productId, value);
      }
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {SelectionMetadataUI});
