/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {Control} from '../../syntool.js';
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {ShapesInfoView} from '../selection_panel/shapes_info_view.js';
import {ShapesExportSheet} from '../selection_panel/shapes_export_sheet.js';

const MIN_ZOOM = 0;
const MAX_ZOOM = 13;

function getDatasetName(dataset) {
  return dataset.name;
}
function getDatasetShape(dataset) {
  return dataset.shape;
}

export class SelectionShapeUI extends Control {
  constructor(options) {
    super(options);

    this._focusOnText = false;
    this._selectionInfo = options.selectionInfo;

    this.view = new ShapesInfoView();
    this.view.hub.on('change:author', this.handleFieldChange.bind(this, 'author'));
    this.view.hub.on('change:text', this.handleFieldChange.bind(this, 'text'));
    this.view.hub.on('change:start', this.handleFieldChange.bind(this, 'start'));
    this.view.hub.on('change:end', this.handleFieldChange.bind(this, 'end'));
    this.view.hub.on('change:arrow', this.handleArrowChange, this);
    this.view.hub.on('change:type', this.handleArrowFieldChange.bind(this, 'directionFlipped'));
    this.view.hub.on('change:flipped', this.handleArrowFieldChange.bind(this, 'sideFlipped'));
    this.view.hub.on('change:ratio', this.handleArrowFieldChange.bind(this, 'ratio'));
    this.view.hub.on('change:density', this.handleArrowFieldChange.bind(this, 'density'));
    this.view.hub.on('change:offset', this.handleArrowFieldChange.bind(this, 'offset'));
    this.view.hub.on('change:size', this.handleArrowSizeChange, this);
    this.view.hub.on('duplicate', this.handleDuplicate, this);
    this.view.hub.on('delete', this.handleDelete, this);
    this.view.setSizeConstraints(
      1,
      MAX_ZOOM - MIN_ZOOM + 1,
      (size) => this.arrowSizeToNormalizedLog(size)
    );
  }

  onActivate() {
    const selectionInfoView = this._selectionInfo.view;
    selectionInfoView.appendChild(this.view);
    selectionInfoView.hub
      .on('will-change:datasets', this.handleWillChangeDatasets, this)
      .on('change:datasets', this.handleChangeDatasets, this)
      .on('render', this.handleRender, this);

    if (this._selectionInfo.syntool.exportShapes) {
      if (!this.exportSheet) {
        this.exportSheet = new ShapesExportSheet();
        this.exportSheet.hub.on('cancel', this.handleExport, this);
        this.exportSheet.hub.on('select', this.handleExport, this);
        this.view.hub.on('export', () => {
          this.exportSheet.show(this._selectionInfo.syntool, this.view.datasets);
        });
      }
      selectionInfoView.appendOverlayChild(this.exportSheet.node);
    }
  }
  onDeactivate() {
    const selectionInfoView = this._selectionInfo.view;
    selectionInfoView.hub
      .off('will-change:datasets', this.handleWillChangeDatasets, this)
      .off('change:datasets', this.handleChangeDatasets, this)
      .off('render', this.handleRender, this);

    selectionInfoView.removeChild(this.view);
    if (this.exportSheet) {
      selectionInfoView.removeOverlayChild(this.exportSheet.node);
    }
  }

  hasUpdatableShape(dataset) {
    return (
      // Check it a shape dataset and its shape is not read-only
      this._selectionInfo.syntool.filters.dataset.isEditableShape(dataset) &&

      // Check its shape has been removed
      this._selectionInfo.syntool.shapesStore.get(dataset.shape.id)
    );
  }
  arrowSizeToNormalizedLog(size) {
    const mapView = this._selectionInfo.syntool.mapView;
    const minSize = Shape.getArrowSizeForResolution(
      mapView.convertToUnit(mapView.getResolutionForZoom(MAX_ZOOM), 'm')
    );
    const normalizedLogValue = Math.log2(size / minSize) / (MAX_ZOOM - MIN_ZOOM);

    return normalizedLogValue;
  }
  arrowSizeFromNormalizedLog(normalizedLogValue) {
    const mapView = this._selectionInfo.syntool.mapView;
    const minSize = Shape.getArrowSizeForResolution(
      mapView.convertToUnit(mapView.getResolutionForZoom(MAX_ZOOM), 'm')
    );
    const size = Math.pow(2, normalizedLogValue * (MAX_ZOOM - MIN_ZOOM)) * minSize;

    return size;
  }

  saveShapeInfoViewChanges() {
    if (this.view.datasets) {
      const modifiableDatasets = (
        this.view.datasets.filter(this.hasUpdatableShape, this)
      );
      const changes = this.view.getChanges();

      if (modifiableDatasets.length > 0 && Object.keys(changes).length > 0) {
        this._selectionInfo.syntool.shapesStore.update(
          modifiableDatasets.map(getDatasetName),
          changes
        );
      }
    }
  }

  handleWillChangeDatasets() {
    this.saveShapeInfoViewChanges();
    this.exportSheet.hide();
  }
  handleChangeDatasets() {
    this._focusOnText = (
      this.view.datasets &&
      this.view.datasets.length === 1 &&
      this.view.datasets[0].shape.type === Shape.TEXT
    );
  }
  handleRender() {
    if (this._focusOnText) {
      this.view.focusOn('text');
      this._focusOnText = false;
    }
  }
  handleFieldChange(name, value) {
    if (!this.view.datasets) {
      return;
    }

    const modifiableDatasets = this.view.datasets.filter(
      this._selectionInfo.syntool.filters.dataset.isEditableShape
    );

    if (modifiableDatasets.length > 0) {
      this._selectionInfo.syntool.shapesStore.update(
        modifiableDatasets.map(getDatasetName),
        {[name]: value}
      );
    }
  }
  handleArrowChange(on) {
    if (
      this.view.datasets &&
      this.view.datasets.length === 1 &&
      !this.view.datasets[0].readOnly
    ) {
      const updates = {arrow: null};
      const {syntool} = this._selectionInfo;

      if (on) {
        updates.arrow = Shape.getInitialShapeArrowAtResolution(
          this.view.datasets[0].shape,
          syntool.mapView.convertToUnit(syntool.mapView.getResolution(), 'm')
        );
      }
      syntool.shapesStore.update(
        this.view.datasets[0].name,
        updates
      );
    }
  }
  handleArrowFieldChange(name, value) {
    if (
      this.view.datasets &&
      this.view.datasets.length === 1 &&
      !this.view.datasets[0].readOnly
    ) {
      this._selectionInfo.syntool.shapesStore.update(
        this.view.datasets[0].name,
        {arrow: {[name]: value}}
      );
    }
  }
  handleArrowSizeChange(normalizedLogValue) {
    if (
      this.view.datasets &&
      this.view.datasets.length === 1 &&
      !this.view.datasets[0].readOnly
    ) {
      const size = this.arrowSizeFromNormalizedLog(normalizedLogValue);

      this._selectionInfo.syntool.shapesStore.update(
        this.view.datasets[0].name,
        (
          this.view.datasets[0].shape.type === Shape.ARROW
          ? {arrowSize: size}
          : {arrow: {size}}
        )
      );
    }
  }
  handleDuplicate(event) {
    this._selectionInfo.syntool.duplicateSelectedShapes();
  }
  handleExport(destination, format) {
    if (this.view.datasets && destination && format) {
      const shapes = this.view.datasets.map(getDatasetShape);
      const toClipboard = (destination === 'clipboard');
      this._selectionInfo.syntool.exportShapes(shapes, {format, toClipboard});
    }
    this.exportSheet.hide();
  }
  handleDelete() {
    this._selectionInfo.syntool.removeSelectedShapes();
  }
}
window.ODL = Object.assign(window.ODL || {}, {SelectionShapeUI});
