/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DOMEvents,
  H,
  textFromHTML,
  renderModifiedKeys,
  Control,
} from '../../syntool.js';
import {setNodeChildren, linkStyleSheet} from '../../utils/dom_utils.js';
import {preventFocusOnClick} from '../utils.js';

linkStyleSheet('css/shapes_toolbar_control.css');

export class ShapesToolbarControl extends Control {
  /*Public*/
  constructor(options) {
    super(Object.assign({autoActivate: false}, options));

    this._automaticActivation = !!(options?.autoActivate ?? true);
    this.node = H('div', {class: 'ShapesToolbarControl'}, [
      this._handlersGroup = H('div', {class: 'ShapesToolbarButtonGroup --handlers'}),
      H('div', {class: 'ShapesToolbarButtonGroup --undo-redo'}, [
        this._undoButton = H('button', {
          class: 'ShapesToolbarButton --undo',
          title: 'Undo',
        }),
        this._redoButton = H('button', {
          class: 'ShapesToolbarButton --redo',
          title: 'Redo',
        }),
      ]),
      this._actionsGroup = H('div', {class: 'ShapesToolbarButtonGroup --actions'}),
    ]);
    this._actionButtons = [
      H('button', {
        class: 'ShapesToolbarButton --done',
        title: 'Commit drawing',
        'data-shortcut-id': 'user-shapes-finish-drawing',
      }, 'Done'),
      this._duplicateButton = H('button', {
        class: 'ShapesToolbarButton --duplicate',
        title: 'Duplicate selected shape(s)',
      }, H('i', {class: 'fa fa-copy'})),
      this._deleteButton = H('button', {
        class: 'ShapesToolbarButton --delete',
        title: 'Remove selected shape(s)',
      }, H('i', {class: 'fa fa-trash'})),
    ];
    this._handlerNodes = new Map();

    preventFocusOnClick(this.node, '[data-shortcut-id]');
    DOMEvents(this.node)
      .on('click', '[data-shortcut-id]', (event) => {
        event.preventDefault();
        event.stopPropagation();
        this.syntool.executeShortcutWithId(event.target.dataset.shortcutId);
      });
  }

  /*Protected - Control overrides*/
  onAdd() {
    if (this._automaticActivation) {
      this.syntool.hub.on(
        'change:shapesDrawingActive',
        this.handleShapesDrawingActiveChange,
        this
      );
      this.handleShapesDrawingActiveChange();
    }
  }
  onRemove() {
    if (this._automaticActivation) {
      this.syntool.hub.off(
        'change:shapesDrawingActive',
        this.handleShapesDrawingActiveChange,
        this
      );
    }
  }
  onActivate() {
    this.handleShapesDrawingHandlersChange();
    this.syntool.usableViewportNode.appendChild(this.node);
    this.syntool.hub
      .on(
        'change:shapesDrawingActiveHandler',
        this.handleShapesDrawingActiveHandlerChange,
        this
      )
      .on(
        'change:shapesDrawingHandlers',
        this.handleShapesDrawingHandlersChange,
        this
      );
    this.syntool.shortcutsStore.hub
      .on('change', this.handleShortcutsChange, this);
  }
  onDeactivate() {
    this.syntool.usableViewportNode.removeChild(this.node);
    this.syntool.hub
      .off(
        'change:shapesDrawingActiveHandler',
        this.handleShapesDrawingActiveHandlerChange,
        this
      )
      .off(
        'change:shapesDrawingHandlers',
        this.handleShapesDrawingHandlersChange,
        this
      );
    this.syntool.shortcutsStore.hub
      .off('change', this.handleShortcutsChange, this);
  }

  /*Protected*/
  handleShapesDrawingActiveChange() {
    this.active = (
      this.syntool.shapesDrawingActive &&
      this.syntool.userShapesProduct
    );
  }
  handleShapesDrawingHandlersChange() {
    const activeHandler = this.syntool.shapesDrawingActiveHandler;
    const newHandlerNodes = new Map();
    for (const handler of this.syntool.shapesDrawingHandlers) {
      const active = (activeHandler === handler);
      let node = this._handlerNodes.get(handler.id);
      if (!node) {
        node = H('button', {
          class: 'ShapesToolbarButton ' + (handler.cssClass ?? ''),
          title: handler.name,
          'data-id': handler.id,
          'data-shortcut-id': (
            this.syntool.getShortcutIdForHandlerWithId(handler.id)
          ),
        });
      }
      node.classList.toggle('-active', active);
      node.classList.toggle('-inactive', !active);
      newHandlerNodes.set(handler.id, node);
    }
    setNodeChildren(this._handlersGroup, newHandlerNodes.values());
    this._handlerNodes = newHandlerNodes;
    this.handleShortcutsChange();
  }
  handleShapesDrawingActiveHandlerChange() {
    const activeHandlerId = this.syntool.shapesDrawingActiveHandlerId;
    for (const node of this._handlerNodes.values()) {
      if (node.dataset.id) {
        const active = (activeHandlerId === node.dataset.id);
        node.classList.toggle('-active', active);
        node.classList.toggle('-inactive', !active);
      }
    }
  }
  handleShortcutsChange() {
    this._duplicateButton.dataset.shortcutId = (
      this.syntool.shapesDrawingActiveHandlerId
      ? ''
      : 'user-shapes-duplicate-selected-shapes'
    );
    this._deleteButton.dataset.shortcutId = (
      this.syntool.shapesDrawingActiveHandlerId
      ? ''
      : 'user-shapes-remove-selected-shapes'
    );
    this._undoButton.dataset.shortcutId = (
      this.syntool.shapesDrawingActiveHandlerId
      ? 'user-shapes-undo-drawing'
      : 'user-shapes-undo'
    );
    this._redoButton.dataset.shortcutId = (
      this.syntool.shapesDrawingActiveHandlerId
      ? 'user-shapes-redo-drawing'
      : 'user-shapes-redo'
    );

    const buttons = this._actionButtons.filter((button) => {
      const shortcut = (
        this.syntool.shortcutsStore.shortcutWithId(button.dataset.shortcutId)
      );
      return shortcut && !shortcut.disabled;
    });
    setNodeChildren(this._actionsGroup, buttons);

    for (const button of this.node.querySelectorAll('[data-shortcut-id')) {
      const shortcut = (
        this.syntool.shortcutsStore.shortcutWithId(button.dataset.shortcutId)
      );
      button.dataset.title ??= button.title;

      const shortcutDisabled = !shortcut || shortcut.disabled;
      button.disabled = shortcutDisabled;

      if (shortcut.keys) {
        const key = textFromHTML(
          renderModifiedKeys(shortcut.modifiers, shortcut.keys).join(' ')
        );
        button.title = `${button.dataset.title} (${key})`;
      } else {
        button.title = button.dataset.title;
      }
    }
  }
}
