/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  h,
  renderModifiedKeys,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {Dialog} from './dialog.js';

linkStyleSheet('css/shortcuts_dialog.css');

function render(groups) {
  return h('table', Array.from(groups, renderGroup));
}
function renderGroup(group) {
  if (group.label === false) {
    return null;
  }
  const renderedShortcuts = group.shortcuts.map(renderShortcut).filter(Boolean);
  if (renderedShortcuts.length === 0) {
    return null;
  }
  return h([
    h('tr', [h('th'), h('th', group.label)]),
    ...renderedShortcuts,
  ]);
}
function renderShortcut(shortcut) {
  return shortcut.keys && h('tr', [
    h('td', renderModifiedKeys(shortcut.modifiers, shortcut.keys).join(' ')),
    h('td', shortcut.description),
  ]);
}

export class ShortcutsDialog extends Dialog {
  constructor(options) {
    super('shortcuts-dialog', {title: 'Shortcuts'}, options);
    this.store = null;
    this._needsRerender = true;
    this._requestUpdate = throttleOncePerAnimationFrame(this._update.bind(this));
  }

  /*Public - Dialog overrides*/
  show() {
    if (this._needsRerender) {
      this._update();
    }
    super.show();
  }

  /*Protected - Control overrides*/
  onAdd() {
    super.onAdd();
    this.store = this.syntool.shortcutsStore;
    this.store.hub.on('change', this.handleStoreChange, this);
    this.handleStoreChange();
  }
  onRemove() {
    super.onRemove();
    this.store.hub.off('change', this.handleStoreChange, this);
    this.store = null;
  }

  /*Protected*/
  handleStoreChange() {
    if (this.shown) {
      this._requestUpdate();
    } else {
      this._needsRerender = true;
    }
  }
  _update() {
    this.setProps(
      this.active && this.store
      ? {content: render(this.store.groups.values())}
      : null
    );
    this._needsRerender = false;
  }
}
window.ODL = Object.assign(window.ODL || {}, {ShortcutsDialog});
