/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import '../../map_view_extensions/map_view_dataset_selection.js';
import {
  PASSIVE,
  formatUTCDate,
  debounce,
  h,
  H,
  DOMEvents,
  normalizeTimeframe,
  Control,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {preventFocusOnClick} from '../utils.js';
import {CoverageTimelineView} from '../timeline_panel/coverage_timeline_view.js';
import {PaginationHelpDialog} from '../timeline_panel/pagination_help.js';
import {DetailsTimelineView} from '../timeline_panel/details_timeline_view.js';
import {CoordinatesView} from './coordinates_view.js';

import '../../features/syntool_clipboard.js';

linkStyleSheet('css/timeline_panel.css');

const NO_DATASETS_HTML = h('span', {class: 'info-icon'}) + ' No datasets';

export class TimelinePanel extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this.node = H('div', {
      class: 'UISidebar TimelinePanel -closed h-hidden-in-screenshot-mode',
    }, [
      H('div', {class: 'lid vertical-lid'}),
      H('div', {class: 'TimelineBar js-bar'}, [
        H('span', {class: 'TimelineBar-Status js-status'}),
        H('button', {
          class: 'UIButton TimelineBar-Button --FindNearest js-find-nearest',
          type : 'button',
          title: 'Jump to nearest date with something to see',
        }, H('i', {class: '-icon svg svg-search-time'})),
        H('button', {
          class: 'UIButton TimelineBar-Button --Toggle --Coloc js-coloc',
          type : 'button',
          title: (
            'Toggle temporal data colocation highlight\n\n' +
            'White: day/month/year where some but not all selected products ' +
            'are available in the map visible area\n' +
            'Red: day/month/year where all selected products are available ' +
            'on the same day in the map visible area\n' +
            'Yellow: month/year where all selected products are available in ' +
            'the map visible area, but not on the same day'
          ),
        }, H('i', {class: '-icon t t-C'})),
        H('span', {class: 'TimelineBar-DatasetsCount js-datasets-count'}),
        H('button', {
          class: 'UIButton TimelineBar-Button --PrevPage js-prev-page',
          type : 'button',
          title: 'Show previous datasets (100 by default)',
        }, H('i', {class: '-icon svg svg-arrow-left'})),
        H('span', {
          class: 'TimelineBar-DatasetsDisplayedCount js-datasets-displayed-count',
        }),
        H('button', {
          class: 'UIButton TimelineBar-Button --NextPage js-next-page',
          type : 'button',
          title: 'Show next datasets (100 by default)',
        }, H('i', {class: '-icon svg svg-arrow-right'})),
        H('span', {class: 'TimelineBar-SelectedDataset js-selected-dataset'}),
        H('span', {class: 'TimelineBar-CurrentTime js-current-time'}),
      ]),
      H('div', {class: 'TimelineCoverage js-coverage'}),
      H('div', {class: 'TimelineDetailsContainer'}, [
        H('div', {class: 'TimelineDetails js-details'}),
        H('div', {class: 'TimelineDetailsCursor'}),
      ]),
    ]);
    this.store = null;

    DOMEvents(this.node)
    .on('transitionend', (event) => {
      if (event.target === this.node) {
        this.syntool?.updateUsableViewportSize();
      }
    })
    .on('click', '.lid', () => (this.withDetails = !this.withDetails));
  }

  get state() {
    return Object.assign(super.state, {
      animationControlsHidden: this.animationControlsHidden,
      coordinatesFormat: this.coordinatesFormat,
      withDetails: this.withDetails,
    });
  }
  set state(state) {
    const {
      animationControlsHidden = true,
      coordinatesFormat = CoordinatesView.DECIMAL,
      withDetails = false,
    } = state;
    super.state = state;
    this.animationControlsHidden = animationControlsHidden;
    this.coordinatesFormat = coordinatesFormat;
    this.withDetails = withDetails;
  }

  get animationControlsHidden() {
    return this._animationControlsHidden;
  }
  set animationControlsHidden(hidden) {
    if (this._animationControlsHidden === !hidden) {
      for (const node of this.node.querySelectorAll('.js-animation')) {
        node.style.display = hidden ? 'none' : '';
      }
      this._animationControlsHidden = !this._animationControlsHidden;
      this.hub.emit('change:state');
    }
  }
  get coordinatesFormat() {
    return this.coordinatesView.format;
  }
  set coordinatesFormat(format) {
    this.coordinatesView.format = format;
  }
  get withDetails() {
    return this.detailsView.open;
  }
  set withDetails(withDetails) {
    if (this.withDetails === !withDetails) {
      this.node.classList.toggle('-closed', !withDetails);
      this.detailsView.open = withDetails;
    }
  }

  /*Protected - Control overrides*/
  onActivate() {
    const {syntool} = this;

    /*
    this.node should be added to the DOM before creating the
    DetailsTimelineView
    */
    syntool.uiOverlayNode.insertBefore(
      this.node,
      syntool.usableViewportNode.nextSibling
    );

    if (!this.coordinatesView) {
      this.initBar();
    }
    if (!this.coverageView) {
      this.initCoverageView();
    }
    if (this.detailsView) {
      this.detailsView.updateWidth();
    } else {
      this.initDetailsView();
    }
    if (!this.node.querySelector('.js-toggle-animation')) {
      this.initAnimationControls();
    }

    this.store = syntool.timelineStore;
    this.coordinatesView.mapView = syntool.mapView;
    this.coordinatesView.activate();
    this.detailsView.store = this.store;
    this.detailsView.filter = syntool.filters.dataset.inTimeline;
    this.coverageView.syntool = syntool;
    this.coverageView.store = this.store;

    this.store.hub
      .on('change:timeframeLength', this.handleTimeframeLengthChange, this)
      .on('change:currentDate', this.handleCurrentDateChange, this)
      .on('change:coverage', this.handleCoverageChange, this);
    syntool.hub
      .on('change:coloc-active', this.handleColocChange, this)
      .on('change:timelineAnimating', this.handleTimelineAnimatingChange, this);
    syntool.mapView.hub
      .on('change:datasetsStatistics', this.handleDatasetsStatisticsChange, this)
      .on('change:selection', this.updateTimelineBar, this)
      .on('add:dataset', this.handleAddDataset, this)
      .on('update:dataset', this.handleUpdateDataset, this)
      .on('remove:dataset', this.handleRemoveDataset, this)
      .on('clear:datasets', this.handleClearDatasets, this)
      .on('changing:in-map-datasets', this.handleChangingInMapDatasets, this)
      .on('change:selection', this.handleChangeSelection, this);

    this.updateTimelineBar();
    this.handleChangeSelection();
    this.handleDatasetsStatisticsChange();
    this.handleTimelineAnimatingChange();
    this.handleColocChange();
    this.handleCoverageChange();
    this.handleCurrentDateChange();
    this.handleTimeframeLengthChange();

    syntool.updateUsableViewportSize();
    syntool.whenInitialised(() => {
      if (this.syntool !== syntool) {
        return;
      }

      if (!this.node.querySelector('.js-timespan')) {
        this.initTimespans(syntool.config);
      }
    });
  }
  onDeactivate() {
    this.store.hub
      .off('change:timeframeLength', this.handleTimeframeLengthChange, this)
      .off('change:currentDate', this.handleCurrentDateChange, this)
      .off('change:coverage', this.handleCoverageChange, this);
    this.syntool.hub
      .off('change:coloc-active', this.handleColocChange, this)
      .off('change:timelineAnimating', this.handleTimelineAnimatingChange, this);
    this.syntool.mapView.hub
      .off('change:datasetsStatistics', this.handleDatasetsStatisticsChange, this)
      .off('change:selection', this.updateTimelineBar, this)
      .off('add:dataset', this.handleAddDataset, this)
      .off('update:dataset', this.handleUpdateDataset, this)
      .off('remove:dataset', this.handleRemoveDataset, this)
      .off('clear:datasets', this.handleClearDatasets, this)
      .off('changing:in-map-datasets', this.handleChangingInMapDatasets, this)
      .off('change:selection', this.handleChangeSelection, this);

    this.store = null;
    this.coordinatesView.deactivate();
    this.coordinatesView.mapView = null;
    this.detailsView.store = null;
    this.detailsView.filter = null;
    this.coverageView.syntool = null;
    this.coverageView.store = null;

    this.syntool.uiOverlayNode.removeChild(this.node);
    this.syntool.updateUsableViewportSize();
  }

  /*Details view*/
  initDetailsView() {
    const detailsNode = this.node.querySelector('.js-details');

    this.detailsView = new DetailsTimelineView(detailsNode);

    this._windowResizing = false;
    const debouncedWindowResize = debounce(250, () => {
      this._windowResizing = false;
      if (this.active && this.detailsView) {
        this.detailsView.updateWidth();
      }
    });
    const handleWindowResize = () => {
      this._windowResizing = true;
      debouncedWindowResize();
    };
    window.addEventListener('resize', handleWindowResize, PASSIVE);

    this.detailsView.hub
      .on('change:currentDate', this.handleDetailsViewDateChange, this)
      .on('pick', (dataset) => this.syntool.mapView.selectClickDataset(dataset))
      .on('change:open', () => this.hub.emit('change:state'));
  }
  handleDetailsViewDateChange(date) {
    if (!this._windowResizing && this.store) {
      this.store.currentDate = date;
    }
  }

  /*Coverage view*/
  initCoverageView() {
    const coverageNode = this.node.querySelector('.js-coverage');
    this.coverageView = new CoverageTimelineView(coverageNode);
  }

  /*Timespans*/
  initTimespans(config) {
    const timespansNode = H('div',
      {class: 'UISegmentedControl TimelineBar-Timespans'},
      config.timeRanges.map(([label, timeframe]) => H([
        H('label', {class: 'UISegmentedControl-Segment TimelineBar-Timespan'}, [
          H('input', {
            type   : 'radio',
            name   : 'TimelineBar-Timespan',
            class  : 'js-timespan',
            value  : label,
            checked: false,

            'data-timeframe': normalizeTimeframe(timeframe),
          }),
          H('span', label),
        ]),
      ]))
    );
    const statusNode = this.node.querySelector('.js-status');
    const barNode = statusNode.parentNode;
    barNode.insertBefore(timespansNode, statusNode);

    preventFocusOnClick(timespansNode);
    timespansNode.addEventListener('change', (event) => {
      this.store.timeframe = event.target.dataset.timeframe;
    });
    this.handleTimeframeLengthChange();
  }
  handleTimeframeLengthChange() {
    const node = this.node.querySelector(
      `.js-timespan[data-timeframe="${this.store.timeframe}"]`
    );
    if (node) {
      node.checked = true;
    }
  }

  /*Animation*/
  initAnimationControls() {
    this._animationControlsHidden = false;
    const barNode = this.node.querySelector('.js-bar');
    barNode.insertBefore(H([
      H('button', {
        class: 'UIButton TimelineBar-Button --Toggle --Animation js-animation js-toggle-animation',
        type : 'button',
        title: 'Start timeline animation',
      }, H('i', {class: '-icon svg svg-play'})),
      H('div', {class: 'UISelect TimelineBar-SelectAnimationSpeed js-animation'},
        H('select', {class: 'js-select-animation-speed'}, [
          H('option', {value: '-8'}, '- 8x'),
          H('option', {value: '-4'}, '- 4x'),
          H('option', {value: '-2'}, '- 2x'),
          H('option', {value: '-1'}, '- 1x'),
          H('option', {value: '-0.5'}, '- 0.5x'),
          H('option', {value: '-0.25'}, '- 0.25x'),
          H('option', {value: '0.25'}, '0.25x'),
          H('option', {value: '0.5'}, '0.5x'),
          H('option', {value: '1', selected: 'selected'}, '1x'),
          H('option', {value: '2'}, '2x'),
          H('option', {value: '4'}, '4x'),
          H('option', {value: '8'}, '8x'),
        ])
      ),
    ]), barNode.firstChild);

    DOMEvents(this.node)
    .on('click', '.js-toggle-animation', () => {
      this.syntool.timelineAnimating = !this.syntool.timelineAnimating;
    })
    .on('change', '.js-select-animation-speed', (event) => {
      this.syntool.timelineAnimationTimeFactor = parseFloat(event.target.value);
    });
  }
  handleTimelineAnimatingChange() {
    const button = this.node.querySelector('.js-toggle-animation');
    if (button) {
      button.title = (
        this.syntool.timelineAnimating
        ? 'Stop timeline animation'
        : 'Start timeline animation'
      );

      const buttonIcon = button.querySelector('.-icon');
      if (buttonIcon) {
        buttonIcon.classList.toggle('svg-play', !this.syntool.timelineAnimating);
        buttonIcon.classList.toggle('svg-pause', this.syntool.timelineAnimating);
      }
    }
  }

  /*Bar*/
  initBar() {
    const barNode = this.node.querySelector('.js-bar');
    this.coordinatesView = new CoordinatesView();
    barNode.appendChild(this.coordinatesView.node);

    this._visibleDatasetsLabel = this.node.querySelector('.js-datasets-count');
    this._displayedDatasetsLabel = this.node.querySelector('.js-datasets-displayed-count');
    this._prevPageButton = this.node.querySelector('.js-prev-page');
    this._nextPageButton = this.node.querySelector('.js-next-page');

    this._paginationHelpDialog = new PaginationHelpDialog(this);

    DOMEvents(this.node)
      .on('click', '.js-find-nearest', () => {
        this.syntool.executeShortcutWithId('jump-to-nearest-dataset');
      })
      .on('click', '.js-coloc', () => {
        this.syntool.colocActive = !this.syntool.colocActive;
      })
      .on('click', '.js-prev-page', () => {
        this.syntool.mapView.incrementDataPageBy(-1);
      })
      .on('click', '.js-datasets-displayed-count', () => {
        this._paginationHelpDialog.toggle();
      })
      .on('click', '.js-next-page', () => {
        this.syntool.mapView.incrementDataPageBy(1);
      });

    this.coordinatesView.hub
      .on('change:format', () => this.hub.emit('change:state'));
  }
  handleCoverageChange() {
    this.node.querySelector('.js-find-nearest').disabled = (
      this.store.getCoverage().isEmpty || this.store.isCoverageLoading()
    );

    const status = this.node.querySelector('.js-status');
    status.classList.remove('-loading', '-complete', '-error', '-incomplete');
    if (this.store.isCoverageLoading()) {
      status.classList.add('-loading');
      status.title = 'Loading data availability...';
    } else if (this.store.isCoverageComplete()) {
      status.classList.add('-complete');
      status.title = 'Data availability is completely loaded';
    } else if (this.store.getCoverage().isEmpty) {
      status.classList.add('-error');
      status.title = 'Data availability is not loaded due to server or network errors';
    } else {
      status.classList.add('-incomplete');
      status.title = 'Data availability is incomplete due to server or network errors';
    }
  }
  handleColocChange() {
    const colocActive = this.syntool.colocActive;
    this.node.querySelector('.js-coloc').classList.toggle('-red', colocActive);
    this.coverageView.colocActive = colocActive;
  }
  handleCurrentDateChange() {
    const timelineTimeNode = this.node.querySelector('.js-current-time');
    timelineTimeNode.textContent = formatUTCDate(this.store.currentDate);
  }

  handleAddDataset(dataset) {
    this.detailsView.addDataset(dataset);
  }
  handleUpdateDataset(dataset) {
    this.detailsView.updateDataset(dataset);
  }
  handleRemoveDataset(dataset) {
    this.detailsView.removeDataset(dataset);
  }
  handleClearDatasets() {
    this.detailsView.clearDatasets();
  }
  handleChangingInMapDatasets() {
    this.detailsView.repaint();
  }
  handleChangeSelection() {
    this.detailsView.repaint();
  }

  handleDatasetsStatisticsChange() {
    this.updateDatasetsStatistics(this.syntool.mapView.datasetsStatistics);
  }
  updateDatasetsStatistics(datasetStatistics) {
    const {
      changing,
      inMapCount,
      // shouldBeInMapCount,
      displayedCount,
      shouldBeDisplayedCount,
      displayableCount,
    } = datasetStatistics;

    this._visibleDatasetsLabel.classList.toggle('-changing', changing);
    this._displayedDatasetsLabel.classList.toggle('-changing', changing);
    if (inMapCount === 0) {
      this._visibleDatasetsLabel.innerHTML = NO_DATASETS_HTML;
      this._visibleDatasetsLabel.title = (
        'No datasets available at this date for the selected products.'
      );
    } else if (inMapCount === 1) {
      this._visibleDatasetsLabel.textContent = '1 dataset';
      this._visibleDatasetsLabel.title = (
        '1 dataset available at this date for the selected products.'
      );
    } else {
      this._visibleDatasetsLabel.textContent = inMapCount + ' datasets';
      this._visibleDatasetsLabel.title = (
        inMapCount +
        ' datasets available at this date for the selected products.'
      );
    }

    if (displayedCount === inMapCount) {
      this._displayedDatasetsLabel.textContent = '';
      this._displayedDatasetsLabel.style.display = 'none';
    } else {
      this._displayedDatasetsLabel.textContent = displayedCount + ' displayed';
      this._displayedDatasetsLabel.style.display = '';
    }

    if (
      displayedCount < inMapCount &&
      shouldBeDisplayedCount < displayableCount
    ) {
      this._prevPageButton.style.display = '';
      this._nextPageButton.style.display = '';
      this._paginationHelpDialog.showIfNeeded();
    } else {
      this._paginationHelpDialog.hide();
      this._prevPageButton.style.display = 'none';
      this._nextPageButton.style.display = 'none';
      this._displayedDatasetsLabel.textContent = (
        '— ' + this._displayedDatasetsLabel.textContent
      );
    }
  }
  updateTimelineBar() {
    const datasets = this.syntool.mapView.getSelectedDatasets();
    const statusMessage = (
      datasets.length === 0
      ? ''
      : (
        datasets.length === 1
        ? [
          datasets[0].name,
          h('span', null, ' from '),
          datasets[0].getProduct().label,
        ].join('')
        : datasets.length + ' datasets selected'
      )
    );

    const node = this.node.querySelector('.js-selected-dataset');
    node.innerHTML = statusMessage;
    node.title = node.textContent.trim();

    if (datasets.length === 1 && datasets[0]) {
      node.onclick = () => {
        this.syntool.copyToClipboard(datasets[0].name, {label: "dataset's ID"});
      };
    } else {
      node.onclick = null;
    }
  }
}
window.ODL = Object.assign(window.ODL || {}, {TimelinePanel});
