/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  DOMEvents,
  removeAllChildrenOfNode,
  h,
  H,
  Control,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';

linkStyleSheet('css/title_bar.css');

export class TitleBar extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this.buttonGroups = [
    ];

    this.node = H('div', {class: 'TitleBar', hidden: true}, [
      this.wrapperNode = H('div', {class: 'TitleBar-Wrapper'}),
    ]);

    this.titleNode = H('div', {class: 'TitleBar-Title'}, this.getTitleHTML());
    this.menuButton = H('button', {
      class: 'UIButton TitleBar-ToggleMenuButton',
      title: 'Menu',
    }, H('i', {class: '-icon fa fa-bars'}));
    this.menuButton.addEventListener('click', () => {
      this.node.classList.toggle('-open');
      this.syntool.updateUsableViewportSize();
    });

    this.render = throttleOncePerAnimationFrame(this.render);
  }
  addGroup(groupId, buttons, index = -1) {
    for (const {id} of this.buttonGroups) {
      if (id === groupId) {
        return false;
      }
    }
    if (index < 0) {
      index += this.buttonGroups.length + 1;
    }
    this.buttonGroups.splice(index, 0, {id: groupId, buttons: buttons || []});
    this.render();
    return true;
  }
  addButton(groupId, button, index = -1) {
    for (const {id, buttons} of this.buttonGroups) {
      if (id === groupId) {
        if (index < 0) {
          index += buttons.length + 1;
        }
        buttons.splice(index, 0, button);
        this.render();
        return true;
      }
    }
    return false;
  }

  /*Protected - Control overrides*/
  onAdd() {
    this.syntool.uiOverlayNode.insertBefore(
      this.node,
      this.syntool.uiOverlayNode.firstChild
    );
  }
  onRemove() {
    this.syntool.uiOverlayNode.removeChild(this.node);
  }
  onActivate() {
    this.node.hidden = false;
    this.syntool.updateUsableViewportSize();
  }
  onDeactivate() {
    this.node.hidden = true;
    this.syntool.updateUsableViewportSize();
  }

  /*Protected*/
  getTitleHTML() {
    const inlinedTitleNode = document.getElementById('TitleBarTitle');
    if (inlinedTitleNode) {
      inlinedTitleNode.remove();
      return inlinedTitleNode.innerHTML.trim();
    }
    return h('span', document.title);
  }
  buttonItemFromEntry([label, onClick]) {
    return {label, onClick};
  }
  render() {
    const wrapperNode = this.wrapperNode;

    DOMEvents(this.node).off('click');
    removeAllChildrenOfNode(wrapperNode);

    wrapperNode.appendChild(this.titleNode);
    wrapperNode.appendChild(this.menuButton);
    wrapperNode.appendChild(H('span', {class: 'TitleBar-Spacer'}));

    for (const buttonGroup of this.buttonGroups) {
      wrapperNode.appendChild(this.renderGroup(buttonGroup));
      wrapperNode.appendChild(H('span', {class: 'TitleBar-Spacer -x2'}));
    }

    if (wrapperNode.lastElementChild.className === 'TitleBar-Spacer -x2') {
      wrapperNode.lastElementChild.remove();
    }
  }
  renderButtonItem({label, description}) {
    return H('option', {title: description}, label);
  }
  renderButton(button) {
    if (button instanceof HTMLElement) {
      button.classList.add('UIButton', 'TitleBar-GroupButton');
      return button;
    }

    const {id, style, items, onClick, onMouseOver, onMouseOut} = button;
    const className = `UIButton TitleBar-GroupButton ${button.class || ''}`;
    const title = button.description || button.label;
    const icon = button.icon && H('i', {class: `-icon ${button.icon}`});
    const label = button.label && H('span', button.label);

    let node;
    if (items) {
      const itemsArray = (
        Array.isArray(items)
        ? items
        : Object.entries(items).map(this.buttonItemFromEntry, this)
      );
      node = H('div', {id, class: className, style}, [
        icon,
        label,
        Object.assign(
          H('select', {title}, [
            H('option', {disabled: true}, button.label || button.description),
            ...itemsArray.map(this.renderButtonItem, this),
          ]),
          {
            selectedIndex: 0,
            onchange(event) {
              itemsArray[this.selectedIndex - 1].onClick();
              this.selectedIndex = 0;
            },
          }
        ),
      ]);
    } else {
      node = H('button', {id, class: className, style, title}, [icon, label]);
      if (onClick) {
        DOMEvents(this.node).on('click', node, onClick);
      }
    }

    if (onMouseOver) {
      DOMEvents(node).on('mouseenter', onMouseOver);
    }
    if (onMouseOut) {
      DOMEvents(node).on('mouseleave', onMouseOut);
    }
    return node;
  }
  renderGroup(group) {
    return H('div', {
      id: group.id ? `TitleBar-Group-${group.id}` : undefined,
      class: 'UIButtonGroup TitleBar-Group',
    }, group.buttons.map(this.renderButton, this));
  }
}
window.ODL = Object.assign(window.ODL || {}, {TitleBar});
