/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {H, DOMEvents, nodeFromHTML, Control} from '../../syntool.js';
import {Dialog} from './dialog.js';

linkStyleSheet('css/tour_control.css');

function moveDialog(dialog, to) {
  const selector = (typeof to === 'function' ? to(dialog) : to);
  const element = (
    typeof selector === 'string' ? document.querySelector(selector) : selector
  );
  if (element && dialog.node.parentElement !== element) {
    element.appendChild(dialog.node);
  }
}
function positionDialog(dialog, to, [xFrom = 'auto', yFrom = 'auto'] = []) {
  const selector = (typeof to === 'function' ? to(dialog) : to);
  const element = (
    typeof selector === 'string' ? document.querySelector(selector) : selector
  );
  if (element) {
    const {
      top: targetTop,
      right: targetRight,
      bottom: targetBottom,
      left: targetLeft,
    } = element.getBoundingClientRect();
    const {
      top: parentTop,
      right: parentRight,
      bottom: parentBottom,
      left: parentLeft,
    } = dialog.node.parentElement.getBoundingClientRect();
    const parentCenterX = (parentRight - parentLeft) / 2;
    const parentCenterY = (parentBottom - parentTop) / 2;
    const targetCenterX = (targetLeft + targetRight) / 2;
    const targetCenterY = (targetTop + targetBottom) / 2;

    if (xFrom === 'auto') {
      xFrom = (
        targetCenterX === parentCenterX ? 'center'
        : targetCenterX > parentCenterX ? 'right'
        : 'left'
      );
    }
    if (yFrom === 'auto') {
      yFrom = (
        targetCenterY === parentCenterY ? 'center'
        : targetCenterY > parentCenterY ? 'bottom'
        : 'top'
      );
    }

    let top = (
      yFrom === 'center' ? '50%'
      : yFrom === 'bottom' ? `calc(100% - ${parentBottom - targetCenterY}px)`
      : `${targetCenterY}px`
    );
    let left = (
      xFrom === 'center' ? '50%'
      : xFrom === 'right' ? `calc(100% - ${parentRight - targetCenterX}px)`
      : `${targetCenterX}px`
    );

    switch (dialog.props.arrow) {
    case 'top':
      top = `${targetBottom}px`;
      break;
    case 'bottom':
      top = `calc(100% - ${parentBottom - targetTop}px)`;
      break;
    case 'left':
      left = `${targetRight}px`;
      break;
    case 'right':
      left = `calc(100% - ${parentRight - targetLeft}px)`;
      break;
    default:
    }
    dialog.node.style.left = left;
    dialog.node.style.top = top;
  }
}

let nextId = 0;

export class TourControl extends Control {
  constructor(steps, options) {
    super(options);
    this.id = `tour-${nextId++}`;
    this.steps = [];
    this._currentStepIndex = -1;
    if (steps) {
      for (const step of steps) {
        this.addStep(step);
      }
    }
  }

  get state() {
    const {currentStepIndex} = this;
    return Object.assign(super.state, {currentStepIndex});
  }
  set state(state) {
    super.state = state;
    this.currentStepIndex = state.currentStepIndex ?? -1;
  }

  get currentStepIndex() {
    return this._currentStepIndex;
  }
  set currentStepIndex(index) {
    index = Math.floor(index);
    if (index < 0 || index >= this.steps.length) {
      index = -1;
    }
    if (this._currentStepIndex === index) {
      return;
    }
    if (this._currentStepIndex > -1) {
      this.steps[this._currentStepIndex].hide();
    }
    this._currentStepIndex = index;
    if (this._currentStepIndex > -1) {
      this.steps[this._currentStepIndex].show();
    }
    this.hub.emit('change:currentStepIndex');
    this.hub.emit('change:state');
  }

  addStep(stepOrNode) {
    let step;
    let styleNodes;
    if (stepOrNode instanceof Element) {
      const node = stepOrNode;
      const moveToControl = node.getAttribute('move-to-control');
      const positionToControl = node.getAttribute('position-to-control');

      styleNodes = [...node.querySelectorAll('style')];
      for (const styleNode of styleNodes) {
        styleNode.remove();
      }

      step = {
        id: node.getAttribute('id') || undefined,
        title: node.getAttribute('title') || 'Tour',
        arrow: node.getAttribute('arrow') || false,
        moveTo: (
          moveToControl
          ? (dialog) => dialog.syntool.controls[moveToControl]?.node
          : node.getAttribute('move-to') || undefined
        ),
        positionTo: (
          positionToControl
          ? (dialog) => dialog.syntool.controls[positionToControl]?.node
          : node.getAttribute('position-to') || undefined
        ),
        positionFrom: node.getAttribute('position-from')?.split(',').reverse(),
        content: H([...node.children]),
        prev: node.getAttribute('prev') || undefined,
        next: node.getAttribute('next') || undefined,
      };
    } else {
      step = stepOrNode;
    }

    const {steps} = this;
    const index = steps.length;
    const {
      id = `${this.id}-step${index + 1}-dialog`,
      positionTo,
      positionFrom = ['auto', 'auto'],
      moveTo,
      content,
      prev = 'Back',
      next = 'Next',
      ...props
    } = step;
    let stepNumberNode;
    const stepDialog = new Dialog(id, {
      autoLocationReset: false,
      content: H([
        H(styleNodes),
        H('article', {class: 'TourStepContent -with-scroll-shadows'}, content),
        H('div', {class: 'action-buttons'}, [
          stepNumberNode = H('span'),
          H('button', {class: 'js-prev'}, prev),
          H('button', {class: 'js-next'}, next),
        ]),
      ]),
      onBeforeShow(dialog) {
        for (const n of dialog.contentNode.querySelectorAll('syntool-shortcut')) {
          const shortcutId = n.getAttribute('shortcut-id');
          const shortcutHTML = dialog.syntool.renderShortcutWithId(shortcutId);
          if (shortcutHTML) {
            n.replaceWith(nodeFromHTML(shortcutHTML));
          } else {
            n.remove();
          }
        }
        stepNumberNode.textContent = `${index + 1}/${steps.length}`;
        if (positionTo) {
          positionDialog(dialog, positionTo, positionFrom);
        } else {
          dialog.resetLocation();
        }
        if (moveTo) {
          for (
            let parent = dialog.node.parentElement;
            !parent.matches('.UIOverlay');
            parent = parent.parentElement
          ) {
            parent.classList.add('-tour-on-top');
          }
        }
      },
      onShow(dialog) {
        dialog.node.querySelector('.js-next').focus();
      },
      onHide(dialog) {
        if (moveTo) {
          for (
            let parent = dialog.node.parentElement;
            !parent.matches('.UIOverlay');
            parent = parent.parentElement
          ) {
            parent.classList.remove('-tour-on-top');
          }
        }
      },
      ...props,
    }, {autoActivate: false});

    stepDialog.node.classList.add('TourStepDialog');
    if (positionTo || moveTo) {
      stepDialog.hub
      .on('change:active', () => {
        if (stepDialog.active) {
          moveDialog(stepDialog, moveTo);
          positionDialog(stepDialog, positionTo, positionFrom);
        }
      })
      .on('change:shown', () => {
        if (this.currentStepIndex === index && !stepDialog.shown) {
          this.currentStepIndex = -1;
        }
      });
    }

    DOMEvents(stepDialog.node)
    .on('click', '.js-prev', () => {
      this.currentStepIndex--;
    })
    .on('click', '.js-next', () => {
      this.currentStepIndex++;
    });

    this.steps.push(stepDialog);
    this.syntool?.addControl(stepDialog);
    stepDialog.active = this.active;

    return stepDialog;
  }
  start() {
    this.currentStepIndex = 0;
  }
  close() {
    this.currentStepIndex = -1;
  }

  /*Protected - Control overrides*/
  onAdd() {
    super.onAdd();
    for (const stepDialog of this.steps) {
      this.syntool.addControl(stepDialog);
    }
  }
  onRemove() {
    for (const stepDialog of this.steps) {
      this.syntool.removeControl(stepDialog);
    }
    super.onRemove();
  }
  onActivate() {
    for (const stepDialog of this.steps) {
      stepDialog.active = true;
    }
    super.onActivate();
  }
  onDeactivate() {
    for (const stepDialog of this.steps) {
      stepDialog.active = false;
    }
    super.onDeactivate();
  }
}
