/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  modulo,
  h,
  HIDModifiers,
  HIDKeys,
  Control,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import '../../map_view_extensions/map_view_mouse_position.js';

linkStyleSheet('css/value_under_cursor_control.css');

function angleCCWEastToCWNorth(ccw) {
  return modulo(90 - ccw, 360);
}

export class ValueUnderCursorControl extends Control {
  /*Public*/
  constructor(options) {
    super(options);

    this._lastXY = null;
    this._withProductName = true;

    this.node = document.createElement('div');
    this.node.className = 'ValueUnderCursorControl';

    this.handleMousePositionChange = this.handleMousePositionChange.bind(this);
    this.render = throttleOncePerAnimationFrame(this.render);

    this.shortcuts = [
      {
        groupLabel : 'Global',
        id         : 'toggle-value-under-cursor-visibility',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.V,
        description: 'Toggle the visibility of the vector fields info panel',
        handler    : () => {
          this.active = !this.active;
          return true;
        },
      },
      {
        groupLabel : 'Global',
        id         : 'toggle-value-under-cursor-product-names',
        modifiers  : [HIDModifiers.shift],
        keys       : HIDKeys.I,
        description: 'Toggle display of product name in vector fields info panel',
        autoDisable: true,
        handler    : () => {
          this.withProductName = !this.withProductName;
          return true;
        },
      },
    ];
  }

  get state() {
    const {withProductName} = this;
    return Object.assign(super.state, {withProductName});
  }
  set state(state) {
    super.state = state;
    this.withProductName = state.withProductName ?? true;
  }

  get withProductName() {
    return this._withProductName;
  }
  set withProductName(withProductName) {
    if (this._withProductName === !withProductName) {
      this._withProductName = withProductName;
      this.render();
      this.hub.emit('change:withProductName');
      this.hub.emit('change:state');
    }
  }

  onActivate() {
    this.syntool.usableViewportNode.appendChild(this.node);
    this.syntool.mapView.hub
      .on('change:mouse-position', this.handleMousePositionChange);
    this.render();
  }
  onDeactivate() {
    this.syntool.mapView.hub
      .off('change:mouse-position', this.handleMousePositionChange);
    this.syntool.usableViewportNode.removeChild(this.node);
  }

  render() {
    var xy = this._lastXY;
    var worldBounds;
    var tooltipHtml = '';
    var valueRows;

    if (xy) {
      worldBounds = this.syntool.mapView.getWorldProjectedBounds();
      tooltipHtml = '';

      valueRows = [];
      this.syntool.mapView.forEachDatasetInMap((dataset) => {
        var values = dataset.getValuesAtProjectedXY?.(
          xy,
          worldBounds,
          ['modulus', 'angle']
        );
        if (values && !values.some(isNaN)) {
          valueRows.push(
            this.renderVectorData(dataset.getProduct(), values)
          );
        }
      });
      if (valueRows.length > 0) {
        valueRows.unshift(this.renderHeader());
        tooltipHtml = h('table', valueRows);
      }
    }
    this.node.innerHTML = tooltipHtml;
    this.node.style.display = (tooltipHtml === '' ? 'none' : '');
  }

  /*Protected*/
  handleMousePositionChange(projectedXY) {
    this._lastXY = projectedXY;
    this.render();
  }

  renderHeader() {
    return (
      h('thead', [
        h('tr', [
          this._withProductName && h('th', 'Product'),
          h('th', 'Modulus'),
          h('th', 'Direction'),
        ]),
        h('tr', {class: this.displayClass + 'Units'}, [
          this._withProductName && h('th', 'N/A'),
          h('th', '_'),
          h('th', '° CW from North'),
        ]),
      ])
    );
  }
  renderVectorData(product, values) {
    return (
      h('tr', {class: this.displayClass + 'Entry'}, [
        this._withProductName && h('td', {class: this.displayClass + 'ProductName'}, product.label),
        h('td', {class: this.displayClass + 'ModulusValue'}, values[0].toFixed(3)),
        h('td', {class: this.displayClass + 'AngleValue'}, angleCCWEastToCWNorth(values[1]).toFixed(2)),
      ])
    );
  }
}
window.ODL = Object.assign(window.ODL || {}, {ValueUnderCursorControl});
