/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {HIDKeys, HIDModifiers, renderModifiedKey, Syntool} from '../syntool.js';
import {cancelCopy, pasteFromClipboard, cancelPaste} from '../utils/clipboard.js';
import {initGlobalDropZone} from '../utils/global_drop_zone.js';
import {filePicker} from '../utils/file_picker.js';

const EXPORT_TO_FILE = (
  'Save your preferences and personal data (stored in this browser) in ' +
  'Syntool JSON format'
);
const EXPORT_TO_CLIPBOARD = (
  'Export your preferences and personal data (stored in this browser) in ' +
  'Syntool JSON format to clipboard'
);
const CANCEL_EXPORT_TO_CLIPBOARD = 'Cancel export to clipboard';

let IMPORT_FROM_FILE = 'Import preferences and data';
let IMPORT_FROM_CLIPBOARD = 'Import preferences and data from clipboard';
const CANCEL_IMPORT_FROM_CLIPBOARD = 'Cancel import from clipboard';

let waitingBackupToClipboard = false;
let waitingImportFromClipboard = false;

Syntool.whenInitialised(function() {
  const target = document.createElement('div');
  target.className = 'DropTarget';
  Syntool.uiOverlayNode.insertBefore(
    target,
    Syntool.uiOverlayNode.firstElementChild
  );

  initGlobalDropZone(target, handleGlobalDrop, canAcceptGlobalDrop);

  const supportedImportFormats = (
    '\n\nSupported formats:\n· ' + Syntool.acceptedImportFormats.join('\n· ')
  );
  IMPORT_FROM_FILE += supportedImportFormats;
  IMPORT_FROM_CLIPBOARD += supportedImportFormats;

  Syntool.controls.titleBar.addButton('tools', {
    id: 'importButton',
    icon: 'svg svg-import',
    description: IMPORT_FROM_FILE,
    onClick: handleImportClick,
  });
  Syntool.controls.titleBar.addButton('tools', {
    id: 'backupButton',
    icon: 'svg svg-export',
    description: EXPORT_TO_FILE,
    onClick: handleSaveClick,
  });

  Syntool.keyboardManager.hub.on('change', updateBackupAndImportButtonsState);
});

function canImportFromDataTransferItem(item) {
  switch (item.kind) {
  case 'string':
    return false;
  case 'file':
    return Syntool.canImportFromFile(item.getAsFile());
  default:
    return false;
  }
}
function canAcceptGlobalDrop(event) {
  const dt = event.dataTransfer;

  if (dt.items) {
    for (const item of dt.items) {
      if (canImportFromDataTransferItem(item)) {
        return true;
      }
    }
  }
  if (dt.files) {
    for (const file of dt.files) {
      if (Syntool.canImportFromFile(file)) {
        return true;
      }
    }
  }
  if (dt.types) {
    for (const type of dt.types) {
      if (
        type === 'Files' ||
        type === 'application/x-moz-file' ||
        Syntool.canImportFromMimeType(type)
      ) {
        return true;
      }
    }
  }
  return false;
}

function updateBackupAndImportButtonsState() {
  const backupButton = document.getElementById('backupButton');
  if (backupButton) {
    const backupClassList = backupButton.querySelector('.-icon').classList;
    if (waitingBackupToClipboard) {
      backupButton.title = CANCEL_EXPORT_TO_CLIPBOARD;
      backupClassList.remove('svg-export', 'mf', 'mf-copy');
      backupClassList.add('svg', 'svg-cross');
    } else if (Syntool.keyboardManager.isShiftDown()) {
      backupButton.title = EXPORT_TO_CLIPBOARD;
      backupClassList.remove('svg', 'svg-cross', 'svg-export');
      backupClassList.add('mf', 'mf-copy');
    } else {
      backupButton.title = EXPORT_TO_FILE;
      backupClassList.remove('svg-cross', 'mf', 'mf-copy');
      backupClassList.add('svg', 'svg-export');
    }
  }

  const importButton = document.getElementById('importButton');
  if (importButton) {
    const importClassList = importButton.querySelector('.-icon').classList;
    if (waitingImportFromClipboard) {
      importButton.title = CANCEL_IMPORT_FROM_CLIPBOARD;
      importClassList.remove('svg', 'svg-import', 'mf', 'mf-paste');
      importClassList.add('svg', 'svg-cross');
    } else if (Syntool.keyboardManager.isShiftDown()) {
      importButton.title = IMPORT_FROM_CLIPBOARD;
      importClassList.remove('svg', 'svg-cross', 'svg', 'svg-import');
      importClassList.add('mf', 'mf-paste');
    } else {
      importButton.title = IMPORT_FROM_FILE;
      importClassList.remove('svg', 'svg-cross', 'mf', 'mf-paste');
      importClassList.add('svg', 'svg-import');
    }
  }
}

function handleImportClick(event) {
  if (waitingImportFromClipboard) {
    cancelPaste();
    return;
  }

  const fromClipboard = Syntool.keyboardManager.isShiftDown();

  if (fromClipboard) {
    let messageId = null;

    pasteFromClipboard(function() {
      messageId = Syntool.notificationCenter.showInfo(
        'Press ' + renderModifiedKey(HIDModifiers.ctrlOrMeta, HIDKeys.V) +
        ' to import data from clipboard',
        -1 /* keep it visible */
      );
      waitingImportFromClipboard = true;
      updateBackupAndImportButtonsState();
      return true;
    }, function(string) {
      if (messageId) {
        Syntool.notificationCenter.closeMessage(messageId);
        messageId = null;
      }
      if (string !== false) {
        Syntool.importFromString(string);
      }
      waitingImportFromClipboard = false;
      updateBackupAndImportButtonsState();
    });
  } else {
    filePicker(
      true,
      Syntool.acceptedImportFileTypes,
      (files) => Syntool.importFromFiles(files)
    );
  }
}
function handleGlobalDrop(event) {
  const dt = event.dataTransfer;

  if (dt.items) {
    for (const item of dt.items) {
      switch (item.kind) {
      case 'string':
        item.getAsString((string) => Syntool.importFromString(string));
        break;
      case 'file':
        Syntool.importFromFile(item.getAsFile());
        break;
      default:
      }
    }
  } else if (dt.files) {
    for (const file of dt.files) {
      Syntool.importFromFile(file);
    }
  }
}
function handleSaveClick(event) {
  if (waitingBackupToClipboard) {
    cancelCopy();
    return;
  }

  Syntool.exportJSON(Syntool.exportBackupToJSON(), {
    label   : 'your local data',
    fileName: (
      Syntool.keyboardManager.isShiftDown()
      ? null // to clipboard
      : 'syntool-local-data-backup-' + (new Date().toISOString()) + '.json'
    ),
    onWaiting() {
      waitingBackupToClipboard = true;
      updateBackupAndImportButtonsState();
    },
    onDone() {
      waitingBackupToClipboard = false;
      updateBackupAndImportButtonsState();
    },
  });
}
