/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  H,
  EventsHub,
  DOMEvents,
} from '../../syntool.js';
import {linkStyleSheet, setNodeChildren} from '../../utils/dom_utils.js';
import {clearTextFieldIcon} from '../utils.js';

linkStyleSheet('css/products_list_header_view.css');

export class ProductsListHeaderView {
  constructor() {
    const node = H('div', {class: 'ProductsListHeader'}, [
      H('div', {class: 'ProductsListHeader-Row'}, [
        H('button', {
          class: 'UIButton ProductsListHeader-Button --unreachable-toggle js-unreachable-toggle',
          type : 'button',
          title: 'Toggle filter for unavailable products',
        }, H('i', {class: '-icon fa fa-eye-slash'})),
        H('span', {class: 'ProductsListHeader-Title'}, 'Products'),
      ]),
      H('div', {class: 'ProductsListHeader-Row'}, [
        H('div', {class: 'UIButton ProductsListHeader-Button --grouping js-group-by-button', hidden: true}, [
          H('i', {class: '-icon fa fa-list-alt'}),
          H('select', {
            class: 'ProductsListHeader-GroupingSelect js-group-by-select',
            title: 'Group by...',
          }, [
            H('option', {value: ''}, 'No groups'),
            H('option', {disabled: true}, 'Group by...'),
          ]),
        ]),
        H('div', {class: 'ProductsListHeader-Search'}, [
          H('i', {class: 'ProductsListHeader-SearchIcon fa fa-search'}),
          H('input', {
            class      : 'ProductsListHeader-SearchInput js-search',
            type       : 'search',
            inputmode  : 'search',
            placeholder: 'Search...',
          }),
          H('button', {
            class: 'ProductsListHeader-SearchClearButton js-clear',
            'aria-label': 'Clear filter',
          }, clearTextFieldIcon),
        ]),
      ]),
    ]);

    this.node = node;
    this._hasUnreachableProducts = false;
    this._hideUnreachableProducts = true;
    this.groupTags = [];
    this.selectedGroupTag = null;

    this.hub = new EventsHub();

    this.render = throttleOncePerAnimationFrame(this.render);

    // Event handlers
    // ==============
    DOMEvents(node)
    .on('click', '.js-unreachable-toggle', () => {
      this.hub.emit('show-unreachable-toggle');
    })
    .on('change', '.js-group-by-select', (event) => {
      const tag = event.target.value || false;

      // Restore the selected tag
      event.target.value = this.selectedGroupTag || '';

      this.hub.emit('group-by-change', tag);
    })
    .on('input', '.js-search', (event) => {
      this.hub.emit('search-change', event.target.value);
    })
    .on('click', '.js-clear', () => {
      this.search = '';
      this.node.querySelector('.js-search').focus();
    });
  }

  get search() {
    const searchNode = this.node.querySelector('.js-search');
    return searchNode.value;
  }
  set search(search) {
    const searchNode = this.node.querySelector('.js-search');
    searchNode.value = search;
    this.hub.emit('search-change', search);
  }
  get hasUnreachableProducts() {
    return this._hasUnreachableProducts;
  }
  set hasUnreachableProducts(hasUnreachableProducts) {
    if (this._hasUnreachableProducts === !hasUnreachableProducts) {
      this._hasUnreachableProducts = !!hasUnreachableProducts;
      this.render();
    }
  }
  get hideUnreachableProducts() {
    return this._hideUnreachableProducts;
  }
  set hideUnreachableProducts(hideUnreachableProducts) {
    if (this._hideUnreachableProducts === !hideUnreachableProducts) {
      this._hideUnreachableProducts = !!hideUnreachableProducts;
      this.render();
    }
  }
  focusOnSearch() {
    const searchNode = this.node.querySelector('.js-search');
    if (searchNode) {
      searchNode.focus();
      searchNode.select();
    }
  }

  render() {
    const node = this.node;
    const hasUnreachableProducts = this._hasUnreachableProducts;
    const hideUnreachableProducts = this._hideUnreachableProducts;
    const groupTags = this.groupTags;
    const selectedGroupTag = (
      this.selectedGroupTag ? this.selectedGroupTag.toLowerCase() : ''
    );

    const unreachableToggle = node.querySelector('.js-unreachable-toggle');
    unreachableToggle.hidden = !hasUnreachableProducts;
    unreachableToggle.classList.toggle('-blue', hideUnreachableProducts);

    const groupByButton = node.querySelector('.js-group-by-button');
    groupByButton.hidden = (groupTags.length === 0);

    const groupBySelect = node.querySelector('.js-group-by-select');
    const groupByOptionsMap = new Map(
      Array.from(groupBySelect.children, function(option) {
        return [option.value, option];
      })
      // Skip the first 2 options
      .slice(2)
    );

    // Compute new options
    const groupByOptions = groupTags.map(function(groupTag) {
      let element = groupByOptionsMap.get(groupTag);
      if (!element) {
        element = document.createElement('option');
        element.value = groupTag;
      }

      const selected = groupTag.toLowerCase() === selectedGroupTag;
      element.selected = selected;
      element.textContent = groupTag + (selected ? ' ✓' : '');
      return element;
    });

    groupBySelect.firstChild.textContent = (
      'No groups' + (selectedGroupTag ? '' : ' ✓')
    );

    // Add the first 2 options
    groupByOptions.unshift(
      groupBySelect.firstChild,
      groupBySelect.firstChild.nextElementSibling
    );

    // Update options
    setNodeChildren(groupBySelect, groupByOptions);
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProductsListHeaderView});
