/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  throttleOncePerAnimationFrame,
  H,
  isFirefox,
  EventsHub,
  DOMEvents,
  RegExpHelpers,
} from '../../syntool.js';
import {linkStyleSheet, setNodeChildren} from '../../utils/dom_utils.js';

linkStyleSheet('css/products_list_view.css');

const productTagRe = new RegExp(RegExpHelpers.productTag);
const productNode = H('div', {class: 'ProductsList-Product js-product'}, [
  H('input', {
    class: 'ProductsList-Checkbox js-product-checkbox',
    type : 'checkbox',
  }),
  H('label', {class: 'ProductsList-ProductLabel js-product-label'}),
  H('div', {class: 'ProductsList-ProductButtons'}, [
    H('button', {
      class: 'js-product-delete ProductsList-ProductButton --delete',
      title: 'Remove',
      type : 'button',
    }, H('i', {class: 'fa fa-trash'})),
    H('button', {
      class: 'js-product-zoom-in ProductsList-ProductButton --zoom-in',
      title: 'Zoom-in to a level where this product is visible',
      type : 'button',
    }, H('i', {class: 'fa fa-search-plus'})),
    H('button', {
      class: 'js-product-zoom-out ProductsList-ProductButton --zoom-out',
      title: 'Zoom-out to a level where this product is visible',
      type : 'button',
    }, H('i', {class: 'fa fa-search-minus'})),
    H('button', {
      class: 'js-product-info ProductsList-ProductButton --info',
      title: 'More info',
      type : 'button',
    }, H('i', {class: 't t-i'})),
  ]),
]);

function renderProductElement(productId, prefix) {
  const node = productNode.cloneNode(true);
  const checkbox = node.querySelector('.js-product-checkbox');
  const label = node.querySelector('.js-product-label');
  const checkboxId = `checkbox-${prefix}-${productId}`;

  node.dataset.productId = productId;
  checkbox.value = productId;
  checkbox.id = checkboxId;
  checkbox.name = checkboxId;
  label.htmlFor = checkboxId;
  return node;
}

export class ProductsListView {
  constructor() {
    var node = document.createElement('div');
    node.className = 'ProductsList';

    this.node = node;
    this.syntool = null;
    this.store = null;
    this.productIds = [];
    this.groupBy = false;
    this.defaultGroupLabel = null;
    this.sortGroups = null;
    this.collapsibleGroups = true;
    this._collapsedGroups = new Set();
    this._elementShouldBeVisibleAfterRender = null;
    this._productElements = new Map();
    this._groupElements = new Map();

    this.hub = new EventsHub();

    this.render = throttleOncePerAnimationFrame(this.render);

    var longPressTimer = null;
    function handleProductLongPress(productId) {
      longPressTimer = null;
      this.handleSelectOneProduct(productId);
    }

    // Event handlers
    // ==============
    var self = this;
    DOMEvents(node)
    .on('click', '.js-group-header', function handleToggleGroup(event) {
      if (self.collapsibleGroups) {
        var groupId = this.parentNode.dataset.id || null;

        if (!self._collapsedGroups.delete(groupId)) {
          self._collapsedGroups.add(groupId);
          /* When collapsing a group, should ensure that the group header
          remains visible after re-render*/
          self._elementShouldBeVisibleAfterRender = this;
        }
        self.render();
        self.hub.emit('change:collapsedGroups');
      }
    })
    .on('touchstart', '.js-product', function(event) {
      var productId = this.getAttribute('data-product-id');
      longPressTimer = setTimeout(
        handleProductLongPress.bind(self, productId),
        500
      );
    })
    .on(['touchmove', 'touchend', 'touchcancel'], '.js-product', function(event) {
      if (longPressTimer === null && event.type === 'touchend') {
        /*
         * Workaround bug rare bug in Firefox Focus (and maybe others) in iOS
         * When long press ends (finger lifted), the checkbox is toggled
         */
        event.preventDefault();
      }
      clearTimeout(longPressTimer);
      longPressTimer = null;
    })
    .on('click', '.js-product-info', function handleInfoClick(event) {
      var productId = this.parentNode.parentNode.getAttribute('data-product-id');
      self.hub.emit('click:product-info', productId);
    })
    .on('click', '.js-product-delete', function handleDeleteClick(event) {
      var productId = this.parentNode.parentNode.getAttribute('data-product-id');
      self.hub.emit('click:product-delete', productId);
    })
    .on('click', '.js-product-zoom-in', function handleZoomInClick(event) {
      var productId = this.parentNode.parentNode.getAttribute('data-product-id');
      self.hub.emit('click:product-zoom-in', productId);
    })
    .on('click', '.js-product-zoom-out', function handleZoomOutClick(event) {
      var productId = this.parentNode.parentNode.getAttribute('data-product-id');
      self.hub.emit('click:product-zoom-out', productId);
    })
    .on('change', '.js-product-checkbox', function handleSelectionChange(event) {
      var checkbox = event.target;
      self.handleProductSelectionChange(checkbox);
    });

    if (isFirefox) {
      /*
       * Workaround bug in Firefox
       * https://bugzilla.mozilla.org/show_bug.cgi?id=559506
       */
      DOMEvents(node)
      .on('click', '.js-product-label', function handleProductClick(event) {
        if (self.syntool?.keyboardManager.isShiftDown()) {
          var checkbox = document.getElementById(this.htmlFor);
          checkbox.checked = !checkbox.checked;
          self.handleProductSelectionChange(checkbox);
        }
      });
    }
  }

  get groupsCollapsible() {
    return this.collapsibleGroups;
  }
  set groupsCollapsible(collapsible) {
    if (this.collapsibleGroups !== !!collapsible) {
      this.collapsibleGroups = !!collapsible;
      this.render();
    }
  }
  get collapsedGroups() {
    return new Set(this._collapsedGroups);
  }
  set collapsedGroups(collapsedGroups) {
    let changed = false;
    for (const groupId of this._collapsedGroups) {
      if (!collapsedGroups.has(groupId)) {
        this._collapsedGroups.delete(groupId);
        changed = true;
      }
    }
    for (const groupId of collapsedGroups) {
      if (!this.collapsedGroups.has(groupId)) {
        this._collapsedGroups.add(groupId);
        changed = true;
      }
    }
    if (changed) {
      this.render();
      this.hub.emit('change:collapsedGroups');
    }
  }

  isProductWithIdSelected(productId) {
    return this.store.isProductSelected(productId);
  }

  handleProductSelectionChange(checkbox) {
    var productId = checkbox.value;
    var selected = checkbox.checked;

    // Restore the checkbox's original state
    this._updateCheckbox(checkbox);

    if (this.syntool?.keyboardManager.isShiftDown()) {
      this.handleSelectOneProduct(productId);
    } else if (selected) {
      this.handleSelectProduct(productId);
    } else {
      this.handleDeselectProduct(productId);
    }
  }
  handleSelectProduct(productId) {
    this.hub.emit('select-product', productId);
  }
  handleDeselectProduct(productId) {
    this.hub.emit('deselect-product', productId);
  }
  handleSelectOneProduct(productId) {
    window.getSelection().removeAllRanges();
    this.hub.emit('select-one-product', productId);
  }

  render() {
    if (!this.store) {
      return;
    }

    var node = this.node;
    var store = this.store;
    var productIds = this.productIds;
    var _productElements = this._productElements;

    // Remove elements of non-existing products
    for (const [productId] of _productElements) {
      if (!store.hasProduct(productId)) {
        _productElements.delete(productId);
      }
    }

    var children = (
      this.groupBy
      ? this.renderWithGroups(productIds)
      : this.renderWithoutGroups(productIds)
    );

    // Update the DOM
    setNodeChildren(node, children);

    node.classList.toggle(
      '-has-scrollbar',
      node.offsetWidth > node.clientWidth /* + borders*/
    );
    if (this._elementShouldBeVisibleAfterRender) {
      var frameRect = node.getBoundingClientRect();
      var elementRect = (
        this._elementShouldBeVisibleAfterRender.getBoundingClientRect()
      );
      if (elementRect.bottom <= frameRect.top) {
        this._elementShouldBeVisibleAfterRender.scrollIntoView(true);
      } else if (elementRect.top >= frameRect.bottom) {
        this._elementShouldBeVisibleAfterRender.scrollIntoView(false);
      }
      this._elementShouldBeVisibleAfterRender = null;
    }
  }
  renderWithoutGroups(productIds) {
    var zoom = this.syntool.mapView.getZoom();
    return productIds.map(
      (productId) => this.renderProductWithId(zoom, productId, 0)
    );
  }
  renderWithGroups(productIds) {
    var store = this.store;
    var groupBy = this.groupBy && this.groupBy.toLowerCase();
    var groups = new Map();
    var defaultGroup = {
      id     : null,
      value  : null,
      label  : this.defaultGroupLabel,
      entries: [],
    };

    productIds.forEach(function(productId) {
      var product = store.getProduct(productId);
      var groupsCount = 0;

      if (!product) {
        return;
      }

      for (var i = 0; i < product.tags.length; i++) {
        var tag = product.tags[i];
        var parts = productTagRe.exec(tag);
        var key = parts[1];
        if (key && key.toLowerCase() === groupBy) {
          key = groupBy;
          var label = parts[2];
          var value = label.toLowerCase();
          var id = key + ':' + value;
          var group = groups.get(id);
          if (group) {
            group.entries.push({productId, index: groupsCount});
          } else {
            group = {
              id,
              value,
              label,
              entries: [{productId, index: groupsCount}],
            };
            groups.set(id, group);
          }

          groupsCount++;
        }
      }

      // Add the product to the default group if it has none
      if (groupsCount === 0) {
        defaultGroup.entries.push({productId, index: groupsCount});
      }
    });

    // Add the default group if it's not empty
    if (defaultGroup.entries.length > 0) {
      groups.set(null, defaultGroup);
    }

    // Render without groups if there isn't at least two groups
    if (groups.size < 2) {
      return this.renderWithoutGroups(productIds);
    }

    var collapsibleGroups = this.collapsibleGroups;
    var collapsedGroups = this._collapsedGroups;
    var _groupElements = this._groupElements;
    var zoom = this.syntool.mapView.getZoom();
    var groupsSet = new Set();
    var groupsList = [...groups.values()];
    if (this.sortGroups) {
      var comparator = this.sortGroups(groupBy);
      groupsList.sort(function(g1, g2) {
        return comparator(g1.value, g2.value);
      });
    }
    var children = groupsList.map(function renderGroup(group) {
      groupsSet.add(group.id);

      var groupCollapsed = collapsibleGroups && collapsedGroups.has(group.id);
      var groupChildren = group.entries.map(function(entry) {
        var element = (
          this.renderProductWithId(zoom, entry.productId, entry.index)
        );
        if (element) {
          element.classList.toggle('is-hidden-by-group', groupCollapsed);
        }
        return element;
      }, this).filter(Boolean);

      if (groupChildren.length === 0) {
        return null;
      }

      var groupElement = _groupElements.get(group.id);
      if (!groupElement) {
        groupElement = document.createElement('div');
        groupElement.className = 'ProductsList-Group';
        groupElement.dataset.id = group.id || '';

        if (group.label !== null) {
          var groupHeader = document.createElement('span');
          groupHeader.className = 'ProductsList-GroupHeader js-group-header';
          if (group.id === null) {
            groupHeader.innerHTML = group.label;
          } else {
            groupHeader.textContent = group.label;
          }
          groupElement.appendChild(groupHeader);
        }

        _groupElements.set(group.id, groupElement);
      }

      groupElement.classList.toggle(
        'is-expanded',
        collapsibleGroups && !groupCollapsed
      );
      groupElement.classList.toggle('is-reduced', groupCollapsed);

      // Add group header element
      if (group.label !== null) {
        groupChildren.unshift(groupElement.firstChild);
      }

      // Update the DOM
      setNodeChildren(groupElement, groupChildren);
      return groupElement;
    }, this).filter(Boolean);

    // Remove elements of non-existing groups
    for (const [groupTag] of _groupElements) {
      if (!groupsSet.has(groupTag)) {
        _groupElements.delete(groupTag);
      }
    }

    return children;
  }
  renderProductWithId(zoom, productId, index) {
    var elementsMapById = this._productElements;
    var elementsList = elementsMapById.get(productId);

    // Create missing product elements map
    if (!elementsList) {
      elementsList = [];
      elementsMapById.set(productId, elementsList);
    }

    var element = elementsList[index];

    // Create missing product element
    if (!element) {
      element = renderProductElement(productId, index);
      elementsList[index] = element;
    }

    var product = this.store.getProduct(productId);

    // Update the elements to represent the current product state
    this._updateElement(element, product, zoom);
    return element;
  }
  _updateCheckbox(checkbox) {
    var productIsSelected = this.isProductWithIdSelected(checkbox.value);

    if (checkbox.checked !== productIsSelected) {
      checkbox.checked = productIsSelected;
    }
  }
  _updateElement(element, product, zoom) {
    const selectedInCatalog = this.isProductWithIdSelected(product.id);
    const zoomIn = zoom < product.mapMinZoom;
    const zoomOut = zoom > product.mapMaxZoom;
    const reachable = this.syntool.filters.product.withReachableServer(product);
    const unreachable = this.syntool.filters.product.withUnreachableServer(product);
    const selected = this.syntool.filters.product.selected(product);
    const invisible = product.opacity === 0;

    const filtersActive = (
      this.store.productsStore.getProductDatasetFilters(product.id)
    );
    const filteredOutWarnings = [];
    for (const {label, filters} of product.datasetFilters) {
      const inactiveFilters = [];
      for (const {id, label: filterLabel} of filters) {
        if (!filtersActive[id]) {
          inactiveFilters.push(`\t· ${filterLabel}`);
        }
      }
      if (inactiveFilters.length === filters.length) {
        filteredOutWarnings.push(
          `⚠️ Select at least one "${label}" of this product in ` +
          'Display data panel to see it:\n' + inactiveFilters.join('\n')
        );
      }
    }
    const filteredOut = filteredOutWarnings.length > 0;

    var checkbox = element.querySelector('.js-product-checkbox');
    var label = element.querySelector('.js-product-label');

    if (label.innerHTML !== product.label) {
      label.innerHTML = product.label;
    }

    const warnings = [
      (
        unreachable
        ? (
          reachable
          ? '⚠️ Partially unavailable due to server or network errors'
          : '⚠️ Unavailable due to server or network errors'
        )
        : ''
      ),
      selectedInCatalog && zoomIn ? '⚠️ Zoom-in to see this product' : '',
      selectedInCatalog && zoomOut ? '⚠️ Zoom-out to see this product' : '',
      (
        selectedInCatalog && !selected
        ? '⚠️ Select this product in Display data panel to see it'
        : ''
      ),
      (
        selectedInCatalog && invisible
        ? "⚠️ Increase this product's opacity in Display data panel to see it"
        : ''
      ),
      (
        selectedInCatalog && filteredOut
        ? filteredOutWarnings.join('\n')
        : ''
      ),
    ].filter(Boolean).join('\n');
    const tooltip = ((product.description || '') + '\n\n' + warnings).trim();
    if (tooltip) {
      label.setAttribute('title', tooltip);
    } else if (label.hasAttribute('title')) {
      label.removeAttribute('title');
    }

    this._updateCheckbox(checkbox);
    element.classList.toggle('has-info', !!product.infoURL);
    element.classList.toggle(
      '-removable',
      this.store.canDeleteProduct(product.id)
    );
    element.classList.toggle('-zoom-in', zoomIn);
    element.classList.toggle('-zoom-out', zoomOut);
    element.classList.toggle('-reachable', reachable);
    element.classList.toggle('-unreachable', unreachable);
    element.classList.toggle('-selected', selected);
    element.classList.toggle('-invisible', invisible);
    element.classList.toggle('-filtered-out', filteredOut);
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProductsListView});
