/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  H as h,
  S as s,
  HIDKeys,
  isModifiedEvent,
  keyFromEvent,
  EventsHub,
  DOMEvents,
  setGlobalCursor,
  preventEventDefault,
  stopEventPropagation,
  isFirefox,
  Pattern,
  Syntool,
} from '../../syntool.js';
import {colorToCSSColor} from '../../utils/color.js';
import {
  COLOR_PALETTE_MISSING,
  COLOR_PALETTE_LOADING,
  COLOR_PALETTE_LOADED,
  COLOR_PALETTE_ERRORED,
  getColorPalette,
  fetchColorPalette,
  getColorPaletteStatus,
  colorPaletteGetCSSClassName,
} from '../../utils/color_palette.js';
import {linkStyleSheet, setNodeChildren} from '../../utils/dom_utils.js';
import {GaugeComponent} from '../components/gauge_component.js';
import {Gestures, clearTextFieldIcon} from '../utils.js';
import '../../features/product_colormaps.js';

linkStyleSheet('css/products_settings_list_view.css');

function descendingNumberComparator(a, b) {
  return b - a;
}
function descendingProductsStackLevelComparator(a, b) {
  return descendingNumberComparator(a.stackLevel, b.stackLevel);
}

export class ProductsSettingsListView {
  constructor(node) {
    this.node = node;
    /** @type {import('../../syntool.js').ProductsStore | null} */
    this.store = null;
    this.hub = new EventsHub();
    this.selectedDatasetsByProduct = new Map();

    node.classList.add('ProductsSettingsListView');

    this.initialize();
  }

  initialize() {
    const self = this;
    let longPressTimer;
    function startLongPressTimer(callback) {
      clearTimeout(longPressTimer);
      longPressTimer = setTimeout(() => {
        longPressTimer = null;
        callback();
      }, 500);
    }
    function cancelLongPressTimer(event) {
      if (longPressTimer === null && event.type === 'touchend') {
        /*
         * Workaround bug rare bug in Firefox Focus (and maybe others) in iOS
         * When long press ends (finger lifted), the checkbox is toggled
         */
        event.preventDefault();
      }
      clearTimeout(longPressTimer);
      longPressTimer = null;
    }

    DOMEvents(this.node)
    .on('touchstart', '.js-product-checkbox', function(event) {
      const productId = self._getProductIdForElement(this);
      startLongPressTimer(() => self.handleSelectOneProduct(productId));
    })
    .on(['touchmove', 'touchend', 'touchcancel'], '.js-checkbox', cancelLongPressTimer)
    .on('change', '.js-product-checkbox', function(event) {
      self.handleProductSelectionChange(this);
    })
    .on('change', '.js-filter-toggle', function(event) {
      const productId = self._getProductIdForElement(this);

      self._commitFilterField(productId);
      if (this.checked) {
        self.hub.emit('change:filter:activate', productId);
      } else {
        self.hub.emit('change:filter:deactivate', productId);
      }
      self._updateItemForProductWithId(productId);
    })
    .on('click', '.js-filter-clear', function() {
      this.previousElementSibling.value = '';
      self._updateFilterFieldState(self._getProductIdForElement(this));
      self._commitFilterField(self._getProductIdForElement(this));
      this.previousElementSibling.focus();
    })
    .on('click', '.js-remove', function() {
      const productId = self._getProductIdForElement(this);
      self.hub.emit('remove-product', productId);
    })
    // On iOS Safari, when double-tapping on the remove button, a dblclick is
    // fired on the map causing an unintended a zoom-in. We capture that event
    // to prevent that behavior.
    .on('dblclick', '.js-remove', preventEventDefault)
    .on('input', '.js-filter-input', function(event) {
      self._updateFilterFieldState(self._getProductIdForElement(this));
    })
    .on('change', '.js-filter-input', function(event) {
      self._commitFilterField(self._getProductIdForElement(this));
    })
    .on('click', '.js-toggle-dataset-filters', function(event) {
      this.parentElement.classList.toggle('-expanded');
    })
    .on('touchstart', '.js-dataset-filter', function(event) {
      const productId = self._getProductIdForElement(this);
      const checkbox = this.querySelector('.js-dataset-filter-checkbox');
      const group = (
        checkbox
        .closest('.js-dataset-filters-group')
        .querySelector('.js-dataset-filters-group-label')
        .textContent
      );
      const {value: filter} = checkbox;
      startLongPressTimer(
        () => self.handleActivateOneProductDatasetFilter(productId, group, filter)
      );
    })
    .on('change', '.js-dataset-filter-checkbox', function(event) {
      self.handleProductDatasetFilterActiveChange(this);
    })
    .on('mousedown', '.js-no-mouse-reorder-handle', stopEventPropagation);

    if (isFirefox) {
      /*
       * Workaround bug in Firefox
       * https://bugzilla.mozilla.org/show_bug.cgi?id=559506
       */
      DOMEvents(this.node)
      .on('click', '.js-dataset-filter-label', function(event) {
        if (Syntool.keyboardManager.isShiftDown()) {
          const checkbox = this.previousElementSibling;
          checkbox.checked = !checkbox.checked;
          self.handleProductDatasetFilterActiveChange(checkbox);
        }
      });
    }

    const reorderOptions = {
      canReorderElementAtIndex(index) {
        return self.store.canSetProductStackLevel(
          self._getProductIdForElement(self.node.children[index])
        );
      },
      getElementForHandle(handle) {
        return handle.closest('.js-product');
      },

      updateElementStyleAtIndex(fromIndex, toIndex, translation) {
        const element = self.node.children[fromIndex];
        element.style.transform = (
          translation === 0 ? '' : 'translate3d(0,' + translation + 'px,0)'
        );

        if ((toIndex - fromIndex) % 2 === 0) {
          element.classList.remove('-flip-alternation');
        } else {
          element.classList.add('-flip-alternation');
        }
      },
      resetElementStyleAtIndex(index) {
        const element = self.node.children[index];
        element.style.transform = '';
        element.classList.remove('-flip-alternation');
      },

      deltaForKeyDown(event) {
        if (isModifiedEvent(event)) {
          return 0;
        }
        switch (keyFromEvent(event)) {
        case HIDKeys.pageUp:
          return -1;
        case HIDKeys.pageDown:
          return 1;
        default:
          return 0;
        }
      },

      onStart(element) {
        setGlobalCursor('grabbing');
        self.node.classList.add('-reordering');
        element.classList.add('-dragging');
        element.querySelector('.js-reorder-handle').focus();
      },
      onEnd(element) {
        self.node.classList.remove('-reordering');
        element.classList.remove('-dragging');
        setGlobalCursor(null);
      },
    };
    const reorderHandler = this.handleReorder.bind(this);
    Gestures.reorder(
      this.node,
      '.-can-move .js-reorder-handle',
      reorderOptions,
      reorderHandler
    );
    Gestures.reorder(
      this.node,
      '.-can-move .js-mouse-reorder-handle',
      Object.assign({touch: false}, reorderOptions),
      reorderHandler
    );
  }

  render(productIds = this.store.getProductIds()) {
    const node = this.node;
    const productIdsSet = new Set(productIds);
    const products = productIds.map(
      (productId) => this.store.getProduct(productId)
    );

    // Sort products by stack level
    products.sort(descendingProductsStackLevelComparator);

    // Remove all the list items corresponding to unselected products
    for (const child of [...node.children]) {
      if (!productIdsSet.has(this._getProductIdForElement(child))) {
        node.removeChild(child);
      }
    }

    // Save scroll offset
    const scrollTop = node.scrollTop;
    // Save focus
    const focusedElement = node.querySelector(':focus');

    // Get the list items (reuse the existing ones)
    products.forEach((product, index) => {
      const item = this._getOrCreateItemForProduct(product);

      // Move the list item to the correct position, if needed
      if (node.children[index] !== item) {
        node.insertBefore(item, node.children[index]);
      }

      // Update the item to represent the current product state
      this._updateItem(item, product);
    });

    // Restore focus
    if (node.querySelector(':focus') !== focusedElement) {
      focusedElement.focus();
    }
    // Restore scroll offset
    node.scrollTop = scrollTop;
  }

  _getAllItems() {
    return [...this.node.children];
  }
  _getItemForProductWithId(productId) {
    return this.node.querySelector(`[data-product-id="${productId}"]`);
  }
  _getElementForProductWithId(productId, childSelector = '') {
    return this.node.querySelector(
      `[data-product-id="${productId}"] ${childSelector}`
    );
  }
  _createItemForProduct(product) {
    // Start loading the product's colormaps, if not loaded already
    if (product.colormaps) {
      product.colormaps.forEach((colormap) => {
        switch (getColorPaletteStatus(colormap.colorPalette)) {
        case COLOR_PALETTE_MISSING:
        case COLOR_PALETTE_LOADING:
          fetchColorPalette(colormap.colorPalette)
            .finally(() => this._updateColormapsForProductWithId(product.id))
            .catch(emptyFunction);
          break;
        default:
        }
      });
    }

    const item = this._renderItem(product);
    const canSetOpacity = this.store.canSetProductOpacity(product.id);
    const opacityNode = item.querySelector('.js-opacity');
    const opacityGauge = new GaugeComponent({
      class   : 'ProductSettings-OpacityGauge',
      tabindex: canSetOpacity ? '0' : undefined,
      title   : 'Opacity: ${value}%',
      value   : Math.round(product.opacity * 100),
      disabled: !canSetOpacity,
      onChange: (opacity) => this.handleProductOpacityChange(product.id, opacity),
      onClick : () => this.handleProductOpacityClick(product.id),
    });
    opacityNode.parentNode.replaceChild(opacityGauge.node, opacityNode);
    item.opacityGauge = opacityGauge;
    return item;
  }
  _getOrCreateItemForProduct(product) {
    return (
      this._getItemForProductWithId(product.id) ??
      this._createItemForProduct(product)
    );
  }

  _renderColormapTick(tick, index, ticks) {
    const x = tick[0] * 100 + '%';
    return h([
      s('line', {x1: x, y1: '25%', x2: x, y2: '50%'}),
      s('text', {x: x, y: '100%', dy: -1}, tick[1]),
    ]);
  }
  _renderColormapTriangle(position, color) {
    return s('svg', {
      class              : 'Colormap-Triangle -' + position,
      viewBox            : '0 0 10 10',
      preserveAspectRatio: 'none',
    }, (
      isNaN(color)
      ? s('polygon', {points: '0.5 0, 0.5 10'})
      : s('polygon', {
        points: '0.5 0, 0.5 10, 9.5 5',
        fill  : colorToCSSColor(color),
      })
    ));
  }
  _renderColormap({class: className = '', url, ticks}) {
    const colorPaletteStatus = getColorPaletteStatus(url);

    switch (colorPaletteStatus) {
    case COLOR_PALETTE_LOADED: {
      const colorPalette = getColorPalette(url);
      return h('div', {class: `Colormap ${className}`}, [
        h('div', {
          class: (
            colorPaletteGetCSSClassName(colorPalette) +
            ' Colormap-Gradiant'
          ),
        }),
        s('svg', {class: 'Colormap-Ticks'},
          ticks.map(this._renderColormapTick, this)
        ),
        this._renderColormapTriangle('left', colorPalette.backgroundColor),
        this._renderColormapTriangle('right', colorPalette.foregroundColor),
      ]);
    }
    case COLOR_PALETTE_LOADING:
      return h('div', {class: `Colormap -loading ${className}`},
        'Loading color palette...'
      );
    case COLOR_PALETTE_ERRORED:
      return h('div', {class: `Colormap -error ${className}`},
        'Could not load color palette :('
      );
    default:
      return null;
    }
  }
  _renderLabeledColormap(colormap) {
    const colormapHTML = this._renderColormap({
      class: 'ProductSettings-Colormap',
      url  : colormap.colorPalette,
      ticks: colormap.ticks,
    });

    if (!colormapHTML) {
      return null;
    }

    const labelHTML = (
      colormap.label
      ? h('span', {class: 'ProductSettings-ColormapLabel'}, colormap.label)
      : ''
    );

    return h([labelHTML, colormapHTML]);
  }
  _renderColormaps(colormaps) {
    return h('div', {
      class: 'ProductSettings-Colormaps js-colormaps js-mouse-reorder-handle',
    }, colormaps.map(this._renderLabeledColormap, this));
  }
  /**
   * @protected
   * @param {DatasetFilter} filter
   * @returns {HTMLElement}
   */
  _renderDatasetFilter({id}) {
    return h('label', {
      class: (
        'ProductSettings-DatasetFilter js-dataset-filter js-checkbox ' +
        'js-no-mouse-reorder-handle'
      ),
    }, [
      h('input', {
        type: 'checkbox',
        class: 'ProductSettings-DatasetFilterCheckbox js-dataset-filter-checkbox',
        value: id,
      }),
      h('span', {class: 'ProductSettings-DatasetFilterLabel js-dataset-filter-label'})
    ]);
  }
  /**
   * @protected
   * @param {DatasetFiltersGroup} group
   * @returns {HTMLElement}
   */
  _renderDatasetFiltersGroup(group) {
    return h('div', {
      class: 'ProductSettings-DatasetFiltersGroup js-dataset-filters-group',
      role: 'group',
      'aria-label': group.label,
    }, [
      h('span', {
        class: 'ProductSettings-DatasetFiltersGroupLabel js-dataset-filters-group-label',
      }, group.label),
      ...group.filters.map(this._renderDatasetFilter, this),
    ]);
  }
  /**
   * @protected
   * @param {DatasetFiltersGroup[]} groups
   * @returns {HTMLElement}
   */
  _renderDatasetFiltersGroups(groups) {
    const count = groups?.length;
    return count > 0 && h('div', {
      class: (
        'ProductSettings-DatasetFilters ' + (count === 1 ? '-single-group ' : '') +
        'js-dataset-filters js-mouse-reorder-handle'
      ),
    }, [
      h('button', {
        class: (
          'ProductSettings-ToggleDatasetFiltersExpanded js-toggle-dataset-filters ' +
          'js-no-mouse-reorder-handle'
        ),
        title: 'Toggle details'
      }),
      ...groups.flatMap(this._renderDatasetFiltersGroup, this),
    ]);
  }

  _renderItem(product) {
    let unit = product.unit;
    let isLog = false;

    if (unit && unit.indexOf('log ') === 0) {
      isLog = true;
      unit = unit.substr(4);
    }

    return h('div', {
      class: 'ProductSettings js-product',
      'data-product-id': product.id,
    }, [
      h('div', {class: 'ProductSettings-TitleBar'}, [
        h('span', {
          class   : 'ProductSettings-DragHandle fa fa-arrows-v js-reorder-handle',
          title   : 'Drag to reorder',
          tabindex: 0,
        }),
        h('input', {
          class  : 'ProductSettings-Checkbox js-checkbox js-product-checkbox',
          type   : 'checkbox',
          checked: true,
        }),
        h('span', {
          class: 'ProductSettings-Title js-mouse-reorder-handle',
          title: product.id,
        }, product.label),
        (isLog || unit) && h('span', {
          class: (
            'ProductSettings-Unit' + (isLog ? ' -is-log' : '') +
            ' js-mouse-reorder-handle'
          ),
          title: unit + (isLog ? ' (log scale)' : ''),
        }, unit.trim()),
        h('label', {
          class: 'ProductSettings-Toggle ProductSettings-FilterToggle',
          title: 'Filter datasets',
        }, [
          h('input', {
            type    : 'checkbox',
            class   : 'ProductSettings-ToggleCheckbox js-filter-toggle',
            checked : false,
            tabindex: 0,
          }),
          h('i', {class: 'fa fa-filter'}),
        ]),
        h('div', {class: 'js-opacity'}),
        h('button', {
          class: 'ProductSettings-Button ProductSettings-RemoveButton js-remove',
          title: 'Remove product from list',
        }, h('i', {class: 'fa fa-remove'})),
      ]),
      this._renderDatasetFiltersGroups(product.datasetFilters),
      h('div', {class: 'ProductSettings-Filter js-filter'}, [
        h('label', {
          for  : 'filter-' + product.id,
          class: 'js-filter-label',
        }, 'ID Contains'),
        h('input', {
          type       : 'text',
          id         : 'filter-' + product.id,
          class      : 'ProductSettings-TextInput ProductSettings-FilterInput js-filter-input',
          value      : '',
          placeholder: 'Datasets ID pattern (Press Enter to validate)',
        }),
        h('button', {
          class: 'ProductSettings-FilterClearButton js-filter-clear',
          'aria-label': 'Clear filter',
        }, clearTextFieldIcon),
      ]),
      product.colormaps && this._renderColormaps(product.colormaps),
    ]);
  }

  _updateItem(item, product) {
    const filterElement = item.querySelector('.js-filter');
    const filterToggle = item.querySelector('.js-filter-toggle');
    const filterField = item.querySelector('.js-filter-input');
    const checkbox = item.querySelector('.js-product-checkbox');
    const isSelected = this.store.isProductSelected(product.id);
    const patternFilter = this.store.getProductPattern(product.id) || '';
    const patternFilterActive = this.store.isProductPatternActive(product.id);
    const canMove = this.store.canSetProductStackLevel(product.id);

    item.classList.toggle('-is-selected', isSelected);
    item.classList.toggle('-can-move', canMove);
    item.classList.toggle('-has-selected', this.selectedDatasetsByProduct.has(product));
    item.opacityGauge.setProps({value: Math.round(product.opacity * 100)});
    filterElement.hidden = !patternFilterActive;
    if (filterToggle.checked !== patternFilterActive) {
      filterToggle.checked = patternFilterActive;
    }
    if (
      filterField.value !== patternFilter &&
      filterField.checkValidity() &&
      !filterField.classList.contains('no-reset')
    ) {
      filterField.value = this.store.getProductPattern(product.id);
      filterField.classList.remove('has-changes');
    }
    filterField.classList.remove('has-changes');
    if (checkbox.checked !== isSelected) {
      checkbox.checked = isSelected;
    }

    let datasetFiltersElement = item.querySelector('.js-dataset-filters');
    if (product.datasetFilters?.length > 0) {
      if (!datasetFiltersElement) {
        datasetFiltersElement = item.insertBefore(
          this._renderDatasetFiltersGroups(product.datasetFilters),
          filterElement
        );
      }
      this._updateProductDatasetFilters(datasetFiltersElement, product);
    } else if (datasetFiltersElement) {
      datasetFiltersElement.remove();
    }
  }
  /**
   * @protected
   * @param {HTMLElement} element
   * @param {Product} product
   * @returns {void}
   */
  _updateProductDatasetFilters(element, product) {
    element.classList.toggle('-single-group', product.datasetFilters.length === 1);

    const groupElementsByGroup = new Map(
      [...element.querySelectorAll('.js-dataset-filters-group')]
      .map((e) => [e.querySelector('.js-dataset-filters-group-label').textContent, e])
    );
    /** @type {HTMLElement[]} */
    const groupElements = [element.firstChild];
    for (const group of product.datasetFilters) {
      let groupElement = groupElementsByGroup.get(group.label);
      if (!groupElement) {
        groupElement = this._renderDatasetFiltersGroup(group);
      }
      this._updateProductDatasetFiltersGroup(groupElement, product, group);
      groupElements.push(groupElement);
    }
    setNodeChildren(element, groupElements);
  }
  /**
   * @protected
   * @param {HTMLElement} element
   * @param {Product} product
   * @param {DatasetFiltersGroup} group
   * @returns {void}
   */
  _updateProductDatasetFiltersGroup(element, product, group) {
    const selectedFilterIds = new Set();
    const selectedDatasets = this.selectedDatasetsByProduct.get(product);
    if (selectedDatasets) {
      for (const {id, pattern} of group.filters) {
        for (const dataset of selectedDatasets) {
          if (dataset.matchesPattern(Pattern.toHybrid(pattern))) {
            selectedFilterIds.add(id);
            break;
          }
        }
      }
    }

    const groupLabel = element.querySelector('.js-dataset-filters-group-label');
    groupLabel.title = group.description ?? '';

    const datasetFilterElementsByValue = new Map(
      [...element.querySelectorAll('.js-dataset-filter')]
      .map((e) => [e.querySelector('.js-dataset-filter-checkbox').value, e])
    );
    /** @type {HTMLElement[]} */
    const filterElements = [element.firstChild];
    for (const filter of group.filters) {
      const {id, label, aliases, description, pattern} = filter;
      let filterElement = datasetFilterElementsByValue.get(id);
      if (!filterElement) {
        filterElement = this._renderDatasetFilter(filter);
      }

      filterElement.title = (
        `${group.label}: ${label}` +
        (description ? `\n${description}` : '') +
        `\n---\nPattern: ${pattern}\nName: ${id}` +
        (aliases ? `\nAliases: ${aliases.join(', ')}` : '')
      );

      const filterLabel = (
        filterElement.querySelector('.js-dataset-filter-label')
      );
      const filterCheckbox = (
        filterElement.querySelector('.js-dataset-filter-checkbox')
      );

      filterElement.classList.toggle('-has-selected', selectedFilterIds.has(id));
      filterCheckbox.checked = (
        this.store.isProductDatasetFilterActive(product.id, id)
      );
      filterLabel.textContent = label;
      filterElements.push(filterElement);
    }
    setNodeChildren(element, filterElements);
  }
  _updateItemForProductWithId(productId) {
    this._updateItem(
      this._getItemForProductWithId(productId),
      this.store.getProduct(productId)
    );
  }
  _updateColormapsForProductWithId(productId) {
    const product = this.store.getProduct(productId);

    if (product.colormaps) {
      const item = this._getItemForProductWithId(productId);
      if (item) {
        const colormapsNode = item.querySelector('.js-colormaps');
        colormapsNode.parentNode.replaceChild(
          this._renderColormaps(product.colormaps),
          colormapsNode
        );
      }
    }
  }

  _checkFilterFieldValidity(filterField) {
    const pattern = filterField.value;
    let message;
    try {
      Pattern.parse(pattern, true);
      message = '';
    } catch (error) {
      message = (
        error.pattern[error.position] === ','
        ? error.issue
        : `${error.issue} at position ${error.position + 1}`
      );
    }
    filterField.setCustomValidity(message);
  }
  _updateFilterFieldState(productId) {
    const patternFilter = this.store.getProductPattern(productId) || '';
    const filterField = this._getElementForProductWithId(productId, '.js-filter-input');
    const hasChanges = (filterField.value !== patternFilter);

    filterField.classList.toggle('has-changes', hasChanges);
    filterField.classList.add('no-reset');
    this._checkFilterFieldValidity(filterField);
  }
  _commitFilterField(productId) {
    const patternFilter = this.store.getProductPattern(productId) || '';
    const filterField = this._getElementForProductWithId(productId, '.js-filter-input');
    const value = filterField.value;
    const hasChanges = (value !== patternFilter);

    if (hasChanges) {
      filterField.classList.toggle('has-changes', hasChanges);
      filterField.classList.remove('no-reset');
      this._checkFilterFieldValidity(filterField);
      if (filterField.reportValidity()) {
        this.hub.emit('change:filterPattern', productId, value);
      }
    }
  }

  _getProductIdForElement(element) {
    const item = element?.closest('.js-product');
    const productId = item?.dataset.productId;
    return productId;
  }

  handleProductSelectionChange(checkbox) {
    const productId = this._getProductIdForElement(checkbox);
    const selected = checkbox.checked;

    if (Syntool.keyboardManager.isShiftDown()) {
      window.getSelection().removeAllRanges();
      this.hub.emit('select-one-product', productId);
    } else if (selected) {
      this.hub.emit('select-product', productId);
    } else {
      this.hub.emit('deselect-product', productId);
    }

    // Restore the checkbox's original state
    this._updateItemForProductWithId(productId);
  }
  handleSelectOneProduct(productId) {
    this.hub.emit('select-one-product', productId);
  }

  handleProductOpacityChange(productId, opacity) {
    this.hub.emit('change:opacity', productId, opacity / 100);
  }
  handleProductOpacityClick(productId) {
    const item = this._getItemForProductWithId(productId);
    if (item.opacityGauge.props.withSlider) {
      item.opacityGauge.setProps({withSlider: false});
    } else {
      this._getAllItems().forEach(function(i) {
        i.opacityGauge.setProps({withSlider: i.dataset.productId === productId});
      });
    }
  }

  handleProductDatasetFilterActiveChange(checkbox) {
    const productId = this._getProductIdForElement(checkbox);
    const group = (
      checkbox
      .closest('.js-dataset-filters-group')
      .querySelector('.js-dataset-filters-group-label')
      .textContent
    );
    const {value: filter, checked} = checkbox;

    if (Syntool.keyboardManager.isShiftDown()) {
      window.getSelection().removeAllRanges();
      this.hub.emit('select-one:dataset-filter', productId, group, filter);
    } else if (checked) {
      this.hub.emit('select:dataset-filter', productId, filter);
    } else {
      this.hub.emit('deselect:dataset-filter', productId, filter);
    }

    // Restore the checkbox's original state
    this._updateItemForProductWithId(productId);
  }
  handleActivateOneProductDatasetFilter(productId, group, filter) {
    this.hub.emit('select-one:dataset-filter', productId, group, filter);
  }

  handleReorder(fromIndex, toIndex) {
    if (fromIndex === toIndex) {
      return;
    }

    const direction = fromIndex < toIndex ? 1 : -1;
    const items = this.node.children;
    const reorderedItem = items[fromIndex];
    const reorderedProductId = this._getProductIdForElement(reorderedItem);
    const stackLevelsByProduct = {
      [reorderedProductId]: this.store.getProduct(reorderedProductId).stackLevel,
    };

    let i = fromIndex;
    while (i !== toIndex) {
      i += direction;

      const productId = this._getProductIdForElement(items[i]);

      stackLevelsByProduct[productId] = stackLevelsByProduct[reorderedProductId];
      stackLevelsByProduct[reorderedProductId] = (
        this.store.getProduct(productId).stackLevel
      );
    }

    // Save focus
    const focusedElement = this.node.querySelector(':focus');

    // Move reorderedItem to the correct position
    if (toIndex === this.node.length - 1) {
      this.node.appendChild(reorderedItem);
    } else {
      this.node.insertBefore(
        reorderedItem,
        items[direction > 0 ? toIndex + 1 : toIndex]
      );
    }

    // Restore focus
    if (this.node.querySelector(':focus') !== focusedElement) {
      focusedElement.focus();
    }
    this.hub.emit('change:stackLevels', stackLevelsByProduct);
  }
}
window.ODL = Object.assign(window.ODL || {}, {ProductsSettingsListView});
