/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  HIDModifiers,
  HIDKeys,
  getNowMs,
  lerp,
  inverseLerp,
  Syntool,
} from '../syntool.js';

function easing(t) {
  // easeInOutQuad
  return t < 0.5 ? 2 * t * t : 1 - 2 * (1 - t) * (1 - t);
}
function animate(elements, keyframes, options) {
  var {duration, beforeStart, cancel, complete} = options;
  if (!Array.isArray(elements)) {
    if (elements.length >= 0) {
      elements = [...elements];
    } else {
      elements = [elements];
    }
  }

  var properties = Object.keys(keyframes);
  var from = [];
  var to = [];
  var offset = [];

  elements.forEach(function(element) {
    var current = window.getComputedStyle(element);
    var f = {};
    var t = {};
    var o = {};
    properties.forEach(function(key) {
      var fv;
      var tv;
      if (Array.isArray(keyframes[key])) {
        fv = keyframes[key][0];
        tv = keyframes[key][1];
      } else {
        fv = current[key];
        tv = keyframes[key];
      }
      f[key] = parseFloat(fv);
      t[key] = parseFloat(tv);
      o[key] = inverseLerp(fv, tv, parseFloat(current[key]));
    });
    from.push(f);
    to.push(t);
    offset.push(o);

    if (beforeStart) {
      beforeStart(element);
    }
  });

  var done = false;
  var startTime = getNowMs();
  var rafId = window.requestAnimationFrame(loop);

  return function() {
    window.cancelAnimationFrame(rafId);
    if (cancel) {
      elements.forEach(cancel);
    }
  };

  function loop(timestamp) {
    var t = (timestamp - startTime) / duration;
    if (t >= 1) {
      t = 1;
      done = true;
    }

    for (var i = 0; i < properties.length; i++) {
      var property = properties[i];
      for (var j = 0; j < elements.length; j++) {
        elements[j].style[property] = lerp(
          from[j][property],
          to[j][property],
          easing(t + offset[j][property])
        );
      }
    }

    if (done) {
      if (complete) {
        elements.forEach(complete);
      }
    } else {
      rafId = window.requestAnimationFrame(loop);
    }
  }
}

Syntool.extend({
  [Syntool.NAME]: 'ScreenshotMode',
  [Syntool.CONSTRUCTOR]() {
    this._screenshotModeActive = false;
  },
  [Syntool.INIT]() {
    var syntool = this;
    var hiddenElementProps = new WeakMap();
    var cancelHideInScreenshotModeAnimation = null;

    function affectsUsableViewportLayout(element) {
      return (
        element.parentNode === syntool.uiOverlayNode &&
        ['absolute', 'fixed'].indexOf(window.getComputedStyle(element).position) === -1
      );
    }
    function beforeHide(element) {
      hiddenElementProps.set(element, {
        affectsUsableViewportLayout: affectsUsableViewportLayout(element),
        display: element.style.display,
      });
    }
    function beforeShow(element) {
      var props = hiddenElementProps.get(element);
      if (props) {
        element.style.display = props.display;
        if (props.affectsUsableViewportLayout) {
          syntool.updateUsableViewportSize();
        }
      }
    }
    function hideComplete(element) {
      var props = hiddenElementProps.get(element);
      if (props) {
        element.style.display = 'none';
        if (props.affectsUsableViewportLayout) {
          syntool.updateUsableViewportSize();
        }
      }
    }
    function showComplete(element) {
      var props = hiddenElementProps.get(element);
      if (props) {
        hiddenElementProps.delete(element);
        element.style.display = props.display;
        element.style.opacity = '';
        if (props.affectsUsableViewportLayout) {
          syntool.updateUsableViewportSize();
        }
      }
    }

    this.hub.on('change:screenshot-mode', () => {
      const isActive = this.screenshotModeActive;

      this.node.classList.toggle('-screenshot-mode', isActive);

      if (cancelHideInScreenshotModeAnimation) {
        cancelHideInScreenshotModeAnimation();
      }
      cancelHideInScreenshotModeAnimation = animate(
        this.node.querySelectorAll('.h-hidden-in-screenshot-mode'),
        {opacity: isActive ? [1, 0] : [0, 1]},
        {
          duration   : 400,
          beforeStart: isActive ? beforeHide : beforeShow,
          complete   : isActive ? hideComplete : showComplete,
        }
      );
    });

    this.shortcutsStore.add('Global', {
      id         : 'toggle-screenshot-mode',
      modifiers  : [HIDModifiers.shift],
      keys       : HIDKeys.F,
      description: 'Toggle the visibility of the user interface\'s panels',
      context    : this,
      handler    : function() {
        this.screenshotModeActive = !this.screenshotModeActive;
        return true;
      },
    });
  },
  [Syntool.STATE]: {
    get: ({screenshotModeActive}) => ({screenshotModeActive}),
    set({screenshotModeActive}) {
      this.screenshotModeActive = screenshotModeActive;
    },
  },
  get screenshotModeActive() {
    return this._screenshotModeActive;
  },
  set screenshotModeActive(active) {
    if (this.screenshotModeActive === !active) {
      this._screenshotModeActive = !!active;
      this.hub.emit('change:screenshot-mode');
      this._invalidateState();
    }
  },
});
