/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  isIterable,
  asArray,
  formatUTCDate,
  H,
  objectToDataAttributes,
  EventsHub,
  DOMEvents,
  Quantity,
} from '../../syntool.js';
import {isModifiedEvent, linkStyleSheet} from '../../utils/dom_utils.js';
import {opensInNewTabIcon} from '../utils.js';

linkStyleSheet('css/dataset_metadata_view.css');

const ROW_RENDER = Symbol('ROW_RENDER');

function updateElementChildren(node, children) {
  const childrenSet = (
    children instanceof Set
    ? children
    : new Set(children)
  );
  if (childrenSet.size === 0) {
    node.innerHTML = '';
    return;
  }

  for (const child of [...node.children]) {
    if (!childrenSet.has(child)) {
      node.removeChild(child);
    }
  }

  let index = 0;
  for (const child of childrenSet) {
    if (node.children[index] !== child) {
      node.insertBefore(child, node.children[index]);
    }
    index++;
  }
}
function updateElementAttributes(node, props) {
  for (const attributeName of node.getAttributeNames()) {
    if (!props || !Object.hasOwn(props, attributeName)) {
      node.removeAttribute(attributeName);
    }
  }
  return H(node, props);
}
function updateElement(node, props, children) {
  if (
    arguments.length === 2 &&
    (Array.isArray(props) || props instanceof Node || typeof props !== 'object')
  ) {
    children = props;
    props = null;
  } else {
    for (const attributeName of node.getAttributeNames()) {
      if (!props || !Object.hasOwn(props, attributeName)) {
        node.removeAttribute(attributeName);
      }
    }
  }

  if (
    children instanceof Node &&
    (node.childElementCount > 1 || node.firstElementChild !== children)
  ) {
    node.innerHTML = '';
  }

  if (children == null) {
    children = '';
  } else if (typeof children !== 'string' && isIterable(children)) {
    const childrenSet = new Set();
    for (const child of children) {
      if (!child) {
        continue;
      }
      if (!(child instanceof Node)) {
        throw new TypeError(
          'children list must only contain optional Node instances'
        );
      }
      childrenSet.add(child);
    }

    updateElementChildren(node, childrenSet);
    children = null;
  }

  return H(node, props, children);
}

export class DatasetMetadataView {
  constructor() {
    const node = H('div', {class: 'DatasetMetadataView'});

    this.syntool = null;
    this.node = node;
    this.hub = new EventsHub();
    this._collapsedBlocks = new Map();
    this.dataset = null;
    this._granuleInfo = null;
    this.isEmpty = true;
    this.shouldNotify = false;

    this.renderGranuleInfoResourceRowByType = {
      link: ({url, description, label, order, ...rest}) => (
        this.renderRow(
          objectToDataAttributes(Object.assign({type: 'link'}, rest)),
          this.renderLink({
            href  : url,
            title : description,
            target: '_blank',
            rel   : 'noopener',
          }, `${label} ${opensInNewTabIcon}`)
        )
      ),
      spectrum: ({id, description, label, order, ...rest}) => (
        this.renderRow(
          objectToDataAttributes(Object.assign({type: 'button'}, rest)),
          H('button', {
            type     : 'button',
            title    : description,
            class    : 'js-resource',
            'data-id': id,
          }, 'Show ' + label)
        )
      ),
      image: ({id, url, description, label, inline, order, ...rest}) => {
        const linkProps = {
          href     : url,
          title    : description,
          target   : '_blank',
          rel      : 'noopener',
          class    : 'js-resource',
          'data-id': id,
        };
        let type;
        let children;
        if (inline) {
          if (!linkProps.title && label) {
            linkProps.title = label;
          }
          type = 'image';
          children = H('img', {
            class: 'DatasetMetadataView-Image',
            src  : typeof inline === 'string' ? inline : url,
          });
        } else {
          type = 'link';
          children = 'Show ' + label;
        }
        return this.renderRow(
          objectToDataAttributes(Object.assign({type}, rest)),
          this.renderLink(linkProps, children)
        );
      },
    };
    this.createGranuleInfoResourceRowByType = {
      image(row, {config}) {
        row.inline = config.inline;
        return row;
      },
    };

    this.blocksRenderers = [
      () => this.renderInfoBlock(this.dataset),
      () => this.renderMetricsBlock(this.dataset),
      () => this.renderGranuleInfoBlocks(this._granuleInfo),
    ];

    this.handleBlockToggle = this.handleBlockToggle.bind(this);

    DOMEvents(node)
    .on('click', '.js-resource', (event) => {
      let target = event.target;

      while (target !== this.node && !target.hasAttribute('data-id')) {
        target = target.parentNode;
      }
      if (target === this.node) {
        return;
      }
      if (
        target instanceof HTMLAnchorElement &&
        (
          event.button !== 0 || // ignore right clicks
          isModifiedEvent(event) // ignore clicks with modifier keys
        )
      ) {
        return;
      }

      event.preventDefault();

      const configuredResource = (
        this._granuleInfo.getConfiguredResourceWithId(target.dataset.id)
      );

      if (configuredResource) {
        this.hub.emit('click:resource', configuredResource);
      }
    })
    .on('click', '.js-copy', (event) => {
      this.hub.emit(
        'click:copy',
        event.target.dataset.label,
        event.target.dataset.value
      );
    })
    .on('click', '.js-filter', (event) => {
      this.hub.emit(
        'click:filter',
        event.target.dataset.filter,
        event.target.dataset.value
      );
    });

    this.initialize();
  }
  initialize() {
    /* empty */
  }
  setDatasets(datasets) {
    const prevDataset = this.dataset;

    this.dataset = (datasets.length === 1 ? datasets[0] : null);
    this._granuleInfo = null;

    (async () => {
      const dataset = this.dataset;
      if (!dataset) {
        return;
      }
      let metadata;
      try {
        metadata = await this.dataset.getMetadata?.();
      } catch (_) {
        // error loading metadata!
        return;
      }
      if (this.dataset !== dataset || !metadata) {
        // dataset changed or has no metadata!
        return;
      }

      this._granuleInfo = metadata;
      this.render();

      const autoResourceId = metadata.autoResourceId;
      if (prevDataset !== dataset && autoResourceId) {
        // Immediately show resource
        this.hub.emit(
          'click:resource',
          metadata.getConfiguredResourceWithId(autoResourceId)
        );
      }
    })();
  }
  render() {
    const node = this.node;

    // Save scroll offset
    const scrollTop = node.scrollTop;
    // Save focus
    const focusedElement = node.querySelector(':focus');

    this.renderBlocks();

    // Restore focus
    if (node.querySelector(':focus') !== focusedElement) {
      focusedElement.focus();
    }
    // Restore scroll offset
    node.scrollTop = scrollTop;

    this.isEmpty = !node.hasChildNodes();
    this.shouldNotify = node.querySelector('[data-has-new]');
    this.hub.emit('render');
  }

  handleBlockToggle(event) {
    const target = event.target;
    if (!target.open) {
      const frameRect = this.node.getBoundingClientRect();
      const elementRect = target.getBoundingClientRect();
      if (elementRect.bottom <= frameRect.top) {
        target.scrollIntoView(true);
      } else if (elementRect.top >= frameRect.bottom) {
        target.scrollIntoView(false);
      }
    }

    const id = target.dataset.id;
    if (id) {
      this._collapsedBlocks.set(id, !target.open);
      this.hub.emit('change:collapsedBlocks');
    }
  }

  get collapsedBlocks() {
    return new Map(this._collapsedBlocks);
  }
  set collapsedBlocks(collapsedBlocks) {
    let changed = false;
    for (const [blockId, collapsed] of collapsedBlocks) {
      if (this.collapsedBlocks.get(blockId) !== collapsed) {
        this._collapsedBlocks.set(blockId, collapsed);
        changed = true;
      }
    }
    if (changed) {
      this.render();
      this.hub.emit('change:collapsedBlocks');
    }
  }

  renderBlocks() {
    const node = this.node;

    const renderedBlocks = new Set();
    for (const blocksRenderer of this.blocksRenderers) {
      for (const block of asArray(blocksRenderer(this))) {
        if (block instanceof Node) {
          renderedBlocks.add(block);
        } else if (block) {
          throw new TypeError('expected block to be an instance of Node');
        }
      }
    }
    updateElementChildren(node, renderedBlocks);
  }
  renderBlock({label, order, defaultCollapsed, children, ...dataAttributes}) {
    if (!children || children.length === 0) {
      return null;
    }

    const {id} = dataAttributes;
    dataAttributes = objectToDataAttributes(dataAttributes);

    const open = !(this._collapsedBlocks.get(id) ?? defaultCollapsed);
    const attributes = Object.assign(dataAttributes, {
      class: 'DatasetMetadataView-Block',
      open,
    });

    const existingBlock = this.node.querySelector(`details[data-id="${id}"]`);
    if (existingBlock) {
      updateElementAttributes(existingBlock, attributes);
      existingBlock.children[0].innerHTML = label;
      updateElement(existingBlock.children[1], children);
      return existingBlock;
    }

    const block = H('details', attributes, [
      H('summary', {class: 'DatasetMetadataView-BlockTitle'}, label),
      H('ul', {class: 'DatasetMetadataView-BlockRows'}, children),
    ]);

    block.ontoggle = this.handleBlockToggle;

    return block;
  }
  renderRow(props, children) {
    if (!children) {
      children = props;
      props = {};
    }
    return H(
      'li',
      Object.assign({class: 'DatasetMetadataView-BlockRow'}, props),
      children
    );
  }
  renderLink(props, children) {
    const ftpWarning = (
      props.href?.toLowerCase().startsWith('ftp://')
      ? (
        '\n\n⚠️ Modern browsers do not support FTP links anymore: use a ' +
        'FTP client such as lftp or Filezilla to open FTP links.'
      )
      : ''
    );
    if (ftpWarning) {
      props = Object.assign({}, props, {
        title: (props.title ?? '') + ftpWarning
      });
    }
    return H('a', props, children);
  }

  renderInfoBlock(dataset, overwrites) {
    if (!dataset) {
      return null;
    }

    const hasStart = dataset.start > DATE_MIN_TIME;
    const hasEnd = dataset.end < DATE_MAX_TIME;
    return this.renderBlock(Object.assign({
      id: '_info',
      label: 'Info',
      children: [
        {
          label: 'ID',
          value: dataset.name,
          copyable: true,
          filterable: 'datasetId',
        },
        {
          label: 'Product',
          value: dataset.getProduct().label,
          description: dataset.productId,
          copyable: true,
        },
        hasStart && {
          label: 'From',
          value: formatUTCDate(dataset.start),
          copyable: true,
          alignValue: 'right',
        },
        hasEnd && {
          label: 'Till',
          value: formatUTCDate(dataset.end),
          copyable: true,
          alignValue: 'right',
        },
      ].filter(Boolean).map(this.renderGranuleInfoMetadataRow, this),
    }, overwrites));
  }
  renderMetricsBlock(dataset, overwrites) {
    if (!dataset) {
      return null;
    }

    const rows = [];

    const length = new Quantity(dataset.getLength ? dataset.getLength() : 0, 'km');
    if (length > 0) {
      if (length < 1) {
        const lengthM = length.convert('m');
        rows.push({
          label: 'Length',
          value: lengthM.toString({exp: 0}),
          description: lengthM.toString(),
          copyable: true,
        });
      } else {
        rows.push({
          label: 'Length',
          value: length.toString({exp: -2}),
          description: length.toString(),
          copyable: true,
        });
      }
      const lengthNM = length.convert('NM');
      if (lengthNM < 1) {
        const lengthFt = length.convert('ft');
        rows.push({
          label: 'Length',
          value: lengthFt.toString({exp: 0}),
          description: lengthFt.toString(),
          copyable: true,
        });
      } else {
        rows.push({
          label: 'Length',
          value: lengthNM.toString({exp: -2}),
          description: lengthNM.toString(),
          copyable: true,
        });
      }
    }

    const area = new Quantity(dataset.getArea ? dataset.getArea() : 0, 'km2');
    if (area > 0) {
      if (area < 1) {
        const areaM = area.convert('m2');
        rows.push({
          label: 'Area',
          value: areaM.toString({exp: 0}),
          description: areaM.toString(),
          copyable: true,
        });
      } else {
        rows.push({
          label: 'Area',
          value: area.toString({exp: -2}),
          description: area.toString(),
          copyable: true,
        });
      }
      const areaNM = area.convert('NM2');
      if (areaNM < 1) {
        const areaFt = area.convert('ft2');
        rows.push({
          label: 'Area',
          value: areaFt.toString({exp: 0}),
          description: areaFt.toString(),
          copyable: true,
        });
      } else {
        rows.push({
          label: 'Area',
          value: areaNM.toString({exp: -2}),
          description: areaNM.toString(),
          copyable: true,
        });
      }
    }

    return this.renderBlock(Object.assign({
      id: '_metrics',
      label: 'Metrics',
      children: rows.map(this.renderGranuleInfoMetadataRow, this),
    }, overwrites));
  }
  renderGranuleInfoBlocks(granuleInfo) {
    return (
      granuleInfo
      ? (
        Array.from(
          granuleInfo.configuredGroups(),
          this.blockForGranuleInfoGroup,
          this
        )
        .sort(this.compareGranuleInfoItems, this)
        .map(this.renderBlock, this)
      )
      : null
    );
  }
  renderGranuleInfoMetadataRow({description, label, value, order, copyable, filterable, ...rest}) {
    const props = Object.assign(
      {title: description},
      objectToDataAttributes(rest),
    );
    return this.renderRow(props, [
      H('span', {class: 'DatasetMetadataView-Label'}, label),
      H('span', {class: 'DatasetMetadataView-Value'}, value),
      H('div', {class: 'DatasetMetadataView-Buttons'}, [
        filterable && H('button', {
          type: 'button',
          class: 'DatasetMetadataView-Button -Filter js-filter',
          title: `Filter using ${label}`,
          'data-filter': filterable,
          'data-value': value,
        }, [
          H('i', {class: 'fa fa-filter', 'aria-hidden': 'true'}),
          H('span', {class: 'sr-only'}, `Filter using ${label}`),
        ]),
        copyable && H('button', {
          type: 'button',
          class: 'DatasetMetadataView-Button -Copy js-copy',
          title: `Copy ${label} to clipboard`,
          'data-label': label,
          'data-value': value,
        }, [
          H('i', {class: 'fa fa-clipboard', 'aria-hidden': 'true'}),
          H('span', {class: 'sr-only'}, `Copy ${label} to clipboard`),
        ]),
      ]),
    ]);
  }
  renderGranuleInfoRow(row) {
    return (
      row[ROW_RENDER]
      ? row[ROW_RENDER](row)
      : this.renderGranuleInfoMetadataRow(row)
    );
  }

  compareGranuleInfoItems(item1, item2) {
    let result = item1.order - item2.order;
    if (result === 0) {
      result = (
        item1.label.localeCompare(item2.label, undefined, {sensitivity: 'base'})
      );
    }
    if (result === 0) {
      if (item1.id < item2.id) {
        result = -1;
      } else if (item1.id === item2.id) {
        result = 0;
      } else {
        result = 1;
      }
    }
    return result;
  }

  blockForGranuleInfoGroup([groupId, group]) {
    const rows = [];

    for (const [key, configuredMetadataField] of group.configuredMetadataFields()) {
      const row = this.blockRowForGranuleInfoMetadataField(
        [`${groupId}/metadata/${key}`, configuredMetadataField]
      );
      if (row) {
        rows.push(row);
      }
    }
    for (const [resourceId, configuredResource] of group.configuredResources()) {
      const createRow = (
        this.createGranuleInfoResourceRowByType[configuredResource.config.type]
      );
      const renderRow = (
        this.renderGranuleInfoResourceRowByType[configuredResource.config.type]
      );
      let row = this.blockRowForGranuleInfoResource(
        [`${groupId}/${resourceId}`, configuredResource]
      );
      if (row) {
        if (createRow) {
          row = createRow(row, configuredResource);
        }
        row[ROW_RENDER] = renderRow;
        rows.push(row);
      }
    }

    return {
      id: groupId,
      label: group.config.title,
      order: group.config.order ?? 0,
      defaultCollapsed: group.config.defaultCollapsed,
      children: (
        rows
        .sort(this.compareGranuleInfoItems, this)
        .map(this.renderGranuleInfoRow, this)
      ),
    };
  }
  blockRowForGranuleInfoMetadataField([id, {value, config}]) {
    return {
      id,
      label: config.label,
      description: config.description,
      value: config.value ?? value,
      order: config.order ?? 0,
      copyable: config.copyable ?? true,
      hasNew: true,
    };
  }
  blockRowForGranuleInfoResource([id, {resource, config}]) {
    return {
      id,
      label: config.label,
      description: resource.description,
      url: resource.url,
      order: config.order ?? -1,
      hasNew: true,
    };
  }
}
window.ODL = Object.assign(window.ODL || {}, {DatasetMetadataView});
