/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {H, EventsHub, throttleOncePerAnimationFrame} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';

linkStyleSheet('css/selection_info_view.css');

export class SelectionInfoView {
  constructor() {
    this.node = H('div', {class: 'SelectionInfoView'}, [
      H('div', {class: 'SelectionInfoView-Title'}, 'Selection info'),
      this._wrapperNode = H('div', {class: 'SelectionInfoView-Wrapper'}),
      this._overlayNode = H('div', {class: 'SelectionInfoView-Overlay'}),
    ]);
    this.children = [];
    this.datasets = [];
    this.isEmpty = true;
    this.shouldNotify = false;
    this.hub = new EventsHub();
    this._ignoreChildRender = false;

    this.render = throttleOncePerAnimationFrame(this.render);
  }

  appendChild(child) {
    this._wrapperNode.appendChild(child.node);
    this.children.push(child);
    child.hub?.off('render', this.handleChildRender, this);
    child.hub?.on('render', this.handleChildRender, this);
  }
  appendOverlayChild(node) {
    this._overlayNode.appendChild(node);
  }
  removeChild(child) {
    child.hub?.off('render', this.handleChildRender, this);
    this._wrapperNode.removeChild(child.node);
    const index = this.children.indexOf(child);
    if (index !== -1) {
      this.children.splice(index, 1);
    }
  }
  removeOverlayChild(node) {
    this._overlayNode.removeChild(node);
  }

  setDatasets(datasets) {
    const oldDatasets = this.datasets;
    for (const dataset of oldDatasets) {
      dataset.hub.off('change', this.render, this);
      dataset.hub.off('changing', this.render, this);
    }
    this.hub.emit('will-change:datasets', oldDatasets, datasets);
    this.datasets = datasets;
    for (const dataset of this.datasets) {
      dataset.hub.on('change', this.render, this);
      dataset.hub.on('changing', this.render, this);
    }
    for (const child of this.children) {
      child.setDatasets(datasets);
    }
    this.hub.emit('change:datasets', oldDatasets, datasets);
    this.render();
  }

  handleChildRender() {
    if (this._ignoreChildRender) {
      return;
    }

    let isEmpty = true;
    let shouldNotify = false;
    for (const child of this.children) {
      isEmpty &&= child.isEmpty;
      shouldNotify ||= !child.isEmpty && child.shouldNotify;
    }
    this.node.classList.toggle(
      '-has-scrollbar',
      this.node.offsetWidth > this.node.clientWidth /* + borders*/
    );
    this.isEmpty = isEmpty;
    this.shouldNotify = shouldNotify;
    this.hub.emit('render');
  }

  render() {
    this._ignoreChildRender = true;
    for (const child of this.children) {
      child.render();
    }
    this._ignoreChildRender = false;
    this.handleChildRender();
  }
}
window.ODL = Object.assign(window.ODL || {}, {SelectionInfoView});
