/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  H,
  EventsHub,
  DOMEvents,
  preventDefaultAndStopEvent,
} from '../../syntool.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import '../../datasets/shape_dataset/syntool_filters.js';

linkStyleSheet('css/shapes_export_sheet.css');

function renderExportAction(format, exporter) {
  return H('div', {class: 'ShapesExportSheet-Action', 'data-format': format}, [
    H('button', {
      type: 'button',
      class: 'ShapesExportSheet-ActionButton --File js-export-to-file',
    }, [
      H('i', {class: 'fa fa-file', 'aria-hidden': 'true'}),
      ` ${exporter.label ?? format}`,
      H('span', {class: 'ShapesExportSheet-ActionDescription'}),
    ]),
    H('button', {
      type: 'button',
      class: 'ShapesExportSheet-ActionButton --Copy js-export-to-clipboard',
      title: 'Copy to Clipboard',
    }, [
      H('i', {class: 'fa fa-clipboard', 'aria-hidden': 'true'}),
      H('span', {class: 'sr-only'}, 'Copy to Clipboard'),
    ]),
  ]);
}
function renderExportSheet() {
  return H('div', {class: 'ShapesExportSheet', 'aria-hidden': 'true'},
    H('div', {class: 'ShapesExportSheet-Wrapper'}, [
      H('div', {class: 'ShapesExportSheet-Header'}, [
        H('span', {class: 'ShapesExportSheet-Title'}),
        H('button', {type: 'button', class: 'ShapesExportSheet-Cancel'}, 'Cancel'),
      ]),
      H('div', {class: 'ShapesExportSheet-Actions'})
    ])
  );
}

export class ShapesExportSheet {
  constructor() {
    this.node = renderExportSheet();
    this._exportActions = new Map();
    this.hub = new EventsHub();

    DOMEvents(this.node)
    .on('click', (event) => {
      preventDefaultAndStopEvent(event);
      this.hub.emit('cancel');
    })
    .on('click', '.ShapesExportSheet-Header', (event) => {
      preventDefaultAndStopEvent(event);
    })
    .on('click', '.ShapesExportSheet-Header button', (event) => {
      preventDefaultAndStopEvent(event);
      this.hub.emit('cancel');
    })
    .on('click', '.js-export-to-file', (event) => {
      preventDefaultAndStopEvent(event);
      const format = event.target.parentElement.dataset.format;
      this.hub.emit('select', 'file', format);
    })
    .on('click', '.js-export-to-clipboard', (event) => {
      preventDefaultAndStopEvent(event);
      const format = event.target.parentElement.dataset.format;
      this.hub.emit('select', 'clipboard', format);
    });
  }

  show(syntool, datasets) {
    if (!datasets || datasets.length === 0) {
      return false;
    }

    const node = this.node;
    const exportActions = this._exportActions;
    const shapesCount = datasets.length;

    const title = node.querySelector('.ShapesExportSheet-Title');
    title.textContent = `Export ${shapesCount > 1 ? shapesCount : ''} shape${shapesCount > 1 ? 's' : ''}`;

    const sheetWrapper = node.querySelector('.ShapesExportSheet-Actions');
    for (const [format, exporter] of syntool.shapesExporters) {
      let exportAction = exportActions.get(format);
      if (!exportAction) {
        exportAction = renderExportAction(format, exporter);
        exportActions.set(format, exportAction);
        sheetWrapper.appendChild(exportAction);
      }

      let unsupportedCount = 0;
      let unsupportedTypes = new Set();
      for (const dataset of datasets) {
        const shape = dataset.shape;
        if (!exporter.canExportShapeOfType(shape.type)) {
          unsupportedCount++;
          unsupportedTypes.add(shape.type);
        }
      }

      exportAction.classList.toggle(
        '-is-supported',
        unsupportedCount < shapesCount
      );

      const descriptionNode = (
        exportAction.querySelector('.ShapesExportSheet-ActionDescription')
      );

      if (unsupportedCount === 0) {
        descriptionNode.textContent = 'Supports all selected shapes';
      } else if (unsupportedCount < shapesCount) {
        const unsupportedTypesLabels = (
          Array.from(unsupportedTypes, (type) => `${type.toLowerCase()}s`)
          .sort()
        );
        const unsupportedTypesMessage = (
          unsupportedTypesLabels.length === 1
          ? unsupportedTypesLabels[0]
          : (
            unsupportedTypesLabels.slice(0, -1).join(', ') + ' or ' +
            unsupportedTypesLabels[unsupportedTypesLabels.length - 1]
          )
        );

        descriptionNode.textContent = `Doesn't support ${unsupportedTypesMessage}`;
      }
    }

    node.removeAttribute('aria-hidden');
    return true;
  }
  hide() {
    this.node.setAttribute('aria-hidden', 'true');
    return true;
  }
}
window.ODL = Object.assign(window.ODL || {}, {ShapesExportSheet});
