/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DATE_MIN_TIME,
  DATE_MAX_TIME,
  emptyFunctionReturningArgument,
  isValidDate,
  dateToTimestamp,
  formatUTCDate,
  createClass,
  round,
  clamp,
  H,
  HIDKeys,
  EventsHub,
  DOMEvents,
  preventDefaultAndStopEvent,
} from '../../syntool.js';
import {Shape} from '../../datasets/shape_dataset/shape.js';
import {linkStyleSheet} from '../../utils/dom_utils.js';
import {Gestures} from '../utils.js';
import '../../datasets/shape_dataset/syntool_filters.js';

linkStyleSheet('css/shapes_info_view.css');

const dateStringRe = (
  /^\d+-0*(?:[1-9]|1[012])-0*[123]?\d 0*\d{1,2}:0*\d{1,2}:0*\d{1,2}$/
);

function formatDate(date) {
  return (
    date.getTime() === DATE_MIN_TIME || date.getTime() === DATE_MAX_TIME
    ? ''
    : formatUTCDate(date, '$<date> $<time>')
  );
}
function parseDate(dateString) {
  if (dateStringRe.test(dateString)) {
    const date = new Date(dateString.replace(' ', 'T') + 'Z');
    return isValidDate(date) ? date : null;
  }
  return null;
}
function validateDateString(string) {
  return parseDate(string) !== null;
}

function idOfField(name) {
  return `ShapesInfoView-${name}`;
}
function renderField(name, label, child) {
  const id = idOfField(name);
  return H([
    H('label', {for: id, 'data-property': name}, label),
    H(child, {id, 'data-property': name}),
  ]);
}
function renderTextField(name, label, tooltip, placeholder) {
  return renderField(name, label,
    H('input', {type: 'text', title: tooltip, placeholder})
  );
}
function renderDateField(name, label, placeholder) {
  return renderTextField(name, label, 'YYYY-MM-DD hh:mm:ss', placeholder);
}
function renderSwitchField(name, label, choices) {
  return renderField(name, label,
    H('span', {class: 'Switch'}, Array.from(choices, renderSwitchChoice))
  );

  function renderSwitchChoice([value, choiceLabel]) {
    const id = `${idOfField(name)}-${value}`;
    return H('span', {class: 'Switch-Choice'}, [
      H('input', {type: 'radio', id, name, value}),
      H('label', {for: id}, choiceLabel),
    ]);
  }
}
function renderNumberField(name, label, min, max, step) {
  return renderField(name, label,
    H('span', {class: 'Numeric'}, [
      H('input', {type: 'range', min, max, step, tabindex: -1}),
      H('input', {type: 'number', min, max, step}),
    ])
  );
}
function renderButton(name, icon, label, tooltip) {
  return H('button', {type: 'button', id: idOfField(name), title: tooltip}, [
    icon ? H('i', {class: icon}) : null,
    label ? H('span', label) : null,
  ]);
}

function renderPanel() {
  return H('form', {class: 'ShapesInfoView'}, [
    renderTextField('author', 'Author'),
    renderTextField('text', 'Text'),
    renderDateField('start', 'From', 'the beginning of time'),
    renderDateField('end', 'Till', 'the end of time'),
    renderSwitchField('arrow', 'Arrow', [['on', 'ON'], ['off', 'OFF']]),
    renderSwitchField('type', 'Type', [['on', 'CCW'], ['off', 'CW']]),
    renderSwitchField('flipped', 'Hot', [['on', 'Right'], ['off', 'Left'], ['none', 'None']]),
    renderNumberField('size', 'Size', 1, 11, 1),
    renderNumberField('ratio', 'Ratio', 0, 1, 0.1),
    renderNumberField('density', 'Density', 0, 1, 0.1),
    renderNumberField('offset', 'Offset', 0, 1, 1),
    H('em', 'Available only in this browser'),
    H('div', {class: 'ShapesInfoView-Actions'}, [
      renderButton('duplicate', 'fa fa-copy', 'Duplicate'),
      renderButton('delete', 'fa fa-trash', 'Delete'),
      renderButton('export', 'fa fa-download', '', 'Export'),
    ]),
  ]);
}

function addTextFieldHandlers(onChange, node, name, getOriginalValue) {
  const selector = `#${idOfField(name)}`;
  DOMEvents(node)
  .on('focus', selector, function(event) {
    this.select();
  })
  .on('input', selector, function(event) {
    this.classList.toggle('-warning', getOriginalValue() !== this.value);
    this.classList.toggle('-error', this.validationMessage !== '');
  })
  .on('keypress', selector, function(event) {
    if (event.which === HIDKeys.enter.code) {
      preventDefaultAndStopEvent(event);
      this.value = this.value.trim();
      if (this.validationMessage === '') {
        onChange(name, this.value);
      } else {
        this.classList.add('-error');
      }
    }
  });
}
function addDateFieldHandlers(onChange, node, name, getOriginalValue) {
  const selector = `#${idOfField(name)}`;
  DOMEvents(node)
  .on('focus', selector, function(event) {
    if (this.value === '---') {
      this.select();
    }
  })
  .on('input', selector, function(event) {
    const value = this.value.trim();
    const valid = (value === '' || validateDateString(value));
    const originalValue = getOriginalValue();
    const changed = (originalValue && value !== formatDate(originalValue));

    this.classList.toggle('-warning', changed);
    this.classList.toggle('-error', !valid);
  })
  .on('keypress', selector, function(event) {
    if (event.which === HIDKeys.enter.code) {
      preventDefaultAndStopEvent(event);

      const value = this.value.trim();

      if (value === '') {
        onChange(name, undefined);
      } else {
        const date = parseDate(value);
        if (date) {
          onChange(name, date);
        }
      }
    }
  });
}
function addSwitchFieldHandlers(onChange, node, name) {
  DOMEvents(node)
  .on('change', `#${idOfField(name)} input`, function(event) {
    onChange(
      name,
      (
        this.value === 'on' ? true
        : this.value === 'off' ? false
        : this.value
      )
    );
  });
}
function addNumberFieldHandlers(onChange, node, name) {
  const selector = `#${idOfField(name)} input`;
  DOMEvents(node)
  .on('input', selector, function(event) {
    preventDefaultAndStopEvent(event);

    const min = parseFloat(this.min);
    const max = parseFloat(this.max) - min;
    const value = clamp(parseFloat(this.value) - min, 0, max);

    onChange(name, value / max);
  })
  .on('wheel', selector, Gestures.getWheelHandler({step: 1}, function(event, wheelDelta) {
    const min = parseFloat(this.min);
    const max = parseFloat(this.max) - min;
    const step = parseFloat(this.step);
    const delta = wheelDelta * step;
    const value = parseFloat(this.value) - min;
    const newValue = clamp(value + delta, 0, max);

    if (newValue !== value) {
      onChange(name, newValue / max);
    }
  }));
}
function addButtonHandlers(onClick, node, name) {
  DOMEvents(node)
  .on('click', `#${idOfField(name)}`, function(event) {
    preventDefaultAndStopEvent(event);
    onClick(name, event);
  });
}

function updateTextField(node, name, value, {readOnly, pattern = '', required}) {
  const field = node.querySelector(`#${idOfField(name)}`);

  if (field.pattern !== pattern) {
    if (pattern) {
      field.pattern = pattern;
    } else {
      field.removeAttribute('pattern');
    }
  }
  if (field.required !== required) {
    field.required = required;
  }
  if (field.readOnly !== readOnly) {
    field.readOnly = readOnly;
  }
  if (field.value !== value) {
    field.value = value;
  }
  field.classList.remove('-warning', '-error');
}
function updateDateField(node, name, value, props) {
  updateTextField(node, name, value ? formatDate(value) : '---', props);
}
function updateSwitchField(node, name, on, props) {
  const {readOnly} = props;
  const buttons = node.querySelectorAll(`#${idOfField(name)} input`);
  const selectedValue = (
    on === true ? 'on'
    : on === false ? 'off'
    : on
  );

  for (const button of buttons) {
    button.disabled = readOnly;
    button.checked = (button.value === selectedValue);

    const label = props[`${button.value}Label`];
    if (label) {
      button.nextElementSibling.innerHTML = label;
    }
  }
}
function updateNumberFieldInput(input, value, {readOnly, min, max}) {
  const currentMin = parseFloat(input.min);
  const currentMax = parseFloat(input.max);
  const currentValue = parseFloat(input.value);

  if (min == null) {
    min = currentMin;
  }
  if (max == null) {
    max = currentMax;
  }
  if (value == null) {
    value = currentValue;
  } else {
    value = value * (max - min) + min;
    value = round(value, -6);
  }

  if (readOnly != null && input.disabled !== readOnly) {
    input.disabled = readOnly;
  }
  if (currentMin !== min) {
    input.min = min;
  }
  if (currentMax !== max) {
    input.max = max;
  }
  if (currentValue !== value) {
    input.value = value;
  }
}
function updateNumberField(node, name, value, props) {
  for (const input of node.querySelectorAll(`#${idOfField(name)} input`)) {
    updateNumberFieldInput(input, value, props);
  }
}

function getTextFieldValue(node, name) {
  return node.querySelector(`#${idOfField(name)}`)?.value;
}
function getDateFieldValue(node, name) {
  const value = node.querySelector(`#${idOfField(name)}`)?.value.trim();

  return value ? parseDate(value) : undefined;
}

function isShapeDataset(dataset) {
  return dataset.syntool.filters.dataset.isShape(dataset);
}
function isUserShape(dataset) {
  return dataset.getProduct().type === 'USER_SHAPES';
}
function isReadOnly(dataset) {
  return dataset.readOnly;
}
function sameDate(d1, d2) {
  return dateToTimestamp(d1) === dateToTimestamp(d2);
}

export const ShapesInfoView = createClass([
  /*Public*/ {
    constructor: function ShapesInfoView() {
      const node = renderPanel();

      this.node = node;
      this.datasets = null;
      this.isEmpty = true;
      this.shouldNotify = false;
      this._rendering = false;
      this._state = null;
      this.minSize = 1;
      this.maxSize = 11;
      this.convertSize = emptyFunctionReturningArgument;

      this.hub = new EventsHub();

      // Event handlers
      // ==============
      const handleChange = this.handleChange.bind(this);
      const handleClick = this.handleClick.bind(this);

      addTextFieldHandlers(handleChange, node, 'author', this.getValue.bind(this, 'author'));
      addTextFieldHandlers(handleChange, node, 'text', this.getValue.bind(this, 'text'));
      addDateFieldHandlers(handleChange, node, 'start', this.getValue.bind(this, 'start'));
      addDateFieldHandlers(handleChange, node, 'end', this.getValue.bind(this, 'end'));
      addSwitchFieldHandlers(handleChange, node, 'arrow');
      addSwitchFieldHandlers(handleChange, node, 'type');
      addSwitchFieldHandlers(handleChange, node, 'flipped');
      addNumberFieldHandlers(handleChange, node, 'ratio');
      addNumberFieldHandlers(handleChange, node, 'density');
      addNumberFieldHandlers(handleChange, node, 'offset');
      addNumberFieldHandlers(handleChange, node, 'size');
      addButtonHandlers(handleClick, node, 'duplicate');
      addButtonHandlers(handleClick, node, 'delete');
      addButtonHandlers(handleClick, node, 'export');
    },

    handleChange(name, value) {
      if (!this._rendering) {
        this.resetField(name);
        this.hub.emit('change:' + name, value);
      }
    },
    handleClick(name, event) {
      this.hub.emit(name, event);
    },

    _updateState() {
      if (
        !this.datasets ||
        this.datasets.length === 0 ||
        (this.datasets.length === 1 && !this.datasets[0])
      ) {
        this._state = null;
        return;
      }

      this._state = {
        local        : this.datasets.some(isUserShape),
        readOnly     : this.datasets.every(isReadOnly),
        single       : this.datasets.length === 1,
        multiReadOnly: this.datasets.length > 1 && this.datasets.some(isReadOnly),
        exportable   : !!this.datasets[0]?.syntool?.exportShapes,

        shape: this._getStateShape(),
      };
    },
    _getStateShape() {
      const datasets = this.datasets;

      if (datasets.length === 1) {
        return datasets[0].shape;
      }

      let {start, end} = datasets[0].shape;

      for (let i = 1; i < datasets.length && (start || end); i++) {
        if (start && !sameDate(start, datasets[i].shape.start)) {
          start = null;
        }
        if (end && !sameDate(end, datasets[i].shape.end)) {
          end = null;
        }
      }

      return {start, end};
    },

    setDatasets(datasets) {
      datasets = datasets.filter(isShapeDataset);
      this.datasets = (datasets.length > 0 ? datasets : null);
    },

    getValue(name) {
      return this._state?.shape[name];
    },
    getChanges() {
      const changes = {};
      let value;

      if (!this._state) {
        return changes;
      }

      if (this._state.single) {
        if (this.getValue('type') === Shape.WKT) {
          value = getTextFieldValue(this.node, 'author');
          if (value !== this.getValue('author')) {
            changes.author = value;
          }
        }

        value = getTextFieldValue(this.node, 'text');
        if (value !== this.getValue('text')) {
          changes.text = value;
        }
      }
      ['start', 'end'].forEach((name) => {
        //eslint-disable-next-line no-shadow
        const value = getDateFieldValue(this.node, name);
        const originalValue = this.getValue(name);

        if (
          // value is valid date (Date | undefined)
          value !== null &&
          // value is different than the original one
          value !== originalValue &&
          (!value || !originalValue || !sameDate(value, originalValue))
        ) {
          changes[name] = value;
        }
      });
      return changes;
    },

    focusOn(name) {
      this.node.querySelector(`#${idOfField(name)}`)?.focus();
    },

    setSizeConstraints(min, max, convertValue) {
      this.minSize = min;
      this.maxSize = max;
      this.convertSize = convertValue;
      this.render();
    },

    resetField(name) {
      if (!this._state) {
        return;
      }

      this._rendering = true;

      const {readOnly, single, shape} = this._state;

      if (single) {
        switch (name) {
        case 'author':
          updateTextField(this.node, 'author', shape.author, {readOnly});
          break;
        case 'text':
          updateTextField(this.node, 'text', shape.text, {
            readOnly,
            pattern: shape.type === Shape.TEXT ? '.*\\S+.*' : '',
            required: shape.type === Shape.TEXT,
          });
          break;
        case 'start':
          updateDateField(this.node, 'start', shape.start, {readOnly});
          break;
        case 'end':
          updateDateField(this.node, 'end', shape.end, {readOnly});
          break;
        case 'arrow':
          updateSwitchField(this.node, 'arrow', !!shape.arrow, {readOnly});
          break;
        case 'type':
          updateSwitchField(this.node, 'type', shape.arrow.directionFlipped, {readOnly});
          break;
        case 'flipped':
          updateSwitchField(this.node, 'flipped', shape.arrow.sideFlipped, {readOnly});
          break;
        case 'ratio':
          updateNumberField(this.node, 'ratio', shape.arrow.ratio, {readOnly});
          break;
        case 'density':
          updateNumberField(this.node, 'density', shape.arrow.density, {readOnly});
          break;
        case 'offset':
          updateNumberField(this.node, 'offset', shape.arrow.offset, {
            readOnly,
            min: 0,
            max: (shape.type === Shape.LINE ? shape.points.length - 2 : 18),
          });
          break;
        case 'size': {
          const value = this.convertSize(
            shape.type === Shape.ARROW
            ? shape.arrowSize
            : shape.arrow.size
          );

          updateNumberField(this.node, 'size', value, {
            readOnly,
            min: this.minSize,
            max: this.maxSize,
          });
          break;
        }
        default:
        }
      } else {
        switch (name) {
        case 'start':
          updateDateField(this.node, 'start', shape.start, {readOnly});
          break;
        case 'end':
          updateDateField(this.node, 'end', shape.end, {readOnly});
          break;
        default:
        }
      }

      this._rendering = false;
    },
    render() {
      const {node} = this;

      this._updateState();

      if (!this._state) {
        node.setAttribute('aria-hidden', 'true');
        this.isEmpty = true;
        this.shouldNotify = false;
        return;
      }

      this._rendering = true;

      const {
        local,
        multiReadOnly,
        exportable,
        readOnly,
        single,
        shape,
      } = this._state;

      node.classList.toggle('is-local', local);
      node.classList.toggle('is-readonly', readOnly);
      node.classList.toggle('is-multi-readonly', multiReadOnly);
      node.classList.toggle('is-exportable', exportable);
      node.classList.toggle('has-arrow', !!shape.arrow);
      node.removeAttribute('aria-hidden');
      if (shape.type) {
        node.setAttribute('data-shape-type', shape.type);
      } else {
        node.removeAttribute('data-shape-type');
      }

      if (single) {
        if (shape.type === Shape.WKT) {
          updateTextField(node, 'author', shape.author, {readOnly});
        }
        updateTextField(node, 'text', shape.text, {
          readOnly,
          pattern: shape.type === Shape.TEXT ? '.*\\S+.*' : '',
          required: shape.type === Shape.TEXT,
        });
        updateDateField(node, 'start', shape.start, {readOnly});
        updateDateField(node, 'end', shape.end, {readOnly});
        if (
          shape.type === Shape.LINE ||
          shape.type === Shape.ELLIPSE
        ) {
          updateSwitchField(node, 'arrow', !!shape.arrow, {readOnly});
        }
        if (shape.arrow) {
          updateSwitchField(node, 'type', shape.arrow.directionFlipped, {
            readOnly,
            onLabel: (shape.type === Shape.LINE ? '—◀—' : '⟲ CCW'),
            offLabel: (shape.type === Shape.LINE ? '—▶—' : '⟳ CW'),
          });
          updateSwitchField(node, 'flipped', shape.arrow.sideFlipped, {
            readOnly,
            onLabel: (shape.type === Shape.LINE ? 'Left' : 'Out'),
            offLabel: (shape.type === Shape.LINE ? 'Right' : 'In'),
          });
          updateNumberField(node, 'ratio', shape.arrow.ratio, {readOnly});
          updateNumberField(node, 'density', shape.arrow.density, {readOnly});
          updateNumberField(node, 'offset', shape.arrow.offset, {
            readOnly,
            min: 0,
            max: (shape.type === Shape.LINE ? shape.points.length - 2 : 18),
          });
        }
        if (shape.arrow || shape.type === Shape.ARROW) {
          const value = this.convertSize(
            shape.type === Shape.ARROW
            ? shape.arrowSize
            : shape.arrow.size
          );

          updateNumberField(node, 'size', value, {
            readOnly,
            min: this.minSize,
            max: this.maxSize,
          });
        }
      } else {
        updateDateField(node, 'start', shape.start, {readOnly});
        updateDateField(node, 'end', shape.end, {readOnly});
      }

      this._rendering = false;
      this.isEmpty = false;
      this.shouldNotify = true;
      this.hub.emit('render');
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {ShapesInfoView});
