/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DAY_IN_MILLISECONDS,
  MONTH_IN_MILLISECONDS,
  NONPASSIVE,
  parseDecimal,
  inverseLerp,
  createClass,
  throttleOncePerEventLoop,
  throttleOncePerAnimationFrame,
  HIDKeys,
  h,
} from '../../syntool.js';
import {createStyleSheet} from '../../utils/dom_utils.js';
import {Gestures} from '../utils.js';

const styleSheet = createStyleSheet();
const MONTH_NAMES = [
  'January', 'February', 'March', 'April', 'May', 'June', 'July', 'August',
  'September', 'October', 'November', 'December'
];

function cloneDateForYear(date, year) {
  var clone = new Date(date.getTime());

  clone.setUTCFullYear(year);
  return clone;
}
function cloneDateForMonth(date, month) {
  var clone = new Date(date.getTime());

  clone.setUTCMonth(month);
  if (month !== clone.getUTCMonth()) {
    // Oct 31th + 1 month = December 1st
    clone.setUTCDate(0);
    // now its December 0th => November 30th
  }
  return clone;
}
function cloneDateForDay(date, day) {
  var clone = new Date(date.getTime());

  clone.setUTCDate(day);
  return clone;
}

function progressInMonth(date) {
  return inverseLerp(
    Date.UTC(date.getUTCFullYear(), date.getUTCMonth()),
    Date.UTC(date.getUTCFullYear(), date.getUTCMonth() + 1) - 1,
    date.getTime()
  );
}

let instanceCounter = 0;

export const CoverageTimelineView = createClass([
  /*Public*/ {
    constructor: function CoverageTimelineView(node) {
      this.node = node;
      this.node.dataset.id = `CoverageTimelineView-${instanceCounter++}`;
      this.currentDate = new Date();
      this.minYear = this.currentDate.getUTCFullYear();
      this.maxYear = this.minYear + 1;
      this.syntool = null;
      this._store = null;
      this._colocActive = false;

      this._panned = false;
      this._panStartValue = null;

      this.handleElementClick = this.handleElementClick.bind(this);
      this.handleCurrentDateChange = throttleOncePerEventLoop(this.handleCurrentDateChange);
      this.handleCoverageChange = throttleOncePerEventLoop(this.handleCoverageChange);
      this.updateHighlights = throttleOncePerAnimationFrame(this.updateHighlights);
      this.updateVisibleDays = throttleOncePerAnimationFrame(this.updateVisibleDays);
      this.handleMouseWheel = Gestures.getWheelHandler(this.handleMouseWheel, this);

      node.addEventListener('click', this.handleElementClick, NONPASSIVE);
      node.addEventListener('wheel', this.handleMouseWheel, NONPASSIVE);

      Gestures.drag.touch(
        this.node,
        {direction: 'x', moveTolerance: 10},
        this.handleDragChange,
        this
      );
    },

    get store() {
      return this._store;
    },
    set store(store) {
      if (this._store !== store) {
        if (this._store) {
          this._store.hub
            .off('change:currentDate', this.handleCurrentDateChange, this)
            .off('change:minYear', this.handleAbsoluteRangeChange, this)
            .off('change:maxYear', this.handleAbsoluteRangeChange, this)
            .off('change:timeframeLength', this.handleCurrentTimeframeChange, this)
            .off('change:coverage', this.handleCoverageChange, this);
        }
        this._store = store;
        if (store) {
          this.store.hub
            .on('change:currentDate', this.handleCurrentDateChange, this)
            .on('change:minYear', this.handleAbsoluteRangeChange, this)
            .on('change:maxYear', this.handleAbsoluteRangeChange, this)
            .on('change:timeframeLength', this.handleCurrentTimeframeChange, this)
            .on('change:coverage', this.handleCoverageChange, this);

          this.currentDate = new Date(Date.UTC(
            this.store.currentDate.getUTCFullYear(),
            this.store.currentDate.getUTCMonth()
          ));
          this.handleAbsoluteRangeChange();
          this.updateCoverage();
        }
      }
    },

    get colocActive() {
      return this._colocActive;
    },
    set colocActive(colocActive) {
      if (this._colocActive === !colocActive) {
        this._colocActive = !!colocActive;
        this.updateCoverage();
      }
    },

    render() {
      var yearsCount = this.maxYear - this.minYear;
      var yearsHtml = h('div', {class: 'timespan'});
      yearsHtml += h('div', {class: 'current'});
      for (var year = this.minYear; year < this.maxYear; year++) {
        yearsHtml += h('li', {'data-value': year, title: year}, [
          h('span', {class: 'short'}, year.toString().substr(-2)),
          h('span', {class: 'full'}, year),
        ]);
      }

      var monthsHtml = h('div', {class: 'timespan'});
      monthsHtml += h('div', {class: 'current'});
      for (var month = 0; month < 12; month++) {
        monthsHtml += h('li', {'data-value': month, title: month}, [
          h('span', {class: 'short'}, MONTH_NAMES[month].substring(0, 3)),
          h('span', {class: 'full'}, MONTH_NAMES[month]),
        ]);
      }

      var daysHtml = h('div', {class: 'timespan'});
      daysHtml += h('div', {class: 'current'});
      for (var day = 0; day < 31; day++) {
        daysHtml += h('li', {'data-value': day}, h('span', day + 1));
      }

      this.node.innerHTML = [
        h('ul', {class: 'years', style: `--count: ${yearsCount}`}, yearsHtml),
        h('ul', {class: 'months'}, monthsHtml),
        h('ul', {class: 'days'}, daysHtml),
      ].join('');

      const selector = `[data-id="${this.node.dataset.id}"]`;
      for (let i = styleSheet.cssRules.length - 1; i >= 0; i--) {
        const rule = styleSheet.cssRules[i];
        if (rule.cssText.includes(selector)) {
          styleSheet.deleteRule(i);
        }
      }
      styleSheet.insertRule(`
        @media screen and (max-width: ${30 * yearsCount}px) {
          ${selector} ul.years li {
            --show-as: var(--show-as--short);
          }
        }
      `);
      styleSheet.insertRule(`
        @media screen and (max-width: ${20 * yearsCount}px) {
          ${selector} ul.years li:nth-child(2n+2) {
            --show-as: var(--show-as--dot);
          }
        }
      `);
    },
    updateVisibleDays() {
      if (!this.syntool || !this.store) {
        return;
      }

      var currentDate = this.store.currentDate;
      var currentYear = currentDate.getUTCFullYear();
      var currentMonth = currentDate.getUTCMonth();
      var endOfMonth = new Date(Date.UTC(currentYear, currentMonth + 1, 0));
      var lastDay = endOfMonth.getUTCDate() - 1;
      var lastDayNode = this.node.querySelector('.days [data-value="' + lastDay + '"]');
      var tmpNode = lastDayNode;

      while (tmpNode && tmpNode.style.visibility === 'hidden') {
        tmpNode.style.visibility = '';
        tmpNode = tmpNode.previousElementSibling;
      }

      tmpNode = lastDayNode.nextElementSibling;
      while (tmpNode && tmpNode.style.visibility !== 'hidden') {
        tmpNode.style.visibility = 'hidden';
        tmpNode = tmpNode.nextElementSibling;
      }
    },
    updateHighlights() {
      if (!this.syntool || !this.store) {
        return;
      }

      const {minDate, maxDate, currentDate} = this.store;
      const currentYear = currentDate.getUTCFullYear();
      const currentMonth = currentDate.getUTCMonth();

      const leftYear = inverseLerp(this.minYear, this.maxYear, currentYear) * 100;
      this.node.querySelector('.years .current').style.left = leftYear + '%';

      const minFromStartOfFirstYear = (
        minDate.getUTCFullYear() - this.minYear
        + minDate.getUTCMonth() / 12
        + progressInMonth(minDate) / 12
      );
      const maxFromStartOfFirstYear = (
        maxDate.getUTCFullYear() - this.minYear
        + maxDate.getUTCMonth() / 12
        + progressInMonth(maxDate) / 12
      );
      const yearsWidth = maxFromStartOfFirstYear - minFromStartOfFirstYear;
      const deltaYears = this.maxYear - this.minYear;
      const yearsTimespanNode = this.node.querySelector('.years .timespan');
      yearsTimespanNode.style.left = (minFromStartOfFirstYear / deltaYears * 100) + '%';
      yearsTimespanNode.style.setProperty('--width', yearsWidth / deltaYears * 100);

      const leftMonth = currentDate.getUTCMonth() * 100 / 12;
      this.node.querySelector('.months .current').style.left = leftMonth + '%';

      const minFromStartOfYear = minFromStartOfFirstYear - (currentYear - this.minYear);
      const monthsTimespanNode = this.node.querySelector('.months .timespan');
      monthsTimespanNode.style.left = (minFromStartOfYear * 100) + '%';
      monthsTimespanNode.style.setProperty('--width', yearsWidth * 100);

      const leftDay = (currentDate.getUTCDate() - 1) * 100 / 31;
      this.node.querySelector('.days .current').style.left = leftDay + '%';

      const startOfMonth = Date.UTC(currentYear, currentMonth);
      const percentFromStartOfMonth = (
        (minDate.getTime() - startOfMonth) / MONTH_IN_MILLISECONDS * 100
      );
      const percentWidth = (
        (maxDate.getTime() - minDate.getTime()) / MONTH_IN_MILLISECONDS * 100
      );
      const daysTimespanNode = this.node.querySelector('.days .timespan');
      daysTimespanNode.style.left = percentFromStartOfMonth + '%';
      daysTimespanNode.style.setProperty('--width', percentWidth);
    },
    updateCoverage() {
      for (const node of this.node.querySelectorAll('[data-value]')) {
        node.setAttribute('title', 'No data');
        node.classList.remove('has-some', 'has-coloc', 'has-every');
      }

      if (!this.syntool || !this.store) {
        return;
      }

      const coverage = this.store.getCoverage();
      if (coverage.isEmpty) {
        return;
      }

      // type AvailabilityElement = {
      //   hasColoc: bool,
      //   hasEvery: bool,
      //   products: string[],
      // };
      // type Availability = {
      //   years: {[year: number]: AvailabilityElement},
      //   months: {[month: number]: AvailabilityElement},
      //   days: {[day: number]: AvailabilityElement},
      // };
      const availability = {years: {}, months: {}, days: {}};
      const productIds = coverage.getProductIds();
      const colocCount = (
        this.colocActive ? this.countColocProductIds(productIds) : 0
      );

      const currentDate = this.store.currentDate;
      const currentYear = currentDate.getUTCFullYear();
      const currentMonth = currentDate.getUTCMonth();

      let minYear =  Infinity;
      let maxYear = -Infinity;

      for (const productId of productIds) {
        const yearCoverages = coverage.coverageByProduct[productId];
        const years = Object.keys(yearCoverages);

        minYear = Math.min(minYear, ...years);
        maxYear = Math.max(maxYear, ...years);
      }

      if (!isFinite(minYear) || !isFinite(maxYear)) {
        return;
      }

      for (let year = minYear; year <= maxYear; year++) {
        const productsInYear = new Set();
        let colocInYear = false;

        for (let month = 0; month < 12; month++) {
          const productsInMonth = new Set();
          let colocInMonth = false;

          for (let day = 0; day < 31; day++) {
            const productsInDay = new Set();
            let colocInDay = false;

            for (const productId of productIds) {
              const yearCoverages = coverage.coverageByProduct[productId];
              const yearCoverage = yearCoverages[year];
              const monthCoverage = yearCoverage ? yearCoverage[month] : 0;
              //eslint-disable-next-line no-bitwise
              const dayCovered = monthCoverage & (1 << day);

              if (dayCovered) {
                productsInYear.add(productId);
                productsInMonth.add(productId);
                productsInDay.add(productId);
              }
            }

            colocInDay = this.countColocProductIds(productsInDay) === colocCount;
            colocInMonth = colocInMonth || colocInDay;
            colocInYear = colocInYear || colocInDay;

            if (productsInDay.size > 0 && year === currentYear && month === currentMonth) {
              availability.days[day] = {
                hasColoc: colocInDay,
                hasEvery: colocInDay,
                products: [...productsInDay],
              };
            }
          }
          if (productsInMonth.size > 0 && year === currentYear) {
            availability.months[month] = {
              hasColoc: colocInMonth,
              hasEvery: this.countColocProductIds(productsInMonth) === colocCount,
              products: [...productsInMonth],
            };
          }
        }
        if (productsInYear.size > 0) {
          availability.years[year] = {
            hasColoc: colocInYear,
            hasEvery: this.countColocProductIds(productsInYear) === colocCount,
            products: [...productsInYear],
          };
        }
      }

      for (const timeScale of ['years', 'months', 'days']) {
        for (const node of this.node.querySelectorAll(`.${timeScale} [data-value]`)) {
          const timeValue = node.dataset.value;
          const availabilityElement = availability[timeScale][timeValue];
          if (availabilityElement && availabilityElement.products.length > 0) {
            node.classList.add('has-some');
            let dataMessage = 'Contains data for products:';
            if (colocCount > 1) {
              if (availabilityElement.hasEvery) {
                node.classList.add('has-every');
                dataMessage = (
                  'Contains data for all selected products, but not on the same days:'
                );
              }
              if (availabilityElement.hasColoc) {
                node.classList.add('has-coloc');
                dataMessage = (
                  timeScale === 'days'
                  ? 'Contains data for all selected products:'
                  : 'Contains one or more days with data for all selected products:'
                );
              }
            }
            node.title = (
              dataMessage + '\n\n' +
              availabilityElement.products.map(this.labelForProductWithId, this)
              .join('\n')
            );
          }
        }
      }
    },
  },
  /*Event handlers*/ {
    handleElementClick(event) {
      if (!this.syntool || !this.store) {
        return;
      }

      var target = event.target;

      while (
        target !== this.node &&
        !target.hasAttribute('data-value')
      ) {
        target = target.parentNode;
      }

      if (target === this.node) {
        return;
      }

      var date = this.store.currentDate;
      var timeScale = target.parentNode.className;
      var timeValue = parseDecimal(target.getAttribute('data-value'));

      switch (timeScale) {
      case 'years':
        date = cloneDateForYear(date, timeValue);
        break;
      case 'months':
        date = cloneDateForMonth(date, timeValue);
        break;
      case 'days':
        date = cloneDateForDay(date, timeValue + 1);
        break;
      default:
        return;
      }

      date.setUTCHours(12);
      date.setUTCMinutes(0);
      date.setUTCSeconds(0);

      if (timeScale === 'years') {
        date = this.findNearestCoveredDay(date, true, true);
      } else if (timeScale === 'months') {
        date = this.findNearestCoveredDay(date, false, true);
      }

      this.store.currentDate = date;
      event.preventDefault();
    },
    handleMouseWheel(event, delta, horizontal) {
      if (!this.syntool || !this.store) {
        return;
      }
      event.preventDefault();
      if (this.syntool.keyboardManager.isKeyDown(HIDKeys.shift)) {
        delta *= 2;
      } else if (this.syntool.keyboardManager.isKeyDown(HIDKeys.alt)) {
        delta /= 2;
      }

      delta *= this.store.timeframeLength / 3;
      this.store.currentTime += (horizontal ? 1 : -1) * delta;
    },
    handleDragChange(props) {
      if (!this.syntool || !this.store) {
        return;
      }

      switch (props.state) {
      case 'possible':
        this._panStartValue = this.store.currentTime;
        break;
      case 'cancel':
        if (this._panned) {
          this.store.currentTime = this._panStartValue;
        }
        break;
      default:
        this._panned = true;

        var speedMsPerPx = (31 * DAY_IN_MILLISECONDS) / this.node.offsetWidth;
        var delta = props.movement * speedMsPerPx / 2;
        var value = this._panStartValue + delta;

        this.store.currentTime = value;
      }
    },
    handleCurrentDateChange() {
      if (!this.syntool || !this.store) {
        return;
      }

      var prevDate = this.currentDate;
      var currentDate = this.store.currentDate;

      if (
        currentDate.getUTCMonth()    !== prevDate.getUTCMonth()    ||
        currentDate.getUTCFullYear() !== prevDate.getUTCFullYear()
      ) {
        currentDate = new Date(Date.UTC(
          currentDate.getUTCFullYear(),
          currentDate.getUTCMonth()
        ));
        this.currentDate = currentDate;
        this.updateCoverage();
        this.updateVisibleDays();
      }
      this.updateHighlights();
    },
    handleAbsoluteRangeChange() {
      if (!this.syntool || !this.store) {
        return;
      }

      this.minYear = this.store.minYear;
      this.maxYear = this.store.maxYear;
      this.render();
      this.updateHighlights();
      this.updateVisibleDays();
    },
    handleCurrentTimeframeChange() {
      this.updateHighlights();
    },
    handleCoverageChange() {
      this.updateCoverage();
    },
  },
  /*Protected*/ {
    labelForProductWithId(productId) {
      return '• ' + this.syntool.productsStore.getProduct(productId).textLabel;
    },
    countColocProductIds(productIds) {
      let count = 0;
      for (const productId of productIds) {
        if (this.syntool.filters.product.includedInColoc.withId(productId)) {
          count++;
        }
      }
      return count;
    },
    findNearestCoveredDay(currentDate, month, day) {
      var coverage = this.store.getCoverage();
      var withColoc = (this.colocActive && coverage.hasColoc());
      var productIds = (
        coverage.getProductIds()
        .filter(this.syntool.filters.product.nonPermanentVisible.withId)
      );

      if (withColoc) {
        productIds = (
          productIds.filter(this.syntool.filters.product.includedInColoc.withId)
        );
      }

      var dateHasData = coverage.hasDataAtDate(currentDate, withColoc, productIds);
      if (dateHasData) {
        return currentDate;
      }

      var nearestDate = coverage.findNearest(currentDate, {
        productIds: productIds,

        coloc : withColoc,
        month : month,
        day   : day,
        past  : true,
        future: true,
      });

      return nearestDate || currentDate;
    },
  },
]);
window.ODL = Object.assign(window.ODL || {}, {CoverageTimelineView});
