/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  DAY_IN_MILLISECONDS,
  NONPASSIVE,
  emptyFunction,
  wrapMethods,
  extendClass,
  throttleOncePerEventLoop,
  throttleOncePerAnimationFrame,
  clamp,
  EventsHub,
  DOMEvents,
} from '../../syntool.js';
import {SimileAjax, Timeline} from '../../vendors/simile_timeline.js';
import {Gestures} from '../utils.js';
import {TimelineCustomDateLabeller} from './timeline_custom_date_labeller.js';
import {TimelineCustomEventSource} from './timeline_custom_event_source.js';
import {TimelineLabelessEventPainter} from './timeline_labeless_event_painter.js';

const MIN_BAND_DRAG_MOVEMENT = 5; // px

extendClass(Timeline._Band, {
  _onMouseDown(innerFrame, evt, target) {
    window.original();
    if (evt.target === this._eventPainter._eventLayer) {
      this._dragAfterMovement = false;
    } else {
      this._dragging = false;
      this._dragAfterMovement = true;
    }
  },
  _onMouseMove(innerFrame, evt, target) {
    if (!this._dragging && this._dragAfterMovement) {
      const diffX = evt.clientX - this._dragX;
      const diffY = evt.clientY - this._dragY;

      if (Math.hypot(diffX, diffY) < MIN_BAND_DRAG_MOVEMENT) {
        return;
      }
      this._dragging = true;
      this._dragAfterMovement = false;
    }
    window.original();
  },
  _onMouseUp(innerFrame, evt, target) {
    this._dragging = false;
    this._dragAfterMovement = false;
  },
  _onMouseWheel(event, delta, horizontal) {
    event.preventDefault();
    this._moveEther((horizontal ? -1 : 1) * delta * 50);
  },
  _onDblClick: emptyFunction,
  _onKeyDown: emptyFunction,
  _onKeyUp: emptyFunction,

  _addTrackingListeners(event) {
    /**
     * If the target element is removed from the document events won't
     * necessarily bubble up.
     * If there is any risk of an element being removed while it is being
     * touched, the best practice is to attach the touch listeners directly
     * to the target.
     */

    if (
      event.target !== event.currentTarget ||
      !this._elementsWithTrackingTouchListeners.has(event.target)
    ) {
      DOMEvents(event.target)
        .on('touchmove+passive', this._onTouchMove, this)
        .on('touchend+passive', this._onTouchEnd, this)
        .on('touchcancel+passive', this._onTouchCancel, this);
      this._elementsWithTrackingTouchListeners.add(event.target);
    }
  },
  _removeTrackingListeners(event) {
    if (
      event.target !== event.currentTarget &&
      this._elementsWithTrackingTouchListeners.has(event.target)
    ) {
      DOMEvents(event.target)
        .off('touchmove+passive', this._onTouchMove, this)
        .off('touchend+passive', this._onTouchEnd, this)
        .off('touchcancel+passive', this._onTouchCancel, this);
      this._elementsWithTrackingTouchListeners.delete(event.target);
    }
  },

  _onTouchStart(event) {
    if (event.targetTouches.length === 1) {
      var touch = event.targetTouches[0];

      this._dragging = true;
      this._dragX = touch.clientX;
      this._dragY = touch.clientY;
      this._addTrackingListeners(event);
    }
  },
  _onTouchMove(event) {
    if (this._dragging && event.targetTouches.length === 1) {
      var touch = event.targetTouches[0];

      var diffX = touch.clientX - this._dragX;
      var diffY = touch.clientY - this._dragY;

      this._dragX = touch.clientX;
      this._dragY = touch.clientY;

      this._moveEther(this._timeline.isHorizontal() ? diffX : diffY);
      this._positionHighlight();
    }
  },
  _onTouchEnd(event) {
    this._dragging = false;
    this._removeTrackingListeners(event);
  },
  _onMouseCancel(event) {
    this._dragging = false;
    this._removeTrackingListeners(event);
  },
});

function limitBandBetween(band) {
  wrapMethods(band, {
    _moveEther(shift) {
      var length = this.getViewLength();
      var offset = this.getViewOffset();
      var adjustedShift = clamp(
        shift,
        Math.ceil(length / 2 - this.dateToPixelOffset(this.__maxDate)) - offset,
        Math.floor(length / 2 - this.dateToPixelOffset(this.__minDate)) - offset
      );
      return window.original(adjustedShift);
    },
  });
}
function addTouchSupportToBand(band) {
  band._elementsWithTrackingTouchListeners = new WeakSet();
  DOMEvents(band._div).on('touchstart', band._onTouchStart, band);
}
function addScrollSupportToBand(band) {
  const handleMouseWheel = Gestures.getWheelHandler(band._onMouseWheel, band);
  band._div.addEventListener('wheel', handleMouseWheel, NONPASSIVE);
}

const MIN_UNIT_WIDTH = 75; // px
const TICK_MULTIPLES_60 = [1, 2, 5, 10, 15, 20];
const TICK_MULTIPLES_24 = [1, 2, 3, 4, 6, 12];
const TICK_MULTIPLES_12 = [1, 2, 3, 4, 6];
const TICK_MULTIPLES_10 = [1, 2, 5, 10, 15];

function getTicksUnitFromTimeframeLength(length, width) {
  const fitUnitsCount = width / MIN_UNIT_WIDTH;

  // Find biggest unit that fits
  for (
    let unit = SimileAjax.DateTime.MILLISECOND;
    unit < SimileAjax.DateTime.MILLENNIUM;
    unit++
  ) {
    // Select good multiples of unit
    let multiples;
    switch (unit) {
    case SimileAjax.DateTime.SECOND:
    case SimileAjax.DateTime.MINUTE:
      multiples = TICK_MULTIPLES_60;
      break;
    case SimileAjax.DateTime.HOUR:
      multiples = TICK_MULTIPLES_24;
      break;
    case SimileAjax.DateTime.WEEK:
      // Skip week unit
      continue;
    case SimileAjax.DateTime.MONTH:
      multiples = TICK_MULTIPLES_12;
      break;
    default:
      multiples = TICK_MULTIPLES_10;
    }

    // Find best multiple of unit that fits
    const multiple = (
      length / (fitUnitsCount * SimileAjax.DateTime.gregorianUnitLengths[unit])
    );
    for (let i = 0; i < multiples.length; i++) {
      if (multiple <= multiples[i]) {
        return [unit, multiples[i]];
      }
    }
  }
  return [SimileAjax.DateTime.MILLENNIUM, 1];
}

export class DetailsTimelineView {
  constructor(node) {
    var width = node.getBoundingClientRect().width;
    var eventSource = new TimelineCustomEventSource();

    var theme = Timeline.ClassicTheme.create();
    theme.event.track.offset = 2;
    theme.event.track.height = 0;
    theme.event.track.gap = 2;
    theme.event.tape.height = 5;
    theme.event.instant.iconWidth = 10;
    theme.event.instant.iconHeight = 10;
    theme.mouseWheel = null;

    var intervalLength = DAY_IN_MILLISECONDS;
    const [intervalUnit, multiple] = (
      getTicksUnitFromTimeframeLength(intervalLength, width)
    );
    var bandInfo = Timeline.createBandInfo({
      date          : new Date(),
      width         : '100%',
      intervalUnit,
      multiple,
      intervalPixels: width,
      eventSource   : eventSource,
      showEventText : false,
      theme         : theme,
      eventPainter  : TimelineLabelessEventPainter,
    });
    bandInfo.ether._interval = intervalLength;
    bandInfo.labeller = new TimelineCustomDateLabeller();

    var tl = Timeline.create(node, [bandInfo]);


    this.node = node;
    this._store = null;
    this._filter = null;
    this._open = false;

    this.tl = tl;
    this.eventSource = eventSource;
    this.hub = new EventsHub();

    this._ignoreScroll = false;

    this.repaint = throttleOncePerAnimationFrame(this.repaint);
    this.filterMatcher = this.filterMatcher.bind(this);
    this.handleClick = this.handleClick.bind(this);
    this.handleScroll = this.handleScroll.bind(this);
    this.handleCurrentDateChange = (
      throttleOncePerEventLoop(this.handleCurrentDateChange)
    );
    this.handleAbsoluteRangeChange = (
      throttleOncePerEventLoop(this.handleAbsoluteRangeChange)
    );
    this.handleTimeframeLengthChange = (
      throttleOncePerEventLoop(this.handleTimeframeLengthChange)
    );


    var band = tl.getBand(0);
    band._keyboardInput.tabIndex = -1;
    limitBandBetween(band);
    addTouchSupportToBand(band);
    addScrollSupportToBand(band);

    var etherPainter = band.getEtherPainter();
    wrapMethods(etherPainter, {
      paint() {
        window.original();
        // Move the date marks above the events
        // z-index of the events is 150
        this._markerLayer.parentElement.style.zIndex = '120';
        // Make the date marks non-interactive
        this._markerLayer.parentElement.style.pointerEvents = 'none';
      },
    });

    var eventPainter = band.getEventPainter();
    eventPainter.setFilterMatcher(this.filterMatcher);
    var throttledPaint = throttleOncePerAnimationFrame(
      eventPainter.constructor.prototype.paint.bind(eventPainter)
    );
    eventPainter.paint = () => {
      if (this.open) {
        throttledPaint();
      }
    };


    band.addOnScrollListener(this.handleScroll);
    eventPainter.addOnSelectListener(this.handleClick);
  }

  get store() {
    return this._store;
  }
  set store(store) {
    if (this._store !== store) {
      if (this._store) {
        this._store.hub
          .off('change:currentDate', this.handleCurrentDateChange, this)
          .off('change:minYear', this.handleAbsoluteRangeChange, this)
          .off('change:maxYear', this.handleAbsoluteRangeChange, this)
          .off('change:timeframeLength', this.handleTimeframeLengthChange, this);
      }
      this._store = store;
      if (store) {
        this.store.hub
          .on('change:currentDate', this.handleCurrentDateChange, this)
          .on('change:minYear', this.handleAbsoluteRangeChange, this)
          .on('change:maxYear', this.handleAbsoluteRangeChange, this)
          .on('change:timeframeLength', this.handleTimeframeLengthChange, this);
        this.handleCurrentDateChange();
        this.handleAbsoluteRangeChange();
        this.handleTimeframeLengthChange();
      }
    }
  }

  get open() {
    return this._open;
  }
  set open(open) {
    if (this._open === !open) {
      this._open = !!open;
      if (open) {
        this.repaint();
      }
      this.hub.emit('change:open');
    }
  }

  get filter() {
    return this._filter;
  }
  set filter(filter) {
    if (this._open !== filter) {
      this._filter = filter;
      this.repaint();
      this.hub.emit('change:filter');
    }
  }

  addDataset(dataset) {
    this.eventSource.addDataset(dataset);
  }
  updateDataset(dataset) {
    this.eventSource.addDataset(dataset);
  }
  removeDataset(dataset) {
    this.eventSource.remove(dataset.datasetId);
  }
  clearDatasets() {
    this.eventSource.clear();
  }

  updateWidth(width) {
    var band = this.tl.getBand(0);
    var ether = band.getEther();

    if (!width) {
      width = this.tl.getPixelLength();
    }

    if (ether._pixelsPerInterval !== width) {
      ether._pixelsPerInterval = width;
      band.setViewLength(width);
      const etherPainter = band.getEtherPainter();
      if (this.store) {
        [etherPainter._unit, etherPainter._multiple] = (
          getTicksUnitFromTimeframeLength(this.store.timeframeLength, width)
        );
      }
      this.relayout();
    }
  }
  relayout() {
    if (this._open) {
      this._ignoreScroll = true;
      this.tl.layout();
      this._ignoreScroll = false;
    }
  }
  repaint() {
    if (this._open) {
      this._ignoreScroll = true;
      this.tl.paint();
      this._ignoreScroll = false;
    }
  }

  filterMatcher(event) {
    if (this._filter) {
      return this._filter(event.getDataset());
    }
    return true;
  }
  handleClick(eventId) {
    if (eventId) {
      var event = this.eventSource.getEvent(eventId);
      this.hub.emit('pick', event.getDataset());
    } else {
      this.hub.emit('pick', null);
    }
  }
  handleScroll(band) {
    if (this._ignoreScroll) {
      return;
    }
    this.hub.emit('change:currentDate', band.getCenterVisibleDate());
    this.handleCurrentDateChange();
  }
  handleCurrentDateChange() {
    if (!this.store) {
      return;
    }

    const band = this.tl.getBand(0);
    const centerDate = band.getCenterVisibleDate();
    const {currentDate} = this.store;

    this._ignoreScroll = true;
    if (centerDate.getTime() !== currentDate.getTime()) {
      band.setCenterVisibleDate(currentDate);
    }
    this._ignoreScroll = false;
  }
  handleAbsoluteRangeChange() {
    if (!this.store) {
      return;
    }

    const band = this.tl.getBand(0);
    const {absoluteMinDate, absoluteMaxDate, currentDate} = this.store;

    this.eventSource._minDate = absoluteMinDate;
    this.eventSource._maxDate = absoluteMaxDate;

    band.__minDate = absoluteMinDate;
    band.__maxDate = absoluteMaxDate;

    this._ignoreScroll = true;
    band.setCenterVisibleDate(currentDate);
    this._ignoreScroll = false;
  }
  handleTimeframeLengthChange() {
    if (!this.store) {
      return;
    }

    const band = this.tl.getBand(0);
    const ether = band.getEther();
    const etherPainter = band.getEtherPainter();
    const {currentDate, timeframeLength} = this.store;

    ether._interval = timeframeLength;
    [etherPainter._unit, etherPainter._multiple] = (
      getTicksUnitFromTimeframeLength(timeframeLength, ether._pixelsPerInterval)
    );

    this._ignoreScroll = true;
    band.setCenterVisibleDate(currentDate);
    this._ignoreScroll = false;
  }
}
window.ODL = Object.assign(window.ODL || {}, {DetailsTimelineView});
