/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/*global SimileAjax: false,Timeline: false*/
import {
  emptyFunctionReturning,
  emptyFunctionReturningFalse,
  emptyFunctionReturningNull,
  createClass,
  clamp,
  Syntool,
} from '../../syntool.js';

function sameDate(dateA, dateB) {
  if (dateA === dateB) {
    return true;
  }
  if (!dateA && !dateB) {
    return true;
  }
  if (!dateA || !dateB) {
    return false;
  }

  var timeA = dateA.getTime();
  var timeB = dateB.getTime();

  if (isNaN(timeA) && isNaN(timeB)) {
    return true;
  }
  return timeA === timeB;
}

var TimelineEvent = createClass([
  /*Public*/ {
    constructor: function TimelineEvent(args) {
      // args is a hash/object. It supports the following keys. Most are optional
      //   id
      //   start
      //   end
      //   caption - tooltip-like caption on the Timeline. Uses HTML title attribute
      //   dataset - dataset object

      var id = args.id ? args.id.trim() : '';
      this._id = id.length > 0 ? id : Timeline.EventUtils.getNewEventID();
      this._start = args.start;
      this._end = args.end || args.start;
      this._title = args.caption || null;
      this._dataset = args.dataset;
    },

    getID() {
      return this._id;
    },
    getStart() {
      return this._start;
    },
    getEnd() {
      return this._end;
    },
    getClassName() {
      return [
        this.isInMap() ? 'in-map' : 'not-in-map',
        this._dataset.isSelected() ? 'is-selected' : 'is-not-selected',
      ].join(' ');
    },
    getDataset() {
      return this._dataset;
    },
    isInMap() {
      return Syntool.mapView.isDatasetInMap(this._dataset);
    },

    updateDates(args) {
      this._start = args.start;
      this._end = args.end || args.start;
    },

    isImprecise   : emptyFunctionReturningFalse,
    getText       : emptyFunctionReturning(''),
    getDescription: emptyFunctionReturning(''),
    getEventID    : emptyFunctionReturningNull,
    getImage      : emptyFunctionReturningNull,
    getLink       : emptyFunctionReturningNull,
    getIcon       : emptyFunctionReturningNull,
    getColor      : emptyFunctionReturningNull,
    getTextColor  : emptyFunctionReturningNull,
    getTapeImage  : emptyFunctionReturningNull,
    getTapeRepeat : emptyFunctionReturningNull,
    getTrackNum   : emptyFunctionReturningNull,
    getProperty   : emptyFunctionReturningNull,
    getLatestStart() {
      return this._start;
    },
    getEarliestEnd() {
      return this._end;
    },
  },
]);

export const TimelineCustomEventSource = createClass([
  /*Public*/ {
    constructor: function TimelineCustomEventSource() {
      this._events = new SimileAjax.EventIndex();
      this._listeners = [];
      this._minDate = null;
      this._maxDate = null;
    },

    addListener(listener) {
      this._listeners.push(listener);
    },
    removeListener(listener) {
      for (var i = 0; i < this._listeners.length; i++) {
        if (this._listeners[i] === listener) {
          this._listeners.splice(i, 1);
          break;
        }
      }
    },
    addDataset(dataset) {
      var id = dataset.datasetId;
      var event = this.getEvent(id);
      var start = dataset.start;
      var end = dataset.end;

      /* Truncate events exceeding timeline boundaries */
      if (this._minDate && start) {
        start = clamp(start, this._minDate, this._maxDate);
      }
      if (this._maxDate && end) {
        end = clamp(end, this._minDate, this._maxDate);
      }

      if (
        event &&
        sameDate(dataset.start, event.getStart()) &&
        sameDate(dataset.end,   event.getEnd())
      ) {
        // dataset already exists in the timeline
        return;
      }

      if (event) {
        // Remove the event before updating it
        this._events._events.remove(event);

        event.updateDates({
          start: start,
          end  : end,
        });

        // Re-add the dataset so that it's index is updated in the sorted array
        this._events.add(event);
      } else {
        event = new TimelineEvent({
          id     : id,
          start  : start,
          end    : end,
          caption: (
            'Dataset:\n' + dataset.name + '\nProduct:\n' +
            dataset.getProduct().textLabel
          ),
          dataset: dataset,
        });

        this.add(event);
      }
    },
    add(event) {
      this._events.add(event);
      this._fire('onAddMany', [event]);
    },
    remove(eventId) {
      if (!(eventId in this._events._idToEvent)) {
        return;
      }
      this._events._events.remove(this._events._idToEvent[eventId]);
      delete this._events._idToEvent[eventId];
      this._events._indexed = false;
      this._fire('onClear', []);
    },
    addMany(events) {
      for (var i = 0; i < events.length; i++) {
        this._events.add(events[i]);
      }
      this._fire('onAddMany', []);
    },
    clear() {
      this._events.removeAll();
      this._fire('onClear', []);
    },
    getEvent(id) {
      return this._events.getEvent(id);
    },
    getEventIterator(startDate, endDate) {
      return this._events.getIterator(startDate, endDate);
    },
    getEventReverseIterator(startDate, endDate) {
      return this._events.getReverseIterator(startDate, endDate);
    },
    getAllEventIterator() {
      return this._events.getAllIterator();
    },
    getCount() {
      return this._events.getCount();
    },
    getEarliestDate() {
      return this._events.getEarliestDate();
    },
    getLatestDate() {
      return this._events.getLatestDate();
    },
    _fire(handlerName, args) {
      for (var i = 0; i < this._listeners.length; i++) {
        var listener = this._listeners[i];
        if (handlerName in listener) {
          listener[handlerName].apply(listener, args);
        }
      }
    },
  },
]);
