/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
/*global SimileAjax: false,Timeline: false*/
import {
  emptyFunctionReturningTrue,
  binarySearch,
  createClass,
  removeAllChildrenOfNode,
  EventsHub,
} from '../../syntool.js';

/*==================================================
 *  Labeless Event Painter
 *==================================================
 */

/*==================================================
 *
 * To enable a single event listener to monitor everything
 * on a Timeline, we need a way to map from an event's icon,
 * label or tape element to the associated timeline, band and
 * specific event.
 *
 * Thus a set format is used for the id's of the
 * events' elements on the Timeline--
 *
 * element id format for icons, tapes:
 *    icons: icon-tl-<timelineID>-<band_index>-<evt.id>
 *    tapes: tape1-tl-<timelineID>-<band_index>-<evt.id>
 *           tape2-tl-<timelineID>-<band_index>-<evt.id>
 *           // some events have more than one tape
 * You can then retrieve the band/timeline objects and event object
 * by using Timeline.EventUtils.decodeEventElID
 *
 *==================================================
 */

/*
 *    eventPaintListener functions receive calls about painting.
 *    function(band, op, evt, els)
 *       context: 'this' will be an TimelineLabelessEventPainter object.
 *                It has properties and methods for obtaining
 *                the relevant band, timeline, etc
 *       band = the band being painted
 *       op = 'paintStarting' // the painter is about to remove
 *            all previously painted events, if any. It will
 *            then start painting all of the visible events that
 *            pass the filter.
 *            evt = null, els = null
 *       op = 'paintEnded' // the painter has finished painting
 *            all of the visible events that passed the filter
 *            evt = null, els = null
 *       op = 'paintedEvent' // the painter just finished painting an event
 *            evt = event just painted
 *            els = array of painted elements' divs. Depending on the event,
 *                  the array could be just a tape or icon (if no label).
 *                  Or could include label, multiple tape divs (imprecise event).
 *                  The array is not ordered. The meaning of
 *                  each el is available by decoding the el's id
 *      Note that there may be no paintedEvent calls if no events were visible
 *      or passed the filter.
 */

function indexOfIntervalContaining(intervals, value) {
  return binarySearch(intervals, function(interval) {
    if (value < interval[0]) {
      return -1;
    }
    if (value > interval[1]) {
      return 1;
    }
    return 0;
  });
}
function addInterval(intervals, start, end) {
  var indexOfStart = indexOfIntervalContaining(intervals, start);
  var indexOfEnd = indexOfIntervalContaining(intervals, end);

  if (indexOfStart < 0) {
    indexOfStart = -(indexOfStart + 1);
  } else {
    start = intervals[indexOfStart][0];
  }
  if (indexOfEnd < 0) {
    indexOfEnd = -(indexOfEnd + 1);
  } else {
    end = intervals[indexOfEnd][1];
    indexOfEnd += 1;
  }

  var interval = [start, end];
  intervals.splice(indexOfStart, indexOfEnd - indexOfStart, interval);
}
function hasIntervalsInRange(intervals, start, end) {
  var indexOfStart = indexOfIntervalContaining(intervals, start);
  var indexOfEnd = indexOfIntervalContaining(intervals, end);

  if (indexOfStart < 0 && indexOfStart < 0) {
    return indexOfStart - indexOfEnd;
    /// The above is equivalent to the following
    // indexOfStart = -(indexOfStart + 1);
    // indexOfEnd = -(indexOfEnd + 1);
    // return indexOfEnd - indexOfStart;
  }
  return true;
}

export const TimelineLabelessEventPainter = createClass([
  /*Public*/ {
    constructor: function TimelineLabelessEventPainter(params) {
      this._params = params;
      this._hub = new EventsHub();
      this._filterMatcher = null;
      this._eventIdToElmt = {};
    },

    initialize(band, timeline) {
      this._band = band;
      this._timeline = timeline;
      this._eventLayer = null;
      this._lineLayer = null;
      this._eventIdToElmt = null;
    },
    getType() {
      return 'labeless';
    },
    addOnSelectListener(listener) {
      this._hub.on('select', listener);
    },
    removeOnSelectListener(listener) {
      this._hub.off('select', listener);
    },
    addEventPaintListener(listener) {
      this._hub.on('event-paint', listener);
    },
    removeEventPaintListener(listener) {
      this._hub.off('event-paint', listener);
    },
    getFilterMatcher() {
      return this._filterMatcher;
    },
    setFilterMatcher(filterMatcher) {
      this._filterMatcher = filterMatcher;
    },

    showBubble() {
      /*empty*/
    },

    softPaint() {
      this.paint();
    },
    paint() {
      // Paints the events for a given section of the band--what is
      // visible on screen and some extra.
      var eventSource = this._band.getEventSource();
      if (eventSource == null) {
        return;
      }

      this._eventIdToElmt = {};
      this._fireEventPaintListeners('paintStarting', null, null);
      this._prepareForPainting();

      var eventTheme = this._params.theme.event;
      var trackHeight = Math.max(eventTheme.track.height, eventTheme.tape.height);
      var metrics = {
        trackOffset        : eventTheme.track.offset,
        trackHeight        : trackHeight,
        trackGap           : eventTheme.track.gap,
        trackIncrement     : trackHeight + eventTheme.track.gap,
        iconWidth          : eventTheme.instant.iconWidth,
        iconHeight         : eventTheme.instant.iconHeight,
        labelWidth         : eventTheme.label.width,
        maxLabelChar       : eventTheme.label.maxLabelChar,
        impreciseIconMargin: eventTheme.instant.impreciseIconMargin
      };
      metrics.maxNumberOfVisibleTracks = Math.ceil(
        (this._band.getViewWidth() - metrics.trackOffset) / metrics.trackIncrement
      );

      var minDate = this._band.getMinDate();
      var maxDate = this._band.getMaxDate();

      var filterMatcher = this._filterMatcher || emptyFunctionReturningTrue;
      var iterator = eventSource.getEventIterator(minDate, maxDate);
      var iteratorStartIndex = iterator._currentIndex;
      var evt;
      while (iterator.hasNext()) {
        evt = iterator.next();
        if (evt.isInMap() && filterMatcher(evt)) {
          this.paintEvent(evt, metrics, this._params.theme);
        }
      }

      // restart the iterator
      iterator._currentIndex = iteratorStartIndex - 1;
      iterator._findNext();
      //

      while (iterator.hasNext()) {
        evt = iterator.next();
        if (!evt.isInMap() && filterMatcher(evt)) {
          this.paintEvent(evt, metrics, this._params.theme);
        }
      }

      this._lineLayer.style.display = 'block';
      this._eventLayer.style.display = 'block';
      // update the band object for max number of tracks in this section of the ether
      this._band.updateEventTrackInfo(
        Math.max(this._durationTracks.length, this._instantTracks.length),
        metrics.trackIncrement
      );
      this._fireEventPaintListeners('paintEnded', null, null);
    },
    paintEvent(evt, metrics, theme) {
      this.paintDurationEvent(evt, metrics, theme);
    },
    paintInstantEvent(evt, metrics, theme) {
      var startDate = evt.getStart();
      var startPixel = Math.round(this._band.dateToPixelOffset(startDate));
      var iconRightEdge = Math.round(startPixel + metrics.iconWidth / 2);
      var iconLeftEdge = Math.round(startPixel - metrics.iconWidth / 2);

      var track = this._findFreeTrack(evt, iconLeftEdge, iconRightEdge, this._instantTracks);

      if (track >= metrics.maxNumberOfVisibleTracks) {
        return;
      }

      var iconElmtData = this._paintEventIcon(evt, track, iconLeftEdge, metrics, theme);
      var els = [iconElmtData.elmt];

      var self = this;
      SimileAjax.DOM.registerEvent(
        iconElmtData.elmt,
        'click',
        function clickHandler(elmt, domEvt, target) {
          return self._onClickEvent(iconElmtData.elmt, domEvt, evt);
        }
      );

      this._fireEventPaintListeners('paintedEvent', evt, els);
      this._eventIdToElmt[evt.getID()] = iconElmtData.elmt;
      if (this._instantTracks[track]) {
        addInterval(this._instantTracks[track], iconLeftEdge, iconRightEdge);
      } else {
        this._instantTracks[track] = [[iconLeftEdge, iconRightEdge]];
      }
    },
    paintDurationEvent(evt, metrics, theme) {
      var startDate = evt.getStart();
      var endDate = evt.getEnd();
      var startPixel = Math.round(this._band.dateToPixelOffset(startDate));
      var endPixel = Math.round(this._band.dateToPixelOffset(endDate));

      if (Math.abs(endPixel - startPixel) < metrics.iconWidth / 2) {
        this.paintInstantEvent(evt, metrics, theme);
        return;
      }

      var leftEdge = Math.min(startPixel, endPixel);
      var rightEdge = Math.max(startPixel, endPixel);
      var track = this._findFreeTrack(evt, leftEdge, rightEdge, this._durationTracks);

      if (track >= metrics.maxNumberOfVisibleTracks) {
        return;
      }

      var color = evt.getColor();
      color = color == null ? theme.event.duration.color : color;

      var tapeElmtData = this._paintEventTape(evt, track, leftEdge, rightEdge, color, 100, metrics, theme);
      var els = [tapeElmtData.elmt];

      var self = this;
      SimileAjax.DOM.registerEvent(
        tapeElmtData.elmt,
        'click',
        function clickHandler(elmt, domEvt, target) {
          return self._onClickEvent(tapeElmtData.elmt, domEvt, evt);
        }
      );

      this._fireEventPaintListeners('paintedEvent', evt, els);
      this._eventIdToElmt[evt.getID()] = tapeElmtData.elmt;
      if (this._durationTracks[track]) {
        addInterval(this._durationTracks[track], leftEdge, rightEdge);
      } else {
        this._durationTracks[track] = [[leftEdge, rightEdge]];
      }
    },

    _prepareForPainting() {
      // Remove everything previously painted: line and event layers.
      // Prepare blank layers for painting.
      var band = this._band;

      this._durationTracks = [];
      this._instantTracks = [];

      if (this._lineLayer != null) {
        band.removeLayerDiv(this._lineLayer);
      }
      this._lineLayer = band.createLayerDiv(110, 'timeline-band-lines');
      this._lineLayer.style.display = 'none';

      /* The eventLayer is the element that receives the touches when a user
       * wants to drag the timeline band.
       * But if we remove it while the user is dragging, the touches associated
       * with this element will disappear (stop firing events).
       * So we should reuse it instead of removing it and creating a new one.
       */
      if (this._eventLayer) {
        removeAllChildrenOfNode(this._eventLayer);
      } else {
        this._eventLayer = band.createLayerDiv(115, 'timeline-band-events');
        var self = this;
        SimileAjax.DOM.registerEvent(
          this._eventLayer,
          'click',
          function clickHandler(elmt, domEvt, target) {
            return self._onClick(domEvt);
          }
        );
      }
      this._eventLayer.style.display = 'none';
    },
    _encodeEventElID(elType, evt) {
      return Timeline.EventUtils.encodeEventElID(this._timeline, this._band, elType, evt);
    },
    _findFreeTrack(event, leftEdge, rightEdge, tracks) {
      var trackAttribute = event.getTrackNum();
      if (trackAttribute != null) {
        return trackAttribute; // early return since event includes track number
      }

      // normal case: find an open track
      var i;
      for (i = 0; i < tracks.length; i++) {
        var t = tracks[i];
        if (!hasIntervalsInRange(t, leftEdge, rightEdge)) {
          break;
        }
      }
      return i;
    },
    _paintEventIcon(evt, iconTrack, left, metrics, theme) {
      var top = Math.round(metrics.trackOffset + iconTrack * metrics.iconHeight / 2);
      var iconDiv = this._timeline.getDocument().createElement('div');
      iconDiv.className = this._getElClassName('timeline-event-icon', evt, 'icon');
      iconDiv.id = this._encodeEventElID('icon', evt);
      iconDiv.style.left = left + 'px';
      iconDiv.style.top = top + 'px';
      iconDiv.style.width = metrics.iconWidth + 'px';
      iconDiv.style.height = metrics.iconHeight + 'px';

      if (evt._title != null) {
        iconDiv.title = evt._title;
      }

      this._eventLayer.appendChild(iconDiv);

      return {
        left  : left,
        top   : top,
        width : metrics.iconWidth,
        height: metrics.iconHeight,
        elmt  : iconDiv
      };
    },
    _paintEventTape(
      evt, tapeTrack, startPixel, endPixel, color, opacity, metrics, theme) {

      // On Firefox (tested on v89) the width of an element is capped at
      // 17895700px, other browser might have a different cap.
      // This might be related to https://stackoverflow.com/a/26639355/2168921
      // and https://bugzilla.mozilla.org/show_bug.cgi?id=791428
      //
      // So as a workaround we clip the tape to the band's totalViewLength
      var totalViewLength = this._band.getTotalViewLength();

      if (startPixel < 0) {
        startPixel = 0;
      }
      if (endPixel > totalViewLength) {
        endPixel = totalViewLength;
      }

      var tapeWidth = endPixel - startPixel;
      var tapeHeight = theme.event.tape.height;
      var top = metrics.trackOffset + tapeTrack * metrics.trackIncrement;

      var tapeDiv = this._timeline.getDocument().createElement('div');
      tapeDiv.className = this._getElClassName('timeline-event-tape', evt, 'tape');
      tapeDiv.id = this._encodeEventElID('tape', evt);
      tapeDiv.style.left = startPixel + 'px';
      tapeDiv.style.width = tapeWidth + 'px';
      tapeDiv.style.height = tapeHeight + 'px';
      tapeDiv.style.top = top + 'px';

      if (evt._title != null) {
        tapeDiv.title = evt._title;
      }

      if (color != null) {
        tapeDiv.style.backgroundColor = color;
      }

      var backgroundImage = evt.getTapeImage();
      var backgroundRepeat = evt.getTapeRepeat();
      backgroundRepeat = backgroundRepeat == null ? 'repeat' : backgroundRepeat;
      if (backgroundImage != null) {
        tapeDiv.style.backgroundImage = 'url(' + backgroundImage + ')';
        tapeDiv.style.backgroundRepeat = backgroundRepeat;
      }

      SimileAjax.Graphics.setOpacity(tapeDiv, opacity);

      this._eventLayer.appendChild(tapeDiv);

      return {
        left  : startPixel,
        top   : top,
        width : tapeWidth,
        height: tapeHeight,
        elmt  : tapeDiv
      };
    },
    _getElClassName(elClassName, evt, prefix) {
      // Prefix and '_' is added to the event's classname. Set to null for no prefix
      var evtClassname = evt.getClassName();
      var pieces = [];

      if (evtClassname) {
        if (prefix) {
          pieces.push(prefix + '-' + evtClassname + ' ');
        }
        pieces.push(evtClassname + ' ');
      }
      pieces.push(elClassName);
      return (pieces.join(''));
    },
    _onClick(domEvt) {
      this._hub.emit('select', null);

      domEvt.cancelBubble = true;
      SimileAjax.DOM.cancelEvent(domEvt);
      return false;
    },
    _onClickEvent(icon, domEvt, evt) {
      var eventID = evt.getID();
      this._hub.emit('select', eventID);

      domEvt.cancelBubble = true;
      SimileAjax.DOM.cancelEvent(domEvt);
      return false;
    },
    _fireEventPaintListeners(op, evt, els) {
      this._hub.emit('event-paint', this._band, op, evt, els);
    },
  },
]);
