/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  NONPASSIVE,
  PASSIVE,
  PASSIVE_CAPTURING,
  emptyFunction,
  inRange,
  inRangeInclusive,
  wrapMethods,
  debounce,
  throttleOncePerAnimationFrame,
  clamp,
  getXYMovement,
  getXYLength,
  bisectLeft,
  HIDKeys,
  keyFromEvent,
  isMacOS,
  preventDefaultAndStopEvent,
  DOMEvents,
  EventsHub,
  Syntool,
  decoratedFn,
} from '../../syntool.js';

/**
 * @typedef {[number, number]} XY
 */
/**
 * @callback KeyHandler
 * @param {KeyboardEvent} event
 * @returns {void}
 */
/**
 * @typedef {Object} EventHub
 * @property {(type: string, callback: (event: *) => void, options?: *) => void} addEventListener
 * @property {(type: string, callback: (event: *) => void, options?: *) => void} removeEventListener
 */

/** @type {XY} */
var INVALID_LOCATION = [NaN, NaN];

var keyDownHub = new EventsHub();
var keyDownListenersCount = 0;

/**
 * @param {KeyboardEvent} event
 * @returns {void}
 */
function handleGlobalKeyDown(event) {
  keyDownHub.emit(keyFromEvent(event));
}
/**
 * @param {string} key
 * @param {KeyHandler} handler
 * @param {*} [thisArg]
 * @returns {void}
 */
function onKeyDown(key, handler, thisArg) {
  if (keyDownListenersCount === 0) {
    window.addEventListener('keydown', handleGlobalKeyDown, PASSIVE_CAPTURING);
  }
  keyDownHub.on(key, handler, thisArg);
  keyDownListenersCount++;
}
/**
 * @param {string} key
 * @param {KeyHandler} handler
 * @param {*} [thisArg]
 * @returns {void}
 */
function offKeyDown(key, handler, thisArg) {
  keyDownHub.off(key, handler, thisArg);
  keyDownListenersCount--;
  if (keyDownListenersCount === 0) {
    window.removeEventListener(
      'keydown',
      handleGlobalKeyDown,
      PASSIVE_CAPTURING
    );
  }
}

/**
 * @typedef {'x'|'y'|'xy'} MovementDirection
 */

/**
 * @typedef {{clientX: number, clientY: number}} Pointer
 */

/**
 * @param {Pointer} pointer
 * @returns {XY}
 */
function getLocationFromPointer(pointer) {
  return [pointer.clientX, pointer.clientY];
}
/**
 * @param {MovementDirection} direction
 * @param {XY} l1
 * @param {XY} l2
 * @returns {number|XY}
 */
function getMovementInDirection(direction, l1, l2) {
  switch (direction) {
  case 'x':
    return l2[0] - l1[0];
  case 'y':
    return l2[1] - l1[1];
  default:
  case 'xy':
    return getXYMovement(l1, l2);
  }
}
/**
 * @param {number|XY} m
 * @returns {number}
 */
function getDistance(m) {
  return Array.isArray(m) ? getXYLength(m) : Math.abs(m);
}

/**
 * @param {HTMLElement} node
 * @param {string} selector
 * @returns {EventHub}
 */
function eventHubForSelector(node, selector) {
  const delegator = DOMEvents(node);
  return {
    addEventListener(eventType, callback) {
      delegator.on(eventType, selector, callback);
    },
    removeEventListener(eventType, callback) {
      delegator.off(eventType, selector, callback);
    },
  };
}

/**
 * @typedef {Object} PartialDragGestureOptions
 * @property {MovementDirection} [direction='xy']
 * @property {number} [moveTolerance=0]
 */
/**
 * @typedef {Object} DragGestureOptions
 * @property {MovementDirection} direction
 * @property {number} moveTolerance
 */
/**
 * @typedef {Object} DragGestureChangePropsWithoutPointer
 * @property {DragGestureState} state
 * @property {Event|null} event
 * @property {() => void} stop
 * @property {XY} startLocation
 */
/**
 * @typedef {Object} DragGestureChangePropsWithPointer
 * @property {DragGestureState} state
 * @property {Event|null} event
 * @property {Pointer} pointer
 * @property {() => void} stop
 * @property {XY} startLocation
 * @property {XY} currentLocation
 * @property {number|XY} movement
 * @property {number} distance
 */
/**
 * @typedef {DragGestureChangePropsWithoutPointer | DragGestureChangePropsWithPointer} DragGestureChangeProps
 */
/**
 * @callback DragGestureOnChange
 * @param {DragGestureChangeProps} props
 * @returns {void}
 */
/** @typedef {'possible'|'start'|'change'|'end'|'cancel'} DragGestureState */

class DragGesture {
  /**
   * @param {?PartialDragGestureOptions} [options=null]
   * @param {DragGestureOnChange} onChange
   * @param {*} [thisArg]
   */
  constructor(options, onChange, thisArg) {
    if (typeof options === 'function') {
      thisArg = onChange;
      onChange = options;
      options = null;
    }

    /** @type {DragGestureOptions} */
    var opts = Object.assign({
      direction    : 'xy',
      moveTolerance: 0,
    }, options);

    this.options = opts;
    this.onChange = onChange;
    this.thisArg = thisArg;

    this.started = false;
    this.startLocation = INVALID_LOCATION;
    /** @type {*} */
    this.target = null;
    this.getMovement = getMovementInDirection.bind(null, opts.direction);
  }

  /**
   * @returns {void}
   */
  cancel() {
    if (this.startLocation !== INVALID_LOCATION) {
      this.startLocation = INVALID_LOCATION;
      this.started = false;
      this.emitChange('cancel', null);
      this.target = null;
    }
  }

  /**
   * @param {DragGestureState} state
   * @param {?Event} event
   * @param {Pointer} [pointer]
   * @returns {void}
   */
  emitChange(state, event, pointer) {
    const startLocation = this.startLocation;
    const stop = this.cancel.bind(this);
    let props;
    if (pointer) {
      const currentLocation = getLocationFromPointer(pointer);
      const movement = this.getMovement(this.startLocation, currentLocation);
      const distance = getDistance(movement);
      props = {
        state,
        event,
        pointer,
        stop,

        startLocation,
        currentLocation,
        movement,
        distance,
      };
    } else {
      props = {state, event, stop, startLocation};
    }
    this.onChange.call(this.thisArg || this.target, props);
  }

  /**
   * @param {HTMLElement} target
   * @param {Event} event
   * @param {*} [pointer]
   * @returns {void}
   */
  handleStart(target, event, pointer) {
    if (this.startLocation === INVALID_LOCATION) {
      this.startLocation = getLocationFromPointer(pointer);
      this.target = target;
      this.started = false;
      this.emitChange('possible', event, pointer);
      if (this.options.moveTolerance <= 0) {
        if (event.cancelable) {
          event.preventDefault();
        }
        event.stopPropagation();

        this.emitChange('start', event, pointer);
        this.started = true;
      }
    }
  }

  /**
   * @param {HTMLElement} target
   * @param {Event} event
   * @param {*} [pointer]
   * @returns {void}
   */
  handleMove(target, event, pointer) {
    if (this.startLocation !== INVALID_LOCATION) {
      var currentLocation = getLocationFromPointer(pointer);
      var distance = getDistance(
        this.getMovement(this.startLocation, currentLocation)
      );

      if (this.started || distance >= this.options.moveTolerance) {
        if (event.cancelable) {
          event.preventDefault();
        }
        event.stopPropagation();

        this.emitChange(this.started ? 'change' : 'start', event, pointer);
        this.started = true;
      }
    }
  }

  /**
   * @param {HTMLElement} target
   * @param {Event} event
   * @param {*} [pointer]
   * @returns {void}
   */
  handleEnd(target, event, pointer) {
    if (this.started) {
      if (event.cancelable) {
        event.preventDefault();
      }
      event.stopPropagation();

      this.emitChange('end', event, pointer);
    } else {
      this.emitChange('cancel', event);
    }
    this.startLocation = INVALID_LOCATION;
    this.target = null;
  }

  /**
   * @param {HTMLElement} target
   * @param {Event} event
   * @returns {void}
   */
  handleCancel(target, event) {
    this.emitChange('cancel', event);
    this.startLocation = INVALID_LOCATION;
    this.target = null;
  }
}

/**
 * @typedef {PartialDragGestureOptions} PartialTouchDragOptions
 * @property {number} [moveTolerance=1]
 */
/**
 * @param  {EventHub} eventHub
 * @param  {?PartialTouchDragOptions} [options=null]
 * @param  {DragGestureOnChange} onChange
 * @param  {*} [thisArg]
 * @return {() => void}
 */
function touchDrag(eventHub, options, onChange, thisArg) {
  options = Object.assign({moveTolerance: 1}, options);
  var dragGesture = new DragGesture(options, onChange, thisArg);

  onKeyDown(HIDKeys.escape, dragGesture.cancel, dragGesture);
  eventHub.addEventListener('touchstart', handleTouchStart, NONPASSIVE);
  eventHub.addEventListener('touchmove', handleTouchMove, NONPASSIVE);
  eventHub.addEventListener('touchend', handleTouchEnd, NONPASSIVE);
  eventHub.addEventListener('touchcancel', handleTouchCancel, NONPASSIVE);

  return remove;

  function remove() {
    dragGesture.cancel();
    offKeyDown(HIDKeys.escape, dragGesture.cancel, dragGesture);
    eventHub.removeEventListener('touchstart', handleTouchStart, NONPASSIVE);
    eventHub.removeEventListener('touchmove', handleTouchMove, NONPASSIVE);
    eventHub.removeEventListener('touchend', handleTouchEnd, NONPASSIVE);
    eventHub.removeEventListener('touchcancel', handleTouchCancel, NONPASSIVE);
  }

  /**
   * @this {*}
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchStart(event) {
    dragGesture.handleStart(this, event, event.targetTouches[0]);
  }
  /**
   * @this {*}
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchMove(event) {
    dragGesture.handleMove(this, event, event.targetTouches[0]);
  }
  /**
   * @this {*}
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchEnd(event) {
    dragGesture.handleEnd(this, event, event.changedTouches[0]);
  }
  /**
   * @this {*}
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchCancel(event) {
    dragGesture.handleCancel(this, event);
  }
}

/**
 * @typedef {PartialDragGestureOptions} PartialMouseDragOptions
 * @property {number} [moveTolerance=3]
 */
/**
 * @param  {EventHub} eventHub
 * @param  {?PartialMouseDragOptions} [options=null]
 * @param  {DragGestureOnChange} onChange
 * @param  {*} [thisArg]
 * @return {() => void}
 */
function mouseDrag(eventHub, options, onChange, thisArg) {
  options = Object.assign({moveTolerance: 3}, options);
  var dragGesture = new DragGesture(options, onChange, thisArg);

  wrapMethods(dragGesture, {
    cancel() {
      if (dragGesture.startLocation !== INVALID_LOCATION) {
        removeMouseListeners();
      }
      window.original();
    },
  });

  onKeyDown(HIDKeys.escape, dragGesture.cancel, dragGesture);
  eventHub.addEventListener('mousedown', handleMouseDown, NONPASSIVE);

  return remove;

  function remove() {
    dragGesture.cancel();
    offKeyDown(HIDKeys.escape, dragGesture.cancel, dragGesture);
    eventHub.removeEventListener('mousedown', handleMouseDown, NONPASSIVE);
  }

  function addMouseListeners() {
    window.addEventListener('mousemove', handleMouseMove, NONPASSIVE);
    window.addEventListener('mouseup', handleMouseUp, NONPASSIVE);
  }
  function removeMouseListeners() {
    window.removeEventListener('mousemove', handleMouseMove, NONPASSIVE);
    window.removeEventListener('mouseup', handleMouseUp, NONPASSIVE);
  }

  /**
   * @this {*}
   * @param {MouseEvent} event
   * @returns {void}
   */
  function handleMouseDown(event) {
    if (event.button === 0) {
      addMouseListeners();
      dragGesture.handleStart(this, event, event);
    }
  }
  /**
   * @this {*}
   * @param {MouseEvent} event
   * @returns {void}
   */
  function handleMouseMove(event) {
    dragGesture.handleMove(this, event, event);
  }
  /**
   * @this {*}
   * @param {MouseEvent} event
   * @returns {void}
   */
  function handleMouseUp(event) {
    if (event.button === 0) {
      removeMouseListeners();
      dragGesture.handleEnd(this, event, event);
    }
  }
}

/**
 * @typedef {Object} PartialDragOptions
 * @property {boolean|PartialTouchDragOptions} [touch=true]
 * @property {boolean|PartialMouseDragOptions} [mouse=true]
 */
/**
 * @typedef {Object} DragOptions
 * @property {boolean|DragGestureOptions} touch
 * @property {boolean|DragGestureOptions} mouse
 */
/**
 * @param {EventHub} eventHub
 * @param {?PartialDragOptions} [options=null]
 * @param {DragGestureOnChange} onChange
 * @param {*} [thisArg]
 * @return {() => void}
 */
function drag(eventHub, options, onChange, thisArg) {
  if (typeof options === 'function') {
    thisArg = onChange;
    onChange = options;
    options = null;
  }
  options = Object.assign({
    touch: true,
    mouse: true,
  }, options);

  const removeTouchDrag = (
    options.touch
    ? touchDrag(
      eventHub,
      Object.assign({}, options, options.touch),
      onChange,
      thisArg
    )
    : emptyFunction
  );
  const removeMouseDrag = (
    options.mouse
    ? mouseDrag(
      eventHub,
      Object.assign({}, options, options.mouse),
      onChange,
      thisArg
    )
    : emptyFunction
  );

  return function remove() {
    removeTouchDrag();
    removeMouseDrag();
  };
}
drag.DragGesture = DragGesture;
drag.touch = touchDrag;
drag.mouse = mouseDrag;

/**
 * @typedef {Object} PartialTapOptions
 * @property {number} [maxTaps=Infinity]
 * @property {number} [moveTolerance=20]
 * @property {number} [tapCompletionTimeout=350]
 * @property {number} [nextTapToStartTimeout=750]
 */
/**
 * @typedef {Object} TapOptions
 * @property {number} maxTaps
 * @property {number} moveTolerance
 * @property {number} tapCompletionTimeout
 * @property {number} nextTapToStartTimeout
 */
/**
 * @typedef {Object} TapProps
 * @property {TouchEvent} event
 * @property {Touch} pointer
 * @property {number} count
 * @property {() => void} stop
 */
/**
 * @callback TapOnTap
 * @param {TapProps} props
 * @returns {void}
 */

/**
 * @param {EventHub} eventHub
 * @param {?PartialTapOptions} [options=null]
 * @param {TapOnTap} onTap
 * @param {*} [thisArg]
 * @return {() => void}
 */
function tap(eventHub, options, onTap, thisArg) {
  if (typeof options === 'function') {
    thisArg = onTap;
    onTap = options;
    options = null;
  }
  /** @type {TapOptions} */
  var opts = Object.assign({
    maxTaps              : Infinity,
    moveTolerance        : 20,
    tapCompletionTimeout : 350,
    nextTapToStartTimeout: 750,
  }, options);

  var moved = false;
  var startLocation = INVALID_LOCATION;
  var startTime = 0;
  var lastTime = 0;
  var count = 0;
  /** @type {*} */
  var target = null;
  var getMovement = getMovementInDirection.bind(null, 'xy');

  eventHub.addEventListener('touchstart', handleTouchStart, NONPASSIVE);
  eventHub.addEventListener('touchmove', handleTouchMove, NONPASSIVE);
  eventHub.addEventListener('touchend', handleTouchEnd, NONPASSIVE);
  eventHub.addEventListener('touchcancel', handleTouchCancel, NONPASSIVE);

  return remove;

  function remove() {
    cancel();
    eventHub.removeEventListener('touchstart', handleTouchStart, NONPASSIVE);
    eventHub.removeEventListener('touchmove', handleTouchMove, NONPASSIVE);
    eventHub.removeEventListener('touchend', handleTouchEnd, NONPASSIVE);
    eventHub.removeEventListener('touchcancel', handleTouchCancel, NONPASSIVE);
  }
  function cancel() {
    startLocation = INVALID_LOCATION;
    lastTime = 0;
    count = 0;
    target = null;
  }

  /**
   * @this {HTMLElement}
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchStart(event) {
    if (startLocation === INVALID_LOCATION) {
      startLocation = getLocationFromPointer(event.targetTouches[0]);
      target = this;
      moved = false;
      startTime = Date.now();
    }
  }
  /**
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchMove(event) {
    var touch = event.targetTouches[0];
    if (startLocation !== INVALID_LOCATION) {
      var currentLocation = getLocationFromPointer(touch);
      var distance = getDistance(getMovement(startLocation, currentLocation));

      if (moved || distance >= opts.moveTolerance) {
        moved = true;
      }
    }
  }
  /**
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchEnd(event) {
    if (moved) {
      count = 0;
    } else {
      var now = Date.now();
      if (now - startTime < opts.tapCompletionTimeout) {
        if (
          count < opts.maxTaps &&
          now - lastTime < opts.nextTapToStartTimeout
        ) {
          count++;
        } else {
          count = 1;
        }
        onTap.call(thisArg || target, {
          event  : event,
          pointer: event.changedTouches[0],
          count  : count,
          stop   : cancel,
        });
        lastTime = now;

        if (event.cancelable) {
          event.preventDefault();
        }
        event.stopPropagation();
      } else {
        count = 0;
        lastTime = 0;
      }
    }
    startLocation = INVALID_LOCATION;
  }
  /**
   * @param {TouchEvent} event
   * @returns {void}
   */
  function handleTouchCancel(event) {
    cancel();
  }
}

/**
 * @typedef {Object} PartialReorderOptions
 * @property {boolean|PartialTouchDragOptions} [touch=true]
 * @property {boolean|PartialMouseDragOptions} [mouse=true]
 *
 * @property {(handle: HTMLElement) => HTMLElement} [getElementForHandle]
 * @property {(index: number) => boolean} [canReorderElementAtIndex]
 * @property {() => number} [getElementsCount]
 * @property {(element: HTMLElement) => number} [getIndexOfElement]
 * @property {(index: number) => number} [getOffsetOfElementAtIndex]
 *
 * @property {(index: number, toIndex: number, translation: number) => void} [updateElementStyleAtIndex]
 * @property {(index: number) => void} [resetElementStyleAtIndex]
 * @property {(event: KeyboardEvent) => number} [deltaForKeyDown]
 * @property {(element: HTMLElement) => void} [onStart]
 * @property {(element: HTMLElement) => void} [onEnd]
 */
/**
 * @typedef {Object} ReorderOptions
 * @property {boolean|PartialTouchDragOptions} [touch]
 * @property {boolean|PartialMouseDragOptions} [mouse]
 *
 * @property {(handle: HTMLElement) => HTMLElement} getElementForHandle
 * @property {(index: number) => boolean} canReorderElementAtIndex
 * @property {() => number} getElementsCount
 * @property {(element: HTMLElement) => number} getIndexOfElement
 * @property {(index: number) => number} getOffsetOfElementAtIndex
 *
 * @property {(index: number, toIndex: number, translation: number) => void} updateElementStyleAtIndex
 * @property {(index: number) => void} resetElementStyleAtIndex
 * @property {(event: KeyboardEvent) => number} deltaForKeyDown
 * @property {(element: HTMLElement) => void} onStart
 * @property {(element: HTMLElement) => void} onEnd
 */
/**
 * @callback ReorderOnChange
 * @param {number} fromIndex
 * @param {number} toIndex
 * @returns {void}
 */

/**
 * A gesture that handles reordering elements inside a container with drag-n-drop.
 *
 * @param  {HTMLElement} container          The container element
 * @param  {string|EventHub} handleEventHubOrSelector The selector or event hub of the drag handle
 * @param  {?PartialReorderOptions} [options=null] Customization options
 * @param  {ReorderOnChange} onChange       A function called when the order of elements change
 * @param  {*} [thisArg]                    The context with which `onChange` is called
 * @return {() => void} A function that remove the reorder gesture when called
 */
function reorder(container, handleEventHubOrSelector, options, onChange, thisArg) {
  if (typeof options === 'function') {
    thisArg = onChange;
    onChange = options;
    options = null;
  }
  var handleEventHub = (
    typeof handleEventHubOrSelector === 'string'
    ? eventHubForSelector(container, handleEventHubOrSelector)
    : handleEventHubOrSelector
  );
  /** @type {ReorderOptions} */
  var opts = Object.assign({
    direction: 'y',

    getElementsCount() {
      return container.childElementCount;
    },

    /**
     * @param {HTMLElement} element
     */
    getIndexOfElement(element) {
      return Array.prototype.indexOf.call(container.children, element);
    },
    /**
     * @param {number} index
     * @returns {number}
     */
    getOffsetOfElementAtIndex(index) {
      const element = container.children[index];
      return element.offsetTop + element.offsetHeight / 2;
    },
    /**
     * @param {number} index
     * @returns {boolean}
     */
    canReorderElementAtIndex(index) {
      return true;
    },
    /**
     * @param {HTMLElement} handle
     * @returns {HTMLElement}
     */
    getElementForHandle(handle) {
      return handle;
    },

    /**
     * @param {number} index
     * @param {number} toIndex
     * @param {number} translation
     */
    updateElementStyleAtIndex(index, toIndex, translation) {
      container.children[index].style.transform = (
        translation === 0 ? '' : 'translate3d(0,' + translation + 'px,0)'
      );
    },
    /**
     * @param {number} index
     */
    resetElementStyleAtIndex(index) {
      container.children[index].style.transform = '';
    },

    /**
     * @param {KeyboardEvent} event
     */
    deltaForKeyDown(event) {
      if (Syntool.keyboardManager.isShiftDown()) {
        switch (keyFromEvent(event)) {
        case HIDKeys.up:
          return -1;
        case HIDKeys.down:
          return 1;
        default:
        }
      }
      return 0;
    },

    onStart() {
      // empty
    },
    onEnd() {
      // empty
    },
  }, options);

  var getElementForHandle = opts.getElementForHandle;
  var canReorderElementAtIndex = opts.canReorderElementAtIndex;
  var getElementsCount = opts.getElementsCount;
  var getIndexOfElement = opts.getIndexOfElement;
  var getOffsetOfElementAtIndex = opts.getOffsetOfElementAtIndex;
  var updateElementStyleAtIndex = opts.updateElementStyleAtIndex;
  var resetElementStyleAtIndex = opts.resetElementStyleAtIndex;
  var deltaForKeyDown = opts.deltaForKeyDown;
  var onStart = opts.onStart;
  var onEnd = opts.onEnd;

  /** @type {() => void} */
  var handleThrottledContainerScroll = throttleOncePerAnimationFrame(
    handleContainerScroll
  );

  /** @type {() => void} */
  var removeDragGesture = emptyFunction;
  /** @type {number|undefined} */
  var autoScrollInterval;
  var reordering = false;
  /** @type {HTMLElement} */
  var currentElement;
  /** @type {number} */
  var currentIndex;
  /** @type {number} */
  var currentRealTop;
  /** @type {number} */
  var currentHeight;
  /** @type {number} */
  var minMovementInVisible;
  /** @type {number} */
  var maxMovementInVisible;
  /** @type {number} */
  var minMovement;
  /** @type {number} */
  var maxMovement;
  /** @type {number} */
  var startLocation;
  /** @type {number} */
  var startClientLocation;
  /** @type {number} */
  var lastClientLocation;
  /** @type {{index: number[], y: number[]}} */
  var dropSlots;

  container.addEventListener(
    'scroll',
    handleThrottledContainerScroll,
    PASSIVE
  );
  handleEventHub.addEventListener(
    'keydown',
    handleKeyDown,
    NONPASSIVE
  );
  removeDragGesture = drag(
    handleEventHub,
    opts,
    handleDragGestureChange
  );

  return remove;

  function remove() {
    container.removeEventListener(
      'scroll',
      handleThrottledContainerScroll,
      PASSIVE
    );
    handleEventHub.removeEventListener(
      'keydown',
      handleKeyDown,
      NONPASSIVE
    );
    removeDragGesture();
  }

  function handleContainerScroll() {
    if (reordering) {
      reorderUpdate('change', true);
    }
  }
  /**
   * @this {HTMLElement}
   * @param {KeyboardEvent} event
   */
  function handleKeyDown(event) {
    if (reordering) {
      return;
    }

    var delta = deltaForKeyDown(event);
    if (delta === 0) {
      return;
    }

    var element = getElementForHandle(this);
    var index = getIndexOfElement(element);
    var toIndex = index + delta;

    if (
      inRange(toIndex, 0, getElementsCount()) &&
      canReorderElementAtIndex(toIndex)
    ) {
      onChange.call(thisArg, index, toIndex);
    }

    preventDefaultAndStopEvent(event);
  }
  /**
   * @this {HTMLElement}
   * @param {DragGestureChangeProps} props
   */
  function handleDragGestureChange(props) {
    switch (props.state) {
    case 'possible':
      var containerRealHeight = container.scrollHeight;
      var containerVisibleOffset = container.scrollTop;
      var containerVisibleHeight = container.clientHeight;
      var containerClientRealTop = (
        container.getBoundingClientRect().top - containerVisibleOffset
      );

      currentElement = getElementForHandle(this);
      currentIndex = getIndexOfElement(currentElement);
      currentRealTop = currentElement.offsetTop;
      currentHeight = currentElement.offsetHeight;
      minMovementInVisible = containerVisibleOffset - currentRealTop;
      maxMovementInVisible = (
        minMovementInVisible + containerVisibleHeight - currentHeight
      );
      minMovement = -currentRealTop;
      maxMovement = minMovement + containerRealHeight - currentHeight;
      startClientLocation = props.startLocation[1];
      startLocation = startClientLocation - containerClientRealTop;

      dropSlots = {
        index: [],
        y    : [],
      };
      var elementsCount = getElementsCount();
      for (var index = 0; index < elementsCount; index++) {
        if (index !== currentIndex && canReorderElementAtIndex(index)) {
          dropSlots.index.push(index);
          dropSlots.y.push(getOffsetOfElementAtIndex(index));
        }
      }
      dropSlots.index.push(elementsCount);
      dropSlots.y.push(containerRealHeight);
      reordering = true;
      onStart.call(thisArg, currentElement);
      break;
    case 'end':
      onChange.call(thisArg, currentIndex, getIndexToMoveTo(getMovement()));
      /* fallthrough */
    case 'cancel':
      reordering = false;
      onEnd.call(thisArg, currentElement);
      break;
    default:
    }

    lastClientLocation = (props.currentLocation || props.startLocation)[1];
    reorderUpdate(props.state);
  }
  function getMovementInClient() {
    return lastClientLocation - startClientLocation;
  }
  function getMovement() {
    var containerClientRealTop = (
      container.getBoundingClientRect().top - container.scrollTop
    );
    var currentLocation = lastClientLocation - containerClientRealTop;
    return currentLocation - startLocation;
  }
  /**
   * @param {number} movement
   */
  function getIndexToMoveTo(movement) {
    var currentTop = currentRealTop + movement;
    var y = currentTop + (movement > 0 ? currentHeight : 0);
    var toSlotIndex = bisectLeft(dropSlots.y, y);
    var toIndex = dropSlots.index[
      clamp(toSlotIndex, 0, dropSlots.index.length - 1)
    ];

    return currentIndex < toIndex ? toIndex - 1 : toIndex;
  }
  /**
   * @param {DragGestureState} state
   * @param {boolean} [scrolling]
   */
  function reorderUpdate(state, scrolling) {
    var elementsCount = getElementsCount();
    var index;

    switch (state) {
    case 'end':
    case 'cancel':
      for (index = 0; index < elementsCount; index++) {
        resetElementStyleAtIndex(index);
      }
      clearInterval(autoScrollInterval);
      break;
    default:
      var movement = getMovement();
      var toIndex = getIndexToMoveTo(movement);

      for (index = 0; index < elementsCount; index++) {
        var translation = 0;
        var newIndex = index;

        if (index === currentIndex) {
          translation = clamp(movement, minMovement, maxMovement);
          newIndex = toIndex;
        } else if (inRangeInclusive(index, toIndex, currentIndex)) {
          translation = currentHeight;
          newIndex = index + 1;
        } else if (inRangeInclusive(index, currentIndex, toIndex)) {
          translation = -currentHeight;
          newIndex = index - 1;
        }

        updateElementStyleAtIndex(index, newIndex, translation);
      }

      if (!scrolling) {
        autoScrollUpdate(getMovementInClient());
      }
    }
  }
  /**
   * @param {number} movementInClient
   */
  function autoScrollUpdate(movementInClient) {
    var offset = 0;
    var acceleration = 2;
    var scrollTop = container.scrollTop;
    var scrollHeight = container.scrollHeight;
    var clientHeight = container.clientHeight;
    var isTop = scrollTop === 0;
    var isBottom = scrollHeight - scrollTop - clientHeight === 0;

    if (movementInClient < minMovementInVisible && !isTop) {
      offset = (
        acceleration *
        (movementInClient - minMovementInVisible) / currentHeight
      );
    } else if (movementInClient > maxMovementInVisible && !isBottom) {
      offset = (
        acceleration *
        (movementInClient - maxMovementInVisible) / currentHeight
      );
    }

    clearInterval(autoScrollInterval);
    if (offset !== 0) {
      autoScrollInterval = setInterval(function() {
        container.scrollTop += offset;
      }, 5);
    }
  }
}

export {drag, tap, reorder};

/**
 * @callback WheelHandler
 * @param {WheelEvent} event     The original wheel event
 * @param {number} delta         Normalized scroll amount
 * @param {boolean} isHorizontal `true` if the scroll is horizontal, `false` otherwise
 * @returns {void}
 */
/**
 * Wraps (or decorates) a wheel event handler handling all the platform/browser
 * inconsistencies, and calls the handler with a normalized scroll amount.
 *
 * Supports calling the handler with stepped increments.
 *
 * @param {Object} [options]
 * @param {number} [options.step=0] The increments of the `handler`'s `delta`
 * parameter. When set to `0` the `delta` is not stepped.
 * @param {string} [options.direction='xy'] The directions of interest, in order
 * of priority (will check first then second).
 * @param {WheelHandler} handler The function that will be called to handle a
 * `WheelEvent`.
 * @param {*} [thisArg] If given, the context with which `handler` is called,
 * otherwise will pass the context of the event listener.
 * @returns {GlobalEventHandlers['onwheel']} The 'wheel' event handler to add to
 * the element of your choosing.
 */
export function getWheelHandler(options, handler, thisArg) {
  if (typeof options === 'function') {
    thisArg = handler;
    handler = options;
    options = {};
  }

  const {step = 0, direction = 'xy'} = options;
  const directions = [];
  let wheelingDirection;
  let deltaAccumulator;
  let lastDelta;

  for (const char of direction) {
    switch (char.toLowerCase()) {
    case 'x':
      directions.push(0);
      break;
    case 'y':
      directions.push(1);
      break;
    default:
    }
  }

  const debouncedReset = debounce(100, () => {
    wheelingDirection = null;
    deltaAccumulator = 0;
    lastDelta = 0;
  });

  debouncedReset.invoke();

  return decoratedFn(handler, function onwheel(event) {
    // The decision to use `wheelDelta` properties, except on macOS, is
    // motivated by the wheel deltas survey:
    // https://docs.google.com/spreadsheets/d/1YBTbMSq12_6jmXGm7GJi11RCatYTGk_WkOtfBptgqKI
    //
    // If you want to test yourself, you can use this code pen:
    // https://codepen.io/ziadkh0/embed/pomBxwJ

    let wheelDelta;
    if (isMacOS) {
      // Note that deltaMode MUST be accessed BEFORE delta* in order to get
      // non-pixel values in Firefox.
      // See https://github.com/w3c/uievents/issues/181

      switch (event.deltaMode) {
      default:
      case event.DOM_DELTA_PIXEL:
        wheelDelta = [event.deltaX / 120, -event.deltaY / 120];
        break;
      case event.DOM_DELTA_LINE:
      case event.DOM_DELTA_PAGE:
        // Discard the delta value, just take the sign
        wheelDelta = [Math.sign(event.deltaX), Math.sign(-event.deltaY)];
        break;
      }
    } else {
      wheelDelta = [-event.wheelDeltaX / 120, event.wheelDeltaY / 120];
    }

    if (wheelingDirection === null) {
      for (const dir of directions) {
        if (wheelDelta[dir] !== 0) {
          wheelingDirection = dir;
          break;
        }
      }
    }
    if (wheelingDirection === null) {
      return;
    }

    let delta = wheelDelta[wheelingDirection];
    if (step) {
      deltaAccumulator += delta;
      delta = Math.round(deltaAccumulator);
      if (delta === 0 && deltaAccumulator !== 0) {
        delta = Math.sign(deltaAccumulator);
      }
      [delta, lastDelta] = [Math.sign(delta - lastDelta), delta];
      delta *= step;
    }

    preventDefaultAndStopEvent(event);
    handler.call(thisArg || this, event, delta, wheelingDirection === 0);

    debouncedReset();
  });
}
