/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {test, assert} from '../core/utils/test_utils.js';
import {Enum} from './enum.js';

test('Enum - Color: basic features', () => {
  class Color extends Enum {
    static red = new Color();
    static orange = new Color();
    static yellow = new Color();
    static green = new Color();
    static blue = new Color();
    static purple = new Color();
    static _ = this.closeEnum();
  }

  // Instance properties
  assert.equal(
    Color.red.enumKey, 'red');
  assert.equal(
    Color.red.enumOrdinal, 0);

  // Prototype methods
  assert.equal(
    'Color: ' + Color.red, // .toString()
    'Color: Color.red');

  // Static `.enumKeys` and static `.enumValues`
  assert.deepEqual(
    Color.enumKeys,
    ['red', 'orange', 'yellow', 'green', 'blue', 'purple']);
  assert.deepEqual(
    Color.enumValues,
    [
      Color.red, Color.orange, Color.yellow,
      Color.green, Color.blue, Color.purple
    ]);

  // Static `.enumValueOf()`
  assert.equal(
    Color.enumValueOf('yellow'),
    Color.yellow);

  // Iterability
  const result = [];
  const iterated = [...Color];
  for (const c of Color) {
    result.push('Color: ' + c);
  }
  assert.deepEqual(
    iterated, [
      Color.red,
      Color.orange,
      Color.yellow,
      Color.green,
      Color.blue,
      Color.purple,
    ]);
});

test('Enum - Instance properties', () => {
  class Weekday extends Enum {
    static monday = new Weekday(true);
    static tuesday = new Weekday(true);
    static wednesday = new Weekday(true);
    static thursday = new Weekday(true);
    static friday = new Weekday(true);
    static saturday = new Weekday(false);
    static sunday = new Weekday(false);
    static _ = this.closeEnum();

    constructor(isWorkDay) {
      super();
      this.isWorkDay = isWorkDay;
    }
  }
  assert.equal(Weekday.sunday.isWorkDay, false);
  assert.equal(Weekday.wednesday.isWorkDay, true);
});

test('Enum - switch', () => {
  class Weekday extends Enum {
    static monday = new Weekday();
    static tuesday = new Weekday();
    static wednesday = new Weekday();
    static thursday = new Weekday();
    static friday = new Weekday();
    static saturday = new Weekday();
    static sunday = new Weekday();
    static _ = this.closeEnum();
  }
  function nextDay(weekday) {
    switch (weekday) {
    case Weekday.monday:
      return Weekday.tuesday;
    case Weekday.tuesday:
      return Weekday.wednesday;
    case Weekday.wednesday:
      return Weekday.thursday;
    case Weekday.thursday:
      return Weekday.friday;
    case Weekday.friday:
      return Weekday.saturday;
    case Weekday.saturday:
      return Weekday.sunday;
    case Weekday.sunday:
      return Weekday.monday;
    default:
      throw new Error();
    }
  }
  assert.equal(
    nextDay(Weekday.tuesday), Weekday.wednesday);
  assert.equal(
    nextDay(Weekday.sunday), Weekday.monday);
});

test('Enum - Instance getters', () => {
  class Weekday extends Enum {
    static monday = new Weekday({
      get nextDay() { return Weekday.tuesday; }
    });
    static tuesday = new Weekday({
      get nextDay() { return Weekday.wednesday; }
    });
    static wednesday = new Weekday({
      get nextDay() { return Weekday.thursday; }
    });
    static thursday = new Weekday({
      get nextDay() { return Weekday.friday; }
    });
    static friday = new Weekday({
      get nextDay() { return Weekday.saturday; }
    });
    static saturday = new Weekday({
      get nextDay() { return Weekday.sunday; }
    });
    static sunday = new Weekday({
      get nextDay() { return Weekday.monday; }
    });
    static _ = this.closeEnum();
    constructor(props) {
      super();
      Object.defineProperties(this, Object.getOwnPropertyDescriptors(props));
    }
  }
  assert.equal(
    Weekday.friday.nextDay, Weekday.saturday);
  assert.equal(
    Weekday.sunday.nextDay, Weekday.monday);
});

test('Enum - Arbitrary enum values', () => {
  class Mode extends Enum {
    static user_r = new Mode(0b100000000);
    static user_w = new Mode(0b010000000);
    static user_x = new Mode(0b001000000);
    static group_r = new Mode(0b000100000);
    static group_w = new Mode(0b000010000);
    static group_x = new Mode(0b000001000);
    static all_r = new Mode(0b000000100);
    static all_w = new Mode(0b000000010);
    static all_x = new Mode(0b000000001);
    static _ = this.closeEnum();

    constructor(n) {
      super();
      this.n = n;
    }
  }
  /*eslint-disable no-bitwise*/
  assert.equal(
    Mode.user_r.n | Mode.user_w.n | Mode.user_x.n |
    Mode.group_r.n | Mode.group_x.n |
    Mode.all_r.n | Mode.all_x.n,
    0o755);
  assert.equal(
    Mode.user_r.n | Mode.user_w.n | Mode.user_x.n |
    Mode.group_r.n,
    0o740);
  /*eslint-enable no-bitwise*/
});

class State extends Enum {
  static start = new State({
    done: false,
    accept(x) {
      if (x === '1') {
        return State.one;
      } else {
        return State.start;
      }
    },
  });
  static one = new State({
    done: false,
    accept(x) {
      if (x === '1') {
        return State.two;
      } else {
        return State.start;
      }
    },
  });
  static two = new State({
    done: false,
    accept(x) {
      if (x === '1') {
        return State.three;
      } else {
        return State.start;
      }
    },
  });
  static three = new State({
    done: true,
  });
  static _ = this.closeEnum();

  accept(x) {
    throw new Error('Must be overridden');
  }

  constructor(props) {
    super();
    Object.defineProperties(this, Object.getOwnPropertyDescriptors(props));
  }
}

function run(state, inputString) {
  const trace = [];
  for (const ch of inputString) {
    if (state.done) {
      break;
    }
    state = state.accept(ch);
    trace.push(`${ch} --> ${state}`);
  }
  return trace;
}

test('Enum - run', () => {
  assert.deepEqual(
    run(State.start, '01011100'),
    [
      '0 --> State.start',
      '1 --> State.one',
      '0 --> State.start',
      '1 --> State.one',
      '1 --> State.two',
      '1 --> State.three',
    ]
  );
});
