/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  isValidDate,
  parseDecimal,
  parseURLQueryString,
  stringifyQueryString,
  Pattern,
  RegExpHelpers,
} from '../syntool.js';

// These chars don't need escaping "~!*()_'.-"
const NUM_SEP = '_';
const ID_SEP = '!';
const FILTER_SEP = '!';
const NUM_SEP_RE = /[,_!\n]/;
const ID_SEP_RE = /[,!\n]/;
const TRAILING_COMMAS_RE = /,+$/g;
const TRAILING_FILTER_SEP_RE = /!+$/g;
const DATE_RE = new RegExp(
  RegExpHelpers.or(
    '^(?<time>\\d+)$',
    '^(?<Y>(?:[+-]\\d{2})?\\d{4})[./\\ -]?(?<j>\\d{1,3})$',
    `^${
      [
        '(?<y>(?:[+-]\\d{2})?\\d{4})',
        '(?<m>\\d{2})',
        '(?<d>\\d{2})',
      ].join('[./\\ -]?') +
      `(?:[./\\\\ T-]?(?<H>\\d{2})${
        `(?:[./\\ :-]?(?<M>\\d{2})${
          `(?:${
            '[./\\ :-]?(?<S>\\d{2})' +
            '(?:\\.(?<MS>\\d{3}))?' +
            '(?<tz>.*)?'
          })?`
        })?`
      })?`
    }$`
  ),
  'i'
);
const STR_SPLITTER = {
  [Symbol.split](string) {
    const lines = string.split('\n');
    return lines.length > 1 ? lines : string.split(',');
  },
};

class Field {
  constructor(options) {
    this.required = !!options?.required;
    if (options?.get) {
      this.get = options.get;
    }
    if (options?.set) {
      this.set = options.set;
    }
  }
  applyToState(state, value) {
    return this.set(state, value);
  }
  fromState(state) {
    return this.get(state);
  }
  toJSON(value) {
    return value;
  }
  fromJSON(value) {
    return value == null ? undefined : value;
  }
  toURL(value) {
    return value;
  }
  fromURL(value) {
    return value == null ? undefined : value;
  }
  validate(value) {
    return (!this.required || value != null);
  }
}

class StringField extends Field {}

class DateField extends Field {
  toJSON(value) {
    return value.getTime();
  }
  fromJSON(value) {
    return new Date(value);
  }
  toURL(value) {
    return value.toISOString().split('.')[0];
  }
  fromURL(value) {
    if (!value) {
      return undefined;
    }
    const {Y, j, time, y = Y, m, d = j, H, M, S, MS = time} = (
      value.match(DATE_RE)?.groups ?? {}
    );
    return new Date(Date.UTC(
      y == null ? 1970 : parseDecimal(y),
      m == null ? 0 : parseDecimal(m) - 1,
      d == null ? 1 : parseDecimal(d),
      H == null ? 0 : parseDecimal(H),
      M == null ? 0 : parseDecimal(M),
      S == null ? 0 : parseDecimal(S),
      MS == null ? 0 : parseDecimal(MS)
    ));
  }
  validate(value) {
    return value == null ? !this.required : isValidDate(value);
  }
}

class IntField extends Field {
  fromURL(value) {
    return value == null ? undefined : parseDecimal(value);
  }
  validate(value) {
    return value == null ? !this.required : isFinite(value);
  }
}

class FloatField extends Field {
  fromURL(value) {
    return value == null ? undefined : parseFloat(value);
  }
  validate(value) {
    return value == null ? !this.required : isFinite(value);
  }
}

class ArrayField extends Field {
  constructor(options) {
    super(options);

    const {
      elementType,
      separator,
      splitSeparator = separator,
      length = 0,
      removeTrailing = false,
    } = options;
    this.elementType = elementType;
    this.separator = separator ?? NUM_SEP;
    this.splitSeparator = splitSeparator ?? NUM_SEP_RE;
    this.length = length;
    this.removeTrailing = removeTrailing;
  }
  toJSON(array) {
    return array?.map((element) => this.elementType.toJSON(element));
  }
  fromJSON(array) {
    return array?.map((element) => this.elementType.fromJSON(element));
  }
  toURL(array) {
    let string = (
      array?.map((element) => this.elementType.toURL(element))
      .join(this.separator)
    );
    if (string && this.removeTrailing) {
      string = string.replace(this.removeTrailing, '');
    }
    if (!string && !this.required) {
      string = undefined;
    }
    return string;
  }
  fromURL(string) {
    return (
      string?.split(this.splitSeparator)
      .map((element) => this.elementType.fromURL(element))
    );
  }
  validate(array) {
    return array == null ? !this.required : (
      Array.isArray(array) &&
      array.length >= this.length &&
      array.every((value) => this.elementType.validate(value))
    );
  }
}

class PatternField extends StringField {
  validate(value) {
    return value == null ? !this.required : Pattern.canParse(value, true);
  }
}

class FilterField extends StringField {
  toURL([id, active]) {
    // Spaces ' ' will be encoded as pluses '+'
    return `${active ? ' ' : '-'}${id}`;
  }
  fromURL(value) {
    return value ? [value.slice(1), value[0] !== '-'] : undefined;
  }
}

const types = {
  Field,
  StringField,
  DateField,
  IntField,
  FloatField,
  ArrayField,
  PatternField,
  FilterField,
};

const extensions = new Map();
const fields = {
  date: new DateField({
    get(state) {
      return new Date(state.timeline?.currentTime);
    },
    set(state, currentDate) {
      state.timeline = Object.assign(state.timeline || {}, {
        currentTime: currentDate.getTime(),
      });
      return state;
    },
  }),
  timespan: new StringField({
    get(state) {
      return state.timeline?.timeframe;
    },
    set(state, timeframe) {
      state.timeline = Object.assign(state.timeline || {}, {timeframe});
      return state;
    },
  }),
  extent: new ArrayField({
    elementType: new FloatField({required: true}),
    length: 4,

    get(state) {
      return state.mapView?.extent;
    },
    set(state, extent) {
      state.mapView = Object.assign(state.mapView || {}, {extent});
      return state;
    },
  }),
  center: new ArrayField({
    elementType: new FloatField({required: true}),
    length: 2,

    get(state) {
      return state.mapView?.center;
    },
    set(state, center) {
      state.mapView = Object.assign(state.mapView || {}, {center});
      return state;
    },
  }),
  zoom: new IntField({
    get(state) {
      return state.mapView?.zoom;
    },
    set(state, zoom) {
      state.mapView = Object.assign(state.mapView || {}, {zoom});
      return state;
    },
  }),
  products: new ArrayField({
    elementType: new StringField({required: true}),
    separator: ID_SEP,
    splitSeparator: ID_SEP_RE,
    required: true,

    get(state) {
      return state.selection;
    },
    set(state, selection) {
      return Object.assign(state, {selection});
    },
  }),
  opacity: new ArrayField({
    elementType: new FloatField({required: true}),

    get(state) {
      const {products, selection: selectedProductIds} = state;
      return selectedProductIds?.map(
        (productId) => products[productId].opacity * 100
      );
    },
    set(state, opacities) {
      if (!opacities) {
        return state;
      }
      const {products = {}, selection = []} = state;
      for (let i = 0; i < selection.length; i++) {
        products[selection[i]] = Object.assign(products[selection[i]] || {}, {
          opacity: opacities[i] / 100,
        });
      }
      return Object.assign(state, {products});
    },
  }),
  stackLevel: new ArrayField({
    elementType: new FloatField({required: true}),

    get(state) {
      const {products, selection: selectedProductIds} = state;
      return selectedProductIds?.map(
        (productId) => products[productId].stackLevel
      );
    },
    set(state, stackLevels) {
      if (!stackLevels) {
        return state;
      }
      const {products = {}, selection = []} = state;
      for (let i = 0; i < selection.length; i++) {
        products[selection[i]] = Object.assign(products[selection[i]] || {}, {
          stackLevel: stackLevels[i],
        });
      }
      return Object.assign(state, {products});
    },
  }),
  pattern: new ArrayField({
    elementType: new PatternField(),
    separator: ',',
    splitSeparator: STR_SPLITTER,
    removeTrailing: TRAILING_COMMAS_RE,

    get(state) {
      const {products, selection: selectedProductIds} = state;
      return selectedProductIds?.map(
        (productId) => products[productId].pattern
      );
    },
    set(state, patterns) {
      if (!patterns) {
        return state;
      }
      const {products = {}, selection = []} = state;
      for (let i = 0; i < selection.length; i++) {
        products[selection[i]] = Object.assign(products[selection[i]] || {}, {
          pattern: patterns[i],
        });
      }
      return Object.assign(state, {products});
    },
  }),
  filter: new ArrayField({
    elementType: new ArrayField({
      elementType: new FilterField(),
      separator: ',',
      splitSeparator: ',',
      removeTrailing: TRAILING_COMMAS_RE,
    }),
    separator: FILTER_SEP,
    splitSeparator: FILTER_SEP,
    removeTrailing: TRAILING_FILTER_SEP_RE,

    get(state) {
      const {products, selection: selectedProductIds} = state;
      return selectedProductIds?.map(
        (productId) => Object.entries(products[productId].filters ?? {})
      );
    },
    set(state, filters) {
      if (!filters) {
        return state;
      }
      const {products = {}, selection = []} = state;
      for (let i = 0; i < selection.length; i++) {
        const filtersEntries = filters[i]?.filter(Boolean);
        if (filtersEntries) {
          products[selection[i]] = Object.assign(products[selection[i]] || {}, {
            filters: Object.fromEntries(filtersEntries),
          });
        }
      }
      return Object.assign(state, {products});
    },
  }),
};

function appStateToJSON(state) {
  let json = {};

  for (const [key, type] of Object.entries(fields)) {
    if (type) {
      const value = type.fromState(state);
      if (type.validate(value)) {
        json[key] = type.toJSON(value);
      }
    }
  }

  for (const extension of extensions.values()) {
    if (typeof extension.appStateToJSON === 'function') {
      json = extension.appStateToJSON(json, state);
    }
  }

  return json;
}
function appStateFromJSON(json) {
  let state = {};

  for (const [key, type] of Object.entries(fields)) {
    if (type && Object.hasOwn(json, key)) {
      const value = type.fromJSON(json[key]);
      if (type.validate(value)) {
        state = type.applyToState(state, value);
      }
    }
  }

  for (const extension of extensions.values()) {
    if (typeof extension.appStateFromJSON === 'function') {
      state = extension.appStateFromJSON(state, json);
    }
  }

  return state;
}
function appStateJSONToURLParameters(json) {
  let urlParams = {};

  for (const [key, type] of Object.entries(fields)) {
    if (type) {
      const value = type.fromJSON(json[key]);
      if (type.validate(value)) {
        urlParams[key] = type.toURL(value);
      }
    }
  }

  for (const extension of extensions.values()) {
    if (typeof extension.appStateJSONToURLParameters === 'function') {
      urlParams = extension.appStateJSONToURLParameters(urlParams, json);
    }
  }

  return urlParams;
}

function appStateJSONFromURLParameters(urlParams) {
  let json = {};

  for (const [key, type] of Object.entries(fields)) {
    if (type && Object.hasOwn(urlParams, key)) {
      const value = type.fromURL(urlParams[key]);
      if (type.validate(value)) {
        json[key] = type.toJSON(value);
      }
    }
  }

  for (const extension of extensions.values()) {
    if (typeof extension.appStateJSONFromURLParameters === 'function') {
      json = extension.appStateJSONFromURLParameters(json, urlParams);
    }
  }

  return json;
}

function appStateToURL(state, baseUrl) {
  var urlParams = appStateJSONToURLParameters(appStateToJSON(state));
  var url = new URL(baseUrl || window.location, window.location.origin);
  url.search = stringifyQueryString(urlParams);
  return url;
}

function appStateFromURL(url) {
  return appStateFromJSON(appStateJSONFromURLParameters(parseURLQueryString(url)));
}

function extendAppState(extension) {
  if (extension) {
    extensions.set(extension.id || {}, extension);
  }
}

export {
  appStateToJSON,
  appStateFromJSON,
  appStateJSONToURLParameters,
  appStateJSONFromURLParameters,
  appStateToURL,
  appStateFromURL,
  extendAppState,
  types as appStateFieldTypes,
  fields as appStateFields,
};
window.ODL = Object.assign(window.ODL || {}, {
  appStateToJSON,
  appStateFromJSON,
  appStateJSONToURLParameters,
  appStateJSONFromURLParameters,
  appStateToURL,
  appStateFromURL,
  extendAppState,
  appStateFieldTypes: types,
  appStateFields: fields,
});
