/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {EventsHub} from '../syntool.js';
import {matchSorter} from './match_sorter.js';

const PRODUCTS_SEARCH_OPTIONS = {keys: ['textLabel', 'searchTags']};

function idFromProductOrId(productOrId) {
  return (typeof productOrId === 'string' ? productOrId : productOrId.id);
}

export class CatalogStore {
  constructor(productsStore) {
    this._store = productsStore;
    this._selection = new Set(productsStore.selectedProductIds());
    this.hub = new EventsHub();

    productsStore.hub
      .on('change:products', (productId) => {
        this.hub.emit('change:products', productId);
      })
      .on('change:selection', () => {
        var count = this._selection.size;
        for (const productId of productsStore.selectedProductIds()) {
          this._selection.add(productId);
        }
        if (this._selection.size !== count) {
          this.hub.emit('change:selection');
          this.hub.emit('change:state');
        }
      });
  }

  /**
   * @public
   * @returns {import('../syntool.js').ProductsStore}
   */
  get productsStore() {
    return this._store;
  }

  get state() {
    return {selection: [...this.selectedProductIds()]};
  }
  set state(state) {
    if (state.selection) {
      const selection = new Set(state.selection);
      const toDeselect = new Set();
      for (const productId of this.productsStore.selectedProductIds()) {
        if (!selection.has(productId)) {
          toDeselect.add(productId);
        }
      }
      this.productsStore.deselectProducts(toDeselect);
    }
  }

  getProductIds() {
    return this._store.getProductIds();
  }
  getProducts() {
    return this._store.getProducts();
  }
  getProduct(productId) {
    return this._store.getProduct(productId);
  }
  hasProduct(productId) {
    return this._store.hasProduct(productId);
  }
  forEachProduct(callback, thisArg) {
    return this._store.forEachProduct(callback, thisArg);
  }
  productIds() {
    return this._store.productIds();
  }
  products() {
    return this._store.products();
  }
  canDeleteProduct(productId) {
    return this._store.canDeleteProduct(productId);
  }
  deleteProduct(productId) {
    this.deselectProduct(productId);
    return this._store.deleteProduct(productId);
  }

  searchProducts(search) {
    var products = this.getProducts();
    return (
      search
      ? matchSorter(products, search, PRODUCTS_SEARCH_OPTIONS)
      : products
    );
  }

  getSelectedProductsCount() {
    return this._selection.size;
  }
  getSelectedProductIds() {
    return [...this._selection.values()];
  }
  getSelectedProducts() {
    return this.getSelectedProductIds().map(this.getProduct, this);
  }
  isProductSelected(productId) {
    return this._selection.has(productId);
  }
  forEachSelectedProduct(callback, thisArg) {
    for (const productId of this._selection) {
      callback.call(thisArg, this.getProduct(productId), productId);
    }
  }
  selectedProductIds() {
    return this._selection.values();
  }
  *selectedProducts() {
    for (const productId of this._selection) {
      yield this.getProduct(productId);
    }
  }

  selectProduct(productOrId) {
    const productId = idFromProductOrId(productOrId);
    if (!this._selection.has(productId) && this._store.hasProduct(productId)) {
      this._store.selectProduct(productId);
      return true;
    }
    return false;
  }
  selectProducts(productsOrIds) {
    this._store.selectProducts(productsOrIds);
  }
  selectOnlyProducts(productsOrIds) {
    const productIdsToSelect = new Set();
    let changed = false;
    for (const productOrId of productsOrIds) {
      const productId = idFromProductOrId(productOrId);
      if (this._store.hasProduct(productId)) {
        productIdsToSelect.add(productId);
        if (!this._selection.has(productId)) {
          this._selection.add(productId);
          changed = true;
        }
      }
    }
    for (const productId of this.selectedProductIds()) {
      if (!productIdsToSelect.has(productId)) {
        this._selection.delete(productId);
        changed = true;
      }
    }
    this._store.selectOnlyProducts(productIdsToSelect);
    if (changed) {
      this.hub.emit('change:selection');
      this.hub.emit('change:state');
    }
  }
  deselectProduct(productOrId) {
    const productId = idFromProductOrId(productOrId);
    if (this._selection.delete(productId)) {
      this._store.deselectProduct(productId);
      this.hub.emit('change:selection');
      this.hub.emit('change:state');
      return true;
    }
    return false;
  }
  deselectProducts(productsOrIds) {
    var changed = false;
    for (const productOrId of productsOrIds) {
      const productId = idFromProductOrId(productOrId);
      if (this.isProductSelected(productId)) {
        this._selection.delete(productId);
        changed = true;
      }
    }
    this._store.deselectProducts(productsOrIds);
    if (changed) {
      this.hub.emit('change:selection');
      this.hub.emit('change:state');
    }
  }
  deselectAllProducts() {
    if (this._selection.size === 0) {
      return;
    }
    this._selection.clear();
    this._store.deselectAllProducts();
    this.hub.emit('change:selection');
    this.hub.emit('change:state');
  }
}
window.ODL = Object.assign(window.ODL || {}, {CatalogStore});
