/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {NONPASSIVE, PASSIVE, scheduleTask} from '../syntool.js';

var element = createElement();
var container = createContainer(element);

var waitingCopyIntervention = false;
var waitingPasteIntervention = false;
var onCopyFinish = null;
var onPasteFinish = null;

document.addEventListener('selectionchange', handleSelectionChange, NONPASSIVE);
document.addEventListener('paste', handlePaste, NONPASSIVE);
element.addEventListener('cut', handleCopy, PASSIVE);
element.addEventListener('copy', handleCopy, PASSIVE);

function createElement() {
  var isRTL = document.documentElement.getAttribute('dir') === 'rtl';
  var pre = document.createElement('pre');

  // Reset box model
  pre.style.border = '0';
  pre.style.padding = '0';
  pre.style.margin = '0';
  pre.style.width = '1px';
  pre.style.height = '1px';
  // Move element out of screen horizontally
  pre.style.position = 'absolute';
  pre.style[isRTL ? 'right' : 'left'] = '-9999px';

  return pre;
}
function createContainer(el) {
  var cntnr = document.createElement('div');

  cntnr.style.position = 'fixed';
  cntnr.style.display = 'block';
  cntnr.style.left = '0';
  cntnr.style.top = '0';
  cntnr.style.width = '0';
  cntnr.style.height = '0';
  cntnr.style.zIndex = '100';
  cntnr.style.opacity = '0';

  cntnr.appendChild(el);

  return cntnr;
}

function handleSelectionChange(event) {
  if (waitingCopyIntervention) {
    window.getSelection().selectAllChildren(element);
  }
}
function handlePaste(event) {
  if (waitingPasteIntervention) {
    var pasteData = (event.clipboardData || window.clipboardData);
    finishPaste(pasteData.getData('text'));
    event.preventDefault();
  }
}
function handleCopy(event) {
  if (waitingCopyIntervention) {
    scheduleTask(finishCopy, true);
  }
}

function copyWithClipboard(text, requestIntervention) {
  if (navigator.clipboard && navigator.clipboard.writeText) {
    navigator.clipboard.writeText(text).then(function() {
      finishCopy(true);
    }, function() {
      copyWithCommand(text, requestIntervention);
    });
  } else {
    copyWithCommand(text, requestIntervention);
  }
}
function copyWithCommand(text, requestIntervention) {
  element.textContent = text;

  // Move element to the same position vertically
  element.style.top = (
    (window.pageYOffset || document.documentElement.scrollTop) + 'px'
  );

  document.body.appendChild(container);
  window.getSelection().selectAllChildren(element);

  try {
    if (document.execCommand('copy')) {
      finishCopy(true);
      return;
    }
  } catch (error) {
    /*ignore error*/
  }
  copyWithIntervention(text, requestIntervention);
}
function copyWithIntervention(text, requestIntervention) {
  if (requestIntervention && requestIntervention()) {
    waitingCopyIntervention = true;
  } else {
    finishCopy(false);
  }
}

function pasteWithClipboard(requestIntervention) {
  if (navigator.clipboard && navigator.clipboard.readText) {
    navigator.clipboard.readText().then(finishPaste, function() {
      pasteWithIntervention(requestIntervention);
    });
  } else {
    pasteWithIntervention(requestIntervention);
  }
}
function pasteWithIntervention(requestIntervention) {
  if (requestIntervention && requestIntervention()) {
    waitingPasteIntervention = true;
  } else {
    finishPaste(false);
  }
}

export function copyToClipboard(text, requestIntervention, done) {
  if (text) {
    // cancel the last copy request, if any
    finishCopy(false);

    // start a new copy request
    onCopyFinish = done;
    copyWithClipboard(text, requestIntervention);
  }
}
function finishCopy(succeeded) {
  waitingCopyIntervention = false;
  if (container.parentNode) {
    container.remove();
  }
  if (onCopyFinish) {
    onCopyFinish(succeeded);
  }
  onCopyFinish = null;
}
export function pasteFromClipboard(requestIntervention, done) {
  // cancel the last paste request, if any
  finishPaste(false);

  // start a new paste request
  onPasteFinish = done;
  pasteWithClipboard(requestIntervention);
}
function finishPaste(pastedTextOrFalse) {
  waitingPasteIntervention = false;
  if (onPasteFinish) {
    onPasteFinish(pastedTextOrFalse);
  }
  onPasteFinish = null;
}
export function cancelCopy() {
  finishCopy(false);
}
export function cancelPaste() {
  finishPaste(false);
}
window.ODL = Object.assign(window.ODL || {}, {
  Clipboard: {
    copy : copyToClipboard,
    paste: pasteFromClipboard,
    cancelCopy,
    cancelPaste,
  },
});
