/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {negate, clamp, lerp, parseHex, RegExpHelpers, Types} from '../syntool.js';

const {
  hexDigit,
  uint8,
  ufloat,
  requiredWhiteSpace,
  funcCall,
} = RegExpHelpers;

const cssColorHexRGB = '#' + hexDigit + '{3}';
const cssColorHexRGBA = '#' + hexDigit + '{4}';
const cssColorHexShort =
  '(?:' + [cssColorHexRGB, cssColorHexRGBA].join('|') + ')';
const cssColorHexRRGGBB = '#' + hexDigit + '{6}';
const cssColorHexRRGGBBAA = '#' + hexDigit + '{8}';
const cssColorHexLong =
  '(?:' + [cssColorHexRRGGBB, cssColorHexRRGGBBAA].join('|') + ')';
const cssColorHex = '(?:' + [cssColorHexShort, cssColorHexLong].join('|') + ')';
const cssColorRGB = funcCall('rgb', uint8, uint8, uint8);
const cssColorRGBA = funcCall('rgba', uint8, uint8, uint8, ufloat);
const cssColor =
  '(?:' + [cssColorHex, cssColorRGB, cssColorRGBA].join('|') + ')';

const Color = Types.string.match(new RegExp('^' + cssColor + '$', 'i'));

var cssColorHexShortRegex = new RegExp('^' + cssColorHexShort + '$', 'i');
var cssColorHexLongRegex = new RegExp('^' + cssColorHexLong + '$', 'i');
var cssColorRGBRegex = new RegExp('^' + cssColorRGB + '$', 'i');
var cssColorRGBARegex = new RegExp('^' + cssColorRGBA + '$', 'i');

var HALF = 0xf;
var BYTE = 0xff;
var R_OFFSET = 3 * 8;
var G_OFFSET = 2 * 8;
var B_OFFSET = 1 * 8;
var A_OFFSET = 0 * 8;

const colorIsInvalid = Number.isNaN;
const colorIsValid = negate(colorIsInvalid);

const COLOR_INVALID = NaN;
const COLOR_TRANSPARENT = createColor(0, 0, 0, 0);
const COLOR_BLACK = createColor(0, 0, 0);
const COLOR_WHITE = createColor(255, 255, 255);

function createColor(r, g, b, a) {
  if (arguments.length < 4 || a === undefined) {
    a = 1;
  }

  /*eslint-disable no-bitwise*/
  return (
    ((r & BYTE) << R_OFFSET) |
    ((g & BYTE) << G_OFFSET) |
    ((b & BYTE) << B_OFFSET) |
    (((a * BYTE) & BYTE) << A_OFFSET)
  );
  /*eslint-enable no-bitwise*/
}

function _colorGetRed(color) {
  //eslint-disable-next-line no-bitwise
  return (color >>> R_OFFSET) & BYTE;
}
function _colorGetGreen(color) {
  //eslint-disable-next-line no-bitwise
  return (color >>> G_OFFSET) & BYTE;
}
function _colorGetBlue(color) {
  //eslint-disable-next-line no-bitwise
  return (color >>> B_OFFSET) & BYTE;
}
function _colorGetAlpha(color) {
  //eslint-disable-next-line no-bitwise
  return ((color >>> A_OFFSET) & BYTE) / BYTE;
}

function colorGetRed(color) {
  return colorIsInvalid(color) ? NaN : _colorGetRed(color);
}
function colorGetGreen(color) {
  return colorIsInvalid(color) ? NaN : _colorGetGreen(color);
}
function colorGetBlue(color) {
  return colorIsInvalid(color) ? NaN : _colorGetBlue(color);
}
function colorGetAlpha(color) {
  return colorIsInvalid(color) ? NaN : _colorGetAlpha(color);
}

function colorSetRed(color, red) {
  return colorIsInvalid(color) ? NaN : createColor(
    red,
    colorGetGreen(color),
    colorGetBlue(color),
    colorGetAlpha(color)
  );
}
function colorSetGreen(color, green) {
  return colorIsInvalid(color) ? NaN : createColor(
    colorGetRed(color),
    green,
    colorGetBlue(color),
    colorGetAlpha(color)
  );
}
function colorSetBlue(color, blue) {
  return colorIsInvalid(color) ? NaN : createColor(
    colorGetRed(color),
    colorGetGreen(color),
    blue,
    colorGetAlpha(color)
  );
}
function colorSetAlpha(color, alpha) {
  return colorIsInvalid(color) ? NaN : createColor(
    colorGetRed(color),
    colorGetGreen(color),
    colorGetBlue(color),
    alpha
  );
}

function colorToArray(color) {
  return [
    colorGetRed(color),
    colorGetGreen(color),
    colorGetBlue(color),
    colorGetAlpha(color),
  ];
}
function colorFromArray(components) {
  return createColor(...components);
}

function colorToNormalizedRGBArray(color) {
  return [
    colorGetRed(color) / BYTE,
    colorGetGreen(color) / BYTE,
    colorGetBlue(color) / BYTE,
  ];
}
function colorFromNormalizedRGBArray(components) {
  return createColor(
    components[0] * BYTE,
    components[1] * BYTE,
    components[2] * BYTE
  );
}

function colorToNormalizedRGBAArray(color) {
  return [
    colorGetRed(color) / BYTE,
    colorGetGreen(color) / BYTE,
    colorGetBlue(color) / BYTE,
    colorGetAlpha(color),
  ];
}
function colorFromNormalizedRGBAArray(components) {
  return createColor(
    components[0] * BYTE,
    components[1] * BYTE,
    components[2] * BYTE,
    components[3]
  );
}

function uint32(number) {
  //eslint-disable-next-line no-bitwise
  return number >>> 0;
}
function colorToHexCSSColor(color) {
  if (colorIsInvalid(color)) {
    return '#00000000';
  }
  if (_colorGetAlpha(color) === 1) {
    // eslint-disable-next-line no-bitwise
    return '#' + uint32(color >>> 8).toString(16).padStart(6, '0');
  }
  return '#' + uint32(color).toString(16).padStart(8, '0');
}
function colorToCSSColor(color) {
  if (colorIsInvalid(color)) {
    return 'rgba(0, 0, 0, 0)';
  }
  if (_colorGetAlpha(color) === 1) {
    return (
      'rgb(' +
      _colorGetRed(color) +
      ', ' +
      _colorGetGreen(color) +
      ', ' +
      _colorGetBlue(color) +
      ')'
    );
  }
  return (
    'rgba(' +
    _colorGetRed(color) +
    ', ' +
    _colorGetGreen(color) +
    ', ' +
    _colorGetBlue(color) +
    ', ' +
    _colorGetAlpha(color) +
    ')'
  );
}
function colorFromCSSColor(cssColor) {
  if (typeof cssColor !== 'string') {
    throw new TypeError(
      'cssColor expected to be a string but got ' + cssColor
    );
  }

  cssColor = (
    cssColor
    .replace(new RegExp(requiredWhiteSpace, 'g'), '')
    .toLowerCase()
  );
  if (cssColorHexLongRegex.test(cssColor)) {
    return createColor(
      parseHex(cssColor.substr(1, 2)),
      parseHex(cssColor.substr(3, 2)),
      parseHex(cssColor.substr(5, 2)),
      parseHex(cssColor.substr(7, 2) || 'FF') / BYTE
    );
  }
  if (cssColorHexShortRegex.test(cssColor)) {
    return createColor(
      (parseHex(cssColor.substr(1, 1)) / HALF) * BYTE,
      (parseHex(cssColor.substr(2, 1)) / HALF) * BYTE,
      (parseHex(cssColor.substr(3, 1)) / HALF) * BYTE,
      parseHex(cssColor.substr(4, 1) || 'F') / HALF
    );
  }

  var colorComponents;
  if (cssColorRGBRegex.test(cssColor)) {
    colorComponents = cssColor.substring(4, cssColor.length - 1).split(',');
    colorComponents.push(1);
  } else if (cssColorRGBARegex.test(cssColor)) {
    colorComponents = cssColor.substring(5, cssColor.length - 1).split(',');
  }
  if (colorComponents) {
    return createColor(
      clamp(parseFloat(colorComponents[0]), 0, BYTE),
      clamp(parseFloat(colorComponents[1]), 0, BYTE),
      clamp(parseFloat(colorComponents[2]), 0, BYTE),
      clamp(parseFloat(colorComponents[3]), 0, 1)
    );
  }
  return NaN;
}

function _colorIsDark(color) {
  var red = _colorGetRed(color);
  var green = _colorGetGreen(color);
  var blue = _colorGetBlue(color);
  // YIQ equation from http://24ways.org/2010/calculating-color-contrast
  var yiq = (red * 299 + green * 587 + blue * 114) / 1000;
  return yiq < 128;
}
function colorIsDark(color) {
  return !colorIsInvalid(color) && _colorIsDark(color);
}
function colorIsLight(color) {
  return !colorIsInvalid(color) && !_colorIsDark(color);
}

function colorLerp(a, b, t) {
  if (colorIsInvalid(a) || colorIsInvalid(b)) {
    return NaN;
  }
  return createColor(
    lerp(_colorGetRed(a), _colorGetRed(b), t),
    lerp(_colorGetGreen(a), _colorGetGreen(b), t),
    lerp(_colorGetBlue(a), _colorGetBlue(b), t),
    lerp(_colorGetAlpha(a), _colorGetAlpha(b), t)
  );
}
function colorInverseLerp(a, b, color) {
  if (colorIsInvalid(a) || colorIsInvalid(b) || colorIsInvalid(color)) {
    return NaN;
  }
  if (color === a) {
    return [0, 0];
  }
  if (color === b) {
    return [1, 0];
  }

  const aRed = _colorGetRed(a);
  const aGreen = _colorGetGreen(a);
  const aBlue = _colorGetBlue(a);
  const aAlpha = _colorGetAlpha(a);
  const aColorRed = aRed - _colorGetRed(color);
  const aColorGreen = aGreen - _colorGetGreen(color);
  const aColorBlue = aBlue - _colorGetBlue(color);
  const aColorAlpha = aAlpha - _colorGetAlpha(color);
  const abRed = aRed - _colorGetRed(b);
  const abGreen = aGreen - _colorGetGreen(b);
  const abBlue = aBlue - _colorGetBlue(b);
  const abAlpha = aAlpha - _colorGetAlpha(b);
  const t = (
    (
      aColorRed * abRed +
      aColorGreen * abGreen +
      aColorBlue * abBlue +
      aColorAlpha * abAlpha
    ) /
    (
      abRed * abRed +
      abGreen * abGreen +
      abBlue * abBlue +
      abAlpha * abAlpha
    )
  );
  const dist = Math.hypot(
    aColorRed - t * abRed,
    aColorGreen - t * abGreen,
    aColorBlue - t * abBlue,
    aColorAlpha - t * abAlpha
  );

  return [t, dist];
}

export {
  cssColorHexRGB,
  cssColorHexRGBA,
  cssColorHexShort,
  cssColorHexRRGGBB,
  cssColorHexRRGGBBAA,
  cssColorHexLong,
  cssColorHex,
  cssColorRGB,
  cssColorRGBA,
  cssColor,

  Color,

  colorIsInvalid,
  colorIsValid,

  COLOR_INVALID,
  COLOR_TRANSPARENT,
  COLOR_BLACK,
  COLOR_WHITE,

  createColor,
  colorGetRed,
  colorGetGreen,
  colorGetBlue,
  colorGetAlpha,
  colorSetRed,
  colorSetGreen,
  colorSetBlue,
  colorSetAlpha,
  colorToArray,
  colorFromArray,
  colorToNormalizedRGBArray,
  colorFromNormalizedRGBArray,
  colorToNormalizedRGBAArray,
  colorFromNormalizedRGBAArray,
  colorToHexCSSColor,
  colorToCSSColor,
  colorFromCSSColor,
  colorIsDark,
  colorIsLight,
  colorLerp,
  colorInverseLerp,
};
Object.assign(RegExpHelpers, {
  cssColorHexRGB,
  cssColorHexRGBA,
  cssColorHexShort,
  cssColorHexRRGGBB,
  cssColorHexRRGGBBAA,
  cssColorHexLong,
  cssColorHex,
  cssColorRGB,
  cssColorRGBA,
  cssColor,
});
Object.assign(Types, {
  Color,
});
window.ODL = Object.assign(window.ODL || {}, {
  colorIsInvalid,
  colorIsValid,

  COLOR_INVALID,
  COLOR_TRANSPARENT,
  COLOR_BLACK,
  COLOR_WHITE,

  createColor,
  colorGetRed,
  colorGetGreen,
  colorGetBlue,
  colorGetAlpha,
  colorSetRed,
  colorSetGreen,
  colorSetBlue,
  colorSetAlpha,
  colorToArray,
  colorFromArray,
  colorToNormalizedRGBArray,
  colorFromNormalizedRGBArray,
  colorToNormalizedRGBAArray,
  colorFromNormalizedRGBAArray,
  colorToHexCSSColor,
  colorToCSSColor,
  colorFromCSSColor,
  colorIsDark,
  colorIsLight,
  colorLerp,
  colorInverseLerp,
});
