/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunctionReturningNull,
  parseDecimal,
  inRangeInclusive,
  lerp,
  inverseLerp,
  ajax,
  isAbortError,
  ajaxRetriable,
  fetchResource,
  RegExpHelpers,
} from '../syntool.js';
import {
  createColor,
  colorFromCSSColor,
  colorFromArray,
  colorToCSSColor,
  colorGetAlpha,
  colorSetAlpha,
  colorLerp,
  colorInverseLerp,
  colorIsInvalid,
} from './color.js';
import {
  SVG_NS,
  createStyleSheet,
  createSVGDefs,
  addCSSRuleToSheet,
} from './dom_utils.js';

const {
  optionalWhiteSpace,
  requiredWhiteSpace,
  uint8,
  ieeeFloat,
} = RegExpHelpers;

var rgbLine =
  '^' + optionalWhiteSpace + '(' + uint8 + ')' + requiredWhiteSpace +
  '(' + uint8 + ')' + requiredWhiteSpace + '(' + uint8 + ')' +
  optionalWhiteSpace + '$';
var cptRgbColorSeparator = '(?:' + requiredWhiteSpace + '|\\/)';
var cptRgbColor =
  '(' + uint8 + ')' + cptRgbColorSeparator + '(' + uint8 + ')' +
  cptRgbColorSeparator + '(' + uint8 + ')';
var cptStop =
  '(' + ieeeFloat + ')' + requiredWhiteSpace + '(?:-|' + cptRgbColor + ')';
var cptSliceLine =
  '^' + optionalWhiteSpace + cptStop + requiredWhiteSpace + cptStop + '.*$';
var cptSpecialLine =
  '^' + optionalWhiteSpace + '([BFN])' + requiredWhiteSpace + cptRgbColor + '.*$';
var cptSliceLineRegex = new RegExp(cptSliceLine, 'i');
var cptSpecialLineRegex = new RegExp(cptSpecialLine, 'i');

const COLOR_PALETTE_MISSING = 'COLOR_PALETTE_MISSING';
const COLOR_PALETTE_LOADING = 'COLOR_PALETTE_LOADING';
const COLOR_PALETTE_LOADED = 'COLOR_PALETTE_LOADED';
const COLOR_PALETTE_ERRORED = 'COLOR_PALETTE_ERRORED';

var colorPalettesStyleSheet = createStyleSheet();
var colorPalettesSVGDefs = createSVGDefs();
var colorPalettes = new Map();
var colorPalettesStatus = new Map();

function parseDecimalWithDefault(defaultValue, string) {
  var value = parseDecimal(string);
  return isNaN(value) ? defaultValue : value;
}

function addColorPaletteToCSS(selector, palette) {
  return addCSSRuleToSheet(
    colorPalettesStyleSheet,
    `${selector} {background: ${colorPaletteToCSSLinearGradient(palette)}}`
  );
}
function addColorPaletteToSVG(id, palette) {
  var paletteElement = colorPaletteToSVGLinearGradient(palette);
  paletteElement.id = id;
  colorPalettesSVGDefs.appendChild(paletteElement);
  return true;
}

function _normalizedColorStops(stops) {
  if (!stops || stops.length < 2) {
    return stops;
  }

  const normalized = [];
  const minZ = stops[0][0];
  const maxZ = stops[stops.length - 1][0];
  const deltaZ = maxZ - minZ;
  for (let [z, color] of stops) {
    // Normalize the z value to be between 0 and 1
    z = (z - minZ) / deltaZ;
    if (
      normalized.length > 1 &&
      Object.is(color, normalized[normalized.length - 1][1]) &&
      Object.is(color, normalized[normalized.length - 2][1])
    ) {
      normalized[normalized.length - 1][0] = z;
    } else if (
      normalized.length === 0 ||
      normalized[normalized.length - 1][0] !== z ||
      !Object.is(color, normalized[normalized.length - 1][1])
    ) {
      normalized.push([z, color]);
    }
  }

  const simplified = [normalized[0]];
  for (let i = 1; i < normalized.length - 1; i++) {
    const prevColor = simplified[simplified.length - 1][1];
    const currentColor = normalized[i][1];
    const nextColor = normalized[i + 1][1];

    if (
      !Object.is(currentColor, prevColor) &&
      !Object.is(currentColor, nextColor)
    ) {
      // the middle color is different from the other two
      const [, dist] = colorInverseLerp(prevColor, nextColor, currentColor);

      if (dist === 0) {
        // and it can be losslessly recreated by linearly interpolating from
        // the other two
        // => skip the middle one
        continue;
      }
    } else if (Object.is(prevColor, nextColor)) {
      // all three colors are the same
      // => skip the middle one
      continue;
    }

    simplified.push(normalized[i]);
  }
  simplified.push(normalized[normalized.length - 1]);

  return simplified;
}

function createColorPalette(props) {
  var stops = _normalizedColorStops(props.stops);
  if (!stops || stops.length < 2) {
    throw new TypeError(
      'Cannot create a ColorPalette with less than two color stops'
    );
  }

  return Object.assign({
    backgroundColor: stops[0][1],
    foregroundColor: stops[stops.length - 1][1],
    nullColor      : NaN,

    toString() {
      return colorPaletteToCSSLinearGradient(this);
    },
  }, props, {stops});
}
function colorPaletteFromColors(colors, stepped) {
  const stops = [];
  let z = 0;
  for (const color of colors) {
    stops.push([z++, color]);
    if (stepped) {
      stops.push([z, color]);
    }
  }
  return createColorPalette({stops});
}
function colorPaletteFromCSSColors(colors, stepped) {
  return colorPaletteFromColors(colors.map(colorFromCSSColor), stepped);
}

function colorPaletteFromRGB(rgbText) {
  var colors = [];
  var colorLineRegex = new RegExp(rgbLine, 'gm');
  var rgb;

  while ((rgb = colorLineRegex.exec(rgbText)) !== null) {
    var colorComponents = rgb
      .slice(1)
      .map(parseDecimalWithDefault.bind(null, 0));

    colors.push(colorFromArray(colorComponents));
  }

  return colorPaletteFromColors(colors);
}

/**
 * Parse a [CPT][1] color palette table.
 *
 * [1]: https://gmt.soest.hawaii.edu/doc/latest/GMT_Docs.html#color-palette-tables
 *
 * @param  {string}  cptText - The CPT.
 * @return {ColorPalette} The corresponding color palette.
 * @note Assumes RGB color model.
 * @note Assumes Regular CPTs.
 */
function colorPaletteFromCPT(cptText) {
  var stops = [];
  var palette = {stops};
  var cptLines = cptText.split(/[\n\r]+/);
  var matched;

  for (var i = 0, linesCount = cptLines.length; i < linesCount; i++) {
    var cptLine = cptLines[i].trim();

    if (cptLine[0] === '#') {
      // Ignore comments
      continue;
    }

    matched = cptSpecialLineRegex.exec(cptLine);
    if (matched) {
      var cat = matched[1];
      var r = matched[2];
      var g = matched[3];
      var b = matched[4];
      var color = createColor(
        parseDecimal(r),
        parseDecimal(g),
        parseDecimal(b)
      );

      if (colorIsInvalid(color)) {
        console.warn(
          'Error parsing CPT: Found an invalid ' + cat + ' color. (line: ' +
          (i + 1) + ')\n' +
          'Ignoring and continuing.'
        );
        continue;
      }
      switch (cat.toUpperCase()) {
      case 'B':
        palette.backgroundColor = color;
        break;
      case 'F':
        palette.foregroundColor = color;
        break;
      case 'N':
        palette.nullColor = color;
        break;
      default:
      }
      continue;
    }
    matched = cptSliceLineRegex.exec(cptLine);
    if (matched) {
      matched = matched.slice(1).map(parseFloat);
      var z0 = matched[0];
      var r0 = matched[1];
      var g0 = matched[2];
      var b0 = matched[3];
      var z1 = matched[4];
      var r1 = matched[5];
      var g1 = matched[6];
      var b1 = matched[7];
      var color0 = createColor(r0, g0, b0);
      var color1 = createColor(r1, g1, b1);

      if (colorIsInvalid(color0) && !colorIsInvalid(color1)) {
        console.warn(
          'Error parsing CPT: Found a z-slice with min color set to - but ' +
          'not the max color. (line: ' + (i + 1) + ')\n' +
          'Expected a z-slice to have both min and max colors set to -, ' +
          'only max color set to -, or none of min and max set to -.\n' +
          'Ignoring z-slice and continuing.'
        );
        continue;
      }
      if (stops.length > 0 && z0 !== stops[stops.length - 1][0]) {
        console.warn(
          'Error parsing CPT: Found a gap between z-slices. (line: ' + (i + 1) +
          ')\n' +
          'Expected all z-slices to form a contiguous range of z values.\n' +
          'Filling the gap with a z-slice between the two and continuing.'
        );
      }

      stops.push([z0, color0]);

      if (z0 < z1) {
        stops.push([z1, colorIsInvalid(color1) ? color0 : color1]);
      } else {
        console.warn(
          'Error parsing CPT: Found z-slice with z0 >= z1. (line: ' + (i + 1) +
          ')\n' +
          'Expected a z-slice to have z0 < z1.\n' +
          'Ignoring z-slice and continuing.'
        );
      }
    }
  }

  return createColorPalette(palette);
}

function colorPaletteFromPNG(png) {
  var width = png.metadata.width;
  var numberOfChannels = png.metadata.channels;
  var colors = [];

  for (var i = 0; i < width; i++) {
    var index = i * numberOfChannels;
    var color = (
      numberOfChannels < 3
      ? createColor(
        png.pixels[index],
        png.pixels[index],
        png.pixels[index]
      )
      : createColor(
        png.pixels[index],
        png.pixels[index + 1],
        png.pixels[index + 2]
      )
    );
    colors.push(color);
  }

  return colorPaletteFromColors(colors);
}

function addColorPalette(url, id, palette) {
  palette.id = id;
  palette.className = 'ColorPalette -' + palette.id;
  palette.svgId = 'ColorPalette-' + palette.id;
  colorPalettes.set(url, palette);
  colorPalettesStatus.set(url, COLOR_PALETTE_LOADED);
  addColorPaletteToCSS('.ColorPalette.-' + palette.id, palette);
  addColorPaletteToSVG(palette.svgId, palette);
  logColorPalette(palette);
  return palette;
}
function getColorPalette(url) {
  return colorPalettes.get(url) || null;
}
function getColorPaletteStatus(url) {
  return colorPalettesStatus.get(url) || COLOR_PALETTE_MISSING;
}
function fetchColorPalette(url, signal) {
  if (colorPalettes.has(url)) {
    return Promise.resolve(colorPalettes.get(url));
  }

  var [promise, abort] = fetchResource(_fetchColorPalette, url);

  colorPalettesStatus.set(url, COLOR_PALETTE_LOADING);

  if (signal) {
    signal.addEventListener('abort', function handleAbort() {
      signal.removeEventListener('abort', handleAbort);
      abort();
    });
  }
  return promise;
}
function _fetchColorPalette(url, signal) {
  var match = /(?:^|\/)([^/?#]+)\.([^/?#]+)(?:[?#]|$)/.exec(url);
  var baseName = match && match[1];
  var extension = match && match[2].toUpperCase();

  if (!baseName || !extension) {
    return Promise.reject(new Error('Invalid palette URL: "' + url + '"'));
  }

  var ajaxFetcher;
  var parseColorPalette;
  switch (extension) {
  case 'PNG':
    ajaxFetcher = ajax.getPNG;
    parseColorPalette = colorPaletteFromPNG;
    break;
  case 'CPT':
    ajaxFetcher = ajax.get;
    parseColorPalette = colorPaletteFromCPT;
    break;
  case 'RGB':
    ajaxFetcher = ajax.get;
    parseColorPalette = colorPaletteFromRGB;
    break;
  default:
    return Promise.reject(
      new Error('Unknown color palette format: "' + extension + '"')
    );
  }

  return (
    ajaxRetriable(ajaxFetcher)(url, signal)
    .then(parseColorPalette)
    .then(addColorPalette.bind(null, url, baseName))
    .catch(function(error) {
      if (isAbortError(error)) {
        colorPalettesStatus.delete(url);
      } else {
        console.error(
          'ColorPalette: Could not load color palette from URL "' + url + '"!',
          error
        );
        colorPalettesStatus.set(url, COLOR_PALETTE_ERRORED);
      }
      throw error;
    })
  );
}
async function withColorPalette(url, callback) {
  let colorPalette;

  switch (getColorPaletteStatus(url)) {
  case COLOR_PALETTE_LOADED:
    colorPalette = getColorPalette(url);
    break;
  case COLOR_PALETTE_ERRORED:
    colorPalette = null;
    break;
  default:
    colorPalette = (
      await fetchColorPalette(url).catch(emptyFunctionReturningNull)
    );
    break;
  }
  await callback?.(colorPalette);
  return colorPalette;
}

function colorPaletteGetColor(palette, value) {
  var stops = palette.stops;
  var stopsCount = stops.length;

  if (isNaN(value) || stopsCount === 0) {
    return palette.nullColor;
  }
  if (value < 0) {
    return palette.backgroundColor;
  }
  if (value > 1) {
    return palette.foregroundColor;
  }
  if (stopsCount === 1) {
    return stops[0][1];
  }

  for (var i = 1; i < stopsCount; i++) {
    var prevValue = stops[i - 1][0];
    var prevColor = stops[i - 1][1];
    var nextValue = stops[i][0];
    var nextColor = stops[i][1];

    if (prevValue <= value && value < nextValue) {
      var valuePosition = inverseLerp(prevValue, nextValue, value);
      var color = colorLerp(prevColor, nextColor, valuePosition);

      return color;
    }
  }

  return stops[stopsCount - 1][1];
}
function colorPaletteGetColorBetween(palette, min, max, value) {
  return colorPaletteGetColor(palette, inverseLerp(min, max, value));
}

function colorPaletteGetValue(palette, color) {
  const stops = palette.stops;
  const stopsCount = stops.length;

  if (colorIsInvalid(color) || stopsCount === 0) {
    return NaN;
  }

  let minDist = Infinity;
  let minValue;
  for (let i = 1; i < stopsCount; i++) {
    const prevValue = stops[i - 1][0];
    const prevColor = stops[i - 1][1];
    const nextValue = stops[i][0];
    const nextColor = stops[i][1];
    const [position, dist] = colorInverseLerp(prevColor, nextColor, color);

    if (dist < minDist && inRangeInclusive(position, 0, 1)) {
      minDist = dist;
      if (position === 0) {
        minValue = prevValue;
      } else if (position === 1) {
        minValue = nextValue;
      } else {
        minValue = (nextValue - prevValue) * position + prevValue;
      }
    }
  }

  if (minDist < 5) {
    return minValue;
  }
  if (color === palette.backgroundColor) {
    return -Infinity;
  }
  if (color === palette.foregroundColor) {
    return Infinity;
  }
  return NaN;
}
function colorPaletteGetValueBetween(palette, min, max, color) {
  return lerp(min, max, colorPaletteGetValue(palette, color));
}

function colorPaletteGetCSSClassName(palette) {
  return palette.className;
}
function colorPaletteGetSVGGradiantId(palette) {
  return palette.svgId;
}
function colorPaletteGetSVGFill(palette) {
  return 'url(#' + palette.svgId + ')';
}

function colorStopsToSVGStops(stops) {
  var stopsElement = document.createDocumentFragment();
  for (var i = 0, l = stops.length; i < l; i++) {
    var stop = stops[i];
    var position = stop[0];
    var colorWithAlpha = stop[1];
    var colorWithoutAlpha = colorSetAlpha(colorWithAlpha, 1);
    var offset = position * 100 + '%';
    var color = colorToCSSColor(colorWithoutAlpha);
    var opacity = colorGetAlpha(colorWithAlpha);

    var stopElement = document.createElementNS(SVG_NS, 'stop');

    stopElement.setAttribute('offset', offset);
    stopElement.setAttribute('stop-color', color);
    if (opacity < 1) {
      stopElement.setAttribute('stop-opacity', opacity);
    }
    stopsElement.appendChild(stopElement);
  }
  return stopsElement;
}
function colorStopsToCSSStops(stops) {
  return stops
    .map(function(stop) {
      var offset = stop[0] * 100 + '%';
      var color = colorToCSSColor(stop[1]);
      return color + ' ' + offset;
    })
    .join(',');
}
function colorPaletteToSVGStops(palette) {
  return colorStopsToSVGStops(palette.stops);
}
function colorPaletteToSVGLinearGradient(palette) {
  var linearGradientElement = document.createElementNS(SVG_NS, 'linearGradient');
  linearGradientElement.appendChild(colorPaletteToSVGStops(palette));
  return linearGradientElement;
}
function colorPaletteToCSSStops(palette) {
  return colorStopsToCSSStops(palette.stops);
}
function colorPaletteToCSSLinearGradient(palette) {
  return 'linear-gradient(to right, ' + colorPaletteToCSSStops(palette) + ')';
}
function logRect(width, background) {
  return (
    'margin-right: 1px; padding: 0 ' + width / 2 + 'ch; background: ' +
    background
  );
}
function logColor(color) {
  return (
    colorIsInvalid(color)
    ? (
      'linear-gradient(45deg, transparent 45%, red 50%, transparent 55%),' +
      'linear-gradient(-45deg, transparent 45%, red 50%, transparent 55%)'
    )
    : colorToCSSColor(color)
  );
}
function logColorPalette(palette, id) {
  console.log(
    'ColorPalette: %c %c %c %c %c' + (id || palette.id || ' '),
    logRect(1, logColor(palette.nullColor)),
    logRect(1, logColor(palette.backgroundColor)),
    logRect(30, colorPaletteToCSSLinearGradient(palette)),
    logRect(1, logColor(palette.foregroundColor)),
    'background: none'
  );
}

export {
  COLOR_PALETTE_MISSING,
  COLOR_PALETTE_LOADING,
  COLOR_PALETTE_LOADED,
  COLOR_PALETTE_ERRORED,

  createColorPalette,
  colorPaletteFromColors,
  colorPaletteFromCSSColors,
  colorPaletteFromRGB,
  colorPaletteFromCPT,
  colorPaletteFromPNG,
  addColorPalette,
  getColorPalette,
  getColorPaletteStatus,
  fetchColorPalette,
  withColorPalette,
  colorPaletteGetColor,
  colorPaletteGetColorBetween,
  colorPaletteGetValue,
  colorPaletteGetValueBetween,
  colorPaletteGetCSSClassName,
  colorPaletteGetSVGGradiantId,
  colorPaletteGetSVGFill,
  colorStopsToSVGStops,
  colorStopsToCSSStops,
  colorPaletteToSVGLinearGradient,
  colorPaletteToCSSLinearGradient,
  logColorPalette,
};
window.ODL = Object.assign(window.ODL || {}, {
  COLOR_PALETTE_MISSING,
  COLOR_PALETTE_LOADING,
  COLOR_PALETTE_LOADED,
  COLOR_PALETTE_ERRORED,

  createColorPalette,
  colorPaletteFromColors,
  colorPaletteFromCSSColors,
  colorPaletteFromRGB,
  colorPaletteFromCPT,
  colorPaletteFromPNG,
  addColorPalette,
  getColorPalette,
  getColorPaletteStatus,
  fetchColorPalette,
  withColorPalette,
  colorPaletteGetColor,
  colorPaletteGetColorBetween,
  colorPaletteGetValue,
  colorPaletteGetValueBetween,
  colorPaletteGetCSSClassName,
  colorPaletteGetSVGGradiantId,
  colorPaletteGetSVGFill,
  colorStopsToSVGStops,
  colorStopsToCSSStops,
  colorPaletteToSVGLinearGradient,
  colorPaletteToCSSLinearGradient,
  logColorPalette,
});
