/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  NONPASSIVE_CAPTURING,
  preventEventDefault,
  removeAllChildrenOfNode,
} from '../syntool.js';

function removeNodes(nodes) {
  var parent;
  for (const node of nodes) {
    if (node && (parent = node.parentNode)) {
      parent.removeChild(node);
    }
  }
}
function appendChildrenToNode(node, children) {
  for (const child of children) {
    node.appendChild(child);
  }
}
/**
 * @param {Readonly<Node>} node
 * @param {Readonly<Iterable<Node | DocumentFragment>>} children
 * @returns {void}
 */
function setNodeChildren(node, children) {
  children = Array.isArray(children) ? children : [...children];

  if (children.length === 0) {
    removeAllChildrenOfNode(node);
    return;
  }

  // Update the DOM
  let current = node.firstChild;
  for (const child of children) {
    if (child) {
      if (child instanceof DocumentFragment) {
        for (const grandchild of child.children) {
          if (grandchild === current) {
            current = current.nextSibling;
          } else {
            node.insertBefore(grandchild, current);
          }
        }
      } else if (child === current) {
        current = current.nextSibling;
      } else {
        node.insertBefore(child, current);
      }
    }
  }

  // Remove the remaining children that are missing from the `children` param
  while (current) {
    const next = current.nextSibling;
    node.removeChild(current);
    current = next;
  }
}

const SVG_NS = 'http://www.w3.org/2000/svg';

function createSVGDefs() {
  var element = (
    document.body &&
    document.body.insertBefore(
      document.createElementNS(SVG_NS, 'svg'),
      document.body.firstChild
    )
  );
  if (element) {
    element.setAttribute('width', 0);
    element.setAttribute('height', 0);
    element.style.position = 'fixed';
    return element.appendChild(document.createElementNS(SVG_NS, 'defs'));
  }
  return null;
}
function createStyleSheet() {
  var element = document.head?.appendChild(document.createElement('style'));

  if (element) {
    element.type = 'text/css';
    return element.sheet;
  }
  return null;
}
function addCSSRuleToSheet(sheet, rule) {
  if (sheet) {
    sheet.insertRule(rule, sheet.cssRules.length);
    return true;
  }
  return false;
}

var linkStyleSheetPromises = new Map();
var linkStyleSheetsAfter = (
  document.querySelector('link[rel="stylesheet"][data-syntool]')
);

function linkStyleSheet(href, baseUrl) {
  if (baseUrl) {
    href = (new URL(href, baseUrl)).href;
  }
  if (linkStyleSheetPromises.has(href)) {
    return Promise.resolve(linkStyleSheetPromises.get(href));
  }

  var promise = new Promise(function(resolve, reject) {
    var link = document.createElement('link');
    link.rel = 'stylesheet';
    link.type = 'text/css';
    link.onerror = function(event) {
      if (linkStyleSheetsAfter === link) {
        linkStyleSheetsAfter = link.previousSibling;
      }
      link.remove();
      link = undefined;
      reject(new URIError(`Could not load: ${event.target.href}`));
    };
    link.onload = function(event) {
      link.onerror = null;
      link.onload = null;
      link = undefined;
      resolve(event.target);
    };
    link.href = href;
    if (linkStyleSheetsAfter.insertAdjacentElement('afterend', link)) {
      linkStyleSheetsAfter = link;
    } else {
      link = undefined;
      reject(new URIError(`Could add link to DOM to load: ${href}`));
    }
  });
  linkStyleSheetPromises.set(href, promise);
  return promise;
}

var EMPTY_ELEMENT = document.createElement('div');

function supportsCSSProperty(property, values) {
  var style = EMPTY_ELEMENT.style;
  style.cssText = property + ':' + values.join(';' + property + ':') + ';';
  return !!style[property];
}
const supportsPixelated = (
  supportsCSSProperty('-ms-interpolation-mode', ['nearest-neighbor']) ||
  supportsCSSProperty('image-rendering', [
    'optimizeSpeed',
    '-moz-crisp-edges',
    '-o-crisp-edges',
    '-webkit-optimize-contrast',
    'pixelated',
    'optimize-contrast',
  ])
);

function disableImageDrag(element) {
  (element || window).addEventListener(
    'dragstart',
    preventEventDefault,
    NONPASSIVE_CAPTURING
  );
}
function enableImageDrag(element) {
  (element || window).removeEventListener(
    'dragstart',
    preventEventDefault,
    NONPASSIVE_CAPTURING
  );
}

function isModifiedEvent(event) {
  return !!(event.metaKey || event.altKey || event.ctrlKey || event.shiftKey);
}
function isMouseEvent(event) {
  return event.type.indexOf('mouse') === 0 || event instanceof MouseEvent;
}

function importScript(src, globalName) {
  return new Promise(function(resolve, reject) {
    var script = document.createElement('script');
    script.async = 1;
    script.onerror = function(event) {
      reject(new URIError(`Could not load: ${event.target.src}`));
    };
    script.onload = function(event) {
      script.onload = null;
      script = undefined;
      resolve(window[globalName]);
    };
    script.src = src;

    var prior = document.getElementsByTagName('script')[0];
    prior.parentNode.insertBefore(script, prior);
  });
}

export {
  SVG_NS,

  removeNodes,
  appendChildrenToNode,
  setNodeChildren,

  createSVGDefs,
  createStyleSheet,
  addCSSRuleToSheet,

  linkStyleSheet,

  supportsCSSProperty,
  supportsPixelated,

  disableImageDrag,
  enableImageDrag,

  isModifiedEvent,
  isMouseEvent,

  importScript,
};
window.ODL = Object.assign(window.ODL || {}, {
  SVG_NS,

  removeNodes,
  appendChildrenToNode,
  setNodeChildren,

  createSVGDefs,
  createStyleSheet,
  addCSSRuleToSheet,

  linkStyleSheet,

  supportsCSSProperty,
  supportsPixelated,

  disableImageDrag,
  enableImageDrag,

  isModifiedEvent,
  isMouseEvent,

  importScript,
});
