/* eslint-disable max-lines */
/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  idlRangeWorldsAway,
  idlRangeWrapCount,
  upsert,
  isObjectEmpty,
  asArray,
  shallowEqual,
  areXYsEqual,
  bisectRight,
  removeAllChildrenOfNode,
  isSafari,
  S,
  Bounds,
  Geometry,
} from '../syntool.js';

// The maximum safe value of a translate in SVG. This limit varies between
// browsers but in general it's due to the float32's limit to 23 bits of
// significand precision.
//
// eslint-disable-next-line no-bitwise
const MAX_TRANSLATE = 1 << 23;

const NODE_STACK_LEVEL = Symbol('ODL.FeatureRenderer:NODE_STACK_LEVEL');
const NODE_FEATURE_ID = Symbol('ODL.FeatureRenderer:NODE_FEATURE_ID');
const NODE_STYLE = Symbol('ODL.FeatureRenderer:NODE_STYLE');
const NODE_PROPS = Symbol('ODL.FeatureRenderer:NODE_PROPS');
const NODE_SYMBOL_NAME = Symbol('ODL.FeatureRenderer:NODE_SYMBOL_NAME');
const XLINK_NS = 'http://www.w3.org/1999/xlink';
const HIDDEN_STYLE = {display: 'none'};
const LABEL_ALIGN = {
  l: 'start',
  r: 'end',
  b: 'bottom',
  t: 'hanging'
};
const LABEL_VSHIFT = {
  // according to
  // http://www.w3.org/Graphics/SVG/Test/20061213/htmlObjectHarness/full-text-align-02-b.html
  // a baseline-shift of -70% shifts the text exactly from the
  // bottom to the top of the baseline, so -35% moves the text to
  // the center of the baseline.
  t: '-70%',
  b: '0'
};
const LABEL_VFACTOR = {
  t: 0,
  b: -1
};

const tempSet = new Set();

/**
 * stroke - {Boolean} Set to false if no stroke is desired.
 * strokeColor - {String} Hex stroke color.  Default is '#000000'.
 * strokeOpacity - {Number} Stroke opacity (0-1).  Default is 1.
 * strokeWidth - {Number} Pixel stroke width.  Default is 1.
 * strokeLinecap - {String} Stroke cap type.
 *      Supported values include 'butt', 'round' (default), 'square'.
 * strokeLinejoin - {String} Stroke join type.  Default is 'round'.
 *      Supported values include 'arcs', 'bevel', 'miter', 'miter-clip', 'round' (default).
 * strokeDashstyle - {String} Stroke dash style.
 *      Supported values include 'dot', 'dash', 'dashdot', 'longdash', 'longdashdot', 'solid' (default).
 * fill - {Boolean} Set to false if no fill is desired.
 * fillColor - {String} Hex fill color.  Default is '#FFFFFF'.
 * fillOpacity - {Number} Fill opacity (0-1).  Default is 0.4
 * graphic - {Boolean} Set to false if no graphic is desired.
 * pointRadius - {Number} Pixel point radius.  Default is 6.
 * externalGraphic - {String} URL to an external graphic that will be used for rendering points.
 * graphicName - {String} Named graphic to use when rendering points.
 *      Supported values include 'circle' (default), 'square', 'star', 'x', 'cross', 'triangle'.
 * graphicOpacity - {Number} Opacity (0-1) for an external graphic.
 * graphicWidth - {Number} Pixel width for sizing an external graphic.
 * graphicHeight - {Number} Pixel height for sizing an external graphic.
 * graphicXOffset - {Number} Pixel offset along the positive x axis for displacing an external graphic.
 * graphicYOffset - {Number} Pixel offset along the positive y axis for displacing an external graphic.
 * graphicHFlip - {Boolean} Set to true to flip horizontally the external graphic.
 * graphicVFlip - {Boolean} Set to true to flip vertically the external graphic.
 * graphicRotation - {Number} For point, this is the rotation of a graphic in the clockwise direction about its
 *      center point (or any point off center as specified by graphicXOffset and graphicYOffset).
 * pointerEvents - {String} Sets the 'pointer-events' SVG attribute.
 * cursor - {String} Sets the 'cursor' SVG attribute.
 * label - {String} The text for an optional label.
 * labelAlign - {String} Label alignment. This specifies the insertion point relative to the text. It is a string
 *      composed of two characters. The first character is for the horizontal alignment, the second for the vertical
 *      alignment. Valid values for horizontal alignment: 'l'=left, 'c'=center, 'r'=right. Valid values for vertical
 *      alignment: 't'=top, 'm'=middle, 'b'=bottom. Example values: 'lt', 'cm', 'rb'. Default is 'cm'.
 * labelXOffset - {Number} Pixel offset along the positive x axis for displacing the label.
 * labelYOffset - {Number} Pixel offset along the positive y axis for displacing the label.
 * labelSelect - {Boolean} If set to true, labels will be selectable.
 * labelOutlineColor - {String} The color of the label outline. Default is '#FFFFFF'.
 * labelOutlineWidth - {Number} The width of the label outline. Default is 3, set to 0 or null to disable.
 * fontColor - {String} The font color for the label, to be provided like CSS. Default is '#000000'.
 * fontOpacity - {Number} Opacity (0-1) for the label
 * fontFamily - {String} The font family for the label, to be provided like in CSS.
 * fontSize - {String} The font size for the label, to be provided like in CSS.
 * fontStyle - {String} The font style for the label, to be provided like in CSS.
 * fontWeight - {String} The font weight for the label, to be provided like in CSS.
 * display - {String} Element will not be drawn if display is set to 'none'. All other values have no effect.
 * opacity - {Number}
 * lineHitWidth - {Number} Pixel stroke width of the hit area of lines.  Default is 0.
 * noWrap - {Boolean} Set to true to disable wrapping features.  Default false.
 * __show_virtual - {Boolean} Used internally
 */
const DEFAULT_STYLE = {
  stroke: true,
  strokeColor: '#000000',
  strokeOpacity: 1,
  strokeWidth: 2,
  strokeLinecap: 'round',
  strokeLinejoin: 'round',
  strokeDashstyle: 'solid',

  fill: true,
  fillColor: '#FFFFFF',
  fillOpacity: 1,

  graphic: true,
  pointRadius: 6,
  externalGraphic: false,
  graphicName: 'circle',
  graphicOpacity: 1,
  graphicWidth: 0,
  graphicHeight: 0,
  graphicXOffset: undefined,
  graphicYOffset: undefined,
  graphicHFlip: false,
  graphicVFlip: false,
  graphicRotation: 0,
  pointerEvents: null,
  cursor: null,

  label: '',
  labelAlign: 'cm',
  labelXOffset: 0,
  labelYOffset: 0,
  labelSelect: false,
  labelOutlineColor: '#FFFFFF',
  labelOutlineWidth: 2,
  fontColor: '#000000',
  fontOpacity: null,
  fontFamily: null,
  fontSize: null,
  fontStyle: null,
  fontWeight: null,

  display: '',
  opacity: 1,
  lineHitWidth: 0,
  noWrap: false,
  __show_virtual: false,
};

function findNextElement(container, stackLevel) {
  if (
    !container.lastElementChild ||
    container.lastElementChild[NODE_STACK_LEVEL] <= stackLevel
  ) {
    return null;
  }
  const index = bisectRight(
    container.children,
    (node) => stackLevel - node[NODE_STACK_LEVEL]
  );
  return container.children[index];
}
function removeNodes(nodes) {
  if (nodes) {
    for (const node of nodes) {
      node?.remove();
    }
    return true;
  }
  return false;
}
function getDep(key) {
  return this?.[key] ?? DEFAULT_STYLE[key];
}
function createUpdateMap(props, ...maps) {
  const depsMap = {};
  for (const [key, value] of Object.entries(props)) {
    const deps = (
      typeof value === 'string' ? [value]
      : typeof value === 'function' ? [key]
      : asArray(value[0])
    );
    const mapper = (
      typeof value === 'string' ? (v) => v
      : typeof value === 'function' ? value
      : value[1]
    );
    // eslint-disable-next-line func-style
    const handler = (out, obj) => {
      out[key] = mapper(...deps.map(getDep, obj));
    };

    for (const dep of deps) {
      upsert(depsMap, dep, []).push(handler);
    }
  }
  for (const map of maps) {
    for (const key of map.keys) {
      upsert(depsMap, key, []).push(...map.depsMap[key]);
    }
  }

  const keys = Object.keys(depsMap);
  getUpdates.keys = keys;
  getUpdates.depsMap = depsMap;

  function getUpdates(prev, next) {
    const updates = {};
    for (const key of keys) {
      if (!Object.is(prev?.[key], next?.[key])) {
        for (const update of depsMap[key]) {
          update(updates, next);
        }
      }
    }
    return updates;
  }

  return getUpdates;
}

const textAttributes = createUpdateMap({
  fill: 'fontColor',
  stroke: 'labelOutlineColor',
  'stroke-width': 'labelOutlineWidth',
  opacity: 'fontOpacity',
  'font-family': 'fontFamily',
  'font-size': [
    'fontSize',
    (fontSize) => (typeof fontSize === 'number' ? `${fontSize}px` : fontSize),
  ],
  'font-style': 'fontStyle',
  'font-weight': 'fontWeight',
  'pointer-events': [
    'labelSelect',
    (labelSelect) => (labelSelect ? 'visible' : 'none'),
  ],
  cursor: [
    ['cursor', 'labelSelect'],
    (cursor, labelSelect) => (labelSelect ? cursor : null),
  ],
  'text-anchor': [
    'labelAlign',
    (labelAlign) => LABEL_ALIGN[labelAlign[0]] || 'middle',
  ],
  'dominant-baseline': [
    'labelAlign',
    (labelAlign) => LABEL_ALIGN[labelAlign[1]] || 'central',
  ],
  style: [
    'opacity',
    (opacity) => (opacity === 1 ? undefined : `opacity: ${opacity}`),
  ],
});
const geometryAttributes = createUpdateMap({
  'pointer-events': [
    'pointerEvents',
    (pointerEvents) => pointerEvents || undefined,
  ],
  cursor: (cursor) => cursor || undefined,
});
const fillAttributes = createUpdateMap({
  fill: [
    ['fill', 'fillColor'],
    (fill, fillColor) => (fill ? fillColor : 'none'),
  ],
  'fill-opacity': [
    ['fill', 'fillOpacity'],
    (fill, fillOpacity) => (fill ? fillOpacity : undefined),
  ],
});
const strokeAttributes = createUpdateMap({
  stroke: [
    ['stroke', 'strokeColor'],
    (stroke, strokeColor) => (stroke ? strokeColor : 'none'),
  ],
  'stroke-opacity': [
    ['stroke', 'strokeOpacity'],
    (stroke, strokeOpacity) => (stroke ? strokeOpacity : undefined),
  ],
  'stroke-width': [
    ['stroke', 'strokeWidth', '_graphicScale'],
    (stroke, strokeWidth, scale = 1) => (
      stroke ? (strokeWidth / scale) : undefined
    ),
  ],
  'stroke-linecap': [
    ['stroke', 'strokeLinecap'],
    (stroke, strokeLinecap) => (stroke ? strokeLinecap : undefined),
  ],
  'stroke-linejoin': [
    ['stroke', 'strokeLinejoin'],
    (stroke, strokeLinejoin) => (stroke ? strokeLinejoin : undefined),
  ],
  'stroke-dasharray': [
    ['stroke', 'strokeWidth', 'strokeDashstyle', '_graphicScale'],
    (stroke, strokeWidth, strokeDashstyle, scale = 1) => {
      if (stroke && strokeWidth && strokeDashstyle) {
        const w = strokeWidth / scale;
        switch (strokeDashstyle) {
        case 'solid':
          return 'none';
        case 'dot':
          return [1, 4 * w].join();
        case 'dash':
          return [4 * w, 4 * w].join();
        case 'dashdot':
          return [4 * w, 4 * w, 1, 4 * w].join();
        case 'longdash':
          return [8 * w, 4 * w].join();
        case 'longdashdot':
          return [8 * w, 4 * w, 1, 4 * w].join();
        default:
          return strokeDashstyle.trim().replace(/\s+/g, ',');
        }
      }
      return undefined;
    },
  ],
});
const opacityAttributes = createUpdateMap({
  style: [
    'opacity',
    (opacity) => (opacity === 1 ? undefined : `opacity: ${opacity}`),
  ],
});
const pointAttributes = createUpdateMap({
  style: [
    ['graphic', 'graphicOpacity', 'opacity'],
    (graphic, graphicOpacity, opacity) => (
      (graphic ? '' : 'visibility: hidden;') +
      (
        opacity === 1 && graphicOpacity === 1
        ? ''
        : `opacity: ${opacity * graphicOpacity}`
      )
    ) || undefined,
  ],
});
const imageAttributes = createUpdateMap({
  x: [
    ['graphicXOffset', 'graphicWidth', 'graphicHeight'],
    (graphicXOffset, graphicWidth, graphicHeight) => (
      graphicXOffset ?? -(graphicWidth || graphicHeight) / 2
    ),
  ],
  y: [
    ['graphicYOffset', 'graphicWidth', 'graphicHeight'],
    (graphicYOffset, graphicWidth, graphicHeight) => (
      graphicYOffset ?? -(graphicHeight || graphicWidth) / 2
    ),
  ],
  width: [
    ['graphicWidth', 'graphicHeight'],
    (graphicWidth, graphicHeight) => (graphicWidth || graphicHeight),
  ],
  height: [
    ['graphicWidth', 'graphicHeight'],
    (graphicWidth, graphicHeight) => (graphicHeight || graphicWidth),
  ],
  href: 'externalGraphic',
  preserveAspectRatio: [
    ['graphicWidth', 'graphicHeight'],
    (graphicWidth, graphicHeight) => (
      graphicWidth && graphicHeight ? 'none' : undefined
    ),
  ],
});

const symbolMap = new Map([
  /*eslint-disable comma-spacing */
  ['star', [
    231,-38, 260,48, 350,48, 278,102, 304,188, 231,137, 158,188,
    184,102, 112,48, 202,48, 231,-38
  ]],
  ['cross', [
    -1,-5, 1,-5, 1,-1, 5,-1, 5,1, 1,1, 1,5, -1,5, -1,1, -5,1,
    -5,-1, -1,-1, -1,-5
  ]],
  ['x', [
    -50,-50, -25,-50, 0,-15, 25,-50, 50,-50, 15,0, 50,50, 25,50, 0,15,
    -25,50, -50,50, -15,0, -50,-50
  ]],
  ['square', [-.5,-.5, -.5,.5, .5,.5, .5,-.5, -.5,-.5]],
  ['triangle', [-5,5, 5,5, 0,-5, -5,5]],
  /*eslint-enable comma-spacing */

  ['circle', {node: S('circle', {r: 10}), width: 20, height: 20}],
]);

export class FeaturesRenderer {
  static symbols = symbolMap;
  static defaultStyle = DEFAULT_STYLE;

  constructor(container) {
    this.container = container;
    this.mapView = null;
    this.extent = null;
    this.size = null;

    this._containerId = container.id;
    this._root = this.container.appendChild(
      this._get('svg', null, {'xmlns:xlink': XLINK_NS})
    );
    this._vectorRoot = null;
    this._textRoot = null;

    this._resolution = null;
    this._xOffset = 0;
    this._yOffset = 0;
    this._featureDx = 0;
    this._featureDy = 0;
    this._visible = false;
    this._appliesDatasetProperties = false;

    this._symbols = new Map();
    this._wrapMap = new WeakMap();

    this._productGroups = new Map();
    this._datasetGroups = new Map();

    this._datasetEventHandlers = new WeakMap();

    this._dx = 0;
    this._dy = 0;
  }

  get appliesDatasetProperties() {
    return this._appliesDatasetProperties;
  }
  set appliesDatasetProperties(appliesDatasetProperties) {
    appliesDatasetProperties = !!appliesDatasetProperties;
    if (this._appliesDatasetProperties !== appliesDatasetProperties) {
      if (this._appliesDatasetProperties) {
        // Restore opacity to 100% and visibility to true
        for (const dataset of this._datasetGroups.keys()) {
          if (dataset) {
            this._setGroupsOpacity(this._datasetGroups, dataset, 1);
            this._setGroupsVisibility(this._datasetGroups, dataset, true);
          }
        }
      }

      this._appliesDatasetProperties = appliesDatasetProperties;

      if (this._appliesDatasetProperties) {
        // Apply dataset properties
        for (const dataset of this._datasetGroups.keys()) {
          if (dataset) {
            this._handleDatasetPropertiesChange(dataset);
          }
        }
      }
    }
  }

  destroy() {
    this.clear();

    this.container = null;
    this.mapView = null;
    this.extent = null;
    this.size = null;

    this._root = null;
    this._vectorRoot = null;
    this._textRoot = null;

    this._resolution = null;
    this._symbols = null;
    this._wrapMap = null;

    this._productGroups = null;
    this._datasetGroups = null;

    this._datasetEventHandlers = null;
  }

  setExtent(extent) {
    this._resolution = null;
    this.extent = Bounds.createFromBounds(extent);
    this.setOffset(0, 0);
  }
  setSize(width, height) {
    this.size = [width, height];
    this._resolution = null;

    this._update(this._root, {width, height});
    this.setOffset(0, 0);
  }

  resetOffset() {
    this.setOffset(-this._dx, -this._dy);
  }
  setOffset(dx, dy) {
    this._dx += dx;
    this._dy += dy;

    if (this.extent && this.size) {
      const resolution = this._getResolution();
      const globalLeft = Bounds.getLeft(this.extent) / resolution + this._dx;
      const globalTop = -Bounds.getTop(this.extent)  / resolution + this._dy;
      const xOffset = Math.round(globalLeft / MAX_TRANSLATE) * MAX_TRANSLATE;
      const yOffset = Math.round(globalTop / MAX_TRANSLATE) * MAX_TRANSLATE;
      const left = globalLeft - xOffset;
      const top = globalTop - yOffset;

      this._xOffset = xOffset * resolution;
      this._yOffset = yOffset * resolution;

      this._update(this._root, {
        viewBox: `${left} ${top} ${this.size[0]} ${this.size[1]}`,
      });
    }
  }

  setVisibility(visible) {
    if (!!visible !== this._visible) {
      this._visible = !!visible;
      if (this._productGroups.size > 0) {
        if (this._visible) {
          this._observeProductsStackLevel();
        } else {
          this._unobserveProductsStackLevel();
        }
      }
    }
  }
  getVisibility() {
    return this._visible;
  }

  invalidate() {
    /* empty */
  }

  isFeatureRendered(feature) {
    return feature?.rendererInfo === this;
  }
  isFeatureVisible(feature) {
    return (
      this._visible &&
      !!(
        this._getById(this._id(feature.id)) ||
        this._getById(this._featureTextId(feature.id))
      )
    );
  }
  drawFeature(feature, style) {
    if (this._drawFeature(feature, style)) {
      feature.rendererInfo = this;
      return true;
    }

    if (feature?.rendererInfo === this) {
      this._removeFeatureNodes(feature);
      this._removeAllEmptyGroups([feature.dataset]);
      feature.rendererInfo = null;
    }
    return false;
  }
  eraseFeature(feature) {
    if (feature?.rendererInfo === this) {
      this._removeFeatureNodes(feature);
      this._removeAllEmptyGroups([feature.dataset]);
    }
  }
  eraseFeatures(features) {
    const datasets = tempSet;
    for (const feature of asArray(features)) {
      if (feature?.rendererInfo === this) {
        this._removeFeatureNodes(feature);
        datasets.add(feature.dataset);
      }
    }
    this._removeAllEmptyGroups(datasets);
    datasets.clear();
  }
  eraseAllFeaturesOfDataset(dataset) {
    this._removeDatasetGroups(dataset);
    this._removeAllEmptyGroups([dataset]);
  }
  eraseAllFeaturesOfProductWithId(productId) {
    this._removeProductGroups(productId);
  }
  clear() {
    if (this._vectorRoot) {
      removeAllChildrenOfNode(this._vectorRoot);
    }
    if (this._textRoot) {
      removeAllChildrenOfNode(this._textRoot);
    }
    for (const dataset of this._datasetGroups.keys()) {
      this._unobserveDatasetProperties(dataset);
    }
    this._unobserveProductsStackLevel();

    this._wrapMap = new WeakMap();

    this._productGroups.clear();
    this._datasetGroups.clear();
  }

  getFeatureIdFromEvent(event) {
    let target = event.target;
    if (!this._root.contains(target)) {
      return null;
    }
    while (
      target &&
      !this._getNodeFeatureId(target) &&
      target !== this._root &&
      target !== this.mapView.node
    ) {
      target = target.parentNode;
    }
    return target && this._getNodeFeatureId(target);
  }

  /*Protected*/

  _observeDatasetProperties(dataset) {
    let handler = this._datasetEventHandlers.get(dataset);
    if (!handler) {
      handler = () => this._handleDatasetPropertiesChange(dataset);
      dataset.hub
        .on('change:opacity', handler)
        .on('change:visibility', handler)
        .on('change:stackLevel', handler);
      this._datasetEventHandlers.set(dataset, handler);
    }
  }
  _unobserveDatasetProperties(dataset) {
    const handler = this._datasetEventHandlers.get(dataset);
    if (handler) {
      dataset.hub
        .off('change:opacity', handler)
        .off('change:visibility', handler)
        .off('change:stackLevel', handler);
      this._datasetEventHandlers.delete(dataset);
    }
  }
  _handleDatasetPropertiesChange(dataset) {
    if (this._appliesDatasetProperties) {
      const opacity = dataset.getOpacity();
      const visible = dataset.isVisible();
      this._setGroupsOpacity(this._datasetGroups, dataset, opacity);
      this._setGroupsVisibility(this._datasetGroups, dataset, visible);
    }
    const stackLevel = dataset.getStackLevel();
    this._setGroupsStackLevel(this._datasetGroups, dataset, stackLevel);
  }
  _observeProductsStackLevel() {
    this.mapView.syntool.productsStore.hub.on(
      'change:stackLevel',
      this._handleProductsStackLevelChange,
      this
    );
  }
  _unobserveProductsStackLevel() {
    this.mapView.syntool.productsStore.hub.off(
      'change:stackLevel',
      this._handleProductsStackLevelChange,
      this
    );
  }
  _getStackLevelOfProductWithId(productId) {
    const product = this.mapView.syntool.productsStore.getProduct(productId);
    return product ? product.stackLevel : 0;
  }
  _handleProductsStackLevelChange(productId) {
    if (productId) {
      const stackLevel = this._getStackLevelOfProductWithId(productId);
      this._setGroupsStackLevel(this._productGroups, productId, stackLevel);
    } else {
      for (const id of this._productGroups.keys()) {
        if (id) {
          this._handleProductsStackLevelChange(id);
        }
      }
    }
  }

  _getResolution() {
    if (!this._resolution) {
      this._resolution = this.mapView.getResolution();
    }
    return this._resolution;
  }

  _setNodeFeatureId(node, featureId) {
    node[NODE_FEATURE_ID] = featureId;
  }
  _getNodeFeatureId(node) {
    return node[NODE_FEATURE_ID];
  }
  _unsetNodeFeatureId(node) {
    node[NODE_FEATURE_ID] = undefined;
  }

  _saveNodeStyle(node, style) {
    node[NODE_STYLE] = style;
  }
  _getNodeStyle(node) {
    return node[NODE_STYLE];
  }
  _nodeStyleChanged(node, style, props) {
    return !shallowEqual(node[NODE_STYLE], style, props);
  }

  _saveNodeProps(node, props) {
    node[NODE_PROPS] = props;
  }
  _getNodeProps(node) {
    return node[NODE_PROPS];
  }
  _nodePropsChanged(node, props) {
    return !shallowEqual(node[NODE_PROPS], props, Object.keys(props));
  }

  _id(id, suffix = '') {
    return `${this._containerId}.${id}.${suffix}`;
  }
  _featureTextId(featureId) {
    return this._id(featureId, 'label');
  }
  _getById(id) {
    return (
      document.contains(this._root)
      ? document.getElementById(id)
      : this._root.querySelector('#' + CSS.escape(id))
    );
  }
  _get(tag, id, props) {
    const node = id && this._getById(id);
    if (node) {
      if (node.nodeName === tag) {
        return node;
      }
      this._removeWithWraps(node);
    }
    if (!props) {
      props = {};
    }
    if (!props.$parent) {
      props.$parent = this._root;
    }
    if (id) {
      props.id = id;
    }
    return S(tag, props, null);
  }
  _update(node, props) {
    if (isObjectEmpty(props)) {
      return node;
    }
    if (!node.ownerSVGElement) {
      if (!props) {
        props = {};
      }
      if (!props.$parent) {
        props.$parent = this._root;
      }
    }
    return S(node, props, null);
  }

  _getGroup(groupMap, key, container, index) {
    let groups = groupMap.get(key);
    if (!groups) {
      groups = [null, null];
      groupMap.set(key, groups);

      if (key) {
        if (groupMap === this._datasetGroups) {
          this._observeDatasetProperties(key);
        } else if (groupMap.size === 1) {
          // groupMap === this._productGroups
          this._observeProductsStackLevel();
        }
      }
    }
    let group = groups[index];
    let referenceNode = null;
    if (!group) {
      group = this._get('g', key?.id ? this._id(key.id, index) : null);
      groups[index] = group;

      let stackLevel = 0;
      if (key) {
        if (groupMap === this._datasetGroups) {
          if (this._appliesDatasetProperties) {
            const opacity = key.getOpacity();
            if (opacity < 1) {
              group.style.opacity = opacity;
            }
            const visible = key.isVisible();
            if (!visible) {
              group.style.display = 'none';
            }
          }
          stackLevel = key.getStackLevel();
        } else {
          // groupMap === this._productGroups
          stackLevel = this._getStackLevelOfProductWithId(key);
        }
      }
      group[NODE_STACK_LEVEL] = stackLevel;
      referenceNode = findNextElement(container, stackLevel);
    }
    if (group.parentNode !== container) {
      container.insertBefore(group, referenceNode);
    }
    return group;
  }
  _setGroupsOpacity(groupMap, key, opacity) {
    const groups = groupMap.get(key);
    if (groups) {
      const opacityStyle = (opacity < 1 ? opacity : '');
      for (const group of groups) {
        if (group) {
          group.style.opacity = opacityStyle;
        }
      }
    }
  }
  _setGroupsVisibility(groupMap, key, visible) {
    const groups = groupMap.get(key);
    if (groups) {
      const display = (visible ? '' : 'none');
      for (const group of groups) {
        if (group) {
          group.style.display = display;
        }
      }
    }
  }
  _setGroupsStackLevel(groupMap, key, stackLevel) {
    const groups = groupMap.get(key);
    if (groups) {
      for (const group of groups) {
        if (group && group[NODE_STACK_LEVEL] !== stackLevel) {
          const referenceNode = findNextElement(group.parentNode, stackLevel);
          if (referenceNode !== group.nextElementSibling) {
            group.parentNode.insertBefore(group, referenceNode);
          }
          group[NODE_STACK_LEVEL] = stackLevel;
        }
      }
    }
  }
  _removeGroups(groupMap, key) {
    if (removeNodes(groupMap.get(key))) {
      this._unobserveDatasetProperties(key);
      groupMap.delete(key);
    }
  }
  _removeEmptyGroups(groups) {
    let allRemoved = true;
    if (groups) {
      for (let i = 0; i < groups.length; i++) {
        const node = groups[i];
        if (node) {
          if (node.firstElementChild) {
            allRemoved = false;
          } else {
            node.remove();
            groups[i] = null;
          }
        }
      }
    }
    return allRemoved;
  }
  _removeAllEmptyGroups(datasets = this._datasetGroups.keys()) {
    const productIds = tempSet;
    for (const dataset of datasets) {
      if (this._removeEmptyGroups(this._datasetGroups.get(dataset))) {
        this._unobserveDatasetProperties(dataset);
        this._datasetGroups.delete(dataset);
        productIds.add(dataset?.productId);
      }
    }
    for (const productId of productIds) {
      if (this._removeEmptyGroups(this._productGroups.get(productId))) {
        this._productGroups.delete(productId);
      }
    }
    if (this._productGroups.size === 0) {
      this._unobserveProductsStackLevel();
    }
    productIds.clear();
  }
  _removeFeatureNodes(feature) {
    this._removeWithWrapsById(this._id(feature.id));
    this._removeWithWrapsById(this._featureTextId(feature.id));
  }
  _getDatasetGeometryGroup(dataset) {
    const container = this._getProductGeometryGroup(dataset?.productId);
    return this._getGroup(this._datasetGroups, dataset, container, 0);
  }
  _getDatasetTextGroup(dataset) {
    const container = this._getProductTextGroup(dataset?.productId);
    return this._getGroup(this._datasetGroups, dataset, container, 1);
  }
  _removeDatasetGroups(dataset) {
    this._removeGroups(this._datasetGroups, dataset);
  }
  _getProductGeometryGroup(productId) {
    const container = this._vectorRoot || (
      this._vectorRoot = this._root.appendChild(this._get('g'))
    );
    return this._getGroup(this._productGroups, productId, container, 0);
  }
  _getProductTextGroup(productId) {
    const container = this._textRoot || (
      this._textRoot = this._root.appendChild(
        this._get('g', null, {'paint-order': 'stroke'})
      )
    );
    return this._getGroup(this._productGroups, productId, container, 1);
  }
  _removeProductGroups(productId) {
    this._removeGroups(this._productGroups, productId);
    if (this._productGroups.size === 0) {
      this._unobserveProductsStackLevel();
    }
  }

  _updateWraps(element, times, wrapWidth) {
    if (times > 0) {
      const {id} = element;
      let wrapSet = this._wrapMap.get(element);
      if (!wrapSet) {
        wrapSet = new Set();
        this._wrapMap.set(element, wrapSet);
      }

      const parent = element.parentNode;
      const featureId = this._getNodeFeatureId(element);
      const wrapsIter = wrapSet.values();
      const maxX = times * wrapWidth;

      const elementsBeforeSibling = new Set();
      let sibling = element;
      do {
        elementsBeforeSibling.add(sibling);
        sibling = sibling.nextElementSibling;
      } while (sibling && !(sibling instanceof SVGUseElement));

      for (let x = wrapWidth; x <= maxX; x += wrapWidth) {
        const {value} = wrapsIter.next();
        const wrap = (
          value
          ? this._update(value, {x: -x})
          : this._get('use', null, {'xlink:href': `#${id}`, x: -x})
        );

        if (!elementsBeforeSibling.has(wrap)) {
          parent.insertBefore(wrap, sibling);
        }
        if (!value) {
          this._setNodeFeatureId(wrap, featureId);
          wrapSet.add(wrap);
        }
      }

      for (const wrap of wrapsIter) {
        wrap.remove();
        wrapSet.delete(wrap);
      }
    } else {
      this._removeWraps(element);
    }
  }
  _removeWraps(element) {
    if (removeNodes(this._wrapMap.get(element))) {
      this._wrapMap.delete(element);
    }
  }
  _removeWithWraps(element) {
    this._removeWraps(element);
    element.remove();
  }
  _removeWithWrapsById(id) {
    const element = this._getById(id);
    if (element) {
      this._removeWithWraps(element);
    }
  }

  _xyToSVG(xy) {
    const resolution = this._getResolution();
    xy[0] = (xy[0] - this._featureDx) / resolution;
    xy[1] = (-xy[1] - this._featureDy) / resolution;
    return xy;
  }
  _xysToSVGPath(xy) {
    return this._xyToSVG(xy).join(',');
  }
  _getFeatureBounds(feature, style) {
    const resolution = this._getResolution();
    if (Geometry.isPoint(feature.geometry) && !style.noWrap) {
      const point = feature.geometry;
      const hidden = point.__virtual && !style.__show_virtual;

      if (!hidden) {
        if (point.style) {
          style = Object.assign({}, style, point.style);
        }
        const width = resolution * (
          style.externalGraphic
          ? (style.graphicWidth || style.graphicHeight)
          : (style.pointRadius ?? DEFAULT_STYLE.pointRadius) * 2
        );
        const height = resolution * (
          style.externalGraphic
          ? (style.graphicHeight || style.graphicWidth)
          : (style.pointRadius ?? DEFAULT_STYLE.pointRadius) * 2
        );
        const [x, y] = Geometry.pointToXY(feature.geometry);
        return Bounds.create(x - width, y - height, x + width, y + height);
      }
    }
    return Geometry.getBounds(feature.geometry);
  }
  _drawFeature(feature, style) {
    if (
      !feature.geometry ||
      style.display === 'none' ||
      !this._visible ||
      !this.extent
    ) {
      return false;
    }

    const resolution = this._getResolution();
    const worldBounds = this.mapView.getWorldProjectedBounds();

    const bounds = this._getFeatureBounds(feature, style);
    if (!Bounds.intersects(bounds, this.extent, worldBounds)) {
      return false;
    }

    this._featureDx = this._xOffset;
    this._featureDy = this._yOffset;

    let wrapWidth;
    let wrapCount;
    if (worldBounds) {
      const worldWidth = Bounds.getWidth(worldBounds);
      const viewCenterX = Bounds.getCenterX(
        this.mapView.shouldWrapDateLine() ? this.extent : worldBounds
      );
      const [leftWorldsAway, rightWorldsAway] = idlRangeWorldsAway(
        Bounds.getLeft(bounds),
        Bounds.getRight(bounds),
        viewCenterX,
        worldWidth
      );
      this._featureDx += leftWorldsAway * worldWidth;
      wrapWidth = worldWidth / resolution;
      wrapCount = (
        style.noWrap
        ? 0
        : idlRangeWrapCount(leftWorldsAway, rightWorldsAway)
      );
    } else {
      wrapWidth = 0;
      wrapCount = 0;
    }

    const geometryId = this._id(feature.id);
    const result = this._drawFeatureGeometry(
      this._getDatasetGeometryGroup(feature.dataset),
      geometryId,
      feature.geometry,
      style,
      feature.id
    );
    if (result === true || result === false) {
      this._removeWithWrapsById(geometryId);
    } else {
      for (const node of result) {
        this._updateWraps(node, wrapCount, wrapWidth);
      }
    }

    if (!result) {
      return false;
    }

    const textId = this._featureTextId(feature.id);
    if (style.label) {
      const location = this._xyToSVG(Geometry.getCentroidXY(feature.geometry));
      if (!isNaN(style.labelXOffset)) {
        location[0] += style.labelXOffset;
      }
      if (!isNaN(style.labelYOffset)) {
        location[1] += style.labelYOffset;
      }

      const textNode = this._drawText(
        this._getDatasetTextGroup(feature.dataset),
        textId,
        style,
        feature.id,
        location
      );
      this._updateWraps(textNode, wrapCount, wrapWidth);
    } else {
      this._removeWithWrapsById(textId);
    }
    return true;
  }
  _drawFeatureGeometry(container, id, geometry, style, featureId) {
    if (Geometry.isCollection(geometry)) {
      const hidden = geometry.__virtual && !style.__show_virtual;

      if (hidden) {
        style = HIDDEN_STYLE;
      } else if (geometry.style) {
        style = Object.assign({}, style, geometry.style);
      }

      if (style.display === 'none') {
        return hidden;
      }

      const node = this._get('g', id);
      const children = new Set(node.children);
      let rendered = false;
      for (const {geometry: geom} of Geometry.geometries(geometry)) {
        const result = (
          this._drawGeometryWithHitArea(node, geom.id, geom, style, featureId)
        );
        if (result === true || result === false) {
          this._removeWithWrapsById(id);
        }
        for (const child of result) {
          children.delete(child);
        }
        if (result) {
          rendered = [node];
        }
      }
      removeNodes(children);
      if (rendered && this._nodeStyleChanged(node, style)) {
        container.appendChild(node);
      }
      // node._style = style;
      return rendered;
    }

    return (
      this._drawGeometryWithHitArea(container, id, geometry, style, featureId)
    );
  }
  _drawGeometryWithHitArea(container, id, geometry, style, featureId) {
    const result = (
      this._drawGeometry(container, id, geometry, style, featureId)
    );
    if (result === true || result === false) {
      return result;
    }

    if (
      Geometry.isPolyline(geometry) && style.lineHitWidth > style.strokeWidth
    ) {
      const hitNode = this._drawGeometry(
        container,
        `${id}hit`,
        geometry,
        Object.assign({}, style, {
          strokeWidth: style.lineHitWidth,
          strokeOpacity: 0,
        }),
        featureId
      );
      return [result, hitNode];
    }
    return [result];
  }
  _drawGeometry(container, id, geometry, style, featureId) {
    const hidden = geometry.__virtual && !style.__show_virtual;

    if (hidden) {
      style = HIDDEN_STYLE;
    } else if (geometry.style) {
      style = Object.assign({}, style, geometry.style);
    }

    if (style.display === 'none') {
      return hidden;
    }

    let node;
    if (Geometry.isPoint(geometry)) {
      node = this._drawPoint(id, geometry, style);
    } else if (Geometry.isPolyline(geometry)) {
      node = this._drawPolyline(id, geometry, style);
    } else if (Geometry.isPolygon(geometry)) {
      node = this._drawPolygon(id, geometry, style);
    } else {
      return hidden;
    }

    // node._style = style;
    if (this._nodeStyleChanged(node, style)) {
      this._update(node, geometryAttributes(this._getNodeStyle(node), style));

      this._setNodeFeatureId(node, featureId);
      this._saveNodeStyle(node, style);
      if (node.parentNode !== container) {
        container.appendChild(node);
      }
    }

    return node;
  }
  _drawText(container, id, style, featureId, [x, y]) {
    const node = this._get('text', id);

    const styleChanged = this._nodeStyleChanged(node, style);
    const positionChanged = this._nodePropsChanged(node, {x, y});
    this._saveNodeProps(node, {x, y});

    if (styleChanged) {
      this._update(node, textAttributes(this._getNodeStyle(node), style));

      if (this._nodeStyleChanged(node, style, ['labelSelect'])) {
        if (style.labelSelect) {
          this._setNodeFeatureId(node, featureId);
        } else {
          this._unsetNodeFeatureId(node);
        }
      }

      if (this._nodeStyleChanged(node, style, ['label'])) {
        const labelRows = String(style.label ?? '').split('\n');
        const numRows = labelRows.length;
        while (node.childNodes.length > numRows) {
          node.removeChild(node.lastChild);
        }

        const align = style.labelAlign ?? 'cm';
        const vshift = LABEL_VSHIFT[align[1]] || '-35%';
        const vfactor = LABEL_VFACTOR[align[1]] ?? -.5;
        for (let i = 0; i < numRows; i++) {
          const tspan = (
            node.childNodes[i] ||
            node.appendChild(this._get('tspan', `${id}_tspan_${i}`))
          );
          this._update(tspan, {
            x,
            dy: i === 0 ? `${vfactor * (numRows - 1)}em` : '1em',
            'baseline-shift': isSafari ? vshift : undefined,
          });
          tspan.textContent = (labelRows[i] === '') ? ' ' : labelRows[i];
        }
      }

      this._saveNodeStyle(node, style);
      container.appendChild(node);
    } else if (positionChanged) {
      for (const tspan of node.children) {
        this._update(tspan, {x});
      }
    }
    if (positionChanged) {
      this._update(node, {x, y});
    }
    return node;
  }

  _drawPoint(id, geometry, style) {
    const node = (
      style.externalGraphic
      ? this._get('image', id)
      : this._drawSymbol(id, style)
    );

    const [x, y] = this._xyToSVG(Geometry.pointToXY(geometry));

    const styleChanged = this._nodeStyleChanged(node, style);
    const positionChanged = this._nodePropsChanged(node, {x, y});
    this._saveNodeProps(node, {x, y});

    const updates = {};
    if (styleChanged) {
      const prevStyle = this._getNodeStyle(node);
      Object.assign(
        updates,
        style.externalGraphic && imageAttributes(prevStyle, style),
        pointAttributes(prevStyle, style)
      );
    }
    if (styleChanged || positionChanged) {
      const scaleFactor = style._graphicScale ?? 1;
      const rotation = Math.round((style.graphicRotation ?? 0) * 100) / 100;
      const translate = `translate(${x} ${y})`;
      const scale = scaleFactor === 1 ? '' : `scale(${scaleFactor})`;
      const rotate = (
        rotation === 0 || !isFinite(rotation) ? '' : `rotate(${rotation})`
      );
      const hflip = style.graphicHFlip ? 'scale(-1, 1)' : '';
      const vflip = style.graphicVFlip ? 'scale(1, -1)' : '';

      updates.transform = `${translate} ${scale} ${rotate} ${hflip} ${vflip}`;
      this._update(node, updates);
    }

    return node;
  }
  _drawSymbol(id, style) {
    const symbolName = style.graphicName ?? DEFAULT_STYLE.graphicName;
    const {node: symbolNode, size: symbolSize} = this._importSymbol(symbolName);
    const scale = ((style.pointRadius ?? DEFAULT_STYLE.pointRadius) * 2) / symbolSize;

    style._graphicScale = scale;

    let node = this._getById(id);
    if (node && node[NODE_SYMBOL_NAME] !== symbolName) {
      this._removeWithWraps(node);
      node = null;
    }
    if (!node) {
      const clone = symbolNode.cloneNode(true);
      if (symbolNode instanceof DocumentFragment) {
        node = this._get('g', id);
        node.appendChild(clone);
      } else {
        node = this._update(clone, {id});
      }
      node[NODE_SYMBOL_NAME] = symbolName;
    }

    if (this._nodeStyleChanged(node, style)) {
      const prevStyle = this._getNodeStyle(node);
      const prevGraphicStyle = (prevStyle?.graphic !== false && prevStyle);
      const graphicStyle = (style.graphic !== false && style);
      this._update(node, Object.assign(
        fillAttributes(prevGraphicStyle, graphicStyle),
        strokeAttributes(prevGraphicStyle, graphicStyle)
      ));
    }
    return node;
  }
  _drawPolyline(id, line, style) {
    const node = this._get('polyline', id, fillAttributes(null, {fill: false}));

    const resolution = this._getResolution();
    const featureDx = this._featureDx;
    const featureDy = this._featureDy;
    const points = Geometry.polylineToXYs(line);
    const prevProps = this._getNodeProps(node);
    const redrawPoints = (
      !prevProps ||
      prevProps.resolution !== resolution ||
      prevProps.featureDx !== featureDx ||
      prevProps.featureDy !== featureDy ||
      !areXYsEqual(prevProps.points, points)
    );

    const updates = {};

    if (redrawPoints) {
      updates.points = points.map(this._xysToSVGPath, this).join(' ');
      this._saveNodeProps(node, {resolution, featureDx, featureDy, points});
    }
    if (this._nodeStyleChanged(node, style)) {
      const prevStyle = this._getNodeStyle(node);
      Object.assign(
        updates,
        strokeAttributes(prevStyle, style),
        opacityAttributes(prevStyle, style)
      );
    }

    this._update(node, updates);

    return node;
  }
  _drawPolygon(id, polygon, style) {
    let d = '';
    if (polygon.__bezier) {
      const [xs, ys] = polygon.__bezier;
      const len = xs.length;

      if (len > 0) {
        let lastX = null;
        let lastY = null;

        const coords = (...xy) => this._xysToSVGPath(xy);
        for (let i = 0; i < len; i++) {
          if (xs[i][0] !== lastX || ys[i][0] !== lastY) {
            d += `M${coords(xs[i][0], ys[i][0])} `;
          }

          d += `C${coords(xs[i][1], ys[i][1])} ${coords(xs[i][2], ys[i][2])} ${
            coords(xs[i][3], ys[i][3])
          }`;

          lastX = xs[i][3];
          lastY = ys[i][3];
        }
      }
    } else {
      for (const ring of Geometry.polygonRings(polygon)) {
        const points = Array.from(
          Geometry.polylinePoints(ring),
          this._xysToSVGPath,
          this
        );
        if (ring.__open) {
          points.pop();
        }
        d += `M${points.join(' ')}`;
      }
    }

    const node = this._get('path', id, {'fill-rule': 'evenodd'});

    const prevProps = this._getNodeProps(node);
    const redrawPath = prevProps?.d !== d;
    const updates = {};

    if (redrawPath) {
      updates.d = d;
      this._saveNodeProps(node, {d});
    }

    if (this._nodeStyleChanged(node, style)) {
      const prevStyle = this._getNodeStyle(node);
      Object.assign(
        updates,
        fillAttributes(prevStyle, style),
        strokeAttributes(prevStyle, style),
        opacityAttributes(prevStyle, style)
      );
    }

    this._update(node, updates);

    return node;
  }

  _importSymbol(graphicName) {
    let symbol = this._symbols.get(graphicName);
    if (!symbol) {
      const symbolDefinition = FeaturesRenderer.symbols.get(graphicName);
      if (!symbolDefinition) {
        throw new Error(graphicName + ' is not a valid symbol name');
      }

      const {node, width, height} = (
        Array.isArray(symbolDefinition)
        ? this._createPolygonSymbol(symbolDefinition)
        : symbolDefinition
      );
      symbol = {node, size: Math.max(width, height)};

      this._symbols.set(graphicName, symbol);
    }
    return symbol;
  }
  _createPolygonSymbol(symbol) {
    let minX = Number.MAX_VALUE;
    let minY = Number.MAX_VALUE;
    let maxX = 0;
    let maxY = 0;

    for (var i = 0; i < symbol.length; i += 2) {
      const x = symbol[i];
      const y = symbol[i + 1];
      if (x < minX) {
        minX = x;
      } else if (x > maxX) {
        maxX = x;
      }
      if (y < minY) {
        minY = y;
      } else if (y > maxY) {
        maxY = y;
      }
    }

    return {
      node: this._get('polygon', null, {points: symbol.join(' ')}),
      width: maxX - minX,
      height: maxY - minY,
    };
  }
}
window.ODL = Object.assign(window.ODL || {}, {FeaturesRenderer});
