/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
// @ts-check

/**
 * From Stack Overflow
 * Original question: https://stackoverflow.com/questions/113511/best-implementation-for-hashcode-method/113600#comment25262_113600
 * Question author: Ominpotent - https://stackoverflow.com/users/11193/omnipotent
 * Answer: https://stackoverflow.com/a/113600
 * Answer author: dmeister - https://stackoverflow.com/users/4194/dmeister
 *
 * Algorithm proposed in Josh Bloch's Effective Java in item 8.
 * See https://stackoverflow.com/a/113600/2168921
 *
 * @param {string} string
 * @returns {number}
 */
export function hashString(string) {
  let hash = 0;

  if (string) {
    for (let i = 0, l = string.length; i < l; i++) {
      hash = (hash * 37 + string.charCodeAt(i)) | 0;
    }
  }
  return hash;
}

/**
 * @param {object?} object
 * @returns {number}
 */
export function hashObject(object) {
  return hashString(JSON.stringify(object));
}

/**
 * @template I
 * @template {{id: I}} T
 */
export class HashMap {
  /**
   * @param {(object: T, id: I) => boolean} equals
   */
  constructor(equals) {
    /** @type {Map<number, Set<I>>} */
    this._store = new Map();
    /** @type {(object: T, id: I) => boolean} */
    this.equals = equals;
  }

  /**
   * @param {T} object
   */
  add(object) {
    const hash = hashObject(object);
    let objects = this._store.get(hash);

    if (!objects) {
      objects = new Set();
      this._store.set(hash, objects);
    }
    objects.add(object.id);
  }

  /**
   * @param {T} object
   * @returns {boolean}
   */
  has(object) {
    const hash = hashObject(object);
    const objects = this._store.get(hash);

    if (objects) {
      for (const id of objects) {
        if (this.equals(object, id)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * @param {T} object
   */
  delete(object) {
    const hash = hashObject(object);
    const objects = this._store.get(hash);

    if (objects) {
      objects.delete(object.id);
      if (objects.size === 0) {
        this._store.delete(hash);
      }
    }
  }

  clear() {
    this._store.clear();
  }
}
Object.assign(HashMap, {hashString, hashObject});
// @ts-expect-error
window.ODL = Object.assign(window.ODL || {}, {HashMap});
