/**
 * @author <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

/**
 * Inspired by [Kent C. Dodds's match-sorter](https://github.com/kentcdodds/match-sorter)
 *
 * Just in case, here is the MIT license for match-sorter:
 * The MIT License (MIT)
 * Copyright (c) 2020 Kent C. Dodds
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import {asArray, RegExpHelpers} from '../syntool.js';

export const rankings = {
  CASE_SENSITIVE_EQUAL: 5,
  EQUAL: 4,
  STARTS_WITH: 3,
  WORD_STARTS_WITH: 2,
  CONTAINS: 1,
  NO_MATCH: 0,
};

export function matchSorter(items, query, {keys, threshold = rankings.CONTAINS} = {}) {
  const queryParts = query.split(/\s+/).filter(Boolean);

  if (queryParts.length === 0) {
    return items;
  }

  const rankedItems = items.map(toRankingItem);

  for (const queryPart of queryParts) {
    for (const ranked of rankedItems) {
      if (ranked.rank !== rankings.NO_MATCH) {
        const rank = getHighestRanking(ranked.item, keys, queryPart, threshold);
        ranked.rank = (
          rank < threshold
          ? rankings.NO_MATCH
          : (ranked.rank ?? 0) + rank
        );
      }
    }
  }

  return (
    rankedItems
    .filter(({rank}) => rank >= threshold)
    .sort(compareRankedItems)
    .map(fromRankingItem)
  );
}

function toRankingItem(item) {
  return {item, rank: undefined};
}
function fromRankingItem({item}) {
  return item;
}
function compareRankedItems({rank: aRank}, {rank: bRank}) {
  return bRank - aRank;
}

/**
 * Gets the highest ranking for query for the given item based on its values
 * for the given keys
 *
 * @param {*} item - the item to rank
 * @param {Array} keys - the keys to get values from the item for the ranking
 * @param {String} query - the value to rank against
 * @return {Number} - the highest ranking
 */
function getHighestRanking(item, keys, query) {
  if (!keys) {
    // if keys is not specified, then we assume the item given is ready to be
    // matched
    return getMatchRanking(item, query);
  }

  let rank = rankings.NO_MATCH;
  for (const key of keys) {
    const value = item[key];
    if (value) {
      for (const itemValue of asArray(value)) {
        const newRank = getMatchRanking(itemValue, query);
        if (newRank > rank) {
          rank = newRank;
        }
      }
    }
  }
  return rank;
}

/**
 * Gives a rankings score based on how well the two strings match.
 *
 * @param {String} testString - the string to test against
 * @param {String} stringToRank - the string to rank
 * @returns {Number} the ranking for how well stringToRank matches testString
 */
function getMatchRanking(testString, stringToRank) {
  testString = String(testString);
  stringToRank = String(stringToRank);

  // too long
  if (stringToRank.length > testString.length) {
    return rankings.NO_MATCH;
  }

  // case sensitive equals
  if (testString === stringToRank) {
    return rankings.CASE_SENSITIVE_EQUAL;
  }

  // Lower casing before further comparison
  testString = testString.toLowerCase();
  stringToRank = stringToRank.toLowerCase();

  // case insensitive equals
  if (testString === stringToRank) {
    return rankings.EQUAL;
  }

  // starts with
  if (testString.startsWith(stringToRank)) {
    return rankings.STARTS_WITH;
  }

  // word starts with
  if (testString.match(new RegExp(`[^a-z0-9]${RegExpHelpers.escape(stringToRank)}`))) {
    return rankings.WORD_STARTS_WITH;
  }

  // contains
  if (testString.includes(stringToRank)) {
    return rankings.CONTAINS;
  }

  return rankings.NO_MATCH;
}
