/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {parseDecimal, clamp} from '../syntool.js';
import {OpenLayers} from '../vendors/openlayers.js';

export class OLLayer extends OpenLayers.Layer {
  constructor(name, options) {
    super(name, options);

    this.alwaysInRange = options?.alwaysInRange ?? true;
    this.renderer = options?.renderer ?? null;
  }

  clone() {
    throw new Error('Not Implemented');
  }
  destroy() {
    this.renderer?.destroy();
    this.renderer = null;
    super.destroy();
  }

  setMap(map) {
    super.setMap(map);

    if (this.renderer) {
      this.renderer.mapView = this.map.mapView;
      this._setRendererSize();
    }
  }
  afterAdd() {
    this.renderer?.setVisibility(true);
  }
  removeMap(map) {
    this.renderer?.setVisibility(false);
  }
  onMapResize() {
    this._setRendererSize();
  }

  moveTo(bounds, zoomChanged, dragging) {
    super.moveTo(bounds, zoomChanged, dragging);

    if (!dragging) {
      if (!this.isFixed) {
        const offsetLeft = -parseDecimal(this.map.layerContainerDiv.style.left);
        const offsetTop = -parseDecimal(this.map.layerContainerDiv.style.top);
        this.div.style.left = offsetLeft + 'px';
        this.div.style.top = offsetTop + 'px';
      }

      this.renderer?.invalidate();
      this.renderer?.resetOffset();
      this._setRendererSize();
      this.renderer?.setExtent(this.map.getExtent());
    }
  }
  moveByPx(dx, dy) {
    if (this.isFixed) {
      this.renderer?.setOffset(dx, dy);
    }
  }
  display(display) {
    super.display(display);
    this.renderer?.setVisibility(this.div.style.display !== 'none');
  }

  setRenderer(renderer) {
    const opacityValue = clamp(parseFloat(this.opacity), 0, 1);
    const opacity = (opacityValue === 1 ? '' : opacityValue);

    for (const childNode of this.div.childNodes) {
      (childNode.firstChild || childNode).style.opacity = opacity;
    }

    this.renderer = renderer;
    if (this.map) {
      this.renderer.invalidate();
      this.renderer.resetOffset();
      this._setRendererSize();
      this.renderer.setExtent(this.map.getExtent());
      this.renderer.setVisibility(this.getVisibility());
    }
  }

  _setRendererSize() {
    const {w: width, h: height} = this.map.getSize();
    this.renderer?.setSize(width, height);
  }
}
window.ODL = Object.assign(window.ODL || {}, {OLLayer});
