/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {shallowEqual} from '../syntool.js';
import {OLLayer} from './ol_layer.js';
import {FeaturesRenderer} from './features_renderer.js';

const MapMoveRedrawMode = {
  NONE: 'NONE',
  RENDERED: 'RENDERED',
  ALL: 'ALL',
};

export class OLVectorLayer extends OLLayer {
  static MapMoveRedrawMode = MapMoveRedrawMode;

  constructor(name, options) {
    super(name, options);

    this.renderer = options?.renderer ?? new FeaturesRenderer(this.div);
    this.style = options?.style ?? null;
    this.resolutionDependentStyle = !!options?.resolutionDependentStyle;
    this.mapMoveRedrawMode = (
      options?.mapMoveRedrawMode ?? MapMoveRedrawMode.ALL
    );

    this.div.dataset.name = name;
    this.featureById = new Map();
    this.featuresByDataset = new Map();
    this.featuresByProductId = new Map();
  }

  get appliesDatasetProperties() {
    return this.renderer.appliesDatasetProperties;
  }
  set appliesDatasetProperties(appliesDatasetProperties) {
    this.renderer.appliesDatasetProperties = appliesDatasetProperties;
  }

  destroy() {
    this.destroyFeatures();
    this.featureById = null;
    this.featuresByDataset = null;
    this.featuresByProductId = null;
    super.destroy();
  }

  moveTo(bounds, zoomChanged, dragging) {
    super.moveTo(bounds, zoomChanged, dragging);
    if (!dragging && this.mapMoveRedrawMode !== MapMoveRedrawMode.NONE) {
      for (const feature of this.featureById.values()) {
        const shouldRedrawFeature = (
          this.mapMoveRedrawMode === MapMoveRedrawMode.ALL ||
          this.renderer.isFeatureRendered(feature)
        );
        if (shouldRedrawFeature) {
          const {layerInfo} = feature;
          const prevStyle = (
            layerInfo?.layer === this && !(this.resolutionDependentStyle && zoomChanged)
            ? layerInfo.style
            : null
          );
          const style = prevStyle || this._getFeatureStyle(feature);
          this.renderer.drawFeature(feature, style);
        }
      }
    }
  }

  addFeature(feature) {
    if (!feature || feature.layer === this) {
      return;
    }

    //give feature reference to its layer
    feature.layer = this;

    this.featureById.set(feature.id, feature);

    const dataset = feature.dataset;
    let datasetFeatures = this.featuresByDataset.get(dataset);
    if (!datasetFeatures) {
      datasetFeatures = new Set();
      this.featuresByDataset.set(dataset, datasetFeatures);
    }
    datasetFeatures.add(feature);

    const productId = dataset?.productId;
    let productFeatures = this.featuresByProductId.get(productId);
    if (!productFeatures) {
      productFeatures = new Set();
      this.featuresByProductId.set(productId, productFeatures);
    }
    productFeatures.add(feature);

    this.drawFeature(feature, true /*force*/);
  }
  addFeatures(features) {
    for (const feature of features) {
      this.addFeature(feature);
    }
  }

  getFeatureWithId(featureId) {
    return this.featureById.get(featureId);
  }
  isFeatureVisible(feature) {
    return feature?.layer === this && this.renderer.isFeatureVisible(feature);
  }

  removeFeature(feature) {
    if (feature?.layer === this) {
      // feature has no layer at this point
      feature.layer = null;

      this.renderer.eraseFeature(feature);
      feature.rendererInfo = null;

      this.featureById.delete(feature.id);

      const dataset = feature.dataset;
      const productId = dataset?.productId;
      const datasetFeatures = this.featuresByDataset.get(dataset);
      if (datasetFeatures?.delete(feature) && datasetFeatures.size === 0) {
        this.featuresByDataset.delete(dataset);
      }
      const productFeatures = this.featuresByProductId.get(productId);
      if (productFeatures?.delete(feature) && productFeatures.size === 0) {
        this.featuresByProductId.delete(productId);
      }
    }
  }
  removeFeatures(features) {
    for (const feature of features) {
      this.removeFeature(feature);
    }
  }
  removeAllFeaturesOfDataset(dataset) {
    const datasetFeatures = this.featuresByDataset.get(dataset);
    if (!datasetFeatures) {
      return;
    }

    this.renderer.eraseAllFeaturesOfDataset(dataset);

    const productId = dataset?.productId;
    const productFeatures = this.featuresByProductId.get(productId);

    for (const feature of datasetFeatures) {
      // feature has no layer at this point
      feature.layer = null;
      feature.rendererInfo = null;

      this.featureById.delete(feature.id);
      if (productFeatures.delete(feature) && productFeatures.size === 0) {
        this.featuresByProductId.delete(productId);
      }
    }
    this.featuresByDataset.delete(dataset);
  }
  removeAllFeaturesOfProductWithId(productId) {
    const productFeatures = this.featuresByProductId.get(productId);
    if (!productFeatures) {
      return;
    }

    this.renderer.eraseAllFeaturesOfProductWithId(productId);

    for (const feature of productFeatures) {
      // feature has no layer at this point
      feature.layer = null;
      feature.rendererInfo = null;

      this.featureById.delete(feature.id);
      this.featuresByDataset.delete(feature.dataset);
    }
    this.featuresByProductId.delete(productId);
  }
  removeAllFeatures() {
    for (const feature of this.featureById.values()) {
      // feature has no layer at this point
      feature.layer = null;
      feature.rendererInfo = null;
    }
    this.renderer.clear();
    this.featureById.clear();
    this.featuresByDataset.clear();
    this.featuresByProductId.clear();
  }

  destroyFeatures(features) {
    if (features) {
      this.removeFeatures(features);
    } else {
      features = this.featureById.values();
      this.removeAllFeatures();
    }
    for (const feature of features) {
      feature.destroy();
    }
  }

  drawFeature(feature, force) {
    if (feature?.layer === this) {
      const {layerInfo} = feature;
      const prevStyle = (
        layerInfo?.layer === this && this.renderer.isFeatureRendered(feature)
        ? layerInfo.style
        : null
      );
      const style = this._getFeatureStyle(feature);
      if (force || prevStyle !== style) {
        this.renderer.drawFeature(feature, style);
      }
    }
  }
  drawFeatures(features, force) {
    for (const feature of features) {
      this.drawFeature(feature, force);
    }
  }
  redrawAllFeatures(force) {
    for (const feature of this.featureById.values()) {
      this.drawFeature(feature, force);
    }
  }
  eraseFeature(feature) {
    if (feature?.layer === this) {
      this.renderer.eraseFeature(feature);
    }
  }
  eraseFeatures(features) {
    this.renderer.eraseFeatures(features);
  }

  getFeatureFromEvent(evt) {
    if (!this.renderer) {
      throw new Error(
        'getFeatureFromEvent called on layer with no renderer. This ' +
        'usually means you destroyed a layer, but not some handler which ' +
        'is associated with it.'
      );
    }
    const featureId = this.renderer.getFeatureIdFromEvent(evt);
    return featureId ? this.featureById.get(featureId) : null;
  }

  get features() {
    return this.featureById.values();
  }
  getFeaturesCount() {
    return this.featureById.size;
  }
  hasFeatures(predicate, thisArg) {
    if (predicate) {
      for (const feature of this.featureById.values()) {
        if (predicate.call(thisArg, feature)) {
          return true;
        }
      }
      return false;
    }
    return this.featureById.size > 0;
  }
  getFeatures(predicate, thisArg) {
    var features = Array.from(this.featureById.values());
    if (predicate) {
      return features.filter(predicate, thisArg);
    }
    return features;
  }
  forEachFeature(callback, thisArg) {
    this.featureById.forEach(callback, thisArg);
  }

  _getStyle(feature, intent) {
    const {style} = this;

    if (typeof style === 'function') {
      return style(feature, intent);
    }

    const styleForIntent = style?.[intent];
    if (typeof styleForIntent === 'function') {
      return styleForIntent(feature, intent);
    }

    return styleForIntent;
  }
  _getFeatureStyle(feature) {
    const baseStyle = this._getStyle(feature, 'base');
    const renderIntent = feature.renderIntent;
    const intentStyle = (
      renderIntent && renderIntent !== 'base'
      ? this._getStyle(feature, renderIntent)
      : null
    );
    const featureStyle = feature.style;

    const prevProps = feature.layerInfo;
    if (
      prevProps &&
      prevProps.layer === this &&
      prevProps.renderIntent === renderIntent &&
      shallowEqual(prevProps.baseStyle, baseStyle) &&
      shallowEqual(prevProps.intentStyle, intentStyle) &&
      shallowEqual(prevProps.featureStyle, featureStyle)
    ) {
      return prevProps.style;
    }

    const style = Object.assign(
      {},
      baseStyle,
      intentStyle,
      featureStyle
    );

    feature.layerInfo = {
      layer: this,
      baseStyle,
      renderIntent,
      intentStyle,
      featureStyle: featureStyle ? Object.assign({}, featureStyle) : null,
      style,
    };

    return style;
  }
}
window.ODL = Object.assign(window.ODL || {}, {OLVectorLayer});
