/**
 * @author: <sylvain.gerard@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
import {
  emptyFunction,
  formatUTCDate,
  ajax,
  isAbortError,
  fetchResource,
  ajaxRetriable,
  Syntool,
} from '../syntool.js';
import {d3} from '../vendors/d3.js';
import {colorToCSSColor} from './color.js';
import {
  getColorPalette,
  fetchColorPalette,
  colorPaletteGetColorBetween,
} from './color_palette.js';
import {linkStyleSheet} from './dom_utils.js';

linkStyleSheet('css/spectrum_plot.css');

var SPECTRUM_COLOR_PALETTE_URL = 'images/palettes/spectrum_palette.png';

function createElement(className) {
  var div = document.createElement('div');
  div.className = className;
  return div;
}

export class SpectrumBuilder {
  constructor() {
    this._abortLoadJson = null;

    this.chartSize = 400;
    this.fontSize = 14;
    this.element = createElement('spectrum-plot');
    this.unitsElement = this.element.appendChild(createElement('units'));
    this.detailsElement = this.element.appendChild(createElement('details'));
    this.container = null;
    this.palette = null;
    this.svgContainer = null;

    this.rows = 1;
    this.columns = 1;

    this.width = this.chartSize * this.columns;
    this.height = this.chartSize * this.rows;

    this.scaleX = null;
    this.scaleY = null;
    this.zoom = null;

    this.stepAngle = null;
    this.stepFrequency = null;
    this.nbSegments = 8;

    this.minValue = 0.0;
    this.maxValue = 0.25;

    this.markers = [6.25, 12.5, 25, 50, 100, 200, 400, 800, 1600];
    this.markerType = 'wavelength';
    this.markerNodes = [];

    //this.markers = [2, 4, 8, 16, 32];
    //this.markerType = "second";

    this.data = null;

    this.setPalette = this.setPalette.bind(this);
    this.getColor = this.getColor.bind(this);
    this.checkZoomLevel = this.checkZoomLevel.bind(this);
    this.handleSpectrumData = this.handleSpectrumData.bind(this);
  }

  /**
   *
   */
  checkZoomLevel() {
    var scale;           //
    var center;          //
    var ratio;           //
    var isVisible;       //
    var screenRadius;    //
    var markerNode;      //
    var i;               //
    var l;               //

    //this.container.attr("transform", "translate(" + d3.event.translate + ") scale(" + d3.event.scale + ")");
    scale = this.zoom.scale();
    center =  this.chartSize / 2 * (1 - scale);
    this.container.attr('transform', 'translate(' + [center, center] + ') scale(' + scale + ')');
    //console.log("zoom:", this.zoom.scale());
    screenRadius = this.chartSize / 2 / scale;
    for (i = 0, l = this.markerNodes.length; i < l; i++) {
      markerNode = this.markerNodes[i];
      ratio = markerNode.radius / screenRadius;
      isVisible = (0.2 <= ratio && ratio <= 1.0);
      d3.select(markerNode.circle).attr('visibility', isVisible ? 'visible' : 'hidden')
        .attr('stroke-width', 1 / scale);
      d3.select(markerNode.text)
        .attr('visibility', isVisible ? 'visible' : 'hidden')
        .attr('font-size', this.fontSize / scale + 'px');
    }
    d3.selectAll('.line')
      .attr('stroke-width', 1 / scale);
  }

  /**
   *
   */
  initCanvas() {
    var svgContainer;

    this.scaleX = d3.scale.linear()
      .domain([-this.width / 2, this.width / 2])
      .range([0, this.width]);

    this.scaleY = d3.scale.linear()
      .domain([-this.height / 2, this.height / 2])
      .range([this.height, 0]);

    this.zoom = d3.behavior.zoom()
      .scaleExtent([1, 10])
      .x(this.scaleX)
      .y(this.scaleY)
      .center([0, 0])
      .scale(1)
      .on('zoom', this.checkZoomLevel);

    svgContainer = d3.select(this.element).append('svg')
      .attr('class', 'plot')
      .attr('viewBox', [0, 0, this.width, this.height].join(' '))
      .call(this.zoom)
      .on('mousedown.zoom', null);
    this.svgContainer = svgContainer;
  }

  /**
   *
   */
  initPalette() {
    this.palette = getColorPalette(SPECTRUM_COLOR_PALETTE_URL);
    if (!this.palette) {
      fetchColorPalette(SPECTRUM_COLOR_PALETTE_URL)
        .then(this.setPalette, emptyFunction);
    }
  }

  /**
   *
   */
  clear() {
    this.unitsElement.style.display = 'none';
    this.detailsElement.style.display = 'none';
    if (this.svgContainer) {
      this.svgContainer.remove();
    }
  }

  /**
   *
   */
  loadJson(spectrumUrl, datasetId) {
    if (this._abortLoadJson) {
      this._abortLoadJson();
    }

    var [promise, abort] = fetchResource(ajaxRetriable(ajax.getJSON), spectrumUrl);

    this._abortLoadJson = abort;
    promise
    .then(this.handleSpectrumData, function(error) {
      if (!isAbortError(error)) {
        console.error(
          'Dataset(' + datasetId + ') could not load spectrum data from ' +
          'URL "' + spectrumUrl + '"!',
          error
        );
        Syntool.notificationCenter.showError(
          'Could not load spectrum for dataset "' + datasetId + '" due to server or network errors!'
        );
      }
    });
  }

  handleSpectrumData(data) {
    this.data = data;
    this.clear();
    this.initCanvas();
    this.buildSpectrum();
  }

  /**
   *
   */
  setPalette(palette) {
    this.palette = palette;
    if (this.data) {
      this.handleSpectrumData(this.data);
    }
  }

  /**
   *
   */
  getColor(d) {
    //var value = Math.round((1 - d / 500) * 256);
    return colorToCSSColor(colorPaletteGetColorBetween(
      this.palette,
      this.minValue,
      this.maxValue,
      d
    ));
  }

  /**
   *
   */
  getPoint(angleIndex, frequency, offset) {
    var angle;   //
    var radius;  //
    var x;       //
    var y;       //

    angle = this.stepAngle * angleIndex;
    radius = this.stepFrequency * frequency;
    x = Math.cos(angle) * radius + offset.x;
    y = Math.sin(angle) * radius + offset.y;
    return [this.scaleX(x), this.scaleY(y)];
  }

  /**
   *
   */
  addCircle(value, radius, offset, color, dash, where) {
    var circle;  //
    var text;    //
    var textPos; //

    color = color || 'black';
    circle = this.container.append('circle')
      .attr('cx', this.scaleX(offset.x))
      .attr('cy', this.scaleY(offset.y))
      .attr('r', radius)
      .style('stroke-opacity', 0.8)
      .style('stroke', color)
      .style('fill', 'none');

    if (dash) {
      circle.style('stroke-dasharray', ('1, 1'));
    }

    textPos = ('above' === where) ? 1 : -1;

    text = this.container.append('text')
      .attr('x', this.scaleX(offset.x))
      .attr('y', this.scaleY(offset.y + textPos * (radius + 1)))
      .text(value)
      .attr('text-anchor', 'middle')
      .attr('alignment-baseline', 'hanging')
      .attr('font-family', 'sans-serif')
      .attr('font-size', this.fontSize + 'px')
      .attr('fill', color);

    this.markerNodes.push({
      radius: radius,
      circle: circle[0][0],
      text  : text[0][0]
    });
  }

  /**
   *
   */
  showPeak(offset) {
    var peakPeriod = 1 / this.data.principal_freq;
    var peakWavelength = 1.56 * peakPeriod * peakPeriod;
    var label = peakWavelength.toFixed(3) + 'm';
    this.addLine(
      this.data.principal_freq,  // start from peak frequency
      null,  // go to the edge of the plot
      this.data.mean_dir,
      offset,
      1.0,  // full opacity
      'blue',
      false /* no dashes */
    );
    this.addCircle(
      label,
      this.stepFrequency * this.data.principal_freq, //radius (px)
      offset,
      'blue',
      false, // no dashes
      'above' /* text above the circle*/
    );
  }

  /**
   *
   */
  addLine(freqMin, freqMax, direction, offset, opacity, color, dash) {
    var horizontalSize;  //
    var verticalSize;    //
    var x1; //
    var y1; //
    var x2; //
    var y2; //

    color = color || 'black';
    if (Number.isFinite(direction)) {
      // Clockwise from North -> counter-clockwise from East (trigo)
      var dirRadians = 0.5 * Math.PI - Math.PI * direction / 180.0;
      var fMin = freqMin || 0;
      var fMax = freqMax || this.data.metadata.a.max;
      var radiusMin = fMin * this.stepFrequency;
      var radiusMax = fMax * this.stepFrequency;
      x1 = radiusMin * Math.cos(dirRadians);
      y1 = radiusMin * Math.sin(dirRadians);
      x2 = radiusMax * Math.cos(dirRadians);
      y2 = radiusMax * Math.sin(dirRadians);
    } else {
      horizontalSize = this.chartSize / 2 * (direction === 'horizontal' ? 1 : 0);
      verticalSize = this.chartSize / 2 * (direction === 'vertical' ? 1 : 0);
      x1 = -1 * horizontalSize;
      y1 = -1 * verticalSize;
      x2 = horizontalSize;
      y2 = verticalSize;
    }

    var line = this.container.append('line')
      .attr('x1', this.scaleX(offset.x + x1))
      .attr('y1', this.scaleY(offset.y + y1))
      .attr('x2', this.scaleX(offset.x + x2))
      .attr('y2', this.scaleY(offset.y + y2))
      .attr('class', 'line')
      .style('stroke-opacity', opacity)
      .style('stroke', color);

    if (dash) {
      line.style('stroke-dasharray', ('2, 1'));
    }
  }

  /**
   *
   */
  buildSpectrum() {
    var nbAngles;        //
    var nbFrequencies;   //
    var datetime;        //
    var data;            //
    var polygons;        //
    var i;               //
    var j;               //
    var k;               //
    var l;               //
    var offset;          //
    var value;           //
    var radius;          //
    var markerValue;     //

    this.zoom.scale(1);
    this.container = this.svgContainer.append('g');

    nbAngles = this.data.y; // 24 // 36
    nbFrequencies = this.data.abCount; // 64 // 46 // 30
    this.stepAngle = Math.PI  * 2 / nbAngles;
    this.stepFrequency = (this.chartSize * 0.9) / 2 / this.data.metadata.a.max;

    this.minValue = this.data.metadata.x.min;
    this.maxValue = this.data.metadata.x.max;

    datetime = new Date(Math.round(this.data.t) * 1000);

    this.unitsElement.style.display = '';
    this.unitsElement.textContent = 'hs: ' + this.data.hs.toFixed(3) + ' m';
    this.detailsElement.style.display = '';
    this.detailsElement.textContent = formatUTCDate(datetime);

    data = [];
    for (i = 0; i < nbAngles; i++) {
      for (j = 0; j < nbFrequencies; j++) {
        data.push(this.data.x[i * nbFrequencies + j]);
      }
    }

    /*
    // Debug: compute energy and mean direction for each frequency

    var a1s = Array.from({length: nbFrequencies}, (v, i) => 0);
    var b1s = Array.from({length: nbFrequencies}, (v, i) => 0);
    var etot = Array.from({length: nbFrequencies}, (v, i) => 0);
    for (j=0; j <nbFrequencies; j++)
    {
      for (i=0; i<nbAngles; i++)
      {
        var radAngle = i * this.stepAngle;
        var energy = this.data.x[i * nbFrequencies + j];
        a1s[j] = a1s[j] + this.stepAngle * Math.cos(radAngle) * energy;
        b1s[j] = b1s[j] + this.stepAngle * Math.sin(radAngle) * energy;
        etot[j] = etot[j] + this.stepAngle * this.data.b[j] * energy;
      }

      if (0.0 >= etot[j]) {
        continue;
      }
      console.log(this.data.a[j], this.data.b[j], this.stepAngle);
      console.log("\tA1, B1", a1s[j], b1s[j]);
      console.log("\tetot, direction", etot[j], 270 - 180 * Math.atan2(b1s[j], a1s[j]) / Math.PI);
    }
    */
    polygons = this.container.selectAll('polygon').data(data);

    for (i = 0; i < this.rows; i++) {
      for (j = 0; j < this.columns; j++) {
        offset = {
          x: (i - this.rows / 2 + 1 / 2) * this.chartSize,
          y: (j - this.columns / 2 + 1 / 2) * this.chartSize,
        };

        polygons.enter().append('polygon')
          .attr('points', this.getPolygonPoints.bind(this, nbFrequencies, offset))
          .attr('fill', this.getColor)
          .attr('stroke', this.getColor)
          .attr('stroke-opacity', 1)
          .attr('stroke-width', 0);

        this.addLine(null, null, 'horizontal', offset, 0.8);
        this.addLine(null, null, 'vertical', offset, 0.8);

        // Add ~transparent ticks every 30 degrees
        this.addLine(null, null, 30.0, offset, 0.1);
        this.addLine(null, null, 60.0, offset, 0.1);
        this.addLine(null, null, 120.0, offset, 0.1);
        this.addLine(null, null, 150.0, offset, 0.1);
        this.addLine(null, null, 210.0, offset, 0.1);
        this.addLine(null, null, 240.0, offset, 0.1);
        this.addLine(null, null, 300.0, offset, 0.1);
        this.addLine(null, null, 330.0, offset, 0.1);

        for (k = 0, l = this.markers.length; k < l; k++) {
          markerValue = this.markers[k];
          value = markerValue;
          radius = markerValue;
          if (this.markerType === 'wavelength') {
            value += 'm';
            // wavelength ~= g/2pi * (1 / freq) ^ 2
            radius = this.stepFrequency / Math.sqrt(markerValue / 1.56);
          } else if (this.markerType === 'second') {
            value += 's';
            radius = 1 / markerValue * this.stepFrequency;
          }
          this.addCircle(value, radius, offset, 'black', true);
        }

        // Display information about the peak value
        this.showPeak(offset);
      }
    }

    this.zoom.scale(4);
    this.checkZoomLevel();
  }

  /**
   *
   */
  getPolygonPoints(nbFrequencies, offset, d, index) {
    var angleIndex = Math.floor(index / nbFrequencies);
    var frequencyIndex = index % nbFrequencies;
    var polygon = [];
    var frequencyStart = this.data.a[frequencyIndex] - this.data.b[frequencyIndex] / 2;
    var frequencyEnd = this.data.a[frequencyIndex] + this.data.b[frequencyIndex] / 2;
    var segmentId;
    var angleStart = angleIndex - 0.5;
    var angleEnd = angleIndex + 0.5;

    // first point
    polygon.push(this.getPoint(angleStart, frequencyStart, offset));
    // second point
    polygon.push(this.getPoint(angleStart, frequencyEnd, offset));
    // arc between first and second point
    for (segmentId = 0; segmentId < this.nbSegments; segmentId++) {
      polygon.push(this.getPoint(angleStart + segmentId / this.nbSegments, frequencyEnd, offset));
    }
    // third point
    polygon.push(this.getPoint(angleEnd, frequencyEnd, offset));
    // fourth point
    polygon.push(this.getPoint(angleEnd, frequencyStart, offset));
    // arc between fourth and first point
    for (segmentId = 0; segmentId < this.nbSegments; segmentId++) {
      polygon.push(this.getPoint(angleEnd - segmentId / this.nbSegments, frequencyStart, offset));
    }
    return polygon;
  }

  /**
   *
   */
  init() {
    this.initCanvas();
    this.initPalette();
  }
}
window.ODL = Object.assign(window.ODL || {}, {SpectrumBuilder});
