/**
 * @author: <ziad.khoury.hanna@oceandatalab.com>
 *
 * Copyright (C) 2014-2024 OceanDataLab
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

/** An event which takes place in the DOM. */
type DOMEvent = Event;

/**
 * Based on https://github.com/DefinitelyTyped/DefinitelyTyped/blob/master/types/openlayers/v2/index.d.ts
 */
export namespace OpenLayers {
  /**
   * How many times should we try to reload an image before giving up?
   * Default is 0
   */
  export let IMAGE_RELOAD_ATTEMPTS: number;

  export interface LonLatLike {
    lon: number;
    lat: number;
  }
  export interface PointLike {
    x: number;
    y: number;
  }
  export interface PixelLike {
    x: number;
    y: number;
  }
  export interface SizeLike {
    w: number;
    h: number;
  }
  export interface BoundsLike {
    left: number;
    bottom: number;
    right: number;
    top: number;
  }

  export class LonLat {
    lon: number;
    lat: number;

    constructor(lon: number, lat: number);
    constructor(lonlat: [lat: number, lon: number]);

    /** Shortened String representation of OpenLayers.LonLat object. */
    toShortString(): string;

    /** New OpenLayers.LonLat object with the same lon and lat values */
    clone(): LonLat;

    /**
     * A new OpenLayers.LonLat object with the lon and lat passed-in added to
     * this’s.
     */
    add(lon: number, lat: number): LonLat;

    /**
     * Boolean value indicating whether the passed-in OpenLayers.LonLat object
     * has the same lon and lat components as this.
     * Note: if ll passed in is null, returns false.
     */
    equals(ll: LonLat | LonLatLike): boolean;

    /**
     * Transform the LonLat object from source to dest.
     * This transformation is in place: if you want a new lonlat, use .clone() first.
     */
    transform(source: Projection, dest: Projection): LonLat;

    /**
     * @returns A copy of this lonlat, but wrapped around the "dateline" (as
     * specified by the borders of maxExtent).
     */
    wrapDateLine(maxExtent: Bounds): LonLat;
  }

  export class Pixel {
    x: number;
    y: number;

    constructor(x: number, y: number);
    clone(): Pixel;
    equals(px: PointLike | null): boolean;
    distanceTo(px: Pixel): number;

    /** Returns a new Pixel with this pixel's x & y augmented by the values passed in. */
    add(x: number, y: number): Pixel;

    /**
     * Returns a new Pixel with this pixel's x & y augmented by the x & y
     * values of the pixel passed in.
     */
    offset(px: { x: number, y: number }): Pixel;
  }

  export interface BoundsOptions {
    /**
     * Whether or not to include the border. Default is true.
     */
    inclusive?: boolean | undefined;

    /**
     * If a worldBounds is provided, the two bounds will be wrapped around the
     * dateline.
     */
    worldBounds?: Bounds | false | null | undefined;
  }
  export interface WrapDateLineOptions {
    /**
     * Allow for a margin of error with the 'left' value of this bound.
     * Default is 0.
     */
    leftTolerance?: number | undefined;

    /**
     * Allow for a margin of error with the 'right' value of this bound.
     * Default is 0.
     */
    rightTolerance?: number | undefined;
  }

  export class Bounds {
    /** Minimum horizontal coordinate. */
    left: number;

    /** Minimum vertical coordinate. */
    bottom: number;

    /** Maximum horizontal coordinate. */
    right: number;

    /** Maximum vertical coordinate. */
    top: number;

    /**
     * A cached center location.
     * This should not be accessed directly.
     * Use getCenterLonLa() instead.
     */
    centerLonLat: LonLat | null;

    constructor(left?: number, bottom?: number, right?: number, top?: number);
    constructor(bounds: [left: number, bottom: number, right: number, top: number]);

    /** Create a cloned instance of this bounds. */
    clone(): Bounds;

    /** Test a two bounds for equivalence. */
    equals(bounds: Bounds | BoundsLike): boolean;

    /** @returns A string representation of the bounds object. */
    toString(): string;

    /** @returns An array representation of the bounds object. */
    toArray(reverseAxisOrder?: boolean): number[];

    /** @returns A boundingbox-string representation of the bounds object. */
    toBBOX(decimal?: number, reverseAxisOrder?: boolean): string;

    /** Create a new polygon geometry based on this bounds. */
    toGeometry(): Geometry.Polygon;

    /** @returns The width of the bounds. */
    getWidth(): number;

    /** @returns The height of the bounds. */
    getHeight(): number;

    getSize(): Size;

    /** @returns The Pixel object which represents the center of the bounds. */
    getCenterPixel(): Pixel;

    /** @returns The LonLat object which represents the center of the bounds. */
    getCenterLonLat(): LonLat;

    /**
     * Scales the bounds around a pixel or lonlat. Note that the new
     * bounds may return non-integer properties, even if a pixel
     * is passed.
     */
    scale(ratio: number, origin?: Pixel): void;

    /**
     * Scales the bounds around a pixel or lonlat. Note that the new
     * bounds may return non-integer properties, even if a pixel
     * is passed.
     */
    scale(ratio: number, origin?: LonLat): void;

    /**
     * Shifts the coordinates of the bound by the given horizontal and vertical
     * deltas.
     */
    add(x: number, y: number): Bounds;

    extend(object: LonLat): void;
    extend(object: Geometry.Point): void;
    extend(object: Bounds): void;

    extendXY(x: number, y: number): void;

    containsBounds(bounds: Bounds, partial: boolean, inclusive: boolean): boolean;
    containsLonLat(ll: LonLat, options?: BoundsOptions | null): boolean;
    containsPixel(px: Pixel, inclusive: boolean): boolean;
    contains(x: number, y: number, inclusive?: boolean): boolean;

    /**
     * Determine whether the target bounds intersects this bounds. Bounds are
     * considered intersecting if any of their edges intersect or if one
     * bounds contains the other.
     */
    intersectsBounds(bounds: Bounds, options?: BoundsOptions | boolean | null): boolean;

    /** @returns The quadrant ("br", "tr", "tl", "bl") in which the given LonLat lies. */
    determineQuadrant(lonlat: LonLat): string;

    /** Transform the Bounds object from source to dest. */
    transform(source: Projection, dest: Projection): Bounds;

    /** Wraps the bounds object around the dateline. */
    wrapDateLine(maxExtent: Bounds, options: WrapDateLineOptions): Bounds;

    /**
     * Alternative constructor that builds a new OpenLayers.Bounds from a
     * parameter string.
     */
    static fromString(str: string, reverseAxisOrder: boolean): Bounds;

    /**
     * Alternative constructor that builds a new OpenLayers.Bounds from an array.
     */
    static fromArray(bbox: number[], reverseAxisOrder: boolean): Bounds;

    /**
     * Alternative constructor that builds a new OpenLayers.Bounds from a size.
     */
    static fromSize(size: Size): Bounds;

    /**
     * Get the opposite quadrant for a given quadrant string.
     */
    static oppositeQuadrant(quadrant: string): string;
  }

  export class Size {
    /** Width */
    w: number;

    /** Height */
    h: number;

    constructor(width: number, height: number);

    /**
     * @returns The string representation of OpenLayers.Size object.
     * (e.g. <i>"w=55,h=66"</i>)
     */
    toString(): string;

    /**
     * Create a clone of this size object
     *
     * @returns A new Size object with the same w and h values.
     */
    clone(): Size;

    /**
     * Determine where this size is equal to another
     * Note that if size passed in is null, returns false.
     *
     * @returns The passed in size has the same h and w properties as this one.
     */
    equals(size: Size | SizeLike): boolean;
  }

  /** This class offers several methods for interacting with a wrapped pro4js projection object. */
  export class Projection {
    /** Proj4js.Proj instance. */
    proj: null;

    projCode: string;

    constructor(projCode: string, options?: object);

    /** Get the string SRS code. */
    getCode(): string;

    /** Get the units string for the projection -- returns null if proj4js is not available. */
    getUnits(): string | null;

    /**
     * Set a custom transform method between two projections.
     * Use this method in cases where the proj4js lib is not available or where custom projections need to be handled.
     */
    static addTransform(from: string, to: string, method: (point: PointLike) => void): void;

    /**
     * Transform a point coordinate from one projection to another.
     * Note that the input point is transformed in place.
     */
    static transform<T extends PointLike>(point: T, source: Projection, dest: Projection): T;

    /**
     * A null transformation useful for defining projection aliases when proj4js is not available.
     */
    static nullTransform<T extends PointLike>(point: T): T;
  }

  export abstract class Geometry {
    /** A unique identifier for this geometry. */
    id: string;

    /** This is set when a Geometry is added as component of another geometry */
    parent: Geometry | null;

    /** The bounds of this geometry */
    bounds: Bounds | null;

    destroy(): void;
    clone(): this;
    setBounds(bounds: Bounds): void;

    /** Nullify this components bounds and that of its parent as well. */
    clearBounds(): void;

    /**
     * Extend the existing bounds to include the new bounds.
     * If geometry's bounds is not yet set, then set a new Bounds.
     */
    extendBounds(newBounds: Bounds): void;

    /**
     * Get the bounds for this Geometry.
     * If bounds is not set, it is calculated again, this makes queries faster.
     */
    getBounds(): Bounds;

    /**
     * Calculate the closest distance between two geometries (on the x-y plane).
     */
    distanceTo(target: Geometry, options?: Geometry.DistanceOptions & {edge?: true | undefined, details: true}): Geometry.DistanceResult;
    distanceTo(target: Geometry, options?: Geometry.DistanceOptions): number;

    /**
     * @param nodes - For lines, only return vertices that are endpoints.
     * If false, for lines, only vertices that are not endpoints will be returned.
     * If not provided, all vertices will be returned.
     *
     * @returns A list of all points in this geometry.
     */
    getVertices(nodes?: boolean): Geometry.Point[];

    /** @returns Whether or not the geometry is at the specified location */
    atPoint(lonlat: LonLat, toleranceLon?: number, toleranceLat?: number): boolean;

    /** @returns The length of the collection by summing its parts */
    getLength(): number;

    /** @returns The area of the collection by summing its parts */
    getArea(): number;

    /**
     * @returns A text representation of the geometry. If the WKT format is
     * included in a build, this will be the Well-Known Text representation.
     */
    toString(): string;

    /**
     * Calculate the centroid of this geometry.
     * This method is defined in subclasses.
     */
    getCentroid(): Geometry.Point | false | null;

    CLASS_NAME: string;
  }

  namespace Geometry {
    export interface DistanceOptions {
      /**
       * Return details from the distance calculation.
       * Default is false.
       */
      details?: boolean | undefined;

      /**
       * Calculate the distance from this geometry to the nearest edge of the
       * target geometry.
       * Default is true.
       * If true, calling distanceTo from a geometry that is wholly contained
       * within the target will result in a non-zero distance.
       * If false, whenever geometries intersect, calling distanceTo will return 0.
       * If false, details cannot be returned.
       */
      edge?: boolean | undefined;
    }
    export interface DistanceResult {
      /** The distance between the two geometries. */
      distance: number;

      /** The X coordinates of the closest point on this geometry. */
      x0: number;

      /** The Y coordinates of the closest point on this geometry. */
      y0: number;

      /** The X coordinates of the closest point on the target geometry. */
      x1: number;

      /** The Y coordinates of the closest point on the target geometry. */
      y1: number;
    }

    /** Generate a geometry given a Well-Known Text string. */
    export function fromWKT(wkt: string): Geometry;

    export class Point extends Geometry {
      x: number;
      y: number;

      constructor(x: number, y: number);
      clone(out?: Point): this;

      /**
       * Recalculate the bounds by iterating through the components and
       * calling calling extendBounds() on each item.
       */
      calculateBounds(): void;

      /**
       * Determine whether another geometry is equivalent to this one.
       * Geometries are considered equivalent if all components have the same coordinates.
       */
      equals(geom: Point | null): boolean;

      /**
       * Moves a geometry by the given displacement along positive x and y axes.
       * This modifies the position of the geometry and clears the cached bounds.
       */
      move(x: number, y: number): void;

      /** Rotate a point around another. */
      rotate(angle: number, origin: Point): void;

      /**
       * Resize a point relative to some origin.
       * For points, this has the effect of scaling a vector (from the origin to the point).
       * This method is more useful on geometry collection subclasses.
       */
      resize(scale: number, origin: Point, ratio?: number): this;

      /** Determine if the input geometry intersects this one. */
      intersects(geometry: Geometry): boolean;

      /** Translate the x,y properties of the point from source to dest. */
      transform(source: Projection, dest: Projection): this;

      getVertices(nodes: boolean): [this];

      // CLASS_NAME: 'OpenLayers.Geometry.Point';
    }

    export class Collection<T extends Geometry> extends Geometry {
      /** The component parts of this geometry */
      components: T[];

      /**
       * An array of class names representing the types of
       * components that the collection can include. A null value means the
       * component types are not restricted.
       */
      componentTypes: string[];

      constructor(components?: T[]);

      /** Get a string representing the components for this collection */
      getComponentsString(): string;

      /**
       * Recalculate the bounds by iterating through the components and
       * calling calling extendBounds() on each item.
       */
      calculateBounds(): void;

      /** Add components to this geometry. */
      addComponents(components: T[]): void;

      /**
       * Add a new component (geometry) to the collection. If this.componentTypes
       * is set, then the component class name must be in the componentTypes array.
       */
      addComponent(component: T, index?: number): boolean;

      /** Remove components from this geometry. */
      removeComponents(components: T[]): boolean;

      /** Remove a component from this geometry. */
      removeComponent(component: T): boolean;

      /**
       * Calculate the approximate area of the polygon were it projected onto
       * the earth.
       */
      getGeodesicArea(projection: Projection): number;

      /** Compute the centroid for this geometry collection. */
      getCentroid(weighted?: boolean): Point | false | null;

      /**
       * Calculate the approximate length of the geometry were it projected onto
       * the earth.
       */
      getGeodesicLength(projection: Projection): number;

      /**
       * Moves a geometry by the given displacement along positive x and y axes.
       * This modifies the position of the geometry and clears the cached
       * bounds.
       */
      move(x: number, y: number): void;

      /** Rotate a geometry around some origin */
      rotate(angle: number, origin: Point): void;

      /**
       * Resize a geometry relative to some origin. Use this method to apply
       * a uniform scaling to a geometry.
       */
      resize(scale: number, origin: Point, ratio?: number): this;

      /**
       * Determine whether another geometry is equivalent to this one. Geometries
       * are considered equivalent if all components have the same coordinates.
       */
      equals(geometry: Geometry | null): boolean;

      /** Reproject the components geometry from source to dest. */
      transform(source: Projection, dest: Projection): this;

      /** Determine if the input geometry intersects this one. */
      intersects(geometry: Geometry): boolean;

      // CLASS_NAME: 'OpenLayers.Geometry.Collection';
    }

    export class LineString extends Curve {
      /**
       * This function will return a simplified LineString.
       * Simplification is based on the Douglas-Peucker algorithm.
       */
      simplify(tolerance: number): LineString;

      /**
       * Return a list of all points in this geometry.
       *
       * @param nodes - For lines, only return vertices that are endpoints.
       * If false, for lines, only vertices that are not endpoints will be returned.
       * If not provided, all vertices will be returned.
       */
      getVertices(nodes?: boolean): Point[];

      // CLASS_NAME: 'OpenLayers.Geometry.LineString';
    }

    export class LinearRing extends LineString {
      /**
       * Test if a point is inside a linear ring.
       *
       * @returns true if the point is inside the ring, 1 if the point is on
       * the ring, and false otherwise.
       */
      containsPoint(point: Point): boolean | 1;

      // CLASS_NAME: 'OpenLayers.Geometry.LinearRing';
    }

    export class Polygon extends Collection<LinearRing> {
      /**
       * Create a regular polygon around a radius. Useful for creating circles
       * and the like.
       *
       * @param origin - Center of polygon.
       * @param radius - Distance to vertex, in map units.
       * @param sides - Number of sides. 20 approximates a circle.
       * @param rotation - Original angle of rotation, in degrees.
       */
      static createRegularPolygon(origin: Point, radius: number, sides: number, rotation: number): Polygon;

      /**
       * Test if a point is inside a polygon.
       *
       * @returns true if the point is inside the polygon, 1 if the point is on
       * an edge, and false otherwise.
       */
      containsPoint(point: Point): boolean | 1;

      // CLASS_NAME: 'OpenLayers.Geometry.Polygon';
    }

    export class MultiPoint extends Collection<Point> {
      addPoint(point: Point, index?: number): void;
      removePoint(point: Point): void;
      // CLASS_NAME: 'OpenLayers.Geometry.MultiPoint';
    }
    export class Curve extends MultiPoint {
      // CLASS_NAME: 'OpenLayers.Geometry.Curve';
    }
    export class MultiLineString extends Collection<LineString> {
      // CLASS_NAME: 'OpenLayers.Geometry.MultiLineString';
    }
    export class MultiPolygon extends Collection<Polygon> {
      // CLASS_NAME: 'OpenLayers.Geometry.MultiPolygon';
    }
  }

  export const Util: {
    /**
     * @constant
     * The number of significant digits to retain to avoid floating point
     * precision errors.
     *
     * We use 14 as a "safe" default because, although IEEE 754 double floats
     * (standard on most modern operating systems) support up to about 16
     * significant digits, 14 significant digits are sufficient to represent
     * sub-millimeter accuracy in any coordinate system that anyone is likely to
     * use with OpenLayers.
     *
     * If DEFAULT_PRECISION is set to 0, the original non-truncating behavior
     * of OpenLayers <2.8 is preserved. Be aware that this will cause problems
     * with certain projections, e.g. spherical Mercator.
     */
    DEFAULT_PRECISION: number;

    /**
     * Tests that the provided object is an array.
     * This test handles the cross-IFRAME case not caught
     * by "a instanceof Array" and should be used instead.
     *
     * @param a - the object test.
     * @returns true if the object is an array.
     */
    isArray(a: any): a is any[];
  }

  export const String: {
    /**
     * Test whether a string starts with another string.
     *
     * @param str The string to test.
     * @param sub The substring to look for.
     * @returns The first string starts with the second.
     */
    startsWith(str: string, sub: string): boolean;

    /**
     * Test whether a string contains another string.
     *
     * @param str The string to test.
     * @param sub The substring to look for.
     * @returns The first string contains the second.
     */
    contains(str: string, sub: string): boolean;

    /**
     * Removes leading and trailing whitespace characters from a string.
     *
     * @param str The (potentially) space padded string.
     * @returns A trimmed version of the string with all leading and trailing
     *     spaces removed.
     */
    trim(str: string): string;
  };

  /**
   * Contains convenience functions for function manipulation.
   */
  export const Function: {
    /**
     * Bind a function to an object.  Method to easily create closures with
     *     'this' altered.
     *
     * @param func Input function.
     * @param object The object to bind to the input function (as this).
     * @returns A closure with 'this' set to the passed in object.
     */
    bind(func: Function, object: any): Function;

    /**
     * Bind a function to an object, and configure it to receive the event
     *     object as first parameter when called.
     *
     * @param func Input function to serve as an event listener.
     * @param object A reference to this.
     */
    bindAsEventListener(func: Function, object: any): Function;

    /**
     * A simple function to that just does "return false". We use this to
     * avoid attaching anonymous functions to DOM event handlers, which
     * causes "issues" on IE<8.
     *
     * @example
     * document.onclick = OpenLayers.Function.False;
     */
    False(): false;

    /**
     * A simple function to that just does "return true". We use this to
     * avoid attaching anonymous functions to DOM event handlers, which
     * causes "issues" on IE<8.
     *
     * @example
     * document.onclick = OpenLayers.Function.True;
     */
    True(): true;

    /**
     * A reusable function that returns `undefined`.
     */
    Void(): void;
  };

  export const Event: {
    KEY_SPACE: 32;
    KEY_BACKSPACE: 8;
    KEY_TAB: 9;
    KEY_RETURN: 13;
    KEY_ESC: 27;
    KEY_LEFT: 37;
    KEY_UP: 38;
    KEY_RIGHT: 39;
    KEY_DOWN: 40;
    KEY_DELETE: 46;

    /**
     * A hashtable cache of the event observers. Keyed by element._eventCacheID
     */
    observers: Record<string, object[]>;

    /**
     * Cross browser event element detection.
     *
     * @param event
     * @returns The element that caused the event
     */
    element(event: DOMEvent): HTMLElement;

    /**
     * Determine whether event was caused by a single touch
     *
     * @param event
     */
    isSingleTouch(event: DOMEvent): boolean;

    /**
     * Determine whether event was caused by a multi touch
     *
     * @param event
     */
    isMultiTouch(event: DOMEvent): boolean;

    /**
     * Determine whether event was caused by a left click.
     *
     * @param event
     */
    isLeftClick(event: DOMEvent): boolean;

    /**
     * Determine whether event was caused by a right mouse click.
     *
     * @param event
     */
    isRightClick(event: DOMEvent): boolean;

    /**
     * Stops an event from propagating.
     *
     * @param event
     * @param allowDefault If true, we stop the event chain but
     *     still allow the default browser behavior (text selection,
     *     radio-button clicking, etc).  Default is false.
     */
    stop(event: DOMEvent, allowDefault?: boolean): void;

    /**
     * @param event
     * @param tagName
     * @returns The first node with the given tagName, starting from the
     * node the event was triggered on and traversing the DOM upwards
     */
    findElement(event: DOMEvent, tagName: string): HTMLElement | Window;

    /**
     * @param elementParam
     * @param name
     * @param observer
     * @param useCapture
     */
    observe(elementParam: EventTarget | string, name: string, observer: Function, useCapture: boolean): void;

    /**
     * Given the id of an element to stop observing, cycle through the
     *   element's cached observers, calling stopObserving on each one,
     *   skipping those entries which can no longer be removed.
     *
     * @param elementParam
     */
    stopObservingElement(elementParam: EventTarget | string): void;

    /**
     * @param elementParam
     * @param name
     * @param observer
     * @param useCapture
     * @returns Whether or not the event observer was removed
     */
    stopObserving(elementParam: EventTarget | string, name: string, observer: Function, useCapture?: boolean): boolean;

    /**
     * Cycle through all the element entries in the events cache and call
     *   stopObservingElement on each.
     */
    unloadCache(): void;
  }

  export class Events<T extends {}> {
    object: T;

    /**
     * @param element a DOM element to attach browser events to
     */
    attachToElement(element: HTMLElement): void;

    /**
     * Convenience method for registering listeners with a common scope.
     * Internally, this method calls <register> as shown in the examples below.
     *
     * @param object
     *
     * @example
     * // register a single listener for the "loadstart" event
     * events.on({"loadstart": loadStartListener});
     *
     * // this is equivalent to the following
     * events.register("loadstart", undefined, loadStartListener);
     *
     * // register multiple listeners to be called with the same `this` object
     * events.on({
     *     "loadstart": loadStartListener,
     *     "loadend": loadEndListener,
     *     scope: object
     * });
     *
     * // this is equivalent to the following
     * events.register("loadstart", object, loadStartListener);
     * events.register("loadend", object, loadEndListener);
     */
    on<O extends {scope: {}}>(object: O & ThisType<O['scope']>): void;
    on(object: Record<string, () => void> & ThisType<T>): void;

    /**
     * Register an event on the events object.
     *
     * When the event is triggered, the 'func' function will be called, in the
     * context of 'obj'. Imagine we were to register an event, specifying an
     * OpenLayers.Bounds Object as 'obj'. When the event is triggered, the
     * context in the callback function will be our Bounds object. This means
     * that within our callback function, we can access the properties and
     * methods of the Bounds object through the "this" variable. So our
     * callback could execute something like:
     * :    leftStr = "Left: " + this.left;
     *
     *                   or
     *
     * :    centerStr = "Center: " + this.getCenterLonLat();
     *
     * @param type - Name of the event to register
     * @param obj - The object to bind the context to for the callback#.
     *     If no object is specified, default is the Events's 'object' property.
     * @param func - The callback function. If no callback is
     *     specified, this function does nothing.
     * @param priority - If true, adds the new listener to the
     *     *front* of the events queue instead of to the end.
     * @param priority.extension - If true, then the event will be registered as
     *     extension event. Extension events are handled before all other
     *     events.
     */
    register<TThis>(
      type: string,
      obj: TThis,
      func: (this: TThis extends null | undefined ? T : TThis) => void,
      priority?: boolean | {extension: boolean}
    ): void;

    /**
     * Same as register() but adds the new listener to the *front* of the
     *     events queue instead of to the end.
     *
     * @param type - Name of the event to register
     * @param obj - The object to bind the context to for the callback#.
     *                If no object is specified, default is the Events's
     *                'object' property.
     * @param func - The callback function. If no callback is
     *                   specified, this function does nothing.
     */
    registerPriority<TThis>(
      type: string,
      obj: TThis,
      func: (this: TThis extends null | undefined ? T : TThis) => void
    ): void;

    /**
     * Convenience method for unregistering listeners with a common scope.
     *     Internally, this method calls <unregister> as shown in the examples
     *     below.
     *
     * @example
     * // unregister a single listener for the "loadstart" event
     * events.un({"loadstart": loadStartListener});
     *
     * // this is equivalent to the following
     * events.unregister("loadstart", undefined, loadStartListener);
     *
     * // unregister multiple listeners with the same `this` object
     * events.un({
     *     "loadstart": loadStartListener,
     *     "loadend": loadEndListener,
     *     scope: object
     * });
     *
     * // this is equivalent to the following
     * events.unregister("loadstart", object, loadStartListener);
     * events.unregister("loadend", object, loadEndListener);
     */
    un(object: any): void;

    /**
     * @param type
     * @param obj - If none specified, defaults to this.object
     * @param func
     */
    unregister(type: string, obj: any, func: () => void): void;

    /**
     * Remove all listeners for a given event type. If type is not registered,
     *     does nothing.
     *
     * @param type
     */
    remove(type: string): void;

    /**
     * Trigger a specified registered event.
     *
     * @param type
     * @param evt - will be passed to the listeners.
     * @returns The last listener return.  If a listener returns false, the
     *     chain of listeners will stop getting called.
     */
    triggerEvent(type: string, evt?: unknown): boolean;

    /**
     * Basically just a wrapper to the triggerEvent() function, but takes
     *     care to set a property 'xy' on the event with the current mouse
     *     position.
     *
     * @param evt
     */
    handleBrowserEvent(evt: Event): void;

    /**
     * @param evt
     * @returns The current xy coordinate of the mouse, adjusted for offsets
     */
    getMousePosition(evt: Event): Pixel;
  }

  // Layer
  // Layer.Google
  // Layer.Google.cache
  // Layer.Google.v3
  // Layer.Grid
  // Layer.Image
  // Layer.prototype.moveTo.apply
  // Layer.TMS
  // Layer.WMS
  // Layer.WMTS
  // Layer.XYZ

  export interface LayerOptions {}

  export class Layer {
    id: string;
    name: string;
    div: HTMLElement;

    /** The layer's opacity. Float number between 0.0 and 1.0. */
    opacity: number;

    /**
     * If a layer's display should not be scale-based, this should
     * be set to true. This will cause the layer, as an overlay, to always
     * be 'active', by always returning true from the calculateInRange()
     * function.
     *
     * If not explicitly specified for a layer, its value will be
     * determined on startup in initResolutions() based on whether or not
     * any scale-specific properties have been set as options on the
     * layer. If no scale-specific options have been set on the layer, we
     * assume that it should always be in range.
     */
    alwaysInRange: boolean;

    /** The properties that are used for calculating resolutions information. */
    RESOLUTION_PROPERTIES: string[];

    /**
     * Register a listener for a particular event with the following syntax:
     * ```
     * layer.events.register(type, obj, listener);
     * ```
     *
     * Listeners will be called with a reference to an event object. The
     * properties of this event depends on exactly what happened.
     *
     * All event objects have at least the following properties:
     * object - {Object} A reference to layer.events.object.
     * element - {DOMElement} A reference to layer.events.element.
     *
     * Supported map event types:
     * loadstart - Triggered when layer loading starts. When using a Vector
     * layer with a Fixed or BBOX strategy, the event object includes
     * a *filter* property holding the OpenLayers.Filter used when
     * calling read on the protocol.
     * loadend - Triggered when layer loading ends. When using a Vector layer
     * with a Fixed or BBOX strategy, the event object includes a
     * *response* property holding an OpenLayers.Protocol.Response object.
     * visibilitychanged - Triggered when the layer's visibility property is
     * changed, e.g. by turning the layer on or off in the layer switcher.
     * Note that the actual visibility of the layer can also change if it
     * gets out of range (see <calculateInRange>). If you also want to catch
     * these cases, register for the map's 'changelayer' event instead.
     * move - Triggered when layer moves (triggered with every mousemove
     * during a drag).
     * moveend - Triggered when layer is done moving, object passed as
     * argument has a zoomChanged boolean property which tells that the
     * zoom has changed.
     * added - Triggered after the layer is added to a map. Listeners will
     * receive an object with a *map* property referencing the map and a
     * *layer* property referencing the layer.
     * removed - Triggered after the layer is removed from the map. Listeners
     * will receive an object with a *map* property referencing the map and
     * a *layer* property referencing the layer.
     */
    events: Events<this>;

    /**
     * This variable is set when the layer is added to
     * the map, via the accessor function setMap()
     */
    map: Map;

    /**
     * Whether or not the layer is a base layer. This should be set individually
     * by all subclasses. Default is false
     */
    isBaseLayer: boolean;

    /** The layer's images have an alpha channel.  Default is false. */
    alpha: boolean;

    /** Display the layer's name in the layer switcher.  Default is true */
    displayInLayerSwitcher: boolean;

    /** The layer should be displayed in the map.  Default is true. */
    visibility: boolean;

    /**
     * Attribution string, displayed when an <OpenLayers.Control.Attribution>
     * has been added to the map.
     */
    attribution: string;

    /**
     * The current map resolution is within the layer's min/max
     * range. This is set in <OpenLayers.Map.setCenter> whenever the zoom
     * changes.
     */
    inRange: boolean;

    /**
     * For layers with a gutter, the image is larger than
     * the tile by twice the gutter in each dimension.
     */
    imageSize: Size;

    /**
     * An optional object whose properties will be set on the layer.
     * Any of the layer properties can be set as a property of the options
     * object and sent to the constructor when the layer is created.
     */
    options: LayerOptions;

    /**
     * An optional object whose properties will be set on the layer.
     * Any of the layer properties can be set as a property of the options
     * object and sent to the constructor when the layer is created.
     */
    eventListeners: object;

    /**
     * Determines the width (in pixels) of the gutter around image
     * tiles to ignore. By setting this property to a non-zero value,
     * images will be requested that are wider and taller than the tile
     * size by a value of 2 x gutter. This allows artifacts of rendering
     * at tile edges to be ignored. Set a gutter value that is equal to
     * half the size of the widest symbol that needs to be displayed.
     * Defaults to zero. Non-tiled layers always have zero gutter.
     */
    gutter: number;

    /**
     * Specifies the projection of the layer.
     * Can be set in the layer options. If not specified in the layer options,
     * it is set to the default projection specified in the map,
     * when the layer is added to the map.
     * Projection along with default maxExtent and resolutions
     * are set automatically with commercial baselayers in EPSG:3857,
     * such as Google, Bing and OpenStreetMap, and do not need to be specified.
     * Otherwise, if specifying projection, also set maxExtent,
     * maxResolution or resolutions as appropriate.
     * When using vector layers with strategies, layer projection should be set
     * to the projection of the source data if that is different from the map default.
     */
    projection: Projection;

    /**
     * The layer map units. Defaults to null. Possible values
     * are 'degrees' (or 'dd'), 'm', 'ft', 'km', 'mi', 'inches'.
     * Normally taken from the projection.
     * Only required if both map and layers do not define a projection,
     * or if they define a projection which does not define units.
     */
    units: string;

    /**
     * An array of map scales in descending order. The values in the
     * array correspond to the map scale denominator. Note that these
     * values only make sense if the display (monitor) resolution of the
     * client is correctly guessed by whomever is configuring the
     * application. In addition, the units property must also be set.
     * Use <resolutions> instead wherever possible.
     */
    scales: any[];

    /**
     * A list of map resolutions (map units per pixel) in descending
     * order. If this is not set in the layer constructor, it will be set
     * based on other resolution related properties (maxExtent,
     * maxResolution, maxScale, etc.).
     */
    resolutions: any[];

    /**
     * The maximum extent for the layer.  Defaults to null.
     */
    maxExtent: Bounds;

    /**
     * The minimum extent for the layer.  Defaults to null.
     */
    minExtent: Bounds;

    /**
     * Default max is 360 deg / 256 px, which corresponds to
     * zoom level 0 on gmaps. Specify a different value in the layer
     * options if you are not using the default <OpenLayers.Map.tileSize>
     * and displaying the whole world.
     */
    maxResolution: number;
    minResolution: number;
    numZoomLevels: number;
    minScale: number;
    maxScale: number;

    /**
     * Request map tiles that are completely outside of the max
     * extent for this layer. Defaults to false.
     */
    displayOutsideMaxExtent: boolean;

    /**
     * Wraps the world at the international dateline, so the map can
     * be panned infinitely in longitudinal direction. Only use this on the
     * base layer, and only if the layer's maxExtent equals the world bounds.
     */
    wrapDateLine: boolean;

    /**
     * This object can be used to store additional information on a
     * layer object.
     */
    metadata: Object;

    /**
     * @param name The layer name
     * @param options Hashtable of extra options to tag onto the layer
     */
    constructor(name: string, options: LayerOptions);

    /**
     * Destroy is a destructor: this is to alleviate cyclic references which
     * the Javascript garbage cleaner can not take care of on its own.
     *
     * @param setNewBaseLayer Set a new base layer when this layer has
     * been destroyed. Default is true.
     */
    destroy(setNewBaseLayer?: boolean): void;

    /**
     * @param obj The layer to be cloned
     * @returns An exact clone of this <OpenLayers.Layer>
     */
    clone(): Layer;

    /**
     * Extracts an object from the layer with the properties that were set as
     * options, but updates them with the values currently set on the
     * instance.
     *
     * @returns the <options> of the layer, representing the current state.
     */
    private getOptions(): LayerOptions;

    /**
     * Sets the new layer name for this layer. Can trigger a changelayer event
     * on the map.
     *
     * @param newName The new name.
     */
    setName(newName: string): void;

    /**
     * @param newOptions
     * @param reinitialize If set to true, and if resolution options of the
     * current baseLayer were changed, the map will be recentered to make
     * sure that it is displayed with a valid resolution, and a
     * changebaselayer event will be triggered.
     */
    addOptions(newOptions: LayerOptions, reinitialize: boolean): void;

    /** This function can be implemented by subclasses */
    onMapResize(): void;

    /**
     * Redraws the layer. Returns true if the layer was redrawn, false if not.
     *
     * @returns The layer was redrawn.
     */
    redraw(): void;

    /**
     * @param bounds
     * @param zoomChanged Tells when zoom has changed, as layers have to
     * do some init work in that case.
     * @param dragging
     */
    moveTo(bounds: Bounds, zoomChanged: boolean, dragging: boolean): void;

    /**
     * Move the layer based on pixel vector. To be implemented by subclasses.
     *
     * @param dx The x coord of the displacement vector.
     * @param dy The y coord of the displacement vector.
     */
    moveByPx(dx: number, dy: number): void;

    /**
     * Set the map property for the layer. This is done through an accessor
     * so that subclasses can override this and take special action once
     * they have their map variable set.
     *
     * Here we take care to bring over any of the necessary default
     * properties from the map.
     *
     * @param map
     */
    setMap(map: Map): void;

    /**
     * Called at the end of the map.addLayer sequence. At this point, the map
     * will have a base layer. To be overridden by subclasses.
     */
    private afterAdd(): void;

    /**
     * Just as setMap() allows each layer the possibility to take a
     * personalized action on being added to the map, removeMap() allows
     * each layer to take a personalized action on being removed from it.
     * For now, this will be mostly unused, except for the EventPane layer,
     * which needs this hook so that it can remove the special invisible
     * pane.
     *
     * @param map
     */
    removeMap(map: Map): void;

    /**
     * @param bounds optional tile bounds, can be used by subclasses that have
     * to deal with different tile sizes at the layer extent edges (e.g. Zoomify)
     * @returns The size that the image should be, taking into account gutters.
     */
    getImageSize(bounds: Bounds): Size;

    /**
     * Set the tile size based on the map size. This also sets layer.imageSize
     * or use by Tile.Image.
     *
     * @param size
     */
    setTileSize(size: Size): void;

    /**
     * @returns The layer should be displayed (if in range).
     */
    getVisibility(): boolean;

    /**
     * Set the visibility flag for the layer and hide/show & redraw
     * accordingly. Fire event unless otherwise specified
     *
     * Note that visibility is no longer simply whether or not the layer's
     * style.display is set to "block". Now we store a 'visibility' state
     * property on the layer class, this allows us to remember whether or
     * not we *desire* for a layer to be visible. In the case where the
     * map's resolution is out of the layer's range, this desire may be
     * subverted.
     *
     * @param visibility Whether or not to display the layer (if in range)
     */
    setVisibility(visibility: boolean): void;

    /**
     * Hide or show the Layer. This is designed to be used internally, and
     * is not generally the way to enable or disable the layer. For that,
     * use the setVisibility function instead..
     *
     * @param display
     */
    display(display: boolean): void;

    /**
     * @returns The layer is displayable at the current map's current
     * resolution. Note that if 'alwaysInRange' is true for the layer,
     * this function will always return true.
     */
    calculateInRange(): boolean;

    /**
     * @param isBaseLayer
     */
    setIsBaseLayer(isBaseLayer: boolean): void;

    /********************************************************/
    /*                                                      */
    /*               Baselayer Functions                    */
    /*                                                      */
    /********************************************************/

    /**
     * This method's responsibility is to set up the 'resolutions' array
     * for the layer -- this array is what the layer will use to interface
     * between the zoom levels of the map and the resolution display
     * of the layer.
     *
     * The user has several options that determine how the array is set up.
     *
     * For a detailed explanation, see the following wiki from the
     * openlayers.org homepage:
     * http://trac.openlayers.org/wiki/SettingZoomLevels
     */
    initResolutions(): void;

    /**
     * Derive resolutions from scales.
     *
     * @param scales Scales
     * @returns Resolutions
     */
    private resolutionsFromScales(scales: number[]): number[];

    /**
     * Calculate resolutions based on the provided properties.
     *
     * @param props Properties
     * @returns Array of resolutions.
     */
    calculateResolutions(props: object): number[];

    /**
     * @returns The currently selected resolution of the map, taken from the
     * resolutions array, indexed by current zoom level.
     */
    getResolution(): number;

    /**
     * @returns A Bounds object which represents the lon/lat bounds of the
     * current viewPort.
     */
    getExtent(): Bounds;

    /**
     * @param extent
     * @param closest Find the zoom level that most closely fits the
     * specified bounds. Note that this may result in a zoom that does
     * not exactly contain the entire extent.
     * Default is false.
     * @returns The index of the zoomLevel (entry in the resolutions array)
     * for the passed-in extent. We do this by calculating the ideal
     * resolution for the given extent (based on the map size) and then
     * calling getZoomForResolution(), passing along the 'closest'
     * parameter.
     */
    getZoomForExtent(extent: Bounds, closest?: boolean): number;

    /**
     * Calculates the max extent which includes all of the data for the layer.
     * This function is to be implemented by subclasses.
     */
    private getDataExtent(): Bounds;

    /**
     * @param zoom
     * @returns A suitable resolution for the specified zoom.
     */
    getResolutionForZoom(zoom: number): number;

    /**
     * @param resolution
     * @param closest Find the zoom level that corresponds to the absolute
     * closest resolution, which may result in a zoom whose corresponding
     * resolution is actually smaller than we would have desired (if this
     * is being called from a getZoomForExtent() call, then this means that
     * the returned zoom index might not actually contain the entire
     * extent specified... but it'll be close).
     * Default is false.
     * @returns The index of the zoomLevel (entry in the resolutions array)
     * that corresponds to the best fit resolution given the passed in
     * value and the 'closest' specification.
     */
    getZoomForResolution(resolution: number, closest?: boolean): number;

    /**
     * @param viewPortPx An OpenLayers.Pixel or an object with a 'x' and 'y'
     * properties.
     * @returns An OpenLayers.LonLat which is the passed-in
     * view port <OpenLayers.Pixel>, translated into lon/lat by the layer.
     */
    getLonLatFromViewPortPx(viewPortPx: PixelLike | null): LonLat | null;

    /**
     * Returns a pixel location given a map location. This method will return
     * fractional pixel values.
     *
     * @param lonlat An OpenLayers.LonLat or an object with a 'lon' and 'lat'
     * properties.
     * @param resolution
     * @returns An <OpenLayers.Pixel> which is the passed-in
     * lonlat translated into view port pixels.
     */
    getViewPortPxFromLonLat(lonlat: LonLatLike | null, resolution?: number): Pixel | null;

    /**
     * Sets the opacity for the entire layer (all images)
     *
     * @param opacity
     */
    setOpacity(opacity: number): void;

    /**
     * @returns the z-index of this layer
     */
    getZIndex(): number;

    /**
     * @param zIndex
     */
    setZIndex(zIndex: number): void;

    /**
     * This function will take a bounds, and if wrapDateLine option is set
     * on the layer, it will return a bounds which is wrapped around the
     * world. We do not wrap for bounds which *cross* the
     * maxExtent.left/right, only bounds which are entirely to the left
     * or entirely to the right.
     *
     * @param bounds
     */
    adjustBounds(bounds: Bounds): Bounds;
  }
  namespace Layer {
    export interface WMSGetMapParams {
      version?: string | undefined;
      exceptions?: string | undefined;
      transparent?: string | undefined;
      format?: string | undefined;
      styles?: string | undefined;
      layers: string;
      service?: string | undefined;
    }

    export interface WMSOptions {
      opacity?: number | undefined;
      singleTile?: boolean | undefined;
      isBaseLayer?: boolean | undefined;
      encodeBBOX?: boolean | undefined;
      noMagic?: boolean | undefined;
      yx?: Object | undefined;
    }

    export interface TileOptions {
      crossOriginKeyword?: string | undefined;
    }

    export class ArcGIS93Rest {}
    export class ArcGISCache {}
    export class ArcIMS {}
    export class Bing {}
    export class Boxes {}
    export class EventPane {}
    export class FixedZoomLevels {}
    export class GeoRSS {}
    export class Google {}

    export class Grid extends HTTPRequest {
      /**
       * APIProperty: tileSize
       * {<OpenLayers.Size>}
       */
      tileSize: Size;

      /**
       * Property: tileOriginCorner
       * {String} If the <tileOrigin> property is not provided, the tile origin
       * will be derived from the layer's <maxExtent>. The corner of the
       * <maxExtent> used is determined by this property. Acceptable values
       * are "tl" (top left), "tr" (top right), "bl" (bottom left), and "br"
       * (bottom right). Default is "bl".
       */
      tileOriginCorner: string;

      /**
       * APIProperty: tileOrigin
       * {<OpenLayers.LonLat>} Optional origin for aligning the grid of tiles.
       * If provided, requests for tiles at all resolutions will be aligned
       * with this location (no tiles shall overlap this location). If
       * not provided, the grid of tiles will be aligned with the layer's
       * <maxExtent>. Default is ``null``.
       */
      tileOrigin: LonLat;

      /** APIProperty: tileOptions
       * {Object} optional configuration options for <OpenLayers.Tile> instances
       * created by this Layer, if supported by the tile class.
       */
      tileOptions: Object;

      /**
       * APIProperty: tileClass
       * {<OpenLayers.Tile>} The tile class to use for this layer.
       * Defaults is OpenLayers.Tile.Image.
       */
      tileClass: OpenLayers.Tile;

      /**
       * Property: grid
       * {Array(Array(<OpenLayers.Tile>))} This is an array of rows, each row is
       * an array of tiles.
       */
      grid: OpenLayers.Tile[][];

      /**
       * APIProperty: singleTile
       * {Boolean} Moves the layer into single-tile mode, meaning that one tile
       * will be loaded. The tile's size will be determined by the 'ratio'
       * property. When the tile is dragged such that it does not cover the
       * entire viewport, it is reloaded.
       */
      singleTile: boolean;

      /** APIProperty: ratio
       * {Float} Used only when in single-tile mode, this specifies the
       * ratio of the size of the single tile to the size of the map.
       * Default value is 1.5.
       */
      ratio: number;

      /**
       * APIProperty: buffer
       * {Integer} Used only when in gridded mode, this specifies the number of
       * extra rows and columns of tiles on each side which will
       * surround the minimum grid tiles to cover the map.
       * For very slow loading layers, a larger value may increase
       * performance somewhat when dragging, but will increase bandwidth
       * use significantly.
       */
      buffer: number;

      /**
       * APIProperty: transitionEffect
       * {String} The transition effect to use when the map is zoomed.
       * Two posible values:
       *
       * "resize" - Existing tiles are resized on zoom to provide a visual
       * effect of the zoom having taken place immediately. As the
       * new tiles become available, they are drawn on top of the
       * resized tiles (this is the default setting).
       * "map-resize" - Existing tiles are resized on zoom and placed below the
       * base layer. New tiles for the base layer will cover existing tiles.
       * This setting is recommended when having an overlay duplicated during
       * the transition is undesirable (e.g. street labels or big transparent
       * fills).
       * null - No transition effect.
       *
       * Using "resize" on non-opaque layers can cause undesired visual
       * effects. Set transitionEffect to null in this case.
       */
      transitionEffect: string;

      /**
       * APIProperty: numLoadingTiles
       * {Integer} How many tiles are still loading?
       */
      numLoadingTiles: number;

      /**
       * Property: serverResolutions
       * {Array(Number}} This property is documented in subclasses as
       * an API property.
       */
      serverResolutions: number[];

      /**
       * Property: loading
       * {Boolean} Indicates if tiles are being loaded.
       */
      loading: boolean;

      /**
       * Property: backBuffer
       * {DOMElement} The back buffer.
       */
      backBuffer: HTMLElement;

      /**
       * Property: gridResolution
       * {Number} The resolution of the current grid. Used for backbuffer and
       * client zoom. This property is updated every time the grid is
       * initialized.
       */
      gridResolution: number;

      /**
       * Property: backBufferResolution
       * {Number} The resolution of the current back buffer. This property is
       * updated each time a back buffer is created.
       */
      backBufferResolution: number;

      /**
       * Property: backBufferLonLat
       * {Object} The top-left corner of the current back buffer. Includes lon
       * and lat properties. This object is updated each time a back buffer
       * is created.
       */
      backBufferLonLat: { lon: number; lat: number };

      /**
       * Property: backBufferTimerId
       * {Number} The id of the back buffer timer. This timer is used to
       * delay the removal of the back buffer, thereby preventing
       * flash effects caused by tile animation.
       */
      backBufferTimerId: number;

      /**
       * APIProperty: removeBackBufferDelay
       * {Number} Delay for removing the backbuffer when all tiles have finished
       * loading. Can be set to 0 when no css opacity transitions for the
       * olTileImage class are used. Default is 0 for <singleTile> layers,
       * 2500 for tiled layers. See <className> for more information on
       * tile animation.
       */
      removeBackBufferDelay: number;

      /**
       * APIProperty: className
       * {String} Name of the class added to the layer div. If not set in the
       * options passed to the constructor then className defaults to
       * "olLayerGridSingleTile" for single tile layers (see <singleTile>),
       * and "olLayerGrid" for non single tile layers.
       *
       * Note:
       *
       * The displaying of tiles is not animated by default for single tile
       * layers - OpenLayers' default theme (style.css) includes this:
       * (code)
       * .olLayerGrid .olTileImage {
       * -webkit-transition: opacity 0.2s linear;
       * -moz-transition: opacity 0.2s linear;
       * -o-transition: opacity 0.2s linear;
       * transition: opacity 0.2s linear;
       * }
       * (end)
       * To animate tile displaying for any grid layer the following
       * CSS rule can be used:
       * (code)
       * .olTileImage {
       * -webkit-transition: opacity 0.2s linear;
       * -moz-transition: opacity 0.2s linear;
       * -o-transition: opacity 0.2s linear;
       * transition: opacity 0.2s linear;
       * }
       * (end)
       * In that case, to avoid flash effects, <removeBackBufferDelay>
       * should not be zero.
       */
      className: string;

      /**
       * Property: gridLayout
       * {Object} Object containing properties tilelon, tilelat, startcol,
       * startrow
       */
      gridLayout: {
        tilelon: number;
        tilelat: number;
        startcol: number;
        startrow: number;
      };

      /**
       * Property: rowSign
       * {Number} 1 for grids starting at the top, -1 for grids starting at the
       * bottom. This is used for several grid index and offset calculations.
       */
      rowSign: number;

      /**
       * Property: transitionendEvents
       * {Array} Event names for transitionend
       */
      transitionendEvents: string[];

      /**
       * Constructor: OpenLayers.Layer.Grid
       * Create a new grid layer
       *
       * Parameters:
       * name - {String}
       * url - {String}
       * params - {Object}
       * options - {Object} Hashtable of extra options to tag onto the layer
       */
      constructor(name: string, url: string, params: {}, options: {});

      /**
       * Method: initProperties
       * Set any properties that depend on the value of singleTile.
       * Currently sets removeBackBufferDelay and className
       */
      private initProperties(): void;

      /**
       * Method: setMap
       *
       * Parameters:
       * map - {<OpenLayers.Map>} The map.
       */
      setMap(map: Map): void;

      /**
       * Method: removeMap
       * Called when the layer is removed from the map.
       *
       * Parameters:
       * map - {<OpenLayers.Map>} The map.
       */
      removeMap(map: Map): void;

      /**
       * APIMethod: destroy
       * Deconstruct the layer and clear the grid.
       */
      destroy(): void;

      /**
       * Method: clearGrid
       * Go through and remove all tiles from the grid, calling
       * destroy() on each of them to kill circular references
       */
      private clearGrid(): void;

      /**
       * APIMethod: addOptions
       *
       * Parameters:
       * newOptions - {Object}
       * reinitialize - {Boolean} If set to true, and if resolution options of the
       * current baseLayer were changed, the map will be recentered to make
       * sure that it is displayed with a valid resolution, and a
       * changebaselayer event will be triggered.
       */
      addOptions(newOptions: {}, reinitialize: boolean): void;

      /**
       * APIMethod: clone
       * Create a clone of this layer
       *
       * Parameters:
       * obj - {Object} Is this ever used?
       *
       * Returns:
       * {<OpenLayers.Layer.Grid>} An exact clone of this OpenLayers.Layer.Grid
       */
      clone(obj?: Object): Layer.Grid;

      /**
       * Method: moveTo
       * This function is called whenever the map is moved. All the moving
       * of actual 'tiles' is done by the map, but moveTo's role is to accept
       * a bounds and make sure the data that that bounds requires is pre-loaded.
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       * zoomChanged - {Boolean}
       * dragging - {Boolean}
       */
      moveTo(bounds: Bounds, zoomChanged: boolean, dragging: boolean): void;

      /**
       * Method: getTileData
       * Given a map location, retrieve a tile and the pixel offset within that
       * tile corresponding to the location. If there is not an existing
       * tile in the grid that covers the given location, null will be
       * returned.
       *
       * Parameters:
       * loc - {<OpenLayers.LonLat>} map location
       *
       * Returns:
       * {Object} Object with the following properties: tile ({<OpenLayers.Tile>}),
       * i ({Number} x-pixel offset from top left), and j ({Integer} y-pixel
       * offset from top left).
       */
      private getTileData(loc: LonLat): { tile: Tile; i: number; j: number };

      /**
       * Method: destroyTile
       *
       * Parameters:
       * tile - {<OpenLayers.Tile>}
       */
      private destroyTile(tile: Tile): void;

      /**
       * Method: getServerResolution
       * Return the closest server-supported resolution.
       *
       * Parameters:
       * resolution - {Number} The base resolution. If undefined the
       * map resolution is used.
       *
       * Returns:
       * {Number} The closest server resolution value.
       */
      private getServerResolution(resolution: number): number;

      /**
       * Method: getServerZoom
       * Return the zoom value corresponding to the best matching server
       * resolution, taking into account <serverResolutions> and <zoomOffset>.
       *
       * Returns:
       * {Number} The closest server supported zoom. This is not the map zoom
       * level, but an index of the server's resolutions array.
       */
      private getServerZoom(): number;

      /**
       * Method: applyBackBuffer
       * Create, insert, scale and position a back buffer for the layer.
       *
       * Parameters:
       * resolution - {Number} The resolution to transition to.
       */
      private applyBackBuffer(resolution: number): void;

      /**
       * Method: createBackBuffer
       * Create a back buffer.
       *
       * Returns:
       * {DOMElement} The DOM element for the back buffer, undefined if the
       * grid isn't initialized yet.
       */
      private createBackBuffer(): HTMLElement;

      /**
       * Method: removeBackBuffer
       * Remove back buffer from DOM.
       */
      private removeBackBuffer(): void;

      /**
       * Method: moveByPx
       * Move the layer based on pixel vector.
       *
       * Parameters:
       * dx - {Number}
       * dy - {Number}
       */
      moveByPx(dx: number, dy: number): void;

      /**
       * APIMethod: setTileSize
       * Check if we are in singleTile mode and if so, set the size as a ratio
       * of the map size (as specified by the layer's 'ratio' property).
       *
       * Parameters:
       * size - {<OpenLayers.Size>}
       */
      setTileSize(size: Size): void;

      /**
       * APIMethod: getTilesBounds
       * Return the bounds of the tile grid.
       *
       * Returns:
       * {<OpenLayers.Bounds>} A Bounds object representing the bounds of all the
       * currently loaded tiles (including those partially or not at all seen
       * onscreen).
       */
      getTilesBounds(): Bounds;

      /**
       * Method: initSingleTile
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       */
      private initSingleTile(bounds: Bounds);

      /**
       * Method: calculateGridLayout
       * Generate parameters for the grid layout.
       *
       * Parameters:
       * bounds - {<OpenLayers.Bound>|Object} OpenLayers.Bounds or an
       * object with a 'left' and 'top' properties.
       * origin - {<OpenLayers.LonLat>|Object} OpenLayers.LonLat or an
       * object with a 'lon' and 'lat' properties.
       * resolution - {Number}
       *
       * Returns:
       * {Object} Object containing properties tilelon, tilelat, startcol,
       * startrow
       */
      private calculateGridLayout(
        bounds: Bounds,
        origin: LonLat,
        resolution: number
      ): {
        tilelon: number;
        tilelat: number;
        startcol: number;
        startrow: number;
      };

      getImageSize(): Size;

      /**
       * Method: getTileOrigin
       * Determine the origin for aligning the grid of tiles. If a <tileOrigin>
       * property is supplied, that will be returned. Otherwise, the origin
       * will be derived from the layer's <maxExtent> property. In this case,
       * the tile origin will be the corner of the <maxExtent> given by the
       * <tileOriginCorner> property.
       *
       * Returns:
       * {<OpenLayers.LonLat>} The tile origin.
       */
      private getTileOrigin(): LonLat;

      /**
       * Method: getTileBoundsForGridIndex
       *
       * Parameters:
       * row - {Number} The row of the grid
       * col - {Number} The column of the grid
       *
       * Returns:
       * {<OpenLayers.Bounds>} The bounds for the tile at (row, col)
       */
      private getTileBoundsForGridIndex(row: number, col: number): Bounds;

      /**
       * Method: initGriddedTiles
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       */
      private initGriddedTiles(bounds: Bounds): void;

      /**
       * Method: getMaxExtent
       * Get this layer's maximum extent. (Implemented as a getter for
       * potential specific implementations in sub-classes.)
       *
       * Returns:
       * {<OpenLayers.Bounds>}
       */
      private getMaxExtent(): Bounds;

      /**
       * APIMethod: addTile
       * Create a tile, initialize it, and add it to the layer div.
       *
       * Parameters
       * bounds - {<OpenLayers.Bounds>}
       * position - {<OpenLayers.Pixel>}
       *
       * Returns:
       * {<OpenLayers.Tile>} The added OpenLayers.Tile
       */
      addTile(bounds: Bounds, position: number): Tile;

      /**
       * Method: addTileMonitoringHooks
       * This function takes a tile as input and adds the appropriate hooks to
       * the tile so that the layer can keep track of the loading tiles.
       *
       * Parameters:
       * tile - {<OpenLayers.Tile>}
       */
      private addTileMonitoringHooks(tile: Tile): void;

      /**
       * Method: removeTileMonitoringHooks
       * This function takes a tile as input and removes the tile hooks
       * that were added in addTileMonitoringHooks()
       *
       * Parameters:
       * tile - {<OpenLayers.Tile>}
       */
      private removeTileMonitoringHooks(tile: Tile): void;

      /**
       * Method: moveGriddedTiles
       */
      private moveGriddedTiles(): void;

      /**
       * Method: shiftRow
       * Shifty grid work
       *
       * Parameters:
       * prepend - {Boolean} if true, prepend to beginning.
       * if false, then append to end
       * tileSize - {Object} rendered tile size; object with w and h properties
       */
      private shiftRow(
        prepend: boolean,
        tileSize: { w: number; h: number }
      ): void;

      /**
       * Method: shiftColumn
       * Shift grid work in the other dimension
       *
       * Parameters:
       * prepend - {Boolean} if true, prepend to beginning.
       * if false, then append to end
       * tileSize - {Object} rendered tile size; object with w and h properties
       */
      private shiftColumn(
        prepend: boolean,
        tileSize: { w: number; h: number }
      ): void;

      /**
       * Method: removeExcessTiles
       * When the size of the map or the buffer changes, we may need to
       * remove some excess rows and columns.
       *
       * Parameters:
       * rows - {Integer} Maximum number of rows we want our grid to have.
       * columns - {Integer} Maximum number of columns we want our grid to have.
       */
      private removeExcessTiles(rows: number, columns: number): void;

      /**
       * Method: onMapResize
       * For singleTile layers, this will set a new tile size according to the
       * dimensions of the map pane.
       */
      onMapResize(): void;

      /**
       * APIMethod: getTileBounds
       * Returns The tile bounds for a layer given a pixel location.
       *
       * Parameters:
       * viewPortPx - {<OpenLayers.Pixel>} The location in the viewport.
       *
       * Returns:
       * {<OpenLayers.Bounds>} Bounds of the tile at the given pixel location.
       */
      getTileBounds(viewPortPx: Pixel): Bounds;
    }

    export class HTTPRequest extends Layer {
      /**
       * Constant: URL_HASH_FACTOR
       * {Float} Used to hash URL param strings for multi-WMS server selection.
       * Set to the Golden Ratio per Knuth's recommendation.
       */
      static URL_HASH_FACTOR: number;

      /**
       * Property: url
       * {Array(String) or String} This is either an array of url strings or
       * a single url string.
       */
      url: string[];

      /**
       * Property: params
       * {Object} Hashtable of key/value parameters
       */
      params: Object;

      /**
       * APIProperty: reproject
       * *Deprecated*. See http://docs.openlayers.org/library/spherical_mercator.html
       * for information on the replacement for this functionality.
       * {Boolean} Whether layer should reproject itself based on base layer
       * locations. This allows reprojection onto commercial layers.
       * Default is false: Most layers can't reproject, but layers
       * which can create non-square geographic pixels can, like WMS.
       */
      reproject: boolean;

      /**
       * Constructor: OpenLayers.Layer.HTTPRequest
       *
       * Parameters:
       * name - {String}
       * url - {Array(String) or String}
       * params - {Object}
       * options - {Object} Hashtable of extra options to tag onto the layer
       */
      constructor(name: string, url: string, params: Object, options: Object);
      constructor(name: string, url: string[], params: Object, options: Object);

      /**
       * APIMethod: destroy
       */
      destroy(): void;

      /**
       * APIMethod: clone
       *
       * Parameters:
       * obj - {Object}
       *
       * Returns:
       * {<OpenLayers.Layer.HTTPRequest>} An exact clone of this
       * <OpenLayers.Layer.HTTPRequest>
       */
      clone(obj?: Object): HTTPRequest;

      /**
       * APIMethod: setUrl
       *
       * Parameters:
       * newUrl - {String}
       */
      setUrl(newUrl: string): void;

      /**
       * APIMethod: mergeNewParams
       *
       * Parameters:
       * newParams - {Object}
       *
       * Returns:
       * redrawn: {Boolean} whether the layer was actually redrawn.
       */
      mergeNewParams(newParams: Object): boolean;

      /**
       * APIMethod: redraw
       * Redraws the layer. Returns true if the layer was redrawn, false if not.
       *
       * Parameters:
       * force - {Boolean} Force redraw by adding random parameter.
       *
       * Returns:
       * {Boolean} The layer was redrawn.
       */
      redraw(force?: boolean): boolean;

      /**
       * Method: selectUrl
       * selectUrl() implements the standard floating-point multiplicative
       * hash function described by Knuth, and hashes the contents of the
       * given param string into a float between 0 and 1. This float is then
       * scaled to the size of the provided urls array, and used to select
       * a URL.
       *
       * Parameters:
       * paramString - {String}
       * urls - {Array(String)}
       *
       * Returns:
       * {String} An entry from the urls array, deterministically selected based
       * on the paramString.
       */
      private selectUrl(paramString: string, urls: string[]): string;

      /**
       * Method: getFullRequestString
       * Combine url with layer's params and these newParams.
       *
       * does checking on the serverPath variable, allowing for cases when it
       * is supplied with trailing ? or &, as well as cases where not.
       *
       * return in formatted string like this:
       * "server?key1=value1&key2=value2&key3=value3"
       *
       * WARNING: The altUrl parameter is deprecated and will be removed in 3.0.
       *
       * Parameters:
       * newParams - {Object}
       * altUrl - {String} Use this as the url instead of the layer's url
       *
       * Returns:
       * {String}
       */
      getFullRequestString(newParams: Object, altUrl: string): string;
    }

    export class Image extends Layer {
      /**
       * Property: isBaseLayer
       * {Boolean} The layer is a base layer. Default is true. Set this property
       * in the layer options
       */
      isBaseLayer: boolean;

      /**
       * Property: url
       * {String} URL of the image to use
       */
      url: string;

      /**
       * Property: extent
       * {<OpenLayers.Bounds>} The image bounds in map units. This extent will
       * also be used as the default maxExtent for the layer. If you wish
       * to have a maxExtent that is different than the image extent, set the
       * maxExtent property of the options argument (as with any other layer).
       */
      extent: OpenLayers.Bounds;

      /**
       * Property: size
       * {<OpenLayers.Size>} The image size in pixels
       */
      size: OpenLayers.Size;

      /**
       * Property: tile
       * {<OpenLayers.Tile.Image>}
       */
      tile: OpenLayers.Tile.Image;

      /**
       * Property: aspectRatio
       * {Float} The ratio of height/width represented by a single pixel in the
       * graphic
       */
      aspectRatio: number;

      /**
       * Constructor: OpenLayers.Layer.Image
       * Create a new image layer
       *
       * Parameters:
       * name - {String} A name for the layer.
       * url - {String} Relative or absolute path to the image
       * extent - {<OpenLayers.Bounds>} The extent represented by the image
       * size - {<OpenLayers.Size>} The size (in pixels) of the image
       * options - {Object} Hashtable of extra options to tag onto the layer
       */
      constructor(
        name: string,
        url: string,
        extent: OpenLayers.Bounds,
        size: OpenLayers.Size,
        options: any
      );

      /**
       * Method: destroy
       * Destroy this layer
       */
      destroy(): void;

      /**
       * Method: clone
       * Create a clone of this layer
       *
       * Parameters:
       * obj - {Object} An optional layer (is this ever used?)
       *
       * Returns:
       * {<OpenLayers.Layer.Image>} An exact copy of this layer
       */
      clone(obj?: any): OpenLayers.Layer.Image;

      /**
       * APIMethod: setMap
       *
       * Parameters:
       * map - {<OpenLayers.Map>}
       */
      setMap(map: OpenLayers.Map): void;

      /**
       * Method: moveTo
       * Create the tile for the image or resize it for the new resolution
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       * zoomChanged - {Boolean}
       * dragging - {Boolean}
       */
      moveTo(
        bounds: OpenLayers.Bounds,
        zoomChanged: boolean,
        dragging: boolean
      ): void;

      /**
       * Set the tile size based on the map size.
       */
      setTileSize(): void;

      /**
       * Method: addTileMonitoringHooks
       * This function takes a tile as input and adds the appropriate hooks to
       * the tile so that the layer can keep track of the loading tiles.
       *
       * Parameters:
       * tile - {<OpenLayers.Tile>}
       */
      addTileMonitoringHooks(tile: OpenLayers.Tile): void;

      /**
       * Method: removeTileMonitoringHooks
       * This function takes a tile as input and removes the tile hooks
       * that were added in <addTileMonitoringHooks>.
       *
       * Parameters:
       * tile - {<OpenLayers.Tile>}
       */
      removeTileMonitoringHooks(tile: OpenLayers.Tile): void;

      /**
       * APIMethod: setUrl
       *
       * Parameters:
       * newUrl - {String}
       */
      setUrl(newUrl: string): void;

      /**
       * APIMethod: getURL
       * The url we return is always the same (the image itself never changes)
       * so we can ignore the bounds parameter (it will always be the same,
       * anyways)
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       */
      getURL(bounds: OpenLayers.Bounds): string;

      CLASS_NAME: string;
    }
    export class KaMap {
      // TODO
    }
    export class KaMapCache {
      // TODO
    }
    export class MapGuide {
      // TODO
    }
    export class MapServer {
      // TODO
    }
    export class Markers {
      // TODO
    }

    export class OSM extends Layer.XYZ {
      /**
       * The layer name. Defaults to "OpenStreetMap" if the first
       * argument to the constructor is null or undefined.
       */
      name: string;

      /**
       * The tileset URL scheme. Defaults to
       * : http://[a|b|c].tile.openstreetmap.org/${z}/${x}/${y}.png
       * (the official OSM tileset) if the second argument to the constructor
       * is null or undefined. To use another tileset you can have something
       * like this:
       * new OpenLayers.Layer.OSM("OpenCycleMap",
       * ["http://a.tile.opencyclemap.org/cycle/${z}/${x}/${y}.png",
       * "http://b.tile.opencyclemap.org/cycle/${z}/${x}/${y}.png",
       * "http://c.tile.opencyclemap.org/cycle/${z}/${x}/${y}.png"]);
       */
      url: string[];

      /**
       * The layer attribution.
       */
      attribution: string;

      sphericalMercator: boolean;

      wrapDateLine: boolean;

      /**
       * optional configuration options for <OpenLayers.Tile> instances
       * created by this Layer.
       */
      tileOptions: TileOptions;

      constructor();

      constructor(name: string, url: string, options: TileOptions);

      /**
       * Create a clone of this layer
       */
      clone(obj?: Object): Layer.OSM;

      static CLASS_NAME: string;
    }

    export class PointGrid {
      // TODO
    }
    export class PointTrack {
      // TODO
    }
    export class SphericalMercator {
      // TODO
    }
    export class TMS {
      // TODO
    }
    export class Text {
      // TODO
    }
    export class TileCache {
      // TODO
    }
    export class UTFGrid {
      // TODO
    }
    export class Vector {
      // TODO
    }

    export class WMS extends Layer.Grid {
      /**
       * Default is true for WMS layer
       */
      isBaseLayer: boolean;

      /**
       * Should the BBOX commas be encoded? The WMS spec says 'no',
       * but some services want it that way. Default false.
       */
      encodeBBOX: boolean;

      /**
       * If true, the image format will not be automatically switched
       * from image/jpeg to image/png or image/gif when using
       * TRANSPARENT=TRUE. Also isBaseLayer will not changed by the
       * constructor. Default false.
       */
      noMagic: boolean;

      /**
       * Keys in this object are EPSG codes for which the axis order
       * is to be reversed (yx instead of xy, LatLon instead of LonLat), with
       * true as value. This is only relevant for WMS versions >= 1.3.0, and
       * only if yx is not set in <OpenLayers.Projection.defaults> for the
       * used projection.
       */
      yx: Object;

      /**
       * Constructor: OpenLayers.Layer.WMS
       * Create a new WMS layer object
       *
       * Examples:
       *
       * The code below creates a simple WMS layer using the image/jpeg format.
       * (code)
       * var wms = new OpenLayers.Layer.WMS("NASA Global Mosaic",
       * "http://wms.jpl.nasa.gov/wms.cgi",
       * {layers: "modis,global_mosaic"});
       * (end)
       * Note the 3rd argument (params). Properties added to this object will be
       * added to the WMS GetMap requests used for this layer's tiles. The only
       * mandatory parameter is "layers". Other common WMS params include
       * "transparent", "styles" and "format". Note that the "srs" param will
       * always be ignored. Instead, it will be derived from the baseLayer's or
       * map's projection.
       *
       * The code below creates a transparent WMS layer with additional options.
       * (code)
       * var wms = new OpenLayers.Layer.WMS("NASA Global Mosaic",
       * "http://wms.jpl.nasa.gov/wms.cgi",
       * {
       * layers: "modis,global_mosaic",
       * transparent: true
       * }, {
       * opacity: 0.5,
       * singleTile: true
       * });
       * (end)
       * Note that by default, a WMS layer is configured as baseLayer. Setting
       * the "transparent" param to true will apply some magic (see <noMagic>).
       * The default image format changes from image/jpeg to image/png, and the
       * layer is not configured as baseLayer.
       *
       * Parameters:
       * name - {String} A name for the layer
       * url - {String} Base url for the WMS
       * (e.g. http://wms.jpl.nasa.gov/wms.cgi)
       * params - {Object} An object with key/value pairs representing the
       * GetMap query string parameters and parameter values.
       * options - {Object} Hashtable of extra options to tag onto the layer.
       * These options include all properties listed above, plus the ones
       * inherited from superclasses.
       */
      constructor(
        name: string,
        url: string,
        params: WMSGetMapParams,
        options: WMSOptions
      );

      /**
       * Create a clone of this layer
       */
      clone(): Layer.WMS;

      /**
       * Returns true if the axis order is reversed for the WMS version and
       * projection of the layer.
       */
      reverseAxisOrder(): boolean;

      /**
       * Return a GetMap query string for this layer
       */
      getURL(bounds: Bounds): string;

      /**
       * Catch changeParams and uppercase the new params to be merged in
       * before calling changeParams on the super class.
       * Once params have been changed, the tiles will be reloaded with
       * the new parameters.
       */
      mergeNewParams(newParams: Object): boolean;

      /**
       * Combine the layer's url with its params and these newParams.
       *
       * Add the SRS parameter from projection -- this is probably
       * more eloquently done via a setProjection() method, but this
       * works for now and always.
       */
      getFullRequestString(newParams: Object, altUrl: string): string;

      static CLASS_NAME: string;
    }

    export class WMTS {
      // TODO
    }

    export class WorldWind {
      // TODO
    }

    export class XYZ extends Layer.Grid {
      /**
       * APIProperty: isBaseLayer
       * Default is true, as this is designed to be a base tile source.
       */
      isBaseLayer: boolean;

      /**
       * APIProperty: sphericalMercator
       * Whether the tile extents should be set to the defaults for
       * spherical mercator. Useful for things like OpenStreetMap.
       * Default is false, except for the OSM subclass.
       */
      sphericalMercator: boolean;

      /**
       * APIProperty: zoomOffset
       * {Number} If your cache has more zoom levels than you want to provide
       * access to with this layer, supply a zoomOffset. This zoom offset
       * is added to the current map zoom level to determine the level
       * for a requested tile. For example, if you supply a zoomOffset
       * of 3, when the map is at the zoom 0, tiles will be requested from
       * level 3 of your cache. Default is 0 (assumes cache level and map
       * zoom are equivalent). Using <zoomOffset> is an alternative to
       * setting <serverResolutions> if you only want to expose a subset
       * of the server resolutions.
       */
      zoomOffset: number;

      /**
       * APIProperty: serverResolutions
       * {Array} A list of all resolutions available on the server. Only set this
       * property if the map resolutions differ from the server. This
       * property serves two purposes. (a) <serverResolutions> can include
       * resolutions that the server supports and that you don't want to
       * provide with this layer; you can also look at <zoomOffset>, which is
       * an alternative to <serverResolutions> for that specific purpose.
       * (b) The map can work with resolutions that aren't supported by
       * the server, i.e. that aren't in <serverResolutions>. When the
       * map is displayed in such a resolution data for the closest
       * server-supported resolution is loaded and the layer div is
       * stretched as necessary.
       */
      serverResolutions: number[];

      /**
       * Constructor: OpenLayers.Layer.XYZ
       *
       * Parameters:
       * name - {String}
       * url - {String}
       * options - {Object} Hashtable of extra options to tag onto the layer
       */
      constructor(name: string, url: string, options?: any);

      /**
       * APIMethod: clone
       * Create a clone of this layer
       *
       * Parameters:
       * obj - {Object} Is this ever used?
       *
       * Returns:
       * {<OpenLayers.Layer.XYZ>} An exact clone of this OpenLayers.Layer.XYZ
       */
      clone(obj?: any): Layer.XYZ;

      /**
       * Method: getURL
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       *
       * Returns:
       * {String} A string with the layer's url and parameters and also the
       * passed-in bounds and appropriate tile size specified as
       * parameters
       */
      private getURL(bounds: Bounds): string;

      /**
       * Method: getXYZ
       * Calculates x, y and z for the given bounds.
       *
       * Parameters:
       * bounds - {<OpenLayers.Bounds>}
       *
       * Returns:
       * {Object} - an object with x, y and z properties.
       */
      private getXYZ(bounds: Bounds): { x: number; y: number; z: number };

      /* APIMethod: setMap
       * When the layer is added to a map, then we can fetch our origin
       * (if we don't have one.)
       *
       * Parameters:
       * map - {<OpenLayers.Map>}
       */
      setMap(map: Map): void;
    }

    export class Zoomify {
      // TODO
    }

    namespace Google {
      export class v3 {
        // TODO
      }
    }

    namespace Vector {
      export class RootContainer {
        // TODO
      }
    }
  }

  /**
   * This is a class designed to designate a single tile, however
   *     it is explicitly designed to do relatively little. Tiles store
   *     information about themselves -- such as the URL that they are related
   *     to, and their size - but do not add themselves to the layer div
   *     automatically, for example. Create a new tile with the
   *     <OpenLayers.Tile> constructor, or a subclass.
   */
  export class Tile {
    /**
     * An events object that handles all events on the tile.
     *
     * Register a listener for a particular event with the following syntax:
     * ```
     * tile.events.register(type, obj, listener);
     * ```
     *
     * Supported event types:
     * beforedraw - Triggered before the tile is drawn. Used to defer
     *     drawing to an animation queue. To defer drawing, listeners need
     *     to return false, which will abort drawing. The queue handler needs
     *     to call <draw>(true) to actually draw the tile.
     * loadstart - Triggered when tile loading starts.
     * loadend - Triggered when tile loading ends.
     * loaderror - Triggered before the loadend event (i.e. when the tile is
     *     still hidden) if the tile could not be loaded.
     * reload - Triggered when an already loading tile is reloaded.
     * unload - Triggered before a tile is unloaded.
     */
    events: Events<this>;

    /**
     * If set as an option at construction, the eventListeners
     *     object will be registered with <OpenLayers.Events.on>.  Object
     *     structure must be a listeners object as shown in the example for
     *     the events.on method.
     *
     * This options can be set in the `tileOptions` option from
     * <OpenLayers.Layer.Grid>. For example, to be notified of the
     * `loadend` event of each tiles:
     * ```
     * new OpenLayers.Layer.OSM('osm', 'http://tile.openstreetmap.org/${z}/${x}/${y}.png', {
     *     tileOptions: {
     *         eventListeners: {
     *             'loadend': function(evt) {
     *                 // do something on loadend
     *             }
     *         }
     *     }
     * });
     * ```
     */
    eventListeners: object;

    id: string;

    /** layer the tile is attached to */
    layer: Layer;

    /** url of the request. */
    url: string;

    bounds: Bounds;
    size: Size;

    /** Top Left pixel of the tile */
    position: Pixel;

    /** Is the tile loading? */
    isLoading: boolean;

    /**
     * Constructor for a new <OpenLayers.Tile> instance.
     *
     * @param layer layer that the tile will go in.
     * @param position
     * @param bounds
     * @param url
     * @param size
     * @param options
     */
    constructor(layer: Layer, position: Pixel, bounds: Bounds, url: string, size: string, options: object);

    /**
     * Call immediately before destroying if you are listening to tile
     * events, so that counters are properly handled if tile is still
     * loading at destroy-time. Will only fire an event if the tile is
     * still loading.
     */
    unload(): void;

    /** Nullify references to prevent circular references and memory leaks. */
    destroy(): void;

    /**
     * Clear whatever is currently in the tile, then return whether or not
     *     it should actually be re-drawn. This is an example implementation
     *     that can be overridden by subclasses. The minimum thing to do here
     *     is to call <clear> and return the result from <shouldDraw>.
     *
     * @param deferred When drawing was aborted by returning false from a
     *     *beforedraw* listener, the queue manager needs to pass true, so the
     *     tile will not be cleared and immediately be drawn. Otherwise, the
     *     tile will be cleared and a *beforedraw* event will be fired.
     *
     * @returns Whether or not the tile should actually be drawn.
     */
    draw(deferred: boolean): boolean;

    /**
     * Return whether or not the tile should actually be (re-)drawn. The only
     * case where we *wouldn't* want to draw the tile is if the tile is outside
     * its layer's maxExtent
     *
     * @returns Whether or not the tile should actually be drawn.
     */
    shouldDraw(): boolean;

    /**
     * Sets the bounds on this instance
     *
     * @param bounds
     */
    setBounds(bounds: Bounds): void;

    /**
     * Reposition the tile.
     *
     * @param bounds - {<OpenLayers.Bounds>}
     * @param position - {<OpenLayers.Pixel>}
     * @param redraw - {Boolean} Call draw method on tile after moving.
     *     Default is true
     */
    moveTo(bounds: Bounds, position: Pixel, redraw?: boolean): void;

    /**
     * Clear the tile of any bounds/position-related data so that it can
     *     be reused in a new location.
     */
    clear(): void;
  }

  namespace Tile {
    /**
     * Instances of OpenLayers.Tile.Image are used to manage the image tiles
     * used by various layers.
     */
    export class Image extends Tile {
      /**
       * The URL of the image being requested. No default. Filled in by
       * layer.getURL() function. May be modified by loadstart listeners.
       */
      url: string;

      /** The image for this tile. */
      imgDiv: HTMLImageElement;

      /**
       * The image element is appended to the frame.  Any gutter on
       * the image will be hidden behind the frame. If no gutter is set,
       * this will be null.
       */
      frame: HTMLElement;

      /** Attempts to load the image. */
      imageReloadAttempts: number;

      /** True if the png alpha hack needs to be applied on the layer's div. */
      layerAlphaHack: boolean

      /**
       * ID of an request to see if request is still valid. This is a
       * number which increments by 1 for each asynchronous request.
       */
      asyncRequestId: number

      /**
       * Using a data scheme url is not supported by all browsers, but
       * we don't care because we either set it as css backgroundImage, or the
       * image's display style is set to "none" when we use it.
       */
      blankImageUrl: string;

      /**
       * If set, requests that would result in GET urls with more
       * characters than the number provided will be made using form-encoded
       * HTTP POST. It is good practice to avoid urls that are longer than 2048
       * characters.
       *
       * Caution:
       * Older versions of Gecko based browsers (e.g. Firefox < 3.5) and most
       * Opera versions do not fully support this option. On all browsers,
       * transition effects are not supported if POST requests are used.
       */
      maxGetUrlLength: number | null;

      /** A canvas context associated with the tile image. */
      canvasContext: CanvasRenderingContext2D;

      /**
       * The value of the crossorigin keyword to use when loading images. This is
       * only relevant when using <getCanvasContext> for tiles from remote
       * origins and should be set to either 'anonymous' or 'use-credentials'
       * for servers that send Access-Control-Allow-Origin headers with their
       * tiles.
       */
      crossOriginKeyword: string;

      /**
       * Check that a tile should be drawn, and draw it.
       *
       * @returns Was a tile drawn?
       */
      draw(): boolean;

      /**
       * Internal function to actually initialize the image tile,
       *     position it correctly, and set its url.
       */
      renderTile(): void;

      /**
       * Using the properties currently set on the layer, position the tile correctly.
       * This method is used both by the async and non-async versions of the Tile.Image
       * code.
       */
      positionTile(): void;

      /**
       * Remove the tile from the DOM, clear it of any image related data so that
       * it can be reused in a new location.
       */
      clear(): void;

      /**
       * Returns or creates and returns the tile image.
       */
      getImage(): HTMLImageElement;

      /**
       * Creates the content for the frame on the tile.
       */
      initImage(): void;

      /**
       * Sets the source for the tile image
       *
       * @param url or undefined to hide the image
       */
      setImgSrc(url: string): void;

      /**
       * Get the tile's markup.
       *
       * @returns The tile's markup
       */
      getTile(): HTMLElement

      /**
       * Create a backbuffer for this tile. A backbuffer isn't exactly a clone
       * of the tile's markup, because we want to avoid the reloading of the
       * image. So we clone the frame, and steal the image from the tile.
       *
       * @returns The markup, or undefined if the tile has no image
       * or if it's currently loading.
       */
      createBackBuffer(): HTMLElement;

      /**
       * Handler for the image onload event
       */
      onImageLoad(): void;

      /**
       * Handler for the image onerror event
       */
      onImageError(): void;

      /**
       * Returns a canvas context associated with the tile image (with
       * the image drawn on it).
       * Returns undefined if the browser does not support canvas, if
       * the tile has no image or if it's currently loading.
       *
       * The function returns a canvas context instance but the
       * underlying canvas is still available in the 'canvas' property:
       * ```
       * var context = tile.getCanvasContext();
       * if (context) {
       *     var data = context.canvas.toDataURL('image/jpeg');
       * }
       * ```
       */
      getCanvasContext(): CanvasRenderingContext2D | undefined;
    }
  }

  type _MapOptionProperties = (
    | 'size'
    | 'tileSize'
    | 'units'
    | 'resolutions'
    | 'maxResolution'
    | 'minResolution'
    | 'maxScale'
    | 'minScale'
    | 'numZoomLevels'
    | 'theme'
    | 'displayProjection'
    | 'fallThrough'
    | 'panTween'
    | 'eventListeners'
    | 'panMethod'
    | 'panDuration'
    | 'paddingForPopups'
    | 'minPx'
    | 'maxPx'
    | 'layers'
    | 'controls'
  );
  export interface MapOptions extends Partial<Pick<Map, _MapOptionProperties>> {
    /** The element that contains the map (or an id for that element). */
    div?: HTMLElement | string;

    /**
     * Set in the map options to specify the default projection
     * for layers added to this map. When using a projection other than EPSG:4326
     * (CRS:84, Geographic) or EPSG:3857 (EPSG:900913, Web Mercator),
     * also set maxExtent, maxResolution or resolutions. Default is "EPSG:4326".
     * Note that the projection of the map is usually determined
     * by that of the current baseLayer (see <baseLayer> and <getProjectionObject>).
     */
    projection?: Projection | string;

    /**
     * If provided as an array, the array should consist of four values (left,
     * bottom, right, top).
     * The maximum extent for the map.
     * Default depends on projection; if this is one of those defined in
     * OpenLayers.Projection.defaults (EPSG:4326 or web mercator), maxExtent
     * will be set to the value defined there; else, defaults to `null`.
     * To restrict user panning and zooming of the map, use <restrictedExtent> instead.
     * The value for <maxExtent> will change calculations for tile URLs.
     */
    maxExtent?: Bounds | [left: number, bottom: number, right: number, top: number];

    /**
     * If provided as an array, the array should consist of four values (left,
     * bottom, right, top).
     * The minimum extent for the map. Defaults to null.
     */
    minExtent?: Bounds | [left: number, bottom: number, right: number, top: number];

    /**
     * Limit map navigation to this extent where possible.
     * If a non-null restrictedExtent is set, panning will be restricted
     * to the given bounds. In addition, zooming to a resolution that
     * displays more than the restricted extent will center the map
     * on the restricted extent. If you wish to limit the zoom level
     * or resolution, use maxResolution.
     */
    restrictedExtent?: Bounds | [left: number, bottom: number, right: number, top: number];

    /**
     * The default initial center of the map.
     * Only specify if <layers> is provided.
     * Note that if an ArgParser/Permalink control is present,
     * and the querystring contains coordinates, center will be set
     * by that, and this option will be ignored.
     */
    center?: LonLatLike | [lon: number, lat: number];

    /**
     * The initial zoom level for the map. Only specify if <layers> is provided.
     * Note that if an ArgParser/Permalink control is present,
     * and the querystring contains a zoom level, zoom will be set
     * by that, and this option will be ignored.
     */
    zoom?: number;
  }

  export class Map {
    /** Unique identifier for the map */
    id: string;

    /**
     * For a base layer that supports it, allow the map resolution
     * to be set to a value between one of the values in the resolutions
     * array. Default is false.
     */
    fractionalZoom: boolean;

    /** An events object that handles all events on the map */
    events: Events<this>;

    /**
     * Allow the map to function with "overlays" only. Defaults to
     * false. If true, the lowest layer in the draw order will act as
     * the base layer. In addition, if set to true, all layers will
     * have isBaseLayer set to false when they are added to the map.
     */
    allOverlays: boolean;

    /** The element that contains the map (or an id for that element). */
    div: HTMLDivElement;

    /** The map is currently being dragged. */
    dragging: boolean;

    /** Size of the main div (this.div) */
    size: Size;

    /** The element that represents the map viewport */
    viewPortDiv: HTMLDivElement;

    /** The lonlat at which the later container was re-initialized (on-zoom) */
    layerContainerOrigin: LonLat;

    /** The element that contains the layers. */
    layerContainerDiv: HTMLDivElement;

    /** Ordered list of layers in the map */
    layers: Layer[];

    /** List of controls associated with the map. */
    controls: Control[];

    /** List of popups associated with the map */
    popups: Popup[];

    /**
     * The currently selected base layer. This determines min/max zoom level,
     * projection, etc.
     */
    baseLayer: Layer | null;

    /** The current center of the map */
    center: LonLat;

    /** The resolution of the map. */
    resolution: number;

    /** The current zoom level of the map */
    zoom: number;

    /** The ratio of the current extent within which panning will tween. */
    panRatio: number;

    /** The options object passed to the class constructor. Read-only. */
    readonly options: MapOptions;

    /** Set in the map options to override the default tile size for this map. */
    tileSize: Size;

    /**
     * Set in the map options to specify the default projection
     * for layers added to this map. When using a projection other than EPSG:4326
     * (CRS:84, Geographic) or EPSG:3857 (EPSG:900913, Web Mercator),
     * also set maxExtent, maxResolution or resolutions. Default is "EPSG:4326".
     * Note that the projection of the map is usually determined
     * by that of the current baseLayer (see <baseLayer> and <getProjectionObject>).
     */
    projection: string;

    /**
     * The map units. Possible values are 'degrees' (or 'dd'), 'm',
     * 'ft', 'km', 'mi', 'inches'. Normally taken from the projection.
     * Only required if both map and layers do not define a projection,
     * or if they define a projection which does not define units
     */
    units: string;

    /**
     * A list of map resolutions (map units per pixel) in
     * descending order. If this is not set in the layer constructor, it
     * will be set based on other resolution related properties
     * (maxExtent, maxResolution, maxScale, etc.).
     */
    resolutions: number[];

    /**
     * Required if you are not displaying the whole world on a tile
     * with the size specified in <tileSize>.
     */
    maxResolution: number;
    minResolution: number;
    maxScale: number;
    minScale: number;

    /**
     * If provided as an array, the array should consist of four values (left,
     * bottom, right, top).
     * The maximum extent for the map.
     * Default depends on projection; if this is one of those defined in
     * OpenLayers.Projection.defaults (EPSG:4326 or web mercator), maxExtent
     * will be set to the value defined there; else, defaults to `null`.
     * To restrict user panning and zooming of the map, use <restrictedExtent> instead.
     * The value for <maxExtent> will change calculations for tile URLs.
     */
    maxExtent: Bounds;

    /**
     * If provided as an array, the array should consist of four values (left,
     * bottom, right, top).
     * The minimum extent for the map. Defaults to null.
     */
    minExtent: Bounds;

    /**
     * Limit map navigation to this extent where possible.
     * If a non-null restrictedExtent is set, panning will be restricted
     * to the given bounds. In addition, zooming to a resolution that
     * displays more than the restricted extent will center the map
     * on the restricted extent. If you wish to limit the zoom level
     * or resolution, use maxResolution.
     */
    restrictedExtent: Bounds;

    /**
     * Number of zoom levels for the map. Defaults to 16. Set a
     * different value in the map options if needed.
     */
    numZoomLevels: number;

    /**
     * Relative path to a CSS file from which to load theme styles.
     * Specify null in the map options (e.g. {theme: null}) if you
     * want to get cascading style declarations - by putting links to
     * stylesheets or style declarations directly in your page.
     */
    theme: string | null;

    /**
     * Requires proj4js support for projections other
     * than EPSG:4326 or EPSG:900913/EPSG:3857. Projection used by
     * several controls to display data to user. If this property is set,
     * it will be set on any control which has a null displayProjection
     * property at the time the control is added to the map.
     */
    displayProjection: Projection;

    /**
     * Should OpenLayers allow events on the map to fall through to
     * other elements on the page, or should it swallow them? (#457)
     * Default is to swallow.
     */
    fallThrough: boolean;

    /** Animated panning tween object, see panTo() */
    panTween: Tween;

    /**
     * If set as an option at construction, the eventListeners
     * object will be registered with <OpenLayers.Events.on>. Object
     * structure must be a listeners object as shown in the example for
     * the events.on method.
     */
    eventListeners: object;

    /**
     * The Easing function to be used for tweening. Default is
     * OpenLayers.Easing.Expo.easeOut. Setting this to 'null' turns off
     * animated panning.
     */
    panMethod(): void;

    /**
     * The number of steps to be passed to the OpenLayers.Tween.start() method
     * when the map is panned.
     * Default is 50.
     */
    panDuration: number;

    /**
     * Outside margin of the popup. Used to prevent
     * the popup from getting too close to the map border.
     */
    paddingForPopups: Bounds;

    /**
     * An object with a 'x' and 'y' values that is the lower
     * left of maxExtent in viewport pixel space.
     * Used to verify in moveByPx that the new location we're moving to is valid.
     * It is also used in the getLonLatFromViewPortPx function of Layer.
     */
    minPx: PointLike;

    /**
     * An object with a 'x' and 'y' values that is the top
     * right of maxExtent in viewport pixel space.
     * Used to verify in moveByPx that the new location we're moving to is valid.
     */
    maxPx: PointLike;

    /**
     * Constructor for a new OpenLayers.Map instance. There are two possible
     * ways to call the map constructor. See the examples below.
     *
     * @param div The element or id of an element in your page
     * that will contain the map. May be omitted if the <div> option is
     * provided or if you intend to call the <render> method later.
     * @param options Optional object with properties to tag onto the map.
     *
     * @example
     * // create a map with default options in an element with the id "map1"
     * var map = new OpenLayers.Map("map1");
     *
     * // create a map with non-default options in an element with id "map2"
     * var options = {
     *   projection: "EPSG:3857",
     *   maxExtent: new OpenLayers.Bounds(-200000, -200000, 200000, 200000),
     *   center: new OpenLayers.LonLat(-12356463.476333, 5621521.4854095)
     * };
     * var map = new OpenLayers.Map("map2", options);
     *
     * // map with non-default options - same as above but with a single argument,
     * // a restricted extent, and using arrays for bounds and center
     * var map = new OpenLayers.Map({
     *   div: "map_id",
     *   projection: "EPSG:3857",
     *   maxExtent: [-18924313.432222, -15538711.094146, 18924313.432222, 15538711.094146],
     *   restrictedExtent: [-13358338.893333, -9608371.5085962, 13358338.893333, 9608371.5085962],
     *   center: [-12356463.476333, 5621521.4854095]
     * });
     *
     * // create a map without a reference to a container - call render later
     * var map = new OpenLayers.Map({
     *   projection: "EPSG:3857",
     *   maxExtent: new OpenLayers.Bounds(-200000, -200000, 200000, 200000)
     * });
     */
    constructor(id?: HTMLElement | string | MapOptions, options?: MapOptions);

    /**
     * APIMethod: getViewport
     * Get the DOMElement representing the view port.
     *
     * Returns:
     * {DOMElement}
     */
    getViewport(): HTMLElement;

    /**
     * APIMethod: render
     * Render the map to a specified container.
     *
     * Parameters:
     * div - {String|DOMElement} The container that the map should be rendered
     * to. If different than the current container, the map viewport
     * will be moved from the current to the new container.
     */
    render(div: HTMLElement | string): void;

    /**
     * Method: unloadDestroy
     * Function that is called to destroy the map on page unload. stored here
     * so that if map is manually destroyed, we can unregister this.
     */
    private unloadDestroy(): void;

    /**
     * Method: updateSizeDestroy
     * When the map is destroyed, we need to stop listening to updateSize
     * events: this method stores the function we need to unregister in
     * non-IE browsers.
     */
    private updateSizeDestroy(): void;

    /**
     * APIMethod: destroy
     * Destroy this map.
     * Note that if you are using an application which removes a container
     * of the map from the DOM, you need to ensure that you destroy the
     * map *before* this happens; otherwise, the page unload handler
     * will fail because the DOM elements that map.destroy() wants
     * to clean up will be gone. (See
     * http://trac.osgeo.org/openlayers/ticket/2277 for more information).
     * This will apply to GeoExt and also to other applications which
     * modify the DOM of the container of the OpenLayers Map.
     */
    destroy(): void;

    /**
     * Change the map options
     *
     * @param options Hashtable of options to tag to the map
     */
    setOptions(options: MapOptions): void;

    /** Get the tile size for the map */
    getTileSize(): Size;

    /**
     * Get a list of objects given a property and a match item.
     *
     * @param array A property on the map whose value is an array.
     * @param property A property on each item of the given array.
     * @param match A string to match. Can also be a regular
     * expression literal or object. In addition, it can be any object
     * with a method named test. For regular expressions or other, if
     * match.test(map[array][i][property]) evaluates to true, the item will
     * be included in the array returned. If no items are found, an empty
     * array is returned.
     * @returns An array of items where the given property matches the given
     * criteria.
     */
    getBy(array: string, property: string, match: string | object): any[];

    /**
     * Get a list of layers with properties matching the given criteria.
     *
     * @param property A layer property to be matched.
     * @param match A string to match. Can also be a regular
     * expression literal or object. In addition, it can be any object
     * with a method named test. For regular expressions or other, if
     * match.test(layer[property]) evaluates to true, the layer will be
     * included in the array returned. If no layers are found, an empty
     * array is returned.
     * @returns A list of layers matching the given criteria.
     * An empty array is returned if no matches are found.
     */
    getLayersBy(property: string, match: string | object): Layer[];

    /**
     * Get a list of layers with names matching the given name.
     *
     * @param match A layer name. The name can also be a regular
     * expression literal or object. In addition, it can be any object
     * with a method named test. For regular expressions or other, if
     * name.test(layer.name) evaluates to true, the layer will be included
     * in the list of layers returned. If no layers are found, an empty
     * array is returned.
     *
     * @returns A list of layers matching the given name.
     * An empty array is returned if no matches are found.
     */
    getLayersByName(match: string | object): Layer[];

    /**
     * Get a list of layers of a given class (CLASS_NAME).
     *
     * @param match A layer class name. The match can also be a
     * regular expression literal or object. In addition, it can be any
     * object with a method named test. For regular expressions or other,
     * if type.test(layer.CLASS_NAME) evaluates to true, the layer will
     * be included in the list of layers returned. If no layers are
     * found, an empty array is returned.
     *
     * @returns A list of layers matching the given class.
     * An empty array is returned if no matches are found.
     */
    getLayersByClass(match: string | object): Layer[];

    /**
     * Get a list of controls with properties matching the given criteria.
     *
     * @param property A control property to be matched.
     * @param match A string to match. Can also be a regular
     * expression literal or object. In addition, it can be any object
     * with a method named test. For regular expressions or other, if
     * match.test(layer[property]) evaluates to true, the layer will be
     * included in the array returned. If no layers are found, an empty
     * array is returned.
     *
     * @returns A list of controls matching the given
     * criteria. An empty array is returned if no matches are found.
     */
    getControlsBy(property: string, match: string | object): Control[];

    /**
     * Get a list of controls of a given class (CLASS_NAME).
     *
     * @param match A control class name. The match can also be a
     * regular expression literal or object. In addition, it can be any
     * object with a method named test. For regular expressions or other,
     * if type.test(control.CLASS_NAME) evaluates to true, the control will
     * be included in the list of controls returned. If no controls are
     * found, an empty array is returned.
     *
     * @returns A list of controls matching the given class.
     * An empty array is returned if no matches are found.
     */
    getControlsByClass(match: string | object): Control[];

    /**
     * Get a layer based on its id
     *
     * @param id A layer id
     *
     * @returns The Layer with the corresponding id from the map's
     * layer collection, or null if not found.
     */
    getLayer(id: string): Layer | null;

    /**
     * @param layer
     * @param zIdx
     */
    private setLayerZIndex(layer: Layer, zIdx: number): void;

    /** Reset each layer's z-index based on layer's array index */
    private resetLayersZIndex(): void;

    /**
     * @param layer
     * @returns `true` if the layer has been added to the map.
     */
    addLayer(layer: Layer): boolean;

    /**
     * @param layers
     */
    addLayers(layers: Layer[]): void;

    /**
     * Removes a layer from the map by removing its visual element (the
     * layer.div property), then removing it from the map's internal list
     * of layers, setting the layer's map property to null.
     *
     * a "removelayer" event is triggered.
     *
     * very worthy of mention is that simply removing a layer from a map
     * will not cause the removal of any popups which may have been created
     * by the layer. this is due to the fact that it was decided at some
     * point that popups would not belong to layers. thus there is no way
     * for us to know here to which layer the popup belongs.
     *
     * A simple solution to this is simply to call destroy() on the layer.
     * the default OpenLayers.Layer class's destroy() function
     * automatically takes care to remove itself from whatever map it has
     * been attached to.
     *
     * The correct solution is for the layer itself to register an
     * event-handler on "removelayer" and when it is called, if it
     * recognizes itself as the layer being removed, then it cycles through
     * its own personal list of popups, removing them from the map.
     *
     * @param layer
     * @param setNewBaseLayer Default is true
     */
    removeLayer(layer: Layer, setNewBaseLayer?: boolean): void;

    /**
     * @returns The number of layers attached to the map.
     */
    getNumLayers(): number;

    /**
     * @param layer
     * @returns The current (zero-based) index of the given layer in the map's
     * layer stack. Returns -1 if the layer isn't on the map.
     */
    getLayerIndex(layer: Layer): number;

    /**
     * Move the given layer to the specified (zero-based) index in the layer
     * list, changing its z-index in the map display. Use
     * map.getLayerIndex() to find out the current index of a layer. Note
     * that this cannot (or at least should not) be effectively used to
     * raise base layers above overlays.
     *
     * @param layer
     * @param idx
     */
    setLayerIndex(layer: Layer, idx: number): void;

    /**
     * Change the index of the given layer by delta. If delta is positive,
     * the layer is moved up the map's layer stack; if delta is negative,
     * the layer is moved down. Again, note that this cannot (or at least
     * should not) be effectively used to raise base layers above overlays.
     *
     * @param layer
     * @param delta
     */
    raiseLayer(layer: Layer, delta: number): void;

    /**
     * Allows user to specify one of the currently-loaded layers as the Map's
     * new base layer.
     *
     * @param newBaseLayer
     */
    setBaseLayer(newBaseLayer: Layer): void;

    /**
     * Add the passed over control to the map. Optionally
     * position the control at the given pixel.
     *
     * @param control
     * @param px
     */
    addControl(control: Control, px: Pixel): void;

    /**
     * Add all of the passed over controls to the map.
     * You can pass over an optional second array
     * with pixel-objects to position the controls.
     * The indices of the two arrays should match and
     * you can add null as pixel for those controls
     * you want to be autopositioned.
     *
     * @param controls
     * @param pixels
     */
    addControls(controls: Control[], pixels: Pixel[]): void;

    /**
     * @param control
     * @param px
     */
    private addControlToMap(control: Control, px: Pixel): void;

    /**
     * @param id ID of the control to return.
     * @returns The control from the map's list of controls
     * which has a matching 'id'. If none found,
     * returns null.
     */
    getControl(id: string): Control | null;

    /**
     * Remove a control from the map. Removes the control both from the map
     * object's internal array of controls, as well as from the map's
     * viewPort (assuming the control was not added outsideViewport)
     *
     * @param control The control to remove.
     */
    removeControl(control: Control): void;

    /**
     * @param popup
     * @param exclusive If true, closes all other popups first
     */
    addPopup(popup: Popup, exclusive: boolean): void;

    /**
     * @param popup
     */
    removePopup(popup: Popup): void;

    /**
     * @returns An <OpenLayers.Size> object that represents the
     * size, in pixels, of the div into which OpenLayers
     * has been loaded.
     * Note - A clone() of this locally cached variable is
     * returned, so as not to allow users to modify it.
     */
    getSize(): Size;

    /**
     * This function should be called by any external code which dynamically
     * changes the size of the map div (because mozilla wont let us catch
     * the "onresize" for an element)
     */
    updateSize(): void;

    /**
     * @returns A new <OpenLayers.Size> object with the dimensions of the map div
     */
    private getCurrentSize(): Size;

    /**
     * @param center Default is this.getCenter()
     * @param resolution Default is this.getResolution()
     * @returns A bounds based on resolution, center, and current mapsize.
     */
    calculateBounds(center?: LonLat, resolution?: number): Bounds;

    getCenter(): LonLat | null;
    getCachedCenter(): LonLat | null;
    getZoom(): number;

    /**
     * Allows user to pan by a value of screen pixels
     *
     * @param dx
     * @param dy
     * @param options Options to configure panning:
     * @param options.animate Use panTo instead of setCenter. Default is true.
     * @param options.dragging Call setCenter with dragging true. Default is false.
     */
    pan(
      dx: number,
      dy: number,
      options?: {
        animate?: boolean;
        dragging?: boolean;
      }
    ): void;

    /**
     * Allows user to pan to a new lonlat
     * If the new lonlat is in the current extent the map will slide smoothly
     *
     * @param lonlat
     */
    panTo(lonlat: LonLat): void;

    /**
     * Set the map center (and optionally, the zoom level).
     *
     * @param lonlat The new center location.
     * If provided as array, the first value is the x coordinate,
     * and the 2nd value is the y coordinate.
     * @param zoom Optional zoom level.
     * @param dragging Specifies whether or not to trigger
     * movestart/end events
     * @param forceZoomChange Specifies whether or not to trigger zoom
     * change events (needed on baseLayer change)
     */
    setCenter(
      lonlat: LonLat | [lon: number, lat: number] | null | undefined,
      zoom?: number,
      dragging?: boolean,
      forceZoomChange?: boolean
    ): void;

    /**
     * Drag the map by pixels
     *
     * @param dx
     * @param dy
     */
    moveByPx(dx: number, dy: number): void;

    /**
     * @param zoom The zoom level to adjust
     * @returns Adjusted zoom level that shows a map not wider than its
     * <baseLayer>'s maxExtent.
     */
    adjustZoom(zoom: number | null | undefined): number;

    /**
     * @param lonlat
     * @param zoom
     * @param options
     */
    moveTo(
      lonlat: LonLat | [lon: number, lat: number] | null | undefined,
      zoom?: number | string | null | undefined,
      options?: {dragging?: boolean, forceZoomChange?: boolean} | undefined
    ): void;

    /**
     * This function takes care to recenter the layerContainerDiv.
     *
     * @param lonlat
     */
    private centerLayerContainer(lonlat: LonLat): void;

    /**
     * @param zoomLevel
     * @returns Whether or not the zoom level passed in is non-null and
     *           within the min/max range of zoom levels.
     */
    isValidZoomLevel(zoomLevel: number | null | undefined): boolean;

    /**
     * @param lonlat
     * @returns Whether or not the lonlat passed in is non-null and within
     *           the maxExtent bounds
     */
    isValidLonLat(lonlat: LonLat | null | undefined): boolean;

    /**
     * This method returns a string representing the projection. In
     * the case of projection support, this will be the srsCode which
     * is loaded -- otherwise it will simply be the string value that
     * was passed to the projection at startup.
     *
     * @returns The Projection string from the base layer or null.
     */
    getProjection(): string | null;

    /**
     * Returns the projection object from the baselayer.
     *
     * @returns The Projection of the base layer.
     */
    getProjectionObject(): Projection | null;

    /**
     * @returns The Map's Maximum Resolution
     */
    getMaxResolution(): string | null;

    /**
     * @param options
     * @param options.restricted If true, returns restricted extent (if it is
     * available.)
     * @returns The maxExtent property as set on the current
     * baselayer, unless the 'restricted' option is set, in which case
     * the 'restrictedExtent' option from the map is returned (if it
     * is set).
     */
    getMaxExtent(options?: { restricted: boolean }): Bounds | null;

    /**
     * @returns The total number of zoom levels that can be displayed by the
     * current baseLayer.
     */
    getNumZoomLevels(): number | null;

    /**
     * @returns A Bounds object which represents the lon/lat
     * bounds of the current viewPort.
     * If no baselayer is set, returns null.
     */
    getExtent(): Bounds | null;

    /**
     * @returns The current resolution of the map.
     * If no baselayer is set, returns null.
     */
    getResolution(): number | null;

    /**
     * @returns The current units of the map.
     * If no baselayer is set, returns null.
     */
    getUnits(): string | null;

    /**
     * @returns The current scale denominator of the map.
     * If no baselayer is set, returns null.
     */
    getScale(): number | null;

    /**
     * @param bounds
     * @param closest Find the zoom level that most closely fits the
     * specified bounds. Note that this may result in a zoom that does
     * not exactly contain the entire extent.
     * Default is false.
     * @returns A suitable zoom level for the specified bounds.
     * If no baselayer is set, returns null.
     */
    getZoomForExtent(bounds: Bounds, closest?: boolean): number | null;

    /**
     * @param zoom
     * @returns A suitable resolution for the specified zoom. If no baselayer
     * is set, returns null.
     */
    getResolutionForZoom(zoom: number): number | null;

    /**
     * @param resolution
     * @param closest Find the zoom level that corresponds to the absolute
     * closest resolution, which may result in a zoom whose corresponding
     * resolution is actually smaller than we would have desired (if this
     * is being called from a getZoomForExtent() call, then this means that
     * the returned zoom index might not actually contain the entire
     * extent specified... but it'll be close).
     * Default is false.
     * @returns A suitable zoom level for the specified resolution.
     * If no baselayer is set, returns null.
     */
    getZoomForResolution(resolution: number, closest?: boolean): number | null;

    /**
     * Zoom to a specific zoom level. Zooming will be animated unless the map
     * is configured with {zoomMethod: null}. To zoom without animation, use
     * <setCenter> without a lonlat argument.
     *
     * @param zoom
     */
    zoomTo(zoom: number): void;
    zoomIn(): void;
    zoomOut(): void;

    /**
     * Zoom to the passed in bounds, recenter
     *
     * @param bounds If provided as an array, the array
     * should consist of four values (left, bottom, right, top).
     * @param closest Find the zoom level that most closely fits the
     * specified bounds. Note that this may result in a zoom that does
     * not exactly contain the entire extent.
     * Default is false.
     */
    zoomToExtent(bounds: Bounds | [left: number, bottom: number, right: number, top: number], closest?: boolean): void;

    /**
     * Zoom to the full extent and recenter.
     *
     * @param options
     * @param options.restricted True to zoom to restricted extent if it is
     * set. Defaults to true.
     */
    zoomToMaxExtent(options?: { restricted: boolean }): void;

    /**
     * Zoom to a specified scale
     *
     * @param scale
     * @param closest Find the zoom level that most closely fits the
     * specified scale. Note that this may result in a zoom that does
     * not exactly contain the entire extent.
     * Default is false.
     */
    zoomToScale(scale: number, closest?: boolean): void;

    /**
     * @param viewPortPx An OpenLayers.Pixel or an object with a 'x' and 'y'
     * properties.
     * @returns An OpenLayers.LonLat which is the passed-in view port
     * <OpenLayers.Pixel>, translated into lon/lat by the current base layer.
     */
    getLonLatFromViewPortPx(viewPortPx: PixelLike | null): LonLat | null;

    /**
     * @param lonlat
     * @returns An OpenLayers.Pixel which is the passed-in <OpenLayers.LonLat>,
     * translated into view port pixels by the current base layer.
     */
    getViewPortPxFromLonLat(lonlat: LonLat | null): Pixel | null;

    /**
     * @param px An OpenLayers.Pixel or an object with
     * a 'x' and 'y' properties.
     * @returns An OpenLayers.LonLat corresponding to the given
     * OpenLayers.Pixel, translated into lon/lat by the
     * current base layer
     */
    getLonLatFromPixel(px: PixelLike | null): LonLat | null;

    /**
     * Returns a pixel location given a map location. The map location is
     * translated to an integer pixel location (in viewport pixel
     * coordinates) by the current base layer.
     *
     * @param lonlat A map location.
     * @returns An OpenLayers.Pixel corresponding to the
     * <OpenLayers.LonLat> translated into view port pixels by the current
     * base layer.
     */
    getPixelFromLonLat(lonlat: LonLat | null): Pixel | null;

    /**
     * @param layerPx
     * @returns Layer Pixel translated into ViewPort Pixel coordinates
     */
    getViewPortPxFromLayerPx(layerPx: Pixel | null): Pixel | null;

    /**
     * @param viewPortPx
     * @returns ViewPort Pixel translated into Layer Pixel coordinates
     */
    getLayerPxFromViewPortPx(viewPortPx: Pixel | null): Pixel | null;

    /**
     * @param lonlat
     * @returns An OpenLayers.Pixel which is the passed-in <OpenLayers.LonLat>,
     * translated into layer pixels by the current base layer
     */
    getLayerPxFromLonLat(lonlat: LonLat | null): Pixel | null;

    /**
     * @param px The pixel to get the geodesic length for. If not provided, the
     * center pixel of the map viewport will be used.
     * @returns The geodesic size of the pixel in kilometers.
     */
    getGeodesicPixelSize(px?: Pixel | null): Size;

    /** 256 Default tile width (unless otherwise specified) */
    static TILE_WIDTH: number;

    /** 256 Default tile height (unless otherwise specified) */
    static TILE_HEIGHT: number;
  }

  export class Control {
    // TODO
  }
  export class Popup {
    // TODO
  }
  export class Tween {
    // TODO
  }
}
