/*
 * From: https://github.com/simile-widgets/ancient-simile-widgets
 *
 *  (c) Copyright The SIMILE Project 2006. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import {SimileAjax} from './simile/init.js';
import {Timeline} from './init.js';

/*==================================================
 *  Gregorian Ether Painter
 *==================================================
 */

Timeline.GregorianEtherPainter = function(params) {
  this._params = params;
  this._theme = params.theme;
  this._unit = params.unit;
  this._multiple = ('multiple' in params) ? params.multiple : 1;
};

Timeline.GregorianEtherPainter.prototype.initialize = function(band, timeline) {
  this._band = band;
  this._timeline = timeline;

  this._backgroundLayer = band.createLayerDiv(0);
  this._backgroundLayer.setAttribute('name', 'ether-background'); // for debugging
  this._backgroundLayer.className = 'timeline-ether-bg';
  //  this._backgroundLayer.style.background = this._theme.ether.backgroundColors[band.getIndex()];


  this._markerLayer = null;
  this._lineLayer = null;

  var align = ('align' in this._params && this._params.align != null) ? this._params.align
        : this._theme.ether.interval.marker[timeline.isHorizontal() ? 'hAlign' : 'vAlign'];
  var showLine = ('showLine' in this._params) ? this._params.showLine
        : this._theme.ether.interval.line.show;

  this._intervalMarkerLayout = new Timeline.EtherIntervalMarkerLayout(
    this._timeline, this._band, this._theme, align, showLine);

  this._highlight = new Timeline.EtherHighlight(
    this._timeline, this._band, this._theme, this._backgroundLayer);
};

Timeline.GregorianEtherPainter.prototype.setHighlight = function(startDate, endDate) {
  this._highlight.position(startDate, endDate);
};

Timeline.GregorianEtherPainter.prototype.paint = function() {
  if (this._markerLayer) {
    this._band.removeLayerDiv(this._markerLayer);
  }
  this._markerLayer = this._band.createLayerDiv(100);
  this._markerLayer.setAttribute('name', 'ether-markers'); // for debugging
  this._markerLayer.style.display = 'none';

  if (this._lineLayer) {
    this._band.removeLayerDiv(this._lineLayer);
  }
  this._lineLayer = this._band.createLayerDiv(1);
  this._lineLayer.setAttribute('name', 'ether-lines'); // for debugging
  this._lineLayer.style.display = 'none';

  var minDate = this._band.getMinDate();
  var maxDate = this._band.getMaxDate();

  var timeZone = this._band.getTimeZone();
  var labeller = this._band.getLabeller();

  SimileAjax.DateTime.roundDownToInterval(minDate, this._unit, timeZone, this._multiple, this._theme.firstDayOfWeek);

  while (minDate.getTime() < maxDate.getTime()) {
    this._intervalMarkerLayout.createIntervalMarker(
      minDate, labeller, this._unit, this._markerLayer, this._lineLayer);

    for (var i = 0; i < this._multiple; i++) {
      SimileAjax.DateTime.incrementByInterval(minDate, this._unit);
    }
    if (this._unit === SimileAjax.DateTime.DAY && this._multiple > 1) {
      var nextMonth = new Date(minDate.getTime());
      SimileAjax.DateTime.roundDownToInterval(nextMonth, SimileAjax.DateTime.MONTH, timeZone, 1, this._theme.firstDayOfWeek);
      nextMonth.setUTCMonth(nextMonth.getUTCMonth() + 1);

      var nextIncrement = this._multiple * SimileAjax.DateTime.gregorianUnitLengths[this._unit];

      // Go to start of next month if next step will cross it
      if (
        nextMonth.getTime() < maxDate.getTime() &&
        nextMonth.getTime() < minDate.getTime() + nextIncrement
      ) {
        if (nextMonth.getTime() > minDate.getTime() + 0.75 * nextIncrement) {
          this._intervalMarkerLayout.createIntervalMarker(
            minDate, labeller, this._unit, this._markerLayer, this._lineLayer);
        }

        minDate.setTime(nextMonth.getTime());
      }
    }
  }
  this._markerLayer.style.display = 'block';
  this._lineLayer.style.display = 'block';
};

Timeline.GregorianEtherPainter.prototype.softPaint = function() {
};

Timeline.GregorianEtherPainter.prototype.zoom = function(netIntervalChange) {
  if (netIntervalChange !== 0) {
    this._unit += netIntervalChange;
  }
};

/*==================================================
 *  Ether Interval Marker Layout
 *==================================================
 */

Timeline.EtherIntervalMarkerLayout = function(timeline, band, theme, align, showLine) {
  var horizontal = timeline.isHorizontal();
  if (horizontal) {
    if (align === 'Top') {
      this.positionDiv = function(div, offset) {
        div.style.left = offset + 'px';
        div.style.top = '0px';
      };
    } else {
      this.positionDiv = function(div, offset) {
        div.style.left = offset + 'px';
        div.style.bottom = '0px';
      };
    }
  } else if (align === 'Left') {
    this.positionDiv = function(div, offset) {
      div.style.top = offset + 'px';
      div.style.left = '0px';
    };
  } else {
    this.positionDiv = function(div, offset) {
      div.style.top = offset + 'px';
      div.style.right = '0px';
    };
  }

  var lineTheme = theme.ether.interval.line;
  var weekendTheme = theme.ether.interval.weekend;

  var day = SimileAjax.DateTime.gregorianUnitLengths[SimileAjax.DateTime.DAY];

  this.createIntervalMarker = function(date, labeller, unit, markerDiv, lineDiv) {
    var offset = Math.round(band.dateToPixelOffset(date));

    if (showLine && unit !== SimileAjax.DateTime.WEEK) {
      var divLine = timeline.getDocument().createElement('div');
      divLine.className = 'timeline-ether-lines';

      if (lineTheme.opacity < 100) {
        SimileAjax.Graphics.setOpacity(divLine, lineTheme.opacity);
      }

      if (horizontal) {
        //divLine.className += " timeline-ether-lines-vertical";
        divLine.style.left = offset + 'px';
      } else {
        //divLine.className += " timeline-ether-lines-horizontal";
        divLine.style.top = offset + 'px';
      }
      lineDiv.appendChild(divLine);
    }
    if (unit === SimileAjax.DateTime.WEEK) {
      var firstDayOfWeek = theme.firstDayOfWeek;

      var saturday = new Date(date.getTime() + (6 - firstDayOfWeek - 7) * day);
      var monday = new Date(saturday.getTime() + 2 * day);

      var saturdayPixel = Math.round(band.dateToPixelOffset(saturday));
      var mondayPixel = Math.round(band.dateToPixelOffset(monday));
      var length = Math.max(1, mondayPixel - saturdayPixel);

      var divWeekend = timeline.getDocument().createElement('div');
      divWeekend.className = 'timeline-ether-weekends';

      if (weekendTheme.opacity < 100) {
        SimileAjax.Graphics.setOpacity(divWeekend, weekendTheme.opacity);
      }

      if (horizontal) {
        divWeekend.style.left = saturdayPixel + 'px';
        divWeekend.style.width = length + 'px';
      } else {
        divWeekend.style.top = saturdayPixel + 'px';
        divWeekend.style.height = length + 'px';
      }
      lineDiv.appendChild(divWeekend);
    }

    var label = labeller.labelInterval(date, unit);

    var div = timeline.getDocument().createElement('div');
    div.innerHTML = label.text;

    div.className = 'timeline-date-label';
    if (label.emphasized) {
      div.className += ' timeline-date-label-em';
    }

    this.positionDiv(div, offset);
    markerDiv.appendChild(div);

    return div;
  };
};

/*==================================================
 *  Ether Highlight Layout
 *==================================================
 */

Timeline.EtherHighlight = function(timeline, band, theme, backgroundLayer) {
  var horizontal = timeline.isHorizontal();

  this._highlightDiv = null;
  this._createHighlightDiv = function() {
    if (this._highlightDiv == null) {
      this._highlightDiv = timeline.getDocument().createElement('div');
      this._highlightDiv.setAttribute('name', 'ether-highlight'); // for debugging
      this._highlightDiv.className = 'timeline-ether-highlight';

      var opacity = theme.ether.highlightOpacity;
      if (opacity < 100) {
        SimileAjax.Graphics.setOpacity(this._highlightDiv, opacity);
      }

      backgroundLayer.appendChild(this._highlightDiv);
    }
  };

  this.position = function(startDate, endDate) {
    this._createHighlightDiv();

    var startPixel = Math.round(band.dateToPixelOffset(startDate));
    var endPixel = Math.round(band.dateToPixelOffset(endDate));
    var length = Math.max(endPixel - startPixel, 3);
    if (horizontal) {
      this._highlightDiv.style.left = startPixel + 'px';
      this._highlightDiv.style.width = length + 'px';
      this._highlightDiv.style.height = (band.getViewWidth() - 4) + 'px';
    } else {
      this._highlightDiv.style.top = startPixel + 'px';
      this._highlightDiv.style.height = length + 'px';
      this._highlightDiv.style.width = (band.getViewWidth() - 4) + 'px';
    }
  };
};
