/*
 * From: https://github.com/simile-widgets/ancient-simile-widgets
 *
 *  (c) Copyright The SIMILE Project 2006. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import {SimileAjax} from './simile/init.js';
import {Timeline} from './init.js';

/*==================================================
 *  Gregorian Date Labeller
 *==================================================
 */

Timeline.GregorianDateLabeller = function(locale, timeZone) {
  this._locale = locale;
  this._timeZone = timeZone;
};

Timeline.GregorianDateLabeller.monthNames = [];
Timeline.GregorianDateLabeller.dayNames = [];
Timeline.GregorianDateLabeller.labelIntervalFunctions = [];

Timeline.GregorianDateLabeller.monthNames.en =
  ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
Timeline.GregorianDateLabeller.dayNames.en =
  ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'];

Timeline.GregorianDateLabeller.getMonthName = function(month, locale) {
  return Timeline.GregorianDateLabeller.monthNames[locale][month];
};

Timeline.GregorianDateLabeller.prototype.labelInterval = function(date, intervalUnit) {
  var f = Timeline.GregorianDateLabeller.labelIntervalFunctions[this._locale];
  if (f == null) {
    f = Timeline.GregorianDateLabeller.prototype.defaultLabelInterval;
  }
  return f.call(this, date, intervalUnit);
};

Timeline.GregorianDateLabeller.prototype.labelPrecise = function(date) {
  return SimileAjax.DateTime.removeTimeZoneOffset(
    date,
    this._timeZone //+ (new Date().getTimezoneOffset() / 60)
  ).toUTCString();
};

Timeline.GregorianDateLabeller.prototype.defaultLabelInterval = function(date, intervalUnit) {
  var text;
  var emphasized = false;
  var m;

  date = SimileAjax.DateTime.removeTimeZoneOffset(date, this._timeZone);

  switch (intervalUnit) {
  case SimileAjax.DateTime.MILLISECOND:
    text = date.getUTCMilliseconds();
    break;
  case SimileAjax.DateTime.SECOND:
    text = date.getUTCSeconds();
    break;
  case SimileAjax.DateTime.MINUTE:
    m = date.getUTCMinutes();
    if (m === 0) {
      text = date.getUTCHours() + ':00';
      emphasized = true;
    } else {
      text = m;
    }
    break;
  case SimileAjax.DateTime.HOUR:
    text = date.getUTCHours() + 'hr';
    break;
  case SimileAjax.DateTime.DAY:
    text = Timeline.GregorianDateLabeller.getMonthName(date.getUTCMonth(), this._locale) + ' ' + date.getUTCDate();
    break;
  case SimileAjax.DateTime.WEEK:
    text = Timeline.GregorianDateLabeller.getMonthName(date.getUTCMonth(), this._locale) + ' ' + date.getUTCDate();
    break;
  case SimileAjax.DateTime.MONTH:
    m = date.getUTCMonth();
    if (m !== 0) {
      text = Timeline.GregorianDateLabeller.getMonthName(m, this._locale);
      break;
    } // else, fall through
  case SimileAjax.DateTime.YEAR:
  case SimileAjax.DateTime.DECADE:
  case SimileAjax.DateTime.CENTURY:
  case SimileAjax.DateTime.MILLENNIUM:
    var y = date.getUTCFullYear();
    if (y > 0) {
      text = date.getUTCFullYear();
    } else {
      text = (1 - y) + 'BC';
    }
    emphasized = (
      (intervalUnit === SimileAjax.DateTime.MONTH) ||
      (intervalUnit === SimileAjax.DateTime.DECADE && y % 100 === 0) ||
      (intervalUnit === SimileAjax.DateTime.CENTURY && y % 1000 === 0)
    );
    break;
  default:
    text = date.toUTCString();
  }
  return {text: text, emphasized: emphasized};
};

