/*
 * From: https://github.com/simile-widgets/ancient-simile-widgets
 *
 *  (c) Copyright The SIMILE Project 2006. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import {SimileAjax} from './init.js';

/**
 * A sorted array data structure
 *
 * @constructor
 */
SimileAjax.SortedArray = function(compare, initialArray) {
  this._a = (initialArray instanceof Array) ? initialArray : [];
  this._compare = compare;
};

SimileAjax.SortedArray.prototype.add = function(elmt) {
  var sa = this;
  var index = this.find(function(elmt2) {
    return sa._compare(elmt2, elmt);
  });

  if (index < this._a.length) {
    this._a.splice(index, 0, elmt);
  } else {
    this._a.push(elmt);
  }
};

SimileAjax.SortedArray.prototype.remove = function(elmt) {
  var sa = this;
  var index = this.find(function(elmt2) {
    return sa._compare(elmt2, elmt);
  });

  while (index < this._a.length && this._compare(this._a[index], elmt) === 0) {
    if (this._a[index] === elmt) {
      this._a.splice(index, 1);
      return true;
    } else {
      index++;
    }
  }
  return false;
};

SimileAjax.SortedArray.prototype.removeAll = function() {
  this._a = [];
};

SimileAjax.SortedArray.prototype.elementAt = function(index) {
  return this._a[index];
};

SimileAjax.SortedArray.prototype.length = function() {
  return this._a.length;
};

SimileAjax.SortedArray.prototype.find = function(compare) {
  var a = 0;
  var b = this._a.length;

  while (a < b) {
    var mid = Math.floor((a + b) / 2);
    var c = compare(this._a[mid]);
    if (mid === a) {
      return c < 0 ? a + 1 : a;
    } else if (c < 0) {
      a = mid;
    } else {
      b = mid;
    }
  }
  return a;
};

SimileAjax.SortedArray.prototype.getFirst = function() {
  return (this._a.length > 0) ? this._a[0] : null;
};

SimileAjax.SortedArray.prototype.getLast = function() {
  return (this._a.length > 0) ? this._a[this._a.length - 1] : null;
};

/*==================================================
 *  Event Index
 *==================================================
 */

SimileAjax.EventIndex = function(unit) {
  var eventIndex = this;

  this._unit = (unit == null) ? SimileAjax.NativeDateUnit : unit;
  this._events = new SimileAjax.SortedArray(
    function(event1, event2) {
      return eventIndex._unit.compare(event1.getStart(), event2.getStart());
    }
  );
  this._idToEvent = {};
  this._indexed = true;
};

SimileAjax.EventIndex.prototype.getUnit = function() {
  return this._unit;
};

SimileAjax.EventIndex.prototype.getEvent = function(id) {
  return this._idToEvent[id];
};

SimileAjax.EventIndex.prototype.add = function(evt) {
  this._events.add(evt);
  this._idToEvent[evt.getID()] = evt;
  this._indexed = false;
};

SimileAjax.EventIndex.prototype.removeAll = function() {
  this._events.removeAll();
  this._idToEvent = {};
  this._indexed = false;
};

SimileAjax.EventIndex.prototype.getCount = function() {
  return this._events.length();
};

SimileAjax.EventIndex.prototype.getIterator = function(startDate, endDate) {
  if (!this._indexed) {
    this._index();
  }
  return new SimileAjax.EventIndex._Iterator(this._events, startDate, endDate, this._unit);
};

SimileAjax.EventIndex.prototype.getReverseIterator = function(startDate, endDate) {
  if (!this._indexed) {
    this._index();
  }
  return new SimileAjax.EventIndex._ReverseIterator(this._events, startDate, endDate, this._unit);
};

SimileAjax.EventIndex.prototype.getAllIterator = function() {
  return new SimileAjax.EventIndex._AllIterator(this._events);
};

SimileAjax.EventIndex.prototype.getEarliestDate = function() {
  var evt = this._events.getFirst();
  return (evt == null) ? null : evt.getStart();
};

SimileAjax.EventIndex.prototype.getLatestDate = function() {
  var evt = this._events.getLast();
  if (evt == null) {
    return null;
  }

  if (!this._indexed) {
    this._index();
  }

  var index = evt._earliestOverlapIndex;
  var date = this._events.elementAt(index).getEnd();
  for (var i = index + 1; i < this._events.length(); i++) {
    date = this._unit.later(date, this._events.elementAt(i).getEnd());
  }

  return date;
};

SimileAjax.EventIndex.prototype._index = function() {
  /*
   *  For each event, we want to find the earliest preceding
   *  event that overlaps with it, if any.
   */

  var l = this._events.length();
  var i;
  var evt;
  for (i = 0; i < l; i++) {
    evt = this._events.elementAt(i);
    evt._earliestOverlapIndex = i;
  }

  var toIndex = 1;
  for (i = 0; i < l; i++) {
    evt = this._events.elementAt(i);
    var end = evt.getEnd();

    toIndex = Math.max(toIndex, i + 1);
    while (toIndex < l) {
      var evt2 = this._events.elementAt(toIndex);
      var start2 = evt2.getStart();

      if (this._unit.compare(start2, end) < 0) {
        evt2._earliestOverlapIndex = i;
        toIndex++;
      } else {
        break;
      }
    }
  }
  this._indexed = true;
};

SimileAjax.EventIndex._Iterator = function(events, startDate, endDate, unit) {
  this._events = events;
  this._startDate = startDate;
  this._endDate = endDate;
  this._unit = unit;

  this._currentIndex = events.find(function(evt) {
    return unit.compare(evt.getStart(), startDate);
  });
  if (this._currentIndex - 1 >= 0) {
    this._currentIndex = this._events.elementAt(this._currentIndex - 1)._earliestOverlapIndex;
  }
  this._currentIndex--;

  this._maxIndex = events.find(function(evt) {
    return unit.compare(evt.getStart(), endDate);
  });

  this._hasNext = false;
  this._next = null;
  this._findNext();
};

SimileAjax.EventIndex._Iterator.prototype = {
  hasNext: function() { return this._hasNext; },
  next: function() {
    if (this._hasNext) {
      var next = this._next;
      this._findNext();

      return next;
    } else {
      return null;
    }
  },
  _findNext: function() {
    var unit = this._unit;
    while ((++this._currentIndex) < this._maxIndex) {
      var evt = this._events.elementAt(this._currentIndex);
      if (unit.compare(evt.getStart(), this._endDate) < 0 &&
                unit.compare(evt.getEnd(), this._startDate) > 0) {

        this._next = evt;
        this._hasNext = true;
        return;
      }
    }
    this._next = null;
    this._hasNext = false;
  }
};

SimileAjax.EventIndex._ReverseIterator = function(events, startDate, endDate, unit) {
  this._events = events;
  this._startDate = startDate;
  this._endDate = endDate;
  this._unit = unit;

  this._minIndex = events.find(function(evt) {
    return unit.compare(evt.getStart(), startDate);
  });
  if (this._minIndex - 1 >= 0) {
    this._minIndex = this._events.elementAt(this._minIndex - 1)._earliestOverlapIndex;
  }

  this._maxIndex = events.find(function(evt) {
    return unit.compare(evt.getStart(), endDate);
  });

  this._currentIndex = this._maxIndex;
  this._hasNext = false;
  this._next = null;
  this._findNext();
};

SimileAjax.EventIndex._ReverseIterator.prototype = {
  hasNext: function() { return this._hasNext; },
  next: function() {
    if (this._hasNext) {
      var next = this._next;
      this._findNext();

      return next;
    } else {
      return null;
    }
  },
  _findNext: function() {
    var unit = this._unit;
    while ((--this._currentIndex) >= this._minIndex) {
      var evt = this._events.elementAt(this._currentIndex);
      if (unit.compare(evt.getStart(), this._endDate) < 0 &&
                unit.compare(evt.getEnd(), this._startDate) > 0) {

        this._next = evt;
        this._hasNext = true;
        return;
      }
    }
    this._next = null;
    this._hasNext = false;
  }
};

SimileAjax.EventIndex._AllIterator = function(events) {
  this._events = events;
  this._index = 0;
};

SimileAjax.EventIndex._AllIterator.prototype = {
  hasNext: function() {
    return this._index < this._events.length();
  },
  next: function() {
    return this._index < this._events.length()
            ? this._events.elementAt(this._index++) : null;
  }
};
